
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (3)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (4)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*3+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[10],
		T io_hv[10]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6567643262e-1,.2847044317,.2847044317,.6567643273e-1,.2847044315,1.558845727,.2847044317,.2847044314,.2847044318,.6567643269e-1},
			{.4551941938e-1,-.3486902923,-.3486902923,.4551941936e-1,-.1456296199,1.272792206,-.1456296200,1.191700496,1.191700498,.6983490945e-1},
			{.1632993161,0.,0.,.1632993161,-.5238966154,0.,-.5238966156,1.329162352,1.329162353,.5123596360},
			{-.1883469763e-1,-1.035891454,1.035891454,.1883469765e-1,-1.193181575,0.,1.193181574,-.1572901205,.1572901205,0.},
			{.4071380639,0.,0.,-.4071380640,-1.039593595,0.,1.039593596,-1.039593595,1.039593596,0.},
			{.6180460035,0.,0.,.6180460033,1.221587794,0.,1.221587793,.8475937972e-2,.8475937847e-2,.4543791643e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6567643262e-1,.2847044317,.2847044317,.6567643273e-1,.2847044315,1.558845727,.2847044317,.2847044314,.2847044318,.6567643269e-1},
			{.4551941938e-1,-.3486902923,-.3486902923,.4551941936e-1,-.1456296199,1.272792206,-.1456296200,1.191700496,1.191700498,.6983490945e-1},
			{.1632993161,0.,0.,.1632993161,-.5238966154,0.,-.5238966156,1.329162352,1.329162353,.5123596360},
			{-.1883469763e-1,-1.035891454,1.035891454,.1883469765e-1,-1.193181575,0.,1.193181574,-.1572901205,.1572901205,0.},
			{.4071380639,0.,0.,-.4071380640,-1.039593595,0.,1.039593596,-1.039593595,1.039593596,0.},
			{.6180460035,0.,0.,.6180460033,1.221587794,0.,1.221587793,.8475937972e-2,.8475937847e-2,.4543791643e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5687745911e-1,.2465612703,.2465612703,.5687745920e-1,.2465612703,1.350000000,.2465612704,.2465612703,.2465612705,.5687745920e-1},
			{-.4031920378e-1,-.8052657365,-.8052657365,-.4031920378e-1,-.5707914690,-.7348469229,-.5707914692,.9734243377,.9734243378,-.1224209429e-1},
			{.2519125866,1.162300974,1.162300974,.2519125869,0.,0.,0.,0.,0.,.7071067805},
			{.3283821632e-1,.1423522158,.1423522158,.3283821637e-1,.1423522158,.7794228636,.1423522159,.1423522157,.1423522159,.3283821636e-1},
			{.1060196464e-1,-.8430102043,.4943199122,.3491745472e-1,-.8430102047,.6363961029,.6973805849,.4943199122,.6973805851,.3491745472e-1},
			{.4138637121,.9265294839,-.9265294839,-.2505643959,0.,0.,-.5238966156,0.,1.329162353,.2561798179},
			{.2624354259e-1,-.2928671112,.7430243425,.4507824022e-1,-.3977271917,.8215838362,.7954543828,-.5243004016e-1,.1048600803,.3252177514e-1},
			{.5197967977,-.5197967977,-.5197967977,.1126587337,0.,0.,1.039593596,0.,1.039593596,-.9091029830e-1},
			{.2635850852,-.6065559276,.6065559276,.3544609182,0.,0.,1.221587793,0.,.8475937847e-2,.2271895821e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5687745911e-1,.2465612703,.2465612703,.5687745920e-1,.2465612703,1.350000000,.2465612704,.2465612703,.2465612705,.5687745920e-1},
			{-.4031920378e-1,-.8052657365,-.8052657365,-.4031920378e-1,-.5707914690,-.7348469229,-.5707914692,.9734243377,.9734243378,-.1224209429e-1},
			{.2519125866,1.162300974,1.162300974,.2519125869,0.,0.,0.,0.,0.,.7071067805},
			{.3283821632e-1,.1423522158,.1423522158,.3283821637e-1,.1423522158,.7794228636,.1423522159,.1423522157,.1423522159,.3283821636e-1},
			{.1060196464e-1,-.8430102043,.4943199122,.3491745472e-1,-.8430102047,.6363961029,.6973805849,.4943199122,.6973805851,.3491745472e-1},
			{.4138637121,.9265294839,-.9265294839,-.2505643959,0.,0.,-.5238966156,0.,1.329162353,.2561798179},
			{.2624354259e-1,-.2928671112,.7430243425,.4507824022e-1,-.3977271917,.8215838362,.7954543828,-.5243004016e-1,.1048600803,.3252177514e-1},
			{.5197967977,-.5197967977,-.5197967977,.1126587337,0.,0.,1.039593596,0.,1.039593596,-.9091029830e-1},
			{.2635850852,-.6065559276,.6065559276,.3544609182,0.,0.,1.221587793,0.,.8475937847e-2,.2271895821e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_dofs_to_nodes(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.505360805,1.843682925,-2.128901666,-1.187112542,1.370759491,3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,1.505360805,-1.843682925,2.128901666,-1.187112542,1.370759491,-3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,-2.484731210,-1.258649951,2.509295895,2.817981979,3.185720379,-2.761597556},
			{1.414213562,0.,-1.360827635,.8380524812,0.,0.,0.,-1.217161239,-1.873942317,0.},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,2.484731210,1.258649951,-2.509295895,2.817981979,3.185720379,2.761597556},
			{1.414213562,2.303680699,.9961259674,-1.348290938,-.9793704052,-3.102332876,-4.427892772,.4377975572,2.032696517,-.1691075797},
			{1.414213562,2.303680699,.9961259674,-1.348290938,.9793704052,3.102332876,4.427892772,.4377975572,2.032696517,.1691075797},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_dofs_to_nodes(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.505360805,1.843682925,-2.128901666,-1.187112542,1.370759491,3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,1.505360805,-1.843682925,2.128901666,-1.187112542,1.370759491,-3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,-2.484731210,-1.258649951,2.509295895,2.817981979,3.185720379,-2.761597556},
			{1.414213562,0.,-1.360827635,.8380524812,0.,0.,0.,-1.217161239,-1.873942317,0.},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,2.484731210,1.258649951,-2.509295895,2.817981979,3.185720379,2.761597556},
			{1.414213562,2.303680699,.9961259674,-1.348290938,-.9793704052,-3.102332876,-4.427892772,.4377975572,2.032696517,-.1691075797},
			{1.414213562,2.303680699,.9961259674,-1.348290938,.9793704052,3.102332876,4.427892772,.4377975572,2.032696517,.1691075797},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.341640786,-1.095445115,1.264911064,2.506647478,1.172637045,-2.525628468,2.867912544,3.095507552,2.835318997},
			{1.414213562,2.341640786,1.095445115,-1.264911064,.9574541388,3.070003641,4.485220266,.4184228013,1.964136704,.1580069126},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.341640786,-1.095445115,1.264911064,2.506647478,1.172637045,-2.525628468,2.867912544,3.095507552,2.835318997},
			{1.414213562,2.341640786,1.095445115,-1.264911064,.9574541388,3.070003641,4.485220266,.4184228013,1.964136704,.1580069126},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.549193338,1.897366595,-2.190890230,-1.095445115,1.264911064,3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,1.549193338,-1.897366595,2.190890230,-1.095445115,1.264911064,-3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.549193338,1.897366595,-2.190890230,-1.095445115,1.264911064,3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,1.549193338,-1.897366595,2.190890230,-1.095445115,1.264911064,-3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.341640786,1.095445115,-1.264911064,-.9574541388,-3.070003641,-4.485220266,.4184228013,1.964136704,-.1580069126},
			{1.414213562,-.341640786,-1.095445115,1.264911064,-2.506647478,-1.172637045,2.525628468,2.867912544,3.095507552,-2.835318997},
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[4]
	) {
		static const T matrix[4][10] = 
		{
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.341640786,1.095445115,-1.264911064,-.9574541388,-3.070003641,-4.485220266,.4184228013,1.964136704,-.1580069126},
			{1.414213562,-.341640786,-1.095445115,1.264911064,-2.506647478,-1.172637045,2.525628468,2.867912544,3.095507552,-2.835318997},
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1666666667,-.8333333333,-.8333333333,-.1666666667},
			{.2357022604,.2013137632,-1.379825065,-.4714045206},
			{-.2886751347,.6454972238,-.6454972245,-.8660254040},
			{0.,0.,0.,-1.},
			{-.4082482906,-1.477056191,-.5641852611,0.},
			{.5000000000,-.6909830067,-1.809016994,0.},
			{-.2309401077,.7136441777,-1.868344718,-.3464101616},
			{-.6454972245,-1.689933672,-.2465579999,0.},
			{.5962847939,-1.490711985,-1.490711985,.1490711985},
			{-.8567194721,-1.727071357,-.3676284881e-1,-.2519763153e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1666666667,-.8333333333,-.8333333333,-.1666666667},
			{.2357022604,.2013137632,-1.379825065,-.4714045206},
			{-.2886751347,.6454972238,-.6454972245,-.8660254040},
			{0.,0.,0.,-1.},
			{-.4082482906,-1.477056191,-.5641852611,0.},
			{.5000000000,-.6909830067,-1.809016994,0.},
			{-.2309401077,.7136441777,-1.868344718,-.3464101616},
			{-.6454972245,-1.689933672,-.2465579999,0.},
			{.5962847939,-1.490711985,-1.490711985,.1490711985},
			{-.8567194721,-1.727071357,-.3676284881e-1,-.2519763153e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{.1666666667,.8333333333,.8333333333,.1666666667},
			{-.2041241452,-1.020620727,-1.020620726,-.2041241452},
			{.2357022604,1.178511302,1.178511302,.2357022604},
			{.2886751347,.6454972238,-.6454972245,-.2886751347},
			{-.3535533905,-.7905694138,.7905694148,.3535533905},
			{.4082482906,.9128709283,-.9128709292,-.4082482906},
			{-.4564354646,.4564354645,.4564354645,-.4564354646},
			{.5270462768,-.5270462766,-.5270462766,.5270462768},
			{.2672612419,-.5976143042,.5976143046,-.2672612419}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{.1666666667,.8333333333,.8333333333,.1666666667},
			{-.2041241452,-1.020620727,-1.020620726,-.2041241452},
			{.2357022604,1.178511302,1.178511302,.2357022604},
			{.2886751347,.6454972238,-.6454972245,-.2886751347},
			{-.3535533905,-.7905694138,.7905694148,.3535533905},
			{.4082482906,.9128709283,-.9128709292,-.4082482906},
			{-.4564354646,.4564354645,.4564354645,-.4564354646},
			{.5270462768,-.5270462766,-.5270462766,.5270462768},
			{.2672612419,-.5976143042,.5976143046,-.2672612419}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{-.3333333333,-.9756836608,.1423503275,.1666666667},
			{-.6123724358,-.4564354642,.4564354645,-.2041241452},
			{-.7071067810,0.,0.,0.},
			{0.,.3989392250,1.044436449,.2886751347},
			{0.,1.279168182,.4885987688,-.3535533905},
			{.2449489743,1.321119218,-.5046226386,.1632993162},
			{0.,-.1743428331,-1.194963560,-.4564354646},
			{.1054092553,-1.054092553,-1.054092553,.4216370212},
			{.1781741613e-1,.2599526072e-1,1.221223869,.6057921484}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[4],
		T o_output[10]
	) {
		static const T matrix[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{-.3333333333,-.9756836608,.1423503275,.1666666667},
			{-.6123724358,-.4564354642,.4564354645,-.2041241452},
			{-.7071067810,0.,0.,0.},
			{0.,.3989392250,1.044436449,.2886751347},
			{0.,1.279168182,.4885987688,-.3535533905},
			{.2449489743,1.321119218,-.5046226386,.1632993162},
			{0.,-.1743428331,-1.194963560,-.4564354646},
			{.1054092553,-1.054092553,-1.054092553,.4216370212},
			{.1781741613e-1,.2599526072e-1,1.221223869,.6057921484}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,-.4082482906,1.750000000,-7.678758582,.3227486122,-2.459674775,-.2645751311},
			{0.,-.7500000000,2.857738034,-12.37436867,-.7216878367,2.828427124,-13.26806944,.9128709293,-5.850213671,-.9354143468},
			{0.,0.,1.166666667,-7.216878367,0.,1.587713241,-9.500000000,.9316949905,-6.067673910,-1.636634177},
			{0.,0.,0.,-1.875000000,0.,0.,-2.901185103,0.,-2.627379873,-1.218935425},
			{0.,.4330127020,-.7071067810,2.245365598,-.2500000000,0.,.4949747467,.3162277660,-1.004158022,-.3240370350},
			{0.,0.,-.8660254040,3.500000000,0.,-.2500000000,1.558845727,.5809475019,-1.565247584,-1.020504077},
			{0.,0.,0.,1.659882024,0.,0.,1.275000000,0.,-.5486726408,-.9983468483},
			{0.,0.,.3726779962,-.6454972245,0.,-.3227486122,.2236067977,.8333333333e-1,.2886751347,-.1463850110},
			{0.,0.,0.,-.8385254914,0.,0.,.9682458365e-1,0.,.6250000000,-.2746751327},
			{0.,0.,0.,.3307189139,0.,0.,-.3436931772,0.,.1479019946,-.2500000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,-.4082482906,1.750000000,-7.678758582,.3227486122,-2.459674775,-.2645751311},
			{0.,-.7500000000,2.857738034,-12.37436867,-.7216878367,2.828427124,-13.26806944,.9128709293,-5.850213671,-.9354143468},
			{0.,0.,1.166666667,-7.216878367,0.,1.587713241,-9.500000000,.9316949905,-6.067673910,-1.636634177},
			{0.,0.,0.,-1.875000000,0.,0.,-2.901185103,0.,-2.627379873,-1.218935425},
			{0.,.4330127020,-.7071067810,2.245365598,-.2500000000,0.,.4949747467,.3162277660,-1.004158022,-.3240370350},
			{0.,0.,-.8660254040,3.500000000,0.,-.2500000000,1.558845727,.5809475019,-1.565247584,-1.020504077},
			{0.,0.,0.,1.659882024,0.,0.,1.275000000,0.,-.5486726408,-.9983468483},
			{0.,0.,.3726779962,-.6454972245,0.,-.3227486122,.2236067977,.8333333333e-1,.2886751347,-.1463850110},
			{0.,0.,0.,-.8385254914,0.,0.,.9682458365e-1,0.,.6250000000,-.2746751327},
			{0.,0.,0.,.3307189139,0.,0.,-.3436931772,0.,.1479019946,-.2500000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,2.904737510,1.341640786,-11.64130577},
			{0.,-.2500000000,0.,0.,.7216878367,.3535533905,.1632993162,-2.464751509,-1.581138830,10.85080642},
			{0.,0.,.1666666667,0.,0.,-.4330127020,-.2000000000,1.155301788,1.290994449,-6.197388082},
			{0.,0.,0.,-.1250000000,0.,0.,.3031088914,0.,-.7267220925,1.974864371},
			{0.,-.4330127020,0.,0.,-.7500000000,.6123724358,.2828427124,3.320391543,-1.278019301,-15.76980237},
			{0.,0.,.2886751347,0.,0.,-.2500000000,-.3464101616,-1.484643616,.4472135954,8.919961562},
			{0.,0.,0.,-.2165063510,0.,0.,.3250000000,0.,.3227486122e-1,-2.776822654},
			{0.,0.,.3726779962,0.,0.,.9682458365,-.4472135954,1.083333333,-4.041451885,-8.002380602},
			{0.,0.,0.,-.2795084971,0.,0.,-.9682458365e-1,0.,1.375000000,2.387560769},
			{0.,0.,0.,-.3307189139,0.,0.,-1.031079532,0.,-1.922725929,-1.575000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.5000000000,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,2.904737510,1.341640786,-11.64130577},
			{0.,-.2500000000,0.,0.,.7216878367,.3535533905,.1632993162,-2.464751509,-1.581138830,10.85080642},
			{0.,0.,.1666666667,0.,0.,-.4330127020,-.2000000000,1.155301788,1.290994449,-6.197388082},
			{0.,0.,0.,-.1250000000,0.,0.,.3031088914,0.,-.7267220925,1.974864371},
			{0.,-.4330127020,0.,0.,-.7500000000,.6123724358,.2828427124,3.320391543,-1.278019301,-15.76980237},
			{0.,0.,.2886751347,0.,0.,-.2500000000,-.3464101616,-1.484643616,.4472135954,8.919961562},
			{0.,0.,0.,-.2165063510,0.,0.,.3250000000,0.,.3227486122e-1,-2.776822654},
			{0.,0.,.3726779962,0.,0.,.9682458365,-.4472135954,1.083333333,-4.041451885,-8.002380602},
			{0.,0.,0.,-.2795084971,0.,0.,-.9682458365e-1,0.,1.375000000,2.387560769},
			{0.,0.,0.,-.3307189139,0.,0.,-1.031079532,0.,-1.922725929,-1.575000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,.4082482906,.1250000000,-.2453738645,-.1613743061,.5590169942e-1,.6614378278e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,.7216878367,-.7071067810,.3163924252,.3651483717,.3952847075e-1,-.1403121520},
			{0.,0.,.8333333333e-1,-.2165063510,0.,-.2165063510,.4250000000,.5776508940,-.6132223630,.3218713882},
			{0.,0.,0.,-.3125000000e-1,0.,0.,.7577722285e-1,0.,-.1816805231,.4937160930},
			{0.,-.4330127020,.7071067810,-.4592793268,-.7500000000,0.,.1237436867,-.3162277660,-.1141088661,.1890216038},
			{0.,0.,.1443375673,-.3750000000,0.,-.1250000000,.3897114318,-.7423218081,.2795084971,-.3496171374},
			{0.,0.,0.,-.5412658775e-1,0.,0.,.8125000000e-1,0.,.8068715303e-2,-.6942056633},
			{0.,0.,.1863389981,-.4841229182,0.,.4841229182,-.3913118960,.5416666667,.3608439183,.1341862601},
			{0.,0.,0.,-.6987712428e-1,0.,0.,-.2420614591e-1,0.,.3437500000,.5968901920},
			{0.,0.,0.,-.8267972847e-1,0.,0.,-.2577698829,0.,-.4806814823,-.3937500000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,.4082482906,.1250000000,-.2453738645,-.1613743061,.5590169942e-1,.6614378278e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,.7216878367,-.7071067810,.3163924252,.3651483717,.3952847075e-1,-.1403121520},
			{0.,0.,.8333333333e-1,-.2165063510,0.,-.2165063510,.4250000000,.5776508940,-.6132223630,.3218713882},
			{0.,0.,0.,-.3125000000e-1,0.,0.,.7577722285e-1,0.,-.1816805231,.4937160930},
			{0.,-.4330127020,.7071067810,-.4592793268,-.7500000000,0.,.1237436867,-.3162277660,-.1141088661,.1890216038},
			{0.,0.,.1443375673,-.3750000000,0.,-.1250000000,.3897114318,-.7423218081,.2795084971,-.3496171374},
			{0.,0.,0.,-.5412658775e-1,0.,0.,.8125000000e-1,0.,.8068715303e-2,-.6942056633},
			{0.,0.,.1863389981,-.4841229182,0.,.4841229182,-.3913118960,.5416666667,.3608439183,.1341862601},
			{0.,0.,0.,-.6987712428e-1,0.,0.,-.2420614591e-1,0.,.3437500000,.5968901920},
			{0.,0.,0.,-.8267972847e-1,0.,0.,-.2577698829,0.,-.4806814823,-.3937500000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,-.4082482906,-.1250000000,.2453738645,.1613743061,-.5590169942e-1,-.6614378278e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,-.7216878367,-.3535533905,.5103103630e-1,.4564354646,.1185854122,-.2338535867},
			{0.,0.,.5833333333,.2165063510,0.,.7938566203,.3750000000,.4658474951,.9682458365e-1,-.4091585443},
			{0.,0.,0.,-.4687500000,0.,0.,-.7252962758,0.,-.6568449682,-.3047338563},
			{0.,.4330127020,.7071067810,.2551551816,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2966830521,-.8100925875e-1},
			{0.,0.,-.4330127020,-.6250000000,0.,-.1250000000,-.4330127020e-1,.2904737510,.6149186937,-.2551260193},
			{0.,0.,0.,.4149705060,0.,0.,.3187500000,0.,-.1371681602,-.2495867120},
			{0.,0.,.1863389981,.4841229182,0.,-.1613743061,-.5031152948,.4166666667e-1,.2165063510,-.3659625275e-1},
			{0.,0.,0.,-.2096313728,0.,0.,.2420614591e-1,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,.8267972847e-1,0.,0.,-.8592329430e-1,0.,.3697549864e-1,-.6250000000e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,-.4082482906,-.1250000000,.2453738645,.1613743061,-.5590169942e-1,-.6614378278e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,-.7216878367,-.3535533905,.5103103630e-1,.4564354646,.1185854122,-.2338535867},
			{0.,0.,.5833333333,.2165063510,0.,.7938566203,.3750000000,.4658474951,.9682458365e-1,-.4091585443},
			{0.,0.,0.,-.4687500000,0.,0.,-.7252962758,0.,-.6568449682,-.3047338563},
			{0.,.4330127020,.7071067810,.2551551816,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2966830521,-.8100925875e-1},
			{0.,0.,-.4330127020,-.6250000000,0.,-.1250000000,-.4330127020e-1,.2904737510,.6149186937,-.2551260193},
			{0.,0.,0.,.4149705060,0.,0.,.3187500000,0.,-.1371681602,-.2495867120},
			{0.,0.,.1863389981,.4841229182,0.,-.1613743061,-.5031152948,.4166666667e-1,.2165063510,-.3659625275e-1},
			{0.,0.,0.,-.2096313728,0.,0.,.2420614591e-1,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,.8267972847e-1,0.,0.,-.8592329430e-1,0.,.3697549864e-1,-.6250000000e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-32.86335345,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,6.,-19.59591794,42.42640686,3.464101616,-25.45584412,63.68673332,-10.95445115,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,42.42640686,-117.5755077,32.86335345,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,24.49489743,-127.2792206,0.,21.21320343,-161.6663230,5.477225575,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,205.7571384,0.,265.6313234,89.79977729},
			{0.,0.,0.,98.99494934,0.,0.,102.8785692,0.,44.27188724,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-32.86335345,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,6.,-19.59591794,42.42640686,3.464101616,-25.45584412,63.68673332,-10.95445115,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,42.42640686,-117.5755077,32.86335345,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,24.49489743,-127.2792206,0.,21.21320343,-161.6663230,5.477225575,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,205.7571384,0.,265.6313234,89.79977729},
			{0.,0.,0.,98.99494934,0.,0.,102.8785692,0.,44.27188724,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 3200;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[10][10] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6567643262e-1,.2847044317,.2847044317,.6567643273e-1,.2847044315,1.558845727,.2847044317,.2847044314,.2847044318,.6567643269e-1},
			{.4551941938e-1,-.3486902923,-.3486902923,.4551941936e-1,-.1456296199,1.272792206,-.1456296200,1.191700496,1.191700498,.6983490945e-1},
			{.1632993161,0.,0.,.1632993161,-.5238966154,0.,-.5238966156,1.329162352,1.329162353,.5123596360},
			{-.1883469763e-1,-1.035891454,1.035891454,.1883469765e-1,-1.193181575,0.,1.193181574,-.1572901205,.1572901205,0.},
			{.4071380639,0.,0.,-.4071380640,-1.039593595,0.,1.039593596,-1.039593595,1.039593596,0.},
			{.6180460035,0.,0.,.6180460033,1.221587794,0.,1.221587793,.8475937972e-2,.8475937847e-2,.4543791643e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[10][10] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5687745911e-1,.2465612703,.2465612703,.5687745920e-1,.2465612703,1.350000000,.2465612704,.2465612703,.2465612705,.5687745920e-1},
			{-.4031920378e-1,-.8052657365,-.8052657365,-.4031920378e-1,-.5707914690,-.7348469229,-.5707914692,.9734243377,.9734243378,-.1224209429e-1},
			{.2519125866,1.162300974,1.162300974,.2519125869,0.,0.,0.,0.,0.,.7071067805},
			{.3283821632e-1,.1423522158,.1423522158,.3283821637e-1,.1423522158,.7794228636,.1423522159,.1423522157,.1423522159,.3283821636e-1},
			{.1060196464e-1,-.8430102043,.4943199122,.3491745472e-1,-.8430102047,.6363961029,.6973805849,.4943199122,.6973805851,.3491745472e-1},
			{.4138637121,.9265294839,-.9265294839,-.2505643959,0.,0.,-.5238966156,0.,1.329162353,.2561798179},
			{.2624354259e-1,-.2928671112,.7430243425,.4507824022e-1,-.3977271917,.8215838362,.7954543828,-.5243004016e-1,.1048600803,.3252177514e-1},
			{.5197967977,-.5197967977,-.5197967977,.1126587337,0.,0.,1.039593596,0.,1.039593596,-.9091029830e-1},
			{.2635850852,-.6065559276,.6065559276,.3544609182,0.,0.,1.221587793,0.,.8475937847e-2,.2271895821e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_dofs_to_nodes:" << std::endl;
		static const T timestep_stiffness_dofs_to_nodes[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.505360805,1.843682925,-2.128901666,-1.187112542,1.370759491,3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,1.505360805,-1.843682925,2.128901666,-1.187112542,1.370759491,-3.342659338},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,-2.484731210,-1.258649951,2.509295895,2.817981979,3.185720379,-2.761597556},
			{1.414213562,0.,-1.360827635,.8380524812,0.,0.,0.,-1.217161239,-1.873942317,0.},
			{1.414213562,-.3036806991,-1.132775699,1.229948799,2.484731210,1.258649951,-2.509295895,2.817981979,3.185720379,2.761597556},
			{1.414213562,2.303680699,.9961259674,-1.348290938,-.9793704052,-3.102332876,-4.427892772,.4377975572,2.032696517,-.1691075797},
			{1.414213562,2.303680699,.9961259674,-1.348290938,.9793704052,3.102332876,4.427892772,.4377975572,2.032696517,.1691075797},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_dofs_to_nodes[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_dofs_to_nodes[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774},
			{1.414213562,-.341640786,-1.095445115,1.264911064,2.506647478,1.172637045,-2.525628468,2.867912544,3.095507552,2.835318997},
			{1.414213562,2.341640786,1.095445115,-1.264911064,.9574541388,3.070003641,4.485220266,.4184228013,1.964136704,.1580069126},
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[4][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{1.414213562,-2.,2.449489743,-2.828427124,-1.549193338,1.897366595,-2.190890230,-1.095445115,1.264911064,3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,1.549193338,-1.897366595,2.190890230,-1.095445115,1.264911064,-3.346640106},
			{1.414213562,-2.,2.449489743,-2.828427124,3.464101616,-4.242640686,4.898979486,5.477225575,-6.324555320,7.483314774}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[4][10] = 
		{
			{1.414213562,4.,7.348469229,11.31370850,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.341640786,1.095445115,-1.264911064,-.9574541388,-3.070003641,-4.485220266,.4184228013,1.964136704,-.1580069126},
			{1.414213562,-.341640786,-1.095445115,1.264911064,-2.506647478,-1.172637045,2.525628468,2.867912544,3.095507552,-2.835318997},
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[10][4] = 
		{
			{-.1666666667,-.8333333333,-.8333333333,-.1666666667},
			{.2357022604,.2013137632,-1.379825065,-.4714045206},
			{-.2886751347,.6454972238,-.6454972245,-.8660254040},
			{0.,0.,0.,-1.},
			{-.4082482906,-1.477056191,-.5641852611,0.},
			{.5000000000,-.6909830067,-1.809016994,0.},
			{-.2309401077,.7136441777,-1.868344718,-.3464101616},
			{-.6454972245,-1.689933672,-.2465579999,0.},
			{.5962847939,-1.490711985,-1.490711985,.1490711985},
			{-.8567194721,-1.727071357,-.3676284881e-1,-.2519763153e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{.1666666667,.8333333333,.8333333333,.1666666667},
			{-.2041241452,-1.020620727,-1.020620726,-.2041241452},
			{.2357022604,1.178511302,1.178511302,.2357022604},
			{.2886751347,.6454972238,-.6454972245,-.2886751347},
			{-.3535533905,-.7905694138,.7905694148,.3535533905},
			{.4082482906,.9128709283,-.9128709292,-.4082482906},
			{-.4564354646,.4564354645,.4564354645,-.4564354646},
			{.5270462768,-.5270462766,-.5270462766,.5270462768},
			{.2672612419,-.5976143042,.5976143046,-.2672612419}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[10][4] = 
		{
			{-.1178511302,-.5892556509,-.5892556509,-.1178511302},
			{-.3333333333,-.9756836608,.1423503275,.1666666667},
			{-.6123724358,-.4564354642,.4564354645,-.2041241452},
			{-.7071067810,0.,0.,0.},
			{0.,.3989392250,1.044436449,.2886751347},
			{0.,1.279168182,.4885987688,-.3535533905},
			{.2449489743,1.321119218,-.5046226386,.1632993162},
			{0.,-.1743428331,-1.194963560,-.4564354646},
			{.1054092553,-1.054092553,-1.054092553,.4216370212},
			{.1781741613e-1,.2599526072e-1,1.221223869,.6057921484}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[10][10] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,-.4082482906,1.750000000,-7.678758582,.3227486122,-2.459674775,-.2645751311},
			{0.,-.7500000000,2.857738034,-12.37436867,-.7216878367,2.828427124,-13.26806944,.9128709293,-5.850213671,-.9354143468},
			{0.,0.,1.166666667,-7.216878367,0.,1.587713241,-9.500000000,.9316949905,-6.067673910,-1.636634177},
			{0.,0.,0.,-1.875000000,0.,0.,-2.901185103,0.,-2.627379873,-1.218935425},
			{0.,.4330127020,-.7071067810,2.245365598,-.2500000000,0.,.4949747467,.3162277660,-1.004158022,-.3240370350},
			{0.,0.,-.8660254040,3.500000000,0.,-.2500000000,1.558845727,.5809475019,-1.565247584,-1.020504077},
			{0.,0.,0.,1.659882024,0.,0.,1.275000000,0.,-.5486726408,-.9983468483},
			{0.,0.,.3726779962,-.6454972245,0.,-.3227486122,.2236067977,.8333333333e-1,.2886751347,-.1463850110},
			{0.,0.,0.,-.8385254914,0.,0.,.9682458365e-1,0.,.6250000000,-.2746751327},
			{0.,0.,0.,.3307189139,0.,0.,-.3436931772,0.,.1479019946,-.2500000000e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[10][10] = 
		{
			{.5000000000,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,2.904737510,1.341640786,-11.64130577},
			{0.,-.2500000000,0.,0.,.7216878367,.3535533905,.1632993162,-2.464751509,-1.581138830,10.85080642},
			{0.,0.,.1666666667,0.,0.,-.4330127020,-.2000000000,1.155301788,1.290994449,-6.197388082},
			{0.,0.,0.,-.1250000000,0.,0.,.3031088914,0.,-.7267220925,1.974864371},
			{0.,-.4330127020,0.,0.,-.7500000000,.6123724358,.2828427124,3.320391543,-1.278019301,-15.76980237},
			{0.,0.,.2886751347,0.,0.,-.2500000000,-.3464101616,-1.484643616,.4472135954,8.919961562},
			{0.,0.,0.,-.2165063510,0.,0.,.3250000000,0.,.3227486122e-1,-2.776822654},
			{0.,0.,.3726779962,0.,0.,.9682458365,-.4472135954,1.083333333,-4.041451885,-8.002380602},
			{0.,0.,0.,-.2795084971,0.,0.,-.9682458365e-1,0.,1.375000000,2.387560769},
			{0.,0.,0.,-.3307189139,0.,0.,-1.031079532,0.,-1.922725929,-1.575000000}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[10][10] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,.4082482906,.1250000000,-.2453738645,-.1613743061,.5590169942e-1,.6614378278e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,.7216878367,-.7071067810,.3163924252,.3651483717,.3952847075e-1,-.1403121520},
			{0.,0.,.8333333333e-1,-.2165063510,0.,-.2165063510,.4250000000,.5776508940,-.6132223630,.3218713882},
			{0.,0.,0.,-.3125000000e-1,0.,0.,.7577722285e-1,0.,-.1816805231,.4937160930},
			{0.,-.4330127020,.7071067810,-.4592793268,-.7500000000,0.,.1237436867,-.3162277660,-.1141088661,.1890216038},
			{0.,0.,.1443375673,-.3750000000,0.,-.1250000000,.3897114318,-.7423218081,.2795084971,-.3496171374},
			{0.,0.,0.,-.5412658775e-1,0.,0.,.8125000000e-1,0.,.8068715303e-2,-.6942056633},
			{0.,0.,.1863389981,-.4841229182,0.,.4841229182,-.3913118960,.5416666667,.3608439183,.1341862601},
			{0.,0.,0.,-.6987712428e-1,0.,0.,-.2420614591e-1,0.,.3437500000,.5968901920},
			{0.,0.,0.,-.8267972847e-1,0.,0.,-.2577698829,0.,-.4806814823,-.3937500000}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[10][10] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,-.4082482906,-.1250000000,.2453738645,.1613743061,-.5590169942e-1,-.6614378278e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,-.7216878367,-.3535533905,.5103103630e-1,.4564354646,.1185854122,-.2338535867},
			{0.,0.,.5833333333,.2165063510,0.,.7938566203,.3750000000,.4658474951,.9682458365e-1,-.4091585443},
			{0.,0.,0.,-.4687500000,0.,0.,-.7252962758,0.,-.6568449682,-.3047338563},
			{0.,.4330127020,.7071067810,.2551551816,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2966830521,-.8100925875e-1},
			{0.,0.,-.4330127020,-.6250000000,0.,-.1250000000,-.4330127020e-1,.2904737510,.6149186937,-.2551260193},
			{0.,0.,0.,.4149705060,0.,0.,.3187500000,0.,-.1371681602,-.2495867120},
			{0.,0.,.1863389981,.4841229182,0.,-.1613743061,-.5031152948,.4166666667e-1,.2165063510,-.3659625275e-1},
			{0.,0.,0.,-.2096313728,0.,0.,.2420614591e-1,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,.8267972847e-1,0.,0.,-.8592329430e-1,0.,.3697549864e-1,-.6250000000e-2}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[10][10] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[10][10] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-32.86335345,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,6.,-19.59591794,42.42640686,3.464101616,-25.45584412,63.68673332,-10.95445115,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,42.42640686,-117.5755077,32.86335345,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,24.49489743,-127.2792206,0.,21.21320343,-161.6663230,5.477225575,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,205.7571384,0.,265.6313234,89.79977729},
			{0.,0.,0.,98.99494934,0.,0.,102.8785692,0.,44.27188724,7.483314774}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][10] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
