
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (2)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (18)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (18)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (9)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (6)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*2+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[18],
		T io_hv[18]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{648.,-27.,108.,-27.,108.,108.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,216.,-9.,36.,-9.,36.,36.},
			{-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.},
			{108.,-27.,648.,108.,-27.,108.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,36.,-9.,216.,36.,-9.,36.},
			{-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.},
			{108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.},
			{108.,108.,108.,-27.,-27.,648.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,36.,36.,36.,-9.,-9.,216.},
			{-108.,4.500000000,-18.,4.500000000,-18.,-18.,162.,-6.750000000,27.,-6.750000000,27.,27.,-108.,4.500000000,-18.,4.500000000,-18.,-18.},
			{4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-6.750000000,43.87500000,-6.750000000,-15.18750000,-15.18750000,27.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.},
			{-18.,4.500000000,-108.,-18.,4.500000000,-18.,27.,-6.750000000,162.,27.,-6.750000000,27.,-18.,4.500000000,-108.,-18.,4.500000000,-18.},
			{4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-6.750000000,-15.18750000,27.,43.87500000,-15.18750000,-6.750000000,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000},
			{-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,27.,-15.18750000,-6.750000000,-15.18750000,43.87500000,-6.750000000,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000},
			{-18.,-18.,-18.,4.500000000,4.500000000,-108.,27.,27.,27.,-6.750000000,-6.750000000,162.,-18.,-18.,-18.,4.500000000,4.500000000,-108.},
			{216.,-9.,36.,-9.,36.,36.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,648.,-27.,108.,-27.,108.,108.},
			{-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.},
			{36.,-9.,216.,36.,-9.,36.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,108.,-27.,648.,108.,-27.,108.},
			{-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.},
			{36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.},
			{36.,36.,36.,-9.,-9.,216.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,108.,108.,108.,-27.,-27.,648.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{648.,-27.,108.,-27.,108.,108.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,216.,-9.,36.,-9.,36.,36.},
			{-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.},
			{108.,-27.,648.,108.,-27.,108.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,36.,-9.,216.,36.,-9.,36.},
			{-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.},
			{108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.},
			{108.,108.,108.,-27.,-27.,648.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,36.,36.,36.,-9.,-9.,216.},
			{-108.,4.500000000,-18.,4.500000000,-18.,-18.,162.,-6.750000000,27.,-6.750000000,27.,27.,-108.,4.500000000,-18.,4.500000000,-18.,-18.},
			{4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-6.750000000,43.87500000,-6.750000000,-15.18750000,-15.18750000,27.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.},
			{-18.,4.500000000,-108.,-18.,4.500000000,-18.,27.,-6.750000000,162.,27.,-6.750000000,27.,-18.,4.500000000,-108.,-18.,4.500000000,-18.},
			{4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-6.750000000,-15.18750000,27.,43.87500000,-15.18750000,-6.750000000,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000},
			{-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,27.,-15.18750000,-6.750000000,-15.18750000,43.87500000,-6.750000000,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000},
			{-18.,-18.,-18.,4.500000000,4.500000000,-108.,27.,27.,27.,-6.750000000,-6.750000000,162.,-18.,-18.,-18.,4.500000000,4.500000000,-108.},
			{216.,-9.,36.,-9.,36.,36.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,648.,-27.,108.,-27.,108.,108.},
			{-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.},
			{36.,-9.,216.,36.,-9.,36.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,108.,-27.,648.,108.,-27.,108.},
			{-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.},
			{36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.},
			{36.,36.,36.,-9.,-9.,216.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,108.,108.,108.,-27.,-27.,648.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.},
			{-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2},
			{.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2},
			{-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.5333333333e-1,0.,-.5333333333e-1,.7111111111e-1,-.7111111111e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.},
			{-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.1777777778e-1,.5333333333e-1,.3555555556e-1,-.1777777778e-1,.5333333333e-1,-.1777777778e-1,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2},
			{.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.1777777778e-1,-.7111111111e-1,.1777777778e-1,-.1422222222,-.1422222222,-.3555555556e-1,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2},
			{-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.1777777778e-1,.7111111111e-1,-.1777777778e-1,.1422222222,.1422222222,.3555555556e-1,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.},
			{.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2},
			{-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2},
			{.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.},
			{-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2},
			{.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2},
			{-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.5333333333e-1,0.,-.5333333333e-1,.7111111111e-1,-.7111111111e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.},
			{-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.1777777778e-1,.5333333333e-1,.3555555556e-1,-.1777777778e-1,.5333333333e-1,-.1777777778e-1,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2},
			{.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.1777777778e-1,-.7111111111e-1,.1777777778e-1,-.1422222222,-.1422222222,-.3555555556e-1,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2},
			{-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.1777777778e-1,.7111111111e-1,-.1777777778e-1,.1422222222,.1422222222,.3555555556e-1,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.},
			{.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2},
			{-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2},
			{.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2},
			{-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2},
			{-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.1777777778e-1,-.1422222222,-.3555555556e-1,-.7111111111e-1,-.1422222222,.1777777778e-1,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.5333333333e-1,.7111111111e-1,0.,0.,-.7111111111e-1,-.5333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2},
			{-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.1777777778e-1,.1422222222,.3555555556e-1,.7111111111e-1,.1422222222,-.1777777778e-1,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2},
			{-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.1777777778e-1,-.1777777778e-1,-.1777777778e-1,.5333333333e-1,.5333333333e-1,.3555555556e-1,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1},
			{.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2},
			{.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2},
			{-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2},
			{-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.1777777778e-1,-.1422222222,-.3555555556e-1,-.7111111111e-1,-.1422222222,.1777777778e-1,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.5333333333e-1,.7111111111e-1,0.,0.,-.7111111111e-1,-.5333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2},
			{-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.1777777778e-1,.1422222222,.3555555556e-1,.7111111111e-1,.1422222222,-.1777777778e-1,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2},
			{-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.1777777778e-1,-.1777777778e-1,-.1777777778e-1,.5333333333e-1,.5333333333e-1,.3555555556e-1,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1},
			{.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2},
			{.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_dofs_to_nodes(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_dofs_to_nodes(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[18],
		T o_output[9]
	) {
		static const T matrix[9][18] = 
		{
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 9; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.2514157444e-1,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.1571348402e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.1005662977,-.5028314888e-1,.2514157444e-1,-.1257078722e-1,-.6285393608e-2,.3142696804e-2},
			{.6285393608e-2,.3142696804e-2,-.1571348402e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.2514157444e-1,-.1257078722e-1,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.1005662977,-.1257078722e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,.3142696804e-2,.2514157444e-1,.3142696804e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.4022651909,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2},
			{.3142696804e-2,.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.1257078722e-1,-.1005662977,-.1257078722e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6285393608e-2,-.1257078722e-1,-.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,-.1571348402e-2,.3142696804e-2,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.2514157444e-1,-.5028314888e-1,-.1005662977,.3142696804e-2,-.6285393608e-2,-.1257078722e-1},
			{-.1571348402e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.2514157444e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.2514157444e-1,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.1571348402e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.1005662977,-.5028314888e-1,.2514157444e-1,-.1257078722e-1,-.6285393608e-2,.3142696804e-2},
			{.6285393608e-2,.3142696804e-2,-.1571348402e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.2514157444e-1,-.1257078722e-1,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.1005662977,-.1257078722e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,.3142696804e-2,.2514157444e-1,.3142696804e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.4022651909,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2},
			{.3142696804e-2,.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.1257078722e-1,-.1005662977,-.1257078722e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6285393608e-2,-.1257078722e-1,-.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,-.1571348402e-2,.3142696804e-2,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.2514157444e-1,-.5028314888e-1,-.1005662977,.3142696804e-2,-.6285393608e-2,-.1257078722e-1},
			{-.1571348402e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.2514157444e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[9],
		T o_output[18]
	) {
		static const T matrix[18][9] = 
		{
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 9; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,1.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,.5000000000,.2500000000,.1250000000,0.,0.,0.,.2500000000,.1250000000,.6250000000e-1,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,.5000000000,.5000000000,.5000000000,0.,0.,0.,.2500000000,.2500000000,.2500000000,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,.5000000000,0.,0.,.2500000000,0.,.1250000000,.2500000000,0.,0.,.1250000000,0.,.6250000000e-1},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,.5000000000,.2500000000,.1250000000,.2500000000,.1250000000,.1250000000,.2500000000,.1250000000,.6250000000e-1,.1250000000,.6250000000e-1,.6250000000e-1},
			{1.,0.,0.,1.,0.,1.,.5000000000,0.,0.,.5000000000,0.,.5000000000,.2500000000,0.,0.,.2500000000,0.,.2500000000},
			{1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000},
			{1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,1.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,.5000000000,.2500000000,.1250000000,0.,0.,0.,.2500000000,.1250000000,.6250000000e-1,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,.5000000000,.5000000000,.5000000000,0.,0.,0.,.2500000000,.2500000000,.2500000000,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,.5000000000,0.,0.,.2500000000,0.,.1250000000,.2500000000,0.,0.,.1250000000,0.,.6250000000e-1},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,.5000000000,.2500000000,.1250000000,.2500000000,.1250000000,.1250000000,.2500000000,.1250000000,.6250000000e-1,.1250000000,.6250000000e-1,.6250000000e-1},
			{1.,0.,0.,1.,0.,1.,.5000000000,0.,0.,.5000000000,0.,.5000000000,.2500000000,0.,0.,.2500000000,0.,.2500000000},
			{1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000},
			{1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,4.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,-4.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,4.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,-4.,0.,-4.,4.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,-4.,0.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,0.,0.,0.,4.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.},
			{9.,-12.,3.,0.,0.,0.,-12.,16.,-4.,0.,0.,0.,3.,-4.,1.,0.,0.,0.},
			{-6.,12.,-6.,0.,0.,0.,8.,-16.,8.,0.,0.,0.,-2.,4.,-2.,0.,0.,0.},
			{9.,0.,0.,-12.,0.,3.,-12.,0.,0.,16.,0.,-4.,3.,0.,0.,-4.,0.,1.},
			{-12.,12.,0.,12.,-12.,0.,16.,-16.,0.,-16.,16.,0.,-4.,4.,0.,4.,-4.,0.},
			{-6.,0.,0.,12.,0.,-6.,8.,0.,0.,-16.,0.,8.,-2.,0.,0.,4.,0.,-2.},
			{2.,0.,0.,0.,0.,0.,-4.,0.,0.,0.,0.,0.,2.,0.,0.,0.,0.,0.},
			{-6.,8.,-2.,0.,0.,0.,12.,-16.,4.,0.,0.,0.,-6.,8.,-2.,0.,0.,0.},
			{4.,-8.,4.,0.,0.,0.,-8.,16.,-8.,0.,0.,0.,4.,-8.,4.,0.,0.,0.},
			{-6.,0.,0.,8.,0.,-2.,12.,0.,0.,-16.,0.,4.,-6.,0.,0.,8.,0.,-2.},
			{8.,-8.,0.,-8.,8.,0.,-16.,16.,0.,16.,-16.,0.,8.,-8.,0.,-8.,8.,0.},
			{4.,0.,0.,-8.,0.,4.,-8.,0.,0.,16.,0.,-8.,4.,0.,0.,-8.,0.,4.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,4.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,-4.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,4.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,-4.,0.,-4.,4.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,-4.,0.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,0.,0.,0.,4.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.},
			{9.,-12.,3.,0.,0.,0.,-12.,16.,-4.,0.,0.,0.,3.,-4.,1.,0.,0.,0.},
			{-6.,12.,-6.,0.,0.,0.,8.,-16.,8.,0.,0.,0.,-2.,4.,-2.,0.,0.,0.},
			{9.,0.,0.,-12.,0.,3.,-12.,0.,0.,16.,0.,-4.,3.,0.,0.,-4.,0.,1.},
			{-12.,12.,0.,12.,-12.,0.,16.,-16.,0.,-16.,16.,0.,-4.,4.,0.,4.,-4.,0.},
			{-6.,0.,0.,12.,0.,-6.,8.,0.,0.,-16.,0.,8.,-2.,0.,0.,4.,0.,-2.},
			{2.,0.,0.,0.,0.,0.,-4.,0.,0.,0.,0.,0.,2.,0.,0.,0.,0.,0.},
			{-6.,8.,-2.,0.,0.,0.,12.,-16.,4.,0.,0.,0.,-6.,8.,-2.,0.,0.,0.},
			{4.,-8.,4.,0.,0.,0.,-8.,16.,-8.,0.,0.,0.,4.,-8.,4.,0.,0.,0.},
			{-6.,0.,0.,8.,0.,-2.,12.,0.,0.,-16.,0.,4.,-6.,0.,0.,8.,0.,-2.},
			{8.,-8.,0.,-8.,8.,0.,-16.,16.,0.,16.,-16.,0.,8.,-8.,0.,-8.,8.,0.},
			{4.,0.,0.,-8.,0.,4.,-8.,0.,0.,16.,0.,-8.,4.,0.,0.,-8.,0.,4.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[18],
		T o_output[1]
	) {
		static const T matrix[1][18] = 
		{
			{0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.,0.,.2222222222,0.,.2222222222,.2222222222,0.,0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[18],
		T o_output[1]
	) {
		static const T matrix[1][18] = 
		{
			{0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.,0.,.2222222222,0.,.2222222222,.2222222222,0.,0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8333333333e-2,0.,.1388888889e-2,0.,.5555555556e-2,.1388888889e-2,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2,.2777777778e-2,0.,-.4629629630e-3,0.,-.1851851852e-2,-.4629629630e-3},
			{0.,-.4444444444e-1,0.,-.2222222222e-1,-.2222222222e-1,.5555555556e-2,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2,0.,.1481481481e-1,0.,.7407407407e-2,.7407407407e-2,-.1851851852e-2},
			{.1388888889e-2,0.,-.8333333333e-2,.5555555556e-2,0.,.1388888889e-2,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2,-.4629629630e-3,0.,.2777777778e-2,-.1851851852e-2,0.,-.4629629630e-3},
			{0.,-.2222222222e-1,.5555555556e-2,-.4444444444e-1,-.2222222222e-1,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.,0.,.7407407407e-2,-.1851851852e-2,.1481481481e-1,.7407407407e-2,0.},
			{.5555555556e-2,-.2222222222e-1,0.,-.2222222222e-1,-.4444444444e-1,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.,-.1851851852e-2,.7407407407e-2,0.,.7407407407e-2,.1481481481e-1,0.},
			{.1388888889e-2,.5555555556e-2,.1388888889e-2,0.,0.,-.8333333333e-2,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1,-.4629629630e-3,-.1851851852e-2,-.4629629630e-3,0.,0.,.2777777778e-2},
			{.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,0.,0.,0.,0.,0.,0.,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2},
			{0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,0.,0.,0.,0.,0.,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2},
			{-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,0.,0.,0.,0.,0.,0.,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2},
			{0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,0.,0.,0.,0.,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.},
			{-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,0.,0.,0.,0.,0.,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.},
			{-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,0.,0.,0.,0.,0.,0.,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1},
			{-.2777777778e-2,0.,.4629629630e-3,0.,.1851851852e-2,.4629629630e-3,.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,.8333333333e-2,0.,-.1388888889e-2,0.,-.5555555556e-2,-.1388888889e-2},
			{0.,-.1481481481e-1,0.,-.7407407407e-2,-.7407407407e-2,.1851851852e-2,0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,.4444444444e-1,0.,.2222222222e-1,.2222222222e-1,-.5555555556e-2},
			{.4629629630e-3,0.,-.2777777778e-2,.1851851852e-2,0.,.4629629630e-3,-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,-.1388888889e-2,0.,.8333333333e-2,-.5555555556e-2,0.,-.1388888889e-2},
			{0.,-.7407407407e-2,.1851851852e-2,-.1481481481e-1,-.7407407407e-2,0.,0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,.2222222222e-1,-.5555555556e-2,.4444444444e-1,.2222222222e-1,0.},
			{.1851851852e-2,-.7407407407e-2,0.,-.7407407407e-2,-.1481481481e-1,0.,-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,-.5555555556e-2,.2222222222e-1,0.,.2222222222e-1,.4444444444e-1,0.},
			{.4629629630e-3,.1851851852e-2,.4629629630e-3,0.,0.,-.2777777778e-2,-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,-.1388888889e-2,-.5555555556e-2,-.1388888889e-2,0.,0.,.8333333333e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[18],
		T o_output[18]
	) {
		static const T matrix[18][18] = 
		{
			{-.8333333333e-2,0.,.1388888889e-2,0.,.5555555556e-2,.1388888889e-2,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2,.2777777778e-2,0.,-.4629629630e-3,0.,-.1851851852e-2,-.4629629630e-3},
			{0.,-.4444444444e-1,0.,-.2222222222e-1,-.2222222222e-1,.5555555556e-2,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2,0.,.1481481481e-1,0.,.7407407407e-2,.7407407407e-2,-.1851851852e-2},
			{.1388888889e-2,0.,-.8333333333e-2,.5555555556e-2,0.,.1388888889e-2,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2,-.4629629630e-3,0.,.2777777778e-2,-.1851851852e-2,0.,-.4629629630e-3},
			{0.,-.2222222222e-1,.5555555556e-2,-.4444444444e-1,-.2222222222e-1,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.,0.,.7407407407e-2,-.1851851852e-2,.1481481481e-1,.7407407407e-2,0.},
			{.5555555556e-2,-.2222222222e-1,0.,-.2222222222e-1,-.4444444444e-1,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.,-.1851851852e-2,.7407407407e-2,0.,.7407407407e-2,.1481481481e-1,0.},
			{.1388888889e-2,.5555555556e-2,.1388888889e-2,0.,0.,-.8333333333e-2,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1,-.4629629630e-3,-.1851851852e-2,-.4629629630e-3,0.,0.,.2777777778e-2},
			{.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,0.,0.,0.,0.,0.,0.,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2},
			{0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,0.,0.,0.,0.,0.,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2},
			{-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,0.,0.,0.,0.,0.,0.,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2},
			{0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,0.,0.,0.,0.,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.},
			{-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,0.,0.,0.,0.,0.,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.},
			{-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,0.,0.,0.,0.,0.,0.,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1},
			{-.2777777778e-2,0.,.4629629630e-3,0.,.1851851852e-2,.4629629630e-3,.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,.8333333333e-2,0.,-.1388888889e-2,0.,-.5555555556e-2,-.1388888889e-2},
			{0.,-.1481481481e-1,0.,-.7407407407e-2,-.7407407407e-2,.1851851852e-2,0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,.4444444444e-1,0.,.2222222222e-1,.2222222222e-1,-.5555555556e-2},
			{.4629629630e-3,0.,-.2777777778e-2,.1851851852e-2,0.,.4629629630e-3,-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,-.1388888889e-2,0.,.8333333333e-2,-.5555555556e-2,0.,-.1388888889e-2},
			{0.,-.7407407407e-2,.1851851852e-2,-.1481481481e-1,-.7407407407e-2,0.,0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,.2222222222e-1,-.5555555556e-2,.4444444444e-1,.2222222222e-1,0.},
			{.1851851852e-2,-.7407407407e-2,0.,-.7407407407e-2,-.1481481481e-1,0.,-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,-.5555555556e-2,.2222222222e-1,0.,.2222222222e-1,.4444444444e-1,0.},
			{.4629629630e-3,.1851851852e-2,.4629629630e-3,0.,0.,-.2777777778e-2,-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,-.1388888889e-2,-.5555555556e-2,-.1388888889e-2,0.,0.,.8333333333e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[18],
		T o_output[6]
	) {
		static const T matrix[6][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[18],
		T o_output[6]
	) {
		static const T matrix[6][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[18],
		T o_output[6]
	) {
		static const T matrix[6][18] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[18],
		T o_output[6]
	) {
		static const T matrix[6][18] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 6; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 18; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[6],
		T o_output[18]
	) {
		static const T matrix[18][6] = 
		{
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[6],
		T o_output[18]
	) {
		static const T matrix[18][6] = 
		{
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[6],
		T o_output[18]
	) {
		static const T matrix[18][6] = 
		{
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[6],
		T o_output[18]
	) {
		static const T matrix[18][6] = 
		{
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 18; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 6; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 15120;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[18][18] = 
		{
			{648.,-27.,108.,-27.,108.,108.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,216.,-9.,36.,-9.,36.,36.},
			{-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.},
			{108.,-27.,648.,108.,-27.,108.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,36.,-9.,216.,36.,-9.,36.},
			{-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.},
			{108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.},
			{108.,108.,108.,-27.,-27.,648.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,36.,36.,36.,-9.,-9.,216.},
			{-108.,4.500000000,-18.,4.500000000,-18.,-18.,162.,-6.750000000,27.,-6.750000000,27.,27.,-108.,4.500000000,-18.,4.500000000,-18.,-18.},
			{4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-6.750000000,43.87500000,-6.750000000,-15.18750000,-15.18750000,27.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.},
			{-18.,4.500000000,-108.,-18.,4.500000000,-18.,27.,-6.750000000,162.,27.,-6.750000000,27.,-18.,4.500000000,-108.,-18.,4.500000000,-18.},
			{4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-6.750000000,-15.18750000,27.,43.87500000,-15.18750000,-6.750000000,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000},
			{-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,27.,-15.18750000,-6.750000000,-15.18750000,43.87500000,-6.750000000,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000},
			{-18.,-18.,-18.,4.500000000,4.500000000,-108.,27.,27.,27.,-6.750000000,-6.750000000,162.,-18.,-18.,-18.,4.500000000,4.500000000,-108.},
			{216.,-9.,36.,-9.,36.,36.,-108.,4.500000000,-18.,4.500000000,-18.,-18.,648.,-27.,108.,-27.,108.,108.},
			{-9.,58.50000000,-9.,-20.25000000,-20.25000000,36.,4.500000000,-29.25000000,4.500000000,10.12500000,10.12500000,-18.,-27.,175.5000000,-27.,-60.75000000,-60.75000000,108.},
			{36.,-9.,216.,36.,-9.,36.,-18.,4.500000000,-108.,-18.,4.500000000,-18.,108.,-27.,648.,108.,-27.,108.},
			{-9.,-20.25000000,36.,58.50000000,-20.25000000,-9.,4.500000000,10.12500000,-18.,-29.25000000,10.12500000,4.500000000,-27.,-60.75000000,108.,175.5000000,-60.75000000,-27.},
			{36.,-20.25000000,-9.,-20.25000000,58.50000000,-9.,-18.,10.12500000,4.500000000,10.12500000,-29.25000000,4.500000000,108.,-60.75000000,-27.,-60.75000000,175.5000000,-27.},
			{36.,36.,36.,-9.,-9.,216.,-18.,-18.,-18.,4.500000000,4.500000000,-108.,108.,108.,108.,-27.,-27.,648.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.},
			{-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2},
			{.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2},
			{-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.5333333333e-1,0.,-.5333333333e-1,.7111111111e-1,-.7111111111e-1,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.},
			{-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.1777777778e-1,.5333333333e-1,.3555555556e-1,-.1777777778e-1,.5333333333e-1,-.1777777778e-1,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2},
			{.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.1777777778e-1,-.7111111111e-1,.1777777778e-1,-.1422222222,-.1422222222,-.3555555556e-1,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2},
			{-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.1777777778e-1,.7111111111e-1,-.1777777778e-1,.1422222222,.1422222222,.3555555556e-1,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.3333333333e-2,0.,.3333333333e-2,-.4444444444e-2,.4444444444e-2,0.,.6666666667e-2,0.,-.6666666667e-2,.8888888889e-2,-.8888888889e-2,0.,.1333333333e-1,0.,-.1333333333e-1,.1777777778e-1,-.1777777778e-1,0.},
			{.1111111111e-2,-.3333333333e-2,-.2222222222e-2,.1111111111e-2,-.3333333333e-2,.1111111111e-2,-.2222222222e-2,.6666666667e-2,.4444444444e-2,-.2222222222e-2,.6666666667e-2,-.2222222222e-2,-.4444444444e-2,.1333333333e-1,.8888888889e-2,-.4444444444e-2,.1333333333e-1,-.4444444444e-2},
			{-.1111111111e-2,.4444444444e-2,-.1111111111e-2,.8888888889e-2,.8888888889e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,.2222222222e-2,-.1777777778e-1,-.1777777778e-1,-.4444444444e-2,.4444444444e-2,-.1777777778e-1,.4444444444e-2,-.3555555556e-1,-.3555555556e-1,-.8888888889e-2},
			{.1111111111e-2,-.4444444444e-2,.1111111111e-2,-.8888888889e-2,-.8888888889e-2,-.2222222222e-2,-.2222222222e-2,.8888888889e-2,-.2222222222e-2,.1777777778e-1,.1777777778e-1,.4444444444e-2,-.4444444444e-2,.1777777778e-1,-.4444444444e-2,.3555555556e-1,.3555555556e-1,.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[18][18] = 
		{
			{-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2},
			{.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2},
			{-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2},
			{-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2},
			{-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.3555555556e-1,-.5333333333e-1,.1777777778e-1,-.5333333333e-1,.1777777778e-1,.1777777778e-1,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2},
			{.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.1777777778e-1,-.1422222222,-.3555555556e-1,-.7111111111e-1,-.1422222222,.1777777778e-1,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.5333333333e-1,.7111111111e-1,0.,0.,-.7111111111e-1,-.5333333333e-1,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2},
			{-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.1777777778e-1,.1422222222,.3555555556e-1,.7111111111e-1,.1422222222,-.1777777778e-1,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2},
			{-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.1777777778e-1,-.1777777778e-1,-.1777777778e-1,.5333333333e-1,.5333333333e-1,.3555555556e-1,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2},
			{.2222222222e-2,.3333333333e-2,-.1111111111e-2,.3333333333e-2,-.1111111111e-2,-.1111111111e-2,-.4444444444e-2,-.6666666667e-2,.2222222222e-2,-.6666666667e-2,.2222222222e-2,.2222222222e-2,-.8888888889e-2,-.1333333333e-1,.4444444444e-2,-.1333333333e-1,.4444444444e-2,.4444444444e-2},
			{-.1111111111e-2,.8888888889e-2,.2222222222e-2,.4444444444e-2,.8888888889e-2,-.1111111111e-2,.2222222222e-2,-.1777777778e-1,-.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,.4444444444e-2,-.3555555556e-1,-.8888888889e-2,-.1777777778e-1,-.3555555556e-1,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.3333333333e-2,-.4444444444e-2,0.,0.,.4444444444e-2,.3333333333e-2,.6666666667e-2,.8888888889e-2,0.,0.,-.8888888889e-2,-.6666666667e-2,.1333333333e-1,.1777777778e-1,0.,0.,-.1777777778e-1,-.1333333333e-1},
			{.1111111111e-2,-.8888888889e-2,-.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1111111111e-2,-.2222222222e-2,.1777777778e-1,.4444444444e-2,.8888888889e-2,.1777777778e-1,-.2222222222e-2,-.4444444444e-2,.3555555556e-1,.8888888889e-2,.1777777778e-1,.3555555556e-1,-.4444444444e-2},
			{.1111111111e-2,.1111111111e-2,.1111111111e-2,-.3333333333e-2,-.3333333333e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,-.2222222222e-2,.6666666667e-2,.6666666667e-2,.4444444444e-2,-.4444444444e-2,-.4444444444e-2,-.4444444444e-2,.1333333333e-1,.1333333333e-1,.8888888889e-2}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_dofs_to_nodes:" << std::endl;
		static const T timestep_stiffness_dofs_to_nodes[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_stiffness_dofs_to_nodes[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_stiffness_dofs_to_nodes[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[9][18] = 
		{
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[9][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[9][18] = 
		{
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 9; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[18][9] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.2514157444e-1,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.1571348402e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.1005662977,-.5028314888e-1,.2514157444e-1,-.1257078722e-1,-.6285393608e-2,.3142696804e-2},
			{.6285393608e-2,.3142696804e-2,-.1571348402e-2,-.1257078722e-1,-.6285393608e-2,.3142696804e-2,-.2514157444e-1,-.1257078722e-1,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1257078722e-1,-.1005662977,-.1257078722e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,.3142696804e-2,.2514157444e-1,.3142696804e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.4022651909,-.5028314888e-1,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2},
			{.3142696804e-2,.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.5028314888e-1,-.6285393608e-2,-.1257078722e-1,-.1005662977,-.1257078722e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.6285393608e-2,-.1257078722e-1,-.2514157444e-1,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,-.1571348402e-2,.3142696804e-2,.6285393608e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.2514157444e-1,-.5028314888e-1,-.1005662977,.3142696804e-2,-.6285393608e-2,-.1257078722e-1},
			{-.1571348402e-2,.3142696804e-2,.6285393608e-2,.3142696804e-2,-.6285393608e-2,-.1257078722e-1,.6285393608e-2,-.1257078722e-1,-.2514157444e-1}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[18][9] = 
		{
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[18][9] = 
		{
			{.4444444444e-2,.2222222222e-2,-.1111111111e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.1777777778e-1,-.8888888889e-2,.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.4444444444e-2,.2222222222e-2,-.7111111111e-1,-.3555555556e-1,.1777777778e-1,-.8888888889e-2,-.4444444444e-2,.2222222222e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.1777777778e-1,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.4444444444e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.1111111111e-2},
			{.2222222222e-2,.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.8888888889e-2,-.7111111111e-1,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.2844444444,-.3555555556e-1,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.8888888889e-2,-.7111111111e-1,-.8888888889e-2,-.4444444444e-2,-.3555555556e-1,-.4444444444e-2,.2222222222e-2,.1777777778e-1,.2222222222e-2},
			{-.1111111111e-2,.2222222222e-2,.4444444444e-2,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.4444444444e-2,-.8888888889e-2,-.1777777778e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2222222222e-2,-.4444444444e-2,-.8888888889e-2,.1777777778e-1,-.3555555556e-1,-.7111111111e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4444444444e-2,-.8888888889e-2,-.1777777778e-1,.2222222222e-2,-.4444444444e-2,-.8888888889e-2,-.1111111111e-2,.2222222222e-2,.4444444444e-2}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 9; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[18][18] = 
		{
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,1.,0.,-1.,-.5000000000,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,0.,0.,-4.,0.,1.500000000}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[18][18] = 
		{
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-1.,.5000000000,1.,-.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.,-4.,1.500000000,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000,0.,.7500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[18][18] = 
		{
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,.7500000000,.3750000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,-.1250000000,.5000000000,.2500000000,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,1.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,.5000000000,.2500000000,.1250000000,0.,0.,0.,.2500000000,.1250000000,.6250000000e-1,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,.5000000000,.5000000000,.5000000000,0.,0.,0.,.2500000000,.2500000000,.2500000000,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,.5000000000,0.,0.,.2500000000,0.,.1250000000,.2500000000,0.,0.,.1250000000,0.,.6250000000e-1},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,.5000000000,.2500000000,.1250000000,.2500000000,.1250000000,.1250000000,.2500000000,.1250000000,.6250000000e-1,.1250000000,.6250000000e-1,.6250000000e-1},
			{1.,0.,0.,1.,0.,1.,.5000000000,0.,0.,.5000000000,0.,.5000000000,.2500000000,0.,0.,.2500000000,0.,.2500000000},
			{1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.,1.,.5000000000,.2500000000,0.,0.,0.},
			{1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.,1.,1.,1.,0.,0.,0.},
			{1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000,1.,0.,0.,.5000000000,0.,.2500000000},
			{1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000,1.,.5000000000,.2500000000,.5000000000,.2500000000,.2500000000},
			{1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.,1.,0.,0.,1.,0.,1.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[18][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,4.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,-4.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,4.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,-4.,0.,-4.,4.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,-4.,0.,2.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.,0.,0.,0.,0.,0.,4.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.},
			{9.,-12.,3.,0.,0.,0.,-12.,16.,-4.,0.,0.,0.,3.,-4.,1.,0.,0.,0.},
			{-6.,12.,-6.,0.,0.,0.,8.,-16.,8.,0.,0.,0.,-2.,4.,-2.,0.,0.,0.},
			{9.,0.,0.,-12.,0.,3.,-12.,0.,0.,16.,0.,-4.,3.,0.,0.,-4.,0.,1.},
			{-12.,12.,0.,12.,-12.,0.,16.,-16.,0.,-16.,16.,0.,-4.,4.,0.,4.,-4.,0.},
			{-6.,0.,0.,12.,0.,-6.,8.,0.,0.,-16.,0.,8.,-2.,0.,0.,4.,0.,-2.},
			{2.,0.,0.,0.,0.,0.,-4.,0.,0.,0.,0.,0.,2.,0.,0.,0.,0.,0.},
			{-6.,8.,-2.,0.,0.,0.,12.,-16.,4.,0.,0.,0.,-6.,8.,-2.,0.,0.,0.},
			{4.,-8.,4.,0.,0.,0.,-8.,16.,-8.,0.,0.,0.,4.,-8.,4.,0.,0.,0.},
			{-6.,0.,0.,8.,0.,-2.,12.,0.,0.,-16.,0.,4.,-6.,0.,0.,8.,0.,-2.},
			{8.,-8.,0.,-8.,8.,0.,-16.,16.,0.,16.,-16.,0.,8.,-8.,0.,-8.,8.,0.},
			{4.,0.,0.,-8.,0.,4.,-8.,0.,0.,16.,0.,-8.,4.,0.,0.,-8.,0.,4.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][18] = 
		{
			{0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.,0.,.2222222222,0.,.2222222222,.2222222222,0.,0.,.5555555556e-1,0.,.5555555556e-1,.5555555556e-1,0.}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[18][18] = 
		{
			{-.8333333333e-2,0.,.1388888889e-2,0.,.5555555556e-2,.1388888889e-2,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2,.2777777778e-2,0.,-.4629629630e-3,0.,-.1851851852e-2,-.4629629630e-3},
			{0.,-.4444444444e-1,0.,-.2222222222e-1,-.2222222222e-1,.5555555556e-2,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2,0.,.1481481481e-1,0.,.7407407407e-2,.7407407407e-2,-.1851851852e-2},
			{.1388888889e-2,0.,-.8333333333e-2,.5555555556e-2,0.,.1388888889e-2,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2,-.4629629630e-3,0.,.2777777778e-2,-.1851851852e-2,0.,-.4629629630e-3},
			{0.,-.2222222222e-1,.5555555556e-2,-.4444444444e-1,-.2222222222e-1,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.,0.,.7407407407e-2,-.1851851852e-2,.1481481481e-1,.7407407407e-2,0.},
			{.5555555556e-2,-.2222222222e-1,0.,-.2222222222e-1,-.4444444444e-1,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.,-.1851851852e-2,.7407407407e-2,0.,.7407407407e-2,.1481481481e-1,0.},
			{.1388888889e-2,.5555555556e-2,.1388888889e-2,0.,0.,-.8333333333e-2,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1,-.4629629630e-3,-.1851851852e-2,-.4629629630e-3,0.,0.,.2777777778e-2},
			{.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,0.,0.,0.,0.,0.,0.,-.1111111111e-1,0.,.1851851852e-2,0.,.7407407407e-2,.1851851852e-2},
			{0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,0.,0.,0.,0.,0.,0.,-.5925925926e-1,0.,-.2962962963e-1,-.2962962963e-1,.7407407407e-2},
			{-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,0.,0.,0.,0.,0.,0.,.1851851852e-2,0.,-.1111111111e-1,.7407407407e-2,0.,.1851851852e-2},
			{0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,0.,0.,0.,0.,0.,0.,-.2962962963e-1,.7407407407e-2,-.5925925926e-1,-.2962962963e-1,0.},
			{-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,0.,0.,0.,0.,0.,0.,.7407407407e-2,-.2962962963e-1,0.,-.2962962963e-1,-.5925925926e-1,0.},
			{-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,0.,0.,0.,0.,0.,0.,.1851851852e-2,.7407407407e-2,.1851851852e-2,0.,0.,-.1111111111e-1},
			{-.2777777778e-2,0.,.4629629630e-3,0.,.1851851852e-2,.4629629630e-3,.1111111111e-1,0.,-.1851851852e-2,0.,-.7407407407e-2,-.1851851852e-2,.8333333333e-2,0.,-.1388888889e-2,0.,-.5555555556e-2,-.1388888889e-2},
			{0.,-.1481481481e-1,0.,-.7407407407e-2,-.7407407407e-2,.1851851852e-2,0.,.5925925926e-1,0.,.2962962963e-1,.2962962963e-1,-.7407407407e-2,0.,.4444444444e-1,0.,.2222222222e-1,.2222222222e-1,-.5555555556e-2},
			{.4629629630e-3,0.,-.2777777778e-2,.1851851852e-2,0.,.4629629630e-3,-.1851851852e-2,0.,.1111111111e-1,-.7407407407e-2,0.,-.1851851852e-2,-.1388888889e-2,0.,.8333333333e-2,-.5555555556e-2,0.,-.1388888889e-2},
			{0.,-.7407407407e-2,.1851851852e-2,-.1481481481e-1,-.7407407407e-2,0.,0.,.2962962963e-1,-.7407407407e-2,.5925925926e-1,.2962962963e-1,0.,0.,.2222222222e-1,-.5555555556e-2,.4444444444e-1,.2222222222e-1,0.},
			{.1851851852e-2,-.7407407407e-2,0.,-.7407407407e-2,-.1481481481e-1,0.,-.7407407407e-2,.2962962963e-1,0.,.2962962963e-1,.5925925926e-1,0.,-.5555555556e-2,.2222222222e-1,0.,.2222222222e-1,.4444444444e-1,0.},
			{.4629629630e-3,.1851851852e-2,.4629629630e-3,0.,0.,-.2777777778e-2,-.1851851852e-2,-.7407407407e-2,-.1851851852e-2,0.,0.,.1111111111e-1,-.1388888889e-2,-.5555555556e-2,-.1388888889e-2,0.,0.,.8333333333e-2}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[6][18] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 6; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[6][18] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 6; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 18; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[18][6] = 
		{
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 6; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[18][6] = 
		{
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.},
			{-.1666666667e-1,0.,.2777777778e-2,0.,.1111111111e-1,.2777777778e-2},
			{0.,-.8888888889e-1,0.,-.4444444444e-1,-.4444444444e-1,.1111111111e-1},
			{.2777777778e-2,0.,-.1666666667e-1,.1111111111e-1,0.,.2777777778e-2},
			{0.,-.4444444444e-1,.1111111111e-1,-.8888888889e-1,-.4444444444e-1,0.},
			{.1111111111e-1,-.4444444444e-1,0.,-.4444444444e-1,-.8888888889e-1,0.},
			{.2777777778e-2,.1111111111e-1,.2777777778e-2,0.,0.,-.1666666667e-1}
		};
		for (int j = 0; j < 18; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 6; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
