
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (3)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (20)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (20)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (16)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (10)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*3+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[20],
		T io_hv[20]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,7.838367178,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,-2.993325910,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,7.838367178,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,-2.993325910,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,-2.828427124,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,3.919183589,0.,0.,4.200000000,0.,0.,-5.091168823,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,2.529822128,0.,0.,-1.807392228,0.,0.,2.190890230,0.,0.,6.260990336,0.,0.,8.082903771,0.,0.,0.},
			{2.645751311,0.,0.,0.,-1.496662955,0.,0.,.3055050464,0.,0.,5.184592560,0.,0.,-1.058300524,0.,0.,6.831300514,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,-2.828427124,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,3.919183589,0.,0.,4.200000000,0.,0.,-5.091168823,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,2.529822128,0.,0.,-1.807392228,0.,0.,2.190890230,0.,0.,6.260990336,0.,0.,8.082903771,0.,0.,0.},
			{2.645751311,0.,0.,0.,-1.496662955,0.,0.,.3055050464,0.,0.,5.184592560,0.,0.,-1.058300524,0.,0.,6.831300514,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_dofs_to_nodes(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.505360805,1.166047465,.6732178181,1.843682925,-1.428110653,-2.128901666,-1.187112542,.9195334206,1.370759491,3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.505360805,-1.166047465,-.6732178181,-1.843682925,1.428110653,2.128901666,-1.187112542,.9195334206,1.370759491,-3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,-2.484731210,1.924664519,1.111205578,-1.258649951,.9749460596,2.509295895,2.817981978,-2.182799455,3.185720379,-2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,0.,0.,0.,-1.360827635,1.054092553,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,.9428090414,-1.873942317,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,2.484731210,-1.924664519,-1.111205578,1.258649951,-.9749460596,-2.509295895,2.817981978,-2.182799455,3.185720379,2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,-.9793704052,.7586170537,.4379877603,-3.102332876,2.403056713,-4.427892772,.4377975572,-.3391165296,2.032696517,-.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,.9793704052,-.7586170537,-.4379877603,3.102332876,-2.403056713,4.427892772,.4377975572,-.3391165296,2.032696517,.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_dofs_to_nodes
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_dofs_to_nodes(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.505360805,1.166047465,.6732178181,1.843682925,-1.428110653,-2.128901666,-1.187112542,.9195334206,1.370759491,3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.505360805,-1.166047465,-.6732178181,-1.843682925,1.428110653,2.128901666,-1.187112542,.9195334206,1.370759491,-3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,-2.484731210,1.924664519,1.111205578,-1.258649951,.9749460596,2.509295895,2.817981978,-2.182799455,3.185720379,-2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,0.,0.,0.,-1.360827635,1.054092553,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,.9428090414,-1.873942317,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,2.484731210,-1.924664519,-1.111205578,1.258649951,-.9749460596,-2.509295895,2.817981978,-2.182799455,3.185720379,2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,-.9793704052,.7586170537,.4379877603,-3.102332876,2.403056713,-4.427892772,.4377975572,-.3391165296,2.032696517,-.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,.9793704052,-.7586170537,-.4379877603,3.102332876,-2.403056713,4.427892772,.4377975572,-.3391165296,2.032696517,.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,2.506647478,-4.341640786,5.605034154,1.172637045,-2.031066944,-2.525628468,2.867912544,-4.967370234,3.095507552,2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,2.506647478,-1.941640786,-1.121006831,1.172637045,-.9083207499,-2.525628468,2.867912544,-2.221475504,3.095507552,2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,2.506647478,1.941640786,-1.121006831,1.172637045,.9083207499,-2.525628468,2.867912544,2.221475504,3.095507552,2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,2.506647478,4.341640786,5.605034154,1.172637045,2.031066944,-2.525628468,2.867912544,4.967370234,3.095507552,2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,.9574541388,-1.658359214,2.140932538,3.070003641,-5.317402286,4.485220266,.4184228013,-.7247295507,1.964136704,.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,.9574541388,-.7416407864,-.4281865076,3.070003641,-2.378014595,4.485220266,.4184228013,-.3241089081,1.964136704,.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,.9574541388,.7416407864,-.4281865076,3.070003641,2.378014595,4.485220266,.4184228013,.3241089081,1.964136704,.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,.9574541388,1.658359214,2.140932538,3.070003641,5.317402286,4.485220266,.4184228013,.7247295507,1.964136704,.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,2.506647478,-4.341640786,5.605034154,1.172637045,-2.031066944,-2.525628468,2.867912544,-4.967370234,3.095507552,2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,2.506647478,-1.941640786,-1.121006831,1.172637045,-.9083207499,-2.525628468,2.867912544,-2.221475504,3.095507552,2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,2.506647478,1.941640786,-1.121006831,1.172637045,.9083207499,-2.525628468,2.867912544,2.221475504,3.095507552,2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,2.506647478,4.341640786,5.605034154,1.172637045,2.031066944,-2.525628468,2.867912544,4.967370234,3.095507552,2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,.9574541388,-1.658359214,2.140932538,3.070003641,-5.317402286,4.485220266,.4184228013,-.7247295507,1.964136704,.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,.9574541388,-.7416407864,-.4281865076,3.070003641,-2.378014595,4.485220266,.4184228013,-.3241089081,1.964136704,.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,.9574541388,.7416407864,-.4281865076,3.070003641,2.378014595,4.485220266,.4184228013,.3241089081,1.964136704,.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,.9574541388,1.658359214,2.140932538,3.070003641,5.317402286,4.485220266,.4184228013,.7247295507,1.964136704,.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.549193338,2.683281572,-3.464101616,1.897366595,-3.286335344,-2.190890230,-1.095445115,1.897366595,1.264911064,3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.549193338,1.200000000,.6928203232,1.897366595,-1.469693846,-2.190890230,-1.095445115,.8485281372,1.264911064,3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-1.549193338,-1.200000000,.6928203232,1.897366595,1.469693846,-2.190890230,-1.095445115,-.8485281372,1.264911064,3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.549193338,-2.683281572,-3.464101616,1.897366595,3.286335344,-2.190890230,-1.095445115,-1.897366595,1.264911064,3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.549193338,-2.683281572,3.464101616,-1.897366595,3.286335344,2.190890230,-1.095445115,1.897366595,1.264911064,-3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.549193338,-1.200000000,-.6928203232,-1.897366595,1.469693846,2.190890230,-1.095445115,.8485281372,1.264911064,-3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,1.549193338,1.200000000,-.6928203232,-1.897366595,-1.469693846,2.190890230,-1.095445115,-.8485281372,1.264911064,-3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.549193338,2.683281572,3.464101616,-1.897366595,-3.286335344,2.190890230,-1.095445115,-1.897366595,1.264911064,-3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.549193338,2.683281572,-3.464101616,1.897366595,-3.286335344,-2.190890230,-1.095445115,1.897366595,1.264911064,3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.549193338,1.200000000,.6928203232,1.897366595,-1.469693846,-2.190890230,-1.095445115,.8485281372,1.264911064,3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-1.549193338,-1.200000000,.6928203232,1.897366595,1.469693846,-2.190890230,-1.095445115,-.8485281372,1.264911064,3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.549193338,-2.683281572,-3.464101616,1.897366595,3.286335344,-2.190890230,-1.095445115,-1.897366595,1.264911064,3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.549193338,-2.683281572,3.464101616,-1.897366595,3.286335344,2.190890230,-1.095445115,1.897366595,1.264911064,-3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.549193338,-1.200000000,-.6928203232,-1.897366595,1.469693846,2.190890230,-1.095445115,.8485281372,1.264911064,-3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,1.549193338,1.200000000,-.6928203232,-1.897366595,-1.469693846,2.190890230,-1.095445115,-.8485281372,1.264911064,-3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.549193338,2.683281572,3.464101616,-1.897366595,-3.286335344,2.190890230,-1.095445115,-1.897366595,1.264911064,-3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,-.9574541388,1.658359214,-2.140932538,-3.070003641,5.317402286,-4.485220266,.4184228013,-.7247295507,1.964136704,-.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,-.9574541388,.7416407864,.4281865076,-3.070003641,2.378014595,-4.485220266,.4184228013,-.3241089081,1.964136704,-.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,-.9574541388,-.7416407864,.4281865076,-3.070003641,-2.378014595,-4.485220266,.4184228013,.3241089081,1.964136704,-.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,-.9574541388,-1.658359214,-2.140932538,-3.070003641,-5.317402286,-4.485220266,.4184228013,.7247295507,1.964136704,-.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,-2.506647478,4.341640786,-5.605034154,-1.172637045,2.031066944,2.525628468,2.867912544,-4.967370234,3.095507552,-2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,-2.506647478,1.941640786,1.121006831,-1.172637045,.9083207499,2.525628468,2.867912544,-2.221475504,3.095507552,-2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,-2.506647478,-1.941640786,1.121006831,-1.172637045,-.9083207499,2.525628468,2.867912544,2.221475504,3.095507552,-2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,-2.506647478,-4.341640786,-5.605034154,-1.172637045,-2.031066944,2.525628468,2.867912544,4.967370234,3.095507552,-2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[20],
		T o_output[16]
	) {
		static const T matrix[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,-.9574541388,1.658359214,-2.140932538,-3.070003641,5.317402286,-4.485220266,.4184228013,-.7247295507,1.964136704,-.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,-.9574541388,.7416407864,.4281865076,-3.070003641,2.378014595,-4.485220266,.4184228013,-.3241089081,1.964136704,-.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,-.9574541388,-.7416407864,.4281865076,-3.070003641,-2.378014595,-4.485220266,.4184228013,.3241089081,1.964136704,-.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,-.9574541388,-1.658359214,-2.140932538,-3.070003641,-5.317402286,-4.485220266,.4184228013,.7247295507,1.964136704,-.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,-2.506647478,4.341640786,-5.605034154,-1.172637045,2.031066944,2.525628468,2.867912544,-4.967370234,3.095507552,-2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,-2.506647478,1.941640786,1.121006831,-1.172637045,.9083207499,2.525628468,2.867912544,-2.221475504,3.095507552,-2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,-2.506647478,-1.941640786,1.121006831,-1.172637045,-.9083207499,2.525628468,2.867912544,2.221475504,3.095507552,-2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,-2.506647478,-4.341640786,-5.605034154,-1.172637045,-2.031066944,2.525628468,2.867912544,4.967370234,3.095507552,-2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 16; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1},
			{.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1,.1202813062,.2689571766,-.2689571769,-.1202813062,.1202813061,.2689571765,-.2689571768,-.1202813061,.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1},
			{-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1},
			{.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.1677614693e-1,.8388073465e-1,.8388073465e-1,.1677614693e-1,-.1149854221,-.5749271105,-.5749271104,-.1149854221,-.3928371006e-1,-.1964185503,-.1964185503,-.3928371006e-1},
			{-.3402069088e-1,-.7607257735e-1,.7607257741e-1,.3402069088e-1,-.2905713885e-1,-.6497373762e-1,.6497373767e-1,.2905713885e-1,.1991605932,.4453366242,-.4453366250,-.1991605932,.6804138176e-1,.1521451547,-.1521451548,-.6804138176e-1},
			{.4392052305e-1,-.4392052305e-1,-.4392052305e-1,.4392052305e-1,.3751260493e-1,-.3751260493e-1,-.3751260493e-1,.3751260493e-1,-.2571152203,.2571152203,.2571152203,-.2571152203,-.8784104609e-1,.8784104609e-1,.8784104609e-1,-.8784104609e-1},
			{-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.7216878367e-1,-.3608439185,-.3608439183,-.7216878367e-1},
			{.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.1250000000,.2795084969,-.2795084971,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,-.4166666667,-.4166666667,-.8333333333e-1},
			{-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1,-.1230880159,-.6154400797,-.6154400794,-.1230880159,-.4701543842e-1,-.2350771922,-.2350771921,-.4701543842e-1,0.,0.,0.,0.},
			{.5892556509e-1,.1317615690,-.1317615691,-.5892556509e-1,.2131946973,.4767178350,-.4767178355,-.2131946973,.8143312809e-1,.1820900098,-.1820900102,-.8143312809e-1,0.,0.,0.,0.},
			{-.7607257741e-1,.7607257741e-1,.7607257741e-1,-.7607257741e-1,-.2752331706,.2752331706,.2752331706,-.2752331706,-.1051297163,.1051297163,.1051297163,-.1051297163,0.,0.,0.,0.},
			{.4166666667e-1,.2083333333,.2083333333,.4166666667e-1,-.5758191722e-1,-.2879095862,-.2879095861,-.5758191722e-1,-.1507514162,-.7537570812,-.7537570810,-.1507514162,0.,0.,0.,0.},
			{-.7216878367e-1,-.1613743059,.1613743061,.7216878367e-1,.9973480625e-1,.2230138062,-.2230138064,-.9973480625e-1,.2611091120,.5838577232,-.5838577246,-.2611091120,0.,0.,0.,0.},
			{-.1924500898e-1,-.9622504495e-1,-.9622504490e-1,-.1924500898e-1,.5947034815e-1,.2973517409,.2973517408,.5947034815e-1,-.1556953932,-.7784769662,-.7784769659,-.1556953932,-.2886751347e-1,-.1443375674,-.1443375673,-.2886751347e-1},
			{-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.1408278060,-.7041390305,-.7041390302,-.1408278060,-.2054649999e-1,-.1027325000,-.1027325000,-.2054649999e-1,0.,0.,0.,0.},
			{.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.2439209152,.5454237466,-.5454237475,-.2439209152,.355875819e-1,.7957625215e-1,-.7957625224e-1,-.355875819e-1,0.,0.,0.,0.},
			{.4969039948e-1,.2484519974,.2484519974,.4969039948e-1,-.1242259987,-.6211299939,-.6211299937,-.1242259987,-.1242259987,-.6211299937,-.6211299937,-.1242259987,.1242259987e-1,.6211299937e-1,.6211299937e-1,.1242259987e-1},
			{-.7139328933e-1,-.3569664467,-.3569664467,-.7139328933e-1,-.1439226131,-.7196130654,-.7196130654,-.1439226131,-.3063570734e-2,-.1531785368e-1,-.1531785367e-1,-.3063570734e-2,-.2099802628e-2,-.1049901314e-1,-.1049901314e-1,-.2099802628e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1},
			{.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1,.1202813062,.2689571766,-.2689571769,-.1202813062,.1202813061,.2689571765,-.2689571768,-.1202813061,.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1},
			{-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1},
			{.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.1677614693e-1,.8388073465e-1,.8388073465e-1,.1677614693e-1,-.1149854221,-.5749271105,-.5749271104,-.1149854221,-.3928371006e-1,-.1964185503,-.1964185503,-.3928371006e-1},
			{-.3402069088e-1,-.7607257735e-1,.7607257741e-1,.3402069088e-1,-.2905713885e-1,-.6497373762e-1,.6497373767e-1,.2905713885e-1,.1991605932,.4453366242,-.4453366250,-.1991605932,.6804138176e-1,.1521451547,-.1521451548,-.6804138176e-1},
			{.4392052305e-1,-.4392052305e-1,-.4392052305e-1,.4392052305e-1,.3751260493e-1,-.3751260493e-1,-.3751260493e-1,.3751260493e-1,-.2571152203,.2571152203,.2571152203,-.2571152203,-.8784104609e-1,.8784104609e-1,.8784104609e-1,-.8784104609e-1},
			{-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.7216878367e-1,-.3608439185,-.3608439183,-.7216878367e-1},
			{.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.1250000000,.2795084969,-.2795084971,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,-.4166666667,-.4166666667,-.8333333333e-1},
			{-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1,-.1230880159,-.6154400797,-.6154400794,-.1230880159,-.4701543842e-1,-.2350771922,-.2350771921,-.4701543842e-1,0.,0.,0.,0.},
			{.5892556509e-1,.1317615690,-.1317615691,-.5892556509e-1,.2131946973,.4767178350,-.4767178355,-.2131946973,.8143312809e-1,.1820900098,-.1820900102,-.8143312809e-1,0.,0.,0.,0.},
			{-.7607257741e-1,.7607257741e-1,.7607257741e-1,-.7607257741e-1,-.2752331706,.2752331706,.2752331706,-.2752331706,-.1051297163,.1051297163,.1051297163,-.1051297163,0.,0.,0.,0.},
			{.4166666667e-1,.2083333333,.2083333333,.4166666667e-1,-.5758191722e-1,-.2879095862,-.2879095861,-.5758191722e-1,-.1507514162,-.7537570812,-.7537570810,-.1507514162,0.,0.,0.,0.},
			{-.7216878367e-1,-.1613743059,.1613743061,.7216878367e-1,.9973480625e-1,.2230138062,-.2230138064,-.9973480625e-1,.2611091120,.5838577232,-.5838577246,-.2611091120,0.,0.,0.,0.},
			{-.1924500898e-1,-.9622504495e-1,-.9622504490e-1,-.1924500898e-1,.5947034815e-1,.2973517409,.2973517408,.5947034815e-1,-.1556953932,-.7784769662,-.7784769659,-.1556953932,-.2886751347e-1,-.1443375674,-.1443375673,-.2886751347e-1},
			{-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.1408278060,-.7041390305,-.7041390302,-.1408278060,-.2054649999e-1,-.1027325000,-.1027325000,-.2054649999e-1,0.,0.,0.,0.},
			{.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.2439209152,.5454237466,-.5454237475,-.2439209152,.355875819e-1,.7957625215e-1,-.7957625224e-1,-.355875819e-1,0.,0.,0.,0.},
			{.4969039948e-1,.2484519974,.2484519974,.4969039948e-1,-.1242259987,-.6211299939,-.6211299937,-.1242259987,-.1242259987,-.6211299937,-.6211299937,-.1242259987,.1242259987e-1,.6211299937e-1,.6211299937e-1,.1242259987e-1},
			{-.7139328933e-1,-.3569664467,-.3569664467,-.7139328933e-1,-.1439226131,-.7196130654,-.7196130654,-.1439226131,-.3063570734e-2,-.1531785368e-1,-.1531785367e-1,-.3063570734e-2,-.2099802628e-2,-.1049901314e-1,-.1049901314e-1,-.2099802628e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1,-.1202813062,-.2689571766,.2689571769,.1202813062,-.1202813061,-.2689571765,.2689571768,.1202813061,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.4252586363,-.4252586361,-.8505172722e-1,-.8505172718e-1,-.4252586361,-.4252586359,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1,.1473139128,.3294039226,-.3294039229,-.1473139128,.1473139128,.3294039225,-.3294039229,-.1473139128,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.9820927513e-1,.4910463760,.4910463757,.9820927513e-1,.9820927513e-1,.4910463757,.4910463757,.9820927513e-1,.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1},
			{.2405626122e-1,.1202813062,.1202813061,.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1},
			{-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1},
			{.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1,.1202813059,-.1202813059,-.1202813059,.1202813059,-.1202813061,.1202813061,.1202813061,-.1202813061,-.5379143537e-1,.5379143537e-1,.5379143537e-1,-.5379143537e-1},
			{-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1,-.6588078447e-1,-.3294039226,-.3294039225,-.6588078447e-1,.6588078455e-1,.3294039229,.3294039229,.6588078455e-1,.2946278254e-1,.1473139128,.1473139128,.2946278254e-1},
			{.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1,.1141088660,.2551551810,-.2551551812,-.1141088660,-.1141088661,-.2551551812,.2551551816,.1141088661,-.5103103630e-1,-.1141088660,.1141088661,.5103103630e-1},
			{.3402069088e-1,.1701034544,.1701034544,.3402069088e-1,.7607257735e-1,.3803628869,.3803628867,.7607257735e-1,-.7607257742e-1,-.3803628873,-.3803628871,-.7607257742e-1,-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1},
			{-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.4392052306e-1,.2196026152,.2196026152,.4392052306e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,.4392052306e-1,.2196026152,.2196026152,.4392052306e-1},
			{.2227177016e-1,.1113588508,.1113588508,.2227177016e-1,-.4980119202e-1,-.2490059603,-.2490059601,-.4980119202e-1,.4980119204e-1,.2490059603,.2490059602,.4980119204e-1,-.2227177016e-1,-.1113588508,-.1113588508,-.2227177016e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1,-.1202813062,-.2689571766,.2689571769,.1202813062,-.1202813061,-.2689571765,.2689571768,.1202813061,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.4252586363,-.4252586361,-.8505172722e-1,-.8505172718e-1,-.4252586361,-.4252586359,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1,.1473139128,.3294039226,-.3294039229,-.1473139128,.1473139128,.3294039225,-.3294039229,-.1473139128,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.9820927513e-1,.4910463760,.4910463757,.9820927513e-1,.9820927513e-1,.4910463757,.4910463757,.9820927513e-1,.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1},
			{.2405626122e-1,.1202813062,.1202813061,.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1},
			{-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1},
			{.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1,.1202813059,-.1202813059,-.1202813059,.1202813059,-.1202813061,.1202813061,.1202813061,-.1202813061,-.5379143537e-1,.5379143537e-1,.5379143537e-1,-.5379143537e-1},
			{-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1,-.6588078447e-1,-.3294039226,-.3294039225,-.6588078447e-1,.6588078455e-1,.3294039229,.3294039229,.6588078455e-1,.2946278254e-1,.1473139128,.1473139128,.2946278254e-1},
			{.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1,.1141088660,.2551551810,-.2551551812,-.1141088660,-.1141088661,-.2551551812,.2551551816,.1141088661,-.5103103630e-1,-.1141088660,.1141088661,.5103103630e-1},
			{.3402069088e-1,.1701034544,.1701034544,.3402069088e-1,.7607257735e-1,.3803628869,.3803628867,.7607257735e-1,-.7607257742e-1,-.3803628873,-.3803628871,-.7607257742e-1,-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1},
			{-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.4392052306e-1,.2196026152,.2196026152,.4392052306e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,.4392052306e-1,.2196026152,.2196026152,.4392052306e-1},
			{.2227177016e-1,.1113588508,.1113588508,.2227177016e-1,-.4980119202e-1,-.2490059603,-.2490059601,-.4980119202e-1,.4980119204e-1,.2490059603,.2490059602,.4980119204e-1,-.2227177016e-1,-.1113588508,-.1113588508,-.2227177016e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{-.2777777778e-1,-.1388888889,-.1388888889,-.2777777778e-1,-.8130697173e-1,-.4065348587,-.4065348587,-.8130697173e-1,.1186252729e-1,.5931263656e-1,.5931263656e-1,.1186252729e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{.4811252245e-1,.1075828706,-.1075828707,-.4811252245e-1,.1408278060,.3149005471,-.3149005473,-.1408278060,-.2054650002e-1,-.4594337068e-1,.4594337076e-1,.2054650002e-1,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{-.6211299937e-1,.6211299937e-1,.6211299937e-1,-.6211299937e-1,-.1818079158,.1818079158,.1818079158,-.1818079158,.2652541750e-1,-.2652541750e-1,-.2652541750e-1,.2652541750e-1,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.5103103630e-1,-.2551551817,-.2551551816,-.5103103630e-1,-.3803628868e-1,-.1901814436,-.1901814434,-.3803628868e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.8838834762e-1,.1976423534,-.1976423537,-.8838834762e-1,.6588078456e-1,.1473139125,-.1473139127,-.6588078456e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{-.5892556509e-1,-.2946278254,-.2946278254,-.5892556509e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.3324493542e-1,.1662246771,.1662246771,.3324493542e-1,.8703637069e-1,.4351818534,.4351818534,.8703637069e-1,.2405626122e-1,.1202813062,.1202813061,.2405626122e-1},
			{0.,0.,0.,0.,-.5758191722e-1,-.1287570810,.1287570811,.5758191722e-1,-.1507514162,-.3370904137,.3370904144,.1507514162,-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1},
			{0.,0.,0.,0.,.7433793548e-1,-.7433793548e-1,-.7433793548e-1,.7433793548e-1,.1946192414,-.1946192414,-.1946192414,.1946192414,.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1},
			{0.,0.,0.,0.,.1065973485,.5329867431,.5329867429,.1065973485,.4071656407e-1,.2035828203,.2035828202,.4071656407e-1,-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1},
			{0.,0.,0.,0.,-.1846320236,-.4128497551,.4128497560,.1846320236,-.7052315768e-1,-.1576945743,.1576945745,.7052315768e-1,.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1},
			{.2041241452e-1,.1020620727,.1020620726,.2041241452e-1,.1100932682,.5504663412,.5504663412,.1100932682,-.4205188656e-1,-.2102594328,-.2102594328,-.4205188656e-1,.1360827635e-1,.6804138178e-1,.6804138176e-1,.1360827635e-1},
			{0.,0.,0.,0.,-.1452856955e-1,-.7264284777e-1,-.7264284777e-1,-.1452856955e-1,-.9958029666e-1,-.4979014835,-.4979014833,-.9958029666e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{0.,0.,0.,0.,.2516422062e-1,.5626890782e-1,-.5626890792e-1,-.2516422062e-1,.1724781332,.3856728304,-.3856728304,-.1724781332,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.8784104612e-2,.4392052305e-1,.4392052305e-1,.8784104612e-2,-.8784104609e-1,-.4392052307,-.4392052305,-.8784104609e-1,-.8784104609e-1,-.4392052305,-.4392052305,-.8784104609e-1,.3513641844e-1,.1756820922,.1756820922,.3513641844e-1},
			{.1484784677e-2,.7423923388e-2,.7423923386e-2,.1484784677e-2,.2166271727e-2,.1083135864e-1,.1083135863e-1,.2166271727e-2,.1017686557,.5088432789,.5088432787,.1017686557,.5048267902e-1,.2524133952,.2524133952,.5048267902e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[16],
		T o_output[20]
	) {
		static const T matrix[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{-.2777777778e-1,-.1388888889,-.1388888889,-.2777777778e-1,-.8130697173e-1,-.4065348587,-.4065348587,-.8130697173e-1,.1186252729e-1,.5931263656e-1,.5931263656e-1,.1186252729e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{.4811252245e-1,.1075828706,-.1075828707,-.4811252245e-1,.1408278060,.3149005471,-.3149005473,-.1408278060,-.2054650002e-1,-.4594337068e-1,.4594337076e-1,.2054650002e-1,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{-.6211299937e-1,.6211299937e-1,.6211299937e-1,-.6211299937e-1,-.1818079158,.1818079158,.1818079158,-.1818079158,.2652541750e-1,-.2652541750e-1,-.2652541750e-1,.2652541750e-1,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.5103103630e-1,-.2551551817,-.2551551816,-.5103103630e-1,-.3803628868e-1,-.1901814436,-.1901814434,-.3803628868e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.8838834762e-1,.1976423534,-.1976423537,-.8838834762e-1,.6588078456e-1,.1473139125,-.1473139127,-.6588078456e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{-.5892556509e-1,-.2946278254,-.2946278254,-.5892556509e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.3324493542e-1,.1662246771,.1662246771,.3324493542e-1,.8703637069e-1,.4351818534,.4351818534,.8703637069e-1,.2405626122e-1,.1202813062,.1202813061,.2405626122e-1},
			{0.,0.,0.,0.,-.5758191722e-1,-.1287570810,.1287570811,.5758191722e-1,-.1507514162,-.3370904137,.3370904144,.1507514162,-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1},
			{0.,0.,0.,0.,.7433793548e-1,-.7433793548e-1,-.7433793548e-1,.7433793548e-1,.1946192414,-.1946192414,-.1946192414,.1946192414,.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1},
			{0.,0.,0.,0.,.1065973485,.5329867431,.5329867429,.1065973485,.4071656407e-1,.2035828203,.2035828202,.4071656407e-1,-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1},
			{0.,0.,0.,0.,-.1846320236,-.4128497551,.4128497560,.1846320236,-.7052315768e-1,-.1576945743,.1576945745,.7052315768e-1,.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1},
			{.2041241452e-1,.1020620727,.1020620726,.2041241452e-1,.1100932682,.5504663412,.5504663412,.1100932682,-.4205188656e-1,-.2102594328,-.2102594328,-.4205188656e-1,.1360827635e-1,.6804138178e-1,.6804138176e-1,.1360827635e-1},
			{0.,0.,0.,0.,-.1452856955e-1,-.7264284777e-1,-.7264284777e-1,-.1452856955e-1,-.9958029666e-1,-.4979014835,-.4979014833,-.9958029666e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{0.,0.,0.,0.,.2516422062e-1,.5626890782e-1,-.5626890792e-1,-.2516422062e-1,.1724781332,.3856728304,-.3856728304,-.1724781332,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.8784104612e-2,.4392052305e-1,.4392052305e-1,.8784104612e-2,-.8784104609e-1,-.4392052307,-.4392052305,-.8784104609e-1,-.8784104609e-1,-.4392052305,-.4392052305,-.8784104609e-1,.3513641844e-1,.1756820922,.1756820922,.3513641844e-1},
			{.1484784677e-2,.7423923388e-2,.7423923386e-2,.1484784677e-2,.2166271727e-2,.1083135864e-1,.1083135863e-1,.2166271727e-2,.1017686557,.5088432789,.5088432787,.1017686557,.5048267902e-1,.2524133952,.2524133952,.5048267902e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 16; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,-8.500000000,-.4082482906,0.,0.,1.750000000,0.,-7.678758582,.3227486122,0.,-2.459674775,-.2645751311},
			{0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,0.,-.4082482906,0.,0.,1.750000000,0.,0.,.3227486122,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,-12.37436867,-.7216878367,0.,0.,2.828427124,0.,-13.26806944,.9128709293,0.,-5.850213671,-.9354143468},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,0.,-.7216878367,0.,0.,2.828427124,0.,0.,.9128709295,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,-7.216878367,0.,0.,0.,1.587713241,0.,-9.500000000,.9316949905,0.,-6.067673910,-1.636634177},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,1.587713241,0.,0.,.9316949905,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,-2.901185103,0.,0.,-2.627379873,-1.218935425},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,2.245365598,-.2500000000,0.,0.,0.,0.,.4949747467,.3162277660,0.,-1.004158022,-.3240370350},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,3.500000000,0.,0.,0.,-.2500000000,0.,1.558845727,.5809475019,0.,-1.565247584,-1.020504077},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,-.2500000000,0.,0.,.5809475019,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,1.275000000,0.,0.,-.5486726408,-.9983468483},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,-.6454972245,0.,0.,0.,-.3227486122,0.,.2236067977,.8333333333e-1,0.,.2886751347,-.1463850110},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.8333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6250000000,-.2746751327},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.1479019946,-.2500000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,-8.500000000,-.4082482906,0.,0.,1.750000000,0.,-7.678758582,.3227486122,0.,-2.459674775,-.2645751311},
			{0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,0.,-.4082482906,0.,0.,1.750000000,0.,0.,.3227486122,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,-12.37436867,-.7216878367,0.,0.,2.828427124,0.,-13.26806944,.9128709293,0.,-5.850213671,-.9354143468},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,0.,-.7216878367,0.,0.,2.828427124,0.,0.,.9128709295,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,-7.216878367,0.,0.,0.,1.587713241,0.,-9.500000000,.9316949905,0.,-6.067673910,-1.636634177},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,1.587713241,0.,0.,.9316949905,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,-2.901185103,0.,0.,-2.627379873,-1.218935425},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,2.245365598,-.2500000000,0.,0.,0.,0.,.4949747467,.3162277660,0.,-1.004158022,-.3240370350},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,3.500000000,0.,0.,0.,-.2500000000,0.,1.558845727,.5809475019,0.,-1.565247584,-1.020504077},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,-.2500000000,0.,0.,.5809475019,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,1.275000000,0.,0.,-.5486726408,-.9983468483},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,-.6454972245,0.,0.,0.,-.3227486122,0.,.2236067977,.8333333333e-1,0.,.2886751347,-.1463850110},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.8333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6250000000,-.2746751327},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.1479019946,-.2500000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,-.1154700539,2.904737510,0.,1.341640786,-11.64130577},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,0.,2.904737510,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,.1632993162,-2.464751509,0.,-1.581138830,10.85080642},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,0.,-2.464751509,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,-.2000000000,1.155301788,0.,1.290994449,-6.197388082},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,0.,1.155301788,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.3031088914,0.,0.,-.7267220925,1.974864371},
			{0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,.2828427124,3.320391543,0.,-1.278019301,-15.76980237},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,0.,3.320391543,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,-.3464101616,-1.484643616,0.,.4472135954,8.919961562},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-1.484643616,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.3250000000,0.,0.,.3227486122e-1,-2.776822654},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,-.4472135954,1.083333333,0.,-4.041451885,-8.002380602},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,0.,1.083333333,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,1.375000000,2.387560769},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,-1.031079532,0.,0.,-1.922725929,-1.575000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,-.1154700539,2.904737510,0.,1.341640786,-11.64130577},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,0.,2.904737510,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,.1632993162,-2.464751509,0.,-1.581138830,10.85080642},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,0.,-2.464751509,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,-.2000000000,1.155301788,0.,1.290994449,-6.197388082},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,0.,1.155301788,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.3031088914,0.,0.,-.7267220925,1.974864371},
			{0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,.2828427124,3.320391543,0.,-1.278019301,-15.76980237},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,0.,3.320391543,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,-.3464101616,-1.484643616,0.,.4472135954,8.919961562},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-1.484643616,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.3250000000,0.,0.,.3227486122e-1,-2.776822654},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,-.4472135954,1.083333333,0.,-4.041451885,-8.002380602},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,0.,1.083333333,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,1.375000000,2.387560769},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,-1.031079532,0.,0.,-1.922725929,-1.575000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,.1250000000,.4082482906,0.,0.,.1250000000,0.,-.2453738645,-.1613743061,0.,.5590169942e-1,.6614378278e-1},
			{0.,1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,0.,.4082482906,0.,0.,.1250000000,0.,0.,-.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,-.2651650429,.7216878367,0.,0.,-.7071067810,0.,.3163924252,.3651483717,0.,.3952847075e-1,-.1403121520},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,0.,.7216878367,0.,0.,-.7071067810,0.,0.,.3651483718,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,-.2165063510,0.,0.,0.,-.2165063510,0.,.4250000000,.5776508940,0.,-.6132223630,.3218713882},
			{0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.5776508940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.1816805231,.4937160930},
			{0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,-.4592793268,-.7500000000,0.,0.,0.,0.,.1237436867,-.3162277660,0.,-.1141088661,.1890216038},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,-.3750000000,0.,0.,0.,-.1250000000,0.,.3897114318,-.7423218081,0.,.2795084971,-.3496171374},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.7423218081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,.8068715303e-2,-.6942056633},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,-.4841229182,0.,0.,0.,.4841229182,0.,-.3913118960,.5416666667,0.,.3608439183,.1341862601},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,.4841229182,0.,0.,.5416666667,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,.3437500000,.5968901920},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,-.2577698829,0.,0.,-.4806814823,-.3937500000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,.1250000000,.4082482906,0.,0.,.1250000000,0.,-.2453738645,-.1613743061,0.,.5590169942e-1,.6614378278e-1},
			{0.,1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,0.,.4082482906,0.,0.,.1250000000,0.,0.,-.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,-.2651650429,.7216878367,0.,0.,-.7071067810,0.,.3163924252,.3651483717,0.,.3952847075e-1,-.1403121520},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,0.,.7216878367,0.,0.,-.7071067810,0.,0.,.3651483718,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,-.2165063510,0.,0.,0.,-.2165063510,0.,.4250000000,.5776508940,0.,-.6132223630,.3218713882},
			{0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.5776508940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.1816805231,.4937160930},
			{0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,-.4592793268,-.7500000000,0.,0.,0.,0.,.1237436867,-.3162277660,0.,-.1141088661,.1890216038},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,-.3750000000,0.,0.,0.,-.1250000000,0.,.3897114318,-.7423218081,0.,.2795084971,-.3496171374},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.7423218081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,.8068715303e-2,-.6942056633},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,-.4841229182,0.,0.,0.,.4841229182,0.,-.3913118960,.5416666667,0.,.3608439183,.1341862601},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,.4841229182,0.,0.,.5416666667,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,.3437500000,.5968901920},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,-.2577698829,0.,0.,-.4806814823,-.3937500000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,-.1250000000,-.4082482906,0.,0.,-.1250000000,0.,.2453738645,.1613743061,0.,-.5590169942e-1,-.6614378278e-1},
			{0.,1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,0.,-.4082482906,0.,0.,-.1250000000,0.,0.,.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,.2651650429,-.7216878367,0.,0.,-.3535533905,0.,.5103103630e-1,.4564354646,0.,.1185854122,-.2338535867},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,0.,-.7216878367,0.,0.,-.3535533905,0.,0.,.4564354646,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,.2165063510,0.,0.,0.,.7938566203,0.,.3750000000,.4658474951,0.,.9682458365e-1,-.4091585443},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,.7938566203,0.,0.,.4658474951,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.6568449682,-.3047338563},
			{0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,.2551551816,-.2500000000,0.,0.,-.6123724358,0.,-.4065863991,.1581138830,0.,.2966830521,-.8100925875e-1},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,0.,-.2500000000,0.,0.,-.6123724358,0.,0.,.1581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,-.6250000000,0.,0.,0.,-.1250000000,0.,-.4330127020e-1,.2904737510,0.,.6149186937,-.2551260193},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.2904737510,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,.3187500000,0.,0.,-.1371681602,-.2495867120},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,.4841229182,0.,0.,0.,-.1613743061,0.,-.5031152948,.4166666667e-1,0.,.2165063510,-.3659625275e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,-.1613743061,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,.3697549864e-1,-.6250000000e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,-.1250000000,-.4082482906,0.,0.,-.1250000000,0.,.2453738645,.1613743061,0.,-.5590169942e-1,-.6614378278e-1},
			{0.,1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,0.,-.4082482906,0.,0.,-.1250000000,0.,0.,.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,.2651650429,-.7216878367,0.,0.,-.3535533905,0.,.5103103630e-1,.4564354646,0.,.1185854122,-.2338535867},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,0.,-.7216878367,0.,0.,-.3535533905,0.,0.,.4564354646,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,.2165063510,0.,0.,0.,.7938566203,0.,.3750000000,.4658474951,0.,.9682458365e-1,-.4091585443},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,.7938566203,0.,0.,.4658474951,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.6568449682,-.3047338563},
			{0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,.2551551816,-.2500000000,0.,0.,-.6123724358,0.,-.4065863991,.1581138830,0.,.2966830521,-.8100925875e-1},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,0.,-.2500000000,0.,0.,-.6123724358,0.,0.,.1581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,-.6250000000,0.,0.,0.,-.1250000000,0.,-.4330127020e-1,.2904737510,0.,.6149186937,-.2551260193},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.2904737510,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,.3187500000,0.,0.,-.1371681602,-.2495867120},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,.4841229182,0.,0.,0.,-.1613743061,0.,-.5031152948,.4166666667e-1,0.,.2165063510,-.3659625275e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,-.1613743061,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,.3697549864e-1,-.6250000000e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1,.3535533905,.1178511302,.5892556509e-1,.1178511302,.2946278254e-1,.5892556509e-1,.2357022604,.7856742012e-1,.7856742012e-1,.1767766952},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.2041241452,.6804138176e-1,.6804138176e-1,.1837117307},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.1756820922e-1,.7905694150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000,0.,-.4166666667e-1,-.3333333333e-1,.8333333333e-1,.8333333333e-2,.6666666667e-1,0.,-.2777777778e-1,.5555555556e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,.1242259987e-1,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1,0.,0.,.6804138176e-2,0.,-.1020620726e-1,.2041241452e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.,0.,.7216878367e-1,.5773502693e-1,0.,.1443375673e-1,0.,0.,.4811252245e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.,0.,0.,-.1178511302e-1,0.,.1178511302e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.,0.,0.,.1521451549e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1,.3535533905,.1178511302,.5892556509e-1,.1178511302,.2946278254e-1,.5892556509e-1,.2357022604,.7856742012e-1,.7856742012e-1,.1767766952},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.2041241452,.6804138176e-1,.6804138176e-1,.1837117307},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.1756820922e-1,.7905694150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000,0.,-.4166666667e-1,-.3333333333e-1,.8333333333e-1,.8333333333e-2,.6666666667e-1,0.,-.2777777778e-1,.5555555556e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,.1242259987e-1,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1,0.,0.,.6804138176e-2,0.,-.1020620726e-1,.2041241452e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.,0.,.7216878367e-1,.5773502693e-1,0.,.1443375673e-1,0.,0.,.4811252245e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.,0.,0.,-.1178511302e-1,0.,.1178511302e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.,0.,0.,.1521451549e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-8.485281372,14.69693846,9.797958972,-32.86335345,56.92099788,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,0.,0.,0.,6.,-10.39230485,13.41640786,-19.59591794,33.94112549,42.42640686,3.464101616,-6.,7.745966692,-25.45584412,44.09081537,63.68673332,-10.95445115,18.97366596,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,-117.5755077,32.86335345,-56.92099788,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,-127.2792206,0.,0.,0.,21.21320343,-36.74234614,-161.6663230,5.477225575,-9.486832980,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,0.,0.,265.6313234,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,0.,0.,0.,0.,0.,102.8785692,0.,0.,44.27188724,7.483314774},
			{0.,4.898979486,-18.97366596,44.89988864,0.,-6.928203232,26.83281572,0.,8.485281372,0.,0.,-12.,46.47580015,0.,14.69693846,0.,0.,18.97366596,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,0.,-29.39387692,0.,0.,-113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,20.78460970,-80.49844717,0.,-67.88225098,0.,0.,12.,-46.47580015,0.,-88.18163075,0.,0.,-37.94733192,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,0.,0.,0.,0.,0.,73.48469229,0.,0.,18.97366596,0.,0.},
			{0.,0.,18.97366596,-112.2497216,0.,0.,-26.83281572,0.,0.,0.,0.,0.,-46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,80.49844717,0.,0.,0.,0.,0.,46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-8.485281372,14.69693846,9.797958972,-32.86335345,56.92099788,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,0.,0.,0.,6.,-10.39230485,13.41640786,-19.59591794,33.94112549,42.42640686,3.464101616,-6.,7.745966692,-25.45584412,44.09081537,63.68673332,-10.95445115,18.97366596,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,-117.5755077,32.86335345,-56.92099788,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,-127.2792206,0.,0.,0.,21.21320343,-36.74234614,-161.6663230,5.477225575,-9.486832980,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,0.,0.,265.6313234,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,0.,0.,0.,0.,0.,102.8785692,0.,0.,44.27188724,7.483314774},
			{0.,4.898979486,-18.97366596,44.89988864,0.,-6.928203232,26.83281572,0.,8.485281372,0.,0.,-12.,46.47580015,0.,14.69693846,0.,0.,18.97366596,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,0.,-29.39387692,0.,0.,-113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,20.78460970,-80.49844717,0.,-67.88225098,0.,0.,12.,-46.47580015,0.,-88.18163075,0.,0.,-37.94733192,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,0.,0.,0.,0.,0.,73.48469229,0.,0.,18.97366596,0.,0.},
			{0.,0.,18.97366596,-112.2497216,0.,0.,-26.83281572,0.,0.,0.,0.,0.,-46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,80.49844717,0.,0.,0.,0.,0.,46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[20],
		T o_output[1]
	) {
		static const T matrix[1][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[20],
		T o_output[1]
	) {
		static const T matrix[1][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[20],
		T o_output[10]
	) {
		static const T matrix[10][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[20],
		T o_output[10]
	) {
		static const T matrix[10][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[20],
		T o_output[10]
	) {
		static const T matrix[10][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.505360805,-2.607361398,-3.366089089,1.843682925,3.193352500,-2.128901666,-1.187112542,-2.056139236,1.370759491,3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.505360805,2.607361398,3.366089089,-1.843682925,-3.193352500,2.128901666,-1.187112542,-2.056139236,1.370759491,-3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,-2.484731210,-4.303680699,-5.556027890,-1.258649951,-2.180045665,2.509295895,2.817981978,4.880887957,3.185720379,-2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,0.,0.,0.,-1.360827635,-2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-1.873942317,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,2.484731210,4.303680699,5.556027890,1.258649951,2.180045665,-2.509295895,2.817981978,4.880887957,3.185720379,2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,-.9793704052,-1.696319301,-2.189938801,-3.102332876,-5.373398163,-4.427892772,.4377975572,.7582876121,2.032696517,-.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,.9793704052,1.696319301,2.189938801,3.102332876,5.373398163,4.427892772,.4377975572,.7582876121,2.032696517,.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[20],
		T o_output[10]
	) {
		static const T matrix[10][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.505360805,-2.607361398,-3.366089089,1.843682925,3.193352500,-2.128901666,-1.187112542,-2.056139236,1.370759491,3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.505360805,2.607361398,3.366089089,-1.843682925,-3.193352500,2.128901666,-1.187112542,-2.056139236,1.370759491,-3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,-2.484731210,-4.303680699,-5.556027890,-1.258649951,-2.180045665,2.509295895,2.817981978,4.880887957,3.185720379,-2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,0.,0.,0.,-1.360827635,-2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-1.873942317,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,2.484731210,4.303680699,5.556027890,1.258649951,2.180045665,-2.509295895,2.817981978,4.880887957,3.185720379,2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,-.9793704052,-1.696319301,-2.189938801,-3.102332876,-5.373398163,-4.427892772,.4377975572,.7582876121,2.032696517,-.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,.9793704052,1.696319301,2.189938801,3.102332876,5.373398163,4.427892772,.4377975572,.7582876121,2.032696517,.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[10],
		T o_output[20]
	) {
		static const T matrix[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{.2322012545e-1,.1006582171,.1006582171,.2322012547e-1,.1006582171,.5511351922,.1006582171,.1006582170,.1006582172,.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{.3546932749e-1,.1537579664,.1537579663,.3546932754e-1,.1537579662,.8418729121,.1537579664,.1537579663,.1537579664,.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{-.1985351453e-2,-.1423522158,-.1423522158,-.1985351454e-2,-.9261271237e-1,0.,-.9261271240e-1,.2349649283,.2349649284,.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{.2594854185e-1,.9132364797e-1,.9132364797e-1,.2594854188e-1,-.6311588137e-1,-.2272843224,-.6311588142e-1,.1330298246e-1,.1330298247e-1,.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{-.3438729585e-2,-.1891270389,.1891270390,.3438729588e-2,-.2178441547,0.,.2178441546,-.2871711568e-1,.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{.5751898119e-1,.6618067742e-1,-.6618067742e-1,-.5751898117e-1,-.6756787835e-1,0.,.6756787845e-1,-.1337485559,.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{.6776309270e-1,-.7425668538e-1,-.7425668538e-1,.6776309264e-1,.9841195758e-1,-.2032892782,.9841195768e-1,.1297307049e-1,.1297307048e-1,-.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[10],
		T o_output[20]
	) {
		static const T matrix[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{.2322012545e-1,.1006582171,.1006582171,.2322012547e-1,.1006582171,.5511351922,.1006582171,.1006582170,.1006582172,.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{.3546932749e-1,.1537579664,.1537579663,.3546932754e-1,.1537579662,.8418729121,.1537579664,.1537579663,.1537579664,.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{-.1985351453e-2,-.1423522158,-.1423522158,-.1985351454e-2,-.9261271237e-1,0.,-.9261271240e-1,.2349649283,.2349649284,.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{.2594854185e-1,.9132364797e-1,.9132364797e-1,.2594854188e-1,-.6311588137e-1,-.2272843224,-.6311588142e-1,.1330298246e-1,.1330298247e-1,.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{-.3438729585e-2,-.1891270389,.1891270390,.3438729588e-2,-.2178441547,0.,.2178441546,-.2871711568e-1,.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{.5751898119e-1,.6618067742e-1,-.6618067742e-1,-.5751898117e-1,-.6756787835e-1,0.,.6756787845e-1,-.1337485559,.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{.6776309270e-1,-.7425668538e-1,-.7425668538e-1,.6776309264e-1,.9841195758e-1,-.2032892782,.9841195768e-1,.1297307049e-1,.1297307048e-1,-.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[10],
		T o_output[20]
	) {
		static const T matrix[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{-.2322012545e-1,-.1006582171,-.1006582171,-.2322012547e-1,-.1006582171,-.5511351922,-.1006582171,-.1006582170,-.1006582172,-.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{-.3546932749e-1,-.1537579664,-.1537579663,-.3546932754e-1,-.1537579662,-.8418729121,-.1537579664,-.1537579663,-.1537579664,-.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{.1985351453e-2,.1423522158,.1423522158,.1985351454e-2,.9261271237e-1,0.,.9261271240e-1,-.2349649283,-.2349649284,-.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{-.2594854185e-1,-.9132364797e-1,-.9132364797e-1,-.2594854188e-1,.6311588137e-1,.2272843224,.6311588142e-1,-.1330298246e-1,-.1330298247e-1,-.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{.3438729585e-2,.1891270389,-.1891270390,-.3438729588e-2,.2178441547,0.,-.2178441546,.2871711568e-1,-.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{-.5751898119e-1,-.6618067742e-1,.6618067742e-1,.5751898117e-1,.6756787835e-1,0.,-.6756787845e-1,.1337485559,-.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{-.6776309270e-1,.7425668538e-1,.7425668538e-1,-.6776309264e-1,-.9841195758e-1,.2032892782,-.9841195768e-1,-.1297307049e-1,-.1297307048e-1,.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[10],
		T o_output[20]
	) {
		static const T matrix[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{-.2322012545e-1,-.1006582171,-.1006582171,-.2322012547e-1,-.1006582171,-.5511351922,-.1006582171,-.1006582170,-.1006582172,-.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{-.3546932749e-1,-.1537579664,-.1537579663,-.3546932754e-1,-.1537579662,-.8418729121,-.1537579664,-.1537579663,-.1537579664,-.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{.1985351453e-2,.1423522158,.1423522158,.1985351454e-2,.9261271237e-1,0.,.9261271240e-1,-.2349649283,-.2349649284,-.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{-.2594854185e-1,-.9132364797e-1,-.9132364797e-1,-.2594854188e-1,.6311588137e-1,.2272843224,.6311588142e-1,-.1330298246e-1,-.1330298247e-1,-.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{.3438729585e-2,.1891270389,-.1891270390,-.3438729588e-2,.2178441547,0.,-.2178441546,.2871711568e-1,-.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{-.5751898119e-1,-.6618067742e-1,.6618067742e-1,.5751898117e-1,.6756787835e-1,0.,-.6756787845e-1,.1337485559,-.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{-.6776309270e-1,.7425668538e-1,.7425668538e-1,-.6776309264e-1,-.9841195758e-1,.2032892782,-.9841195768e-1,-.1297307049e-1,-.1297307048e-1,.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 23600;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[20][20] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,7.838367178,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,-2.993325910,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,-2.828427124,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,3.919183589,0.,0.,4.200000000,0.,0.,-5.091168823,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,2.529822128,0.,0.,-1.807392228,0.,0.,2.190890230,0.,0.,6.260990336,0.,0.,8.082903771,0.,0.,0.},
			{2.645751311,0.,0.,0.,-1.496662955,0.,0.,.3055050464,0.,0.,5.184592560,0.,0.,-1.058300524,0.,0.,6.831300514,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_dofs_to_nodes:" << std::endl;
		static const T timestep_stiffness_dofs_to_nodes[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.505360805,1.166047465,.6732178181,1.843682925,-1.428110653,-2.128901666,-1.187112542,.9195334206,1.370759491,3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.505360805,-1.166047465,-.6732178181,-1.843682925,1.428110653,2.128901666,-1.187112542,.9195334206,1.370759491,-3.342659338},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,-2.484731210,1.924664519,1.111205578,-1.258649951,.9749460596,2.509295895,2.817981978,-2.182799455,3.185720379,-2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,0.,0.,0.,-1.360827635,1.054092553,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,.9428090414,-1.873942317,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.3036806991,.2352300578,.1358101372,-1.132775699,.8774442833,1.229948799,2.484731210,-1.924664519,-1.111205578,1.258649951,-.9749460596,-2.509295895,2.817981978,-2.182799455,3.185720379,2.761597556},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,-.9793704052,.7586170537,.4379877603,-3.102332876,2.403056713,-4.427892772,.4377975572,-.3391165296,2.032696517,-.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.303680699,-1.784423396,-1.030237328,.9961259674,-.7715958560,-1.348290938,.9793704052,-.7586170537,-.4379877603,3.102332876,-2.403056713,4.427892772,.4377975572,-.3391165296,2.032696517,.1691075797},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_dofs_to_nodes[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_dofs_to_nodes[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,2.506647478,-4.341640786,5.605034154,1.172637045,-2.031066944,-2.525628468,2.867912544,-4.967370234,3.095507552,2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,2.506647478,-1.941640786,-1.121006831,1.172637045,-.9083207499,-2.525628468,2.867912544,-2.221475504,3.095507552,2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,2.506647478,1.941640786,-1.121006831,1.172637045,.9083207499,-2.525628468,2.867912544,2.221475504,3.095507552,2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,2.506647478,4.341640786,5.605034154,1.172637045,2.031066944,-2.525628468,2.867912544,4.967370234,3.095507552,2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,.9574541388,-1.658359214,2.140932538,3.070003641,-5.317402286,4.485220266,.4184228013,-.7247295507,1.964136704,.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,.9574541388,-.7416407864,-.4281865076,3.070003641,-2.378014595,4.485220266,.4184228013,-.3241089081,1.964136704,.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,.9574541388,.7416407864,-.4281865076,3.070003641,2.378014595,4.485220266,.4184228013,.3241089081,1.964136704,.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,.9574541388,1.658359214,2.140932538,3.070003641,5.317402286,4.485220266,.4184228013,.7247295507,1.964136704,.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 16; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.549193338,2.683281572,-3.464101616,1.897366595,-3.286335344,-2.190890230,-1.095445115,1.897366595,1.264911064,3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-1.549193338,1.200000000,.6928203232,1.897366595,-1.469693846,-2.190890230,-1.095445115,.8485281372,1.264911064,3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-1.549193338,-1.200000000,.6928203232,1.897366595,1.469693846,-2.190890230,-1.095445115,-.8485281372,1.264911064,3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.549193338,-2.683281572,-3.464101616,1.897366595,3.286335344,-2.190890230,-1.095445115,-1.897366595,1.264911064,3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.549193338,-2.683281572,3.464101616,-1.897366595,3.286335344,2.190890230,-1.095445115,1.897366595,1.264911064,-3.346640106},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,1.549193338,-1.200000000,-.6928203232,-1.897366595,1.469693846,2.190890230,-1.095445115,.8485281372,1.264911064,-3.346640106},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,1.549193338,1.200000000,-.6928203232,-1.897366595,-1.469693846,2.190890230,-1.095445115,-.8485281372,1.264911064,-3.346640106},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.549193338,2.683281572,3.464101616,-1.897366595,-3.286335344,2.190890230,-1.095445115,-1.897366595,1.264911064,-3.346640106},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,3.464101616,-2.683281572,-1.549193338,-4.242640686,3.286335344,4.898979486,5.477225575,-4.242640686,-6.324555320,7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,3.464101616,2.683281572,-1.549193338,-4.242640686,-3.286335344,4.898979486,5.477225575,4.242640686,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774}
		};
		for (int j = 0; j < 16; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[16][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,4.,-3.098386677,-1.788854382,7.348469229,-5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,4.,3.098386677,-1.788854382,7.348469229,5.692099786,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.341640786,-4.055840816,5.236067976,1.095445115,-1.897366595,-1.264911064,-.9574541388,1.658359214,-2.140932538,-3.070003641,5.317402286,-4.485220266,.4184228013,-.7247295507,1.964136704,-.1580069126},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,2.341640786,-1.813827154,-1.047213595,1.095445115,-.8485281372,-1.264911064,-.9574541388,.7416407864,.4281865076,-3.070003641,2.378014595,-4.485220266,.4184228013,-.3241089081,1.964136704,-.1580069126},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,2.341640786,1.813827154,-1.047213595,1.095445115,.8485281372,-1.264911064,-.9574541388,-.7416407864,.4281865076,-3.070003641,-2.378014595,-4.485220266,.4184228013,.3241089081,1.964136704,-.1580069126},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.341640786,4.055840816,5.236067976,1.095445115,1.897366595,-1.264911064,-.9574541388,-1.658359214,-2.140932538,-3.070003641,-5.317402286,-4.485220266,.4184228013,.7247295507,1.964136704,-.1580069126},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.341640786,.5917391998,-.7639320216,-1.095445115,1.897366595,1.264911064,-2.506647478,4.341640786,-5.605034154,-1.172637045,2.031066944,2.525628468,2.867912544,-4.967370234,3.095507552,-2.835318997},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-.341640786,.2646338150,.1527864043,-1.095445115,.8485281372,1.264911064,-2.506647478,1.941640786,1.121006831,-1.172637045,.9083207499,2.525628468,2.867912544,-2.221475504,3.095507552,-2.835318997},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-.341640786,-.2646338150,.1527864043,-1.095445115,-.8485281372,1.264911064,-2.506647478,-1.941640786,1.121006831,-1.172637045,-.9083207499,2.525628468,2.867912544,2.221475504,3.095507552,-2.835318997},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.341640786,-.5917391998,-.7639320216,-1.095445115,-1.897366595,1.264911064,-2.506647478,-4.341640786,-5.605034154,-1.172637045,-2.031066944,2.525628468,2.867912544,4.967370234,3.095507552,-2.835318997},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-1.095445115,-.6324555318,1.673320053,-2.,1.549193338,.8944271908,2.449489743,-1.897366595,-2.828427124,-3.464101616,2.683281572,1.549193338,4.242640686,-3.286335344,-4.898979486,5.477225575,-4.242640686,-6.324555320,-7.483314774},
			{1.414213562,1.095445115,-.6324555318,-1.673320053,-2.,-1.549193338,.8944271908,2.449489743,1.897366595,-2.828427124,-3.464101616,-2.683281572,1.549193338,4.242640686,3.286335344,-4.898979486,5.477225575,4.242640686,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774}
		};
		for (int j = 0; j < 16; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[20][16] = 
		{
			{-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.6944444444e-1,-.3472222222,-.3472222222,-.6944444444e-1,-.1388888889e-1,-.6944444444e-1,-.6944444444e-1,-.1388888889e-1},
			{.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1,.1202813062,.2689571766,-.2689571769,-.1202813062,.1202813061,.2689571765,-.2689571768,-.1202813061,.2405626122e-1,.5379143531e-1,-.5379143537e-1,-.2405626122e-1},
			{-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.3105649968e-1,.3105649968e-1,.3105649968e-1,-.3105649968e-1},
			{.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.7874259854e-1,-.1760738030,.1760738030,-.7874259854e-1,.1574851971e-1,-.3521476059e-1,.3521476060e-1,-.1574851971e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.1677614693e-1,.8388073465e-1,.8388073465e-1,.1677614693e-1,-.1149854221,-.5749271105,-.5749271104,-.1149854221,-.3928371006e-1,-.1964185503,-.1964185503,-.3928371006e-1},
			{-.3402069088e-1,-.7607257735e-1,.7607257741e-1,.3402069088e-1,-.2905713885e-1,-.6497373762e-1,.6497373767e-1,.2905713885e-1,.1991605932,.4453366242,-.4453366250,-.1991605932,.6804138176e-1,.1521451547,-.1521451548,-.6804138176e-1},
			{.4392052305e-1,-.4392052305e-1,-.4392052305e-1,.4392052305e-1,.3751260493e-1,-.3751260493e-1,-.3751260493e-1,.3751260493e-1,-.2571152203,.2571152203,.2571152203,-.2571152203,-.8784104609e-1,.8784104609e-1,.8784104609e-1,-.8784104609e-1},
			{-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.7216878367e-1,-.3608439185,-.3608439183,-.7216878367e-1},
			{.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.1250000000,.2795084969,-.2795084971,-.1250000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,-.4166666667,-.4166666667,-.8333333333e-1},
			{-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1,-.1230880159,-.6154400797,-.6154400794,-.1230880159,-.4701543842e-1,-.2350771922,-.2350771921,-.4701543842e-1,0.,0.,0.,0.},
			{.5892556509e-1,.1317615690,-.1317615691,-.5892556509e-1,.2131946973,.4767178350,-.4767178355,-.2131946973,.8143312809e-1,.1820900098,-.1820900102,-.8143312809e-1,0.,0.,0.,0.},
			{-.7607257741e-1,.7607257741e-1,.7607257741e-1,-.7607257741e-1,-.2752331706,.2752331706,.2752331706,-.2752331706,-.1051297163,.1051297163,.1051297163,-.1051297163,0.,0.,0.,0.},
			{.4166666667e-1,.2083333333,.2083333333,.4166666667e-1,-.5758191722e-1,-.2879095862,-.2879095861,-.5758191722e-1,-.1507514162,-.7537570812,-.7537570810,-.1507514162,0.,0.,0.,0.},
			{-.7216878367e-1,-.1613743059,.1613743061,.7216878367e-1,.9973480625e-1,.2230138062,-.2230138064,-.9973480625e-1,.2611091120,.5838577232,-.5838577246,-.2611091120,0.,0.,0.,0.},
			{-.1924500898e-1,-.9622504495e-1,-.9622504490e-1,-.1924500898e-1,.5947034815e-1,.2973517409,.2973517408,.5947034815e-1,-.1556953932,-.7784769662,-.7784769659,-.1556953932,-.2886751347e-1,-.1443375674,-.1443375673,-.2886751347e-1},
			{-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.1408278060,-.7041390305,-.7041390302,-.1408278060,-.2054649999e-1,-.1027325000,-.1027325000,-.2054649999e-1,0.,0.,0.,0.},
			{.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.2439209152,.5454237466,-.5454237475,-.2439209152,.355875819e-1,.7957625215e-1,-.7957625224e-1,-.355875819e-1,0.,0.,0.,0.},
			{.4969039948e-1,.2484519974,.2484519974,.4969039948e-1,-.1242259987,-.6211299939,-.6211299937,-.1242259987,-.1242259987,-.6211299937,-.6211299937,-.1242259987,.1242259987e-1,.6211299937e-1,.6211299937e-1,.1242259987e-1},
			{-.7139328933e-1,-.3569664467,-.3569664467,-.7139328933e-1,-.1439226131,-.7196130654,-.7196130654,-.1439226131,-.3063570734e-2,-.1531785368e-1,-.1531785367e-1,-.3063570734e-2,-.2099802628e-2,-.1049901314e-1,-.1049901314e-1,-.2099802628e-2}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 16; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.6944444444e-1,.3472222222,.3472222222,.6944444444e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1,-.1202813062,-.2689571766,.2689571769,.1202813062,-.1202813061,-.2689571765,.2689571768,.1202813061,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1,.1552824984,-.1552824984,-.1552824984,.1552824984,.1552824984,-.1552824984,-.1552824984,.1552824984,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.4252586363,-.4252586361,-.8505172722e-1,-.8505172718e-1,-.4252586361,-.4252586359,-.8505172718e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1,.1473139128,.3294039226,-.3294039229,-.1473139128,.1473139128,.3294039225,-.3294039229,-.1473139128,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1,.9820927513e-1,.4910463760,.4910463757,.9820927513e-1,.9820927513e-1,.4910463757,.4910463757,.9820927513e-1,.1964185503e-1,.9820927513e-1,.9820927513e-1,.1964185503e-1},
			{.2405626122e-1,.1202813062,.1202813061,.2405626122e-1,.5379143531e-1,.2689571766,.2689571765,.5379143531e-1,-.5379143537e-1,-.2689571769,-.2689571768,-.5379143537e-1,-.2405626122e-1,-.1202813062,-.1202813061,-.2405626122e-1},
			{-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1,-.9316949897e-1,-.2083333333,.2083333333,.9316949897e-1,.9316949905e-1,.2083333333,-.2083333333,-.9316949905e-1,.4166666667e-1,.9316949897e-1,-.9316949905e-1,-.4166666667e-1},
			{.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1,.1202813059,-.1202813059,-.1202813059,.1202813059,-.1202813061,.1202813061,.1202813061,-.1202813061,-.5379143537e-1,.5379143537e-1,.5379143537e-1,-.5379143537e-1},
			{-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1,-.6588078447e-1,-.3294039226,-.3294039225,-.6588078447e-1,.6588078455e-1,.3294039229,.3294039229,.6588078455e-1,.2946278254e-1,.1473139128,.1473139128,.2946278254e-1},
			{.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1,.1141088660,.2551551810,-.2551551812,-.1141088660,-.1141088661,-.2551551812,.2551551816,.1141088661,-.5103103630e-1,-.1141088660,.1141088661,.5103103630e-1},
			{.3402069088e-1,.1701034544,.1701034544,.3402069088e-1,.7607257735e-1,.3803628869,.3803628867,.7607257735e-1,-.7607257742e-1,-.3803628873,-.3803628871,-.7607257742e-1,-.3402069088e-1,-.1701034544,-.1701034544,-.3402069088e-1},
			{-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.4392052306e-1,.2196026152,.2196026152,.4392052306e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,-.4392052305e-1,-.2196026152,-.2196026152,-.4392052305e-1,.4392052306e-1,.2196026152,.2196026152,.4392052306e-1},
			{.2227177016e-1,.1113588508,.1113588508,.2227177016e-1,-.4980119202e-1,-.2490059603,-.2490059601,-.4980119202e-1,.4980119204e-1,.2490059603,.2490059602,.4980119204e-1,-.2227177016e-1,-.1113588508,-.1113588508,-.2227177016e-1}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 16; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[20][16] = 
		{
			{-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.2455231880,-.2455231878,-.4910463757e-1,-.4910463757e-1,-.2455231878,-.2455231878,-.4910463757e-1,-.9820927513e-2,-.4910463757e-1,-.4910463757e-1,-.9820927513e-2},
			{.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1,.8505172722e-1,.1901814434,-.1901814436,-.8505172722e-1,.8505172718e-1,.1901814434,-.1901814435,-.8505172718e-1,.1701034544e-1,.3803628867e-1,-.3803628871e-1,-.1701034544e-1},
			{-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1,-.1098013076,.1098013076,.1098013076,-.1098013076,-.1098013076,.1098013076,.1098013076,-.1098013076,-.2196026153e-1,.2196026152e-1,.2196026152e-1,-.2196026153e-1},
			{.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1,.5567942541e-1,-.1245029801,.1245029801,-.5567942541e-1,.5567942540e-1,-.1245029801,.1245029801,-.5567942540e-1,.1113588508e-1,-.2490059601e-1,.2490059602e-1,-.1113588508e-1},
			{-.2777777778e-1,-.1388888889,-.1388888889,-.2777777778e-1,-.8130697173e-1,-.4065348587,-.4065348587,-.8130697173e-1,.1186252729e-1,.5931263656e-1,.5931263656e-1,.1186252729e-1,.1388888889e-1,.6944444444e-1,.6944444444e-1,.1388888889e-1},
			{.4811252245e-1,.1075828706,-.1075828707,-.4811252245e-1,.1408278060,.3149005471,-.3149005473,-.1408278060,-.2054650002e-1,-.4594337068e-1,.4594337076e-1,.2054650002e-1,-.2405626122e-1,-.5379143531e-1,.5379143537e-1,.2405626122e-1},
			{-.6211299937e-1,.6211299937e-1,.6211299937e-1,-.6211299937e-1,-.1818079158,.1818079158,.1818079158,-.1818079158,.2652541750e-1,-.2652541750e-1,-.2652541750e-1,.2652541750e-1,.3105649968e-1,-.3105649968e-1,-.3105649968e-1,.3105649968e-1},
			{-.5103103630e-1,-.2551551817,-.2551551816,-.5103103630e-1,-.3803628868e-1,-.1901814436,-.1901814434,-.3803628868e-1,.3803628871e-1,.1901814436,.1901814435,.3803628871e-1,-.1701034544e-1,-.8505172722e-1,-.8505172718e-1,-.1701034544e-1},
			{.8838834762e-1,.1976423534,-.1976423537,-.8838834762e-1,.6588078456e-1,.1473139125,-.1473139127,-.6588078456e-1,-.6588078455e-1,-.1473139125,.1473139128,.6588078455e-1,.2946278254e-1,.6588078447e-1,-.6588078455e-1,-.2946278254e-1},
			{-.5892556509e-1,-.2946278254,-.2946278254,-.5892556509e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.3324493542e-1,.1662246771,.1662246771,.3324493542e-1,.8703637069e-1,.4351818534,.4351818534,.8703637069e-1,.2405626122e-1,.1202813062,.1202813061,.2405626122e-1},
			{0.,0.,0.,0.,-.5758191722e-1,-.1287570810,.1287570811,.5758191722e-1,-.1507514162,-.3370904137,.3370904144,.1507514162,-.4166666667e-1,-.9316949897e-1,.9316949905e-1,.4166666667e-1},
			{0.,0.,0.,0.,.7433793548e-1,-.7433793548e-1,-.7433793548e-1,.7433793548e-1,.1946192414,-.1946192414,-.1946192414,.1946192414,.5379143537e-1,-.5379143537e-1,-.5379143537e-1,.5379143537e-1},
			{0.,0.,0.,0.,.1065973485,.5329867431,.5329867429,.1065973485,.4071656407e-1,.2035828203,.2035828202,.4071656407e-1,-.2946278254e-1,-.1473139128,-.1473139128,-.2946278254e-1},
			{0.,0.,0.,0.,-.1846320236,-.4128497551,.4128497560,.1846320236,-.7052315768e-1,-.1576945743,.1576945745,.7052315768e-1,.5103103630e-1,.1141088660,-.1141088661,-.5103103630e-1},
			{.2041241452e-1,.1020620727,.1020620726,.2041241452e-1,.1100932682,.5504663412,.5504663412,.1100932682,-.4205188656e-1,-.2102594328,-.2102594328,-.4205188656e-1,.1360827635e-1,.6804138178e-1,.6804138176e-1,.1360827635e-1},
			{0.,0.,0.,0.,-.1452856955e-1,-.7264284777e-1,-.7264284777e-1,-.1452856955e-1,-.9958029666e-1,-.4979014835,-.4979014833,-.9958029666e-1,-.3803628871e-1,-.1901814436,-.1901814435,-.3803628871e-1},
			{0.,0.,0.,0.,.2516422062e-1,.5626890782e-1,-.5626890792e-1,-.2516422062e-1,.1724781332,.3856728304,-.3856728304,-.1724781332,.6588078457e-1,.1473139125,-.1473139128,-.6588078457e-1},
			{.8784104612e-2,.4392052305e-1,.4392052305e-1,.8784104612e-2,-.8784104609e-1,-.4392052307,-.4392052305,-.8784104609e-1,-.8784104609e-1,-.4392052305,-.4392052305,-.8784104609e-1,.3513641844e-1,.1756820922,.1756820922,.3513641844e-1},
			{.1484784677e-2,.7423923388e-2,.7423923386e-2,.1484784677e-2,.2166271727e-2,.1083135864e-1,.1083135863e-1,.2166271727e-2,.1017686557,.5088432789,.5088432787,.1017686557,.5048267902e-1,.2524133952,.2524133952,.5048267902e-1}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 16; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[20][20] = 
		{
			{.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,-8.500000000,-.4082482906,0.,0.,1.750000000,0.,-7.678758582,.3227486122,0.,-2.459674775,-.2645751311},
			{0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,2.309401077,0.,0.,-.4082482906,0.,0.,1.750000000,0.,0.,.3227486122,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,-12.37436867,-.7216878367,0.,0.,2.828427124,0.,-13.26806944,.9128709293,0.,-5.850213671,-.9354143468},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,2.857738034,0.,0.,-.7216878367,0.,0.,2.828427124,0.,0.,.9128709295,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,-7.216878367,0.,0.,0.,1.587713241,0.,-9.500000000,.9316949905,0.,-6.067673910,-1.636634177},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,1.587713241,0.,0.,.9316949905,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,-2.901185103,0.,0.,-2.627379873,-1.218935425},
			{0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,2.245365598,-.2500000000,0.,0.,0.,0.,.4949747467,.3162277660,0.,-1.004158022,-.3240370350},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,-.7071067810,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,3.500000000,0.,0.,0.,-.2500000000,0.,1.558845727,.5809475019,0.,-1.565247584,-1.020504077},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,-.2500000000,0.,0.,.5809475019,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,1.275000000,0.,0.,-.5486726408,-.9983468483},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,-.6454972245,0.,0.,0.,-.3227486122,0.,.2236067977,.8333333333e-1,0.,.2886751347,-.1463850110},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.8333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.6250000000,-.2746751327},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.1479019946,-.2500000000e-1}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,-.1154700539,2.904737510,0.,1.341640786,-11.64130577},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,-.2500000000,0.,0.,2.904737510,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,.1632993162,-2.464751509,0.,-1.581138830,10.85080642},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,.3535533905,0.,0.,-2.464751509,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,-.2000000000,1.155301788,0.,1.290994449,-6.197388082},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,-.4330127020,0.,0.,1.155301788,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.3031088914,0.,0.,-.7267220925,1.974864371},
			{0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,.2828427124,3.320391543,0.,-1.278019301,-15.76980237},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,.6123724358,0.,0.,3.320391543,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,-.3464101616,-1.484643616,0.,.4472135954,8.919961562},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-1.484643616,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.3250000000,0.,0.,.3227486122e-1,-2.776822654},
			{0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,-.4472135954,1.083333333,0.,-4.041451885,-8.002380602},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,.9682458365,0.,0.,1.083333333,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,1.375000000,2.387560769},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,-1.031079532,0.,0.,-1.922725929,-1.575000000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[20][20] = 
		{
			{1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,.1250000000,.4082482906,0.,0.,.1250000000,0.,-.2453738645,-.1613743061,0.,.5590169942e-1,.6614378278e-1},
			{0.,1.,0.,0.,0.,-.7071067810,0.,0.,.1443375673,0.,0.,.4082482906,0.,0.,.1250000000,0.,0.,-.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,-.2651650429,.7216878367,0.,0.,-.7071067810,0.,.3163924252,.3651483717,0.,.3952847075e-1,-.1403121520},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,.4082482906,0.,0.,.7216878367,0.,0.,-.7071067810,0.,0.,.3651483718,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,-.2165063510,0.,0.,0.,-.2165063510,0.,.4250000000,.5776508940,0.,-.6132223630,.3218713882},
			{0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.5776508940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.1816805231,.4937160930},
			{0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,-.4592793268,-.7500000000,0.,0.,0.,0.,.1237436867,-.3162277660,0.,-.1141088661,.1890216038},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,.7071067810,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.3162277660,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,-.3750000000,0.,0.,0.,-.1250000000,0.,.3897114318,-.7423218081,0.,.2795084971,-.3496171374},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.7423218081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,.8068715303e-2,-.6942056633},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,-.4841229182,0.,0.,0.,.4841229182,0.,-.3913118960,.5416666667,0.,.3608439183,.1341862601},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,.4841229182,0.,0.,.5416666667,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,.3437500000,.5968901920},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,-.2577698829,0.,0.,-.4806814823,-.3937500000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[20][20] = 
		{
			{1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,-.1250000000,-.4082482906,0.,0.,-.1250000000,0.,.2453738645,.1613743061,0.,-.5590169942e-1,-.6614378278e-1},
			{0.,1.,0.,0.,0.,.7071067810,0.,0.,-.1443375673,0.,0.,-.4082482906,0.,0.,-.1250000000,0.,0.,.1613743061,0.,0.},
			{0.,0.,1.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,.2651650429,-.7216878367,0.,0.,-.3535533905,0.,.5103103630e-1,.4564354646,0.,.1185854122,-.2338535867},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,-.4082482906,0.,0.,-.7216878367,0.,0.,-.3535533905,0.,0.,.4564354646,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,.2165063510,0.,0.,0.,.7938566203,0.,.3750000000,.4658474951,0.,.9682458365e-1,-.4091585443},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,.7938566203,0.,0.,.4658474951,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.6568449682,-.3047338563},
			{0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,.2551551816,-.2500000000,0.,0.,-.6123724358,0.,-.4065863991,.1581138830,0.,.2966830521,-.8100925875e-1},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,.7071067810,0.,0.,-.2500000000,0.,0.,-.6123724358,0.,0.,.1581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,-.6250000000,0.,0.,0.,-.1250000000,0.,-.4330127020e-1,.2904737510,0.,.6149186937,-.2551260193},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.2904737510,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,.3187500000,0.,0.,-.1371681602,-.2495867120},
			{0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,.4841229182,0.,0.,0.,-.1613743061,0.,-.5031152948,.4166666667e-1,0.,.2165063510,-.3659625275e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,-.1613743061,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1562500000,-.6866878320e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,.3697549864e-1,-.6250000000e-2}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[20][20] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302,.2357022604e-1,.7071067810e-1,.3535533905,.1178511302,.5892556509e-1,.1178511302,.2946278254e-1,.5892556509e-1,.2357022604,.7856742012e-1,.7856742012e-1,.1767766952},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.2041241452,.6804138176e-1,.6804138176e-1,.1837117307},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.1756820922e-1,.7905694150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,.1666666667,.1666666667e-1,0.,.1333333333,.1666666667e-1,.1000000000,0.,-.4166666667e-1,-.3333333333e-1,.8333333333e-1,.8333333333e-2,.6666666667e-1,0.,-.2777777778e-1,.5555555556e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2405626122e-1,.4811252245e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,.1242259987e-1,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,.4082482906e-1,-.5832118436e-2,.5248906592e-1,0.,0.,.6804138176e-2,0.,-.1020620726e-1,.2041241452e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,0.,0.,.3367175148e-2,0.,-.5050762721e-2,.1010152544e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,0.,.2886751347e-1,.1924500898e-1,0.,.9622504490e-2,0.,0.,.7216878367e-1,.5773502693e-1,0.,.1443375673e-1,0.,0.,.4811252245e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.4166666667e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,0.,.1346870059e-1,0.,0.,0.,-.1178511302e-1,0.,.1178511302e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,0.,0.,-.4860098696e-2,0.,.4860098696e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,0.,0.,.4347004424e-2,0.,0.,0.,0.,0.,.1521451549e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,0.,0.,.3764616260e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[20][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-8.485281372,14.69693846,9.797958972,-32.86335345,56.92099788,37.94733192,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,-37.94733192,-224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955},
			{0.,0.,0.,0.,6.,-10.39230485,13.41640786,-19.59591794,33.94112549,42.42640686,3.464101616,-6.,7.745966692,-25.45584412,44.09081537,63.68673332,-10.95445115,18.97366596,56.92099788,22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,-117.5755077,32.86335345,-56.92099788,-303.5786554,-179.5995546},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,224.4994432},
			{0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,-127.2792206,0.,0.,0.,21.21320343,-36.74234614,-161.6663230,5.477225575,-9.486832980,-94.86832980,-22.44994432},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,0.,0.,265.6313234,89.79977729},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,0.,0.,0.,0.,0.,102.8785692,0.,0.,44.27188724,7.483314774},
			{0.,4.898979486,-18.97366596,44.89988864,0.,-6.928203232,26.83281572,0.,8.485281372,0.,0.,-12.,46.47580015,0.,14.69693846,0.,0.,18.97366596,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,0.,-29.39387692,0.,0.,-113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,20.78460970,-80.49844717,0.,-67.88225098,0.,0.,12.,-46.47580015,0.,-88.18163075,0.,0.,-37.94733192,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,0.,0.,113.8419958,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,0.,0.,0.,0.,0.,73.48469229,0.,0.,18.97366596,0.,0.},
			{0.,0.,18.97366596,-112.2497216,0.,0.,-26.83281572,0.,0.,0.,0.,0.,-46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,80.49844717,0.,0.,0.,0.,0.,46.47580015,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[10][20] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-3.464101616,6.,-7.745966692,4.242640686,-7.348469229,-4.898979486,5.477225575,-9.486832980,-6.324555320,-7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,-1.505360805,2.607361398,-3.366089089,1.843682925,-3.193352500,-2.128901666,-1.187112542,2.056139236,1.370759491,3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,1.505360805,-2.607361398,3.366089089,-1.843682925,3.193352500,2.128901666,-1.187112542,2.056139236,1.370759491,-3.342659338},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-2.,3.464101616,-4.472135954,2.449489743,-4.242640686,-2.828427124,3.464101616,-6.,7.745966692,-4.242640686,7.348469229,4.898979486,5.477225575,-9.486832980,-6.324555320,7.483314774},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,-2.484731210,4.303680699,-5.556027890,-1.258649951,2.180045665,2.509295895,2.817981978,-4.880887957,3.185720379,-2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,0.,0.,0.,-1.360827635,2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-1.873942317,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,-.3036806991,.5259903999,-.6790506860,-1.132775699,1.962025064,1.229948799,2.484731210,-4.303680699,5.556027890,1.258649951,-2.180045665,-2.509295895,2.817981978,-4.880887957,3.185720379,2.761597556},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,-.9793704052,1.696319301,-2.189938801,-3.102332876,5.373398163,-4.427892772,.4377975572,-.7582876121,2.032696517,-.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,2.303680699,-3.990092015,5.151186639,.9961259674,-1.725340785,-1.348290938,.9793704052,-1.696319301,2.189938801,3.102332876,-5.373398163,4.427892772,.4377975572,-.7582876121,2.032696517,.1691075797},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.,-6.928203232,8.944271908,7.348469229,-12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[10][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-3.464101616,-6.,-7.745966692,4.242640686,7.348469229,-4.898979486,5.477225575,9.486832980,-6.324555320,-7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,-1.505360805,-2.607361398,-3.366089089,1.843682925,3.193352500,-2.128901666,-1.187112542,-2.056139236,1.370759491,3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,1.505360805,2.607361398,3.366089089,-1.843682925,-3.193352500,2.128901666,-1.187112542,-2.056139236,1.370759491,-3.342659338},
			{1.414213562,2.449489743,3.162277660,3.741657387,-2.,-3.464101616,-4.472135954,2.449489743,4.242640686,-2.828427124,3.464101616,6.,7.745966692,-4.242640686,-7.348469229,4.898979486,5.477225575,9.486832980,-6.324555320,7.483314774},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,-2.484731210,-4.303680699,-5.556027890,-1.258649951,-2.180045665,2.509295895,2.817981978,4.880887957,3.185720379,-2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,0.,0.,0.,-1.360827635,-2.357022604,.8380524812,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-1.873942317,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,-.3036806991,-.5259903999,-.6790506860,-1.132775699,-1.962025064,1.229948799,2.484731210,4.303680699,5.556027890,1.258649951,2.180045665,-2.509295895,2.817981978,4.880887957,3.185720379,2.761597556},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,-.9793704052,-1.696319301,-2.189938801,-3.102332876,-5.373398163,-4.427892772,.4377975572,.7582876121,2.032696517,-.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.303680699,3.990092015,5.151186639,.9961259674,1.725340785,-1.348290938,.9793704052,1.696319301,2.189938801,3.102332876,5.373398163,4.427892772,.4377975572,.7582876121,2.032696517,.1691075797},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.,6.928203232,8.944271908,7.348469229,12.72792206,11.31370850,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{.2322012545e-1,.1006582171,.1006582171,.2322012547e-1,.1006582171,.5511351922,.1006582171,.1006582170,.1006582172,.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{.3546932749e-1,.1537579664,.1537579663,.3546932754e-1,.1537579662,.8418729121,.1537579664,.1537579663,.1537579664,.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{-.1985351453e-2,-.1423522158,-.1423522158,-.1985351454e-2,-.9261271237e-1,0.,-.9261271240e-1,.2349649283,.2349649284,.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{.2594854185e-1,.9132364797e-1,.9132364797e-1,.2594854188e-1,-.6311588137e-1,-.2272843224,-.6311588142e-1,.1330298246e-1,.1330298247e-1,.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{-.3438729585e-2,-.1891270389,.1891270390,.3438729588e-2,-.2178441547,0.,.2178441546,-.2871711568e-1,.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{.5751898119e-1,.6618067742e-1,-.6618067742e-1,-.5751898117e-1,-.6756787835e-1,0.,.6756787845e-1,-.1337485559,.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{.6776309270e-1,-.7425668538e-1,-.7425668538e-1,.6776309264e-1,.9841195758e-1,-.2032892782,.9841195768e-1,.1297307049e-1,.1297307048e-1,-.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[20][10] = 
		{
			{-.1340614568e-1,-.5811504871e-1,-.5811504871e-1,-.1340614568e-1,-.5811504871e-1,-.3181980514,-.5811504874e-1,-.5811504871e-1,-.5811504876e-1,-.1340614568e-1},
			{-.2322012545e-1,-.1006582171,-.1006582171,-.2322012547e-1,-.1006582171,-.5511351922,-.1006582171,-.1006582170,-.1006582172,-.2322012546e-1},
			{-.2997705306e-1,-.1299491994,-.1299491994,-.2997705307e-1,-.1299491994,-.7115124735,-.1299491995,-.1299491994,-.1299491995,-.2997705306e-1},
			{-.3546932749e-1,-.1537579664,-.1537579663,-.3546932754e-1,-.1537579662,-.8418729121,-.1537579664,-.1537579663,-.1537579664,-.3546932753e-1},
			{.1146243195e-2,.8218709003e-1,.8218709011e-1,.1146243196e-2,.5346997435e-1,0.,.5346997442e-1,-.1356570646,-.1356570646,-.2292486392e-2},
			{.1985351453e-2,.1423522158,.1423522158,.1985351454e-2,.9261271237e-1,0.,.9261271240e-1,-.2349649283,-.2349649284,-.3970702907e-2},
			{.2563077703e-2,.1837759203,.1837759203,.2563077706e-2,.1195624975,0.,.1195624975,-.3033384178,-.3033384180,-.5126155411e-2},
			{-.1498139764e-1,-.5272573275e-1,-.5272573275e-1,-.1498139764e-1,.3643997113e-1,.1312226648,.3643997114e-1,-.7680480510e-2,-.7680480512e-2,-.5332738507e-1},
			{-.2594854185e-1,-.9132364797e-1,-.9132364797e-1,-.2594854188e-1,.6311588137e-1,.2272843224,.6311588142e-1,-.1330298246e-1,-.1330298247e-1,-.9236574035e-1},
			{.1100121325e-1,.1245322472e-1,.1245322472e-1,.1100121325e-1,-.3530856516e-1,-.4545686449e-1,-.3530856518e-1,.6021501464e-1,.6021501465e-1,-.5126491036e-1},
			{.1985351453e-2,.1091925468,-.1091925469,-.1985351454e-2,.1257723813,0.,-.1257723814,.1657983447e-1,-.1657983447e-1,0.},
			{.3438729585e-2,.1891270389,-.1891270390,-.3438729588e-2,.2178441547,0.,-.2178441546,.2871711568e-1,-.2871711570e-1,0.},
			{.4439380807e-2,.2441619572,-.2441619574,-.4439380807e-2,.2812355943,0.,-.2812355944,.3707363693e-1,-.3707363695e-1,0.},
			{-.3320859927e-1,-.3820943187e-1,.3820943193e-1,.3320859926e-1,.3901033278e-1,0.,-.3901033278e-1,.7721976470e-1,-.7721976474e-1,0.},
			{-.5751898119e-1,-.6618067742e-1,.6618067742e-1,.5751898117e-1,.6756787835e-1,0.,-.6756787845e-1,.1337485559,-.1337485559,0.},
			{.2156961794e-1,.5515056451e-2,-.5515056451e-2,-.2156961794e-1,-.5515056450e-1,0.,.5515056457e-1,.5515056449e-1,-.5515056457e-1,0.},
			{-.3912303979e-1,.4287211731e-1,.4287211729e-1,-.3912303978e-1,-.5681817023e-1,.1173691195,-.5681817020e-1,-.7490005736e-2,-.7490005739e-2,.3749077515e-2},
			{-.6776309270e-1,.7425668538e-1,.7425668538e-1,-.6776309264e-1,-.9841195758e-1,.2032892782,-.9841195768e-1,-.1297307049e-1,-.1297307048e-1,.6493592735e-2},
			{.3109305337e-1,-.2784625702e-1,-.2784625702e-1,.3109305338e-1,-.4920597878e-1,.1016446391,-.4920597882e-1,-.6486535244e-2,-.6486535242e-2,.3246796368e-2},
			{.3294813564e-1,-.7581949096e-1,.7581949096e-1,-.3294813565e-1,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
