
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (5)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (56)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (56)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (144)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (52)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*5+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[56],
		T io_hv[56]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,0.,7.668115805,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.098386677,0.,0.,0.,0.,0.,7.668115806,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,0.,0.,0.,0.,0.,7.314285714,0.,0.,0.,0.,11.33763824,0.,0.,0.,12.92995257,0.,0.,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780458,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.533303756,0.,0.,0.,0.,6.337932448,0.,0.,0.,13.55261854,0.,0.,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357805,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,0.,0.,0.,0.,0.,5.935526613,0.,0.,0.,0.,-.3207134902,0.,0.,0.,-5.966396197,0.,0.,2.530240385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.792842914,0.,0.,0.,9.200874127,0.,0.,19.80226324,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.857142857,0.,0.,0.,0.,9.972922524,0.,0.,0.,-7.273098319,0.,0.,1.490788040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.320493800,0.,0.,21.25401924,0.,0.,0.,0.,0.,0.},
			{5.416025602,0.,0.,0.,0.,0.,-4.376809636,0.,0.,0.,0.,2.010178182,0.,0.,0.,-.5158119621,0.,0.,.5766953058e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,15.73213272,0.,0.,0.,-4.036867140,0.,0.,.4513354669,0.,0.,0.,0.,0.,0.,0.,0.,21.79908256,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,0.,7.668115805,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.098386677,0.,0.,0.,0.,0.,7.668115806,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,0.,0.,0.,0.,0.,7.314285714,0.,0.,0.,0.,11.33763824,0.,0.,0.,12.92995257,0.,0.,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780458,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.533303756,0.,0.,0.,0.,6.337932448,0.,0.,0.,13.55261854,0.,0.,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357805,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,0.,0.,0.,0.,0.,5.935526613,0.,0.,0.,0.,-.3207134902,0.,0.,0.,-5.966396197,0.,0.,2.530240385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.792842914,0.,0.,0.,9.200874127,0.,0.,19.80226324,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.857142857,0.,0.,0.,0.,9.972922524,0.,0.,0.,-7.273098319,0.,0.,1.490788040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.320493800,0.,0.,21.25401924,0.,0.,0.,0.,0.,0.},
			{5.416025602,0.,0.,0.,0.,0.,-4.376809636,0.,0.,0.,0.,2.010178182,0.,0.,0.,-.5158119621,0.,0.,.5766953058e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,15.73213272,0.,0.,0.,-4.036867140,0.,0.,.4513354669,0.,0.,0.,0.,0.,0.,0.,0.,21.79908256,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,0.,0.,0.,0.,0.,-4.618802155,0.,0.,0.,0.,12.72792206,0.,0.,0.,-3.265986323,0.,0.,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,0.,3.834057902,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.549193338,0.,0.,0.,0.,0.,3.834057903,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,0.,0.,0.,0.,0.,3.657142857,0.,0.,0.,0.,5.668819119,0.,0.,0.,6.464976283,0.,0.,4.969293464,0.,0.,-6.730254569,0.,0.,0.,0.,9.455027815,0.,0.,0.,-6.998542123,0.,0.,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.236067977,0.,0.,0.,0.,0.,2.529822127,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890229,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,0.,0.,0.,0.,0.,3.836097790,0.,0.,0.,0.,3.659207005,0.,0.,0.,.5216405308,0.,0.,-3.207665141,0.,0.,.7666518779,0.,0.,0.,0.,3.168966225,0.,0.,0.,6.776309270,0.,0.,8.333758491,0.,0.,8.636804253,0.,0.,0.,-8.748177653,0.,0.,15.06237033,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178903,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178902,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.160246899,0.,0.,0.,0.,0.,2.967763307,0.,0.,0.,0.,-.1603567451,0.,0.,0.,-2.983198099,0.,0.,1.265120192,0.,0.,1.814229470,0.,0.,0.,0.,5.416047582,0.,0.,0.,5.345224839,0.,0.,-3.286878676,0.,0.,.8964214566,0.,0.,0.,4.600437064,0.,0.,9.901131622,0.,0.,-9.525793445,0.,0.,13.38877363,0.,0.,0.,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225576,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,.9897433188,0.,0.,0.,0.,-2.727411870,0.,0.,0.,1.516350793,0.,0.,-.2869022921,0.,0.,3.428571429,0.,0.,0.,0.,4.986461263,0.,0.,0.,-3.636549159,0.,0.,.7453940198,0.,0.,3.049339172,0.,0.,0.,8.346248498,0.,0.,-2.245365598,0.,0.,2.160246900,0.,0.,10.62700962,0.,0.,12.04989627,0.,0.,0.},
			{2.708012801,0.,0.,0.,0.,0.,-2.188404818,0.,0.,0.,0.,1.005089091,0.,0.,0.,-.2579059811,0.,0.,.2883476528e-1,0.,0.,5.685642496,0.,0.,0.,0.,-2.611298058,0.,0.,0.,.6700593942,0.,0.,-.7491491772e-1,0.,0.,7.866066360,0.,0.,0.,-2.018433570,0.,0.,.2256677334,0.,0.,9.552971226,0.,0.,-1.068054652,0.,0.,10.89954128,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,0.,0.,0.,0.,0.,-4.618802155,0.,0.,0.,0.,12.72792206,0.,0.,0.,-3.265986323,0.,0.,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,0.,3.834057902,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.549193338,0.,0.,0.,0.,0.,3.834057903,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,0.,0.,0.,0.,0.,3.657142857,0.,0.,0.,0.,5.668819119,0.,0.,0.,6.464976283,0.,0.,4.969293464,0.,0.,-6.730254569,0.,0.,0.,0.,9.455027815,0.,0.,0.,-6.998542123,0.,0.,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.236067977,0.,0.,0.,0.,0.,2.529822127,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890229,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,0.,0.,0.,0.,0.,3.836097790,0.,0.,0.,0.,3.659207005,0.,0.,0.,.5216405308,0.,0.,-3.207665141,0.,0.,.7666518779,0.,0.,0.,0.,3.168966225,0.,0.,0.,6.776309270,0.,0.,8.333758491,0.,0.,8.636804253,0.,0.,0.,-8.748177653,0.,0.,15.06237033,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178903,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178902,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.160246899,0.,0.,0.,0.,0.,2.967763307,0.,0.,0.,0.,-.1603567451,0.,0.,0.,-2.983198099,0.,0.,1.265120192,0.,0.,1.814229470,0.,0.,0.,0.,5.416047582,0.,0.,0.,5.345224839,0.,0.,-3.286878676,0.,0.,.8964214566,0.,0.,0.,4.600437064,0.,0.,9.901131622,0.,0.,-9.525793445,0.,0.,13.38877363,0.,0.,0.,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225576,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,.9897433188,0.,0.,0.,0.,-2.727411870,0.,0.,0.,1.516350793,0.,0.,-.2869022921,0.,0.,3.428571429,0.,0.,0.,0.,4.986461263,0.,0.,0.,-3.636549159,0.,0.,.7453940198,0.,0.,3.049339172,0.,0.,0.,8.346248498,0.,0.,-2.245365598,0.,0.,2.160246900,0.,0.,10.62700962,0.,0.,12.04989627,0.,0.,0.},
			{2.708012801,0.,0.,0.,0.,0.,-2.188404818,0.,0.,0.,0.,1.005089091,0.,0.,0.,-.2579059811,0.,0.,.2883476528e-1,0.,0.,5.685642496,0.,0.,0.,0.,-2.611298058,0.,0.,0.,.6700593942,0.,0.,-.7491491772e-1,0.,0.,7.866066360,0.,0.,0.,-2.018433570,0.,0.,.2256677334,0.,0.,9.552971226,0.,0.,-1.068054652,0.,0.,10.89954128,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,3.43216369764802,-5.83506573884863,7.25392389703222,-8.09901689119009,8.47552946493044,-4.00974905232829,6.81702604882233,-8.4746582765388,9.46196865399936,4.32546942586304,-7.35378632026677,9.14193750783762,-4.4144312457315,7.50503145913621,4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,7.27823584467215,-12.3738225688191,15.3826488429769,-7.46210387346585,12.6864189798666,7.31197439196321,9.14177937313724,-15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,3.29802783242322,-5.60701962552938,6.97042595115334,-7.78249101017965,8.14428871473084,-3.07100849759875,5.22106113152625,-6.4906175528327,7.24678421547914,2.20598407348367,-3.75042195543293,4.66237685613718,-.923194488569861,1.5695348492102,-.507379758856801,4.96464287844228,-8.44045334766827,10.4928391542811,-11.7152706263276,-3.80884940873901,6.47547397319737,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,6.45780747384475,-10.9790017210618,13.6486625036235,-4.10479708970297,6.97861844197595,1.32544341327556,7.79428043908965,-13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,3.06556135691961,-5.21180037431139,6.47910494479513,-7.23393037100344,7.57022621762746,-1.5947619510804,2.71127534923244,-3.37055072311448,3.76322492874006,-.444911232504622,.756399320675517,-.940325842964601,2.09440936915946,-3.56073236366527,-2.63186985254377,4.2894282422279,-7.29251224169724,9.06576398575723,-10.1219390639593,-.964148801340397,1.63916180436132,-2.03774139204767,-2.01722273728899,3.42950637731078,3.46190525493913,5.18623489409548,-8.8171847889118,10.9611768422514,.205483038001612,-.349344361529711,-3.63947886955732,5.81833621896754,-9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,2.74931534527504,-4.67414645388234,5.81071476784009,-6.48767173775818,6.78927500842464,.106757773657142,-.181500267083405,.225633983149896,-.251920680005671,-2.26282036321899,3.84705006415361,-4.78250111498198,2.15044627554921,-3.65600143048661,-.119501749587716,3.45007465931297,-5.86551639682902,7.29177895707792,-8.14128212331119,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429708,1.33170919606025,3.7410698100085,-6.36024100270088,7.90680110023492,3.5848146864382,-6.09458965574695,-3.59724489879787,3.76406389114812,-6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,2.36915365304765,-4.02782866109789,5.00723794478697,-5.59058866190052,5.85048772793476,1.68415540504794,-2.86325431258853,3.55948498424847,-3.97417031416821,-2.32831086437971,3.95839130925952,-4.9209161654707,-.381749900992996,.649017927110068,2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,2.3938769796025,-4.06986110773884,5.05949102727518,4.93737708601548,-8.39409842544749,.508166439915615,2.07554270972349,-3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,1.94896143440088,-3.31345445440235,4.11915606854152,-4.59904391748787,4.81284738096772,2.83314651369184,-4.81666890668704,5.98789306701478,-6.6854915743852,-.637278791777431,1.08344588754871,-1.34689725343212,-2.63766838605638,4.48433402043889,.808065843843811,1.73374759666501,-2.94756663562365,3.66429874441148,-4.09119503456583,4.12740556141282,-7.01705539432372,8.72332687917667,1.98940127224618,-3.38220674597771,-2.96832239843811,1.33269841309056,-2.26573774926528,2.81667592802333,4.37531622836173,-7.43853151653138,4.52786743647191,.950542947793588,-1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,1.51514018159912,-2.57590935107797,3.20226904625418,-3.57533818439405,3.74155092353066,3.36446686788906,-5.7199734894835,7.11084574520705,-7.93926992086969,1.91964817870762,-3.26361861024795,4.05720211125474,-1.95173675811295,3.31817281869637,-2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898136,3.55510980002441,-6.0440879938959,7.51377213010338,4.5904502888745,-7.8042837062112,1.24891795372526,.626153181688298,-1.06453109468611,1.32338312764,2.84473124098147,-4.83636425064192,5.48872263996487,.347192252628366,-.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,1.09494796295235,-1.86153514438243,2.31418717000874,-2.5837934399814,2.70391057656363,3.24473228828999,-5.51641118734948,6.85778510905627,-7.65672734172539,4.05017972204568,-6.88576274670089,8.56010903534823,1.63294619261043,-2.776192869346,-2.28560008620039,.547225517982272,-.930345156393644,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070527,4.87179403320197,-8.28259983241364,5.49730994822616,.236321146004053,-.40177262626832,.499467904136592,1.3620925765208,-2.31571114636374,3.77635721863399,.0946962283839075,-.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,.714786270724957,-1.21521735159799,1.51071034695561,-1.68671036412375,1.76512329607374,2.59853682213183,-4.41780594598826,5.49204234459095,-6.13187350041214,4.7549644652192,-8.08397637227704,10.0496810203089,5.36731919542456,-9.12504854432544,3.25729414076559,.233201739437629,-.396469282990197,.492874996717866,-.550295671800538,1.22672710731592,-2.08557456680069,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984458,.0657431866024185,-.111770838907699,.138949104550329,.451524561772771,-.76764272750461,1.59193561065049,.0171974308138372,-.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,.398540259080393,-.677563431168936,.842320170000572,-.940451730878483,.984172086870924,1.67165845871656,-2.84200809304711,3.53307252095548,-3.9446807593569,3.84755914106515,-6.54128488547504,8.13186772631248,6.3113399865174,-10.7299904588883,7.98777931164405,.07249762486996,-.123254146474207,.153224700235958,-.171075607231601,.434860760346183,-.739312383715058,.919084035475439,1.40707187748205,-2.3921810358159,3.23180916099,.0113956405123911,-.0193738753136992,.0240848387004708,.0887334889355571,-.150856948291044,.36808995603673,.00166206328996398,-.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,.166073783576779,-.282344179950942,.350999163642368,-.391891091702269,.410109589767539,.764836159193308,-1.30030781284792,1.61649145672597,-1.80481513164277,2.02218309922343,-3.43793955015111,4.27391103771759,4.02885967082133,-6.84951625494143,6.68010475273478,.0125887190510937,-.0214022435166571,.0266064261611111,-.0297061146457814,.0823390061849577,-.139985605707145,.174024591276674,.299004079114616,-.508340689012672,.796298871799331,.000824565926335038,-.00140185515920491,.00174273111828065,.00697737063218123,-.0118623177482726,.0321260785532874,5.01144631634162e-005,-8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,.0319379183519753,-.054298066631689,.0675012177634823,-.0753652106918314,.07886883956794,.154660030302817,-.262939511058368,.326876043550518,-.364957592022403,.43925822557731,-.746788570738901,.928378137184539,.961797763577864,-1.63516477461472,1.79684767953686,.000465578153648566,-.000791535419924754,.000984005657525639,-.00109864378994572,.00319092434145458,-.00542493160168594,.0067440612905754,.0123137463285649,-.0209347588552144,.0353622090543065,5.86464816381893e-006,-9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,-8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,-1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,3.43216369764802,-5.37468949190064,5.572814208608,-4.46269573688388,2.39877236199698,-4.00974905232829,6.27917489029831,-6.51064123734671,5.21370528571982,4.32546942586304,-6.77358573667404,7.02327732701019,-4.4144312457315,6.91289790468793,4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,7.27823584467215,-11.3975501068473,11.8176927763525,-7.46210387346585,11.6854832203471,7.31197439196321,9.14177937313724,-14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,3.29802783242322,-5.16463580891203,5.35501741292457,-4.28828460541245,2.30502351007762,-3.07100849759875,4.80912874954551,-4.98640545923911,3.99310107128126,2.20598407348367,-3.45452037324197,3.58186277642043,-.923194488569861,1.44570135775136,-.507379758856801,4.96464287844228,-7.77451667983119,8.06110512843119,-6.45531290809181,-3.80884940873901,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,6.45780747384475,-10.1127781284249,10.4855608369157,-4.10479708970297,6.42801793274623,1.32544341327556,7.79428043908965,-12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,3.06556135691961,-4.80059864950596,4.9775609184691,-3.98601838486204,2.14255044478909,-1.5947619510804,2.49736057507455,-2.5894196337366,2.07360079279396,-.444911232504622,.696720767591157,-.722403666316151,2.09440936915946,-3.27979695088037,-2.63186985254377,4.2894282422279,-6.71714607327472,6.9647571533504,-5.57736018045176,-.964148801340397,1.50983487071459,-1.56549122161564,-2.01722273728899,3.15892446006539,3.46190525493913,5.18623489409548,-8.12152468439221,8.42090472917074,.205483038001612,-.321781716436834,-3.63947886955732,5.81833621896754,-9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,2.74931534527504,-4.30536466145163,4.46407134676943,-3.5748172149004,1.9215230524055,.106757773657142,-.16718022074295,.173342908623153,-.13881256945113,-2.26282036321899,3.54352469706742,-3.67414802589247,2.15044627554921,-3.3675494587808,-.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155853,-4.48598459696891,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761006,1.33170919606025,3.7410698100085,-5.8584293670558,6.07438596565011,3.5848146864382,-5.61373743608464,-3.59724489879787,3.76406389114812,-5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,2.36915365304765,-3.71004018615404,3.84680169804461,-3.08050921776368,1.65582437345554,1.68415540504794,-2.63734867013603,2.73456804580791,-2.18983527963514,-2.32831086437971,3.64608131714141,-3.78048514370595,-.381749900992996,.597811573855878,2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,2.3938769796025,-3.74875634795766,3.88694503628021,4.93737708601548,-7.73181907803741,.508166439915615,2.07554270972349,-3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,1.94896143440088,-3.05202882623936,3.1645343667906,-2.53415123836086,1.36214797290775,2.83314651369184,-4.43664234772504,4.60018826185061,-3.68381930162858,-.637278791777431,.997963945475626,-1.03475143356998,-2.63766838605638,4.13052808940267,.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633232,4.12740556141282,-6.46342227832349,6.70168045166518,1.98940127224618,-3.11535668501805,-2.96832239843811,1.33269841309056,-2.0869750946082,2.16390630168515,4.37531622836173,-6.85164468857581,4.52786743647191,.950542947793588,-1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,1.51514018159912,-2.37267471198348,2.46013753301879,-1.97007200846434,1.0589461087365,3.36446686788906,-5.26867781509235,5.46289467143737,-4.37467244553793,1.91964817870762,-3.0061249410246,3.11693834864261,-1.95173675811295,3.05637492003247,-2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002441,-5.56722026499826,5.77244215424071,4.5904502888745,-7.18853968487789,1.24891795372526,.626153181688298,-.980541495928357,1.01668674800429,2.84473124098147,-4.45478376427994,5.48872263996487,.347192252628366,-.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,1.09494796295235,-1.7146633520688,1.77787020176478,-1.42371402906152,.765269708188713,3.24473228828999,-5.08117621439183,5.26848128513793,-4.21898669004741,4.05017972204568,-6.34248838569497,6.576288631721,1.63294619261043,-2.55715621899877,-2.28560008620039,.547225517982272,-.856942588087487,.888531670338269,-.711533948435091,2.39319484717323,-3.74768814487269,3.88583745543936,4.87179403320197,-7.62911752450547,5.49730994822616,.236321146004053,-.370073484889949,.383715334269613,1.3620925765208,-2.13300568008583,3.77635721863399,.0946962283839075,-.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,.714786270724957,-1.11933887677121,1.16060055303996,-.929406031924799,.499571029238757,2.59853682213183,-4.06924896099705,4.21925182103647,-3.37876634868698,4.7549644652192,-7.44616510000515,7.72064967553202,5.36731919542456,-8.40509854032022,3.25729414076559,.233201739437629,-.365188565907959,.378650344801755,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064311,5.89679444984458,.0657431866024185,-.102952319693431,.106747403902078,.451524561772771,-.707077089737652,1.59193561065049,.0171974308138372,-.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,.398540259080393,-.624104888716883,.647110981340288,-.518204861963161,.278543636855169,1.67165845871656,-2.61777873930361,2.71427671758158,-2.17358603453451,3.84755914106515,-6.02518921139496,6.24729300320555,6.3113399865174,-9.88341340931671,7.98777931164405,.07249762486996,-.113529614838378,.117714605047621,-.0942655625163295,.434860760346183,-.68098195938567,.706084685269141,1.40707187748205,-2.20344223207683,3.23180916099,.0113956405123911,-.0178453112143479,.0185031347484812,.0887334889355571,-.138954604977606,.36808995603673,.00166206328996398,-.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,.166073783576779,-.260067729310804,.269654486884814,-.215938641412748,.11607057156663,.764836159193308,-1.1977158528687,1.24186670359493,-.994483762912882,2.02218309922343,-3.16669227065688,3.28342459824587,4.02885967082133,-6.30910167728468,6.68010475273478,.0125887190510937,-.019713644795161,.02044033985683,-.0163685732441573,.0823390061849577,-.12894099187304,.133694084564441,.299004079114616,-.468233518150739,.796298871799331,.000824565926335038,-.00129125129528224,.00133885010038651,.00697737063218123,-.0109264021014796,.0321260785532874,5.01144631634162e-005,-7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,.0319379183519753,-.0500140463222003,.0518576912013895,-.0415275099413171,.0223217196472694,.154660030302817,-.24219405407586,.251121942525088,-.201097826062499,.43925822557731,-.687868288629403,.713224862474738,.961797763577864,-1.50615319946369,1.79684767953686,.000465578153648566,-.000729084691473618,.000755960606624651,-.000605371370823784,.00319092434145458,-.00499691420436583,.0051811131626592,.0123137463285649,-.0192830438371028,.0353622090543065,5.86464816381893e-006,-9.18390427744748e-006,9.52244633459444e-006,5.19109369737154e-005,-8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,-1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,3.43216369764802,-4.57682649415414,2.98636125886026,.126713592860271,-3.19855198842569,-4.00974905232829,5.34704266035319,-3.48892427303922,-.148037726033603,4.32546942586304,-5.76805910494959,3.7636358461794,-4.4144312457315,5.8866906319401,4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,7.27823584467215,-9.70560427278211,6.33286855712841,-7.46210387346585,9.9507942329456,7.31197439196321,9.14177937313724,-12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,3.29802783242322,-4.39795490297745,2.86964825021048,.121761370614664,-3.07354613898791,-3.07100849759875,4.09522222779324,-2.67211637191124,-.11337994190694,2.20598407348367,-2.94170303097044,1.91944964063812,-.923194488569861,1.23108958885198,-.507379758856801,4.96464287844228,-6.62040364808343,4.31978730278896,.183291880031408,-3.80884940873901,5.07914086475798,-3.31411940634736,2.02329330850018,-2.6980829707225,.0162698697481229,6.45780747384475,-8.61155438605811,5.61900531555837,-4.10479708970297,5.47379022059323,1.32544341327556,7.79428043908965,-10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,3.06556135691961,-4.08795840577754,2.66737675689462,.113178836410129,-2.85690259486619,-1.5947619510804,2.12663188500128,-1.38761892782618,-.0588777326765052,-.444911232504622,.593293821521023,-.387121881618236,2.09440936915946,-2.79291698671264,-2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.15836332764069,-.964148801340397,1.28570259705192,-.838915879897342,-2.01722273728899,2.68998779997356,3.46190525493913,5.18623489409548,-6.91589893787678,4.51259681489189,.205483038001612,-.274013798814207,-3.63947886955732,5.81833621896754,-7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,2.74931534527504,-3.6662410134057,2.39220781955879,.101503207887319,-2.56218200503226,.106757773657142,-.142362617365695,.0928910470423959,.0039414381903617,-2.26282036321899,3.01749482284837,-1.96890348630228,2.15044627554921,-2.86764279249033,-.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,3.7410698100085,-4.98875605403389,3.25514367429686,3.5848146864382,-4.78038820054912,-3.59724489879787,3.76406389114812,-5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,2.36915365304765,-3.15929138677771,2.06142518510909,.0874678476499899,-2.20789618310877,1.68415540504794,-2.24583899888368,1.46540110005889,.0621780897499632,-2.32831086437971,3.10482709694771,-2.02588745115913,-.381749900992996,.50906751987416,2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217327,3.12121330296807,-1.52344991306454,2.03153653003234,-2.69130383032704,2.3938769796025,-3.19226020392739,2.0829372085896,4.93737708601548,-6.58404442855622,.508166439915615,2.07554270972349,-2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,1.94896143440088,-2.59896063091726,1.69581157410879,.0719545824309833,-1.81630452989169,2.83314651369184,-3.77803076303363,2.46515019062261,.104598208499715,-.637278791777431,.849817990926997,-.55450289173293,-2.63766838605638,3.51735861632833,.808065843843811,1.73374759666501,-2.31197070864004,1.50855178059943,.0640090061407144,4.12740556141282,-5.50393885468462,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843811,1.33269841309056,-1.77716739172311,1.15959472274786,4.37531622836173,-5.83453034569908,4.52786743647191,.950542947793588,-1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,1.51514018159912,-2.02045541424857,1.31833919902212,.0559381407795183,-1.41201150863478,3.36446686788906,-4.4865520602846,2.92745754610215,.124214263271286,1.91964817870762,-2.55987109498181,1.67030580640284,-1.95173675811295,2.60266155489591,-2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981012,.0386847883193903,3.55510980002441,-4.74077641992772,3.0933379389912,4.5904502888745,-6.12141388513832,1.24891795372526,.626153181688298,-.834981872710091,.544822382907868,2.84473124098147,-3.79347911889773,5.48872263996487,.347192252628366,-.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,1.09494796295235,-1.46012465838812,.952725588021822,.0404248755605117,-1.02041985541771,3.24473228828999,-4.32688473530227,2.82327524550586,.119793728554495,4.05017972204568,-5.4009573811209,3.52410341735766,1.63294619261043,-2.17755097302669,-2.28560008620039,.547225517982272,-.729731000653747,.476146603536493,.0202032646552823,2.39319484717323,-3.19135057367102,2.08234367785444,4.87179403320197,-6.49658873611034,5.49730994822616,.236321146004053,-.315136741011863,.205625482170815,1.3620925765208,-1.81636481949506,3.77635721863399,.0946962283839075,-.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,.714786270724957,-.953175031760129,.621942953572119,.0263895153231826,-.666134033494214,2.59853682213183,-3.46517626442722,2.26101386266496,.0959365479342438,4.7549644652192,-6.34079527029531,4.13734393811778,5.36731919542456,-7.15737676442222,3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831033,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.40719665133859,5.89679444984458,.0657431866024185,-.0876692328212795,.0572038290823807,.451524561772771,-.602112766253273,1.59193561065049,.0171974308138372,-.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,.398540259080393,-.531457639388288,.346774016236295,.0147138868003722,-.371413443660288,1.67165845871656,-2.22917418911978,1.45452737733324,.0617167093760001,3.84755914106515,-5.13076069911761,3.34780114662786,6.3113399865174,-8.41623844737545,7.98777931164405,.07249762486996,-.0966763475074086,.0630809359431453,.00267657237098111,.434860760346183,-.579891411113904,.37837686142582,1.40707187748205,-1.8763451919045,3.23180916099,.0113956405123911,-.0151962068293123,.00991546509750554,.0887334889355571,-.118327043544488,.36808995603673,.00166206328996398,-.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,.166073783576779,-.221461142188379,.144502522920426,.00613135259583728,-.154769899538562,.764836159193308,-1.01991708658492,.665491881382992,.0282373296477263,2.02218309922343,-2.69660249304008,1.75952250506474,4.02885967082133,-5.37252686958436,6.68010475273478,.0125887190510937,-.0167871896471047,.0109535751204636,.000464768572193082,.0823390061849577,-.109799933312678,.0716440239592634,.299004079114616,-.398725093710571,.796298871799331,.000824565926335038,-.00109956736063812,.0007174633715654,.00697737063218123,-.00930439734224517,.0321260785532874,5.01144631634162e-005,-6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,.0319379183519753,-.0425895510116904,.0277895142706456,.00117913035023069,-.029764050100784,.154660030302817,-.206240781926363,.134571297792104,.00570996311627599,.43925822557731,-.585755477154869,.382203141612095,.961797763577864,-1.28256746328293,1.79684767953686,.000465578153648566,-.000620853379055138,.000405104384309318,1.7188888943926e-005,.00319092434145458,-.00425513126926133,.00277645639124963,.0123137463285649,-.0164205106275597,.0353622090543065,5.86464816381893e-006,-7.82057018924326e-006,5.1028912428631e-006,5.19109369737154e-005,-6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,-9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,3.43216369764802,-3.49141841400203,.133641330308316,3.40070220695544,-4.09773034500049,-4.00974905232829,4.07897551472765,-.156131305216531,-3.97299303311745,4.32546942586304,-4.40014665135235,.168424801181435,-4.4144312457315,4.49064435720703,4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,7.27823584467215,-7.40389121595287,.283399396550352,-7.46210387346585,7.59093364435532,7.31197439196321,9.14177937313724,-9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,3.29802783242322,-3.35496675520006,.128418358139773,3.26779592010949,-3.93758279561022,-3.07100849759875,3.12402803860063,-.119578696522851,-3.04285759691612,2.20598407348367,-2.24406936639925,.0858964409190727,-.923194488569861,.939133013954983,-.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.9191336688493,-3.80884940873901,3.87460743053655,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,6.45780747384475,-6.5692985292465,.251453618731751,-4.10479708970297,4.17566451229335,1.32544341327556,7.79428043908965,-7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,3.06556135691961,-3.11848685368264,.119366596109985,3.03746038662953,-3.6600363220783,-1.5947619510804,1.62229477839744,-.0620967201893262,-1.58014330533833,-.444911232504622,.452592418790165,-.0173239199033356,2.09440936915946,-2.13056837715884,-2.63186985254377,4.2894282422279,-4.36348323472718,.167021432333127,4.25010849784967,-.964148801340397,.980794384058671,-.0375419530619817,-2.01722273728899,2.05204915464121,3.46190525493913,5.18623489409548,-5.27577283777776,.201941221845807,.205483038001612,-.209030607582029,-3.63947886955732,5.81833621896754,-5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,2.74931534527504,-2.7967810011421,.107052632842487,2.72411329584901,-3.28246374904242,.106757773657142,-.108600897231078,.00415692611322449,.105779161098541,-2.26282036321899,2.30188690834251,-.0881095280677934,2.15044627554921,-2.18757273479106,-.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,3.7410698100085,-3.80565764777524,.14566949316231,3.5848146864382,-3.64670485341608,-3.59724489879787,3.76406389114812,-3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,2.36915365304765,-2.41005599340123,.0922499256416775,2.34743641804055,-2.82858093939833,1.68415540504794,-1.71323156899255,.0655775157362847,1.66871731961812,-2.32831086437971,2.36850807266163,-.090659592246103,-.381749900992996,.388340636274758,2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622404,2.53843733881177,3.58714638367868,-3.64907680538233,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,2.3938769796025,-2.43520615704218,.0932126006787416,4.93737708601548,-5.02261861703304,.508166439915615,2.07554270972349,-2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,1.94896143440088,-1.98260935074574,.0758885128327144,1.93109596018978,-2.32690486658693,2.83314651369184,-2.88205947758122,.110316844590442,2.80717601371158,-.637278791777431,.648281114842891,-.0248143133692,-2.63766838605638,2.68320650944122,.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336551,1.7178549162266,4.12740556141282,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843811,1.33269841309056,-1.35570683401507,.0518925099589829,4.37531622836173,-4.45085403872016,4.52786743647191,.950542947793588,-.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,1.51514018159912,-1.54129837497396,.0589964137232872,1.50125140095791,-1.80895681150723,3.36446686788906,-3.42255283189436,.131005356341758,3.33362593315312,1.91964817870762,-1.95279001506978,.074747115516283,-1.95173675811295,1.98543258954328,-2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063146,1.03821099255249,3.55510980002441,-3.61648712494208,.138428596350645,4.5904502888745,-4.66970228947897,1.24891795372526,.626153181688298,-.636963426396326,.0243811052053838,2.84473124098147,-2.89384420968738,5.48872263996487,.347192252628366,-.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,1.09494796295235,-1.11385173231848,.042635000914324,1.08491094310714,-1.30728073869583,3.24473228828999,-3.30075109017009,.126343140340609,3.21498892012256,4.05017972204568,-4.12010419835037,.157705591526293,1.63294619261043,-1.66113825253464,-2.28560008620039,.547225517982272,-.556673113172381,.0213078258129724,.542209285918407,2.39319484717323,-2.43451224831939,.093186039847633,4.87179403320197,-4.95590330339709,5.49730994822616,.236321146004053,-.240401121148629,.00920185489986045,1.3620925765208,-1.3856084746561,3.77635721863399,.0946962283839075,-.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,.714786270724957,-.727126724577601,.027832293713515,.70823406529868,-.853397929051741,2.59853682213183,-2.6433993582315,.101181630171365,2.57471691018525,4.7549644652192,-4.83705672356223,.185148446522551,5.36731919542456,-5.45998347614275,3.25729414076559,.233201739437629,-.237227859491313,.00908039168484126,.231064057614915,1.22672710731592,-1.24790598293054,.0477662072687216,3.30495822058058,-3.36201679531997,5.89679444984458,.0657431866024185,-.0668782122732,.00255990322490874,.451524561772771,-.459319924425253,1.59193561065049,.0171974308138372,-.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,.398540259080393,-.405420872037066,.0155183304460168,.394886974518159,-.47582535601586,1.67165845871656,-1.70051886866409,.0650909105855692,1.6563349285852,3.84755914106515,-3.91398546692074,.149815966677054,6.3113399865174,-6.42030234909073,7.98777931164405,.07249762486996,-.0737492627952914,.00282290703160039,.0718330635538729,.434860760346183,-.442368430061906,.0169325753770764,1.40707187748205,-1.43136432250755,3.23180916099,.0113956405123911,-.0115923809684749,.000443722587923622,.0887334889355571,-.0902654315677738,.36808995603673,.00166206328996398,-.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,.166073783576779,-.168940970519646,.00646656841622833,.164551441038204,-.198278882483944,.764836159193308,-.778040701653472,.0297811324981325,.757825163693696,2.02218309922343,-2.05709515348012,.0787396124921655,4.02885967082133,-4.09841606758444,6.68010475273478,.0125887190510937,-.0128060574566477,.0004901785898768,.0124733225023729,.0823390061849577,-.0837605510092491,.00320610999159926,.299004079114616,-.304166246215339,.796298871799331,.000824565926335038,-.000838801675193727,3.21068856418541e-005,.00697737063218123,-.00709783170793108,.0321260785532874,5.01144631634162e-005,-5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,.0319379183519753,-.0324893117176766,.00124359624768518,.0316451541922531,-.0381313330936754,.154660030302817,-.157330164177211,.00602214055814436,.153242313889452,.43925822557731,-.446841815317497,.0171038035473819,.961797763577864,-.978402756631806,1.79684767953686,.000465578153648566,-.000473616144898038,1.81286469184544e-005,.00046131035548159,.00319092434145458,-.00324601417302372,.000124247970525588,.0123137463285649,-.0125263374623454,.0353622090543065,5.86464816381893e-006,-5.96589859842733e-006,2.283572264044e-007,5.19109369737154e-005,-5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,-6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,3.43216369764802,-2.18664125535227,-2.27972534004502,3.88041880521056,-.538361161121001,-4.00974905232829,2.55462252942968,2.66337136913609,-4.53343925473611,4.32546942586304,-2.75576887572707,-2.87308040304256,-4.4144312457315,2.81244670509951,4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,7.27823584467215,-4.63698476184224,-4.83437858766286,-7.46210387346585,4.75412760808657,7.31197439196321,9.14177937313724,-5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,3.29802783242322,-2.10118291403716,-2.19062908534965,3.72876422817853,-.517320923384123,-3.07100849759875,1.95654825118803,2.0398374063157,-3.47209520918979,2.20598407348367,-1.40543872850124,-1.4652674627761,-.923194488569861,.588169835027997,-.507379758856801,4.96464287844228,-3.16298810157525,-3.29763472137029,5.61304625700717,-3.80884940873901,2.42662879416799,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,6.45780747384475,-4.11428751282488,-4.28943041158218,-4.10479708970297,2.61517790433631,1.32544341327556,7.79428043908965,-4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,3.06556135691961,-1.95307786118934,-2.03621928395244,3.46593664692318,-.48085677636235,-1.5947619510804,1.01602737631215,1.05927908845788,-1.80304461367463,-.444911232504622,.283454211851926,.295520697836744,2.09440936915946,-1.33435416739534,-2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340397,.614261944257266,.640410728572749,-2.01722273728899,1.28517834515454,3.46190525493913,5.18623489409548,-3.30416500488362,-3.44482144542742,.205483038001612,-.130913828127032,-3.63947886955732,5.81833621896754,-3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,2.74931534527504,-1.75159662753591,-1.82616110784378,3.10838756746075,-.431251167473262,.106757773657142,-.0680156813349527,-.070911070513513,.120700790879792,-2.26282036321899,1.44164929037179,1.50301948784968,2.15044627554921,-1.37005543987348,-.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,3.7410698100085,-2.38344621859102,-2.48490817973373,3.5848146864382,-2.28389563582347,-3.59724489879787,3.76406389114812,-2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,2.36915365304765,-1.50939453196029,-1.57364860569271,2.6785751489707,-.371619894588022,1.68415540504794,-1.07298020016939,-1.11865636189275,1.9041132311976,-2.32831086437971,1.48337347495979,1.54651984745118,-.381749900992996,.243213947948102,2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001398,-2.69130383032704,2.3938769796025,-1.52514583383108,-1.59007043171884,4.93737708601548,-3.14561698874583,.508166439915615,2.07554270972349,-1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,1.94896143440088,-1.24168887412682,-1.29454686902578,2.20350404785819,-.305709527061058,2.83314651369184,-1.80500570449659,-1.88184377891507,3.20316744314248,-.637278791777431,.406012131085817,.423295838593301,-2.63766838605638,1.68046603255685,.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141282,-2.62958182530229,-2.74152163963163,1.98940127224618,-1.26745321062054,-2.96832239843811,1.33269841309056,-.849065950254408,-.885210208841134,4.37531622836173,-2.78752641722509,4.52786743647191,.950542947793588,-.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,1.51514018159912,-.965300119862265,-1.00639240141108,1.71302390303713,-.237661341149246,3.36446686788906,-2.14351141382237,-2.23475948529486,3.80388048506198,1.91964817870762,-1.22301331540983,-1.27507630270228,-1.95173675811295,1.24345704101385,-2.75500656898115,1.04781596872157,-.66756653456346,-.695984465482169,1.18466517033957,3.55510980002441,-2.2649705675878,-2.36138908644034,4.5904502888745,-2.92458893939423,1.24891795372526,.626153181688298,-.398923973340749,-.415905941872819,2.84473124098147,-1.81238636712449,5.48872263996487,.347192252628366,-.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,1.09494796295235,-.697594462028796,-.727290664744143,1.23795280192461,-.171750973622283,3.24473228828999,-2.06722817250133,-2.15522890943138,3.66850806244401,4.05017972204568,-2.58038102231816,-2.69022638759908,1.63294619261043,-1.04035466481729,-2.28560008620039,.547225517982272,-.348638934287124,-.363480297113412,.61869548740089,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320197,-3.10383383936015,5.49730994822616,.236321146004053,-.150560874408055,-.156970165907753,1.3620925765208,-.867793055767215,3.77635721863399,.0946962283839075,-.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,.714786270724957,-.455392366453167,-.474778162593069,.80814038343456,-.112119700737043,2.59853682213183,-1.65553520251249,-1.7260104019959,2.93791673259187,4.7549644652192,-3.02940138707022,-3.15836129611741,5.36731919542456,-3.41953432868287,3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066826,3.30495822058058,-2.10559828531327,5.89679444984458,.0657431866024185,-.0418851712113071,-.0436681993244439,.451524561772771,-.287667582930082,1.59193561065049,.0171974308138372,-.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,.398540259080393,-.253911132799744,-.264719986484413,.450591303972135,-.0625140918479549,1.67165845871656,-1.06501836010644,-1.11035559079056,1.88998413076671,3.84755914106515,-2.45129087378787,-2.55564094422791,6.3113399865174,-4.0209726602962,7.98777931164405,.07249762486996,-.0461884430503283,-.0481546590241492,.0819661216087146,.434860760346183,-.277051027559316,-.288844933452007,1.40707187748205,-.896449496423069,3.23180916099,.0113956405123911,-.00726019498888195,-.00756925739444607,.0887334889355571,-.056532358254944,.36808995603673,.00166206328996398,-.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,.166073783576779,-.105806079951921,-.110310185087204,.187763722716786,-.026049944826181,.764836159193308,-.487279293067771,-.508022497640507,.864726999689721,2.02218309922343,-1.28833860494792,-1.34318245298509,4.02885967082133,-2.56679795782664,6.68010475273478,.0125887190510937,-.00802030872060961,-.00836172873999097,.0142328590555998,.0823390061849577,-.0524584150880736,-.0546915402140795,.299004079114616,-.190496349504476,.796298871799331,.000824565926335038,-.000525333297401842,-.000547696439558021,.00697737063218123,-.00444530268095806,.0321260785532874,5.01144631634162e-005,-3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,.0319379183519753,-.0203477386368136,-.0212139303918346,.0361091456847505,-.00500970708930343,.154660030302817,-.0985343453312713,-.102728896816887,.174859285048361,.43925822557731,-.279852665057626,-.291765835225991,.961797763577864,-.612764091356553,1.79684767953686,.000465578153648566,-.000296621166194705,-.000309248161967433,.000526384631617522,.00319092434145458,-.00203294697585556,-.00211948838134787,.0123137463285649,-.00784512281993512,.0353622090543065,5.86464816381893e-006,-3.73638402898161e-006,-3.89543978958814e-006,5.19109369737154e-005,-3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,-4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,3.43216369764802,-.744472777196595,-3.65673100407672,1.66121346489508,3.26669916121878,-4.00974905232829,.869757178740834,4.27210791134071,-1.94077255839989,4.32546942586304,-.938240282077712,-4.60848594454081,-4.4144312457315,.957537046434959,4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,7.27823584467215,-1.5787266955735,-7.75445258971519,-7.46210387346585,1.61860962555059,7.31197439196321,9.14177937313724,-1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,3.29802783242322,-.715377282662343,-3.51381859652975,1.5962898991603,3.13902998310834,-3.07100849759875,.666134376965478,3.2719453330298,-1.4864094837286,2.20598407348367,-.47850138692202,-2.35032214902362,-.923194488569861,.200250694746914,-.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390441,-3.80884940873901,.82617991094677,4.05806335333585,2.02329330850018,-.438873818984882,.0162698697481229,6.45780747384475,-1.40076706380652,-6.88034339064389,-4.10479708970297,.890374108024764,1.32544341327556,7.79428043908965,-1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,3.06556135691961,-.66495283386875,-3.26614178293108,1.48377299342298,2.91777071127945,-1.5947619510804,.345920813804923,1.69910761471712,-.771886268054593,-.444911232504622,.0965059740773432,.474021882875009,2.09440936915946,-.454299648071273,-2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340397,.209134120399515,1.02723329226541,-2.01722273728899,.437557047580714,3.46190525493913,5.18623489409548,-1.1249494591216,-5.52557150711589,.205483038001612,-.0445714544901277,-3.63947886955732,5.81833621896754,-1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,2.74931534527504,-.596355713420295,-2.92920371774241,1.33070566358573,2.61677091290561,.106757773657142,-.0231568955595295,-.113742960840148,.0516722006304461,-2.26282036321899,.490829782123398,2.41087725104767,2.15044627554921,-.466454648534285,-.119501749587716,3.45007465931297,-.74835785526233,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239337,-1.88618542328239,-3.42380239520456,.742659122293478,1.33170919606025,3.7410698100085,-.811477795596299,-3.98584891938638,3.5848146864382,-.777584399208989,-3.59724489879787,3.76406389114812,-.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,2.36915365304765,-.513894602666682,-2.52416795342767,1.1467022833279,2.25493680750505,1.68415540504794,-.365311203903165,-1.79435010419287,.815153904032968,-2.32831086437971,.505035367744937,2.48065281165978,-.381749900992996,.0828056101286271,2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.330452474967379,-2.69130383032704,2.3938769796025,-.519257354978935,-2.55050893377231,4.93737708601548,-1.07096955656365,.508166439915615,2.07554270972349,-.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,1.94896143440088,-.422750445356608,-2.07648245560297,.943323589025389,1.85500204648398,2.83314651369184,-.614539585033937,-3.01851998146924,1.37128107917545,-.637278791777431,.138232541833329,.678976098432196,-2.63766838605638,.57213830168982,.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.8391571928101,4.12740556141282,-.89527812555293,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843811,1.33269841309056,-.289076447479325,-1.41989719504135,4.37531622836173,-.949052583847877,4.52786743647191,.950542947793588,-.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,1.51514018159912,-.328650005712205,-1.6142761726,.733348258592925,1.44209530674498,3.36446686788906,-.729788615894315,-3.58460475610227,1.62844728825442,1.91964817870762,-.416392088662432,-2.04525122687055,-1.95173675811295,.423352442647985,-2.75500656898115,1.04781596872157,-.227282418061714,-1.11637482313293,.507157044419472,3.55510980002441,-.771141093489059,-3.78772150021559,4.5904502888745,-.995717447724869,1.24891795372526,.626153181688298,-.135819278812355,-.66712253696142,2.84473124098147,-.617052435457252,5.48872263996487,.347192252628366,-.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,1.09494796295235,-.237505848402132,-1.1665906747753,.529969564290411,1.04216054572391,3.24473228828999,-.703816913228947,-3.45703591374062,1.57049413873188,4.05017972204568,-.878527020016876,-4.31518396831288,1.63294619261043,-.354203381414383,-2.28560008620039,.547225517982272,-.118699029853337,-.583030617100297,.264864522602584,2.39319484717323,-.519109393293246,-2.54978216897294,4.87179403320197,-1.05674389554722,5.49730994822616,.236321146004053,-.0512605677924316,-.251783696179213,1.3620925765208,-.295452354005303,3.77635721863399,.0946962283839075,-.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,.714786270724957,-.155044737648518,-.76155491046056,.345966184032581,.680326440323354,2.59853682213183,-.563650249872674,-2.76855972053751,1.25772682793123,4.7549644652192,-1.03140232004452,-5.06608294899263,5.36731919542456,-1.16422856843223,3.25729414076559,.233201739437629,-.0505839353643847,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929027,5.89679444984458,.0657431866024185,-.0142603957822122,-.0700447793236106,.451524561772771,-.0979404755348809,1.59193561065049,.0171974308138372,-.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,.398540259080393,-.0864476172000634,-.4246168452719,.192898854195334,.379326641949509,1.67165845871656,-.362600483446047,-1.78103547961334,.809105213656782,3.84755914106515,-.834576462901105,-4.09930587333892,6.3113399865174,-1.3689967095037,7.98777931164405,.07249762486996,-.0157255052185161,-.0772411621484507,.0350898270822693,.434860760346183,-.0943259199045095,-.463313805821701,1.40707187748205,-.305208842367266,3.23180916099,.0113956405123911,-.00247183552086188,-.0121412600482659,.0887334889355571,-.0192472366688018,.36808995603673,.00166206328996398,-.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,.166073783576779,-.0360231684064709,-.176940031673225,.080381948458016,.158067370120622,.764836159193308,-.165901090401825,-.814879575736018,.370191004490028,2.02218309922343,-.43863300248611,-2.1544950322241,4.02885967082133,-.873902474623302,6.68010475273478,.0125887190510937,-.00273062693415816,-.0134124020091826,.0060931096085216,.0823390061849577,-.0178602054051424,-.0877264674162827,.299004079114616,-.0648571621038786,.796298871799331,.000824565926335038,-.000178857111523577,-.000878517475844129,.00697737063218123,-.00151346583454046,.0321260785532874,5.01144631634162e-005,-1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,.0319379183519753,-.00692767387221854,-.0340276241262534,.0154583827232305,.0303981920101794,.154660030302817,-.0335474040556334,-.16477947382797,.0748575381590809,.43925822557731,-.0952797768044285,-.467998998326769,.961797763577864,-.208624155307574,1.79684767953686,.000465578153648566,-.000100988848916915,-.000496041045803229,.000225346098359972,.00319092434145458,-.000692145397560786,-.00339970729838099,.0123137463285649,-.00267098242974684,.0353622090543065,5.86464816381893e-006,-1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,-1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,-1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,3.43216369764802,.744472777196595,-3.65673100407672,-1.66121346489509,3.26669916121882,-4.00974905232829,-.869757178740834,4.27210791134071,1.9407725583999,4.32546942586304,.938240282077712,-4.60848594454081,-4.4144312457315,-.957537046434959,4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,7.27823584467215,1.5787266955735,-7.75445258971519,-7.46210387346585,-1.61860962555059,7.31197439196321,9.14177937313724,1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,3.29802783242322,.715377282662343,-3.51381859652975,-1.5962898991603,3.13902998310838,-3.07100849759875,-.666134376965478,3.2719453330298,1.48640948372861,2.20598407348367,.47850138692202,-2.35032214902362,-.923194488569861,-.200250694746914,-.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873901,-.82617991094677,4.05806335333585,2.02329330850018,.438873818984882,.0162698697481229,6.45780747384475,1.40076706380652,-6.8803433906439,-4.10479708970297,-.890374108024764,1.32544341327556,7.79428043908965,1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,3.06556135691961,.66495283386875,-3.26614178293108,-1.48377299342298,2.91777071127949,-1.5947619510804,-.345920813804923,1.69910761471712,.771886268054594,-.444911232504622,-.0965059740773432,.474021882875009,2.09440936915946,.454299648071273,-2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340397,-.209134120399515,1.02723329226541,-2.01722273728899,-.437557047580714,3.46190525493913,5.18623489409548,1.1249494591216,-5.52557150711589,.205483038001612,.0445714544901277,-3.63947886955732,5.81833621896754,1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,2.74931534527504,.596355713420295,-2.92920371774242,-1.33070566358573,2.61677091290564,.106757773657142,.0231568955595295,-.113742960840148,-.0516722006304461,-2.26282036321899,-.490829782123398,2.41087725104768,2.15044627554921,.466454648534285,-.119501749587716,3.45007465931297,.74835785526233,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239337,-1.88618542328239,-3.42380239520456,-.742659122293478,1.33170919606025,3.7410698100085,.811477795596299,-3.98584891938638,3.5848146864382,.777584399208989,-3.59724489879787,3.76406389114812,.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,2.36915365304765,.513894602666682,-2.52416795342767,-1.14670228332791,2.25493680750508,1.68415540504794,.365311203903165,-1.79435010419287,-.81515390403297,-2.32831086437971,-.505035367744937,2.48065281165978,-.381749900992996,-.0828056101286271,2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.330452474967379,-2.69130383032704,2.3938769796025,.519257354978935,-2.55050893377231,4.93737708601548,1.07096955656365,.508166439915615,2.07554270972349,.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,1.94896143440088,.422750445356608,-2.07648245560297,-.94332358902539,1.85500204648401,2.83314651369184,.614539585033937,-3.01851998146924,-1.37128107917545,-.637278791777431,-.138232541833329,.678976098432196,-2.63766838605638,-.57213830168982,.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810101,4.12740556141282,.89527812555293,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843811,1.33269841309056,.289076447479325,-1.41989719504135,4.37531622836173,.949052583847877,4.52786743647191,.950542947793588,.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,1.51514018159912,.328650005712205,-1.6142761726,-.733348258592926,1.44209530674499,3.36446686788906,.729788615894315,-3.58460475610227,-1.62844728825442,1.91964817870762,.416392088662432,-2.04525122687055,-1.95173675811295,-.423352442647985,-2.75500656898115,1.04781596872157,.227282418061714,-1.11637482313293,-.507157044419472,3.55510980002441,.771141093489059,-3.78772150021559,4.5904502888745,.995717447724869,1.24891795372526,.626153181688298,.135819278812355,-.66712253696142,2.84473124098147,.617052435457252,5.48872263996487,.347192252628366,.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,1.09494796295235,.237505848402132,-1.1665906747753,-.529969564290411,1.04216054572392,3.24473228828999,.703816913228947,-3.45703591374062,-1.57049413873189,4.05017972204568,.878527020016876,-4.31518396831288,1.63294619261043,.354203381414383,-2.28560008620039,.547225517982272,.118699029853337,-.583030617100298,-.264864522602585,2.39319484717323,.519109393293246,-2.54978216897294,4.87179403320197,1.05674389554722,5.49730994822616,.236321146004053,.0512605677924316,-.251783696179213,1.3620925765208,.295452354005303,3.77635721863399,.0946962283839075,.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,.714786270724957,.155044737648518,-.761554910460561,-.345966184032582,.680326440323361,2.59853682213183,.563650249872674,-2.76855972053751,-1.25772682793124,4.7549644652192,1.03140232004452,-5.06608294899263,5.36731919542456,1.16422856843223,3.25729414076559,.233201739437629,.0505839353643847,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929027,5.89679444984458,.0657431866024185,.0142603957822122,-.0700447793236106,.451524561772771,.0979404755348809,1.59193561065049,.0171974308138372,.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,.398540259080393,.0864476172000634,-.4246168452719,-.192898854195334,.379326641949514,1.67165845871656,.362600483446047,-1.78103547961334,-.809105213656783,3.84755914106515,.834576462901105,-4.09930587333892,6.3113399865174,1.3689967095037,7.98777931164405,.07249762486996,.0157255052185161,-.0772411621484508,-.0350898270822694,.434860760346183,.0943259199045095,-.463313805821702,1.40707187748205,.305208842367266,3.23180916099,.0113956405123911,.00247183552086188,-.0121412600482659,.0887334889355571,.0192472366688018,.36808995603673,.00166206328996398,.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,.166073783576779,.0360231684064709,-.176940031673225,-.0803819484580161,.158067370120624,.764836159193308,.165901090401825,-.814879575736019,-.370191004490028,2.02218309922343,.43863300248611,-2.15449503222411,4.02885967082133,.873902474623302,6.68010475273478,.0125887190510937,.00273062693415816,-.0134124020091826,-.00609310960852161,.0823390061849577,.0178602054051424,-.0877264674162827,.299004079114616,.0648571621038786,.796298871799331,.000824565926335038,.000178857111523577,-.000878517475844129,.00697737063218123,.00151346583454046,.0321260785532874,5.01144631634162e-005,1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,.0319379183519753,.00692767387221854,-.0340276241262535,-.0154583827232306,.0303981920101797,.154660030302817,.0335474040556334,-.16477947382797,-.074857538159081,.43925822557731,.0952797768044285,-.467998998326769,.961797763577864,.208624155307574,1.79684767953686,.000465578153648566,.000100988848916915,-.00049604104580323,-.000225346098359973,.00319092434145458,.000692145397560786,-.00339970729838099,.0123137463285649,.00267098242974684,.0353622090543065,5.86464816381893e-006,1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,3.43216369764802,2.18664125535227,-2.27972534004502,-3.88041880521058,-.538361161120983,-4.00974905232829,-2.55462252942968,2.6633713691361,4.53343925473613,4.32546942586304,2.75576887572707,-2.87308040304257,-4.4144312457315,-2.81244670509951,4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,7.27823584467215,4.63698476184224,-4.83437858766287,-7.46210387346585,-4.75412760808657,7.31197439196321,9.14177937313724,5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,3.29802783242322,2.10118291403716,-2.19062908534965,-3.72876422817854,-.517320923384106,-3.07100849759875,-1.95654825118803,2.0398374063157,3.47209520918981,2.20598407348367,1.40543872850124,-1.46526746277611,-.923194488569861,-.588169835027997,-.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873901,-2.42662879416799,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,6.45780747384475,4.11428751282488,-4.28943041158219,-4.10479708970297,-2.61517790433631,1.32544341327556,7.79428043908965,4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,3.06556135691961,1.95307786118934,-2.03621928395244,-3.46593664692319,-.480856776362333,-1.5947619510804,-1.01602737631215,1.05927908845788,1.80304461367463,-.444911232504622,-.283454211851926,.295520697836744,2.09440936915946,1.33435416739534,-2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611566,-4.8496457306699,-.964148801340397,-.614261944257266,.640410728572749,-2.01722273728899,-1.28517834515454,3.46190525493913,5.18623489409548,3.30416500488362,-3.44482144542743,.205483038001612,.130913828127032,-3.63947886955732,5.81833621896754,3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,2.74931534527504,1.75159662753591,-1.82616110784379,-3.10838756746077,-.431251167473247,.106757773657142,.0680156813349527,-.0709110705135131,-.120700790879792,-2.26282036321899,-1.44164929037179,1.50301948784969,2.15044627554921,1.37005543987348,-.119501749587716,3.45007465931297,2.19805238043167,-2.2916222307293,-3.90066901628344,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,3.7410698100085,2.38344621859102,-2.48490817973374,3.5848146864382,2.28389563582347,-3.59724489879787,3.76406389114812,2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,2.36915365304765,1.50939453196029,-1.57364860569271,-2.67857514897071,-.371619894588009,1.68415540504794,1.07298020016939,-1.11865636189276,-1.90411323119761,-2.32831086437971,-1.48337347495979,1.54651984745118,-.381749900992996,-.243213947948102,2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188741,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001398,-2.69130383032704,2.3938769796025,1.52514583383108,-1.59007043171884,4.93737708601548,3.14561698874583,.508166439915615,2.07554270972349,1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,1.94896143440088,1.24168887412682,-1.29454686902578,-2.2035040478582,-.305709527061048,2.83314651369184,1.80500570449659,-1.88184377891507,-3.20316744314249,-.637278791777431,-.406012131085817,.423295838593301,-2.63766838605638,-1.68046603255685,.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141282,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843811,1.33269841309056,.849065950254408,-.885210208841136,4.37531622836173,2.78752641722509,4.52786743647191,.950542947793588,.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,1.51514018159912,.965300119862265,-1.00639240141108,-1.71302390303713,-.237661341149238,3.36446686788906,2.14351141382237,-2.23475948529486,-3.80388048506199,1.91964817870762,1.22301331540983,-1.27507630270228,-1.95173675811295,-1.24345704101385,-2.75500656898115,1.04781596872157,.66756653456346,-.69598446548217,-1.18466517033958,3.55510980002441,2.2649705675878,-2.36138908644035,4.5904502888745,2.92458893939423,1.24891795372526,.626153181688298,.398923973340749,-.41590594187282,2.84473124098147,1.81238636712449,5.48872263996487,.347192252628366,.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,1.09494796295235,.697594462028796,-.727290664744144,-1.23795280192461,-.171750973622277,3.24473228828999,2.06722817250133,-2.15522890943138,-3.66850806244403,4.05017972204568,2.58038102231816,-2.69022638759909,1.63294619261043,1.04035466481729,-2.28560008620039,.547225517982272,.348638934287124,-.363480297113413,-.618695487400893,2.39319484717323,1.52471124557473,-1.58961734284588,4.87179403320197,3.10383383936015,5.49730994822616,.236321146004053,.150560874408055,-.156970165907753,1.3620925765208,.867793055767215,3.77635721863399,.0946962283839075,.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,.714786270724957,.455392366453167,-.47477816259307,-.808140383434563,-.112119700737039,2.59853682213183,1.65553520251249,-1.7260104019959,-2.93791673259189,4.7549644652192,3.02940138707022,-3.15836129611742,5.36731919542456,3.41953432868287,3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066827,3.30495822058058,2.10559828531327,5.89679444984458,.0657431866024185,.0418851712113071,-.043668199324444,.451524561772771,.287667582930082,1.59193561065049,.0171974308138372,.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,.398540259080393,.253911132799744,-.264719986484414,-.450591303972137,-.0625140918479527,1.67165845871656,1.06501836010644,-1.11035559079056,-1.88998413076672,3.84755914106515,2.45129087378787,-2.55564094422791,6.3113399865174,4.0209726602962,7.98777931164405,.07249762486996,.0461884430503283,-.0481546590241493,-.081966121608715,.434860760346183,.277051027559316,-.288844933452007,1.40707187748205,.896449496423069,3.23180916099,.0113956405123911,.00726019498888195,-.00756925739444608,.0887334889355571,.056532358254944,.36808995603673,.00166206328996398,.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,.166073783576779,.105806079951921,-.110310185087204,-.187763722716787,-.0260499448261801,.764836159193308,.487279293067771,-.508022497640508,-.864726999689725,2.02218309922343,1.28833860494792,-1.34318245298509,4.02885967082133,2.56679795782664,6.68010475273478,.0125887190510937,.00802030872060961,-.00836172873999098,-.0142328590555999,.0823390061849577,.0524584150880736,-.0546915402140795,.299004079114616,.190496349504476,.796298871799331,.000824565926335038,.000525333297401842,-.000547696439558022,.00697737063218123,.00444530268095806,.0321260785532874,5.01144631634162e-005,3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,.0319379183519753,.0203477386368136,-.0212139303918347,-.0361091456847507,-.00500970708930326,.154660030302817,.0985343453312713,-.102728896816887,-.174859285048362,.43925822557731,.279852665057626,-.291765835225991,.961797763577864,.612764091356553,1.79684767953686,.000465578153648566,.000296621166194705,-.000309248161967434,-.000526384631617524,.00319092434145458,.00203294697585556,-.00211948838134787,.0123137463285649,.00784512281993512,.0353622090543065,5.86464816381893e-006,3.73638402898161e-006,-3.89543978958815e-006,5.19109369737154e-005,3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,3.43216369764802,3.49141841400203,.13364133030831,-3.40070220695543,-4.09773034500053,-4.00974905232829,-4.07897551472765,-.156131305216524,3.97299303311744,4.32546942586304,4.40014665135235,.168424801181427,-4.4144312457315,-4.49064435720703,4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,7.27823584467215,7.40389121595287,.283399396550339,-7.46210387346585,-7.59093364435532,7.31197439196321,9.14177937313724,9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,3.29802783242322,3.35496675520006,.128418358139767,-3.26779592010948,-3.93758279561025,-3.07100849759875,-3.12402803860063,-.119578696522845,3.04285759691611,2.20598407348367,2.24406936639925,.0858964409190686,-.923194488569861,-.939133013954983,-.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873901,-3.87460743053655,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,6.45780747384475,6.5692985292465,.251453618731739,-4.10479708970297,-4.17566451229335,1.32544341327556,7.79428043908965,7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,3.06556135691961,3.11848685368264,.119366596109979,-3.03746038662953,-3.66003632207834,-1.5947619510804,-1.62229477839744,-.0620967201893232,1.58014330533832,-.444911232504622,-.452592418790165,-.0173239199033348,2.09440936915946,2.13056837715884,-2.63186985254377,4.2894282422279,4.36348323472718,.167021432333119,-4.25010849784966,-.964148801340397,-.980794384058671,-.0375419530619799,-2.01722273728899,-2.05204915464121,3.46190525493913,5.18623489409548,5.27577283777776,.201941221845797,.205483038001612,.209030607582029,-3.63947886955732,5.81833621896754,5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,2.74931534527504,2.7967810011421,.107052632842481,-2.72411329584901,-3.28246374904245,.106757773657142,.108600897231078,.00415692611322429,-.105779161098541,-2.26282036321899,-2.30188690834251,-.0881095280677891,2.15044627554921,2.18757273479106,-.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.41844898660569,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,3.7410698100085,3.80565764777524,.145669493162303,3.5848146864382,3.64670485341608,-3.59724489879787,3.76406389114812,3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,2.36915365304765,2.41005599340123,.0922499256416731,-2.34743641804054,-2.82858093939836,1.68415540504794,1.71323156899255,.0655775157362815,-1.66871731961812,-2.32831086437971,-2.36850807266163,-.0906595922460986,-.381749900992996,-.388340636274758,2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622356,-2.53843733881177,3.58714638367868,3.64907680538233,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,2.3938769796025,2.43520615704218,.0932126006787372,4.93737708601548,5.02261861703304,.508166439915615,2.07554270972349,2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,1.94896143440088,1.98260935074574,.0758885128327108,-1.93109596018978,-2.32690486658695,2.83314651369184,2.88205947758122,.110316844590437,-2.80717601371157,-.637278791777431,-.648281114842891,-.0248143133691988,-2.63766838605638,-2.68320650944122,.808065843843811,1.73374759666501,1.76367994615688,.0675085327336519,-1.71785491622659,4.12740556141282,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843811,1.33269841309056,1.35570683401507,.0518925099589804,4.37531622836173,4.45085403872016,4.52786743647191,.950542947793588,.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,1.51514018159912,1.54129837497396,.0589964137232844,-1.50125140095791,-1.80895681150725,3.36446686788906,3.42255283189436,.131005356341752,-3.33362593315312,1.91964817870762,1.95279001506978,.0747471155162795,-1.95173675811295,-1.98543258954328,-2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063126,-1.03821099255249,3.55510980002441,3.61648712494208,.138428596350638,4.5904502888745,4.66970228947897,1.24891795372526,.626153181688298,.636963426396326,.0243811052053826,2.84473124098147,2.89384420968738,5.48872263996487,.347192252628366,.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,1.09494796295235,1.11385173231848,.042635000914322,-1.08491094310714,-1.30728073869584,3.24473228828999,3.30075109017009,.126343140340603,-3.21498892012255,4.05017972204568,4.12010419835037,.157705591526285,1.63294619261043,1.66113825253464,-2.28560008620039,.547225517982272,.556673113172381,.0213078258129714,-.542209285918406,2.39319484717323,2.43451224831939,.0931860398476285,4.87179403320197,4.95590330339709,5.49730994822616,.236321146004053,.240401121148629,.00920185489986001,1.3620925765208,1.3856084746561,3.77635721863399,.0946962283839075,.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,.714786270724957,.727126724577601,.0278322937135137,-.708234065298678,-.853397929051749,2.59853682213183,2.6433993582315,.10118163017136,-2.57471691018525,4.7549644652192,4.83705672356223,.185148446522542,5.36731919542456,5.45998347614275,3.25729414076559,.233201739437629,.237227859491313,.00908039168484082,-.231064057614914,1.22672710731592,1.24790598293054,.0477662072687193,3.30495822058058,3.36201679531997,5.89679444984458,.0657431866024185,.0668782122732,.00255990322490862,.451524561772771,.459319924425253,1.59193561065049,.0171974308138372,.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,.398540259080393,.405420872037066,.015518330446016,-.394886974518158,-.475825356015865,1.67165845871656,1.70051886866409,.0650909105855661,-1.6563349285852,3.84755914106515,3.91398546692074,.149815966677047,6.3113399865174,6.42030234909073,7.98777931164405,.07249762486996,.0737492627952914,.00282290703160026,-.0718330635538727,.434860760346183,.442368430061906,.0169325753770755,1.40707187748205,1.43136432250755,3.23180916099,.0113956405123911,.0115923809684749,.0004437225879236,.0887334889355571,.0902654315677738,.36808995603673,.00166206328996398,.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,.166073783576779,.168940970519646,.00646656841622802,-.164551441038203,-.198278882483946,.764836159193308,.778040701653472,.0297811324981311,-.757825163693694,2.02218309922343,2.05709515348012,.0787396124921617,4.02885967082133,4.09841606758444,6.68010475273478,.0125887190510937,.0128060574566477,.000490178589876777,-.0124733225023729,.0823390061849577,.0837605510092491,.00320610999159911,.299004079114616,.304166246215339,.796298871799331,.000824565926335038,.000838801675193727,3.21068856418526e-005,.00697737063218123,.00709783170793108,.0321260785532874,5.01144631634162e-005,5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,.0319379183519753,.0324893117176766,.00124359624768512,-.031645154192253,-.0381313330936757,.154660030302817,.157330164177211,.00602214055814407,-.153242313889452,.43925822557731,.446841815317497,.0171038035473811,.961797763577864,.978402756631806,1.79684767953686,.000465578153648566,.000473616144898038,1.81286469184535e-005,-.000461310355481589,.00319092434145458,.00324601417302372,.000124247970525582,.0123137463285649,.0125263374623454,.0353622090543065,5.86464816381893e-006,5.96589859842733e-006,2.28357226404389e-007,5.19109369737154e-005,5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,3.43216369764802,4.57682649415414,2.98636125886026,-.126713592860266,-3.19855198842569,-4.00974905232829,-5.34704266035319,-3.48892427303921,.148037726033597,4.32546942586304,5.76805910494959,3.76363584617939,-4.4144312457315,-5.8866906319401,4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,7.27823584467215,9.70560427278211,6.3328685571284,-7.46210387346585,-9.9507942329456,7.31197439196321,9.14177937313724,12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,3.29802783242322,4.39795490297745,2.86964825021048,-.12176137061466,-3.07354613898791,-3.07100849759875,-4.09522222779324,-2.67211637191123,.113379941906936,2.20598407348367,2.94170303097044,1.91944964063812,-.923194488569861,-1.23108958885198,-.507379758856801,4.96464287844228,6.62040364808343,4.31978730278895,-.183291880031401,-3.80884940873901,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,6.45780747384475,8.61155438605811,5.61900531555836,-4.10479708970297,-5.47379022059323,1.32544341327556,7.79428043908965,10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,3.06556135691961,4.08795840577754,2.66737675689461,-.113178836410125,-2.85690259486619,-1.5947619510804,-2.12663188500128,-1.38761892782617,.0588777326765029,-.444911232504622,-.593293821521023,-.387121881618236,2.09440936915946,2.79291698671264,-2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640684,-.964148801340397,-1.28570259705192,-.838915879897341,-2.01722273728899,-2.68998779997356,3.46190525493913,5.18623489409548,6.91589893787678,4.51259681489188,.205483038001612,.274013798814207,-3.63947886955732,5.81833621896754,7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,2.74931534527504,3.6662410134057,2.39220781955879,-.101503207887315,-2.56218200503226,.106757773657142,.142362617365695,.0928910470423957,-.00394143819036155,-2.26282036321899,-3.01749482284837,-1.96890348630228,2.15044627554921,2.86764279249033,-.119501749587716,3.45007465931297,4.60071095140733,3.00194577372104,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,3.7410698100085,4.98875605403389,3.25514367429685,3.5848146864382,4.78038820054912,-3.59724489879787,3.76406389114812,5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,2.36915365304765,3.15929138677771,2.06142518510908,-.0874678476499865,-2.20789618310877,1.68415540504794,2.24583899888368,1.46540110005888,-.0621780897499608,-2.32831086437971,-3.10482709694771,-2.02588745115913,-.381749900992996,-.50906751987416,2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273613,3.58714638367868,4.78349754217327,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,2.3938769796025,3.19226020392739,2.0829372085896,4.93737708601548,6.58404442855622,.508166439915615,2.07554270972349,2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,1.94896143440088,2.59896063091726,1.69581157410878,-.0719545824309805,-1.81630452989169,2.83314651369184,3.77803076303363,2.46515019062261,-.104598208499711,-.637278791777431,-.849817990926997,-.55450289173293,-2.63766838605638,-3.51735861632833,.808065843843811,1.73374759666501,2.31197070864004,1.50855178059943,-.0640090061407119,4.12740556141282,5.50393885468462,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843811,1.33269841309056,1.77716739172311,1.15959472274786,4.37531622836173,5.83453034569908,4.52786743647191,.950542947793588,1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,1.51514018159912,2.02045541424857,1.31833919902212,-.0559381407795162,-1.41201150863478,3.36446686788906,4.4865520602846,2.92745754610214,-.124214263271281,1.91964817870762,2.55987109498181,1.67030580640284,-1.95173675811295,-2.60266155489591,-2.75500656898115,1.04781596872157,1.39727364720762,.91171555098101,-.0386847883193888,3.55510980002441,4.74077641992772,3.0933379389912,4.5904502888745,6.12141388513832,1.24891795372526,.626153181688298,.834981872710091,.544822382907867,2.84473124098147,3.79347911889773,5.48872263996487,.347192252628366,.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,1.09494796295235,1.46012465838812,.95272558802182,-.0404248755605102,-1.02041985541771,3.24473228828999,4.32688473530227,2.82327524550585,-.11979372855449,4.05017972204568,5.4009573811209,3.52410341735765,1.63294619261043,2.17755097302669,-2.28560008620039,.547225517982272,.729731000653747,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367102,2.08234367785444,4.87179403320197,6.49658873611034,5.49730994822616,.236321146004053,.315136741011863,.205625482170815,1.3620925765208,1.81636481949506,3.77635721863399,.0946962283839075,.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,.714786270724957,.953175031760129,.621942953572118,-.0263895153231816,-.666134033494215,2.59853682213183,3.46517626442722,2.26101386266496,-.0959365479342401,4.7549644652192,6.34079527029531,4.13734393811777,5.36731919542456,7.15737676442222,3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831,1.22672710731592,1.63585353791591,1.06738798948578,3.30495822058058,4.40719665133859,5.89679444984458,.0657431866024185,.0876692328212795,.0572038290823806,.451524561772771,.602112766253273,1.59193561065049,.0171974308138372,.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,.398540259080393,.531457639388288,.346774016236294,-.0147138868003716,-.371413443660288,1.67165845871656,2.22917418911978,1.45452737733324,-.0617167093759977,3.84755914106515,5.13076069911761,3.34780114662785,6.3113399865174,8.41623844737545,7.98777931164405,.07249762486996,.0966763475074086,.0630809359431452,-.00267657237098101,.434860760346183,.579891411113904,.378376861425819,1.40707187748205,1.8763451919045,3.23180916099,.0113956405123911,.0151962068293123,.00991546509750553,.0887334889355571,.118327043544488,.36808995603673,.00166206328996398,.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,.166073783576779,.221461142188379,.144502522920425,-.00613135259583705,-.154769899538562,.764836159193308,1.01991708658492,.665491881382991,-.0282373296477252,2.02218309922343,2.69660249304008,1.75952250506473,4.02885967082133,5.37252686958436,6.68010475273478,.0125887190510937,.0167871896471047,.0109535751204636,-.000464768572193064,.0823390061849577,.109799933312678,.0716440239592633,.299004079114616,.398725093710571,.796298871799331,.000824565926335038,.00109956736063812,.000717463371565399,.00697737063218123,.00930439734224517,.0321260785532874,5.01144631634162e-005,6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,.0319379183519753,.0425895510116904,.0277895142706455,-.00117913035023065,-.0297640501007841,.154660030302817,.206240781926363,.134571297792104,-.00570996311627577,.43925822557731,.585755477154869,.382203141612094,.961797763577864,1.28256746328293,1.79684767953686,.000465578153648566,.000620853379055138,.000405104384309317,-1.71888889439253e-005,.00319092434145458,.00425513126926133,.00277645639124963,.0123137463285649,.0164205106275597,.0353622090543065,5.86464816381893e-006,7.82057018924326e-006,5.10289124286309e-006,5.19109369737154e-005,6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,3.43216369764802,5.37468949190064,5.57281420860799,4.46269573688388,2.39877236199692,-4.00974905232829,-6.27917489029831,-6.5106412373467,-5.21370528571982,4.32546942586304,6.77358573667404,7.02327732701018,-4.4144312457315,-6.91289790468793,4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,7.27823584467215,11.3975501068473,11.8176927763525,-7.46210387346585,-11.6854832203471,7.31197439196321,9.14177937313724,14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,3.29802783242322,5.16463580891203,5.35501741292456,4.28828460541245,2.30502351007756,-3.07100849759875,-4.80912874954551,-4.9864054592391,-3.99310107128126,2.20598407348367,3.45452037324197,3.58186277642042,-.923194488569861,-1.44570135775136,-.507379758856801,4.96464287844228,7.77451667983119,8.06110512843118,6.45531290809181,-3.80884940873901,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,6.45780747384475,10.1127781284249,10.4855608369157,-4.10479708970297,-6.42801793274623,1.32544341327556,7.79428043908965,12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,3.06556135691961,4.80059864950596,4.97756091846909,3.98601838486204,2.14255044478903,-1.5947619510804,-2.49736057507455,-2.58941963373659,-2.07360079279396,-.444911232504622,-.696720767591157,-.72240366631615,2.09440936915946,3.27979695088037,-2.63186985254377,4.2894282422279,6.71714607327472,6.96475715335038,5.57736018045176,-.964148801340397,-1.50983487071459,-1.56549122161564,-2.01722273728899,-3.15892446006539,3.46190525493913,5.18623489409548,8.12152468439221,8.42090472917073,.205483038001612,.321781716436834,-3.63947886955732,5.81833621896754,9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,2.74931534527504,4.30536466145163,4.46407134676942,3.5748172149004,1.92152305240544,.106757773657142,.16718022074295,.173342908623152,.138812569451129,-2.26282036321899,-3.54352469706742,-3.67414802589246,2.15044627554921,3.3675494587808,-.119501749587716,3.45007465931297,5.40273764744738,5.60189629155852,4.48598459696891,1.77035093065653,2.7723288820058,2.87452397106014,-3.42380239520456,-5.36159588761006,1.33170919606025,3.7410698100085,5.8584293670558,6.0743859656501,3.5848146864382,5.61373743608464,-3.59724489879787,3.76406389114812,5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,2.36915365304765,3.71004018615404,3.8468016980446,3.08050921776368,1.6558243734555,1.68415540504794,2.63734867013603,2.7345680458079,2.18983527963514,-2.32831086437971,-3.64608131714141,-3.78048514370594,-.381749900992996,-.597811573855878,2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635523,3.33115715545233,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,2.3938769796025,3.74875634795766,3.8869450362802,4.93737708601548,7.73181907803741,.508166439915615,2.07554270972349,3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,1.94896143440088,3.05202882623936,3.16453436679059,2.53415123836086,1.36214797290772,2.83314651369184,4.43664234772504,4.6001882618506,3.68381930162858,-.637278791777431,-.997963945475626,-1.03475143356997,-2.63766838605638,-4.13052808940267,.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141282,6.46342227832349,6.70168045166517,1.98940127224618,3.11535668501805,-2.96832239843811,1.33269841309056,2.0869750946082,2.16390630168514,4.37531622836173,6.85164468857581,4.52786743647191,.950542947793588,1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,1.51514018159912,2.37267471198348,2.46013753301878,1.97007200846434,1.05894610873647,3.36446686788906,5.26867781509235,5.46289467143736,4.37467244553793,1.91964817870762,3.0061249410246,3.1169383486426,-1.95173675811295,-3.05637492003247,-2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591415,1.36243031247302,3.55510980002441,5.56722026499826,5.7724421542407,4.5904502888745,7.18853968487789,1.24891795372526,.626153181688298,.980541495928357,1.01668674800429,2.84473124098147,4.45478376427994,5.48872263996487,.347192252628366,.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,1.09494796295235,1.7146633520688,1.77787020176478,1.42371402906152,.765269708188692,3.24473228828999,5.08117621439183,5.26848128513792,4.21898669004741,4.05017972204568,6.34248838569497,6.57628863172099,1.63294619261043,2.55715621899877,-2.28560008620039,.547225517982272,.856942588087487,.888531670338268,.711533948435091,2.39319484717323,3.74768814487269,3.88583745543936,4.87179403320197,7.62911752450547,5.49730994822616,.236321146004053,.370073484889949,.383715334269613,1.3620925765208,2.13300568008583,3.77635721863399,.0946962283839075,.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,.714786270724957,1.11933887677121,1.16060055303996,.929406031924798,.499571029238743,2.59853682213183,4.06924896099705,4.21925182103647,3.37876634868698,4.7549644652192,7.44616510000515,7.72064967553201,5.36731919542456,8.40509854032022,3.25729414076559,.233201739437629,.365188565907959,.378650344801754,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064311,5.89679444984458,.0657431866024185,.102952319693431,.106747403902078,.451524561772771,.707077089737652,1.59193561065049,.0171974308138372,.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,.398540259080393,.624104888716883,.647110981340287,.518204861963161,.278543636855161,1.67165845871656,2.61777873930361,2.71427671758157,2.17358603453451,3.84755914106515,6.02518921139496,6.24729300320554,6.3113399865174,9.88341340931671,7.98777931164405,.07249762486996,.113529614838378,.11771460504762,.0942655625163294,.434860760346183,.68098195938567,.70608468526914,1.40707187748205,2.20344223207683,3.23180916099,.0113956405123911,.0178453112143479,.0185031347484812,.0887334889355571,.138954604977606,.36808995603673,.00166206328996398,.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,.166073783576779,.260067729310804,.269654486884813,.215938641412748,.116070571566627,.764836159193308,1.1977158528687,1.24186670359493,.994483762912882,2.02218309922343,3.16669227065688,3.28342459824587,4.02885967082133,6.30910167728468,6.68010475273478,.0125887190510937,.019713644795161,.02044033985683,.0163685732441573,.0823390061849577,.12894099187304,.133694084564441,.299004079114616,.468233518150739,.796298871799331,.000824565926335038,.00129125129528224,.00133885010038651,.00697737063218123,.0109264021014796,.0321260785532874,5.01144631634162e-005,7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,.0319379183519753,.0500140463222003,.0518576912013894,.0415275099413171,.0223217196472688,.154660030302817,.24219405407586,.251121942525088,.201097826062499,.43925822557731,.687868288629403,.713224862474736,.961797763577864,1.50615319946369,1.79684767953686,.000465578153648566,.000729084691473618,.00075596060662465,.000605371370823784,.00319092434145458,.00499691420436583,.00518111316265919,.0123137463285649,.0192830438371028,.0353622090543065,5.86464816381893e-006,9.18390427744748e-006,9.52244633459443e-006,5.19109369737154e-005,8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,3.43216369764802,5.83506573884863,7.25392389703223,8.09901689119005,8.47552946493044,-4.00974905232829,-6.81702604882233,-8.4746582765388,-9.46196865399931,4.32546942586304,7.35378632026677,9.14193750783762,-4.4144312457315,-7.50503145913621,4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,7.27823584467215,12.3738225688191,15.3826488429769,-7.46210387346585,-12.6864189798666,7.31197439196321,9.14177937313724,15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,3.29802783242322,5.60701962552938,6.97042595115334,7.78249101017962,8.14428871473084,-3.07100849759875,-5.22106113152625,-6.4906175528327,-7.24678421547911,2.20598407348367,3.75042195543293,4.66237685613718,-.923194488569861,-1.5695348492102,-.507379758856801,4.96464287844228,8.44045334766827,10.4928391542811,11.7152706263275,-3.80884940873901,-6.47547397319737,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,6.45780747384475,10.9790017210618,13.6486625036235,-4.10479708970297,-6.97861844197595,1.32544341327556,7.79428043908965,13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,3.06556135691961,5.21180037431139,6.47910494479514,7.2339303710034,7.57022621762745,-1.5947619510804,-2.71127534923244,-3.37055072311448,-3.76322492874004,-.444911232504622,-.756399320675517,-.940325842964601,2.09440936915946,3.56073236366527,-2.63186985254377,4.2894282422279,7.29251224169724,9.06576398575724,10.1219390639593,-.964148801340397,-1.63916180436132,-2.03774139204767,-2.01722273728899,-3.42950637731078,3.46190525493913,5.18623489409548,8.8171847889118,10.9611768422514,.205483038001612,.349344361529711,-3.63947886955732,5.81833621896754,9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,2.74931534527504,4.67414645388234,5.81071476784009,6.48767173775814,6.78927500842464,.106757773657142,.181500267083405,.225633983149896,.251920680005669,-2.26282036321899,-3.84705006415361,-4.78250111498198,2.15044627554921,3.65600143048661,-.119501749587716,3.45007465931297,5.86551639682902,7.29177895707792,8.14128212331115,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429708,1.33170919606025,3.7410698100085,6.36024100270088,7.90680110023492,3.5848146864382,6.09458965574695,-3.59724489879787,3.76406389114812,6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,2.36915365304765,4.02782866109789,5.00723794478697,5.59058866190049,5.85048772793475,1.68415540504794,2.86325431258853,3.55948498424847,3.97417031416819,-2.32831086437971,-3.95839130925952,-4.9209161654707,-.381749900992996,-.649017927110068,2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830962,-1.52344991306454,-2.59003683420207,-2.69130383032704,2.3938769796025,4.06986110773884,5.05949102727518,4.93737708601548,8.39409842544749,.508166439915615,2.07554270972349,3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,1.94896143440088,3.31345445440235,4.11915606854152,4.59904391748785,4.81284738096772,2.83314651369184,4.81666890668704,5.98789306701478,6.68549157438517,-.637278791777431,-1.08344588754871,-1.34689725343212,-2.63766838605638,-4.48433402043889,.808065843843811,1.73374759666501,2.94756663562365,3.66429874441148,4.09119503456581,4.12740556141282,7.01705539432372,8.72332687917667,1.98940127224618,3.38220674597771,-2.96832239843811,1.33269841309056,2.26573774926528,2.81667592802333,4.37531622836173,7.43853151653138,4.52786743647191,.950542947793588,1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,1.51514018159912,2.57590935107797,3.20226904625419,3.57533818439403,3.74155092353066,3.36446686788906,5.7199734894835,7.11084574520705,7.93926992086965,1.91964817870762,3.26361861024795,4.05720211125474,-1.95173675811295,-3.31817281869637,-2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898135,3.55510980002441,6.0440879938959,7.51377213010338,4.5904502888745,7.8042837062112,1.24891795372526,.626153181688298,1.06453109468611,1.32338312764,2.84473124098147,4.83636425064192,5.48872263996487,.347192252628366,.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,1.09494796295235,1.86153514438243,2.31418717000874,2.58379343998139,2.70391057656362,3.24473228828999,5.51641118734948,6.85778510905627,7.65672734172536,4.05017972204568,6.88576274670089,8.56010903534823,1.63294619261043,2.776192869346,-2.28560008620039,.547225517982272,.930345156393644,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070527,4.87179403320197,8.28259983241364,5.49730994822616,.236321146004053,.40177262626832,.499467904136592,1.3620925765208,2.31571114636374,3.77635721863399,.0946962283839075,.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,.714786270724957,1.21521735159799,1.51071034695561,1.68671036412374,1.76512329607374,2.59853682213183,4.41780594598826,5.49204234459095,6.13187350041211,4.7549644652192,8.08397637227704,10.049681020309,5.36731919542456,9.12504854432544,3.25729414076559,.233201739437629,.396469282990197,.492874996717866,.550295671800536,1.22672710731592,2.08557456680069,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984458,.0657431866024185,.111770838907699,.138949104550329,.451524561772771,.76764272750461,1.59193561065049,.0171974308138372,.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,.398540259080393,.677563431168936,.842320170000573,.940451730878478,.984172086870924,1.67165845871656,2.84200809304711,3.53307252095549,3.94468075935688,3.84755914106515,6.54128488547504,8.13186772631248,6.3113399865174,10.7299904588883,7.98777931164405,.07249762486996,.123254146474207,.153224700235958,.1710756072316,.434860760346183,.739312383715058,.919084035475439,1.40707187748205,2.3921810358159,3.23180916099,.0113956405123911,.0193738753136992,.0240848387004708,.0887334889355571,.150856948291044,.36808995603673,.00166206328996398,.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,.166073783576779,.282344179950942,.350999163642368,.391891091702267,.410109589767539,.764836159193308,1.30030781284792,1.61649145672597,1.80481513164276,2.02218309922343,3.43793955015111,4.2739110377176,4.02885967082133,6.84951625494143,6.68010475273478,.0125887190510937,.0214022435166571,.0266064261611111,.0297061146457812,.0823390061849577,.139985605707145,.174024591276674,.299004079114616,.508340689012672,.796298871799331,.000824565926335038,.00140185515920491,.00174273111828065,.00697737063218123,.0118623177482726,.0321260785532874,5.01144631634162e-005,8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,.0319379183519753,.054298066631689,.0675012177634823,.075365210691831,.07886883956794,.154660030302817,.262939511058368,.326876043550518,.364957592022401,.43925822557731,.746788570738901,.928378137184539,.961797763577864,1.63516477461472,1.79684767953686,.000465578153648566,.000791535419924754,.000984005657525639,.00109864378994572,.00319092434145458,.00542493160168594,.00674406129057541,.0123137463285649,.0209347588552144,.0353622090543065,5.86464816381893e-006,9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,3.43216369764802,-5.83506573884863,7.25392389703222,-8.09901689119009,8.47552946493044,-4.00974905232829,6.81702604882233,-8.4746582765388,9.46196865399936,4.32546942586304,-7.35378632026677,9.14193750783762,-4.4144312457315,7.50503145913621,4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,7.27823584467215,-12.3738225688191,15.3826488429769,-7.46210387346585,12.6864189798666,7.31197439196321,9.14177937313724,-15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,3.29802783242322,-5.60701962552938,6.97042595115334,-7.78249101017965,8.14428871473084,-3.07100849759875,5.22106113152625,-6.4906175528327,7.24678421547914,2.20598407348367,-3.75042195543293,4.66237685613718,-.923194488569861,1.5695348492102,-.507379758856801,4.96464287844228,-8.44045334766827,10.4928391542811,-11.7152706263276,-3.80884940873901,6.47547397319737,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,6.45780747384475,-10.9790017210618,13.6486625036235,-4.10479708970297,6.97861844197595,1.32544341327556,7.79428043908965,-13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,3.06556135691961,-5.21180037431139,6.47910494479513,-7.23393037100344,7.57022621762746,-1.5947619510804,2.71127534923244,-3.37055072311448,3.76322492874006,-.444911232504622,.756399320675517,-.940325842964601,2.09440936915946,-3.56073236366527,-2.63186985254377,4.2894282422279,-7.29251224169724,9.06576398575723,-10.1219390639593,-.964148801340397,1.63916180436132,-2.03774139204767,-2.01722273728899,3.42950637731078,3.46190525493913,5.18623489409548,-8.8171847889118,10.9611768422514,.205483038001612,-.349344361529711,-3.63947886955732,5.81833621896754,-9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,2.74931534527504,-4.67414645388234,5.81071476784009,-6.48767173775818,6.78927500842464,.106757773657142,-.181500267083405,.225633983149896,-.251920680005671,-2.26282036321899,3.84705006415361,-4.78250111498198,2.15044627554921,-3.65600143048661,-.119501749587716,3.45007465931297,-5.86551639682902,7.29177895707792,-8.14128212331119,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429708,1.33170919606025,3.7410698100085,-6.36024100270088,7.90680110023492,3.5848146864382,-6.09458965574695,-3.59724489879787,3.76406389114812,-6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,2.36915365304765,-4.02782866109789,5.00723794478697,-5.59058866190052,5.85048772793476,1.68415540504794,-2.86325431258853,3.55948498424847,-3.97417031416821,-2.32831086437971,3.95839130925952,-4.9209161654707,-.381749900992996,.649017927110068,2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,2.3938769796025,-4.06986110773884,5.05949102727518,4.93737708601548,-8.39409842544749,.508166439915615,2.07554270972349,-3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,1.94896143440088,-3.31345445440235,4.11915606854152,-4.59904391748787,4.81284738096772,2.83314651369184,-4.81666890668704,5.98789306701478,-6.6854915743852,-.637278791777431,1.08344588754871,-1.34689725343212,-2.63766838605638,4.48433402043889,.808065843843811,1.73374759666501,-2.94756663562365,3.66429874441148,-4.09119503456583,4.12740556141282,-7.01705539432372,8.72332687917667,1.98940127224618,-3.38220674597771,-2.96832239843811,1.33269841309056,-2.26573774926528,2.81667592802333,4.37531622836173,-7.43853151653138,4.52786743647191,.950542947793588,-1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,1.51514018159912,-2.57590935107797,3.20226904625418,-3.57533818439405,3.74155092353066,3.36446686788906,-5.7199734894835,7.11084574520705,-7.93926992086969,1.91964817870762,-3.26361861024795,4.05720211125474,-1.95173675811295,3.31817281869637,-2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898136,3.55510980002441,-6.0440879938959,7.51377213010338,4.5904502888745,-7.8042837062112,1.24891795372526,.626153181688298,-1.06453109468611,1.32338312764,2.84473124098147,-4.83636425064192,5.48872263996487,.347192252628366,-.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,1.09494796295235,-1.86153514438243,2.31418717000874,-2.5837934399814,2.70391057656363,3.24473228828999,-5.51641118734948,6.85778510905627,-7.65672734172539,4.05017972204568,-6.88576274670089,8.56010903534823,1.63294619261043,-2.776192869346,-2.28560008620039,.547225517982272,-.930345156393644,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070527,4.87179403320197,-8.28259983241364,5.49730994822616,.236321146004053,-.40177262626832,.499467904136592,1.3620925765208,-2.31571114636374,3.77635721863399,.0946962283839075,-.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,.714786270724957,-1.21521735159799,1.51071034695561,-1.68671036412375,1.76512329607374,2.59853682213183,-4.41780594598826,5.49204234459095,-6.13187350041214,4.7549644652192,-8.08397637227704,10.0496810203089,5.36731919542456,-9.12504854432544,3.25729414076559,.233201739437629,-.396469282990197,.492874996717866,-.550295671800538,1.22672710731592,-2.08557456680069,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984458,.0657431866024185,-.111770838907699,.138949104550329,.451524561772771,-.76764272750461,1.59193561065049,.0171974308138372,-.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,.398540259080393,-.677563431168936,.842320170000572,-.940451730878483,.984172086870924,1.67165845871656,-2.84200809304711,3.53307252095548,-3.9446807593569,3.84755914106515,-6.54128488547504,8.13186772631248,6.3113399865174,-10.7299904588883,7.98777931164405,.07249762486996,-.123254146474207,.153224700235958,-.171075607231601,.434860760346183,-.739312383715058,.919084035475439,1.40707187748205,-2.3921810358159,3.23180916099,.0113956405123911,-.0193738753136992,.0240848387004708,.0887334889355571,-.150856948291044,.36808995603673,.00166206328996398,-.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,.166073783576779,-.282344179950942,.350999163642368,-.391891091702269,.410109589767539,.764836159193308,-1.30030781284792,1.61649145672597,-1.80481513164277,2.02218309922343,-3.43793955015111,4.27391103771759,4.02885967082133,-6.84951625494143,6.68010475273478,.0125887190510937,-.0214022435166571,.0266064261611111,-.0297061146457814,.0823390061849577,-.139985605707145,.174024591276674,.299004079114616,-.508340689012672,.796298871799331,.000824565926335038,-.00140185515920491,.00174273111828065,.00697737063218123,-.0118623177482726,.0321260785532874,5.01144631634162e-005,-8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,.0319379183519753,-.054298066631689,.0675012177634823,-.0753652106918314,.07886883956794,.154660030302817,-.262939511058368,.326876043550518,-.364957592022403,.43925822557731,-.746788570738901,.928378137184539,.961797763577864,-1.63516477461472,1.79684767953686,.000465578153648566,-.000791535419924754,.000984005657525639,-.00109864378994572,.00319092434145458,-.00542493160168594,.0067440612905754,.0123137463285649,-.0209347588552144,.0353622090543065,5.86464816381893e-006,-9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,-8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,-1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,3.43216369764802,-5.37468949190064,5.572814208608,-4.46269573688388,2.39877236199698,-4.00974905232829,6.27917489029831,-6.51064123734671,5.21370528571982,4.32546942586304,-6.77358573667404,7.02327732701019,-4.4144312457315,6.91289790468793,4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,7.27823584467215,-11.3975501068473,11.8176927763525,-7.46210387346585,11.6854832203471,7.31197439196321,9.14177937313724,-14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,3.29802783242322,-5.16463580891203,5.35501741292457,-4.28828460541245,2.30502351007762,-3.07100849759875,4.80912874954551,-4.98640545923911,3.99310107128126,2.20598407348367,-3.45452037324197,3.58186277642043,-.923194488569861,1.44570135775136,-.507379758856801,4.96464287844228,-7.77451667983119,8.06110512843119,-6.45531290809181,-3.80884940873901,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,6.45780747384475,-10.1127781284249,10.4855608369157,-4.10479708970297,6.42801793274623,1.32544341327556,7.79428043908965,-12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,3.06556135691961,-4.80059864950596,4.9775609184691,-3.98601838486204,2.14255044478909,-1.5947619510804,2.49736057507455,-2.5894196337366,2.07360079279396,-.444911232504622,.696720767591157,-.722403666316151,2.09440936915946,-3.27979695088037,-2.63186985254377,4.2894282422279,-6.71714607327472,6.9647571533504,-5.57736018045176,-.964148801340397,1.50983487071459,-1.56549122161564,-2.01722273728899,3.15892446006539,3.46190525493913,5.18623489409548,-8.12152468439221,8.42090472917074,.205483038001612,-.321781716436834,-3.63947886955732,5.81833621896754,-9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,2.74931534527504,-4.30536466145163,4.46407134676943,-3.5748172149004,1.9215230524055,.106757773657142,-.16718022074295,.173342908623153,-.13881256945113,-2.26282036321899,3.54352469706742,-3.67414802589247,2.15044627554921,-3.3675494587808,-.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155853,-4.48598459696891,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761006,1.33170919606025,3.7410698100085,-5.8584293670558,6.07438596565011,3.5848146864382,-5.61373743608464,-3.59724489879787,3.76406389114812,-5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,2.36915365304765,-3.71004018615404,3.84680169804461,-3.08050921776368,1.65582437345554,1.68415540504794,-2.63734867013603,2.73456804580791,-2.18983527963514,-2.32831086437971,3.64608131714141,-3.78048514370595,-.381749900992996,.597811573855878,2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,2.3938769796025,-3.74875634795766,3.88694503628021,4.93737708601548,-7.73181907803741,.508166439915615,2.07554270972349,-3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,1.94896143440088,-3.05202882623936,3.1645343667906,-2.53415123836086,1.36214797290775,2.83314651369184,-4.43664234772504,4.60018826185061,-3.68381930162858,-.637278791777431,.997963945475626,-1.03475143356998,-2.63766838605638,4.13052808940267,.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633232,4.12740556141282,-6.46342227832349,6.70168045166518,1.98940127224618,-3.11535668501805,-2.96832239843811,1.33269841309056,-2.0869750946082,2.16390630168515,4.37531622836173,-6.85164468857581,4.52786743647191,.950542947793588,-1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,1.51514018159912,-2.37267471198348,2.46013753301879,-1.97007200846434,1.0589461087365,3.36446686788906,-5.26867781509235,5.46289467143737,-4.37467244553793,1.91964817870762,-3.0061249410246,3.11693834864261,-1.95173675811295,3.05637492003247,-2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002441,-5.56722026499826,5.77244215424071,4.5904502888745,-7.18853968487789,1.24891795372526,.626153181688298,-.980541495928357,1.01668674800429,2.84473124098147,-4.45478376427994,5.48872263996487,.347192252628366,-.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,1.09494796295235,-1.7146633520688,1.77787020176478,-1.42371402906152,.765269708188713,3.24473228828999,-5.08117621439183,5.26848128513793,-4.21898669004741,4.05017972204568,-6.34248838569497,6.576288631721,1.63294619261043,-2.55715621899877,-2.28560008620039,.547225517982272,-.856942588087487,.888531670338269,-.711533948435091,2.39319484717323,-3.74768814487269,3.88583745543936,4.87179403320197,-7.62911752450547,5.49730994822616,.236321146004053,-.370073484889949,.383715334269613,1.3620925765208,-2.13300568008583,3.77635721863399,.0946962283839075,-.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,.714786270724957,-1.11933887677121,1.16060055303996,-.929406031924799,.499571029238757,2.59853682213183,-4.06924896099705,4.21925182103647,-3.37876634868698,4.7549644652192,-7.44616510000515,7.72064967553202,5.36731919542456,-8.40509854032022,3.25729414076559,.233201739437629,-.365188565907959,.378650344801755,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064311,5.89679444984458,.0657431866024185,-.102952319693431,.106747403902078,.451524561772771,-.707077089737652,1.59193561065049,.0171974308138372,-.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,.398540259080393,-.624104888716883,.647110981340288,-.518204861963161,.278543636855169,1.67165845871656,-2.61777873930361,2.71427671758158,-2.17358603453451,3.84755914106515,-6.02518921139496,6.24729300320555,6.3113399865174,-9.88341340931671,7.98777931164405,.07249762486996,-.113529614838378,.117714605047621,-.0942655625163295,.434860760346183,-.68098195938567,.706084685269141,1.40707187748205,-2.20344223207683,3.23180916099,.0113956405123911,-.0178453112143479,.0185031347484812,.0887334889355571,-.138954604977606,.36808995603673,.00166206328996398,-.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,.166073783576779,-.260067729310804,.269654486884814,-.215938641412748,.11607057156663,.764836159193308,-1.1977158528687,1.24186670359493,-.994483762912882,2.02218309922343,-3.16669227065688,3.28342459824587,4.02885967082133,-6.30910167728468,6.68010475273478,.0125887190510937,-.019713644795161,.02044033985683,-.0163685732441573,.0823390061849577,-.12894099187304,.133694084564441,.299004079114616,-.468233518150739,.796298871799331,.000824565926335038,-.00129125129528224,.00133885010038651,.00697737063218123,-.0109264021014796,.0321260785532874,5.01144631634162e-005,-7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,.0319379183519753,-.0500140463222003,.0518576912013895,-.0415275099413171,.0223217196472694,.154660030302817,-.24219405407586,.251121942525088,-.201097826062499,.43925822557731,-.687868288629403,.713224862474738,.961797763577864,-1.50615319946369,1.79684767953686,.000465578153648566,-.000729084691473618,.000755960606624651,-.000605371370823784,.00319092434145458,-.00499691420436583,.0051811131626592,.0123137463285649,-.0192830438371028,.0353622090543065,5.86464816381893e-006,-9.18390427744748e-006,9.52244633459444e-006,5.19109369737154e-005,-8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,-1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,3.43216369764802,-4.57682649415414,2.98636125886026,.126713592860271,-3.19855198842569,-4.00974905232829,5.34704266035319,-3.48892427303922,-.148037726033603,4.32546942586304,-5.76805910494959,3.7636358461794,-4.4144312457315,5.8866906319401,4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,7.27823584467215,-9.70560427278211,6.33286855712841,-7.46210387346585,9.9507942329456,7.31197439196321,9.14177937313724,-12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,3.29802783242322,-4.39795490297745,2.86964825021048,.121761370614664,-3.07354613898791,-3.07100849759875,4.09522222779324,-2.67211637191124,-.11337994190694,2.20598407348367,-2.94170303097044,1.91944964063812,-.923194488569861,1.23108958885198,-.507379758856801,4.96464287844228,-6.62040364808343,4.31978730278896,.183291880031408,-3.80884940873901,5.07914086475798,-3.31411940634736,2.02329330850018,-2.6980829707225,.0162698697481229,6.45780747384475,-8.61155438605811,5.61900531555837,-4.10479708970297,5.47379022059323,1.32544341327556,7.79428043908965,-10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,3.06556135691961,-4.08795840577754,2.66737675689462,.113178836410129,-2.85690259486619,-1.5947619510804,2.12663188500128,-1.38761892782618,-.0588777326765052,-.444911232504622,.593293821521023,-.387121881618236,2.09440936915946,-2.79291698671264,-2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.15836332764069,-.964148801340397,1.28570259705192,-.838915879897342,-2.01722273728899,2.68998779997356,3.46190525493913,5.18623489409548,-6.91589893787678,4.51259681489189,.205483038001612,-.274013798814207,-3.63947886955732,5.81833621896754,-7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,2.74931534527504,-3.6662410134057,2.39220781955879,.101503207887319,-2.56218200503226,.106757773657142,-.142362617365695,.0928910470423959,.0039414381903617,-2.26282036321899,3.01749482284837,-1.96890348630228,2.15044627554921,-2.86764279249033,-.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,3.7410698100085,-4.98875605403389,3.25514367429686,3.5848146864382,-4.78038820054912,-3.59724489879787,3.76406389114812,-5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,2.36915365304765,-3.15929138677771,2.06142518510909,.0874678476499899,-2.20789618310877,1.68415540504794,-2.24583899888368,1.46540110005889,.0621780897499632,-2.32831086437971,3.10482709694771,-2.02588745115913,-.381749900992996,.50906751987416,2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217327,3.12121330296807,-1.52344991306454,2.03153653003234,-2.69130383032704,2.3938769796025,-3.19226020392739,2.0829372085896,4.93737708601548,-6.58404442855622,.508166439915615,2.07554270972349,-2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,1.94896143440088,-2.59896063091726,1.69581157410879,.0719545824309833,-1.81630452989169,2.83314651369184,-3.77803076303363,2.46515019062261,.104598208499715,-.637278791777431,.849817990926997,-.55450289173293,-2.63766838605638,3.51735861632833,.808065843843811,1.73374759666501,-2.31197070864004,1.50855178059943,.0640090061407144,4.12740556141282,-5.50393885468462,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843811,1.33269841309056,-1.77716739172311,1.15959472274786,4.37531622836173,-5.83453034569908,4.52786743647191,.950542947793588,-1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,1.51514018159912,-2.02045541424857,1.31833919902212,.0559381407795183,-1.41201150863478,3.36446686788906,-4.4865520602846,2.92745754610215,.124214263271286,1.91964817870762,-2.55987109498181,1.67030580640284,-1.95173675811295,2.60266155489591,-2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981012,.0386847883193903,3.55510980002441,-4.74077641992772,3.0933379389912,4.5904502888745,-6.12141388513832,1.24891795372526,.626153181688298,-.834981872710091,.544822382907868,2.84473124098147,-3.79347911889773,5.48872263996487,.347192252628366,-.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,1.09494796295235,-1.46012465838812,.952725588021822,.0404248755605117,-1.02041985541771,3.24473228828999,-4.32688473530227,2.82327524550586,.119793728554495,4.05017972204568,-5.4009573811209,3.52410341735766,1.63294619261043,-2.17755097302669,-2.28560008620039,.547225517982272,-.729731000653747,.476146603536493,.0202032646552823,2.39319484717323,-3.19135057367102,2.08234367785444,4.87179403320197,-6.49658873611034,5.49730994822616,.236321146004053,-.315136741011863,.205625482170815,1.3620925765208,-1.81636481949506,3.77635721863399,.0946962283839075,-.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,.714786270724957,-.953175031760129,.621942953572119,.0263895153231826,-.666134033494214,2.59853682213183,-3.46517626442722,2.26101386266496,.0959365479342438,4.7549644652192,-6.34079527029531,4.13734393811778,5.36731919542456,-7.15737676442222,3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831033,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.40719665133859,5.89679444984458,.0657431866024185,-.0876692328212795,.0572038290823807,.451524561772771,-.602112766253273,1.59193561065049,.0171974308138372,-.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,.398540259080393,-.531457639388288,.346774016236295,.0147138868003722,-.371413443660288,1.67165845871656,-2.22917418911978,1.45452737733324,.0617167093760001,3.84755914106515,-5.13076069911761,3.34780114662786,6.3113399865174,-8.41623844737545,7.98777931164405,.07249762486996,-.0966763475074086,.0630809359431453,.00267657237098111,.434860760346183,-.579891411113904,.37837686142582,1.40707187748205,-1.8763451919045,3.23180916099,.0113956405123911,-.0151962068293123,.00991546509750554,.0887334889355571,-.118327043544488,.36808995603673,.00166206328996398,-.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,.166073783576779,-.221461142188379,.144502522920426,.00613135259583728,-.154769899538562,.764836159193308,-1.01991708658492,.665491881382992,.0282373296477263,2.02218309922343,-2.69660249304008,1.75952250506474,4.02885967082133,-5.37252686958436,6.68010475273478,.0125887190510937,-.0167871896471047,.0109535751204636,.000464768572193082,.0823390061849577,-.109799933312678,.0716440239592634,.299004079114616,-.398725093710571,.796298871799331,.000824565926335038,-.00109956736063812,.0007174633715654,.00697737063218123,-.00930439734224517,.0321260785532874,5.01144631634162e-005,-6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,.0319379183519753,-.0425895510116904,.0277895142706456,.00117913035023069,-.029764050100784,.154660030302817,-.206240781926363,.134571297792104,.00570996311627599,.43925822557731,-.585755477154869,.382203141612095,.961797763577864,-1.28256746328293,1.79684767953686,.000465578153648566,-.000620853379055138,.000405104384309318,1.7188888943926e-005,.00319092434145458,-.00425513126926133,.00277645639124963,.0123137463285649,-.0164205106275597,.0353622090543065,5.86464816381893e-006,-7.82057018924326e-006,5.1028912428631e-006,5.19109369737154e-005,-6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,-9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,3.43216369764802,-3.49141841400203,.133641330308316,3.40070220695544,-4.09773034500049,-4.00974905232829,4.07897551472765,-.156131305216531,-3.97299303311745,4.32546942586304,-4.40014665135235,.168424801181435,-4.4144312457315,4.49064435720703,4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,7.27823584467215,-7.40389121595287,.283399396550352,-7.46210387346585,7.59093364435532,7.31197439196321,9.14177937313724,-9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,3.29802783242322,-3.35496675520006,.128418358139773,3.26779592010949,-3.93758279561022,-3.07100849759875,3.12402803860063,-.119578696522851,-3.04285759691612,2.20598407348367,-2.24406936639925,.0858964409190727,-.923194488569861,.939133013954983,-.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.9191336688493,-3.80884940873901,3.87460743053655,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,6.45780747384475,-6.5692985292465,.251453618731751,-4.10479708970297,4.17566451229335,1.32544341327556,7.79428043908965,-7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,3.06556135691961,-3.11848685368264,.119366596109985,3.03746038662953,-3.6600363220783,-1.5947619510804,1.62229477839744,-.0620967201893262,-1.58014330533833,-.444911232504622,.452592418790165,-.0173239199033356,2.09440936915946,-2.13056837715884,-2.63186985254377,4.2894282422279,-4.36348323472718,.167021432333127,4.25010849784967,-.964148801340397,.980794384058671,-.0375419530619817,-2.01722273728899,2.05204915464121,3.46190525493913,5.18623489409548,-5.27577283777776,.201941221845807,.205483038001612,-.209030607582029,-3.63947886955732,5.81833621896754,-5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,2.74931534527504,-2.7967810011421,.107052632842487,2.72411329584901,-3.28246374904242,.106757773657142,-.108600897231078,.00415692611322449,.105779161098541,-2.26282036321899,2.30188690834251,-.0881095280677934,2.15044627554921,-2.18757273479106,-.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,3.7410698100085,-3.80565764777524,.14566949316231,3.5848146864382,-3.64670485341608,-3.59724489879787,3.76406389114812,-3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,2.36915365304765,-2.41005599340123,.0922499256416775,2.34743641804055,-2.82858093939833,1.68415540504794,-1.71323156899255,.0655775157362847,1.66871731961812,-2.32831086437971,2.36850807266163,-.090659592246103,-.381749900992996,.388340636274758,2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622404,2.53843733881177,3.58714638367868,-3.64907680538233,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,2.3938769796025,-2.43520615704218,.0932126006787416,4.93737708601548,-5.02261861703304,.508166439915615,2.07554270972349,-2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,1.94896143440088,-1.98260935074574,.0758885128327144,1.93109596018978,-2.32690486658693,2.83314651369184,-2.88205947758122,.110316844590442,2.80717601371158,-.637278791777431,.648281114842891,-.0248143133692,-2.63766838605638,2.68320650944122,.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336551,1.7178549162266,4.12740556141282,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843811,1.33269841309056,-1.35570683401507,.0518925099589829,4.37531622836173,-4.45085403872016,4.52786743647191,.950542947793588,-.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,1.51514018159912,-1.54129837497396,.0589964137232872,1.50125140095791,-1.80895681150723,3.36446686788906,-3.42255283189436,.131005356341758,3.33362593315312,1.91964817870762,-1.95279001506978,.074747115516283,-1.95173675811295,1.98543258954328,-2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063146,1.03821099255249,3.55510980002441,-3.61648712494208,.138428596350645,4.5904502888745,-4.66970228947897,1.24891795372526,.626153181688298,-.636963426396326,.0243811052053838,2.84473124098147,-2.89384420968738,5.48872263996487,.347192252628366,-.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,1.09494796295235,-1.11385173231848,.042635000914324,1.08491094310714,-1.30728073869583,3.24473228828999,-3.30075109017009,.126343140340609,3.21498892012256,4.05017972204568,-4.12010419835037,.157705591526293,1.63294619261043,-1.66113825253464,-2.28560008620039,.547225517982272,-.556673113172381,.0213078258129724,.542209285918407,2.39319484717323,-2.43451224831939,.093186039847633,4.87179403320197,-4.95590330339709,5.49730994822616,.236321146004053,-.240401121148629,.00920185489986045,1.3620925765208,-1.3856084746561,3.77635721863399,.0946962283839075,-.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,.714786270724957,-.727126724577601,.027832293713515,.70823406529868,-.853397929051741,2.59853682213183,-2.6433993582315,.101181630171365,2.57471691018525,4.7549644652192,-4.83705672356223,.185148446522551,5.36731919542456,-5.45998347614275,3.25729414076559,.233201739437629,-.237227859491313,.00908039168484126,.231064057614915,1.22672710731592,-1.24790598293054,.0477662072687216,3.30495822058058,-3.36201679531997,5.89679444984458,.0657431866024185,-.0668782122732,.00255990322490874,.451524561772771,-.459319924425253,1.59193561065049,.0171974308138372,-.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,.398540259080393,-.405420872037066,.0155183304460168,.394886974518159,-.47582535601586,1.67165845871656,-1.70051886866409,.0650909105855692,1.6563349285852,3.84755914106515,-3.91398546692074,.149815966677054,6.3113399865174,-6.42030234909073,7.98777931164405,.07249762486996,-.0737492627952914,.00282290703160039,.0718330635538729,.434860760346183,-.442368430061906,.0169325753770764,1.40707187748205,-1.43136432250755,3.23180916099,.0113956405123911,-.0115923809684749,.000443722587923622,.0887334889355571,-.0902654315677738,.36808995603673,.00166206328996398,-.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,.166073783576779,-.168940970519646,.00646656841622833,.164551441038204,-.198278882483944,.764836159193308,-.778040701653472,.0297811324981325,.757825163693696,2.02218309922343,-2.05709515348012,.0787396124921655,4.02885967082133,-4.09841606758444,6.68010475273478,.0125887190510937,-.0128060574566477,.0004901785898768,.0124733225023729,.0823390061849577,-.0837605510092491,.00320610999159926,.299004079114616,-.304166246215339,.796298871799331,.000824565926335038,-.000838801675193727,3.21068856418541e-005,.00697737063218123,-.00709783170793108,.0321260785532874,5.01144631634162e-005,-5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,.0319379183519753,-.0324893117176766,.00124359624768518,.0316451541922531,-.0381313330936754,.154660030302817,-.157330164177211,.00602214055814436,.153242313889452,.43925822557731,-.446841815317497,.0171038035473819,.961797763577864,-.978402756631806,1.79684767953686,.000465578153648566,-.000473616144898038,1.81286469184544e-005,.00046131035548159,.00319092434145458,-.00324601417302372,.000124247970525588,.0123137463285649,-.0125263374623454,.0353622090543065,5.86464816381893e-006,-5.96589859842733e-006,2.283572264044e-007,5.19109369737154e-005,-5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,-6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,3.43216369764802,-2.18664125535227,-2.27972534004502,3.88041880521056,-.538361161121001,-4.00974905232829,2.55462252942968,2.66337136913609,-4.53343925473611,4.32546942586304,-2.75576887572707,-2.87308040304256,-4.4144312457315,2.81244670509951,4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,7.27823584467215,-4.63698476184224,-4.83437858766286,-7.46210387346585,4.75412760808657,7.31197439196321,9.14177937313724,-5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,3.29802783242322,-2.10118291403716,-2.19062908534965,3.72876422817853,-.517320923384123,-3.07100849759875,1.95654825118803,2.0398374063157,-3.47209520918979,2.20598407348367,-1.40543872850124,-1.4652674627761,-.923194488569861,.588169835027997,-.507379758856801,4.96464287844228,-3.16298810157525,-3.29763472137029,5.61304625700717,-3.80884940873901,2.42662879416799,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,6.45780747384475,-4.11428751282488,-4.28943041158218,-4.10479708970297,2.61517790433631,1.32544341327556,7.79428043908965,-4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,3.06556135691961,-1.95307786118934,-2.03621928395244,3.46593664692318,-.48085677636235,-1.5947619510804,1.01602737631215,1.05927908845788,-1.80304461367463,-.444911232504622,.283454211851926,.295520697836744,2.09440936915946,-1.33435416739534,-2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340397,.614261944257266,.640410728572749,-2.01722273728899,1.28517834515454,3.46190525493913,5.18623489409548,-3.30416500488362,-3.44482144542742,.205483038001612,-.130913828127032,-3.63947886955732,5.81833621896754,-3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,2.74931534527504,-1.75159662753591,-1.82616110784378,3.10838756746075,-.431251167473262,.106757773657142,-.0680156813349527,-.070911070513513,.120700790879792,-2.26282036321899,1.44164929037179,1.50301948784968,2.15044627554921,-1.37005543987348,-.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,3.7410698100085,-2.38344621859102,-2.48490817973373,3.5848146864382,-2.28389563582347,-3.59724489879787,3.76406389114812,-2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,2.36915365304765,-1.50939453196029,-1.57364860569271,2.6785751489707,-.371619894588022,1.68415540504794,-1.07298020016939,-1.11865636189275,1.9041132311976,-2.32831086437971,1.48337347495979,1.54651984745118,-.381749900992996,.243213947948102,2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001398,-2.69130383032704,2.3938769796025,-1.52514583383108,-1.59007043171884,4.93737708601548,-3.14561698874583,.508166439915615,2.07554270972349,-1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,1.94896143440088,-1.24168887412682,-1.29454686902578,2.20350404785819,-.305709527061058,2.83314651369184,-1.80500570449659,-1.88184377891507,3.20316744314248,-.637278791777431,.406012131085817,.423295838593301,-2.63766838605638,1.68046603255685,.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141282,-2.62958182530229,-2.74152163963163,1.98940127224618,-1.26745321062054,-2.96832239843811,1.33269841309056,-.849065950254408,-.885210208841134,4.37531622836173,-2.78752641722509,4.52786743647191,.950542947793588,-.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,1.51514018159912,-.965300119862265,-1.00639240141108,1.71302390303713,-.237661341149246,3.36446686788906,-2.14351141382237,-2.23475948529486,3.80388048506198,1.91964817870762,-1.22301331540983,-1.27507630270228,-1.95173675811295,1.24345704101385,-2.75500656898115,1.04781596872157,-.66756653456346,-.695984465482169,1.18466517033957,3.55510980002441,-2.2649705675878,-2.36138908644034,4.5904502888745,-2.92458893939423,1.24891795372526,.626153181688298,-.398923973340749,-.415905941872819,2.84473124098147,-1.81238636712449,5.48872263996487,.347192252628366,-.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,1.09494796295235,-.697594462028796,-.727290664744143,1.23795280192461,-.171750973622283,3.24473228828999,-2.06722817250133,-2.15522890943138,3.66850806244401,4.05017972204568,-2.58038102231816,-2.69022638759908,1.63294619261043,-1.04035466481729,-2.28560008620039,.547225517982272,-.348638934287124,-.363480297113412,.61869548740089,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320197,-3.10383383936015,5.49730994822616,.236321146004053,-.150560874408055,-.156970165907753,1.3620925765208,-.867793055767215,3.77635721863399,.0946962283839075,-.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,.714786270724957,-.455392366453167,-.474778162593069,.80814038343456,-.112119700737043,2.59853682213183,-1.65553520251249,-1.7260104019959,2.93791673259187,4.7549644652192,-3.02940138707022,-3.15836129611741,5.36731919542456,-3.41953432868287,3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066826,3.30495822058058,-2.10559828531327,5.89679444984458,.0657431866024185,-.0418851712113071,-.0436681993244439,.451524561772771,-.287667582930082,1.59193561065049,.0171974308138372,-.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,.398540259080393,-.253911132799744,-.264719986484413,.450591303972135,-.0625140918479549,1.67165845871656,-1.06501836010644,-1.11035559079056,1.88998413076671,3.84755914106515,-2.45129087378787,-2.55564094422791,6.3113399865174,-4.0209726602962,7.98777931164405,.07249762486996,-.0461884430503283,-.0481546590241492,.0819661216087146,.434860760346183,-.277051027559316,-.288844933452007,1.40707187748205,-.896449496423069,3.23180916099,.0113956405123911,-.00726019498888195,-.00756925739444607,.0887334889355571,-.056532358254944,.36808995603673,.00166206328996398,-.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,.166073783576779,-.105806079951921,-.110310185087204,.187763722716786,-.026049944826181,.764836159193308,-.487279293067771,-.508022497640507,.864726999689721,2.02218309922343,-1.28833860494792,-1.34318245298509,4.02885967082133,-2.56679795782664,6.68010475273478,.0125887190510937,-.00802030872060961,-.00836172873999097,.0142328590555998,.0823390061849577,-.0524584150880736,-.0546915402140795,.299004079114616,-.190496349504476,.796298871799331,.000824565926335038,-.000525333297401842,-.000547696439558021,.00697737063218123,-.00444530268095806,.0321260785532874,5.01144631634162e-005,-3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,.0319379183519753,-.0203477386368136,-.0212139303918346,.0361091456847505,-.00500970708930343,.154660030302817,-.0985343453312713,-.102728896816887,.174859285048361,.43925822557731,-.279852665057626,-.291765835225991,.961797763577864,-.612764091356553,1.79684767953686,.000465578153648566,-.000296621166194705,-.000309248161967433,.000526384631617522,.00319092434145458,-.00203294697585556,-.00211948838134787,.0123137463285649,-.00784512281993512,.0353622090543065,5.86464816381893e-006,-3.73638402898161e-006,-3.89543978958814e-006,5.19109369737154e-005,-3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,-4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,3.43216369764802,-.744472777196595,-3.65673100407672,1.66121346489508,3.26669916121878,-4.00974905232829,.869757178740834,4.27210791134071,-1.94077255839989,4.32546942586304,-.938240282077712,-4.60848594454081,-4.4144312457315,.957537046434959,4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,7.27823584467215,-1.5787266955735,-7.75445258971519,-7.46210387346585,1.61860962555059,7.31197439196321,9.14177937313724,-1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,3.29802783242322,-.715377282662343,-3.51381859652975,1.5962898991603,3.13902998310834,-3.07100849759875,.666134376965478,3.2719453330298,-1.4864094837286,2.20598407348367,-.47850138692202,-2.35032214902362,-.923194488569861,.200250694746914,-.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390441,-3.80884940873901,.82617991094677,4.05806335333585,2.02329330850018,-.438873818984882,.0162698697481229,6.45780747384475,-1.40076706380652,-6.88034339064389,-4.10479708970297,.890374108024764,1.32544341327556,7.79428043908965,-1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,3.06556135691961,-.66495283386875,-3.26614178293108,1.48377299342298,2.91777071127945,-1.5947619510804,.345920813804923,1.69910761471712,-.771886268054593,-.444911232504622,.0965059740773432,.474021882875009,2.09440936915946,-.454299648071273,-2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340397,.209134120399515,1.02723329226541,-2.01722273728899,.437557047580714,3.46190525493913,5.18623489409548,-1.1249494591216,-5.52557150711589,.205483038001612,-.0445714544901277,-3.63947886955732,5.81833621896754,-1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,2.74931534527504,-.596355713420295,-2.92920371774241,1.33070566358573,2.61677091290561,.106757773657142,-.0231568955595295,-.113742960840148,.0516722006304461,-2.26282036321899,.490829782123398,2.41087725104767,2.15044627554921,-.466454648534285,-.119501749587716,3.45007465931297,-.74835785526233,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239337,-1.88618542328239,-3.42380239520456,.742659122293478,1.33170919606025,3.7410698100085,-.811477795596299,-3.98584891938638,3.5848146864382,-.777584399208989,-3.59724489879787,3.76406389114812,-.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,2.36915365304765,-.513894602666682,-2.52416795342767,1.1467022833279,2.25493680750505,1.68415540504794,-.365311203903165,-1.79435010419287,.815153904032968,-2.32831086437971,.505035367744937,2.48065281165978,-.381749900992996,.0828056101286271,2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.330452474967379,-2.69130383032704,2.3938769796025,-.519257354978935,-2.55050893377231,4.93737708601548,-1.07096955656365,.508166439915615,2.07554270972349,-.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,1.94896143440088,-.422750445356608,-2.07648245560297,.943323589025389,1.85500204648398,2.83314651369184,-.614539585033937,-3.01851998146924,1.37128107917545,-.637278791777431,.138232541833329,.678976098432196,-2.63766838605638,.57213830168982,.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.8391571928101,4.12740556141282,-.89527812555293,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843811,1.33269841309056,-.289076447479325,-1.41989719504135,4.37531622836173,-.949052583847877,4.52786743647191,.950542947793588,-.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,1.51514018159912,-.328650005712205,-1.6142761726,.733348258592925,1.44209530674498,3.36446686788906,-.729788615894315,-3.58460475610227,1.62844728825442,1.91964817870762,-.416392088662432,-2.04525122687055,-1.95173675811295,.423352442647985,-2.75500656898115,1.04781596872157,-.227282418061714,-1.11637482313293,.507157044419472,3.55510980002441,-.771141093489059,-3.78772150021559,4.5904502888745,-.995717447724869,1.24891795372526,.626153181688298,-.135819278812355,-.66712253696142,2.84473124098147,-.617052435457252,5.48872263996487,.347192252628366,-.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,1.09494796295235,-.237505848402132,-1.1665906747753,.529969564290411,1.04216054572391,3.24473228828999,-.703816913228947,-3.45703591374062,1.57049413873188,4.05017972204568,-.878527020016876,-4.31518396831288,1.63294619261043,-.354203381414383,-2.28560008620039,.547225517982272,-.118699029853337,-.583030617100297,.264864522602584,2.39319484717323,-.519109393293246,-2.54978216897294,4.87179403320197,-1.05674389554722,5.49730994822616,.236321146004053,-.0512605677924316,-.251783696179213,1.3620925765208,-.295452354005303,3.77635721863399,.0946962283839075,-.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,.714786270724957,-.155044737648518,-.76155491046056,.345966184032581,.680326440323354,2.59853682213183,-.563650249872674,-2.76855972053751,1.25772682793123,4.7549644652192,-1.03140232004452,-5.06608294899263,5.36731919542456,-1.16422856843223,3.25729414076559,.233201739437629,-.0505839353643847,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929027,5.89679444984458,.0657431866024185,-.0142603957822122,-.0700447793236106,.451524561772771,-.0979404755348809,1.59193561065049,.0171974308138372,-.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,.398540259080393,-.0864476172000634,-.4246168452719,.192898854195334,.379326641949509,1.67165845871656,-.362600483446047,-1.78103547961334,.809105213656782,3.84755914106515,-.834576462901105,-4.09930587333892,6.3113399865174,-1.3689967095037,7.98777931164405,.07249762486996,-.0157255052185161,-.0772411621484507,.0350898270822693,.434860760346183,-.0943259199045095,-.463313805821701,1.40707187748205,-.305208842367266,3.23180916099,.0113956405123911,-.00247183552086188,-.0121412600482659,.0887334889355571,-.0192472366688018,.36808995603673,.00166206328996398,-.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,.166073783576779,-.0360231684064709,-.176940031673225,.080381948458016,.158067370120622,.764836159193308,-.165901090401825,-.814879575736018,.370191004490028,2.02218309922343,-.43863300248611,-2.1544950322241,4.02885967082133,-.873902474623302,6.68010475273478,.0125887190510937,-.00273062693415816,-.0134124020091826,.0060931096085216,.0823390061849577,-.0178602054051424,-.0877264674162827,.299004079114616,-.0648571621038786,.796298871799331,.000824565926335038,-.000178857111523577,-.000878517475844129,.00697737063218123,-.00151346583454046,.0321260785532874,5.01144631634162e-005,-1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,.0319379183519753,-.00692767387221854,-.0340276241262534,.0154583827232305,.0303981920101794,.154660030302817,-.0335474040556334,-.16477947382797,.0748575381590809,.43925822557731,-.0952797768044285,-.467998998326769,.961797763577864,-.208624155307574,1.79684767953686,.000465578153648566,-.000100988848916915,-.000496041045803229,.000225346098359972,.00319092434145458,-.000692145397560786,-.00339970729838099,.0123137463285649,-.00267098242974684,.0353622090543065,5.86464816381893e-006,-1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,-1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,-1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,3.43216369764802,.744472777196595,-3.65673100407672,-1.66121346489509,3.26669916121882,-4.00974905232829,-.869757178740834,4.27210791134071,1.9407725583999,4.32546942586304,.938240282077712,-4.60848594454081,-4.4144312457315,-.957537046434959,4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,7.27823584467215,1.5787266955735,-7.75445258971519,-7.46210387346585,-1.61860962555059,7.31197439196321,9.14177937313724,1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,3.29802783242322,.715377282662343,-3.51381859652975,-1.5962898991603,3.13902998310838,-3.07100849759875,-.666134376965478,3.2719453330298,1.48640948372861,2.20598407348367,.47850138692202,-2.35032214902362,-.923194488569861,-.200250694746914,-.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873901,-.82617991094677,4.05806335333585,2.02329330850018,.438873818984882,.0162698697481229,6.45780747384475,1.40076706380652,-6.8803433906439,-4.10479708970297,-.890374108024764,1.32544341327556,7.79428043908965,1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,3.06556135691961,.66495283386875,-3.26614178293108,-1.48377299342298,2.91777071127949,-1.5947619510804,-.345920813804923,1.69910761471712,.771886268054594,-.444911232504622,-.0965059740773432,.474021882875009,2.09440936915946,.454299648071273,-2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340397,-.209134120399515,1.02723329226541,-2.01722273728899,-.437557047580714,3.46190525493913,5.18623489409548,1.1249494591216,-5.52557150711589,.205483038001612,.0445714544901277,-3.63947886955732,5.81833621896754,1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,2.74931534527504,.596355713420295,-2.92920371774242,-1.33070566358573,2.61677091290564,.106757773657142,.0231568955595295,-.113742960840148,-.0516722006304461,-2.26282036321899,-.490829782123398,2.41087725104768,2.15044627554921,.466454648534285,-.119501749587716,3.45007465931297,.74835785526233,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239337,-1.88618542328239,-3.42380239520456,-.742659122293478,1.33170919606025,3.7410698100085,.811477795596299,-3.98584891938638,3.5848146864382,.777584399208989,-3.59724489879787,3.76406389114812,.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,2.36915365304765,.513894602666682,-2.52416795342767,-1.14670228332791,2.25493680750508,1.68415540504794,.365311203903165,-1.79435010419287,-.81515390403297,-2.32831086437971,-.505035367744937,2.48065281165978,-.381749900992996,-.0828056101286271,2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.330452474967379,-2.69130383032704,2.3938769796025,.519257354978935,-2.55050893377231,4.93737708601548,1.07096955656365,.508166439915615,2.07554270972349,.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,1.94896143440088,.422750445356608,-2.07648245560297,-.94332358902539,1.85500204648401,2.83314651369184,.614539585033937,-3.01851998146924,-1.37128107917545,-.637278791777431,-.138232541833329,.678976098432196,-2.63766838605638,-.57213830168982,.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810101,4.12740556141282,.89527812555293,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843811,1.33269841309056,.289076447479325,-1.41989719504135,4.37531622836173,.949052583847877,4.52786743647191,.950542947793588,.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,1.51514018159912,.328650005712205,-1.6142761726,-.733348258592926,1.44209530674499,3.36446686788906,.729788615894315,-3.58460475610227,-1.62844728825442,1.91964817870762,.416392088662432,-2.04525122687055,-1.95173675811295,-.423352442647985,-2.75500656898115,1.04781596872157,.227282418061714,-1.11637482313293,-.507157044419472,3.55510980002441,.771141093489059,-3.78772150021559,4.5904502888745,.995717447724869,1.24891795372526,.626153181688298,.135819278812355,-.66712253696142,2.84473124098147,.617052435457252,5.48872263996487,.347192252628366,.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,1.09494796295235,.237505848402132,-1.1665906747753,-.529969564290411,1.04216054572392,3.24473228828999,.703816913228947,-3.45703591374062,-1.57049413873189,4.05017972204568,.878527020016876,-4.31518396831288,1.63294619261043,.354203381414383,-2.28560008620039,.547225517982272,.118699029853337,-.583030617100298,-.264864522602585,2.39319484717323,.519109393293246,-2.54978216897294,4.87179403320197,1.05674389554722,5.49730994822616,.236321146004053,.0512605677924316,-.251783696179213,1.3620925765208,.295452354005303,3.77635721863399,.0946962283839075,.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,.714786270724957,.155044737648518,-.761554910460561,-.345966184032582,.680326440323361,2.59853682213183,.563650249872674,-2.76855972053751,-1.25772682793124,4.7549644652192,1.03140232004452,-5.06608294899263,5.36731919542456,1.16422856843223,3.25729414076559,.233201739437629,.0505839353643847,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929027,5.89679444984458,.0657431866024185,.0142603957822122,-.0700447793236106,.451524561772771,.0979404755348809,1.59193561065049,.0171974308138372,.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,.398540259080393,.0864476172000634,-.4246168452719,-.192898854195334,.379326641949514,1.67165845871656,.362600483446047,-1.78103547961334,-.809105213656783,3.84755914106515,.834576462901105,-4.09930587333892,6.3113399865174,1.3689967095037,7.98777931164405,.07249762486996,.0157255052185161,-.0772411621484508,-.0350898270822694,.434860760346183,.0943259199045095,-.463313805821702,1.40707187748205,.305208842367266,3.23180916099,.0113956405123911,.00247183552086188,-.0121412600482659,.0887334889355571,.0192472366688018,.36808995603673,.00166206328996398,.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,.166073783576779,.0360231684064709,-.176940031673225,-.0803819484580161,.158067370120624,.764836159193308,.165901090401825,-.814879575736019,-.370191004490028,2.02218309922343,.43863300248611,-2.15449503222411,4.02885967082133,.873902474623302,6.68010475273478,.0125887190510937,.00273062693415816,-.0134124020091826,-.00609310960852161,.0823390061849577,.0178602054051424,-.0877264674162827,.299004079114616,.0648571621038786,.796298871799331,.000824565926335038,.000178857111523577,-.000878517475844129,.00697737063218123,.00151346583454046,.0321260785532874,5.01144631634162e-005,1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,.0319379183519753,.00692767387221854,-.0340276241262535,-.0154583827232306,.0303981920101797,.154660030302817,.0335474040556334,-.16477947382797,-.074857538159081,.43925822557731,.0952797768044285,-.467998998326769,.961797763577864,.208624155307574,1.79684767953686,.000465578153648566,.000100988848916915,-.00049604104580323,-.000225346098359973,.00319092434145458,.000692145397560786,-.00339970729838099,.0123137463285649,.00267098242974684,.0353622090543065,5.86464816381893e-006,1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,3.43216369764802,2.18664125535227,-2.27972534004502,-3.88041880521058,-.538361161120983,-4.00974905232829,-2.55462252942968,2.6633713691361,4.53343925473613,4.32546942586304,2.75576887572707,-2.87308040304257,-4.4144312457315,-2.81244670509951,4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,7.27823584467215,4.63698476184224,-4.83437858766287,-7.46210387346585,-4.75412760808657,7.31197439196321,9.14177937313724,5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,3.29802783242322,2.10118291403716,-2.19062908534965,-3.72876422817854,-.517320923384106,-3.07100849759875,-1.95654825118803,2.0398374063157,3.47209520918981,2.20598407348367,1.40543872850124,-1.46526746277611,-.923194488569861,-.588169835027997,-.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873901,-2.42662879416799,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,6.45780747384475,4.11428751282488,-4.28943041158219,-4.10479708970297,-2.61517790433631,1.32544341327556,7.79428043908965,4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,3.06556135691961,1.95307786118934,-2.03621928395244,-3.46593664692319,-.480856776362333,-1.5947619510804,-1.01602737631215,1.05927908845788,1.80304461367463,-.444911232504622,-.283454211851926,.295520697836744,2.09440936915946,1.33435416739534,-2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611566,-4.8496457306699,-.964148801340397,-.614261944257266,.640410728572749,-2.01722273728899,-1.28517834515454,3.46190525493913,5.18623489409548,3.30416500488362,-3.44482144542743,.205483038001612,.130913828127032,-3.63947886955732,5.81833621896754,3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,2.74931534527504,1.75159662753591,-1.82616110784379,-3.10838756746077,-.431251167473247,.106757773657142,.0680156813349527,-.0709110705135131,-.120700790879792,-2.26282036321899,-1.44164929037179,1.50301948784969,2.15044627554921,1.37005543987348,-.119501749587716,3.45007465931297,2.19805238043167,-2.2916222307293,-3.90066901628344,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,3.7410698100085,2.38344621859102,-2.48490817973374,3.5848146864382,2.28389563582347,-3.59724489879787,3.76406389114812,2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,2.36915365304765,1.50939453196029,-1.57364860569271,-2.67857514897071,-.371619894588009,1.68415540504794,1.07298020016939,-1.11865636189276,-1.90411323119761,-2.32831086437971,-1.48337347495979,1.54651984745118,-.381749900992996,-.243213947948102,2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188741,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001398,-2.69130383032704,2.3938769796025,1.52514583383108,-1.59007043171884,4.93737708601548,3.14561698874583,.508166439915615,2.07554270972349,1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,1.94896143440088,1.24168887412682,-1.29454686902578,-2.2035040478582,-.305709527061048,2.83314651369184,1.80500570449659,-1.88184377891507,-3.20316744314249,-.637278791777431,-.406012131085817,.423295838593301,-2.63766838605638,-1.68046603255685,.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141282,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843811,1.33269841309056,.849065950254408,-.885210208841136,4.37531622836173,2.78752641722509,4.52786743647191,.950542947793588,.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,1.51514018159912,.965300119862265,-1.00639240141108,-1.71302390303713,-.237661341149238,3.36446686788906,2.14351141382237,-2.23475948529486,-3.80388048506199,1.91964817870762,1.22301331540983,-1.27507630270228,-1.95173675811295,-1.24345704101385,-2.75500656898115,1.04781596872157,.66756653456346,-.69598446548217,-1.18466517033958,3.55510980002441,2.2649705675878,-2.36138908644035,4.5904502888745,2.92458893939423,1.24891795372526,.626153181688298,.398923973340749,-.41590594187282,2.84473124098147,1.81238636712449,5.48872263996487,.347192252628366,.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,1.09494796295235,.697594462028796,-.727290664744144,-1.23795280192461,-.171750973622277,3.24473228828999,2.06722817250133,-2.15522890943138,-3.66850806244403,4.05017972204568,2.58038102231816,-2.69022638759909,1.63294619261043,1.04035466481729,-2.28560008620039,.547225517982272,.348638934287124,-.363480297113413,-.618695487400893,2.39319484717323,1.52471124557473,-1.58961734284588,4.87179403320197,3.10383383936015,5.49730994822616,.236321146004053,.150560874408055,-.156970165907753,1.3620925765208,.867793055767215,3.77635721863399,.0946962283839075,.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,.714786270724957,.455392366453167,-.47477816259307,-.808140383434563,-.112119700737039,2.59853682213183,1.65553520251249,-1.7260104019959,-2.93791673259189,4.7549644652192,3.02940138707022,-3.15836129611742,5.36731919542456,3.41953432868287,3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066827,3.30495822058058,2.10559828531327,5.89679444984458,.0657431866024185,.0418851712113071,-.043668199324444,.451524561772771,.287667582930082,1.59193561065049,.0171974308138372,.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,.398540259080393,.253911132799744,-.264719986484414,-.450591303972137,-.0625140918479527,1.67165845871656,1.06501836010644,-1.11035559079056,-1.88998413076672,3.84755914106515,2.45129087378787,-2.55564094422791,6.3113399865174,4.0209726602962,7.98777931164405,.07249762486996,.0461884430503283,-.0481546590241493,-.081966121608715,.434860760346183,.277051027559316,-.288844933452007,1.40707187748205,.896449496423069,3.23180916099,.0113956405123911,.00726019498888195,-.00756925739444608,.0887334889355571,.056532358254944,.36808995603673,.00166206328996398,.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,.166073783576779,.105806079951921,-.110310185087204,-.187763722716787,-.0260499448261801,.764836159193308,.487279293067771,-.508022497640508,-.864726999689725,2.02218309922343,1.28833860494792,-1.34318245298509,4.02885967082133,2.56679795782664,6.68010475273478,.0125887190510937,.00802030872060961,-.00836172873999098,-.0142328590555999,.0823390061849577,.0524584150880736,-.0546915402140795,.299004079114616,.190496349504476,.796298871799331,.000824565926335038,.000525333297401842,-.000547696439558022,.00697737063218123,.00444530268095806,.0321260785532874,5.01144631634162e-005,3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,.0319379183519753,.0203477386368136,-.0212139303918347,-.0361091456847507,-.00500970708930326,.154660030302817,.0985343453312713,-.102728896816887,-.174859285048362,.43925822557731,.279852665057626,-.291765835225991,.961797763577864,.612764091356553,1.79684767953686,.000465578153648566,.000296621166194705,-.000309248161967434,-.000526384631617524,.00319092434145458,.00203294697585556,-.00211948838134787,.0123137463285649,.00784512281993512,.0353622090543065,5.86464816381893e-006,3.73638402898161e-006,-3.89543978958815e-006,5.19109369737154e-005,3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,3.43216369764802,3.49141841400203,.13364133030831,-3.40070220695543,-4.09773034500053,-4.00974905232829,-4.07897551472765,-.156131305216524,3.97299303311744,4.32546942586304,4.40014665135235,.168424801181427,-4.4144312457315,-4.49064435720703,4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,7.27823584467215,7.40389121595287,.283399396550339,-7.46210387346585,-7.59093364435532,7.31197439196321,9.14177937313724,9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,3.29802783242322,3.35496675520006,.128418358139767,-3.26779592010948,-3.93758279561025,-3.07100849759875,-3.12402803860063,-.119578696522845,3.04285759691611,2.20598407348367,2.24406936639925,.0858964409190686,-.923194488569861,-.939133013954983,-.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873901,-3.87460743053655,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,6.45780747384475,6.5692985292465,.251453618731739,-4.10479708970297,-4.17566451229335,1.32544341327556,7.79428043908965,7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,3.06556135691961,3.11848685368264,.119366596109979,-3.03746038662953,-3.66003632207834,-1.5947619510804,-1.62229477839744,-.0620967201893232,1.58014330533832,-.444911232504622,-.452592418790165,-.0173239199033348,2.09440936915946,2.13056837715884,-2.63186985254377,4.2894282422279,4.36348323472718,.167021432333119,-4.25010849784966,-.964148801340397,-.980794384058671,-.0375419530619799,-2.01722273728899,-2.05204915464121,3.46190525493913,5.18623489409548,5.27577283777776,.201941221845797,.205483038001612,.209030607582029,-3.63947886955732,5.81833621896754,5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,2.74931534527504,2.7967810011421,.107052632842481,-2.72411329584901,-3.28246374904245,.106757773657142,.108600897231078,.00415692611322429,-.105779161098541,-2.26282036321899,-2.30188690834251,-.0881095280677891,2.15044627554921,2.18757273479106,-.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.41844898660569,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,3.7410698100085,3.80565764777524,.145669493162303,3.5848146864382,3.64670485341608,-3.59724489879787,3.76406389114812,3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,2.36915365304765,2.41005599340123,.0922499256416731,-2.34743641804054,-2.82858093939836,1.68415540504794,1.71323156899255,.0655775157362815,-1.66871731961812,-2.32831086437971,-2.36850807266163,-.0906595922460986,-.381749900992996,-.388340636274758,2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622356,-2.53843733881177,3.58714638367868,3.64907680538233,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,2.3938769796025,2.43520615704218,.0932126006787372,4.93737708601548,5.02261861703304,.508166439915615,2.07554270972349,2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,1.94896143440088,1.98260935074574,.0758885128327108,-1.93109596018978,-2.32690486658695,2.83314651369184,2.88205947758122,.110316844590437,-2.80717601371157,-.637278791777431,-.648281114842891,-.0248143133691988,-2.63766838605638,-2.68320650944122,.808065843843811,1.73374759666501,1.76367994615688,.0675085327336519,-1.71785491622659,4.12740556141282,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843811,1.33269841309056,1.35570683401507,.0518925099589804,4.37531622836173,4.45085403872016,4.52786743647191,.950542947793588,.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,1.51514018159912,1.54129837497396,.0589964137232844,-1.50125140095791,-1.80895681150725,3.36446686788906,3.42255283189436,.131005356341752,-3.33362593315312,1.91964817870762,1.95279001506978,.0747471155162795,-1.95173675811295,-1.98543258954328,-2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063126,-1.03821099255249,3.55510980002441,3.61648712494208,.138428596350638,4.5904502888745,4.66970228947897,1.24891795372526,.626153181688298,.636963426396326,.0243811052053826,2.84473124098147,2.89384420968738,5.48872263996487,.347192252628366,.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,1.09494796295235,1.11385173231848,.042635000914322,-1.08491094310714,-1.30728073869584,3.24473228828999,3.30075109017009,.126343140340603,-3.21498892012255,4.05017972204568,4.12010419835037,.157705591526285,1.63294619261043,1.66113825253464,-2.28560008620039,.547225517982272,.556673113172381,.0213078258129714,-.542209285918406,2.39319484717323,2.43451224831939,.0931860398476285,4.87179403320197,4.95590330339709,5.49730994822616,.236321146004053,.240401121148629,.00920185489986001,1.3620925765208,1.3856084746561,3.77635721863399,.0946962283839075,.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,.714786270724957,.727126724577601,.0278322937135137,-.708234065298678,-.853397929051749,2.59853682213183,2.6433993582315,.10118163017136,-2.57471691018525,4.7549644652192,4.83705672356223,.185148446522542,5.36731919542456,5.45998347614275,3.25729414076559,.233201739437629,.237227859491313,.00908039168484082,-.231064057614914,1.22672710731592,1.24790598293054,.0477662072687193,3.30495822058058,3.36201679531997,5.89679444984458,.0657431866024185,.0668782122732,.00255990322490862,.451524561772771,.459319924425253,1.59193561065049,.0171974308138372,.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,.398540259080393,.405420872037066,.015518330446016,-.394886974518158,-.475825356015865,1.67165845871656,1.70051886866409,.0650909105855661,-1.6563349285852,3.84755914106515,3.91398546692074,.149815966677047,6.3113399865174,6.42030234909073,7.98777931164405,.07249762486996,.0737492627952914,.00282290703160026,-.0718330635538727,.434860760346183,.442368430061906,.0169325753770755,1.40707187748205,1.43136432250755,3.23180916099,.0113956405123911,.0115923809684749,.0004437225879236,.0887334889355571,.0902654315677738,.36808995603673,.00166206328996398,.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,.166073783576779,.168940970519646,.00646656841622802,-.164551441038203,-.198278882483946,.764836159193308,.778040701653472,.0297811324981311,-.757825163693694,2.02218309922343,2.05709515348012,.0787396124921617,4.02885967082133,4.09841606758444,6.68010475273478,.0125887190510937,.0128060574566477,.000490178589876777,-.0124733225023729,.0823390061849577,.0837605510092491,.00320610999159911,.299004079114616,.304166246215339,.796298871799331,.000824565926335038,.000838801675193727,3.21068856418526e-005,.00697737063218123,.00709783170793108,.0321260785532874,5.01144631634162e-005,5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,.0319379183519753,.0324893117176766,.00124359624768512,-.031645154192253,-.0381313330936757,.154660030302817,.157330164177211,.00602214055814407,-.153242313889452,.43925822557731,.446841815317497,.0171038035473811,.961797763577864,.978402756631806,1.79684767953686,.000465578153648566,.000473616144898038,1.81286469184535e-005,-.000461310355481589,.00319092434145458,.00324601417302372,.000124247970525582,.0123137463285649,.0125263374623454,.0353622090543065,5.86464816381893e-006,5.96589859842733e-006,2.28357226404389e-007,5.19109369737154e-005,5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,3.43216369764802,4.57682649415414,2.98636125886026,-.126713592860266,-3.19855198842569,-4.00974905232829,-5.34704266035319,-3.48892427303921,.148037726033597,4.32546942586304,5.76805910494959,3.76363584617939,-4.4144312457315,-5.8866906319401,4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,7.27823584467215,9.70560427278211,6.3328685571284,-7.46210387346585,-9.9507942329456,7.31197439196321,9.14177937313724,12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,3.29802783242322,4.39795490297745,2.86964825021048,-.12176137061466,-3.07354613898791,-3.07100849759875,-4.09522222779324,-2.67211637191123,.113379941906936,2.20598407348367,2.94170303097044,1.91944964063812,-.923194488569861,-1.23108958885198,-.507379758856801,4.96464287844228,6.62040364808343,4.31978730278895,-.183291880031401,-3.80884940873901,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,6.45780747384475,8.61155438605811,5.61900531555836,-4.10479708970297,-5.47379022059323,1.32544341327556,7.79428043908965,10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,3.06556135691961,4.08795840577754,2.66737675689461,-.113178836410125,-2.85690259486619,-1.5947619510804,-2.12663188500128,-1.38761892782617,.0588777326765029,-.444911232504622,-.593293821521023,-.387121881618236,2.09440936915946,2.79291698671264,-2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640684,-.964148801340397,-1.28570259705192,-.838915879897341,-2.01722273728899,-2.68998779997356,3.46190525493913,5.18623489409548,6.91589893787678,4.51259681489188,.205483038001612,.274013798814207,-3.63947886955732,5.81833621896754,7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,2.74931534527504,3.6662410134057,2.39220781955879,-.101503207887315,-2.56218200503226,.106757773657142,.142362617365695,.0928910470423957,-.00394143819036155,-2.26282036321899,-3.01749482284837,-1.96890348630228,2.15044627554921,2.86764279249033,-.119501749587716,3.45007465931297,4.60071095140733,3.00194577372104,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,3.7410698100085,4.98875605403389,3.25514367429685,3.5848146864382,4.78038820054912,-3.59724489879787,3.76406389114812,5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,2.36915365304765,3.15929138677771,2.06142518510908,-.0874678476499865,-2.20789618310877,1.68415540504794,2.24583899888368,1.46540110005888,-.0621780897499608,-2.32831086437971,-3.10482709694771,-2.02588745115913,-.381749900992996,-.50906751987416,2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273613,3.58714638367868,4.78349754217327,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,2.3938769796025,3.19226020392739,2.0829372085896,4.93737708601548,6.58404442855622,.508166439915615,2.07554270972349,2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,1.94896143440088,2.59896063091726,1.69581157410878,-.0719545824309805,-1.81630452989169,2.83314651369184,3.77803076303363,2.46515019062261,-.104598208499711,-.637278791777431,-.849817990926997,-.55450289173293,-2.63766838605638,-3.51735861632833,.808065843843811,1.73374759666501,2.31197070864004,1.50855178059943,-.0640090061407119,4.12740556141282,5.50393885468462,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843811,1.33269841309056,1.77716739172311,1.15959472274786,4.37531622836173,5.83453034569908,4.52786743647191,.950542947793588,1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,1.51514018159912,2.02045541424857,1.31833919902212,-.0559381407795162,-1.41201150863478,3.36446686788906,4.4865520602846,2.92745754610214,-.124214263271281,1.91964817870762,2.55987109498181,1.67030580640284,-1.95173675811295,-2.60266155489591,-2.75500656898115,1.04781596872157,1.39727364720762,.91171555098101,-.0386847883193888,3.55510980002441,4.74077641992772,3.0933379389912,4.5904502888745,6.12141388513832,1.24891795372526,.626153181688298,.834981872710091,.544822382907867,2.84473124098147,3.79347911889773,5.48872263996487,.347192252628366,.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,1.09494796295235,1.46012465838812,.95272558802182,-.0404248755605102,-1.02041985541771,3.24473228828999,4.32688473530227,2.82327524550585,-.11979372855449,4.05017972204568,5.4009573811209,3.52410341735765,1.63294619261043,2.17755097302669,-2.28560008620039,.547225517982272,.729731000653747,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367102,2.08234367785444,4.87179403320197,6.49658873611034,5.49730994822616,.236321146004053,.315136741011863,.205625482170815,1.3620925765208,1.81636481949506,3.77635721863399,.0946962283839075,.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,.714786270724957,.953175031760129,.621942953572118,-.0263895153231816,-.666134033494215,2.59853682213183,3.46517626442722,2.26101386266496,-.0959365479342401,4.7549644652192,6.34079527029531,4.13734393811777,5.36731919542456,7.15737676442222,3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831,1.22672710731592,1.63585353791591,1.06738798948578,3.30495822058058,4.40719665133859,5.89679444984458,.0657431866024185,.0876692328212795,.0572038290823806,.451524561772771,.602112766253273,1.59193561065049,.0171974308138372,.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,.398540259080393,.531457639388288,.346774016236294,-.0147138868003716,-.371413443660288,1.67165845871656,2.22917418911978,1.45452737733324,-.0617167093759977,3.84755914106515,5.13076069911761,3.34780114662785,6.3113399865174,8.41623844737545,7.98777931164405,.07249762486996,.0966763475074086,.0630809359431452,-.00267657237098101,.434860760346183,.579891411113904,.378376861425819,1.40707187748205,1.8763451919045,3.23180916099,.0113956405123911,.0151962068293123,.00991546509750553,.0887334889355571,.118327043544488,.36808995603673,.00166206328996398,.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,.166073783576779,.221461142188379,.144502522920425,-.00613135259583705,-.154769899538562,.764836159193308,1.01991708658492,.665491881382991,-.0282373296477252,2.02218309922343,2.69660249304008,1.75952250506473,4.02885967082133,5.37252686958436,6.68010475273478,.0125887190510937,.0167871896471047,.0109535751204636,-.000464768572193064,.0823390061849577,.109799933312678,.0716440239592633,.299004079114616,.398725093710571,.796298871799331,.000824565926335038,.00109956736063812,.000717463371565399,.00697737063218123,.00930439734224517,.0321260785532874,5.01144631634162e-005,6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,.0319379183519753,.0425895510116904,.0277895142706455,-.00117913035023065,-.0297640501007841,.154660030302817,.206240781926363,.134571297792104,-.00570996311627577,.43925822557731,.585755477154869,.382203141612094,.961797763577864,1.28256746328293,1.79684767953686,.000465578153648566,.000620853379055138,.000405104384309317,-1.71888889439253e-005,.00319092434145458,.00425513126926133,.00277645639124963,.0123137463285649,.0164205106275597,.0353622090543065,5.86464816381893e-006,7.82057018924326e-006,5.10289124286309e-006,5.19109369737154e-005,6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,3.43216369764802,5.37468949190064,5.57281420860799,4.46269573688388,2.39877236199692,-4.00974905232829,-6.27917489029831,-6.5106412373467,-5.21370528571982,4.32546942586304,6.77358573667404,7.02327732701018,-4.4144312457315,-6.91289790468793,4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,7.27823584467215,11.3975501068473,11.8176927763525,-7.46210387346585,-11.6854832203471,7.31197439196321,9.14177937313724,14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,3.29802783242322,5.16463580891203,5.35501741292456,4.28828460541245,2.30502351007756,-3.07100849759875,-4.80912874954551,-4.9864054592391,-3.99310107128126,2.20598407348367,3.45452037324197,3.58186277642042,-.923194488569861,-1.44570135775136,-.507379758856801,4.96464287844228,7.77451667983119,8.06110512843118,6.45531290809181,-3.80884940873901,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,6.45780747384475,10.1127781284249,10.4855608369157,-4.10479708970297,-6.42801793274623,1.32544341327556,7.79428043908965,12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,3.06556135691961,4.80059864950596,4.97756091846909,3.98601838486204,2.14255044478903,-1.5947619510804,-2.49736057507455,-2.58941963373659,-2.07360079279396,-.444911232504622,-.696720767591157,-.72240366631615,2.09440936915946,3.27979695088037,-2.63186985254377,4.2894282422279,6.71714607327472,6.96475715335038,5.57736018045176,-.964148801340397,-1.50983487071459,-1.56549122161564,-2.01722273728899,-3.15892446006539,3.46190525493913,5.18623489409548,8.12152468439221,8.42090472917073,.205483038001612,.321781716436834,-3.63947886955732,5.81833621896754,9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,2.74931534527504,4.30536466145163,4.46407134676942,3.5748172149004,1.92152305240544,.106757773657142,.16718022074295,.173342908623152,.138812569451129,-2.26282036321899,-3.54352469706742,-3.67414802589246,2.15044627554921,3.3675494587808,-.119501749587716,3.45007465931297,5.40273764744738,5.60189629155852,4.48598459696891,1.77035093065653,2.7723288820058,2.87452397106014,-3.42380239520456,-5.36159588761006,1.33170919606025,3.7410698100085,5.8584293670558,6.0743859656501,3.5848146864382,5.61373743608464,-3.59724489879787,3.76406389114812,5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,2.36915365304765,3.71004018615404,3.8468016980446,3.08050921776368,1.6558243734555,1.68415540504794,2.63734867013603,2.7345680458079,2.18983527963514,-2.32831086437971,-3.64608131714141,-3.78048514370594,-.381749900992996,-.597811573855878,2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635523,3.33115715545233,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,2.3938769796025,3.74875634795766,3.8869450362802,4.93737708601548,7.73181907803741,.508166439915615,2.07554270972349,3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,1.94896143440088,3.05202882623936,3.16453436679059,2.53415123836086,1.36214797290772,2.83314651369184,4.43664234772504,4.6001882618506,3.68381930162858,-.637278791777431,-.997963945475626,-1.03475143356997,-2.63766838605638,-4.13052808940267,.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141282,6.46342227832349,6.70168045166517,1.98940127224618,3.11535668501805,-2.96832239843811,1.33269841309056,2.0869750946082,2.16390630168514,4.37531622836173,6.85164468857581,4.52786743647191,.950542947793588,1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,1.51514018159912,2.37267471198348,2.46013753301878,1.97007200846434,1.05894610873647,3.36446686788906,5.26867781509235,5.46289467143736,4.37467244553793,1.91964817870762,3.0061249410246,3.1169383486426,-1.95173675811295,-3.05637492003247,-2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591415,1.36243031247302,3.55510980002441,5.56722026499826,5.7724421542407,4.5904502888745,7.18853968487789,1.24891795372526,.626153181688298,.980541495928357,1.01668674800429,2.84473124098147,4.45478376427994,5.48872263996487,.347192252628366,.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,1.09494796295235,1.7146633520688,1.77787020176478,1.42371402906152,.765269708188692,3.24473228828999,5.08117621439183,5.26848128513792,4.21898669004741,4.05017972204568,6.34248838569497,6.57628863172099,1.63294619261043,2.55715621899877,-2.28560008620039,.547225517982272,.856942588087487,.888531670338268,.711533948435091,2.39319484717323,3.74768814487269,3.88583745543936,4.87179403320197,7.62911752450547,5.49730994822616,.236321146004053,.370073484889949,.383715334269613,1.3620925765208,2.13300568008583,3.77635721863399,.0946962283839075,.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,.714786270724957,1.11933887677121,1.16060055303996,.929406031924798,.499571029238743,2.59853682213183,4.06924896099705,4.21925182103647,3.37876634868698,4.7549644652192,7.44616510000515,7.72064967553201,5.36731919542456,8.40509854032022,3.25729414076559,.233201739437629,.365188565907959,.378650344801754,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064311,5.89679444984458,.0657431866024185,.102952319693431,.106747403902078,.451524561772771,.707077089737652,1.59193561065049,.0171974308138372,.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,.398540259080393,.624104888716883,.647110981340287,.518204861963161,.278543636855161,1.67165845871656,2.61777873930361,2.71427671758157,2.17358603453451,3.84755914106515,6.02518921139496,6.24729300320554,6.3113399865174,9.88341340931671,7.98777931164405,.07249762486996,.113529614838378,.11771460504762,.0942655625163294,.434860760346183,.68098195938567,.70608468526914,1.40707187748205,2.20344223207683,3.23180916099,.0113956405123911,.0178453112143479,.0185031347484812,.0887334889355571,.138954604977606,.36808995603673,.00166206328996398,.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,.166073783576779,.260067729310804,.269654486884813,.215938641412748,.116070571566627,.764836159193308,1.1977158528687,1.24186670359493,.994483762912882,2.02218309922343,3.16669227065688,3.28342459824587,4.02885967082133,6.30910167728468,6.68010475273478,.0125887190510937,.019713644795161,.02044033985683,.0163685732441573,.0823390061849577,.12894099187304,.133694084564441,.299004079114616,.468233518150739,.796298871799331,.000824565926335038,.00129125129528224,.00133885010038651,.00697737063218123,.0109264021014796,.0321260785532874,5.01144631634162e-005,7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,.0319379183519753,.0500140463222003,.0518576912013894,.0415275099413171,.0223217196472688,.154660030302817,.24219405407586,.251121942525088,.201097826062499,.43925822557731,.687868288629403,.713224862474736,.961797763577864,1.50615319946369,1.79684767953686,.000465578153648566,.000729084691473618,.00075596060662465,.000605371370823784,.00319092434145458,.00499691420436583,.00518111316265919,.0123137463285649,.0192830438371028,.0353622090543065,5.86464816381893e-006,9.18390427744748e-006,9.52244633459443e-006,5.19109369737154e-005,8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,3.43216369764802,5.83506573884863,7.25392389703223,8.09901689119005,8.47552946493044,-4.00974905232829,-6.81702604882233,-8.4746582765388,-9.46196865399931,4.32546942586304,7.35378632026677,9.14193750783762,-4.4144312457315,-7.50503145913621,4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,7.27823584467215,12.3738225688191,15.3826488429769,-7.46210387346585,-12.6864189798666,7.31197439196321,9.14177937313724,15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,3.29802783242322,5.60701962552938,6.97042595115334,7.78249101017962,8.14428871473084,-3.07100849759875,-5.22106113152625,-6.4906175528327,-7.24678421547911,2.20598407348367,3.75042195543293,4.66237685613718,-.923194488569861,-1.5695348492102,-.507379758856801,4.96464287844228,8.44045334766827,10.4928391542811,11.7152706263275,-3.80884940873901,-6.47547397319737,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,6.45780747384475,10.9790017210618,13.6486625036235,-4.10479708970297,-6.97861844197595,1.32544341327556,7.79428043908965,13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,3.06556135691961,5.21180037431139,6.47910494479514,7.2339303710034,7.57022621762745,-1.5947619510804,-2.71127534923244,-3.37055072311448,-3.76322492874004,-.444911232504622,-.756399320675517,-.940325842964601,2.09440936915946,3.56073236366527,-2.63186985254377,4.2894282422279,7.29251224169724,9.06576398575724,10.1219390639593,-.964148801340397,-1.63916180436132,-2.03774139204767,-2.01722273728899,-3.42950637731078,3.46190525493913,5.18623489409548,8.8171847889118,10.9611768422514,.205483038001612,.349344361529711,-3.63947886955732,5.81833621896754,9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,2.74931534527504,4.67414645388234,5.81071476784009,6.48767173775814,6.78927500842464,.106757773657142,.181500267083405,.225633983149896,.251920680005669,-2.26282036321899,-3.84705006415361,-4.78250111498198,2.15044627554921,3.65600143048661,-.119501749587716,3.45007465931297,5.86551639682902,7.29177895707792,8.14128212331115,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429708,1.33170919606025,3.7410698100085,6.36024100270088,7.90680110023492,3.5848146864382,6.09458965574695,-3.59724489879787,3.76406389114812,6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,2.36915365304765,4.02782866109789,5.00723794478697,5.59058866190049,5.85048772793475,1.68415540504794,2.86325431258853,3.55948498424847,3.97417031416819,-2.32831086437971,-3.95839130925952,-4.9209161654707,-.381749900992996,-.649017927110068,2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830962,-1.52344991306454,-2.59003683420207,-2.69130383032704,2.3938769796025,4.06986110773884,5.05949102727518,4.93737708601548,8.39409842544749,.508166439915615,2.07554270972349,3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,1.94896143440088,3.31345445440235,4.11915606854152,4.59904391748785,4.81284738096772,2.83314651369184,4.81666890668704,5.98789306701478,6.68549157438517,-.637278791777431,-1.08344588754871,-1.34689725343212,-2.63766838605638,-4.48433402043889,.808065843843811,1.73374759666501,2.94756663562365,3.66429874441148,4.09119503456581,4.12740556141282,7.01705539432372,8.72332687917667,1.98940127224618,3.38220674597771,-2.96832239843811,1.33269841309056,2.26573774926528,2.81667592802333,4.37531622836173,7.43853151653138,4.52786743647191,.950542947793588,1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,1.51514018159912,2.57590935107797,3.20226904625419,3.57533818439403,3.74155092353066,3.36446686788906,5.7199734894835,7.11084574520705,7.93926992086965,1.91964817870762,3.26361861024795,4.05720211125474,-1.95173675811295,-3.31817281869637,-2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898135,3.55510980002441,6.0440879938959,7.51377213010338,4.5904502888745,7.8042837062112,1.24891795372526,.626153181688298,1.06453109468611,1.32338312764,2.84473124098147,4.83636425064192,5.48872263996487,.347192252628366,.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,1.09494796295235,1.86153514438243,2.31418717000874,2.58379343998139,2.70391057656362,3.24473228828999,5.51641118734948,6.85778510905627,7.65672734172536,4.05017972204568,6.88576274670089,8.56010903534823,1.63294619261043,2.776192869346,-2.28560008620039,.547225517982272,.930345156393644,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070527,4.87179403320197,8.28259983241364,5.49730994822616,.236321146004053,.40177262626832,.499467904136592,1.3620925765208,2.31571114636374,3.77635721863399,.0946962283839075,.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,.714786270724957,1.21521735159799,1.51071034695561,1.68671036412374,1.76512329607374,2.59853682213183,4.41780594598826,5.49204234459095,6.13187350041211,4.7549644652192,8.08397637227704,10.049681020309,5.36731919542456,9.12504854432544,3.25729414076559,.233201739437629,.396469282990197,.492874996717866,.550295671800536,1.22672710731592,2.08557456680069,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984458,.0657431866024185,.111770838907699,.138949104550329,.451524561772771,.76764272750461,1.59193561065049,.0171974308138372,.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,.398540259080393,.677563431168936,.842320170000573,.940451730878478,.984172086870924,1.67165845871656,2.84200809304711,3.53307252095549,3.94468075935688,3.84755914106515,6.54128488547504,8.13186772631248,6.3113399865174,10.7299904588883,7.98777931164405,.07249762486996,.123254146474207,.153224700235958,.1710756072316,.434860760346183,.739312383715058,.919084035475439,1.40707187748205,2.3921810358159,3.23180916099,.0113956405123911,.0193738753136992,.0240848387004708,.0887334889355571,.150856948291044,.36808995603673,.00166206328996398,.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,.166073783576779,.282344179950942,.350999163642368,.391891091702267,.410109589767539,.764836159193308,1.30030781284792,1.61649145672597,1.80481513164276,2.02218309922343,3.43793955015111,4.2739110377176,4.02885967082133,6.84951625494143,6.68010475273478,.0125887190510937,.0214022435166571,.0266064261611111,.0297061146457812,.0823390061849577,.139985605707145,.174024591276674,.299004079114616,.508340689012672,.796298871799331,.000824565926335038,.00140185515920491,.00174273111828065,.00697737063218123,.0118623177482726,.0321260785532874,5.01144631634162e-005,8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,.0319379183519753,.054298066631689,.0675012177634823,.075365210691831,.07886883956794,.154660030302817,.262939511058368,.326876043550518,.364957592022401,.43925822557731,.746788570738901,.928378137184539,.961797763577864,1.63516477461472,1.79684767953686,.000465578153648566,.000791535419924754,.000984005657525639,.00109864378994572,.00319092434145458,.00542493160168594,.00674406129057541,.0123137463285649,.0209347588552144,.0353622090543065,5.86464816381893e-006,9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.40022577929605,5.78076767221695,-7.18642267926874,8.02365168049826,-8.3966606253625,4.1644090826311,-7.0799655598807,8.80153432008931,-9.82692624602176,-4.80864541143983,8.17524004082037,-10.1631364184159,5.37622900930936,-9.14019623375092,-5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375092,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,-6.67435503772231,11.3471570254948,-14.1063386773585,7.46215578440288,-12.6865072343198,-8.17438210188192,7.8090615264033,-13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.13195404884644,5.32467544557844,-6.61942678751097,7.39059991847739,-7.73417912496331,3.83584465679206,-6.52136894437417,8.10710900955867,-9.05159934712192,-4.42925189189755,7.53022822838542,-9.36128313844683,4.95205415939119,-8.41905110415162,-5.42470353822284,3.97724364248013,-6.7617631798718,8.40595765692319,-9.38526430991047,-4.59252537503604,7.80781158453306,-9.70636382862251,5.13459946240545,-8.72939873011817,-5.62467189980939,-3.67768288095471,6.25247606762144,-7.77283197754114,4.11177446033512,-6.99048075972417,-4.5042232468252,2.21014640326355,-3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.66702109783921,4.53423694314245,-5.63678477479456,6.29347864012496,-6.58605413075654,3.26642040979696,-5.55328344227955,6.90362324406996,-7.70790568809696,-3.77173740709444,6.41237937831348,-7.97161747704437,4.21693061735794,-7.16925809522303,-4.61941604516583,2.13132645129358,-3.62349557083132,4.50458697357607,-5.02937810056321,-2.46104380066281,4.18405228663708,-5.2014490017318,2.75153061636385,-4.67791266762517,-3.01415077313091,.104423971232213,-.177532539414032,.220701460427408,-.116749549066047,.198487413238653,.127892723210502,-2.94703586087065,5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.03452907455009,3.45892910228435,-4.30000442088448,4.80096137363443,-5.0241517123509,2.49177904847469,-4.23630567890486,5.26640836144105,-5.87995282040647,-2.87725860980962,4.89166444644702,-6.08112457054741,3.21687291987534,-5.46904711383883,-3.52390772571971,.0953780462518065,-.162153445774676,.201582776983911,-.225067472325601,-.110133081351697,.187238671144273,-.232767741029294,.123132528211026,-.209339198368165,-.134884926556002,2.8024998850757,-4.76456617608066,5.92312100550576,-3.1332901246654,5.32694692824228,3.43234736114769,-3.77550789188145,6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-1.2742056900953,2.16629351671545,-2.69305077477823,3.00679522191912,-3.14657715137113,1.56057688324205,-2.65315687476095,3.2983001248078,-3.68255702755718,-1.80199896789671,3.06360167061167,-3.80854893001879,2.01469609360342,-3.42521079645606,-2.20698899398908,-1.62700976129611,2.76610026595298,-3.43870690112598,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865961,-2.10046056933792,3.57102008849817,2.30093927043685,3.19783168117628,-5.43667485806875,6.75866004623775,-3.57528450949468,6.07838727908375,3.91652795089531,-.49602746920004,.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-.433821252801768,.737545103324376,-.916887022287339,1.02370573309382,-1.07129645743706,.531320354197217,-.903304582796468,1.12295267819227,-1.25377834648449,-.613515899259544,1.04304628770802,-1.29667406213985,.68593162794344,-1.16616120174253,-.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.3691862575809,5.72799698473071,3.69075862820297,1.36899683098764,-2.32744915738298,2.89339311993396,-1.53058498738027,2.60216726588949,1.67667184761831,3.00982432351404,-5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,.433821252801768,-.737545103324376,.916887022287339,-1.02370573309382,1.07129645743706,-.531320354197217,.903304582796468,-1.12295267819227,1.25377834648449,.613515899259544,-1.04304628770802,1.29667406213985,-.68593162794344,1.16616120174253,.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.36918625758091,5.72799698473071,3.69075862820298,-1.36899683098765,2.32744915738298,-2.89339311993396,1.53058498738027,-2.60216726588949,-1.67667184761832,3.00982432351407,-5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,1.2742056900953,-2.16629351671545,2.69305077477823,-3.00679522191912,3.14657715137113,-1.56057688324205,2.65315687476095,-3.2983001248078,3.68255702755718,1.80199896789671,-3.06360167061167,3.80854893001879,-2.01469609360342,3.42521079645606,2.20698899398908,-1.62700976129612,2.76610026595298,-3.43870690112599,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865962,-2.10046056933793,3.57102008849817,2.30093927043686,-3.1978316811763,5.43667485806878,-6.75866004623778,3.5752845094947,-6.07838727908378,-3.91652795089533,-.496027469200023,.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.03452907455009,-3.45892910228435,4.30000442088448,-4.80096137363443,5.0241517123509,-2.49177904847469,4.23630567890486,-5.26640836144105,5.87995282040647,2.87725860980962,-4.89166444644702,6.08112457054741,-3.21687291987534,5.46904711383883,3.52390772571971,.0953780462518019,-.162153445774668,.201582776983901,-.225067472325591,-.110133081351691,.187238671144264,-.232767741029283,.12313252821102,-.209339198368155,-.134884926555995,-2.8024998850757,4.76456617608065,-5.92312100550575,3.13329012466539,-5.32694692824227,-3.43234736114768,-3.77550789188148,6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.66702109783921,-4.53423694314245,5.63678477479456,-6.29347864012496,6.58605413075654,-3.26642040979696,5.55328344227955,-6.90362324406996,7.70790568809696,3.77173740709444,-6.41237937831348,7.97161747704437,-4.21693061735794,7.16925809522303,4.61941604516583,2.13132645129358,-3.62349557083131,4.50458697357606,-5.0293781005632,-2.46104380066281,4.18405228663707,-5.20144900173179,2.75153061636384,-4.67791266762516,-3.0141507731309,-.104423971232209,.177532539414025,-.2207014604274,.116749549066043,-.198487413238645,-.127892723210497,-2.94703586087065,5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.13195404884644,-5.32467544557844,6.61942678751097,-7.39059991847739,7.73417912496331,-3.83584465679206,6.52136894437417,-8.10710900955867,9.05159934712192,4.42925189189755,-7.53022822838542,9.36128313844683,-4.95205415939119,8.41905110415162,5.42470353822284,3.97724364248012,-6.76176317987178,8.40595765692318,-9.38526430991045,-4.59252537503603,7.80781158453304,-9.70636382862249,5.13459946240544,-8.72939873011815,-5.62467189980938,3.67768288095471,-6.25247606762144,7.77283197754114,-4.11177446033511,6.99048075972417,4.5042232468252,2.21014640326349,-3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.40022577929605,-5.78076767221695,7.18642267926874,-8.02365168049826,8.3966606253625,-4.1644090826311,7.0799655598807,-8.80153432008931,9.82692624602176,4.80864541143983,-8.17524004082037,10.1631364184159,-5.37622900930936,9.14019623375092,5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375093,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,6.67435503772228,-11.3471570254948,14.1063386773585,-7.46215578440284,12.6865072343197,8.17438210188188,7.8090615264033,-13.2762861543698,-8.55439830449838,8.51143945889045},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.40022577929605,5.32467544557844,-5.52095651740661,4.42116822694257,-2.37645064234971,4.1644090826311,-6.52136894437417,6.7617631798718,-5.41480311178232,-4.80864541143983,7.53022822838542,-7.80781158453306,5.37622900930936,-8.41905110415162,-5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375092,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,-6.67435503772231,10.4518866380257,-10.8371697481941,7.46215578440288,-11.6855645116818,-8.17438210188192,7.8090615264033,-12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.13195404884644,4.90456807960123,-5.08536292603976,4.0723459639997,-2.18895293851099,3.83584465679206,-6.00684460241421,6.22827216283404,-4.98758483419414,-4.42925189189755,6.93610669392436,-7.19178921991208,4.95205415939119,-7.75480303503604,-5.42470353822284,3.97724364248013,-6.22827216283404,6.45786210779303,-5.17143989055336,-4.59252537503604,7.19178921991208,-7.45689685002025,5.13459946240545,-8.04066478857437,-5.62467189980939,-3.67768288095471,5.75916689254577,-5.97146442400448,4.11177446033512,-6.43894433484766,-4.5042232468252,2.21014640326355,-3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.66702109783921,4.17649376078907,-4.33044993712881,3.46781352289888,-1.86400680793392,3.26642040979696,-5.11513931437816,5.30369635131817,-4.24718682732847,-3.77173740709444,5.90645411811629,-6.12418103253549,4.21693061735794,-6.60361645843633,-4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812298,-2.77127267544969,-2.46104380066281,3.85393804714674,-3.99600399916797,2.75153061636385,-4.3088337276685,-3.01415077313091,.104423971232213,-.163525539687803,.169553506762539,-.116749549066047,.182827111459215,.127892723210502,-2.94703586087065,4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.03452907455009,3.18602578468041,-3.30347079372947,2.6454111829756,-1.42195202316674,2.49177904847469,-3.9020687402541,4.04590891241332,-3.23995377923585,-2.87725860980962,4.50572087357673,-4.67181319960176,3.21687291987534,-5.03754908153942,-3.52390772571971,.0953780462518065,-.149359829020983,.154865612009825,-.12401599635522,-.110133081351697,.172465874990319,-.178823405501172,.123132528211026,-.192822710157554,-.134884926556002,2.8024998850757,-4.38865042934355,4.55042724011644,-3.1332901246654,4.90666034634693,3.43234736114769,-3.77550789188145,5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-1.2742056900953,1.99537683408524,-2.06893149627983,1.65679518870216,-.890554665266828,1.56057688324205,-2.4438275442558,2.53391323933003,-2.02915141041228,-1.80199896789671,2.82188898006397,-2.92591098178889,2.01469609360342,-3.15496779285464,-2.20698899398908,-1.62700976129611,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294204,3.05046536107998,-2.10046056933792,3.2892729926303,2.30093927043685,3.19783168117628,-5.00773093883056,5.19232863088549,-3.57528450949468,5.59881339795158,3.91652795089531,-.49602746920004,.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-.433821252801768,.679354114255872,-.704396833771809,.564079229896515,-.303201864171254,.531320354197217,-.832035467367304,.862706409586763,-.690853143909356,-.613515899259544,.960751801781151,-.996167555656712,.68593162794344,-1.07415316937022,-.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.3933599570385,3.01349180095816,-4.71905892262039,4.89301542900031,-3.3691862575809,5.27606827091984,3.69075862820297,1.36899683098764,-2.14381758303679,2.2228441487308,-1.53058498738027,2.39686092429589,1.67667184761831,3.00982432351404,-4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,.433821252801768,-.679354114255872,.704396833771809,-.564079229896515,.303201864171254,-.531320354197217,.832035467367304,-.862706409586763,.690853143909356,.613515899259544,-.960751801781151,.996167555656712,-.68593162794344,1.07415316937022,.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.39335995703851,3.01349180095816,-4.71905892262039,4.89301542900031,-3.36918625758091,5.27606827091984,3.69075862820298,-1.36899683098765,2.1438175830368,-2.2228441487308,1.53058498738027,-2.3968609242959,-1.67667184761832,3.00982432351407,-4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,1.2742056900953,-1.99537683408524,2.06893149627983,-1.65679518870216,.890554665266828,-1.56057688324205,2.4438275442558,-2.53391323933003,2.02915141041228,1.80199896789671,-2.82188898006397,2.92591098178889,-2.01469609360342,3.15496779285464,2.20698899398908,-1.62700976129612,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294205,3.05046536107998,-2.10046056933793,3.28927299263031,2.30093927043686,-3.1978316811763,5.00773093883059,-5.19232863088552,3.5752845094947,-5.5988133979516,-3.91652795089533,-.496027469200023,.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.03452907455009,-3.18602578468041,3.30347079372947,-2.6454111829756,1.42195202316674,-2.49177904847469,3.9020687402541,-4.04590891241332,3.23995377923585,2.87725860980962,-4.50572087357673,4.67181319960176,-3.21687291987534,5.03754908153942,3.52390772571971,.0953780462518019,-.149359829020976,.154865612009818,-.124015996355214,-.110133081351691,.172465874990311,-.178823405501163,.12313252821102,-.192822710157544,-.134884926555995,-2.8024998850757,4.38865042934354,-4.55042724011643,3.13329012466539,-4.90666034634692,-3.43234736114768,-3.77550789188148,5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.66702109783921,-4.17649376078907,4.33044993712881,-3.46781352289888,1.86400680793392,-3.26642040979696,5.11513931437816,-5.30369635131817,4.24718682732847,3.77173740709444,-5.90645411811629,6.12418103253549,-4.21693061735794,6.60361645843633,4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812297,-2.77127267544968,-2.46104380066281,3.85393804714673,-3.99600399916796,2.75153061636384,-4.30883372766849,-3.0141507731309,-.104423971232209,.163525539687797,-.169553506762532,.116749549066043,-.182827111459208,-.127892723210497,-2.94703586087065,4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.13195404884644,-4.90456807960123,5.08536292603976,-4.0723459639997,2.18895293851099,-3.83584465679206,6.00684460241421,-6.22827216283404,4.98758483419414,4.42925189189755,-6.93610669392436,7.19178921991208,-4.95205415939119,7.75480303503604,5.42470353822284,3.97724364248012,-6.22827216283403,6.45786210779302,-5.17143989055335,-4.59252537503603,7.19178921991206,-7.45689685002024,5.13459946240544,-8.04066478857435,-5.62467189980938,3.67768288095471,-5.75916689254577,5.97146442400447,-4.11177446033511,6.43894433484766,4.5042232468252,2.21014640326349,-3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.40022577929605,-5.32467544557844,5.52095651740661,-4.42116822694257,2.37645064234971,-4.1644090826311,6.52136894437417,-6.7617631798718,5.41480311178232,4.80864541143983,-7.53022822838542,7.80781158453306,-5.37622900930936,8.41905110415162,5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375093,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,6.67435503772228,-10.4518866380257,10.8371697481941,-7.46215578440284,11.6855645116817,8.17438210188188,7.8090615264033,-12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.40022577929605,4.53423694314245,-2.95857174458962,-.12553446251004,3.1687879383249,4.1644090826311,-5.55328344227955,3.62349557083132,.153747689149879,-4.80864541143983,6.41237937831348,-4.18405228663708,5.37622900930936,-7.16925809522303,-5.88936380548032,5.17702934684997,-6.90362324406996,4.50458697357607,.191133071440505,-5.97791857375092,7.97161747704437,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,-6.67435503772231,8.90032284672419,-5.80742559866954,7.46215578440288,-9.95086345672774,-8.17438210188192,7.8090615264033,-10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.13195404884644,4.17649376078907,-2.72514572729006,-.115630018018827,2.91877623944935,3.83584465679206,-5.11513931437816,3.33760825329423,.141617271554667,-4.42925189189755,5.90645411811629,-3.85393804714674,4.95205415939119,-6.60361645843633,-5.42470353822284,3.97724364248013,-5.30369635131817,3.46064097812298,.146837644201687,-4.59252537503604,6.12418103253549,-3.99600399916797,5.13459946240545,-6.84704254830277,-5.62467189980939,-3.67768288095471,4.90422891550851,-3.19999004936569,4.11177446033512,-5.48309461793542,-4.5042232468252,2.21014640326355,-2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.66702109783921,3.55650076638925,-2.32060274065832,-.0984649496097572,2.4854891512059,3.26642040979696,-4.35580607412106,2.84214630515942,.120594442052505,-3.77173740709444,5.02965161709108,-3.28182786886397,4.21693061735794,-5.62332146066281,-4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898143,.0786873984258951,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636385,-3.66919510297595,-3.01415077313091,.104423971232213,-.139250467146397,.0908603812984511,-.116749549066047,.155686755269708,.127892723210502,-2.94703586087065,3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.03452907455009,2.71306598164557,-1.77026486598667,-.0751136925641364,1.89604797153805,2.49177904847469,-3.32281364706153,2.16812281562257,.0919951097438821,-2.87725860980962,3.83685470584401,-2.50353258258141,3.21687291987534,-4.28973397193188,-3.52390772571971,.0953780462518065,-.127187630774895,.0829894280920859,.00352130492348755,-.110133081351697,.146863625737388,-.0958279372639542,.123132528211026,-.164198525301527,-.134884926556002,2.8024998850757,-3.73716315870324,2.43848424018079,-3.1332901246654,4.17827543429579,3.43234736114769,-3.77550789188145,5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-1.2742056900953,1.69916672838959,-1.10869959708727,-.0470429720894781,1.18747632769106,1.56057688324205,-2.08104573641859,1.35787414544697,.0576156388045318,-1.80199896789671,2.40298463138773,-1.56793800686758,2.01469609360342,-2.68661849290085,-2.20698899398908,-1.62700976129611,2.16963467924844,-1.41567808207906,-.0600683040611749,1.87870904723461,-2.50527833232423,1.63468424297244,-2.10046056933792,2.8009863270915,2.30093927043685,3.19783168117628,-4.26434227893045,2.78246653954405,-3.57528450949468,4.76767960906116,3.91652795089531,-.49602746920004,.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-.433821252801768,.578505216668687,-.377472375086663,-.016016441651465,.404293021256913,.531320354197217,-.70852129725097,.462307355479536,.0196160547715705,-.613515899259544,.818129923100605,-.533826552292297,.68593162794344,-.914697061432424,-.751400451068813,-2.60976045384023,3.48014309433045,-2.27077966086395,-.0963509182288022,3.01349180095816,-4.01852310483628,2.62207049601641,-3.3691862575809,4.49284541622379,3.69075862820297,1.36899683098764,-1.82557171487995,1.19117835294064,-1.53058498738027,2.04105122680137,1.67667184761831,3.00982432351404,-4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,.433821252801768,-.578505216668687,.377472375086663,.016016441651465,-.404293021256913,-.531320354197217,.70852129725097,-.462307355479536,-.0196160547715705,.613515899259544,-.818129923100605,.533826552292297,-.68593162794344,.914697061432424,.751400451068813,-2.60976045384023,3.48014309433046,-2.27077966086395,-.0963509182288023,3.01349180095816,-4.01852310483628,2.62207049601641,-3.36918625758091,4.49284541622379,3.69075862820298,-1.36899683098765,1.82557171487995,-1.19117835294064,1.53058498738027,-2.04105122680137,-1.67667184761832,3.00982432351407,-4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,1.2742056900953,-1.69916672838959,1.10869959708727,.0470429720894781,-1.18747632769106,-1.56057688324205,2.08104573641859,-1.35787414544697,-.0576156388045318,1.80199896789671,-2.40298463138773,1.56793800686758,-2.01469609360342,2.68661849290085,2.20698899398908,-1.62700976129612,2.16963467924844,-1.41567808207907,-.060068304061175,1.87870904723461,-2.50527833232424,1.63468424297245,-2.10046056933793,2.8009863270915,2.30093927043686,-3.1978316811763,4.26434227893047,-2.78246653954406,3.5752845094947,-4.76767960906118,-3.91652795089533,-.496027469200023,.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.03452907455009,-2.71306598164557,1.77026486598667,.0751136925641364,-1.89604797153805,-2.49177904847469,3.32281364706153,-2.16812281562257,-.0919951097438821,2.87725860980962,-3.83685470584401,2.50353258258141,-3.21687291987534,4.28973397193188,3.52390772571971,.0953780462518019,-.127187630774889,.0829894280920819,.00352130492348739,-.110133081351691,.146863625737381,-.0958279372639496,.12313252821102,-.164198525301519,-.134884926555995,-2.8024998850757,3.73716315870323,-2.43848424018078,3.13329012466539,-4.17827543429578,-3.43234736114768,-3.77550789188148,5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.66702109783921,-3.55650076638925,2.32060274065832,.0984649496097572,-2.4854891512059,-3.26642040979696,4.35580607412106,-2.84214630515942,-.120594442052505,3.77173740709444,-5.02965161709108,3.28182786886397,-4.21693061735794,5.62332146066281,4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898142,.0786873984258949,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636384,-3.66919510297594,-3.0141507731309,-.104423971232209,.139250467146391,-.0908603812984476,.116749549066043,-.155686755269702,-.127892723210497,-2.94703586087065,3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.13195404884644,-4.17649376078907,2.72514572729006,.115630018018827,-2.91877623944935,-3.83584465679206,5.11513931437816,-3.33760825329423,-.141617271554667,4.42925189189755,-5.90645411811629,3.85393804714674,-4.95205415939119,6.60361645843633,5.42470353822284,3.97724364248012,-5.30369635131817,3.46064097812297,.146837644201687,-4.59252537503603,6.12418103253548,-3.99600399916796,5.13459946240544,-6.84704254830276,-5.62467189980938,3.67768288095471,-4.90422891550851,3.19999004936568,-4.11177446033511,5.48309461793542,4.5042232468252,2.21014640326349,-2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.40022577929605,-4.53423694314245,2.95857174458962,.12553446251004,-3.1687879383249,-4.1644090826311,5.55328344227955,-3.62349557083132,-.153747689149879,4.80864541143983,-6.41237937831348,4.18405228663708,-5.37622900930936,7.16925809522303,5.88936380548032,5.17702934684997,-6.90362324406997,4.50458697357607,.191133071440505,-5.97791857375093,7.97161747704438,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,6.67435503772228,-8.90032284672415,5.80742559866952,-7.46215578440284,9.95086345672769,8.17438210188188,7.8090615264033,-10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.40022577929605,3.45892910228435,-.132397734060631,-3.36905705276318,4.05959901190681,4.1644090826311,-4.23630567890486,.162153445774676,4.1262353470069,-4.80864541143983,4.89166444644702,-.187238671144273,5.37622900930936,-5.46904711383883,-5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375092,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,-6.67435503772231,6.78958468653045,-.259885531387107,7.46215578440288,-7.59098645151066,-8.17438210188192,7.8090615264033,-7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.13195404884644,3.18602578468041,-.121951789723545,-3.10324447907128,3.73930391312628,3.83584465679206,-3.9020687402541,.149359829020983,3.80068276060982,-4.42925189189755,4.50572087357673,-.172465874990319,4.95205415939119,-5.03754908153942,-5.42470353822284,3.97724364248013,-4.04590891241332,.154865612009825,3.94078558921029,-4.59252537503604,4.67181319960176,-.178823405501172,5.13459946240545,-5.22324594675678,-5.62467189980939,-3.67768288095471,3.74117637274605,-.143201337715524,4.11177446033512,-4.18276234400124,-4.5042232468252,2.21014640326355,-2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.66702109783921,2.71306598164557,-.103848265663968,-2.64257341211137,3.18421096606244,3.26642040979696,-3.32281364706153,.127187630774895,3.23647823392353,-3.77173740709444,3.83685470584401,-.146863625737388,4.21693061735794,-4.28973397193188,-4.61941604516583,2.13132645129358,-2.16812281562257,.0829894280920859,2.11178929936587,-2.46104380066281,2.50353258258141,-.0958279372639542,2.75153061636385,-2.79903451954298,-3.01415077313091,.104423971232213,-.106226802736416,.0040660526897138,-.116749549066047,.118765176014247,.127892723210502,-2.94703586087065,2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.03452907455009,2.0696542765645,-.0792203391289715,-2.01587923055033,2.42906581999068,2.49177904847469,-2.53479846100042,.0970247040581402,2.46893774908671,-2.87725860980962,2.92693314656882,-.112034478016931,3.21687291987534,-3.2724107413517,-3.52390772571971,.0953780462518065,-.0970247040581402,.00371382314810282,.0945037478171115,-.110133081351697,.112034478016931,-.00428835358704871,.123132528211026,-.12525835434705,-.134884926556002,2.8024998850757,-2.85088374774363,.109123528450137,-3.1332901246654,3.18738492899079,3.43234736114769,-3.77550789188145,3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-1.2742056900953,1.29620426108275,-.0496149247273535,-1.2625254749334,1.52130020070251,1.56057688324205,-1.58751952117754,.0607656246043247,1.54627160050444,-1.80199896789671,1.83310964514541,-.0701660994486395,2.01469609360342,-2.04947888880939,-2.20698899398908,-1.62700976129611,1.65509933148238,-.0633523829764578,-1.61209551065421,1.87870904723461,-1.91114408921758,.0731530307047499,-2.10046056933792,2.13672404935301,2.30093927043685,3.19783168117628,-3.25304076422429,.124516928010565,-3.57528450949468,3.63701014237694,3.91652795089531,-.49602746920004,.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-.433821252801768,.441310975771777,-.0168920991094272,-.429844559231871,.517948055079696,.531320354197217,-.540493354313145,.0206885117513158,.526449919441546,-.613515899259544,.624107966995901,-.0238890356585565,.68593162794344,-.697773919897946,-.751400451068813,-2.60976045384023,2.65481675969736,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.3691862575809,3.42735369966595,3.69075862820297,1.36899683098764,-1.39263192728723,.0533058950082319,-1.53058498738027,1.5570098290328,1.67667184761831,3.00982432351404,-3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,.433821252801768,-.441310975771777,.0168920991094272,.429844559231871,-.517948055079696,-.531320354197217,.540493354313145,-.0206885117513158,-.526449919441546,.613515899259544,-.624107966995901,.0238890356585565,-.68593162794344,.697773919897946,.751400451068813,-2.60976045384023,2.65481675969737,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.36918625758091,3.42735369966595,3.69075862820298,-1.36899683098765,1.39263192728723,-.0533058950082319,1.53058498738027,-1.5570098290328,-1.67667184761832,3.00982432351407,-3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,1.2742056900953,-1.29620426108275,.0496149247273535,1.2625254749334,-1.52130020070251,-1.56057688324205,1.58751952117754,-.0607656246043247,-1.54627160050444,1.80199896789671,-1.83310964514541,.0701660994486395,-2.01469609360342,2.04947888880939,2.20698899398908,-1.62700976129612,1.65509933148239,-.0633523829764579,-1.61209551065421,1.87870904723461,-1.91114408921759,.07315303070475,-2.10046056933793,2.13672404935301,2.30093927043686,-3.1978316811763,3.2530407642243,-.124516928010566,3.5752845094947,-3.63701014237696,-3.91652795089533,-.496027469200023,.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.03452907455009,-2.0696542765645,.0792203391289715,2.01587923055033,-2.42906581999068,-2.49177904847469,2.53479846100042,-.0970247040581402,-2.46893774908671,2.87725860980962,-2.92693314656882,.112034478016931,-3.21687291987534,3.2724107413517,3.52390772571971,.0953780462518019,-.0970247040581356,.00371382314810264,.0945037478171069,-.110133081351691,.112034478016925,-.00428835358704851,.12313252821102,-.125258354347044,-.134884926555995,-2.8024998850757,2.85088374774362,-.109123528450137,3.13329012466539,-3.18738492899078,-3.43234736114768,-3.77550789188148,3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.66702109783921,-2.71306598164557,.103848265663968,2.64257341211137,-3.18421096606244,-3.26642040979696,3.32281364706153,-.127187630774895,-3.23647823392353,3.77173740709444,-3.83685470584401,.146863625737388,-4.21693061735794,4.28973397193188,4.61941604516583,2.13132645129358,-2.16812281562256,.0829894280920858,2.11178929936586,-2.46104380066281,2.5035325825814,-.095827937263954,2.75153061636384,-2.79903451954297,-3.0141507731309,-.104423971232209,.106226802736412,-.00406605268971365,.116749549066043,-.118765176014243,-.127892723210497,-2.94703586087065,2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.13195404884644,-3.18602578468041,.121951789723545,3.10324447907128,-3.73930391312628,-3.83584465679206,3.9020687402541,-.149359829020983,-3.80068276060982,4.42925189189755,-4.50572087357673,.172465874990319,-4.95205415939119,5.03754908153942,5.42470353822284,3.97724364248012,-4.04590891241331,.154865612009825,3.94078558921028,-4.59252537503603,4.67181319960176,-.178823405501172,5.13459946240544,-5.22324594675677,-5.62467189980938,3.67768288095471,-3.74117637274605,.143201337715524,-4.11177446033511,4.18276234400124,4.5042232468252,2.21014640326349,-2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.40022577929605,-3.45892910228435,.132397734060631,3.36905705276318,-4.05959901190681,-4.1644090826311,4.23630567890486,-.162153445774676,-4.1262353470069,4.80864541143983,-4.89166444644702,.187238671144273,-5.37622900930936,5.46904711383883,5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375093,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,6.67435503772228,-6.78958468653042,.259885531387105,-7.46215578440284,7.59098645151062,8.17438210188188,7.8090615264033,-7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.40022577929605,2.16629351671545,2.25851140965318,-3.84430965952581,.533351454031698,4.1644090826311,-2.65315687476095,-2.76610026595298,4.70829853978447,-4.80864541143983,3.06360167061167,3.19401746644684,5.37622900930936,-3.42521079645606,-5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375092,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,-6.67435503772231,4.25225058180805,4.43326649059391,7.46215578440288,-4.75416068069224,-8.17438210188192,7.8090615264033,-4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.13195404884644,1.99537683408524,2.08031890026244,-3.54100050546174,.491270978557942,3.83584465679206,-2.4438275442558,-2.54785990395621,4.33682220887952,-4.42925189189755,2.82188898006397,2.94201520294204,4.95205415939119,-3.15496779285464,-5.42470353822284,3.97724364248013,-2.53391323933003,-2.64178049699053,4.49668850050967,-4.59252537503604,2.92591098178889,3.05046536107998,5.13459946240545,-3.27126792601698,-5.62467189980939,-3.67768288095471,2.34306222576705,2.44280506311383,4.11177446033512,-2.61962320701725,-4.5042232468252,2.21014640326355,-1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.66702109783921,1.69916672838959,1.77149929746803,-3.01534534295104,.418342684514395,3.26642040979696,-2.08104573641859,-2.16963467924844,3.69302874444133,-3.77173740709444,2.40298463138773,2.50527833232423,4.21693061735794,-2.68661849290085,-4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636385,-1.75300798410243,-3.01415077313091,.104423971232213,-.0665288091384509,-.0693609030179036,-.116749549066047,.0743814698720828,.127892723210502,-2.94703586087065,1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.03452907455009,1.29620426108275,1.35138294525071,-2.30024718402619,.319131466736219,2.49177904847469,-1.58751952117754,-1.65509933148238,2.81721594171208,-2.87725860980962,1.83310964514541,1.91114408921758,3.21687291987534,-2.04947888880939,-3.52390772571971,.0953780462518065,-.0607656246043247,-.0633523829764578,.107834822891095,-.110133081351697,.0701660994486395,.0731530307047499,.123132528211026,-.0784480840492681,-.134884926556002,2.8024998850757,-1.78548064936278,-1.86148755350589,-3.1332901246654,1.99622805289337,3.43234736114769,-3.77550789188145,2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-1.2742056900953,.811800069931489,.846357940948566,-1.44062234704609,.199868920965739,1.56057688324205,-.994247972331933,-1.03657254753862,1.76439483124641,-1.80199896789671,1.14805866852966,1.19693087876444,2.01469609360342,-1.28356861276539,-2.20698899398908,-1.62700976129611,1.03657254753862,1.08069885633786,-1.83950412433748,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933792,1.338209404774,2.30093927043685,3.19783168117628,-2.03734766130249,-2.1240764020783,-3.57528450949468,2.27782393297861,3.91652795089531,-.49602746920004,.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-.433821252801768,.27638875426455,.2881544676147,-.490480144821059,.0680481859118126,.531320354197217,-.338505709325782,-.352915706379786,.6007130419195,-.613515899259544,.390872724665213,.407511956177088,.68593162794344,-.437008991543009,-.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.3691862575809,2.14651815042224,3.69075862820297,1.36899683098764,-.872191775558762,-.909320487484575,-1.53058498738027,.975140050100603,1.67667184761831,3.00982432351404,-1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,.433821252801768,-.27638875426455,-.2881544676147,.490480144821059,-.0680481859118126,-.531320354197217,.338505709325782,.352915706379786,-.6007130419195,.613515899259544,-.390872724665213,-.407511956177088,-.68593162794344,.437008991543009,.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.36918625758091,2.14651815042224,3.69075862820298,-1.36899683098765,.872191775558764,.909320487484576,1.53058498738027,-.975140050100604,-1.67667184761832,3.00982432351407,-1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,1.2742056900953,-.811800069931489,-.846357940948566,1.44062234704609,-.199868920965739,-1.56057688324205,.994247972331933,1.03657254753862,-1.76439483124641,1.80199896789671,-1.14805866852966,-1.19693087876444,-2.01469609360342,1.28356861276539,2.20698899398908,-1.62700976129612,1.03657254753862,1.08069885633786,-1.83950412433749,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933793,1.338209404774,2.30093927043686,-3.1978316811763,2.0373476613025,2.12407640207831,3.5752845094947,-2.27782393297862,-3.91652795089533,-.496027469200023,.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.03452907455009,-1.29620426108275,-1.35138294525071,2.30024718402619,-.319131466736219,-2.49177904847469,1.58751952117754,1.65509933148238,-2.81721594171208,2.87725860980962,-1.83310964514541,-1.91114408921758,-3.21687291987534,2.04947888880939,3.52390772571971,.0953780462518019,-.0607656246043218,-.0633523829764547,.10783482289109,-.110133081351691,.0701660994486362,.0731530307047464,.12313252821102,-.0784480840492644,-.134884926555995,-2.8024998850757,1.78548064936277,1.86148755350588,3.13329012466539,-1.99622805289336,-3.43234736114768,-3.77550789188148,2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.66702109783921,-1.69916672838959,-1.77149929746803,3.01534534295104,-.418342684514395,-3.26642040979696,2.08104573641859,2.16963467924844,-3.69302874444133,3.77173740709444,-2.40298463138773,-2.50527833232423,-4.21693061735794,2.68661849290085,4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636384,-1.75300798410242,-3.0141507731309,-.104423971232209,.0665288091384484,.0693609030179009,.116749549066043,-.0743814698720799,-.127892723210497,-2.94703586087065,1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.13195404884644,-1.99537683408524,-2.08031890026244,3.54100050546174,-.491270978557942,-3.83584465679206,2.4438275442558,2.54785990395621,-4.33682220887952,4.42925189189755,-2.82188898006397,-2.94201520294204,-4.95205415939119,3.15496779285464,5.42470353822284,3.97724364248012,-2.53391323933002,-2.64178049699052,4.49668850050967,-4.59252537503603,2.92591098178889,3.05046536107998,5.13459946240544,-3.27126792601697,-5.62467189980938,3.67768288095471,-2.34306222576705,-2.44280506311383,-4.11177446033511,2.61962320701725,4.5042232468252,2.21014640326349,-1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.40022577929605,-2.16629351671545,-2.25851140965318,3.84430965952581,-.533351454031698,-4.1644090826311,2.65315687476095,2.76610026595298,-4.70829853978447,4.80864541143983,-3.06360167061167,-3.19401746644684,-5.37622900930936,3.42521079645606,5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375093,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,6.67435503772228,-4.25225058180803,-4.43326649059389,-7.46215578440284,4.75416068069222,8.17438210188188,7.8090615264033,-4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.40022577929605,.737545103324376,3.62270337995047,-1.64575508217185,-3.2363009692086,4.1644090826311,-.903304582796468,-4.43688738516868,2.01563009655897,-4.80864541143983,1.04304628770802,5.12327625260686,5.37622900930936,-1.16616120174253,-5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375092,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394474,-7.32142511479571,-6.67435503772231,1.44773853151531,7.11105971439921,7.46215578440288,-1.61862088558446,-8.17438210188192,7.8090615264033,-1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.13195404884644,.679354114255872,3.33687856485652,-1.51590795070228,-2.98096261298772,3.83584465679206,-.832035467367304,-4.08682490876582,1.85660048821863,-4.42925189189755,.960751801781151,4.71905892262039,4.95205415939119,-1.07415316937022,-5.42470353822284,3.97724364248013,-.862706409586763,-4.2374756641165,1.92503950203906,-4.59252537503604,.996167555656712,4.89301542900031,5.13459946240545,-1.1137491858232,-5.62467189980939,-3.67768288095471,.797728497114731,3.91831456811703,4.11177446033512,-.891887573859297,-4.5042232468252,2.21014640326355,-.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.66702109783921,.578505216668687,2.84152493765918,-1.29087413922765,-2.53844406932994,3.26642040979696,-.70852129725097,-3.48014309433045,1.58099148171138,-3.77173740709444,.818129923100605,4.01852310483628,4.21693061735794,-.914697061432424,-4.61941604516583,2.13132645129358,-.462307355479536,-2.27077966086395,1.03159071439798,-2.46104380066281,.533826552292297,2.62207049601641,2.75153061636385,-.596836229618421,-3.01415077313091,.104423971232213,-.0226506690028153,-.111256457118347,-.116749549066047,.0253242178213241,.127892723210502,-2.94703586087065,.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.03452907455009,.441310975771777,2.16764880728185,-.984739479553152,-1.93644447258226,2.49177904847469,-.540493354313145,-2.65481675969736,1.20605462730079,-2.87725860980962,.624107966995901,3.06551834185512,3.21687291987534,-.697773919897946,-3.52390772571971,.0953780462518065,-.0206885117513158,-.101618655082187,.0461642592626141,-.110133081351697,.0238890356585565,.117339115691434,.123132528211026,-.0267087538273405,-.134884926556002,2.8024998850757,-.607892005333882,-2.98586813553316,-3.1332901246654,.6796439236741,3.43234736114769,-3.77550789188145,.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-1.2742056900953,.27638875426455,1.35757727865237,-.616732719037493,-1.21277626178114,1.56057688324205,-.338505709325782,-1.66268580954775,.755340234698916,-1.80199896789671,.390872724665213,1.91990419952445,2.01469609360342,-.437008991543009,-2.20698899398908,-1.62700976129611,.352915706379786,1.73346540682961,-.78749464258244,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933792,.455612217872568,2.30093927043685,3.19783168117628,-.693643672830323,-3.40706658739604,-3.57528450949468,.775517202558345,3.91652795089531,-.49602746920004,.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-.433821252801768,.0941004396444027,.462206283002972,-.209975330432464,-.412906739739006,.531320354197217,-.115249030860378,-.566084774633034,.257166209078973,-.613515899259544,.133078117935277,.653658394065701,.68593162794344,-.148785859041837,-.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.3691862575809,.730812301668709,3.69075862820297,1.36899683098764,-.296949960039812,-1.45857062726747,-1.53058498738027,.332000148390628,1.67667184761831,3.00982432351404,-.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,.433821252801768,-.0941004396444027,-.462206283002972,.209975330432464,.412906739739006,-.531320354197217,.115249030860378,.566084774633034,-.257166209078973,.613515899259544,-.133078117935277,-.653658394065701,-.68593162794344,.148785859041837,.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.36918625758091,.730812301668709,3.69075862820298,-1.36899683098765,.296949960039813,1.45857062726747,1.53058498738027,-.332000148390629,-1.67667184761832,3.00982432351407,-.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,1.2742056900953,-.27638875426455,-1.35757727865237,.616732719037493,1.21277626178114,-1.56057688324205,.338505709325782,1.66268580954775,-.755340234698916,1.80199896789671,-.390872724665213,-1.91990419952445,-2.01469609360342,.437008991543009,2.20698899398908,-1.62700976129612,.352915706379786,1.73346540682961,-.787494642582441,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933793,.455612217872569,2.30093927043686,-3.1978316811763,.693643672830326,3.40706658739606,3.5752845094947,-.775517202558348,-3.91652795089533,-.496027469200023,.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.03452907455009,-.441310975771777,-2.16764880728185,.984739479553152,1.93644447258226,-2.49177904847469,.540493354313145,2.65481675969736,-1.20605462730079,2.87725860980962,-.624107966995901,-3.06551834185512,-3.21687291987534,.697773919897946,3.52390772571971,.0953780462518019,-.0206885117513148,-.101618655082182,.0461642592626119,-.110133081351691,.0238890356585554,.117339115691428,.12313252821102,-.0267087538273392,-.134884926555995,-2.8024998850757,.607892005333881,2.98586813553315,3.13329012466539,-.679643923674099,-3.43234736114768,-3.77550789188148,.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.66702109783921,-.578505216668687,-2.84152493765918,1.29087413922765,2.53844406932994,-3.26642040979696,.70852129725097,3.48014309433045,-1.58099148171138,3.77173740709444,-.818129923100605,-4.01852310483628,-4.21693061735794,.914697061432424,4.61941604516583,2.13132645129358,-.462307355479535,-2.27077966086395,1.03159071439797,-2.46104380066281,.533826552292296,2.62207049601641,2.75153061636384,-.59683622961842,-3.0141507731309,-.104423971232209,.0226506690028144,.111256457118342,.116749549066043,-.0253242178213231,-.127892723210497,-2.94703586087065,.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.13195404884644,-.679354114255872,-3.33687856485652,1.51590795070228,2.98096261298772,-3.83584465679206,.832035467367304,4.08682490876582,-1.85660048821863,4.42925189189755,-.960751801781151,-4.71905892262039,-4.95205415939119,1.07415316937022,5.42470353822284,3.97724364248012,-.862706409586761,-4.23747566411649,1.92503950203906,-4.59252537503603,.996167555656711,4.8930154290003,5.13459946240544,-1.1137491858232,-5.62467189980938,3.67768288095471,-.797728497114731,-3.91831456811703,-4.11177446033511,.891887573859297,4.5042232468252,2.21014640326349,-.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.40022577929605,-.737545103324376,-3.62270337995047,1.64575508217185,3.2363009692086,-4.1644090826311,.903304582796468,4.43688738516868,-2.01563009655897,4.80864541143983,-1.04304628770802,-5.12327625260686,-5.37622900930936,1.16616120174253,5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375093,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394475,-7.32142511479571,6.67435503772228,-1.4477385315153,-7.11105971439917,-7.46215578440284,1.61862088558445,8.17438210188188,7.8090615264033,-1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.40022577929605,-.737545103324376,3.62270337995047,1.64575508217186,-3.23630096920864,4.1644090826311,.903304582796468,-4.43688738516868,-2.01563009655898,-4.80864541143983,-1.04304628770802,5.12327625260686,5.37622900930936,1.16616120174253,-5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575887,-5.97791857375092,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394474,-7.32142511479571,-6.67435503772231,-1.44773853151531,7.11105971439921,7.46215578440288,1.61862088558446,-8.17438210188192,7.8090615264033,1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.13195404884644,-.679354114255872,3.33687856485653,1.51590795070228,-2.98096261298775,3.83584465679206,.832035467367304,-4.08682490876582,-1.85660048821863,-4.42925189189755,-.960751801781151,4.71905892262039,4.95205415939119,1.07415316937022,-5.42470353822284,3.97724364248013,.862706409586763,-4.2374756641165,-1.92503950203907,-4.59252537503604,-.996167555656712,4.89301542900031,5.13459946240545,1.1137491858232,-5.62467189980939,-3.67768288095471,-.797728497114731,3.91831456811704,4.11177446033512,.891887573859297,-4.5042232468252,2.21014640326355,.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.66702109783921,-.578505216668687,2.84152493765918,1.29087413922765,-2.53844406932997,3.26642040979696,.70852129725097,-3.48014309433046,-1.58099148171138,-3.77173740709444,-.818129923100605,4.01852310483628,4.21693061735794,.914697061432424,-4.61941604516583,2.13132645129358,.462307355479536,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292297,2.62207049601641,2.75153061636385,.596836229618421,-3.01415077313091,.104423971232213,.0226506690028153,-.111256457118347,-.116749549066047,-.0253242178213241,.127892723210502,-2.94703586087065,-.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.03452907455009,-.441310975771777,2.16764880728185,.984739479553153,-1.93644447258228,2.49177904847469,.540493354313145,-2.65481675969737,-1.20605462730079,-2.87725860980962,-.624107966995901,3.06551834185512,3.21687291987534,.697773919897946,-3.52390772571971,.0953780462518065,.0206885117513158,-.101618655082187,-.0461642592626142,-.110133081351697,-.0238890356585565,.117339115691434,.123132528211026,.0267087538273405,-.134884926556002,2.8024998850757,.607892005333882,-2.98586813553316,-3.1332901246654,-.6796439236741,3.43234736114769,-3.77550789188145,-.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-1.2742056900953,-.27638875426455,1.35757727865237,.616732719037494,-1.21277626178116,1.56057688324205,.338505709325782,-1.66268580954775,-.755340234698916,-1.80199896789671,-.390872724665213,1.91990419952445,2.01469609360342,.437008991543009,-2.20698899398908,-1.62700976129611,-.352915706379786,1.73346540682961,.787494642582441,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933792,-.455612217872568,2.30093927043685,3.19783168117628,.693643672830323,-3.40706658739604,-3.57528450949468,-.775517202558345,3.91652795089531,-.49602746920004,-.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-.433821252801768,-.0941004396444027,.462206283002972,.209975330432464,-.41290673973901,.531320354197217,.115249030860378,-.566084774633034,-.257166209078973,-.613515899259544,-.133078117935277,.653658394065701,.68593162794344,.148785859041837,-.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353978,-3.3691862575809,-.730812301668709,3.69075862820297,1.36899683098764,.296949960039812,-1.45857062726747,-1.53058498738027,-.332000148390628,1.67667184761831,3.00982432351404,.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,.433821252801768,.0941004396444027,-.462206283002972,-.209975330432464,.41290673973901,-.531320354197217,-.115249030860378,.566084774633034,.257166209078973,.613515899259544,.133078117935277,-.653658394065701,-.68593162794344,-.148785859041837,.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353979,-3.36918625758091,-.730812301668709,3.69075862820298,-1.36899683098765,-.296949960039813,1.45857062726747,1.53058498738027,.332000148390629,-1.67667184761832,3.00982432351407,.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,1.2742056900953,.27638875426455,-1.35757727865237,-.616732719037494,1.21277626178116,-1.56057688324205,-.338505709325782,1.66268580954775,.755340234698916,1.80199896789671,.390872724665213,-1.91990419952445,-2.01469609360342,-.437008991543009,2.20698899398908,-1.62700976129612,-.352915706379786,1.73346540682961,.787494642582442,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933793,-.455612217872569,2.30093927043686,-3.1978316811763,-.693643672830326,3.40706658739606,3.5752845094947,.775517202558348,-3.91652795089533,-.496027469200023,-.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.03452907455009,.441310975771777,-2.16764880728185,-.984739479553153,1.93644447258228,-2.49177904847469,-.540493354313145,2.65481675969737,1.20605462730079,2.87725860980962,.624107966995901,-3.06551834185512,-3.21687291987534,-.697773919897946,3.52390772571971,.0953780462518019,.0206885117513148,-.101618655082182,-.046164259262612,-.110133081351691,-.0238890356585554,.117339115691428,.12313252821102,.0267087538273392,-.134884926555995,-2.8024998850757,-.607892005333881,2.98586813553315,3.13329012466539,.679643923674099,-3.43234736114768,-3.77550789188148,-.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.66702109783921,.578505216668687,-2.84152493765918,-1.29087413922765,2.53844406932997,-3.26642040979696,-.70852129725097,3.48014309433046,1.58099148171138,3.77173740709444,.818129923100605,-4.01852310483628,-4.21693061735794,-.914697061432424,4.61941604516583,2.13132645129358,.462307355479535,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292296,2.62207049601641,2.75153061636384,.59683622961842,-3.0141507731309,-.104423971232209,-.0226506690028144,.111256457118342,.116749549066043,.0253242178213231,-.127892723210497,-2.94703586087065,-.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.13195404884644,.679354114255872,-3.33687856485653,-1.51590795070228,2.98096261298775,-3.83584465679206,-.832035467367304,4.08682490876582,1.85660048821863,4.42925189189755,.960751801781151,-4.71905892262039,-4.95205415939119,-1.07415316937022,5.42470353822284,3.97724364248012,.862706409586761,-4.23747566411649,-1.92503950203906,-4.59252537503603,-.996167555656711,4.89301542900031,5.13459946240544,1.1137491858232,-5.62467189980938,3.67768288095471,.797728497114731,-3.91831456811703,-4.11177446033511,-.891887573859297,4.5042232468252,2.21014640326349,.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.40022577929605,.737545103324376,-3.62270337995047,-1.64575508217186,3.23630096920864,-4.1644090826311,-.903304582796468,4.43688738516868,2.01563009655898,4.80864541143983,1.04304628770802,-5.12327625260686,-5.37622900930936,-1.16616120174253,5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575888,-5.97791857375093,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394475,-7.32142511479571,6.67435503772228,1.4477385315153,-7.11105971439918,-7.46215578440284,-1.61862088558445,8.17438210188188,7.8090615264033,1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.40022577929605,-2.16629351671545,2.25851140965319,3.84430965952583,.533351454031679,4.1644090826311,2.65315687476095,-2.76610026595298,-4.70829853978449,-4.80864541143983,-3.06360167061167,3.19401746644684,5.37622900930936,3.42521079645606,-5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375092,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,-6.67435503772231,-4.25225058180805,4.43326649059392,7.46215578440288,4.75416068069224,-8.17438210188192,7.8090615264033,4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.13195404884644,-1.99537683408524,2.08031890026245,3.54100050546176,.491270978557926,3.83584465679206,2.4438275442558,-2.54785990395621,-4.33682220887954,-4.42925189189755,-2.82188898006397,2.94201520294205,4.95205415939119,3.15496779285464,-5.42470353822284,3.97724364248013,2.53391323933003,-2.64178049699053,-4.4966885005097,-4.59252537503604,-2.92591098178889,3.05046536107998,5.13459946240545,3.27126792601698,-5.62467189980939,-3.67768288095471,-2.34306222576705,2.44280506311384,4.11177446033512,2.61962320701725,-4.5042232468252,2.21014640326355,1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.66702109783921,-1.69916672838959,1.77149929746803,3.01534534295106,.418342684514381,3.26642040979696,2.08104573641859,-2.16963467924844,-3.69302874444135,-3.77173740709444,-2.40298463138773,2.50527833232424,4.21693061735794,2.68661849290085,-4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207907,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297245,2.75153061636385,1.75300798410243,-3.01415077313091,.104423971232213,.0665288091384509,-.0693609030179037,-.116749549066047,-.0743814698720828,.127892723210502,-2.94703586087065,-1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.03452907455009,-1.29620426108275,1.35138294525072,2.3002471840262,.319131466736208,2.49177904847469,1.58751952117754,-1.65509933148239,-2.8172159417121,-2.87725860980962,-1.83310964514541,1.91114408921759,3.21687291987534,2.04947888880939,-3.52390772571971,.0953780462518065,.0607656246043247,-.0633523829764579,-.107834822891096,-.110133081351697,-.0701660994486395,.07315303070475,.123132528211026,.0784480840492681,-.134884926556002,2.8024998850757,1.78548064936278,-1.86148755350589,-3.1332901246654,-1.99622805289337,3.43234736114769,-3.77550789188145,-2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-1.2742056900953,-.811800069931489,.846357940948568,1.4406223470461,.199868920965732,1.56057688324205,.994247972331933,-1.03657254753862,-1.76439483124642,-1.80199896789671,-1.14805866852966,1.19693087876444,2.01469609360342,1.28356861276539,-2.20698899398908,-1.62700976129611,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.24788355079949,-2.10046056933792,-1.338209404774,2.30093927043685,3.19783168117628,2.03734766130249,-2.12407640207831,-3.57528450949468,-2.27782393297861,3.91652795089531,-.49602746920004,-.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-.433821252801768,-.27638875426455,.2881544676147,.490480144821061,.0680481859118103,.531320354197217,.338505709325782,-.352915706379786,-.600713041919503,-.613515899259544,-.390872724665213,.407511956177088,.68593162794344,.437008991543009,-.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.3691862575809,-2.14651815042224,3.69075862820297,1.36899683098764,.872191775558762,-.909320487484577,-1.53058498738027,-.975140050100603,1.67667184761831,3.00982432351404,1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,.433821252801768,.27638875426455,-.2881544676147,-.490480144821061,-.0680481859118103,-.531320354197217,-.338505709325782,.352915706379786,.600713041919503,.613515899259544,.390872724665213,-.407511956177088,-.68593162794344,-.437008991543009,.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.36918625758091,-2.14651815042224,3.69075862820298,-1.36899683098765,-.872191775558764,.909320487484578,1.53058498738027,.975140050100604,-1.67667184761832,3.00982432351407,1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,1.2742056900953,.811800069931489,-.846357940948568,-1.4406223470461,-.199868920965732,-1.56057688324205,-.994247972331933,1.03657254753862,1.76439483124642,1.80199896789671,1.14805866852966,-1.19693087876444,-2.01469609360342,-1.28356861276539,2.20698899398908,-1.62700976129612,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.2478835507995,-2.10046056933793,-1.338209404774,2.30093927043686,-3.1978316811763,-2.0373476613025,2.12407640207832,3.5752845094947,2.27782393297862,-3.91652795089533,-.496027469200023,-.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.03452907455009,1.29620426108275,-1.35138294525072,-2.3002471840262,-.319131466736208,-2.49177904847469,-1.58751952117754,1.65509933148239,2.8172159417121,2.87725860980962,1.83310964514541,-1.91114408921759,-3.21687291987534,-2.04947888880939,3.52390772571971,.0953780462518019,.0607656246043218,-.0633523829764548,-.10783482289109,-.110133081351691,-.0701660994486362,.0731530307047465,.12313252821102,.0784480840492644,-.134884926555995,-2.8024998850757,-1.78548064936277,1.86148755350589,3.13329012466539,1.99622805289336,-3.43234736114768,-3.77550789188148,-2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.66702109783921,1.69916672838959,-1.77149929746803,-3.01534534295106,-.418342684514381,-3.26642040979696,-2.08104573641859,2.16963467924844,3.69302874444135,3.77173740709444,2.40298463138773,-2.50527833232424,-4.21693061735794,-2.68661849290085,4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207906,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297244,2.75153061636384,1.75300798410242,-3.0141507731309,-.104423971232209,-.0665288091384484,.069360903017901,.116749549066043,.0743814698720799,-.127892723210497,-2.94703586087065,-1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.13195404884644,1.99537683408524,-2.08031890026245,-3.54100050546176,-.491270978557926,-3.83584465679206,-2.4438275442558,2.54785990395621,4.33682220887954,4.42925189189755,2.82188898006397,-2.94201520294205,-4.95205415939119,-3.15496779285464,5.42470353822284,3.97724364248012,2.53391323933002,-2.64178049699052,-4.49668850050969,-4.59252537503603,-2.92591098178889,3.05046536107998,5.13459946240544,3.27126792601697,-5.62467189980938,3.67768288095471,2.34306222576705,-2.44280506311384,-4.11177446033511,-2.61962320701725,4.5042232468252,2.21014640326349,1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.40022577929605,2.16629351671545,-2.25851140965319,-3.84430965952583,-.533351454031679,-4.1644090826311,-2.65315687476095,2.76610026595298,4.70829853978449,4.80864541143983,3.06360167061167,-3.19401746644684,-5.37622900930936,-3.42521079645606,5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375093,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,6.67435503772228,4.25225058180803,-4.4332664905939,-7.46215578440284,-4.75416068069222,8.17438210188188,7.8090615264033,4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.40022577929605,-3.45892910228435,-.132397734060625,3.36905705276318,4.05959901190685,4.1644090826311,4.23630567890486,.162153445774668,-4.1262353470069,-4.80864541143983,-4.89166444644702,-.187238671144264,5.37622900930936,5.46904711383883,-5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375092,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,-6.67435503772231,-6.78958468653045,-.259885531387094,7.46215578440288,7.59098645151066,-8.17438210188192,7.8090615264033,7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.13195404884644,-3.18602578468041,-.121951789723539,3.10324447907128,3.73930391312631,3.83584465679206,3.9020687402541,.149359829020976,-3.80068276060981,-4.42925189189755,-4.50572087357673,-.172465874990311,4.95205415939119,5.03754908153942,-5.42470353822284,3.97724364248013,4.04590891241332,.154865612009818,-3.94078558921028,-4.59252537503604,-4.67181319960176,-.178823405501163,5.13459946240545,5.22324594675678,-5.62467189980939,-3.67768288095471,-3.74117637274605,-.143201337715517,4.11177446033512,4.18276234400124,-4.5042232468252,2.21014640326355,2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.66702109783921,-2.71306598164557,-.103848265663963,2.64257341211137,3.18421096606247,3.26642040979696,3.32281364706153,.127187630774889,-3.23647823392352,-3.77173740709444,-3.83685470584401,-.146863625737381,4.21693061735794,4.28973397193188,-4.61941604516583,2.13132645129358,2.16812281562257,.0829894280920819,-2.11178929936586,-2.46104380066281,-2.50353258258141,-.0958279372639496,2.75153061636385,2.79903451954298,-3.01415077313091,.104423971232213,.106226802736416,.00406605268971361,-.116749549066047,-.118765176014247,.127892723210502,-2.94703586087065,-2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.03452907455009,-2.0696542765645,-.0792203391289677,2.01587923055033,2.4290658199907,2.49177904847469,2.53479846100042,.0970247040581356,-2.46893774908671,-2.87725860980962,-2.92693314656882,-.112034478016925,3.21687291987534,3.2724107413517,-3.52390772571971,.0953780462518065,.0970247040581402,.00371382314810264,-.0945037478171113,-.110133081351697,-.112034478016931,-.00428835358704851,.123132528211026,.12525835434705,-.134884926556002,2.8024998850757,2.85088374774363,.109123528450132,-3.1332901246654,-3.18738492899079,3.43234736114769,-3.77550789188145,-3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-1.2742056900953,-1.29620426108275,-.0496149247273511,1.2625254749334,1.52130020070252,1.56057688324205,1.58751952117754,.0607656246043218,-1.54627160050443,-1.80199896789671,-1.83310964514541,-.0701660994486362,2.01469609360342,2.04947888880939,-2.20698899398908,-1.62700976129611,-1.65509933148238,-.0633523829764547,1.6120955106542,1.87870904723461,1.91114408921758,.0731530307047464,-2.10046056933792,-2.13672404935301,2.30093927043685,3.19783168117628,3.25304076422429,.124516928010559,-3.57528450949468,-3.63701014237694,3.91652795089531,-.49602746920004,-.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-.433821252801768,-.441310975771777,-.0168920991094264,.42984455923187,.5179480550797,.531320354197217,.540493354313145,.0206885117513148,-.526449919441544,-.613515899259544,-.624107966995901,-.0238890356585554,.68593162794344,.697773919897946,-.751400451068813,-2.60976045384023,-2.65481675969736,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.3691862575809,-3.42735369966595,3.69075862820297,1.36899683098764,1.39263192728723,.0533058950082293,-1.53058498738027,-1.5570098290328,1.67667184761831,3.00982432351404,3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,.433821252801768,.441310975771777,.0168920991094264,-.42984455923187,-.5179480550797,-.531320354197217,-.540493354313145,-.0206885117513148,.526449919441544,.613515899259544,.624107966995901,.0238890356585554,-.68593162794344,-.697773919897946,.751400451068813,-2.60976045384023,-2.65481675969737,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.36918625758091,-3.42735369966595,3.69075862820298,-1.36899683098765,-1.39263192728723,-.0533058950082294,1.53058498738027,1.5570098290328,-1.67667184761832,3.00982432351407,3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,1.2742056900953,1.29620426108275,.0496149247273511,-1.2625254749334,-1.52130020070252,-1.56057688324205,-1.58751952117754,-.0607656246043218,1.54627160050443,1.80199896789671,1.83310964514541,.0701660994486362,-2.01469609360342,-2.04947888880939,2.20698899398908,-1.62700976129612,-1.65509933148239,-.0633523829764548,1.61209551065421,1.87870904723461,1.91114408921759,.0731530307047465,-2.10046056933793,-2.13672404935301,2.30093927043686,-3.1978316811763,-3.2530407642243,-.12451692801056,3.5752845094947,3.63701014237696,-3.91652795089533,-.496027469200023,-.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.03452907455009,2.0696542765645,.0792203391289677,-2.01587923055033,-2.4290658199907,-2.49177904847469,-2.53479846100042,-.0970247040581356,2.46893774908671,2.87725860980962,2.92693314656882,.112034478016925,-3.21687291987534,-3.2724107413517,3.52390772571971,.0953780462518019,.0970247040581356,.00371382314810246,-.0945037478171067,-.110133081351691,-.112034478016925,-.0042883535870483,.12313252821102,.125258354347044,-.134884926555995,-2.8024998850757,-2.85088374774362,-.109123528450131,3.13329012466539,3.18738492899078,-3.43234736114768,-3.77550789188148,-3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.66702109783921,2.71306598164557,.103848265663963,-2.64257341211137,-3.18421096606247,-3.26642040979696,-3.32281364706153,-.127187630774889,3.23647823392352,3.77173740709444,3.83685470584401,.146863625737381,-4.21693061735794,-4.28973397193188,4.61941604516583,2.13132645129358,2.16812281562256,.0829894280920818,-2.11178929936586,-2.46104380066281,-2.5035325825814,-.0958279372639495,2.75153061636384,2.79903451954297,-3.0141507731309,-.104423971232209,-.106226802736412,-.00406605268971345,.116749549066043,.118765176014243,-.127892723210497,-2.94703586087065,-2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.13195404884644,3.18602578468041,.121951789723539,-3.10324447907128,-3.73930391312631,-3.83584465679206,-3.9020687402541,-.149359829020976,3.80068276060981,4.42925189189755,4.50572087357673,.172465874990311,-4.95205415939119,-5.03754908153942,5.42470353822284,3.97724364248012,4.04590891241331,.154865612009818,-3.94078558921027,-4.59252537503603,-4.67181319960176,-.178823405501163,5.13459946240544,5.22324594675677,-5.62467189980938,3.67768288095471,3.74117637274605,.143201337715517,-4.11177446033511,-4.18276234400124,4.5042232468252,2.21014640326349,2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.40022577929605,3.45892910228435,.132397734060625,-3.36905705276318,-4.05959901190685,-4.1644090826311,-4.23630567890486,-.162153445774668,4.1262353470069,4.80864541143983,4.89166444644702,.187238671144264,-5.37622900930936,-5.46904711383883,5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375093,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,6.67435503772228,6.78958468653042,.259885531387093,-7.46215578440284,-7.59098645151062,8.17438210188188,7.8090615264033,7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.40022577929605,-4.53423694314245,-2.95857174458961,.125534462510035,3.16878793832491,4.1644090826311,5.55328344227955,3.62349557083131,-.153747689149873,-4.80864541143983,-6.41237937831348,-4.18405228663707,5.37622900930936,7.16925809522303,-5.88936380548032,5.17702934684997,6.90362324406996,4.50458697357606,-.191133071440498,-5.97791857375092,-7.97161747704437,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,-6.67435503772231,-8.90032284672419,-5.80742559866953,7.46215578440288,9.95086345672774,-8.17438210188192,7.8090615264033,10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.13195404884644,-4.17649376078907,-2.72514572729005,.115630018018823,2.91877623944935,3.83584465679206,5.11513931437816,3.33760825329423,-.141617271554661,-4.42925189189755,-5.90645411811629,-3.85393804714673,4.95205415939119,6.60361645843633,-5.42470353822284,3.97724364248013,5.30369635131817,3.46064097812297,-.146837644201682,-4.59252537503604,-6.12418103253549,-3.99600399916796,5.13459946240545,6.84704254830277,-5.62467189980939,-3.67768288095471,-4.90422891550851,-3.19999004936568,4.11177446033512,5.48309461793542,-4.5042232468252,2.21014640326355,2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.66702109783921,-3.55650076638925,-2.32060274065832,.0984649496097535,2.4854891512059,3.26642040979696,4.35580607412106,2.84214630515942,-.120594442052501,-3.77173740709444,-5.02965161709108,-3.28182786886397,4.21693061735794,5.62332146066281,-4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258921,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636385,3.66919510297595,-3.01415077313091,.104423971232213,.139250467146397,.0908603812984509,-.116749549066047,-.155686755269708,.127892723210502,-2.94703586087065,-3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.03452907455009,-2.71306598164557,-1.77026486598667,.0751136925641335,1.89604797153805,2.49177904847469,3.32281364706153,2.16812281562256,-.0919951097438786,-2.87725860980962,-3.83685470584401,-2.5035325825814,3.21687291987534,4.28973397193188,-3.52390772571971,.0953780462518065,.127187630774895,.0829894280920858,-.00352130492348742,-.110133081351697,-.146863625737388,-.095827937263954,.123132528211026,.164198525301527,-.134884926556002,2.8024998850757,3.73716315870324,2.43848424018078,-3.1332901246654,-4.17827543429579,3.43234736114769,-3.77550789188145,-5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-1.2742056900953,-1.69916672838959,-1.10869959708726,.0470429720894763,1.18747632769106,1.56057688324205,2.08104573641859,1.35787414544697,-.0576156388045296,-1.80199896789671,-2.40298463138773,-1.56793800686758,2.01469609360342,2.68661849290085,-2.20698899398908,-1.62700976129611,-2.16963467924844,-1.41567808207906,.0600683040611726,1.87870904723461,2.50527833232423,1.63468424297244,-2.10046056933792,-2.8009863270915,2.30093927043685,3.19783168117628,4.26434227893045,2.78246653954404,-3.57528450949468,-4.76767960906116,3.91652795089531,-.49602746920004,-.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-.433821252801768,-.578505216668687,-.377472375086662,.0160164416514644,.404293021256913,.531320354197217,.70852129725097,.462307355479535,-.0196160547715697,-.613515899259544,-.818129923100605,-.533826552292296,.68593162794344,.914697061432424,-.751400451068813,-2.60976045384023,-3.48014309433045,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.3691862575809,-4.49284541622379,3.69075862820297,1.36899683098764,1.82557171487995,1.19117835294064,-1.53058498738027,-2.04105122680137,1.67667184761831,3.00982432351404,4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,.433821252801768,.578505216668687,.377472375086662,-.0160164416514644,-.404293021256913,-.531320354197217,-.70852129725097,-.462307355479535,.0196160547715697,.613515899259544,.818129923100605,.533826552292296,-.68593162794344,-.914697061432424,.751400451068813,-2.60976045384023,-3.48014309433046,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.36918625758091,-4.49284541622379,3.69075862820298,-1.36899683098765,-1.82557171487995,-1.19117835294064,1.53058498738027,2.04105122680137,-1.67667184761832,3.00982432351407,4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,1.2742056900953,1.69916672838959,1.10869959708726,-.0470429720894763,-1.18747632769106,-1.56057688324205,-2.08104573641859,-1.35787414544697,.0576156388045296,1.80199896789671,2.40298463138773,1.56793800686758,-2.01469609360342,-2.68661849290085,2.20698899398908,-1.62700976129612,-2.16963467924844,-1.41567808207906,.0600683040611727,1.87870904723461,2.50527833232424,1.63468424297244,-2.10046056933793,-2.8009863270915,2.30093927043686,-3.1978316811763,-4.26434227893047,-2.78246653954406,3.5752845094947,4.76767960906118,-3.91652795089533,-.496027469200023,-.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.03452907455009,2.71306598164557,1.77026486598667,-.0751136925641335,-1.89604797153805,-2.49177904847469,-3.32281364706153,-2.16812281562256,.0919951097438786,2.87725860980962,3.83685470584401,2.5035325825814,-3.21687291987534,-4.28973397193188,3.52390772571971,.0953780462518019,.127187630774889,.0829894280920818,-.00352130492348725,-.110133081351691,-.146863625737381,-.0958279372639495,.12313252821102,.164198525301519,-.134884926555995,-2.8024998850757,-3.73716315870323,-2.43848424018078,3.13329012466539,4.17827543429578,-3.43234736114768,-3.77550789188148,-5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.66702109783921,3.55650076638925,2.32060274065832,-.0984649496097535,-2.4854891512059,-3.26642040979696,-4.35580607412106,-2.84214630515942,.120594442052501,3.77173740709444,5.02965161709108,3.28182786886397,-4.21693061735794,-5.62332146066281,4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258919,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636384,3.66919510297594,-3.0141507731309,-.104423971232209,-.139250467146391,-.0908603812984474,.116749549066043,.155686755269702,-.127892723210497,-2.94703586087065,-3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.13195404884644,4.17649376078907,2.72514572729005,-.115630018018823,-2.91877623944935,-3.83584465679206,-5.11513931437816,-3.33760825329423,.141617271554661,4.42925189189755,5.90645411811629,3.85393804714673,-4.95205415939119,-6.60361645843633,5.42470353822284,3.97724364248012,5.30369635131817,3.46064097812296,-.146837644201681,-4.59252537503603,-6.12418103253548,-3.99600399916795,5.13459946240544,6.84704254830276,-5.62467189980938,3.67768288095471,4.90422891550851,3.19999004936568,-4.11177446033511,-5.48309461793542,4.5042232468252,2.21014640326349,2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.40022577929605,4.53423694314245,2.95857174458961,-.125534462510035,-3.16878793832491,-4.1644090826311,-5.55328344227955,-3.62349557083131,.153747689149873,4.80864541143983,6.41237937831348,4.18405228663707,-5.37622900930936,-7.16925809522303,5.88936380548032,5.17702934684997,6.90362324406997,4.50458697357606,-.191133071440498,-5.97791857375093,-7.97161747704438,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,6.67435503772228,8.90032284672415,5.8074255986695,-7.46215578440284,-9.95086345672769,8.17438210188188,7.8090615264033,10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.40022577929605,-5.32467544557844,-5.5209565174066,-4.42116822694256,-2.37645064234965,4.1644090826311,6.52136894437417,6.76176317987178,5.41480311178231,-4.80864541143983,-7.53022822838542,-7.80781158453304,5.37622900930936,8.41905110415162,-5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394288,-5.97791857375092,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,-6.67435503772231,-10.4518866380257,-10.8371697481941,7.46215578440288,11.6855645116818,-8.17438210188192,7.8090615264033,12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.13195404884644,-4.90456807960123,-5.08536292603975,-4.0723459639997,-2.18895293851093,3.83584465679206,6.00684460241421,6.22827216283403,4.98758483419414,-4.42925189189755,-6.93610669392436,-7.19178921991206,4.95205415939119,7.75480303503604,-5.42470353822284,3.97724364248013,6.22827216283404,6.45786210779302,5.17143989055336,-4.59252537503604,-7.19178921991208,-7.45689685002024,5.13459946240545,8.04066478857437,-5.62467189980939,-3.67768288095471,-5.75916689254577,-5.97146442400447,4.11177446033512,6.43894433484766,-4.5042232468252,2.21014640326355,3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.66702109783921,-4.17649376078907,-4.3304499371288,-3.46781352289888,-1.86400680793386,3.26642040979696,5.11513931437816,5.30369635131817,4.24718682732847,-3.77173740709444,-5.90645411811629,-6.12418103253548,4.21693061735794,6.60361645843633,-4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812297,2.77127267544969,-2.46104380066281,-3.85393804714674,-3.99600399916796,2.75153061636385,4.3088337276685,-3.01415077313091,.104423971232213,.163525539687803,.169553506762539,-.116749549066047,-.182827111459215,.127892723210502,-2.94703586087065,-4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.03452907455009,-3.18602578468041,-3.30347079372946,-2.6454111829756,-1.4219520231667,2.49177904847469,3.9020687402541,4.04590891241331,3.23995377923585,-2.87725860980962,-4.50572087357673,-4.67181319960176,3.21687291987534,5.03754908153942,-3.52390772571971,.0953780462518065,.149359829020983,.154865612009825,.12401599635522,-.110133081351697,-.172465874990319,-.178823405501172,.123132528211026,.192822710157554,-.134884926556002,2.8024998850757,4.38865042934355,4.55042724011643,-3.1332901246654,-4.90666034634693,3.43234736114769,-3.77550789188145,-5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-1.2742056900953,-1.99537683408524,-2.06893149627982,-1.65679518870216,-.890554665266804,1.56057688324205,2.4438275442558,2.53391323933002,2.02915141041227,-1.80199896789671,-2.82188898006397,-2.92591098178889,2.01469609360342,3.15496779285464,-2.20698899398908,-1.62700976129611,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294204,3.05046536107998,-2.10046056933792,-3.2892729926303,2.30093927043685,3.19783168117628,5.00773093883056,5.19232863088549,-3.57528450949468,-5.59881339795158,3.91652795089531,-.49602746920004,-.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-.433821252801768,-.679354114255872,-.704396833771808,-.564079229896515,-.303201864171246,.531320354197217,.832035467367304,.862706409586761,.690853143909355,-.613515899259544,-.960751801781151,-.996167555656711,.68593162794344,1.07415316937022,-.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.8930154290003,-3.3691862575809,-5.27606827091984,3.69075862820297,1.36899683098764,2.14381758303679,2.22284414873079,-1.53058498738027,-2.39686092429589,1.67667184761831,3.00982432351404,4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,.433821252801768,.679354114255872,.704396833771808,.564079229896515,.303201864171246,-.531320354197217,-.832035467367304,-.862706409586761,-.690853143909355,.613515899259544,.960751801781151,.996167555656711,-.68593162794344,-1.07415316937022,.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.89301542900031,-3.36918625758091,-5.27606827091984,3.69075862820298,-1.36899683098765,-2.1438175830368,-2.22284414873079,1.53058498738027,2.3968609242959,-1.67667184761832,3.00982432351407,4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,1.2742056900953,1.99537683408524,2.06893149627982,1.65679518870216,.890554665266804,-1.56057688324205,-2.4438275442558,-2.53391323933002,-2.02915141041227,1.80199896789671,2.82188898006397,2.92591098178889,-2.01469609360342,-3.15496779285464,2.20698899398908,-1.62700976129612,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294205,3.05046536107998,-2.10046056933793,-3.28927299263031,2.30093927043686,-3.1978316811763,-5.00773093883059,-5.19232863088551,3.5752845094947,5.5988133979516,-3.91652795089533,-.496027469200023,-.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.03452907455009,3.18602578468041,3.30347079372946,2.6454111829756,1.4219520231667,-2.49177904847469,-3.9020687402541,-4.04590891241331,-3.23995377923585,2.87725860980962,4.50572087357673,4.67181319960176,-3.21687291987534,-5.03754908153942,3.52390772571971,.0953780462518019,.149359829020976,.154865612009818,.124015996355214,-.110133081351691,-.172465874990311,-.178823405501163,.12313252821102,.192822710157544,-.134884926555995,-2.8024998850757,-4.38865042934354,-4.55042724011642,3.13329012466539,4.90666034634692,-3.43234736114768,-3.77550789188148,-5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.66702109783921,4.17649376078907,4.3304499371288,3.46781352289888,1.86400680793386,-3.26642040979696,-5.11513931437816,-5.30369635131817,-4.24718682732847,3.77173740709444,5.90645411811629,6.12418103253548,-4.21693061735794,-6.60361645843633,4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812296,2.77127267544968,-2.46104380066281,-3.85393804714673,-3.99600399916795,2.75153061636384,4.30883372766849,-3.0141507731309,-.104423971232209,-.163525539687797,-.169553506762532,.116749549066043,.182827111459208,-.127892723210497,-2.94703586087065,-4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.13195404884644,4.90456807960123,5.08536292603975,4.0723459639997,2.18895293851093,-3.83584465679206,-6.00684460241421,-6.22827216283403,-4.98758483419414,4.42925189189755,6.93610669392436,7.19178921991206,-4.95205415939119,-7.75480303503604,5.42470353822284,3.97724364248012,6.22827216283403,6.45786210779301,5.17143989055335,-4.59252537503603,-7.19178921991206,-7.45689685002023,5.13459946240544,8.04066478857435,-5.62467189980938,3.67768288095471,5.75916689254577,5.97146442400446,-4.11177446033511,-6.43894433484766,4.5042232468252,2.21014640326349,3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.40022577929605,5.32467544557844,5.5209565174066,4.42116822694256,2.37645064234965,-4.1644090826311,-6.52136894437417,-6.76176317987178,-5.41480311178231,4.80864541143983,7.53022822838542,7.80781158453304,-5.37622900930936,-8.41905110415162,5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394289,-5.97791857375093,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,6.67435503772228,10.4518866380257,10.8371697481941,-7.46215578440284,-11.6855645116817,8.17438210188188,7.8090615264033,12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.40022577929605,-5.78076767221695,-7.18642267926874,-8.02365168049822,-8.3966606253625,4.1644090826311,7.0799655598807,8.80153432008931,9.82692624602171,-4.80864541143983,-8.17524004082037,-10.1631364184159,5.37622900930936,9.14019623375092,-5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375092,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,-6.67435503772231,-11.3471570254948,-14.1063386773585,7.46215578440288,12.6865072343198,-8.17438210188192,7.8090615264033,13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.13195404884644,-5.32467544557844,-6.61942678751097,-7.39059991847735,-7.7341791249633,3.83584465679206,6.52136894437417,8.10710900955867,9.05159934712187,-4.42925189189755,-7.53022822838542,-9.36128313844683,4.95205415939119,8.41905110415162,-5.42470353822284,3.97724364248013,6.7617631798718,8.40595765692319,9.38526430991042,-4.59252537503604,-7.80781158453306,-9.70636382862251,5.13459946240545,8.72939873011817,-5.62467189980939,-3.67768288095471,-6.25247606762144,-7.77283197754114,4.11177446033512,6.99048075972417,-4.5042232468252,2.21014640326355,3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.66702109783921,-4.53423694314245,-5.63678477479456,-6.29347864012493,-6.58605413075653,3.26642040979696,5.55328344227955,6.90362324406997,7.70790568809692,-3.77173740709444,-6.41237937831348,-7.97161747704438,4.21693061735794,7.16925809522303,-4.61941604516583,2.13132645129358,3.62349557083132,4.50458697357607,5.02937810056318,-2.46104380066281,-4.18405228663708,-5.2014490017318,2.75153061636385,4.67791266762517,-3.01415077313091,.104423971232213,.177532539414032,.220701460427408,-.116749549066047,-.198487413238653,.127892723210502,-2.94703586087065,-5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.03452907455009,-3.45892910228435,-4.30000442088448,-4.80096137363441,-5.0241517123509,2.49177904847469,4.23630567890486,5.26640836144105,5.87995282040644,-2.87725860980962,-4.89166444644702,-6.08112457054741,3.21687291987534,5.46904711383883,-3.52390772571971,.0953780462518065,.162153445774676,.201582776983911,.2250674723256,-.110133081351697,-.187238671144273,-.232767741029294,.123132528211026,.209339198368165,-.134884926556002,2.8024998850757,4.76456617608066,5.92312100550577,-3.1332901246654,-5.32694692824228,3.43234736114769,-3.77550789188145,-6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-1.2742056900953,-2.16629351671545,-2.69305077477823,-3.00679522191911,-3.14657715137113,1.56057688324205,2.65315687476095,3.2983001248078,3.68255702755716,-1.80199896789671,-3.06360167061167,-3.80854893001879,2.01469609360342,3.42521079645606,-2.20698899398908,-1.62700976129611,-2.76610026595298,-3.43870690112598,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865961,-2.10046056933792,-3.57102008849817,2.30093927043685,3.19783168117628,5.43667485806875,6.75866004623775,-3.57528450949468,-6.07838727908375,3.91652795089531,-.49602746920004,-.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-.433821252801768,-.737545103324376,-.916887022287339,-1.02370573309381,-1.07129645743706,.531320354197217,.903304582796468,1.12295267819227,1.25377834648448,-.613515899259544,-1.04304628770802,-1.29667406213985,.68593162794344,1.16616120174253,-.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.3691862575809,-5.72799698473071,3.69075862820297,1.36899683098764,2.32744915738298,2.89339311993396,-1.53058498738027,-2.60216726588949,1.67667184761831,3.00982432351404,5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,.433821252801768,.737545103324376,.916887022287339,1.02370573309381,1.07129645743706,-.531320354197217,-.903304582796468,-1.12295267819227,-1.25377834648448,.613515899259544,1.04304628770802,1.29667406213985,-.68593162794344,-1.16616120174253,.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.36918625758091,-5.72799698473071,3.69075862820298,-1.36899683098765,-2.32744915738298,-2.89339311993396,1.53058498738027,2.60216726588949,-1.67667184761832,3.00982432351407,5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,1.2742056900953,2.16629351671545,2.69305077477823,3.00679522191911,3.14657715137113,-1.56057688324205,-2.65315687476095,-3.2983001248078,-3.68255702755716,1.80199896789671,3.06360167061167,3.80854893001879,-2.01469609360342,-3.42521079645606,2.20698899398908,-1.62700976129612,-2.76610026595298,-3.43870690112599,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865962,-2.10046056933793,-3.57102008849817,2.30093927043686,-3.1978316811763,-5.43667485806878,-6.75866004623778,3.5752845094947,6.07838727908378,-3.91652795089533,-.496027469200023,-.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.03452907455009,3.45892910228435,4.30000442088448,4.80096137363441,5.0241517123509,-2.49177904847469,-4.23630567890486,-5.26640836144105,-5.87995282040644,2.87725860980962,4.89166444644702,6.08112457054741,-3.21687291987534,-5.46904711383883,3.52390772571971,.0953780462518019,.162153445774668,.201582776983901,.225067472325589,-.110133081351691,-.187238671144264,-.232767741029283,.12313252821102,.209339198368155,-.134884926555995,-2.8024998850757,-4.76456617608065,-5.92312100550575,3.13329012466539,5.32694692824227,-3.43234736114768,-3.77550789188148,-6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.66702109783921,4.53423694314245,5.63678477479456,6.29347864012493,6.58605413075653,-3.26642040979696,-5.55328344227955,-6.90362324406997,-7.70790568809692,3.77173740709444,6.41237937831348,7.97161747704438,-4.21693061735794,-7.16925809522303,4.61941604516583,2.13132645129358,3.62349557083131,4.50458697357606,5.02937810056317,-2.46104380066281,-4.18405228663707,-5.20144900173179,2.75153061636384,4.67791266762516,-3.0141507731309,-.104423971232209,-.177532539414025,-.2207014604274,.116749549066043,.198487413238645,-.127892723210497,-2.94703586087065,-5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.13195404884644,5.32467544557844,6.61942678751097,7.39059991847735,7.7341791249633,-3.83584465679206,-6.52136894437417,-8.10710900955867,-9.05159934712187,4.42925189189755,7.53022822838542,9.36128313844683,-4.95205415939119,-8.41905110415162,5.42470353822284,3.97724364248012,6.76176317987178,8.40595765692318,9.3852643099104,-4.59252537503603,-7.80781158453304,-9.70636382862249,5.13459946240544,8.72939873011815,-5.62467189980938,3.67768288095471,6.25247606762144,7.77283197754114,-4.11177446033511,-6.99048075972417,4.5042232468252,2.21014640326349,3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.40022577929605,5.78076767221695,7.18642267926874,8.02365168049822,8.3966606253625,-4.1644090826311,-7.0799655598807,-8.80153432008931,-9.82692624602171,4.80864541143983,8.17524004082037,10.1631364184159,-5.37622900930936,-9.14019623375092,5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375093,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,6.67435503772228,11.3471570254948,14.1063386773585,-7.46215578440284,-12.6865072343197,8.17438210188188,7.8090615264033,13.2762861543698,-8.55439830449838,8.51143945889045}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.40022577929605,5.78076767221695,-7.18642267926874,8.02365168049826,-8.3966606253625,4.1644090826311,-7.0799655598807,8.80153432008931,-9.82692624602176,-4.80864541143983,8.17524004082037,-10.1631364184159,5.37622900930936,-9.14019623375092,-5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375092,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,-6.67435503772231,11.3471570254948,-14.1063386773585,7.46215578440288,-12.6865072343198,-8.17438210188192,7.8090615264033,-13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.13195404884644,5.32467544557844,-6.61942678751097,7.39059991847739,-7.73417912496331,3.83584465679206,-6.52136894437417,8.10710900955867,-9.05159934712192,-4.42925189189755,7.53022822838542,-9.36128313844683,4.95205415939119,-8.41905110415162,-5.42470353822284,3.97724364248013,-6.7617631798718,8.40595765692319,-9.38526430991047,-4.59252537503604,7.80781158453306,-9.70636382862251,5.13459946240545,-8.72939873011817,-5.62467189980939,-3.67768288095471,6.25247606762144,-7.77283197754114,4.11177446033512,-6.99048075972417,-4.5042232468252,2.21014640326355,-3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.66702109783921,4.53423694314245,-5.63678477479456,6.29347864012496,-6.58605413075654,3.26642040979696,-5.55328344227955,6.90362324406996,-7.70790568809696,-3.77173740709444,6.41237937831348,-7.97161747704437,4.21693061735794,-7.16925809522303,-4.61941604516583,2.13132645129358,-3.62349557083132,4.50458697357607,-5.02937810056321,-2.46104380066281,4.18405228663708,-5.2014490017318,2.75153061636385,-4.67791266762517,-3.01415077313091,.104423971232213,-.177532539414032,.220701460427408,-.116749549066047,.198487413238653,.127892723210502,-2.94703586087065,5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.03452907455009,3.45892910228435,-4.30000442088448,4.80096137363443,-5.0241517123509,2.49177904847469,-4.23630567890486,5.26640836144105,-5.87995282040647,-2.87725860980962,4.89166444644702,-6.08112457054741,3.21687291987534,-5.46904711383883,-3.52390772571971,.0953780462518065,-.162153445774676,.201582776983911,-.225067472325601,-.110133081351697,.187238671144273,-.232767741029294,.123132528211026,-.209339198368165,-.134884926556002,2.8024998850757,-4.76456617608066,5.92312100550576,-3.1332901246654,5.32694692824228,3.43234736114769,-3.77550789188145,6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-1.2742056900953,2.16629351671545,-2.69305077477823,3.00679522191912,-3.14657715137113,1.56057688324205,-2.65315687476095,3.2983001248078,-3.68255702755718,-1.80199896789671,3.06360167061167,-3.80854893001879,2.01469609360342,-3.42521079645606,-2.20698899398908,-1.62700976129611,2.76610026595298,-3.43870690112598,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865961,-2.10046056933792,3.57102008849817,2.30093927043685,3.19783168117628,-5.43667485806875,6.75866004623775,-3.57528450949468,6.07838727908375,3.91652795089531,-.49602746920004,.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-.433821252801768,.737545103324376,-.916887022287339,1.02370573309382,-1.07129645743706,.531320354197217,-.903304582796468,1.12295267819227,-1.25377834648449,-.613515899259544,1.04304628770802,-1.29667406213985,.68593162794344,-1.16616120174253,-.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.3691862575809,5.72799698473071,3.69075862820297,1.36899683098764,-2.32744915738298,2.89339311993396,-1.53058498738027,2.60216726588949,1.67667184761831,3.00982432351404,-5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,.433821252801768,-.737545103324376,.916887022287339,-1.02370573309382,1.07129645743706,-.531320354197217,.903304582796468,-1.12295267819227,1.25377834648449,.613515899259544,-1.04304628770802,1.29667406213985,-.68593162794344,1.16616120174253,.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.36918625758091,5.72799698473071,3.69075862820298,-1.36899683098765,2.32744915738298,-2.89339311993396,1.53058498738027,-2.60216726588949,-1.67667184761832,3.00982432351407,-5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,1.2742056900953,-2.16629351671545,2.69305077477823,-3.00679522191912,3.14657715137113,-1.56057688324205,2.65315687476095,-3.2983001248078,3.68255702755718,1.80199896789671,-3.06360167061167,3.80854893001879,-2.01469609360342,3.42521079645606,2.20698899398908,-1.62700976129612,2.76610026595298,-3.43870690112599,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865962,-2.10046056933793,3.57102008849817,2.30093927043686,-3.1978316811763,5.43667485806878,-6.75866004623778,3.5752845094947,-6.07838727908378,-3.91652795089533,-.496027469200023,.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.03452907455009,-3.45892910228435,4.30000442088448,-4.80096137363443,5.0241517123509,-2.49177904847469,4.23630567890486,-5.26640836144105,5.87995282040647,2.87725860980962,-4.89166444644702,6.08112457054741,-3.21687291987534,5.46904711383883,3.52390772571971,.0953780462518019,-.162153445774668,.201582776983901,-.225067472325591,-.110133081351691,.187238671144264,-.232767741029283,.12313252821102,-.209339198368155,-.134884926555995,-2.8024998850757,4.76456617608065,-5.92312100550575,3.13329012466539,-5.32694692824227,-3.43234736114768,-3.77550789188148,6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.66702109783921,-4.53423694314245,5.63678477479456,-6.29347864012496,6.58605413075654,-3.26642040979696,5.55328344227955,-6.90362324406996,7.70790568809696,3.77173740709444,-6.41237937831348,7.97161747704437,-4.21693061735794,7.16925809522303,4.61941604516583,2.13132645129358,-3.62349557083131,4.50458697357606,-5.0293781005632,-2.46104380066281,4.18405228663707,-5.20144900173179,2.75153061636384,-4.67791266762516,-3.0141507731309,-.104423971232209,.177532539414025,-.2207014604274,.116749549066043,-.198487413238645,-.127892723210497,-2.94703586087065,5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.13195404884644,-5.32467544557844,6.61942678751097,-7.39059991847739,7.73417912496331,-3.83584465679206,6.52136894437417,-8.10710900955867,9.05159934712192,4.42925189189755,-7.53022822838542,9.36128313844683,-4.95205415939119,8.41905110415162,5.42470353822284,3.97724364248012,-6.76176317987178,8.40595765692318,-9.38526430991045,-4.59252537503603,7.80781158453304,-9.70636382862249,5.13459946240544,-8.72939873011815,-5.62467189980938,3.67768288095471,-6.25247606762144,7.77283197754114,-4.11177446033511,6.99048075972417,4.5042232468252,2.21014640326349,-3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.40022577929605,-5.78076767221695,7.18642267926874,-8.02365168049826,8.3966606253625,-4.1644090826311,7.0799655598807,-8.80153432008931,9.82692624602176,4.80864541143983,-8.17524004082037,10.1631364184159,-5.37622900930936,9.14019623375092,5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375093,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,6.67435503772228,-11.3471570254948,14.1063386773585,-7.46215578440284,12.6865072343197,8.17438210188188,7.8090615264033,-13.2762861543698,-8.55439830449838,8.51143945889045},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.40022577929605,5.32467544557844,-5.52095651740661,4.42116822694257,-2.37645064234971,4.1644090826311,-6.52136894437417,6.7617631798718,-5.41480311178232,-4.80864541143983,7.53022822838542,-7.80781158453306,5.37622900930936,-8.41905110415162,-5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375092,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,-6.67435503772231,10.4518866380257,-10.8371697481941,7.46215578440288,-11.6855645116818,-8.17438210188192,7.8090615264033,-12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.13195404884644,4.90456807960123,-5.08536292603976,4.0723459639997,-2.18895293851099,3.83584465679206,-6.00684460241421,6.22827216283404,-4.98758483419414,-4.42925189189755,6.93610669392436,-7.19178921991208,4.95205415939119,-7.75480303503604,-5.42470353822284,3.97724364248013,-6.22827216283404,6.45786210779303,-5.17143989055336,-4.59252537503604,7.19178921991208,-7.45689685002025,5.13459946240545,-8.04066478857437,-5.62467189980939,-3.67768288095471,5.75916689254577,-5.97146442400448,4.11177446033512,-6.43894433484766,-4.5042232468252,2.21014640326355,-3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.66702109783921,4.17649376078907,-4.33044993712881,3.46781352289888,-1.86400680793392,3.26642040979696,-5.11513931437816,5.30369635131817,-4.24718682732847,-3.77173740709444,5.90645411811629,-6.12418103253549,4.21693061735794,-6.60361645843633,-4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812298,-2.77127267544969,-2.46104380066281,3.85393804714674,-3.99600399916797,2.75153061636385,-4.3088337276685,-3.01415077313091,.104423971232213,-.163525539687803,.169553506762539,-.116749549066047,.182827111459215,.127892723210502,-2.94703586087065,4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.03452907455009,3.18602578468041,-3.30347079372947,2.6454111829756,-1.42195202316674,2.49177904847469,-3.9020687402541,4.04590891241332,-3.23995377923585,-2.87725860980962,4.50572087357673,-4.67181319960176,3.21687291987534,-5.03754908153942,-3.52390772571971,.0953780462518065,-.149359829020983,.154865612009825,-.12401599635522,-.110133081351697,.172465874990319,-.178823405501172,.123132528211026,-.192822710157554,-.134884926556002,2.8024998850757,-4.38865042934355,4.55042724011644,-3.1332901246654,4.90666034634693,3.43234736114769,-3.77550789188145,5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-1.2742056900953,1.99537683408524,-2.06893149627983,1.65679518870216,-.890554665266828,1.56057688324205,-2.4438275442558,2.53391323933003,-2.02915141041228,-1.80199896789671,2.82188898006397,-2.92591098178889,2.01469609360342,-3.15496779285464,-2.20698899398908,-1.62700976129611,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294204,3.05046536107998,-2.10046056933792,3.2892729926303,2.30093927043685,3.19783168117628,-5.00773093883056,5.19232863088549,-3.57528450949468,5.59881339795158,3.91652795089531,-.49602746920004,.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-.433821252801768,.679354114255872,-.704396833771809,.564079229896515,-.303201864171254,.531320354197217,-.832035467367304,.862706409586763,-.690853143909356,-.613515899259544,.960751801781151,-.996167555656712,.68593162794344,-1.07415316937022,-.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.3933599570385,3.01349180095816,-4.71905892262039,4.89301542900031,-3.3691862575809,5.27606827091984,3.69075862820297,1.36899683098764,-2.14381758303679,2.2228441487308,-1.53058498738027,2.39686092429589,1.67667184761831,3.00982432351404,-4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,.433821252801768,-.679354114255872,.704396833771809,-.564079229896515,.303201864171254,-.531320354197217,.832035467367304,-.862706409586763,.690853143909356,.613515899259544,-.960751801781151,.996167555656712,-.68593162794344,1.07415316937022,.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.39335995703851,3.01349180095816,-4.71905892262039,4.89301542900031,-3.36918625758091,5.27606827091984,3.69075862820298,-1.36899683098765,2.1438175830368,-2.2228441487308,1.53058498738027,-2.3968609242959,-1.67667184761832,3.00982432351407,-4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,1.2742056900953,-1.99537683408524,2.06893149627983,-1.65679518870216,.890554665266828,-1.56057688324205,2.4438275442558,-2.53391323933003,2.02915141041228,1.80199896789671,-2.82188898006397,2.92591098178889,-2.01469609360342,3.15496779285464,2.20698899398908,-1.62700976129612,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294205,3.05046536107998,-2.10046056933793,3.28927299263031,2.30093927043686,-3.1978316811763,5.00773093883059,-5.19232863088552,3.5752845094947,-5.5988133979516,-3.91652795089533,-.496027469200023,.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.03452907455009,-3.18602578468041,3.30347079372947,-2.6454111829756,1.42195202316674,-2.49177904847469,3.9020687402541,-4.04590891241332,3.23995377923585,2.87725860980962,-4.50572087357673,4.67181319960176,-3.21687291987534,5.03754908153942,3.52390772571971,.0953780462518019,-.149359829020976,.154865612009818,-.124015996355214,-.110133081351691,.172465874990311,-.178823405501163,.12313252821102,-.192822710157544,-.134884926555995,-2.8024998850757,4.38865042934354,-4.55042724011643,3.13329012466539,-4.90666034634692,-3.43234736114768,-3.77550789188148,5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.66702109783921,-4.17649376078907,4.33044993712881,-3.46781352289888,1.86400680793392,-3.26642040979696,5.11513931437816,-5.30369635131817,4.24718682732847,3.77173740709444,-5.90645411811629,6.12418103253549,-4.21693061735794,6.60361645843633,4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812297,-2.77127267544968,-2.46104380066281,3.85393804714673,-3.99600399916796,2.75153061636384,-4.30883372766849,-3.0141507731309,-.104423971232209,.163525539687797,-.169553506762532,.116749549066043,-.182827111459208,-.127892723210497,-2.94703586087065,4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.13195404884644,-4.90456807960123,5.08536292603976,-4.0723459639997,2.18895293851099,-3.83584465679206,6.00684460241421,-6.22827216283404,4.98758483419414,4.42925189189755,-6.93610669392436,7.19178921991208,-4.95205415939119,7.75480303503604,5.42470353822284,3.97724364248012,-6.22827216283403,6.45786210779302,-5.17143989055335,-4.59252537503603,7.19178921991206,-7.45689685002024,5.13459946240544,-8.04066478857435,-5.62467189980938,3.67768288095471,-5.75916689254577,5.97146442400447,-4.11177446033511,6.43894433484766,4.5042232468252,2.21014640326349,-3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.40022577929605,-5.32467544557844,5.52095651740661,-4.42116822694257,2.37645064234971,-4.1644090826311,6.52136894437417,-6.7617631798718,5.41480311178232,4.80864541143983,-7.53022822838542,7.80781158453306,-5.37622900930936,8.41905110415162,5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375093,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,6.67435503772228,-10.4518866380257,10.8371697481941,-7.46215578440284,11.6855645116817,8.17438210188188,7.8090615264033,-12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.40022577929605,4.53423694314245,-2.95857174458962,-.12553446251004,3.1687879383249,4.1644090826311,-5.55328344227955,3.62349557083132,.153747689149879,-4.80864541143983,6.41237937831348,-4.18405228663708,5.37622900930936,-7.16925809522303,-5.88936380548032,5.17702934684997,-6.90362324406996,4.50458697357607,.191133071440505,-5.97791857375092,7.97161747704437,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,-6.67435503772231,8.90032284672419,-5.80742559866954,7.46215578440288,-9.95086345672774,-8.17438210188192,7.8090615264033,-10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.13195404884644,4.17649376078907,-2.72514572729006,-.115630018018827,2.91877623944935,3.83584465679206,-5.11513931437816,3.33760825329423,.141617271554667,-4.42925189189755,5.90645411811629,-3.85393804714674,4.95205415939119,-6.60361645843633,-5.42470353822284,3.97724364248013,-5.30369635131817,3.46064097812298,.146837644201687,-4.59252537503604,6.12418103253549,-3.99600399916797,5.13459946240545,-6.84704254830277,-5.62467189980939,-3.67768288095471,4.90422891550851,-3.19999004936569,4.11177446033512,-5.48309461793542,-4.5042232468252,2.21014640326355,-2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.66702109783921,3.55650076638925,-2.32060274065832,-.0984649496097572,2.4854891512059,3.26642040979696,-4.35580607412106,2.84214630515942,.120594442052505,-3.77173740709444,5.02965161709108,-3.28182786886397,4.21693061735794,-5.62332146066281,-4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898143,.0786873984258951,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636385,-3.66919510297595,-3.01415077313091,.104423971232213,-.139250467146397,.0908603812984511,-.116749549066047,.155686755269708,.127892723210502,-2.94703586087065,3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.03452907455009,2.71306598164557,-1.77026486598667,-.0751136925641364,1.89604797153805,2.49177904847469,-3.32281364706153,2.16812281562257,.0919951097438821,-2.87725860980962,3.83685470584401,-2.50353258258141,3.21687291987534,-4.28973397193188,-3.52390772571971,.0953780462518065,-.127187630774895,.0829894280920859,.00352130492348755,-.110133081351697,.146863625737388,-.0958279372639542,.123132528211026,-.164198525301527,-.134884926556002,2.8024998850757,-3.73716315870324,2.43848424018079,-3.1332901246654,4.17827543429579,3.43234736114769,-3.77550789188145,5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-1.2742056900953,1.69916672838959,-1.10869959708727,-.0470429720894781,1.18747632769106,1.56057688324205,-2.08104573641859,1.35787414544697,.0576156388045318,-1.80199896789671,2.40298463138773,-1.56793800686758,2.01469609360342,-2.68661849290085,-2.20698899398908,-1.62700976129611,2.16963467924844,-1.41567808207906,-.0600683040611749,1.87870904723461,-2.50527833232423,1.63468424297244,-2.10046056933792,2.8009863270915,2.30093927043685,3.19783168117628,-4.26434227893045,2.78246653954405,-3.57528450949468,4.76767960906116,3.91652795089531,-.49602746920004,.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-.433821252801768,.578505216668687,-.377472375086663,-.016016441651465,.404293021256913,.531320354197217,-.70852129725097,.462307355479536,.0196160547715705,-.613515899259544,.818129923100605,-.533826552292297,.68593162794344,-.914697061432424,-.751400451068813,-2.60976045384023,3.48014309433045,-2.27077966086395,-.0963509182288022,3.01349180095816,-4.01852310483628,2.62207049601641,-3.3691862575809,4.49284541622379,3.69075862820297,1.36899683098764,-1.82557171487995,1.19117835294064,-1.53058498738027,2.04105122680137,1.67667184761831,3.00982432351404,-4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,.433821252801768,-.578505216668687,.377472375086663,.016016441651465,-.404293021256913,-.531320354197217,.70852129725097,-.462307355479536,-.0196160547715705,.613515899259544,-.818129923100605,.533826552292297,-.68593162794344,.914697061432424,.751400451068813,-2.60976045384023,3.48014309433046,-2.27077966086395,-.0963509182288023,3.01349180095816,-4.01852310483628,2.62207049601641,-3.36918625758091,4.49284541622379,3.69075862820298,-1.36899683098765,1.82557171487995,-1.19117835294064,1.53058498738027,-2.04105122680137,-1.67667184761832,3.00982432351407,-4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,1.2742056900953,-1.69916672838959,1.10869959708727,.0470429720894781,-1.18747632769106,-1.56057688324205,2.08104573641859,-1.35787414544697,-.0576156388045318,1.80199896789671,-2.40298463138773,1.56793800686758,-2.01469609360342,2.68661849290085,2.20698899398908,-1.62700976129612,2.16963467924844,-1.41567808207907,-.060068304061175,1.87870904723461,-2.50527833232424,1.63468424297245,-2.10046056933793,2.8009863270915,2.30093927043686,-3.1978316811763,4.26434227893047,-2.78246653954406,3.5752845094947,-4.76767960906118,-3.91652795089533,-.496027469200023,.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.03452907455009,-2.71306598164557,1.77026486598667,.0751136925641364,-1.89604797153805,-2.49177904847469,3.32281364706153,-2.16812281562257,-.0919951097438821,2.87725860980962,-3.83685470584401,2.50353258258141,-3.21687291987534,4.28973397193188,3.52390772571971,.0953780462518019,-.127187630774889,.0829894280920819,.00352130492348739,-.110133081351691,.146863625737381,-.0958279372639496,.12313252821102,-.164198525301519,-.134884926555995,-2.8024998850757,3.73716315870323,-2.43848424018078,3.13329012466539,-4.17827543429578,-3.43234736114768,-3.77550789188148,5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.66702109783921,-3.55650076638925,2.32060274065832,.0984649496097572,-2.4854891512059,-3.26642040979696,4.35580607412106,-2.84214630515942,-.120594442052505,3.77173740709444,-5.02965161709108,3.28182786886397,-4.21693061735794,5.62332146066281,4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898142,.0786873984258949,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636384,-3.66919510297594,-3.0141507731309,-.104423971232209,.139250467146391,-.0908603812984476,.116749549066043,-.155686755269702,-.127892723210497,-2.94703586087065,3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.13195404884644,-4.17649376078907,2.72514572729006,.115630018018827,-2.91877623944935,-3.83584465679206,5.11513931437816,-3.33760825329423,-.141617271554667,4.42925189189755,-5.90645411811629,3.85393804714674,-4.95205415939119,6.60361645843633,5.42470353822284,3.97724364248012,-5.30369635131817,3.46064097812297,.146837644201687,-4.59252537503603,6.12418103253548,-3.99600399916796,5.13459946240544,-6.84704254830276,-5.62467189980938,3.67768288095471,-4.90422891550851,3.19999004936568,-4.11177446033511,5.48309461793542,4.5042232468252,2.21014640326349,-2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.40022577929605,-4.53423694314245,2.95857174458962,.12553446251004,-3.1687879383249,-4.1644090826311,5.55328344227955,-3.62349557083132,-.153747689149879,4.80864541143983,-6.41237937831348,4.18405228663708,-5.37622900930936,7.16925809522303,5.88936380548032,5.17702934684997,-6.90362324406997,4.50458697357607,.191133071440505,-5.97791857375093,7.97161747704438,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,6.67435503772228,-8.90032284672415,5.80742559866952,-7.46215578440284,9.95086345672769,8.17438210188188,7.8090615264033,-10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.40022577929605,3.45892910228435,-.132397734060631,-3.36905705276318,4.05959901190681,4.1644090826311,-4.23630567890486,.162153445774676,4.1262353470069,-4.80864541143983,4.89166444644702,-.187238671144273,5.37622900930936,-5.46904711383883,-5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375092,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,-6.67435503772231,6.78958468653045,-.259885531387107,7.46215578440288,-7.59098645151066,-8.17438210188192,7.8090615264033,-7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.13195404884644,3.18602578468041,-.121951789723545,-3.10324447907128,3.73930391312628,3.83584465679206,-3.9020687402541,.149359829020983,3.80068276060982,-4.42925189189755,4.50572087357673,-.172465874990319,4.95205415939119,-5.03754908153942,-5.42470353822284,3.97724364248013,-4.04590891241332,.154865612009825,3.94078558921029,-4.59252537503604,4.67181319960176,-.178823405501172,5.13459946240545,-5.22324594675678,-5.62467189980939,-3.67768288095471,3.74117637274605,-.143201337715524,4.11177446033512,-4.18276234400124,-4.5042232468252,2.21014640326355,-2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.66702109783921,2.71306598164557,-.103848265663968,-2.64257341211137,3.18421096606244,3.26642040979696,-3.32281364706153,.127187630774895,3.23647823392353,-3.77173740709444,3.83685470584401,-.146863625737388,4.21693061735794,-4.28973397193188,-4.61941604516583,2.13132645129358,-2.16812281562257,.0829894280920859,2.11178929936587,-2.46104380066281,2.50353258258141,-.0958279372639542,2.75153061636385,-2.79903451954298,-3.01415077313091,.104423971232213,-.106226802736416,.0040660526897138,-.116749549066047,.118765176014247,.127892723210502,-2.94703586087065,2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.03452907455009,2.0696542765645,-.0792203391289715,-2.01587923055033,2.42906581999068,2.49177904847469,-2.53479846100042,.0970247040581402,2.46893774908671,-2.87725860980962,2.92693314656882,-.112034478016931,3.21687291987534,-3.2724107413517,-3.52390772571971,.0953780462518065,-.0970247040581402,.00371382314810282,.0945037478171115,-.110133081351697,.112034478016931,-.00428835358704871,.123132528211026,-.12525835434705,-.134884926556002,2.8024998850757,-2.85088374774363,.109123528450137,-3.1332901246654,3.18738492899079,3.43234736114769,-3.77550789188145,3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-1.2742056900953,1.29620426108275,-.0496149247273535,-1.2625254749334,1.52130020070251,1.56057688324205,-1.58751952117754,.0607656246043247,1.54627160050444,-1.80199896789671,1.83310964514541,-.0701660994486395,2.01469609360342,-2.04947888880939,-2.20698899398908,-1.62700976129611,1.65509933148238,-.0633523829764578,-1.61209551065421,1.87870904723461,-1.91114408921758,.0731530307047499,-2.10046056933792,2.13672404935301,2.30093927043685,3.19783168117628,-3.25304076422429,.124516928010565,-3.57528450949468,3.63701014237694,3.91652795089531,-.49602746920004,.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-.433821252801768,.441310975771777,-.0168920991094272,-.429844559231871,.517948055079696,.531320354197217,-.540493354313145,.0206885117513158,.526449919441546,-.613515899259544,.624107966995901,-.0238890356585565,.68593162794344,-.697773919897946,-.751400451068813,-2.60976045384023,2.65481675969736,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.3691862575809,3.42735369966595,3.69075862820297,1.36899683098764,-1.39263192728723,.0533058950082319,-1.53058498738027,1.5570098290328,1.67667184761831,3.00982432351404,-3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,.433821252801768,-.441310975771777,.0168920991094272,.429844559231871,-.517948055079696,-.531320354197217,.540493354313145,-.0206885117513158,-.526449919441546,.613515899259544,-.624107966995901,.0238890356585565,-.68593162794344,.697773919897946,.751400451068813,-2.60976045384023,2.65481675969737,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.36918625758091,3.42735369966595,3.69075862820298,-1.36899683098765,1.39263192728723,-.0533058950082319,1.53058498738027,-1.5570098290328,-1.67667184761832,3.00982432351407,-3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,1.2742056900953,-1.29620426108275,.0496149247273535,1.2625254749334,-1.52130020070251,-1.56057688324205,1.58751952117754,-.0607656246043247,-1.54627160050444,1.80199896789671,-1.83310964514541,.0701660994486395,-2.01469609360342,2.04947888880939,2.20698899398908,-1.62700976129612,1.65509933148239,-.0633523829764579,-1.61209551065421,1.87870904723461,-1.91114408921759,.07315303070475,-2.10046056933793,2.13672404935301,2.30093927043686,-3.1978316811763,3.2530407642243,-.124516928010566,3.5752845094947,-3.63701014237696,-3.91652795089533,-.496027469200023,.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.03452907455009,-2.0696542765645,.0792203391289715,2.01587923055033,-2.42906581999068,-2.49177904847469,2.53479846100042,-.0970247040581402,-2.46893774908671,2.87725860980962,-2.92693314656882,.112034478016931,-3.21687291987534,3.2724107413517,3.52390772571971,.0953780462518019,-.0970247040581356,.00371382314810264,.0945037478171069,-.110133081351691,.112034478016925,-.00428835358704851,.12313252821102,-.125258354347044,-.134884926555995,-2.8024998850757,2.85088374774362,-.109123528450137,3.13329012466539,-3.18738492899078,-3.43234736114768,-3.77550789188148,3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.66702109783921,-2.71306598164557,.103848265663968,2.64257341211137,-3.18421096606244,-3.26642040979696,3.32281364706153,-.127187630774895,-3.23647823392353,3.77173740709444,-3.83685470584401,.146863625737388,-4.21693061735794,4.28973397193188,4.61941604516583,2.13132645129358,-2.16812281562256,.0829894280920858,2.11178929936586,-2.46104380066281,2.5035325825814,-.095827937263954,2.75153061636384,-2.79903451954297,-3.0141507731309,-.104423971232209,.106226802736412,-.00406605268971365,.116749549066043,-.118765176014243,-.127892723210497,-2.94703586087065,2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.13195404884644,-3.18602578468041,.121951789723545,3.10324447907128,-3.73930391312628,-3.83584465679206,3.9020687402541,-.149359829020983,-3.80068276060982,4.42925189189755,-4.50572087357673,.172465874990319,-4.95205415939119,5.03754908153942,5.42470353822284,3.97724364248012,-4.04590891241331,.154865612009825,3.94078558921028,-4.59252537503603,4.67181319960176,-.178823405501172,5.13459946240544,-5.22324594675677,-5.62467189980938,3.67768288095471,-3.74117637274605,.143201337715524,-4.11177446033511,4.18276234400124,4.5042232468252,2.21014640326349,-2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.40022577929605,-3.45892910228435,.132397734060631,3.36905705276318,-4.05959901190681,-4.1644090826311,4.23630567890486,-.162153445774676,-4.1262353470069,4.80864541143983,-4.89166444644702,.187238671144273,-5.37622900930936,5.46904711383883,5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375093,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,6.67435503772228,-6.78958468653042,.259885531387105,-7.46215578440284,7.59098645151062,8.17438210188188,7.8090615264033,-7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.40022577929605,2.16629351671545,2.25851140965318,-3.84430965952581,.533351454031698,4.1644090826311,-2.65315687476095,-2.76610026595298,4.70829853978447,-4.80864541143983,3.06360167061167,3.19401746644684,5.37622900930936,-3.42521079645606,-5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375092,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,-6.67435503772231,4.25225058180805,4.43326649059391,7.46215578440288,-4.75416068069224,-8.17438210188192,7.8090615264033,-4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.13195404884644,1.99537683408524,2.08031890026244,-3.54100050546174,.491270978557942,3.83584465679206,-2.4438275442558,-2.54785990395621,4.33682220887952,-4.42925189189755,2.82188898006397,2.94201520294204,4.95205415939119,-3.15496779285464,-5.42470353822284,3.97724364248013,-2.53391323933003,-2.64178049699053,4.49668850050967,-4.59252537503604,2.92591098178889,3.05046536107998,5.13459946240545,-3.27126792601698,-5.62467189980939,-3.67768288095471,2.34306222576705,2.44280506311383,4.11177446033512,-2.61962320701725,-4.5042232468252,2.21014640326355,-1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.66702109783921,1.69916672838959,1.77149929746803,-3.01534534295104,.418342684514395,3.26642040979696,-2.08104573641859,-2.16963467924844,3.69302874444133,-3.77173740709444,2.40298463138773,2.50527833232423,4.21693061735794,-2.68661849290085,-4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636385,-1.75300798410243,-3.01415077313091,.104423971232213,-.0665288091384509,-.0693609030179036,-.116749549066047,.0743814698720828,.127892723210502,-2.94703586087065,1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.03452907455009,1.29620426108275,1.35138294525071,-2.30024718402619,.319131466736219,2.49177904847469,-1.58751952117754,-1.65509933148238,2.81721594171208,-2.87725860980962,1.83310964514541,1.91114408921758,3.21687291987534,-2.04947888880939,-3.52390772571971,.0953780462518065,-.0607656246043247,-.0633523829764578,.107834822891095,-.110133081351697,.0701660994486395,.0731530307047499,.123132528211026,-.0784480840492681,-.134884926556002,2.8024998850757,-1.78548064936278,-1.86148755350589,-3.1332901246654,1.99622805289337,3.43234736114769,-3.77550789188145,2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-1.2742056900953,.811800069931489,.846357940948566,-1.44062234704609,.199868920965739,1.56057688324205,-.994247972331933,-1.03657254753862,1.76439483124641,-1.80199896789671,1.14805866852966,1.19693087876444,2.01469609360342,-1.28356861276539,-2.20698899398908,-1.62700976129611,1.03657254753862,1.08069885633786,-1.83950412433748,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933792,1.338209404774,2.30093927043685,3.19783168117628,-2.03734766130249,-2.1240764020783,-3.57528450949468,2.27782393297861,3.91652795089531,-.49602746920004,.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-.433821252801768,.27638875426455,.2881544676147,-.490480144821059,.0680481859118126,.531320354197217,-.338505709325782,-.352915706379786,.6007130419195,-.613515899259544,.390872724665213,.407511956177088,.68593162794344,-.437008991543009,-.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.3691862575809,2.14651815042224,3.69075862820297,1.36899683098764,-.872191775558762,-.909320487484575,-1.53058498738027,.975140050100603,1.67667184761831,3.00982432351404,-1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,.433821252801768,-.27638875426455,-.2881544676147,.490480144821059,-.0680481859118126,-.531320354197217,.338505709325782,.352915706379786,-.6007130419195,.613515899259544,-.390872724665213,-.407511956177088,-.68593162794344,.437008991543009,.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.36918625758091,2.14651815042224,3.69075862820298,-1.36899683098765,.872191775558764,.909320487484576,1.53058498738027,-.975140050100604,-1.67667184761832,3.00982432351407,-1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,1.2742056900953,-.811800069931489,-.846357940948566,1.44062234704609,-.199868920965739,-1.56057688324205,.994247972331933,1.03657254753862,-1.76439483124641,1.80199896789671,-1.14805866852966,-1.19693087876444,-2.01469609360342,1.28356861276539,2.20698899398908,-1.62700976129612,1.03657254753862,1.08069885633786,-1.83950412433749,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933793,1.338209404774,2.30093927043686,-3.1978316811763,2.0373476613025,2.12407640207831,3.5752845094947,-2.27782393297862,-3.91652795089533,-.496027469200023,.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.03452907455009,-1.29620426108275,-1.35138294525071,2.30024718402619,-.319131466736219,-2.49177904847469,1.58751952117754,1.65509933148238,-2.81721594171208,2.87725860980962,-1.83310964514541,-1.91114408921758,-3.21687291987534,2.04947888880939,3.52390772571971,.0953780462518019,-.0607656246043218,-.0633523829764547,.10783482289109,-.110133081351691,.0701660994486362,.0731530307047464,.12313252821102,-.0784480840492644,-.134884926555995,-2.8024998850757,1.78548064936277,1.86148755350588,3.13329012466539,-1.99622805289336,-3.43234736114768,-3.77550789188148,2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.66702109783921,-1.69916672838959,-1.77149929746803,3.01534534295104,-.418342684514395,-3.26642040979696,2.08104573641859,2.16963467924844,-3.69302874444133,3.77173740709444,-2.40298463138773,-2.50527833232423,-4.21693061735794,2.68661849290085,4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636384,-1.75300798410242,-3.0141507731309,-.104423971232209,.0665288091384484,.0693609030179009,.116749549066043,-.0743814698720799,-.127892723210497,-2.94703586087065,1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.13195404884644,-1.99537683408524,-2.08031890026244,3.54100050546174,-.491270978557942,-3.83584465679206,2.4438275442558,2.54785990395621,-4.33682220887952,4.42925189189755,-2.82188898006397,-2.94201520294204,-4.95205415939119,3.15496779285464,5.42470353822284,3.97724364248012,-2.53391323933002,-2.64178049699052,4.49668850050967,-4.59252537503603,2.92591098178889,3.05046536107998,5.13459946240544,-3.27126792601697,-5.62467189980938,3.67768288095471,-2.34306222576705,-2.44280506311383,-4.11177446033511,2.61962320701725,4.5042232468252,2.21014640326349,-1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.40022577929605,-2.16629351671545,-2.25851140965318,3.84430965952581,-.533351454031698,-4.1644090826311,2.65315687476095,2.76610026595298,-4.70829853978447,4.80864541143983,-3.06360167061167,-3.19401746644684,-5.37622900930936,3.42521079645606,5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375093,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,6.67435503772228,-4.25225058180803,-4.43326649059389,-7.46215578440284,4.75416068069222,8.17438210188188,7.8090615264033,-4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.40022577929605,.737545103324376,3.62270337995047,-1.64575508217185,-3.2363009692086,4.1644090826311,-.903304582796468,-4.43688738516868,2.01563009655897,-4.80864541143983,1.04304628770802,5.12327625260686,5.37622900930936,-1.16616120174253,-5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375092,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394474,-7.32142511479571,-6.67435503772231,1.44773853151531,7.11105971439921,7.46215578440288,-1.61862088558446,-8.17438210188192,7.8090615264033,-1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.13195404884644,.679354114255872,3.33687856485652,-1.51590795070228,-2.98096261298772,3.83584465679206,-.832035467367304,-4.08682490876582,1.85660048821863,-4.42925189189755,.960751801781151,4.71905892262039,4.95205415939119,-1.07415316937022,-5.42470353822284,3.97724364248013,-.862706409586763,-4.2374756641165,1.92503950203906,-4.59252537503604,.996167555656712,4.89301542900031,5.13459946240545,-1.1137491858232,-5.62467189980939,-3.67768288095471,.797728497114731,3.91831456811703,4.11177446033512,-.891887573859297,-4.5042232468252,2.21014640326355,-.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.66702109783921,.578505216668687,2.84152493765918,-1.29087413922765,-2.53844406932994,3.26642040979696,-.70852129725097,-3.48014309433045,1.58099148171138,-3.77173740709444,.818129923100605,4.01852310483628,4.21693061735794,-.914697061432424,-4.61941604516583,2.13132645129358,-.462307355479536,-2.27077966086395,1.03159071439798,-2.46104380066281,.533826552292297,2.62207049601641,2.75153061636385,-.596836229618421,-3.01415077313091,.104423971232213,-.0226506690028153,-.111256457118347,-.116749549066047,.0253242178213241,.127892723210502,-2.94703586087065,.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.03452907455009,.441310975771777,2.16764880728185,-.984739479553152,-1.93644447258226,2.49177904847469,-.540493354313145,-2.65481675969736,1.20605462730079,-2.87725860980962,.624107966995901,3.06551834185512,3.21687291987534,-.697773919897946,-3.52390772571971,.0953780462518065,-.0206885117513158,-.101618655082187,.0461642592626141,-.110133081351697,.0238890356585565,.117339115691434,.123132528211026,-.0267087538273405,-.134884926556002,2.8024998850757,-.607892005333882,-2.98586813553316,-3.1332901246654,.6796439236741,3.43234736114769,-3.77550789188145,.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-1.2742056900953,.27638875426455,1.35757727865237,-.616732719037493,-1.21277626178114,1.56057688324205,-.338505709325782,-1.66268580954775,.755340234698916,-1.80199896789671,.390872724665213,1.91990419952445,2.01469609360342,-.437008991543009,-2.20698899398908,-1.62700976129611,.352915706379786,1.73346540682961,-.78749464258244,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933792,.455612217872568,2.30093927043685,3.19783168117628,-.693643672830323,-3.40706658739604,-3.57528450949468,.775517202558345,3.91652795089531,-.49602746920004,.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-.433821252801768,.0941004396444027,.462206283002972,-.209975330432464,-.412906739739006,.531320354197217,-.115249030860378,-.566084774633034,.257166209078973,-.613515899259544,.133078117935277,.653658394065701,.68593162794344,-.148785859041837,-.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.3691862575809,.730812301668709,3.69075862820297,1.36899683098764,-.296949960039812,-1.45857062726747,-1.53058498738027,.332000148390628,1.67667184761831,3.00982432351404,-.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,.433821252801768,-.0941004396444027,-.462206283002972,.209975330432464,.412906739739006,-.531320354197217,.115249030860378,.566084774633034,-.257166209078973,.613515899259544,-.133078117935277,-.653658394065701,-.68593162794344,.148785859041837,.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.36918625758091,.730812301668709,3.69075862820298,-1.36899683098765,.296949960039813,1.45857062726747,1.53058498738027,-.332000148390629,-1.67667184761832,3.00982432351407,-.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,1.2742056900953,-.27638875426455,-1.35757727865237,.616732719037493,1.21277626178114,-1.56057688324205,.338505709325782,1.66268580954775,-.755340234698916,1.80199896789671,-.390872724665213,-1.91990419952445,-2.01469609360342,.437008991543009,2.20698899398908,-1.62700976129612,.352915706379786,1.73346540682961,-.787494642582441,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933793,.455612217872569,2.30093927043686,-3.1978316811763,.693643672830326,3.40706658739606,3.5752845094947,-.775517202558348,-3.91652795089533,-.496027469200023,.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.03452907455009,-.441310975771777,-2.16764880728185,.984739479553152,1.93644447258226,-2.49177904847469,.540493354313145,2.65481675969736,-1.20605462730079,2.87725860980962,-.624107966995901,-3.06551834185512,-3.21687291987534,.697773919897946,3.52390772571971,.0953780462518019,-.0206885117513148,-.101618655082182,.0461642592626119,-.110133081351691,.0238890356585554,.117339115691428,.12313252821102,-.0267087538273392,-.134884926555995,-2.8024998850757,.607892005333881,2.98586813553315,3.13329012466539,-.679643923674099,-3.43234736114768,-3.77550789188148,.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.66702109783921,-.578505216668687,-2.84152493765918,1.29087413922765,2.53844406932994,-3.26642040979696,.70852129725097,3.48014309433045,-1.58099148171138,3.77173740709444,-.818129923100605,-4.01852310483628,-4.21693061735794,.914697061432424,4.61941604516583,2.13132645129358,-.462307355479535,-2.27077966086395,1.03159071439797,-2.46104380066281,.533826552292296,2.62207049601641,2.75153061636384,-.59683622961842,-3.0141507731309,-.104423971232209,.0226506690028144,.111256457118342,.116749549066043,-.0253242178213231,-.127892723210497,-2.94703586087065,.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.13195404884644,-.679354114255872,-3.33687856485652,1.51590795070228,2.98096261298772,-3.83584465679206,.832035467367304,4.08682490876582,-1.85660048821863,4.42925189189755,-.960751801781151,-4.71905892262039,-4.95205415939119,1.07415316937022,5.42470353822284,3.97724364248012,-.862706409586761,-4.23747566411649,1.92503950203906,-4.59252537503603,.996167555656711,4.8930154290003,5.13459946240544,-1.1137491858232,-5.62467189980938,3.67768288095471,-.797728497114731,-3.91831456811703,-4.11177446033511,.891887573859297,4.5042232468252,2.21014640326349,-.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.40022577929605,-.737545103324376,-3.62270337995047,1.64575508217185,3.2363009692086,-4.1644090826311,.903304582796468,4.43688738516868,-2.01563009655897,4.80864541143983,-1.04304628770802,-5.12327625260686,-5.37622900930936,1.16616120174253,5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375093,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394475,-7.32142511479571,6.67435503772228,-1.4477385315153,-7.11105971439917,-7.46215578440284,1.61862088558445,8.17438210188188,7.8090615264033,-1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.40022577929605,-.737545103324376,3.62270337995047,1.64575508217186,-3.23630096920864,4.1644090826311,.903304582796468,-4.43688738516868,-2.01563009655898,-4.80864541143983,-1.04304628770802,5.12327625260686,5.37622900930936,1.16616120174253,-5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575887,-5.97791857375092,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394474,-7.32142511479571,-6.67435503772231,-1.44773853151531,7.11105971439921,7.46215578440288,1.61862088558446,-8.17438210188192,7.8090615264033,1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.13195404884644,-.679354114255872,3.33687856485653,1.51590795070228,-2.98096261298775,3.83584465679206,.832035467367304,-4.08682490876582,-1.85660048821863,-4.42925189189755,-.960751801781151,4.71905892262039,4.95205415939119,1.07415316937022,-5.42470353822284,3.97724364248013,.862706409586763,-4.2374756641165,-1.92503950203907,-4.59252537503604,-.996167555656712,4.89301542900031,5.13459946240545,1.1137491858232,-5.62467189980939,-3.67768288095471,-.797728497114731,3.91831456811704,4.11177446033512,.891887573859297,-4.5042232468252,2.21014640326355,.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.66702109783921,-.578505216668687,2.84152493765918,1.29087413922765,-2.53844406932997,3.26642040979696,.70852129725097,-3.48014309433046,-1.58099148171138,-3.77173740709444,-.818129923100605,4.01852310483628,4.21693061735794,.914697061432424,-4.61941604516583,2.13132645129358,.462307355479536,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292297,2.62207049601641,2.75153061636385,.596836229618421,-3.01415077313091,.104423971232213,.0226506690028153,-.111256457118347,-.116749549066047,-.0253242178213241,.127892723210502,-2.94703586087065,-.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.03452907455009,-.441310975771777,2.16764880728185,.984739479553153,-1.93644447258228,2.49177904847469,.540493354313145,-2.65481675969737,-1.20605462730079,-2.87725860980962,-.624107966995901,3.06551834185512,3.21687291987534,.697773919897946,-3.52390772571971,.0953780462518065,.0206885117513158,-.101618655082187,-.0461642592626142,-.110133081351697,-.0238890356585565,.117339115691434,.123132528211026,.0267087538273405,-.134884926556002,2.8024998850757,.607892005333882,-2.98586813553316,-3.1332901246654,-.6796439236741,3.43234736114769,-3.77550789188145,-.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-1.2742056900953,-.27638875426455,1.35757727865237,.616732719037494,-1.21277626178116,1.56057688324205,.338505709325782,-1.66268580954775,-.755340234698916,-1.80199896789671,-.390872724665213,1.91990419952445,2.01469609360342,.437008991543009,-2.20698899398908,-1.62700976129611,-.352915706379786,1.73346540682961,.787494642582441,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933792,-.455612217872568,2.30093927043685,3.19783168117628,.693643672830323,-3.40706658739604,-3.57528450949468,-.775517202558345,3.91652795089531,-.49602746920004,-.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-.433821252801768,-.0941004396444027,.462206283002972,.209975330432464,-.41290673973901,.531320354197217,.115249030860378,-.566084774633034,-.257166209078973,-.613515899259544,-.133078117935277,.653658394065701,.68593162794344,.148785859041837,-.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353978,-3.3691862575809,-.730812301668709,3.69075862820297,1.36899683098764,.296949960039812,-1.45857062726747,-1.53058498738027,-.332000148390628,1.67667184761831,3.00982432351404,.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,.433821252801768,.0941004396444027,-.462206283002972,-.209975330432464,.41290673973901,-.531320354197217,-.115249030860378,.566084774633034,.257166209078973,.613515899259544,.133078117935277,-.653658394065701,-.68593162794344,-.148785859041837,.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353979,-3.36918625758091,-.730812301668709,3.69075862820298,-1.36899683098765,-.296949960039813,1.45857062726747,1.53058498738027,.332000148390629,-1.67667184761832,3.00982432351407,.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,1.2742056900953,.27638875426455,-1.35757727865237,-.616732719037494,1.21277626178116,-1.56057688324205,-.338505709325782,1.66268580954775,.755340234698916,1.80199896789671,.390872724665213,-1.91990419952445,-2.01469609360342,-.437008991543009,2.20698899398908,-1.62700976129612,-.352915706379786,1.73346540682961,.787494642582442,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933793,-.455612217872569,2.30093927043686,-3.1978316811763,-.693643672830326,3.40706658739606,3.5752845094947,.775517202558348,-3.91652795089533,-.496027469200023,-.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.03452907455009,.441310975771777,-2.16764880728185,-.984739479553153,1.93644447258228,-2.49177904847469,-.540493354313145,2.65481675969737,1.20605462730079,2.87725860980962,.624107966995901,-3.06551834185512,-3.21687291987534,-.697773919897946,3.52390772571971,.0953780462518019,.0206885117513148,-.101618655082182,-.046164259262612,-.110133081351691,-.0238890356585554,.117339115691428,.12313252821102,.0267087538273392,-.134884926555995,-2.8024998850757,-.607892005333881,2.98586813553315,3.13329012466539,.679643923674099,-3.43234736114768,-3.77550789188148,-.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.66702109783921,.578505216668687,-2.84152493765918,-1.29087413922765,2.53844406932997,-3.26642040979696,-.70852129725097,3.48014309433046,1.58099148171138,3.77173740709444,.818129923100605,-4.01852310483628,-4.21693061735794,-.914697061432424,4.61941604516583,2.13132645129358,.462307355479535,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292296,2.62207049601641,2.75153061636384,.59683622961842,-3.0141507731309,-.104423971232209,-.0226506690028144,.111256457118342,.116749549066043,.0253242178213231,-.127892723210497,-2.94703586087065,-.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.13195404884644,.679354114255872,-3.33687856485653,-1.51590795070228,2.98096261298775,-3.83584465679206,-.832035467367304,4.08682490876582,1.85660048821863,4.42925189189755,.960751801781151,-4.71905892262039,-4.95205415939119,-1.07415316937022,5.42470353822284,3.97724364248012,.862706409586761,-4.23747566411649,-1.92503950203906,-4.59252537503603,-.996167555656711,4.89301542900031,5.13459946240544,1.1137491858232,-5.62467189980938,3.67768288095471,.797728497114731,-3.91831456811703,-4.11177446033511,-.891887573859297,4.5042232468252,2.21014640326349,.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.40022577929605,.737545103324376,-3.62270337995047,-1.64575508217186,3.23630096920864,-4.1644090826311,-.903304582796468,4.43688738516868,2.01563009655898,4.80864541143983,1.04304628770802,-5.12327625260686,-5.37622900930936,-1.16616120174253,5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575888,-5.97791857375093,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394475,-7.32142511479571,6.67435503772228,1.4477385315153,-7.11105971439918,-7.46215578440284,-1.61862088558445,8.17438210188188,7.8090615264033,1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.40022577929605,-2.16629351671545,2.25851140965319,3.84430965952583,.533351454031679,4.1644090826311,2.65315687476095,-2.76610026595298,-4.70829853978449,-4.80864541143983,-3.06360167061167,3.19401746644684,5.37622900930936,3.42521079645606,-5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375092,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,-6.67435503772231,-4.25225058180805,4.43326649059392,7.46215578440288,4.75416068069224,-8.17438210188192,7.8090615264033,4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.13195404884644,-1.99537683408524,2.08031890026245,3.54100050546176,.491270978557926,3.83584465679206,2.4438275442558,-2.54785990395621,-4.33682220887954,-4.42925189189755,-2.82188898006397,2.94201520294205,4.95205415939119,3.15496779285464,-5.42470353822284,3.97724364248013,2.53391323933003,-2.64178049699053,-4.4966885005097,-4.59252537503604,-2.92591098178889,3.05046536107998,5.13459946240545,3.27126792601698,-5.62467189980939,-3.67768288095471,-2.34306222576705,2.44280506311384,4.11177446033512,2.61962320701725,-4.5042232468252,2.21014640326355,1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.66702109783921,-1.69916672838959,1.77149929746803,3.01534534295106,.418342684514381,3.26642040979696,2.08104573641859,-2.16963467924844,-3.69302874444135,-3.77173740709444,-2.40298463138773,2.50527833232424,4.21693061735794,2.68661849290085,-4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207907,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297245,2.75153061636385,1.75300798410243,-3.01415077313091,.104423971232213,.0665288091384509,-.0693609030179037,-.116749549066047,-.0743814698720828,.127892723210502,-2.94703586087065,-1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.03452907455009,-1.29620426108275,1.35138294525072,2.3002471840262,.319131466736208,2.49177904847469,1.58751952117754,-1.65509933148239,-2.8172159417121,-2.87725860980962,-1.83310964514541,1.91114408921759,3.21687291987534,2.04947888880939,-3.52390772571971,.0953780462518065,.0607656246043247,-.0633523829764579,-.107834822891096,-.110133081351697,-.0701660994486395,.07315303070475,.123132528211026,.0784480840492681,-.134884926556002,2.8024998850757,1.78548064936278,-1.86148755350589,-3.1332901246654,-1.99622805289337,3.43234736114769,-3.77550789188145,-2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-1.2742056900953,-.811800069931489,.846357940948568,1.4406223470461,.199868920965732,1.56057688324205,.994247972331933,-1.03657254753862,-1.76439483124642,-1.80199896789671,-1.14805866852966,1.19693087876444,2.01469609360342,1.28356861276539,-2.20698899398908,-1.62700976129611,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.24788355079949,-2.10046056933792,-1.338209404774,2.30093927043685,3.19783168117628,2.03734766130249,-2.12407640207831,-3.57528450949468,-2.27782393297861,3.91652795089531,-.49602746920004,-.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-.433821252801768,-.27638875426455,.2881544676147,.490480144821061,.0680481859118103,.531320354197217,.338505709325782,-.352915706379786,-.600713041919503,-.613515899259544,-.390872724665213,.407511956177088,.68593162794344,.437008991543009,-.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.3691862575809,-2.14651815042224,3.69075862820297,1.36899683098764,.872191775558762,-.909320487484577,-1.53058498738027,-.975140050100603,1.67667184761831,3.00982432351404,1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,.433821252801768,.27638875426455,-.2881544676147,-.490480144821061,-.0680481859118103,-.531320354197217,-.338505709325782,.352915706379786,.600713041919503,.613515899259544,.390872724665213,-.407511956177088,-.68593162794344,-.437008991543009,.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.36918625758091,-2.14651815042224,3.69075862820298,-1.36899683098765,-.872191775558764,.909320487484578,1.53058498738027,.975140050100604,-1.67667184761832,3.00982432351407,1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,1.2742056900953,.811800069931489,-.846357940948568,-1.4406223470461,-.199868920965732,-1.56057688324205,-.994247972331933,1.03657254753862,1.76439483124642,1.80199896789671,1.14805866852966,-1.19693087876444,-2.01469609360342,-1.28356861276539,2.20698899398908,-1.62700976129612,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.2478835507995,-2.10046056933793,-1.338209404774,2.30093927043686,-3.1978316811763,-2.0373476613025,2.12407640207832,3.5752845094947,2.27782393297862,-3.91652795089533,-.496027469200023,-.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.03452907455009,1.29620426108275,-1.35138294525072,-2.3002471840262,-.319131466736208,-2.49177904847469,-1.58751952117754,1.65509933148239,2.8172159417121,2.87725860980962,1.83310964514541,-1.91114408921759,-3.21687291987534,-2.04947888880939,3.52390772571971,.0953780462518019,.0607656246043218,-.0633523829764548,-.10783482289109,-.110133081351691,-.0701660994486362,.0731530307047465,.12313252821102,.0784480840492644,-.134884926555995,-2.8024998850757,-1.78548064936277,1.86148755350589,3.13329012466539,1.99622805289336,-3.43234736114768,-3.77550789188148,-2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.66702109783921,1.69916672838959,-1.77149929746803,-3.01534534295106,-.418342684514381,-3.26642040979696,-2.08104573641859,2.16963467924844,3.69302874444135,3.77173740709444,2.40298463138773,-2.50527833232424,-4.21693061735794,-2.68661849290085,4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207906,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297244,2.75153061636384,1.75300798410242,-3.0141507731309,-.104423971232209,-.0665288091384484,.069360903017901,.116749549066043,.0743814698720799,-.127892723210497,-2.94703586087065,-1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.13195404884644,1.99537683408524,-2.08031890026245,-3.54100050546176,-.491270978557926,-3.83584465679206,-2.4438275442558,2.54785990395621,4.33682220887954,4.42925189189755,2.82188898006397,-2.94201520294205,-4.95205415939119,-3.15496779285464,5.42470353822284,3.97724364248012,2.53391323933002,-2.64178049699052,-4.49668850050969,-4.59252537503603,-2.92591098178889,3.05046536107998,5.13459946240544,3.27126792601697,-5.62467189980938,3.67768288095471,2.34306222576705,-2.44280506311384,-4.11177446033511,-2.61962320701725,4.5042232468252,2.21014640326349,1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.40022577929605,2.16629351671545,-2.25851140965319,-3.84430965952583,-.533351454031679,-4.1644090826311,-2.65315687476095,2.76610026595298,4.70829853978449,4.80864541143983,3.06360167061167,-3.19401746644684,-5.37622900930936,-3.42521079645606,5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375093,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,6.67435503772228,4.25225058180803,-4.4332664905939,-7.46215578440284,-4.75416068069222,8.17438210188188,7.8090615264033,4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.40022577929605,-3.45892910228435,-.132397734060625,3.36905705276318,4.05959901190685,4.1644090826311,4.23630567890486,.162153445774668,-4.1262353470069,-4.80864541143983,-4.89166444644702,-.187238671144264,5.37622900930936,5.46904711383883,-5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375092,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,-6.67435503772231,-6.78958468653045,-.259885531387094,7.46215578440288,7.59098645151066,-8.17438210188192,7.8090615264033,7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.13195404884644,-3.18602578468041,-.121951789723539,3.10324447907128,3.73930391312631,3.83584465679206,3.9020687402541,.149359829020976,-3.80068276060981,-4.42925189189755,-4.50572087357673,-.172465874990311,4.95205415939119,5.03754908153942,-5.42470353822284,3.97724364248013,4.04590891241332,.154865612009818,-3.94078558921028,-4.59252537503604,-4.67181319960176,-.178823405501163,5.13459946240545,5.22324594675678,-5.62467189980939,-3.67768288095471,-3.74117637274605,-.143201337715517,4.11177446033512,4.18276234400124,-4.5042232468252,2.21014640326355,2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.66702109783921,-2.71306598164557,-.103848265663963,2.64257341211137,3.18421096606247,3.26642040979696,3.32281364706153,.127187630774889,-3.23647823392352,-3.77173740709444,-3.83685470584401,-.146863625737381,4.21693061735794,4.28973397193188,-4.61941604516583,2.13132645129358,2.16812281562257,.0829894280920819,-2.11178929936586,-2.46104380066281,-2.50353258258141,-.0958279372639496,2.75153061636385,2.79903451954298,-3.01415077313091,.104423971232213,.106226802736416,.00406605268971361,-.116749549066047,-.118765176014247,.127892723210502,-2.94703586087065,-2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.03452907455009,-2.0696542765645,-.0792203391289677,2.01587923055033,2.4290658199907,2.49177904847469,2.53479846100042,.0970247040581356,-2.46893774908671,-2.87725860980962,-2.92693314656882,-.112034478016925,3.21687291987534,3.2724107413517,-3.52390772571971,.0953780462518065,.0970247040581402,.00371382314810264,-.0945037478171113,-.110133081351697,-.112034478016931,-.00428835358704851,.123132528211026,.12525835434705,-.134884926556002,2.8024998850757,2.85088374774363,.109123528450132,-3.1332901246654,-3.18738492899079,3.43234736114769,-3.77550789188145,-3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-1.2742056900953,-1.29620426108275,-.0496149247273511,1.2625254749334,1.52130020070252,1.56057688324205,1.58751952117754,.0607656246043218,-1.54627160050443,-1.80199896789671,-1.83310964514541,-.0701660994486362,2.01469609360342,2.04947888880939,-2.20698899398908,-1.62700976129611,-1.65509933148238,-.0633523829764547,1.6120955106542,1.87870904723461,1.91114408921758,.0731530307047464,-2.10046056933792,-2.13672404935301,2.30093927043685,3.19783168117628,3.25304076422429,.124516928010559,-3.57528450949468,-3.63701014237694,3.91652795089531,-.49602746920004,-.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-.433821252801768,-.441310975771777,-.0168920991094264,.42984455923187,.5179480550797,.531320354197217,.540493354313145,.0206885117513148,-.526449919441544,-.613515899259544,-.624107966995901,-.0238890356585554,.68593162794344,.697773919897946,-.751400451068813,-2.60976045384023,-2.65481675969736,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.3691862575809,-3.42735369966595,3.69075862820297,1.36899683098764,1.39263192728723,.0533058950082293,-1.53058498738027,-1.5570098290328,1.67667184761831,3.00982432351404,3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,.433821252801768,.441310975771777,.0168920991094264,-.42984455923187,-.5179480550797,-.531320354197217,-.540493354313145,-.0206885117513148,.526449919441544,.613515899259544,.624107966995901,.0238890356585554,-.68593162794344,-.697773919897946,.751400451068813,-2.60976045384023,-2.65481675969737,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.36918625758091,-3.42735369966595,3.69075862820298,-1.36899683098765,-1.39263192728723,-.0533058950082294,1.53058498738027,1.5570098290328,-1.67667184761832,3.00982432351407,3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,1.2742056900953,1.29620426108275,.0496149247273511,-1.2625254749334,-1.52130020070252,-1.56057688324205,-1.58751952117754,-.0607656246043218,1.54627160050443,1.80199896789671,1.83310964514541,.0701660994486362,-2.01469609360342,-2.04947888880939,2.20698899398908,-1.62700976129612,-1.65509933148239,-.0633523829764548,1.61209551065421,1.87870904723461,1.91114408921759,.0731530307047465,-2.10046056933793,-2.13672404935301,2.30093927043686,-3.1978316811763,-3.2530407642243,-.12451692801056,3.5752845094947,3.63701014237696,-3.91652795089533,-.496027469200023,-.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.03452907455009,2.0696542765645,.0792203391289677,-2.01587923055033,-2.4290658199907,-2.49177904847469,-2.53479846100042,-.0970247040581356,2.46893774908671,2.87725860980962,2.92693314656882,.112034478016925,-3.21687291987534,-3.2724107413517,3.52390772571971,.0953780462518019,.0970247040581356,.00371382314810246,-.0945037478171067,-.110133081351691,-.112034478016925,-.0042883535870483,.12313252821102,.125258354347044,-.134884926555995,-2.8024998850757,-2.85088374774362,-.109123528450131,3.13329012466539,3.18738492899078,-3.43234736114768,-3.77550789188148,-3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.66702109783921,2.71306598164557,.103848265663963,-2.64257341211137,-3.18421096606247,-3.26642040979696,-3.32281364706153,-.127187630774889,3.23647823392352,3.77173740709444,3.83685470584401,.146863625737381,-4.21693061735794,-4.28973397193188,4.61941604516583,2.13132645129358,2.16812281562256,.0829894280920818,-2.11178929936586,-2.46104380066281,-2.5035325825814,-.0958279372639495,2.75153061636384,2.79903451954297,-3.0141507731309,-.104423971232209,-.106226802736412,-.00406605268971345,.116749549066043,.118765176014243,-.127892723210497,-2.94703586087065,-2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.13195404884644,3.18602578468041,.121951789723539,-3.10324447907128,-3.73930391312631,-3.83584465679206,-3.9020687402541,-.149359829020976,3.80068276060981,4.42925189189755,4.50572087357673,.172465874990311,-4.95205415939119,-5.03754908153942,5.42470353822284,3.97724364248012,4.04590891241331,.154865612009818,-3.94078558921027,-4.59252537503603,-4.67181319960176,-.178823405501163,5.13459946240544,5.22324594675677,-5.62467189980938,3.67768288095471,3.74117637274605,.143201337715517,-4.11177446033511,-4.18276234400124,4.5042232468252,2.21014640326349,2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.40022577929605,3.45892910228435,.132397734060625,-3.36905705276318,-4.05959901190685,-4.1644090826311,-4.23630567890486,-.162153445774668,4.1262353470069,4.80864541143983,4.89166444644702,.187238671144264,-5.37622900930936,-5.46904711383883,5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375093,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,6.67435503772228,6.78958468653042,.259885531387093,-7.46215578440284,-7.59098645151062,8.17438210188188,7.8090615264033,7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.40022577929605,-4.53423694314245,-2.95857174458961,.125534462510035,3.16878793832491,4.1644090826311,5.55328344227955,3.62349557083131,-.153747689149873,-4.80864541143983,-6.41237937831348,-4.18405228663707,5.37622900930936,7.16925809522303,-5.88936380548032,5.17702934684997,6.90362324406996,4.50458697357606,-.191133071440498,-5.97791857375092,-7.97161747704437,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,-6.67435503772231,-8.90032284672419,-5.80742559866953,7.46215578440288,9.95086345672774,-8.17438210188192,7.8090615264033,10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.13195404884644,-4.17649376078907,-2.72514572729005,.115630018018823,2.91877623944935,3.83584465679206,5.11513931437816,3.33760825329423,-.141617271554661,-4.42925189189755,-5.90645411811629,-3.85393804714673,4.95205415939119,6.60361645843633,-5.42470353822284,3.97724364248013,5.30369635131817,3.46064097812297,-.146837644201682,-4.59252537503604,-6.12418103253549,-3.99600399916796,5.13459946240545,6.84704254830277,-5.62467189980939,-3.67768288095471,-4.90422891550851,-3.19999004936568,4.11177446033512,5.48309461793542,-4.5042232468252,2.21014640326355,2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.66702109783921,-3.55650076638925,-2.32060274065832,.0984649496097535,2.4854891512059,3.26642040979696,4.35580607412106,2.84214630515942,-.120594442052501,-3.77173740709444,-5.02965161709108,-3.28182786886397,4.21693061735794,5.62332146066281,-4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258921,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636385,3.66919510297595,-3.01415077313091,.104423971232213,.139250467146397,.0908603812984509,-.116749549066047,-.155686755269708,.127892723210502,-2.94703586087065,-3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.03452907455009,-2.71306598164557,-1.77026486598667,.0751136925641335,1.89604797153805,2.49177904847469,3.32281364706153,2.16812281562256,-.0919951097438786,-2.87725860980962,-3.83685470584401,-2.5035325825814,3.21687291987534,4.28973397193188,-3.52390772571971,.0953780462518065,.127187630774895,.0829894280920858,-.00352130492348742,-.110133081351697,-.146863625737388,-.095827937263954,.123132528211026,.164198525301527,-.134884926556002,2.8024998850757,3.73716315870324,2.43848424018078,-3.1332901246654,-4.17827543429579,3.43234736114769,-3.77550789188145,-5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-1.2742056900953,-1.69916672838959,-1.10869959708726,.0470429720894763,1.18747632769106,1.56057688324205,2.08104573641859,1.35787414544697,-.0576156388045296,-1.80199896789671,-2.40298463138773,-1.56793800686758,2.01469609360342,2.68661849290085,-2.20698899398908,-1.62700976129611,-2.16963467924844,-1.41567808207906,.0600683040611726,1.87870904723461,2.50527833232423,1.63468424297244,-2.10046056933792,-2.8009863270915,2.30093927043685,3.19783168117628,4.26434227893045,2.78246653954404,-3.57528450949468,-4.76767960906116,3.91652795089531,-.49602746920004,-.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-.433821252801768,-.578505216668687,-.377472375086662,.0160164416514644,.404293021256913,.531320354197217,.70852129725097,.462307355479535,-.0196160547715697,-.613515899259544,-.818129923100605,-.533826552292296,.68593162794344,.914697061432424,-.751400451068813,-2.60976045384023,-3.48014309433045,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.3691862575809,-4.49284541622379,3.69075862820297,1.36899683098764,1.82557171487995,1.19117835294064,-1.53058498738027,-2.04105122680137,1.67667184761831,3.00982432351404,4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,.433821252801768,.578505216668687,.377472375086662,-.0160164416514644,-.404293021256913,-.531320354197217,-.70852129725097,-.462307355479535,.0196160547715697,.613515899259544,.818129923100605,.533826552292296,-.68593162794344,-.914697061432424,.751400451068813,-2.60976045384023,-3.48014309433046,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.36918625758091,-4.49284541622379,3.69075862820298,-1.36899683098765,-1.82557171487995,-1.19117835294064,1.53058498738027,2.04105122680137,-1.67667184761832,3.00982432351407,4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,1.2742056900953,1.69916672838959,1.10869959708726,-.0470429720894763,-1.18747632769106,-1.56057688324205,-2.08104573641859,-1.35787414544697,.0576156388045296,1.80199896789671,2.40298463138773,1.56793800686758,-2.01469609360342,-2.68661849290085,2.20698899398908,-1.62700976129612,-2.16963467924844,-1.41567808207906,.0600683040611727,1.87870904723461,2.50527833232424,1.63468424297244,-2.10046056933793,-2.8009863270915,2.30093927043686,-3.1978316811763,-4.26434227893047,-2.78246653954406,3.5752845094947,4.76767960906118,-3.91652795089533,-.496027469200023,-.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.03452907455009,2.71306598164557,1.77026486598667,-.0751136925641335,-1.89604797153805,-2.49177904847469,-3.32281364706153,-2.16812281562256,.0919951097438786,2.87725860980962,3.83685470584401,2.5035325825814,-3.21687291987534,-4.28973397193188,3.52390772571971,.0953780462518019,.127187630774889,.0829894280920818,-.00352130492348725,-.110133081351691,-.146863625737381,-.0958279372639495,.12313252821102,.164198525301519,-.134884926555995,-2.8024998850757,-3.73716315870323,-2.43848424018078,3.13329012466539,4.17827543429578,-3.43234736114768,-3.77550789188148,-5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.66702109783921,3.55650076638925,2.32060274065832,-.0984649496097535,-2.4854891512059,-3.26642040979696,-4.35580607412106,-2.84214630515942,.120594442052501,3.77173740709444,5.02965161709108,3.28182786886397,-4.21693061735794,-5.62332146066281,4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258919,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636384,3.66919510297594,-3.0141507731309,-.104423971232209,-.139250467146391,-.0908603812984474,.116749549066043,.155686755269702,-.127892723210497,-2.94703586087065,-3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.13195404884644,4.17649376078907,2.72514572729005,-.115630018018823,-2.91877623944935,-3.83584465679206,-5.11513931437816,-3.33760825329423,.141617271554661,4.42925189189755,5.90645411811629,3.85393804714673,-4.95205415939119,-6.60361645843633,5.42470353822284,3.97724364248012,5.30369635131817,3.46064097812296,-.146837644201681,-4.59252537503603,-6.12418103253548,-3.99600399916795,5.13459946240544,6.84704254830276,-5.62467189980938,3.67768288095471,4.90422891550851,3.19999004936568,-4.11177446033511,-5.48309461793542,4.5042232468252,2.21014640326349,2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.40022577929605,4.53423694314245,2.95857174458961,-.125534462510035,-3.16878793832491,-4.1644090826311,-5.55328344227955,-3.62349557083131,.153747689149873,4.80864541143983,6.41237937831348,4.18405228663707,-5.37622900930936,-7.16925809522303,5.88936380548032,5.17702934684997,6.90362324406997,4.50458697357606,-.191133071440498,-5.97791857375093,-7.97161747704438,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,6.67435503772228,8.90032284672415,5.8074255986695,-7.46215578440284,-9.95086345672769,8.17438210188188,7.8090615264033,10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.40022577929605,-5.32467544557844,-5.5209565174066,-4.42116822694256,-2.37645064234965,4.1644090826311,6.52136894437417,6.76176317987178,5.41480311178231,-4.80864541143983,-7.53022822838542,-7.80781158453304,5.37622900930936,8.41905110415162,-5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394288,-5.97791857375092,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,-6.67435503772231,-10.4518866380257,-10.8371697481941,7.46215578440288,11.6855645116818,-8.17438210188192,7.8090615264033,12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.13195404884644,-4.90456807960123,-5.08536292603975,-4.0723459639997,-2.18895293851093,3.83584465679206,6.00684460241421,6.22827216283403,4.98758483419414,-4.42925189189755,-6.93610669392436,-7.19178921991206,4.95205415939119,7.75480303503604,-5.42470353822284,3.97724364248013,6.22827216283404,6.45786210779302,5.17143989055336,-4.59252537503604,-7.19178921991208,-7.45689685002024,5.13459946240545,8.04066478857437,-5.62467189980939,-3.67768288095471,-5.75916689254577,-5.97146442400447,4.11177446033512,6.43894433484766,-4.5042232468252,2.21014640326355,3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.66702109783921,-4.17649376078907,-4.3304499371288,-3.46781352289888,-1.86400680793386,3.26642040979696,5.11513931437816,5.30369635131817,4.24718682732847,-3.77173740709444,-5.90645411811629,-6.12418103253548,4.21693061735794,6.60361645843633,-4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812297,2.77127267544969,-2.46104380066281,-3.85393804714674,-3.99600399916796,2.75153061636385,4.3088337276685,-3.01415077313091,.104423971232213,.163525539687803,.169553506762539,-.116749549066047,-.182827111459215,.127892723210502,-2.94703586087065,-4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.03452907455009,-3.18602578468041,-3.30347079372946,-2.6454111829756,-1.4219520231667,2.49177904847469,3.9020687402541,4.04590891241331,3.23995377923585,-2.87725860980962,-4.50572087357673,-4.67181319960176,3.21687291987534,5.03754908153942,-3.52390772571971,.0953780462518065,.149359829020983,.154865612009825,.12401599635522,-.110133081351697,-.172465874990319,-.178823405501172,.123132528211026,.192822710157554,-.134884926556002,2.8024998850757,4.38865042934355,4.55042724011643,-3.1332901246654,-4.90666034634693,3.43234736114769,-3.77550789188145,-5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-1.2742056900953,-1.99537683408524,-2.06893149627982,-1.65679518870216,-.890554665266804,1.56057688324205,2.4438275442558,2.53391323933002,2.02915141041227,-1.80199896789671,-2.82188898006397,-2.92591098178889,2.01469609360342,3.15496779285464,-2.20698899398908,-1.62700976129611,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294204,3.05046536107998,-2.10046056933792,-3.2892729926303,2.30093927043685,3.19783168117628,5.00773093883056,5.19232863088549,-3.57528450949468,-5.59881339795158,3.91652795089531,-.49602746920004,-.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-.433821252801768,-.679354114255872,-.704396833771808,-.564079229896515,-.303201864171246,.531320354197217,.832035467367304,.862706409586761,.690853143909355,-.613515899259544,-.960751801781151,-.996167555656711,.68593162794344,1.07415316937022,-.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.8930154290003,-3.3691862575809,-5.27606827091984,3.69075862820297,1.36899683098764,2.14381758303679,2.22284414873079,-1.53058498738027,-2.39686092429589,1.67667184761831,3.00982432351404,4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,.433821252801768,.679354114255872,.704396833771808,.564079229896515,.303201864171246,-.531320354197217,-.832035467367304,-.862706409586761,-.690853143909355,.613515899259544,.960751801781151,.996167555656711,-.68593162794344,-1.07415316937022,.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.89301542900031,-3.36918625758091,-5.27606827091984,3.69075862820298,-1.36899683098765,-2.1438175830368,-2.22284414873079,1.53058498738027,2.3968609242959,-1.67667184761832,3.00982432351407,4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,1.2742056900953,1.99537683408524,2.06893149627982,1.65679518870216,.890554665266804,-1.56057688324205,-2.4438275442558,-2.53391323933002,-2.02915141041227,1.80199896789671,2.82188898006397,2.92591098178889,-2.01469609360342,-3.15496779285464,2.20698899398908,-1.62700976129612,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294205,3.05046536107998,-2.10046056933793,-3.28927299263031,2.30093927043686,-3.1978316811763,-5.00773093883059,-5.19232863088551,3.5752845094947,5.5988133979516,-3.91652795089533,-.496027469200023,-.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.03452907455009,3.18602578468041,3.30347079372946,2.6454111829756,1.4219520231667,-2.49177904847469,-3.9020687402541,-4.04590891241331,-3.23995377923585,2.87725860980962,4.50572087357673,4.67181319960176,-3.21687291987534,-5.03754908153942,3.52390772571971,.0953780462518019,.149359829020976,.154865612009818,.124015996355214,-.110133081351691,-.172465874990311,-.178823405501163,.12313252821102,.192822710157544,-.134884926555995,-2.8024998850757,-4.38865042934354,-4.55042724011642,3.13329012466539,4.90666034634692,-3.43234736114768,-3.77550789188148,-5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.66702109783921,4.17649376078907,4.3304499371288,3.46781352289888,1.86400680793386,-3.26642040979696,-5.11513931437816,-5.30369635131817,-4.24718682732847,3.77173740709444,5.90645411811629,6.12418103253548,-4.21693061735794,-6.60361645843633,4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812296,2.77127267544968,-2.46104380066281,-3.85393804714673,-3.99600399916795,2.75153061636384,4.30883372766849,-3.0141507731309,-.104423971232209,-.163525539687797,-.169553506762532,.116749549066043,.182827111459208,-.127892723210497,-2.94703586087065,-4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.13195404884644,4.90456807960123,5.08536292603975,4.0723459639997,2.18895293851093,-3.83584465679206,-6.00684460241421,-6.22827216283403,-4.98758483419414,4.42925189189755,6.93610669392436,7.19178921991206,-4.95205415939119,-7.75480303503604,5.42470353822284,3.97724364248012,6.22827216283403,6.45786210779301,5.17143989055335,-4.59252537503603,-7.19178921991206,-7.45689685002023,5.13459946240544,8.04066478857435,-5.62467189980938,3.67768288095471,5.75916689254577,5.97146442400446,-4.11177446033511,-6.43894433484766,4.5042232468252,2.21014640326349,3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.40022577929605,5.32467544557844,5.5209565174066,4.42116822694256,2.37645064234965,-4.1644090826311,-6.52136894437417,-6.76176317987178,-5.41480311178231,4.80864541143983,7.53022822838542,7.80781158453304,-5.37622900930936,-8.41905110415162,5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394289,-5.97791857375093,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,6.67435503772228,10.4518866380257,10.8371697481941,-7.46215578440284,-11.6855645116817,8.17438210188188,7.8090615264033,12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.40022577929605,-5.78076767221695,-7.18642267926874,-8.02365168049822,-8.3966606253625,4.1644090826311,7.0799655598807,8.80153432008931,9.82692624602171,-4.80864541143983,-8.17524004082037,-10.1631364184159,5.37622900930936,9.14019623375092,-5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375092,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,-6.67435503772231,-11.3471570254948,-14.1063386773585,7.46215578440288,12.6865072343198,-8.17438210188192,7.8090615264033,13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.13195404884644,-5.32467544557844,-6.61942678751097,-7.39059991847735,-7.7341791249633,3.83584465679206,6.52136894437417,8.10710900955867,9.05159934712187,-4.42925189189755,-7.53022822838542,-9.36128313844683,4.95205415939119,8.41905110415162,-5.42470353822284,3.97724364248013,6.7617631798718,8.40595765692319,9.38526430991042,-4.59252537503604,-7.80781158453306,-9.70636382862251,5.13459946240545,8.72939873011817,-5.62467189980939,-3.67768288095471,-6.25247606762144,-7.77283197754114,4.11177446033512,6.99048075972417,-4.5042232468252,2.21014640326355,3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.66702109783921,-4.53423694314245,-5.63678477479456,-6.29347864012493,-6.58605413075653,3.26642040979696,5.55328344227955,6.90362324406997,7.70790568809692,-3.77173740709444,-6.41237937831348,-7.97161747704438,4.21693061735794,7.16925809522303,-4.61941604516583,2.13132645129358,3.62349557083132,4.50458697357607,5.02937810056318,-2.46104380066281,-4.18405228663708,-5.2014490017318,2.75153061636385,4.67791266762517,-3.01415077313091,.104423971232213,.177532539414032,.220701460427408,-.116749549066047,-.198487413238653,.127892723210502,-2.94703586087065,-5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.03452907455009,-3.45892910228435,-4.30000442088448,-4.80096137363441,-5.0241517123509,2.49177904847469,4.23630567890486,5.26640836144105,5.87995282040644,-2.87725860980962,-4.89166444644702,-6.08112457054741,3.21687291987534,5.46904711383883,-3.52390772571971,.0953780462518065,.162153445774676,.201582776983911,.2250674723256,-.110133081351697,-.187238671144273,-.232767741029294,.123132528211026,.209339198368165,-.134884926556002,2.8024998850757,4.76456617608066,5.92312100550577,-3.1332901246654,-5.32694692824228,3.43234736114769,-3.77550789188145,-6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-1.2742056900953,-2.16629351671545,-2.69305077477823,-3.00679522191911,-3.14657715137113,1.56057688324205,2.65315687476095,3.2983001248078,3.68255702755716,-1.80199896789671,-3.06360167061167,-3.80854893001879,2.01469609360342,3.42521079645606,-2.20698899398908,-1.62700976129611,-2.76610026595298,-3.43870690112598,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865961,-2.10046056933792,-3.57102008849817,2.30093927043685,3.19783168117628,5.43667485806875,6.75866004623775,-3.57528450949468,-6.07838727908375,3.91652795089531,-.49602746920004,-.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-.433821252801768,-.737545103324376,-.916887022287339,-1.02370573309381,-1.07129645743706,.531320354197217,.903304582796468,1.12295267819227,1.25377834648448,-.613515899259544,-1.04304628770802,-1.29667406213985,.68593162794344,1.16616120174253,-.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.3691862575809,-5.72799698473071,3.69075862820297,1.36899683098764,2.32744915738298,2.89339311993396,-1.53058498738027,-2.60216726588949,1.67667184761831,3.00982432351404,5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,.433821252801768,.737545103324376,.916887022287339,1.02370573309381,1.07129645743706,-.531320354197217,-.903304582796468,-1.12295267819227,-1.25377834648448,.613515899259544,1.04304628770802,1.29667406213985,-.68593162794344,-1.16616120174253,.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.36918625758091,-5.72799698473071,3.69075862820298,-1.36899683098765,-2.32744915738298,-2.89339311993396,1.53058498738027,2.60216726588949,-1.67667184761832,3.00982432351407,5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,1.2742056900953,2.16629351671545,2.69305077477823,3.00679522191911,3.14657715137113,-1.56057688324205,-2.65315687476095,-3.2983001248078,-3.68255702755716,1.80199896789671,3.06360167061167,3.80854893001879,-2.01469609360342,-3.42521079645606,2.20698899398908,-1.62700976129612,-2.76610026595298,-3.43870690112599,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865962,-2.10046056933793,-3.57102008849817,2.30093927043686,-3.1978316811763,-5.43667485806878,-6.75866004623778,3.5752845094947,6.07838727908378,-3.91652795089533,-.496027469200023,-.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.03452907455009,3.45892910228435,4.30000442088448,4.80096137363441,5.0241517123509,-2.49177904847469,-4.23630567890486,-5.26640836144105,-5.87995282040644,2.87725860980962,4.89166444644702,6.08112457054741,-3.21687291987534,-5.46904711383883,3.52390772571971,.0953780462518019,.162153445774668,.201582776983901,.225067472325589,-.110133081351691,-.187238671144264,-.232767741029283,.12313252821102,.209339198368155,-.134884926555995,-2.8024998850757,-4.76456617608065,-5.92312100550575,3.13329012466539,5.32694692824227,-3.43234736114768,-3.77550789188148,-6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.66702109783921,4.53423694314245,5.63678477479456,6.29347864012493,6.58605413075653,-3.26642040979696,-5.55328344227955,-6.90362324406997,-7.70790568809692,3.77173740709444,6.41237937831348,7.97161747704438,-4.21693061735794,-7.16925809522303,4.61941604516583,2.13132645129358,3.62349557083131,4.50458697357606,5.02937810056317,-2.46104380066281,-4.18405228663707,-5.20144900173179,2.75153061636384,4.67791266762516,-3.0141507731309,-.104423971232209,-.177532539414025,-.2207014604274,.116749549066043,.198487413238645,-.127892723210497,-2.94703586087065,-5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.13195404884644,5.32467544557844,6.61942678751097,7.39059991847735,7.7341791249633,-3.83584465679206,-6.52136894437417,-8.10710900955867,-9.05159934712187,4.42925189189755,7.53022822838542,9.36128313844683,-4.95205415939119,-8.41905110415162,5.42470353822284,3.97724364248012,6.76176317987178,8.40595765692318,9.3852643099104,-4.59252537503603,-7.80781158453304,-9.70636382862249,5.13459946240544,8.72939873011815,-5.62467189980938,3.67768288095471,6.25247606762144,7.77283197754114,-4.11177446033511,-6.99048075972417,4.5042232468252,2.21014640326349,3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.40022577929605,5.78076767221695,7.18642267926874,8.02365168049822,8.3966606253625,-4.1644090826311,-7.0799655598807,-8.80153432008931,-9.82692624602171,4.80864541143983,8.17524004082037,10.1631364184159,-5.37622900930936,-9.14019623375092,5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375093,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,6.67435503772228,11.3471570254948,14.1063386773585,-7.46215578440284,-12.6865072343197,8.17438210188188,7.8090615264033,13.2762861543698,-8.55439830449838,8.51143945889045}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,-.0319379183519753,.054298066631689,-.0675012177634823,.0753652106918314,-.07886883956794,-.154660030302817,.262939511058368,-.326876043550518,.364957592022403,-.43925822557731,.746788570738901,-.928378137184539,-.961797763577864,1.63516477461472,-1.79684767953686,.000465578153648587,-.000791535419924791,.000984005657525685,-.00109864378994577,.00319092434145473,-.0054249316016862,.00674406129057572,.0123137463285655,-.0209347588552154,.0353622090543082,-5.86464816610367e-006,9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,-1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,-.166073783576779,.282344179950942,-.350999163642368,.391891091702269,-.410109589767539,-.764836159193308,1.30030781284792,-1.61649145672597,1.80481513164277,-2.02218309922343,3.43793955015111,-4.27391103771759,-4.02885967082133,6.84951625494143,-6.68010475273478,.0125887190510936,-.021402243516657,.0266064261611109,-.0297061146457811,.0823390061849571,-.139985605707144,.174024591276672,.299004079114613,-.508340689012668,.796298871799325,-.000824565926326729,.00140185515919078,-.00174273111826309,-.00697737063211093,.0118623177481531,-.0321260785529637,5.01144631645353e-005,-8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,-.398540259080393,.677563431168936,-.842320170000572,.940451730878483,-.984172086870924,-1.67165845871656,2.84200809304711,-3.53307252095548,3.9446807593569,-3.84755914106515,6.54128488547504,-8.13186772631248,-6.3113399865174,10.7299904588883,-7.98777931164405,.0724976248699598,-.123254146474207,.153224700235957,-.171075607231601,.434860760346183,-.739312383715057,.919084035475437,1.40707187748205,-2.39218103581589,3.23180916098999,-.0113956405123911,.0193738753136992,-.0240848387004708,-.0887334889355571,.150856948291044,-.36808995603673,.00166206328995187,-.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,-.714786270724957,1.21521735159799,-1.51071034695561,1.68671036412375,-1.76512329607374,-2.59853682213183,4.41780594598826,-5.49204234459095,6.13187350041214,-4.7549644652192,8.08397637227704,-10.0496810203089,-5.36731919542456,9.12504854432544,-3.25729414076559,.233201739437629,-.396469282990198,.492874996717867,-.550295671800539,1.22672710731592,-2.0855745668007,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984459,-.0657431866024218,.111770838907705,-.138949104550336,-.451524561772794,.767642727504649,-1.59193561065057,.017197430813833,-.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,-1.09494796295235,1.86153514438243,-2.31418717000874,2.5837934399814,-2.70391057656363,-3.24473228828999,5.51641118734948,-6.85778510905627,7.65672734172539,-4.05017972204568,6.88576274670089,-8.56010903534823,-1.63294619261043,2.776192869346,2.28560008620039,.547225517982271,-.930345156393643,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070526,4.87179403320196,-8.28259983241363,5.49730994822615,-.236321146004056,.401772626268325,-.499467904136599,-1.36209257652082,2.31571114636377,-3.77635721863404,.0946962283839159,-.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,-1.51514018159912,2.57590935107797,-3.20226904625418,3.57533818439405,-3.74155092353066,-3.36446686788906,5.7199734894835,-7.11084574520705,7.93926992086969,-1.91964817870762,3.26361861024795,-4.05720211125474,1.95173675811295,-3.31817281869637,2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898137,3.55510980002442,-6.04408799389592,7.51377213010339,4.59045028887451,-7.80428370621122,1.24891795372526,-.626153181688296,1.06453109468611,-1.32338312764,-2.84473124098146,4.83636425064191,-5.48872263996485,.347192252628353,-.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,-1.94896143440088,3.31345445440235,-4.11915606854152,4.59904391748787,-4.81284738096772,-2.83314651369184,4.81666890668704,-5.98789306701478,6.6854915743852,.637278791777431,-1.08344588754871,1.34689725343212,2.63766838605638,-4.48433402043889,-.808065843843811,1.73374759666501,-2.94756663562366,3.66429874441148,-4.09119503456583,4.12740556141283,-7.01705539432373,8.72332687917668,1.98940127224618,-3.38220674597772,-2.96832239843812,-1.33269841309056,2.26573774926528,-2.81667592802333,-4.37531622836174,7.4385315165314,-4.52786743647192,.95054294779362,-1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,-2.36915365304765,4.02782866109789,-5.00723794478697,5.59058866190052,-5.85048772793476,-1.68415540504794,2.86325431258853,-3.55948498424847,3.97417031416821,2.32831086437971,-3.95839130925952,4.9209161654707,.381749900992996,-.649017927110068,-2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,-2.39387697960251,4.06986110773887,-5.05949102727521,-4.93737708601551,8.39409842544753,-.508166439915618,2.07554270972356,-3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,-2.74931534527504,4.67414645388234,-5.81071476784009,6.48767173775818,-6.78927500842464,-.106757773657142,.181500267083405,-.225633983149896,.251920680005671,2.26282036321899,-3.84705006415361,4.78250111498198,-2.15044627554921,3.65600143048661,.119501749587716,3.45007465931297,-5.86551639682903,7.29177895707793,-8.1412821233112,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429709,1.33170919606025,-3.74106981000851,6.36024100270089,-7.90680110023493,-3.58481468643821,6.09458965574696,3.59724489879787,3.76406389114815,-6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,-3.06556135691961,5.21180037431139,-6.47910494479513,7.23393037100344,-7.57022621762746,1.5947619510804,-2.71127534923244,3.37055072311448,-3.76322492874006,.444911232504622,-.756399320675517,.940325842964601,-2.09440936915946,3.56073236366527,2.63186985254377,4.2894282422279,-7.29251224169723,9.06576398575723,-10.1219390639593,-.964148801340396,1.63916180436132,-2.03774139204767,-2.01722273728898,3.42950637731078,3.46190525493913,-5.18623489409549,8.81718478891181,-10.9611768422514,-.205483038001613,.349344361529711,3.63947886955733,5.81833621896758,-9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,-3.29802783242322,5.60701962552938,-6.97042595115334,7.78249101017965,-8.14428871473084,3.07100849759875,-5.22106113152625,6.4906175528327,-7.24678421547914,-2.20598407348367,3.75042195543293,-4.66237685613718,.923194488569861,-1.5695348492102,.507379758856801,4.96464287844228,-8.44045334766828,10.4928391542811,-11.7152706263276,-3.80884940873902,6.47547397319738,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,-6.45780747384476,10.9790017210618,-13.6486625036235,4.10479708970298,-6.97861844197597,-1.32544341327556,7.79428043908953,-13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,-3.43216369764802,5.83506573884863,-7.25392389703222,8.09901689119009,-8.47552946493044,4.00974905232829,-6.81702604882233,8.4746582765388,-9.46196865399936,-4.32546942586304,7.35378632026677,-9.14193750783762,4.4144312457315,-7.50503145913621,-4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,-7.27823584467216,12.3738225688191,-15.3826488429769,7.46210387346587,-12.6864189798666,-7.31197439196322,9.14177937313726,-15.5420569434015,-8.9987004064276,-10.9691712594614},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,-.0319379183519753,.0500140463222003,-.0518576912013895,.0415275099413171,-.0223217196472694,-.154660030302817,.24219405407586,-.251121942525088,.201097826062499,-.43925822557731,.687868288629403,-.713224862474738,-.961797763577864,1.50615319946369,-1.79684767953686,.000465578153648587,-.000729084691473652,.000755960606624687,-.000605371370823812,.00319092434145473,-.00499691420436607,.00518111316265945,.0123137463285655,-.0192830438371037,.0353622090543082,-5.86464816610367e-006,9.18390428102533e-006,-9.52244633830418e-006,-5.19109369939388e-005,8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,-1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,-.166073783576779,.260067729310804,-.269654486884814,.215938641412748,-.11607057156663,-.764836159193308,1.1977158528687,-1.24186670359493,.994483762912882,-2.02218309922343,3.16669227065688,-3.28342459824587,-4.02885967082133,6.30910167728468,-6.68010475273478,.0125887190510936,-.0197136447951608,.0204403398568299,-.0163685732441572,.0823390061849571,-.128940991873039,.13369408456444,.299004079114613,-.468233518150735,.796298871799325,-.000824565926326729,.00129125129526923,-.00133885010037302,-.00697737063211093,.0109264021013695,-.0321260785529637,5.01144631645353e-005,-7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,-.398540259080393,.624104888716883,-.647110981340288,.518204861963161,-.278543636855169,-1.67165845871656,2.61777873930361,-2.71427671758158,2.17358603453451,-3.84755914106515,6.02518921139496,-6.24729300320555,-6.3113399865174,9.88341340931671,-7.98777931164405,.0724976248699598,-.113529614838378,.11771460504762,-.0942655625163293,.434860760346183,-.680981959385669,.70608468526914,1.40707187748205,-2.20344223207683,3.23180916098999,-.0113956405123911,.0178453112143479,-.0185031347484812,-.0887334889355571,.138954604977606,-.36808995603673,.00166206328995187,-.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,-.714786270724957,1.11933887677121,-1.16060055303996,.929406031924799,-.499571029238757,-2.59853682213183,4.06924896099705,-4.21925182103647,3.37876634868698,-4.7549644652192,7.44616510000515,-7.72064967553202,-5.36731919542456,8.40509854032022,-3.25729414076559,.233201739437629,-.365188565907959,.378650344801756,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064312,5.89679444984459,-.0657431866024218,.102952319693436,-.106747403902083,-.451524561772794,.707077089737688,-1.59193561065057,.017197430813833,-.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,-1.09494796295235,1.7146633520688,-1.77787020176478,1.42371402906152,-.765269708188713,-3.24473228828999,5.08117621439183,-5.26848128513793,4.21898669004741,-4.05017972204568,6.34248838569497,-6.576288631721,-1.63294619261043,2.55715621899877,2.28560008620039,.547225517982271,-.856942588087486,.888531670338268,-.71153394843509,2.39319484717323,-3.74768814487268,3.88583745543936,4.87179403320196,-7.62911752450546,5.49730994822615,-.236321146004056,.370073484889954,-.383715334269619,-1.36209257652082,2.13300568008586,-3.77635721863404,.0946962283839159,-.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,-1.51514018159912,2.37267471198348,-2.46013753301879,1.97007200846434,-1.0589461087365,-3.36446686788906,5.26867781509235,-5.46289467143737,4.37467244553793,-1.91964817870762,3.0061249410246,-3.11693834864261,1.95173675811295,-3.05637492003247,2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002442,-5.56722026499827,5.77244215424072,4.59045028887451,-7.1885396848779,1.24891795372526,-.626153181688296,.980541495928355,-1.01668674800428,-2.84473124098146,4.45478376427993,-5.48872263996485,.347192252628353,-.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,-1.94896143440088,3.05202882623936,-3.1645343667906,2.53415123836086,-1.36214797290775,-2.83314651369184,4.43664234772504,-4.60018826185061,3.68381930162858,.637278791777431,-.997963945475626,1.03475143356998,2.63766838605638,-4.13052808940267,-.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633233,4.12740556141283,-6.4634222783235,6.70168045166519,1.98940127224618,-3.11535668501805,-2.96832239843812,-1.33269841309056,2.08697509460821,-2.16390630168515,-4.37531622836174,6.85164468857583,-4.52786743647192,.95054294779362,-1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,-2.36915365304765,3.71004018615404,-3.84680169804461,3.08050921776368,-1.65582437345554,-1.68415540504794,2.63734867013603,-2.73456804580791,2.18983527963514,2.32831086437971,-3.64608131714141,3.78048514370595,.381749900992996,-.597811573855878,-2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,-2.39387697960251,3.74875634795768,-3.88694503628023,-4.93737708601551,7.73181907803745,-.508166439915618,2.07554270972356,-3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,-2.74931534527504,4.30536466145163,-4.46407134676943,3.5748172149004,-1.9215230524055,-.106757773657142,.16718022074295,-.173342908623153,.13881256945113,2.26282036321899,-3.54352469706742,3.67414802589247,-2.15044627554921,3.3675494587808,.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155854,-4.48598459696892,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761007,1.33170919606025,-3.74106981000851,5.85842936705581,-6.07438596565012,-3.58481468643821,5.61373743608465,3.59724489879787,3.76406389114815,-5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,-3.06556135691961,4.80059864950596,-4.9775609184691,3.98601838486204,-2.14255044478909,1.5947619510804,-2.49736057507455,2.5894196337366,-2.07360079279396,.444911232504622,-.696720767591157,.722403666316151,-2.09440936915946,3.27979695088037,2.63186985254377,4.2894282422279,-6.71714607327471,6.96475715335039,-5.57736018045175,-.964148801340396,1.50983487071459,-1.56549122161564,-2.01722273728898,3.15892446006539,3.46190525493913,-5.18623489409549,8.12152468439221,-8.42090472917075,-.205483038001613,.321781716436834,3.63947886955733,5.81833621896758,-9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,-3.29802783242322,5.16463580891203,-5.35501741292457,4.28828460541245,-2.30502351007762,3.07100849759875,-4.80912874954551,4.98640545923911,-3.99310107128126,-2.20598407348367,3.45452037324197,-3.58186277642043,.923194488569861,-1.44570135775136,.507379758856801,4.96464287844228,-7.7745166798312,8.0611051284312,-6.45531290809182,-3.80884940873902,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,-6.45780747384476,10.1127781284249,-10.4855608369157,4.10479708970298,-6.42801793274624,-1.32544341327556,7.79428043908953,-12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,-3.43216369764802,5.37468949190064,-5.572814208608,4.46269573688388,-2.39877236199698,4.00974905232829,-6.27917489029831,6.51064123734671,-5.21370528571982,-4.32546942586304,6.77358573667404,-7.02327732701019,4.4144312457315,-6.91289790468793,-4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,-7.27823584467216,11.3975501068473,-11.8176927763526,7.46210387346587,-11.6854832203471,-7.31197439196322,9.14177937313726,-14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,-.0319379183519753,.0425895510116904,-.0277895142706456,-.00117913035023069,.029764050100784,-.154660030302817,.206240781926363,-.134571297792104,-.00570996311627599,-.43925822557731,.585755477154869,-.382203141612095,-.961797763577864,1.28256746328293,-1.79684767953686,.000465578153648587,-.000620853379055167,.000405104384309337,1.71888889439268e-005,.00319092434145473,-.00425513126926153,.00277645639124976,.0123137463285655,-.0164205106275604,.0353622090543082,-5.86464816610367e-006,7.82057019228999e-006,-5.10289124485107e-006,-5.19109369939388e-005,6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,-9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,-.166073783576779,.221461142188379,-.144502522920426,-.00613135259583728,.154769899538562,-.764836159193308,1.01991708658492,-.665491881382992,-.0282373296477263,-2.02218309922343,2.69660249304008,-1.75952250506474,-4.02885967082133,5.37252686958436,-6.68010475273478,.0125887190510936,-.0167871896471046,.0109535751204635,.000464768572193078,.0823390061849571,-.109799933312678,.0716440239592628,.299004079114613,-.398725093710568,.796298871799325,-.000824565926326729,.00109956736062704,-.000717463371558171,-.00697737063211093,.00930439734215142,-.0321260785529637,5.01144631645353e-005,-6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,-.398540259080393,.531457639388288,-.346774016236295,-.0147138868003722,.371413443660288,-1.67165845871656,2.22917418911978,-1.45452737733324,-.0617167093760001,-3.84755914106515,5.13076069911761,-3.34780114662786,-6.3113399865174,8.41623844737545,-7.98777931164405,.0724976248699598,-.0966763475074085,.0630809359431452,.0026765723709811,.434860760346183,-.579891411113903,.378376861425819,1.40707187748205,-1.87634519190449,3.23180916098999,-.0113956405123911,.0151962068293123,-.00991546509750554,-.0887334889355571,.118327043544488,-.36808995603673,.00166206328995187,-.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,-.714786270724957,.953175031760129,-.621942953572119,-.0263895153231826,.666134033494214,-2.59853682213183,3.46517626442722,-2.26101386266496,-.0959365479342438,-4.7549644652192,6.34079527029531,-4.13734393811778,-5.36731919542456,7.15737676442222,-3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831034,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.4071966513386,5.89679444984459,-.0657431866024218,.0876692328212839,-.0572038290823836,-.451524561772794,.602112766253303,-1.59193561065057,.017197430813833,-.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,-1.09494796295235,1.46012465838812,-.952725588021822,-.0404248755605117,1.02041985541771,-3.24473228828999,4.32688473530227,-2.82327524550586,-.119793728554495,-4.05017972204568,5.4009573811209,-3.52410341735766,-1.63294619261043,2.17755097302669,2.28560008620039,.547225517982271,-.729731000653746,.476146603536492,.0202032646552823,2.39319484717323,-3.19135057367101,2.08234367785444,4.87179403320196,-6.49658873611033,5.49730994822615,-.236321146004056,.315136741011867,-.205625482170818,-1.36209257652082,1.81636481949508,-3.77635721863404,.0946962283839159,-.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,-1.51514018159912,2.02045541424857,-1.31833919902212,-.0559381407795183,1.41201150863478,-3.36446686788906,4.4865520602846,-2.92745754610215,-.124214263271286,-1.91964817870762,2.55987109498181,-1.67030580640284,1.95173675811295,-2.60266155489591,2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981014,.0386847883193904,3.55510980002442,-4.74077641992773,3.09333793899121,4.59045028887451,-6.12141388513833,1.24891795372526,-.626153181688296,.834981872710088,-.544822382907866,-2.84473124098146,3.79347911889772,-5.48872263996485,.347192252628353,-.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,-1.94896143440088,2.59896063091726,-1.69581157410879,-.0719545824309833,1.81630452989169,-2.83314651369184,3.77803076303363,-2.46515019062261,-.104598208499715,.637278791777431,-.849817990926997,.55450289173293,2.63766838605638,-3.51735861632833,-.808065843843811,1.73374759666501,-2.31197070864005,1.50855178059943,.0640090061407145,4.12740556141283,-5.50393885468463,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843812,-1.33269841309056,1.77716739172311,-1.15959472274786,-4.37531622836174,5.8345303456991,-4.52786743647192,.95054294779362,-1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,-2.36915365304765,3.15929138677771,-2.06142518510909,-.0874678476499899,2.20789618310877,-1.68415540504794,2.24583899888368,-1.46540110005889,-.0621780897499632,2.32831086437971,-3.10482709694771,2.02588745115913,.381749900992996,-.50906751987416,-2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217328,3.12121330296808,-1.52344991306454,2.03153653003234,-2.69130383032704,-2.39387697960251,3.19226020392741,-2.08293720858962,-4.93737708601551,6.58404442855625,-.508166439915618,2.07554270972356,-2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,-2.74931534527504,3.6662410134057,-2.39220781955879,-.101503207887319,2.56218200503226,-.106757773657142,.142362617365695,-.0928910470423959,-.0039414381903617,2.26282036321899,-3.01749482284837,1.96890348630228,-2.15044627554921,2.86764279249033,.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,-3.74106981000851,4.9887560540339,-3.25514367429686,-3.58481468643821,4.78038820054912,3.59724489879787,3.76406389114815,-5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,-3.06556135691961,4.08795840577754,-2.66737675689462,-.113178836410129,2.85690259486619,1.5947619510804,-2.12663188500128,1.38761892782618,.0588777326765052,.444911232504622,-.593293821521023,.387121881618236,-2.09440936915946,2.79291698671264,2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.158363327640689,-.964148801340396,1.28570259705192,-.838915879897341,-2.01722273728898,2.68998779997356,3.46190525493913,-5.18623489409549,6.91589893787678,-4.51259681489189,-.205483038001613,.274013798814207,3.63947886955733,5.81833621896758,-7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,-3.29802783242322,4.39795490297745,-2.86964825021048,-.121761370614664,3.07354613898791,3.07100849759875,-4.09522222779324,2.67211637191124,.11337994190694,-2.20598407348367,2.94170303097044,-1.91944964063812,.923194488569861,-1.23108958885198,.507379758856801,4.96464287844228,-6.62040364808344,4.31978730278896,.183291880031408,-3.80884940873902,5.07914086475798,-3.31411940634737,2.02329330850018,-2.6980829707225,.0162698697481229,-6.45780747384476,8.61155438605813,-5.61900531555838,4.10479708970298,-5.47379022059324,-1.32544341327556,7.79428043908953,-10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,-3.43216369764802,4.57682649415414,-2.98636125886026,-.126713592860271,3.19855198842569,4.00974905232829,-5.34704266035319,3.48892427303922,.148037726033603,-4.32546942586304,5.76805910494959,-3.7636358461794,4.4144312457315,-5.8866906319401,-4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,-7.27823584467216,9.70560427278213,-6.33286855712843,7.46210387346587,-9.95079423294563,-7.31197439196322,9.14177937313726,-12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,-.0319379183519753,.0324893117176766,-.00124359624768518,-.0316451541922531,.0381313330936754,-.154660030302817,.157330164177211,-.00602214055814436,-.153242313889452,-.43925822557731,.446841815317497,-.0171038035473819,-.961797763577864,.978402756631806,-1.79684767953686,.000465578153648587,-.000473616144898061,1.81286469184553e-005,.000461310355481612,.00319092434145473,-.00324601417302388,.000124247970525594,.0123137463285655,-.012526337462346,.0353622090543082,-5.86464816610367e-006,5.96589860075151e-006,-2.28357226493363e-007,-5.19109369939388e-005,5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,-6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,-.166073783576779,.168940970519646,-.00646656841622833,-.164551441038204,.198278882483944,-.764836159193308,.778040701653472,-.0297811324981325,-.757825163693696,-2.02218309922343,2.05709515348012,-.0787396124921655,-4.02885967082133,4.09841606758444,-6.68010475273478,.0125887190510936,-.0128060574566476,.000490178589876797,.0124733225023729,.0823390061849571,-.0837605510092484,.00320610999159924,.299004079114613,-.304166246215336,.796298871799325,-.000824565926326729,.000838801675185275,-3.21068856415306e-005,-.00697737063211093,.00709783170785957,-.0321260785529637,5.01144631645353e-005,-5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,-.398540259080393,.405420872037066,-.0155183304460168,-.394886974518159,.47582535601586,-1.67165845871656,1.70051886866409,-.0650909105855692,-1.6563349285852,-3.84755914106515,3.91398546692074,-.149815966677054,-6.3113399865174,6.42030234909073,-7.98777931164405,.0724976248699598,-.0737492627952912,.00282290703160039,.0718330635538727,.434860760346183,-.442368430061905,.0169325753770763,1.40707187748205,-1.43136432250754,3.23180916098999,-.0113956405123911,.0115923809684749,-.000443722587923622,-.0887334889355571,.0902654315677738,-.36808995603673,.00166206328995187,-.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,-.714786270724957,.727126724577601,-.027832293713515,-.70823406529868,.853397929051741,-2.59853682213183,2.6433993582315,-.101181630171365,-2.57471691018525,-4.7549644652192,4.83705672356223,-.185148446522551,-5.36731919542456,5.45998347614275,-3.25729414076559,.233201739437629,-.237227859491313,.00908039168484127,.231064057614915,1.22672710731592,-1.24790598293055,.0477662072687217,3.30495822058058,-3.36201679531997,5.89679444984459,-.0657431866024218,.0668782122732034,-.00255990322490887,-.451524561772794,.459319924425276,-1.59193561065057,.017197430813833,-.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,-1.09494796295235,1.11385173231848,-.042635000914324,-1.08491094310714,1.30728073869583,-3.24473228828999,3.30075109017009,-.126343140340609,-3.21498892012256,-4.05017972204568,4.12010419835037,-.157705591526293,-1.63294619261043,1.66113825253464,2.28560008620039,.547225517982271,-.55667311317238,.0213078258129724,.542209285918406,2.39319484717323,-2.43451224831938,.0931860398476328,4.87179403320196,-4.95590330339708,5.49730994822615,-.236321146004056,.240401121148633,-.00920185489986058,-1.36209257652082,1.38560847465611,-3.77635721863404,.0946962283839159,-.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,-1.51514018159912,1.54129837497396,-.0589964137232872,-1.50125140095791,1.80895681150723,-3.36446686788906,3.42255283189436,-.131005356341758,-3.33362593315312,-1.91964817870762,1.95279001506978,-.074747115516283,1.95173675811295,-1.98543258954328,2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063147,1.0382109925525,3.55510980002442,-3.61648712494209,.138428596350645,4.59045028887451,-4.66970228947898,1.24891795372526,-.626153181688296,.636963426396325,-.0243811052053837,-2.84473124098146,2.89384420968737,-5.48872263996485,.347192252628353,-.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,-1.94896143440088,1.98260935074574,-.0758885128327144,-1.93109596018978,2.32690486658693,-2.83314651369184,2.88205947758122,-.110316844590442,-2.80717601371158,.637278791777431,-.648281114842891,.0248143133692,2.63766838605638,-2.68320650944122,-.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336552,1.7178549162266,4.12740556141283,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843812,-1.33269841309056,1.35570683401508,-.051892509958983,-4.37531622836174,4.45085403872017,-4.52786743647192,.95054294779362,-.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,-2.36915365304765,2.41005599340123,-.0922499256416775,-2.34743641804055,2.82858093939833,-1.68415540504794,1.71323156899255,-.0655775157362847,-1.66871731961812,2.32831086437971,-2.36850807266163,.090659592246103,.381749900992996,-.388340636274758,-2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622405,2.53843733881177,3.58714638367868,-3.64907680538234,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,-2.39387697960251,2.4352061570422,-.0932126006787421,-4.93737708601551,5.02261861703307,-.508166439915618,2.07554270972356,-2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,-2.74931534527504,2.7967810011421,-.107052632842487,-2.72411329584901,3.28246374904242,-.106757773657142,.108600897231078,-.00415692611322449,-.105779161098541,2.26282036321899,-2.30188690834251,.0881095280677934,-2.15044627554921,2.18757273479106,.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,-3.74106981000851,3.80565764777524,-.14566949316231,-3.58481468643821,3.64670485341609,3.59724489879787,3.76406389114815,-3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,-3.06556135691961,3.11848685368264,-.119366596109985,-3.03746038662953,3.6600363220783,1.5947619510804,-1.62229477839744,.0620967201893262,1.58014330533833,.444911232504622,-.452592418790165,.0173239199033356,-2.09440936915946,2.13056837715884,2.63186985254377,4.2894282422279,-4.36348323472717,.167021432333127,4.25010849784966,-.964148801340396,.98079438405867,-.0375419530619817,-2.01722273728898,2.05204915464121,3.46190525493913,-5.18623489409549,5.27577283777777,-.201941221845807,-.205483038001613,.20903060758203,3.63947886955733,5.81833621896758,-5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,-3.29802783242322,3.35496675520006,-.128418358139773,-3.26779592010949,3.93758279561022,3.07100849759875,-3.12402803860063,.119578696522851,3.04285759691612,-2.20598407348367,2.24406936639925,-.0858964409190727,.923194488569861,-.939133013954983,.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.91913366884931,-3.80884940873902,3.87460743053656,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,-6.45780747384476,6.56929852924652,-.251453618731752,4.10479708970298,-4.17566451229335,-1.32544341327556,7.79428043908953,-7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,-3.43216369764802,3.49141841400203,-.133641330308316,-3.40070220695544,4.09773034500049,4.00974905232829,-4.07897551472765,.156131305216531,3.97299303311745,-4.32546942586304,4.40014665135235,-.168424801181435,4.4144312457315,-4.49064435720703,-4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,-7.27823584467216,7.40389121595289,-.283399396550353,7.46210387346587,-7.59093364435534,-7.31197439196322,9.14177937313726,-9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,-.0319379183519753,.0203477386368136,.0212139303918346,-.0361091456847505,.00500970708930343,-.154660030302817,.0985343453312713,.102728896816887,-.174859285048361,-.43925822557731,.279852665057626,.291765835225991,-.961797763577864,.612764091356553,-1.79684767953686,.000465578153648587,-.000296621166194719,-.000309248161967448,.000526384631617547,.00319092434145473,-.00203294697585566,-.00211948838134797,.0123137463285655,-.00784512281993549,.0353622090543082,-5.86464816610367e-006,3.73638403043722e-006,3.89543979110572e-006,-5.19109369939388e-005,3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,-4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,-.166073783576779,.105806079951921,.110310185087204,-.187763722716786,.026049944826181,-.764836159193308,.487279293067771,.508022497640507,-.864726999689721,-2.02218309922343,1.28833860494792,1.34318245298509,-4.02885967082133,2.56679795782664,-6.68010475273478,.0125887190510936,-.00802030872060955,-.00836172873999091,.0142328590555997,.0823390061849571,-.0524584150880732,-.054691540214079,.299004079114613,-.190496349504475,.796298871799325,-.000824565926326729,.000525333297396549,.000547696439552502,-.00697737063211093,.00444530268091327,-.0321260785529637,5.01144631645353e-005,-3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,-.398540259080393,.253911132799744,.264719986484413,-.450591303972135,.0625140918479549,-1.67165845871656,1.06501836010644,1.11035559079056,-1.88998413076671,-3.84755914106515,2.45129087378787,2.55564094422791,-6.3113399865174,4.0209726602962,-7.98777931164405,.0724976248699598,-.0461884430503282,-.0481546590241491,.0819661216087145,.434860760346183,-.277051027559315,-.288844933452006,1.40707187748205,-.896449496423068,3.23180916098999,-.0113956405123911,.00726019498888195,.00756925739444607,-.0887334889355571,.056532358254944,-.36808995603673,.00166206328995187,-.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,-.714786270724957,.455392366453167,.474778162593069,-.80814038343456,.112119700737043,-2.59853682213183,1.65553520251249,1.7260104019959,-2.93791673259187,-4.7549644652192,3.02940138707022,3.15836129611741,-5.36731919542456,3.41953432868287,-3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066827,3.30495822058058,-2.10559828531328,5.89679444984459,-.0657431866024218,.0418851712113093,.0436681993244461,-.451524561772794,.287667582930097,-1.59193561065057,.017197430813833,-.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,-1.09494796295235,.697594462028796,.727290664744143,-1.23795280192461,.171750973622283,-3.24473228828999,2.06722817250133,2.15522890943138,-3.66850806244401,-4.05017972204568,2.58038102231816,2.69022638759908,-1.63294619261043,1.04035466481729,2.28560008620039,.547225517982271,-.348638934287124,-.363480297113411,.618695487400889,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320196,-3.10383383936015,5.49730994822615,-.236321146004056,.150560874408057,.156970165907755,-1.36209257652082,.867793055767227,-3.77635721863404,.0946962283839159,-.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,-1.51514018159912,.965300119862265,1.00639240141108,-1.71302390303713,.237661341149246,-3.36446686788906,2.14351141382237,2.23475948529486,-3.80388048506198,-1.91964817870762,1.22301331540983,1.27507630270228,1.95173675811295,-1.24345704101385,2.75500656898115,1.04781596872157,-.667566534563461,-.69598446548217,1.18466517033957,3.55510980002442,-2.2649705675878,-2.36138908644035,4.59045028887451,-2.92458893939423,1.24891795372526,-.626153181688296,.398923973340748,.415905941872818,-2.84473124098146,1.81238636712449,-5.48872263996485,.347192252628353,-.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,-1.94896143440088,1.24168887412682,1.29454686902578,-2.20350404785819,.305709527061058,-2.83314651369184,1.80500570449659,1.88184377891507,-3.20316744314248,.637278791777431,-.406012131085817,-.423295838593301,2.63766838605638,-1.68046603255685,-.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141283,-2.62958182530229,-2.74152163963164,1.98940127224618,-1.26745321062054,-2.96832239843812,-1.33269841309056,.849065950254409,.885210208841136,-4.37531622836174,2.78752641722509,-4.52786743647192,.95054294779362,-.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,-2.36915365304765,1.50939453196029,1.57364860569271,-2.6785751489707,.371619894588022,-1.68415540504794,1.07298020016939,1.11865636189275,-1.9041132311976,2.32831086437971,-1.48337347495979,-1.54651984745118,.381749900992996,-.243213947948102,-2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001399,-2.69130383032704,-2.39387697960251,1.52514583383108,1.59007043171885,-4.93737708601551,3.14561698874585,-.508166439915618,2.07554270972356,-1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,-2.74931534527504,1.75159662753591,1.82616110784378,-3.10838756746075,.431251167473262,-.106757773657142,.0680156813349527,.070911070513513,-.120700790879792,2.26282036321899,-1.44164929037179,-1.50301948784968,-2.15044627554921,1.37005543987348,.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,-3.74106981000851,2.38344621859103,2.48490817973374,-3.58481468643821,2.28389563582348,3.59724489879787,3.76406389114815,-2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,-3.06556135691961,1.95307786118934,2.03621928395244,-3.46593664692318,.48085677636235,1.5947619510804,-1.01602737631215,-1.05927908845788,1.80304461367463,.444911232504622,-.283454211851926,-.295520697836744,-2.09440936915946,1.33435416739534,2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340396,.614261944257265,.640410728572748,-2.01722273728898,1.28517834515454,3.46190525493913,-5.18623489409549,3.30416500488362,3.44482144542742,-.205483038001613,.130913828127032,3.63947886955733,5.81833621896758,-3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,-3.29802783242322,2.10118291403716,2.19062908534965,-3.72876422817853,.517320923384123,3.07100849759875,-1.95654825118803,-2.0398374063157,3.47209520918979,-2.20598407348367,1.40543872850124,1.4652674627761,.923194488569861,-.588169835027997,.507379758856801,4.96464287844228,-3.16298810157525,-3.2976347213703,5.61304625700718,-3.80884940873902,2.426628794168,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,-6.45780747384476,4.11428751282488,4.28943041158219,4.10479708970298,-2.61517790433632,-1.32544341327556,7.79428043908953,-4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,-3.43216369764802,2.18664125535227,2.27972534004502,-3.88041880521056,.538361161121001,4.00974905232829,-2.55462252942968,-2.66337136913609,4.53343925473611,-4.32546942586304,2.75576887572707,2.87308040304256,4.4144312457315,-2.81244670509951,-4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,-7.27823584467216,4.63698476184225,4.83437858766287,7.46210387346587,-4.75412760808659,-7.31197439196322,9.14177937313726,-5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,-.0319379183519753,.00692767387221854,.0340276241262534,-.0154583827232305,-.0303981920101794,-.154660030302817,.0335474040556334,.16477947382797,-.0748575381590809,-.43925822557731,.0952797768044285,.467998998326769,-.961797763577864,.208624155307574,-1.79684767953686,.000465578153648587,-.00010098884891692,-.000496041045803253,.000225346098359983,.00319092434145473,-.000692145397560818,-.00339970729838115,.0123137463285655,-.00267098242974697,.0353622090543082,-5.86464816610367e-006,1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,-1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,-.166073783576779,.0360231684064709,.176940031673225,-.080381948458016,-.158067370120622,-.764836159193308,.165901090401825,.814879575736018,-.370191004490028,-2.02218309922343,.43863300248611,2.1544950322241,-4.02885967082133,.873902474623302,-6.68010475273478,.0125887190510936,-.00273062693415814,-.0134124020091825,.00609310960852155,.0823390061849571,-.0178602054051423,-.087726467416282,.299004079114613,-.0648571621038782,.796298871799325,-.000824565926326729,.000178857111521775,.000878517475835277,-.00697737063211093,.00151346583452521,-.0321260785529637,5.01144631645353e-005,-1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,-.398540259080393,.0864476172000634,.4246168452719,-.192898854195334,-.379326641949509,-1.67165845871656,.362600483446047,1.78103547961334,-.809105213656782,-3.84755914106515,.834576462901105,4.09930587333892,-6.3113399865174,1.3689967095037,-7.98777931164405,.0724976248699598,-.0157255052185161,-.0772411621484506,.0350898270822692,.434860760346183,-.0943259199045093,-.4633138058217,1.40707187748205,-.305208842367266,3.23180916098999,-.0113956405123911,.00247183552086188,.0121412600482659,-.0887334889355571,.0192472366688018,-.36808995603673,.00166206328995187,-.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,-.714786270724957,.155044737648518,.76155491046056,-.345966184032581,-.680326440323354,-2.59853682213183,.563650249872674,2.76855972053751,-1.25772682793123,-4.7549644652192,1.03140232004452,5.06608294899263,-5.36731919542456,1.16422856843223,-3.25729414076559,.233201739437629,-.0505839353643848,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929028,5.89679444984459,-.0657431866024218,.0142603957822129,.0700447793236141,-.451524561772794,.0979404755348858,-1.59193561065057,.017197430813833,-.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,-1.09494796295235,.237505848402132,1.1665906747753,-.529969564290411,-1.04216054572391,-3.24473228828999,.703816913228947,3.45703591374062,-1.57049413873188,-4.05017972204568,.878527020016876,4.31518396831288,-1.63294619261043,.354203381414383,2.28560008620039,.547225517982271,-.118699029853337,-.583030617100296,.264864522602584,2.39319484717323,-.519109393293245,-2.54978216897294,4.87179403320196,-1.05674389554722,5.49730994822615,-.236321146004056,.0512605677924324,.251783696179217,-1.36209257652082,.295452354005307,-3.77635721863404,.0946962283839159,-.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,-1.51514018159912,.328650005712205,1.6142761726,-.733348258592925,-1.44209530674498,-3.36446686788906,.729788615894315,3.58460475610227,-1.62844728825442,-1.91964817870762,.416392088662432,2.04525122687055,1.95173675811295,-.423352442647985,2.75500656898115,1.04781596872157,-.227282418061715,-1.11637482313293,.507157044419473,3.55510980002442,-.771141093489061,-3.7877215002156,4.59045028887451,-.995717447724871,1.24891795372526,-.626153181688296,.135819278812355,.667122536961418,-2.84473124098146,.61705243545725,-5.48872263996485,.347192252628353,-.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,-1.94896143440088,.422750445356608,2.07648245560297,-.943323589025389,-1.85500204648398,-2.83314651369184,.614539585033937,3.01851998146924,-1.37128107917545,.637278791777431,-.138232541833329,-.678976098432196,2.63766838605638,-.57213830168982,-.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.839157192810101,4.12740556141283,-.895278125552931,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843812,-1.33269841309056,.289076447479325,1.41989719504135,-4.37531622836174,.949052583847879,-4.52786743647192,.95054294779362,-.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,-2.36915365304765,.513894602666682,2.52416795342767,-1.1467022833279,-2.25493680750505,-1.68415540504794,.365311203903165,1.79435010419287,-.815153904032968,2.32831086437971,-.505035367744937,-2.48065281165978,.381749900992996,-.0828056101286271,-2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.33045247496738,-2.69130383032704,-2.39387697960251,.519257354978938,2.55050893377232,-4.93737708601551,1.07096955656365,-.508166439915618,2.07554270972356,-.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,-2.74931534527504,.596355713420295,2.92920371774241,-1.33070566358573,-2.61677091290561,-.106757773657142,.0231568955595295,.113742960840148,-.0516722006304461,2.26282036321899,-.490829782123398,-2.41087725104767,-2.15044627554921,.466454648534285,.119501749587716,3.45007465931297,-.748357855262331,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239338,-1.88618542328239,-3.42380239520456,.742659122293479,1.33170919606025,-3.74106981000851,.8114777955963,3.98584891938639,-3.58481468643821,.777584399208991,3.59724489879787,3.76406389114815,-.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,-3.06556135691961,.66495283386875,3.26614178293108,-1.48377299342298,-2.91777071127945,1.5947619510804,-.345920813804923,-1.69910761471712,.771886268054593,.444911232504622,-.0965059740773432,-.474021882875009,-2.09440936915946,.454299648071273,2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340396,.209134120399514,1.02723329226541,-2.01722273728898,.437557047580714,3.46190525493913,-5.18623489409549,1.1249494591216,5.5255715071159,-.205483038001613,.0445714544901277,3.63947886955733,5.81833621896758,-1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,-3.29802783242322,.715377282662343,3.51381859652975,-1.5962898991603,-3.13902998310834,3.07100849759875,-.666134376965478,-3.2719453330298,1.4864094837286,-2.20598407348367,.47850138692202,2.35032214902362,.923194488569861,-.200250694746914,.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390442,-3.80884940873902,.826179910946771,4.05806335333585,2.02329330850018,-.438873818984883,.0162698697481229,-6.45780747384476,1.40076706380653,6.88034339064391,4.10479708970298,-.890374108024765,-1.32544341327556,7.79428043908953,-1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,-3.43216369764802,.744472777196595,3.65673100407672,-1.66121346489508,-3.26669916121878,4.00974905232829,-.869757178740834,-4.27210791134071,1.94077255839989,-4.32546942586304,.938240282077712,4.60848594454081,4.4144312457315,-.957537046434959,-4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,-7.27823584467216,1.57872669557351,7.75445258971521,7.46210387346587,-1.6186096255506,-7.31197439196322,9.14177937313726,-1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,-.0319379183519753,-.00692767387221854,.0340276241262535,.0154583827232306,-.0303981920101797,-.154660030302817,-.0335474040556334,.16477947382797,.074857538159081,-.43925822557731,-.0952797768044285,.467998998326769,-.961797763577864,-.208624155307574,-1.79684767953686,.000465578153648587,.00010098884891692,-.000496041045803253,-.000225346098359983,.00319092434145473,.000692145397560818,-.00339970729838115,.0123137463285655,.00267098242974697,.0353622090543082,-5.86464816610367e-006,-1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,-1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,-.166073783576779,-.0360231684064709,.176940031673225,.0803819484580161,-.158067370120624,-.764836159193308,-.165901090401825,.814879575736019,.370191004490028,-2.02218309922343,-.43863300248611,2.15449503222411,-4.02885967082133,-.873902474623302,-6.68010475273478,.0125887190510936,.00273062693415814,-.0134124020091825,-.00609310960852156,.0823390061849571,.0178602054051423,-.087726467416282,.299004079114613,.0648571621038782,.796298871799325,-.000824565926326729,-.000178857111521775,.000878517475835277,-.00697737063211093,-.00151346583452521,-.0321260785529637,5.01144631645353e-005,1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,-.398540259080393,-.0864476172000634,.4246168452719,.192898854195334,-.379326641949514,-1.67165845871656,-.362600483446047,1.78103547961334,.809105213656783,-3.84755914106515,-.834576462901105,4.09930587333892,-6.3113399865174,-1.3689967095037,-7.98777931164405,.0724976248699598,.0157255052185161,-.0772411621484506,-.0350898270822693,.434860760346183,.0943259199045093,-.463313805821701,1.40707187748205,.305208842367266,3.23180916098999,-.0113956405123911,-.00247183552086188,.0121412600482659,-.0887334889355571,-.0192472366688018,-.36808995603673,.00166206328995187,.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,-.714786270724957,-.155044737648518,.761554910460561,.345966184032582,-.680326440323361,-2.59853682213183,-.563650249872674,2.76855972053751,1.25772682793124,-4.7549644652192,-1.03140232004452,5.06608294899263,-5.36731919542456,-1.16422856843223,-3.25729414076559,.233201739437629,.0505839353643848,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929028,5.89679444984459,-.0657431866024218,-.0142603957822129,.0700447793236141,-.451524561772794,-.0979404755348858,-1.59193561065057,.017197430813833,.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,-1.09494796295235,-.237505848402132,1.1665906747753,.529969564290411,-1.04216054572392,-3.24473228828999,-.703816913228947,3.45703591374062,1.57049413873189,-4.05017972204568,-.878527020016876,4.31518396831288,-1.63294619261043,-.354203381414383,2.28560008620039,.547225517982271,.118699029853337,-.583030617100297,-.264864522602584,2.39319484717323,.519109393293245,-2.54978216897294,4.87179403320196,1.05674389554722,5.49730994822615,-.236321146004056,-.0512605677924324,.251783696179217,-1.36209257652082,-.295452354005307,-3.77635721863404,.0946962283839159,.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,-1.51514018159912,-.328650005712205,1.6142761726,.733348258592926,-1.44209530674499,-3.36446686788906,-.729788615894315,3.58460475610227,1.62844728825442,-1.91964817870762,-.416392088662432,2.04525122687055,1.95173675811295,.423352442647985,2.75500656898115,1.04781596872157,.227282418061715,-1.11637482313293,-.507157044419473,3.55510980002442,.771141093489061,-3.7877215002156,4.59045028887451,.995717447724871,1.24891795372526,-.626153181688296,-.135819278812355,.667122536961419,-2.84473124098146,-.61705243545725,-5.48872263996485,.347192252628353,.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,-1.94896143440088,-.422750445356608,2.07648245560297,.94332358902539,-1.85500204648401,-2.83314651369184,-.614539585033937,3.01851998146924,1.37128107917545,.637278791777431,.138232541833329,-.678976098432196,2.63766838605638,.57213830168982,-.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810102,4.12740556141283,.895278125552931,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843812,-1.33269841309056,-.289076447479325,1.41989719504135,-4.37531622836174,-.949052583847879,-4.52786743647192,.95054294779362,.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,-2.36915365304765,-.513894602666682,2.52416795342767,1.14670228332791,-2.25493680750508,-1.68415540504794,-.365311203903165,1.79435010419287,.81515390403297,2.32831086437971,.505035367744937,-2.48065281165978,.381749900992996,.0828056101286271,-2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.33045247496738,-2.69130383032704,-2.39387697960251,-.519257354978938,2.55050893377233,-4.93737708601551,-1.07096955656365,-.508166439915618,2.07554270972356,.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,-2.74931534527504,-.596355713420295,2.92920371774242,1.33070566358573,-2.61677091290564,-.106757773657142,-.0231568955595295,.113742960840148,.0516722006304461,2.26282036321899,.490829782123398,-2.41087725104768,-2.15044627554921,-.466454648534285,.119501749587716,3.45007465931297,.748357855262331,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239338,-1.88618542328239,-3.42380239520456,-.742659122293479,1.33170919606025,-3.74106981000851,-.8114777955963,3.98584891938639,-3.58481468643821,-.777584399208991,3.59724489879787,3.76406389114815,.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,-3.06556135691961,-.66495283386875,3.26614178293108,1.48377299342298,-2.91777071127949,1.5947619510804,.345920813804923,-1.69910761471712,-.771886268054594,.444911232504622,.0965059740773432,-.474021882875009,-2.09440936915946,-.454299648071273,2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340396,-.209134120399514,1.02723329226541,-2.01722273728898,-.437557047580714,3.46190525493913,-5.18623489409549,-1.1249494591216,5.5255715071159,-.205483038001613,-.0445714544901277,3.63947886955733,5.81833621896758,1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,-3.29802783242322,-.715377282662343,3.51381859652975,1.5962898991603,-3.13902998310838,3.07100849759875,.666134376965478,-3.2719453330298,-1.48640948372861,-2.20598407348367,-.47850138692202,2.35032214902362,.923194488569861,.200250694746914,.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873902,-.826179910946771,4.05806335333585,2.02329330850018,.438873818984883,.0162698697481229,-6.45780747384476,-1.40076706380653,6.88034339064391,4.10479708970298,.890374108024765,-1.32544341327556,7.79428043908953,1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,-3.43216369764802,-.744472777196595,3.65673100407672,1.66121346489509,-3.26669916121882,4.00974905232829,.869757178740834,-4.27210791134071,-1.9407725583999,-4.32546942586304,-.938240282077712,4.60848594454081,4.4144312457315,.957537046434959,-4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,-7.27823584467216,-1.57872669557351,7.75445258971521,7.46210387346587,1.6186096255506,-7.31197439196322,9.14177937313726,1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,-.0319379183519753,-.0203477386368136,.0212139303918347,.0361091456847507,.00500970708930326,-.154660030302817,-.0985343453312713,.102728896816887,.174859285048362,-.43925822557731,-.279852665057626,.291765835225991,-.961797763577864,-.612764091356553,-1.79684767953686,.000465578153648587,.000296621166194719,-.000309248161967448,-.000526384631617549,.00319092434145473,.00203294697585566,-.00211948838134797,.0123137463285655,.00784512281993549,.0353622090543082,-5.86464816610367e-006,-3.73638403043722e-006,3.89543979110573e-006,-5.19109369939388e-005,-3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,-.166073783576779,-.105806079951921,.110310185087204,.187763722716787,.0260499448261801,-.764836159193308,-.487279293067771,.508022497640508,.864726999689725,-2.02218309922343,-1.28833860494792,1.34318245298509,-4.02885967082133,-2.56679795782664,-6.68010475273478,.0125887190510936,.00802030872060955,-.00836172873999092,-.0142328590555998,.0823390061849571,.0524584150880732,-.0546915402140791,.299004079114613,.190496349504475,.796298871799325,-.000824565926326729,-.000525333297396549,.000547696439552503,-.00697737063211093,-.00444530268091327,-.0321260785529637,5.01144631645353e-005,3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,-.398540259080393,-.253911132799744,.264719986484414,.450591303972137,.0625140918479527,-1.67165845871656,-1.06501836010644,1.11035559079056,1.88998413076672,-3.84755914106515,-2.45129087378787,2.55564094422791,-6.3113399865174,-4.0209726602962,-7.98777931164405,.0724976248699598,.0461884430503282,-.0481546590241492,-.0819661216087149,.434860760346183,.277051027559315,-.288844933452006,1.40707187748205,.896449496423068,3.23180916098999,-.0113956405123911,-.00726019498888195,.00756925739444608,-.0887334889355571,-.056532358254944,-.36808995603673,.00166206328995187,.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,-.714786270724957,-.455392366453167,.47477816259307,.808140383434563,.112119700737039,-2.59853682213183,-1.65553520251249,1.7260104019959,2.93791673259189,-4.7549644652192,-3.02940138707022,3.15836129611742,-5.36731919542456,-3.41953432868287,-3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066828,3.30495822058058,2.10559828531328,5.89679444984459,-.0657431866024218,-.0418851712113093,.0436681993244462,-.451524561772794,-.287667582930097,-1.59193561065057,.017197430813833,.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,-1.09494796295235,-.697594462028796,.727290664744144,1.23795280192461,.171750973622277,-3.24473228828999,-2.06722817250133,2.15522890943138,3.66850806244403,-4.05017972204568,-2.58038102231816,2.69022638759909,-1.63294619261043,-1.04035466481729,2.28560008620039,.547225517982271,.348638934287124,-.363480297113412,-.618695487400892,2.39319484717323,1.52471124557473,-1.58961734284587,4.87179403320196,3.10383383936015,5.49730994822615,-.236321146004056,-.150560874408057,.156970165907756,-1.36209257652082,-.867793055767227,-3.77635721863404,.0946962283839159,.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,-1.51514018159912,-.965300119862265,1.00639240141108,1.71302390303713,.237661341149238,-3.36446686788906,-2.14351141382237,2.23475948529486,3.80388048506199,-1.91964817870762,-1.22301331540983,1.27507630270228,1.95173675811295,1.24345704101385,2.75500656898115,1.04781596872157,.667566534563461,-.695984465482171,-1.18466517033958,3.55510980002442,2.2649705675878,-2.36138908644035,4.59045028887451,2.92458893939423,1.24891795372526,-.626153181688296,-.398923973340748,.415905941872819,-2.84473124098146,-1.81238636712449,-5.48872263996485,.347192252628353,.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,-1.94896143440088,-1.24168887412682,1.29454686902578,2.2035040478582,.305709527061048,-2.83314651369184,-1.80500570449659,1.88184377891507,3.20316744314249,.637278791777431,.406012131085817,-.423295838593301,2.63766838605638,1.68046603255685,-.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141283,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843812,-1.33269841309056,-.849065950254409,.885210208841138,-4.37531622836174,-2.78752641722509,-4.52786743647192,.95054294779362,.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,-2.36915365304765,-1.50939453196029,1.57364860569271,2.67857514897071,.371619894588009,-1.68415540504794,-1.07298020016939,1.11865636189276,1.90411323119761,2.32831086437971,1.48337347495979,-1.54651984745118,.381749900992996,.243213947948102,-2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188742,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001399,-2.69130383032704,-2.39387697960251,-1.52514583383108,1.59007043171885,-4.93737708601551,-3.14561698874585,-.508166439915618,2.07554270972356,1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,-2.74931534527504,-1.75159662753591,1.82616110784379,3.10838756746077,.431251167473247,-.106757773657142,-.0680156813349527,.0709110705135131,.120700790879792,2.26282036321899,1.44164929037179,-1.50301948784969,-2.15044627554921,-1.37005543987348,.119501749587716,3.45007465931297,2.19805238043167,-2.29162223072931,-3.90066901628345,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,-3.74106981000851,-2.38344621859103,2.48490817973374,-3.58481468643821,-2.28389563582348,3.59724489879787,3.76406389114815,2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,-3.06556135691961,-1.95307786118934,2.03621928395244,3.46593664692319,.480856776362333,1.5947619510804,1.01602737631215,-1.05927908845788,-1.80304461367463,.444911232504622,.283454211851926,-.295520697836744,-2.09440936915946,-1.33435416739534,2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611565,-4.84964573066989,-.964148801340396,-.614261944257265,.640410728572749,-2.01722273728898,-1.28517834515454,3.46190525493913,-5.18623489409549,-3.30416500488362,3.44482144542743,-.205483038001613,-.130913828127032,3.63947886955733,5.81833621896758,3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,-3.29802783242322,-2.10118291403716,2.19062908534965,3.72876422817854,.517320923384106,3.07100849759875,1.95654825118803,-2.0398374063157,-3.47209520918981,-2.20598407348367,-1.40543872850124,1.46526746277611,.923194488569861,.588169835027997,.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873902,-2.426628794168,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,-6.45780747384476,-4.11428751282488,4.2894304115822,4.10479708970298,2.61517790433632,-1.32544341327556,7.79428043908953,4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,-3.43216369764802,-2.18664125535227,2.27972534004502,3.88041880521058,.538361161120983,4.00974905232829,2.55462252942968,-2.6633713691361,-4.53343925473613,-4.32546942586304,-2.75576887572707,2.87308040304257,4.4144312457315,2.81244670509951,-4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,-7.27823584467216,-4.63698476184225,4.83437858766288,7.46210387346587,4.75412760808659,-7.31197439196322,9.14177937313726,5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,-.0319379183519753,-.0324893117176766,-.00124359624768512,.031645154192253,.0381313330936757,-.154660030302817,-.157330164177211,-.00602214055814407,.153242313889452,-.43925822557731,-.446841815317497,-.0171038035473811,-.961797763577864,-.978402756631806,-1.79684767953686,.000465578153648587,.000473616144898061,1.81286469184544e-005,-.000461310355481611,.00319092434145473,.00324601417302388,.000124247970525588,.0123137463285655,.012526337462346,.0353622090543082,-5.86464816610367e-006,-5.96589860075151e-006,-2.28357226493352e-007,-5.19109369939388e-005,-5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,-.166073783576779,-.168940970519646,-.00646656841622802,.164551441038203,.198278882483946,-.764836159193308,-.778040701653472,-.0297811324981311,.757825163693694,-2.02218309922343,-2.05709515348012,-.0787396124921617,-4.02885967082133,-4.09841606758444,-6.68010475273478,.0125887190510936,.0128060574566476,.000490178589876773,-.0124733225023728,.0823390061849571,.0837605510092484,.00320610999159909,.299004079114613,.304166246215336,.796298871799325,-.000824565926326729,-.000838801675185275,-3.21068856415291e-005,-.00697737063211093,-.00709783170785957,-.0321260785529637,5.01144631645353e-005,5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,-.398540259080393,-.405420872037066,-.015518330446016,.394886974518158,.475825356015865,-1.67165845871656,-1.70051886866409,-.0650909105855661,1.6563349285852,-3.84755914106515,-3.91398546692074,-.149815966677047,-6.3113399865174,-6.42030234909073,-7.98777931164405,.0724976248699598,.0737492627952912,.00282290703160025,-.0718330635538726,.434860760346183,.442368430061905,.0169325753770755,1.40707187748205,1.43136432250754,3.23180916098999,-.0113956405123911,-.0115923809684749,-.0004437225879236,-.0887334889355571,-.0902654315677738,-.36808995603673,.00166206328995187,.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,-.714786270724957,-.727126724577601,-.0278322937135137,.708234065298678,.853397929051749,-2.59853682213183,-2.6433993582315,-.10118163017136,2.57471691018525,-4.7549644652192,-4.83705672356223,-.185148446522542,-5.36731919542456,-5.45998347614275,-3.25729414076559,.233201739437629,.237227859491313,.00908039168484083,-.231064057614915,1.22672710731592,1.24790598293055,.0477662072687194,3.30495822058058,3.36201679531997,5.89679444984459,-.0657431866024218,-.0668782122732034,-.00255990322490875,-.451524561772794,-.459319924425276,-1.59193561065057,.017197430813833,.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,-1.09494796295235,-1.11385173231848,-.042635000914322,1.08491094310714,1.30728073869584,-3.24473228828999,-3.30075109017009,-.126343140340603,3.21498892012255,-4.05017972204568,-4.12010419835037,-.157705591526285,-1.63294619261043,-1.66113825253464,2.28560008620039,.547225517982271,.55667311317238,.0213078258129713,-.542209285918405,2.39319484717323,2.43451224831938,.0931860398476283,4.87179403320196,4.95590330339708,5.49730994822615,-.236321146004056,-.240401121148633,-.00920185489986014,-1.36209257652082,-1.38560847465611,-3.77635721863404,.0946962283839159,.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,-1.51514018159912,-1.54129837497396,-.0589964137232844,1.50125140095791,1.80895681150725,-3.36446686788906,-3.42255283189436,-.131005356341752,3.33362593315312,-1.91964817870762,-1.95279001506978,-.0747471155162795,1.95173675811295,1.98543258954328,2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063127,-1.03821099255249,3.55510980002442,3.61648712494209,.138428596350638,4.59045028887451,4.66970228947898,1.24891795372526,-.626153181688296,-.636963426396325,-.0243811052053825,-2.84473124098146,-2.89384420968737,-5.48872263996485,.347192252628353,.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,-1.94896143440088,-1.98260935074574,-.0758885128327108,1.93109596018978,2.32690486658695,-2.83314651369184,-2.88205947758122,-.110316844590437,2.80717601371157,.637278791777431,.648281114842891,.0248143133691988,2.63766838605638,2.68320650944122,-.808065843843811,1.73374759666501,1.76367994615688,.067508532733652,-1.71785491622659,4.12740556141283,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843812,-1.33269841309056,-1.35570683401508,-.0518925099589806,-4.37531622836174,-4.45085403872017,-4.52786743647192,.95054294779362,.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,-2.36915365304765,-2.41005599340123,-.0922499256416731,2.34743641804054,2.82858093939836,-1.68415540504794,-1.71323156899255,-.0655775157362815,1.66871731961812,2.32831086437971,2.36850807266163,.0906595922460986,.381749900992996,.388340636274758,-2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622357,-2.53843733881177,3.58714638367868,3.64907680538234,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,-2.39387697960251,-2.4352061570422,-.0932126006787377,-4.93737708601551,-5.02261861703307,-.508166439915618,2.07554270972356,2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,-2.74931534527504,-2.7967810011421,-.107052632842481,2.72411329584901,3.28246374904245,-.106757773657142,-.108600897231078,-.00415692611322429,.105779161098541,2.26282036321899,2.30188690834251,.0881095280677891,-2.15044627554921,-2.18757273479106,.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.4184489866057,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,-3.74106981000851,-3.80565764777524,-.145669493162303,-3.58481468643821,-3.64670485341609,3.59724489879787,3.76406389114815,3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,-3.06556135691961,-3.11848685368264,-.119366596109979,3.03746038662953,3.66003632207834,1.5947619510804,1.62229477839744,.0620967201893232,-1.58014330533832,.444911232504622,.452592418790165,.0173239199033348,-2.09440936915946,-2.13056837715884,2.63186985254377,4.2894282422279,4.36348323472717,.167021432333119,-4.25010849784966,-.964148801340396,-.98079438405867,-.0375419530619799,-2.01722273728898,-2.05204915464121,3.46190525493913,-5.18623489409549,-5.27577283777777,-.201941221845797,-.205483038001613,-.20903060758203,3.63947886955733,5.81833621896758,5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,-3.29802783242322,-3.35496675520006,-.128418358139767,3.26779592010948,3.93758279561025,3.07100849759875,3.12402803860063,.119578696522845,-3.04285759691611,-2.20598407348367,-2.24406936639925,-.0858964409190686,.923194488569861,.939133013954983,.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873902,-3.87460743053656,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,-6.45780747384476,-6.56929852924652,-.25145361873174,4.10479708970298,4.17566451229335,-1.32544341327556,7.79428043908953,7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,-3.43216369764802,-3.49141841400203,-.13364133030831,3.40070220695543,4.09773034500053,4.00974905232829,4.07897551472765,.156131305216524,-3.97299303311744,-4.32546942586304,-4.40014665135235,-.168424801181427,4.4144312457315,4.49064435720703,-4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,-7.27823584467216,-7.40389121595289,-.28339939655034,7.46210387346587,7.59093364435534,-7.31197439196322,9.14177937313726,9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,-.0319379183519753,-.0425895510116904,-.0277895142706455,.00117913035023065,.0297640501007841,-.154660030302817,-.206240781926363,-.134571297792104,.00570996311627577,-.43925822557731,-.585755477154869,-.382203141612094,-.961797763577864,-1.28256746328293,-1.79684767953686,.000465578153648587,.000620853379055167,.000405104384309337,-1.71888889439261e-005,.00319092434145473,.00425513126926153,.00277645639124976,.0123137463285655,.0164205106275604,.0353622090543082,-5.86464816610367e-006,-7.82057019228999e-006,-5.10289124485106e-006,-5.19109369939388e-005,-6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,-.166073783576779,-.221461142188379,-.144502522920425,.00613135259583705,.154769899538562,-.764836159193308,-1.01991708658492,-.665491881382991,.0282373296477252,-2.02218309922343,-2.69660249304008,-1.75952250506473,-4.02885967082133,-5.37252686958436,-6.68010475273478,.0125887190510936,.0167871896471046,.0109535751204635,-.000464768572193061,.0823390061849571,.109799933312678,.0716440239592627,.299004079114613,.398725093710568,.796298871799325,-.000824565926326729,-.00109956736062704,-.00071746337155817,-.00697737063211093,-.00930439734215142,-.0321260785529637,5.01144631645353e-005,6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,-.398540259080393,-.531457639388288,-.346774016236294,.0147138868003716,.371413443660288,-1.67165845871656,-2.22917418911978,-1.45452737733324,.0617167093759977,-3.84755914106515,-5.13076069911761,-3.34780114662785,-6.3113399865174,-8.41623844737545,-7.98777931164405,.0724976248699598,.0966763475074085,.0630809359431451,-.002676572370981,.434860760346183,.579891411113903,.378376861425818,1.40707187748205,1.87634519190449,3.23180916098999,-.0113956405123911,-.0151962068293123,-.00991546509750553,-.0887334889355571,-.118327043544488,-.36808995603673,.00166206328995187,.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,-.714786270724957,-.953175031760129,-.621942953572118,.0263895153231816,.666134033494215,-2.59853682213183,-3.46517626442722,-2.26101386266496,.0959365479342401,-4.7549644652192,-6.34079527029531,-4.13734393811777,-5.36731919542456,-7.15737676442222,-3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831002,1.22672710731592,1.63585353791591,1.06738798948579,3.30495822058058,4.4071966513386,5.89679444984459,-.0657431866024218,-.0876692328212839,-.0572038290823835,-.451524561772794,-.602112766253303,-1.59193561065057,.017197430813833,.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,-1.09494796295235,-1.46012465838812,-.95272558802182,.0404248755605102,1.02041985541771,-3.24473228828999,-4.32688473530227,-2.82327524550585,.11979372855449,-4.05017972204568,-5.4009573811209,-3.52410341735765,-1.63294619261043,-2.17755097302669,2.28560008620039,.547225517982271,.729731000653746,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367101,2.08234367785444,4.87179403320196,6.49658873611033,5.49730994822615,-.236321146004056,-.315136741011867,-.205625482170818,-1.36209257652082,-1.81636481949508,-3.77635721863404,.0946962283839159,.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,-1.51514018159912,-2.02045541424857,-1.31833919902212,.0559381407795162,1.41201150863478,-3.36446686788906,-4.4865520602846,-2.92745754610214,.124214263271281,-1.91964817870762,-2.55987109498181,-1.67030580640284,1.95173675811295,2.60266155489591,2.75500656898115,1.04781596872157,1.39727364720762,.911715550981012,-.0386847883193889,3.55510980002442,4.74077641992773,3.0933379389912,4.59045028887451,6.12141388513833,1.24891795372526,-.626153181688296,-.834981872710088,-.544822382907866,-2.84473124098146,-3.79347911889772,-5.48872263996485,.347192252628353,.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,-1.94896143440088,-2.59896063091726,-1.69581157410878,.0719545824309805,1.81630452989169,-2.83314651369184,-3.77803076303363,-2.46515019062261,.104598208499711,.637278791777431,.849817990926997,.55450289173293,2.63766838605638,3.51735861632833,-.808065843843811,1.73374759666501,2.31197070864005,1.50855178059943,-.064009006140712,4.12740556141283,5.50393885468463,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843812,-1.33269841309056,-1.77716739172311,-1.15959472274786,-4.37531622836174,-5.8345303456991,-4.52786743647192,.95054294779362,1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,-2.36915365304765,-3.15929138677771,-2.06142518510908,.0874678476499865,2.20789618310877,-1.68415540504794,-2.24583899888368,-1.46540110005888,.0621780897499608,2.32831086437971,3.10482709694771,2.02588745115913,.381749900992996,.50906751987416,-2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273614,3.58714638367868,4.78349754217328,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,-2.39387697960251,-3.19226020392741,-2.08293720858961,-4.93737708601551,-6.58404442855625,-.508166439915618,2.07554270972356,2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,-2.74931534527504,-3.6662410134057,-2.39220781955879,.101503207887315,2.56218200503226,-.106757773657142,-.142362617365695,-.0928910470423957,.00394143819036155,2.26282036321899,3.01749482284837,1.96890348630228,-2.15044627554921,-2.86764279249033,.119501749587716,3.45007465931297,4.60071095140733,3.00194577372105,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,-3.74106981000851,-4.9887560540339,-3.25514367429686,-3.58481468643821,-4.78038820054912,3.59724489879787,3.76406389114815,5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,-3.06556135691961,-4.08795840577754,-2.66737675689461,.113178836410125,2.85690259486619,1.5947619510804,2.12663188500128,1.38761892782617,-.0588777326765029,.444911232504622,.593293821521023,.387121881618236,-2.09440936915946,-2.79291698671264,2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640683,-.964148801340396,-1.28570259705192,-.83891587989734,-2.01722273728898,-2.68998779997356,3.46190525493913,-5.18623489409549,-6.91589893787678,-4.51259681489188,-.205483038001613,-.274013798814207,3.63947886955733,5.81833621896758,7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,-3.29802783242322,-4.39795490297745,-2.86964825021048,.12176137061466,3.07354613898791,3.07100849759875,4.09522222779324,2.67211637191123,-.113379941906936,-2.20598407348367,-2.94170303097044,-1.91944964063812,.923194488569861,1.23108958885198,.507379758856801,4.96464287844228,6.62040364808344,4.31978730278895,-.183291880031401,-3.80884940873902,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,-6.45780747384476,-8.61155438605813,-5.61900531555837,4.10479708970298,5.47379022059324,-1.32544341327556,7.79428043908953,10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,-3.43216369764802,-4.57682649415414,-2.98636125886026,.126713592860266,3.19855198842569,4.00974905232829,5.34704266035319,3.48892427303921,-.148037726033597,-4.32546942586304,-5.76805910494959,-3.76363584617939,4.4144312457315,5.8866906319401,-4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,-7.27823584467216,-9.70560427278213,-6.33286855712842,7.46210387346587,9.95079423294563,-7.31197439196322,9.14177937313726,12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,-.0319379183519753,-.0500140463222003,-.0518576912013894,-.0415275099413171,-.0223217196472688,-.154660030302817,-.24219405407586,-.251121942525088,-.201097826062499,-.43925822557731,-.687868288629403,-.713224862474736,-.961797763577864,-1.50615319946369,-1.79684767953686,.000465578153648587,.000729084691473652,.000755960606624685,.000605371370823812,.00319092434145473,.00499691420436607,.00518111316265944,.0123137463285655,.0192830438371037,.0353622090543082,-5.86464816610367e-006,-9.18390428102533e-006,-9.52244633830417e-006,-5.19109369939388e-005,-8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,-.166073783576779,-.260067729310804,-.269654486884813,-.215938641412748,-.116070571566627,-.764836159193308,-1.1977158528687,-1.24186670359493,-.994483762912882,-2.02218309922343,-3.16669227065688,-3.28342459824587,-4.02885967082133,-6.30910167728468,-6.68010475273478,.0125887190510936,.0197136447951608,.0204403398568298,.0163685732441572,.0823390061849571,.128940991873039,.13369408456444,.299004079114613,.468233518150735,.796298871799325,-.000824565926326729,-.00129125129526923,-.00133885010037302,-.00697737063211093,-.0109264021013695,-.0321260785529637,5.01144631645353e-005,7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,-.398540259080393,-.624104888716883,-.647110981340287,-.518204861963161,-.278543636855161,-1.67165845871656,-2.61777873930361,-2.71427671758157,-2.17358603453451,-3.84755914106515,-6.02518921139496,-6.24729300320554,-6.3113399865174,-9.88341340931671,-7.98777931164405,.0724976248699598,.113529614838378,.11771460504762,.0942655625163293,.434860760346183,.680981959385669,.706084685269139,1.40707187748205,2.20344223207683,3.23180916098999,-.0113956405123911,-.0178453112143479,-.0185031347484812,-.0887334889355571,-.138954604977606,-.36808995603673,.00166206328995187,.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,-.714786270724957,-1.11933887677121,-1.16060055303996,-.929406031924798,-.499571029238743,-2.59853682213183,-4.06924896099705,-4.21925182103647,-3.37876634868698,-4.7549644652192,-7.44616510000515,-7.72064967553201,-5.36731919542456,-8.40509854032022,-3.25729414076559,.233201739437629,.365188565907959,.378650344801755,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064312,5.89679444984459,-.0657431866024218,-.102952319693436,-.106747403902083,-.451524561772794,-.707077089737688,-1.59193561065057,.017197430813833,.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,-1.09494796295235,-1.7146633520688,-1.77787020176478,-1.42371402906152,-.765269708188692,-3.24473228828999,-5.08117621439183,-5.26848128513792,-4.21898669004741,-4.05017972204568,-6.34248838569497,-6.57628863172099,-1.63294619261043,-2.55715621899877,2.28560008620039,.547225517982271,.856942588087486,.888531670338266,.711533948435089,2.39319484717323,3.74768814487268,3.88583745543935,4.87179403320196,7.62911752450546,5.49730994822615,-.236321146004056,-.370073484889954,-.383715334269618,-1.36209257652082,-2.13300568008586,-3.77635721863404,.0946962283839159,.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,-1.51514018159912,-2.37267471198348,-2.46013753301878,-1.97007200846434,-1.05894610873647,-3.36446686788906,-5.26867781509235,-5.46289467143736,-4.37467244553793,-1.91964817870762,-3.0061249410246,-3.1169383486426,1.95173675811295,3.05637492003247,2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591416,1.36243031247303,3.55510980002442,5.56722026499827,5.77244215424071,4.59045028887451,7.1885396848779,1.24891795372526,-.626153181688296,-.980541495928355,-1.01668674800428,-2.84473124098146,-4.45478376427993,-5.48872263996485,.347192252628353,.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,-1.94896143440088,-3.05202882623936,-3.16453436679059,-2.53415123836086,-1.36214797290772,-2.83314651369184,-4.43664234772504,-4.6001882618506,-3.68381930162858,.637278791777431,.997963945475626,1.03475143356997,2.63766838605638,4.13052808940267,-.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141283,6.4634222783235,6.70168045166518,1.98940127224618,3.11535668501805,-2.96832239843812,-1.33269841309056,-2.08697509460821,-2.16390630168515,-4.37531622836174,-6.85164468857583,-4.52786743647192,.95054294779362,1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,-2.36915365304765,-3.71004018615404,-3.8468016980446,-3.08050921776368,-1.6558243734555,-1.68415540504794,-2.63734867013603,-2.7345680458079,-2.18983527963514,2.32831086437971,3.64608131714141,3.78048514370594,.381749900992996,.597811573855878,-2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635524,3.33115715545234,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,-2.39387697960251,-3.74875634795768,-3.88694503628022,-4.93737708601551,-7.73181907803745,-.508166439915618,2.07554270972356,3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,-2.74931534527504,-4.30536466145163,-4.46407134676942,-3.5748172149004,-1.92152305240544,-.106757773657142,-.16718022074295,-.173342908623152,-.138812569451129,2.26282036321899,3.54352469706742,3.67414802589246,-2.15044627554921,-3.3675494587808,.119501749587716,3.45007465931297,5.40273764744738,5.60189629155853,4.48598459696892,1.77035093065653,2.7723288820058,2.87452397106015,-3.42380239520456,-5.36159588761007,1.33170919606025,-3.74106981000851,-5.85842936705581,-6.07438596565011,-3.58481468643821,-5.61373743608465,3.59724489879787,3.76406389114815,5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,-3.06556135691961,-4.80059864950596,-4.97756091846909,-3.98601838486204,-2.14255044478903,1.5947619510804,2.49736057507455,2.58941963373659,2.07360079279396,.444911232504622,.696720767591157,.72240366631615,-2.09440936915946,-3.27979695088037,2.63186985254377,4.2894282422279,6.71714607327471,6.96475715335038,5.57736018045175,-.964148801340396,-1.50983487071459,-1.56549122161564,-2.01722273728898,-3.15892446006539,3.46190525493913,-5.18623489409549,-8.12152468439221,-8.42090472917074,-.205483038001613,-.321781716436834,3.63947886955733,5.81833621896758,9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,-3.29802783242322,-5.16463580891203,-5.35501741292456,-4.28828460541245,-2.30502351007756,3.07100849759875,4.80912874954551,4.9864054592391,3.99310107128126,-2.20598407348367,-3.45452037324197,-3.58186277642042,.923194488569861,1.44570135775136,.507379758856801,4.96464287844228,7.7745166798312,8.06110512843118,6.45531290809181,-3.80884940873902,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,-6.45780747384476,-10.1127781284249,-10.4855608369157,4.10479708970298,6.42801793274624,-1.32544341327556,7.79428043908953,12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,-3.43216369764802,-5.37468949190064,-5.57281420860799,-4.46269573688388,-2.39877236199692,4.00974905232829,6.27917489029831,6.5106412373467,5.21370528571982,-4.32546942586304,-6.77358573667404,-7.02327732701018,4.4144312457315,6.91289790468793,-4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,-7.27823584467216,-11.3975501068473,-11.8176927763526,7.46210387346587,11.6854832203471,-7.31197439196322,9.14177937313726,14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,-.0319379183519753,-.054298066631689,-.0675012177634823,-.075365210691831,-.07886883956794,-.154660030302817,-.262939511058368,-.326876043550518,-.364957592022401,-.43925822557731,-.746788570738901,-.928378137184539,-.961797763577864,-1.63516477461472,-1.79684767953686,.000465578153648587,.000791535419924791,.000984005657525685,.00109864378994577,.00319092434145473,.0054249316016862,.00674406129057573,.0123137463285655,.0209347588552154,.0353622090543082,-5.86464816610367e-006,-9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,-8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,-.166073783576779,-.282344179950942,-.350999163642368,-.391891091702267,-.410109589767539,-.764836159193308,-1.30030781284792,-1.61649145672597,-1.80481513164276,-2.02218309922343,-3.43793955015111,-4.2739110377176,-4.02885967082133,-6.84951625494143,-6.68010475273478,.0125887190510936,.021402243516657,.0266064261611109,.029706114645781,.0823390061849571,.139985605707144,.174024591276672,.299004079114613,.508340689012668,.796298871799325,-.000824565926326729,-.00140185515919078,-.00174273111826309,-.00697737063211093,-.0118623177481531,-.0321260785529637,5.01144631645353e-005,8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,-.398540259080393,-.677563431168936,-.842320170000573,-.940451730878478,-.984172086870924,-1.67165845871656,-2.84200809304711,-3.53307252095549,-3.94468075935688,-3.84755914106515,-6.54128488547504,-8.13186772631248,-6.3113399865174,-10.7299904588883,-7.98777931164405,.0724976248699598,.123254146474207,.153224700235957,.1710756072316,.434860760346183,.739312383715057,.919084035475437,1.40707187748205,2.39218103581589,3.23180916098999,-.0113956405123911,-.0193738753136992,-.0240848387004708,-.0887334889355571,-.150856948291044,-.36808995603673,.00166206328995187,.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,-.714786270724957,-1.21521735159799,-1.51071034695561,-1.68671036412374,-1.76512329607374,-2.59853682213183,-4.41780594598826,-5.49204234459095,-6.13187350041211,-4.7549644652192,-8.08397637227704,-10.049681020309,-5.36731919542456,-9.12504854432544,-3.25729414076559,.233201739437629,.396469282990198,.492874996717867,.550295671800536,1.22672710731592,2.0855745668007,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984459,-.0657431866024218,-.111770838907705,-.138949104550336,-.451524561772794,-.767642727504649,-1.59193561065057,.017197430813833,.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,-1.09494796295235,-1.86153514438243,-2.31418717000874,-2.58379343998139,-2.70391057656362,-3.24473228828999,-5.51641118734948,-6.85778510905627,-7.65672734172536,-4.05017972204568,-6.88576274670089,-8.56010903534823,-1.63294619261043,-2.776192869346,2.28560008620039,.547225517982271,.930345156393643,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070526,4.87179403320196,8.28259983241363,5.49730994822615,-.236321146004056,-.401772626268325,-.499467904136599,-1.36209257652082,-2.31571114636377,-3.77635721863404,.0946962283839159,.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,-1.51514018159912,-2.57590935107797,-3.20226904625419,-3.57533818439403,-3.74155092353066,-3.36446686788906,-5.7199734894835,-7.11084574520705,-7.93926992086965,-1.91964817870762,-3.26361861024795,-4.05720211125474,1.95173675811295,3.31817281869637,2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898136,3.55510980002442,6.04408799389592,7.5137721301034,4.59045028887451,7.80428370621122,1.24891795372526,-.626153181688296,-1.06453109468611,-1.32338312764,-2.84473124098146,-4.83636425064191,-5.48872263996485,.347192252628353,.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,-1.94896143440088,-3.31345445440235,-4.11915606854152,-4.59904391748785,-4.81284738096772,-2.83314651369184,-4.81666890668704,-5.98789306701478,-6.68549157438517,.637278791777431,1.08344588754871,1.34689725343212,2.63766838605638,4.48433402043889,-.808065843843811,1.73374759666501,2.94756663562366,3.66429874441148,4.09119503456581,4.12740556141283,7.01705539432373,8.72332687917668,1.98940127224618,3.38220674597772,-2.96832239843812,-1.33269841309056,-2.26573774926528,-2.81667592802333,-4.37531622836174,-7.4385315165314,-4.52786743647192,.95054294779362,1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,-2.36915365304765,-4.02782866109789,-5.00723794478697,-5.59058866190049,-5.85048772793475,-1.68415540504794,-2.86325431258853,-3.55948498424847,-3.97417031416819,2.32831086437971,3.95839130925952,4.9209161654707,.381749900992996,.649017927110068,-2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830963,-1.52344991306454,-2.59003683420207,-2.69130383032704,-2.39387697960251,-4.06986110773887,-5.05949102727521,-4.93737708601551,-8.39409842544753,-.508166439915618,2.07554270972356,3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,-2.74931534527504,-4.67414645388234,-5.81071476784009,-6.48767173775814,-6.78927500842464,-.106757773657142,-.181500267083405,-.225633983149896,-.251920680005669,2.26282036321899,3.84705006415361,4.78250111498198,-2.15044627554921,-3.65600143048661,.119501749587716,3.45007465931297,5.86551639682903,7.29177895707793,8.14128212331116,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429709,1.33170919606025,-3.74106981000851,-6.36024100270089,-7.90680110023493,-3.58481468643821,-6.09458965574696,3.59724489879787,3.76406389114815,6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,-3.06556135691961,-5.21180037431139,-6.47910494479514,-7.2339303710034,-7.57022621762745,1.5947619510804,2.71127534923244,3.37055072311448,3.76322492874004,.444911232504622,.756399320675517,.940325842964601,-2.09440936915946,-3.56073236366527,2.63186985254377,4.2894282422279,7.29251224169723,9.06576398575723,10.1219390639593,-.964148801340396,-1.63916180436132,-2.03774139204767,-2.01722273728898,-3.42950637731078,3.46190525493913,-5.18623489409549,-8.81718478891181,-10.9611768422514,-.205483038001613,-.349344361529711,3.63947886955733,5.81833621896758,9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,-3.29802783242322,-5.60701962552938,-6.97042595115334,-7.78249101017962,-8.14428871473084,3.07100849759875,5.22106113152625,6.4906175528327,7.24678421547911,-2.20598407348367,-3.75042195543293,-4.66237685613718,.923194488569861,1.5695348492102,.507379758856801,4.96464287844228,8.44045334766828,10.4928391542811,11.7152706263275,-3.80884940873902,-6.47547397319738,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,-6.45780747384476,-10.9790017210618,-13.6486625036235,4.10479708970298,6.97861844197597,-1.32544341327556,7.79428043908953,13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,-3.43216369764802,-5.83506573884863,-7.25392389703223,-8.09901689119005,-8.47552946493044,4.00974905232829,6.81702604882233,8.4746582765388,9.46196865399931,-4.32546942586304,-7.35378632026677,-9.14193750783762,4.4144312457315,7.50503145913621,-4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,-7.27823584467216,-12.3738225688191,-15.3826488429769,7.46210387346587,12.6864189798666,-7.31197439196322,9.14177937313726,15.5420569434015,-8.9987004064276,-10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[56],
		T o_output[144]
	) {
		static const T matrix[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,-.0319379183519753,.054298066631689,-.0675012177634823,.0753652106918314,-.07886883956794,-.154660030302817,.262939511058368,-.326876043550518,.364957592022403,-.43925822557731,.746788570738901,-.928378137184539,-.961797763577864,1.63516477461472,-1.79684767953686,.000465578153648587,-.000791535419924791,.000984005657525685,-.00109864378994577,.00319092434145473,-.0054249316016862,.00674406129057572,.0123137463285655,-.0209347588552154,.0353622090543082,-5.86464816610367e-006,9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,-1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,-.166073783576779,.282344179950942,-.350999163642368,.391891091702269,-.410109589767539,-.764836159193308,1.30030781284792,-1.61649145672597,1.80481513164277,-2.02218309922343,3.43793955015111,-4.27391103771759,-4.02885967082133,6.84951625494143,-6.68010475273478,.0125887190510936,-.021402243516657,.0266064261611109,-.0297061146457811,.0823390061849571,-.139985605707144,.174024591276672,.299004079114613,-.508340689012668,.796298871799325,-.000824565926326729,.00140185515919078,-.00174273111826309,-.00697737063211093,.0118623177481531,-.0321260785529637,5.01144631645353e-005,-8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,-.398540259080393,.677563431168936,-.842320170000572,.940451730878483,-.984172086870924,-1.67165845871656,2.84200809304711,-3.53307252095548,3.9446807593569,-3.84755914106515,6.54128488547504,-8.13186772631248,-6.3113399865174,10.7299904588883,-7.98777931164405,.0724976248699598,-.123254146474207,.153224700235957,-.171075607231601,.434860760346183,-.739312383715057,.919084035475437,1.40707187748205,-2.39218103581589,3.23180916098999,-.0113956405123911,.0193738753136992,-.0240848387004708,-.0887334889355571,.150856948291044,-.36808995603673,.00166206328995187,-.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,-.714786270724957,1.21521735159799,-1.51071034695561,1.68671036412375,-1.76512329607374,-2.59853682213183,4.41780594598826,-5.49204234459095,6.13187350041214,-4.7549644652192,8.08397637227704,-10.0496810203089,-5.36731919542456,9.12504854432544,-3.25729414076559,.233201739437629,-.396469282990198,.492874996717867,-.550295671800539,1.22672710731592,-2.0855745668007,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984459,-.0657431866024218,.111770838907705,-.138949104550336,-.451524561772794,.767642727504649,-1.59193561065057,.017197430813833,-.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,-1.09494796295235,1.86153514438243,-2.31418717000874,2.5837934399814,-2.70391057656363,-3.24473228828999,5.51641118734948,-6.85778510905627,7.65672734172539,-4.05017972204568,6.88576274670089,-8.56010903534823,-1.63294619261043,2.776192869346,2.28560008620039,.547225517982271,-.930345156393643,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070526,4.87179403320196,-8.28259983241363,5.49730994822615,-.236321146004056,.401772626268325,-.499467904136599,-1.36209257652082,2.31571114636377,-3.77635721863404,.0946962283839159,-.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,-1.51514018159912,2.57590935107797,-3.20226904625418,3.57533818439405,-3.74155092353066,-3.36446686788906,5.7199734894835,-7.11084574520705,7.93926992086969,-1.91964817870762,3.26361861024795,-4.05720211125474,1.95173675811295,-3.31817281869637,2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898137,3.55510980002442,-6.04408799389592,7.51377213010339,4.59045028887451,-7.80428370621122,1.24891795372526,-.626153181688296,1.06453109468611,-1.32338312764,-2.84473124098146,4.83636425064191,-5.48872263996485,.347192252628353,-.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,-1.94896143440088,3.31345445440235,-4.11915606854152,4.59904391748787,-4.81284738096772,-2.83314651369184,4.81666890668704,-5.98789306701478,6.6854915743852,.637278791777431,-1.08344588754871,1.34689725343212,2.63766838605638,-4.48433402043889,-.808065843843811,1.73374759666501,-2.94756663562366,3.66429874441148,-4.09119503456583,4.12740556141283,-7.01705539432373,8.72332687917668,1.98940127224618,-3.38220674597772,-2.96832239843812,-1.33269841309056,2.26573774926528,-2.81667592802333,-4.37531622836174,7.4385315165314,-4.52786743647192,.95054294779362,-1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,-2.36915365304765,4.02782866109789,-5.00723794478697,5.59058866190052,-5.85048772793476,-1.68415540504794,2.86325431258853,-3.55948498424847,3.97417031416821,2.32831086437971,-3.95839130925952,4.9209161654707,.381749900992996,-.649017927110068,-2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,-2.39387697960251,4.06986110773887,-5.05949102727521,-4.93737708601551,8.39409842544753,-.508166439915618,2.07554270972356,-3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,-2.74931534527504,4.67414645388234,-5.81071476784009,6.48767173775818,-6.78927500842464,-.106757773657142,.181500267083405,-.225633983149896,.251920680005671,2.26282036321899,-3.84705006415361,4.78250111498198,-2.15044627554921,3.65600143048661,.119501749587716,3.45007465931297,-5.86551639682903,7.29177895707793,-8.1412821233112,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429709,1.33170919606025,-3.74106981000851,6.36024100270089,-7.90680110023493,-3.58481468643821,6.09458965574696,3.59724489879787,3.76406389114815,-6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,-3.06556135691961,5.21180037431139,-6.47910494479513,7.23393037100344,-7.57022621762746,1.5947619510804,-2.71127534923244,3.37055072311448,-3.76322492874006,.444911232504622,-.756399320675517,.940325842964601,-2.09440936915946,3.56073236366527,2.63186985254377,4.2894282422279,-7.29251224169723,9.06576398575723,-10.1219390639593,-.964148801340396,1.63916180436132,-2.03774139204767,-2.01722273728898,3.42950637731078,3.46190525493913,-5.18623489409549,8.81718478891181,-10.9611768422514,-.205483038001613,.349344361529711,3.63947886955733,5.81833621896758,-9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,-3.29802783242322,5.60701962552938,-6.97042595115334,7.78249101017965,-8.14428871473084,3.07100849759875,-5.22106113152625,6.4906175528327,-7.24678421547914,-2.20598407348367,3.75042195543293,-4.66237685613718,.923194488569861,-1.5695348492102,.507379758856801,4.96464287844228,-8.44045334766828,10.4928391542811,-11.7152706263276,-3.80884940873902,6.47547397319738,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,-6.45780747384476,10.9790017210618,-13.6486625036235,4.10479708970298,-6.97861844197597,-1.32544341327556,7.79428043908953,-13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,-3.43216369764802,5.83506573884863,-7.25392389703222,8.09901689119009,-8.47552946493044,4.00974905232829,-6.81702604882233,8.4746582765388,-9.46196865399936,-4.32546942586304,7.35378632026677,-9.14193750783762,4.4144312457315,-7.50503145913621,-4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,-7.27823584467216,12.3738225688191,-15.3826488429769,7.46210387346587,-12.6864189798666,-7.31197439196322,9.14177937313726,-15.5420569434015,-8.9987004064276,-10.9691712594614},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,-.0319379183519753,.0500140463222003,-.0518576912013895,.0415275099413171,-.0223217196472694,-.154660030302817,.24219405407586,-.251121942525088,.201097826062499,-.43925822557731,.687868288629403,-.713224862474738,-.961797763577864,1.50615319946369,-1.79684767953686,.000465578153648587,-.000729084691473652,.000755960606624687,-.000605371370823812,.00319092434145473,-.00499691420436607,.00518111316265945,.0123137463285655,-.0192830438371037,.0353622090543082,-5.86464816610367e-006,9.18390428102533e-006,-9.52244633830418e-006,-5.19109369939388e-005,8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,-1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,-.166073783576779,.260067729310804,-.269654486884814,.215938641412748,-.11607057156663,-.764836159193308,1.1977158528687,-1.24186670359493,.994483762912882,-2.02218309922343,3.16669227065688,-3.28342459824587,-4.02885967082133,6.30910167728468,-6.68010475273478,.0125887190510936,-.0197136447951608,.0204403398568299,-.0163685732441572,.0823390061849571,-.128940991873039,.13369408456444,.299004079114613,-.468233518150735,.796298871799325,-.000824565926326729,.00129125129526923,-.00133885010037302,-.00697737063211093,.0109264021013695,-.0321260785529637,5.01144631645353e-005,-7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,-.398540259080393,.624104888716883,-.647110981340288,.518204861963161,-.278543636855169,-1.67165845871656,2.61777873930361,-2.71427671758158,2.17358603453451,-3.84755914106515,6.02518921139496,-6.24729300320555,-6.3113399865174,9.88341340931671,-7.98777931164405,.0724976248699598,-.113529614838378,.11771460504762,-.0942655625163293,.434860760346183,-.680981959385669,.70608468526914,1.40707187748205,-2.20344223207683,3.23180916098999,-.0113956405123911,.0178453112143479,-.0185031347484812,-.0887334889355571,.138954604977606,-.36808995603673,.00166206328995187,-.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,-.714786270724957,1.11933887677121,-1.16060055303996,.929406031924799,-.499571029238757,-2.59853682213183,4.06924896099705,-4.21925182103647,3.37876634868698,-4.7549644652192,7.44616510000515,-7.72064967553202,-5.36731919542456,8.40509854032022,-3.25729414076559,.233201739437629,-.365188565907959,.378650344801756,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064312,5.89679444984459,-.0657431866024218,.102952319693436,-.106747403902083,-.451524561772794,.707077089737688,-1.59193561065057,.017197430813833,-.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,-1.09494796295235,1.7146633520688,-1.77787020176478,1.42371402906152,-.765269708188713,-3.24473228828999,5.08117621439183,-5.26848128513793,4.21898669004741,-4.05017972204568,6.34248838569497,-6.576288631721,-1.63294619261043,2.55715621899877,2.28560008620039,.547225517982271,-.856942588087486,.888531670338268,-.71153394843509,2.39319484717323,-3.74768814487268,3.88583745543936,4.87179403320196,-7.62911752450546,5.49730994822615,-.236321146004056,.370073484889954,-.383715334269619,-1.36209257652082,2.13300568008586,-3.77635721863404,.0946962283839159,-.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,-1.51514018159912,2.37267471198348,-2.46013753301879,1.97007200846434,-1.0589461087365,-3.36446686788906,5.26867781509235,-5.46289467143737,4.37467244553793,-1.91964817870762,3.0061249410246,-3.11693834864261,1.95173675811295,-3.05637492003247,2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002442,-5.56722026499827,5.77244215424072,4.59045028887451,-7.1885396848779,1.24891795372526,-.626153181688296,.980541495928355,-1.01668674800428,-2.84473124098146,4.45478376427993,-5.48872263996485,.347192252628353,-.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,-1.94896143440088,3.05202882623936,-3.1645343667906,2.53415123836086,-1.36214797290775,-2.83314651369184,4.43664234772504,-4.60018826185061,3.68381930162858,.637278791777431,-.997963945475626,1.03475143356998,2.63766838605638,-4.13052808940267,-.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633233,4.12740556141283,-6.4634222783235,6.70168045166519,1.98940127224618,-3.11535668501805,-2.96832239843812,-1.33269841309056,2.08697509460821,-2.16390630168515,-4.37531622836174,6.85164468857583,-4.52786743647192,.95054294779362,-1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,-2.36915365304765,3.71004018615404,-3.84680169804461,3.08050921776368,-1.65582437345554,-1.68415540504794,2.63734867013603,-2.73456804580791,2.18983527963514,2.32831086437971,-3.64608131714141,3.78048514370595,.381749900992996,-.597811573855878,-2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,-2.39387697960251,3.74875634795768,-3.88694503628023,-4.93737708601551,7.73181907803745,-.508166439915618,2.07554270972356,-3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,-2.74931534527504,4.30536466145163,-4.46407134676943,3.5748172149004,-1.9215230524055,-.106757773657142,.16718022074295,-.173342908623153,.13881256945113,2.26282036321899,-3.54352469706742,3.67414802589247,-2.15044627554921,3.3675494587808,.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155854,-4.48598459696892,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761007,1.33170919606025,-3.74106981000851,5.85842936705581,-6.07438596565012,-3.58481468643821,5.61373743608465,3.59724489879787,3.76406389114815,-5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,-3.06556135691961,4.80059864950596,-4.9775609184691,3.98601838486204,-2.14255044478909,1.5947619510804,-2.49736057507455,2.5894196337366,-2.07360079279396,.444911232504622,-.696720767591157,.722403666316151,-2.09440936915946,3.27979695088037,2.63186985254377,4.2894282422279,-6.71714607327471,6.96475715335039,-5.57736018045175,-.964148801340396,1.50983487071459,-1.56549122161564,-2.01722273728898,3.15892446006539,3.46190525493913,-5.18623489409549,8.12152468439221,-8.42090472917075,-.205483038001613,.321781716436834,3.63947886955733,5.81833621896758,-9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,-3.29802783242322,5.16463580891203,-5.35501741292457,4.28828460541245,-2.30502351007762,3.07100849759875,-4.80912874954551,4.98640545923911,-3.99310107128126,-2.20598407348367,3.45452037324197,-3.58186277642043,.923194488569861,-1.44570135775136,.507379758856801,4.96464287844228,-7.7745166798312,8.0611051284312,-6.45531290809182,-3.80884940873902,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,-6.45780747384476,10.1127781284249,-10.4855608369157,4.10479708970298,-6.42801793274624,-1.32544341327556,7.79428043908953,-12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,-3.43216369764802,5.37468949190064,-5.572814208608,4.46269573688388,-2.39877236199698,4.00974905232829,-6.27917489029831,6.51064123734671,-5.21370528571982,-4.32546942586304,6.77358573667404,-7.02327732701019,4.4144312457315,-6.91289790468793,-4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,-7.27823584467216,11.3975501068473,-11.8176927763526,7.46210387346587,-11.6854832203471,-7.31197439196322,9.14177937313726,-14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,-.0319379183519753,.0425895510116904,-.0277895142706456,-.00117913035023069,.029764050100784,-.154660030302817,.206240781926363,-.134571297792104,-.00570996311627599,-.43925822557731,.585755477154869,-.382203141612095,-.961797763577864,1.28256746328293,-1.79684767953686,.000465578153648587,-.000620853379055167,.000405104384309337,1.71888889439268e-005,.00319092434145473,-.00425513126926153,.00277645639124976,.0123137463285655,-.0164205106275604,.0353622090543082,-5.86464816610367e-006,7.82057019228999e-006,-5.10289124485107e-006,-5.19109369939388e-005,6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,-9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,-.166073783576779,.221461142188379,-.144502522920426,-.00613135259583728,.154769899538562,-.764836159193308,1.01991708658492,-.665491881382992,-.0282373296477263,-2.02218309922343,2.69660249304008,-1.75952250506474,-4.02885967082133,5.37252686958436,-6.68010475273478,.0125887190510936,-.0167871896471046,.0109535751204635,.000464768572193078,.0823390061849571,-.109799933312678,.0716440239592628,.299004079114613,-.398725093710568,.796298871799325,-.000824565926326729,.00109956736062704,-.000717463371558171,-.00697737063211093,.00930439734215142,-.0321260785529637,5.01144631645353e-005,-6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,-.398540259080393,.531457639388288,-.346774016236295,-.0147138868003722,.371413443660288,-1.67165845871656,2.22917418911978,-1.45452737733324,-.0617167093760001,-3.84755914106515,5.13076069911761,-3.34780114662786,-6.3113399865174,8.41623844737545,-7.98777931164405,.0724976248699598,-.0966763475074085,.0630809359431452,.0026765723709811,.434860760346183,-.579891411113903,.378376861425819,1.40707187748205,-1.87634519190449,3.23180916098999,-.0113956405123911,.0151962068293123,-.00991546509750554,-.0887334889355571,.118327043544488,-.36808995603673,.00166206328995187,-.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,-.714786270724957,.953175031760129,-.621942953572119,-.0263895153231826,.666134033494214,-2.59853682213183,3.46517626442722,-2.26101386266496,-.0959365479342438,-4.7549644652192,6.34079527029531,-4.13734393811778,-5.36731919542456,7.15737676442222,-3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831034,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.4071966513386,5.89679444984459,-.0657431866024218,.0876692328212839,-.0572038290823836,-.451524561772794,.602112766253303,-1.59193561065057,.017197430813833,-.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,-1.09494796295235,1.46012465838812,-.952725588021822,-.0404248755605117,1.02041985541771,-3.24473228828999,4.32688473530227,-2.82327524550586,-.119793728554495,-4.05017972204568,5.4009573811209,-3.52410341735766,-1.63294619261043,2.17755097302669,2.28560008620039,.547225517982271,-.729731000653746,.476146603536492,.0202032646552823,2.39319484717323,-3.19135057367101,2.08234367785444,4.87179403320196,-6.49658873611033,5.49730994822615,-.236321146004056,.315136741011867,-.205625482170818,-1.36209257652082,1.81636481949508,-3.77635721863404,.0946962283839159,-.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,-1.51514018159912,2.02045541424857,-1.31833919902212,-.0559381407795183,1.41201150863478,-3.36446686788906,4.4865520602846,-2.92745754610215,-.124214263271286,-1.91964817870762,2.55987109498181,-1.67030580640284,1.95173675811295,-2.60266155489591,2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981014,.0386847883193904,3.55510980002442,-4.74077641992773,3.09333793899121,4.59045028887451,-6.12141388513833,1.24891795372526,-.626153181688296,.834981872710088,-.544822382907866,-2.84473124098146,3.79347911889772,-5.48872263996485,.347192252628353,-.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,-1.94896143440088,2.59896063091726,-1.69581157410879,-.0719545824309833,1.81630452989169,-2.83314651369184,3.77803076303363,-2.46515019062261,-.104598208499715,.637278791777431,-.849817990926997,.55450289173293,2.63766838605638,-3.51735861632833,-.808065843843811,1.73374759666501,-2.31197070864005,1.50855178059943,.0640090061407145,4.12740556141283,-5.50393885468463,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843812,-1.33269841309056,1.77716739172311,-1.15959472274786,-4.37531622836174,5.8345303456991,-4.52786743647192,.95054294779362,-1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,-2.36915365304765,3.15929138677771,-2.06142518510909,-.0874678476499899,2.20789618310877,-1.68415540504794,2.24583899888368,-1.46540110005889,-.0621780897499632,2.32831086437971,-3.10482709694771,2.02588745115913,.381749900992996,-.50906751987416,-2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217328,3.12121330296808,-1.52344991306454,2.03153653003234,-2.69130383032704,-2.39387697960251,3.19226020392741,-2.08293720858962,-4.93737708601551,6.58404442855625,-.508166439915618,2.07554270972356,-2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,-2.74931534527504,3.6662410134057,-2.39220781955879,-.101503207887319,2.56218200503226,-.106757773657142,.142362617365695,-.0928910470423959,-.0039414381903617,2.26282036321899,-3.01749482284837,1.96890348630228,-2.15044627554921,2.86764279249033,.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,-3.74106981000851,4.9887560540339,-3.25514367429686,-3.58481468643821,4.78038820054912,3.59724489879787,3.76406389114815,-5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,-3.06556135691961,4.08795840577754,-2.66737675689462,-.113178836410129,2.85690259486619,1.5947619510804,-2.12663188500128,1.38761892782618,.0588777326765052,.444911232504622,-.593293821521023,.387121881618236,-2.09440936915946,2.79291698671264,2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.158363327640689,-.964148801340396,1.28570259705192,-.838915879897341,-2.01722273728898,2.68998779997356,3.46190525493913,-5.18623489409549,6.91589893787678,-4.51259681489189,-.205483038001613,.274013798814207,3.63947886955733,5.81833621896758,-7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,-3.29802783242322,4.39795490297745,-2.86964825021048,-.121761370614664,3.07354613898791,3.07100849759875,-4.09522222779324,2.67211637191124,.11337994190694,-2.20598407348367,2.94170303097044,-1.91944964063812,.923194488569861,-1.23108958885198,.507379758856801,4.96464287844228,-6.62040364808344,4.31978730278896,.183291880031408,-3.80884940873902,5.07914086475798,-3.31411940634737,2.02329330850018,-2.6980829707225,.0162698697481229,-6.45780747384476,8.61155438605813,-5.61900531555838,4.10479708970298,-5.47379022059324,-1.32544341327556,7.79428043908953,-10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,-3.43216369764802,4.57682649415414,-2.98636125886026,-.126713592860271,3.19855198842569,4.00974905232829,-5.34704266035319,3.48892427303922,.148037726033603,-4.32546942586304,5.76805910494959,-3.7636358461794,4.4144312457315,-5.8866906319401,-4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,-7.27823584467216,9.70560427278213,-6.33286855712843,7.46210387346587,-9.95079423294563,-7.31197439196322,9.14177937313726,-12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,-.0319379183519753,.0324893117176766,-.00124359624768518,-.0316451541922531,.0381313330936754,-.154660030302817,.157330164177211,-.00602214055814436,-.153242313889452,-.43925822557731,.446841815317497,-.0171038035473819,-.961797763577864,.978402756631806,-1.79684767953686,.000465578153648587,-.000473616144898061,1.81286469184553e-005,.000461310355481612,.00319092434145473,-.00324601417302388,.000124247970525594,.0123137463285655,-.012526337462346,.0353622090543082,-5.86464816610367e-006,5.96589860075151e-006,-2.28357226493363e-007,-5.19109369939388e-005,5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,-6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,-.166073783576779,.168940970519646,-.00646656841622833,-.164551441038204,.198278882483944,-.764836159193308,.778040701653472,-.0297811324981325,-.757825163693696,-2.02218309922343,2.05709515348012,-.0787396124921655,-4.02885967082133,4.09841606758444,-6.68010475273478,.0125887190510936,-.0128060574566476,.000490178589876797,.0124733225023729,.0823390061849571,-.0837605510092484,.00320610999159924,.299004079114613,-.304166246215336,.796298871799325,-.000824565926326729,.000838801675185275,-3.21068856415306e-005,-.00697737063211093,.00709783170785957,-.0321260785529637,5.01144631645353e-005,-5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,-.398540259080393,.405420872037066,-.0155183304460168,-.394886974518159,.47582535601586,-1.67165845871656,1.70051886866409,-.0650909105855692,-1.6563349285852,-3.84755914106515,3.91398546692074,-.149815966677054,-6.3113399865174,6.42030234909073,-7.98777931164405,.0724976248699598,-.0737492627952912,.00282290703160039,.0718330635538727,.434860760346183,-.442368430061905,.0169325753770763,1.40707187748205,-1.43136432250754,3.23180916098999,-.0113956405123911,.0115923809684749,-.000443722587923622,-.0887334889355571,.0902654315677738,-.36808995603673,.00166206328995187,-.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,-.714786270724957,.727126724577601,-.027832293713515,-.70823406529868,.853397929051741,-2.59853682213183,2.6433993582315,-.101181630171365,-2.57471691018525,-4.7549644652192,4.83705672356223,-.185148446522551,-5.36731919542456,5.45998347614275,-3.25729414076559,.233201739437629,-.237227859491313,.00908039168484127,.231064057614915,1.22672710731592,-1.24790598293055,.0477662072687217,3.30495822058058,-3.36201679531997,5.89679444984459,-.0657431866024218,.0668782122732034,-.00255990322490887,-.451524561772794,.459319924425276,-1.59193561065057,.017197430813833,-.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,-1.09494796295235,1.11385173231848,-.042635000914324,-1.08491094310714,1.30728073869583,-3.24473228828999,3.30075109017009,-.126343140340609,-3.21498892012256,-4.05017972204568,4.12010419835037,-.157705591526293,-1.63294619261043,1.66113825253464,2.28560008620039,.547225517982271,-.55667311317238,.0213078258129724,.542209285918406,2.39319484717323,-2.43451224831938,.0931860398476328,4.87179403320196,-4.95590330339708,5.49730994822615,-.236321146004056,.240401121148633,-.00920185489986058,-1.36209257652082,1.38560847465611,-3.77635721863404,.0946962283839159,-.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,-1.51514018159912,1.54129837497396,-.0589964137232872,-1.50125140095791,1.80895681150723,-3.36446686788906,3.42255283189436,-.131005356341758,-3.33362593315312,-1.91964817870762,1.95279001506978,-.074747115516283,1.95173675811295,-1.98543258954328,2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063147,1.0382109925525,3.55510980002442,-3.61648712494209,.138428596350645,4.59045028887451,-4.66970228947898,1.24891795372526,-.626153181688296,.636963426396325,-.0243811052053837,-2.84473124098146,2.89384420968737,-5.48872263996485,.347192252628353,-.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,-1.94896143440088,1.98260935074574,-.0758885128327144,-1.93109596018978,2.32690486658693,-2.83314651369184,2.88205947758122,-.110316844590442,-2.80717601371158,.637278791777431,-.648281114842891,.0248143133692,2.63766838605638,-2.68320650944122,-.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336552,1.7178549162266,4.12740556141283,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843812,-1.33269841309056,1.35570683401508,-.051892509958983,-4.37531622836174,4.45085403872017,-4.52786743647192,.95054294779362,-.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,-2.36915365304765,2.41005599340123,-.0922499256416775,-2.34743641804055,2.82858093939833,-1.68415540504794,1.71323156899255,-.0655775157362847,-1.66871731961812,2.32831086437971,-2.36850807266163,.090659592246103,.381749900992996,-.388340636274758,-2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622405,2.53843733881177,3.58714638367868,-3.64907680538234,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,-2.39387697960251,2.4352061570422,-.0932126006787421,-4.93737708601551,5.02261861703307,-.508166439915618,2.07554270972356,-2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,-2.74931534527504,2.7967810011421,-.107052632842487,-2.72411329584901,3.28246374904242,-.106757773657142,.108600897231078,-.00415692611322449,-.105779161098541,2.26282036321899,-2.30188690834251,.0881095280677934,-2.15044627554921,2.18757273479106,.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,-3.74106981000851,3.80565764777524,-.14566949316231,-3.58481468643821,3.64670485341609,3.59724489879787,3.76406389114815,-3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,-3.06556135691961,3.11848685368264,-.119366596109985,-3.03746038662953,3.6600363220783,1.5947619510804,-1.62229477839744,.0620967201893262,1.58014330533833,.444911232504622,-.452592418790165,.0173239199033356,-2.09440936915946,2.13056837715884,2.63186985254377,4.2894282422279,-4.36348323472717,.167021432333127,4.25010849784966,-.964148801340396,.98079438405867,-.0375419530619817,-2.01722273728898,2.05204915464121,3.46190525493913,-5.18623489409549,5.27577283777777,-.201941221845807,-.205483038001613,.20903060758203,3.63947886955733,5.81833621896758,-5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,-3.29802783242322,3.35496675520006,-.128418358139773,-3.26779592010949,3.93758279561022,3.07100849759875,-3.12402803860063,.119578696522851,3.04285759691612,-2.20598407348367,2.24406936639925,-.0858964409190727,.923194488569861,-.939133013954983,.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.91913366884931,-3.80884940873902,3.87460743053656,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,-6.45780747384476,6.56929852924652,-.251453618731752,4.10479708970298,-4.17566451229335,-1.32544341327556,7.79428043908953,-7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,-3.43216369764802,3.49141841400203,-.133641330308316,-3.40070220695544,4.09773034500049,4.00974905232829,-4.07897551472765,.156131305216531,3.97299303311745,-4.32546942586304,4.40014665135235,-.168424801181435,4.4144312457315,-4.49064435720703,-4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,-7.27823584467216,7.40389121595289,-.283399396550353,7.46210387346587,-7.59093364435534,-7.31197439196322,9.14177937313726,-9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,-.0319379183519753,.0203477386368136,.0212139303918346,-.0361091456847505,.00500970708930343,-.154660030302817,.0985343453312713,.102728896816887,-.174859285048361,-.43925822557731,.279852665057626,.291765835225991,-.961797763577864,.612764091356553,-1.79684767953686,.000465578153648587,-.000296621166194719,-.000309248161967448,.000526384631617547,.00319092434145473,-.00203294697585566,-.00211948838134797,.0123137463285655,-.00784512281993549,.0353622090543082,-5.86464816610367e-006,3.73638403043722e-006,3.89543979110572e-006,-5.19109369939388e-005,3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,-4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,-.166073783576779,.105806079951921,.110310185087204,-.187763722716786,.026049944826181,-.764836159193308,.487279293067771,.508022497640507,-.864726999689721,-2.02218309922343,1.28833860494792,1.34318245298509,-4.02885967082133,2.56679795782664,-6.68010475273478,.0125887190510936,-.00802030872060955,-.00836172873999091,.0142328590555997,.0823390061849571,-.0524584150880732,-.054691540214079,.299004079114613,-.190496349504475,.796298871799325,-.000824565926326729,.000525333297396549,.000547696439552502,-.00697737063211093,.00444530268091327,-.0321260785529637,5.01144631645353e-005,-3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,-.398540259080393,.253911132799744,.264719986484413,-.450591303972135,.0625140918479549,-1.67165845871656,1.06501836010644,1.11035559079056,-1.88998413076671,-3.84755914106515,2.45129087378787,2.55564094422791,-6.3113399865174,4.0209726602962,-7.98777931164405,.0724976248699598,-.0461884430503282,-.0481546590241491,.0819661216087145,.434860760346183,-.277051027559315,-.288844933452006,1.40707187748205,-.896449496423068,3.23180916098999,-.0113956405123911,.00726019498888195,.00756925739444607,-.0887334889355571,.056532358254944,-.36808995603673,.00166206328995187,-.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,-.714786270724957,.455392366453167,.474778162593069,-.80814038343456,.112119700737043,-2.59853682213183,1.65553520251249,1.7260104019959,-2.93791673259187,-4.7549644652192,3.02940138707022,3.15836129611741,-5.36731919542456,3.41953432868287,-3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066827,3.30495822058058,-2.10559828531328,5.89679444984459,-.0657431866024218,.0418851712113093,.0436681993244461,-.451524561772794,.287667582930097,-1.59193561065057,.017197430813833,-.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,-1.09494796295235,.697594462028796,.727290664744143,-1.23795280192461,.171750973622283,-3.24473228828999,2.06722817250133,2.15522890943138,-3.66850806244401,-4.05017972204568,2.58038102231816,2.69022638759908,-1.63294619261043,1.04035466481729,2.28560008620039,.547225517982271,-.348638934287124,-.363480297113411,.618695487400889,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320196,-3.10383383936015,5.49730994822615,-.236321146004056,.150560874408057,.156970165907755,-1.36209257652082,.867793055767227,-3.77635721863404,.0946962283839159,-.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,-1.51514018159912,.965300119862265,1.00639240141108,-1.71302390303713,.237661341149246,-3.36446686788906,2.14351141382237,2.23475948529486,-3.80388048506198,-1.91964817870762,1.22301331540983,1.27507630270228,1.95173675811295,-1.24345704101385,2.75500656898115,1.04781596872157,-.667566534563461,-.69598446548217,1.18466517033957,3.55510980002442,-2.2649705675878,-2.36138908644035,4.59045028887451,-2.92458893939423,1.24891795372526,-.626153181688296,.398923973340748,.415905941872818,-2.84473124098146,1.81238636712449,-5.48872263996485,.347192252628353,-.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,-1.94896143440088,1.24168887412682,1.29454686902578,-2.20350404785819,.305709527061058,-2.83314651369184,1.80500570449659,1.88184377891507,-3.20316744314248,.637278791777431,-.406012131085817,-.423295838593301,2.63766838605638,-1.68046603255685,-.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141283,-2.62958182530229,-2.74152163963164,1.98940127224618,-1.26745321062054,-2.96832239843812,-1.33269841309056,.849065950254409,.885210208841136,-4.37531622836174,2.78752641722509,-4.52786743647192,.95054294779362,-.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,-2.36915365304765,1.50939453196029,1.57364860569271,-2.6785751489707,.371619894588022,-1.68415540504794,1.07298020016939,1.11865636189275,-1.9041132311976,2.32831086437971,-1.48337347495979,-1.54651984745118,.381749900992996,-.243213947948102,-2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001399,-2.69130383032704,-2.39387697960251,1.52514583383108,1.59007043171885,-4.93737708601551,3.14561698874585,-.508166439915618,2.07554270972356,-1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,-2.74931534527504,1.75159662753591,1.82616110784378,-3.10838756746075,.431251167473262,-.106757773657142,.0680156813349527,.070911070513513,-.120700790879792,2.26282036321899,-1.44164929037179,-1.50301948784968,-2.15044627554921,1.37005543987348,.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,-3.74106981000851,2.38344621859103,2.48490817973374,-3.58481468643821,2.28389563582348,3.59724489879787,3.76406389114815,-2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,-3.06556135691961,1.95307786118934,2.03621928395244,-3.46593664692318,.48085677636235,1.5947619510804,-1.01602737631215,-1.05927908845788,1.80304461367463,.444911232504622,-.283454211851926,-.295520697836744,-2.09440936915946,1.33435416739534,2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340396,.614261944257265,.640410728572748,-2.01722273728898,1.28517834515454,3.46190525493913,-5.18623489409549,3.30416500488362,3.44482144542742,-.205483038001613,.130913828127032,3.63947886955733,5.81833621896758,-3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,-3.29802783242322,2.10118291403716,2.19062908534965,-3.72876422817853,.517320923384123,3.07100849759875,-1.95654825118803,-2.0398374063157,3.47209520918979,-2.20598407348367,1.40543872850124,1.4652674627761,.923194488569861,-.588169835027997,.507379758856801,4.96464287844228,-3.16298810157525,-3.2976347213703,5.61304625700718,-3.80884940873902,2.426628794168,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,-6.45780747384476,4.11428751282488,4.28943041158219,4.10479708970298,-2.61517790433632,-1.32544341327556,7.79428043908953,-4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,-3.43216369764802,2.18664125535227,2.27972534004502,-3.88041880521056,.538361161121001,4.00974905232829,-2.55462252942968,-2.66337136913609,4.53343925473611,-4.32546942586304,2.75576887572707,2.87308040304256,4.4144312457315,-2.81244670509951,-4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,-7.27823584467216,4.63698476184225,4.83437858766287,7.46210387346587,-4.75412760808659,-7.31197439196322,9.14177937313726,-5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,-.0319379183519753,.00692767387221854,.0340276241262534,-.0154583827232305,-.0303981920101794,-.154660030302817,.0335474040556334,.16477947382797,-.0748575381590809,-.43925822557731,.0952797768044285,.467998998326769,-.961797763577864,.208624155307574,-1.79684767953686,.000465578153648587,-.00010098884891692,-.000496041045803253,.000225346098359983,.00319092434145473,-.000692145397560818,-.00339970729838115,.0123137463285655,-.00267098242974697,.0353622090543082,-5.86464816610367e-006,1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,-1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,-.166073783576779,.0360231684064709,.176940031673225,-.080381948458016,-.158067370120622,-.764836159193308,.165901090401825,.814879575736018,-.370191004490028,-2.02218309922343,.43863300248611,2.1544950322241,-4.02885967082133,.873902474623302,-6.68010475273478,.0125887190510936,-.00273062693415814,-.0134124020091825,.00609310960852155,.0823390061849571,-.0178602054051423,-.087726467416282,.299004079114613,-.0648571621038782,.796298871799325,-.000824565926326729,.000178857111521775,.000878517475835277,-.00697737063211093,.00151346583452521,-.0321260785529637,5.01144631645353e-005,-1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,-.398540259080393,.0864476172000634,.4246168452719,-.192898854195334,-.379326641949509,-1.67165845871656,.362600483446047,1.78103547961334,-.809105213656782,-3.84755914106515,.834576462901105,4.09930587333892,-6.3113399865174,1.3689967095037,-7.98777931164405,.0724976248699598,-.0157255052185161,-.0772411621484506,.0350898270822692,.434860760346183,-.0943259199045093,-.4633138058217,1.40707187748205,-.305208842367266,3.23180916098999,-.0113956405123911,.00247183552086188,.0121412600482659,-.0887334889355571,.0192472366688018,-.36808995603673,.00166206328995187,-.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,-.714786270724957,.155044737648518,.76155491046056,-.345966184032581,-.680326440323354,-2.59853682213183,.563650249872674,2.76855972053751,-1.25772682793123,-4.7549644652192,1.03140232004452,5.06608294899263,-5.36731919542456,1.16422856843223,-3.25729414076559,.233201739437629,-.0505839353643848,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929028,5.89679444984459,-.0657431866024218,.0142603957822129,.0700447793236141,-.451524561772794,.0979404755348858,-1.59193561065057,.017197430813833,-.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,-1.09494796295235,.237505848402132,1.1665906747753,-.529969564290411,-1.04216054572391,-3.24473228828999,.703816913228947,3.45703591374062,-1.57049413873188,-4.05017972204568,.878527020016876,4.31518396831288,-1.63294619261043,.354203381414383,2.28560008620039,.547225517982271,-.118699029853337,-.583030617100296,.264864522602584,2.39319484717323,-.519109393293245,-2.54978216897294,4.87179403320196,-1.05674389554722,5.49730994822615,-.236321146004056,.0512605677924324,.251783696179217,-1.36209257652082,.295452354005307,-3.77635721863404,.0946962283839159,-.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,-1.51514018159912,.328650005712205,1.6142761726,-.733348258592925,-1.44209530674498,-3.36446686788906,.729788615894315,3.58460475610227,-1.62844728825442,-1.91964817870762,.416392088662432,2.04525122687055,1.95173675811295,-.423352442647985,2.75500656898115,1.04781596872157,-.227282418061715,-1.11637482313293,.507157044419473,3.55510980002442,-.771141093489061,-3.7877215002156,4.59045028887451,-.995717447724871,1.24891795372526,-.626153181688296,.135819278812355,.667122536961418,-2.84473124098146,.61705243545725,-5.48872263996485,.347192252628353,-.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,-1.94896143440088,.422750445356608,2.07648245560297,-.943323589025389,-1.85500204648398,-2.83314651369184,.614539585033937,3.01851998146924,-1.37128107917545,.637278791777431,-.138232541833329,-.678976098432196,2.63766838605638,-.57213830168982,-.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.839157192810101,4.12740556141283,-.895278125552931,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843812,-1.33269841309056,.289076447479325,1.41989719504135,-4.37531622836174,.949052583847879,-4.52786743647192,.95054294779362,-.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,-2.36915365304765,.513894602666682,2.52416795342767,-1.1467022833279,-2.25493680750505,-1.68415540504794,.365311203903165,1.79435010419287,-.815153904032968,2.32831086437971,-.505035367744937,-2.48065281165978,.381749900992996,-.0828056101286271,-2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.33045247496738,-2.69130383032704,-2.39387697960251,.519257354978938,2.55050893377232,-4.93737708601551,1.07096955656365,-.508166439915618,2.07554270972356,-.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,-2.74931534527504,.596355713420295,2.92920371774241,-1.33070566358573,-2.61677091290561,-.106757773657142,.0231568955595295,.113742960840148,-.0516722006304461,2.26282036321899,-.490829782123398,-2.41087725104767,-2.15044627554921,.466454648534285,.119501749587716,3.45007465931297,-.748357855262331,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239338,-1.88618542328239,-3.42380239520456,.742659122293479,1.33170919606025,-3.74106981000851,.8114777955963,3.98584891938639,-3.58481468643821,.777584399208991,3.59724489879787,3.76406389114815,-.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,-3.06556135691961,.66495283386875,3.26614178293108,-1.48377299342298,-2.91777071127945,1.5947619510804,-.345920813804923,-1.69910761471712,.771886268054593,.444911232504622,-.0965059740773432,-.474021882875009,-2.09440936915946,.454299648071273,2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340396,.209134120399514,1.02723329226541,-2.01722273728898,.437557047580714,3.46190525493913,-5.18623489409549,1.1249494591216,5.5255715071159,-.205483038001613,.0445714544901277,3.63947886955733,5.81833621896758,-1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,-3.29802783242322,.715377282662343,3.51381859652975,-1.5962898991603,-3.13902998310834,3.07100849759875,-.666134376965478,-3.2719453330298,1.4864094837286,-2.20598407348367,.47850138692202,2.35032214902362,.923194488569861,-.200250694746914,.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390442,-3.80884940873902,.826179910946771,4.05806335333585,2.02329330850018,-.438873818984883,.0162698697481229,-6.45780747384476,1.40076706380653,6.88034339064391,4.10479708970298,-.890374108024765,-1.32544341327556,7.79428043908953,-1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,-3.43216369764802,.744472777196595,3.65673100407672,-1.66121346489508,-3.26669916121878,4.00974905232829,-.869757178740834,-4.27210791134071,1.94077255839989,-4.32546942586304,.938240282077712,4.60848594454081,4.4144312457315,-.957537046434959,-4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,-7.27823584467216,1.57872669557351,7.75445258971521,7.46210387346587,-1.6186096255506,-7.31197439196322,9.14177937313726,-1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,-.0319379183519753,-.00692767387221854,.0340276241262535,.0154583827232306,-.0303981920101797,-.154660030302817,-.0335474040556334,.16477947382797,.074857538159081,-.43925822557731,-.0952797768044285,.467998998326769,-.961797763577864,-.208624155307574,-1.79684767953686,.000465578153648587,.00010098884891692,-.000496041045803253,-.000225346098359983,.00319092434145473,.000692145397560818,-.00339970729838115,.0123137463285655,.00267098242974697,.0353622090543082,-5.86464816610367e-006,-1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,-1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,-.166073783576779,-.0360231684064709,.176940031673225,.0803819484580161,-.158067370120624,-.764836159193308,-.165901090401825,.814879575736019,.370191004490028,-2.02218309922343,-.43863300248611,2.15449503222411,-4.02885967082133,-.873902474623302,-6.68010475273478,.0125887190510936,.00273062693415814,-.0134124020091825,-.00609310960852156,.0823390061849571,.0178602054051423,-.087726467416282,.299004079114613,.0648571621038782,.796298871799325,-.000824565926326729,-.000178857111521775,.000878517475835277,-.00697737063211093,-.00151346583452521,-.0321260785529637,5.01144631645353e-005,1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,-.398540259080393,-.0864476172000634,.4246168452719,.192898854195334,-.379326641949514,-1.67165845871656,-.362600483446047,1.78103547961334,.809105213656783,-3.84755914106515,-.834576462901105,4.09930587333892,-6.3113399865174,-1.3689967095037,-7.98777931164405,.0724976248699598,.0157255052185161,-.0772411621484506,-.0350898270822693,.434860760346183,.0943259199045093,-.463313805821701,1.40707187748205,.305208842367266,3.23180916098999,-.0113956405123911,-.00247183552086188,.0121412600482659,-.0887334889355571,-.0192472366688018,-.36808995603673,.00166206328995187,.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,-.714786270724957,-.155044737648518,.761554910460561,.345966184032582,-.680326440323361,-2.59853682213183,-.563650249872674,2.76855972053751,1.25772682793124,-4.7549644652192,-1.03140232004452,5.06608294899263,-5.36731919542456,-1.16422856843223,-3.25729414076559,.233201739437629,.0505839353643848,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929028,5.89679444984459,-.0657431866024218,-.0142603957822129,.0700447793236141,-.451524561772794,-.0979404755348858,-1.59193561065057,.017197430813833,.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,-1.09494796295235,-.237505848402132,1.1665906747753,.529969564290411,-1.04216054572392,-3.24473228828999,-.703816913228947,3.45703591374062,1.57049413873189,-4.05017972204568,-.878527020016876,4.31518396831288,-1.63294619261043,-.354203381414383,2.28560008620039,.547225517982271,.118699029853337,-.583030617100297,-.264864522602584,2.39319484717323,.519109393293245,-2.54978216897294,4.87179403320196,1.05674389554722,5.49730994822615,-.236321146004056,-.0512605677924324,.251783696179217,-1.36209257652082,-.295452354005307,-3.77635721863404,.0946962283839159,.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,-1.51514018159912,-.328650005712205,1.6142761726,.733348258592926,-1.44209530674499,-3.36446686788906,-.729788615894315,3.58460475610227,1.62844728825442,-1.91964817870762,-.416392088662432,2.04525122687055,1.95173675811295,.423352442647985,2.75500656898115,1.04781596872157,.227282418061715,-1.11637482313293,-.507157044419473,3.55510980002442,.771141093489061,-3.7877215002156,4.59045028887451,.995717447724871,1.24891795372526,-.626153181688296,-.135819278812355,.667122536961419,-2.84473124098146,-.61705243545725,-5.48872263996485,.347192252628353,.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,-1.94896143440088,-.422750445356608,2.07648245560297,.94332358902539,-1.85500204648401,-2.83314651369184,-.614539585033937,3.01851998146924,1.37128107917545,.637278791777431,.138232541833329,-.678976098432196,2.63766838605638,.57213830168982,-.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810102,4.12740556141283,.895278125552931,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843812,-1.33269841309056,-.289076447479325,1.41989719504135,-4.37531622836174,-.949052583847879,-4.52786743647192,.95054294779362,.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,-2.36915365304765,-.513894602666682,2.52416795342767,1.14670228332791,-2.25493680750508,-1.68415540504794,-.365311203903165,1.79435010419287,.81515390403297,2.32831086437971,.505035367744937,-2.48065281165978,.381749900992996,.0828056101286271,-2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.33045247496738,-2.69130383032704,-2.39387697960251,-.519257354978938,2.55050893377233,-4.93737708601551,-1.07096955656365,-.508166439915618,2.07554270972356,.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,-2.74931534527504,-.596355713420295,2.92920371774242,1.33070566358573,-2.61677091290564,-.106757773657142,-.0231568955595295,.113742960840148,.0516722006304461,2.26282036321899,.490829782123398,-2.41087725104768,-2.15044627554921,-.466454648534285,.119501749587716,3.45007465931297,.748357855262331,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239338,-1.88618542328239,-3.42380239520456,-.742659122293479,1.33170919606025,-3.74106981000851,-.8114777955963,3.98584891938639,-3.58481468643821,-.777584399208991,3.59724489879787,3.76406389114815,.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,-3.06556135691961,-.66495283386875,3.26614178293108,1.48377299342298,-2.91777071127949,1.5947619510804,.345920813804923,-1.69910761471712,-.771886268054594,.444911232504622,.0965059740773432,-.474021882875009,-2.09440936915946,-.454299648071273,2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340396,-.209134120399514,1.02723329226541,-2.01722273728898,-.437557047580714,3.46190525493913,-5.18623489409549,-1.1249494591216,5.5255715071159,-.205483038001613,-.0445714544901277,3.63947886955733,5.81833621896758,1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,-3.29802783242322,-.715377282662343,3.51381859652975,1.5962898991603,-3.13902998310838,3.07100849759875,.666134376965478,-3.2719453330298,-1.48640948372861,-2.20598407348367,-.47850138692202,2.35032214902362,.923194488569861,.200250694746914,.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873902,-.826179910946771,4.05806335333585,2.02329330850018,.438873818984883,.0162698697481229,-6.45780747384476,-1.40076706380653,6.88034339064391,4.10479708970298,.890374108024765,-1.32544341327556,7.79428043908953,1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,-3.43216369764802,-.744472777196595,3.65673100407672,1.66121346489509,-3.26669916121882,4.00974905232829,.869757178740834,-4.27210791134071,-1.9407725583999,-4.32546942586304,-.938240282077712,4.60848594454081,4.4144312457315,.957537046434959,-4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,-7.27823584467216,-1.57872669557351,7.75445258971521,7.46210387346587,1.6186096255506,-7.31197439196322,9.14177937313726,1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,-.0319379183519753,-.0203477386368136,.0212139303918347,.0361091456847507,.00500970708930326,-.154660030302817,-.0985343453312713,.102728896816887,.174859285048362,-.43925822557731,-.279852665057626,.291765835225991,-.961797763577864,-.612764091356553,-1.79684767953686,.000465578153648587,.000296621166194719,-.000309248161967448,-.000526384631617549,.00319092434145473,.00203294697585566,-.00211948838134797,.0123137463285655,.00784512281993549,.0353622090543082,-5.86464816610367e-006,-3.73638403043722e-006,3.89543979110573e-006,-5.19109369939388e-005,-3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,-.166073783576779,-.105806079951921,.110310185087204,.187763722716787,.0260499448261801,-.764836159193308,-.487279293067771,.508022497640508,.864726999689725,-2.02218309922343,-1.28833860494792,1.34318245298509,-4.02885967082133,-2.56679795782664,-6.68010475273478,.0125887190510936,.00802030872060955,-.00836172873999092,-.0142328590555998,.0823390061849571,.0524584150880732,-.0546915402140791,.299004079114613,.190496349504475,.796298871799325,-.000824565926326729,-.000525333297396549,.000547696439552503,-.00697737063211093,-.00444530268091327,-.0321260785529637,5.01144631645353e-005,3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,-.398540259080393,-.253911132799744,.264719986484414,.450591303972137,.0625140918479527,-1.67165845871656,-1.06501836010644,1.11035559079056,1.88998413076672,-3.84755914106515,-2.45129087378787,2.55564094422791,-6.3113399865174,-4.0209726602962,-7.98777931164405,.0724976248699598,.0461884430503282,-.0481546590241492,-.0819661216087149,.434860760346183,.277051027559315,-.288844933452006,1.40707187748205,.896449496423068,3.23180916098999,-.0113956405123911,-.00726019498888195,.00756925739444608,-.0887334889355571,-.056532358254944,-.36808995603673,.00166206328995187,.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,-.714786270724957,-.455392366453167,.47477816259307,.808140383434563,.112119700737039,-2.59853682213183,-1.65553520251249,1.7260104019959,2.93791673259189,-4.7549644652192,-3.02940138707022,3.15836129611742,-5.36731919542456,-3.41953432868287,-3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066828,3.30495822058058,2.10559828531328,5.89679444984459,-.0657431866024218,-.0418851712113093,.0436681993244462,-.451524561772794,-.287667582930097,-1.59193561065057,.017197430813833,.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,-1.09494796295235,-.697594462028796,.727290664744144,1.23795280192461,.171750973622277,-3.24473228828999,-2.06722817250133,2.15522890943138,3.66850806244403,-4.05017972204568,-2.58038102231816,2.69022638759909,-1.63294619261043,-1.04035466481729,2.28560008620039,.547225517982271,.348638934287124,-.363480297113412,-.618695487400892,2.39319484717323,1.52471124557473,-1.58961734284587,4.87179403320196,3.10383383936015,5.49730994822615,-.236321146004056,-.150560874408057,.156970165907756,-1.36209257652082,-.867793055767227,-3.77635721863404,.0946962283839159,.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,-1.51514018159912,-.965300119862265,1.00639240141108,1.71302390303713,.237661341149238,-3.36446686788906,-2.14351141382237,2.23475948529486,3.80388048506199,-1.91964817870762,-1.22301331540983,1.27507630270228,1.95173675811295,1.24345704101385,2.75500656898115,1.04781596872157,.667566534563461,-.695984465482171,-1.18466517033958,3.55510980002442,2.2649705675878,-2.36138908644035,4.59045028887451,2.92458893939423,1.24891795372526,-.626153181688296,-.398923973340748,.415905941872819,-2.84473124098146,-1.81238636712449,-5.48872263996485,.347192252628353,.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,-1.94896143440088,-1.24168887412682,1.29454686902578,2.2035040478582,.305709527061048,-2.83314651369184,-1.80500570449659,1.88184377891507,3.20316744314249,.637278791777431,.406012131085817,-.423295838593301,2.63766838605638,1.68046603255685,-.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141283,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843812,-1.33269841309056,-.849065950254409,.885210208841138,-4.37531622836174,-2.78752641722509,-4.52786743647192,.95054294779362,.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,-2.36915365304765,-1.50939453196029,1.57364860569271,2.67857514897071,.371619894588009,-1.68415540504794,-1.07298020016939,1.11865636189276,1.90411323119761,2.32831086437971,1.48337347495979,-1.54651984745118,.381749900992996,.243213947948102,-2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188742,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001399,-2.69130383032704,-2.39387697960251,-1.52514583383108,1.59007043171885,-4.93737708601551,-3.14561698874585,-.508166439915618,2.07554270972356,1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,-2.74931534527504,-1.75159662753591,1.82616110784379,3.10838756746077,.431251167473247,-.106757773657142,-.0680156813349527,.0709110705135131,.120700790879792,2.26282036321899,1.44164929037179,-1.50301948784969,-2.15044627554921,-1.37005543987348,.119501749587716,3.45007465931297,2.19805238043167,-2.29162223072931,-3.90066901628345,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,-3.74106981000851,-2.38344621859103,2.48490817973374,-3.58481468643821,-2.28389563582348,3.59724489879787,3.76406389114815,2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,-3.06556135691961,-1.95307786118934,2.03621928395244,3.46593664692319,.480856776362333,1.5947619510804,1.01602737631215,-1.05927908845788,-1.80304461367463,.444911232504622,.283454211851926,-.295520697836744,-2.09440936915946,-1.33435416739534,2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611565,-4.84964573066989,-.964148801340396,-.614261944257265,.640410728572749,-2.01722273728898,-1.28517834515454,3.46190525493913,-5.18623489409549,-3.30416500488362,3.44482144542743,-.205483038001613,-.130913828127032,3.63947886955733,5.81833621896758,3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,-3.29802783242322,-2.10118291403716,2.19062908534965,3.72876422817854,.517320923384106,3.07100849759875,1.95654825118803,-2.0398374063157,-3.47209520918981,-2.20598407348367,-1.40543872850124,1.46526746277611,.923194488569861,.588169835027997,.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873902,-2.426628794168,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,-6.45780747384476,-4.11428751282488,4.2894304115822,4.10479708970298,2.61517790433632,-1.32544341327556,7.79428043908953,4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,-3.43216369764802,-2.18664125535227,2.27972534004502,3.88041880521058,.538361161120983,4.00974905232829,2.55462252942968,-2.6633713691361,-4.53343925473613,-4.32546942586304,-2.75576887572707,2.87308040304257,4.4144312457315,2.81244670509951,-4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,-7.27823584467216,-4.63698476184225,4.83437858766288,7.46210387346587,4.75412760808659,-7.31197439196322,9.14177937313726,5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,-.0319379183519753,-.0324893117176766,-.00124359624768512,.031645154192253,.0381313330936757,-.154660030302817,-.157330164177211,-.00602214055814407,.153242313889452,-.43925822557731,-.446841815317497,-.0171038035473811,-.961797763577864,-.978402756631806,-1.79684767953686,.000465578153648587,.000473616144898061,1.81286469184544e-005,-.000461310355481611,.00319092434145473,.00324601417302388,.000124247970525588,.0123137463285655,.012526337462346,.0353622090543082,-5.86464816610367e-006,-5.96589860075151e-006,-2.28357226493352e-007,-5.19109369939388e-005,-5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,-.166073783576779,-.168940970519646,-.00646656841622802,.164551441038203,.198278882483946,-.764836159193308,-.778040701653472,-.0297811324981311,.757825163693694,-2.02218309922343,-2.05709515348012,-.0787396124921617,-4.02885967082133,-4.09841606758444,-6.68010475273478,.0125887190510936,.0128060574566476,.000490178589876773,-.0124733225023728,.0823390061849571,.0837605510092484,.00320610999159909,.299004079114613,.304166246215336,.796298871799325,-.000824565926326729,-.000838801675185275,-3.21068856415291e-005,-.00697737063211093,-.00709783170785957,-.0321260785529637,5.01144631645353e-005,5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,-.398540259080393,-.405420872037066,-.015518330446016,.394886974518158,.475825356015865,-1.67165845871656,-1.70051886866409,-.0650909105855661,1.6563349285852,-3.84755914106515,-3.91398546692074,-.149815966677047,-6.3113399865174,-6.42030234909073,-7.98777931164405,.0724976248699598,.0737492627952912,.00282290703160025,-.0718330635538726,.434860760346183,.442368430061905,.0169325753770755,1.40707187748205,1.43136432250754,3.23180916098999,-.0113956405123911,-.0115923809684749,-.0004437225879236,-.0887334889355571,-.0902654315677738,-.36808995603673,.00166206328995187,.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,-.714786270724957,-.727126724577601,-.0278322937135137,.708234065298678,.853397929051749,-2.59853682213183,-2.6433993582315,-.10118163017136,2.57471691018525,-4.7549644652192,-4.83705672356223,-.185148446522542,-5.36731919542456,-5.45998347614275,-3.25729414076559,.233201739437629,.237227859491313,.00908039168484083,-.231064057614915,1.22672710731592,1.24790598293055,.0477662072687194,3.30495822058058,3.36201679531997,5.89679444984459,-.0657431866024218,-.0668782122732034,-.00255990322490875,-.451524561772794,-.459319924425276,-1.59193561065057,.017197430813833,.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,-1.09494796295235,-1.11385173231848,-.042635000914322,1.08491094310714,1.30728073869584,-3.24473228828999,-3.30075109017009,-.126343140340603,3.21498892012255,-4.05017972204568,-4.12010419835037,-.157705591526285,-1.63294619261043,-1.66113825253464,2.28560008620039,.547225517982271,.55667311317238,.0213078258129713,-.542209285918405,2.39319484717323,2.43451224831938,.0931860398476283,4.87179403320196,4.95590330339708,5.49730994822615,-.236321146004056,-.240401121148633,-.00920185489986014,-1.36209257652082,-1.38560847465611,-3.77635721863404,.0946962283839159,.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,-1.51514018159912,-1.54129837497396,-.0589964137232844,1.50125140095791,1.80895681150725,-3.36446686788906,-3.42255283189436,-.131005356341752,3.33362593315312,-1.91964817870762,-1.95279001506978,-.0747471155162795,1.95173675811295,1.98543258954328,2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063127,-1.03821099255249,3.55510980002442,3.61648712494209,.138428596350638,4.59045028887451,4.66970228947898,1.24891795372526,-.626153181688296,-.636963426396325,-.0243811052053825,-2.84473124098146,-2.89384420968737,-5.48872263996485,.347192252628353,.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,-1.94896143440088,-1.98260935074574,-.0758885128327108,1.93109596018978,2.32690486658695,-2.83314651369184,-2.88205947758122,-.110316844590437,2.80717601371157,.637278791777431,.648281114842891,.0248143133691988,2.63766838605638,2.68320650944122,-.808065843843811,1.73374759666501,1.76367994615688,.067508532733652,-1.71785491622659,4.12740556141283,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843812,-1.33269841309056,-1.35570683401508,-.0518925099589806,-4.37531622836174,-4.45085403872017,-4.52786743647192,.95054294779362,.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,-2.36915365304765,-2.41005599340123,-.0922499256416731,2.34743641804054,2.82858093939836,-1.68415540504794,-1.71323156899255,-.0655775157362815,1.66871731961812,2.32831086437971,2.36850807266163,.0906595922460986,.381749900992996,.388340636274758,-2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622357,-2.53843733881177,3.58714638367868,3.64907680538234,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,-2.39387697960251,-2.4352061570422,-.0932126006787377,-4.93737708601551,-5.02261861703307,-.508166439915618,2.07554270972356,2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,-2.74931534527504,-2.7967810011421,-.107052632842481,2.72411329584901,3.28246374904245,-.106757773657142,-.108600897231078,-.00415692611322429,.105779161098541,2.26282036321899,2.30188690834251,.0881095280677891,-2.15044627554921,-2.18757273479106,.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.4184489866057,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,-3.74106981000851,-3.80565764777524,-.145669493162303,-3.58481468643821,-3.64670485341609,3.59724489879787,3.76406389114815,3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,-3.06556135691961,-3.11848685368264,-.119366596109979,3.03746038662953,3.66003632207834,1.5947619510804,1.62229477839744,.0620967201893232,-1.58014330533832,.444911232504622,.452592418790165,.0173239199033348,-2.09440936915946,-2.13056837715884,2.63186985254377,4.2894282422279,4.36348323472717,.167021432333119,-4.25010849784966,-.964148801340396,-.98079438405867,-.0375419530619799,-2.01722273728898,-2.05204915464121,3.46190525493913,-5.18623489409549,-5.27577283777777,-.201941221845797,-.205483038001613,-.20903060758203,3.63947886955733,5.81833621896758,5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,-3.29802783242322,-3.35496675520006,-.128418358139767,3.26779592010948,3.93758279561025,3.07100849759875,3.12402803860063,.119578696522845,-3.04285759691611,-2.20598407348367,-2.24406936639925,-.0858964409190686,.923194488569861,.939133013954983,.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873902,-3.87460743053656,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,-6.45780747384476,-6.56929852924652,-.25145361873174,4.10479708970298,4.17566451229335,-1.32544341327556,7.79428043908953,7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,-3.43216369764802,-3.49141841400203,-.13364133030831,3.40070220695543,4.09773034500053,4.00974905232829,4.07897551472765,.156131305216524,-3.97299303311744,-4.32546942586304,-4.40014665135235,-.168424801181427,4.4144312457315,4.49064435720703,-4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,-7.27823584467216,-7.40389121595289,-.28339939655034,7.46210387346587,7.59093364435534,-7.31197439196322,9.14177937313726,9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,-.0319379183519753,-.0425895510116904,-.0277895142706455,.00117913035023065,.0297640501007841,-.154660030302817,-.206240781926363,-.134571297792104,.00570996311627577,-.43925822557731,-.585755477154869,-.382203141612094,-.961797763577864,-1.28256746328293,-1.79684767953686,.000465578153648587,.000620853379055167,.000405104384309337,-1.71888889439261e-005,.00319092434145473,.00425513126926153,.00277645639124976,.0123137463285655,.0164205106275604,.0353622090543082,-5.86464816610367e-006,-7.82057019228999e-006,-5.10289124485106e-006,-5.19109369939388e-005,-6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,-.166073783576779,-.221461142188379,-.144502522920425,.00613135259583705,.154769899538562,-.764836159193308,-1.01991708658492,-.665491881382991,.0282373296477252,-2.02218309922343,-2.69660249304008,-1.75952250506473,-4.02885967082133,-5.37252686958436,-6.68010475273478,.0125887190510936,.0167871896471046,.0109535751204635,-.000464768572193061,.0823390061849571,.109799933312678,.0716440239592627,.299004079114613,.398725093710568,.796298871799325,-.000824565926326729,-.00109956736062704,-.00071746337155817,-.00697737063211093,-.00930439734215142,-.0321260785529637,5.01144631645353e-005,6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,-.398540259080393,-.531457639388288,-.346774016236294,.0147138868003716,.371413443660288,-1.67165845871656,-2.22917418911978,-1.45452737733324,.0617167093759977,-3.84755914106515,-5.13076069911761,-3.34780114662785,-6.3113399865174,-8.41623844737545,-7.98777931164405,.0724976248699598,.0966763475074085,.0630809359431451,-.002676572370981,.434860760346183,.579891411113903,.378376861425818,1.40707187748205,1.87634519190449,3.23180916098999,-.0113956405123911,-.0151962068293123,-.00991546509750553,-.0887334889355571,-.118327043544488,-.36808995603673,.00166206328995187,.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,-.714786270724957,-.953175031760129,-.621942953572118,.0263895153231816,.666134033494215,-2.59853682213183,-3.46517626442722,-2.26101386266496,.0959365479342401,-4.7549644652192,-6.34079527029531,-4.13734393811777,-5.36731919542456,-7.15737676442222,-3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831002,1.22672710731592,1.63585353791591,1.06738798948579,3.30495822058058,4.4071966513386,5.89679444984459,-.0657431866024218,-.0876692328212839,-.0572038290823835,-.451524561772794,-.602112766253303,-1.59193561065057,.017197430813833,.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,-1.09494796295235,-1.46012465838812,-.95272558802182,.0404248755605102,1.02041985541771,-3.24473228828999,-4.32688473530227,-2.82327524550585,.11979372855449,-4.05017972204568,-5.4009573811209,-3.52410341735765,-1.63294619261043,-2.17755097302669,2.28560008620039,.547225517982271,.729731000653746,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367101,2.08234367785444,4.87179403320196,6.49658873611033,5.49730994822615,-.236321146004056,-.315136741011867,-.205625482170818,-1.36209257652082,-1.81636481949508,-3.77635721863404,.0946962283839159,.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,-1.51514018159912,-2.02045541424857,-1.31833919902212,.0559381407795162,1.41201150863478,-3.36446686788906,-4.4865520602846,-2.92745754610214,.124214263271281,-1.91964817870762,-2.55987109498181,-1.67030580640284,1.95173675811295,2.60266155489591,2.75500656898115,1.04781596872157,1.39727364720762,.911715550981012,-.0386847883193889,3.55510980002442,4.74077641992773,3.0933379389912,4.59045028887451,6.12141388513833,1.24891795372526,-.626153181688296,-.834981872710088,-.544822382907866,-2.84473124098146,-3.79347911889772,-5.48872263996485,.347192252628353,.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,-1.94896143440088,-2.59896063091726,-1.69581157410878,.0719545824309805,1.81630452989169,-2.83314651369184,-3.77803076303363,-2.46515019062261,.104598208499711,.637278791777431,.849817990926997,.55450289173293,2.63766838605638,3.51735861632833,-.808065843843811,1.73374759666501,2.31197070864005,1.50855178059943,-.064009006140712,4.12740556141283,5.50393885468463,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843812,-1.33269841309056,-1.77716739172311,-1.15959472274786,-4.37531622836174,-5.8345303456991,-4.52786743647192,.95054294779362,1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,-2.36915365304765,-3.15929138677771,-2.06142518510908,.0874678476499865,2.20789618310877,-1.68415540504794,-2.24583899888368,-1.46540110005888,.0621780897499608,2.32831086437971,3.10482709694771,2.02588745115913,.381749900992996,.50906751987416,-2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273614,3.58714638367868,4.78349754217328,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,-2.39387697960251,-3.19226020392741,-2.08293720858961,-4.93737708601551,-6.58404442855625,-.508166439915618,2.07554270972356,2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,-2.74931534527504,-3.6662410134057,-2.39220781955879,.101503207887315,2.56218200503226,-.106757773657142,-.142362617365695,-.0928910470423957,.00394143819036155,2.26282036321899,3.01749482284837,1.96890348630228,-2.15044627554921,-2.86764279249033,.119501749587716,3.45007465931297,4.60071095140733,3.00194577372105,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,-3.74106981000851,-4.9887560540339,-3.25514367429686,-3.58481468643821,-4.78038820054912,3.59724489879787,3.76406389114815,5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,-3.06556135691961,-4.08795840577754,-2.66737675689461,.113178836410125,2.85690259486619,1.5947619510804,2.12663188500128,1.38761892782617,-.0588777326765029,.444911232504622,.593293821521023,.387121881618236,-2.09440936915946,-2.79291698671264,2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640683,-.964148801340396,-1.28570259705192,-.83891587989734,-2.01722273728898,-2.68998779997356,3.46190525493913,-5.18623489409549,-6.91589893787678,-4.51259681489188,-.205483038001613,-.274013798814207,3.63947886955733,5.81833621896758,7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,-3.29802783242322,-4.39795490297745,-2.86964825021048,.12176137061466,3.07354613898791,3.07100849759875,4.09522222779324,2.67211637191123,-.113379941906936,-2.20598407348367,-2.94170303097044,-1.91944964063812,.923194488569861,1.23108958885198,.507379758856801,4.96464287844228,6.62040364808344,4.31978730278895,-.183291880031401,-3.80884940873902,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,-6.45780747384476,-8.61155438605813,-5.61900531555837,4.10479708970298,5.47379022059324,-1.32544341327556,7.79428043908953,10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,-3.43216369764802,-4.57682649415414,-2.98636125886026,.126713592860266,3.19855198842569,4.00974905232829,5.34704266035319,3.48892427303921,-.148037726033597,-4.32546942586304,-5.76805910494959,-3.76363584617939,4.4144312457315,5.8866906319401,-4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,-7.27823584467216,-9.70560427278213,-6.33286855712842,7.46210387346587,9.95079423294563,-7.31197439196322,9.14177937313726,12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,-.0319379183519753,-.0500140463222003,-.0518576912013894,-.0415275099413171,-.0223217196472688,-.154660030302817,-.24219405407586,-.251121942525088,-.201097826062499,-.43925822557731,-.687868288629403,-.713224862474736,-.961797763577864,-1.50615319946369,-1.79684767953686,.000465578153648587,.000729084691473652,.000755960606624685,.000605371370823812,.00319092434145473,.00499691420436607,.00518111316265944,.0123137463285655,.0192830438371037,.0353622090543082,-5.86464816610367e-006,-9.18390428102533e-006,-9.52244633830417e-006,-5.19109369939388e-005,-8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,-.166073783576779,-.260067729310804,-.269654486884813,-.215938641412748,-.116070571566627,-.764836159193308,-1.1977158528687,-1.24186670359493,-.994483762912882,-2.02218309922343,-3.16669227065688,-3.28342459824587,-4.02885967082133,-6.30910167728468,-6.68010475273478,.0125887190510936,.0197136447951608,.0204403398568298,.0163685732441572,.0823390061849571,.128940991873039,.13369408456444,.299004079114613,.468233518150735,.796298871799325,-.000824565926326729,-.00129125129526923,-.00133885010037302,-.00697737063211093,-.0109264021013695,-.0321260785529637,5.01144631645353e-005,7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,-.398540259080393,-.624104888716883,-.647110981340287,-.518204861963161,-.278543636855161,-1.67165845871656,-2.61777873930361,-2.71427671758157,-2.17358603453451,-3.84755914106515,-6.02518921139496,-6.24729300320554,-6.3113399865174,-9.88341340931671,-7.98777931164405,.0724976248699598,.113529614838378,.11771460504762,.0942655625163293,.434860760346183,.680981959385669,.706084685269139,1.40707187748205,2.20344223207683,3.23180916098999,-.0113956405123911,-.0178453112143479,-.0185031347484812,-.0887334889355571,-.138954604977606,-.36808995603673,.00166206328995187,.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,-.714786270724957,-1.11933887677121,-1.16060055303996,-.929406031924798,-.499571029238743,-2.59853682213183,-4.06924896099705,-4.21925182103647,-3.37876634868698,-4.7549644652192,-7.44616510000515,-7.72064967553201,-5.36731919542456,-8.40509854032022,-3.25729414076559,.233201739437629,.365188565907959,.378650344801755,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064312,5.89679444984459,-.0657431866024218,-.102952319693436,-.106747403902083,-.451524561772794,-.707077089737688,-1.59193561065057,.017197430813833,.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,-1.09494796295235,-1.7146633520688,-1.77787020176478,-1.42371402906152,-.765269708188692,-3.24473228828999,-5.08117621439183,-5.26848128513792,-4.21898669004741,-4.05017972204568,-6.34248838569497,-6.57628863172099,-1.63294619261043,-2.55715621899877,2.28560008620039,.547225517982271,.856942588087486,.888531670338266,.711533948435089,2.39319484717323,3.74768814487268,3.88583745543935,4.87179403320196,7.62911752450546,5.49730994822615,-.236321146004056,-.370073484889954,-.383715334269618,-1.36209257652082,-2.13300568008586,-3.77635721863404,.0946962283839159,.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,-1.51514018159912,-2.37267471198348,-2.46013753301878,-1.97007200846434,-1.05894610873647,-3.36446686788906,-5.26867781509235,-5.46289467143736,-4.37467244553793,-1.91964817870762,-3.0061249410246,-3.1169383486426,1.95173675811295,3.05637492003247,2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591416,1.36243031247303,3.55510980002442,5.56722026499827,5.77244215424071,4.59045028887451,7.1885396848779,1.24891795372526,-.626153181688296,-.980541495928355,-1.01668674800428,-2.84473124098146,-4.45478376427993,-5.48872263996485,.347192252628353,.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,-1.94896143440088,-3.05202882623936,-3.16453436679059,-2.53415123836086,-1.36214797290772,-2.83314651369184,-4.43664234772504,-4.6001882618506,-3.68381930162858,.637278791777431,.997963945475626,1.03475143356997,2.63766838605638,4.13052808940267,-.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141283,6.4634222783235,6.70168045166518,1.98940127224618,3.11535668501805,-2.96832239843812,-1.33269841309056,-2.08697509460821,-2.16390630168515,-4.37531622836174,-6.85164468857583,-4.52786743647192,.95054294779362,1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,-2.36915365304765,-3.71004018615404,-3.8468016980446,-3.08050921776368,-1.6558243734555,-1.68415540504794,-2.63734867013603,-2.7345680458079,-2.18983527963514,2.32831086437971,3.64608131714141,3.78048514370594,.381749900992996,.597811573855878,-2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635524,3.33115715545234,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,-2.39387697960251,-3.74875634795768,-3.88694503628022,-4.93737708601551,-7.73181907803745,-.508166439915618,2.07554270972356,3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,-2.74931534527504,-4.30536466145163,-4.46407134676942,-3.5748172149004,-1.92152305240544,-.106757773657142,-.16718022074295,-.173342908623152,-.138812569451129,2.26282036321899,3.54352469706742,3.67414802589246,-2.15044627554921,-3.3675494587808,.119501749587716,3.45007465931297,5.40273764744738,5.60189629155853,4.48598459696892,1.77035093065653,2.7723288820058,2.87452397106015,-3.42380239520456,-5.36159588761007,1.33170919606025,-3.74106981000851,-5.85842936705581,-6.07438596565011,-3.58481468643821,-5.61373743608465,3.59724489879787,3.76406389114815,5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,-3.06556135691961,-4.80059864950596,-4.97756091846909,-3.98601838486204,-2.14255044478903,1.5947619510804,2.49736057507455,2.58941963373659,2.07360079279396,.444911232504622,.696720767591157,.72240366631615,-2.09440936915946,-3.27979695088037,2.63186985254377,4.2894282422279,6.71714607327471,6.96475715335038,5.57736018045175,-.964148801340396,-1.50983487071459,-1.56549122161564,-2.01722273728898,-3.15892446006539,3.46190525493913,-5.18623489409549,-8.12152468439221,-8.42090472917074,-.205483038001613,-.321781716436834,3.63947886955733,5.81833621896758,9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,-3.29802783242322,-5.16463580891203,-5.35501741292456,-4.28828460541245,-2.30502351007756,3.07100849759875,4.80912874954551,4.9864054592391,3.99310107128126,-2.20598407348367,-3.45452037324197,-3.58186277642042,.923194488569861,1.44570135775136,.507379758856801,4.96464287844228,7.7745166798312,8.06110512843118,6.45531290809181,-3.80884940873902,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,-6.45780747384476,-10.1127781284249,-10.4855608369157,4.10479708970298,6.42801793274624,-1.32544341327556,7.79428043908953,12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,-3.43216369764802,-5.37468949190064,-5.57281420860799,-4.46269573688388,-2.39877236199692,4.00974905232829,6.27917489029831,6.5106412373467,5.21370528571982,-4.32546942586304,-6.77358573667404,-7.02327732701018,4.4144312457315,6.91289790468793,-4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,-7.27823584467216,-11.3975501068473,-11.8176927763526,7.46210387346587,11.6854832203471,-7.31197439196322,9.14177937313726,14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,-.0319379183519753,-.054298066631689,-.0675012177634823,-.075365210691831,-.07886883956794,-.154660030302817,-.262939511058368,-.326876043550518,-.364957592022401,-.43925822557731,-.746788570738901,-.928378137184539,-.961797763577864,-1.63516477461472,-1.79684767953686,.000465578153648587,.000791535419924791,.000984005657525685,.00109864378994577,.00319092434145473,.0054249316016862,.00674406129057573,.0123137463285655,.0209347588552154,.0353622090543082,-5.86464816610367e-006,-9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,-8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,-.166073783576779,-.282344179950942,-.350999163642368,-.391891091702267,-.410109589767539,-.764836159193308,-1.30030781284792,-1.61649145672597,-1.80481513164276,-2.02218309922343,-3.43793955015111,-4.2739110377176,-4.02885967082133,-6.84951625494143,-6.68010475273478,.0125887190510936,.021402243516657,.0266064261611109,.029706114645781,.0823390061849571,.139985605707144,.174024591276672,.299004079114613,.508340689012668,.796298871799325,-.000824565926326729,-.00140185515919078,-.00174273111826309,-.00697737063211093,-.0118623177481531,-.0321260785529637,5.01144631645353e-005,8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,-.398540259080393,-.677563431168936,-.842320170000573,-.940451730878478,-.984172086870924,-1.67165845871656,-2.84200809304711,-3.53307252095549,-3.94468075935688,-3.84755914106515,-6.54128488547504,-8.13186772631248,-6.3113399865174,-10.7299904588883,-7.98777931164405,.0724976248699598,.123254146474207,.153224700235957,.1710756072316,.434860760346183,.739312383715057,.919084035475437,1.40707187748205,2.39218103581589,3.23180916098999,-.0113956405123911,-.0193738753136992,-.0240848387004708,-.0887334889355571,-.150856948291044,-.36808995603673,.00166206328995187,.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,-.714786270724957,-1.21521735159799,-1.51071034695561,-1.68671036412374,-1.76512329607374,-2.59853682213183,-4.41780594598826,-5.49204234459095,-6.13187350041211,-4.7549644652192,-8.08397637227704,-10.049681020309,-5.36731919542456,-9.12504854432544,-3.25729414076559,.233201739437629,.396469282990198,.492874996717867,.550295671800536,1.22672710731592,2.0855745668007,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984459,-.0657431866024218,-.111770838907705,-.138949104550336,-.451524561772794,-.767642727504649,-1.59193561065057,.017197430813833,.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,-1.09494796295235,-1.86153514438243,-2.31418717000874,-2.58379343998139,-2.70391057656362,-3.24473228828999,-5.51641118734948,-6.85778510905627,-7.65672734172536,-4.05017972204568,-6.88576274670089,-8.56010903534823,-1.63294619261043,-2.776192869346,2.28560008620039,.547225517982271,.930345156393643,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070526,4.87179403320196,8.28259983241363,5.49730994822615,-.236321146004056,-.401772626268325,-.499467904136599,-1.36209257652082,-2.31571114636377,-3.77635721863404,.0946962283839159,.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,-1.51514018159912,-2.57590935107797,-3.20226904625419,-3.57533818439403,-3.74155092353066,-3.36446686788906,-5.7199734894835,-7.11084574520705,-7.93926992086965,-1.91964817870762,-3.26361861024795,-4.05720211125474,1.95173675811295,3.31817281869637,2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898136,3.55510980002442,6.04408799389592,7.5137721301034,4.59045028887451,7.80428370621122,1.24891795372526,-.626153181688296,-1.06453109468611,-1.32338312764,-2.84473124098146,-4.83636425064191,-5.48872263996485,.347192252628353,.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,-1.94896143440088,-3.31345445440235,-4.11915606854152,-4.59904391748785,-4.81284738096772,-2.83314651369184,-4.81666890668704,-5.98789306701478,-6.68549157438517,.637278791777431,1.08344588754871,1.34689725343212,2.63766838605638,4.48433402043889,-.808065843843811,1.73374759666501,2.94756663562366,3.66429874441148,4.09119503456581,4.12740556141283,7.01705539432373,8.72332687917668,1.98940127224618,3.38220674597772,-2.96832239843812,-1.33269841309056,-2.26573774926528,-2.81667592802333,-4.37531622836174,-7.4385315165314,-4.52786743647192,.95054294779362,1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,-2.36915365304765,-4.02782866109789,-5.00723794478697,-5.59058866190049,-5.85048772793475,-1.68415540504794,-2.86325431258853,-3.55948498424847,-3.97417031416819,2.32831086437971,3.95839130925952,4.9209161654707,.381749900992996,.649017927110068,-2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830963,-1.52344991306454,-2.59003683420207,-2.69130383032704,-2.39387697960251,-4.06986110773887,-5.05949102727521,-4.93737708601551,-8.39409842544753,-.508166439915618,2.07554270972356,3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,-2.74931534527504,-4.67414645388234,-5.81071476784009,-6.48767173775814,-6.78927500842464,-.106757773657142,-.181500267083405,-.225633983149896,-.251920680005669,2.26282036321899,3.84705006415361,4.78250111498198,-2.15044627554921,-3.65600143048661,.119501749587716,3.45007465931297,5.86551639682903,7.29177895707793,8.14128212331116,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429709,1.33170919606025,-3.74106981000851,-6.36024100270089,-7.90680110023493,-3.58481468643821,-6.09458965574696,3.59724489879787,3.76406389114815,6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,-3.06556135691961,-5.21180037431139,-6.47910494479514,-7.2339303710034,-7.57022621762745,1.5947619510804,2.71127534923244,3.37055072311448,3.76322492874004,.444911232504622,.756399320675517,.940325842964601,-2.09440936915946,-3.56073236366527,2.63186985254377,4.2894282422279,7.29251224169723,9.06576398575723,10.1219390639593,-.964148801340396,-1.63916180436132,-2.03774139204767,-2.01722273728898,-3.42950637731078,3.46190525493913,-5.18623489409549,-8.81718478891181,-10.9611768422514,-.205483038001613,-.349344361529711,3.63947886955733,5.81833621896758,9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,-3.29802783242322,-5.60701962552938,-6.97042595115334,-7.78249101017962,-8.14428871473084,3.07100849759875,5.22106113152625,6.4906175528327,7.24678421547911,-2.20598407348367,-3.75042195543293,-4.66237685613718,.923194488569861,1.5695348492102,.507379758856801,4.96464287844228,8.44045334766828,10.4928391542811,11.7152706263275,-3.80884940873902,-6.47547397319738,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,-6.45780747384476,-10.9790017210618,-13.6486625036235,4.10479708970298,6.97861844197597,-1.32544341327556,7.79428043908953,13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,-3.43216369764802,-5.83506573884863,-7.25392389703223,-8.09901689119005,-8.47552946493044,4.00974905232829,6.81702604882233,8.4746582765388,9.46196865399931,-4.32546942586304,-7.35378632026677,-9.14193750783762,4.4144312457315,7.50503145913621,-4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,-7.27823584467216,-12.3738225688191,-15.3826488429769,7.46210387346587,12.6864189798666,-7.31197439196322,9.14177937313726,15.5420569434015,-8.9987004064276,-10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 144; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{.0018918113337895,.00428844840289335,.00641941367947712,.00814735927306806,.00936344787276585,.00999122031360907,.00999122031360907,.00936344787276585,.00814735927306806,.00641941367947712,.00428844840289335,.0018918113337895,.00395009749714678,.00895426991072093,.0134037202629815,.0170116664902499,.0195508565501554,.0208616438908705,.0208616438908705,.0195508565501554,.0170116664902499,.0134037202629815,.00895426991072092,.00395009749714678,.0050351690836213,.0114139620739561,.0170856536889082,.0216846842981607,.0249213768851943,.0265922308066324,.0265922308066324,.0249213768851943,.0216846842981607,.0170856536889082,.0114139620739561,.0050351690836213,.00487498195709245,.0110508422349653,.0165420966158804,.0209948152570779,.0241285368263833,.0257462347794558,.0257462347794558,.0241285368263833,.0209948152570779,.0165420966158804,.0110508422349653,.00487498195709245,.0035088724497807,.00795407987279395,.0119065275703486,.0151114669740713,.0173670300462221,.0185314027207393,.0185314027207393,.0173670300462221,.0151114669740713,.0119065275703486,.00795407987279395,.0035088724497807,.00127473997181245,.00288964152956687,.00432552816798997,.00548985215628924,.00630927675725883,.00673228227015115,.00673228227015115,.00630927675725883,.00548985215628924,.00432552816798997,.00288964152956687,.00127473997181245,-.00127473997181245,-.00288964152956687,-.00432552816798997,-.00548985215628924,-.00630927675725883,-.00673228227015115,-.00673228227015115,-.00630927675725883,-.00548985215628924,-.00432552816798997,-.00288964152956687,-.00127473997181245,-.0035088724497807,-.00795407987279395,-.0119065275703486,-.0151114669740713,-.0173670300462221,-.0185314027207393,-.0185314027207393,-.0173670300462221,-.0151114669740713,-.0119065275703486,-.00795407987279395,-.0035088724497807,-.00487498195709245,-.0110508422349653,-.0165420966158804,-.0209948152570779,-.0241285368263833,-.0257462347794558,-.0257462347794558,-.0241285368263833,-.0209948152570779,-.0165420966158804,-.0110508422349653,-.00487498195709245,-.0050351690836213,-.0114139620739561,-.0170856536889082,-.0216846842981607,-.0249213768851943,-.0265922308066324,-.0265922308066324,-.0249213768851943,-.0216846842981607,-.0170856536889082,-.0114139620739561,-.0050351690836213,-.00395009749714678,-.00895426991072092,-.0134037202629815,-.0170116664902499,-.0195508565501554,-.0208616438908705,-.0208616438908705,-.0195508565501554,-.0170116664902499,-.0134037202629815,-.00895426991072093,-.00395009749714678,-.0018918113337895,-.00428844840289335,-.00641941367947713,-.00814735927306806,-.00936344787276585,-.00999122031360907,-.00999122031360907,-.00936344787276585,-.00814735927306806,-.00641941367947713,-.00428844840289335,-.0018918113337895},
			{-.00235182533596431,-.00533123014211866,-.00798036224002773,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002773,-.00533123014211866,-.00235182533596431,-.00409570813081086,-.0092843470585992,-.013897815459963,-.0176387597554457,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554457,-.013897815459963,-.0092843470585992,-.00409570813081086,-.00328542799229539,-.00744756529085685,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480003,-.0173513258480003,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085686,-.00328542799229539,-.000186600113929722,-.000422994061970596,-.000633183289769704,-.000803620393549264,-.000923570130184837,-.000985490897277736,-.000985490897277736,-.000923570130184837,-.000803620393549264,-.000633183289769704,-.000422994061970597,-.000186600113929722,.00365824316969289,.00829268056399677,.0124133817294243,.0157547535948168,.0181063347139945,.019320274076138,.019320274076138,.0181063347139945,.0157547535948168,.0124133817294243,.00829268056399677,.00365824316969289,.00626131850458325,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686322,.0330678918746395,.0330678918746395,.0309901565686322,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458325,.00626131850458326,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686323,.0330678918746395,.0330678918746395,.0309901565686323,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458326,.00365824316969289,.00829268056399679,.0124133817294243,.0157547535948168,.0181063347139946,.019320274076138,.019320274076138,.0181063347139946,.0157547535948168,.0124133817294243,.00829268056399678,.00365824316969289,-.000186600113929713,-.000422994061970576,-.000633183289769673,-.000803620393549226,-.000923570130184793,-.000985490897277688,-.000985490897277688,-.000923570130184793,-.000803620393549226,-.000633183289769673,-.000422994061970576,-.000186600113929713,-.00328542799229539,-.00744756529085684,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480002,-.0173513258480002,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085684,-.00328542799229539,-.00409570813081086,-.00928434705859918,-.013897815459963,-.0176387597554456,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554456,-.013897815459963,-.00928434705859918,-.00409570813081086,-.00235182533596432,-.00533123014211866,-.00798036224002774,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002774,-.00533123014211866,-.00235182533596432},
			{.002625816508753,.00595232643558409,.00891008638917558,.0113084587769704,.0129963784252772,.0138677207253597,.0138677207253597,.0129963784252772,.0113084587769704,.00891008638917558,.00595232643558409,.002625816508753,.00327983286928533,.00743488199844395,.0111293360027156,.01412507487634,.0162334073987287,.0173217763333754,.0173217763333754,.0162334073987287,.01412507487634,.0111293360027156,.00743488199844395,.00327983286928533,-.000139403223161069,-.000316005892895126,-.000473031819685113,-.000600360153589864,-.000689970923659479,-.000736230029999612,-.000736230029999612,-.000689970923659479,-.000600360153589864,-.000473031819685113,-.000316005892895126,-.000139403223161069,-.00474831713981733,-.010763712369717,-.0161122895594359,-.0204493148918112,-.0235016141556679,-.0250772800730455,-.0250772800730455,-.0235016141556679,-.0204493148918112,-.0161122895594359,-.010763712369717,-.00474831713981733,-.0062268534463654,-.0141153291769463,-.0211293523195189,-.0268168454550162,-.0308195730805799,-.0328858716152058,-.0328858716152058,-.0308195730805799,-.0268168454550162,-.0211293523195189,-.0141153291769463,-.0062268534463654,-.00284444948144652,-.00644793411369711,-.00965196559807224,-.0122500333122624,-.0140784939652993,-.0150223867108016,-.0150223867108016,-.0140784939652993,-.0122500333122624,-.00965196559807224,-.00644793411369711,-.00284444948144652,.00284444948144652,.00644793411369712,.00965196559807225,.0122500333122624,.0140784939652993,.0150223867108016,.0150223867108016,.0140784939652993,.0122500333122624,.00965196559807225,.00644793411369712,.00284444948144652,.00622685344636543,.0141153291769464,.021129352319519,.0268168454550163,.0308195730805801,.0328858716152059,.0328858716152059,.0308195730805801,.0268168454550163,.021129352319519,.0141153291769464,.00622685344636543,.00474831713981732,.0107637123697169,.0161122895594359,.0204493148918111,.0235016141556678,.0250772800730455,.0250772800730455,.0235016141556678,.0204493148918111,.0161122895594359,.0107637123697169,.00474831713981732,.000139403223161063,.000316005892895114,.000473031819685095,.000600360153589841,.000689970923659453,.000736230029999584,.000736230029999584,.000689970923659453,.000600360153589841,.000473031819685095,.000316005892895114,.000139403223161063,-.00327983286928533,-.00743488199844395,-.0111293360027156,-.01412507487634,-.0162334073987287,-.0173217763333754,-.0173217763333754,-.0162334073987287,-.01412507487634,-.0111293360027156,-.00743488199844395,-.00327983286928533,-.00262581650875299,-.00595232643558406,-.00891008638917554,-.0113084587769704,-.0129963784252772,-.0138677207253597,-.0138677207253597,-.0129963784252772,-.0113084587769704,-.00891008638917554,-.00595232643558406,-.00262581650875299},
			{-.00274788724207344,-.00622904221162262,-.00932430450984375,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984375,-.00622904221162262,-.00274788724207344,-.00176296411460538,-.00399637136462942,-.00598220115906353,-.00759246007816475,-.00872572348723494,-.00931073969132328,-.00931073969132328,-.00872572348723494,-.00759246007816475,-.00598220115906353,-.00399637136462941,-.00176296411460538,.00351886839094271,.00797673915025254,.0119404464291104,.0151545159411986,.0174165045748593,.0185841942695623,.0185841942695623,.0174165045748593,.0151545159411986,.0119404464291104,.00797673915025254,.00351886839094271,.00572156044449334,.0129699068358449,.0194147601979758,.024640728021481,.028318644685133,.0302172684545957,.0302172684545957,.028318644685133,.024640728021481,.0194147601979758,.0129699068358449,.00572156044449334,.000863900578693703,.0019583311458107,.00293144199610321,.00372051285723621,.00427584288738108,.00456251680948201,.00456251680948201,.00427584288738108,.00372051285723621,.00293144199610321,.0019583311458107,.000863900578693703,-.00559347785892409,-.012679563105626,-.0189801422806282,-.0240891218318903,-.027684704824447,-.0295408260906547,-.0295408260906547,-.027684704824447,-.0240891218318903,-.0189801422806282,-.012679563105626,-.00559347785892409,-.00559347785892416,-.0126795631056261,-.0189801422806285,-.0240891218318906,-.0276847048244473,-.029540826090655,-.029540826090655,-.0276847048244473,-.0240891218318906,-.0189801422806285,-.0126795631056261,-.00559347785892416,.000863900578693674,.00195833114581063,.00293144199610311,.00372051285723608,.00427584288738094,.00456251680948185,.00456251680948185,.00427584288738094,.00372051285723608,.00293144199610311,.00195833114581063,.000863900578693674,.00572156044449339,.012969906835845,.0194147601979759,.0246407280214812,.0283186446851333,.030217268454596,.030217268454596,.0283186446851333,.0246407280214812,.0194147601979759,.012969906835845,.00572156044449339,.00351886839094272,.00797673915025255,.0119404464291104,.0151545159411987,.0174165045748594,.0185841942695623,.0185841942695623,.0174165045748594,.0151545159411987,.0119404464291104,.00797673915025255,.00351886839094272,-.00176296411460533,-.0039963713646293,-.00598220115906337,-.00759246007816454,-.00872572348723469,-.00931073969132302,-.00931073969132302,-.00872572348723469,-.00759246007816454,-.00598220115906337,-.00399637136462931,-.00176296411460533,-.00274788724207344,-.00622904221162261,-.00932430450984374,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984374,-.00622904221162261,-.00274788724207344},
			{.00273408778348797,.00619776093897419,.00927747931558112,.0117747446895576,.0135322629603741,.0144395336435885,.0144395336435885,.0135322629603741,.0117747446895576,.00927747931558112,.00619776093897419,.00273408778348797,-.000117317226705007,-.00026594029991732,-.00039808822185238,-.000505243613786289,-.000580657128546422,-.000619587290580092,-.000619587290580092,-.000580657128546422,-.000505243613786289,-.00039808822185238,-.00026594029991732,-.000117317226705007,-.00525141148444389,-.0119041506894287,-.0178194210583757,-.0226159634897508,-.0259916603811327,-.0277342714684962,-.0277342714684962,-.0259916603811327,-.0226159634897508,-.0178194210583757,-.0119041506894287,-.00525141148444389,-.00192519773143665,-.00436413028570642,-.00653270251221464,-.00829114262588265,-.00952869257156061,-.0101675438446148,-.0101675438446148,-.00952869257156061,-.00829114262588265,-.00653270251221464,-.00436413028570642,-.00192519773143665,.00561226510581592,.0127221509353366,.0190438923531344,.0241700318293333,.0277776915847029,.0296400471490286,.0296400471490286,.0277776915847029,.0241700318293333,.0190438923531344,.0127221509353366,.00561226510581592,.00424652299470629,.00962622140444603,.0144095699795955,.0182882657912371,.0210180032179986,.0224271554193144,.0224271554193144,.0210180032179986,.0182882657912371,.0144095699795955,.00962622140444603,.00424652299470629,-.00424652299470642,-.00962622140444635,-.014409569979596,-.0182882657912377,-.0210180032179993,-.0224271554193151,-.0224271554193151,-.0210180032179993,-.0182882657912377,-.014409569979596,-.00962622140444634,-.00424652299470642,-.00561226510581605,-.0127221509353369,-.0190438923531349,-.0241700318293339,-.0277776915847036,-.0296400471490293,-.0296400471490293,-.0277776915847035,-.0241700318293339,-.0190438923531349,-.0127221509353369,-.00561226510581605,.00192519773143709,.00436413028570742,.00653270251221612,.00829114262588453,.00952869257156278,.0101675438446171,.0101675438446171,.00952869257156278,.00829114262588453,.00653270251221612,.00436413028570742,.00192519773143709,.00525141148444307,.0119041506894268,.0178194210583729,.0226159634897473,.0259916603811286,.0277342714684918,.0277342714684918,.0259916603811286,.0226159634897473,.0178194210583729,.0119041506894268,.00525141148444307,.000117317226705136,.000265940299917613,.000398088221852818,.000505243613786846,.000580657128547061,.000619587290580775,.000619587290580775,.000580657128547061,.000505243613786846,.000398088221852818,.000265940299917613,.000117317226705136,-.00273408778348764,-.00619776093897344,-.00927747931557999,-.0117747446895562,-.0135322629603725,-.0144395336435867,-.0144395336435867,-.0135322629603725,-.0117747446895562,-.00927747931557999,-.00619776093897344,-.00273408778348764},
			{.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671589,-.00310382375628885,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.00519219719506989,.0103637745886009,.0118657346942092,.008761335787885,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00519219719506989,.0103637745886009,.0118657346942092,.00876133578788499,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671588,-.00310382375628885},
			{-.00260142552913706,-.00519252000263166,-.00594504099050645,-.00438965658114597,-.00068523266475808,.00441059027814434,.00971912899628547,.0139271476391209,.0159117625644931,.0150234628799625,.0112573018965134,.00527685077526242,-.00543177022292261,-.0108419692266112,-.012413231232108,-.00916559234130155,-.00143076414931682,.00920930185008349,.0202935178745168,.0290798506392448,.0332237218109789,.0313689542147427,.0235052191758168,.0110180516762225,-.00692384973169649,-.0138202027405243,-.0158230823854093,-.0116833336957715,-.0018237877459474,.0117390499831386,.0258680434782914,.0370679369308766,.0423501083267716,.0399858455548738,.0299619827063265,.0140446541387724,-.0067035767727117,-.0133805315937793,-.0153196923188487,-.0113116441612351,-.00176576639381999,.0113655879099196,.0250450864961337,.0358886704147581,.0410027966380503,.0387137497036944,.0290087825585011,.0135978423729273,-.00482504264831925,-.00963092357798896,-.0110266759528246,-.00814179763298104,-.00127094809920464,.00818062479896092,.0180267362591577,.0258316375293235,.0295126391755971,.0278650487239994,.0208796912105726,.00978733765555977,-.00175289208073003,-.00349882289141084,-.00400588644770123,-.00295783760558683,-.000461723350547117,.0029719431455217,.00654894174688664,.00938438810964969,.0107216609806272,.0101231070474515,.00758539313302242,.00355564663741111,.00175289208073003,.00349882289141084,.00400588644770123,.00295783760558683,.000461723350547117,-.0029719431455217,-.00654894174688664,-.00938438810964969,-.0107216609806272,-.0101231070474515,-.00758539313302242,-.00355564663741111,.00482504264831925,.00963092357798896,.0110266759528246,.00814179763298104,.00127094809920464,-.00818062479896092,-.0180267362591577,-.0258316375293235,-.0295126391755971,-.0278650487239994,-.0208796912105726,-.00978733765555977,.0067035767727117,.0133805315937793,.0153196923188487,.0113116441612351,.00176576639381999,-.0113655879099196,-.0250450864961337,-.0358886704147581,-.0410027966380503,-.0387137497036944,-.0290087825585011,-.0135978423729273,.00692384973169649,.0138202027405243,.0158230823854093,.0116833336957715,.0018237877459474,-.0117390499831386,-.0258680434782914,-.0370679369308766,-.0423501083267716,-.0399858455548738,-.0299619827063265,-.0140446541387724,.00543177022292261,.0108419692266112,.012413231232108,.00916559234130155,.00143076414931682,-.00920930185008349,-.0202935178745168,-.0290798506392448,-.0332237218109789,-.0313689542147427,-.0235052191758168,-.0110180516762225,.00260142552913706,.00519252000263166,.00594504099050645,.00438965658114597,.00068523266475808,-.00441059027814434,-.00971912899628547,-.0139271476391209,-.0159117625644931,-.0150234628799625,-.0112573018965134,-.00527685077526242},
			{.00323398975244044,.00645513634342751,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132602,-.00655997303832753,.00563199907811109,.0112416317670242,.0128708144833851,.00950345936923257,.00148350575212864,-.00954878012159535,-.0210415885190764,-.0301518078398598,-.0344484326345773,-.0325252935908768,-.0243716812928383,-.0114242050632393,.00451778467430605,.00901762784526014,.0103244989234578,.00762332921148292,.00119001431965068,-.00765968386950273,-.016878796465716,-.0241866828229713,-.02763327885035,-.0260906067053993,-.0195500756844557,-.0091640814984193,.000256593398763391,.000512167786738112,.000586393212665884,.000432976800198454,6.75883958332245e-005,-.00043504160981969,-.000958652982468414,-.00137371379952092,-.00156946765956192,-.00148184960837375,-.00111037172587836,-.000520485810578236,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163794,-.00132505164225417,.0085288693781531,.0187941242444344,.0269312757552921,.0307689755634805,.0290512481165313,.0217685278774507,.0102039790939625,-.00860992505308021,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.032167352592636,.0460946108327734,.052663080917845,.0497230798980886,.0372582357416281,.0174647665990918,-.00860992505308022,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.0321673525926361,.0460946108327734,.052663080917845,.0497230798980887,.0372582357416281,.0174647665990918,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163795,-.00132505164225417,.00852886937815311,.0187941242444344,.0269312757552922,.0307689755634806,.0290512481165313,.0217685278774507,.0102039790939625,.000256593398763379,.000512167786738088,.000586393212665856,.000432976800198433,6.75883958332213e-005,-.00043504160981967,-.000958652982468369,-.00137371379952085,-.00156946765956185,-.00148184960837368,-.00111037172587831,-.000520485810578211,.00451778467430605,.00901762784526012,.0103244989234578,.00762332921148291,.00119001431965068,-.00765968386950272,-.016878796465716,-.0241866828229712,-.0276332788503499,-.0260906067053993,-.0195500756844557,-.00916408149841928,.00563199907811108,.0112416317670242,.0128708144833851,.00950345936923255,.00148350575212863,-.00954878012159534,-.0210415885190763,-.0301518078398597,-.0344484326345772,-.0325252935908767,-.0243716812928383,-.0114242050632393,.00323398975244044,.00645513634342752,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132603,-.00655997303832753},
			{-.00361075440083214,-.00720716939267161,-.00825166151372371,-.00609280244273312,-.000951096555391269,.00612185821915875,.0134900605081838,.0193307511845923,.0220853782133051,.0208524265261428,.0156250301648324,.0073242197194988,-.00451009083357084,-.00900227071838211,-.0103069161796764,-.00761034659166953,-.00118798771118922,.00764663933727674,.016850051675698,.0241454926160516,.0275862190496304,.0260461740935889,.0195167816743332,.00914847495923774,.000191693059983865,.000382624848254669,.000438076387901722,.000323463690539438,5.049321798266e-005,-.000325006246491721,-.000716180246871757,-.00102625945578221,-.00117250116198259,-.00110704440267768,-.000829524667738866,-.000388838988800048,.00652940026539944,.013032870287285,.0149216465305082,.0110177379767064,.00171988715149418,-.0110702801253114,-.0243943494584096,-.0349561886252798,-.0399374364354906,-.0377078649444119,-.0282550583007305,-.0132445347624081,.00856253222944282,.0170910600271185,.0195679655313662,.0144484535618117,.00225542753807154,-.0145173563434682,-.031990289301271,-.0458408857714791,-.0523732001011417,-.0494493821432545,-.0370531499848233,-.0173686328234554,.00391139611196793,.00780725887484712,.00893871838932497,.00660010655391095,.00103028756760808,-.00663158159716375,-.0146132814266631,-.0209402846694332,-.0239242698021651,-.0225886590405194,-.0169260147469522,-.00793405514577347,-.00391139611196793,-.00780725887484712,-.00893871838932498,-.00660010655391096,-.00103028756760808,.00663158159716376,.0146132814266631,.0209402846694332,.0239242698021651,.0225886590405194,.0169260147469522,.00793405514577348,-.00856253222944286,-.0170910600271186,-.0195679655313663,-.0144484535618117,-.00225542753807155,.0145173563434683,.0319902893012711,.0458408857714793,.0523732001011419,.0494493821432547,.0370531499848235,.0173686328234555,-.00652940026539942,-.013032870287285,-.0149216465305081,-.0110177379767064,-.00171988715149417,.0110702801253114,.0243943494584096,.0349561886252797,.0399374364354905,.0377078649444118,.0282550583007304,.0132445347624081,-.000191693059983858,-.000382624848254655,-.000438076387901706,-.000323463690539425,-5.04932179826581e-005,.000325006246491708,.00071618024687173,.00102625945578217,.00117250116198255,.00110704440267764,.000829524667738834,.000388838988800033,.00451009083357083,.00900227071838211,.0103069161796764,.00761034659166953,.00118798771118921,-.00764663933727674,-.016850051675698,-.0241454926160516,-.0275862190496304,-.0260461740935889,-.0195167816743332,-.00914847495923773,.00361075440083212,.00720716939267157,.00825166151372367,.00609280244273308,.000951096555391264,-.00612185821915872,-.0134900605081837,-.0193307511845922,-.022085378213305,-.0208524265261427,-.0156250301648324,-.00732421971949876},
			{.00377861359217487,.00754222115521882,.00863527033203834,.00637604876123444,.00099531177496977,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.023112098204709,-.0218218282260771,-.016351417129166,-.0076647129967619,.00242424800551054,.00483886858146159,.00554013697585819,.00409068646883364,.000638562934915723,-.00411019441654455,-.00905717992717652,-.0129785994288253,-.0148280464803178,-.014000246985618,-.010490591164109,-.00491745576570479,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.00816499167123309,-.00127456872700186,.00820392944654196,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972578,.0209391724674719,.00981521944459481,-.00786770495273284,-.0157041648451191,-.0179800758935133,-.0132759990387789,-.00207240544459477,.0133393105384579,.0293943603166261,.0421210168157691,.0481232507843108,.0454366930684684,.0340463829902474,.0159592133291381,-.00118794775089702,-.00237117779829748,-.00271481846955098,-.00200454812347154,-.000312913283053141,.00201410755078885,.00443826559803358,.00635986828336871,.00726614786377036,.00686050349665649,.00514067626358997,.0024096876655053,.00769157887618114,.015352612142481,.0175775747521701,.0129788031427645,.0020260126728559,-.0130406973541611,-.0287363395359262,-.0411780976955557,-.0470459659340941,-.0444195493741242,-.0332842222466921,-.0156019511230183,.00769157887618123,.0153526121424812,.0175775747521703,.0129788031427646,.00202601267285592,-.0130406973541613,-.0287363395359265,-.0411780976955562,-.0470459659340946,-.0444195493741247,-.0332842222466924,-.0156019511230185,-.00118794775089698,-.0023711777982974,-.00271481846955089,-.00200454812347147,-.00031291328305313,.00201410755078878,.00443826559803343,.00635986828336849,.00726614786377011,.00686050349665626,.00514067626358979,.00240968766550522,-.00786770495273291,-.0157041648451192,-.0179800758935135,-.013275999038779,-.00207240544459478,.0133393105384581,.0293943603166264,.0421210168157695,.0481232507843112,.0454366930684688,.0340463829902477,.0159592133291382,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.0081649916712331,-.00127456872700187,.00820392944654197,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972579,.0209391724674719,.00981521944459482,.00242424800551048,.00483886858146146,.00554013697585804,.00409068646883352,.000638562934915706,-.00411019441654444,-.00905717992717627,-.0129785994288249,-.0148280464803174,-.0140002469856176,-.0104905911641087,-.00491745576570466,.00377861359217486,.00754222115521882,.00863527033203834,.00637604876123444,.000995311774969769,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.0231120982047089,-.0218218282260771,-.016351417129166,-.0076647129967619},
			{-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291795,-.0362874566153928,-.0189020194535391,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133056,.0295240165836947,.0169054630708873,-.00973006417443097,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133057,.0295240165836947,.0169054630708873,-.00973006417443096,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291794,-.0362874566153928,-.0189020194535391,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299},
			{.00303781657536634,.00474973406301401,.00235687098578853,-.00317458288582604,-.00858873504163554,-.010141337275906,-.00580693356275668,.00334222351587521,.013401456434816,.0194775827034515,.0181809574007362,.00947040224160834,.0063429536736534,.00991743325392784,.00492114092696054,-.0066285213996303,-.0179332580268801,-.0211750877428169,-.012124863256468,.00697855463035482,.0279822086740939,.0406691456503908,.0379617951494089,.0197741770113756,.00808533061765968,.0126417014630968,.00627295317887047,-.0084493423379166,-.0228594323180056,-.0269917760821381,-.0154555012012238,.00889552790747716,.0356687783299041,.0518407520280328,.048389706186527,.0252060423353226,.00782810670776326,.0122395227481066,.0060733876175387,-.00818053788513823,-.0221321902500056,-.0261330690598512,-.0149638052352352,.00861252866141587,.0345340241612611,.0501915082853594,.0468502528218308,.0244041460284054,.00563444710207634,.00880965807613161,.00437145053573306,-.00588811697401006,-.0159301169327542,-.0188098349613213,-.0107705186185175,.00619905156756988,.0248566019374114,.0361264107612158,.0337214707333665,.0175654056595871,.00204694101677262,.00320046494945897,.00158810638245006,-.00213909686741309,-.00578725989629442,-.00683343405369158,-.00391282692565031,.00225205644650943,.00903016695720288,.0131243812633644,.012250689435808,.00638134472992193,-.00204694101677262,-.00320046494945897,-.00158810638245006,.00213909686741309,.00578725989629442,.00683343405369158,.00391282692565031,-.00225205644650943,-.00903016695720288,-.0131243812633644,-.012250689435808,-.00638134472992193,-.00563444710207634,-.00880965807613161,-.00437145053573306,.00588811697401006,.0159301169327542,.0188098349613213,.0107705186185175,-.00619905156756988,-.0248566019374114,-.0361264107612158,-.0337214707333665,-.0175654056595871,-.00782810670776326,-.0122395227481066,-.0060733876175387,.00818053788513823,.0221321902500056,.0261330690598512,.0149638052352352,-.00861252866141587,-.0345340241612611,-.0501915082853594,-.0468502528218308,-.0244041460284054,-.00808533061765968,-.0126417014630968,-.00627295317887047,.0084493423379166,.0228594323180056,.0269917760821381,.0154555012012238,-.00889552790747716,-.0356687783299041,-.0518407520280328,-.048389706186527,-.0252060423353226,-.0063429536736534,-.00991743325392784,-.00492114092696054,.0066285213996303,.0179332580268801,.0211750877428169,.012124863256468,-.00697855463035482,-.0279822086740939,-.0406691456503908,-.0379617951494089,-.0197741770113756,-.00303781657536634,-.00474973406301401,-.00235687098578853,.00317458288582604,.00858873504163554,.010141337275906,.00580693356275668,-.00334222351587521,-.013401456434816,-.0194775827034515,-.0181809574007362,-.00947040224160835},
			{-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354764,-.0226018501305136,-.0117732310473289,-.0065767710679063,-.01028301505074,-.0051025466895586,.00687286554608826,.0185943234985903,.0219556552975931,.0125718165337835,-.00723580189137435,-.0290137040080097,-.042168313727554,-.0393611634058515,-.0205031034359056,-.00527564637794148,-.00824866040593754,-.00409307723852653,.00551316259147853,.0149156895386415,.0176120275664321,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.031574092667901,-.016446843331074,-.000299637130226379,-.000468493290714356,-.000232471592992847,.00031312716945722,.000847155796376566,.00100029778711831,.000572770891524226,-.00032966251846565,-.00132186188578481,-.00192118478526154,-.00179329125546946,-.000934119647899669,.00587430233536655,.0091846802486317,.00455754071797623,-.00613877079058787,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431711,.0259147334538654,.0376642916790926,.0351569747115048,.0183131550653389,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939193,.0644648579968841,.0601734237215266,.0313441429038559,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939194,.0644648579968841,.0601734237215266,.0313441429038559,.00587430233536656,.00918468024863171,.00455754071797623,-.00613877079058788,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431712,.0259147334538654,.0376642916790926,.0351569747115049,.0183131550653389,-.000299637130226364,-.000468493290714334,-.000232471592992836,.000313127169457205,.000847155796376526,.00100029778711826,.000572770891524199,-.000329662518465634,-.00132186188578475,-.00192118478526145,-.00179329125546937,-.000934119647899624,-.00527564637794147,-.00824866040593752,-.00409307723852653,.00551316259147852,.0149156895386415,.017612027566432,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.0315740926679009,-.016446843331074,-.00657677106790629,-.01028301505074,-.00510254668955859,.00687286554608825,.0185943234985903,.021955655297593,.0125718165337835,-.00723580189137434,-.0290137040080097,-.0421683137275539,-.0393611634058515,-.0205031034359056,-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354765,-.0226018501305136,-.0117732310473289},
			{.00421646110756717,.00659258663290769,.00327131497264031,-.00440629147242691,-.0119210842286909,-.0140760816664588,-.00805996969011775,.00463897510526447,.0186010966891058,.0270347033472955,.0252349995059512,.0131448300890066,.00526666188843086,.0082346128376237,.00408610667854559,-.0055037736089871,-.014890287939178,-.0175820340707359,-.0100674793638363,.00579441213021712,.0232341018964256,.0337682805916803,.0315203216066774,.0164188341582059,-.000223849711801565,-.000349996971430556,-.000173672398524789,.00023392770644707,.000632884117955103,.000747291803232149,.000427899569386332,-.000246280758644836,-.000987522480778167,-.00143525824110462,-.00133971290677435,-.000697852144734339,-.00762471196281051,-.0119215078435287,-.00591558507707979,.00796798605379006,.0215571378958162,.0254540633801011,.0145750063260015,-.00838875257661241,-.033636739632742,-.0488874012505932,-.0456329603679329,-.0237700623038804,-.00999890330935153,-.0156336403014995,-.00775758658064257,.010449066471062,.0282696236261719,.0333799781302062,.0191133881119304,-.0110008517447921,-.0441105852745744,-.0641100149794998,-.0598422026515798,-.0311716109136999,-.00456753568689175,-.00714150420140017,-.00354369398872959,.00477317186942322,.012913667706615,.0152480963782088,.00873106575768223,-.00502522940526383,-.0201498770593039,-.0292856898648238,-.0273361375479713,-.0142393061380111,.00456753568689175,.00714150420140018,.0035436939887296,-.00477317186942323,-.012913667706615,-.0152480963782088,-.00873106575768224,.00502522940526384,.0201498770593039,.0292856898648239,.0273361375479713,.0142393061380111,.00999890330935158,.0156336403014996,.0077575865806426,-.010449066471062,-.028269623626172,-.0333799781302063,-.0191133881119305,.0110008517447921,.0441105852745746,.0641100149795001,.0598422026515801,.0311716109137,.0076247119628105,.0119215078435287,.00591558507707978,-.00796798605379004,-.0215571378958161,-.025454063380101,-.0145750063260015,.0083887525766124,.033636739632742,.0488874012505931,.0456329603679328,.0237700623038804,.000223849711801557,.000349996971430543,.000173672398524782,-.000233927706447061,-.000632884117955079,-.000747291803232121,-.000427899569386316,.000246280758644826,.000987522480778129,.00143525824110457,.0013397129067743,.000697852144734312,-.00526666188843086,-.00823461283762369,-.00408610667854559,.0055037736089871,.014890287939178,.0175820340707359,.0100674793638363,-.00579441213021711,-.0232341018964256,-.0337682805916803,-.0315203216066774,-.0164188341582059,-.00421646110756715,-.00659258663290766,-.00327131497264029,.00440629147242689,.0119210842286909,.0140760816664587,.00805996969011771,-.00463897510526445,-.0186010966891057,-.0270347033472953,-.025234999505951,-.0131448300890066},
			{.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854622,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854621,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439},
			{-.00327473465962885,-.00326307824304169,.00235463328641995,.00792987603467538,.00656636699674288,-.002385380868931,-.0114262268213642,-.0111099792585027,.000926856258488525,.016554558282902,.0231216395735111,.0141071699093245,-.00683763806148434,-.00681329949179824,.00491646861608254,.0165575620113583,.0137105584327533,-.00498066949411644,-.0238579340109859,-.0231976098634047,.00193527362955397,.0345658777187952,.0482779275954164,.0294557367047126,-.00871590227761381,-.00868487802727276,.0062669974080828,.0211058396406552,.0174767787351192,-.00634883393906744,-.030411586503362,-.0295698746885344,.00246688340095815,.0440609476003131,.0615396274713053,.0375470769767045,-.00843861757917165,-.00840858032359331,.00606762132158154,.0204343857631653,.0169207785452416,-.00614685433349348,-.0294440828159479,-.0286291489294765,.00238840282624089,.0426592078630735,.0595818270621095,.0363525672650008,-.00607387534936655,-.00605225539271983,.00436730012098293,.0147080858685528,.012179091982045,-.00442432977464276,-.0211930078739329,-.0206064417927042,.00171910397817624,.0307048765549653,.0428852933869165,.0261655372015035,-.00220658111756891,-.00219872679304704,.00158659857626521,.00534330763908673,.00442454822513244,-.00160731690676576,-.00769921809541031,-.00748612421965592,.000624534116884189,.0111547894756412,.0155798190061378,.00950569068328115,.00220658111756891,.00219872679304704,-.00158659857626521,-.00534330763908673,-.00442454822513244,.00160731690676576,.00769921809541031,.00748612421965592,-.000624534116884189,-.0111547894756412,-.0155798190061378,-.00950569068328115,.00607387534936655,.00605225539271983,-.00436730012098293,-.0147080858685528,-.012179091982045,.00442432977464276,.0211930078739329,.0206064417927042,-.00171910397817624,-.0307048765549653,-.0428852933869165,-.0261655372015035,.00843861757917165,.00840858032359331,-.00606762132158154,-.0204343857631653,-.0169207785452416,.00614685433349347,.0294440828159479,.0286291489294765,-.00238840282624089,-.0426592078630735,-.0595818270621095,-.0363525672650008,.00871590227761381,.00868487802727276,-.0062669974080828,-.0211058396406552,-.0174767787351192,.00634883393906744,.030411586503362,.0295698746885344,-.00246688340095815,-.0440609476003131,-.0615396274713053,-.0375470769767045,.00683763806148434,.00681329949179824,-.00491646861608254,-.0165575620113583,-.0137105584327533,.00498066949411644,.0238579340109859,.0231976098634047,-.00193527362955397,-.0345658777187952,-.0482779275954164,-.0294557367047126,.00327473465962885,.00326307824304169,-.00235463328641995,-.00792987603467538,-.00656636699674288,.002385380868931,.0114262268213642,.0111099792585027,-.000926856258488525,-.016554558282902,-.0231216395735111,-.0141071699093245},
			{.00407102114440158,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589114,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092,.00708969078054082,.0070644550293116,-.00509770214609633,-.0171679158335966,-.0142159644664865,.00516426963168224,.0247373981014314,.024052732727339,-.00200661273467494,-.0358400638320779,-.0500575747208443,-.0305415500311089,.00568709190283437,.00566684870169624,-.00408919112208648,-.0137714772263088,-.0114035292808877,.004142589136187,.0198434403974258,.0192942267538859,-.00160963170168385,-.0287495947462082,-.0401543645107642,-.0244993198234636,.00032300570868889,.000321855970010235,-.000232250876004936,-.000782168784534644,-.000647678131435131,.000235283684983902,.00112703375255846,.00109584045637403,-9.1421105446067e-005,-.00163287025850404,-.00228061884480666,-.00139147042059047,-.00633243679598169,-.00630989648997895,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615883,-.0214836463932346,.00179228526457281,.0320119658255772,.0447108961296189,.0272793893570287,-.010838372915793,-.0107997937321446,.00779311800518549,.0262454710298765,.0217326719902005,-.00789488312515758,-.0378173257324053,-.0367706427561445,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548538,-.010838372915793,-.0107997937321446,.0077931180051855,.0262454710298765,.0217326719902006,-.00789488312515758,-.0378173257324053,-.0367706427561446,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548539,-.0063324367959817,-.00630989648997896,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615884,-.0214836463932346,.00179228526457282,.0320119658255772,.044710896129619,.0272793893570288,.000323005708688875,.00032185597001022,-.000232250876004925,-.000782168784534606,-.0006476781314351,.00023528368498389,.00112703375255841,.00109584045637398,-9.14211054460626e-005,-.00163287025850396,-.00228061884480655,-.00139147042059041,.00568709190283437,.00566684870169623,-.00408919112208647,-.0137714772263087,-.0114035292808876,.00414258913618699,.0198434403974258,.0192942267538859,-.00160963170168384,-.0287495947462082,-.0401543645107641,-.0244993198234635,.00708969078054081,.00706445502931158,-.00509770214609633,-.0171679158335965,-.0142159644664865,.00516426963168223,.0247373981014314,.0240527327273389,-.00200661273467494,-.0358400638320778,-.0500575747208442,-.0305415500311089,.00407102114440159,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589115,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092},
			{-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419407,-.00398827433495215,-.0145295292180157,-.0111084414324411,-.004451735343494,-.00147019434909273,.00789381658375618,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237627,-.0329362316722039,-.0251811462743288,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083196,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083195,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.004451735343494,-.00147019434909273,.00789381658375617,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237626,-.0329362316722039,-.0251811462743288,-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419408,-.00398827433495215,-.0145295292180157,-.0111084414324411},
			{.00333875638300519,.00110263085932663,-.0059202824228546,-.00595784387271232,.00390263111617153,.0106248423145106,.00253846386580488,-.0119075736105008,-.0122215351790305,.00678051660291391,.0247018398989999,.01888560445932,.00697131526529464,.00230229057164388,-.0123615354026379,-.0124399636192375,.00814868434650274,.0221846450958567,.00530030642192468,-.0248629849443481,-.0255185359412845,.0141577022933105,.0515773820591734,.0394330965061518,.00888629992584825,.00293471802056214,-.0157571859758933,-.0158571580226064,.0103870862740315,.0282786536784102,.00675627349671033,-.0316927197895115,-.0325283472936162,.0180467507280219,.0657454252068613,.0502651665035225,.00860359425560065,.00284135391717673,-.015255892314898,-.0153526838855177,.0100566351063325,.0273790063776703,.00654133175006641,-.0306844585711385,-.0314935017110425,.0174726176464323,.0636538230041544,.0486660478934077,.00619262083804183,.00204512520616675,-.0109807545364729,-.0110504223380172,.00723847804416364,.0197066249734953,.00470826332583461,-.0220857948325456,-.0226681209230377,.0125762899688114,.045816199491271,.0350284280425796,.00224972022366444,.000742974526710104,-.00398920363411266,-.00401451328347947,.00262967019464013,.00715922933157688,.00171046726410221,-.00802355910202843,-.00823511262949449,.0045688464741287,.0166445893043808,.0127254945897002,-.00224972022366444,-.000742974526710104,.00398920363411266,.00401451328347947,-.00262967019464013,-.00715922933157688,-.00171046726410221,.00802355910202843,.00823511262949449,-.0045688464741287,-.0166445893043808,-.0127254945897002,-.00619262083804183,-.00204512520616675,.0109807545364729,.0110504223380172,-.00723847804416364,-.0197066249734953,-.00470826332583461,.0220857948325456,.0226681209230377,-.0125762899688114,-.045816199491271,-.0350284280425796,-.00860359425560065,-.00284135391717673,.015255892314898,.0153526838855177,-.0100566351063325,-.0273790063776703,-.00654133175006641,.0306844585711385,.0314935017110425,-.0174726176464323,-.0636538230041544,-.0486660478934077,-.00888629992584825,-.00293471802056214,.0157571859758933,.0158571580226064,-.0103870862740315,-.0282786536784102,-.00675627349671033,.0316927197895115,.0325283472936162,-.0180467507280219,-.0657454252068613,-.0502651665035225,-.00697131526529464,-.00230229057164388,.0123615354026379,.0124399636192375,-.00814868434650274,-.0221846450958567,-.00530030642192469,.0248629849443481,.0255185359412845,-.0141577022933105,-.0515773820591734,-.0394330965061518,-.00333875638300519,-.00110263085932663,.0059202824228546,.00595784387271232,-.00390263111617154,-.0106248423145106,-.00253846386580488,.0119075736105008,.0122215351790305,-.00678051660291391,-.0247018398989999,-.01888560445932},
			{.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508331,-.0313789443876932,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508332,-.0313789443876932,.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847},
			{-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005},
			{.00459124870047564,.0100009097368206,.0139152296561375,.0158389513870659,.0156860656096203,.0137691389655327,.0107042527083733,.00724960390977444,.00411792158131722,.00180905830503188,.000503600637591741,4.27237565125139e-005,.00958651619965776,.0208818753803107,.0290549659195549,.0330716922483351,.0327524671142169,.0287499289103516,.022350453806482,.0151371554572844,.00859821031784408,.00377730937239729,.00105151691519805,8.92071004289135e-005,.0122198831859258,.0266180197827826,.037036216505918,.0421563170205638,.0417494023742321,.0366473873898747,.0284900091941002,.019295254668408,.0109600947313715,.00481491694443133,.00134036323562505,.000113711834820181,.0118311240516121,.0257712033140444,.0358579591324206,.040815170541386,.0404212013367775,.0354815000912993,.0275836379021353,.0186814021145725,.0106114140709319,.00466173684322829,.00129772138355337,.000110094245863453,.00851570438619621,.018549374357127,.0258095324275119,.0293775912826531,.029094023527884,.0255385679871328,.0198539128865321,.0134463384234132,.00763779206044802,.00335538472168774,.000934062700194595,7.92426863503829e-005,.00309367437106509,.00673881119463904,.00937635753650586,.0106725993661888,.010569581899212,.00927791873371822,.0072127376288433,.00488492679861124,.00277473717702488,.00121897934073137,.000339336090769426,2.87881139056347e-005,-.00309367437106509,-.00673881119463904,-.00937635753650586,-.0106725993661888,-.010569581899212,-.00927791873371822,-.0072127376288433,-.00488492679861124,-.00277473717702488,-.00121897934073137,-.000339336090769426,-2.87881139056347e-005,-.00851570438619621,-.018549374357127,-.0258095324275119,-.0293775912826531,-.029094023527884,-.0255385679871328,-.0198539128865321,-.0134463384234132,-.00763779206044802,-.00335538472168774,-.000934062700194595,-7.92426863503829e-005,-.0118311240516121,-.0257712033140444,-.0358579591324206,-.040815170541386,-.0404212013367775,-.0354815000912993,-.0275836379021353,-.0186814021145725,-.0106114140709319,-.00466173684322829,-.00129772138355337,-.000110094245863453,-.0122198831859258,-.0266180197827826,-.037036216505918,-.0421563170205638,-.0417494023742321,-.0366473873898747,-.0284900091941002,-.019295254668408,-.0109600947313715,-.00481491694443133,-.00134036323562505,-.000113711834820181,-.00958651619965776,-.0208818753803107,-.0290549659195549,-.0330716922483351,-.0327524671142169,-.0287499289103516,-.022350453806482,-.0151371554572845,-.00859821031784408,-.00377730937239729,-.00105151691519805,-8.92071004289135e-005,-.00459124870047563,-.0100009097368206,-.0139152296561375,-.0158389513870659,-.0156860656096203,-.0137691389655327,-.0107042527083733,-.00724960390977444,-.00411792158131722,-.00180905830503188,-.000503600637591741,-4.2723756512514e-005},
			{-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042617,-.0195002989275631,-.0171172512269981,-.0133071053509708,-.00901242203528271,-.00511923791436272,-.00224895003013865,-.000626056476974539,-5.31124913051166e-005,-.00993989955437502,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.033959806304287,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562087,-.00391655059726168,-.00109027850150277,-9.24955009028306e-005,-.00797342564302374,-.0173681530562416,-.0241659853793842,-.0275068307964137,-.0272413205925748,-.0239122759128171,-.0185896187731037,-.0125900776644864,-.00715141863890626,-.00314171433962441,-.000874581731371057,-7.41965242936831e-005,-.000452860977896186,-.000986446620240561,-.00137253826155738,-.00156228587949368,-.00154720588553185,-.00135812850566637,-.00105582133868279,-.000715069674966715,-.000406174031483403,-.000178437711946007,-4.96729957550138e-005,-4.2140871505504e-006,.00887821311745346,.0193390107581648,.0269082296615077,.0306281787690543,.0303325396949503,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930627,.00349822155622026,.000973825222354376,8.26160028005004e-005,.0151956328492542,.0330999609111817,.0460551659608014,.0524220981473091,.0519160928547411,.0455716503331479,.0354278116245859,.0239939777830022,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.0151956328492542,.0330999609111817,.0460551659608015,.0524220981473091,.0519160928547411,.045571650333148,.0354278116245859,.0239939777830023,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.00887821311745347,.0193390107581649,.0269082296615078,.0306281787690543,.0303325396949504,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930629,.00349822155622027,.000973825222354377,8.26160028005006e-005,-.000452860977896164,-.000986446620240513,-.00137253826155732,-.00156228587949361,-.00154720588553178,-.0013581285056663,-.00105582133868274,-.000715069674966681,-.000406174031483384,-.000178437711945999,-4.96729957550114e-005,-4.2140871505502e-006,-.00797342564302373,-.0173681530562415,-.0241659853793842,-.0275068307964136,-.0272413205925747,-.023912275912817,-.0185896187731037,-.0125900776644864,-.00715141863890625,-.0031417143396244,-.000874581731371055,-7.41965242936829e-005,-.009939899554375,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.0339598063042869,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562085,-.00391655059726168,-.00109027850150277,-9.24955009028305e-005,-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042618,-.0195002989275631,-.0171172512269982,-.0133071053509708,-.00901242203528272,-.00511923791436273,-.00224895003013865,-.00062605647697454,-5.31124913051166e-005},
			{.00637261042846754,.0138811695555554,.019314209141412,.0219843169843424,.0217721129493893,.0191114366236973,.0148574030483831,.0100623826962154,.00571563679614984,.00251095607588359,.000698992993903286,5.93001760646835e-005,.00795984680451423,.017338574885438,.0241248304194484,.027459986337079,.0271949283002686,.0238715530231142,.0185579604316863,.0125686366127877,.00713923968808145,.00313636396281389,.000873092308301808,7.40701667321458e-005,-.000338318549950809,-.000736944021350783,-.00102538124737378,-.00116713587426844,-.00115587007319604,-.00101461616061174,-.000788771746176309,-.000534206627103106,-.000303440163905314,-.000133305342938937,-3.71091719441577e-005,-3.14821528841767e-006,-.0115237204206783,-.0251015998648952,-.0349262753137827,-.0397546853696583,-.0393709525182014,-.0345595976658438,-.0268669426491322,-.0181960102941585,-.0103357017040835,-.00454061269426197,-.0012640031786301,-.000107233708624919,-.0151119893856147,-.0329177641310666,-.0458016580195613,-.0521335438038479,-.0516303237875308,-.0453208038751262,-.0352328013276903,-.0238619043492618,-.013554044071107,-.00595447375803616,-.00165758990339312,-.000140624260860408,-.00690321215069344,-.0150369553289623,-.0209223652885514,-.0238147939270004,-.0235849211787824,-.0207027093526006,-.0160944728071212,-.010900205382576,-.00619153701972338,-.00272002544129556,-.000757193144462012,-6.42376778783298e-005,.00690321215069345,.0150369553289623,.0209223652885514,.0238147939270005,.0235849211787825,.0207027093526006,.0160944728071212,.010900205382576,.00619153701972339,.00272002544129557,.000757193144462013,6.42376778783299e-005,.0151119893856147,.0329177641310668,.0458016580195615,.0521335438038482,.051630323787531,.0453208038751264,.0352328013276905,.0238619043492619,.0135540440711071,.00595447375803619,.00165758990339312,.000140624260860409,.0115237204206782,.0251015998648951,.0349262753137827,.0397546853696582,.0393709525182013,.0345595976658438,.0268669426491322,.0181960102941585,.0103357017040834,.00454061269426196,.0012640031786301,.000107233708624919,.000338318549950797,.000736944021350755,.00102538124737374,.0011671358742684,.00115587007319599,.0010146161606117,.000788771746176279,.000534206627103086,.000303440163905302,.000133305342938932,3.71091719441563e-005,3.14821528841755e-006,-.00795984680451422,-.017338574885438,-.0241248304194484,-.027459986337079,-.0271949283002686,-.0238715530231142,-.0185579604316863,-.0125686366127877,-.00713923968808145,-.00313636396281389,-.000873092308301808,-7.40701667321458e-005,-.00637261042846751,-.0138811695555553,-.0193142091414119,-.0219843169843423,-.0217721129493892,-.0191114366236972,-.014857403048383,-.0100623826962154,-.00571563679614981,-.00251095607588358,-.000698992993903282,-5.93001760646832e-005},
			{-.00666886465162736,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907023,-6.20569627520987e-005,-.00427854858305904,-.00931976919031684,-.0129674931619453,-.0147601943251846,-.0146177212706554,-.0128313523954431,-.00997520897819624,-.00675584891158071,-.00383745876041692,-.0016858472177435,-.000469301476561824,-3.98139329439154e-005,.00853996360068238,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172028,.0033649434139527,.000936723622457736,7.94684298987088e-005,.0138856906557926,.0302465729786341,.0420849723059915,.0479030419872811,.0474406569696583,.0416431382277769,.0323737509132328,.0219255727222366,.0124541685613893,.00547128366175477,.00152308078343358,.000129212966948341,.00209660569165992,.00456694151066636,.00635442590918465,.00723289845409743,.00716308276514704,.00628771321435967,.0048881249127316,.00331055053017243,.00188045962840229,.000826111192469611,.000229970544394388,1.95099148220722e-005,-.0135748462316425,-.029569474604983,-.0411428600764948,-.0468306867209362,-.0463786526327775,-.0407109168753721,-.0316490336334332,-.0214347478726881,-.0121753701241565,-.00534880376058466,-.00148898516796897,-.000126320411489674,-.0135748462316426,-.0295694746049834,-.0411428600764953,-.0468306867209367,-.046378652632778,-.0407109168753726,-.0316490336334336,-.0214347478726884,-.0121753701241566,-.00534880376058472,-.00148898516796898,-.000126320411489675,.00209660569165985,.0045669415106662,.00635442590918444,.00723289845409719,.00716308276514679,.00628771321435945,.00488812491273143,.00331055053017232,.00188045962840222,.000826111192469583,.00022997054439438,1.95099148220715e-005,.0138856906557927,.0302465729786343,.0420849723059919,.0479030419872815,.0474406569696587,.0416431382277773,.0323737509132331,.0219255727222367,.0124541685613895,.00547128366175482,.0015230807834336,.000129212966948343,.00853996360068239,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172029,.00336494341395271,.000936723622457737,7.94684298987089e-005,-.00427854858305892,-.00931976919031658,-.012967493161945,-.0147601943251842,-.014617721270655,-.0128313523954428,-.00997520897819597,-.00675584891158053,-.00383745876041681,-.00168584721774346,-.000469301476561811,-3.98139329439144e-005,-.00666886465162735,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907022,-6.20569627520987e-005},
			{.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831813,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932927,-.0625431691900999,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932926,-.0625431691900998,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831814,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342},
			{-.00536389192389492,-.00931249836832228,-.00723896090430979,.000615037191458855,.0111507213412938,.0200157824556806,.0237694861813762,.0214832345431866,.0149703083366938,.0075880103734666,.00231928224558504,.000206890674868773,-.0111997933843804,-.0194444741052602,-.015114932887575,.00128419617057589,.0232826792338887,.0417929054334019,.0496306297482205,.0448569418485149,.0312579676567552,.0158437473362906,.00484265572818485,.000431987229526937,-.0142763193649554,-.0247857717265967,-.0192669276724185,.00163695829281661,.0296783121801492,.0532732028778971,.0632639099894119,.057178914430488,.0398443715568573,.0201959437239261,.00617290849726889,.000550651912820983,-.013822137481785,-.0239972457633677,-.0186539763038293,.00158488066824849,.0287341366282638,.0515783876396646,.0612512538529679,.0553598442367079,.0385767765104859,.0195534369602114,.00597652572281162,.000533133663449908,-.0099487788537054,-.0172725304977166,-.0134265981099973,.00114075173240722,.0206820089326614,.0371246468163075,.0440868989979581,.0398464310178064,.0277664593408513,.0140740041403415,.0043017321169096,.000383734348187365,-.0036142966884799,-.00627494597051114,-.00487775533058912,.000414424249392327,.00751357704253213,.0134870309233169,.016016350890574,.0144758292241828,.0100872904626957,.00511295178092569,.00156277834430738,.000139407037215731,.0036142966884799,.00627494597051114,.00487775533058912,-.000414424249392327,-.00751357704253213,-.0134870309233169,-.016016350890574,-.0144758292241828,-.0100872904626957,-.00511295178092569,-.00156277834430738,-.000139407037215731,.0099487788537054,.0172725304977166,.0134265981099973,-.00114075173240722,-.0206820089326614,-.0371246468163075,-.0440868989979581,-.0398464310178064,-.0277664593408513,-.0140740041403415,-.0043017321169096,-.000383734348187365,.013822137481785,.0239972457633677,.0186539763038293,-.00158488066824849,-.0287341366282638,-.0515783876396646,-.0612512538529679,-.0553598442367079,-.0385767765104859,-.0195534369602114,-.00597652572281162,-.000533133663449908,.0142763193649554,.0247857717265967,.0192669276724185,-.00163695829281661,-.0296783121801492,-.0532732028778971,-.0632639099894119,-.057178914430488,-.0398443715568573,-.0201959437239261,-.00617290849726889,-.000550651912820983,.0111997933843804,.0194444741052602,.015114932887575,-.00128419617057589,-.0232826792338887,-.0417929054334019,-.0496306297482205,-.0448569418485149,-.0312579676567552,-.0158437473362906,-.00484265572818485,-.000431987229526937,.00536389192389492,.00931249836832228,.00723896090430979,-.000615037191458855,-.0111507213412938,-.0200157824556806,-.0237694861813762,-.0214832345431866,-.0149703083366938,-.0075880103734666,-.00231928224558504,-.000206890674868773},
			{.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065728,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700374,-.043333498156298,-.0514601409111785,-.0465104827417968,-.032410215795647,-.0164277881345211,-.00502116832732015,-.000447911376826381,.00931524214857705,.0161726184158759,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248514,-.0412794520598759,-.0373090164275984,-.0259982954865174,-.0131777737243128,-.00402779847824231,-.000359298203996196,.000529071174374757,.000918544690683033,.000714019894739991,-6.06646169956767e-005,-.00109985907972585,-.00197427048868781,-.00234451964109179,-.00211901363606869,-.00147660667381522,-.000748448629543353,-.000228764001733872,-2.0406804214751e-005,-.0103722927557633,-.0180078123747257,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831512,.0459636534116219,.0415426711023014,.0289484618473594,.0146731265551326,.004484854425811,.000400069702861968,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191293,.000684745031277692,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191294,.000684745031277693,-.0103722927557633,-.0180078123747258,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831513,.045963653411622,.0415426711023014,.0289484618473595,.0146731265551327,.00448485442581101,.000400069702861968,.000529071174374731,.000918544690682989,.000714019894739956,-6.06646169956738e-005,-.0010998590797258,-.00197427048868771,-.00234451964109168,-.00211901363606859,-.00147660667381515,-.000748448629543317,-.000228764001733861,-2.04068042147501e-005,.00931524214857704,.0161726184158758,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248513,-.0412794520598758,-.0373090164275984,-.0259982954865174,-.0131777737243127,-.0040277984782423,-.000359298203996196,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700373,-.0433334981562979,-.0514601409111784,-.0465104827417967,-.0324102157956469,-.016427788134521,-.00502116832732014,-.00044791137682638,.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065729,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687},
			{-.00744503202371008,-.0129256609858319,-.0100476103015583,.000853665892443807,.015477097348583,.0277817195939975,.0329918253981113,.0298185294215917,.0207786485173192,.0105320877132218,.00321914215189193,.000287162329453971,-.00929937817926298,-.0161450762523781,-.0125501848339245,.00106628983559657,.0193320029925161,.0347013573819497,.0412091526569903,.0372454787243047,.0259540200768997,.0131553318173118,.0040209390890345,.000358686314837344,.000395252850754388,.000686216571838048,.000533422368407895,-4.53206751167283e-005,-.000821670992004966,-.00147491694238737,-.00175151873070067,-.00158305010934183,-.00110312756683095,-.000559143020444886,-.000170902570797115,-1.52452976675568e-005,.013463002097378,.0233737343785841,.0181692970739038,-.00154370131167062,-.0279875483949175,-.0502382458492594,-.0596597856283979,-.0539214502859268,-.0375744506777747,-.0190453873833324,-.00582123990933241,-.000519281452573281,.0176551267617537,.0306518739553514,.0238268731365797,-.0020243807542107,-.0367023425450716,-.0658814870815153,-.0782367164641202,-.0707115718388728,-.0492744251929199,-.0249757614273064,-.00763386410894635,-.000680975892590434,.00806492662705201,.0140018883901234,.0108842029962012,-.000924744549741412,-.0167657646224416,-.0300949048150996,-.0357388188903564,-.032301305238965,-.0225087380642283,-.0114090193791377,-.00348717711010207,-.000311072284138478,-.00806492662705202,-.0140018883901234,-.0108842029962012,.000924744549741413,.0167657646224416,.0300949048150996,.0357388188903565,.032301305238965,.0225087380642284,.0114090193791378,.00348717711010208,.000311072284138478,-.0176551267617538,-.0306518739553516,-.0238268731365798,.00202438075421071,.0367023425450718,.0658814870815156,.0782367164641205,.0707115718388731,.0492744251929201,.0249757614273065,.00763386410894639,.000680975892590437,-.013463002097378,-.023373734378584,-.0181692970739037,.00154370131167062,.0279875483949174,.0502382458492593,.0596597856283978,.0539214502859267,.0375744506777747,.0190453873833324,.0058212399093324,.00051928145257328,-.000395252850754372,-.000686216571838022,-.000533422368407874,4.53206751167266e-005,.000821670992004935,.00147491694238732,.0017515187307006,.00158305010934177,.00110312756683091,.000559143020444865,.000170902570797108,1.52452976675562e-005,.00929937817926298,.0161450762523781,.0125501848339245,-.00106628983559657,-.0193320029925161,-.0347013573819496,-.0412091526569903,-.0372454787243047,-.0259540200768997,-.0131553318173118,-.00402093908903449,-.000358686314837344,.00744503202371004,.0129256609858318,.0100476103015582,-.000853665892443803,-.0154770973485829,-.0277817195939974,-.0329918253981111,-.0298185294215915,-.020778648517319,-.0105320877132217,-.00321914215189191,-.00028716232945397},
			{-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487695,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487696,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613},
			{.00578623505122275,.00668940613303525,-.00201954593586876,-.013036228022183,-.015415647248866,-.00450228521335004,.0135620449344063,.0268160677531348,.0273936022556361,.0174648825623257,.00613204972269062,.000587601273095628,.0120816448143678,.0139674638522171,-.00421680703729971,-.0272196126304688,-.0321878342299675,-.00940076065336832,.0283175170734496,.0559918847131752,.057197775359769,.0364666326159833,.0128037050133204,.00122691003928565,.0154004108740453,.0178042547598082,-.00537514237082621,-.0346967010520833,-.0410296677235249,-.0119831015407985,.0360961942322037,.0713725691695229,.0729097118112787,.0464838301496112,.0163208173179129,.00156393595415252,.0149104675326585,.0172378363610222,-.0052041394518047,-.033592872213516,-.0397243640751441,-.0116018752957915,.0349478424019757,.0691019469563708,.0705901874741346,.0450050096655201,.0158015924844782,.00151418143699325,.0107321276671759,.0124073011209858,-.00374579059122209,-.0241791877090991,-.0285924600162962,-.00835069769478169,.0251544564601237,.0497376031410035,.0508087960597447,.0323933175358136,.011373533896014,.00108986444975558,.00389887985834265,.00450745443375195,-.00136081007816862,-.00878406881409194,-.0103873686481536,-.00303372901021968,.00913837467102563,.0180691979356344,.018458352129919,.0117681840173745,.00413189663788404,.000395937383830426,-.00389887985834265,-.00450745443375195,.00136081007816862,.00878406881409194,.0103873686481536,.00303372901021968,-.00913837467102562,-.0180691979356344,-.018458352129919,-.0117681840173745,-.00413189663788404,-.000395937383830426,-.0107321276671759,-.0124073011209858,.00374579059122209,.0241791877090991,.0285924600162962,.00835069769478169,-.0251544564601236,-.0497376031410035,-.0508087960597447,-.0323933175358136,-.011373533896014,-.00108986444975558,-.0149104675326585,-.0172378363610222,.0052041394518047,.033592872213516,.0397243640751441,.0116018752957915,-.0349478424019757,-.0691019469563708,-.0705901874741346,-.0450050096655201,-.0158015924844782,-.00151418143699325,-.0154004108740453,-.0178042547598082,.00537514237082622,.0346967010520833,.0410296677235249,.0119831015407985,-.0360961942322037,-.0713725691695229,-.0729097118112787,-.0464838301496112,-.0163208173179129,-.00156393595415252,-.0120816448143678,-.0139674638522171,.00421680703729971,.0272196126304688,.0321878342299675,.00940076065336832,-.0283175170734496,-.0559918847131752,-.057197775359769,-.0364666326159833,-.0128037050133204,-.00122691003928565,-.00578623505122275,-.00668940613303525,.00201954593586876,.013036228022183,.015415647248866,.00450228521335004,-.0135620449344063,-.0268160677531348,-.0273936022556361,-.0174648825623257,-.00613204972269062,-.000587601273095628},
			{-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137491,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099975,.00974729658452533,-.0293613727260863,-.0580558878956585,-.0593062307364416,-.0378108853797437,-.0132756821226898,-.00127213706171361,-.010048707428968,-.0116172060950429,.00350725922283749,.0226394607568784,.0267716965627408,.00781892655072194,-.0235526245439282,-.0465703202274322,-.047573299745961,-.0303305160603543,-.0106492690078639,-.00102046204932158,-.000570729279560872,-.000659814181276736,.000199199304351608,.0012858373297019,.00152053298395657,.000444086002978713,-.00133770164299672,-.00264502131246258,-.00270198682589509,-.00172266072051549,-.000604838947861804,-5.79584562835985e-005,.0111889882847748,.0129354729270487,-.00390524678257324,-.0252084820831874,-.0298096249016592,-.00870618218253045,.0262252674745031,.0518549048349062,.052971697831619,.0337722827805856,.0118576988147661,.00113625936426423,.0191506731907347,.0221398940001207,-.00668408107674887,-.043145938642774,-.0510210905490773,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0191506731907347,.0221398940001207,-.00668408107674888,-.043145938642774,-.0510210905490774,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0111889882847749,.0129354729270488,-.00390524678257325,-.0252084820831874,-.0298096249016593,-.00870618218253046,.0262252674745032,.0518549048349063,.0529716978316191,.0337722827805856,.0118576988147661,.00113625936426423,-.000570729279560845,-.000659814181276704,.000199199304351599,.00128583732970183,.0015205329839565,.000444086002978692,-.00133770164299666,-.00264502131246246,-.00270198682589496,-.00172266072051541,-.000604838947861776,-5.79584562835958e-005,-.010048707428968,-.0116172060950428,.00350725922283748,.0226394607568783,.0267716965627407,.00781892655072193,-.0235526245439282,-.0465703202274321,-.0475732997459609,-.0303305160603542,-.0106492690078639,-.00102046204932158,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099974,.00974729658452531,-.0293613727260862,-.0580558878956584,-.0593062307364415,-.0378108853797436,-.0132756821226898,-.00127213706171361,-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137492,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872},
			{.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503008,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503009},
			{-.00590524039988117,-.00279948660884738,.00950696592933597,.0123888349467185,-.00252754986470693,-.018634756913296,-.0137887462421145,.010811667317975,.0309214103019088,.0286485035369933,.0122170775898605,.00128660900891531,-.0123301276949928,-.00584532127908038,.019850521903057,.025867857452222,-.00527751801398726,-.0389093274357997,-.0287908688565067,.0225747352519459,.0645638300414449,.0598180062049106,.0255092285058617,.00268643650371509,-.0157171507315389,-.00745100114858395,.025303358778414,.0329736256376032,-.00672722522962469,-.0495975207474624,-.0366995733299215,.0287758996059307,.0822991840529711,.0762497066729088,.0325164831532697,.00342438687611848,-.0152171307379511,-.00721395757687207,.0244983664798579,.031924614123907,-.00651320761449607,-.0480196423883553,-.0355320258059113,.027860433095352,.0796809462956736,.0738239248952259,.0314820150123508,.00331544461723926,-.0109528550630685,-.00519240013976796,.0176332228431619,.0229784232957589,-.00468802037820009,-.0345631638655983,-.0255749349500402,.0200531421424059,.0573520640056069,.0531363476787172,.0226598531261651,.00238636211961347,-.00397906802090637,-.001886349561729,.00640598206739025,.00834784252877614,-.00170311319385134,-.0125564685414928,-.00929113050528727,.0072851157217069,.0208354591112203,.0193039294849313,.00823210901756414,.000866942650275162,.00397906802090637,.001886349561729,-.00640598206739025,-.00834784252877614,.00170311319385134,.0125564685414928,.00929113050528727,-.0072851157217069,-.0208354591112203,-.0193039294849313,-.00823210901756414,-.000866942650275162,.0109528550630685,.00519240013976796,-.0176332228431619,-.0229784232957589,.00468802037820009,.0345631638655983,.0255749349500402,-.0200531421424059,-.0573520640056069,-.0531363476787172,-.0226598531261651,-.00238636211961347,.0152171307379511,.00721395757687207,-.0244983664798579,-.031924614123907,.00651320761449607,.0480196423883553,.0355320258059113,-.027860433095352,-.0796809462956736,-.0738239248952259,-.0314820150123508,-.00331544461723926,.0157171507315389,.00745100114858395,-.025303358778414,-.0329736256376032,.00672722522962468,.0495975207474624,.0366995733299215,-.0287758996059307,-.0822991840529711,-.0762497066729088,-.0325164831532697,-.00342438687611848,.0123301276949928,.00584532127908038,-.019850521903057,-.025867857452222,.00527751801398726,.0389093274357997,.0287908688565067,-.0225747352519459,-.0645638300414449,-.0598180062049106,-.0255092285058617,-.00268643650371509,.00590524039988117,.00279948660884738,-.00950696592933597,-.0123888349467185,.00252754986470693,.018634756913296,.0137887462421145,-.010811667317975,-.0309214103019089,-.0286485035369933,-.0122170775898605,-.00128660900891531},
			{-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599055,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599056},
			{-.00423058507690693,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394862,-.00435422423400533,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415894,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415895,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.00423058507690692,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040534,-.00720462946394862,-.00435422423400534,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007},
			{.00719247221789664,.015054738114976,.0194706196180468,.0198760629317139,.0169623740708349,.0122486834132115,.00740267274248333,.00362315688940088,.00134348757825736,.000329082012180178,3.81739176729443e-005,6.2280977297559e-007,.0150178646226051,.0314342568299276,.0406546067448804,.0415011714047538,.0354173960992067,.0255752213886969,.0154567628103504,.00756514283587924,.00280519881915107,.000687122446765829,7.9707047918831e-005,1.30042530201362e-006,.0191431952513503,.0400690864645718,.051822219359513,.0529013310031521,.0451463737261948,.0326006039436125,.0197026565272781,.00964324889396088,.00357577259239878,.000875871469784373,.000101602166457287,1.65764548368426e-006,.0185341802631805,.0387943424159014,.0501735652087417,.051218346368036,.0437101026178578,.0315634596129842,.0190758430317083,.00933646191122561,.00346201420073717,.000848006797988197,9.83698302986383e-005,1.60490972398582e-006,.0133403723495074,.0279230570510056,.0361134958483479,.0368654994163948,.0314612805140234,.0227184746180725,.0137302457033551,.00672011799573213,.00249185870976258,.000610371016113917,7.08037876780903e-005,1.15516807332712e-006,.00484643033229705,.0101441809206661,.0131196893984132,.0133928851388839,.0114295838362937,.00825340564775263,.00498806761185259,.00244135492157639,.000905268557607982,.000221741982078836,2.57223423195746e-005,4.19661568867713e-007,-.00484643033229705,-.0101441809206661,-.0131196893984132,-.0133928851388839,-.0114295838362937,-.00825340564775263,-.00498806761185259,-.00244135492157639,-.000905268557607982,-.000221741982078836,-2.57223423195746e-005,-4.19661568867713e-007,-.0133403723495074,-.0279230570510056,-.0361134958483479,-.0368654994163948,-.0314612805140234,-.0227184746180725,-.0137302457033551,-.00672011799573213,-.00249185870976258,-.000610371016113917,-7.08037876780903e-005,-1.15516807332712e-006,-.0185341802631805,-.0387943424159014,-.0501735652087417,-.051218346368036,-.0437101026178578,-.0315634596129842,-.0190758430317083,-.00933646191122561,-.00346201420073717,-.000848006797988197,-9.83698302986383e-005,-1.60490972398582e-006,-.0191431952513503,-.0400690864645718,-.051822219359513,-.0529013310031521,-.0451463737261948,-.0326006039436125,-.0197026565272781,-.00964324889396088,-.00357577259239877,-.000875871469784373,-.000101602166457287,-1.65764548368426e-006,-.0150178646226051,-.0314342568299276,-.0406546067448804,-.0415011714047538,-.0354173960992067,-.025575221388697,-.0154567628103504,-.00756514283587924,-.00280519881915107,-.000687122446765829,-7.97070479188311e-005,-1.30042530201362e-006,-.00719247221789664,-.015054738114976,-.0194706196180468,-.0198760629317139,-.0169623740708349,-.0122486834132115,-.00740267274248333,-.00362315688940088,-.00134348757825736,-.000329082012180178,-3.81739176729443e-005,-6.2280977297559e-007},
			{-.00894139816848509,-.0187154574574963,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007,-.0155714612898606,-.0325930034464412,-.0421532388985272,-.0430310100838354,-.0367229713548178,-.0265179890644552,-.0160265383805396,-.00784401323233701,-.00290860557878677,-.000712451526903725,-8.26452523302774e-005,-1.34836228448754e-006,-.0124908594969934,-.0261449210871181,-.0338137940058043,-.0345179101026881,-.029457831026034,-.0212717656603367,-.0128559057758339,-.0062921819188208,-.0023331775316716,-.000571502684000991,-6.62950134052773e-005,-1.08160714868461e-006,-.000709434451366774,-.00148493446363173,-.00192049797733382,-.00196048915784558,-.00167309545011213,-.00120815732532064,-.00073016772489491,-.000357372575406282,-.000132515822671862,-3.24592309421519e-005,-3.76531066375748e-006,-6.1431270947059e-008,.0139082644774524,.0291117258431486,.0376508270012981,.038434842937163,.0328005694836934,.0236855872710412,.014314734520114,.0070061896291659,.002597935729815,.000636354166066229,7.38178649914994e-005,1.20434292678016e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102608,.0245005889388461,.0119915442295211,.0044465341160397,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102609,.0245005889388461,.0119915442295211,.00444653411603971,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0139082644774524,.0291117258431486,.0376508270012982,.038434842937163,.0328005694836935,.0236855872710413,.0143147345201141,.00700618962916591,.002597935729815,.00063635416606623,7.38178649914996e-005,1.20434292678016e-006,-.00070943445136674,-.00148493446363166,-.00192049797733373,-.00196048915784549,-.00167309545011205,-.00120815732532058,-.000730167724894874,-.000357372575406265,-.000132515822671856,-3.24592309421504e-005,-3.7653106637573e-006,-6.14312709470561e-008,-.0124908594969934,-.0261449210871181,-.0338137940058042,-.0345179101026881,-.029457831026034,-.0212717656603366,-.0128559057758339,-.00629218191882079,-.0023331775316716,-.00057150268400099,-6.62950134052772e-005,-1.08160714868461e-006,-.0155714612898606,-.0325930034464411,-.0421532388985272,-.0430310100838353,-.0367229713548177,-.0265179890644552,-.0160265383805396,-.007844013232337,-.00290860557878677,-.000712451526903724,-8.26452523302773e-005,-1.34836228448754e-006,-.00894139816848509,-.0187154574574964,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007},
			{.00998308444276627,.0208958362733153,.0270250386669949,.0275877902097846,.0235436172110032,.0170010584848057,.0102748408129391,.00502890801389247,.00186474824444619,.000456762767608005,5.29850421517036e-005,8.64454163607971e-007,.0124695874152247,.0261003959766295,.0337562087138414,.0344591256926668,.0294076639906539,.0212355395912979,.0128340120159257,.00628146625843626,.00232920410268166,.00057052940815652,6.61821122117002e-005,1.07976515889088e-006,-.00052999672436047,-.00110934900341893,-.0014347451482894,-.00146462133295129,-.00124991830660814,-.000902577290542385,-.000545485917243634,-.000266982092533976,-9.89985076235179e-005,-2.42492960998119e-005,-2.8129481365704e-006,-4.5893418782406e-008,-.0180526136574343,-.0377863636688457,-.0488699244135907,-.0498875594186185,-.0425743995300703,-.0307433581628203,-.0185802026068238,-.00909387614760269,-.00337206199330092,-.000825973357633232,-9.58139241505331e-005,-1.5632099607735e-006,-.0236738567072675,-.0495523238883786,-.0640870961854267,-.0654216035175219,-.0558312637160867,-.0403162594435231,-.0243657268943327,-.011925537482675,-.00442205843166394,-.00108316586637731,-.000125648571112395,-2.04996402831038e-006,-.0108143045303454,-.0226356832070683,-.0292752204756728,-.029884828317171,-.0255039259384566,-.0184166151099846,-.0111303533681376,-.00544762924015274,-.00202001250249669,-.000494794138560608,-5.7396727901756e-005,-9.36431083136384e-007,.0108143045303454,.0226356832070683,.0292752204756728,.0298848283171711,.0255039259384567,.0184166151099846,.0111303533681376,.00544762924015275,.00202001250249669,.000494794138560608,5.73967279017561e-005,9.36431083136386e-007,.0236738567072676,.0495523238883789,.064087096185427,.0654216035175222,.055831263716087,.0403162594435233,.0243657268943328,.0119255374826751,.00442205843166396,.00108316586637732,.000125648571112395,2.04996402831039e-006,.0180526136574342,.0377863636688456,.0488699244135906,.0498875594186183,.0425743995300703,.0307433581628202,.0185802026068238,.00909387614760267,.00337206199330092,.00082597335763323,9.58139241505329e-005,1.56320996077349e-006,.00052999672436045,.00110934900341889,.00143474514828935,.00146462133295124,.00124991830660809,.00090257729054235,.000545485917243613,.000266982092533966,9.89985076235141e-005,2.4249296099811e-005,2.81294813657029e-006,4.58934187824042e-008,-.0124695874152247,-.0261003959766295,-.0337562087138414,-.0344591256926668,-.0294076639906538,-.0212355395912979,-.0128340120159257,-.00628146625843626,-.00232920410268166,-.00057052940815652,-6.61821122117001e-005,-1.07976515889088e-006,-.00998308444276622,-.0208958362733152,-.0270250386669947,-.0275877902097844,-.0235436172110031,-.0170010584848056,-.0102748408129391,-.00502890801389244,-.00186474824444618,-.000456762767608003,-5.29850421517033e-005,-8.64454163607967e-007},
			{.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.00942311384888909,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848839,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848838,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.0094231138488891,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006},
			{-.00776915549838746,-.0115499204629525,-.00437647479049313,.0101990855232791,.0237503437017621,.0291595409512613,.0251163517269084,.0158452413384258,.00706723987011998,.0019739247223913,.000249684056884166,4.2685397699263e-006,-.0162219779196966,-.024116206035927,-.00913806879424977,.0212956659438816,.0495906602970113,.0608850510918503,.0524428817548842,.0330848256515858,.0147563797828581,.00412155005372477,.00052133970783773,8.91270073194743e-006,-.0206780723147828,-.0307408045330656,-.0116482495710665,.0271454764985451,.0632129611343108,.0776098633347217,.0668486732438716,.0421730581025391,.0188098818630402,.00525371878089153,.00066454906008391,1.13609740543078e-005,-.0200202272789457,-.0297628272171986,-.0112776761907307,.0262818797044071,.0612019258671093,.075140810003757,.0647219726899549,.0408313790283736,.0182114707916447,.00508657878993488,.00064340732629019,1.09995399578346e-005,-.0144099864483064,-.0214224309688108,-.00811734846026063,.018916944603062,.0440513941259184,.054084203879524,.0465850230556935,.0293892576875637,.0131080953105517,.00366117379238187,.000463106173741249,7.91715396241739e-006,-.00523501095632017,-.0077825653226702,-.00294895545378721,.00687234596731837,.0160034523084138,.019648276623219,.0169238955894816,.0106768376600795,.00476204629431086,.00133006960033359,.000168242066165155,2.87622669770101e-006,.00523501095632017,.0077825653226702,.00294895545378721,-.00687234596731837,-.0160034523084138,-.019648276623219,-.0169238955894816,-.0106768376600795,-.00476204629431086,-.00133006960033359,-.000168242066165155,-2.87622669770101e-006,.0144099864483064,.0214224309688108,.00811734846026063,-.018916944603062,-.0440513941259184,-.054084203879524,-.0465850230556935,-.0293892576875637,-.0131080953105517,-.00366117379238187,-.000463106173741249,-7.91715396241739e-006,.0200202272789457,.0297628272171986,.0112776761907307,-.0262818797044071,-.0612019258671093,-.075140810003757,-.0647219726899549,-.0408313790283736,-.0182114707916447,-.00508657878993488,-.00064340732629019,-1.09995399578346e-005,.0206780723147828,.0307408045330656,.0116482495710665,-.0271454764985451,-.0632129611343108,-.0776098633347217,-.0668486732438716,-.0421730581025391,-.0188098818630402,-.00525371878089153,-.00066454906008391,-1.13609740543078e-005,.0162219779196966,.024116206035927,.00913806879424977,-.0212956659438816,-.0495906602970113,-.0608850510918503,-.0524428817548842,-.0330848256515858,-.0147563797828581,-.00412155005372477,-.00052133970783773,-8.91270073194743e-006,.00776915549838746,.0115499204629525,.00437647479049313,-.0101990855232791,-.0237503437017621,-.0291595409512613,-.0251163517269084,-.0158452413384258,-.00706723987011998,-.0019739247223913,-.000249684056884166,-4.2685397699263e-006},
			{.00965830811000427,.014358406200055,.0054406610822788,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.0312236592031229,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597266e-006,.0168199612662992,.0250051907000728,.00947492124135092,-.0220806783296889,-.0514186981079021,-.0631294288607147,-.0543760596998002,-.034304416435326,-.0153003374562057,-.00427348086675615,-.000540557614849618,-9.24124553932794e-006,.0134923607368192,.0200582538732903,.00760043696398981,-.0177123164923541,-.0412462081514687,-.0506401301294754,-.0436184959823004,-.0275177542970243,-.012273373825647,-.00342803081073059,-.000433615644121537,-7.41299082085032e-006,.00076631600405656,.00113923436056569,.000431676605520518,-.00100599382582219,-.00234263151050095,-.00287617140711524,-.00247736865298144,-.00156290629377859,-.000697082072575792,-.000194699424652437,-2.46277589356807e-005,-4.21030360420172e-007,-.0150234114472303,-.022334371777439,-.00846289940774852,.0197222282696639,.0459266371382019,.0563865379466385,.0485681212230874,.0306403418180061,.0136660995377323,.00381702789659693,.000482820342463611,8.25418274299939e-006,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962322,1.41275647220829e-005,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962323,1.41275647220829e-005,-.0150234114472303,-.022334371777439,-.00846289940774854,.0197222282696639,.045926637138202,.0563865379466386,.0485681212230875,.0306403418180062,.0136660995377323,.00381702789659694,.000482820342463612,8.25418274299941e-006,.000766316004056523,.00113923436056563,.000431676605520497,-.00100599382582214,-.00234263151050084,-.0028761714071151,-.00247736865298132,-.00156290629377852,-.000697082072575759,-.000194699424652428,-2.46277589356795e-005,-4.21030360420152e-007,.0134923607368192,.0200582538732903,.0076004369639898,-.0177123164923541,-.0412462081514686,-.0506401301294753,-.0436184959823003,-.0275177542970243,-.012273373825647,-.00342803081073058,-.000433615644121536,-7.41299082085031e-006,.0168199612662992,.0250051907000728,.0094749212413509,-.0220806783296889,-.051418698107902,-.0631294288607146,-.0543760596998001,-.0343044164353259,-.0153003374562056,-.00427348086675614,-.000540557614849618,-9.24124553932793e-006,.00965830811000427,.014358406200055,.00544066108227881,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.031223659203123,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597267e-006},
			{-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417216e-006,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.0255511390040629,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.025551139004063,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417217e-006},
			{.00794411476554508,.00613541631233657,-.0091565995280034,-.0197247070338324,-.0100866971092938,.0140548407519696,.0324308869658652,.0322559411834705,.0190400394460011,.00638699606694611,.00090669726269933,1.64722539026166e-005,.0165872924470305,.0128107344443715,-.0191189577031385,-.041185140655471,-.0210609992068756,.0293464735505198,.0677156136707053,.0673503272995239,.0397555563854127,.0133360323644821,.00189318169502602,3.43940263715078e-005,.0211437368750014,.0163297776977338,-.0243708376331069,-.052498488222838,-.0268463480086837,.0374077997926938,.0863167465308033,.0858511178610883,.050676204462949,.0169993722707645,.0024132290272223,4.38418895665644e-005,.0204710773480144,.0158102678018768,-.0235955122395174,-.0508283194885088,-.0259922675847922,.0362177209971177,.0835706954313348,.0831198801109617,.0490640092334553,.0164585601249111,.00233645539418102,4.24471188657934e-005,.0147344954209059,.011379778141117,-.0169833742082372,-.0365847692343567,-.0187084900904949,.0260684787183142,.0601517940761608,.0598273101146777,.0353148692221493,.011846400395665,.00168171370375142,3.05522209664918e-005,.0053529019781526,.00413416511271193,-.00616989824205258,-.0132908985350789,-.00679661642648238,.00947043025991277,.0218525744046085,.0217346924690766,.0128295559445646,.0043036845375764,.00061095058597832,1.10993311529656e-005,-.0053529019781526,-.00413416511271193,.00616989824205258,.0132908985350789,.00679661642648238,-.00947043025991277,-.0218525744046085,-.0217346924690766,-.0128295559445646,-.0043036845375764,-.00061095058597832,-1.10993311529656e-005,-.0147344954209059,-.011379778141117,.0169833742082372,.0365847692343567,.0187084900904949,-.0260684787183142,-.0601517940761608,-.0598273101146777,-.0353148692221493,-.011846400395665,-.00168171370375142,-3.05522209664918e-005,-.0204710773480144,-.0158102678018768,.0235955122395174,.0508283194885088,.0259922675847922,-.0362177209971177,-.0835706954313348,-.0831198801109616,-.0490640092334554,-.0164585601249111,-.00233645539418102,-4.24471188657934e-005,-.0211437368750014,-.0163297776977338,.0243708376331069,.052498488222838,.0268463480086837,-.0374077997926938,-.0863167465308033,-.0858511178610884,-.050676204462949,-.0169993722707645,-.0024132290272223,-4.38418895665644e-005,-.0165872924470305,-.0128107344443715,.0191189577031385,.041185140655471,.0210609992068756,-.0293464735505198,-.0677156136707053,-.0673503272995239,-.0397555563854127,-.0133360323644821,-.00189318169502602,-3.43940263715078e-005,-.00794411476554508,-.00613541631233657,.0091565995280034,.0197247070338324,.0100866971092938,-.0140548407519696,-.0324308869658652,-.0322559411834705,-.0190400394460011,-.00638699606694611,-.00090669726269933,-1.64722539026166e-005},
			{.00456148901693467,-2.90196057170823e-005,-.00924310280718151,-.00451266930256049,.0104810735535174,.0123349344803261,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.0154783325752891,-9.84711586118446e-005,-.0313642801168654,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0154783325752891,-9.84711586118446e-005,-.0313642801168653,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.00456148901693467,-2.90196057170823e-005,-.00924310280718152,-.00451266930256049,.0104810735535174,.0123349344803262,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005},
			{-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.0125539393028941,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855103,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855104,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.012553939302894,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009},
			{.00973618795942079,.0195825969134841,.023541414184445,.0215525014129859,.0158497577055902,.00941532722371888,.00442368433837659,.00156467226043237,.000378749981717022,5.17272161167757e-005,2.50040624924348e-006,7.84521387564148e-009,.0203291369483494,.040888415067351,.0491544159693757,.0450015709053905,.0330942558153189,.0196591805070174,.00923664221622893,.0032670319014163,.000790829047220676,.000108006302341081,5.22084221044592e-006,1.63807876277374e-008,.0259134469295932,.052120253632755,.0626568827202963,.0573632723498304,.0421850786840353,.0250594568791694,.0117739006080788,.00416446886110988,.00100806574315481,.00013767508138138,6.65498087262162e-006,2.08805062376629e-008,.0250890454977502,.0504621179227659,.0606635383392967,.0555383370570115,.0408430171912504,.0242622239912202,.01139932980914,.00403198189010626,.000975995488498737,.000133295133992354,6.44326161446257e-006,2.02162210390667e-008,.0180583766900454,.0363211878310999,.0436638781965725,.0399749050398971,.0293976345040205,.0174632542402221,.00820491125204314,.00290210513530733,.000702493611431951,9.59420214212558e-005,4.63767524981805e-006,1.45510571458537e-008,.00656043641434894,.0131951419194234,.0158626714587474,.0145225026138949,.0106798808777874,.00634423409130352,.00298076618282515,.00105430718024807,.000255209244351746,3.4854823431899e-005,1.68482328777663e-006,5.28626059836008e-009,-.00656043641434894,-.0131951419194234,-.0158626714587474,-.0145225026138949,-.0106798808777874,-.00634423409130352,-.00298076618282515,-.00105430718024807,-.000255209244351746,-3.4854823431899e-005,-1.68482328777663e-006,-5.28626059836008e-009,-.0180583766900454,-.0363211878310999,-.0436638781965725,-.0399749050398971,-.0293976345040205,-.0174632542402221,-.00820491125204314,-.00290210513530733,-.000702493611431951,-9.59420214212558e-005,-4.63767524981805e-006,-1.45510571458537e-008,-.0250890454977502,-.0504621179227659,-.0606635383392967,-.0555383370570115,-.0408430171912504,-.0242622239912202,-.01139932980914,-.00403198189010626,-.000975995488498737,-.000133295133992354,-6.44326161446257e-006,-2.02162210390667e-008,-.0259134469295932,-.052120253632755,-.0626568827202963,-.0573632723498304,-.0421850786840353,-.0250594568791694,-.0117739006080788,-.00416446886110988,-.00100806574315481,-.00013767508138138,-6.65498087262162e-006,-2.08805062376629e-008,-.0203291369483494,-.040888415067351,-.0491544159693757,-.0450015709053905,-.0330942558153189,-.0196591805070174,-.00923664221622893,-.0032670319014163,-.000790829047220676,-.000108006302341081,-5.22084221044592e-006,-1.63807876277374e-008,-.00973618795942079,-.0195825969134841,-.023541414184445,-.0215525014129859,-.0158497577055902,-.00941532722371888,-.00442368433837659,-.00156467226043237,-.000378749981717022,-5.17272161167757e-005,-2.50040624924348e-006,-7.84521387564149e-009},
			{-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642852e-009,-.0210785206129219,-.042395665985065,-.0509663726925087,-.0466604432029348,-.0343141942102539,-.0203838678741335,-.0095771283278592,-.00338746300209602,-.000819981016188288,-.000111987689197342,-5.41329572570134e-006,-1.69846250996501e-008,-.0169084220471927,-.0340082601910801,-.0408833691663904,-.0374293092514581,-.0275255976721928,-.0163511969032008,-.00768242376878656,-.00271729952782253,-.000657758926587005,-8.98324482923012e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.000960335605521408,-.00193154293476799,-.00232202360306468,-.00212584582191778,-.0015633517683723,-.000928687285851265,-.000436333151684879,-.000154332526131126,-3.73582653241091e-005,-5.10214959062846e-006,-2.46629292637403e-007,-7.7381807429253e-010,.0188271115999142,.0378673603101058,.0455226248627925,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606287,.00302564611398512,.000732398368229125,.000100026219157067,4.83509846932233e-006,1.51704874410297e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019803,.0146410459204576,.00517858795138701,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019804,.0146410459204576,.00517858795138702,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.0188271115999142,.0378673603101059,.0455226248627926,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606288,.00302564611398512,.000732398368229127,.000100026219157067,4.83509846932234e-006,1.51704874410297e-008,-.000960335605521362,-.0019315429347679,-.00232202360306457,-.00212584582191767,-.00156335176837222,-.000928687285851221,-.000436333151684858,-.000154332526131119,-3.73582653241074e-005,-5.10214959062822e-006,-2.46629292637392e-007,-7.73818074292493e-010,-.0169084220471927,-.03400826019108,-.0408833691663903,-.0374293092514581,-.0275255976721927,-.0163511969032008,-.00768242376878655,-.00271729952782252,-.000657758926587003,-8.9832448292301e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.0210785206129218,-.0423956659850649,-.0509663726925086,-.0466604432029347,-.0343141942102538,-.0203838678741334,-.00957712832785918,-.00338746300209602,-.000819981016188286,-.000111987689197342,-5.41329572570133e-006,-1.69846250996501e-008,-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642853e-009},
			{.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700912,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700913,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008},
			{-.00998215054668693,-.0124473495345263,.000932730855103969,.0206523073750313,.0326901637864827,.0309109950234972,.020097626924235,.00901835704324359,.00260125692755759,.000402780024162244,2.11581155396364e-005,6.94419156685537e-008,-.0208427062366112,-.0259900357702638,.00194753977310106,.0431219679279274,.0682569830460539,.0645420829652721,.0419637964861058,.0188303077287876,.00543141818323182,.000841003717818616,4.41780941541987e-005,1.44994552227784e-007,-.0265680910755723,-.0331293657149398,.00248251899142349,.0549673520444605,.0870068273056559,.0822714631661617,.0534910368290498,.0240028969866065,.00692340098850254,.0010720231392338,5.63135907453441e-005,1.84823814399061e-007,-.0257228630214447,-.0320753995403257,.00240354099144272,.0532186397309965,.0842388222152567,.0796541073115629,.0517892914968888,.0232392771293986,.00670314230567841,.00103791816611484,5.45220496597903e-005,1.78943893536226e-007,-.018514580398419,-.0230869543217517,.00172999999610706,.0383052533138549,.0606326926078268,.0573327460732011,.0372764493748974,.0167269663744151,.00482472993138678,.000747063782032829,3.92434104660595e-005,1.28798691689977e-007,-.00672617088052745,-.00838727081777638,.000628492536515794,.0139159340296408,.0220272802651019,.0208284410901999,.013542179348353,.00607674771591152,.00175277847364417,.000271401163217019,1.42567575958641e-005,4.67913390880377e-008,.00672617088052745,.00838727081777638,-.000628492536515794,-.0139159340296408,-.0220272802651019,-.0208284410901999,-.013542179348353,-.00607674771591152,-.00175277847364417,-.000271401163217019,-1.42567575958641e-005,-4.67913390880377e-008,.018514580398419,.0230869543217517,-.00172999999610706,-.0383052533138549,-.0606326926078268,-.0573327460732011,-.0372764493748974,-.0167269663744151,-.00482472993138678,-.000747063782032829,-3.92434104660595e-005,-1.28798691689977e-007,.0257228630214447,.0320753995403257,-.00240354099144272,-.0532186397309965,-.0842388222152567,-.0796541073115629,-.0517892914968887,-.0232392771293986,-.00670314230567841,-.00103791816611484,-5.45220496597903e-005,-1.78943893536226e-007,.0265680910755723,.0331293657149398,-.00248251899142349,-.0549673520444605,-.0870068273056559,-.0822714631661617,-.0534910368290498,-.0240028969866065,-.00692340098850254,-.0010720231392338,-5.63135907453441e-005,-1.84823814399061e-007,.0208427062366112,.0259900357702638,-.00194753977310106,-.0431219679279274,-.0682569830460539,-.0645420829652721,-.0419637964861058,-.0188303077287876,-.00543141818323182,-.000841003717818616,-4.41780941541987e-005,-1.44994552227784e-007,.00998215054668692,.0124473495345263,-.000932730855103969,-.0206523073750313,-.0326901637864827,-.0309109950234972,-.020097626924235,-.00901835704324359,-.00260125692755759,-.000402780024162244,-2.11581155396364e-005,-6.94419156685537e-008},
			{-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243526,-.0333379180302253,-.0122284658175383,-.00222781405170755,-.000129893528533068,-4.50638134603823e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243525,-.0333379180302253,-.0122284658175383,-.00222781405170754,-.000129893528533068,-4.50638134603823e-007,-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007},
			{-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.00895405768790681,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.0195503931804071,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.019550393180407,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.0089540576879068,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011},
			{.0122290736621916,.0236353054365663,.0264106554398282,.0216849715365018,.0137420800410902,.00671545250392889,.00245286453144479,.000626979702197685,9.90753102188358e-005,7.54445587508782e-006,1.51966644376378e-007,9.16955121940153e-011,.0255342762759209,.0493504606770767,.0551453847819794,.0452781682034031,.0286934299414573,.0140218486117464,.00512157521848587,.00130913210415982,.000206869008490129,1.5752805648851e-005,3.17305986489221e-007,1.91460008035021e-010,.0325484114177003,.0629067799068687,.0702935400351982,.057715849511425,.0365753684430921,.0178735787347226,.00652844574550074,.00166874399986403,.000263694867445998,2.0080020819912e-005,4.04468318661127e-007,2.44053093348749e-010,.0315129275220274,.0609054853832662,.068057245681468,.0558796974508159,.0354117723303635,.0173049548869921,.00632075203208146,.00161565515581309,.000255305770199198,1.94412019873441e-005,3.91600703559772e-007,2.36288872704961e-010,.0226821030656547,.0438380247522649,.0489856570714547,.0402206064787555,.0254883799410055,.0124556111144829,.00454949635967349,.00116290232753188,.00018376178439362,1.39932206199794e-005,2.81862975520663e-007,1.7007396600062e-010,.0082401922088581,.0159259372453132,.0177960230838209,.0146117636086239,.00925968589413191,.00452500499468858,.00165278873606634,.000422471349823863,6.67589076578628e-005,5.0836038966856e-006,1.02398136897977e-007,6.17862534841391e-011,-.0082401922088581,-.0159259372453132,-.0177960230838209,-.0146117636086239,-.00925968589413191,-.00452500499468858,-.00165278873606634,-.000422471349823863,-6.67589076578628e-005,-5.0836038966856e-006,-1.02398136897977e-007,-6.17862534841391e-011,-.0226821030656547,-.0438380247522649,-.0489856570714547,-.0402206064787555,-.0254883799410055,-.0124556111144829,-.00454949635967349,-.00116290232753188,-.00018376178439362,-1.39932206199794e-005,-2.81862975520663e-007,-1.7007396600062e-010,-.0315129275220274,-.0609054853832662,-.068057245681468,-.0558796974508159,-.0354117723303635,-.0173049548869921,-.00632075203208146,-.00161565515581309,-.000255305770199198,-1.94412019873441e-005,-3.91600703559772e-007,-2.36288872704961e-010,-.0325484114177003,-.0629067799068687,-.0702935400351982,-.057715849511425,-.0365753684430921,-.0178735787347226,-.00652844574550074,-.00166874399986403,-.000263694867445998,-2.0080020819912e-005,-4.04468318661127e-007,-2.44053093348749e-010,-.0255342762759209,-.0493504606770767,-.0551453847819794,-.0452781682034031,-.0286934299414573,-.0140218486117464,-.00512157521848587,-.00130913210415982,-.000206869008490129,-1.5752805648851e-005,-3.17305986489221e-007,-1.91460008035021e-010,-.0122290736621916,-.0236353054365663,-.0264106554398282,-.0216849715365018,-.0137420800410902,-.00671545250392889,-.00245286453144479,-.000626979702197685,-9.90753102188358e-005,-7.54445587508782e-006,-1.51966644376378e-007,-9.16955121940154e-011},
			{.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594673,-.00597366639905114,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.024026051223952,.0244245520757115,-.0153331658656943,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597428,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.024026051223952,.0244245520757115,-.0153331658656942,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597429,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594672,-.00597366639905115,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010},
			{-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667211,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122411e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122412e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667212,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{.0018918113337895,.00428844840289335,.00641941367947712,.00814735927306806,.00936344787276585,.00999122031360907,.00999122031360907,.00936344787276585,.00814735927306806,.00641941367947712,.00428844840289335,.0018918113337895,.00395009749714678,.00895426991072093,.0134037202629815,.0170116664902499,.0195508565501554,.0208616438908705,.0208616438908705,.0195508565501554,.0170116664902499,.0134037202629815,.00895426991072092,.00395009749714678,.0050351690836213,.0114139620739561,.0170856536889082,.0216846842981607,.0249213768851943,.0265922308066324,.0265922308066324,.0249213768851943,.0216846842981607,.0170856536889082,.0114139620739561,.0050351690836213,.00487498195709245,.0110508422349653,.0165420966158804,.0209948152570779,.0241285368263833,.0257462347794558,.0257462347794558,.0241285368263833,.0209948152570779,.0165420966158804,.0110508422349653,.00487498195709245,.0035088724497807,.00795407987279395,.0119065275703486,.0151114669740713,.0173670300462221,.0185314027207393,.0185314027207393,.0173670300462221,.0151114669740713,.0119065275703486,.00795407987279395,.0035088724497807,.00127473997181245,.00288964152956687,.00432552816798997,.00548985215628924,.00630927675725883,.00673228227015115,.00673228227015115,.00630927675725883,.00548985215628924,.00432552816798997,.00288964152956687,.00127473997181245,-.00127473997181245,-.00288964152956687,-.00432552816798997,-.00548985215628924,-.00630927675725883,-.00673228227015115,-.00673228227015115,-.00630927675725883,-.00548985215628924,-.00432552816798997,-.00288964152956687,-.00127473997181245,-.0035088724497807,-.00795407987279395,-.0119065275703486,-.0151114669740713,-.0173670300462221,-.0185314027207393,-.0185314027207393,-.0173670300462221,-.0151114669740713,-.0119065275703486,-.00795407987279395,-.0035088724497807,-.00487498195709245,-.0110508422349653,-.0165420966158804,-.0209948152570779,-.0241285368263833,-.0257462347794558,-.0257462347794558,-.0241285368263833,-.0209948152570779,-.0165420966158804,-.0110508422349653,-.00487498195709245,-.0050351690836213,-.0114139620739561,-.0170856536889082,-.0216846842981607,-.0249213768851943,-.0265922308066324,-.0265922308066324,-.0249213768851943,-.0216846842981607,-.0170856536889082,-.0114139620739561,-.0050351690836213,-.00395009749714678,-.00895426991072092,-.0134037202629815,-.0170116664902499,-.0195508565501554,-.0208616438908705,-.0208616438908705,-.0195508565501554,-.0170116664902499,-.0134037202629815,-.00895426991072093,-.00395009749714678,-.0018918113337895,-.00428844840289335,-.00641941367947713,-.00814735927306806,-.00936344787276585,-.00999122031360907,-.00999122031360907,-.00936344787276585,-.00814735927306806,-.00641941367947713,-.00428844840289335,-.0018918113337895},
			{-.00235182533596431,-.00533123014211866,-.00798036224002773,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002773,-.00533123014211866,-.00235182533596431,-.00409570813081086,-.0092843470585992,-.013897815459963,-.0176387597554457,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554457,-.013897815459963,-.0092843470585992,-.00409570813081086,-.00328542799229539,-.00744756529085685,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480003,-.0173513258480003,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085686,-.00328542799229539,-.000186600113929722,-.000422994061970596,-.000633183289769704,-.000803620393549264,-.000923570130184837,-.000985490897277736,-.000985490897277736,-.000923570130184837,-.000803620393549264,-.000633183289769704,-.000422994061970597,-.000186600113929722,.00365824316969289,.00829268056399677,.0124133817294243,.0157547535948168,.0181063347139945,.019320274076138,.019320274076138,.0181063347139945,.0157547535948168,.0124133817294243,.00829268056399677,.00365824316969289,.00626131850458325,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686322,.0330678918746395,.0330678918746395,.0309901565686322,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458325,.00626131850458326,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686323,.0330678918746395,.0330678918746395,.0309901565686323,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458326,.00365824316969289,.00829268056399679,.0124133817294243,.0157547535948168,.0181063347139946,.019320274076138,.019320274076138,.0181063347139946,.0157547535948168,.0124133817294243,.00829268056399678,.00365824316969289,-.000186600113929713,-.000422994061970576,-.000633183289769673,-.000803620393549226,-.000923570130184793,-.000985490897277688,-.000985490897277688,-.000923570130184793,-.000803620393549226,-.000633183289769673,-.000422994061970576,-.000186600113929713,-.00328542799229539,-.00744756529085684,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480002,-.0173513258480002,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085684,-.00328542799229539,-.00409570813081086,-.00928434705859918,-.013897815459963,-.0176387597554456,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554456,-.013897815459963,-.00928434705859918,-.00409570813081086,-.00235182533596432,-.00533123014211866,-.00798036224002774,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002774,-.00533123014211866,-.00235182533596432},
			{.002625816508753,.00595232643558409,.00891008638917558,.0113084587769704,.0129963784252772,.0138677207253597,.0138677207253597,.0129963784252772,.0113084587769704,.00891008638917558,.00595232643558409,.002625816508753,.00327983286928533,.00743488199844395,.0111293360027156,.01412507487634,.0162334073987287,.0173217763333754,.0173217763333754,.0162334073987287,.01412507487634,.0111293360027156,.00743488199844395,.00327983286928533,-.000139403223161069,-.000316005892895126,-.000473031819685113,-.000600360153589864,-.000689970923659479,-.000736230029999612,-.000736230029999612,-.000689970923659479,-.000600360153589864,-.000473031819685113,-.000316005892895126,-.000139403223161069,-.00474831713981733,-.010763712369717,-.0161122895594359,-.0204493148918112,-.0235016141556679,-.0250772800730455,-.0250772800730455,-.0235016141556679,-.0204493148918112,-.0161122895594359,-.010763712369717,-.00474831713981733,-.0062268534463654,-.0141153291769463,-.0211293523195189,-.0268168454550162,-.0308195730805799,-.0328858716152058,-.0328858716152058,-.0308195730805799,-.0268168454550162,-.0211293523195189,-.0141153291769463,-.0062268534463654,-.00284444948144652,-.00644793411369711,-.00965196559807224,-.0122500333122624,-.0140784939652993,-.0150223867108016,-.0150223867108016,-.0140784939652993,-.0122500333122624,-.00965196559807224,-.00644793411369711,-.00284444948144652,.00284444948144652,.00644793411369712,.00965196559807225,.0122500333122624,.0140784939652993,.0150223867108016,.0150223867108016,.0140784939652993,.0122500333122624,.00965196559807225,.00644793411369712,.00284444948144652,.00622685344636543,.0141153291769464,.021129352319519,.0268168454550163,.0308195730805801,.0328858716152059,.0328858716152059,.0308195730805801,.0268168454550163,.021129352319519,.0141153291769464,.00622685344636543,.00474831713981732,.0107637123697169,.0161122895594359,.0204493148918111,.0235016141556678,.0250772800730455,.0250772800730455,.0235016141556678,.0204493148918111,.0161122895594359,.0107637123697169,.00474831713981732,.000139403223161063,.000316005892895114,.000473031819685095,.000600360153589841,.000689970923659453,.000736230029999584,.000736230029999584,.000689970923659453,.000600360153589841,.000473031819685095,.000316005892895114,.000139403223161063,-.00327983286928533,-.00743488199844395,-.0111293360027156,-.01412507487634,-.0162334073987287,-.0173217763333754,-.0173217763333754,-.0162334073987287,-.01412507487634,-.0111293360027156,-.00743488199844395,-.00327983286928533,-.00262581650875299,-.00595232643558406,-.00891008638917554,-.0113084587769704,-.0129963784252772,-.0138677207253597,-.0138677207253597,-.0129963784252772,-.0113084587769704,-.00891008638917554,-.00595232643558406,-.00262581650875299},
			{-.00274788724207344,-.00622904221162262,-.00932430450984375,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984375,-.00622904221162262,-.00274788724207344,-.00176296411460538,-.00399637136462942,-.00598220115906353,-.00759246007816475,-.00872572348723494,-.00931073969132328,-.00931073969132328,-.00872572348723494,-.00759246007816475,-.00598220115906353,-.00399637136462941,-.00176296411460538,.00351886839094271,.00797673915025254,.0119404464291104,.0151545159411986,.0174165045748593,.0185841942695623,.0185841942695623,.0174165045748593,.0151545159411986,.0119404464291104,.00797673915025254,.00351886839094271,.00572156044449334,.0129699068358449,.0194147601979758,.024640728021481,.028318644685133,.0302172684545957,.0302172684545957,.028318644685133,.024640728021481,.0194147601979758,.0129699068358449,.00572156044449334,.000863900578693703,.0019583311458107,.00293144199610321,.00372051285723621,.00427584288738108,.00456251680948201,.00456251680948201,.00427584288738108,.00372051285723621,.00293144199610321,.0019583311458107,.000863900578693703,-.00559347785892409,-.012679563105626,-.0189801422806282,-.0240891218318903,-.027684704824447,-.0295408260906547,-.0295408260906547,-.027684704824447,-.0240891218318903,-.0189801422806282,-.012679563105626,-.00559347785892409,-.00559347785892416,-.0126795631056261,-.0189801422806285,-.0240891218318906,-.0276847048244473,-.029540826090655,-.029540826090655,-.0276847048244473,-.0240891218318906,-.0189801422806285,-.0126795631056261,-.00559347785892416,.000863900578693674,.00195833114581063,.00293144199610311,.00372051285723608,.00427584288738094,.00456251680948185,.00456251680948185,.00427584288738094,.00372051285723608,.00293144199610311,.00195833114581063,.000863900578693674,.00572156044449339,.012969906835845,.0194147601979759,.0246407280214812,.0283186446851333,.030217268454596,.030217268454596,.0283186446851333,.0246407280214812,.0194147601979759,.012969906835845,.00572156044449339,.00351886839094272,.00797673915025255,.0119404464291104,.0151545159411987,.0174165045748594,.0185841942695623,.0185841942695623,.0174165045748594,.0151545159411987,.0119404464291104,.00797673915025255,.00351886839094272,-.00176296411460533,-.0039963713646293,-.00598220115906337,-.00759246007816454,-.00872572348723469,-.00931073969132302,-.00931073969132302,-.00872572348723469,-.00759246007816454,-.00598220115906337,-.00399637136462931,-.00176296411460533,-.00274788724207344,-.00622904221162261,-.00932430450984374,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984374,-.00622904221162261,-.00274788724207344},
			{.00273408778348797,.00619776093897419,.00927747931558112,.0117747446895576,.0135322629603741,.0144395336435885,.0144395336435885,.0135322629603741,.0117747446895576,.00927747931558112,.00619776093897419,.00273408778348797,-.000117317226705007,-.00026594029991732,-.00039808822185238,-.000505243613786289,-.000580657128546422,-.000619587290580092,-.000619587290580092,-.000580657128546422,-.000505243613786289,-.00039808822185238,-.00026594029991732,-.000117317226705007,-.00525141148444389,-.0119041506894287,-.0178194210583757,-.0226159634897508,-.0259916603811327,-.0277342714684962,-.0277342714684962,-.0259916603811327,-.0226159634897508,-.0178194210583757,-.0119041506894287,-.00525141148444389,-.00192519773143665,-.00436413028570642,-.00653270251221464,-.00829114262588265,-.00952869257156061,-.0101675438446148,-.0101675438446148,-.00952869257156061,-.00829114262588265,-.00653270251221464,-.00436413028570642,-.00192519773143665,.00561226510581592,.0127221509353366,.0190438923531344,.0241700318293333,.0277776915847029,.0296400471490286,.0296400471490286,.0277776915847029,.0241700318293333,.0190438923531344,.0127221509353366,.00561226510581592,.00424652299470629,.00962622140444603,.0144095699795955,.0182882657912371,.0210180032179986,.0224271554193144,.0224271554193144,.0210180032179986,.0182882657912371,.0144095699795955,.00962622140444603,.00424652299470629,-.00424652299470642,-.00962622140444635,-.014409569979596,-.0182882657912377,-.0210180032179993,-.0224271554193151,-.0224271554193151,-.0210180032179993,-.0182882657912377,-.014409569979596,-.00962622140444634,-.00424652299470642,-.00561226510581605,-.0127221509353369,-.0190438923531349,-.0241700318293339,-.0277776915847036,-.0296400471490293,-.0296400471490293,-.0277776915847035,-.0241700318293339,-.0190438923531349,-.0127221509353369,-.00561226510581605,.00192519773143709,.00436413028570742,.00653270251221612,.00829114262588453,.00952869257156278,.0101675438446171,.0101675438446171,.00952869257156278,.00829114262588453,.00653270251221612,.00436413028570742,.00192519773143709,.00525141148444307,.0119041506894268,.0178194210583729,.0226159634897473,.0259916603811286,.0277342714684918,.0277342714684918,.0259916603811286,.0226159634897473,.0178194210583729,.0119041506894268,.00525141148444307,.000117317226705136,.000265940299917613,.000398088221852818,.000505243613786846,.000580657128547061,.000619587290580775,.000619587290580775,.000580657128547061,.000505243613786846,.000398088221852818,.000265940299917613,.000117317226705136,-.00273408778348764,-.00619776093897344,-.00927747931557999,-.0117747446895562,-.0135322629603725,-.0144395336435867,-.0144395336435867,-.0135322629603725,-.0117747446895562,-.00927747931557999,-.00619776093897344,-.00273408778348764},
			{.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671589,-.00310382375628885,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.00519219719506989,.0103637745886009,.0118657346942092,.008761335787885,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00519219719506989,.0103637745886009,.0118657346942092,.00876133578788499,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671588,-.00310382375628885},
			{-.00260142552913706,-.00519252000263166,-.00594504099050645,-.00438965658114597,-.00068523266475808,.00441059027814434,.00971912899628547,.0139271476391209,.0159117625644931,.0150234628799625,.0112573018965134,.00527685077526242,-.00543177022292261,-.0108419692266112,-.012413231232108,-.00916559234130155,-.00143076414931682,.00920930185008349,.0202935178745168,.0290798506392448,.0332237218109789,.0313689542147427,.0235052191758168,.0110180516762225,-.00692384973169649,-.0138202027405243,-.0158230823854093,-.0116833336957715,-.0018237877459474,.0117390499831386,.0258680434782914,.0370679369308766,.0423501083267716,.0399858455548738,.0299619827063265,.0140446541387724,-.0067035767727117,-.0133805315937793,-.0153196923188487,-.0113116441612351,-.00176576639381999,.0113655879099196,.0250450864961337,.0358886704147581,.0410027966380503,.0387137497036944,.0290087825585011,.0135978423729273,-.00482504264831925,-.00963092357798896,-.0110266759528246,-.00814179763298104,-.00127094809920464,.00818062479896092,.0180267362591577,.0258316375293235,.0295126391755971,.0278650487239994,.0208796912105726,.00978733765555977,-.00175289208073003,-.00349882289141084,-.00400588644770123,-.00295783760558683,-.000461723350547117,.0029719431455217,.00654894174688664,.00938438810964969,.0107216609806272,.0101231070474515,.00758539313302242,.00355564663741111,.00175289208073003,.00349882289141084,.00400588644770123,.00295783760558683,.000461723350547117,-.0029719431455217,-.00654894174688664,-.00938438810964969,-.0107216609806272,-.0101231070474515,-.00758539313302242,-.00355564663741111,.00482504264831925,.00963092357798896,.0110266759528246,.00814179763298104,.00127094809920464,-.00818062479896092,-.0180267362591577,-.0258316375293235,-.0295126391755971,-.0278650487239994,-.0208796912105726,-.00978733765555977,.0067035767727117,.0133805315937793,.0153196923188487,.0113116441612351,.00176576639381999,-.0113655879099196,-.0250450864961337,-.0358886704147581,-.0410027966380503,-.0387137497036944,-.0290087825585011,-.0135978423729273,.00692384973169649,.0138202027405243,.0158230823854093,.0116833336957715,.0018237877459474,-.0117390499831386,-.0258680434782914,-.0370679369308766,-.0423501083267716,-.0399858455548738,-.0299619827063265,-.0140446541387724,.00543177022292261,.0108419692266112,.012413231232108,.00916559234130155,.00143076414931682,-.00920930185008349,-.0202935178745168,-.0290798506392448,-.0332237218109789,-.0313689542147427,-.0235052191758168,-.0110180516762225,.00260142552913706,.00519252000263166,.00594504099050645,.00438965658114597,.00068523266475808,-.00441059027814434,-.00971912899628547,-.0139271476391209,-.0159117625644931,-.0150234628799625,-.0112573018965134,-.00527685077526242},
			{.00323398975244044,.00645513634342751,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132602,-.00655997303832753,.00563199907811109,.0112416317670242,.0128708144833851,.00950345936923257,.00148350575212864,-.00954878012159535,-.0210415885190764,-.0301518078398598,-.0344484326345773,-.0325252935908768,-.0243716812928383,-.0114242050632393,.00451778467430605,.00901762784526014,.0103244989234578,.00762332921148292,.00119001431965068,-.00765968386950273,-.016878796465716,-.0241866828229713,-.02763327885035,-.0260906067053993,-.0195500756844557,-.0091640814984193,.000256593398763391,.000512167786738112,.000586393212665884,.000432976800198454,6.75883958332245e-005,-.00043504160981969,-.000958652982468414,-.00137371379952092,-.00156946765956192,-.00148184960837375,-.00111037172587836,-.000520485810578236,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163794,-.00132505164225417,.0085288693781531,.0187941242444344,.0269312757552921,.0307689755634805,.0290512481165313,.0217685278774507,.0102039790939625,-.00860992505308021,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.032167352592636,.0460946108327734,.052663080917845,.0497230798980886,.0372582357416281,.0174647665990918,-.00860992505308022,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.0321673525926361,.0460946108327734,.052663080917845,.0497230798980887,.0372582357416281,.0174647665990918,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163795,-.00132505164225417,.00852886937815311,.0187941242444344,.0269312757552922,.0307689755634806,.0290512481165313,.0217685278774507,.0102039790939625,.000256593398763379,.000512167786738088,.000586393212665856,.000432976800198433,6.75883958332213e-005,-.00043504160981967,-.000958652982468369,-.00137371379952085,-.00156946765956185,-.00148184960837368,-.00111037172587831,-.000520485810578211,.00451778467430605,.00901762784526012,.0103244989234578,.00762332921148291,.00119001431965068,-.00765968386950272,-.016878796465716,-.0241866828229712,-.0276332788503499,-.0260906067053993,-.0195500756844557,-.00916408149841928,.00563199907811108,.0112416317670242,.0128708144833851,.00950345936923255,.00148350575212863,-.00954878012159534,-.0210415885190763,-.0301518078398597,-.0344484326345772,-.0325252935908767,-.0243716812928383,-.0114242050632393,.00323398975244044,.00645513634342752,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132603,-.00655997303832753},
			{-.00361075440083214,-.00720716939267161,-.00825166151372371,-.00609280244273312,-.000951096555391269,.00612185821915875,.0134900605081838,.0193307511845923,.0220853782133051,.0208524265261428,.0156250301648324,.0073242197194988,-.00451009083357084,-.00900227071838211,-.0103069161796764,-.00761034659166953,-.00118798771118922,.00764663933727674,.016850051675698,.0241454926160516,.0275862190496304,.0260461740935889,.0195167816743332,.00914847495923774,.000191693059983865,.000382624848254669,.000438076387901722,.000323463690539438,5.049321798266e-005,-.000325006246491721,-.000716180246871757,-.00102625945578221,-.00117250116198259,-.00110704440267768,-.000829524667738866,-.000388838988800048,.00652940026539944,.013032870287285,.0149216465305082,.0110177379767064,.00171988715149418,-.0110702801253114,-.0243943494584096,-.0349561886252798,-.0399374364354906,-.0377078649444119,-.0282550583007305,-.0132445347624081,.00856253222944282,.0170910600271185,.0195679655313662,.0144484535618117,.00225542753807154,-.0145173563434682,-.031990289301271,-.0458408857714791,-.0523732001011417,-.0494493821432545,-.0370531499848233,-.0173686328234554,.00391139611196793,.00780725887484712,.00893871838932497,.00660010655391095,.00103028756760808,-.00663158159716375,-.0146132814266631,-.0209402846694332,-.0239242698021651,-.0225886590405194,-.0169260147469522,-.00793405514577347,-.00391139611196793,-.00780725887484712,-.00893871838932498,-.00660010655391096,-.00103028756760808,.00663158159716376,.0146132814266631,.0209402846694332,.0239242698021651,.0225886590405194,.0169260147469522,.00793405514577348,-.00856253222944286,-.0170910600271186,-.0195679655313663,-.0144484535618117,-.00225542753807155,.0145173563434683,.0319902893012711,.0458408857714793,.0523732001011419,.0494493821432547,.0370531499848235,.0173686328234555,-.00652940026539942,-.013032870287285,-.0149216465305081,-.0110177379767064,-.00171988715149417,.0110702801253114,.0243943494584096,.0349561886252797,.0399374364354905,.0377078649444118,.0282550583007304,.0132445347624081,-.000191693059983858,-.000382624848254655,-.000438076387901706,-.000323463690539425,-5.04932179826581e-005,.000325006246491708,.00071618024687173,.00102625945578217,.00117250116198255,.00110704440267764,.000829524667738834,.000388838988800033,.00451009083357083,.00900227071838211,.0103069161796764,.00761034659166953,.00118798771118921,-.00764663933727674,-.016850051675698,-.0241454926160516,-.0275862190496304,-.0260461740935889,-.0195167816743332,-.00914847495923773,.00361075440083212,.00720716939267157,.00825166151372367,.00609280244273308,.000951096555391264,-.00612185821915872,-.0134900605081837,-.0193307511845922,-.022085378213305,-.0208524265261427,-.0156250301648324,-.00732421971949876},
			{.00377861359217487,.00754222115521882,.00863527033203834,.00637604876123444,.00099531177496977,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.023112098204709,-.0218218282260771,-.016351417129166,-.0076647129967619,.00242424800551054,.00483886858146159,.00554013697585819,.00409068646883364,.000638562934915723,-.00411019441654455,-.00905717992717652,-.0129785994288253,-.0148280464803178,-.014000246985618,-.010490591164109,-.00491745576570479,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.00816499167123309,-.00127456872700186,.00820392944654196,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972578,.0209391724674719,.00981521944459481,-.00786770495273284,-.0157041648451191,-.0179800758935133,-.0132759990387789,-.00207240544459477,.0133393105384579,.0293943603166261,.0421210168157691,.0481232507843108,.0454366930684684,.0340463829902474,.0159592133291381,-.00118794775089702,-.00237117779829748,-.00271481846955098,-.00200454812347154,-.000312913283053141,.00201410755078885,.00443826559803358,.00635986828336871,.00726614786377036,.00686050349665649,.00514067626358997,.0024096876655053,.00769157887618114,.015352612142481,.0175775747521701,.0129788031427645,.0020260126728559,-.0130406973541611,-.0287363395359262,-.0411780976955557,-.0470459659340941,-.0444195493741242,-.0332842222466921,-.0156019511230183,.00769157887618123,.0153526121424812,.0175775747521703,.0129788031427646,.00202601267285592,-.0130406973541613,-.0287363395359265,-.0411780976955562,-.0470459659340946,-.0444195493741247,-.0332842222466924,-.0156019511230185,-.00118794775089698,-.0023711777982974,-.00271481846955089,-.00200454812347147,-.00031291328305313,.00201410755078878,.00443826559803343,.00635986828336849,.00726614786377011,.00686050349665626,.00514067626358979,.00240968766550522,-.00786770495273291,-.0157041648451192,-.0179800758935135,-.013275999038779,-.00207240544459478,.0133393105384581,.0293943603166264,.0421210168157695,.0481232507843112,.0454366930684688,.0340463829902477,.0159592133291382,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.0081649916712331,-.00127456872700187,.00820392944654197,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972579,.0209391724674719,.00981521944459482,.00242424800551048,.00483886858146146,.00554013697585804,.00409068646883352,.000638562934915706,-.00411019441654444,-.00905717992717627,-.0129785994288249,-.0148280464803174,-.0140002469856176,-.0104905911641087,-.00491745576570466,.00377861359217486,.00754222115521882,.00863527033203834,.00637604876123444,.000995311774969769,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.0231120982047089,-.0218218282260771,-.016351417129166,-.0076647129967619},
			{-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291795,-.0362874566153928,-.0189020194535391,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133056,.0295240165836947,.0169054630708873,-.00973006417443097,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133057,.0295240165836947,.0169054630708873,-.00973006417443096,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291794,-.0362874566153928,-.0189020194535391,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299},
			{.00303781657536634,.00474973406301401,.00235687098578853,-.00317458288582604,-.00858873504163554,-.010141337275906,-.00580693356275668,.00334222351587521,.013401456434816,.0194775827034515,.0181809574007362,.00947040224160834,.0063429536736534,.00991743325392784,.00492114092696054,-.0066285213996303,-.0179332580268801,-.0211750877428169,-.012124863256468,.00697855463035482,.0279822086740939,.0406691456503908,.0379617951494089,.0197741770113756,.00808533061765968,.0126417014630968,.00627295317887047,-.0084493423379166,-.0228594323180056,-.0269917760821381,-.0154555012012238,.00889552790747716,.0356687783299041,.0518407520280328,.048389706186527,.0252060423353226,.00782810670776326,.0122395227481066,.0060733876175387,-.00818053788513823,-.0221321902500056,-.0261330690598512,-.0149638052352352,.00861252866141587,.0345340241612611,.0501915082853594,.0468502528218308,.0244041460284054,.00563444710207634,.00880965807613161,.00437145053573306,-.00588811697401006,-.0159301169327542,-.0188098349613213,-.0107705186185175,.00619905156756988,.0248566019374114,.0361264107612158,.0337214707333665,.0175654056595871,.00204694101677262,.00320046494945897,.00158810638245006,-.00213909686741309,-.00578725989629442,-.00683343405369158,-.00391282692565031,.00225205644650943,.00903016695720288,.0131243812633644,.012250689435808,.00638134472992193,-.00204694101677262,-.00320046494945897,-.00158810638245006,.00213909686741309,.00578725989629442,.00683343405369158,.00391282692565031,-.00225205644650943,-.00903016695720288,-.0131243812633644,-.012250689435808,-.00638134472992193,-.00563444710207634,-.00880965807613161,-.00437145053573306,.00588811697401006,.0159301169327542,.0188098349613213,.0107705186185175,-.00619905156756988,-.0248566019374114,-.0361264107612158,-.0337214707333665,-.0175654056595871,-.00782810670776326,-.0122395227481066,-.0060733876175387,.00818053788513823,.0221321902500056,.0261330690598512,.0149638052352352,-.00861252866141587,-.0345340241612611,-.0501915082853594,-.0468502528218308,-.0244041460284054,-.00808533061765968,-.0126417014630968,-.00627295317887047,.0084493423379166,.0228594323180056,.0269917760821381,.0154555012012238,-.00889552790747716,-.0356687783299041,-.0518407520280328,-.048389706186527,-.0252060423353226,-.0063429536736534,-.00991743325392784,-.00492114092696054,.0066285213996303,.0179332580268801,.0211750877428169,.012124863256468,-.00697855463035482,-.0279822086740939,-.0406691456503908,-.0379617951494089,-.0197741770113756,-.00303781657536634,-.00474973406301401,-.00235687098578853,.00317458288582604,.00858873504163554,.010141337275906,.00580693356275668,-.00334222351587521,-.013401456434816,-.0194775827034515,-.0181809574007362,-.00947040224160835},
			{-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354764,-.0226018501305136,-.0117732310473289,-.0065767710679063,-.01028301505074,-.0051025466895586,.00687286554608826,.0185943234985903,.0219556552975931,.0125718165337835,-.00723580189137435,-.0290137040080097,-.042168313727554,-.0393611634058515,-.0205031034359056,-.00527564637794148,-.00824866040593754,-.00409307723852653,.00551316259147853,.0149156895386415,.0176120275664321,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.031574092667901,-.016446843331074,-.000299637130226379,-.000468493290714356,-.000232471592992847,.00031312716945722,.000847155796376566,.00100029778711831,.000572770891524226,-.00032966251846565,-.00132186188578481,-.00192118478526154,-.00179329125546946,-.000934119647899669,.00587430233536655,.0091846802486317,.00455754071797623,-.00613877079058787,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431711,.0259147334538654,.0376642916790926,.0351569747115048,.0183131550653389,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939193,.0644648579968841,.0601734237215266,.0313441429038559,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939194,.0644648579968841,.0601734237215266,.0313441429038559,.00587430233536656,.00918468024863171,.00455754071797623,-.00613877079058788,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431712,.0259147334538654,.0376642916790926,.0351569747115049,.0183131550653389,-.000299637130226364,-.000468493290714334,-.000232471592992836,.000313127169457205,.000847155796376526,.00100029778711826,.000572770891524199,-.000329662518465634,-.00132186188578475,-.00192118478526145,-.00179329125546937,-.000934119647899624,-.00527564637794147,-.00824866040593752,-.00409307723852653,.00551316259147852,.0149156895386415,.017612027566432,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.0315740926679009,-.016446843331074,-.00657677106790629,-.01028301505074,-.00510254668955859,.00687286554608825,.0185943234985903,.021955655297593,.0125718165337835,-.00723580189137434,-.0290137040080097,-.0421683137275539,-.0393611634058515,-.0205031034359056,-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354765,-.0226018501305136,-.0117732310473289},
			{.00421646110756717,.00659258663290769,.00327131497264031,-.00440629147242691,-.0119210842286909,-.0140760816664588,-.00805996969011775,.00463897510526447,.0186010966891058,.0270347033472955,.0252349995059512,.0131448300890066,.00526666188843086,.0082346128376237,.00408610667854559,-.0055037736089871,-.014890287939178,-.0175820340707359,-.0100674793638363,.00579441213021712,.0232341018964256,.0337682805916803,.0315203216066774,.0164188341582059,-.000223849711801565,-.000349996971430556,-.000173672398524789,.00023392770644707,.000632884117955103,.000747291803232149,.000427899569386332,-.000246280758644836,-.000987522480778167,-.00143525824110462,-.00133971290677435,-.000697852144734339,-.00762471196281051,-.0119215078435287,-.00591558507707979,.00796798605379006,.0215571378958162,.0254540633801011,.0145750063260015,-.00838875257661241,-.033636739632742,-.0488874012505932,-.0456329603679329,-.0237700623038804,-.00999890330935153,-.0156336403014995,-.00775758658064257,.010449066471062,.0282696236261719,.0333799781302062,.0191133881119304,-.0110008517447921,-.0441105852745744,-.0641100149794998,-.0598422026515798,-.0311716109136999,-.00456753568689175,-.00714150420140017,-.00354369398872959,.00477317186942322,.012913667706615,.0152480963782088,.00873106575768223,-.00502522940526383,-.0201498770593039,-.0292856898648238,-.0273361375479713,-.0142393061380111,.00456753568689175,.00714150420140018,.0035436939887296,-.00477317186942323,-.012913667706615,-.0152480963782088,-.00873106575768224,.00502522940526384,.0201498770593039,.0292856898648239,.0273361375479713,.0142393061380111,.00999890330935158,.0156336403014996,.0077575865806426,-.010449066471062,-.028269623626172,-.0333799781302063,-.0191133881119305,.0110008517447921,.0441105852745746,.0641100149795001,.0598422026515801,.0311716109137,.0076247119628105,.0119215078435287,.00591558507707978,-.00796798605379004,-.0215571378958161,-.025454063380101,-.0145750063260015,.0083887525766124,.033636739632742,.0488874012505931,.0456329603679328,.0237700623038804,.000223849711801557,.000349996971430543,.000173672398524782,-.000233927706447061,-.000632884117955079,-.000747291803232121,-.000427899569386316,.000246280758644826,.000987522480778129,.00143525824110457,.0013397129067743,.000697852144734312,-.00526666188843086,-.00823461283762369,-.00408610667854559,.0055037736089871,.014890287939178,.0175820340707359,.0100674793638363,-.00579441213021711,-.0232341018964256,-.0337682805916803,-.0315203216066774,-.0164188341582059,-.00421646110756715,-.00659258663290766,-.00327131497264029,.00440629147242689,.0119210842286909,.0140760816664587,.00805996969011771,-.00463897510526445,-.0186010966891057,-.0270347033472953,-.025234999505951,-.0131448300890066},
			{.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854622,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854621,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439},
			{-.00327473465962885,-.00326307824304169,.00235463328641995,.00792987603467538,.00656636699674288,-.002385380868931,-.0114262268213642,-.0111099792585027,.000926856258488525,.016554558282902,.0231216395735111,.0141071699093245,-.00683763806148434,-.00681329949179824,.00491646861608254,.0165575620113583,.0137105584327533,-.00498066949411644,-.0238579340109859,-.0231976098634047,.00193527362955397,.0345658777187952,.0482779275954164,.0294557367047126,-.00871590227761381,-.00868487802727276,.0062669974080828,.0211058396406552,.0174767787351192,-.00634883393906744,-.030411586503362,-.0295698746885344,.00246688340095815,.0440609476003131,.0615396274713053,.0375470769767045,-.00843861757917165,-.00840858032359331,.00606762132158154,.0204343857631653,.0169207785452416,-.00614685433349348,-.0294440828159479,-.0286291489294765,.00238840282624089,.0426592078630735,.0595818270621095,.0363525672650008,-.00607387534936655,-.00605225539271983,.00436730012098293,.0147080858685528,.012179091982045,-.00442432977464276,-.0211930078739329,-.0206064417927042,.00171910397817624,.0307048765549653,.0428852933869165,.0261655372015035,-.00220658111756891,-.00219872679304704,.00158659857626521,.00534330763908673,.00442454822513244,-.00160731690676576,-.00769921809541031,-.00748612421965592,.000624534116884189,.0111547894756412,.0155798190061378,.00950569068328115,.00220658111756891,.00219872679304704,-.00158659857626521,-.00534330763908673,-.00442454822513244,.00160731690676576,.00769921809541031,.00748612421965592,-.000624534116884189,-.0111547894756412,-.0155798190061378,-.00950569068328115,.00607387534936655,.00605225539271983,-.00436730012098293,-.0147080858685528,-.012179091982045,.00442432977464276,.0211930078739329,.0206064417927042,-.00171910397817624,-.0307048765549653,-.0428852933869165,-.0261655372015035,.00843861757917165,.00840858032359331,-.00606762132158154,-.0204343857631653,-.0169207785452416,.00614685433349347,.0294440828159479,.0286291489294765,-.00238840282624089,-.0426592078630735,-.0595818270621095,-.0363525672650008,.00871590227761381,.00868487802727276,-.0062669974080828,-.0211058396406552,-.0174767787351192,.00634883393906744,.030411586503362,.0295698746885344,-.00246688340095815,-.0440609476003131,-.0615396274713053,-.0375470769767045,.00683763806148434,.00681329949179824,-.00491646861608254,-.0165575620113583,-.0137105584327533,.00498066949411644,.0238579340109859,.0231976098634047,-.00193527362955397,-.0345658777187952,-.0482779275954164,-.0294557367047126,.00327473465962885,.00326307824304169,-.00235463328641995,-.00792987603467538,-.00656636699674288,.002385380868931,.0114262268213642,.0111099792585027,-.000926856258488525,-.016554558282902,-.0231216395735111,-.0141071699093245},
			{.00407102114440158,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589114,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092,.00708969078054082,.0070644550293116,-.00509770214609633,-.0171679158335966,-.0142159644664865,.00516426963168224,.0247373981014314,.024052732727339,-.00200661273467494,-.0358400638320779,-.0500575747208443,-.0305415500311089,.00568709190283437,.00566684870169624,-.00408919112208648,-.0137714772263088,-.0114035292808877,.004142589136187,.0198434403974258,.0192942267538859,-.00160963170168385,-.0287495947462082,-.0401543645107642,-.0244993198234636,.00032300570868889,.000321855970010235,-.000232250876004936,-.000782168784534644,-.000647678131435131,.000235283684983902,.00112703375255846,.00109584045637403,-9.1421105446067e-005,-.00163287025850404,-.00228061884480666,-.00139147042059047,-.00633243679598169,-.00630989648997895,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615883,-.0214836463932346,.00179228526457281,.0320119658255772,.0447108961296189,.0272793893570287,-.010838372915793,-.0107997937321446,.00779311800518549,.0262454710298765,.0217326719902005,-.00789488312515758,-.0378173257324053,-.0367706427561445,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548538,-.010838372915793,-.0107997937321446,.0077931180051855,.0262454710298765,.0217326719902006,-.00789488312515758,-.0378173257324053,-.0367706427561446,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548539,-.0063324367959817,-.00630989648997896,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615884,-.0214836463932346,.00179228526457282,.0320119658255772,.044710896129619,.0272793893570288,.000323005708688875,.00032185597001022,-.000232250876004925,-.000782168784534606,-.0006476781314351,.00023528368498389,.00112703375255841,.00109584045637398,-9.14211054460626e-005,-.00163287025850396,-.00228061884480655,-.00139147042059041,.00568709190283437,.00566684870169623,-.00408919112208647,-.0137714772263087,-.0114035292808876,.00414258913618699,.0198434403974258,.0192942267538859,-.00160963170168384,-.0287495947462082,-.0401543645107641,-.0244993198234635,.00708969078054081,.00706445502931158,-.00509770214609633,-.0171679158335965,-.0142159644664865,.00516426963168223,.0247373981014314,.0240527327273389,-.00200661273467494,-.0358400638320778,-.0500575747208442,-.0305415500311089,.00407102114440159,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589115,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092},
			{-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419407,-.00398827433495215,-.0145295292180157,-.0111084414324411,-.004451735343494,-.00147019434909273,.00789381658375618,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237627,-.0329362316722039,-.0251811462743288,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083196,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083195,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.004451735343494,-.00147019434909273,.00789381658375617,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237626,-.0329362316722039,-.0251811462743288,-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419408,-.00398827433495215,-.0145295292180157,-.0111084414324411},
			{.00333875638300519,.00110263085932663,-.0059202824228546,-.00595784387271232,.00390263111617153,.0106248423145106,.00253846386580488,-.0119075736105008,-.0122215351790305,.00678051660291391,.0247018398989999,.01888560445932,.00697131526529464,.00230229057164388,-.0123615354026379,-.0124399636192375,.00814868434650274,.0221846450958567,.00530030642192468,-.0248629849443481,-.0255185359412845,.0141577022933105,.0515773820591734,.0394330965061518,.00888629992584825,.00293471802056214,-.0157571859758933,-.0158571580226064,.0103870862740315,.0282786536784102,.00675627349671033,-.0316927197895115,-.0325283472936162,.0180467507280219,.0657454252068613,.0502651665035225,.00860359425560065,.00284135391717673,-.015255892314898,-.0153526838855177,.0100566351063325,.0273790063776703,.00654133175006641,-.0306844585711385,-.0314935017110425,.0174726176464323,.0636538230041544,.0486660478934077,.00619262083804183,.00204512520616675,-.0109807545364729,-.0110504223380172,.00723847804416364,.0197066249734953,.00470826332583461,-.0220857948325456,-.0226681209230377,.0125762899688114,.045816199491271,.0350284280425796,.00224972022366444,.000742974526710104,-.00398920363411266,-.00401451328347947,.00262967019464013,.00715922933157688,.00171046726410221,-.00802355910202843,-.00823511262949449,.0045688464741287,.0166445893043808,.0127254945897002,-.00224972022366444,-.000742974526710104,.00398920363411266,.00401451328347947,-.00262967019464013,-.00715922933157688,-.00171046726410221,.00802355910202843,.00823511262949449,-.0045688464741287,-.0166445893043808,-.0127254945897002,-.00619262083804183,-.00204512520616675,.0109807545364729,.0110504223380172,-.00723847804416364,-.0197066249734953,-.00470826332583461,.0220857948325456,.0226681209230377,-.0125762899688114,-.045816199491271,-.0350284280425796,-.00860359425560065,-.00284135391717673,.015255892314898,.0153526838855177,-.0100566351063325,-.0273790063776703,-.00654133175006641,.0306844585711385,.0314935017110425,-.0174726176464323,-.0636538230041544,-.0486660478934077,-.00888629992584825,-.00293471802056214,.0157571859758933,.0158571580226064,-.0103870862740315,-.0282786536784102,-.00675627349671033,.0316927197895115,.0325283472936162,-.0180467507280219,-.0657454252068613,-.0502651665035225,-.00697131526529464,-.00230229057164388,.0123615354026379,.0124399636192375,-.00814868434650274,-.0221846450958567,-.00530030642192469,.0248629849443481,.0255185359412845,-.0141577022933105,-.0515773820591734,-.0394330965061518,-.00333875638300519,-.00110263085932663,.0059202824228546,.00595784387271232,-.00390263111617154,-.0106248423145106,-.00253846386580488,.0119075736105008,.0122215351790305,-.00678051660291391,-.0247018398989999,-.01888560445932},
			{.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508331,-.0313789443876932,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508332,-.0313789443876932,.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847},
			{-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005},
			{.00459124870047564,.0100009097368206,.0139152296561375,.0158389513870659,.0156860656096203,.0137691389655327,.0107042527083733,.00724960390977444,.00411792158131722,.00180905830503188,.000503600637591741,4.27237565125139e-005,.00958651619965776,.0208818753803107,.0290549659195549,.0330716922483351,.0327524671142169,.0287499289103516,.022350453806482,.0151371554572844,.00859821031784408,.00377730937239729,.00105151691519805,8.92071004289135e-005,.0122198831859258,.0266180197827826,.037036216505918,.0421563170205638,.0417494023742321,.0366473873898747,.0284900091941002,.019295254668408,.0109600947313715,.00481491694443133,.00134036323562505,.000113711834820181,.0118311240516121,.0257712033140444,.0358579591324206,.040815170541386,.0404212013367775,.0354815000912993,.0275836379021353,.0186814021145725,.0106114140709319,.00466173684322829,.00129772138355337,.000110094245863453,.00851570438619621,.018549374357127,.0258095324275119,.0293775912826531,.029094023527884,.0255385679871328,.0198539128865321,.0134463384234132,.00763779206044802,.00335538472168774,.000934062700194595,7.92426863503829e-005,.00309367437106509,.00673881119463904,.00937635753650586,.0106725993661888,.010569581899212,.00927791873371822,.0072127376288433,.00488492679861124,.00277473717702488,.00121897934073137,.000339336090769426,2.87881139056347e-005,-.00309367437106509,-.00673881119463904,-.00937635753650586,-.0106725993661888,-.010569581899212,-.00927791873371822,-.0072127376288433,-.00488492679861124,-.00277473717702488,-.00121897934073137,-.000339336090769426,-2.87881139056347e-005,-.00851570438619621,-.018549374357127,-.0258095324275119,-.0293775912826531,-.029094023527884,-.0255385679871328,-.0198539128865321,-.0134463384234132,-.00763779206044802,-.00335538472168774,-.000934062700194595,-7.92426863503829e-005,-.0118311240516121,-.0257712033140444,-.0358579591324206,-.040815170541386,-.0404212013367775,-.0354815000912993,-.0275836379021353,-.0186814021145725,-.0106114140709319,-.00466173684322829,-.00129772138355337,-.000110094245863453,-.0122198831859258,-.0266180197827826,-.037036216505918,-.0421563170205638,-.0417494023742321,-.0366473873898747,-.0284900091941002,-.019295254668408,-.0109600947313715,-.00481491694443133,-.00134036323562505,-.000113711834820181,-.00958651619965776,-.0208818753803107,-.0290549659195549,-.0330716922483351,-.0327524671142169,-.0287499289103516,-.022350453806482,-.0151371554572845,-.00859821031784408,-.00377730937239729,-.00105151691519805,-8.92071004289135e-005,-.00459124870047563,-.0100009097368206,-.0139152296561375,-.0158389513870659,-.0156860656096203,-.0137691389655327,-.0107042527083733,-.00724960390977444,-.00411792158131722,-.00180905830503188,-.000503600637591741,-4.2723756512514e-005},
			{-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042617,-.0195002989275631,-.0171172512269981,-.0133071053509708,-.00901242203528271,-.00511923791436272,-.00224895003013865,-.000626056476974539,-5.31124913051166e-005,-.00993989955437502,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.033959806304287,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562087,-.00391655059726168,-.00109027850150277,-9.24955009028306e-005,-.00797342564302374,-.0173681530562416,-.0241659853793842,-.0275068307964137,-.0272413205925748,-.0239122759128171,-.0185896187731037,-.0125900776644864,-.00715141863890626,-.00314171433962441,-.000874581731371057,-7.41965242936831e-005,-.000452860977896186,-.000986446620240561,-.00137253826155738,-.00156228587949368,-.00154720588553185,-.00135812850566637,-.00105582133868279,-.000715069674966715,-.000406174031483403,-.000178437711946007,-4.96729957550138e-005,-4.2140871505504e-006,.00887821311745346,.0193390107581648,.0269082296615077,.0306281787690543,.0303325396949503,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930627,.00349822155622026,.000973825222354376,8.26160028005004e-005,.0151956328492542,.0330999609111817,.0460551659608014,.0524220981473091,.0519160928547411,.0455716503331479,.0354278116245859,.0239939777830022,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.0151956328492542,.0330999609111817,.0460551659608015,.0524220981473091,.0519160928547411,.045571650333148,.0354278116245859,.0239939777830023,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.00887821311745347,.0193390107581649,.0269082296615078,.0306281787690543,.0303325396949504,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930629,.00349822155622027,.000973825222354377,8.26160028005006e-005,-.000452860977896164,-.000986446620240513,-.00137253826155732,-.00156228587949361,-.00154720588553178,-.0013581285056663,-.00105582133868274,-.000715069674966681,-.000406174031483384,-.000178437711945999,-4.96729957550114e-005,-4.2140871505502e-006,-.00797342564302373,-.0173681530562415,-.0241659853793842,-.0275068307964136,-.0272413205925747,-.023912275912817,-.0185896187731037,-.0125900776644864,-.00715141863890625,-.0031417143396244,-.000874581731371055,-7.41965242936829e-005,-.009939899554375,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.0339598063042869,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562085,-.00391655059726168,-.00109027850150277,-9.24955009028305e-005,-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042618,-.0195002989275631,-.0171172512269982,-.0133071053509708,-.00901242203528272,-.00511923791436273,-.00224895003013865,-.00062605647697454,-5.31124913051166e-005},
			{.00637261042846754,.0138811695555554,.019314209141412,.0219843169843424,.0217721129493893,.0191114366236973,.0148574030483831,.0100623826962154,.00571563679614984,.00251095607588359,.000698992993903286,5.93001760646835e-005,.00795984680451423,.017338574885438,.0241248304194484,.027459986337079,.0271949283002686,.0238715530231142,.0185579604316863,.0125686366127877,.00713923968808145,.00313636396281389,.000873092308301808,7.40701667321458e-005,-.000338318549950809,-.000736944021350783,-.00102538124737378,-.00116713587426844,-.00115587007319604,-.00101461616061174,-.000788771746176309,-.000534206627103106,-.000303440163905314,-.000133305342938937,-3.71091719441577e-005,-3.14821528841767e-006,-.0115237204206783,-.0251015998648952,-.0349262753137827,-.0397546853696583,-.0393709525182014,-.0345595976658438,-.0268669426491322,-.0181960102941585,-.0103357017040835,-.00454061269426197,-.0012640031786301,-.000107233708624919,-.0151119893856147,-.0329177641310666,-.0458016580195613,-.0521335438038479,-.0516303237875308,-.0453208038751262,-.0352328013276903,-.0238619043492618,-.013554044071107,-.00595447375803616,-.00165758990339312,-.000140624260860408,-.00690321215069344,-.0150369553289623,-.0209223652885514,-.0238147939270004,-.0235849211787824,-.0207027093526006,-.0160944728071212,-.010900205382576,-.00619153701972338,-.00272002544129556,-.000757193144462012,-6.42376778783298e-005,.00690321215069345,.0150369553289623,.0209223652885514,.0238147939270005,.0235849211787825,.0207027093526006,.0160944728071212,.010900205382576,.00619153701972339,.00272002544129557,.000757193144462013,6.42376778783299e-005,.0151119893856147,.0329177641310668,.0458016580195615,.0521335438038482,.051630323787531,.0453208038751264,.0352328013276905,.0238619043492619,.0135540440711071,.00595447375803619,.00165758990339312,.000140624260860409,.0115237204206782,.0251015998648951,.0349262753137827,.0397546853696582,.0393709525182013,.0345595976658438,.0268669426491322,.0181960102941585,.0103357017040834,.00454061269426196,.0012640031786301,.000107233708624919,.000338318549950797,.000736944021350755,.00102538124737374,.0011671358742684,.00115587007319599,.0010146161606117,.000788771746176279,.000534206627103086,.000303440163905302,.000133305342938932,3.71091719441563e-005,3.14821528841755e-006,-.00795984680451422,-.017338574885438,-.0241248304194484,-.027459986337079,-.0271949283002686,-.0238715530231142,-.0185579604316863,-.0125686366127877,-.00713923968808145,-.00313636396281389,-.000873092308301808,-7.40701667321458e-005,-.00637261042846751,-.0138811695555553,-.0193142091414119,-.0219843169843423,-.0217721129493892,-.0191114366236972,-.014857403048383,-.0100623826962154,-.00571563679614981,-.00251095607588358,-.000698992993903282,-5.93001760646832e-005},
			{-.00666886465162736,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907023,-6.20569627520987e-005,-.00427854858305904,-.00931976919031684,-.0129674931619453,-.0147601943251846,-.0146177212706554,-.0128313523954431,-.00997520897819624,-.00675584891158071,-.00383745876041692,-.0016858472177435,-.000469301476561824,-3.98139329439154e-005,.00853996360068238,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172028,.0033649434139527,.000936723622457736,7.94684298987088e-005,.0138856906557926,.0302465729786341,.0420849723059915,.0479030419872811,.0474406569696583,.0416431382277769,.0323737509132328,.0219255727222366,.0124541685613893,.00547128366175477,.00152308078343358,.000129212966948341,.00209660569165992,.00456694151066636,.00635442590918465,.00723289845409743,.00716308276514704,.00628771321435967,.0048881249127316,.00331055053017243,.00188045962840229,.000826111192469611,.000229970544394388,1.95099148220722e-005,-.0135748462316425,-.029569474604983,-.0411428600764948,-.0468306867209362,-.0463786526327775,-.0407109168753721,-.0316490336334332,-.0214347478726881,-.0121753701241565,-.00534880376058466,-.00148898516796897,-.000126320411489674,-.0135748462316426,-.0295694746049834,-.0411428600764953,-.0468306867209367,-.046378652632778,-.0407109168753726,-.0316490336334336,-.0214347478726884,-.0121753701241566,-.00534880376058472,-.00148898516796898,-.000126320411489675,.00209660569165985,.0045669415106662,.00635442590918444,.00723289845409719,.00716308276514679,.00628771321435945,.00488812491273143,.00331055053017232,.00188045962840222,.000826111192469583,.00022997054439438,1.95099148220715e-005,.0138856906557927,.0302465729786343,.0420849723059919,.0479030419872815,.0474406569696587,.0416431382277773,.0323737509132331,.0219255727222367,.0124541685613895,.00547128366175482,.0015230807834336,.000129212966948343,.00853996360068239,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172029,.00336494341395271,.000936723622457737,7.94684298987089e-005,-.00427854858305892,-.00931976919031658,-.012967493161945,-.0147601943251842,-.014617721270655,-.0128313523954428,-.00997520897819597,-.00675584891158053,-.00383745876041681,-.00168584721774346,-.000469301476561811,-3.98139329439144e-005,-.00666886465162735,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907022,-6.20569627520987e-005},
			{.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831813,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932927,-.0625431691900999,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932926,-.0625431691900998,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831814,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342},
			{-.00536389192389492,-.00931249836832228,-.00723896090430979,.000615037191458855,.0111507213412938,.0200157824556806,.0237694861813762,.0214832345431866,.0149703083366938,.0075880103734666,.00231928224558504,.000206890674868773,-.0111997933843804,-.0194444741052602,-.015114932887575,.00128419617057589,.0232826792338887,.0417929054334019,.0496306297482205,.0448569418485149,.0312579676567552,.0158437473362906,.00484265572818485,.000431987229526937,-.0142763193649554,-.0247857717265967,-.0192669276724185,.00163695829281661,.0296783121801492,.0532732028778971,.0632639099894119,.057178914430488,.0398443715568573,.0201959437239261,.00617290849726889,.000550651912820983,-.013822137481785,-.0239972457633677,-.0186539763038293,.00158488066824849,.0287341366282638,.0515783876396646,.0612512538529679,.0553598442367079,.0385767765104859,.0195534369602114,.00597652572281162,.000533133663449908,-.0099487788537054,-.0172725304977166,-.0134265981099973,.00114075173240722,.0206820089326614,.0371246468163075,.0440868989979581,.0398464310178064,.0277664593408513,.0140740041403415,.0043017321169096,.000383734348187365,-.0036142966884799,-.00627494597051114,-.00487775533058912,.000414424249392327,.00751357704253213,.0134870309233169,.016016350890574,.0144758292241828,.0100872904626957,.00511295178092569,.00156277834430738,.000139407037215731,.0036142966884799,.00627494597051114,.00487775533058912,-.000414424249392327,-.00751357704253213,-.0134870309233169,-.016016350890574,-.0144758292241828,-.0100872904626957,-.00511295178092569,-.00156277834430738,-.000139407037215731,.0099487788537054,.0172725304977166,.0134265981099973,-.00114075173240722,-.0206820089326614,-.0371246468163075,-.0440868989979581,-.0398464310178064,-.0277664593408513,-.0140740041403415,-.0043017321169096,-.000383734348187365,.013822137481785,.0239972457633677,.0186539763038293,-.00158488066824849,-.0287341366282638,-.0515783876396646,-.0612512538529679,-.0553598442367079,-.0385767765104859,-.0195534369602114,-.00597652572281162,-.000533133663449908,.0142763193649554,.0247857717265967,.0192669276724185,-.00163695829281661,-.0296783121801492,-.0532732028778971,-.0632639099894119,-.057178914430488,-.0398443715568573,-.0201959437239261,-.00617290849726889,-.000550651912820983,.0111997933843804,.0194444741052602,.015114932887575,-.00128419617057589,-.0232826792338887,-.0417929054334019,-.0496306297482205,-.0448569418485149,-.0312579676567552,-.0158437473362906,-.00484265572818485,-.000431987229526937,.00536389192389492,.00931249836832228,.00723896090430979,-.000615037191458855,-.0111507213412938,-.0200157824556806,-.0237694861813762,-.0214832345431866,-.0149703083366938,-.0075880103734666,-.00231928224558504,-.000206890674868773},
			{.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065728,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700374,-.043333498156298,-.0514601409111785,-.0465104827417968,-.032410215795647,-.0164277881345211,-.00502116832732015,-.000447911376826381,.00931524214857705,.0161726184158759,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248514,-.0412794520598759,-.0373090164275984,-.0259982954865174,-.0131777737243128,-.00402779847824231,-.000359298203996196,.000529071174374757,.000918544690683033,.000714019894739991,-6.06646169956767e-005,-.00109985907972585,-.00197427048868781,-.00234451964109179,-.00211901363606869,-.00147660667381522,-.000748448629543353,-.000228764001733872,-2.0406804214751e-005,-.0103722927557633,-.0180078123747257,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831512,.0459636534116219,.0415426711023014,.0289484618473594,.0146731265551326,.004484854425811,.000400069702861968,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191293,.000684745031277692,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191294,.000684745031277693,-.0103722927557633,-.0180078123747258,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831513,.045963653411622,.0415426711023014,.0289484618473595,.0146731265551327,.00448485442581101,.000400069702861968,.000529071174374731,.000918544690682989,.000714019894739956,-6.06646169956738e-005,-.0010998590797258,-.00197427048868771,-.00234451964109168,-.00211901363606859,-.00147660667381515,-.000748448629543317,-.000228764001733861,-2.04068042147501e-005,.00931524214857704,.0161726184158758,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248513,-.0412794520598758,-.0373090164275984,-.0259982954865174,-.0131777737243127,-.0040277984782423,-.000359298203996196,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700373,-.0433334981562979,-.0514601409111784,-.0465104827417967,-.0324102157956469,-.016427788134521,-.00502116832732014,-.00044791137682638,.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065729,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687},
			{-.00744503202371008,-.0129256609858319,-.0100476103015583,.000853665892443807,.015477097348583,.0277817195939975,.0329918253981113,.0298185294215917,.0207786485173192,.0105320877132218,.00321914215189193,.000287162329453971,-.00929937817926298,-.0161450762523781,-.0125501848339245,.00106628983559657,.0193320029925161,.0347013573819497,.0412091526569903,.0372454787243047,.0259540200768997,.0131553318173118,.0040209390890345,.000358686314837344,.000395252850754388,.000686216571838048,.000533422368407895,-4.53206751167283e-005,-.000821670992004966,-.00147491694238737,-.00175151873070067,-.00158305010934183,-.00110312756683095,-.000559143020444886,-.000170902570797115,-1.52452976675568e-005,.013463002097378,.0233737343785841,.0181692970739038,-.00154370131167062,-.0279875483949175,-.0502382458492594,-.0596597856283979,-.0539214502859268,-.0375744506777747,-.0190453873833324,-.00582123990933241,-.000519281452573281,.0176551267617537,.0306518739553514,.0238268731365797,-.0020243807542107,-.0367023425450716,-.0658814870815153,-.0782367164641202,-.0707115718388728,-.0492744251929199,-.0249757614273064,-.00763386410894635,-.000680975892590434,.00806492662705201,.0140018883901234,.0108842029962012,-.000924744549741412,-.0167657646224416,-.0300949048150996,-.0357388188903564,-.032301305238965,-.0225087380642283,-.0114090193791377,-.00348717711010207,-.000311072284138478,-.00806492662705202,-.0140018883901234,-.0108842029962012,.000924744549741413,.0167657646224416,.0300949048150996,.0357388188903565,.032301305238965,.0225087380642284,.0114090193791378,.00348717711010208,.000311072284138478,-.0176551267617538,-.0306518739553516,-.0238268731365798,.00202438075421071,.0367023425450718,.0658814870815156,.0782367164641205,.0707115718388731,.0492744251929201,.0249757614273065,.00763386410894639,.000680975892590437,-.013463002097378,-.023373734378584,-.0181692970739037,.00154370131167062,.0279875483949174,.0502382458492593,.0596597856283978,.0539214502859267,.0375744506777747,.0190453873833324,.0058212399093324,.00051928145257328,-.000395252850754372,-.000686216571838022,-.000533422368407874,4.53206751167266e-005,.000821670992004935,.00147491694238732,.0017515187307006,.00158305010934177,.00110312756683091,.000559143020444865,.000170902570797108,1.52452976675562e-005,.00929937817926298,.0161450762523781,.0125501848339245,-.00106628983559657,-.0193320029925161,-.0347013573819496,-.0412091526569903,-.0372454787243047,-.0259540200768997,-.0131553318173118,-.00402093908903449,-.000358686314837344,.00744503202371004,.0129256609858318,.0100476103015582,-.000853665892443803,-.0154770973485829,-.0277817195939974,-.0329918253981111,-.0298185294215915,-.020778648517319,-.0105320877132217,-.00321914215189191,-.00028716232945397},
			{-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487695,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487696,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613},
			{.00578623505122275,.00668940613303525,-.00201954593586876,-.013036228022183,-.015415647248866,-.00450228521335004,.0135620449344063,.0268160677531348,.0273936022556361,.0174648825623257,.00613204972269062,.000587601273095628,.0120816448143678,.0139674638522171,-.00421680703729971,-.0272196126304688,-.0321878342299675,-.00940076065336832,.0283175170734496,.0559918847131752,.057197775359769,.0364666326159833,.0128037050133204,.00122691003928565,.0154004108740453,.0178042547598082,-.00537514237082621,-.0346967010520833,-.0410296677235249,-.0119831015407985,.0360961942322037,.0713725691695229,.0729097118112787,.0464838301496112,.0163208173179129,.00156393595415252,.0149104675326585,.0172378363610222,-.0052041394518047,-.033592872213516,-.0397243640751441,-.0116018752957915,.0349478424019757,.0691019469563708,.0705901874741346,.0450050096655201,.0158015924844782,.00151418143699325,.0107321276671759,.0124073011209858,-.00374579059122209,-.0241791877090991,-.0285924600162962,-.00835069769478169,.0251544564601237,.0497376031410035,.0508087960597447,.0323933175358136,.011373533896014,.00108986444975558,.00389887985834265,.00450745443375195,-.00136081007816862,-.00878406881409194,-.0103873686481536,-.00303372901021968,.00913837467102563,.0180691979356344,.018458352129919,.0117681840173745,.00413189663788404,.000395937383830426,-.00389887985834265,-.00450745443375195,.00136081007816862,.00878406881409194,.0103873686481536,.00303372901021968,-.00913837467102562,-.0180691979356344,-.018458352129919,-.0117681840173745,-.00413189663788404,-.000395937383830426,-.0107321276671759,-.0124073011209858,.00374579059122209,.0241791877090991,.0285924600162962,.00835069769478169,-.0251544564601236,-.0497376031410035,-.0508087960597447,-.0323933175358136,-.011373533896014,-.00108986444975558,-.0149104675326585,-.0172378363610222,.0052041394518047,.033592872213516,.0397243640751441,.0116018752957915,-.0349478424019757,-.0691019469563708,-.0705901874741346,-.0450050096655201,-.0158015924844782,-.00151418143699325,-.0154004108740453,-.0178042547598082,.00537514237082622,.0346967010520833,.0410296677235249,.0119831015407985,-.0360961942322037,-.0713725691695229,-.0729097118112787,-.0464838301496112,-.0163208173179129,-.00156393595415252,-.0120816448143678,-.0139674638522171,.00421680703729971,.0272196126304688,.0321878342299675,.00940076065336832,-.0283175170734496,-.0559918847131752,-.057197775359769,-.0364666326159833,-.0128037050133204,-.00122691003928565,-.00578623505122275,-.00668940613303525,.00201954593586876,.013036228022183,.015415647248866,.00450228521335004,-.0135620449344063,-.0268160677531348,-.0273936022556361,-.0174648825623257,-.00613204972269062,-.000587601273095628},
			{-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137491,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099975,.00974729658452533,-.0293613727260863,-.0580558878956585,-.0593062307364416,-.0378108853797437,-.0132756821226898,-.00127213706171361,-.010048707428968,-.0116172060950429,.00350725922283749,.0226394607568784,.0267716965627408,.00781892655072194,-.0235526245439282,-.0465703202274322,-.047573299745961,-.0303305160603543,-.0106492690078639,-.00102046204932158,-.000570729279560872,-.000659814181276736,.000199199304351608,.0012858373297019,.00152053298395657,.000444086002978713,-.00133770164299672,-.00264502131246258,-.00270198682589509,-.00172266072051549,-.000604838947861804,-5.79584562835985e-005,.0111889882847748,.0129354729270487,-.00390524678257324,-.0252084820831874,-.0298096249016592,-.00870618218253045,.0262252674745031,.0518549048349062,.052971697831619,.0337722827805856,.0118576988147661,.00113625936426423,.0191506731907347,.0221398940001207,-.00668408107674887,-.043145938642774,-.0510210905490773,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0191506731907347,.0221398940001207,-.00668408107674888,-.043145938642774,-.0510210905490774,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0111889882847749,.0129354729270488,-.00390524678257325,-.0252084820831874,-.0298096249016593,-.00870618218253046,.0262252674745032,.0518549048349063,.0529716978316191,.0337722827805856,.0118576988147661,.00113625936426423,-.000570729279560845,-.000659814181276704,.000199199304351599,.00128583732970183,.0015205329839565,.000444086002978692,-.00133770164299666,-.00264502131246246,-.00270198682589496,-.00172266072051541,-.000604838947861776,-5.79584562835958e-005,-.010048707428968,-.0116172060950428,.00350725922283748,.0226394607568783,.0267716965627407,.00781892655072193,-.0235526245439282,-.0465703202274321,-.0475732997459609,-.0303305160603542,-.0106492690078639,-.00102046204932158,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099974,.00974729658452531,-.0293613727260862,-.0580558878956584,-.0593062307364415,-.0378108853797436,-.0132756821226898,-.00127213706171361,-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137492,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872},
			{.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503008,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503009},
			{-.00590524039988117,-.00279948660884738,.00950696592933597,.0123888349467185,-.00252754986470693,-.018634756913296,-.0137887462421145,.010811667317975,.0309214103019088,.0286485035369933,.0122170775898605,.00128660900891531,-.0123301276949928,-.00584532127908038,.019850521903057,.025867857452222,-.00527751801398726,-.0389093274357997,-.0287908688565067,.0225747352519459,.0645638300414449,.0598180062049106,.0255092285058617,.00268643650371509,-.0157171507315389,-.00745100114858395,.025303358778414,.0329736256376032,-.00672722522962469,-.0495975207474624,-.0366995733299215,.0287758996059307,.0822991840529711,.0762497066729088,.0325164831532697,.00342438687611848,-.0152171307379511,-.00721395757687207,.0244983664798579,.031924614123907,-.00651320761449607,-.0480196423883553,-.0355320258059113,.027860433095352,.0796809462956736,.0738239248952259,.0314820150123508,.00331544461723926,-.0109528550630685,-.00519240013976796,.0176332228431619,.0229784232957589,-.00468802037820009,-.0345631638655983,-.0255749349500402,.0200531421424059,.0573520640056069,.0531363476787172,.0226598531261651,.00238636211961347,-.00397906802090637,-.001886349561729,.00640598206739025,.00834784252877614,-.00170311319385134,-.0125564685414928,-.00929113050528727,.0072851157217069,.0208354591112203,.0193039294849313,.00823210901756414,.000866942650275162,.00397906802090637,.001886349561729,-.00640598206739025,-.00834784252877614,.00170311319385134,.0125564685414928,.00929113050528727,-.0072851157217069,-.0208354591112203,-.0193039294849313,-.00823210901756414,-.000866942650275162,.0109528550630685,.00519240013976796,-.0176332228431619,-.0229784232957589,.00468802037820009,.0345631638655983,.0255749349500402,-.0200531421424059,-.0573520640056069,-.0531363476787172,-.0226598531261651,-.00238636211961347,.0152171307379511,.00721395757687207,-.0244983664798579,-.031924614123907,.00651320761449607,.0480196423883553,.0355320258059113,-.027860433095352,-.0796809462956736,-.0738239248952259,-.0314820150123508,-.00331544461723926,.0157171507315389,.00745100114858395,-.025303358778414,-.0329736256376032,.00672722522962468,.0495975207474624,.0366995733299215,-.0287758996059307,-.0822991840529711,-.0762497066729088,-.0325164831532697,-.00342438687611848,.0123301276949928,.00584532127908038,-.019850521903057,-.025867857452222,.00527751801398726,.0389093274357997,.0287908688565067,-.0225747352519459,-.0645638300414449,-.0598180062049106,-.0255092285058617,-.00268643650371509,.00590524039988117,.00279948660884738,-.00950696592933597,-.0123888349467185,.00252754986470693,.018634756913296,.0137887462421145,-.010811667317975,-.0309214103019089,-.0286485035369933,-.0122170775898605,-.00128660900891531},
			{-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599055,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599056},
			{-.00423058507690693,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394862,-.00435422423400533,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415894,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415895,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.00423058507690692,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040534,-.00720462946394862,-.00435422423400534,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007},
			{.00719247221789664,.015054738114976,.0194706196180468,.0198760629317139,.0169623740708349,.0122486834132115,.00740267274248333,.00362315688940088,.00134348757825736,.000329082012180178,3.81739176729443e-005,6.2280977297559e-007,.0150178646226051,.0314342568299276,.0406546067448804,.0415011714047538,.0354173960992067,.0255752213886969,.0154567628103504,.00756514283587924,.00280519881915107,.000687122446765829,7.9707047918831e-005,1.30042530201362e-006,.0191431952513503,.0400690864645718,.051822219359513,.0529013310031521,.0451463737261948,.0326006039436125,.0197026565272781,.00964324889396088,.00357577259239878,.000875871469784373,.000101602166457287,1.65764548368426e-006,.0185341802631805,.0387943424159014,.0501735652087417,.051218346368036,.0437101026178578,.0315634596129842,.0190758430317083,.00933646191122561,.00346201420073717,.000848006797988197,9.83698302986383e-005,1.60490972398582e-006,.0133403723495074,.0279230570510056,.0361134958483479,.0368654994163948,.0314612805140234,.0227184746180725,.0137302457033551,.00672011799573213,.00249185870976258,.000610371016113917,7.08037876780903e-005,1.15516807332712e-006,.00484643033229705,.0101441809206661,.0131196893984132,.0133928851388839,.0114295838362937,.00825340564775263,.00498806761185259,.00244135492157639,.000905268557607982,.000221741982078836,2.57223423195746e-005,4.19661568867713e-007,-.00484643033229705,-.0101441809206661,-.0131196893984132,-.0133928851388839,-.0114295838362937,-.00825340564775263,-.00498806761185259,-.00244135492157639,-.000905268557607982,-.000221741982078836,-2.57223423195746e-005,-4.19661568867713e-007,-.0133403723495074,-.0279230570510056,-.0361134958483479,-.0368654994163948,-.0314612805140234,-.0227184746180725,-.0137302457033551,-.00672011799573213,-.00249185870976258,-.000610371016113917,-7.08037876780903e-005,-1.15516807332712e-006,-.0185341802631805,-.0387943424159014,-.0501735652087417,-.051218346368036,-.0437101026178578,-.0315634596129842,-.0190758430317083,-.00933646191122561,-.00346201420073717,-.000848006797988197,-9.83698302986383e-005,-1.60490972398582e-006,-.0191431952513503,-.0400690864645718,-.051822219359513,-.0529013310031521,-.0451463737261948,-.0326006039436125,-.0197026565272781,-.00964324889396088,-.00357577259239877,-.000875871469784373,-.000101602166457287,-1.65764548368426e-006,-.0150178646226051,-.0314342568299276,-.0406546067448804,-.0415011714047538,-.0354173960992067,-.025575221388697,-.0154567628103504,-.00756514283587924,-.00280519881915107,-.000687122446765829,-7.97070479188311e-005,-1.30042530201362e-006,-.00719247221789664,-.015054738114976,-.0194706196180468,-.0198760629317139,-.0169623740708349,-.0122486834132115,-.00740267274248333,-.00362315688940088,-.00134348757825736,-.000329082012180178,-3.81739176729443e-005,-6.2280977297559e-007},
			{-.00894139816848509,-.0187154574574963,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007,-.0155714612898606,-.0325930034464412,-.0421532388985272,-.0430310100838354,-.0367229713548178,-.0265179890644552,-.0160265383805396,-.00784401323233701,-.00290860557878677,-.000712451526903725,-8.26452523302774e-005,-1.34836228448754e-006,-.0124908594969934,-.0261449210871181,-.0338137940058043,-.0345179101026881,-.029457831026034,-.0212717656603367,-.0128559057758339,-.0062921819188208,-.0023331775316716,-.000571502684000991,-6.62950134052773e-005,-1.08160714868461e-006,-.000709434451366774,-.00148493446363173,-.00192049797733382,-.00196048915784558,-.00167309545011213,-.00120815732532064,-.00073016772489491,-.000357372575406282,-.000132515822671862,-3.24592309421519e-005,-3.76531066375748e-006,-6.1431270947059e-008,.0139082644774524,.0291117258431486,.0376508270012981,.038434842937163,.0328005694836934,.0236855872710412,.014314734520114,.0070061896291659,.002597935729815,.000636354166066229,7.38178649914994e-005,1.20434292678016e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102608,.0245005889388461,.0119915442295211,.0044465341160397,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102609,.0245005889388461,.0119915442295211,.00444653411603971,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0139082644774524,.0291117258431486,.0376508270012982,.038434842937163,.0328005694836935,.0236855872710413,.0143147345201141,.00700618962916591,.002597935729815,.00063635416606623,7.38178649914996e-005,1.20434292678016e-006,-.00070943445136674,-.00148493446363166,-.00192049797733373,-.00196048915784549,-.00167309545011205,-.00120815732532058,-.000730167724894874,-.000357372575406265,-.000132515822671856,-3.24592309421504e-005,-3.7653106637573e-006,-6.14312709470561e-008,-.0124908594969934,-.0261449210871181,-.0338137940058042,-.0345179101026881,-.029457831026034,-.0212717656603366,-.0128559057758339,-.00629218191882079,-.0023331775316716,-.00057150268400099,-6.62950134052772e-005,-1.08160714868461e-006,-.0155714612898606,-.0325930034464411,-.0421532388985272,-.0430310100838353,-.0367229713548177,-.0265179890644552,-.0160265383805396,-.007844013232337,-.00290860557878677,-.000712451526903724,-8.26452523302773e-005,-1.34836228448754e-006,-.00894139816848509,-.0187154574574964,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007},
			{.00998308444276627,.0208958362733153,.0270250386669949,.0275877902097846,.0235436172110032,.0170010584848057,.0102748408129391,.00502890801389247,.00186474824444619,.000456762767608005,5.29850421517036e-005,8.64454163607971e-007,.0124695874152247,.0261003959766295,.0337562087138414,.0344591256926668,.0294076639906539,.0212355395912979,.0128340120159257,.00628146625843626,.00232920410268166,.00057052940815652,6.61821122117002e-005,1.07976515889088e-006,-.00052999672436047,-.00110934900341893,-.0014347451482894,-.00146462133295129,-.00124991830660814,-.000902577290542385,-.000545485917243634,-.000266982092533976,-9.89985076235179e-005,-2.42492960998119e-005,-2.8129481365704e-006,-4.5893418782406e-008,-.0180526136574343,-.0377863636688457,-.0488699244135907,-.0498875594186185,-.0425743995300703,-.0307433581628203,-.0185802026068238,-.00909387614760269,-.00337206199330092,-.000825973357633232,-9.58139241505331e-005,-1.5632099607735e-006,-.0236738567072675,-.0495523238883786,-.0640870961854267,-.0654216035175219,-.0558312637160867,-.0403162594435231,-.0243657268943327,-.011925537482675,-.00442205843166394,-.00108316586637731,-.000125648571112395,-2.04996402831038e-006,-.0108143045303454,-.0226356832070683,-.0292752204756728,-.029884828317171,-.0255039259384566,-.0184166151099846,-.0111303533681376,-.00544762924015274,-.00202001250249669,-.000494794138560608,-5.7396727901756e-005,-9.36431083136384e-007,.0108143045303454,.0226356832070683,.0292752204756728,.0298848283171711,.0255039259384567,.0184166151099846,.0111303533681376,.00544762924015275,.00202001250249669,.000494794138560608,5.73967279017561e-005,9.36431083136386e-007,.0236738567072676,.0495523238883789,.064087096185427,.0654216035175222,.055831263716087,.0403162594435233,.0243657268943328,.0119255374826751,.00442205843166396,.00108316586637732,.000125648571112395,2.04996402831039e-006,.0180526136574342,.0377863636688456,.0488699244135906,.0498875594186183,.0425743995300703,.0307433581628202,.0185802026068238,.00909387614760267,.00337206199330092,.00082597335763323,9.58139241505329e-005,1.56320996077349e-006,.00052999672436045,.00110934900341889,.00143474514828935,.00146462133295124,.00124991830660809,.00090257729054235,.000545485917243613,.000266982092533966,9.89985076235141e-005,2.4249296099811e-005,2.81294813657029e-006,4.58934187824042e-008,-.0124695874152247,-.0261003959766295,-.0337562087138414,-.0344591256926668,-.0294076639906538,-.0212355395912979,-.0128340120159257,-.00628146625843626,-.00232920410268166,-.00057052940815652,-6.61821122117001e-005,-1.07976515889088e-006,-.00998308444276622,-.0208958362733152,-.0270250386669947,-.0275877902097844,-.0235436172110031,-.0170010584848056,-.0102748408129391,-.00502890801389244,-.00186474824444618,-.000456762767608003,-5.29850421517033e-005,-8.64454163607967e-007},
			{.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.00942311384888909,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848839,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848838,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.0094231138488891,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006},
			{-.00776915549838746,-.0115499204629525,-.00437647479049313,.0101990855232791,.0237503437017621,.0291595409512613,.0251163517269084,.0158452413384258,.00706723987011998,.0019739247223913,.000249684056884166,4.2685397699263e-006,-.0162219779196966,-.024116206035927,-.00913806879424977,.0212956659438816,.0495906602970113,.0608850510918503,.0524428817548842,.0330848256515858,.0147563797828581,.00412155005372477,.00052133970783773,8.91270073194743e-006,-.0206780723147828,-.0307408045330656,-.0116482495710665,.0271454764985451,.0632129611343108,.0776098633347217,.0668486732438716,.0421730581025391,.0188098818630402,.00525371878089153,.00066454906008391,1.13609740543078e-005,-.0200202272789457,-.0297628272171986,-.0112776761907307,.0262818797044071,.0612019258671093,.075140810003757,.0647219726899549,.0408313790283736,.0182114707916447,.00508657878993488,.00064340732629019,1.09995399578346e-005,-.0144099864483064,-.0214224309688108,-.00811734846026063,.018916944603062,.0440513941259184,.054084203879524,.0465850230556935,.0293892576875637,.0131080953105517,.00366117379238187,.000463106173741249,7.91715396241739e-006,-.00523501095632017,-.0077825653226702,-.00294895545378721,.00687234596731837,.0160034523084138,.019648276623219,.0169238955894816,.0106768376600795,.00476204629431086,.00133006960033359,.000168242066165155,2.87622669770101e-006,.00523501095632017,.0077825653226702,.00294895545378721,-.00687234596731837,-.0160034523084138,-.019648276623219,-.0169238955894816,-.0106768376600795,-.00476204629431086,-.00133006960033359,-.000168242066165155,-2.87622669770101e-006,.0144099864483064,.0214224309688108,.00811734846026063,-.018916944603062,-.0440513941259184,-.054084203879524,-.0465850230556935,-.0293892576875637,-.0131080953105517,-.00366117379238187,-.000463106173741249,-7.91715396241739e-006,.0200202272789457,.0297628272171986,.0112776761907307,-.0262818797044071,-.0612019258671093,-.075140810003757,-.0647219726899549,-.0408313790283736,-.0182114707916447,-.00508657878993488,-.00064340732629019,-1.09995399578346e-005,.0206780723147828,.0307408045330656,.0116482495710665,-.0271454764985451,-.0632129611343108,-.0776098633347217,-.0668486732438716,-.0421730581025391,-.0188098818630402,-.00525371878089153,-.00066454906008391,-1.13609740543078e-005,.0162219779196966,.024116206035927,.00913806879424977,-.0212956659438816,-.0495906602970113,-.0608850510918503,-.0524428817548842,-.0330848256515858,-.0147563797828581,-.00412155005372477,-.00052133970783773,-8.91270073194743e-006,.00776915549838746,.0115499204629525,.00437647479049313,-.0101990855232791,-.0237503437017621,-.0291595409512613,-.0251163517269084,-.0158452413384258,-.00706723987011998,-.0019739247223913,-.000249684056884166,-4.2685397699263e-006},
			{.00965830811000427,.014358406200055,.0054406610822788,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.0312236592031229,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597266e-006,.0168199612662992,.0250051907000728,.00947492124135092,-.0220806783296889,-.0514186981079021,-.0631294288607147,-.0543760596998002,-.034304416435326,-.0153003374562057,-.00427348086675615,-.000540557614849618,-9.24124553932794e-006,.0134923607368192,.0200582538732903,.00760043696398981,-.0177123164923541,-.0412462081514687,-.0506401301294754,-.0436184959823004,-.0275177542970243,-.012273373825647,-.00342803081073059,-.000433615644121537,-7.41299082085032e-006,.00076631600405656,.00113923436056569,.000431676605520518,-.00100599382582219,-.00234263151050095,-.00287617140711524,-.00247736865298144,-.00156290629377859,-.000697082072575792,-.000194699424652437,-2.46277589356807e-005,-4.21030360420172e-007,-.0150234114472303,-.022334371777439,-.00846289940774852,.0197222282696639,.0459266371382019,.0563865379466385,.0485681212230874,.0306403418180061,.0136660995377323,.00381702789659693,.000482820342463611,8.25418274299939e-006,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962322,1.41275647220829e-005,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962323,1.41275647220829e-005,-.0150234114472303,-.022334371777439,-.00846289940774854,.0197222282696639,.045926637138202,.0563865379466386,.0485681212230875,.0306403418180062,.0136660995377323,.00381702789659694,.000482820342463612,8.25418274299941e-006,.000766316004056523,.00113923436056563,.000431676605520497,-.00100599382582214,-.00234263151050084,-.0028761714071151,-.00247736865298132,-.00156290629377852,-.000697082072575759,-.000194699424652428,-2.46277589356795e-005,-4.21030360420152e-007,.0134923607368192,.0200582538732903,.0076004369639898,-.0177123164923541,-.0412462081514686,-.0506401301294753,-.0436184959823003,-.0275177542970243,-.012273373825647,-.00342803081073058,-.000433615644121536,-7.41299082085031e-006,.0168199612662992,.0250051907000728,.0094749212413509,-.0220806783296889,-.051418698107902,-.0631294288607146,-.0543760596998001,-.0343044164353259,-.0153003374562056,-.00427348086675614,-.000540557614849618,-9.24124553932793e-006,.00965830811000427,.014358406200055,.00544066108227881,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.031223659203123,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597267e-006},
			{-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417216e-006,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.0255511390040629,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.025551139004063,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417217e-006},
			{.00794411476554508,.00613541631233657,-.0091565995280034,-.0197247070338324,-.0100866971092938,.0140548407519696,.0324308869658652,.0322559411834705,.0190400394460011,.00638699606694611,.00090669726269933,1.64722539026166e-005,.0165872924470305,.0128107344443715,-.0191189577031385,-.041185140655471,-.0210609992068756,.0293464735505198,.0677156136707053,.0673503272995239,.0397555563854127,.0133360323644821,.00189318169502602,3.43940263715078e-005,.0211437368750014,.0163297776977338,-.0243708376331069,-.052498488222838,-.0268463480086837,.0374077997926938,.0863167465308033,.0858511178610883,.050676204462949,.0169993722707645,.0024132290272223,4.38418895665644e-005,.0204710773480144,.0158102678018768,-.0235955122395174,-.0508283194885088,-.0259922675847922,.0362177209971177,.0835706954313348,.0831198801109617,.0490640092334553,.0164585601249111,.00233645539418102,4.24471188657934e-005,.0147344954209059,.011379778141117,-.0169833742082372,-.0365847692343567,-.0187084900904949,.0260684787183142,.0601517940761608,.0598273101146777,.0353148692221493,.011846400395665,.00168171370375142,3.05522209664918e-005,.0053529019781526,.00413416511271193,-.00616989824205258,-.0132908985350789,-.00679661642648238,.00947043025991277,.0218525744046085,.0217346924690766,.0128295559445646,.0043036845375764,.00061095058597832,1.10993311529656e-005,-.0053529019781526,-.00413416511271193,.00616989824205258,.0132908985350789,.00679661642648238,-.00947043025991277,-.0218525744046085,-.0217346924690766,-.0128295559445646,-.0043036845375764,-.00061095058597832,-1.10993311529656e-005,-.0147344954209059,-.011379778141117,.0169833742082372,.0365847692343567,.0187084900904949,-.0260684787183142,-.0601517940761608,-.0598273101146777,-.0353148692221493,-.011846400395665,-.00168171370375142,-3.05522209664918e-005,-.0204710773480144,-.0158102678018768,.0235955122395174,.0508283194885088,.0259922675847922,-.0362177209971177,-.0835706954313348,-.0831198801109616,-.0490640092334554,-.0164585601249111,-.00233645539418102,-4.24471188657934e-005,-.0211437368750014,-.0163297776977338,.0243708376331069,.052498488222838,.0268463480086837,-.0374077997926938,-.0863167465308033,-.0858511178610884,-.050676204462949,-.0169993722707645,-.0024132290272223,-4.38418895665644e-005,-.0165872924470305,-.0128107344443715,.0191189577031385,.041185140655471,.0210609992068756,-.0293464735505198,-.0677156136707053,-.0673503272995239,-.0397555563854127,-.0133360323644821,-.00189318169502602,-3.43940263715078e-005,-.00794411476554508,-.00613541631233657,.0091565995280034,.0197247070338324,.0100866971092938,-.0140548407519696,-.0324308869658652,-.0322559411834705,-.0190400394460011,-.00638699606694611,-.00090669726269933,-1.64722539026166e-005},
			{.00456148901693467,-2.90196057170823e-005,-.00924310280718151,-.00451266930256049,.0104810735535174,.0123349344803261,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.0154783325752891,-9.84711586118446e-005,-.0313642801168654,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0154783325752891,-9.84711586118446e-005,-.0313642801168653,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.00456148901693467,-2.90196057170823e-005,-.00924310280718152,-.00451266930256049,.0104810735535174,.0123349344803262,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005},
			{-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.0125539393028941,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855103,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855104,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.012553939302894,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009},
			{.00973618795942079,.0195825969134841,.023541414184445,.0215525014129859,.0158497577055902,.00941532722371888,.00442368433837659,.00156467226043237,.000378749981717022,5.17272161167757e-005,2.50040624924348e-006,7.84521387564148e-009,.0203291369483494,.040888415067351,.0491544159693757,.0450015709053905,.0330942558153189,.0196591805070174,.00923664221622893,.0032670319014163,.000790829047220676,.000108006302341081,5.22084221044592e-006,1.63807876277374e-008,.0259134469295932,.052120253632755,.0626568827202963,.0573632723498304,.0421850786840353,.0250594568791694,.0117739006080788,.00416446886110988,.00100806574315481,.00013767508138138,6.65498087262162e-006,2.08805062376629e-008,.0250890454977502,.0504621179227659,.0606635383392967,.0555383370570115,.0408430171912504,.0242622239912202,.01139932980914,.00403198189010626,.000975995488498737,.000133295133992354,6.44326161446257e-006,2.02162210390667e-008,.0180583766900454,.0363211878310999,.0436638781965725,.0399749050398971,.0293976345040205,.0174632542402221,.00820491125204314,.00290210513530733,.000702493611431951,9.59420214212558e-005,4.63767524981805e-006,1.45510571458537e-008,.00656043641434894,.0131951419194234,.0158626714587474,.0145225026138949,.0106798808777874,.00634423409130352,.00298076618282515,.00105430718024807,.000255209244351746,3.4854823431899e-005,1.68482328777663e-006,5.28626059836008e-009,-.00656043641434894,-.0131951419194234,-.0158626714587474,-.0145225026138949,-.0106798808777874,-.00634423409130352,-.00298076618282515,-.00105430718024807,-.000255209244351746,-3.4854823431899e-005,-1.68482328777663e-006,-5.28626059836008e-009,-.0180583766900454,-.0363211878310999,-.0436638781965725,-.0399749050398971,-.0293976345040205,-.0174632542402221,-.00820491125204314,-.00290210513530733,-.000702493611431951,-9.59420214212558e-005,-4.63767524981805e-006,-1.45510571458537e-008,-.0250890454977502,-.0504621179227659,-.0606635383392967,-.0555383370570115,-.0408430171912504,-.0242622239912202,-.01139932980914,-.00403198189010626,-.000975995488498737,-.000133295133992354,-6.44326161446257e-006,-2.02162210390667e-008,-.0259134469295932,-.052120253632755,-.0626568827202963,-.0573632723498304,-.0421850786840353,-.0250594568791694,-.0117739006080788,-.00416446886110988,-.00100806574315481,-.00013767508138138,-6.65498087262162e-006,-2.08805062376629e-008,-.0203291369483494,-.040888415067351,-.0491544159693757,-.0450015709053905,-.0330942558153189,-.0196591805070174,-.00923664221622893,-.0032670319014163,-.000790829047220676,-.000108006302341081,-5.22084221044592e-006,-1.63807876277374e-008,-.00973618795942079,-.0195825969134841,-.023541414184445,-.0215525014129859,-.0158497577055902,-.00941532722371888,-.00442368433837659,-.00156467226043237,-.000378749981717022,-5.17272161167757e-005,-2.50040624924348e-006,-7.84521387564149e-009},
			{-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642852e-009,-.0210785206129219,-.042395665985065,-.0509663726925087,-.0466604432029348,-.0343141942102539,-.0203838678741335,-.0095771283278592,-.00338746300209602,-.000819981016188288,-.000111987689197342,-5.41329572570134e-006,-1.69846250996501e-008,-.0169084220471927,-.0340082601910801,-.0408833691663904,-.0374293092514581,-.0275255976721928,-.0163511969032008,-.00768242376878656,-.00271729952782253,-.000657758926587005,-8.98324482923012e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.000960335605521408,-.00193154293476799,-.00232202360306468,-.00212584582191778,-.0015633517683723,-.000928687285851265,-.000436333151684879,-.000154332526131126,-3.73582653241091e-005,-5.10214959062846e-006,-2.46629292637403e-007,-7.7381807429253e-010,.0188271115999142,.0378673603101058,.0455226248627925,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606287,.00302564611398512,.000732398368229125,.000100026219157067,4.83509846932233e-006,1.51704874410297e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019803,.0146410459204576,.00517858795138701,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019804,.0146410459204576,.00517858795138702,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.0188271115999142,.0378673603101059,.0455226248627926,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606288,.00302564611398512,.000732398368229127,.000100026219157067,4.83509846932234e-006,1.51704874410297e-008,-.000960335605521362,-.0019315429347679,-.00232202360306457,-.00212584582191767,-.00156335176837222,-.000928687285851221,-.000436333151684858,-.000154332526131119,-3.73582653241074e-005,-5.10214959062822e-006,-2.46629292637392e-007,-7.73818074292493e-010,-.0169084220471927,-.03400826019108,-.0408833691663903,-.0374293092514581,-.0275255976721927,-.0163511969032008,-.00768242376878655,-.00271729952782252,-.000657758926587003,-8.9832448292301e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.0210785206129218,-.0423956659850649,-.0509663726925086,-.0466604432029347,-.0343141942102538,-.0203838678741334,-.00957712832785918,-.00338746300209602,-.000819981016188286,-.000111987689197342,-5.41329572570133e-006,-1.69846250996501e-008,-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642853e-009},
			{.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700912,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700913,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008},
			{-.00998215054668693,-.0124473495345263,.000932730855103969,.0206523073750313,.0326901637864827,.0309109950234972,.020097626924235,.00901835704324359,.00260125692755759,.000402780024162244,2.11581155396364e-005,6.94419156685537e-008,-.0208427062366112,-.0259900357702638,.00194753977310106,.0431219679279274,.0682569830460539,.0645420829652721,.0419637964861058,.0188303077287876,.00543141818323182,.000841003717818616,4.41780941541987e-005,1.44994552227784e-007,-.0265680910755723,-.0331293657149398,.00248251899142349,.0549673520444605,.0870068273056559,.0822714631661617,.0534910368290498,.0240028969866065,.00692340098850254,.0010720231392338,5.63135907453441e-005,1.84823814399061e-007,-.0257228630214447,-.0320753995403257,.00240354099144272,.0532186397309965,.0842388222152567,.0796541073115629,.0517892914968888,.0232392771293986,.00670314230567841,.00103791816611484,5.45220496597903e-005,1.78943893536226e-007,-.018514580398419,-.0230869543217517,.00172999999610706,.0383052533138549,.0606326926078268,.0573327460732011,.0372764493748974,.0167269663744151,.00482472993138678,.000747063782032829,3.92434104660595e-005,1.28798691689977e-007,-.00672617088052745,-.00838727081777638,.000628492536515794,.0139159340296408,.0220272802651019,.0208284410901999,.013542179348353,.00607674771591152,.00175277847364417,.000271401163217019,1.42567575958641e-005,4.67913390880377e-008,.00672617088052745,.00838727081777638,-.000628492536515794,-.0139159340296408,-.0220272802651019,-.0208284410901999,-.013542179348353,-.00607674771591152,-.00175277847364417,-.000271401163217019,-1.42567575958641e-005,-4.67913390880377e-008,.018514580398419,.0230869543217517,-.00172999999610706,-.0383052533138549,-.0606326926078268,-.0573327460732011,-.0372764493748974,-.0167269663744151,-.00482472993138678,-.000747063782032829,-3.92434104660595e-005,-1.28798691689977e-007,.0257228630214447,.0320753995403257,-.00240354099144272,-.0532186397309965,-.0842388222152567,-.0796541073115629,-.0517892914968887,-.0232392771293986,-.00670314230567841,-.00103791816611484,-5.45220496597903e-005,-1.78943893536226e-007,.0265680910755723,.0331293657149398,-.00248251899142349,-.0549673520444605,-.0870068273056559,-.0822714631661617,-.0534910368290498,-.0240028969866065,-.00692340098850254,-.0010720231392338,-5.63135907453441e-005,-1.84823814399061e-007,.0208427062366112,.0259900357702638,-.00194753977310106,-.0431219679279274,-.0682569830460539,-.0645420829652721,-.0419637964861058,-.0188303077287876,-.00543141818323182,-.000841003717818616,-4.41780941541987e-005,-1.44994552227784e-007,.00998215054668692,.0124473495345263,-.000932730855103969,-.0206523073750313,-.0326901637864827,-.0309109950234972,-.020097626924235,-.00901835704324359,-.00260125692755759,-.000402780024162244,-2.11581155396364e-005,-6.94419156685537e-008},
			{-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243526,-.0333379180302253,-.0122284658175383,-.00222781405170755,-.000129893528533068,-4.50638134603823e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243525,-.0333379180302253,-.0122284658175383,-.00222781405170754,-.000129893528533068,-4.50638134603823e-007,-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007},
			{-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.00895405768790681,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.0195503931804071,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.019550393180407,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.0089540576879068,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011},
			{.0122290736621916,.0236353054365663,.0264106554398282,.0216849715365018,.0137420800410902,.00671545250392889,.00245286453144479,.000626979702197685,9.90753102188358e-005,7.54445587508782e-006,1.51966644376378e-007,9.16955121940153e-011,.0255342762759209,.0493504606770767,.0551453847819794,.0452781682034031,.0286934299414573,.0140218486117464,.00512157521848587,.00130913210415982,.000206869008490129,1.5752805648851e-005,3.17305986489221e-007,1.91460008035021e-010,.0325484114177003,.0629067799068687,.0702935400351982,.057715849511425,.0365753684430921,.0178735787347226,.00652844574550074,.00166874399986403,.000263694867445998,2.0080020819912e-005,4.04468318661127e-007,2.44053093348749e-010,.0315129275220274,.0609054853832662,.068057245681468,.0558796974508159,.0354117723303635,.0173049548869921,.00632075203208146,.00161565515581309,.000255305770199198,1.94412019873441e-005,3.91600703559772e-007,2.36288872704961e-010,.0226821030656547,.0438380247522649,.0489856570714547,.0402206064787555,.0254883799410055,.0124556111144829,.00454949635967349,.00116290232753188,.00018376178439362,1.39932206199794e-005,2.81862975520663e-007,1.7007396600062e-010,.0082401922088581,.0159259372453132,.0177960230838209,.0146117636086239,.00925968589413191,.00452500499468858,.00165278873606634,.000422471349823863,6.67589076578628e-005,5.0836038966856e-006,1.02398136897977e-007,6.17862534841391e-011,-.0082401922088581,-.0159259372453132,-.0177960230838209,-.0146117636086239,-.00925968589413191,-.00452500499468858,-.00165278873606634,-.000422471349823863,-6.67589076578628e-005,-5.0836038966856e-006,-1.02398136897977e-007,-6.17862534841391e-011,-.0226821030656547,-.0438380247522649,-.0489856570714547,-.0402206064787555,-.0254883799410055,-.0124556111144829,-.00454949635967349,-.00116290232753188,-.00018376178439362,-1.39932206199794e-005,-2.81862975520663e-007,-1.7007396600062e-010,-.0315129275220274,-.0609054853832662,-.068057245681468,-.0558796974508159,-.0354117723303635,-.0173049548869921,-.00632075203208146,-.00161565515581309,-.000255305770199198,-1.94412019873441e-005,-3.91600703559772e-007,-2.36288872704961e-010,-.0325484114177003,-.0629067799068687,-.0702935400351982,-.057715849511425,-.0365753684430921,-.0178735787347226,-.00652844574550074,-.00166874399986403,-.000263694867445998,-2.0080020819912e-005,-4.04468318661127e-007,-2.44053093348749e-010,-.0255342762759209,-.0493504606770767,-.0551453847819794,-.0452781682034031,-.0286934299414573,-.0140218486117464,-.00512157521848587,-.00130913210415982,-.000206869008490129,-1.5752805648851e-005,-3.17305986489221e-007,-1.91460008035021e-010,-.0122290736621916,-.0236353054365663,-.0264106554398282,-.0216849715365018,-.0137420800410902,-.00671545250392889,-.00245286453144479,-.000626979702197685,-9.90753102188358e-005,-7.54445587508782e-006,-1.51966644376378e-007,-9.16955121940154e-011},
			{.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594673,-.00597366639905114,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.024026051223952,.0244245520757115,-.0153331658656943,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597428,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.024026051223952,.0244245520757115,-.0153331658656942,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597429,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594672,-.00597366639905115,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010},
			{-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667211,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122411e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122412e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667212,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505},
			{-.00189181133459202,-.00428844840471255,-.0064194136822003,-.00814735927652425,-.00936344787673791,-.00999122031784744,-.00999122031784744,-.00936344787673791,-.00814735927652425,-.0064194136822003,-.00428844840471255,-.00189181133459202,-.00395009749828707,-.0089542699133058,-.0134037202668508,-.0170116664951607,-.0195508565557993,-.0208616438968927,-.0208616438968927,-.0195508565557993,-.0170116664951607,-.0134037202668508,-.0089542699133058,-.00395009749828707,-.00503516908507482,-.0114139620772511,-.0170856536938404,-.0216846843044205,-.0249213768923885,-.026592230814309,-.026592230814309,-.0249213768923885,-.0216846843044205,-.0170856536938404,-.0114139620772511,-.00503516908507482,-.00487498195849973,-.0110508422381554,-.0165420966206557,-.0209948152631386,-.0241285368333486,-.0257462347868881,-.0257462347868881,-.0241285368333486,-.0209948152631386,-.0165420966206557,-.0110508422381554,-.00487498195849973,-.00350887245079362,-.00795407987509009,-.0119065275737857,-.0151114669784336,-.0173670300512356,-.0185314027260889,-.0185314027260889,-.0173670300512356,-.0151114669784336,-.0119065275737857,-.00795407987509009,-.00350887245079362,-.00127473997218044,-.00288964153040104,-.00432552816923864,-.00548985215787403,-.00630927675908016,-.00673228227209459,-.00673228227209459,-.00630927675908016,-.00548985215787403,-.00432552816923864,-.00288964153040104,-.00127473997218044,.00127473997218044,.00288964153040104,.00432552816923864,.00548985215787403,.00630927675908016,.00673228227209459,.00673228227209459,.00630927675908016,.00548985215787403,.00432552816923864,.00288964153040104,.00127473997218044,.00350887245079362,.00795407987509009,.0119065275737857,.0151114669784336,.0173670300512356,.0185314027260889,.0185314027260889,.0173670300512356,.0151114669784336,.0119065275737857,.00795407987509009,.00350887245079362,.00487498195849973,.0110508422381554,.0165420966206557,.0209948152631386,.0241285368333486,.0257462347868881,.0257462347868881,.0241285368333486,.0209948152631386,.0165420966206557,.0110508422381554,.00487498195849973,.00503516908507482,.0114139620772511,.0170856536938404,.0216846843044205,.0249213768923885,.026592230814309,.026592230814309,.0249213768923885,.0216846843044205,.0170856536938404,.0114139620772511,.00503516908507482,.00395009749828707,.0089542699133058,.0134037202668508,.0170116664951607,.0195508565557993,.0208616438968927,.0208616438968927,.0195508565557993,.0170116664951607,.0134037202668508,.0089542699133058,.00395009749828707,.00189181133459202,.00428844840471255,.0064194136822003,.00814735927652425,.00936344787673791,.00999122031784744,.00999122031784744,.00936344787673791,.00814735927652425,.0064194136822003,.00428844840471255,.00189181133459202},
			{.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433,.00409570813119402,.00928434705946777,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946777,.00409570813119402,.00328542799260275,.00744756529155359,.0111483217284699,.014149171084577,.0162611002476871,.0173513258496235,.0173513258496235,.0162611002476871,.014149171084577,.0111483217284699,.00744756529155359,.00328542799260275,.000186600113947178,.000422994062010168,.000633183289828939,.000803620393624444,.000923570130271238,.00098549089736993,.00098549089736993,.000923570130271238,.000803620393624444,.000633183289828939,.000422994062010168,.000186600113947178,-.00365824317003512,-.00829268056477257,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477257,-.00365824317003512,-.00626131850516901,-.0141934562204533,-.0212463013334064,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181941,-.0212463013334064,-.0141934562204533,-.00626131850516901,-.00626131850516901,-.0141934562204533,-.0212463013334065,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181942,-.0212463013334065,-.0141934562204533,-.00626131850516901,-.00365824317003513,-.00829268056477258,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477258,-.00365824317003513,.000186600113947169,.000422994062010148,.000633183289828909,.000803620393624405,.000923570130271194,.000985490897369883,.000985490897369883,.000923570130271194,.000803620393624405,.000633183289828908,.000422994062010148,.000186600113947169,.00328542799260274,.00744756529155357,.0111483217284699,.014149171084577,.016261100247687,.0173513258496235,.0173513258496235,.016261100247687,.014149171084577,.0111483217284699,.00744756529155357,.00328542799260274,.00409570813119402,.00928434705946775,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946775,.00409570813119401,.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433},
			{-.002625816509223,-.00595232643664948,-.00891008639077038,-.0113084587789945,-.0129963784276034,-.0138677207278419,-.0138677207278419,-.0129963784276034,-.0113084587789945,-.00891008639077038,-.00595232643664948,-.002625816509223,-.00327983286987238,-.00743488199977471,-.0111293360047077,-.0141250748788682,-.0162334074016343,-.0173217763364758,-.0173217763364758,-.0162334074016343,-.0141250748788682,-.0111293360047077,-.00743488199977471,-.00327983286987238,.00013940322318602,.000316005892951688,.00047303181976978,.000600360153697322,.000689970923782976,.000736230030131388,.000736230030131388,.000689970923782976,.000600360153697322,.00047303181976978,.000316005892951688,.00013940322318602,.00474831714066722,.0107637123716435,.0161122895623198,.0204493148954714,.0235016141598744,.0250772800775341,.0250772800775341,.0235016141598744,.0204493148954714,.0161122895623198,.0107637123716435,.00474831714066722,.00622685344747993,.0141153291794728,.0211293523233008,.0268168454598161,.0308195730860963,.032885871621092,.032885871621092,.0308195730860963,.0268168454598161,.0211293523233008,.0141153291794728,.00622685344747993,.00284444948195564,.00644793411485122,.00965196559979983,.012250033314455,.0140784939678192,.0150223867134904,.0150223867134904,.0140784939678192,.012250033314455,.00965196559979983,.00644793411485122,.00284444948195564,-.00284444948195564,-.00644793411485122,-.00965196559979984,-.012250033314455,-.0140784939678192,-.0150223867134905,-.0150223867134905,-.0140784939678192,-.012250033314455,-.00965196559979984,-.00644793411485122,-.00284444948195564,-.00622685344747996,-.0141153291794729,-.0211293523233009,-.0268168454598162,-.0308195730860964,-.0328858716210921,-.0328858716210921,-.0308195730860964,-.0268168454598162,-.0211293523233009,-.0141153291794729,-.00622685344747996,-.00474831714066721,-.0107637123716435,-.0161122895623198,-.0204493148954713,-.0235016141598743,-.025077280077534,-.025077280077534,-.0235016141598743,-.0204493148954713,-.0161122895623198,-.0107637123716435,-.00474831714066721,-.000139403223186015,-.000316005892951675,-.000473031819769762,-.000600360153697299,-.000689970923782949,-.00073623003013136,-.00073623003013136,-.000689970923782949,-.000600360153697299,-.000473031819769762,-.000316005892951675,-.000139403223186015,.00327983286987238,.00743488199977471,.0111293360047077,.0141250748788682,.0162334074016343,.0173217763364758,.0173217763364758,.0162334074016343,.0141250748788682,.0111293360047077,.00743488199977471,.00327983286987238,.00262581650922298,.00595232643664945,.00891008639077034,.0113084587789945,.0129963784276034,.0138677207278418,.0138677207278418,.0129963784276034,.0113084587789945,.00891008639077034,.00595232643664945,.00262581650922298},
			{.00274788724207344,.00622904221162261,.00932430450984374,.0118341740548762,.0136005628530905,.0145124127031825,.0145124127031825,.0136005628530905,.0118341740548762,.00932430450984374,.00622904221162261,.00274788724207344,.00176296411460539,.00399637136462944,.00598220115906356,.00759246007816479,.00872572348723498,.00931073969132332,.00931073969132332,.00872572348723498,.00759246007816479,.00598220115906356,.00399637136462943,.00176296411460539,-.00351886839094271,-.00797673915025254,-.0119404464291104,-.0151545159411986,-.0174165045748593,-.0185841942695622,-.0185841942695622,-.0174165045748593,-.0151545159411986,-.0119404464291104,-.00797673915025254,-.00351886839094271,-.00572156044449335,-.0129699068358449,-.0194147601979758,-.0246407280214811,-.0283186446851331,-.0302172684545958,-.0302172684545958,-.0283186446851331,-.0246407280214811,-.0194147601979758,-.0129699068358449,-.00572156044449335,-.000863900578693703,-.0019583311458107,-.00293144199610321,-.00372051285723621,-.00427584288738108,-.00456251680948201,-.00456251680948201,-.00427584288738108,-.00372051285723621,-.00293144199610321,-.0019583311458107,-.000863900578693703,.00559347785892407,.0126795631056259,.0189801422806282,.0240891218318902,.0276847048244469,.0295408260906545,.0295408260906545,.0276847048244469,.0240891218318902,.0189801422806282,.0126795631056259,.00559347785892407,.00559347785892411,.012679563105626,.0189801422806283,.0240891218318904,.0276847048244471,.0295408260906548,.0295408260906548,.0276847048244471,.0240891218318904,.0189801422806283,.012679563105626,.00559347785892411,-.000863900578693557,-.00195833114581037,-.00293144199610271,-.00372051285723558,-.00427584288738036,-.00456251680948123,-.00456251680948123,-.00427584288738036,-.00372051285723558,-.00293144199610271,-.00195833114581037,-.000863900578693557,-.00572156044449333,-.0129699068358448,-.0194147601979757,-.024640728021481,-.028318644685133,-.0302172684545956,-.0302172684545956,-.028318644685133,-.024640728021481,-.0194147601979757,-.0129699068358448,-.00572156044449333,-.00351886839094282,-.00797673915025279,-.0119404464291107,-.0151545159411991,-.0174165045748599,-.0185841942695628,-.0185841942695628,-.0174165045748599,-.0151545159411991,-.0119404464291107,-.0079767391502528,-.00351886839094282,.00176296411460562,.00399637136462996,.00598220115906434,.00759246007816578,.00872572348723612,.00931073969132453,.00931073969132453,.00872572348723612,.00759246007816578,.00598220115906434,.00399637136462996,.00176296411460562,.00274788724207347,.00622904221162268,.00932430450984385,.0118341740548764,.0136005628530906,.0145124127031827,.0145124127031827,.0136005628530906,.0118341740548764,.00932430450984385,.00622904221162268,.00274788724207347},
			{-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.00700310156746219,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746218,-.00308935672586509,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00308935672586509,-.00700310156746218,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746219,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261},
			{.00231698622849408,.00525225518720616,.00786214398058468,.00997843648419155,.0114678347596974,.012236695836953,.012236695836953,.0114678347596974,.00997843648419155,.00786214398058468,.00525225518720616,.00231698622849407,.00483786165004334,.0109666961477544,.0164161376459761,.0208349512830896,.0239448112857507,.0255501913584168,.0255501913584168,.0239448112857507,.0208349512830896,.0164161376459761,.0109666961477544,.00483786165004334,.00616679751037299,.0139791915092038,.0209255667251747,.0265582058759676,.03052232852132,.0325686982919874,.0325686982919874,.03052232852132,.0265582058759676,.0209255667251747,.0139791915092038,.00616679751037299,.00597060914873778,.0135344623487989,.0202598479878244,.0257132923061589,.029551301725675,.0315325689967173,.0315325689967173,.029551301725675,.0257132923061589,.0202598479878244,.0135344623487989,.00597060914873778,.0042974735362733,.0097417185286608,.0145824585745998,.0185076916715506,.0212701809756486,.0226962404368325,.0226962404368325,.0212701809756486,.0185076916715506,.0145824585745998,.0097417185286608,.0042974735362733,.00156123124248536,.00353907364270423,.00529766843861862,.00672366827160686,.0077272543489116,.00824532818128076,.00824532818128076,.0077272543489116,.00672366827160686,.00529766843861862,.00353907364270423,.00156123124248536,-.00156123124248536,-.00353907364270423,-.00529766843861862,-.00672366827160686,-.0077272543489116,-.00824532818128076,-.00824532818128076,-.0077272543489116,-.00672366827160686,-.00529766843861862,-.00353907364270423,-.00156123124248536,-.0042974735362733,-.0097417185286608,-.0145824585745998,-.0185076916715506,-.0212701809756486,-.0226962404368325,-.0226962404368325,-.0212701809756486,-.0185076916715506,-.0145824585745998,-.0097417185286608,-.0042974735362733,-.00597060914873778,-.0135344623487989,-.0202598479878244,-.0257132923061589,-.029551301725675,-.0315325689967173,-.0315325689967173,-.029551301725675,-.0257132923061589,-.0202598479878244,-.0135344623487989,-.00597060914873778,-.00616679751037299,-.0139791915092038,-.0209255667251747,-.0265582058759676,-.03052232852132,-.0325686982919874,-.0325686982919874,-.03052232852132,-.0265582058759676,-.0209255667251747,-.0139791915092038,-.00616679751037299,-.00483786165004334,-.0109666961477544,-.0164161376459761,-.0208349512830896,-.0239448112857507,-.0255501913584168,-.0255501913584168,-.0239448112857507,-.0208349512830896,-.0164161376459761,-.0109666961477544,-.00483786165004334,-.00231698622849407,-.00525225518720616,-.00786214398058468,-.00997843648419155,-.0114678347596974,-.012236695836953,-.012236695836953,-.0114678347596974,-.00997843648419155,-.00786214398058468,-.00525225518720616,-.00231698622849408},
			{-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.0264920363666131,-.0264920363666131,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947058,-.004023811085128,-.00912136739713697,-.01365384986449,-.0173291247239877,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239877,-.01365384986449,-.00912136739713697,-.004023811085128,-.000228537532606818,-.000518059808186956,-.000775487987034864,-.000984229955847933,-.00113113778066232,-.00120697492261531,-.00120697492261531,-.00113113778066232,-.000984229955847932,-.000775487987034864,-.000518059808186956,-.000228537532606818,.004480414561763,.0101564179939682,.0152032256144154,.019295553671226,.0221756405874862,.0236624065959092,.0236624065959092,.0221756405874862,.019295553671226,.0152032256144154,.0101564179939682,.004480414561763,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936553,.0404997309936553,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936554,.0404997309936554,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.004480414561763,.0101564179939682,.0152032256144154,.0192955536712261,.0221756405874863,.0236624065959092,.0236624065959092,.0221756405874863,.0192955536712261,.0152032256144154,.0101564179939682,.004480414561763,-.000228537532606807,-.000518059808186931,-.000775487987034827,-.000984229955847885,-.00113113778066226,-.00120697492261525,-.00120697492261525,-.00113113778066226,-.000984229955847886,-.000775487987034827,-.000518059808186931,-.000228537532606807,-.00402381108512799,-.00912136739713696,-.01365384986449,-.0173291247239876,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239876,-.01365384986449,-.00912136739713696,-.00402381108512799,-.00501619752947057,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.026492036366613,-.026492036366613,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.0028803860195172},
			{.00321595530276235,.0072900812758542,.0109125826103317,.0138499768923832,.0159172478252584,.0169844198386609,.0169844198386609,.0159172478252584,.0138499768923832,.0109125826103317,.0072900812758542,.00321595530276235,.00401695848624302,.00910583359827495,.0136305971932344,.0172996130152496,.019881782459596,.0212147567306737,.0212147567306737,.019881782459596,.0172996130152496,.0136305971932344,.00910583359827495,.00401695848624302,-.000170733382645958,-.00038702659670719,-.000579343295245751,-.000735288019200337,-.000845038350279965,-.000901693953533159,-.000901693953533159,-.000845038350279965,-.000735288019200337,-.000579343295245751,-.00038702659670719,-.000170733382645958,-.00581547706554894,-.0131828015237968,-.0197334440070673,-.0250451935407354,-.0287834814106213,-.0307132701622272,-.0307132701622272,-.0287834814106213,-.0250451935407354,-.0197334440070673,-.0131828015237968,-.00581547706554894,-.00762630682441431,-.0172876770198974,-.0258780658927918,-.0328437939425454,-.0377461140742206,-.0402768026076231,-.0402768026076231,-.0377461140742206,-.0328437939425454,-.0258780658927918,-.0172876770198974,-.00762630682441431,-.00348372491482343,-.00789707423793068,-.0118211953667475,-.015003165475677,-.0172425632833398,-.0183985910826998,-.0183985910826998,-.0172425632833398,-.015003165475677,-.0118211953667475,-.00789707423793069,-.00348372491482343,.00348372491482344,.00789707423793069,.0118211953667475,.015003165475677,.0172425632833398,.0183985910826999,.0183985910826999,.0172425632833398,.015003165475677,.0118211953667475,.00789707423793069,.00348372491482344,.00762630682441434,.0172876770198975,.0258780658927919,.0328437939425455,.0377461140742208,.0402768026076233,.0402768026076233,.0377461140742208,.0328437939425455,.0258780658927919,.0172876770198975,.00762630682441434,.00581547706554893,.0131828015237968,.0197334440070673,.0250451935407354,.0287834814106213,.0307132701622271,.0307132701622271,.0287834814106213,.0250451935407354,.0197334440070673,.0131828015237968,.00581547706554893,.000170733382645952,.000387026596707175,.000579343295245729,.000735288019200309,.000845038350279932,.000901693953533124,.000901693953533124,.000845038350279932,.000735288019200309,.000579343295245729,.000387026596707175,.000170733382645952,-.00401695848624301,-.00910583359827495,-.0136305971932344,-.0172996130152496,-.019881782459596,-.0212147567306737,-.0212147567306737,-.019881782459596,-.0172996130152496,-.0136305971932344,-.00910583359827495,-.00401695848624301,-.00321595530276233,-.00729008127585417,-.0109125826103316,-.0138499768923831,-.0159172478252584,-.0169844198386609,-.0169844198386609,-.0159172478252584,-.0138499768923831,-.0109125826103316,-.00729008127585417,-.00321595530276233},
			{.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954},
			{-.00267542524640207,-.00606478189450901,-.00907842189039497,-.0115221059845388,-.0132419149757324,-.0141297192758912,-.0141297192758912,-.0132419149757324,-.0115221059845388,-.00907842189039497,-.00606478189450901,-.00267542524640207,-.00558628145387762,-.0126632499505153,-.0189557229845952,-.0240581294721656,-.0276490864927876,-.0295028197276516,-.0295028197276516,-.0276490864927876,-.0240581294721656,-.0189557229845952,-.0126632499505153,-.00558628145387762,-.00712080440781243,-.0161417799674716,-.0241627631719636,-.0306667746341718,-.0352441491885744,-.0376070934653196,-.0376070934653196,-.0352441491885744,-.0306667746341718,-.0241627631719636,-.0161417799674716,-.00712080440781243,-.00689426560092869,-.0156282509663382,-.0233940573872652,-.029691152479886,-.0341229040244673,-.0364106744098191,-.0364106744098191,-.0341229040244673,-.029691152479886,-.0233940573872652,-.0156282509663382,-.00689426560092869,-.00496229500775375,-.0112487676337471,-.0168383727729163,-.0213708415448263,-.0245606894326591,-.0262073610608292,-.0262073610608292,-.0245606894326591,-.0213708415448263,-.0168383727729163,-.0112487676337471,-.00496229500775375,-.00180275455686753,-.00408657024203424,-.006117220600383,-.0077638233758434,-.00892266476002541,-.00952088489339305,-.00952088489339305,-.00892266476002541,-.0077638233758434,-.006117220600383,-.00408657024203424,-.00180275455686753,.00180275455686753,.00408657024203424,.006117220600383,.0077638233758434,.00892266476002541,.00952088489339305,.00952088489339305,.00892266476002541,.0077638233758434,.006117220600383,.00408657024203424,.00180275455686753,.00496229500775375,.0112487676337471,.0168383727729163,.0213708415448263,.0245606894326591,.0262073610608292,.0262073610608292,.0245606894326591,.0213708415448263,.0168383727729163,.0112487676337471,.00496229500775375,.00689426560092869,.0156282509663382,.0233940573872652,.029691152479886,.0341229040244673,.0364106744098191,.0364106744098191,.0341229040244673,.029691152479886,.0233940573872652,.0156282509663382,.00689426560092869,.00712080440781243,.0161417799674716,.0241627631719636,.0306667746341718,.0352441491885744,.0376070934653196,.0376070934653196,.0352441491885744,.0306667746341718,.0241627631719636,.0161417799674716,.00712080440781243,.00558628145387762,.0126632499505153,.0189557229845952,.0240581294721656,.0276490864927876,.0295028197276516,.0295028197276516,.0276490864927876,.0240581294721656,.0189557229845952,.0126632499505153,.00558628145387762,.00267542524640207,.00606478189450901,.00907842189039497,.0115221059845388,.0132419149757324,.0141297192758912,.0141297192758912,.0132419149757324,.0115221059845388,.00907842189039497,.00606478189450901,.00267542524640207},
			{.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085,.00579220598764257,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001756,.0305903686539505,.0305903686539505,.0286683021001756,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764257,.00464629682613013,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613013,.000263892411929397,.00059820393939992,.000895456396096566,.0011364908598568,.00130612540425466,.00139369459289309,.00139369459289309,.00130612540425466,.0011364908598568,.000895456396096566,.00059820393939992,.000263892411929396,-.0051735371064033,-.011727621325126,-.0175551728012975,-.0222805862115142,-.0256062241241532,-.0273229936344479,-.0273229936344479,-.0256062241241532,-.0222805862115142,-.0175551728012975,-.011727621325126,-.0051735371064033,-.00885484154985533,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985533,-.00885484154985534,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985534,-.0051735371064033,-.011727621325126,-.0175551728012976,-.0222805862115143,-.0256062241241533,-.027322993634448,-.027322993634448,-.0256062241241533,-.0222805862115143,-.0175551728012976,-.011727621325126,-.0051735371064033,.000263892411929384,.000598203939399891,.000895456396096523,.00113649085985675,.0013061254042546,.00139369459289302,.00139369459289302,.0013061254042546,.00113649085985675,.000895456396096523,.000598203939399891,.000263892411929384,.00464629682613012,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613012,.00579220598764256,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001755,.0305903686539505,.0305903686539505,.0286683021001755,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764256,.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085},
			{-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748},
			{.00299121635954409,.0067806322917519,.0101499842366785,.0128821061114311,.014804911017555,.0157975064003936,.0157975064003936,.014804911017555,.0128821061114311,.0101499842366785,.0067806322917519,.00299121635954408,.0062456525355466,.0141579438513248,.0211931425760292,.0268978064529921,.0309126184537943,.0329851552162446,.0329851552162446,.0309126184537943,.0268978064529921,.0211931425760292,.0141579438513248,.0062456525355466,.00796130135439456,.0180470586407878,.0270147904857234,.0342864963629789,.0394041566935386,.042046008708203,.042046008708203,.0394041566935386,.0342864963629789,.0270147904857234,.0180470586407878,.00796130135439456,.00770802326855251,.017472915763368,.0261553512911662,.0331957176344168,.0381505664904682,.0407083715394964,.0407083715394964,.0381505664904682,.0331957176344168,.0261553512911662,.017472915763368,.0077080232685525,.00554801448032919,.0125765045448471,.0188258730735172,.0238933272129638,.0274596855701536,.0293007204185775,.0293007204185775,.0274596855701536,.0238933272129638,.0188258730735172,.0125765045448471,.0055480144803292,.00201554086775423,.00456892442756064,.00683926054723933,.00868021841599365,.0099758424709522,.0106446729127462,.0106446729127462,.0099758424709522,.00868021841599365,.00683926054723933,.00456892442756064,.00201554086775423,-.00201554086775423,-.00456892442756064,-.00683926054723933,-.00868021841599365,-.0099758424709522,-.0106446729127462,-.0106446729127462,-.0099758424709522,-.00868021841599365,-.00683926054723933,-.00456892442756064,-.00201554086775423,-.00554801448032919,-.0125765045448471,-.0188258730735172,-.0238933272129638,-.0274596855701536,-.0293007204185775,-.0293007204185775,-.0274596855701536,-.0238933272129638,-.0188258730735172,-.0125765045448471,-.0055480144803292,-.00770802326855251,-.017472915763368,-.0261553512911662,-.0331957176344168,-.0381505664904682,-.0407083715394964,-.0407083715394964,-.0381505664904682,-.0331957176344168,-.0261553512911662,-.017472915763368,-.0077080232685525,-.00796130135439456,-.0180470586407878,-.0270147904857234,-.0342864963629789,-.0394041566935386,-.042046008708203,-.042046008708203,-.0394041566935386,-.0342864963629789,-.0270147904857234,-.0180470586407878,-.00796130135439456,-.0062456525355466,-.0141579438513248,-.0211931425760292,-.0268978064529921,-.0309126184537943,-.0329851552162446,-.0329851552162446,-.0309126184537943,-.0268978064529921,-.0211931425760292,-.0141579438513248,-.0062456525355466,-.00299121635954408,-.0067806322917519,-.0101499842366785,-.0128821061114311,-.014804911017555,-.0157975064003935,-.0157975064003935,-.014804911017555,-.0128821061114311,-.0101499842366785,-.0067806322917519,-.00299121635954409},
			{.00192735045455837,.0043690101814277,.00654000726824865,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824865,.0043690101814277,.00192735045455837,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00192735045455837,.0043690101814277,.00654000726824864,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824864,.0043690101814277,.00192735045455837},
			{.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202},
			{-.003216292833121,-.00671561166886112,-.00856035585883146,-.00828801966032105,-.00596547927875537,-.00216720185657465,.00216720185657465,.00596547927875537,.00828801966032105,.00856035585883146,.00671561166886112,.003216292833121,-.00671561166886112,-.0140221809477405,-.0178740023617151,-.0173053650368621,-.0124559063286174,-.0045251122431745,.0045251122431745,.0124559063286174,.0173053650368621,.0178740023617151,.0140221809477405,.00671561166886112,-.00856035585883146,-.0178740023617151,-.0227838994245812,-.0220590603339104,-.0158774801127414,-.00576813744045716,.00576813744045716,.0158774801127414,.0220590603339104,.0227838994245812,.0178740023617151,.00856035585883146,-.00828801966032105,-.0173053650368621,-.0220590603339104,-.0213572810231997,-.0153723594557108,-.00558463191230799,.00558463191230799,.0153723594557108,.0213572810231997,.0220590603339104,.0173053650368621,.00828801966032105,-.00596547927875537,-.0124559063286174,-.0158774801127414,-.0153723594557108,-.0110645842501621,-.00401965817140192,.00401965817140192,.0110645842501621,.0153723594557108,.0158774801127414,.0124559063286174,.00596547927875537,-.00216720185657465,-.0045251122431745,-.00576813744045716,-.00558463191230799,-.00401965817140192,-.00146030356402064,.00146030356402064,.00401965817140193,.00558463191230799,.00576813744045716,.0045251122431745,.00216720185657466,.00216720185657465,.0045251122431745,.00576813744045716,.00558463191230799,.00401965817140192,.00146030356402064,-.00146030356402064,-.00401965817140193,-.00558463191230799,-.00576813744045716,-.0045251122431745,-.00216720185657466,.00596547927875537,.0124559063286174,.0158774801127414,.0153723594557108,.0110645842501621,.00401965817140193,-.00401965817140193,-.0110645842501621,-.0153723594557108,-.0158774801127414,-.0124559063286174,-.00596547927875537,.00828801966032105,.0173053650368621,.0220590603339104,.0213572810231997,.0153723594557108,.00558463191230799,-.00558463191230799,-.0153723594557108,-.0213572810231997,-.0220590603339104,-.0173053650368621,-.00828801966032105,.00856035585883146,.0178740023617151,.0227838994245812,.0220590603339104,.0158774801127414,.00576813744045716,-.00576813744045716,-.0158774801127414,-.0220590603339104,-.0227838994245812,-.0178740023617151,-.00856035585883146,.00671561166886112,.0140221809477405,.0178740023617151,.0173053650368621,.0124559063286174,.0045251122431745,-.0045251122431745,-.0124559063286174,-.0173053650368621,-.0178740023617151,-.0140221809477405,-.00671561166886112,.003216292833121,.00671561166886112,.00856035585883146,.00828801966032105,.00596547927875537,.00216720185657466,-.00216720185657466,-.00596547927875537,-.00828801966032105,-.00856035585883146,-.00671561166886112,-.003216292833121},
			{.0039983685679321,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316192,.00269417998711382,-.00269417998711382,-.00741604887316192,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.0039983685679321,.00696316618552943,.0145390741807499,.0185328835230693,.0179432847809911,.0129150626978197,.00469191938293559,-.00469191938293559,-.0129150626978197,-.0179432847809911,-.0185328835230693,-.0145390741807499,-.00696316618552943,.00558559847779338,.011662716133544,.0148664046264906,.0143934499750413,.0103599932305885,.00376368695863054,-.00376368695863054,-.0103599932305885,-.0143934499750413,-.0148664046264906,-.011662716133544,-.00558559847779338,.000317241258906341,.000662398982523078,.000844356596319848,.000817494527800616,.000588409157550378,.000213763447844004,-.000213763447844004,-.000588409157550378,-.000817494527800616,-.000844356596319848,-.000662398982523078,-.000317241258906341,-.00621942636635253,-.0129861472342896,-.0165533754841949,-.0160267521257476,-.0115355973598394,-.00419077275215303,.00419077275215303,.0115355973598394,.0160267521257476,.0165533754841949,.0129861472342895,.00621942636635253,-.0106449482959891,-.0222266263366319,-.0283321669513853,-.0274308171979002,-.019743916918636,-.00717277714038947,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.0106449482959891,-.0222266263366319,-.0283321669513854,-.0274308171979002,-.019743916918636,-.00717277714038948,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.00621942636635254,-.0129861472342896,-.0165533754841949,-.0160267521257477,-.0115355973598394,-.00419077275215304,.00419077275215304,.0115355973598394,.0160267521257477,.0165533754841949,.0129861472342896,.00621942636635254,.000317241258906326,.000662398982523046,.000844356596319808,.000817494527800577,.000588409157550351,.000213763447843993,-.000213763447843993,-.000588409157550351,-.000817494527800577,-.000844356596319808,-.000662398982523046,-.000317241258906326,.00558559847779337,.0116627161335439,.0148664046264906,.0143934499750412,.0103599932305885,.00376368695863053,-.00376368695863053,-.0103599932305885,-.0143934499750412,-.0148664046264906,-.0116627161335439,-.00558559847779337,.00696316618552942,.0145390741807499,.0185328835230693,.017943284780991,.0129150626978197,.00469191938293558,-.00469191938293558,-.0129150626978197,-.017943284780991,-.0185328835230693,-.0145390741807499,-.00696316618552942,.00399836856793211,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316193,.00269417998711382,-.00269417998711382,-.00741604887316193,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.00399836856793211},
			{-.00446418449222142,-.00932120643965728,-.0118816941913244,-.0115036940846361,-.008280029669763,-.0030080559891824,.0030080559891824,.008280029669763,.0115036940846361,.0118816941913244,.00932120643965728,.00446418449222142,-.00557608613676323,-.0116428543884434,-.0148410869615597,-.0143689377575399,-.0103423500381766,-.00375727735471369,.00375727735471369,.0103423500381766,.0143689377575399,.0148410869615597,.0116428543884434,.00557608613676323,.000237001216546111,.000494857967838362,.000630792921502108,.00061072509381806,.00043958243844784,.000159696116976629,-.000159696116976629,-.00043958243844784,-.00061072509381806,-.000630792921502108,-.000494857967838362,-.000237001216546111,.00807267517325051,.0168557262678721,.0214859080939843,.0208023628502637,.014972945241251,.00543952852887958,-.00543952852887958,-.014972945241251,-.0208023628502637,-.0214859080939843,-.0168557262678721,-.00807267517325051,.0105863538057353,.0221042811824758,.0281762142089068,.0272798258819948,.0196352376920871,.00713329501153745,-.00713329501153745,-.0196352376920871,-.0272798258819948,-.0281762142089068,-.0221042811824758,-.0105863538057353,.00483588522718639,.0100973166766822,.0128709979423768,.0124615244685555,.00896944855895662,.0032585153113594,-.00325851531135941,-.00896944855895662,-.0124615244685555,-.0128709979423768,-.0100973166766822,-.00483588522718639,-.0048358852271864,-.0100973166766822,-.0128709979423768,-.0124615244685555,-.00896944855895663,-.00325851531135941,.00325851531135941,.00896944855895663,.0124615244685555,.0128709979423768,.0100973166766822,.0048358852271864,-.0105863538057353,-.0221042811824759,-.0281762142089069,-.0272798258819949,-.0196352376920872,-.00713329501153749,.00713329501153749,.0196352376920872,.0272798258819949,.0281762142089069,.0221042811824759,.0105863538057353,-.00807267517325049,-.0168557262678721,-.0214859080939843,-.0208023628502637,-.014972945241251,-.00543952852887957,.00543952852887957,.014972945241251,.0208023628502637,.0214859080939843,.0168557262678721,.00807267517325049,-.000237001216546102,-.000494857967838343,-.000630792921502084,-.000610725093818036,-.000439582438447823,-.000159696116976623,.000159696116976623,.000439582438447823,.000610725093818036,.000630792921502084,.000494857967838343,.000237001216546102,.00557608613676323,.0116428543884434,.0148410869615597,.0143689377575399,.0103423500381766,.00375727735471369,-.00375727735471369,-.0103423500381766,-.0143689377575399,-.0148410869615597,-.0116428543884434,-.00557608613676323,.0044641844922214,.00932120643965724,.0118816941913244,.011503694084636,.00828002966976296,.00300805598918238,-.00300805598918238,-.00828002966976296,-.011503694084636,-.0118816941913244,-.00932120643965724,-.0044641844922214},
			{.00467171851954842,.00975453698756343,.0124340584252153,.012038485593966,.00866495728795332,.00314789653003547,-.00314789653003547,-.00866495728795332,-.012038485593966,-.0124340584252153,-.00975453698756343,-.00467171851954842,.00299723801522752,.00625822573807254,.00797732835137032,.0077235403881935,.00555918325881647,.0020195983787005,-.0020195983787005,-.00555918325881647,-.0077235403881935,-.00797732835137032,-.00625822573807254,-.00299723801522752,-.00598247350841671,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016646,.00403110922016646,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841671,-.00972729866058341,-.0203105761138428,-.0258897875287452,-.0250661387888921,-.0180418890967918,-.00655444662861003,.00655444662861003,.0180418890967918,.0250661387888921,.0258897875287452,.0203105761138428,.00972729866058341,-.00146872850921155,-.00306670158055196,-.00390910882534958,-.00378474579014224,-.00272415166852724,-.000989658379109305,.000989658379109305,.00272415166852724,.00378474579014224,.00390910882534958,.00306670158055196,.00146872850921155,.00950954380591769,.0198559045031346,.0253102199512128,.02450500937368,.0176380042079596,.00640771909172347,-.00640771909172347,-.0176380042079596,-.02450500937368,-.0253102199512128,-.0198559045031346,-.00950954380591768,.00950954380591779,.0198559045031348,.0253102199512131,.0245050093736803,.0176380042079598,.00640771909172355,-.00640771909172354,-.0176380042079598,-.0245050093736803,-.0253102199512131,-.0198559045031348,-.00950954380591779,-.0014687285092115,-.00306670158055185,-.00390910882534945,-.00378474579014211,-.00272415166852715,-.000989658379109271,.000989658379109271,.00272415166852715,.00378474579014211,.00390910882534945,.00306670158055185,.0014687285092115,-.0097272986605835,-.020310576113843,-.0258897875287454,-.0250661387888923,-.0180418890967919,-.00655444662861008,.00655444662861008,.0180418890967919,.0250661387888923,.0258897875287454,.020310576113843,.0097272986605835,-.00598247350841672,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016647,.00403110922016647,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841672,.00299723801522744,.00625822573807237,.0079773283513701,.00772354038819329,.00555918325881632,.00201959837870045,-.00201959837870045,-.00555918325881632,-.00772354038819329,-.0079773283513701,-.00625822573807237,-.00299723801522744,.00467171851954842,.00975453698756342,.0124340584252153,.012038485593966,.00866495728795331,.00314789653003546,-.00314789653003546,-.00866495728795331,-.012038485593966,-.0124340584252153,-.00975453698756342,-.00467171851954842},
			{-.00231698622849408,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849407,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00231698622849407,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849408},
			{.00393913815260715,.00822491095042321,.0104842519363188,.0101507095738694,.00730619015269516,.00265426935934509,-.00265426935934509,-.00730619015269516,-.0101507095738694,-.0104842519363188,-.00822491095042321,-.00393913815260715,.00822491095042321,.0171735942029902,.0218910927256895,.0211946570783323,.0152553073958586,.00554210801278983,-.00554210801278983,-.0152553073958586,-.0211946570783323,-.0218910927256895,-.0171735942029902,-.00822491095042321,.0104842519363188,.0218910927256895,.0279044639730276,.0270167210140658,.0194458623404232,.00706449674830704,-.00706449674830704,-.0194458623404232,-.0270167210140658,-.0279044639730276,-.0218910927256895,-.0104842519363188,.0101507095738694,.0211946570783323,.0270167210140658,.026157220402348,.0188272184062363,.00683974929381445,-.00683974929381445,-.0188272184062363,-.026157220402348,-.0270167210140658,-.0211946570783323,-.0101507095738694,.00730619015269516,.0152553073958586,.0194458623404232,.0188272184062363,.0135512928156657,.00492305573060758,-.00492305573060758,-.0135512928156657,-.0188272184062363,-.0194458623404232,-.0152553073958586,-.00730619015269516,.00265426935934509,.00554210801278983,.00706449674830704,.00683974929381445,.00492305573060758,.00178849930086745,-.00178849930086745,-.00492305573060758,-.00683974929381445,-.00706449674830704,-.00554210801278983,-.00265426935934509,-.00265426935934509,-.00554210801278983,-.00706449674830704,-.00683974929381445,-.00492305573060758,-.00178849930086745,.00178849930086745,.00492305573060758,.00683974929381445,.00706449674830704,.00554210801278983,.00265426935934509,-.00730619015269516,-.0152553073958586,-.0194458623404232,-.0188272184062363,-.0135512928156657,-.00492305573060758,.00492305573060758,.0135512928156657,.0188272184062363,.0194458623404232,.0152553073958586,.00730619015269516,-.0101507095738694,-.0211946570783323,-.0270167210140658,-.026157220402348,-.0188272184062363,-.00683974929381445,.00683974929381445,.0188272184062363,.026157220402348,.0270167210140658,.0211946570783323,.0101507095738694,-.0104842519363188,-.0218910927256895,-.0279044639730276,-.0270167210140658,-.0194458623404232,-.00706449674830704,.00706449674830704,.0194458623404232,.0270167210140658,.0279044639730276,.0218910927256895,.0104842519363188,-.00822491095042321,-.0171735942029902,-.0218910927256895,-.0211946570783323,-.0152553073958586,-.00554210801278983,.00554210801278983,.0152553073958586,.0211946570783323,.0218910927256895,.0171735942029902,.00822491095042321,-.00393913815260715,-.00822491095042321,-.0104842519363188,-.0101507095738694,-.00730619015269516,-.00265426935934509,.00265426935934509,.00730619015269516,.0101507095738694,.0104842519363188,.00822491095042321,.00393913815260715},
			{-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969,-.00852810207437714,-.0178066565376609,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123493,.00574640420123493,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376609,.00852810207437714,-.00684093308933222,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010758,.00460955630010758,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933222,-.000388539604839473,-.000811269756660397,-.00103412141096871,-.0010012222303148,-.000720651097989894,-.000261805686438008,.000261805686438008,.000720651097989894,.0010012222303148,.00103412141096871,.000811269756660397,.000388539604839473,.00761721054519024,.0159047172243371,.0202736617284932,.0196286824710797,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710797,-.0202736617284932,-.0159047172243371,-.00761721054519024,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632954,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632955,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.00761721054519025,.0159047172243371,.0202736617284932,.0196286824710798,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710798,-.0202736617284932,-.0159047172243371,-.00761721054519026,-.000388539604839454,-.000811269756660358,-.00103412141096866,-.00100122223031475,-.00072065109798986,-.000261805686437995,.000261805686437995,.00072065109798986,.00100122223031475,.00103412141096866,.000811269756660358,.000388539604839454,-.00684093308933221,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010757,.00460955630010757,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933221,-.00852810207437712,-.0178066565376608,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123492,.00574640420123492,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376608,.00852810207437712,-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969},
			{.00546748706207169,.0114160997827322,.0145520440250068,.0140890903329313,.0101409238735274,.00368410114579698,-.00368410114579698,-.0101409238735274,-.0140890903329313,-.0145520440250068,-.0114160997827322,-.00546748706207169,.0068292828987853,.0142595262013292,.0181765451429699,.0175982828267856,.0126667401680366,.00460170617081502,-.00460170617081502,-.0126667401680366,-.0175982828267856,-.0181765451429699,-.0142595262013292,-.0068292828987853,-.000290266024493157,-.000606074758208074,-.000772560395559055,-.000747982426521798,-.00053837633707014,-.000195587000258209,.000195587000258209,.00053837633707014,.000747982426521798,.000772560395559055,.000606074758208074,.000290266024493157,-.00988696751735083,-.0206439643012052,-.0263147557466346,-.0254775872149812,-.0183380378947805,-.00666203466887181,.00666203466887181,.0183380378947805,.0254775872149812,.0263147557466346,.0206439643012052,.00988696751735083,-.0129655825309695,-.0270721050154096,-.0345086738493417,-.0334108268431253,-.0240481566631596,-.00873646648194734,.00873646648194734,.0240481566631596,.0334108268431253,.0345086738493417,.0270721050154096,.0129655825309695,-.00592272563093564,-.0123666368152112,-.0157636887204181,-.0152621881833591,-.0109852861223506,-.00399084991614105,.00399084991614105,.0109852861223506,.0152621881833591,.0157636887204181,.0123666368152112,.00592272563093564,.00592272563093565,.0123666368152113,.0157636887204182,.0152621881833591,.0109852861223506,.00399084991614106,-.00399084991614106,-.0109852861223506,-.0152621881833591,-.0157636887204182,-.0123666368152113,-.00592272563093565,.0129655825309696,.0270721050154097,.0345086738493419,.0334108268431254,.0240481566631597,.00873646648194738,-.00873646648194738,-.0240481566631597,-.0334108268431254,-.0345086738493419,-.0270721050154097,-.0129655825309696,.00988696751735081,.0206439643012052,.0263147557466346,.0254775872149811,.0183380378947805,.00666203466887179,-.00666203466887179,-.0183380378947805,-.0254775872149811,-.0263147557466346,-.0206439643012052,-.00988696751735081,.000290266024493146,.000606074758208051,.000772560395559025,.00074798242652177,.000538376337070119,.000195587000258202,-.000195587000258202,-.000538376337070119,-.00074798242652177,-.000772560395559025,-.000606074758208051,-.000290266024493146,-.0068292828987853,-.0142595262013292,-.0181765451429698,-.0175982828267856,-.0126667401680366,-.00460170617081502,.00460170617081502,.0126667401680366,.0175982828267856,.0181765451429698,.0142595262013292,.0068292828987853,-.00546748706207167,-.0114160997827322,-.0145520440250067,-.0140890903329312,-.0101409238735273,-.00368410114579696,.00368410114579696,.0101409238735273,.0140890903329312,.0145520440250067,.0114160997827322,.00546748706207167},
			{.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207},
			{-.00454852494396312,-.00949730909922885,-.0121061713511056,-.0117210298057487,-.00843646169984582,-.00306488625715946,.00306488625715946,.00843646169984582,.0117210298057487,.0121061713511056,.00949730909922885,.00454852494396312,-.00949730909922885,-.0198303584651126,-.0252776565471576,-.024473481930491,-.0176153116569324,-.00639947510386962,.00639947510386962,.0176153116569324,.024473481930491,.0252776565471576,.0198303584651126,.00949730909922885,-.0121061713511056,-.0252776565471576,-.0322212995614867,-.0311962222891923,-.0224541477058241,-.00815737819577448,.00815737819577448,.0224541477058241,.0311962222891923,.0322212995614867,.0252776565471576,.0121061713511056,-.0117210298057487,-.024473481930491,-.0311962222891923,-.0302037564704542,-.0217397992222051,-.00789786218916395,.00789786218916395,.0217397992222051,.0302037564704542,.0311962222891923,.024473481930491,.0117210298057487,-.00843646169984582,-.0176153116569324,-.0224541477058241,-.0217397992222051,-.0156476851044708,-.00568465510060072,.00568465510060072,.0156476851044708,.0217397992222051,.0224541477058241,.0176153116569324,.00843646169984582,-.00306488625715946,-.00639947510386962,-.00815737819577448,-.00789786218916395,-.00568465510060072,-.00206518110487492,.00206518110487492,.00568465510060072,.00789786218916395,.00815737819577448,.00639947510386962,.00306488625715946,.00306488625715946,.00639947510386962,.00815737819577448,.00789786218916395,.00568465510060072,.00206518110487492,-.00206518110487492,-.00568465510060072,-.00789786218916395,-.00815737819577448,-.00639947510386962,-.00306488625715946,.00843646169984582,.0176153116569324,.0224541477058241,.0217397992222051,.0156476851044708,.00568465510060072,-.00568465510060072,-.0156476851044708,-.0217397992222051,-.0224541477058241,-.0176153116569324,-.00843646169984582,.0117210298057487,.024473481930491,.0311962222891923,.0302037564704542,.0217397992222051,.00789786218916395,-.00789786218916395,-.0217397992222051,-.0302037564704542,-.0311962222891923,-.024473481930491,-.0117210298057487,.0121061713511056,.0252776565471576,.0322212995614867,.0311962222891923,.0224541477058241,.00815737819577448,-.00815737819577448,-.0224541477058241,-.0311962222891923,-.0322212995614867,-.0252776565471576,-.0121061713511056,.00949730909922885,.0198303584651126,.0252776565471576,.024473481930491,.0176153116569324,.00639947510386962,-.00639947510386962,-.0176153116569324,-.024473481930491,-.0252776565471576,-.0198303584651126,-.00949730909922885,.00454852494396312,.00949730909922885,.0121061713511056,.0117210298057487,.00843646169984582,.00306488625715946,-.00306488625715946,-.00843646169984582,-.0117210298057487,-.0121061713511056,-.00949730909922885,-.00454852494396312},
			{.00565454705638537,.0118066805991472,.015049915394736,.0145711225952196,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952196,-.015049915394736,-.0118066805991472,-.00565454705638537,.00984740405706795,.0205613558916723,.0262094552293619,.02537563669192,.0182646568269614,.00663537602520149,-.00663537602520149,-.0182646568269614,-.02537563669192,-.0262094552293619,-.0205613558916723,-.00984740405706795,.00789922912161445,.0164935713308931,.0210242710474368,.0203554121649402,.0146512429334382,.00532265714165691,-.00532265714165691,-.0146512429334382,-.0203554121649402,-.0210242710474368,-.0164935713308931,-.00789922912161445,.000448646890909458,.000936773624827609,.0011941005500474,.00115611184843243,.000832136210868989,.000302307167093963,-.000302307167093963,-.000832136210868989,-.00115611184843243,-.0011941005500474,-.000936773624827609,-.000448646890909458,-.00879559711786466,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317994,.00592664766317994,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786466,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.00879559711786467,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317995,.00592664766317995,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786467,.000448646890909437,.000936773624827564,.00119410055004734,.00115611184843238,.000832136210868949,.000302307167093948,-.000302307167093948,-.000832136210868949,-.00115611184843238,-.00119410055004734,-.000936773624827564,-.000448646890909437,.00789922912161443,.0164935713308931,.0210242710474368,.0203554121649401,.0146512429334382,.0053226571416569,-.0053226571416569,-.0146512429334382,-.0203554121649401,-.0210242710474368,-.0164935713308931,-.00789922912161443,.00984740405706793,.0205613558916723,.0262094552293618,.0253756366919199,.0182646568269614,.00663537602520148,-.00663537602520148,-.0182646568269614,-.0253756366919199,-.0262094552293618,-.0205613558916723,-.00984740405706793,.00565454705638537,.0118066805991472,.015049915394736,.0145711225952197,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952197,-.015049915394736,-.0118066805991472,-.00565454705638537},
			{-.00299121635954409,-.0062456525355466,-.00796130135439456,-.00770802326855251,-.00554801448032919,-.00201554086775423,.00201554086775423,.00554801448032919,.00770802326855251,.00796130135439456,.0062456525355466,.00299121635954408,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.00299121635954408,-.0062456525355466,-.00796130135439456,-.0077080232685525,-.0055480144803292,-.00201554086775423,.00201554086775423,.0055480144803292,.0077080232685525,.00796130135439456,.0062456525355466,.00299121635954409},
			{.00508540548709832,.0106183143768374,.0135351110470164,.013104509708737,.00943225092720051,.00342664700787828,-.00342664700787828,-.00943225092720051,-.013104509708737,-.0135351110470164,-.0106183143768374,-.00508540548709832,.0106183143768374,.0221710147777587,.0282612791816195,.0273621846271071,.0196945171590198,.0071548306777916,-.0071548306777916,-.0196945171590198,-.0273621846271071,-.0282612791816195,-.0221710147777587,-.0106183143768374,.0135351110470164,.0282612791816195,.03602450807902,.0348784368472584,.0251045003288081,.00912022608388363,-.00912022608388363,-.0251045003288081,-.0348784368472584,-.03602450807902,-.0282612791816195,-.0135351110470164,.013104509708737,.0273621846271071,.0348784368472584,.0337688263290031,.0243058344441419,.00883007836780732,-.00883007836780732,-.0243058344441419,-.0337688263290031,-.0348784368472584,-.0273621846271071,-.013104509708737,.00943225092720051,.0196945171590198,.0251045003288081,.0243058344441419,.0174946437957377,.00635563761813038,-.00635563761813038,-.0174946437957377,-.0243058344441419,-.0251045003288081,-.0196945171590198,-.00943225092720051,.00342664700787828,.0071548306777916,.00912022608388363,.00883007836780732,.00635563761813038,.00230894266866042,-.00230894266866042,-.00635563761813038,-.00883007836780732,-.00912022608388363,-.0071548306777916,-.00342664700787828,-.00342664700787828,-.0071548306777916,-.00912022608388363,-.00883007836780732,-.00635563761813038,-.00230894266866042,.00230894266866042,.00635563761813038,.00883007836780732,.00912022608388363,.0071548306777916,.00342664700787828,-.00943225092720051,-.0196945171590198,-.0251045003288081,-.0243058344441419,-.0174946437957377,-.00635563761813038,.00635563761813038,.0174946437957377,.0243058344441419,.0251045003288081,.0196945171590198,.00943225092720051,-.013104509708737,-.0273621846271071,-.0348784368472584,-.0337688263290031,-.0243058344441419,-.00883007836780732,.00883007836780732,.0243058344441419,.0337688263290031,.0348784368472584,.0273621846271071,.013104509708737,-.0135351110470164,-.0282612791816195,-.03602450807902,-.0348784368472584,-.0251045003288081,-.00912022608388363,.00912022608388363,.0251045003288081,.0348784368472584,.03602450807902,.0282612791816195,.0135351110470164,-.0106183143768374,-.0221710147777587,-.0282612791816195,-.0273621846271071,-.0196945171590198,-.0071548306777916,.0071548306777916,.0196945171590198,.0273621846271071,.0282612791816195,.0221710147777587,.0106183143768374,-.00508540548709832,-.0106183143768374,-.0135351110470164,-.013104509708737,-.00943225092720051,-.00342664700787828,.00342664700787828,.00943225092720051,.013104509708737,.0135351110470164,.0106183143768374,.00508540548709832},
			{.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600146,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600146,-.00872116867787892,-.00684176956018095,-.00327671334903246,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600145,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600145,-.00872116867787892,-.00684176956018095,-.00327671334903246},
			{-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847933,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847885,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847932,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847886,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.0028803860195172},
			{.00489698139750969,.00852810207437714,.00684093308933222,.000388539604839473,-.00761721054519024,-.0130373458317453,-.0130373458317453,-.00761721054519025,.000388539604839454,.00684093308933221,.00852810207437712,.00489698139750969,.0102248853327823,.0178066565376609,.0142838517710583,.000811269756660397,-.0159047172243371,-.0272219466141356,-.0272219466141356,-.0159047172243371,.000811269756660358,.0142838517710583,.0178066565376608,.0102248853327823,.0130336090560768,.0226980540469838,.0182075528223322,.00103412141096871,-.0202736617284932,-.0346996761691302,-.0346996761691302,-.0202736617284932,.00103412141096866,.0182075528223322,.0226980540469838,.0130336090560768,.0126189623285648,.0219759460114443,.0176283040385686,.0010012222303148,-.0196286824710797,-.0335957526812189,-.0335957526812189,-.0196286824710798,.00100122223031475,.0176283040385686,.0219759460114443,.0126189623285648,.00908276782339724,.0158176568028604,.0126883485768188,.000720651097989894,-.014128163704906,-.0241812609872885,-.0241812609872885,-.014128163704906,.00072065109798986,.0126883485768188,.0158176568028604,.00908276782339724,.00329968312182452,.00574640420123493,.00460955630010758,.000261805686438008,-.00513262743536863,-.00878482201632954,-.00878482201632955,-.00513262743536863,.000261805686437995,.00460955630010757,.00574640420123492,.00329968312182452,-.00329968312182452,-.00574640420123493,-.00460955630010758,-.000261805686438008,.00513262743536863,.00878482201632955,.00878482201632955,.00513262743536863,-.000261805686437995,-.00460955630010757,-.00574640420123492,-.00329968312182452,-.00908276782339724,-.0158176568028604,-.0126883485768188,-.000720651097989894,.014128163704906,.0241812609872885,.0241812609872885,.014128163704906,-.00072065109798986,-.0126883485768188,-.0158176568028604,-.00908276782339724,-.0126189623285648,-.0219759460114443,-.0176283040385686,-.0010012222303148,.0196286824710797,.0335957526812189,.0335957526812189,.0196286824710798,-.00100122223031475,-.0176283040385686,-.0219759460114443,-.0126189623285648,-.0130336090560768,-.0226980540469838,-.0182075528223322,-.00103412141096871,.0202736617284932,.0346996761691302,.0346996761691302,.0202736617284932,-.00103412141096866,-.0182075528223322,-.0226980540469838,-.0130336090560768,-.0102248853327823,-.0178066565376609,-.0142838517710583,-.000811269756660397,.0159047172243371,.0272219466141356,.0272219466141356,.0159047172243371,-.000811269756660358,-.0142838517710583,-.0178066565376608,-.0102248853327823,-.00489698139750969,-.00852810207437714,-.00684093308933222,-.000388539604839473,.00761721054519024,.0130373458317453,.0130373458317453,.00761721054519026,-.000388539604839454,-.00684093308933221,-.00852810207437712,-.00489698139750969},
			{-.00608773439400276,-.0106018005990698,-.00850437857006723,-.000483017133170316,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170293,-.00850437857006722,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.0184630551643588,-.0148103908586545,-.000841175222238799,.0164910055329813,.0282254167678824,.0282254167678824,.0164910055329813,-.000841175222238758,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00850437857006723,-.0148103908586545,-.0118803564975286,-.000674760147281664,.0132284844204532,.0226414020193,.0226414020193,.0132284844204532,-.000674760147281631,-.0118803564975285,-.0148103908586545,-.00850437857006724,-.000483017133170316,-.000841175222238799,-.000674760147281664,-3.83238715483232e-005,.000751328808837938,.00128594758620117,.00128594758620117,.000751328808837939,-3.83238715483214e-005,-.000674760147281662,-.000841175222238797,-.000483017133170317,.0094694161276367,.0164910055329813,.0132284844204532,.000751328808837938,-.0147295916665949,-.0252106436310925,-.0252106436310925,-.0147295916665949,.000751328808837902,.0132284844204532,.0164910055329813,.0094694161276367,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710571,-.0431496382710572,-.0252106436310926,.0012859475862011,.0226414020192999,.0282254167678823,.0162075148308275,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710572,-.0431496382710572,-.0252106436310926,.00128594758620111,.0226414020192999,.0282254167678823,.0162075148308275,.00946941612763672,.0164910055329813,.0132284844204532,.000751328808837939,-.0147295916665949,-.0252106436310926,-.0252106436310926,-.0147295916665949,.000751328808837903,.0132284844204532,.0164910055329813,.00946941612763672,-.000483017133170293,-.000841175222238758,-.000674760147281631,-3.83238715483214e-005,.000751328808837902,.0012859475862011,.00128594758620111,.000751328808837903,-3.83238715483196e-005,-.00067476014728163,-.000841175222238757,-.000483017133170294,-.00850437857006722,-.0148103908586545,-.0118803564975285,-.000674760147281662,.0132284844204532,.0226414020192999,.0226414020192999,.0132284844204532,-.00067476014728163,-.0118803564975285,-.0148103908586545,-.00850437857006722,-.0106018005990698,-.0184630551643587,-.0148103908586545,-.000841175222238797,.0164910055329813,.0282254167678823,.0282254167678823,.0164910055329813,-.000841175222238757,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.00850437857006724,-.000483017133170317,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170294,-.00850437857006722,-.0106018005990698,-.00608773439400276},
			{.00679696456586262,.0118369262425784,.00949515140679809,.000539289352389854,-.0105726172847424,-.0180957145755819,-.0180957145755819,-.0105726172847425,.000539289352389828,.00949515140679807,.0118369262425784,.00679696456586262,.00848989551256602,.0147851685874852,.0118601241096038,.000673611022753388,-.0132059561546973,-.0226028434433086,-.0226028434433086,-.0132059561546973,.000673611022753355,.0118601241096038,.0147851685874851,.00848989551256602,-.00036084728884685,-.00062841621455069,-.000504092615039049,-2.86305892620399e-005,.000561294714169291,.000960692009068383,.000960692009068383,.000561294714169292,-2.86305892620386e-005,-.000504092615039048,-.000628416214550689,-.00036084728884685,-.0122910885963404,-.0214049533061549,-.017170274472633,-.000975207851248453,.0191186778278397,.0327228469278241,.0327228469278241,.0191186778278397,-.000975207851248406,-.017170274472633,-.0214049533061549,-.0122910885963405,-.0161183015228526,-.0280700516286174,-.0225167737593592,-.00127886916367651,.0250718731324701,.0429121236361326,.0429121236361326,.0250718731324701,-.00127886916367645,-.0225167737593592,-.0280700516286173,-.0161183015228526,-.00736289922404356,-.0128225024865171,-.0102857447979673,-.000584192122199518,.0114529235583944,.0196024153893991,.0196024153893991,.0114529235583944,-.00058419212219949,-.0102857447979673,-.012822502486517,-.00736289922404357,.00736289922404357,.0128225024865171,.0102857447979674,.000584192122199519,-.0114529235583944,-.0196024153893992,-.0196024153893992,-.0114529235583944,.000584192122199491,.0102857447979673,.012822502486517,.00736289922404358,.0161183015228527,.0280700516286175,.0225167737593594,.00127886916367651,-.0250718731324702,-.0429121236361328,-.0429121236361328,-.0250718731324703,.00127886916367645,.0225167737593593,.0280700516286175,.0161183015228527,.0122910885963404,.0214049533061549,.017170274472633,.000975207851248451,-.0191186778278397,-.032722846927824,-.032722846927824,-.0191186778278397,.000975207851248404,.0171702744726329,.0214049533061548,.0122910885963404,.000360847288846836,.000628416214550666,.00050409261503903,2.86305892620388e-005,-.000561294714169269,-.000960692009068346,-.000960692009068346,-.00056129471416927,2.86305892620375e-005,.000504092615039029,.000628416214550665,.000360847288846836,-.00848989551256602,-.0147851685874852,-.0118601241096038,-.000673611022753387,.0132059561546973,.0226028434433086,.0226028434433086,.0132059561546973,-.000673611022753355,-.0118601241096038,-.0147851685874851,-.00848989551256602,-.00679696456586259,-.0118369262425784,-.00949515140679805,-.000539289352389851,.0105726172847424,.0180957145755818,.0180957145755819,.0105726172847424,-.000539289352389826,-.00949515140679803,-.0118369262425784,-.00679696456586259},
			{.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929397,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929396,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085},
			{-.00565454705638537,-.00984740405706795,-.00789922912161445,-.000448646890909458,.00879559711786466,.0150542302516124,.0150542302516124,.00879559711786467,-.000448646890909437,-.00789922912161443,-.00984740405706793,-.00565454705638537,-.0118066805991472,-.0205613558916723,-.0164935713308931,-.000936773624827609,.0183651855425165,.0314331964124508,.0314331964124508,.0183651855425165,-.000936773624827564,-.0164935713308931,-.0205613558916723,-.0118066805991472,-.015049915394736,-.0262094552293619,-.0210242710474368,-.0011941005500474,.0234100081138356,.0400677347558359,.0400677347558359,.0234100081138356,-.00119410055004734,-.0210242710474368,-.0262094552293618,-.015049915394736,-.0145711225952196,-.02537563669192,-.0203554121649402,-.00115611184843243,.0226652502180242,.0387930337099593,.0387930337099593,.0226652502180242,-.00115611184843238,-.0203554121649401,-.0253756366919199,-.0145711225952197,-.0104878768961101,-.0182646568269614,-.0146512429334382,-.000832136210868989,.0163137982370799,.0279221150819346,.0279221150819346,.0163137982370799,-.000832136210868949,-.0146512429334382,-.0182646568269614,-.0104878768961101,-.00381014587741865,-.00663537602520149,-.00532265714165691,-.000302307167093963,.00592664766317994,.0101438387122661,.0101438387122661,.00592664766317995,-.000302307167093948,-.0053226571416569,-.00663537602520148,-.00381014587741865,.00381014587741865,.00663537602520149,.00532265714165691,.000302307167093963,-.00592664766317994,-.0101438387122661,-.0101438387122661,-.00592664766317995,.000302307167093948,.0053226571416569,.00663537602520148,.00381014587741865,.0104878768961101,.0182646568269614,.0146512429334382,.000832136210868989,-.0163137982370799,-.0279221150819346,-.0279221150819346,-.0163137982370799,.000832136210868949,.0146512429334382,.0182646568269614,.0104878768961101,.0145711225952196,.02537563669192,.0203554121649402,.00115611184843243,-.0226652502180242,-.0387930337099593,-.0387930337099593,-.0226652502180242,.00115611184843238,.0203554121649401,.0253756366919199,.0145711225952197,.015049915394736,.0262094552293619,.0210242710474368,.0011941005500474,-.0234100081138356,-.0400677347558359,-.0400677347558359,-.0234100081138356,.00119410055004734,.0210242710474368,.0262094552293618,.015049915394736,.0118066805991472,.0205613558916723,.0164935713308931,.000936773624827609,-.0183651855425165,-.0314331964124508,-.0314331964124508,-.0183651855425165,.000936773624827564,.0164935713308931,.0205613558916723,.0118066805991472,.00565454705638537,.00984740405706795,.00789922912161445,.000448646890909458,-.00879559711786466,-.0150542302516124,-.0150542302516124,-.00879559711786467,.000448646890909437,.00789922912161443,.00984740405706793,.00565454705638537},
			{.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997765,.0122419048552221,.00702951017978813,.0122419048552221,.0213192997308994,.0171015662920701,.000971305481648202,-.0190421729606395,-.0325919039263692,-.0325919039263693,-.0190421729606395,.000971305481648156,.01710156629207,.0213192997308993,.0122419048552221,.00982001050997767,.0171015662920701,.0137182540390003,.000779145905068479,-.015274938076857,-.0261440390921353,-.0261440390921353,-.015274938076857,.000779145905068442,.0137182540390002,.01710156629207,.00982001050997767,.000557740143521656,.000971305481648202,.000779145905068479,4.42525950940344e-005,-.000867559779759337,-.00148488436959081,-.00148488436959081,-.000867559779759338,4.42525950940323e-005,.000779145905068478,.0009713054816482,.000557740143521656,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759337,.017008267421531,.0291107437634532,.0291107437634533,.0170082674215311,-.000867559779759295,-.0152749380768569,-.0190421729606395,-.0109343398968667,-.0187148260943523,-.0325919039263692,-.0261440390921353,-.00148488436959081,.0291107437634532,.0498249105249045,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0187148260943523,-.0325919039263693,-.0261440390921353,-.00148488436959081,.0291107437634533,.0498249105249046,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759338,.0170082674215311,.0291107437634533,.0291107437634533,.0170082674215311,-.000867559779759297,-.015274938076857,-.0190421729606395,-.0109343398968667,.000557740143521629,.000971305481648156,.000779145905068442,4.42525950940323e-005,-.000867559779759295,-.00148488436959074,-.00148488436959074,-.000867559779759297,4.42525950940302e-005,.000779145905068441,.000971305481648154,.000557740143521629,.00982001050997765,.01710156629207,.0137182540390002,.000779145905068478,-.0152749380768569,-.0261440390921352,-.0261440390921352,-.015274938076857,.000779145905068441,.0137182540390002,.01710156629207,.00982001050997766,.0122419048552221,.0213192997308993,.01710156629207,.0009713054816482,-.0190421729606395,-.0325919039263692,-.0325919039263692,-.0190421729606395,.000971305481648154,.01710156629207,.0213192997308993,.0122419048552221,.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997766,.0122419048552221,.00702951017978813},
			{-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471107,-.00842941499026791,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325645,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.052284927874802,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.0522849278748019,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325644,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471108,-.00842941499026791,-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697},
			{.00632197580064374,.011009732437834,.00883160664375298,.000501602473032872,-.00983377653008651,-.0168311410974181,-.0168311410974182,-.00983377653008653,.000501602473032848,.00883160664375296,.011009732437834,.00632197580064374,.0132002702054535,.0229882947439242,.0184403733456156,.0010473447524243,-.0205329016482428,-.0351433819676135,-.0351433819676135,-.0205329016482428,.00104734475242425,.0184403733456155,.0229882947439242,.0132002702054535,.0168263169407732,.0293030617759155,.0235058496220271,.00133504500106871,-.0261731847507426,-.0447970893136278,-.0447970893136278,-.0261731847507427,.00133504500106864,.023505849622027,.0293030617759154,.0168263169407732,.0162910103172928,.0283708243105143,.0227580426576453,.00129257234147053,-.0253405201097541,-.0433719302187032,-.0433719302187032,-.0253405201097541,.00129257234147047,.0227580426576453,.0283708243105143,.0162910103172928,.0117258028408243,.0204205071273958,.0163805875786201,.000930356567112207,-.01823938091645,-.0312178737024467,-.0312178737024467,-.01823938091645,.000930356567112162,.0163805875786201,.0204205071273958,.0117258028408243,.00425987259346659,.00741857592503808,.00595091159541767,.00033798968788685,-.00662619352742934,-.0113411564578213,-.0113411564578213,-.00662619352742935,.000337989687886833,.00595091159541766,.00741857592503807,.00425987259346659,-.00425987259346659,-.00741857592503808,-.00595091159541767,-.00033798968788685,.00662619352742934,.0113411564578213,.0113411564578213,.00662619352742935,-.000337989687886833,-.00595091159541766,-.00741857592503807,-.00425987259346659,-.0117258028408243,-.0204205071273958,-.0163805875786201,-.000930356567112207,.01823938091645,.0312178737024467,.0312178737024467,.01823938091645,-.000930356567112162,-.0163805875786201,-.0204205071273958,-.0117258028408243,-.0162910103172928,-.0283708243105143,-.0227580426576453,-.00129257234147053,.0253405201097541,.0433719302187032,.0433719302187032,.0253405201097541,-.00129257234147047,-.0227580426576453,-.0283708243105143,-.0162910103172928,-.0168263169407732,-.0293030617759155,-.0235058496220271,-.00133504500106871,.0261731847507426,.0447970893136278,.0447970893136278,.0261731847507427,-.00133504500106864,-.023505849622027,-.0293030617759154,-.0168263169407732,-.0132002702054535,-.0229882947439242,-.0184403733456156,-.0010473447524243,.0205329016482428,.0351433819676135,.0351433819676135,.0205329016482428,-.00104734475242425,-.0184403733456155,-.0229882947439242,-.0132002702054535,-.00632197580064374,-.011009732437834,-.00883160664375298,-.000501602473032872,.00983377653008651,.0168311410974181,.0168311410974182,.00983377653008653,-.000501602473032848,-.00883160664375296,-.011009732437834,-.00632197580064374},
			{.00407348097349134,.0070939745774067,.00569052820876414,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876412,.00709397457740669,.00407348097349134,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886327,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886328,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.00407348097349134,.0070939745774067,.00569052820876413,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876413,.00709397457740669,.00407348097349134},
			{.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934987,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934986,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133},
			{-.00631331025337702,-.00788577663899888,.00033517033471411,.0114164867138149,.0149713651270644,.00683897447387038,-.00683897447387039,-.0149713651270645,-.0114164867138149,-.000335170334714097,.00788577663899888,.00631331025337699,-.0131821765636861,-.0164654825796768,.000699834849514613,.0238375966899433,.0312601742324971,.0142797621867037,-.0142797621867037,-.0312601742324972,-.0238375966899433,-.000699834849514586,.0164654825796768,.0131821765636861,-.0168032530681212,-.0209884664598731,.00089207590457249,.0303856626242094,.0398471842676737,.0182023398500645,-.0182023398500645,-.0398471842676739,-.0303856626242093,-.000892075904572456,.0209884664598731,.0168032530681211,-.0162686801907029,-.0203207466521331,.000863695710496402,.0294189836701143,.0385794997386939,.0176232569099964,-.0176232569099964,-.0385794997386941,-.0294189836701142,-.000863695710496369,.0203207466521331,.0162686801907028,-.0117097302549807,-.0146262916897378,.000621663446188857,.0211749422273102,.0277684194425538,.0126847157981637,-.0126847157981637,-.0277684194425539,-.0211749422273102,-.000621663446188833,.0146262916897378,.0117097302549806,-.00425403357597065,-.00531359259224779,.000225844414470279,.00769265501789862,.0100880025489925,.00460823654619044,-.00460823654619044,-.0100880025489926,-.0076926550178986,-.00022584441447027,.00531359259224779,.00425403357597063,.00425403357597065,.00531359259224779,-.000225844414470279,-.00769265501789862,-.0100880025489925,-.00460823654619044,.00460823654619044,.0100880025489926,.0076926550178986,.00022584441447027,-.00531359259224779,-.00425403357597063,.0117097302549807,.0146262916897378,-.000621663446188857,-.0211749422273102,-.0277684194425538,-.0126847157981637,.0126847157981637,.0277684194425539,.0211749422273102,.000621663446188833,-.0146262916897378,-.0117097302549806,.0162686801907029,.0203207466521331,-.000863695710496402,-.0294189836701143,-.0385794997386939,-.0176232569099964,.0176232569099964,.0385794997386941,.0294189836701142,.000863695710496369,-.0203207466521331,-.0162686801907028,.0168032530681212,.0209884664598731,-.00089207590457249,-.0303856626242094,-.0398471842676737,-.0182023398500645,.0182023398500645,.0398471842676739,.0303856626242093,.000892075904572456,-.0209884664598731,-.0168032530681211,.0131821765636861,.0164654825796768,-.000699834849514612,-.0238375966899433,-.0312601742324971,-.0142797621867037,.0142797621867037,.0312601742324972,.0238375966899433,.000699834849514586,-.0164654825796768,-.0131821765636861,.00631331025337702,.00788577663899888,-.00033517033471411,-.0114164867138149,-.0149713651270644,-.00683897447387039,.00683897447387039,.0149713651270645,.0114164867138149,.000335170334714097,-.00788577663899888,-.00631331025337699},
			{.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.00980328691616649,-.00784845864116279,.0136681051009101,.0170724421228125,-.000725632541047572,-.0247163110990939,-.0324125037180548,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990939,.000725632541047544,-.0170724421228125,-.01366810510091,.0109640564380927,.0136948916904454,-.000582076013801829,-.0198265251716479,-.0260001307746002,-.0118769550522377,.0118769550522377,.0260001307746003,.0198265251716478,.000582076013801807,-.0136948916904454,-.0109640564380926,.000622717705357654,.000777819010395326,-3.30597568249531e-005,-.00112607303052629,-.00147671091136544,-.000674566957815848,.000674566957815848,.00147671091136544,.00112607303052629,3.30597568249518e-005,-.000777819010395325,-.000622717705357651,-.0122082068670625,-.0152489246770109,.000648127308443884,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.0289505053954479,-.0220763475742017,-.000648127308443859,.0152489246770109,.0122082068670624,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533856,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533857,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0122082068670625,-.015248924677011,.000648127308443885,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.028950505395448,-.0220763475742018,-.00064812730844386,.0152489246770109,.0122082068670624,.000622717705357624,.000777819010395288,-3.30597568249515e-005,-.00112607303052624,-.00147671091136537,-.000674566957815815,.000674566957815816,.00147671091136537,.00112607303052624,3.30597568249502e-005,-.000777819010395288,-.000622717705357621,.0109640564380927,.0136948916904453,-.000582076013801828,-.0198265251716478,-.0260001307746001,-.0118769550522376,.0118769550522377,.0260001307746002,.0198265251716478,.000582076013801806,-.0136948916904453,-.0109640564380926,.0136681051009101,.0170724421228125,-.000725632541047571,-.0247163110990939,-.0324125037180547,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990938,.000725632541047543,-.0170724421228125,-.01366810510091,.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.0098032869161665,-.0078484586411628},
			{-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084},
			{.007058495447099,.00881656631296387,-.000374731826353174,-.0127640201823163,-.0167384950754978,-.00764620591247183,.00764620591247184,.0167384950754978,.0127640201823163,.00037473182635316,-.00881656631296387,-.00705849544709897,.0147381214487064,.0184089691712473,-.000782439148523992,-.0266512433181539,-.0349499372975651,-.0159652594812031,.0159652594812031,.0349499372975653,.0266512433181539,.000782439148523962,-.0184089691712473,-.0147381214487063,.0187866080578478,.0234658188815225,-.000997371182181883,-.0339722035956245,-.0445505063817578,-.0203508346337808,.0203508346337809,.044550506381758,.0339722035956245,.000997371182181844,-.0234658188815225,-.0187866080578477,.0181889374112355,.0227192854412644,-.000965641160587157,-.0328914236683848,-.043133191990884,-.0197034002242685,.0197034002242685,.0431331919908843,.0328914236683847,.00096564116058712,-.0227192854412643,-.0181889374112354,.0130918764284278,.0163526912438262,-.000695040862629039,-.0236743051276634,-.031046036760756,-.0141819434046015,.0141819434046015,.0310460367607562,.0236743051276633,.000695040862629013,-.0163526912438262,-.0130918764284278,.00475615412876842,.00594077712243872,-.000252501731629378,-.00860064977654093,-.0112787297320448,-.00515216508851936,.00515216508851937,.0112787297320449,.00860064977654092,.000252501731629368,-.00594077712243872,-.00475615412876839,-.00475615412876842,-.00594077712243872,.000252501731629378,.00860064977654093,.0112787297320448,.00515216508851936,-.00515216508851937,-.0112787297320449,-.00860064977654092,-.000252501731629368,.00594077712243872,.00475615412876839,-.0130918764284278,-.0163526912438262,.000695040862629039,.0236743051276634,.031046036760756,.0141819434046015,-.0141819434046015,-.0310460367607562,-.0236743051276633,-.000695040862629013,.0163526912438262,.0130918764284278,-.0181889374112355,-.0227192854412644,.000965641160587157,.0328914236683848,.043133191990884,.0197034002242685,-.0197034002242685,-.0431331919908843,-.0328914236683847,-.00096564116058712,.0227192854412643,.0181889374112354,-.0187866080578478,-.0234658188815225,.000997371182181883,.0339722035956245,.0445505063817578,.0203508346337808,-.0203508346337809,-.044550506381758,-.0339722035956245,-.000997371182181844,.0234658188815225,.0187866080578477,-.0147381214487064,-.0184089691712473,.000782439148523992,.0266512433181539,.0349499372975651,.0159652594812031,-.0159652594812031,-.0349499372975653,-.0266512433181539,-.000782439148523962,.0184089691712473,.0147381214487063,-.007058495447099,-.00881656631296387,.000374731826353174,.0127640201823163,.0167384950754978,.00764620591247183,-.00764620591247184,-.0167384950754978,-.0127640201823163,-.00037473182635316,.00881656631296387,.00705849544709897},
			{.00454804760548224,.00568083717114684,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.00454804760548224,.00568083717114683,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222},
			{-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023209,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698057,.00556381945057915,-.00278749101749906,-.00434479121890393,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023208,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698056,.00556381945057915,-.00278749101749906,-.00434479121890393},
			{.00738663555257316,.00473904940765641,-.00945912116166398,-.015380209620101,-.00232226367616612,.0150359089640387,.0150359089640388,-.00232226367616604,-.0153802096201012,-.00945912116166399,.00473904940765628,.00738663555257316,.0154232771965445,.00989512371934247,-.0197506221436933,-.0321138405466809,-.00484888094803859,.031394941608239,.0313949416082394,-.00484888094803843,-.0321138405466812,-.0197506221436933,.00989512371934219,.0154232771965445,.0196599725865644,.0126132636134246,-.0251760170659189,-.0409353483537532,-.00618084375318826,.0400189715524956,.040018971552496,-.00618084375318805,-.0409353483537535,-.0251760170659189,.0126132636134242,.0196599725865644,.0190345170231114,.0122119896103414,-.0243750759725306,-.0396330453491579,-.0059842085292456,.0387458218422929,.0387458218422933,-.0059842085292454,-.0396330453491583,-.0243750759725306,.0122119896103411,.0190345170231114,.0137005004254646,.00878984051179792,-.0175444818655976,-.0285267314116404,-.00430726198103432,.0278881333311886,.0278881333311889,-.00430726198103417,-.0285267314116406,-.0175444818655976,.00878984051179767,.0137005004254646,.00497726143544053,.00319326541691349,-.006373743314669,-.0103634900715324,-.00156478729132365,.0101314934655784,.0101314934655785,-.0015647872913236,-.0103634900715325,-.006373743314669,.0031932654169134,.00497726143544053,-.00497726143544053,-.00319326541691349,.006373743314669,.0103634900715324,.00156478729132365,-.0101314934655784,-.0101314934655785,.0015647872913236,.0103634900715325,.006373743314669,-.0031932654169134,-.00497726143544053,-.0137005004254646,-.00878984051179792,.0175444818655976,.0285267314116404,.00430726198103432,-.0278881333311886,-.0278881333311889,.00430726198103417,.0285267314116406,.0175444818655976,-.00878984051179767,-.0137005004254646,-.0190345170231114,-.0122119896103414,.0243750759725306,.0396330453491579,.0059842085292456,-.0387458218422929,-.0387458218422933,.0059842085292454,.0396330453491583,.0243750759725306,-.0122119896103411,-.0190345170231114,-.0196599725865644,-.0126132636134246,.0251760170659189,.0409353483537532,.00618084375318826,-.0400189715524956,-.040018971552496,.00618084375318805,.0409353483537535,.0251760170659189,-.0126132636134242,-.0196599725865644,-.0154232771965445,-.00989512371934246,.0197506221436933,.0321138405466809,.00484888094803859,-.031394941608239,-.0313949416082394,.00484888094803843,.0321138405466812,.0197506221436933,-.00989512371934219,-.0154232771965445,-.00738663555257316,-.00473904940765642,.00945912116166398,.015380209620101,.00232226367616612,-.0150359089640386,-.0150359089640388,.00232226367616604,.0153802096201012,.00945912116166399,-.00473904940765628,-.00738663555257316},
			{.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.00740597712796921,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.0074059771279692,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619},
			{.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648773,.00909571150529476,.000203199397318671,-.00473557895480586,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440274,.00068950902635523,-.0160690655456805,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.036404249458381,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.0364042494583809,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440273,.00068950902635523,-.0160690655456805,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648772,.00909571150529476,.000203199397318671,-.00473557895480586}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505},
			{-.00189181133459202,-.00428844840471255,-.0064194136822003,-.00814735927652425,-.00936344787673791,-.00999122031784744,-.00999122031784744,-.00936344787673791,-.00814735927652425,-.0064194136822003,-.00428844840471255,-.00189181133459202,-.00395009749828707,-.0089542699133058,-.0134037202668508,-.0170116664951607,-.0195508565557993,-.0208616438968927,-.0208616438968927,-.0195508565557993,-.0170116664951607,-.0134037202668508,-.0089542699133058,-.00395009749828707,-.00503516908507482,-.0114139620772511,-.0170856536938404,-.0216846843044205,-.0249213768923885,-.026592230814309,-.026592230814309,-.0249213768923885,-.0216846843044205,-.0170856536938404,-.0114139620772511,-.00503516908507482,-.00487498195849973,-.0110508422381554,-.0165420966206557,-.0209948152631386,-.0241285368333486,-.0257462347868881,-.0257462347868881,-.0241285368333486,-.0209948152631386,-.0165420966206557,-.0110508422381554,-.00487498195849973,-.00350887245079362,-.00795407987509009,-.0119065275737857,-.0151114669784336,-.0173670300512356,-.0185314027260889,-.0185314027260889,-.0173670300512356,-.0151114669784336,-.0119065275737857,-.00795407987509009,-.00350887245079362,-.00127473997218044,-.00288964153040104,-.00432552816923864,-.00548985215787403,-.00630927675908016,-.00673228227209459,-.00673228227209459,-.00630927675908016,-.00548985215787403,-.00432552816923864,-.00288964153040104,-.00127473997218044,.00127473997218044,.00288964153040104,.00432552816923864,.00548985215787403,.00630927675908016,.00673228227209459,.00673228227209459,.00630927675908016,.00548985215787403,.00432552816923864,.00288964153040104,.00127473997218044,.00350887245079362,.00795407987509009,.0119065275737857,.0151114669784336,.0173670300512356,.0185314027260889,.0185314027260889,.0173670300512356,.0151114669784336,.0119065275737857,.00795407987509009,.00350887245079362,.00487498195849973,.0110508422381554,.0165420966206557,.0209948152631386,.0241285368333486,.0257462347868881,.0257462347868881,.0241285368333486,.0209948152631386,.0165420966206557,.0110508422381554,.00487498195849973,.00503516908507482,.0114139620772511,.0170856536938404,.0216846843044205,.0249213768923885,.026592230814309,.026592230814309,.0249213768923885,.0216846843044205,.0170856536938404,.0114139620772511,.00503516908507482,.00395009749828707,.0089542699133058,.0134037202668508,.0170116664951607,.0195508565557993,.0208616438968927,.0208616438968927,.0195508565557993,.0170116664951607,.0134037202668508,.0089542699133058,.00395009749828707,.00189181133459202,.00428844840471255,.0064194136822003,.00814735927652425,.00936344787673791,.00999122031784744,.00999122031784744,.00936344787673791,.00814735927652425,.0064194136822003,.00428844840471255,.00189181133459202},
			{.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433,.00409570813119402,.00928434705946777,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946777,.00409570813119402,.00328542799260275,.00744756529155359,.0111483217284699,.014149171084577,.0162611002476871,.0173513258496235,.0173513258496235,.0162611002476871,.014149171084577,.0111483217284699,.00744756529155359,.00328542799260275,.000186600113947178,.000422994062010168,.000633183289828939,.000803620393624444,.000923570130271238,.00098549089736993,.00098549089736993,.000923570130271238,.000803620393624444,.000633183289828939,.000422994062010168,.000186600113947178,-.00365824317003512,-.00829268056477257,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477257,-.00365824317003512,-.00626131850516901,-.0141934562204533,-.0212463013334064,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181941,-.0212463013334064,-.0141934562204533,-.00626131850516901,-.00626131850516901,-.0141934562204533,-.0212463013334065,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181942,-.0212463013334065,-.0141934562204533,-.00626131850516901,-.00365824317003513,-.00829268056477258,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477258,-.00365824317003513,.000186600113947169,.000422994062010148,.000633183289828909,.000803620393624405,.000923570130271194,.000985490897369883,.000985490897369883,.000923570130271194,.000803620393624405,.000633183289828908,.000422994062010148,.000186600113947169,.00328542799260274,.00744756529155357,.0111483217284699,.014149171084577,.016261100247687,.0173513258496235,.0173513258496235,.016261100247687,.014149171084577,.0111483217284699,.00744756529155357,.00328542799260274,.00409570813119402,.00928434705946775,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946775,.00409570813119401,.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433},
			{-.002625816509223,-.00595232643664948,-.00891008639077038,-.0113084587789945,-.0129963784276034,-.0138677207278419,-.0138677207278419,-.0129963784276034,-.0113084587789945,-.00891008639077038,-.00595232643664948,-.002625816509223,-.00327983286987238,-.00743488199977471,-.0111293360047077,-.0141250748788682,-.0162334074016343,-.0173217763364758,-.0173217763364758,-.0162334074016343,-.0141250748788682,-.0111293360047077,-.00743488199977471,-.00327983286987238,.00013940322318602,.000316005892951688,.00047303181976978,.000600360153697322,.000689970923782976,.000736230030131388,.000736230030131388,.000689970923782976,.000600360153697322,.00047303181976978,.000316005892951688,.00013940322318602,.00474831714066722,.0107637123716435,.0161122895623198,.0204493148954714,.0235016141598744,.0250772800775341,.0250772800775341,.0235016141598744,.0204493148954714,.0161122895623198,.0107637123716435,.00474831714066722,.00622685344747993,.0141153291794728,.0211293523233008,.0268168454598161,.0308195730860963,.032885871621092,.032885871621092,.0308195730860963,.0268168454598161,.0211293523233008,.0141153291794728,.00622685344747993,.00284444948195564,.00644793411485122,.00965196559979983,.012250033314455,.0140784939678192,.0150223867134904,.0150223867134904,.0140784939678192,.012250033314455,.00965196559979983,.00644793411485122,.00284444948195564,-.00284444948195564,-.00644793411485122,-.00965196559979984,-.012250033314455,-.0140784939678192,-.0150223867134905,-.0150223867134905,-.0140784939678192,-.012250033314455,-.00965196559979984,-.00644793411485122,-.00284444948195564,-.00622685344747996,-.0141153291794729,-.0211293523233009,-.0268168454598162,-.0308195730860964,-.0328858716210921,-.0328858716210921,-.0308195730860964,-.0268168454598162,-.0211293523233009,-.0141153291794729,-.00622685344747996,-.00474831714066721,-.0107637123716435,-.0161122895623198,-.0204493148954713,-.0235016141598743,-.025077280077534,-.025077280077534,-.0235016141598743,-.0204493148954713,-.0161122895623198,-.0107637123716435,-.00474831714066721,-.000139403223186015,-.000316005892951675,-.000473031819769762,-.000600360153697299,-.000689970923782949,-.00073623003013136,-.00073623003013136,-.000689970923782949,-.000600360153697299,-.000473031819769762,-.000316005892951675,-.000139403223186015,.00327983286987238,.00743488199977471,.0111293360047077,.0141250748788682,.0162334074016343,.0173217763364758,.0173217763364758,.0162334074016343,.0141250748788682,.0111293360047077,.00743488199977471,.00327983286987238,.00262581650922298,.00595232643664945,.00891008639077034,.0113084587789945,.0129963784276034,.0138677207278418,.0138677207278418,.0129963784276034,.0113084587789945,.00891008639077034,.00595232643664945,.00262581650922298},
			{.00274788724207344,.00622904221162261,.00932430450984374,.0118341740548762,.0136005628530905,.0145124127031825,.0145124127031825,.0136005628530905,.0118341740548762,.00932430450984374,.00622904221162261,.00274788724207344,.00176296411460539,.00399637136462944,.00598220115906356,.00759246007816479,.00872572348723498,.00931073969132332,.00931073969132332,.00872572348723498,.00759246007816479,.00598220115906356,.00399637136462943,.00176296411460539,-.00351886839094271,-.00797673915025254,-.0119404464291104,-.0151545159411986,-.0174165045748593,-.0185841942695622,-.0185841942695622,-.0174165045748593,-.0151545159411986,-.0119404464291104,-.00797673915025254,-.00351886839094271,-.00572156044449335,-.0129699068358449,-.0194147601979758,-.0246407280214811,-.0283186446851331,-.0302172684545958,-.0302172684545958,-.0283186446851331,-.0246407280214811,-.0194147601979758,-.0129699068358449,-.00572156044449335,-.000863900578693703,-.0019583311458107,-.00293144199610321,-.00372051285723621,-.00427584288738108,-.00456251680948201,-.00456251680948201,-.00427584288738108,-.00372051285723621,-.00293144199610321,-.0019583311458107,-.000863900578693703,.00559347785892407,.0126795631056259,.0189801422806282,.0240891218318902,.0276847048244469,.0295408260906545,.0295408260906545,.0276847048244469,.0240891218318902,.0189801422806282,.0126795631056259,.00559347785892407,.00559347785892411,.012679563105626,.0189801422806283,.0240891218318904,.0276847048244471,.0295408260906548,.0295408260906548,.0276847048244471,.0240891218318904,.0189801422806283,.012679563105626,.00559347785892411,-.000863900578693557,-.00195833114581037,-.00293144199610271,-.00372051285723558,-.00427584288738036,-.00456251680948123,-.00456251680948123,-.00427584288738036,-.00372051285723558,-.00293144199610271,-.00195833114581037,-.000863900578693557,-.00572156044449333,-.0129699068358448,-.0194147601979757,-.024640728021481,-.028318644685133,-.0302172684545956,-.0302172684545956,-.028318644685133,-.024640728021481,-.0194147601979757,-.0129699068358448,-.00572156044449333,-.00351886839094282,-.00797673915025279,-.0119404464291107,-.0151545159411991,-.0174165045748599,-.0185841942695628,-.0185841942695628,-.0174165045748599,-.0151545159411991,-.0119404464291107,-.0079767391502528,-.00351886839094282,.00176296411460562,.00399637136462996,.00598220115906434,.00759246007816578,.00872572348723612,.00931073969132453,.00931073969132453,.00872572348723612,.00759246007816578,.00598220115906434,.00399637136462996,.00176296411460562,.00274788724207347,.00622904221162268,.00932430450984385,.0118341740548764,.0136005628530906,.0145124127031827,.0145124127031827,.0136005628530906,.0118341740548764,.00932430450984385,.00622904221162268,.00274788724207347},
			{-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.00700310156746219,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746218,-.00308935672586509,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00308935672586509,-.00700310156746218,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746219,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261},
			{.00231698622849408,.00525225518720616,.00786214398058468,.00997843648419155,.0114678347596974,.012236695836953,.012236695836953,.0114678347596974,.00997843648419155,.00786214398058468,.00525225518720616,.00231698622849407,.00483786165004334,.0109666961477544,.0164161376459761,.0208349512830896,.0239448112857507,.0255501913584168,.0255501913584168,.0239448112857507,.0208349512830896,.0164161376459761,.0109666961477544,.00483786165004334,.00616679751037299,.0139791915092038,.0209255667251747,.0265582058759676,.03052232852132,.0325686982919874,.0325686982919874,.03052232852132,.0265582058759676,.0209255667251747,.0139791915092038,.00616679751037299,.00597060914873778,.0135344623487989,.0202598479878244,.0257132923061589,.029551301725675,.0315325689967173,.0315325689967173,.029551301725675,.0257132923061589,.0202598479878244,.0135344623487989,.00597060914873778,.0042974735362733,.0097417185286608,.0145824585745998,.0185076916715506,.0212701809756486,.0226962404368325,.0226962404368325,.0212701809756486,.0185076916715506,.0145824585745998,.0097417185286608,.0042974735362733,.00156123124248536,.00353907364270423,.00529766843861862,.00672366827160686,.0077272543489116,.00824532818128076,.00824532818128076,.0077272543489116,.00672366827160686,.00529766843861862,.00353907364270423,.00156123124248536,-.00156123124248536,-.00353907364270423,-.00529766843861862,-.00672366827160686,-.0077272543489116,-.00824532818128076,-.00824532818128076,-.0077272543489116,-.00672366827160686,-.00529766843861862,-.00353907364270423,-.00156123124248536,-.0042974735362733,-.0097417185286608,-.0145824585745998,-.0185076916715506,-.0212701809756486,-.0226962404368325,-.0226962404368325,-.0212701809756486,-.0185076916715506,-.0145824585745998,-.0097417185286608,-.0042974735362733,-.00597060914873778,-.0135344623487989,-.0202598479878244,-.0257132923061589,-.029551301725675,-.0315325689967173,-.0315325689967173,-.029551301725675,-.0257132923061589,-.0202598479878244,-.0135344623487989,-.00597060914873778,-.00616679751037299,-.0139791915092038,-.0209255667251747,-.0265582058759676,-.03052232852132,-.0325686982919874,-.0325686982919874,-.03052232852132,-.0265582058759676,-.0209255667251747,-.0139791915092038,-.00616679751037299,-.00483786165004334,-.0109666961477544,-.0164161376459761,-.0208349512830896,-.0239448112857507,-.0255501913584168,-.0255501913584168,-.0239448112857507,-.0208349512830896,-.0164161376459761,-.0109666961477544,-.00483786165004334,-.00231698622849407,-.00525225518720616,-.00786214398058468,-.00997843648419155,-.0114678347596974,-.012236695836953,-.012236695836953,-.0114678347596974,-.00997843648419155,-.00786214398058468,-.00525225518720616,-.00231698622849408},
			{-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.0264920363666131,-.0264920363666131,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947058,-.004023811085128,-.00912136739713697,-.01365384986449,-.0173291247239877,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239877,-.01365384986449,-.00912136739713697,-.004023811085128,-.000228537532606818,-.000518059808186956,-.000775487987034864,-.000984229955847933,-.00113113778066232,-.00120697492261531,-.00120697492261531,-.00113113778066232,-.000984229955847932,-.000775487987034864,-.000518059808186956,-.000228537532606818,.004480414561763,.0101564179939682,.0152032256144154,.019295553671226,.0221756405874862,.0236624065959092,.0236624065959092,.0221756405874862,.019295553671226,.0152032256144154,.0101564179939682,.004480414561763,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936553,.0404997309936553,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936554,.0404997309936554,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.004480414561763,.0101564179939682,.0152032256144154,.0192955536712261,.0221756405874863,.0236624065959092,.0236624065959092,.0221756405874863,.0192955536712261,.0152032256144154,.0101564179939682,.004480414561763,-.000228537532606807,-.000518059808186931,-.000775487987034827,-.000984229955847885,-.00113113778066226,-.00120697492261525,-.00120697492261525,-.00113113778066226,-.000984229955847886,-.000775487987034827,-.000518059808186931,-.000228537532606807,-.00402381108512799,-.00912136739713696,-.01365384986449,-.0173291247239876,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239876,-.01365384986449,-.00912136739713696,-.00402381108512799,-.00501619752947057,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.026492036366613,-.026492036366613,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.0028803860195172},
			{.00321595530276235,.0072900812758542,.0109125826103317,.0138499768923832,.0159172478252584,.0169844198386609,.0169844198386609,.0159172478252584,.0138499768923832,.0109125826103317,.0072900812758542,.00321595530276235,.00401695848624302,.00910583359827495,.0136305971932344,.0172996130152496,.019881782459596,.0212147567306737,.0212147567306737,.019881782459596,.0172996130152496,.0136305971932344,.00910583359827495,.00401695848624302,-.000170733382645958,-.00038702659670719,-.000579343295245751,-.000735288019200337,-.000845038350279965,-.000901693953533159,-.000901693953533159,-.000845038350279965,-.000735288019200337,-.000579343295245751,-.00038702659670719,-.000170733382645958,-.00581547706554894,-.0131828015237968,-.0197334440070673,-.0250451935407354,-.0287834814106213,-.0307132701622272,-.0307132701622272,-.0287834814106213,-.0250451935407354,-.0197334440070673,-.0131828015237968,-.00581547706554894,-.00762630682441431,-.0172876770198974,-.0258780658927918,-.0328437939425454,-.0377461140742206,-.0402768026076231,-.0402768026076231,-.0377461140742206,-.0328437939425454,-.0258780658927918,-.0172876770198974,-.00762630682441431,-.00348372491482343,-.00789707423793068,-.0118211953667475,-.015003165475677,-.0172425632833398,-.0183985910826998,-.0183985910826998,-.0172425632833398,-.015003165475677,-.0118211953667475,-.00789707423793069,-.00348372491482343,.00348372491482344,.00789707423793069,.0118211953667475,.015003165475677,.0172425632833398,.0183985910826999,.0183985910826999,.0172425632833398,.015003165475677,.0118211953667475,.00789707423793069,.00348372491482344,.00762630682441434,.0172876770198975,.0258780658927919,.0328437939425455,.0377461140742208,.0402768026076233,.0402768026076233,.0377461140742208,.0328437939425455,.0258780658927919,.0172876770198975,.00762630682441434,.00581547706554893,.0131828015237968,.0197334440070673,.0250451935407354,.0287834814106213,.0307132701622271,.0307132701622271,.0287834814106213,.0250451935407354,.0197334440070673,.0131828015237968,.00581547706554893,.000170733382645952,.000387026596707175,.000579343295245729,.000735288019200309,.000845038350279932,.000901693953533124,.000901693953533124,.000845038350279932,.000735288019200309,.000579343295245729,.000387026596707175,.000170733382645952,-.00401695848624301,-.00910583359827495,-.0136305971932344,-.0172996130152496,-.019881782459596,-.0212147567306737,-.0212147567306737,-.019881782459596,-.0172996130152496,-.0136305971932344,-.00910583359827495,-.00401695848624301,-.00321595530276233,-.00729008127585417,-.0109125826103316,-.0138499768923831,-.0159172478252584,-.0169844198386609,-.0169844198386609,-.0159172478252584,-.0138499768923831,-.0109125826103316,-.00729008127585417,-.00321595530276233},
			{.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954},
			{-.00267542524640207,-.00606478189450901,-.00907842189039497,-.0115221059845388,-.0132419149757324,-.0141297192758912,-.0141297192758912,-.0132419149757324,-.0115221059845388,-.00907842189039497,-.00606478189450901,-.00267542524640207,-.00558628145387762,-.0126632499505153,-.0189557229845952,-.0240581294721656,-.0276490864927876,-.0295028197276516,-.0295028197276516,-.0276490864927876,-.0240581294721656,-.0189557229845952,-.0126632499505153,-.00558628145387762,-.00712080440781243,-.0161417799674716,-.0241627631719636,-.0306667746341718,-.0352441491885744,-.0376070934653196,-.0376070934653196,-.0352441491885744,-.0306667746341718,-.0241627631719636,-.0161417799674716,-.00712080440781243,-.00689426560092869,-.0156282509663382,-.0233940573872652,-.029691152479886,-.0341229040244673,-.0364106744098191,-.0364106744098191,-.0341229040244673,-.029691152479886,-.0233940573872652,-.0156282509663382,-.00689426560092869,-.00496229500775375,-.0112487676337471,-.0168383727729163,-.0213708415448263,-.0245606894326591,-.0262073610608292,-.0262073610608292,-.0245606894326591,-.0213708415448263,-.0168383727729163,-.0112487676337471,-.00496229500775375,-.00180275455686753,-.00408657024203424,-.006117220600383,-.0077638233758434,-.00892266476002541,-.00952088489339305,-.00952088489339305,-.00892266476002541,-.0077638233758434,-.006117220600383,-.00408657024203424,-.00180275455686753,.00180275455686753,.00408657024203424,.006117220600383,.0077638233758434,.00892266476002541,.00952088489339305,.00952088489339305,.00892266476002541,.0077638233758434,.006117220600383,.00408657024203424,.00180275455686753,.00496229500775375,.0112487676337471,.0168383727729163,.0213708415448263,.0245606894326591,.0262073610608292,.0262073610608292,.0245606894326591,.0213708415448263,.0168383727729163,.0112487676337471,.00496229500775375,.00689426560092869,.0156282509663382,.0233940573872652,.029691152479886,.0341229040244673,.0364106744098191,.0364106744098191,.0341229040244673,.029691152479886,.0233940573872652,.0156282509663382,.00689426560092869,.00712080440781243,.0161417799674716,.0241627631719636,.0306667746341718,.0352441491885744,.0376070934653196,.0376070934653196,.0352441491885744,.0306667746341718,.0241627631719636,.0161417799674716,.00712080440781243,.00558628145387762,.0126632499505153,.0189557229845952,.0240581294721656,.0276490864927876,.0295028197276516,.0295028197276516,.0276490864927876,.0240581294721656,.0189557229845952,.0126632499505153,.00558628145387762,.00267542524640207,.00606478189450901,.00907842189039497,.0115221059845388,.0132419149757324,.0141297192758912,.0141297192758912,.0132419149757324,.0115221059845388,.00907842189039497,.00606478189450901,.00267542524640207},
			{.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085,.00579220598764257,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001756,.0305903686539505,.0305903686539505,.0286683021001756,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764257,.00464629682613013,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613013,.000263892411929397,.00059820393939992,.000895456396096566,.0011364908598568,.00130612540425466,.00139369459289309,.00139369459289309,.00130612540425466,.0011364908598568,.000895456396096566,.00059820393939992,.000263892411929396,-.0051735371064033,-.011727621325126,-.0175551728012975,-.0222805862115142,-.0256062241241532,-.0273229936344479,-.0273229936344479,-.0256062241241532,-.0222805862115142,-.0175551728012975,-.011727621325126,-.0051735371064033,-.00885484154985533,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985533,-.00885484154985534,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985534,-.0051735371064033,-.011727621325126,-.0175551728012976,-.0222805862115143,-.0256062241241533,-.027322993634448,-.027322993634448,-.0256062241241533,-.0222805862115143,-.0175551728012976,-.011727621325126,-.0051735371064033,.000263892411929384,.000598203939399891,.000895456396096523,.00113649085985675,.0013061254042546,.00139369459289302,.00139369459289302,.0013061254042546,.00113649085985675,.000895456396096523,.000598203939399891,.000263892411929384,.00464629682613012,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613012,.00579220598764256,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001755,.0305903686539505,.0305903686539505,.0286683021001755,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764256,.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085},
			{-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748},
			{.00299121635954409,.0067806322917519,.0101499842366785,.0128821061114311,.014804911017555,.0157975064003936,.0157975064003936,.014804911017555,.0128821061114311,.0101499842366785,.0067806322917519,.00299121635954408,.0062456525355466,.0141579438513248,.0211931425760292,.0268978064529921,.0309126184537943,.0329851552162446,.0329851552162446,.0309126184537943,.0268978064529921,.0211931425760292,.0141579438513248,.0062456525355466,.00796130135439456,.0180470586407878,.0270147904857234,.0342864963629789,.0394041566935386,.042046008708203,.042046008708203,.0394041566935386,.0342864963629789,.0270147904857234,.0180470586407878,.00796130135439456,.00770802326855251,.017472915763368,.0261553512911662,.0331957176344168,.0381505664904682,.0407083715394964,.0407083715394964,.0381505664904682,.0331957176344168,.0261553512911662,.017472915763368,.0077080232685525,.00554801448032919,.0125765045448471,.0188258730735172,.0238933272129638,.0274596855701536,.0293007204185775,.0293007204185775,.0274596855701536,.0238933272129638,.0188258730735172,.0125765045448471,.0055480144803292,.00201554086775423,.00456892442756064,.00683926054723933,.00868021841599365,.0099758424709522,.0106446729127462,.0106446729127462,.0099758424709522,.00868021841599365,.00683926054723933,.00456892442756064,.00201554086775423,-.00201554086775423,-.00456892442756064,-.00683926054723933,-.00868021841599365,-.0099758424709522,-.0106446729127462,-.0106446729127462,-.0099758424709522,-.00868021841599365,-.00683926054723933,-.00456892442756064,-.00201554086775423,-.00554801448032919,-.0125765045448471,-.0188258730735172,-.0238933272129638,-.0274596855701536,-.0293007204185775,-.0293007204185775,-.0274596855701536,-.0238933272129638,-.0188258730735172,-.0125765045448471,-.0055480144803292,-.00770802326855251,-.017472915763368,-.0261553512911662,-.0331957176344168,-.0381505664904682,-.0407083715394964,-.0407083715394964,-.0381505664904682,-.0331957176344168,-.0261553512911662,-.017472915763368,-.0077080232685525,-.00796130135439456,-.0180470586407878,-.0270147904857234,-.0342864963629789,-.0394041566935386,-.042046008708203,-.042046008708203,-.0394041566935386,-.0342864963629789,-.0270147904857234,-.0180470586407878,-.00796130135439456,-.0062456525355466,-.0141579438513248,-.0211931425760292,-.0268978064529921,-.0309126184537943,-.0329851552162446,-.0329851552162446,-.0309126184537943,-.0268978064529921,-.0211931425760292,-.0141579438513248,-.0062456525355466,-.00299121635954408,-.0067806322917519,-.0101499842366785,-.0128821061114311,-.014804911017555,-.0157975064003935,-.0157975064003935,-.014804911017555,-.0128821061114311,-.0101499842366785,-.0067806322917519,-.00299121635954409},
			{.00192735045455837,.0043690101814277,.00654000726824865,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824865,.0043690101814277,.00192735045455837,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00192735045455837,.0043690101814277,.00654000726824864,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824864,.0043690101814277,.00192735045455837},
			{.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202},
			{-.003216292833121,-.00671561166886112,-.00856035585883146,-.00828801966032105,-.00596547927875537,-.00216720185657465,.00216720185657465,.00596547927875537,.00828801966032105,.00856035585883146,.00671561166886112,.003216292833121,-.00671561166886112,-.0140221809477405,-.0178740023617151,-.0173053650368621,-.0124559063286174,-.0045251122431745,.0045251122431745,.0124559063286174,.0173053650368621,.0178740023617151,.0140221809477405,.00671561166886112,-.00856035585883146,-.0178740023617151,-.0227838994245812,-.0220590603339104,-.0158774801127414,-.00576813744045716,.00576813744045716,.0158774801127414,.0220590603339104,.0227838994245812,.0178740023617151,.00856035585883146,-.00828801966032105,-.0173053650368621,-.0220590603339104,-.0213572810231997,-.0153723594557108,-.00558463191230799,.00558463191230799,.0153723594557108,.0213572810231997,.0220590603339104,.0173053650368621,.00828801966032105,-.00596547927875537,-.0124559063286174,-.0158774801127414,-.0153723594557108,-.0110645842501621,-.00401965817140192,.00401965817140192,.0110645842501621,.0153723594557108,.0158774801127414,.0124559063286174,.00596547927875537,-.00216720185657465,-.0045251122431745,-.00576813744045716,-.00558463191230799,-.00401965817140192,-.00146030356402064,.00146030356402064,.00401965817140193,.00558463191230799,.00576813744045716,.0045251122431745,.00216720185657466,.00216720185657465,.0045251122431745,.00576813744045716,.00558463191230799,.00401965817140192,.00146030356402064,-.00146030356402064,-.00401965817140193,-.00558463191230799,-.00576813744045716,-.0045251122431745,-.00216720185657466,.00596547927875537,.0124559063286174,.0158774801127414,.0153723594557108,.0110645842501621,.00401965817140193,-.00401965817140193,-.0110645842501621,-.0153723594557108,-.0158774801127414,-.0124559063286174,-.00596547927875537,.00828801966032105,.0173053650368621,.0220590603339104,.0213572810231997,.0153723594557108,.00558463191230799,-.00558463191230799,-.0153723594557108,-.0213572810231997,-.0220590603339104,-.0173053650368621,-.00828801966032105,.00856035585883146,.0178740023617151,.0227838994245812,.0220590603339104,.0158774801127414,.00576813744045716,-.00576813744045716,-.0158774801127414,-.0220590603339104,-.0227838994245812,-.0178740023617151,-.00856035585883146,.00671561166886112,.0140221809477405,.0178740023617151,.0173053650368621,.0124559063286174,.0045251122431745,-.0045251122431745,-.0124559063286174,-.0173053650368621,-.0178740023617151,-.0140221809477405,-.00671561166886112,.003216292833121,.00671561166886112,.00856035585883146,.00828801966032105,.00596547927875537,.00216720185657466,-.00216720185657466,-.00596547927875537,-.00828801966032105,-.00856035585883146,-.00671561166886112,-.003216292833121},
			{.0039983685679321,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316192,.00269417998711382,-.00269417998711382,-.00741604887316192,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.0039983685679321,.00696316618552943,.0145390741807499,.0185328835230693,.0179432847809911,.0129150626978197,.00469191938293559,-.00469191938293559,-.0129150626978197,-.0179432847809911,-.0185328835230693,-.0145390741807499,-.00696316618552943,.00558559847779338,.011662716133544,.0148664046264906,.0143934499750413,.0103599932305885,.00376368695863054,-.00376368695863054,-.0103599932305885,-.0143934499750413,-.0148664046264906,-.011662716133544,-.00558559847779338,.000317241258906341,.000662398982523078,.000844356596319848,.000817494527800616,.000588409157550378,.000213763447844004,-.000213763447844004,-.000588409157550378,-.000817494527800616,-.000844356596319848,-.000662398982523078,-.000317241258906341,-.00621942636635253,-.0129861472342896,-.0165533754841949,-.0160267521257476,-.0115355973598394,-.00419077275215303,.00419077275215303,.0115355973598394,.0160267521257476,.0165533754841949,.0129861472342895,.00621942636635253,-.0106449482959891,-.0222266263366319,-.0283321669513853,-.0274308171979002,-.019743916918636,-.00717277714038947,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.0106449482959891,-.0222266263366319,-.0283321669513854,-.0274308171979002,-.019743916918636,-.00717277714038948,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.00621942636635254,-.0129861472342896,-.0165533754841949,-.0160267521257477,-.0115355973598394,-.00419077275215304,.00419077275215304,.0115355973598394,.0160267521257477,.0165533754841949,.0129861472342896,.00621942636635254,.000317241258906326,.000662398982523046,.000844356596319808,.000817494527800577,.000588409157550351,.000213763447843993,-.000213763447843993,-.000588409157550351,-.000817494527800577,-.000844356596319808,-.000662398982523046,-.000317241258906326,.00558559847779337,.0116627161335439,.0148664046264906,.0143934499750412,.0103599932305885,.00376368695863053,-.00376368695863053,-.0103599932305885,-.0143934499750412,-.0148664046264906,-.0116627161335439,-.00558559847779337,.00696316618552942,.0145390741807499,.0185328835230693,.017943284780991,.0129150626978197,.00469191938293558,-.00469191938293558,-.0129150626978197,-.017943284780991,-.0185328835230693,-.0145390741807499,-.00696316618552942,.00399836856793211,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316193,.00269417998711382,-.00269417998711382,-.00741604887316193,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.00399836856793211},
			{-.00446418449222142,-.00932120643965728,-.0118816941913244,-.0115036940846361,-.008280029669763,-.0030080559891824,.0030080559891824,.008280029669763,.0115036940846361,.0118816941913244,.00932120643965728,.00446418449222142,-.00557608613676323,-.0116428543884434,-.0148410869615597,-.0143689377575399,-.0103423500381766,-.00375727735471369,.00375727735471369,.0103423500381766,.0143689377575399,.0148410869615597,.0116428543884434,.00557608613676323,.000237001216546111,.000494857967838362,.000630792921502108,.00061072509381806,.00043958243844784,.000159696116976629,-.000159696116976629,-.00043958243844784,-.00061072509381806,-.000630792921502108,-.000494857967838362,-.000237001216546111,.00807267517325051,.0168557262678721,.0214859080939843,.0208023628502637,.014972945241251,.00543952852887958,-.00543952852887958,-.014972945241251,-.0208023628502637,-.0214859080939843,-.0168557262678721,-.00807267517325051,.0105863538057353,.0221042811824758,.0281762142089068,.0272798258819948,.0196352376920871,.00713329501153745,-.00713329501153745,-.0196352376920871,-.0272798258819948,-.0281762142089068,-.0221042811824758,-.0105863538057353,.00483588522718639,.0100973166766822,.0128709979423768,.0124615244685555,.00896944855895662,.0032585153113594,-.00325851531135941,-.00896944855895662,-.0124615244685555,-.0128709979423768,-.0100973166766822,-.00483588522718639,-.0048358852271864,-.0100973166766822,-.0128709979423768,-.0124615244685555,-.00896944855895663,-.00325851531135941,.00325851531135941,.00896944855895663,.0124615244685555,.0128709979423768,.0100973166766822,.0048358852271864,-.0105863538057353,-.0221042811824759,-.0281762142089069,-.0272798258819949,-.0196352376920872,-.00713329501153749,.00713329501153749,.0196352376920872,.0272798258819949,.0281762142089069,.0221042811824759,.0105863538057353,-.00807267517325049,-.0168557262678721,-.0214859080939843,-.0208023628502637,-.014972945241251,-.00543952852887957,.00543952852887957,.014972945241251,.0208023628502637,.0214859080939843,.0168557262678721,.00807267517325049,-.000237001216546102,-.000494857967838343,-.000630792921502084,-.000610725093818036,-.000439582438447823,-.000159696116976623,.000159696116976623,.000439582438447823,.000610725093818036,.000630792921502084,.000494857967838343,.000237001216546102,.00557608613676323,.0116428543884434,.0148410869615597,.0143689377575399,.0103423500381766,.00375727735471369,-.00375727735471369,-.0103423500381766,-.0143689377575399,-.0148410869615597,-.0116428543884434,-.00557608613676323,.0044641844922214,.00932120643965724,.0118816941913244,.011503694084636,.00828002966976296,.00300805598918238,-.00300805598918238,-.00828002966976296,-.011503694084636,-.0118816941913244,-.00932120643965724,-.0044641844922214},
			{.00467171851954842,.00975453698756343,.0124340584252153,.012038485593966,.00866495728795332,.00314789653003547,-.00314789653003547,-.00866495728795332,-.012038485593966,-.0124340584252153,-.00975453698756343,-.00467171851954842,.00299723801522752,.00625822573807254,.00797732835137032,.0077235403881935,.00555918325881647,.0020195983787005,-.0020195983787005,-.00555918325881647,-.0077235403881935,-.00797732835137032,-.00625822573807254,-.00299723801522752,-.00598247350841671,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016646,.00403110922016646,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841671,-.00972729866058341,-.0203105761138428,-.0258897875287452,-.0250661387888921,-.0180418890967918,-.00655444662861003,.00655444662861003,.0180418890967918,.0250661387888921,.0258897875287452,.0203105761138428,.00972729866058341,-.00146872850921155,-.00306670158055196,-.00390910882534958,-.00378474579014224,-.00272415166852724,-.000989658379109305,.000989658379109305,.00272415166852724,.00378474579014224,.00390910882534958,.00306670158055196,.00146872850921155,.00950954380591769,.0198559045031346,.0253102199512128,.02450500937368,.0176380042079596,.00640771909172347,-.00640771909172347,-.0176380042079596,-.02450500937368,-.0253102199512128,-.0198559045031346,-.00950954380591768,.00950954380591779,.0198559045031348,.0253102199512131,.0245050093736803,.0176380042079598,.00640771909172355,-.00640771909172354,-.0176380042079598,-.0245050093736803,-.0253102199512131,-.0198559045031348,-.00950954380591779,-.0014687285092115,-.00306670158055185,-.00390910882534945,-.00378474579014211,-.00272415166852715,-.000989658379109271,.000989658379109271,.00272415166852715,.00378474579014211,.00390910882534945,.00306670158055185,.0014687285092115,-.0097272986605835,-.020310576113843,-.0258897875287454,-.0250661387888923,-.0180418890967919,-.00655444662861008,.00655444662861008,.0180418890967919,.0250661387888923,.0258897875287454,.020310576113843,.0097272986605835,-.00598247350841672,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016647,.00403110922016647,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841672,.00299723801522744,.00625822573807237,.0079773283513701,.00772354038819329,.00555918325881632,.00201959837870045,-.00201959837870045,-.00555918325881632,-.00772354038819329,-.0079773283513701,-.00625822573807237,-.00299723801522744,.00467171851954842,.00975453698756342,.0124340584252153,.012038485593966,.00866495728795331,.00314789653003546,-.00314789653003546,-.00866495728795331,-.012038485593966,-.0124340584252153,-.00975453698756342,-.00467171851954842},
			{-.00231698622849408,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849407,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00231698622849407,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849408},
			{.00393913815260715,.00822491095042321,.0104842519363188,.0101507095738694,.00730619015269516,.00265426935934509,-.00265426935934509,-.00730619015269516,-.0101507095738694,-.0104842519363188,-.00822491095042321,-.00393913815260715,.00822491095042321,.0171735942029902,.0218910927256895,.0211946570783323,.0152553073958586,.00554210801278983,-.00554210801278983,-.0152553073958586,-.0211946570783323,-.0218910927256895,-.0171735942029902,-.00822491095042321,.0104842519363188,.0218910927256895,.0279044639730276,.0270167210140658,.0194458623404232,.00706449674830704,-.00706449674830704,-.0194458623404232,-.0270167210140658,-.0279044639730276,-.0218910927256895,-.0104842519363188,.0101507095738694,.0211946570783323,.0270167210140658,.026157220402348,.0188272184062363,.00683974929381445,-.00683974929381445,-.0188272184062363,-.026157220402348,-.0270167210140658,-.0211946570783323,-.0101507095738694,.00730619015269516,.0152553073958586,.0194458623404232,.0188272184062363,.0135512928156657,.00492305573060758,-.00492305573060758,-.0135512928156657,-.0188272184062363,-.0194458623404232,-.0152553073958586,-.00730619015269516,.00265426935934509,.00554210801278983,.00706449674830704,.00683974929381445,.00492305573060758,.00178849930086745,-.00178849930086745,-.00492305573060758,-.00683974929381445,-.00706449674830704,-.00554210801278983,-.00265426935934509,-.00265426935934509,-.00554210801278983,-.00706449674830704,-.00683974929381445,-.00492305573060758,-.00178849930086745,.00178849930086745,.00492305573060758,.00683974929381445,.00706449674830704,.00554210801278983,.00265426935934509,-.00730619015269516,-.0152553073958586,-.0194458623404232,-.0188272184062363,-.0135512928156657,-.00492305573060758,.00492305573060758,.0135512928156657,.0188272184062363,.0194458623404232,.0152553073958586,.00730619015269516,-.0101507095738694,-.0211946570783323,-.0270167210140658,-.026157220402348,-.0188272184062363,-.00683974929381445,.00683974929381445,.0188272184062363,.026157220402348,.0270167210140658,.0211946570783323,.0101507095738694,-.0104842519363188,-.0218910927256895,-.0279044639730276,-.0270167210140658,-.0194458623404232,-.00706449674830704,.00706449674830704,.0194458623404232,.0270167210140658,.0279044639730276,.0218910927256895,.0104842519363188,-.00822491095042321,-.0171735942029902,-.0218910927256895,-.0211946570783323,-.0152553073958586,-.00554210801278983,.00554210801278983,.0152553073958586,.0211946570783323,.0218910927256895,.0171735942029902,.00822491095042321,-.00393913815260715,-.00822491095042321,-.0104842519363188,-.0101507095738694,-.00730619015269516,-.00265426935934509,.00265426935934509,.00730619015269516,.0101507095738694,.0104842519363188,.00822491095042321,.00393913815260715},
			{-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969,-.00852810207437714,-.0178066565376609,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123493,.00574640420123493,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376609,.00852810207437714,-.00684093308933222,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010758,.00460955630010758,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933222,-.000388539604839473,-.000811269756660397,-.00103412141096871,-.0010012222303148,-.000720651097989894,-.000261805686438008,.000261805686438008,.000720651097989894,.0010012222303148,.00103412141096871,.000811269756660397,.000388539604839473,.00761721054519024,.0159047172243371,.0202736617284932,.0196286824710797,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710797,-.0202736617284932,-.0159047172243371,-.00761721054519024,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632954,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632955,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.00761721054519025,.0159047172243371,.0202736617284932,.0196286824710798,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710798,-.0202736617284932,-.0159047172243371,-.00761721054519026,-.000388539604839454,-.000811269756660358,-.00103412141096866,-.00100122223031475,-.00072065109798986,-.000261805686437995,.000261805686437995,.00072065109798986,.00100122223031475,.00103412141096866,.000811269756660358,.000388539604839454,-.00684093308933221,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010757,.00460955630010757,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933221,-.00852810207437712,-.0178066565376608,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123492,.00574640420123492,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376608,.00852810207437712,-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969},
			{.00546748706207169,.0114160997827322,.0145520440250068,.0140890903329313,.0101409238735274,.00368410114579698,-.00368410114579698,-.0101409238735274,-.0140890903329313,-.0145520440250068,-.0114160997827322,-.00546748706207169,.0068292828987853,.0142595262013292,.0181765451429699,.0175982828267856,.0126667401680366,.00460170617081502,-.00460170617081502,-.0126667401680366,-.0175982828267856,-.0181765451429699,-.0142595262013292,-.0068292828987853,-.000290266024493157,-.000606074758208074,-.000772560395559055,-.000747982426521798,-.00053837633707014,-.000195587000258209,.000195587000258209,.00053837633707014,.000747982426521798,.000772560395559055,.000606074758208074,.000290266024493157,-.00988696751735083,-.0206439643012052,-.0263147557466346,-.0254775872149812,-.0183380378947805,-.00666203466887181,.00666203466887181,.0183380378947805,.0254775872149812,.0263147557466346,.0206439643012052,.00988696751735083,-.0129655825309695,-.0270721050154096,-.0345086738493417,-.0334108268431253,-.0240481566631596,-.00873646648194734,.00873646648194734,.0240481566631596,.0334108268431253,.0345086738493417,.0270721050154096,.0129655825309695,-.00592272563093564,-.0123666368152112,-.0157636887204181,-.0152621881833591,-.0109852861223506,-.00399084991614105,.00399084991614105,.0109852861223506,.0152621881833591,.0157636887204181,.0123666368152112,.00592272563093564,.00592272563093565,.0123666368152113,.0157636887204182,.0152621881833591,.0109852861223506,.00399084991614106,-.00399084991614106,-.0109852861223506,-.0152621881833591,-.0157636887204182,-.0123666368152113,-.00592272563093565,.0129655825309696,.0270721050154097,.0345086738493419,.0334108268431254,.0240481566631597,.00873646648194738,-.00873646648194738,-.0240481566631597,-.0334108268431254,-.0345086738493419,-.0270721050154097,-.0129655825309696,.00988696751735081,.0206439643012052,.0263147557466346,.0254775872149811,.0183380378947805,.00666203466887179,-.00666203466887179,-.0183380378947805,-.0254775872149811,-.0263147557466346,-.0206439643012052,-.00988696751735081,.000290266024493146,.000606074758208051,.000772560395559025,.00074798242652177,.000538376337070119,.000195587000258202,-.000195587000258202,-.000538376337070119,-.00074798242652177,-.000772560395559025,-.000606074758208051,-.000290266024493146,-.0068292828987853,-.0142595262013292,-.0181765451429698,-.0175982828267856,-.0126667401680366,-.00460170617081502,.00460170617081502,.0126667401680366,.0175982828267856,.0181765451429698,.0142595262013292,.0068292828987853,-.00546748706207167,-.0114160997827322,-.0145520440250067,-.0140890903329312,-.0101409238735273,-.00368410114579696,.00368410114579696,.0101409238735273,.0140890903329312,.0145520440250067,.0114160997827322,.00546748706207167},
			{.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207},
			{-.00454852494396312,-.00949730909922885,-.0121061713511056,-.0117210298057487,-.00843646169984582,-.00306488625715946,.00306488625715946,.00843646169984582,.0117210298057487,.0121061713511056,.00949730909922885,.00454852494396312,-.00949730909922885,-.0198303584651126,-.0252776565471576,-.024473481930491,-.0176153116569324,-.00639947510386962,.00639947510386962,.0176153116569324,.024473481930491,.0252776565471576,.0198303584651126,.00949730909922885,-.0121061713511056,-.0252776565471576,-.0322212995614867,-.0311962222891923,-.0224541477058241,-.00815737819577448,.00815737819577448,.0224541477058241,.0311962222891923,.0322212995614867,.0252776565471576,.0121061713511056,-.0117210298057487,-.024473481930491,-.0311962222891923,-.0302037564704542,-.0217397992222051,-.00789786218916395,.00789786218916395,.0217397992222051,.0302037564704542,.0311962222891923,.024473481930491,.0117210298057487,-.00843646169984582,-.0176153116569324,-.0224541477058241,-.0217397992222051,-.0156476851044708,-.00568465510060072,.00568465510060072,.0156476851044708,.0217397992222051,.0224541477058241,.0176153116569324,.00843646169984582,-.00306488625715946,-.00639947510386962,-.00815737819577448,-.00789786218916395,-.00568465510060072,-.00206518110487492,.00206518110487492,.00568465510060072,.00789786218916395,.00815737819577448,.00639947510386962,.00306488625715946,.00306488625715946,.00639947510386962,.00815737819577448,.00789786218916395,.00568465510060072,.00206518110487492,-.00206518110487492,-.00568465510060072,-.00789786218916395,-.00815737819577448,-.00639947510386962,-.00306488625715946,.00843646169984582,.0176153116569324,.0224541477058241,.0217397992222051,.0156476851044708,.00568465510060072,-.00568465510060072,-.0156476851044708,-.0217397992222051,-.0224541477058241,-.0176153116569324,-.00843646169984582,.0117210298057487,.024473481930491,.0311962222891923,.0302037564704542,.0217397992222051,.00789786218916395,-.00789786218916395,-.0217397992222051,-.0302037564704542,-.0311962222891923,-.024473481930491,-.0117210298057487,.0121061713511056,.0252776565471576,.0322212995614867,.0311962222891923,.0224541477058241,.00815737819577448,-.00815737819577448,-.0224541477058241,-.0311962222891923,-.0322212995614867,-.0252776565471576,-.0121061713511056,.00949730909922885,.0198303584651126,.0252776565471576,.024473481930491,.0176153116569324,.00639947510386962,-.00639947510386962,-.0176153116569324,-.024473481930491,-.0252776565471576,-.0198303584651126,-.00949730909922885,.00454852494396312,.00949730909922885,.0121061713511056,.0117210298057487,.00843646169984582,.00306488625715946,-.00306488625715946,-.00843646169984582,-.0117210298057487,-.0121061713511056,-.00949730909922885,-.00454852494396312},
			{.00565454705638537,.0118066805991472,.015049915394736,.0145711225952196,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952196,-.015049915394736,-.0118066805991472,-.00565454705638537,.00984740405706795,.0205613558916723,.0262094552293619,.02537563669192,.0182646568269614,.00663537602520149,-.00663537602520149,-.0182646568269614,-.02537563669192,-.0262094552293619,-.0205613558916723,-.00984740405706795,.00789922912161445,.0164935713308931,.0210242710474368,.0203554121649402,.0146512429334382,.00532265714165691,-.00532265714165691,-.0146512429334382,-.0203554121649402,-.0210242710474368,-.0164935713308931,-.00789922912161445,.000448646890909458,.000936773624827609,.0011941005500474,.00115611184843243,.000832136210868989,.000302307167093963,-.000302307167093963,-.000832136210868989,-.00115611184843243,-.0011941005500474,-.000936773624827609,-.000448646890909458,-.00879559711786466,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317994,.00592664766317994,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786466,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.00879559711786467,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317995,.00592664766317995,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786467,.000448646890909437,.000936773624827564,.00119410055004734,.00115611184843238,.000832136210868949,.000302307167093948,-.000302307167093948,-.000832136210868949,-.00115611184843238,-.00119410055004734,-.000936773624827564,-.000448646890909437,.00789922912161443,.0164935713308931,.0210242710474368,.0203554121649401,.0146512429334382,.0053226571416569,-.0053226571416569,-.0146512429334382,-.0203554121649401,-.0210242710474368,-.0164935713308931,-.00789922912161443,.00984740405706793,.0205613558916723,.0262094552293618,.0253756366919199,.0182646568269614,.00663537602520148,-.00663537602520148,-.0182646568269614,-.0253756366919199,-.0262094552293618,-.0205613558916723,-.00984740405706793,.00565454705638537,.0118066805991472,.015049915394736,.0145711225952197,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952197,-.015049915394736,-.0118066805991472,-.00565454705638537},
			{-.00299121635954409,-.0062456525355466,-.00796130135439456,-.00770802326855251,-.00554801448032919,-.00201554086775423,.00201554086775423,.00554801448032919,.00770802326855251,.00796130135439456,.0062456525355466,.00299121635954408,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.00299121635954408,-.0062456525355466,-.00796130135439456,-.0077080232685525,-.0055480144803292,-.00201554086775423,.00201554086775423,.0055480144803292,.0077080232685525,.00796130135439456,.0062456525355466,.00299121635954409},
			{.00508540548709832,.0106183143768374,.0135351110470164,.013104509708737,.00943225092720051,.00342664700787828,-.00342664700787828,-.00943225092720051,-.013104509708737,-.0135351110470164,-.0106183143768374,-.00508540548709832,.0106183143768374,.0221710147777587,.0282612791816195,.0273621846271071,.0196945171590198,.0071548306777916,-.0071548306777916,-.0196945171590198,-.0273621846271071,-.0282612791816195,-.0221710147777587,-.0106183143768374,.0135351110470164,.0282612791816195,.03602450807902,.0348784368472584,.0251045003288081,.00912022608388363,-.00912022608388363,-.0251045003288081,-.0348784368472584,-.03602450807902,-.0282612791816195,-.0135351110470164,.013104509708737,.0273621846271071,.0348784368472584,.0337688263290031,.0243058344441419,.00883007836780732,-.00883007836780732,-.0243058344441419,-.0337688263290031,-.0348784368472584,-.0273621846271071,-.013104509708737,.00943225092720051,.0196945171590198,.0251045003288081,.0243058344441419,.0174946437957377,.00635563761813038,-.00635563761813038,-.0174946437957377,-.0243058344441419,-.0251045003288081,-.0196945171590198,-.00943225092720051,.00342664700787828,.0071548306777916,.00912022608388363,.00883007836780732,.00635563761813038,.00230894266866042,-.00230894266866042,-.00635563761813038,-.00883007836780732,-.00912022608388363,-.0071548306777916,-.00342664700787828,-.00342664700787828,-.0071548306777916,-.00912022608388363,-.00883007836780732,-.00635563761813038,-.00230894266866042,.00230894266866042,.00635563761813038,.00883007836780732,.00912022608388363,.0071548306777916,.00342664700787828,-.00943225092720051,-.0196945171590198,-.0251045003288081,-.0243058344441419,-.0174946437957377,-.00635563761813038,.00635563761813038,.0174946437957377,.0243058344441419,.0251045003288081,.0196945171590198,.00943225092720051,-.013104509708737,-.0273621846271071,-.0348784368472584,-.0337688263290031,-.0243058344441419,-.00883007836780732,.00883007836780732,.0243058344441419,.0337688263290031,.0348784368472584,.0273621846271071,.013104509708737,-.0135351110470164,-.0282612791816195,-.03602450807902,-.0348784368472584,-.0251045003288081,-.00912022608388363,.00912022608388363,.0251045003288081,.0348784368472584,.03602450807902,.0282612791816195,.0135351110470164,-.0106183143768374,-.0221710147777587,-.0282612791816195,-.0273621846271071,-.0196945171590198,-.0071548306777916,.0071548306777916,.0196945171590198,.0273621846271071,.0282612791816195,.0221710147777587,.0106183143768374,-.00508540548709832,-.0106183143768374,-.0135351110470164,-.013104509708737,-.00943225092720051,-.00342664700787828,.00342664700787828,.00943225092720051,.013104509708737,.0135351110470164,.0106183143768374,.00508540548709832},
			{.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600146,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600146,-.00872116867787892,-.00684176956018095,-.00327671334903246,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600145,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600145,-.00872116867787892,-.00684176956018095,-.00327671334903246},
			{-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847933,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847885,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847932,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847886,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.0028803860195172},
			{.00489698139750969,.00852810207437714,.00684093308933222,.000388539604839473,-.00761721054519024,-.0130373458317453,-.0130373458317453,-.00761721054519025,.000388539604839454,.00684093308933221,.00852810207437712,.00489698139750969,.0102248853327823,.0178066565376609,.0142838517710583,.000811269756660397,-.0159047172243371,-.0272219466141356,-.0272219466141356,-.0159047172243371,.000811269756660358,.0142838517710583,.0178066565376608,.0102248853327823,.0130336090560768,.0226980540469838,.0182075528223322,.00103412141096871,-.0202736617284932,-.0346996761691302,-.0346996761691302,-.0202736617284932,.00103412141096866,.0182075528223322,.0226980540469838,.0130336090560768,.0126189623285648,.0219759460114443,.0176283040385686,.0010012222303148,-.0196286824710797,-.0335957526812189,-.0335957526812189,-.0196286824710798,.00100122223031475,.0176283040385686,.0219759460114443,.0126189623285648,.00908276782339724,.0158176568028604,.0126883485768188,.000720651097989894,-.014128163704906,-.0241812609872885,-.0241812609872885,-.014128163704906,.00072065109798986,.0126883485768188,.0158176568028604,.00908276782339724,.00329968312182452,.00574640420123493,.00460955630010758,.000261805686438008,-.00513262743536863,-.00878482201632954,-.00878482201632955,-.00513262743536863,.000261805686437995,.00460955630010757,.00574640420123492,.00329968312182452,-.00329968312182452,-.00574640420123493,-.00460955630010758,-.000261805686438008,.00513262743536863,.00878482201632955,.00878482201632955,.00513262743536863,-.000261805686437995,-.00460955630010757,-.00574640420123492,-.00329968312182452,-.00908276782339724,-.0158176568028604,-.0126883485768188,-.000720651097989894,.014128163704906,.0241812609872885,.0241812609872885,.014128163704906,-.00072065109798986,-.0126883485768188,-.0158176568028604,-.00908276782339724,-.0126189623285648,-.0219759460114443,-.0176283040385686,-.0010012222303148,.0196286824710797,.0335957526812189,.0335957526812189,.0196286824710798,-.00100122223031475,-.0176283040385686,-.0219759460114443,-.0126189623285648,-.0130336090560768,-.0226980540469838,-.0182075528223322,-.00103412141096871,.0202736617284932,.0346996761691302,.0346996761691302,.0202736617284932,-.00103412141096866,-.0182075528223322,-.0226980540469838,-.0130336090560768,-.0102248853327823,-.0178066565376609,-.0142838517710583,-.000811269756660397,.0159047172243371,.0272219466141356,.0272219466141356,.0159047172243371,-.000811269756660358,-.0142838517710583,-.0178066565376608,-.0102248853327823,-.00489698139750969,-.00852810207437714,-.00684093308933222,-.000388539604839473,.00761721054519024,.0130373458317453,.0130373458317453,.00761721054519026,-.000388539604839454,-.00684093308933221,-.00852810207437712,-.00489698139750969},
			{-.00608773439400276,-.0106018005990698,-.00850437857006723,-.000483017133170316,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170293,-.00850437857006722,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.0184630551643588,-.0148103908586545,-.000841175222238799,.0164910055329813,.0282254167678824,.0282254167678824,.0164910055329813,-.000841175222238758,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00850437857006723,-.0148103908586545,-.0118803564975286,-.000674760147281664,.0132284844204532,.0226414020193,.0226414020193,.0132284844204532,-.000674760147281631,-.0118803564975285,-.0148103908586545,-.00850437857006724,-.000483017133170316,-.000841175222238799,-.000674760147281664,-3.83238715483232e-005,.000751328808837938,.00128594758620117,.00128594758620117,.000751328808837939,-3.83238715483214e-005,-.000674760147281662,-.000841175222238797,-.000483017133170317,.0094694161276367,.0164910055329813,.0132284844204532,.000751328808837938,-.0147295916665949,-.0252106436310925,-.0252106436310925,-.0147295916665949,.000751328808837902,.0132284844204532,.0164910055329813,.0094694161276367,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710571,-.0431496382710572,-.0252106436310926,.0012859475862011,.0226414020192999,.0282254167678823,.0162075148308275,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710572,-.0431496382710572,-.0252106436310926,.00128594758620111,.0226414020192999,.0282254167678823,.0162075148308275,.00946941612763672,.0164910055329813,.0132284844204532,.000751328808837939,-.0147295916665949,-.0252106436310926,-.0252106436310926,-.0147295916665949,.000751328808837903,.0132284844204532,.0164910055329813,.00946941612763672,-.000483017133170293,-.000841175222238758,-.000674760147281631,-3.83238715483214e-005,.000751328808837902,.0012859475862011,.00128594758620111,.000751328808837903,-3.83238715483196e-005,-.00067476014728163,-.000841175222238757,-.000483017133170294,-.00850437857006722,-.0148103908586545,-.0118803564975285,-.000674760147281662,.0132284844204532,.0226414020192999,.0226414020192999,.0132284844204532,-.00067476014728163,-.0118803564975285,-.0148103908586545,-.00850437857006722,-.0106018005990698,-.0184630551643587,-.0148103908586545,-.000841175222238797,.0164910055329813,.0282254167678823,.0282254167678823,.0164910055329813,-.000841175222238757,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.00850437857006724,-.000483017133170317,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170294,-.00850437857006722,-.0106018005990698,-.00608773439400276},
			{.00679696456586262,.0118369262425784,.00949515140679809,.000539289352389854,-.0105726172847424,-.0180957145755819,-.0180957145755819,-.0105726172847425,.000539289352389828,.00949515140679807,.0118369262425784,.00679696456586262,.00848989551256602,.0147851685874852,.0118601241096038,.000673611022753388,-.0132059561546973,-.0226028434433086,-.0226028434433086,-.0132059561546973,.000673611022753355,.0118601241096038,.0147851685874851,.00848989551256602,-.00036084728884685,-.00062841621455069,-.000504092615039049,-2.86305892620399e-005,.000561294714169291,.000960692009068383,.000960692009068383,.000561294714169292,-2.86305892620386e-005,-.000504092615039048,-.000628416214550689,-.00036084728884685,-.0122910885963404,-.0214049533061549,-.017170274472633,-.000975207851248453,.0191186778278397,.0327228469278241,.0327228469278241,.0191186778278397,-.000975207851248406,-.017170274472633,-.0214049533061549,-.0122910885963405,-.0161183015228526,-.0280700516286174,-.0225167737593592,-.00127886916367651,.0250718731324701,.0429121236361326,.0429121236361326,.0250718731324701,-.00127886916367645,-.0225167737593592,-.0280700516286173,-.0161183015228526,-.00736289922404356,-.0128225024865171,-.0102857447979673,-.000584192122199518,.0114529235583944,.0196024153893991,.0196024153893991,.0114529235583944,-.00058419212219949,-.0102857447979673,-.012822502486517,-.00736289922404357,.00736289922404357,.0128225024865171,.0102857447979674,.000584192122199519,-.0114529235583944,-.0196024153893992,-.0196024153893992,-.0114529235583944,.000584192122199491,.0102857447979673,.012822502486517,.00736289922404358,.0161183015228527,.0280700516286175,.0225167737593594,.00127886916367651,-.0250718731324702,-.0429121236361328,-.0429121236361328,-.0250718731324703,.00127886916367645,.0225167737593593,.0280700516286175,.0161183015228527,.0122910885963404,.0214049533061549,.017170274472633,.000975207851248451,-.0191186778278397,-.032722846927824,-.032722846927824,-.0191186778278397,.000975207851248404,.0171702744726329,.0214049533061548,.0122910885963404,.000360847288846836,.000628416214550666,.00050409261503903,2.86305892620388e-005,-.000561294714169269,-.000960692009068346,-.000960692009068346,-.00056129471416927,2.86305892620375e-005,.000504092615039029,.000628416214550665,.000360847288846836,-.00848989551256602,-.0147851685874852,-.0118601241096038,-.000673611022753387,.0132059561546973,.0226028434433086,.0226028434433086,.0132059561546973,-.000673611022753355,-.0118601241096038,-.0147851685874851,-.00848989551256602,-.00679696456586259,-.0118369262425784,-.00949515140679805,-.000539289352389851,.0105726172847424,.0180957145755818,.0180957145755819,.0105726172847424,-.000539289352389826,-.00949515140679803,-.0118369262425784,-.00679696456586259},
			{.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929397,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929396,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085},
			{-.00565454705638537,-.00984740405706795,-.00789922912161445,-.000448646890909458,.00879559711786466,.0150542302516124,.0150542302516124,.00879559711786467,-.000448646890909437,-.00789922912161443,-.00984740405706793,-.00565454705638537,-.0118066805991472,-.0205613558916723,-.0164935713308931,-.000936773624827609,.0183651855425165,.0314331964124508,.0314331964124508,.0183651855425165,-.000936773624827564,-.0164935713308931,-.0205613558916723,-.0118066805991472,-.015049915394736,-.0262094552293619,-.0210242710474368,-.0011941005500474,.0234100081138356,.0400677347558359,.0400677347558359,.0234100081138356,-.00119410055004734,-.0210242710474368,-.0262094552293618,-.015049915394736,-.0145711225952196,-.02537563669192,-.0203554121649402,-.00115611184843243,.0226652502180242,.0387930337099593,.0387930337099593,.0226652502180242,-.00115611184843238,-.0203554121649401,-.0253756366919199,-.0145711225952197,-.0104878768961101,-.0182646568269614,-.0146512429334382,-.000832136210868989,.0163137982370799,.0279221150819346,.0279221150819346,.0163137982370799,-.000832136210868949,-.0146512429334382,-.0182646568269614,-.0104878768961101,-.00381014587741865,-.00663537602520149,-.00532265714165691,-.000302307167093963,.00592664766317994,.0101438387122661,.0101438387122661,.00592664766317995,-.000302307167093948,-.0053226571416569,-.00663537602520148,-.00381014587741865,.00381014587741865,.00663537602520149,.00532265714165691,.000302307167093963,-.00592664766317994,-.0101438387122661,-.0101438387122661,-.00592664766317995,.000302307167093948,.0053226571416569,.00663537602520148,.00381014587741865,.0104878768961101,.0182646568269614,.0146512429334382,.000832136210868989,-.0163137982370799,-.0279221150819346,-.0279221150819346,-.0163137982370799,.000832136210868949,.0146512429334382,.0182646568269614,.0104878768961101,.0145711225952196,.02537563669192,.0203554121649402,.00115611184843243,-.0226652502180242,-.0387930337099593,-.0387930337099593,-.0226652502180242,.00115611184843238,.0203554121649401,.0253756366919199,.0145711225952197,.015049915394736,.0262094552293619,.0210242710474368,.0011941005500474,-.0234100081138356,-.0400677347558359,-.0400677347558359,-.0234100081138356,.00119410055004734,.0210242710474368,.0262094552293618,.015049915394736,.0118066805991472,.0205613558916723,.0164935713308931,.000936773624827609,-.0183651855425165,-.0314331964124508,-.0314331964124508,-.0183651855425165,.000936773624827564,.0164935713308931,.0205613558916723,.0118066805991472,.00565454705638537,.00984740405706795,.00789922912161445,.000448646890909458,-.00879559711786466,-.0150542302516124,-.0150542302516124,-.00879559711786467,.000448646890909437,.00789922912161443,.00984740405706793,.00565454705638537},
			{.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997765,.0122419048552221,.00702951017978813,.0122419048552221,.0213192997308994,.0171015662920701,.000971305481648202,-.0190421729606395,-.0325919039263692,-.0325919039263693,-.0190421729606395,.000971305481648156,.01710156629207,.0213192997308993,.0122419048552221,.00982001050997767,.0171015662920701,.0137182540390003,.000779145905068479,-.015274938076857,-.0261440390921353,-.0261440390921353,-.015274938076857,.000779145905068442,.0137182540390002,.01710156629207,.00982001050997767,.000557740143521656,.000971305481648202,.000779145905068479,4.42525950940344e-005,-.000867559779759337,-.00148488436959081,-.00148488436959081,-.000867559779759338,4.42525950940323e-005,.000779145905068478,.0009713054816482,.000557740143521656,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759337,.017008267421531,.0291107437634532,.0291107437634533,.0170082674215311,-.000867559779759295,-.0152749380768569,-.0190421729606395,-.0109343398968667,-.0187148260943523,-.0325919039263692,-.0261440390921353,-.00148488436959081,.0291107437634532,.0498249105249045,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0187148260943523,-.0325919039263693,-.0261440390921353,-.00148488436959081,.0291107437634533,.0498249105249046,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759338,.0170082674215311,.0291107437634533,.0291107437634533,.0170082674215311,-.000867559779759297,-.015274938076857,-.0190421729606395,-.0109343398968667,.000557740143521629,.000971305481648156,.000779145905068442,4.42525950940323e-005,-.000867559779759295,-.00148488436959074,-.00148488436959074,-.000867559779759297,4.42525950940302e-005,.000779145905068441,.000971305481648154,.000557740143521629,.00982001050997765,.01710156629207,.0137182540390002,.000779145905068478,-.0152749380768569,-.0261440390921352,-.0261440390921352,-.015274938076857,.000779145905068441,.0137182540390002,.01710156629207,.00982001050997766,.0122419048552221,.0213192997308993,.01710156629207,.0009713054816482,-.0190421729606395,-.0325919039263692,-.0325919039263692,-.0190421729606395,.000971305481648154,.01710156629207,.0213192997308993,.0122419048552221,.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997766,.0122419048552221,.00702951017978813},
			{-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471107,-.00842941499026791,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325645,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.052284927874802,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.0522849278748019,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325644,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471108,-.00842941499026791,-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697},
			{.00632197580064374,.011009732437834,.00883160664375298,.000501602473032872,-.00983377653008651,-.0168311410974181,-.0168311410974182,-.00983377653008653,.000501602473032848,.00883160664375296,.011009732437834,.00632197580064374,.0132002702054535,.0229882947439242,.0184403733456156,.0010473447524243,-.0205329016482428,-.0351433819676135,-.0351433819676135,-.0205329016482428,.00104734475242425,.0184403733456155,.0229882947439242,.0132002702054535,.0168263169407732,.0293030617759155,.0235058496220271,.00133504500106871,-.0261731847507426,-.0447970893136278,-.0447970893136278,-.0261731847507427,.00133504500106864,.023505849622027,.0293030617759154,.0168263169407732,.0162910103172928,.0283708243105143,.0227580426576453,.00129257234147053,-.0253405201097541,-.0433719302187032,-.0433719302187032,-.0253405201097541,.00129257234147047,.0227580426576453,.0283708243105143,.0162910103172928,.0117258028408243,.0204205071273958,.0163805875786201,.000930356567112207,-.01823938091645,-.0312178737024467,-.0312178737024467,-.01823938091645,.000930356567112162,.0163805875786201,.0204205071273958,.0117258028408243,.00425987259346659,.00741857592503808,.00595091159541767,.00033798968788685,-.00662619352742934,-.0113411564578213,-.0113411564578213,-.00662619352742935,.000337989687886833,.00595091159541766,.00741857592503807,.00425987259346659,-.00425987259346659,-.00741857592503808,-.00595091159541767,-.00033798968788685,.00662619352742934,.0113411564578213,.0113411564578213,.00662619352742935,-.000337989687886833,-.00595091159541766,-.00741857592503807,-.00425987259346659,-.0117258028408243,-.0204205071273958,-.0163805875786201,-.000930356567112207,.01823938091645,.0312178737024467,.0312178737024467,.01823938091645,-.000930356567112162,-.0163805875786201,-.0204205071273958,-.0117258028408243,-.0162910103172928,-.0283708243105143,-.0227580426576453,-.00129257234147053,.0253405201097541,.0433719302187032,.0433719302187032,.0253405201097541,-.00129257234147047,-.0227580426576453,-.0283708243105143,-.0162910103172928,-.0168263169407732,-.0293030617759155,-.0235058496220271,-.00133504500106871,.0261731847507426,.0447970893136278,.0447970893136278,.0261731847507427,-.00133504500106864,-.023505849622027,-.0293030617759154,-.0168263169407732,-.0132002702054535,-.0229882947439242,-.0184403733456156,-.0010473447524243,.0205329016482428,.0351433819676135,.0351433819676135,.0205329016482428,-.00104734475242425,-.0184403733456155,-.0229882947439242,-.0132002702054535,-.00632197580064374,-.011009732437834,-.00883160664375298,-.000501602473032872,.00983377653008651,.0168311410974181,.0168311410974182,.00983377653008653,-.000501602473032848,-.00883160664375296,-.011009732437834,-.00632197580064374},
			{.00407348097349134,.0070939745774067,.00569052820876414,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876412,.00709397457740669,.00407348097349134,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886327,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886328,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.00407348097349134,.0070939745774067,.00569052820876413,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876413,.00709397457740669,.00407348097349134},
			{.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934987,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934986,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133},
			{-.00631331025337702,-.00788577663899888,.00033517033471411,.0114164867138149,.0149713651270644,.00683897447387038,-.00683897447387039,-.0149713651270645,-.0114164867138149,-.000335170334714097,.00788577663899888,.00631331025337699,-.0131821765636861,-.0164654825796768,.000699834849514613,.0238375966899433,.0312601742324971,.0142797621867037,-.0142797621867037,-.0312601742324972,-.0238375966899433,-.000699834849514586,.0164654825796768,.0131821765636861,-.0168032530681212,-.0209884664598731,.00089207590457249,.0303856626242094,.0398471842676737,.0182023398500645,-.0182023398500645,-.0398471842676739,-.0303856626242093,-.000892075904572456,.0209884664598731,.0168032530681211,-.0162686801907029,-.0203207466521331,.000863695710496402,.0294189836701143,.0385794997386939,.0176232569099964,-.0176232569099964,-.0385794997386941,-.0294189836701142,-.000863695710496369,.0203207466521331,.0162686801907028,-.0117097302549807,-.0146262916897378,.000621663446188857,.0211749422273102,.0277684194425538,.0126847157981637,-.0126847157981637,-.0277684194425539,-.0211749422273102,-.000621663446188833,.0146262916897378,.0117097302549806,-.00425403357597065,-.00531359259224779,.000225844414470279,.00769265501789862,.0100880025489925,.00460823654619044,-.00460823654619044,-.0100880025489926,-.0076926550178986,-.00022584441447027,.00531359259224779,.00425403357597063,.00425403357597065,.00531359259224779,-.000225844414470279,-.00769265501789862,-.0100880025489925,-.00460823654619044,.00460823654619044,.0100880025489926,.0076926550178986,.00022584441447027,-.00531359259224779,-.00425403357597063,.0117097302549807,.0146262916897378,-.000621663446188857,-.0211749422273102,-.0277684194425538,-.0126847157981637,.0126847157981637,.0277684194425539,.0211749422273102,.000621663446188833,-.0146262916897378,-.0117097302549806,.0162686801907029,.0203207466521331,-.000863695710496402,-.0294189836701143,-.0385794997386939,-.0176232569099964,.0176232569099964,.0385794997386941,.0294189836701142,.000863695710496369,-.0203207466521331,-.0162686801907028,.0168032530681212,.0209884664598731,-.00089207590457249,-.0303856626242094,-.0398471842676737,-.0182023398500645,.0182023398500645,.0398471842676739,.0303856626242093,.000892075904572456,-.0209884664598731,-.0168032530681211,.0131821765636861,.0164654825796768,-.000699834849514612,-.0238375966899433,-.0312601742324971,-.0142797621867037,.0142797621867037,.0312601742324972,.0238375966899433,.000699834849514586,-.0164654825796768,-.0131821765636861,.00631331025337702,.00788577663899888,-.00033517033471411,-.0114164867138149,-.0149713651270644,-.00683897447387039,.00683897447387039,.0149713651270645,.0114164867138149,.000335170334714097,-.00788577663899888,-.00631331025337699},
			{.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.00980328691616649,-.00784845864116279,.0136681051009101,.0170724421228125,-.000725632541047572,-.0247163110990939,-.0324125037180548,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990939,.000725632541047544,-.0170724421228125,-.01366810510091,.0109640564380927,.0136948916904454,-.000582076013801829,-.0198265251716479,-.0260001307746002,-.0118769550522377,.0118769550522377,.0260001307746003,.0198265251716478,.000582076013801807,-.0136948916904454,-.0109640564380926,.000622717705357654,.000777819010395326,-3.30597568249531e-005,-.00112607303052629,-.00147671091136544,-.000674566957815848,.000674566957815848,.00147671091136544,.00112607303052629,3.30597568249518e-005,-.000777819010395325,-.000622717705357651,-.0122082068670625,-.0152489246770109,.000648127308443884,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.0289505053954479,-.0220763475742017,-.000648127308443859,.0152489246770109,.0122082068670624,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533856,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533857,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0122082068670625,-.015248924677011,.000648127308443885,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.028950505395448,-.0220763475742018,-.00064812730844386,.0152489246770109,.0122082068670624,.000622717705357624,.000777819010395288,-3.30597568249515e-005,-.00112607303052624,-.00147671091136537,-.000674566957815815,.000674566957815816,.00147671091136537,.00112607303052624,3.30597568249502e-005,-.000777819010395288,-.000622717705357621,.0109640564380927,.0136948916904453,-.000582076013801828,-.0198265251716478,-.0260001307746001,-.0118769550522376,.0118769550522377,.0260001307746002,.0198265251716478,.000582076013801806,-.0136948916904453,-.0109640564380926,.0136681051009101,.0170724421228125,-.000725632541047571,-.0247163110990939,-.0324125037180547,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990938,.000725632541047543,-.0170724421228125,-.01366810510091,.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.0098032869161665,-.0078484586411628},
			{-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084},
			{.007058495447099,.00881656631296387,-.000374731826353174,-.0127640201823163,-.0167384950754978,-.00764620591247183,.00764620591247184,.0167384950754978,.0127640201823163,.00037473182635316,-.00881656631296387,-.00705849544709897,.0147381214487064,.0184089691712473,-.000782439148523992,-.0266512433181539,-.0349499372975651,-.0159652594812031,.0159652594812031,.0349499372975653,.0266512433181539,.000782439148523962,-.0184089691712473,-.0147381214487063,.0187866080578478,.0234658188815225,-.000997371182181883,-.0339722035956245,-.0445505063817578,-.0203508346337808,.0203508346337809,.044550506381758,.0339722035956245,.000997371182181844,-.0234658188815225,-.0187866080578477,.0181889374112355,.0227192854412644,-.000965641160587157,-.0328914236683848,-.043133191990884,-.0197034002242685,.0197034002242685,.0431331919908843,.0328914236683847,.00096564116058712,-.0227192854412643,-.0181889374112354,.0130918764284278,.0163526912438262,-.000695040862629039,-.0236743051276634,-.031046036760756,-.0141819434046015,.0141819434046015,.0310460367607562,.0236743051276633,.000695040862629013,-.0163526912438262,-.0130918764284278,.00475615412876842,.00594077712243872,-.000252501731629378,-.00860064977654093,-.0112787297320448,-.00515216508851936,.00515216508851937,.0112787297320449,.00860064977654092,.000252501731629368,-.00594077712243872,-.00475615412876839,-.00475615412876842,-.00594077712243872,.000252501731629378,.00860064977654093,.0112787297320448,.00515216508851936,-.00515216508851937,-.0112787297320449,-.00860064977654092,-.000252501731629368,.00594077712243872,.00475615412876839,-.0130918764284278,-.0163526912438262,.000695040862629039,.0236743051276634,.031046036760756,.0141819434046015,-.0141819434046015,-.0310460367607562,-.0236743051276633,-.000695040862629013,.0163526912438262,.0130918764284278,-.0181889374112355,-.0227192854412644,.000965641160587157,.0328914236683848,.043133191990884,.0197034002242685,-.0197034002242685,-.0431331919908843,-.0328914236683847,-.00096564116058712,.0227192854412643,.0181889374112354,-.0187866080578478,-.0234658188815225,.000997371182181883,.0339722035956245,.0445505063817578,.0203508346337808,-.0203508346337809,-.044550506381758,-.0339722035956245,-.000997371182181844,.0234658188815225,.0187866080578477,-.0147381214487064,-.0184089691712473,.000782439148523992,.0266512433181539,.0349499372975651,.0159652594812031,-.0159652594812031,-.0349499372975653,-.0266512433181539,-.000782439148523962,.0184089691712473,.0147381214487063,-.007058495447099,-.00881656631296387,.000374731826353174,.0127640201823163,.0167384950754978,.00764620591247183,-.00764620591247184,-.0167384950754978,-.0127640201823163,-.00037473182635316,.00881656631296387,.00705849544709897},
			{.00454804760548224,.00568083717114684,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.00454804760548224,.00568083717114683,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222},
			{-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023209,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698057,.00556381945057915,-.00278749101749906,-.00434479121890393,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023208,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698056,.00556381945057915,-.00278749101749906,-.00434479121890393},
			{.00738663555257316,.00473904940765641,-.00945912116166398,-.015380209620101,-.00232226367616612,.0150359089640387,.0150359089640388,-.00232226367616604,-.0153802096201012,-.00945912116166399,.00473904940765628,.00738663555257316,.0154232771965445,.00989512371934247,-.0197506221436933,-.0321138405466809,-.00484888094803859,.031394941608239,.0313949416082394,-.00484888094803843,-.0321138405466812,-.0197506221436933,.00989512371934219,.0154232771965445,.0196599725865644,.0126132636134246,-.0251760170659189,-.0409353483537532,-.00618084375318826,.0400189715524956,.040018971552496,-.00618084375318805,-.0409353483537535,-.0251760170659189,.0126132636134242,.0196599725865644,.0190345170231114,.0122119896103414,-.0243750759725306,-.0396330453491579,-.0059842085292456,.0387458218422929,.0387458218422933,-.0059842085292454,-.0396330453491583,-.0243750759725306,.0122119896103411,.0190345170231114,.0137005004254646,.00878984051179792,-.0175444818655976,-.0285267314116404,-.00430726198103432,.0278881333311886,.0278881333311889,-.00430726198103417,-.0285267314116406,-.0175444818655976,.00878984051179767,.0137005004254646,.00497726143544053,.00319326541691349,-.006373743314669,-.0103634900715324,-.00156478729132365,.0101314934655784,.0101314934655785,-.0015647872913236,-.0103634900715325,-.006373743314669,.0031932654169134,.00497726143544053,-.00497726143544053,-.00319326541691349,.006373743314669,.0103634900715324,.00156478729132365,-.0101314934655784,-.0101314934655785,.0015647872913236,.0103634900715325,.006373743314669,-.0031932654169134,-.00497726143544053,-.0137005004254646,-.00878984051179792,.0175444818655976,.0285267314116404,.00430726198103432,-.0278881333311886,-.0278881333311889,.00430726198103417,.0285267314116406,.0175444818655976,-.00878984051179767,-.0137005004254646,-.0190345170231114,-.0122119896103414,.0243750759725306,.0396330453491579,.0059842085292456,-.0387458218422929,-.0387458218422933,.0059842085292454,.0396330453491583,.0243750759725306,-.0122119896103411,-.0190345170231114,-.0196599725865644,-.0126132636134246,.0251760170659189,.0409353483537532,.00618084375318826,-.0400189715524956,-.040018971552496,.00618084375318805,.0409353483537535,.0251760170659189,-.0126132636134242,-.0196599725865644,-.0154232771965445,-.00989512371934246,.0197506221436933,.0321138405466809,.00484888094803859,-.031394941608239,-.0313949416082394,.00484888094803843,.0321138405466812,.0197506221436933,-.00989512371934219,-.0154232771965445,-.00738663555257316,-.00473904940765642,.00945912116166398,.015380209620101,.00232226367616612,-.0150359089640386,-.0150359089640388,.00232226367616604,.0153802096201012,.00945912116166399,-.00473904940765628,-.00738663555257316},
			{.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.00740597712796921,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.0074059771279692,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619},
			{.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648773,.00909571150529476,.000203199397318671,-.00473557895480586,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440274,.00068950902635523,-.0160690655456805,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.036404249458381,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.0364042494583809,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440273,.00068950902635523,-.0160690655456805,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648772,.00909571150529476,.000203199397318671,-.00473557895480586}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371},
			{.00373129696748193,.00796011451098884,.0106231924821284,.0112513152129516,.0098479805408067,.00687246202231334,.00311875829553411,-.000484532664068795,-.00310395593642735,-.00420377879992814,-.00367166610627629,-.00183948563288991,.0077909390577761,.016620699876881,.0221812002498267,.0234927189949964,.0205625595883267,.0143496841069869,.00651195978990583,-.00101170303252743,-.00648105249983563,-.00877747998297588,-.00766642996357517,-.00384084155948903,.0099310701835657,.0211863211550268,.0282742625507885,.0299460487897454,.0262109895753331,.0182914689643539,.00830076184995504,-.00128961268294456,-.00826136448532482,-.0111886088569481,-.00977235907777577,-.00489590109849087,.00961512655393934,.0205123068664937,.0273747549478629,.0289933555580273,.0253771222247394,.0177095505014919,.00803668428539622,-.00124858539139083,-.00799854029488872,-.0108326583272072,-.0094614646283383,-.00474014459543961,.00692069282783456,.0147641712479721,.0197035649160317,.0208685872972819,.0182657260709564,.0127468274548747,.00578457527121418,-.000898696019720848,-.00575712031884831,-.00779703734224594,-.00681009137288203,-.00341182037704094,.00251422184947983,.00536368292374106,.00715811764181876,.00758135918698483,.00663576447137033,.00463080112004091,.00210148115205368,-.000326487712290173,-.00209150702911081,-.00283258947258012,-.00247404139334002,-.0012394818772994,-.00251422184947983,-.00536368292374106,-.00715811764181876,-.00758135918698483,-.00663576447137033,-.00463080112004091,-.00210148115205368,.000326487712290173,.00209150702911081,.00283258947258012,.00247404139334002,.0012394818772994,-.00692069282783456,-.0147641712479721,-.0197035649160317,-.0208685872972819,-.0182657260709564,-.0127468274548747,-.00578457527121418,.000898696019720848,.00575712031884831,.00779703734224594,.00681009137288203,.00341182037704094,-.00961512655393934,-.0205123068664937,-.0273747549478629,-.0289933555580273,-.0253771222247394,-.0177095505014919,-.00803668428539622,.00124858539139083,.00799854029488872,.0108326583272072,.0094614646283383,.00474014459543961,-.0099310701835657,-.0211863211550268,-.0282742625507885,-.0299460487897454,-.026210989575333,-.0182914689643539,-.00830076184995504,.00128961268294456,.00826136448532482,.0111886088569481,.00977235907777577,.00489590109849087,-.0077909390577761,-.016620699876881,-.0221812002498267,-.0234927189949964,-.0205625595883267,-.0143496841069869,-.00651195978990583,.00101170303252743,.00648105249983563,.00877747998297588,.00766642996357517,.00384084155948903,-.00373129696748193,-.00796011451098884,-.0106231924821284,-.0112513152129516,-.00984798054080671,-.00687246202231334,-.00311875829553411,.000484532664068795,.00310395593642735,.00420377879992814,.00367166610627629,.00183948563288991},
			{-.00463860142102606,-.00989570082574293,-.0132063344657042,-.0139871919040607,-.0122426215144303,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173,-.0080781328720134,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782068,.00104899697753615,.00671996056648802,.00910104020299664,.00794903405616202,.00398242474081937,-.006479984172588,-.0138239910928359,-.0184488449315262,-.0195396788666562,-.0171025674430431,-.011935111442324,-.00541621440729949,.000841467195356086,.00539050778207919,.00730052320305626,.00637642580128229,.00319455617998525,-.000368039046268343,-.000785151377213537,-.00104782590705614,-.00110978122522199,-.00097136234330704,-.000677870024886959,-.000307620872482971,4.77922130358021e-005,.00030616083159755,.000414642617227202,.000362157315203368,.000181438932321165,.00721530281433017,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.013289452703208,.00603082137473739,-.00093695300190748,-.00600219768054942,-.00812895281996605,-.00709999311788209,-.00355705964429505,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450082,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450083,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.00721530281433019,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.0132894527032081,.0060308213747374,-.000936953001907482,-.00600219768054943,-.00812895281996606,-.0070999931178821,-.00355705964429506,-.000368039046268326,-.000785151377213499,-.00104782590705609,-.00110978122522194,-.000971362343306994,-.000677870024886926,-.000307620872482956,4.77922130357998e-005,.000306160831597535,.000414642617227182,.000362157315203351,.000181438932321156,-.00647998417258799,-.0138239910928359,-.0184488449315262,-.0195396788666561,-.0171025674430431,-.011935111442324,-.00541621440729948,.000841467195356084,.00539050778207918,.00730052320305625,.00637642580128228,.00319455617998524,-.00807813287201338,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782067,.00104899697753615,.00671996056648801,.00910104020299662,.007949034056162,.00398242474081937,-.00463860142102606,-.00989570082574293,-.0132063344657043,-.0139871919040607,-.0122426215144304,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173},
			{.00517900543192415,.0110485647887122,.0147448922047198,.0156167207038177,.0136689052516612,.00953891326647014,.00432880746137177,-.000672526824057758,-.00430826192482315,-.00583480581394941,-.00509623835206271,-.00255318892270115,.00646894868289895,.0138004486724991,.0184174263304009,.0195064025624373,.0170734415684041,.0119147858063732,.00540699053010265,-.000840034166769796,-.00538132768356915,-.00728809032569326,-.00636556667272444,-.00318911581302657,-.000274950685842771,-.000586562517874416,-.000782798604414518,-.000829083522805726,-.000725675020631865,-.000506415909248477,-.000229814120882911,3.57040968488888e-005,.000228723369108405,.000309766784644738,.000270556624922728,.00013554746265675,-.00936530034663045,-.0199793433325387,-.0266634870132945,-.0282400321341923,-.0247177580278924,-.0172494099292266,-.00782787014830752,.00121614386801799,.00779071723872097,.0105511974509746,.00921563096089514,.00461698320596323,-.0122814780526446,-.026200533625503,-.0349659934482049,-.0370334449537273,-.0324144011931623,-.0226205505030159,-.0102653211180761,.00159482810706601,.0102165994939113,.013836641124904,.0120852044460302,.0060546246051647,-.00561022419736452,-.0119684998091909,-.0159725939896759,-.0169170134165105,-.0148070172936393,-.0103331503949069,-.00468923631858351,.000728523325820062,.00466698010205544,.00632062838987608,.00552056569433968,.00276577471540888,.00561022419736453,.0119684998091909,.0159725939896759,.0169170134165105,.0148070172936393,.010333150394907,.00468923631858351,-.000728523325820062,-.00466698010205545,-.00632062838987609,-.00552056569433968,-.00276577471540888,.0122814780526447,.0262005336255031,.034965993448205,.0370334449537275,.0324144011931624,.022620550503016,.0102653211180761,-.00159482810706601,-.0102165994939113,-.0138366411249041,-.0120852044460302,-.00605462460516473,.00936530034663043,.0199793433325386,.0266634870132944,.0282400321341923,.0247177580278923,.0172494099292265,.0078278701483075,-.00121614386801798,-.00779071723872095,-.0105511974509746,-.00921563096089512,-.00461698320596322,.00027495068584276,.000586562517874393,.000782798604414488,.000829083522805695,.000725675020631837,.000506415909248458,.000229814120882903,-3.57040968488874e-005,-.000228723369108396,-.000309766784644726,-.000270556624922718,-.000135547462656745,-.00646894868289895,-.0138004486724991,-.0184174263304009,-.0195064025624373,-.0170734415684041,-.0119147858063732,-.00540699053010265,.000840034166769796,.00538132768356914,.00728809032569326,.00636556667272443,.00318911581302657,-.00517900543192412,-.0110485647887121,-.0147448922047197,-.0156167207038176,-.0136689052516611,-.00953891326647009,-.00432880746137175,.000672526824057755,.00430826192482312,.00583480581394939,.00509623835206268,.00255318892270114},
			{-.00541977053728884,-.0115621979370935,-.015430362720618,-.0163427213723106,-.0143043545587521,-.0099823647144083,-.00453004798877423,.000703791705661616,.0045085473174344,.00610605821077421,.00533315572547085,.0026718832952154,-.00347716631903209,-.00741796815275411,-.00989966958442869,-.0104850122207482,-.00917725566884735,-.00640439334662279,-.00290634634470048,.000451532181612414,.00289255214258342,.00391746842536516,.0034215967881247,.00171420220442671,.00694040822993805,.0148062308480902,.019759695669824,.0209280369217824,.0183177607650484,.012783140124006,.00580105414555624,-.000901256190189091,-.00577352098058376,-.00781924924071369,-.00682949169783763,-.00342153983899534,.0112848679704134,.0240744282936295,.0321285937918819,.0340282769713043,.0297840566287605,.0207849515140105,.0094323169405852,-.00146541194362748,-.00938754894982324,-.0127138335939062,-.0111045214577846,-.00556330752592008,.0017039064892699,.00363500704682817,.00485110854611964,.00513794242893165,.00449710599180968,.00313832770190446,.00142418910757755,-.000221263104428595,-.00141742957169544,-.00191966655001643,-.00167667590101747,-.000840005910576199,-.0110322454417378,-.0235354992633652,-.0314093645879802,-.0332665215482455,-.0291173121245713,-.0203196605576931,-.00922116553296151,.00143260730012424,.00917739971635521,.012429222307352,.0108559361577392,.00543876758281373,-.011032245441738,-.0235354992633655,-.0314093645879806,-.0332665215482459,-.0291173121245716,-.0203196605576934,-.00922116553296161,.00143260730012425,.00917739971635532,.0124292223073521,.0108559361577394,.00543876758281379,.00170390648926984,.00363500704682805,.00485110854611947,.00513794242893148,.00449710599180952,.00313832770190435,.0014241891075775,-.000221263104428587,-.00141742957169539,-.00191966655001636,-.00167667590101741,-.00084000591057617,.0112848679704135,.0240744282936297,.0321285937918822,.0340282769713046,.0297840566287608,.0207849515140107,.00943231694058528,-.0014654119436275,-.00938754894982332,-.0127138335939063,-.0111045214577847,-.00556330752592013,.00694040822993806,.0148062308480902,.0197596956698241,.0209280369217824,.0183177607650485,.012783140124006,.00580105414555625,-.000901256190189092,-.00577352098058377,-.00781924924071369,-.00682949169783764,-.00342153983899535,-.00347716631903199,-.00741796815275391,-.00989966958442842,-.0104850122207479,-.0091772556688471,-.00640439334662262,-.0029063463447004,.000451532181612402,.00289255214258334,.00391746842536505,.0034215967881246,.00171420220442666,-.00541977053728884,-.0115621979370935,-.0154303627206179,-.0163427213723106,-.0143043545587521,-.00998236471440829,-.00453004798877423,.000703791705661616,.00450854731743439,.0061060582107742,.00533315572547085,.0026718832952154},
			{-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204719,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359087,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359088,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204718,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029},
			{.00669658564383886,.0128558782631327,.0137727308070988,.00947626072033451,.00236330891169302,-.00410612209904174,-.00717100835620119,-.0060731527881528,-.00224476908541414,.00166655945599322,.00335856916390389,.00214806069987574,.013982454653438,.0268430427690799,.028757428666868,.0197864095008088,.00493458330070284,-.00857357302734623,-.0149730481312158,-.0126807283562296,-.0046870724296822,.00347977211971043,.00701268430396727,.00448514555420918,.0178233634574797,.0342166893750909,.0366569472911615,.025221635027061,.00629008810395184,-.010928689703139,-.0190860678989135,-.0161640596018723,-.00597458726213122,.00443564772967482,.00893903282793333,.00571719210637408,.0172563371411996,.033128131461881,.0354907558571953,.0244192426596455,.00608997741804402,-.0105810081513981,-.0184788703405621,-.0156498218041611,-.00578451381080864,.00429453356800305,.00865464953138992,.00553530733545099,.0124206174529098,.0238446806208565,.025545230022447,.0175762717825614,.00438339139919734,-.00761590675004047,-.0133005618506412,-.0112642937052734,-.00416352743964371,.00309108231662293,.00622936896392408,.00398415575306398,.00451229213042641,.00866254557198487,.0092803389877543,.00638529228900029,.00159244438452158,-.00276678645200671,-.00483196755688163,-.00409221071607914,-.00151256990016366,.00112296079197981,.00226307046811526,.00144740587326696,-.00451229213042641,-.00866254557198487,-.0092803389877543,-.00638529228900029,-.00159244438452158,.00276678645200671,.00483196755688163,.00409221071607914,.00151256990016366,-.00112296079197981,-.00226307046811526,-.00144740587326696,-.0124206174529098,-.0238446806208565,-.025545230022447,-.0175762717825614,-.00438339139919734,.00761590675004047,.0133005618506412,.0112642937052734,.00416352743964371,-.00309108231662293,-.00622936896392408,-.00398415575306398,-.0172563371411996,-.033128131461881,-.0354907558571953,-.0244192426596455,-.00608997741804402,.0105810081513981,.0184788703405621,.0156498218041611,.00578451381080864,-.00429453356800305,-.00865464953138992,-.00553530733545099,-.0178233634574797,-.0342166893750909,-.0366569472911615,-.025221635027061,-.00629008810395184,.010928689703139,.0190860678989135,.0161640596018723,.00597458726213122,-.00443564772967482,-.00893903282793333,-.00571719210637408,-.013982454653438,-.0268430427690799,-.028757428666868,-.0197864095008088,-.00493458330070284,.00857357302734624,.0149730481312158,.0126807283562296,.0046870724296822,-.00347977211971043,-.00701268430396727,-.00448514555420918,-.00669658564383886,-.0128558782631327,-.0137727308070988,-.00947626072033451,-.00236330891169302,.00410612209904174,.00717100835620119,.00607315278815281,.00224476908541414,-.00166655945599322,-.00335856916390389,-.00214806069987574},
			{-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203508,.00754990434136365,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315,-.0144978834787229,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606307,.00888961672521672,.0155249927504182,.0131481722408983,.00485984983510451,-.00360804536645972,-.00727118967533986,-.00465047942165492,-.0116296744515833,-.0223262550411753,-.0239185137194828,-.0164570175096932,-.00410425771180905,.00713092698429712,.012453584125954,.0105469852216292,.00389839465524694,-.00289424267204597,-.00583268371028218,-.00373044532996883,-.000660522337643703,-.00126804840771953,-.00135848278992925,-.000934697503477068,-.000233106602371594,.000405010181569894,.000707317348663856,.000599029608497395,.000221414344955362,-.000164382239881848,-.000331274782891918,-.000211875446734245,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627044,.00456998989967055,-.00794011160648144,-.0138667592696614,-.0117438083374094,-.00434076645531976,.00322266794806499,.006494549688551,.0041537590171735,.0221636560036439,.0425490359719281,.0455835382498503,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183392,.00551580349401449,.0111158395594773,.0071094257520315,.0221636560036439,.0425490359719281,.0455835382498504,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183393,.0055158034940145,.0111158395594773,.00710942575203151,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627045,.00456998989967055,-.00794011160648145,-.0138667592696614,-.0117438083374094,-.00434076645531976,.003222667948065,.00649454968855101,.0041537590171735,-.000660522337643672,-.00126804840771947,-.00135848278992918,-.000934697503477023,-.000233106602371583,.000405010181569874,.000707317348663822,.000599029608497366,.000221414344955352,-.00016438223988184,-.000331274782891902,-.000211875446734235,-.0116296744515833,-.0223262550411753,-.0239185137194827,-.0164570175096932,-.00410425771180904,.0071309269842971,.012453584125954,.0105469852216291,.00389839465524693,-.00289424267204597,-.00583268371028217,-.00373044532996882,-.0144978834787228,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606306,.0088896167252167,.0155249927504182,.0131481722408983,.0048598498351045,-.00360804536645972,-.00727118967533985,-.00465047942165491,-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203509,.00754990434136366,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315},
			{.00929479849593368,.0178438392786048,.0191164220692825,.0131529616098434,.00328025075555347,-.00569925922554386,-.0099532928015145,-.0084294794994272,-.00311571858085951,.0023131690011613,.00466166271491865,.00298148824255666,.0116098689755076,.0222882331591425,.0238777802016866,.016428991010076,.00409726811134705,-.00711878292950237,-.0124323755217502,-.0105290235783908,-.00389175564205705,.00288931374181348,.00582275057946566,.00372409233650869,-.000493455983937409,-.000947320081473204,-.0010148808353067,-.000698283842916624,-.000174146794559474,.000302570687259703,.000528415101729981,.000447516651629383,.000165411867579778,-.000122804930734209,-.000247485231958592,-.0001582856492233,-.0168079722487348,-.032267375730295,-.034568612948002,-.0237847666975931,-.00593174383418366,.0103060858116714,.01799874082957,.0152431983931265,.0056342169725212,-.0041829503237926,-.00842977904035172,-.00539148553491988,-.0220416574634008,-.0423148273072351,-.045332626345935,-.0311908939779877,-.00777877686962253,.0135152063489619,.0236032088979544,.0199896425729312,.00738860576070614,-.00548544207826128,-.011054653074738,-.00707029233633635,-.0100687099322352,-.0193295682367182,-.0207081099006063,-.0142481147124679,-.00355337379041754,.00617379580587153,.010782032352062,.00913134200774618,.00337514219752845,-.00250577005054177,-.00504980605001451,-.00322973545836477,.0100687099322352,.0193295682367182,.0207081099006063,.0142481147124679,.00355337379041755,-.00617379580587153,-.010782032352062,-.00913134200774619,-.00337514219752846,.00250577005054177,.00504980605001452,.00322973545836478,.0220416574634009,.0423148273072353,.0453326263459352,.0311908939779879,.00777877686962257,-.013515206348962,-.0236032088979545,-.0199896425729313,-.00738860576070618,.0054854420782613,.0110546530747381,.00707029233633639,.0168079722487348,.032267375730295,.0345686129480019,.023784766697593,.00593174383418365,-.0103060858116714,-.01799874082957,-.0152431983931265,-.00563421697252118,.00418295032379259,.0084297790403517,.00539148553491986,.000493455983937391,.000947320081473168,.00101488083530666,.000698283842916597,.000174146794559467,-.000302570687259691,-.000528415101729961,-.000447516651629366,-.000165411867579771,.000122804930734204,.000247485231958582,.000158285649223294,-.0116098689755076,-.0222882331591425,-.0238777802016866,-.016428991010076,-.00409726811134704,.00711878292950237,.0124323755217502,.0105290235783908,.00389175564205705,-.00288931374181348,-.00582275057946566,-.00372409233650869,-.00929479849593364,-.0178438392786047,-.0191164220692824,-.0131529616098433,-.00328025075555346,.00569925922554383,.00995329280151445,.00842947949942716,.00311571858085949,-.00231316900116129,-.00466166271491862,-.00298148824255664},
			{-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993},
			{.00997527550886585,.0163494681388942,.0117058404244761,.00065538634573533,-.00785594167460167,-.0080795624708938,-.00168671898857876,.00464312263238137,.00560726911956695,.00166497716447436,-.00230734475380579,-.00231558708502107,.0208283511742426,.0341376499933582,.0244417665390733,.00136844510727013,-.0164031872460608,-.0168701069287199,-.00352186517506783,.0096948288442119,.011707964381237,.00347646829883996,-.0048177302741764,-.00483494024185036,.0265497927509654,.043515087908134,.031155794841906,.00174434998167423,-.0209090589166153,-.0215042390488418,-.00448930353212625,.0123579487601599,.0149240823364804,.00443143636610296,-.00614113614848276,-.00616307360628593,.0257051468334037,.0421307139622166,.0301646151679837,.0016888558350856,-.0202438653529731,-.0208201106304678,-.00434648238332583,.0119647972554527,.014449292746328,.00429045618329429,-.0059457641685332,-.00596700371564648,.0185018287934532,.0303244817750634,.0217116264332397,.00121559008085388,-.0145709547315911,-.0149857195853513,-.00312847358668079,.00861191853147071,.0104001872586715,.00308814753183857,-.00427959083079407,-.00429487844875196,.00672153834378315,.0110165956718055,.00788762728301515,.000441612309247667,-.00529348920192007,-.00544416932653508,-.0011365446845897,.00312862805450344,.00377828906656089,.00112189461259403,-.00155473462575028,-.0015602884718828,-.00672153834378315,-.0110165956718055,-.00788762728301515,-.000441612309247667,.00529348920192007,.00544416932653508,.00113654468458969,-.00312862805450344,-.00377828906656089,-.00112189461259403,.00155473462575028,.0015602884718828,-.0185018287934532,-.0303244817750634,-.0217116264332397,-.00121559008085388,.0145709547315911,.0149857195853513,.00312847358668079,-.00861191853147071,-.0104001872586715,-.00308814753183857,.00427959083079407,.00429487844875196,-.0257051468334037,-.0421307139622166,-.0301646151679837,-.0016888558350856,.0202438653529731,.0208201106304678,.00434648238332583,-.0119647972554527,-.014449292746328,-.00429045618329429,.0059457641685332,.00596700371564648,-.0265497927509654,-.043515087908134,-.031155794841906,-.00174434998167423,.0209090589166153,.0215042390488418,.00448930353212624,-.0123579487601599,-.0149240823364804,-.00443143636610296,.00614113614848276,.00616307360628593,-.0208283511742426,-.0341376499933582,-.0244417665390733,-.00136844510727013,.0164031872460608,.0168701069287199,.00352186517506784,-.0096948288442119,-.011707964381237,-.00347646829883996,.0048177302741764,.00483494024185036,-.00997527550886584,-.0163494681388943,-.0117058404244761,-.00065538634573533,.00785594167460167,.0080795624708938,.00168671898857876,-.00464312263238137,-.00560726911956695,-.00166497716447436,.00230734475380579,.00231558708502107},
			{-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962,-.0215961371406428,-.0353960505441994,-.0253427521805069,-.00141888947227826,.017007850422056,.0174919819510481,.00365169007740165,-.0100522048780116,-.0121395497079787,-.0036046197569249,.0049953240579315,.00501316842875873,-.0173236351861994,-.0283934234472885,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064034,-.00806351288610232,-.00973790493624807,-.00289149477275801,.00400706714598473,.00402138125078621,-.000983918170479597,-.00161264105089006,-.00115461363289064,-6.46444836215388e-005,.000774876218005063,.000796933209270456,.000166370689198568,-.000457977598884836,-.000553076851722798,-.000164226169402932,.000227586539019653,.000228399527036137,.019289441206072,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433103,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543011,-.0076638869878077,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543012,-.0076638869878077,.0192894412060721,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433104,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,-.00098391817047955,-.00161264105088998,-.00115461363289059,-6.46444836215357e-005,.000774876218005026,.000796933209270418,.00016637068919856,-.000457977598884814,-.000553076851722771,-.000164226169402924,.000227586539019642,.000228399527036126,-.0173236351861993,-.0283934234472884,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064033,-.0080635128861023,-.00973790493624805,-.002891494772758,.00400706714598472,.0040213812507862,-.0215961371406427,-.0353960505441994,-.0253427521805068,-.00141888947227826,.0170078504220559,.017491981951048,.00365169007740164,-.0100522048780115,-.0121395497079786,-.00360461975692489,.00499532405793149,.00501316842875872,-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962},
			{-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758},
			{.0133541389835151,.0174668385049753,.00479454927113329,-.00864193040388231,-.00841992604968446,.0017949650137812,.00751289805161027,.00275957692744254,-.00421283180475702,-.00418627184884443,.000779677757981102,.00236085727977567,.0278834099500396,.0364707166195125,.0100110073002612,-.0180443298147953,-.0175807852593257,.00374788261429831,.0156869129896357,.00576198996068088,-.00879638263519743,-.00874092551138884,.00162796527590073,.00492946429917962,.0355428401014885,.0464890360080293,.0127609798215341,-.0230010149581752,-.0224101370833209,.00477740680633519,.0199960277841051,.00734477914307506,-.0112127039710897,-.0111420130589112,.00207515901375732,.00628356293888253,.0344120924880564,.0450100499065603,.0123550064261315,-.0222692686290634,-.0216971887384138,.00462542003968309,.0193598810769079,.00711111488148247,-.010855986887727,-.010787544911762,.00200914062311667,.00608365984231783,.0247688390097475,.0323969453570202,.00889277992146099,-.0160287820256653,-.0156170152981078,.00332924492618789,.0139346881567349,.00511837691184837,-.00781382857224871,-.00776456599733339,.00144612190203756,.00437884418905171,.00899828352070364,.0117695019702978,.00323066232490896,-.00582310398568677,-.00567351305179212,.00120948300176336,.00506233960976318,.00185945762740475,-.0028386895666607,-.00282079294195926,.00052536232622418,.00159079242634532,-.00899828352070364,-.0117695019702978,-.00323066232490896,.00582310398568677,.00567351305179212,-.00120948300176336,-.00506233960976318,-.00185945762740475,.0028386895666607,.00282079294195926,-.00052536232622418,-.00159079242634532,-.0247688390097475,-.0323969453570202,-.00889277992146099,.0160287820256653,.0156170152981078,-.00332924492618789,-.0139346881567349,-.00511837691184837,.00781382857224871,.00776456599733339,-.00144612190203756,-.00437884418905171,-.0344120924880564,-.0450100499065603,-.0123550064261315,.0222692686290634,.0216971887384138,-.00462542003968309,-.0193598810769079,-.00711111488148247,.010855986887727,.010787544911762,-.00200914062311667,-.00608365984231783,-.0355428401014885,-.0464890360080293,-.0127609798215341,.0230010149581752,.0224101370833208,-.00477740680633519,-.0199960277841051,-.00734477914307506,.0112127039710897,.0111420130589112,-.00207515901375732,-.00628356293888253,-.0278834099500396,-.0364707166195125,-.0100110073002612,.0180443298147953,.0175807852593257,-.00374788261429831,-.0156869129896357,-.00576198996068088,.00879638263519743,.00874092551138884,-.00162796527590073,-.00492946429917962,-.0133541389835151,-.0174668385049753,-.00479454927113329,.00864193040388231,.00841992604968446,-.0017949650137812,-.00751289805161027,-.00275957692744254,.00421283180475702,.00418627184884443,-.000779677757981102,-.00236085727977568},
			{-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292215,-.00120362482446452,.00306032152371413,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292214,-.00120362482446452,.00306032152371413,-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268},
			{1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394712,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394711,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752},
			{-3.02102579557316e-005,-.000356099425944935,-.00127919739538736,-.00291181027531203,-.00512624408687041,-.00756904967962207,-.00973625153619673,-.0110917233656258,-.0111998299356331,-.00983955325421882,-.00707171109480605,-.00324650309107673,-6.30789456599155e-005,-.00074353474146506,-.00267096107256627,-.00607985282341967,-.0107035852745446,-.0158041574533295,-.020329269696504,-.023159491603162,-.0233852178602818,-.0205449634342814,-.0147657156892056,-.00677869061452103,-8.04064095237273e-005,-.000947779933413657,-.00340466042315561,-.00774995730904432,-.0136438054243522,-.0201454787025301,-.0259136161429872,-.0295212855370935,-.0298090176429547,-.0261885598477368,-.0188217822951288,-.00864076226835519,-7.78483878402044e-005,-.000917627590643462,-.00329634573482353,-.0075034028495138,-.0132097461207719,-.0195045774155398,-.0250892093278478,-.0285821055764826,-.0288606838727134,-.0253554060687339,-.0182229926275056,-.00836586804816125,-5.60330408925769e-005,-.000660482069535263,-.00237261529082107,-.00540073456066038,-.00950799708383307,-.0140388364388575,-.0180584946102595,-.0205725813339951,-.0207730940163711,-.0182500954035624,-.0131163883981527,-.00602151231964795,-2.03562705656154e-005,-.000239946850947697,-.000861948558184859,-.00196203547440233,-.00345416486580917,-.00510017568961999,-.00656047925364063,-.0074738230372111,-.00754666738671032,-.00663008599864202,-.0047650590941222,-.00218755812714027,2.03562705656154e-005,.000239946850947697,.000861948558184859,.00196203547440233,.00345416486580917,.00510017568962,.00656047925364063,.0074738230372111,.00754666738671032,.00663008599864202,.0047650590941222,.00218755812714027,5.60330408925769e-005,.000660482069535263,.00237261529082107,.00540073456066038,.00950799708383307,.0140388364388575,.0180584946102595,.0205725813339951,.0207730940163711,.0182500954035624,.0131163883981527,.00602151231964795,7.78483878402044e-005,.000917627590643462,.00329634573482353,.00750340284951379,.0132097461207719,.0195045774155398,.0250892093278478,.0285821055764826,.0288606838727134,.0253554060687339,.0182229926275056,.00836586804816125,8.04064095237273e-005,.000947779933413657,.00340466042315561,.00774995730904432,.0136438054243522,.0201454787025301,.0259136161429872,.0295212855370935,.0298090176429547,.0261885598477368,.0188217822951288,.00864076226835519,6.30789456599155e-005,.00074353474146506,.00267096107256627,.00607985282341966,.0107035852745446,.0158041574533295,.020329269696504,.023159491603162,.0233852178602818,.0205449634342814,.0147657156892056,.00677869061452103,3.02102579557315e-005,.000356099425944935,.00127919739538736,.00291181027531203,.00512624408687041,.00756904967962207,.00973625153619672,.0110917233656258,.0111998299356331,.00983955325421882,.00707171109480605,.00324650309107673},
			{3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491522,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500883,.00403592477071172,6.54041959386344e-005,.00077094332204201,.00276941948707301,.00630397165299627,.0110981466354321,.0163867388699784,.021078658252914,.0240132093332518,.0242472564339873,.0213023030101423,.015310017502792,.00702857038146807,5.24648654853783e-005,.000618422673152875,.0022215275148126,.00505681661629785,.00890252929508159,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327876,2.97980960162188e-006,3.51241121514965e-005,.000126174516176939,.000287208412095994,.000505630616352145,.000746578428543733,.000960341876387736,.00109403977390252,.00110470293989661,.000970531112496787,.000697523094674574,.000320221068507963,-5.84183358335144e-005,-.000688598418638334,-.0024736161852904,-.00563064078419707,-.00991274715595561,-.0146364617863459,-.0188272345384989,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218605,-9.99867395890061e-005,-.00117858048818797,-.00423375321862369,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.032565920170009,-.0234052068248199,-.0107449350355781,-9.99867395890061e-005,-.00117858048818797,-.0042337532186237,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.0325659201700091,-.0234052068248199,-.0107449350355781,-5.84183358335145e-005,-.000688598418638335,-.0024736161852904,-.00563064078419708,-.00991274715595562,-.0146364617863459,-.018827234538499,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218606,2.97980960162174e-006,3.51241121514949e-005,.000126174516176933,.00028720841209598,.000505630616352121,.000746578428543697,.00096034187638769,.00109403977390247,.00110470293989656,.000970531112496741,.000697523094674541,.000320221068507948,5.24648654853782e-005,.000618422673152874,.0022215275148126,.00505681661629784,.00890252929508158,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327875,6.54041959386343e-005,.000770943322042009,.00276941948707301,.00630397165299626,.0110981466354321,.0163867388699784,.0210786582529139,.0240132093332518,.0242472564339873,.0213023030101423,.0153100175027919,.00702857038146806,3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491523,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500884,.00403592477071172},
			{-4.19315566304071e-005,-.000494262686103036,-.00177551406892167,-.00404156553827383,-.00711517904100237,-.0105057704487168,-.0135138264378992,-.0153952087107654,-.0155452596229099,-.0136572082602461,-.00981546912576032,-.00450611604885183,-5.23755171918012e-005,-.000617369491942089,-.00221774422687797,-.00504820479710383,-.008887368181188,-.0131224596692141,-.0168797370239278,-.0192297182193646,-.0194171425546437,-.0170588311884377,-.0122602238803855,-.00562846165395503,2.22612437958035e-006,2.62401471318847e-005,9.4261111981906e-005,.000214564597630484,.000377741128665099,.00055774585065616,.000717441967632789,.000817323567112939,.000825289691448544,.000725054033484014,.000521098114970246,.000239227340925692,7.58256825576655e-005,.000893785219253475,.00321069802758278,.00730844476493686,.0128665222724588,.0189977973412616,.0244373258701412,.0278394675137098,.0281108076152006,.0246966061215671,.0177495114871256,.00814850085580817,9.94363684763003e-005,.00117209306138154,.00421044877365996,.00958415647735547,.0168729143812158,.0249133527446604,.0320466477561979,.0365081520733029,.0368639823593503,.0323866629825667,.0232763742438573,.0106857901742116,4.54228976457494e-005,.000535416407238526,.00192334843497638,.00437807781360722,.00770760914409425,.0113805108641198,.0146390261754792,.0166770577030509,.0168396022821627,.0147943463773532,.0106327330839207,.00488130812483214,-4.54228976457494e-005,-.000535416407238526,-.00192334843497639,-.00437807781360723,-.00770760914409426,-.0113805108641198,-.0146390261754792,-.0166770577030509,-.0168396022821627,-.0147943463773532,-.0106327330839207,-.00488130812483215,-9.94363684763008e-005,-.00117209306138155,-.00421044877365998,-.00958415647735551,-.0168729143812159,-.0249133527446605,-.032046647756198,-.036508152073303,-.0368639823593504,-.0323866629825669,-.0232763742438574,-.0106857901742116,-7.58256825576654e-005,-.000893785219253473,-.00321069802758277,-.00730844476493685,-.0128665222724588,-.0189977973412615,-.0244373258701411,-.0278394675137098,-.0281108076152005,-.0246966061215671,-.0177495114871256,-.00814850085580816,-2.22612437958026e-006,-2.62401471318837e-005,-9.42611119819025e-005,-.000214564597630476,-.000377741128665085,-.000557745850656139,-.000717441967632762,-.000817323567112908,-.000825289691448512,-.000725054033483986,-.000521098114970227,-.000239227340925682,5.23755171918011e-005,.000617369491942089,.00221774422687796,.00504820479710383,.008887368181188,.0131224596692141,.0168797370239278,.0192297182193646,.0194171425546437,.0170588311884377,.0122602238803855,.00562846165395503,4.19315566304069e-005,.000494262686103033,.00177551406892166,.00404156553827381,.00711517904100233,.0105057704487168,.0135138264378992,.0153952087107653,.0155452596229098,.013657208260246,.00981546912576027,.00450611604885181},
			{4.38808991888882e-005,.000517240304071817,.0018580553675488,.00422945256989752,.00744595429549842,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731,2.81527019748646e-005,.000331846256550951,.0011920739999021,.00271349311244975,.00477710657881669,.00705353791336682,.00907313629206733,.0103362898376332,.0104370335006432,.00916940235127242,.00659007199462349,.00302539071720239,-5.61926656806376e-005,-.00066236362564373,-.0023793743066966,-.00541611996712565,-.00953508310299627,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136763,-.0183020855787247,-.0131537538647564,-.00603866617409735,-9.13673651610601e-005,-.00107698075043354,-.00386878177964248,-.00880642704521119,-.0155037211557789,-.0228916988068627,-.0294461454354727,-.0335456102525707,-.0338725658341033,-.0297585693083877,-.0213875568642763,-.00981866602574447,-1.37955930732719e-005,-.000162613731440516,-.000584148826303059,-.00132968575520406,-.00234091272972111,-.0034564262736338,-.0044460846527431,-.00506506439824835,-.00511443154534631,-.00449325765165264,-.00322931531199247,-.00148252410228482,8.93220195809498e-005,.00105287150952592,.00378217541095215,.00860928677962801,.0151566555762327,.0223792463037912,.0287869653955147,.0327946597841923,.033114296153308,.029092395362165,.0209087760126605,.00959886582549863,8.93220195809508e-005,.00105287150952593,.00378217541095219,.00860928677962811,.0151566555762329,.0223792463037915,.028786965395515,.0327946597841926,.0331142961533084,.0290923953621653,.0209087760126607,.00959886582549874,-1.37955930732714e-005,-.00016261373144051,-.00058414882630304,-.00132968575520402,-.00234091272972103,-.00345642627363368,-.00444608465274295,-.00506506439824818,-.00511443154534613,-.00449325765165249,-.00322931531199236,-.00148252410228477,-9.13673651610609e-005,-.00107698075043355,-.00386878177964252,-.00880642704521126,-.0155037211557791,-.0228916988068629,-.029446145435473,-.033545610252571,-.0338725658341036,-.029758569308388,-.0213875568642765,-.00981866602574456,-5.61926656806377e-005,-.000662363625643731,-.0023793743066966,-.00541611996712565,-.00953508310299628,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136764,-.0183020855787247,-.0131537538647564,-.00603866617409736,2.81527019748638e-005,.000331846256550942,.00119207399990207,.00271349311244967,.00477710657881656,.00705353791336663,.00907313629206708,.0103362898376329,.010437033500643,.00916940235127217,.00659007199462331,.0030253907172023,4.38808991888881e-005,.000517240304071817,.0018580553675488,.00422945256989751,.00744595429549841,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731},
			{8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302079,-.00878002632372056,-.00505719415206449,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302078,-.00878002632372056,-.00505719415206449,8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575},
			{-.000146293799182738,-.00163998020354913,-.00536553359148041,-.0105856065426876,-.0151909408292571,-.0168075648663272,-.0141532955069821,-.00788475067656189,-.000434896968818276,.0051187183448384,.00658493074687408,.00379284435342441,-.000305461099423673,-.00342427470479066,-.0112032211823346,-.0221027208990368,-.0317186477684445,-.0350941548540275,-.0295520468412377,-.0164633403725859,-.000908063820704495,.0106878715433549,.0137493194981995,.00791944985099954,-.00038936970167901,-.00436490545862284,-.014280688761482,-.0281742253235845,-.0404315981398701,-.0447343397636272,-.0376698430153202,-.0209857357994469,-.00115750430951867,.0136237752115456,.0175261872670667,.0100948822346398,-.000376982428752617,-.00422604186707806,-.0138263678718428,-.0272779002703818,-.0391453212702289,-.0433111769611684,-.0364714276673539,-.0203181028639926,-.00112067986803374,.0131903531422229,.0169686152112542,.00977372714511676,-.000271341159812294,-.00304178395110495,-.00995182376735971,-.0196338516919752,-.0281756815823874,-.0311741452469417,-.0262510895163342,-.0146243887667217,-.000806633285738924,.00949403857306351,.0122135234447537,.00703484899288287,-9.85756613730145e-005,-.00110505116489299,-.00361540287663619,-.0071327914908855,-.0102359570090312,-.0113252703260409,-.00953677102517345,-.00531290127842359,-.000293042197071052,.00344909387167086,.00443705684789683,.00255569369797207,9.85756613730145e-005,.00110505116489299,.00361540287663619,.0071327914908855,.0102359570090312,.0113252703260409,.00953677102517345,.00531290127842359,.000293042197071052,-.00344909387167086,-.00443705684789683,-.00255569369797207,.000271341159812294,.00304178395110495,.00995182376735971,.0196338516919752,.0281756815823874,.0311741452469417,.0262510895163341,.0146243887667217,.000806633285738924,-.00949403857306351,-.0122135234447537,-.00703484899288287,.000376982428752617,.00422604186707806,.0138263678718428,.0272779002703818,.0391453212702289,.0433111769611684,.0364714276673539,.0203181028639926,.00112067986803374,-.0131903531422229,-.0169686152112542,-.00977372714511676,.00038936970167901,.00436490545862284,.014280688761482,.0281742253235845,.0404315981398701,.0447343397636272,.0376698430153202,.0209857357994469,.00115750430951867,-.0136237752115456,-.0175261872670667,-.0100948822346398,.000305461099423673,.00342427470479066,.0112032211823346,.0221027208990368,.0317186477684445,.0350941548540275,.0295520468412377,.0164633403725859,.000908063820704495,-.0106878715433549,-.0137493194981995,-.00791944985099954,.000146293799182738,.00163998020354913,.00536553359148041,.0105856065426876,.0151909408292571,.0168075648663272,.0141532955069821,.00788475067656189,.000434896968818276,-.0051187183448384,-.00658493074687408,-.00379284435342441},
			{.000181866688959546,.00203875879391254,.00667022002460173,.013159609110584,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.0081861265388698,-.00471511470855014,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053052,.0306414104040703,.0170702209457237,.000941537361279075,-.0110818536551902,-.0142561543633291,-.00821138090239868,.000254062196557116,.00284808361770293,.00931809316299003,.0183835710409382,.026381458519847,.0291889804801732,.0245794241800656,.0136931099430282,.000755267002369574,-.00888945965934217,-.0114357681533554,-.0065868708927751,1.44297896450536e-005,.000161760576944934,.000529233101409994,.00104411859237783,.00149836891174606,.00165782573712333,.00139602005068532,.000777717813756163,4.28963620630315e-005,-.000504888309558718,-.000649509179715462,-.000374109815194419,-.000282891999889151,-.00317127097766542,-.0103754672901093,-.0204696536806732,-.0293751044500413,-.0325012110209999,-.0273685835856313,-.0152469407451353,-.000840971209593475,.00989819443838388,.0127334462466717,.00733431854530109,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.0468430913146029,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.046843091314603,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000282891999889152,-.00317127097766543,-.0103754672901093,-.0204696536806732,-.0293751044500414,-.0325012110209999,-.0273685835856313,-.0152469407451354,-.000840971209593476,.00989819443838389,.0127334462466717,.00733431854530111,1.44297896450529e-005,.000161760576944927,.000529233101409968,.00104411859237778,.00149836891174599,.00165782573712325,.00139602005068526,.000777717813756126,4.28963620630294e-005,-.000504888309558694,-.000649509179715431,-.000374109815194401,.000254062196557116,.00284808361770292,.00931809316299001,.0183835710409382,.026381458519847,.0291889804801731,.0245794241800655,.0136931099430282,.000755267002369572,-.00888945965934215,-.0114357681533554,-.00658687089277509,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053051,.0306414104040702,.0170702209457237,.000941537361279073,-.0110818536551902,-.0142561543633291,-.00821138090239866,.000181866688959546,.00203875879391254,.00667022002460174,.0131596091105841,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.00818612653886981,-.00471511470855015},
			{-.000203054430513827,-.00227627724582951,-.00744731064410978,-.0146927232745112,-.0210848843647913,-.0233287434691147,-.0196446423233178,-.010943960491264,-.000603632941579964,.00710473338089701,.00913982253690271,.00526443263155787,-.000253629525609746,-.00284323329737286,-.00930222433932735,-.0183522636004527,-.0263365305817064,-.0291392712986882,-.0245375651278731,-.0136697904136697,-.00075398077366707,.0088743208036425,.0114162929039563,.00657565337317555,1.07800533648885e-005,.000120846366765943,.000395373821517966,.000780028983233564,.00111938546758022,.00123851077219281,.0010429237719346,.000581009130510076,3.20465567117658e-005,-.000377186574041089,-.000485228391442131,-.000279485971128269,.000367187436559508,.00411623821592978,.0134671225727665,.0265691488808889,.0381282231590321,.0421858389935269,.0355238043246551,.0197901852642516,.00109156166590775,-.0128476331738681,-.0165277260852754,-.00951978008079133,.000481522671607105,.00539795707957059,.0176605302753814,.0348422802025231,.0500006319693174,.0553217127646963,.046585246282749,.0259524753061578,.00143145335939787,-.0168481435739603,-.021674147935839,-.0124840598593624,.000219961321613734,.00246580658242685,.00806739497188598,.0159160813255256,.0228404719819027,.0252711611958884,.0212803112797474,.011855185859552,.000653893142166497,-.00769629374853215,-.00990083023278439,-.00570276430932191,-.000219961321613735,-.00246580658242685,-.00806739497188599,-.0159160813255256,-.0228404719819027,-.0252711611958885,-.0212803112797474,-.0118551858595521,-.000653893142166498,.00769629374853216,.0099008302327844,.00570276430932192,-.000481522671607107,-.00539795707957062,-.0176605302753815,-.0348422802025233,-.0500006319693176,-.0553217127646966,-.0465852462827492,-.0259524753061579,-.00143145335939788,.0168481435739604,.0216741479358391,.0124840598593625,-.000367187436559507,-.00411623821592978,-.0134671225727665,-.0265691488808889,-.038128223159032,-.0421858389935268,-.035523804324655,-.0197901852642516,-.00109156166590775,.0128476331738681,.0165277260852754,.00951978008079131,-1.07800533648881e-005,-.000120846366765938,-.00039537382151795,-.000780028983233534,-.00111938546758017,-.00123851077219277,-.00104292377193456,-.000581009130510054,-3.20465567117645e-005,.000377186574041075,.000485228391442112,.000279485971128258,.000253629525609746,.00284323329737286,.00930222433932735,.0183522636004527,.0263365305817064,.0291392712986881,.0245375651278731,.0136697904136697,.00075398077366707,-.00887432080364249,-.0114162929039563,-.00657565337317555,.000203054430513826,.0022762772458295,.00744731064410975,.0146927232745111,.0210848843647912,.0233287434691146,.0196446423233177,.0109439604912639,.000603632941579961,-.00710473338089698,-.00913982253690267,-.00526443263155784},
			{.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048,.000554003280665635,.00578143006001033,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204796,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204797,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.000554003280665635,.00578143006001032,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048},
			{-.000415496844730152,-.0043360139403437,-.0123495368891892,-.0193702019109772,-.0189618233479971,-.00958981393734538,.00318359640435585,.0109005087031771,.00921800523314784,.00142803462579379,-.00473012443753221,-.00409148604117949,-.000867556408455861,-.00905358663684252,-.0257858032029976,-.0404449348150074,-.0395922413616564,-.0200235083437195,.00664734160455473,.0227602358497139,.0192471726671977,.00298173285024315,-.00987648840327513,-.00854301297387297,-.00110586971823095,-.0115405605969585,-.0328690315056423,-.051554951622511,-.0504680276371612,-.0255238637098843,.00847333235691014,.0290123562694813,.0245342725922579,.00380079961925163,-.0125895092713119,-.0108897349592236,-.00107068796176225,-.0111734131963732,-.0318233475134598,-.0499148002350218,-.0488624552731521,-.0247118563437564,.00820376469397058,.0280893672084472,.0237537477354436,.00367988229564073,-.0121889909806472,-.0105432927002232,-.000770650542793007,-.00804230294180483,-.02290553448866,-.0359272442282916,-.0351697964516905,-.0177868867353227,.0059048349660612,.0202179223629944,.0170972675881759,.00264867392725914,-.00877328675655798,-.00758876024801777,-.000279970008957894,-.00292169213103891,-.0083213627187413,-.0130520259571515,-.0127768523875183,-.00646180669720086,.00214517035484276,.00734497880785625,.00621127462321504,.000962238033926173,-.00318725159515452,-.00275692438613841,.000279970008957894,.00292169213103891,.0083213627187413,.0130520259571515,.0127768523875183,.00646180669720086,-.00214517035484276,-.00734497880785625,-.00621127462321504,-.000962238033926173,.00318725159515452,.00275692438613841,.000770650542793007,.00804230294180483,.02290553448866,.0359272442282916,.0351697964516905,.0177868867353227,-.0059048349660612,-.0202179223629944,-.0170972675881759,-.00264867392725914,.00877328675655798,.00758876024801777,.00107068796176225,.0111734131963732,.0318233475134598,.0499148002350218,.0488624552731521,.0247118563437564,-.00820376469397058,-.0280893672084472,-.0237537477354436,-.00367988229564073,.0121889909806472,.0105432927002232,.00110586971823095,.0115405605969585,.0328690315056423,.051554951622511,.0504680276371612,.0255238637098843,-.00847333235691015,-.0290123562694813,-.0245342725922579,-.00380079961925163,.0125895092713119,.0108897349592236,.000867556408455861,.00905358663684252,.0257858032029976,.0404449348150074,.0395922413616564,.0200235083437195,-.00664734160455473,-.0227602358497139,-.0192471726671977,-.00298173285024315,.00987648840327513,.00854301297387297,.000415496844730152,.0043360139403437,.0123495368891892,.0193702019109772,.0189618233479971,.00958981393734538,-.00318359640435585,-.0109005087031771,-.00921800523314784,-.00142803462579379,.00473012443753221,.00409148604117949},
			{.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703812,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.00588030439408339,.00508637429395636,.000899536743208372,.00938732485666896,.026736333462767,.0419358379330301,.041051712031237,.020761625765838,-.00689237951523805,-.0235992360051222,-.0199566723826478,-.0030916471033844,.0102405608742263,.00885793015048143,.000721575635236956,.00753017033241687,.0214469135896479,.0336394028639592,.0329301892448019,.0166542205347873,-.00552881598728651,-.0189304481891057,-.0160085162284352,-.00248000678059721,.00821460521072995,.00710550916733107,4.09828174776266e-005,.000427685721688117,.00121810507752847,.00191059320784475,.00187031250699067,.000945897903253332,-.000314016224231211,-.00107517918429856,-.000909224295609892,-.000140855178957249,.000466559082044881,.000403566543921213,-.000803456701900855,-.00838465924372367,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752584,.00615620046137393,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733643,-.00791180949317304,-.00137516782835013,-.0143508836460861,-.0408732004597185,-.0641094603592521,-.0627578518723987,-.0317393592123806,.0105367330925572,.0360773591216924,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.00137516782835013,-.0143508836460862,-.0408732004597186,-.0641094603592521,-.0627578518723987,-.0317393592123807,.0105367330925572,.0360773591216925,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.000803456701900856,-.00838465924372368,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752585,.00615620046137394,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733645,-.00791180949317305,4.09828174776246e-005,.000427685721688097,.00121810507752841,.00191059320784466,.00187031250699058,.000945897903253286,-.000314016224231196,-.00107517918429851,-.000909224295609848,-.000140855178957242,.000466559082044858,.000403566543921193,.000721575635236955,.00753017033241686,.0214469135896478,.0336394028639591,.0329301892448018,.0166542205347873,-.0055288159872865,-.0189304481891056,-.0160085162284351,-.00248000678059721,.00821460521072994,.00710550916733106,.00089953674320837,.00938732485666894,.026736333462767,.04193583793303,.0410517120312369,.020761625765838,-.00689237951523804,-.0235992360051222,-.0199566723826478,-.00309164710338439,.0102405608742263,.00885793015048141,.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703813,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.0058803043940834,.00508637429395636},
			{.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454},
			{-.00090976995509462,-.00863877841154346,-.020257551125304,-.0218647389120542,-.00764500327777482,.0097501159735198,.013176762981398,.00178724764942572,-.00876022920331713,-.00672244007828754,.00197953596529396,.0041756355320037,-.00189959746932745,-.0180377484624032,-.0422977178317532,-.0456535220494547,-.0159627483828599,.0203582186081548,.0275130492859463,.00373176877615995,-.0182913374223476,-.0140364386501337,.00413326631535552,.00871871690750996,-.00242140718192188,-.02299262574193,-.0539166846610791,-.0581943111398842,-.0203476337495614,.025950517172659,.0350707432565425,.00475686657924678,-.0233158742926257,-.0178921765820591,.00526865343968954,.0111137038650945,-.00234437337189749,-.0222611463044402,-.0522013979161087,-.0563429374666252,-.0197003011718718,.0251249364009332,.0339550147687404,.0046055332722702,-.0225741111376221,-.0173229610688503,.00510103832266685,.0107601363368395,-.0016874128374327,-.0160229358089307,-.0375730717775044,-.0405540333803145,-.0141797127954073,.0180842099346574,.0244398475527877,.00331493100033043,-.0162481989361725,-.0124685714486963,.00367158135008905,.00774483808976782,-.000613021027013781,-.00582098011077765,-.013649939444666,-.0147328944291874,-.00515135472942478,.00656982138629644,.00887876405495684,.00120428288870561,-.00590281606138012,-.00452971336078232,.00133385056701396,.0028136259808645,.000613021027013781,.00582098011077765,.013649939444666,.0147328944291874,.00515135472942478,-.00656982138629644,-.00887876405495684,-.00120428288870561,.00590281606138012,.00452971336078232,-.00133385056701396,-.0028136259808645,.0016874128374327,.0160229358089307,.0375730717775044,.0405540333803145,.0141797127954073,-.0180842099346574,-.0244398475527877,-.00331493100033043,.0162481989361725,.0124685714486963,-.00367158135008905,-.00774483808976782,.00234437337189749,.0222611463044402,.0522013979161087,.0563429374666252,.0197003011718718,-.0251249364009332,-.0339550147687404,-.0046055332722702,.0225741111376221,.0173229610688503,-.00510103832266685,-.0107601363368395,.00242140718192188,.02299262574193,.0539166846610791,.0581943111398842,.0203476337495614,-.025950517172659,-.0350707432565425,-.00475686657924678,.0233158742926257,.0178921765820591,-.00526865343968954,-.0111137038650945,.00189959746932745,.0180377484624032,.0422977178317532,.0456535220494547,.0159627483828599,-.0203582186081548,-.0275130492859463,-.00373176877615995,.0182913374223476,.0140364386501337,-.00413326631535552,-.00871871690750996,.000909769955094619,.00863877841154346,.020257551125304,.0218647389120542,.00764500327777482,-.00975011597351981,-.013176762981398,-.00178724764942572,.00876022920331713,.00672244007828754,-.00197953596529396,-.00417563553200371},
			{.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622},
			{-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{4.40393013976499e-007,2.6993036058117e-005,.000232696122440507,.000949989177276297,.00256195880647401,.00523448009649574,.00866112710437402,.0119942097336745,.0140544988860133,.0137678071694569,.0106453074128955,.0050858458801123,9.19539549723029e-007,5.63613941066353e-005,.00048586894174179,.00198357510812152,.00534936380130629,.0109295818012743,.0180844124790659,.0250438809603261,.0293457597352286,.0287471481233613,.0222273761718653,.0106192339163871,1.17213236262568e-006,7.1843580904145e-005,.000619334655895749,.00252845304873323,.00681880668738833,.0139318820415032,.0230521081253869,.0319233070161965,.0374068898959917,.0366438427349157,.0283331227625236,.0135362831793791,1.13484254932205e-006,6.95579740866872e-005,.000599631357507937,.00244801371855121,.00660187553145923,.0134886579681297,.022318736335937,.0309077099756012,.0362168400475543,.0354780682047664,.0274317426011938,.0131056445512863,8.16827178275341e-007,5.0065838414078e-005,.000431597484647737,.00176201019189673,.00475184100640955,.00970874984676124,.0160643874648916,.0222464848021472,.0260678446360387,.0255360978134559,.0197445829974339,.00943306775437879,2.96745541228046e-007,1.81884426869712e-005,.000156795259243056,.000640121536048445,.00172629862078525,.00352709643428847,.0058360391029489,.00808193623891563,.00947019990385576,.00927702134312668,.00717301912047857,.00342694375341951,-2.96745541228046e-007,-1.81884426869712e-005,-.000156795259243056,-.000640121536048445,-.00172629862078525,-.00352709643428847,-.0058360391029489,-.00808193623891563,-.00947019990385576,-.00927702134312668,-.00717301912047857,-.00342694375341951,-8.16827178275341e-007,-5.0065838414078e-005,-.000431597484647737,-.00176201019189673,-.00475184100640955,-.00970874984676124,-.0160643874648916,-.0222464848021472,-.0260678446360387,-.0255360978134559,-.0197445829974339,-.00943306775437879,-1.13484254932205e-006,-6.95579740866872e-005,-.000599631357507937,-.00244801371855121,-.00660187553145923,-.0134886579681297,-.022318736335937,-.0309077099756012,-.0362168400475543,-.0354780682047664,-.0274317426011938,-.0131056445512863,-1.17213236262568e-006,-7.1843580904145e-005,-.000619334655895749,-.00252845304873323,-.00681880668738833,-.0139318820415032,-.0230521081253869,-.0319233070161965,-.0374068898959917,-.0366438427349157,-.0283331227625236,-.0135362831793791,-9.19539549723029e-007,-5.63613941066352e-005,-.00048586894174179,-.00198357510812152,-.00534936380130629,-.0109295818012743,-.0180844124790659,-.0250438809603261,-.0293457597352286,-.0287471481233613,-.0222273761718653,-.0106192339163871,-4.40393013976499e-007,-2.6993036058117e-005,-.000232696122440507,-.000949989177276297,-.00256195880647401,-.00523448009649574,-.00866112710437402,-.0119942097336745,-.0140544988860133,-.0137678071694569,-.0106453074128955,-.0050858458801123},
			{-5.47479249038328e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.014910729120128,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278,-9.53436115108903e-007,-5.84390183710293e-005,-.000503779306073239,-.00205669472909975,-.00554655494976578,-.0113324739708158,-.0187510498958539,-.0259670620771616,-.0304275190396141,-.0298068410819887,-.0230467337622952,-.0110106858739491,-7.64811749616544e-007,-4.68776535500915e-005,-.000404113423430025,-.00164980565479233,-.00444924450443136,-.00909049815478573,-.0150414097502034,-.0208298320830514,-.0244078483124377,-.0239099630454571,-.0184872509991657,-.00883237145550259,-4.34384682750355e-008,-2.66247670431932e-006,-2.2952142317351e-005,-9.37028368505082e-005,-.00025270057154655,-.000516306549812956,-.000854296237699806,-.00118305713865876,-.00138627517832104,-.00135799714338606,-.00105000722912861,-.000501645911501147,8.51599050625011e-007,5.21971129219724e-005,.00044997034618045,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801575,.023193505114819,.0271775380818778,.0266231550969231,.0205850987611647,.00983462812913714,1.45756655498971e-006,8.93387163903609e-005,.000770153192219288,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,1.45756655498971e-006,8.93387163903609e-005,.000770153192219289,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,8.51599050625012e-007,5.21971129219725e-005,.000449970346180451,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801576,.0231935051148191,.0271775380818778,.0266231550969231,.0205850987611648,.00983462812913715,-4.34384682750334e-008,-2.6624767043192e-006,-2.29521423173499e-005,-9.37028368505037e-005,-.000252700571546538,-.000516306549812931,-.000854296237699765,-.0011830571386587,-.00138627517832097,-.00135799714338599,-.00105000722912856,-.000501645911501123,-7.64811749616543e-007,-4.68776535500914e-005,-.000404113423430024,-.00164980565479233,-.00444924450443135,-.00909049815478571,-.0150414097502034,-.0208298320830514,-.0244078483124376,-.0239099630454571,-.0184872509991656,-.00883237145550258,-9.53436115108901e-007,-5.84390183710292e-005,-.000503779306073238,-.00205669472909974,-.00554655494976577,-.0113324739708158,-.0187510498958538,-.0259670620771615,-.030427519039614,-.0298068410819887,-.0230467337622952,-.0110106858739491,-5.47479249038329e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.0149107291201281,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278},
			{6.11261401273432e-007,3.74660826168086e-005,.000322980050454363,.00131857612920148,.00355597495952486,.00726540961635832,.0120215637451267,.0166478513879527,.019507513540437,.0191095881083022,.0147755875313232,.00705910670850027,7.63509266142174e-007,4.6797820350488e-005,.000403425213480253,.00164699601620824,.00444166738830656,.00907501692868488,.0150157940511236,.0207943586321327,.0243662814574726,.0238692440950028,.0184557669915978,.00881732982223001,-3.24515476414362e-008,-1.98905470301973e-006,-1.71468417157011e-005,-7.00025160864056e-005,-.000188784848135954,-.000385716791226497,-.000638218522855875,-.000883825710764993,-.00103564367667356,-.00101451802389758,-.000784428203227012,-.000374764277900816,-1.10535636397296e-006,-6.77506755168088e-005,-.000584051362415962,-.00238440790267356,-.00643034149293689,-.0131381872625712,-.0217388370391121,-.0301046466206003,-.0352758315710584,-.0345562549580405,-.0267189939936707,-.0127651255386803,-1.44954346598925e-006,-8.88469567034131e-005,-.000765913929467261,-.00312686750465765,-.00843262842551854,-.0172291707200675,-.0285079004521123,-.0394786651862746,-.0462600594955417,-.045316420311225,-.0350387842542686,-.0167399446189638,-6.62156769174321e-007,-4.05856155279577e-005,-.000349872290759864,-.00142836453897385,-.00385205557811836,-.00787034834568896,-.0130225134342083,-.0180339989827199,-.0211317647632424,-.0207007069245407,-.0160058450967311,-.007646868069241,6.62156769174322e-007,4.05856155279578e-005,.000349872290759864,.00142836453897385,.00385205557811836,.00787034834568897,.0130225134342083,.0180339989827199,.0211317647632424,.0207007069245407,.0160058450967311,.00764686806924101,1.44954346598925e-006,8.88469567034135e-005,.000765913929467265,.00312686750465767,.00843262842551858,.0172291707200676,.0285079004521125,.0394786651862747,.0462600594955419,.0453164203112252,.0350387842542687,.0167399446189638,1.10535636397296e-006,6.77506755168087e-005,.000584051362415961,.00238440790267356,.00643034149293688,.0131381872625712,.0217388370391121,.0301046466206002,.0352758315710583,.0345562549580404,.0267189939936707,.0127651255386803,3.24515476414349e-008,1.98905470301965e-006,1.71468417157005e-005,7.00025160864029e-005,.000188784848135946,.000385716791226483,.000638218522855851,.000883825710764959,.00103564367667352,.00101451802389754,.000784428203226982,.000374764277900801,-7.63509266142173e-007,-4.67978203504879e-005,-.000403425213480253,-.00164699601620824,-.00444166738830656,-.00907501692868487,-.0150157940511236,-.0207943586321327,-.0243662814574726,-.0238692440950028,-.0184557669915978,-.00881732982223001,-6.11261401273429e-007,-3.74660826168084e-005,-.000322980050454361,-.00131857612920147,-.00355597495952484,-.00726540961635829,-.0120215637451267,-.0166478513879526,-.0195075135404369,-.0191095881083021,-.0147755875313231,-.00705910670850024},
			{-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906563,.00323132809181624,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190404,.00732492904924932,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190405,.00732492904924932,-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906564,.00323132809181624},
			{3.01831341787578e-006,.000176553289823538,.00139577555712289,.00499729323775215,.0112042776028942,.0177599426294488,.0206189091483633,.0167940290914578,.0072118425373148,-.00309463500286609,-.00816702708367357,-.00549362253845184,6.30223112790928e-006,.000368642842811124,.00291437599275742,.0104343362129758,.0233945045787828,.037082717323626,.0430522325112947,.0350658921887862,.0150583098027748,-.00646159041306787,-.0170527328289948,-.0114706705943021,8.03342179680519e-006,.000469907146940446,.00371494017739558,.0133005950221825,.0298208553755469,.0472691501765359,.0548784606654209,.0446983134887451,.0191947505157252,-.00823655626282739,-.0217370313501955,-.0146216051595062,7.77784929616953e-006,.000454957683604922,.0035967543563515,.01287745449555,.0288721450037781,.0457653457929115,.0531325763115239,.0432762968137264,.0185840953662182,-.00797452131259557,-.0210454969581169,-.0141564384735731,5.59827325600047e-006,.000327465515948182,.00258884081637867,.00926882308497601,.0207813434103976,.0329405857131029,.0382433073283765,.0311490395153758,.0133762998116843,-.00573983214301874,-.0151479462115431,-.0101893991370919,2.03379940271082e-006,.000118965105897601,.000940501234094081,.00336727522791983,.00754966431306182,.0119670013385727,.0138934296425726,.0113161496526603,.00485948243743287,-.00208522639934011,-.00550310471613919,-.00370171174777645,-2.03379940271082e-006,-.000118965105897601,-.000940501234094081,-.00336727522791983,-.00754966431306182,-.0119670013385727,-.0138934296425726,-.0113161496526603,-.00485948243743287,.00208522639934011,.00550310471613919,.00370171174777645,-5.59827325600047e-006,-.000327465515948182,-.00258884081637867,-.00926882308497601,-.0207813434103976,-.0329405857131029,-.0382433073283765,-.0311490395153758,-.0133762998116843,.00573983214301874,.0151479462115431,.0101893991370919,-7.77784929616953e-006,-.000454957683604922,-.0035967543563515,-.01287745449555,-.0288721450037781,-.0457653457929115,-.0531325763115239,-.0432762968137265,-.0185840953662182,.00797452131259557,.0210454969581169,.0141564384735731,-8.03342179680519e-006,-.000469907146940446,-.00371494017739558,-.0133005950221825,-.0298208553755469,-.0472691501765359,-.0548784606654209,-.0446983134887451,-.0191947505157252,.00823655626282739,.0217370313501955,.0146216051595062,-6.30223112790928e-006,-.000368642842811124,-.00291437599275742,-.0104343362129758,-.0233945045787828,-.037082717323626,-.0430522325112947,-.0350658921887862,-.0150583098027748,.00646159041306787,.0170527328289948,.0114706705943021,-3.01831341787578e-006,-.000176553289823538,-.00139577555712289,-.00499729323775215,-.0112042776028942,-.0177599426294488,-.0206189091483634,-.0167940290914578,-.0072118425373148,.00309463500286609,.00816702708367357,.00549362253845184},
			{-3.75224835608883e-006,-.000219484095846042,-.00173517319595299,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131,-6.5345473857451e-006,-.000382231954981349,-.00302180729935703,-.0108189723668713,-.0242568854796668,-.0384496805377896,-.0446392472280685,-.0363585101022895,-.0156133973760028,.00669978105920017,.0176813399042196,.0118935086675575,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786293,-.0194579906613178,-.0308429342861059,-.0358079794052745,-.029165473474441,-.0125244990989618,.00537432051580024,.0141833273288231,.00954053976870304,-2.97713422859992e-007,-1.7414455344253e-005,-.000137673283428536,-.000492911460431877,-.00110514163839842,-.00175176417356002,-.0020337603052895,-.0016564906264595,-.000711345055883006,.00030524145496262,.000805560341504197,.000541867242857217,5.83658858918833e-006,.000341405538160759,.00269904630894985,.00966339165295146,.02166599347167,.0343428478572752,.039871303339182,.032475036548949,.0139457213459092,-.00598417355813986,-.0157927857332021,-.0106231561080896,9.98969681400165e-006,.000584337539768473,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,9.98969681400166e-006,.000584337539768474,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,5.83658858918834e-006,.000341405538160759,.00269904630894985,.00966339165295147,.02166599347167,.0343428478572752,.039871303339182,.0324750365489491,.0139457213459093,-.00598417355813987,-.0157927857332022,-.0106231561080896,-2.97713422859977e-007,-1.74144553442522e-005,-.00013767328342853,-.000492911460431854,-.00110514163839836,-.00175176417355993,-.0020337603052894,-.00165649062645942,-.000711345055882972,.000305241454962606,.000805560341504158,.000541867242857191,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786291,-.0194579906613177,-.0308429342861059,-.0358079794052744,-.029165473474441,-.0125244990989617,.00537432051580023,.0141833273288231,.00954053976870302,-6.53454738574509e-006,-.000382231954981348,-.00302180729935702,-.0108189723668713,-.0242568854796668,-.0384496805377895,-.0446392472280684,-.0363585101022894,-.0156133973760028,.00669978105920015,.0176813399042196,.0118935086675575,-3.75224835608883e-006,-.000219484095846043,-.001735173195953,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131},
			{-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784,-1.55303998105985e-005,-.000854853930745827,-.00602179902609634,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-1.55303998105985e-005,-.000854853930745827,-.00602179902609635,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784},
			{1.16476424390402e-005,.000641131783107114,.00451628823154087,.0134633410098786,.0228083947504036,.0229321000995076,.00993827320683666,-.00713237192763805,-.0139474741042205,-.00647469362056888,.0043383944810003,.00561733742272306,2.43202492860194e-005,.0013386816149243,.00942999892153632,.0281114235173858,.0476238731611908,.0478821696313965,.0207510904569588,-.0148923753616752,-.0291222922492883,-.0135191446446746,.00905855719998496,.011728986973914,3.10008974207295e-005,.00170641061015528,.0120203714117433,.0358334878300008,.0607059076280363,.0610351568179937,.0264513089096609,-.0189832347320423,-.0371220370342043,-.0172327845581125,.0115468965483827,.0149508797278819,3.00146455997538e-005,.00165212345360114,.0116379594759614,.0346934936503285,.0587746308933795,.0590934054635625,.0256097961229407,-.0183793086724707,-.0359410493961229,-.0166845467145346,.0111795475780318,.0144752376147941,2.16036826314166e-005,.00118915116425069,.00837667005463561,.0249713835102858,.0423042966933997,.0425337415030114,.0184331980818002,-.0132289002122403,-.0258693384205835,-.0120090590732415,.00804671829410513,.0104188616322334,7.84841232696798e-006,.000432007302429133,.0030431645214108,.00907186601040717,.0153687484359428,.0154521035519588,.00669660545930529,-.00480593356555747,-.00939808448469602,-.00436277688733732,.00292329618651467,.00378507328877716,-7.84841232696798e-006,-.000432007302429133,-.0030431645214108,-.00907186601040717,-.0153687484359428,-.0154521035519588,-.00669660545930529,.00480593356555747,.00939808448469602,.00436277688733732,-.00292329618651467,-.00378507328877716,-2.16036826314166e-005,-.00118915116425069,-.00837667005463561,-.0249713835102858,-.0423042966933997,-.0425337415030114,-.0184331980818002,.0132289002122403,.0258693384205835,.0120090590732415,-.00804671829410513,-.0104188616322334,-3.00146455997538e-005,-.00165212345360114,-.0116379594759614,-.0346934936503285,-.0587746308933795,-.0590934054635624,-.0256097961229407,.0183793086724707,.0359410493961229,.0166845467145346,-.0111795475780318,-.0144752376147941,-3.10008974207295e-005,-.00170641061015528,-.0120203714117433,-.0358334878300008,-.0607059076280363,-.0610351568179937,-.0264513089096609,.0189832347320423,.0371220370342043,.0172327845581125,-.0115468965483827,-.0149508797278819,-2.43202492860194e-005,-.0013386816149243,-.00942999892153632,-.0281114235173858,-.0476238731611908,-.0478821696313965,-.0207510904569588,.0148923753616752,.0291222922492883,.0135191446446746,-.00905855719998497,-.011728986973914,-1.16476424390402e-005,-.000641131783107114,-.00451628823154087,-.0134633410098786,-.0228083947504036,-.0229321000995076,-.00993827320683666,.00713237192763805,.0139474741042205,.00647469362056888,-.0043383944810003,-.00561733742272306},
			{-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.00322545981703339,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072408,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072409,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.0032254598170334},
			{3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489},
			{-5.54740393494954e-009,-1.76805421500992e-006,-3.65766652977238e-005,-.000267816680517056,-.00110639036597811,-.00312801719432003,-.00665764172873835,-.0112074711567434,-.0152399199046756,-.0166462936129339,-.0138469870744205,-.00688452453082954,-1.15829660203181e-008,-3.69169293144801e-006,-7.63719888164115e-005,-.000559200582196592,-.00231014041245373,-.00653129234821247,-.0139011398527498,-.0234011727115081,-.0318209159596439,-.0347574208663796,-.0289124755737218,-.0143748705956402,-1.47647475649067e-008,-4.70578210489157e-006,-9.73509836708664e-005,-.000712810123054713,-.00294472417250793,-.00832540496318534,-.0177197118967874,-.0298293552102391,-.0405619588803169,-.0443051066712203,-.0368545847906068,-.0183235740526672,-1.42950269960305e-008,-4.55607398170153e-006,-9.42538931700292e-005,-.000690133028507038,-.00285104173686768,-.00806054341115132,-.0171559831153849,-.0288803744276712,-.0392715347591976,-.0428955993418035,-.0356821057856366,-.0177406342096373,-1.0289151187989e-008,-3.27933161891947e-006,-6.78412539656127e-005,-.000496737996514833,-.00205209822143352,-.0058017483868841,-.0123483854981035,-.0207872667141207,-.0282665264384568,-.0308750243738453,-.0256829582228968,-.0127692006181209,-3.73795071866204e-009,-1.19134997219017e-006,-2.46460820122577e-005,-.000180460187350232,-.000745507756803769,-.00210771998156325,-.00448605094857911,-.00755181619294033,-.0102689600807936,-.0112166025591737,-.00933037433240469,-.00463892907735314,3.73795071866204e-009,1.19134997219017e-006,2.46460820122577e-005,.000180460187350232,.000745507756803769,.00210771998156325,.0044860509485791,.00755181619294033,.0102689600807936,.0112166025591737,.00933037433240469,.00463892907735314,1.0289151187989e-008,3.27933161891947e-006,6.78412539656127e-005,.000496737996514833,.00205209822143352,.0058017483868841,.0123483854981035,.0207872667141207,.0282665264384568,.0308750243738453,.0256829582228968,.0127692006181209,1.42950269960305e-008,4.55607398170153e-006,9.42538931700292e-005,.000690133028507038,.00285104173686769,.00806054341115132,.0171559831153849,.0288803744276712,.0392715347591976,.0428955993418035,.0356821057856366,.0177406342096373,1.47647475649067e-008,4.70578210489157e-006,9.73509836708664e-005,.000712810123054713,.00294472417250793,.00832540496318534,.0177197118967874,.0298293552102391,.0405619588803169,.0443051066712203,.0368545847906068,.0183235740526672,1.15829660203181e-008,3.69169293144801e-006,7.63719888164114e-005,.000559200582196592,.00231014041245373,.00653129234821247,.0139011398527498,.0234011727115081,.0318209159596439,.0347574208663796,.0289124755737218,.0143748705956402,5.54740393494954e-009,1.76805421500992e-006,3.65766652977238e-005,.000267816680517056,.00110639036597811,.00312801719432003,.00665764172873835,.0112074711567434,.0152399199046756,.0166462936129339,.0138469870744205,.00688452453082954},
			{6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978,1.20099435917211e-008,3.82777811718284e-006,7.91872544617432e-005,.000579814137143969,.00239529806043259,.00677205238670959,.014413571204413,.024263799423424,.0329939158106694,.0360386677528614,.0299782629188682,.0149047648667097,9.63393962702057e-009,3.07050428713007e-006,6.35211333748341e-005,.000465105797498382,.00192142092314525,.00543229394429081,.0115620422138201,.0194635367753551,.0264665183938168,.0289089075829344,.0240474714038134,.0119560598918884,5.47172008094484e-010,1.743932453074e-007,3.60776457511335e-006,2.64162827510142e-005,.000109129575813761,.000308534130494273,.000656681077600405,.00110545663708768,.00150319999683172,.00164191863623543,.00136580710768774,.000679059819058227,-1.07271545504665e-008,-3.41893091623005e-006,-7.07292178810733e-005,-.00051788385284143,-.00213945488523406,-.00604872554323789,-.0128740493036516,-.0216721689330841,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859674,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175488,-.0370933112099723,-.0504394866543156,-.0550941546796317,-.0458293038356739,-.0227856763924034,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175489,-.0370933112099723,-.0504394866543156,-.0550941546796318,-.0458293038356739,-.0227856763924034,-1.07271545504666e-008,-3.41893091623005e-006,-7.07292178810734e-005,-.000517883852841431,-.00213945488523406,-.0060487255432379,-.0128740493036517,-.0216721689330842,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859675,5.47172008094458e-010,1.74393245307392e-007,3.60776457511317e-006,2.6416282751013e-005,.000109129575813756,.000308534130494258,.000656681077600374,.00110545663708762,.00150319999683165,.00164191863623535,.00136580710768767,.000679059819058194,9.63393962702055e-009,3.07050428713007e-006,6.3521133374834e-005,.000465105797498381,.00192142092314524,.0054322939442908,.01156204221382,.0194635367753551,.0264665183938168,.0289089075829343,.0240474714038134,.0119560598918883,1.20099435917211e-008,3.82777811718283e-006,7.91872544617431e-005,.000579814137143968,.00239529806043259,.00677205238670958,.014413571204413,.024263799423424,.0329939158106693,.0360386677528614,.0299782629188681,.0149047648667096,6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978},
			{2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387107,.0553680555817044,.0585549188607202,.0369926009098689,.00167071599568083,-.0222958057593638,-.0178801188985916,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387106,.0553680555817044,.0585549188607202,.036992600909869,.00167071599568083,-.0222958057593638,-.0178801188985916,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967},
			{-4.91028494999023e-008,-1.49610469790016e-005,-.000284808486486742,-.00183936641356981,-.00637694142212144,-.0142111682876331,-.0218573742001049,-.0231154364976192,-.0146033865958862,-.000659540312839943,.00880160526598481,.00705844634424949,-1.02526631182363e-007,-3.12386299642582e-005,-.000594679432029528,-.00384059262983655,-.0133150382903681,-.0296728850674858,-.0456381445486889,-.0482649755879333,-.0304918359479907,-.00137711858055358,.0183777305412829,.0147380189220751,-1.30690172571732e-007,-3.98197218991008e-005,-.000758034831541094,-.00489558378913558,-.0169726112318295,-.037823874884499,-.0581747095182799,-.0615231176135874,-.0388677873847604,-.00175540601372305,.0234259991596233,.0187864773676752,-1.26532440654065e-007,-3.85529110484099e-005,-.000733918973770133,-.00473983738085433,-.0164326504524067,-.0366205592199579,-.0563239594442265,-.0595658424432908,-.0376312610492393,-.00169956013435736,.0226807325302158,.0181888108787948,-9.10744283614461e-008,-2.7749281664479e-005,-.000528253866393645,-.00341159925277751,-.0118277513551488,-.0263584301385007,-.0405403735431023,-.0428737881159049,-.0270859043804411,-.00122329472902273,.0163249419621616,.0130917853539995,-3.30864731915692e-008,-1.00810499763263e-005,-.0001919096029833,-.00123940154495863,-.00429690951861456,-.00957576685179106,-.0147279319403104,-.0155756392506594,-.00984005132149962,-.000444411334612696,.00593069607246235,.00475612104229521,3.30864731915692e-008,1.00810499763263e-005,.0001919096029833,.00123940154495863,.00429690951861456,.00957576685179106,.0147279319403104,.0155756392506594,.00984005132149962,.000444411334612696,-.00593069607246235,-.00475612104229521,9.10744283614461e-008,2.7749281664479e-005,.000528253866393645,.00341159925277751,.0118277513551488,.0263584301385007,.0405403735431023,.0428737881159049,.0270859043804411,.00122329472902273,-.0163249419621616,-.0130917853539995,1.26532440654065e-007,3.85529110484099e-005,.000733918973770133,.00473983738085433,.0164326504524067,.0366205592199579,.0563239594442265,.0595658424432908,.0376312610492393,.00169956013435736,-.0226807325302158,-.0181888108787948,1.30690172571732e-007,3.98197218991008e-005,.000758034831541094,.00489558378913558,.0169726112318295,.037823874884499,.0581747095182799,.0615231176135874,.0388677873847604,.00175540601372305,-.0234259991596233,-.0187864773676752,1.02526631182363e-007,3.12386299642582e-005,.000594679432029528,.00384059262983655,.0133150382903681,.0296728850674858,.0456381445486889,.0482649755879333,.0304918359479907,.00137711858055358,-.0183777305412829,-.0147380189220751,4.91028494999023e-008,1.49610469790016e-005,.000284808486486742,.00183936641356981,.00637694142212144,.0142111682876331,.0218573742001049,.0231154364976192,.0146033865958862,.000659540312839943,-.00880160526598481,-.00705844634424949},
			{1.40569330619306e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201856,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016855,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016854,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201857,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,1.40569330619305e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917},
			{-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{6.48385134009702e-011,1.0745664478345e-007,5.33473591100581e-006,7.00568237223647e-005,.000443341599207306,.00173443714397409,.00474854200551721,.00971711798722703,.0153335904273431,.0186751535620109,.0167126847540202,.0086472609164468,1.35382659409353e-010,2.24369214821818e-007,1.11389156998738e-005,.000146278478759263,.000925696188564719,.00362150056830361,.0099149442407531,.0202893188924589,.0320164997847783,.0389936755407666,.0348960454086467,.0180554599121578,1.7257158376693e-010,2.86002290982243e-007,1.41987188917692e-005,.000186460428984299,.00117998019867963,.00461630825846974,.0126385287307287,.0258626910573308,.0408112685822381,.0497050388456101,.0444818106664909,.0230152024363774,1.67081451128867e-010,2.76903513083773e-007,1.37470057631949e-005,.00018052844143149,.00114244071703594,.00446944662526241,.012236450951948,.025039904355241,.0395129130085492,.0481237399429412,.0430666817373268,.0222830047517444,1.20260445247909e-010,1.9930722141311e-007,9.89470119356512e-006,.000129939203901917,.000822296121872419,.00321697972775721,.00880744708519725,.0180230063024996,.0284402635920669,.0346380903052428,.0309981645843277,.0160386688935279,4.36894754025765e-011,7.24064170021469e-008,3.59465078913488e-006,4.72056763219235e-005,.000298732356396325,.00116869812344953,.00319966171749153,.00654758668912568,.0103320771354786,.012583688604042,.0112613382258831,.00582669579070132,-4.36894754025765e-011,-7.24064170021469e-008,-3.59465078913488e-006,-4.72056763219235e-005,-.000298732356396325,-.00116869812344953,-.00319966171749153,-.00654758668912568,-.0103320771354786,-.012583688604042,-.0112613382258831,-.00582669579070132,-1.20260445247909e-010,-1.9930722141311e-007,-9.89470119356512e-006,-.000129939203901917,-.000822296121872419,-.00321697972775721,-.00880744708519725,-.0180230063024996,-.0284402635920669,-.0346380903052428,-.0309981645843277,-.0160386688935279,-1.67081451128867e-010,-2.76903513083773e-007,-1.37470057631949e-005,-.00018052844143149,-.00114244071703594,-.00446944662526241,-.012236450951948,-.025039904355241,-.0395129130085492,-.0481237399429412,-.0430666817373268,-.0222830047517444,-1.7257158376693e-010,-2.86002290982243e-007,-1.41987188917692e-005,-.000186460428984299,-.00117998019867963,-.00461630825846974,-.0126385287307287,-.0258626910573308,-.0408112685822381,-.0497050388456102,-.0444818106664909,-.0230152024363774,-1.35382659409353e-010,-2.24369214821818e-007,-1.11389156998738e-005,-.000146278478759263,-.000925696188564719,-.00362150056830361,-.0099149442407531,-.0202893188924589,-.0320164997847783,-.0389936755407666,-.0348960454086467,-.0180554599121578,-6.48385134009701e-011,-1.0745664478345e-007,-5.33473591100581e-006,-7.00568237223647e-005,-.000443341599207306,-.00173443714397409,-.00474854200551721,-.00971711798722703,-.0153335904273431,-.0186751535620109,-.0167126847540202,-.0086472609164468},
			{-4.13541370363107e-010,-6.55869442144476e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-4.13541370363107e-010,-6.55869442144475e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289},
			{4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906685,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906686,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[144],
		T o_output[56]
	) {
		static const T matrix[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371},
			{.00373129696748193,.00796011451098884,.0106231924821284,.0112513152129516,.0098479805408067,.00687246202231334,.00311875829553411,-.000484532664068795,-.00310395593642735,-.00420377879992814,-.00367166610627629,-.00183948563288991,.0077909390577761,.016620699876881,.0221812002498267,.0234927189949964,.0205625595883267,.0143496841069869,.00651195978990583,-.00101170303252743,-.00648105249983563,-.00877747998297588,-.00766642996357517,-.00384084155948903,.0099310701835657,.0211863211550268,.0282742625507885,.0299460487897454,.0262109895753331,.0182914689643539,.00830076184995504,-.00128961268294456,-.00826136448532482,-.0111886088569481,-.00977235907777577,-.00489590109849087,.00961512655393934,.0205123068664937,.0273747549478629,.0289933555580273,.0253771222247394,.0177095505014919,.00803668428539622,-.00124858539139083,-.00799854029488872,-.0108326583272072,-.0094614646283383,-.00474014459543961,.00692069282783456,.0147641712479721,.0197035649160317,.0208685872972819,.0182657260709564,.0127468274548747,.00578457527121418,-.000898696019720848,-.00575712031884831,-.00779703734224594,-.00681009137288203,-.00341182037704094,.00251422184947983,.00536368292374106,.00715811764181876,.00758135918698483,.00663576447137033,.00463080112004091,.00210148115205368,-.000326487712290173,-.00209150702911081,-.00283258947258012,-.00247404139334002,-.0012394818772994,-.00251422184947983,-.00536368292374106,-.00715811764181876,-.00758135918698483,-.00663576447137033,-.00463080112004091,-.00210148115205368,.000326487712290173,.00209150702911081,.00283258947258012,.00247404139334002,.0012394818772994,-.00692069282783456,-.0147641712479721,-.0197035649160317,-.0208685872972819,-.0182657260709564,-.0127468274548747,-.00578457527121418,.000898696019720848,.00575712031884831,.00779703734224594,.00681009137288203,.00341182037704094,-.00961512655393934,-.0205123068664937,-.0273747549478629,-.0289933555580273,-.0253771222247394,-.0177095505014919,-.00803668428539622,.00124858539139083,.00799854029488872,.0108326583272072,.0094614646283383,.00474014459543961,-.0099310701835657,-.0211863211550268,-.0282742625507885,-.0299460487897454,-.026210989575333,-.0182914689643539,-.00830076184995504,.00128961268294456,.00826136448532482,.0111886088569481,.00977235907777577,.00489590109849087,-.0077909390577761,-.016620699876881,-.0221812002498267,-.0234927189949964,-.0205625595883267,-.0143496841069869,-.00651195978990583,.00101170303252743,.00648105249983563,.00877747998297588,.00766642996357517,.00384084155948903,-.00373129696748193,-.00796011451098884,-.0106231924821284,-.0112513152129516,-.00984798054080671,-.00687246202231334,-.00311875829553411,.000484532664068795,.00310395593642735,.00420377879992814,.00367166610627629,.00183948563288991},
			{-.00463860142102606,-.00989570082574293,-.0132063344657042,-.0139871919040607,-.0122426215144303,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173,-.0080781328720134,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782068,.00104899697753615,.00671996056648802,.00910104020299664,.00794903405616202,.00398242474081937,-.006479984172588,-.0138239910928359,-.0184488449315262,-.0195396788666562,-.0171025674430431,-.011935111442324,-.00541621440729949,.000841467195356086,.00539050778207919,.00730052320305626,.00637642580128229,.00319455617998525,-.000368039046268343,-.000785151377213537,-.00104782590705614,-.00110978122522199,-.00097136234330704,-.000677870024886959,-.000307620872482971,4.77922130358021e-005,.00030616083159755,.000414642617227202,.000362157315203368,.000181438932321165,.00721530281433017,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.013289452703208,.00603082137473739,-.00093695300190748,-.00600219768054942,-.00812895281996605,-.00709999311788209,-.00355705964429505,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450082,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450083,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.00721530281433019,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.0132894527032081,.0060308213747374,-.000936953001907482,-.00600219768054943,-.00812895281996606,-.0070999931178821,-.00355705964429506,-.000368039046268326,-.000785151377213499,-.00104782590705609,-.00110978122522194,-.000971362343306994,-.000677870024886926,-.000307620872482956,4.77922130357998e-005,.000306160831597535,.000414642617227182,.000362157315203351,.000181438932321156,-.00647998417258799,-.0138239910928359,-.0184488449315262,-.0195396788666561,-.0171025674430431,-.011935111442324,-.00541621440729948,.000841467195356084,.00539050778207918,.00730052320305625,.00637642580128228,.00319455617998524,-.00807813287201338,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782067,.00104899697753615,.00671996056648801,.00910104020299662,.007949034056162,.00398242474081937,-.00463860142102606,-.00989570082574293,-.0132063344657043,-.0139871919040607,-.0122426215144304,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173},
			{.00517900543192415,.0110485647887122,.0147448922047198,.0156167207038177,.0136689052516612,.00953891326647014,.00432880746137177,-.000672526824057758,-.00430826192482315,-.00583480581394941,-.00509623835206271,-.00255318892270115,.00646894868289895,.0138004486724991,.0184174263304009,.0195064025624373,.0170734415684041,.0119147858063732,.00540699053010265,-.000840034166769796,-.00538132768356915,-.00728809032569326,-.00636556667272444,-.00318911581302657,-.000274950685842771,-.000586562517874416,-.000782798604414518,-.000829083522805726,-.000725675020631865,-.000506415909248477,-.000229814120882911,3.57040968488888e-005,.000228723369108405,.000309766784644738,.000270556624922728,.00013554746265675,-.00936530034663045,-.0199793433325387,-.0266634870132945,-.0282400321341923,-.0247177580278924,-.0172494099292266,-.00782787014830752,.00121614386801799,.00779071723872097,.0105511974509746,.00921563096089514,.00461698320596323,-.0122814780526446,-.026200533625503,-.0349659934482049,-.0370334449537273,-.0324144011931623,-.0226205505030159,-.0102653211180761,.00159482810706601,.0102165994939113,.013836641124904,.0120852044460302,.0060546246051647,-.00561022419736452,-.0119684998091909,-.0159725939896759,-.0169170134165105,-.0148070172936393,-.0103331503949069,-.00468923631858351,.000728523325820062,.00466698010205544,.00632062838987608,.00552056569433968,.00276577471540888,.00561022419736453,.0119684998091909,.0159725939896759,.0169170134165105,.0148070172936393,.010333150394907,.00468923631858351,-.000728523325820062,-.00466698010205545,-.00632062838987609,-.00552056569433968,-.00276577471540888,.0122814780526447,.0262005336255031,.034965993448205,.0370334449537275,.0324144011931624,.022620550503016,.0102653211180761,-.00159482810706601,-.0102165994939113,-.0138366411249041,-.0120852044460302,-.00605462460516473,.00936530034663043,.0199793433325386,.0266634870132944,.0282400321341923,.0247177580278923,.0172494099292265,.0078278701483075,-.00121614386801798,-.00779071723872095,-.0105511974509746,-.00921563096089512,-.00461698320596322,.00027495068584276,.000586562517874393,.000782798604414488,.000829083522805695,.000725675020631837,.000506415909248458,.000229814120882903,-3.57040968488874e-005,-.000228723369108396,-.000309766784644726,-.000270556624922718,-.000135547462656745,-.00646894868289895,-.0138004486724991,-.0184174263304009,-.0195064025624373,-.0170734415684041,-.0119147858063732,-.00540699053010265,.000840034166769796,.00538132768356914,.00728809032569326,.00636556667272443,.00318911581302657,-.00517900543192412,-.0110485647887121,-.0147448922047197,-.0156167207038176,-.0136689052516611,-.00953891326647009,-.00432880746137175,.000672526824057755,.00430826192482312,.00583480581394939,.00509623835206268,.00255318892270114},
			{-.00541977053728884,-.0115621979370935,-.015430362720618,-.0163427213723106,-.0143043545587521,-.0099823647144083,-.00453004798877423,.000703791705661616,.0045085473174344,.00610605821077421,.00533315572547085,.0026718832952154,-.00347716631903209,-.00741796815275411,-.00989966958442869,-.0104850122207482,-.00917725566884735,-.00640439334662279,-.00290634634470048,.000451532181612414,.00289255214258342,.00391746842536516,.0034215967881247,.00171420220442671,.00694040822993805,.0148062308480902,.019759695669824,.0209280369217824,.0183177607650484,.012783140124006,.00580105414555624,-.000901256190189091,-.00577352098058376,-.00781924924071369,-.00682949169783763,-.00342153983899534,.0112848679704134,.0240744282936295,.0321285937918819,.0340282769713043,.0297840566287605,.0207849515140105,.0094323169405852,-.00146541194362748,-.00938754894982324,-.0127138335939062,-.0111045214577846,-.00556330752592008,.0017039064892699,.00363500704682817,.00485110854611964,.00513794242893165,.00449710599180968,.00313832770190446,.00142418910757755,-.000221263104428595,-.00141742957169544,-.00191966655001643,-.00167667590101747,-.000840005910576199,-.0110322454417378,-.0235354992633652,-.0314093645879802,-.0332665215482455,-.0291173121245713,-.0203196605576931,-.00922116553296151,.00143260730012424,.00917739971635521,.012429222307352,.0108559361577392,.00543876758281373,-.011032245441738,-.0235354992633655,-.0314093645879806,-.0332665215482459,-.0291173121245716,-.0203196605576934,-.00922116553296161,.00143260730012425,.00917739971635532,.0124292223073521,.0108559361577394,.00543876758281379,.00170390648926984,.00363500704682805,.00485110854611947,.00513794242893148,.00449710599180952,.00313832770190435,.0014241891075775,-.000221263104428587,-.00141742957169539,-.00191966655001636,-.00167667590101741,-.00084000591057617,.0112848679704135,.0240744282936297,.0321285937918822,.0340282769713046,.0297840566287608,.0207849515140107,.00943231694058528,-.0014654119436275,-.00938754894982332,-.0127138335939063,-.0111045214577847,-.00556330752592013,.00694040822993806,.0148062308480902,.0197596956698241,.0209280369217824,.0183177607650485,.012783140124006,.00580105414555625,-.000901256190189092,-.00577352098058377,-.00781924924071369,-.00682949169783764,-.00342153983899535,-.00347716631903199,-.00741796815275391,-.00989966958442842,-.0104850122207479,-.0091772556688471,-.00640439334662262,-.0029063463447004,.000451532181612402,.00289255214258334,.00391746842536505,.0034215967881246,.00171420220442666,-.00541977053728884,-.0115621979370935,-.0154303627206179,-.0163427213723106,-.0143043545587521,-.00998236471440829,-.00453004798877423,.000703791705661616,.00450854731743439,.0061060582107742,.00533315572547085,.0026718832952154},
			{-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204719,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359087,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359088,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204718,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029},
			{.00669658564383886,.0128558782631327,.0137727308070988,.00947626072033451,.00236330891169302,-.00410612209904174,-.00717100835620119,-.0060731527881528,-.00224476908541414,.00166655945599322,.00335856916390389,.00214806069987574,.013982454653438,.0268430427690799,.028757428666868,.0197864095008088,.00493458330070284,-.00857357302734623,-.0149730481312158,-.0126807283562296,-.0046870724296822,.00347977211971043,.00701268430396727,.00448514555420918,.0178233634574797,.0342166893750909,.0366569472911615,.025221635027061,.00629008810395184,-.010928689703139,-.0190860678989135,-.0161640596018723,-.00597458726213122,.00443564772967482,.00893903282793333,.00571719210637408,.0172563371411996,.033128131461881,.0354907558571953,.0244192426596455,.00608997741804402,-.0105810081513981,-.0184788703405621,-.0156498218041611,-.00578451381080864,.00429453356800305,.00865464953138992,.00553530733545099,.0124206174529098,.0238446806208565,.025545230022447,.0175762717825614,.00438339139919734,-.00761590675004047,-.0133005618506412,-.0112642937052734,-.00416352743964371,.00309108231662293,.00622936896392408,.00398415575306398,.00451229213042641,.00866254557198487,.0092803389877543,.00638529228900029,.00159244438452158,-.00276678645200671,-.00483196755688163,-.00409221071607914,-.00151256990016366,.00112296079197981,.00226307046811526,.00144740587326696,-.00451229213042641,-.00866254557198487,-.0092803389877543,-.00638529228900029,-.00159244438452158,.00276678645200671,.00483196755688163,.00409221071607914,.00151256990016366,-.00112296079197981,-.00226307046811526,-.00144740587326696,-.0124206174529098,-.0238446806208565,-.025545230022447,-.0175762717825614,-.00438339139919734,.00761590675004047,.0133005618506412,.0112642937052734,.00416352743964371,-.00309108231662293,-.00622936896392408,-.00398415575306398,-.0172563371411996,-.033128131461881,-.0354907558571953,-.0244192426596455,-.00608997741804402,.0105810081513981,.0184788703405621,.0156498218041611,.00578451381080864,-.00429453356800305,-.00865464953138992,-.00553530733545099,-.0178233634574797,-.0342166893750909,-.0366569472911615,-.025221635027061,-.00629008810395184,.010928689703139,.0190860678989135,.0161640596018723,.00597458726213122,-.00443564772967482,-.00893903282793333,-.00571719210637408,-.013982454653438,-.0268430427690799,-.028757428666868,-.0197864095008088,-.00493458330070284,.00857357302734624,.0149730481312158,.0126807283562296,.0046870724296822,-.00347977211971043,-.00701268430396727,-.00448514555420918,-.00669658564383886,-.0128558782631327,-.0137727308070988,-.00947626072033451,-.00236330891169302,.00410612209904174,.00717100835620119,.00607315278815281,.00224476908541414,-.00166655945599322,-.00335856916390389,-.00214806069987574},
			{-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203508,.00754990434136365,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315,-.0144978834787229,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606307,.00888961672521672,.0155249927504182,.0131481722408983,.00485984983510451,-.00360804536645972,-.00727118967533986,-.00465047942165492,-.0116296744515833,-.0223262550411753,-.0239185137194828,-.0164570175096932,-.00410425771180905,.00713092698429712,.012453584125954,.0105469852216292,.00389839465524694,-.00289424267204597,-.00583268371028218,-.00373044532996883,-.000660522337643703,-.00126804840771953,-.00135848278992925,-.000934697503477068,-.000233106602371594,.000405010181569894,.000707317348663856,.000599029608497395,.000221414344955362,-.000164382239881848,-.000331274782891918,-.000211875446734245,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627044,.00456998989967055,-.00794011160648144,-.0138667592696614,-.0117438083374094,-.00434076645531976,.00322266794806499,.006494549688551,.0041537590171735,.0221636560036439,.0425490359719281,.0455835382498503,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183392,.00551580349401449,.0111158395594773,.0071094257520315,.0221636560036439,.0425490359719281,.0455835382498504,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183393,.0055158034940145,.0111158395594773,.00710942575203151,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627045,.00456998989967055,-.00794011160648145,-.0138667592696614,-.0117438083374094,-.00434076645531976,.003222667948065,.00649454968855101,.0041537590171735,-.000660522337643672,-.00126804840771947,-.00135848278992918,-.000934697503477023,-.000233106602371583,.000405010181569874,.000707317348663822,.000599029608497366,.000221414344955352,-.00016438223988184,-.000331274782891902,-.000211875446734235,-.0116296744515833,-.0223262550411753,-.0239185137194827,-.0164570175096932,-.00410425771180904,.0071309269842971,.012453584125954,.0105469852216291,.00389839465524693,-.00289424267204597,-.00583268371028217,-.00373044532996882,-.0144978834787228,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606306,.0088896167252167,.0155249927504182,.0131481722408983,.0048598498351045,-.00360804536645972,-.00727118967533985,-.00465047942165491,-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203509,.00754990434136366,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315},
			{.00929479849593368,.0178438392786048,.0191164220692825,.0131529616098434,.00328025075555347,-.00569925922554386,-.0099532928015145,-.0084294794994272,-.00311571858085951,.0023131690011613,.00466166271491865,.00298148824255666,.0116098689755076,.0222882331591425,.0238777802016866,.016428991010076,.00409726811134705,-.00711878292950237,-.0124323755217502,-.0105290235783908,-.00389175564205705,.00288931374181348,.00582275057946566,.00372409233650869,-.000493455983937409,-.000947320081473204,-.0010148808353067,-.000698283842916624,-.000174146794559474,.000302570687259703,.000528415101729981,.000447516651629383,.000165411867579778,-.000122804930734209,-.000247485231958592,-.0001582856492233,-.0168079722487348,-.032267375730295,-.034568612948002,-.0237847666975931,-.00593174383418366,.0103060858116714,.01799874082957,.0152431983931265,.0056342169725212,-.0041829503237926,-.00842977904035172,-.00539148553491988,-.0220416574634008,-.0423148273072351,-.045332626345935,-.0311908939779877,-.00777877686962253,.0135152063489619,.0236032088979544,.0199896425729312,.00738860576070614,-.00548544207826128,-.011054653074738,-.00707029233633635,-.0100687099322352,-.0193295682367182,-.0207081099006063,-.0142481147124679,-.00355337379041754,.00617379580587153,.010782032352062,.00913134200774618,.00337514219752845,-.00250577005054177,-.00504980605001451,-.00322973545836477,.0100687099322352,.0193295682367182,.0207081099006063,.0142481147124679,.00355337379041755,-.00617379580587153,-.010782032352062,-.00913134200774619,-.00337514219752846,.00250577005054177,.00504980605001452,.00322973545836478,.0220416574634009,.0423148273072353,.0453326263459352,.0311908939779879,.00777877686962257,-.013515206348962,-.0236032088979545,-.0199896425729313,-.00738860576070618,.0054854420782613,.0110546530747381,.00707029233633639,.0168079722487348,.032267375730295,.0345686129480019,.023784766697593,.00593174383418365,-.0103060858116714,-.01799874082957,-.0152431983931265,-.00563421697252118,.00418295032379259,.0084297790403517,.00539148553491986,.000493455983937391,.000947320081473168,.00101488083530666,.000698283842916597,.000174146794559467,-.000302570687259691,-.000528415101729961,-.000447516651629366,-.000165411867579771,.000122804930734204,.000247485231958582,.000158285649223294,-.0116098689755076,-.0222882331591425,-.0238777802016866,-.016428991010076,-.00409726811134704,.00711878292950237,.0124323755217502,.0105290235783908,.00389175564205705,-.00288931374181348,-.00582275057946566,-.00372409233650869,-.00929479849593364,-.0178438392786047,-.0191164220692824,-.0131529616098433,-.00328025075555346,.00569925922554383,.00995329280151445,.00842947949942716,.00311571858085949,-.00231316900116129,-.00466166271491862,-.00298148824255664},
			{-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993},
			{.00997527550886585,.0163494681388942,.0117058404244761,.00065538634573533,-.00785594167460167,-.0080795624708938,-.00168671898857876,.00464312263238137,.00560726911956695,.00166497716447436,-.00230734475380579,-.00231558708502107,.0208283511742426,.0341376499933582,.0244417665390733,.00136844510727013,-.0164031872460608,-.0168701069287199,-.00352186517506783,.0096948288442119,.011707964381237,.00347646829883996,-.0048177302741764,-.00483494024185036,.0265497927509654,.043515087908134,.031155794841906,.00174434998167423,-.0209090589166153,-.0215042390488418,-.00448930353212625,.0123579487601599,.0149240823364804,.00443143636610296,-.00614113614848276,-.00616307360628593,.0257051468334037,.0421307139622166,.0301646151679837,.0016888558350856,-.0202438653529731,-.0208201106304678,-.00434648238332583,.0119647972554527,.014449292746328,.00429045618329429,-.0059457641685332,-.00596700371564648,.0185018287934532,.0303244817750634,.0217116264332397,.00121559008085388,-.0145709547315911,-.0149857195853513,-.00312847358668079,.00861191853147071,.0104001872586715,.00308814753183857,-.00427959083079407,-.00429487844875196,.00672153834378315,.0110165956718055,.00788762728301515,.000441612309247667,-.00529348920192007,-.00544416932653508,-.0011365446845897,.00312862805450344,.00377828906656089,.00112189461259403,-.00155473462575028,-.0015602884718828,-.00672153834378315,-.0110165956718055,-.00788762728301515,-.000441612309247667,.00529348920192007,.00544416932653508,.00113654468458969,-.00312862805450344,-.00377828906656089,-.00112189461259403,.00155473462575028,.0015602884718828,-.0185018287934532,-.0303244817750634,-.0217116264332397,-.00121559008085388,.0145709547315911,.0149857195853513,.00312847358668079,-.00861191853147071,-.0104001872586715,-.00308814753183857,.00427959083079407,.00429487844875196,-.0257051468334037,-.0421307139622166,-.0301646151679837,-.0016888558350856,.0202438653529731,.0208201106304678,.00434648238332583,-.0119647972554527,-.014449292746328,-.00429045618329429,.0059457641685332,.00596700371564648,-.0265497927509654,-.043515087908134,-.031155794841906,-.00174434998167423,.0209090589166153,.0215042390488418,.00448930353212624,-.0123579487601599,-.0149240823364804,-.00443143636610296,.00614113614848276,.00616307360628593,-.0208283511742426,-.0341376499933582,-.0244417665390733,-.00136844510727013,.0164031872460608,.0168701069287199,.00352186517506784,-.0096948288442119,-.011707964381237,-.00347646829883996,.0048177302741764,.00483494024185036,-.00997527550886584,-.0163494681388943,-.0117058404244761,-.00065538634573533,.00785594167460167,.0080795624708938,.00168671898857876,-.00464312263238137,-.00560726911956695,-.00166497716447436,.00230734475380579,.00231558708502107},
			{-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962,-.0215961371406428,-.0353960505441994,-.0253427521805069,-.00141888947227826,.017007850422056,.0174919819510481,.00365169007740165,-.0100522048780116,-.0121395497079787,-.0036046197569249,.0049953240579315,.00501316842875873,-.0173236351861994,-.0283934234472885,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064034,-.00806351288610232,-.00973790493624807,-.00289149477275801,.00400706714598473,.00402138125078621,-.000983918170479597,-.00161264105089006,-.00115461363289064,-6.46444836215388e-005,.000774876218005063,.000796933209270456,.000166370689198568,-.000457977598884836,-.000553076851722798,-.000164226169402932,.000227586539019653,.000228399527036137,.019289441206072,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433103,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543011,-.0076638869878077,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543012,-.0076638869878077,.0192894412060721,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433104,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,-.00098391817047955,-.00161264105088998,-.00115461363289059,-6.46444836215357e-005,.000774876218005026,.000796933209270418,.00016637068919856,-.000457977598884814,-.000553076851722771,-.000164226169402924,.000227586539019642,.000228399527036126,-.0173236351861993,-.0283934234472884,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064033,-.0080635128861023,-.00973790493624805,-.002891494772758,.00400706714598472,.0040213812507862,-.0215961371406427,-.0353960505441994,-.0253427521805068,-.00141888947227826,.0170078504220559,.017491981951048,.00365169007740164,-.0100522048780115,-.0121395497079786,-.00360461975692489,.00499532405793149,.00501316842875872,-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962},
			{-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758},
			{.0133541389835151,.0174668385049753,.00479454927113329,-.00864193040388231,-.00841992604968446,.0017949650137812,.00751289805161027,.00275957692744254,-.00421283180475702,-.00418627184884443,.000779677757981102,.00236085727977567,.0278834099500396,.0364707166195125,.0100110073002612,-.0180443298147953,-.0175807852593257,.00374788261429831,.0156869129896357,.00576198996068088,-.00879638263519743,-.00874092551138884,.00162796527590073,.00492946429917962,.0355428401014885,.0464890360080293,.0127609798215341,-.0230010149581752,-.0224101370833209,.00477740680633519,.0199960277841051,.00734477914307506,-.0112127039710897,-.0111420130589112,.00207515901375732,.00628356293888253,.0344120924880564,.0450100499065603,.0123550064261315,-.0222692686290634,-.0216971887384138,.00462542003968309,.0193598810769079,.00711111488148247,-.010855986887727,-.010787544911762,.00200914062311667,.00608365984231783,.0247688390097475,.0323969453570202,.00889277992146099,-.0160287820256653,-.0156170152981078,.00332924492618789,.0139346881567349,.00511837691184837,-.00781382857224871,-.00776456599733339,.00144612190203756,.00437884418905171,.00899828352070364,.0117695019702978,.00323066232490896,-.00582310398568677,-.00567351305179212,.00120948300176336,.00506233960976318,.00185945762740475,-.0028386895666607,-.00282079294195926,.00052536232622418,.00159079242634532,-.00899828352070364,-.0117695019702978,-.00323066232490896,.00582310398568677,.00567351305179212,-.00120948300176336,-.00506233960976318,-.00185945762740475,.0028386895666607,.00282079294195926,-.00052536232622418,-.00159079242634532,-.0247688390097475,-.0323969453570202,-.00889277992146099,.0160287820256653,.0156170152981078,-.00332924492618789,-.0139346881567349,-.00511837691184837,.00781382857224871,.00776456599733339,-.00144612190203756,-.00437884418905171,-.0344120924880564,-.0450100499065603,-.0123550064261315,.0222692686290634,.0216971887384138,-.00462542003968309,-.0193598810769079,-.00711111488148247,.010855986887727,.010787544911762,-.00200914062311667,-.00608365984231783,-.0355428401014885,-.0464890360080293,-.0127609798215341,.0230010149581752,.0224101370833208,-.00477740680633519,-.0199960277841051,-.00734477914307506,.0112127039710897,.0111420130589112,-.00207515901375732,-.00628356293888253,-.0278834099500396,-.0364707166195125,-.0100110073002612,.0180443298147953,.0175807852593257,-.00374788261429831,-.0156869129896357,-.00576198996068088,.00879638263519743,.00874092551138884,-.00162796527590073,-.00492946429917962,-.0133541389835151,-.0174668385049753,-.00479454927113329,.00864193040388231,.00841992604968446,-.0017949650137812,-.00751289805161027,-.00275957692744254,.00421283180475702,.00418627184884443,-.000779677757981102,-.00236085727977568},
			{-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292215,-.00120362482446452,.00306032152371413,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292214,-.00120362482446452,.00306032152371413,-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268},
			{1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394712,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394711,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752},
			{-3.02102579557316e-005,-.000356099425944935,-.00127919739538736,-.00291181027531203,-.00512624408687041,-.00756904967962207,-.00973625153619673,-.0110917233656258,-.0111998299356331,-.00983955325421882,-.00707171109480605,-.00324650309107673,-6.30789456599155e-005,-.00074353474146506,-.00267096107256627,-.00607985282341967,-.0107035852745446,-.0158041574533295,-.020329269696504,-.023159491603162,-.0233852178602818,-.0205449634342814,-.0147657156892056,-.00677869061452103,-8.04064095237273e-005,-.000947779933413657,-.00340466042315561,-.00774995730904432,-.0136438054243522,-.0201454787025301,-.0259136161429872,-.0295212855370935,-.0298090176429547,-.0261885598477368,-.0188217822951288,-.00864076226835519,-7.78483878402044e-005,-.000917627590643462,-.00329634573482353,-.0075034028495138,-.0132097461207719,-.0195045774155398,-.0250892093278478,-.0285821055764826,-.0288606838727134,-.0253554060687339,-.0182229926275056,-.00836586804816125,-5.60330408925769e-005,-.000660482069535263,-.00237261529082107,-.00540073456066038,-.00950799708383307,-.0140388364388575,-.0180584946102595,-.0205725813339951,-.0207730940163711,-.0182500954035624,-.0131163883981527,-.00602151231964795,-2.03562705656154e-005,-.000239946850947697,-.000861948558184859,-.00196203547440233,-.00345416486580917,-.00510017568961999,-.00656047925364063,-.0074738230372111,-.00754666738671032,-.00663008599864202,-.0047650590941222,-.00218755812714027,2.03562705656154e-005,.000239946850947697,.000861948558184859,.00196203547440233,.00345416486580917,.00510017568962,.00656047925364063,.0074738230372111,.00754666738671032,.00663008599864202,.0047650590941222,.00218755812714027,5.60330408925769e-005,.000660482069535263,.00237261529082107,.00540073456066038,.00950799708383307,.0140388364388575,.0180584946102595,.0205725813339951,.0207730940163711,.0182500954035624,.0131163883981527,.00602151231964795,7.78483878402044e-005,.000917627590643462,.00329634573482353,.00750340284951379,.0132097461207719,.0195045774155398,.0250892093278478,.0285821055764826,.0288606838727134,.0253554060687339,.0182229926275056,.00836586804816125,8.04064095237273e-005,.000947779933413657,.00340466042315561,.00774995730904432,.0136438054243522,.0201454787025301,.0259136161429872,.0295212855370935,.0298090176429547,.0261885598477368,.0188217822951288,.00864076226835519,6.30789456599155e-005,.00074353474146506,.00267096107256627,.00607985282341966,.0107035852745446,.0158041574533295,.020329269696504,.023159491603162,.0233852178602818,.0205449634342814,.0147657156892056,.00677869061452103,3.02102579557315e-005,.000356099425944935,.00127919739538736,.00291181027531203,.00512624408687041,.00756904967962207,.00973625153619672,.0110917233656258,.0111998299356331,.00983955325421882,.00707171109480605,.00324650309107673},
			{3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491522,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500883,.00403592477071172,6.54041959386344e-005,.00077094332204201,.00276941948707301,.00630397165299627,.0110981466354321,.0163867388699784,.021078658252914,.0240132093332518,.0242472564339873,.0213023030101423,.015310017502792,.00702857038146807,5.24648654853783e-005,.000618422673152875,.0022215275148126,.00505681661629785,.00890252929508159,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327876,2.97980960162188e-006,3.51241121514965e-005,.000126174516176939,.000287208412095994,.000505630616352145,.000746578428543733,.000960341876387736,.00109403977390252,.00110470293989661,.000970531112496787,.000697523094674574,.000320221068507963,-5.84183358335144e-005,-.000688598418638334,-.0024736161852904,-.00563064078419707,-.00991274715595561,-.0146364617863459,-.0188272345384989,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218605,-9.99867395890061e-005,-.00117858048818797,-.00423375321862369,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.032565920170009,-.0234052068248199,-.0107449350355781,-9.99867395890061e-005,-.00117858048818797,-.0042337532186237,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.0325659201700091,-.0234052068248199,-.0107449350355781,-5.84183358335145e-005,-.000688598418638335,-.0024736161852904,-.00563064078419708,-.00991274715595562,-.0146364617863459,-.018827234538499,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218606,2.97980960162174e-006,3.51241121514949e-005,.000126174516176933,.00028720841209598,.000505630616352121,.000746578428543697,.00096034187638769,.00109403977390247,.00110470293989656,.000970531112496741,.000697523094674541,.000320221068507948,5.24648654853782e-005,.000618422673152874,.0022215275148126,.00505681661629784,.00890252929508158,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327875,6.54041959386343e-005,.000770943322042009,.00276941948707301,.00630397165299626,.0110981466354321,.0163867388699784,.0210786582529139,.0240132093332518,.0242472564339873,.0213023030101423,.0153100175027919,.00702857038146806,3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491523,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500884,.00403592477071172},
			{-4.19315566304071e-005,-.000494262686103036,-.00177551406892167,-.00404156553827383,-.00711517904100237,-.0105057704487168,-.0135138264378992,-.0153952087107654,-.0155452596229099,-.0136572082602461,-.00981546912576032,-.00450611604885183,-5.23755171918012e-005,-.000617369491942089,-.00221774422687797,-.00504820479710383,-.008887368181188,-.0131224596692141,-.0168797370239278,-.0192297182193646,-.0194171425546437,-.0170588311884377,-.0122602238803855,-.00562846165395503,2.22612437958035e-006,2.62401471318847e-005,9.4261111981906e-005,.000214564597630484,.000377741128665099,.00055774585065616,.000717441967632789,.000817323567112939,.000825289691448544,.000725054033484014,.000521098114970246,.000239227340925692,7.58256825576655e-005,.000893785219253475,.00321069802758278,.00730844476493686,.0128665222724588,.0189977973412616,.0244373258701412,.0278394675137098,.0281108076152006,.0246966061215671,.0177495114871256,.00814850085580817,9.94363684763003e-005,.00117209306138154,.00421044877365996,.00958415647735547,.0168729143812158,.0249133527446604,.0320466477561979,.0365081520733029,.0368639823593503,.0323866629825667,.0232763742438573,.0106857901742116,4.54228976457494e-005,.000535416407238526,.00192334843497638,.00437807781360722,.00770760914409425,.0113805108641198,.0146390261754792,.0166770577030509,.0168396022821627,.0147943463773532,.0106327330839207,.00488130812483214,-4.54228976457494e-005,-.000535416407238526,-.00192334843497639,-.00437807781360723,-.00770760914409426,-.0113805108641198,-.0146390261754792,-.0166770577030509,-.0168396022821627,-.0147943463773532,-.0106327330839207,-.00488130812483215,-9.94363684763008e-005,-.00117209306138155,-.00421044877365998,-.00958415647735551,-.0168729143812159,-.0249133527446605,-.032046647756198,-.036508152073303,-.0368639823593504,-.0323866629825669,-.0232763742438574,-.0106857901742116,-7.58256825576654e-005,-.000893785219253473,-.00321069802758277,-.00730844476493685,-.0128665222724588,-.0189977973412615,-.0244373258701411,-.0278394675137098,-.0281108076152005,-.0246966061215671,-.0177495114871256,-.00814850085580816,-2.22612437958026e-006,-2.62401471318837e-005,-9.42611119819025e-005,-.000214564597630476,-.000377741128665085,-.000557745850656139,-.000717441967632762,-.000817323567112908,-.000825289691448512,-.000725054033483986,-.000521098114970227,-.000239227340925682,5.23755171918011e-005,.000617369491942089,.00221774422687796,.00504820479710383,.008887368181188,.0131224596692141,.0168797370239278,.0192297182193646,.0194171425546437,.0170588311884377,.0122602238803855,.00562846165395503,4.19315566304069e-005,.000494262686103033,.00177551406892166,.00404156553827381,.00711517904100233,.0105057704487168,.0135138264378992,.0153952087107653,.0155452596229098,.013657208260246,.00981546912576027,.00450611604885181},
			{4.38808991888882e-005,.000517240304071817,.0018580553675488,.00422945256989752,.00744595429549842,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731,2.81527019748646e-005,.000331846256550951,.0011920739999021,.00271349311244975,.00477710657881669,.00705353791336682,.00907313629206733,.0103362898376332,.0104370335006432,.00916940235127242,.00659007199462349,.00302539071720239,-5.61926656806376e-005,-.00066236362564373,-.0023793743066966,-.00541611996712565,-.00953508310299627,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136763,-.0183020855787247,-.0131537538647564,-.00603866617409735,-9.13673651610601e-005,-.00107698075043354,-.00386878177964248,-.00880642704521119,-.0155037211557789,-.0228916988068627,-.0294461454354727,-.0335456102525707,-.0338725658341033,-.0297585693083877,-.0213875568642763,-.00981866602574447,-1.37955930732719e-005,-.000162613731440516,-.000584148826303059,-.00132968575520406,-.00234091272972111,-.0034564262736338,-.0044460846527431,-.00506506439824835,-.00511443154534631,-.00449325765165264,-.00322931531199247,-.00148252410228482,8.93220195809498e-005,.00105287150952592,.00378217541095215,.00860928677962801,.0151566555762327,.0223792463037912,.0287869653955147,.0327946597841923,.033114296153308,.029092395362165,.0209087760126605,.00959886582549863,8.93220195809508e-005,.00105287150952593,.00378217541095219,.00860928677962811,.0151566555762329,.0223792463037915,.028786965395515,.0327946597841926,.0331142961533084,.0290923953621653,.0209087760126607,.00959886582549874,-1.37955930732714e-005,-.00016261373144051,-.00058414882630304,-.00132968575520402,-.00234091272972103,-.00345642627363368,-.00444608465274295,-.00506506439824818,-.00511443154534613,-.00449325765165249,-.00322931531199236,-.00148252410228477,-9.13673651610609e-005,-.00107698075043355,-.00386878177964252,-.00880642704521126,-.0155037211557791,-.0228916988068629,-.029446145435473,-.033545610252571,-.0338725658341036,-.029758569308388,-.0213875568642765,-.00981866602574456,-5.61926656806377e-005,-.000662363625643731,-.0023793743066966,-.00541611996712565,-.00953508310299628,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136764,-.0183020855787247,-.0131537538647564,-.00603866617409736,2.81527019748638e-005,.000331846256550942,.00119207399990207,.00271349311244967,.00477710657881656,.00705353791336663,.00907313629206708,.0103362898376329,.010437033500643,.00916940235127217,.00659007199462331,.0030253907172023,4.38808991888881e-005,.000517240304071817,.0018580553675488,.00422945256989751,.00744595429549841,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731},
			{8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302079,-.00878002632372056,-.00505719415206449,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302078,-.00878002632372056,-.00505719415206449,8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575},
			{-.000146293799182738,-.00163998020354913,-.00536553359148041,-.0105856065426876,-.0151909408292571,-.0168075648663272,-.0141532955069821,-.00788475067656189,-.000434896968818276,.0051187183448384,.00658493074687408,.00379284435342441,-.000305461099423673,-.00342427470479066,-.0112032211823346,-.0221027208990368,-.0317186477684445,-.0350941548540275,-.0295520468412377,-.0164633403725859,-.000908063820704495,.0106878715433549,.0137493194981995,.00791944985099954,-.00038936970167901,-.00436490545862284,-.014280688761482,-.0281742253235845,-.0404315981398701,-.0447343397636272,-.0376698430153202,-.0209857357994469,-.00115750430951867,.0136237752115456,.0175261872670667,.0100948822346398,-.000376982428752617,-.00422604186707806,-.0138263678718428,-.0272779002703818,-.0391453212702289,-.0433111769611684,-.0364714276673539,-.0203181028639926,-.00112067986803374,.0131903531422229,.0169686152112542,.00977372714511676,-.000271341159812294,-.00304178395110495,-.00995182376735971,-.0196338516919752,-.0281756815823874,-.0311741452469417,-.0262510895163342,-.0146243887667217,-.000806633285738924,.00949403857306351,.0122135234447537,.00703484899288287,-9.85756613730145e-005,-.00110505116489299,-.00361540287663619,-.0071327914908855,-.0102359570090312,-.0113252703260409,-.00953677102517345,-.00531290127842359,-.000293042197071052,.00344909387167086,.00443705684789683,.00255569369797207,9.85756613730145e-005,.00110505116489299,.00361540287663619,.0071327914908855,.0102359570090312,.0113252703260409,.00953677102517345,.00531290127842359,.000293042197071052,-.00344909387167086,-.00443705684789683,-.00255569369797207,.000271341159812294,.00304178395110495,.00995182376735971,.0196338516919752,.0281756815823874,.0311741452469417,.0262510895163341,.0146243887667217,.000806633285738924,-.00949403857306351,-.0122135234447537,-.00703484899288287,.000376982428752617,.00422604186707806,.0138263678718428,.0272779002703818,.0391453212702289,.0433111769611684,.0364714276673539,.0203181028639926,.00112067986803374,-.0131903531422229,-.0169686152112542,-.00977372714511676,.00038936970167901,.00436490545862284,.014280688761482,.0281742253235845,.0404315981398701,.0447343397636272,.0376698430153202,.0209857357994469,.00115750430951867,-.0136237752115456,-.0175261872670667,-.0100948822346398,.000305461099423673,.00342427470479066,.0112032211823346,.0221027208990368,.0317186477684445,.0350941548540275,.0295520468412377,.0164633403725859,.000908063820704495,-.0106878715433549,-.0137493194981995,-.00791944985099954,.000146293799182738,.00163998020354913,.00536553359148041,.0105856065426876,.0151909408292571,.0168075648663272,.0141532955069821,.00788475067656189,.000434896968818276,-.0051187183448384,-.00658493074687408,-.00379284435342441},
			{.000181866688959546,.00203875879391254,.00667022002460173,.013159609110584,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.0081861265388698,-.00471511470855014,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053052,.0306414104040703,.0170702209457237,.000941537361279075,-.0110818536551902,-.0142561543633291,-.00821138090239868,.000254062196557116,.00284808361770293,.00931809316299003,.0183835710409382,.026381458519847,.0291889804801732,.0245794241800656,.0136931099430282,.000755267002369574,-.00888945965934217,-.0114357681533554,-.0065868708927751,1.44297896450536e-005,.000161760576944934,.000529233101409994,.00104411859237783,.00149836891174606,.00165782573712333,.00139602005068532,.000777717813756163,4.28963620630315e-005,-.000504888309558718,-.000649509179715462,-.000374109815194419,-.000282891999889151,-.00317127097766542,-.0103754672901093,-.0204696536806732,-.0293751044500413,-.0325012110209999,-.0273685835856313,-.0152469407451353,-.000840971209593475,.00989819443838388,.0127334462466717,.00733431854530109,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.0468430913146029,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.046843091314603,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000282891999889152,-.00317127097766543,-.0103754672901093,-.0204696536806732,-.0293751044500414,-.0325012110209999,-.0273685835856313,-.0152469407451354,-.000840971209593476,.00989819443838389,.0127334462466717,.00733431854530111,1.44297896450529e-005,.000161760576944927,.000529233101409968,.00104411859237778,.00149836891174599,.00165782573712325,.00139602005068526,.000777717813756126,4.28963620630294e-005,-.000504888309558694,-.000649509179715431,-.000374109815194401,.000254062196557116,.00284808361770292,.00931809316299001,.0183835710409382,.026381458519847,.0291889804801731,.0245794241800655,.0136931099430282,.000755267002369572,-.00888945965934215,-.0114357681533554,-.00658687089277509,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053051,.0306414104040702,.0170702209457237,.000941537361279073,-.0110818536551902,-.0142561543633291,-.00821138090239866,.000181866688959546,.00203875879391254,.00667022002460174,.0131596091105841,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.00818612653886981,-.00471511470855015},
			{-.000203054430513827,-.00227627724582951,-.00744731064410978,-.0146927232745112,-.0210848843647913,-.0233287434691147,-.0196446423233178,-.010943960491264,-.000603632941579964,.00710473338089701,.00913982253690271,.00526443263155787,-.000253629525609746,-.00284323329737286,-.00930222433932735,-.0183522636004527,-.0263365305817064,-.0291392712986882,-.0245375651278731,-.0136697904136697,-.00075398077366707,.0088743208036425,.0114162929039563,.00657565337317555,1.07800533648885e-005,.000120846366765943,.000395373821517966,.000780028983233564,.00111938546758022,.00123851077219281,.0010429237719346,.000581009130510076,3.20465567117658e-005,-.000377186574041089,-.000485228391442131,-.000279485971128269,.000367187436559508,.00411623821592978,.0134671225727665,.0265691488808889,.0381282231590321,.0421858389935269,.0355238043246551,.0197901852642516,.00109156166590775,-.0128476331738681,-.0165277260852754,-.00951978008079133,.000481522671607105,.00539795707957059,.0176605302753814,.0348422802025231,.0500006319693174,.0553217127646963,.046585246282749,.0259524753061578,.00143145335939787,-.0168481435739603,-.021674147935839,-.0124840598593624,.000219961321613734,.00246580658242685,.00806739497188598,.0159160813255256,.0228404719819027,.0252711611958884,.0212803112797474,.011855185859552,.000653893142166497,-.00769629374853215,-.00990083023278439,-.00570276430932191,-.000219961321613735,-.00246580658242685,-.00806739497188599,-.0159160813255256,-.0228404719819027,-.0252711611958885,-.0212803112797474,-.0118551858595521,-.000653893142166498,.00769629374853216,.0099008302327844,.00570276430932192,-.000481522671607107,-.00539795707957062,-.0176605302753815,-.0348422802025233,-.0500006319693176,-.0553217127646966,-.0465852462827492,-.0259524753061579,-.00143145335939788,.0168481435739604,.0216741479358391,.0124840598593625,-.000367187436559507,-.00411623821592978,-.0134671225727665,-.0265691488808889,-.038128223159032,-.0421858389935268,-.035523804324655,-.0197901852642516,-.00109156166590775,.0128476331738681,.0165277260852754,.00951978008079131,-1.07800533648881e-005,-.000120846366765938,-.00039537382151795,-.000780028983233534,-.00111938546758017,-.00123851077219277,-.00104292377193456,-.000581009130510054,-3.20465567117645e-005,.000377186574041075,.000485228391442112,.000279485971128258,.000253629525609746,.00284323329737286,.00930222433932735,.0183522636004527,.0263365305817064,.0291392712986881,.0245375651278731,.0136697904136697,.00075398077366707,-.00887432080364249,-.0114162929039563,-.00657565337317555,.000203054430513826,.0022762772458295,.00744731064410975,.0146927232745111,.0210848843647912,.0233287434691146,.0196446423233177,.0109439604912639,.000603632941579961,-.00710473338089698,-.00913982253690267,-.00526443263155784},
			{.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048,.000554003280665635,.00578143006001033,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204796,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204797,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.000554003280665635,.00578143006001032,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048},
			{-.000415496844730152,-.0043360139403437,-.0123495368891892,-.0193702019109772,-.0189618233479971,-.00958981393734538,.00318359640435585,.0109005087031771,.00921800523314784,.00142803462579379,-.00473012443753221,-.00409148604117949,-.000867556408455861,-.00905358663684252,-.0257858032029976,-.0404449348150074,-.0395922413616564,-.0200235083437195,.00664734160455473,.0227602358497139,.0192471726671977,.00298173285024315,-.00987648840327513,-.00854301297387297,-.00110586971823095,-.0115405605969585,-.0328690315056423,-.051554951622511,-.0504680276371612,-.0255238637098843,.00847333235691014,.0290123562694813,.0245342725922579,.00380079961925163,-.0125895092713119,-.0108897349592236,-.00107068796176225,-.0111734131963732,-.0318233475134598,-.0499148002350218,-.0488624552731521,-.0247118563437564,.00820376469397058,.0280893672084472,.0237537477354436,.00367988229564073,-.0121889909806472,-.0105432927002232,-.000770650542793007,-.00804230294180483,-.02290553448866,-.0359272442282916,-.0351697964516905,-.0177868867353227,.0059048349660612,.0202179223629944,.0170972675881759,.00264867392725914,-.00877328675655798,-.00758876024801777,-.000279970008957894,-.00292169213103891,-.0083213627187413,-.0130520259571515,-.0127768523875183,-.00646180669720086,.00214517035484276,.00734497880785625,.00621127462321504,.000962238033926173,-.00318725159515452,-.00275692438613841,.000279970008957894,.00292169213103891,.0083213627187413,.0130520259571515,.0127768523875183,.00646180669720086,-.00214517035484276,-.00734497880785625,-.00621127462321504,-.000962238033926173,.00318725159515452,.00275692438613841,.000770650542793007,.00804230294180483,.02290553448866,.0359272442282916,.0351697964516905,.0177868867353227,-.0059048349660612,-.0202179223629944,-.0170972675881759,-.00264867392725914,.00877328675655798,.00758876024801777,.00107068796176225,.0111734131963732,.0318233475134598,.0499148002350218,.0488624552731521,.0247118563437564,-.00820376469397058,-.0280893672084472,-.0237537477354436,-.00367988229564073,.0121889909806472,.0105432927002232,.00110586971823095,.0115405605969585,.0328690315056423,.051554951622511,.0504680276371612,.0255238637098843,-.00847333235691015,-.0290123562694813,-.0245342725922579,-.00380079961925163,.0125895092713119,.0108897349592236,.000867556408455861,.00905358663684252,.0257858032029976,.0404449348150074,.0395922413616564,.0200235083437195,-.00664734160455473,-.0227602358497139,-.0192471726671977,-.00298173285024315,.00987648840327513,.00854301297387297,.000415496844730152,.0043360139403437,.0123495368891892,.0193702019109772,.0189618233479971,.00958981393734538,-.00318359640435585,-.0109005087031771,-.00921800523314784,-.00142803462579379,.00473012443753221,.00409148604117949},
			{.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703812,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.00588030439408339,.00508637429395636,.000899536743208372,.00938732485666896,.026736333462767,.0419358379330301,.041051712031237,.020761625765838,-.00689237951523805,-.0235992360051222,-.0199566723826478,-.0030916471033844,.0102405608742263,.00885793015048143,.000721575635236956,.00753017033241687,.0214469135896479,.0336394028639592,.0329301892448019,.0166542205347873,-.00552881598728651,-.0189304481891057,-.0160085162284352,-.00248000678059721,.00821460521072995,.00710550916733107,4.09828174776266e-005,.000427685721688117,.00121810507752847,.00191059320784475,.00187031250699067,.000945897903253332,-.000314016224231211,-.00107517918429856,-.000909224295609892,-.000140855178957249,.000466559082044881,.000403566543921213,-.000803456701900855,-.00838465924372367,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752584,.00615620046137393,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733643,-.00791180949317304,-.00137516782835013,-.0143508836460861,-.0408732004597185,-.0641094603592521,-.0627578518723987,-.0317393592123806,.0105367330925572,.0360773591216924,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.00137516782835013,-.0143508836460862,-.0408732004597186,-.0641094603592521,-.0627578518723987,-.0317393592123807,.0105367330925572,.0360773591216925,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.000803456701900856,-.00838465924372368,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752585,.00615620046137394,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733645,-.00791180949317305,4.09828174776246e-005,.000427685721688097,.00121810507752841,.00191059320784466,.00187031250699058,.000945897903253286,-.000314016224231196,-.00107517918429851,-.000909224295609848,-.000140855178957242,.000466559082044858,.000403566543921193,.000721575635236955,.00753017033241686,.0214469135896478,.0336394028639591,.0329301892448018,.0166542205347873,-.0055288159872865,-.0189304481891056,-.0160085162284351,-.00248000678059721,.00821460521072994,.00710550916733106,.00089953674320837,.00938732485666894,.026736333462767,.04193583793303,.0410517120312369,.020761625765838,-.00689237951523804,-.0235992360051222,-.0199566723826478,-.00309164710338439,.0102405608742263,.00885793015048141,.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703813,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.0058803043940834,.00508637429395636},
			{.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454},
			{-.00090976995509462,-.00863877841154346,-.020257551125304,-.0218647389120542,-.00764500327777482,.0097501159735198,.013176762981398,.00178724764942572,-.00876022920331713,-.00672244007828754,.00197953596529396,.0041756355320037,-.00189959746932745,-.0180377484624032,-.0422977178317532,-.0456535220494547,-.0159627483828599,.0203582186081548,.0275130492859463,.00373176877615995,-.0182913374223476,-.0140364386501337,.00413326631535552,.00871871690750996,-.00242140718192188,-.02299262574193,-.0539166846610791,-.0581943111398842,-.0203476337495614,.025950517172659,.0350707432565425,.00475686657924678,-.0233158742926257,-.0178921765820591,.00526865343968954,.0111137038650945,-.00234437337189749,-.0222611463044402,-.0522013979161087,-.0563429374666252,-.0197003011718718,.0251249364009332,.0339550147687404,.0046055332722702,-.0225741111376221,-.0173229610688503,.00510103832266685,.0107601363368395,-.0016874128374327,-.0160229358089307,-.0375730717775044,-.0405540333803145,-.0141797127954073,.0180842099346574,.0244398475527877,.00331493100033043,-.0162481989361725,-.0124685714486963,.00367158135008905,.00774483808976782,-.000613021027013781,-.00582098011077765,-.013649939444666,-.0147328944291874,-.00515135472942478,.00656982138629644,.00887876405495684,.00120428288870561,-.00590281606138012,-.00452971336078232,.00133385056701396,.0028136259808645,.000613021027013781,.00582098011077765,.013649939444666,.0147328944291874,.00515135472942478,-.00656982138629644,-.00887876405495684,-.00120428288870561,.00590281606138012,.00452971336078232,-.00133385056701396,-.0028136259808645,.0016874128374327,.0160229358089307,.0375730717775044,.0405540333803145,.0141797127954073,-.0180842099346574,-.0244398475527877,-.00331493100033043,.0162481989361725,.0124685714486963,-.00367158135008905,-.00774483808976782,.00234437337189749,.0222611463044402,.0522013979161087,.0563429374666252,.0197003011718718,-.0251249364009332,-.0339550147687404,-.0046055332722702,.0225741111376221,.0173229610688503,-.00510103832266685,-.0107601363368395,.00242140718192188,.02299262574193,.0539166846610791,.0581943111398842,.0203476337495614,-.025950517172659,-.0350707432565425,-.00475686657924678,.0233158742926257,.0178921765820591,-.00526865343968954,-.0111137038650945,.00189959746932745,.0180377484624032,.0422977178317532,.0456535220494547,.0159627483828599,-.0203582186081548,-.0275130492859463,-.00373176877615995,.0182913374223476,.0140364386501337,-.00413326631535552,-.00871871690750996,.000909769955094619,.00863877841154346,.020257551125304,.0218647389120542,.00764500327777482,-.00975011597351981,-.013176762981398,-.00178724764942572,.00876022920331713,.00672244007828754,-.00197953596529396,-.00417563553200371},
			{.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622},
			{-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{4.40393013976499e-007,2.6993036058117e-005,.000232696122440507,.000949989177276297,.00256195880647401,.00523448009649574,.00866112710437402,.0119942097336745,.0140544988860133,.0137678071694569,.0106453074128955,.0050858458801123,9.19539549723029e-007,5.63613941066353e-005,.00048586894174179,.00198357510812152,.00534936380130629,.0109295818012743,.0180844124790659,.0250438809603261,.0293457597352286,.0287471481233613,.0222273761718653,.0106192339163871,1.17213236262568e-006,7.1843580904145e-005,.000619334655895749,.00252845304873323,.00681880668738833,.0139318820415032,.0230521081253869,.0319233070161965,.0374068898959917,.0366438427349157,.0283331227625236,.0135362831793791,1.13484254932205e-006,6.95579740866872e-005,.000599631357507937,.00244801371855121,.00660187553145923,.0134886579681297,.022318736335937,.0309077099756012,.0362168400475543,.0354780682047664,.0274317426011938,.0131056445512863,8.16827178275341e-007,5.0065838414078e-005,.000431597484647737,.00176201019189673,.00475184100640955,.00970874984676124,.0160643874648916,.0222464848021472,.0260678446360387,.0255360978134559,.0197445829974339,.00943306775437879,2.96745541228046e-007,1.81884426869712e-005,.000156795259243056,.000640121536048445,.00172629862078525,.00352709643428847,.0058360391029489,.00808193623891563,.00947019990385576,.00927702134312668,.00717301912047857,.00342694375341951,-2.96745541228046e-007,-1.81884426869712e-005,-.000156795259243056,-.000640121536048445,-.00172629862078525,-.00352709643428847,-.0058360391029489,-.00808193623891563,-.00947019990385576,-.00927702134312668,-.00717301912047857,-.00342694375341951,-8.16827178275341e-007,-5.0065838414078e-005,-.000431597484647737,-.00176201019189673,-.00475184100640955,-.00970874984676124,-.0160643874648916,-.0222464848021472,-.0260678446360387,-.0255360978134559,-.0197445829974339,-.00943306775437879,-1.13484254932205e-006,-6.95579740866872e-005,-.000599631357507937,-.00244801371855121,-.00660187553145923,-.0134886579681297,-.022318736335937,-.0309077099756012,-.0362168400475543,-.0354780682047664,-.0274317426011938,-.0131056445512863,-1.17213236262568e-006,-7.1843580904145e-005,-.000619334655895749,-.00252845304873323,-.00681880668738833,-.0139318820415032,-.0230521081253869,-.0319233070161965,-.0374068898959917,-.0366438427349157,-.0283331227625236,-.0135362831793791,-9.19539549723029e-007,-5.63613941066352e-005,-.00048586894174179,-.00198357510812152,-.00534936380130629,-.0109295818012743,-.0180844124790659,-.0250438809603261,-.0293457597352286,-.0287471481233613,-.0222273761718653,-.0106192339163871,-4.40393013976499e-007,-2.6993036058117e-005,-.000232696122440507,-.000949989177276297,-.00256195880647401,-.00523448009649574,-.00866112710437402,-.0119942097336745,-.0140544988860133,-.0137678071694569,-.0106453074128955,-.0050858458801123},
			{-5.47479249038328e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.014910729120128,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278,-9.53436115108903e-007,-5.84390183710293e-005,-.000503779306073239,-.00205669472909975,-.00554655494976578,-.0113324739708158,-.0187510498958539,-.0259670620771616,-.0304275190396141,-.0298068410819887,-.0230467337622952,-.0110106858739491,-7.64811749616544e-007,-4.68776535500915e-005,-.000404113423430025,-.00164980565479233,-.00444924450443136,-.00909049815478573,-.0150414097502034,-.0208298320830514,-.0244078483124377,-.0239099630454571,-.0184872509991657,-.00883237145550259,-4.34384682750355e-008,-2.66247670431932e-006,-2.2952142317351e-005,-9.37028368505082e-005,-.00025270057154655,-.000516306549812956,-.000854296237699806,-.00118305713865876,-.00138627517832104,-.00135799714338606,-.00105000722912861,-.000501645911501147,8.51599050625011e-007,5.21971129219724e-005,.00044997034618045,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801575,.023193505114819,.0271775380818778,.0266231550969231,.0205850987611647,.00983462812913714,1.45756655498971e-006,8.93387163903609e-005,.000770153192219288,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,1.45756655498971e-006,8.93387163903609e-005,.000770153192219289,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,8.51599050625012e-007,5.21971129219725e-005,.000449970346180451,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801576,.0231935051148191,.0271775380818778,.0266231550969231,.0205850987611648,.00983462812913715,-4.34384682750334e-008,-2.6624767043192e-006,-2.29521423173499e-005,-9.37028368505037e-005,-.000252700571546538,-.000516306549812931,-.000854296237699765,-.0011830571386587,-.00138627517832097,-.00135799714338599,-.00105000722912856,-.000501645911501123,-7.64811749616543e-007,-4.68776535500914e-005,-.000404113423430024,-.00164980565479233,-.00444924450443135,-.00909049815478571,-.0150414097502034,-.0208298320830514,-.0244078483124376,-.0239099630454571,-.0184872509991656,-.00883237145550258,-9.53436115108901e-007,-5.84390183710292e-005,-.000503779306073238,-.00205669472909974,-.00554655494976577,-.0113324739708158,-.0187510498958538,-.0259670620771615,-.030427519039614,-.0298068410819887,-.0230467337622952,-.0110106858739491,-5.47479249038329e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.0149107291201281,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278},
			{6.11261401273432e-007,3.74660826168086e-005,.000322980050454363,.00131857612920148,.00355597495952486,.00726540961635832,.0120215637451267,.0166478513879527,.019507513540437,.0191095881083022,.0147755875313232,.00705910670850027,7.63509266142174e-007,4.6797820350488e-005,.000403425213480253,.00164699601620824,.00444166738830656,.00907501692868488,.0150157940511236,.0207943586321327,.0243662814574726,.0238692440950028,.0184557669915978,.00881732982223001,-3.24515476414362e-008,-1.98905470301973e-006,-1.71468417157011e-005,-7.00025160864056e-005,-.000188784848135954,-.000385716791226497,-.000638218522855875,-.000883825710764993,-.00103564367667356,-.00101451802389758,-.000784428203227012,-.000374764277900816,-1.10535636397296e-006,-6.77506755168088e-005,-.000584051362415962,-.00238440790267356,-.00643034149293689,-.0131381872625712,-.0217388370391121,-.0301046466206003,-.0352758315710584,-.0345562549580405,-.0267189939936707,-.0127651255386803,-1.44954346598925e-006,-8.88469567034131e-005,-.000765913929467261,-.00312686750465765,-.00843262842551854,-.0172291707200675,-.0285079004521123,-.0394786651862746,-.0462600594955417,-.045316420311225,-.0350387842542686,-.0167399446189638,-6.62156769174321e-007,-4.05856155279577e-005,-.000349872290759864,-.00142836453897385,-.00385205557811836,-.00787034834568896,-.0130225134342083,-.0180339989827199,-.0211317647632424,-.0207007069245407,-.0160058450967311,-.007646868069241,6.62156769174322e-007,4.05856155279578e-005,.000349872290759864,.00142836453897385,.00385205557811836,.00787034834568897,.0130225134342083,.0180339989827199,.0211317647632424,.0207007069245407,.0160058450967311,.00764686806924101,1.44954346598925e-006,8.88469567034135e-005,.000765913929467265,.00312686750465767,.00843262842551858,.0172291707200676,.0285079004521125,.0394786651862747,.0462600594955419,.0453164203112252,.0350387842542687,.0167399446189638,1.10535636397296e-006,6.77506755168087e-005,.000584051362415961,.00238440790267356,.00643034149293688,.0131381872625712,.0217388370391121,.0301046466206002,.0352758315710583,.0345562549580404,.0267189939936707,.0127651255386803,3.24515476414349e-008,1.98905470301965e-006,1.71468417157005e-005,7.00025160864029e-005,.000188784848135946,.000385716791226483,.000638218522855851,.000883825710764959,.00103564367667352,.00101451802389754,.000784428203226982,.000374764277900801,-7.63509266142173e-007,-4.67978203504879e-005,-.000403425213480253,-.00164699601620824,-.00444166738830656,-.00907501692868487,-.0150157940511236,-.0207943586321327,-.0243662814574726,-.0238692440950028,-.0184557669915978,-.00881732982223001,-6.11261401273429e-007,-3.74660826168084e-005,-.000322980050454361,-.00131857612920147,-.00355597495952484,-.00726540961635829,-.0120215637451267,-.0166478513879526,-.0195075135404369,-.0191095881083021,-.0147755875313231,-.00705910670850024},
			{-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906563,.00323132809181624,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190404,.00732492904924932,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190405,.00732492904924932,-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906564,.00323132809181624},
			{3.01831341787578e-006,.000176553289823538,.00139577555712289,.00499729323775215,.0112042776028942,.0177599426294488,.0206189091483633,.0167940290914578,.0072118425373148,-.00309463500286609,-.00816702708367357,-.00549362253845184,6.30223112790928e-006,.000368642842811124,.00291437599275742,.0104343362129758,.0233945045787828,.037082717323626,.0430522325112947,.0350658921887862,.0150583098027748,-.00646159041306787,-.0170527328289948,-.0114706705943021,8.03342179680519e-006,.000469907146940446,.00371494017739558,.0133005950221825,.0298208553755469,.0472691501765359,.0548784606654209,.0446983134887451,.0191947505157252,-.00823655626282739,-.0217370313501955,-.0146216051595062,7.77784929616953e-006,.000454957683604922,.0035967543563515,.01287745449555,.0288721450037781,.0457653457929115,.0531325763115239,.0432762968137264,.0185840953662182,-.00797452131259557,-.0210454969581169,-.0141564384735731,5.59827325600047e-006,.000327465515948182,.00258884081637867,.00926882308497601,.0207813434103976,.0329405857131029,.0382433073283765,.0311490395153758,.0133762998116843,-.00573983214301874,-.0151479462115431,-.0101893991370919,2.03379940271082e-006,.000118965105897601,.000940501234094081,.00336727522791983,.00754966431306182,.0119670013385727,.0138934296425726,.0113161496526603,.00485948243743287,-.00208522639934011,-.00550310471613919,-.00370171174777645,-2.03379940271082e-006,-.000118965105897601,-.000940501234094081,-.00336727522791983,-.00754966431306182,-.0119670013385727,-.0138934296425726,-.0113161496526603,-.00485948243743287,.00208522639934011,.00550310471613919,.00370171174777645,-5.59827325600047e-006,-.000327465515948182,-.00258884081637867,-.00926882308497601,-.0207813434103976,-.0329405857131029,-.0382433073283765,-.0311490395153758,-.0133762998116843,.00573983214301874,.0151479462115431,.0101893991370919,-7.77784929616953e-006,-.000454957683604922,-.0035967543563515,-.01287745449555,-.0288721450037781,-.0457653457929115,-.0531325763115239,-.0432762968137265,-.0185840953662182,.00797452131259557,.0210454969581169,.0141564384735731,-8.03342179680519e-006,-.000469907146940446,-.00371494017739558,-.0133005950221825,-.0298208553755469,-.0472691501765359,-.0548784606654209,-.0446983134887451,-.0191947505157252,.00823655626282739,.0217370313501955,.0146216051595062,-6.30223112790928e-006,-.000368642842811124,-.00291437599275742,-.0104343362129758,-.0233945045787828,-.037082717323626,-.0430522325112947,-.0350658921887862,-.0150583098027748,.00646159041306787,.0170527328289948,.0114706705943021,-3.01831341787578e-006,-.000176553289823538,-.00139577555712289,-.00499729323775215,-.0112042776028942,-.0177599426294488,-.0206189091483634,-.0167940290914578,-.0072118425373148,.00309463500286609,.00816702708367357,.00549362253845184},
			{-3.75224835608883e-006,-.000219484095846042,-.00173517319595299,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131,-6.5345473857451e-006,-.000382231954981349,-.00302180729935703,-.0108189723668713,-.0242568854796668,-.0384496805377896,-.0446392472280685,-.0363585101022895,-.0156133973760028,.00669978105920017,.0176813399042196,.0118935086675575,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786293,-.0194579906613178,-.0308429342861059,-.0358079794052745,-.029165473474441,-.0125244990989618,.00537432051580024,.0141833273288231,.00954053976870304,-2.97713422859992e-007,-1.7414455344253e-005,-.000137673283428536,-.000492911460431877,-.00110514163839842,-.00175176417356002,-.0020337603052895,-.0016564906264595,-.000711345055883006,.00030524145496262,.000805560341504197,.000541867242857217,5.83658858918833e-006,.000341405538160759,.00269904630894985,.00966339165295146,.02166599347167,.0343428478572752,.039871303339182,.032475036548949,.0139457213459092,-.00598417355813986,-.0157927857332021,-.0106231561080896,9.98969681400165e-006,.000584337539768473,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,9.98969681400166e-006,.000584337539768474,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,5.83658858918834e-006,.000341405538160759,.00269904630894985,.00966339165295147,.02166599347167,.0343428478572752,.039871303339182,.0324750365489491,.0139457213459093,-.00598417355813987,-.0157927857332022,-.0106231561080896,-2.97713422859977e-007,-1.74144553442522e-005,-.00013767328342853,-.000492911460431854,-.00110514163839836,-.00175176417355993,-.0020337603052894,-.00165649062645942,-.000711345055882972,.000305241454962606,.000805560341504158,.000541867242857191,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786291,-.0194579906613177,-.0308429342861059,-.0358079794052744,-.029165473474441,-.0125244990989617,.00537432051580023,.0141833273288231,.00954053976870302,-6.53454738574509e-006,-.000382231954981348,-.00302180729935702,-.0108189723668713,-.0242568854796668,-.0384496805377895,-.0446392472280684,-.0363585101022894,-.0156133973760028,.00669978105920015,.0176813399042196,.0118935086675575,-3.75224835608883e-006,-.000219484095846043,-.001735173195953,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131},
			{-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784,-1.55303998105985e-005,-.000854853930745827,-.00602179902609634,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-1.55303998105985e-005,-.000854853930745827,-.00602179902609635,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784},
			{1.16476424390402e-005,.000641131783107114,.00451628823154087,.0134633410098786,.0228083947504036,.0229321000995076,.00993827320683666,-.00713237192763805,-.0139474741042205,-.00647469362056888,.0043383944810003,.00561733742272306,2.43202492860194e-005,.0013386816149243,.00942999892153632,.0281114235173858,.0476238731611908,.0478821696313965,.0207510904569588,-.0148923753616752,-.0291222922492883,-.0135191446446746,.00905855719998496,.011728986973914,3.10008974207295e-005,.00170641061015528,.0120203714117433,.0358334878300008,.0607059076280363,.0610351568179937,.0264513089096609,-.0189832347320423,-.0371220370342043,-.0172327845581125,.0115468965483827,.0149508797278819,3.00146455997538e-005,.00165212345360114,.0116379594759614,.0346934936503285,.0587746308933795,.0590934054635625,.0256097961229407,-.0183793086724707,-.0359410493961229,-.0166845467145346,.0111795475780318,.0144752376147941,2.16036826314166e-005,.00118915116425069,.00837667005463561,.0249713835102858,.0423042966933997,.0425337415030114,.0184331980818002,-.0132289002122403,-.0258693384205835,-.0120090590732415,.00804671829410513,.0104188616322334,7.84841232696798e-006,.000432007302429133,.0030431645214108,.00907186601040717,.0153687484359428,.0154521035519588,.00669660545930529,-.00480593356555747,-.00939808448469602,-.00436277688733732,.00292329618651467,.00378507328877716,-7.84841232696798e-006,-.000432007302429133,-.0030431645214108,-.00907186601040717,-.0153687484359428,-.0154521035519588,-.00669660545930529,.00480593356555747,.00939808448469602,.00436277688733732,-.00292329618651467,-.00378507328877716,-2.16036826314166e-005,-.00118915116425069,-.00837667005463561,-.0249713835102858,-.0423042966933997,-.0425337415030114,-.0184331980818002,.0132289002122403,.0258693384205835,.0120090590732415,-.00804671829410513,-.0104188616322334,-3.00146455997538e-005,-.00165212345360114,-.0116379594759614,-.0346934936503285,-.0587746308933795,-.0590934054635624,-.0256097961229407,.0183793086724707,.0359410493961229,.0166845467145346,-.0111795475780318,-.0144752376147941,-3.10008974207295e-005,-.00170641061015528,-.0120203714117433,-.0358334878300008,-.0607059076280363,-.0610351568179937,-.0264513089096609,.0189832347320423,.0371220370342043,.0172327845581125,-.0115468965483827,-.0149508797278819,-2.43202492860194e-005,-.0013386816149243,-.00942999892153632,-.0281114235173858,-.0476238731611908,-.0478821696313965,-.0207510904569588,.0148923753616752,.0291222922492883,.0135191446446746,-.00905855719998497,-.011728986973914,-1.16476424390402e-005,-.000641131783107114,-.00451628823154087,-.0134633410098786,-.0228083947504036,-.0229321000995076,-.00993827320683666,.00713237192763805,.0139474741042205,.00647469362056888,-.0043383944810003,-.00561733742272306},
			{-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.00322545981703339,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072408,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072409,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.0032254598170334},
			{3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489},
			{-5.54740393494954e-009,-1.76805421500992e-006,-3.65766652977238e-005,-.000267816680517056,-.00110639036597811,-.00312801719432003,-.00665764172873835,-.0112074711567434,-.0152399199046756,-.0166462936129339,-.0138469870744205,-.00688452453082954,-1.15829660203181e-008,-3.69169293144801e-006,-7.63719888164115e-005,-.000559200582196592,-.00231014041245373,-.00653129234821247,-.0139011398527498,-.0234011727115081,-.0318209159596439,-.0347574208663796,-.0289124755737218,-.0143748705956402,-1.47647475649067e-008,-4.70578210489157e-006,-9.73509836708664e-005,-.000712810123054713,-.00294472417250793,-.00832540496318534,-.0177197118967874,-.0298293552102391,-.0405619588803169,-.0443051066712203,-.0368545847906068,-.0183235740526672,-1.42950269960305e-008,-4.55607398170153e-006,-9.42538931700292e-005,-.000690133028507038,-.00285104173686768,-.00806054341115132,-.0171559831153849,-.0288803744276712,-.0392715347591976,-.0428955993418035,-.0356821057856366,-.0177406342096373,-1.0289151187989e-008,-3.27933161891947e-006,-6.78412539656127e-005,-.000496737996514833,-.00205209822143352,-.0058017483868841,-.0123483854981035,-.0207872667141207,-.0282665264384568,-.0308750243738453,-.0256829582228968,-.0127692006181209,-3.73795071866204e-009,-1.19134997219017e-006,-2.46460820122577e-005,-.000180460187350232,-.000745507756803769,-.00210771998156325,-.00448605094857911,-.00755181619294033,-.0102689600807936,-.0112166025591737,-.00933037433240469,-.00463892907735314,3.73795071866204e-009,1.19134997219017e-006,2.46460820122577e-005,.000180460187350232,.000745507756803769,.00210771998156325,.0044860509485791,.00755181619294033,.0102689600807936,.0112166025591737,.00933037433240469,.00463892907735314,1.0289151187989e-008,3.27933161891947e-006,6.78412539656127e-005,.000496737996514833,.00205209822143352,.0058017483868841,.0123483854981035,.0207872667141207,.0282665264384568,.0308750243738453,.0256829582228968,.0127692006181209,1.42950269960305e-008,4.55607398170153e-006,9.42538931700292e-005,.000690133028507038,.00285104173686769,.00806054341115132,.0171559831153849,.0288803744276712,.0392715347591976,.0428955993418035,.0356821057856366,.0177406342096373,1.47647475649067e-008,4.70578210489157e-006,9.73509836708664e-005,.000712810123054713,.00294472417250793,.00832540496318534,.0177197118967874,.0298293552102391,.0405619588803169,.0443051066712203,.0368545847906068,.0183235740526672,1.15829660203181e-008,3.69169293144801e-006,7.63719888164114e-005,.000559200582196592,.00231014041245373,.00653129234821247,.0139011398527498,.0234011727115081,.0318209159596439,.0347574208663796,.0289124755737218,.0143748705956402,5.54740393494954e-009,1.76805421500992e-006,3.65766652977238e-005,.000267816680517056,.00110639036597811,.00312801719432003,.00665764172873835,.0112074711567434,.0152399199046756,.0166462936129339,.0138469870744205,.00688452453082954},
			{6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978,1.20099435917211e-008,3.82777811718284e-006,7.91872544617432e-005,.000579814137143969,.00239529806043259,.00677205238670959,.014413571204413,.024263799423424,.0329939158106694,.0360386677528614,.0299782629188682,.0149047648667097,9.63393962702057e-009,3.07050428713007e-006,6.35211333748341e-005,.000465105797498382,.00192142092314525,.00543229394429081,.0115620422138201,.0194635367753551,.0264665183938168,.0289089075829344,.0240474714038134,.0119560598918884,5.47172008094484e-010,1.743932453074e-007,3.60776457511335e-006,2.64162827510142e-005,.000109129575813761,.000308534130494273,.000656681077600405,.00110545663708768,.00150319999683172,.00164191863623543,.00136580710768774,.000679059819058227,-1.07271545504665e-008,-3.41893091623005e-006,-7.07292178810733e-005,-.00051788385284143,-.00213945488523406,-.00604872554323789,-.0128740493036516,-.0216721689330841,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859674,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175488,-.0370933112099723,-.0504394866543156,-.0550941546796317,-.0458293038356739,-.0227856763924034,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175489,-.0370933112099723,-.0504394866543156,-.0550941546796318,-.0458293038356739,-.0227856763924034,-1.07271545504666e-008,-3.41893091623005e-006,-7.07292178810734e-005,-.000517883852841431,-.00213945488523406,-.0060487255432379,-.0128740493036517,-.0216721689330842,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859675,5.47172008094458e-010,1.74393245307392e-007,3.60776457511317e-006,2.6416282751013e-005,.000109129575813756,.000308534130494258,.000656681077600374,.00110545663708762,.00150319999683165,.00164191863623535,.00136580710768767,.000679059819058194,9.63393962702055e-009,3.07050428713007e-006,6.3521133374834e-005,.000465105797498381,.00192142092314524,.0054322939442908,.01156204221382,.0194635367753551,.0264665183938168,.0289089075829343,.0240474714038134,.0119560598918883,1.20099435917211e-008,3.82777811718283e-006,7.91872544617431e-005,.000579814137143968,.00239529806043259,.00677205238670958,.014413571204413,.024263799423424,.0329939158106693,.0360386677528614,.0299782629188681,.0149047648667096,6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978},
			{2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387107,.0553680555817044,.0585549188607202,.0369926009098689,.00167071599568083,-.0222958057593638,-.0178801188985916,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387106,.0553680555817044,.0585549188607202,.036992600909869,.00167071599568083,-.0222958057593638,-.0178801188985916,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967},
			{-4.91028494999023e-008,-1.49610469790016e-005,-.000284808486486742,-.00183936641356981,-.00637694142212144,-.0142111682876331,-.0218573742001049,-.0231154364976192,-.0146033865958862,-.000659540312839943,.00880160526598481,.00705844634424949,-1.02526631182363e-007,-3.12386299642582e-005,-.000594679432029528,-.00384059262983655,-.0133150382903681,-.0296728850674858,-.0456381445486889,-.0482649755879333,-.0304918359479907,-.00137711858055358,.0183777305412829,.0147380189220751,-1.30690172571732e-007,-3.98197218991008e-005,-.000758034831541094,-.00489558378913558,-.0169726112318295,-.037823874884499,-.0581747095182799,-.0615231176135874,-.0388677873847604,-.00175540601372305,.0234259991596233,.0187864773676752,-1.26532440654065e-007,-3.85529110484099e-005,-.000733918973770133,-.00473983738085433,-.0164326504524067,-.0366205592199579,-.0563239594442265,-.0595658424432908,-.0376312610492393,-.00169956013435736,.0226807325302158,.0181888108787948,-9.10744283614461e-008,-2.7749281664479e-005,-.000528253866393645,-.00341159925277751,-.0118277513551488,-.0263584301385007,-.0405403735431023,-.0428737881159049,-.0270859043804411,-.00122329472902273,.0163249419621616,.0130917853539995,-3.30864731915692e-008,-1.00810499763263e-005,-.0001919096029833,-.00123940154495863,-.00429690951861456,-.00957576685179106,-.0147279319403104,-.0155756392506594,-.00984005132149962,-.000444411334612696,.00593069607246235,.00475612104229521,3.30864731915692e-008,1.00810499763263e-005,.0001919096029833,.00123940154495863,.00429690951861456,.00957576685179106,.0147279319403104,.0155756392506594,.00984005132149962,.000444411334612696,-.00593069607246235,-.00475612104229521,9.10744283614461e-008,2.7749281664479e-005,.000528253866393645,.00341159925277751,.0118277513551488,.0263584301385007,.0405403735431023,.0428737881159049,.0270859043804411,.00122329472902273,-.0163249419621616,-.0130917853539995,1.26532440654065e-007,3.85529110484099e-005,.000733918973770133,.00473983738085433,.0164326504524067,.0366205592199579,.0563239594442265,.0595658424432908,.0376312610492393,.00169956013435736,-.0226807325302158,-.0181888108787948,1.30690172571732e-007,3.98197218991008e-005,.000758034831541094,.00489558378913558,.0169726112318295,.037823874884499,.0581747095182799,.0615231176135874,.0388677873847604,.00175540601372305,-.0234259991596233,-.0187864773676752,1.02526631182363e-007,3.12386299642582e-005,.000594679432029528,.00384059262983655,.0133150382903681,.0296728850674858,.0456381445486889,.0482649755879333,.0304918359479907,.00137711858055358,-.0183777305412829,-.0147380189220751,4.91028494999023e-008,1.49610469790016e-005,.000284808486486742,.00183936641356981,.00637694142212144,.0142111682876331,.0218573742001049,.0231154364976192,.0146033865958862,.000659540312839943,-.00880160526598481,-.00705844634424949},
			{1.40569330619306e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201856,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016855,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016854,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201857,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,1.40569330619305e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917},
			{-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{6.48385134009702e-011,1.0745664478345e-007,5.33473591100581e-006,7.00568237223647e-005,.000443341599207306,.00173443714397409,.00474854200551721,.00971711798722703,.0153335904273431,.0186751535620109,.0167126847540202,.0086472609164468,1.35382659409353e-010,2.24369214821818e-007,1.11389156998738e-005,.000146278478759263,.000925696188564719,.00362150056830361,.0099149442407531,.0202893188924589,.0320164997847783,.0389936755407666,.0348960454086467,.0180554599121578,1.7257158376693e-010,2.86002290982243e-007,1.41987188917692e-005,.000186460428984299,.00117998019867963,.00461630825846974,.0126385287307287,.0258626910573308,.0408112685822381,.0497050388456101,.0444818106664909,.0230152024363774,1.67081451128867e-010,2.76903513083773e-007,1.37470057631949e-005,.00018052844143149,.00114244071703594,.00446944662526241,.012236450951948,.025039904355241,.0395129130085492,.0481237399429412,.0430666817373268,.0222830047517444,1.20260445247909e-010,1.9930722141311e-007,9.89470119356512e-006,.000129939203901917,.000822296121872419,.00321697972775721,.00880744708519725,.0180230063024996,.0284402635920669,.0346380903052428,.0309981645843277,.0160386688935279,4.36894754025765e-011,7.24064170021469e-008,3.59465078913488e-006,4.72056763219235e-005,.000298732356396325,.00116869812344953,.00319966171749153,.00654758668912568,.0103320771354786,.012583688604042,.0112613382258831,.00582669579070132,-4.36894754025765e-011,-7.24064170021469e-008,-3.59465078913488e-006,-4.72056763219235e-005,-.000298732356396325,-.00116869812344953,-.00319966171749153,-.00654758668912568,-.0103320771354786,-.012583688604042,-.0112613382258831,-.00582669579070132,-1.20260445247909e-010,-1.9930722141311e-007,-9.89470119356512e-006,-.000129939203901917,-.000822296121872419,-.00321697972775721,-.00880744708519725,-.0180230063024996,-.0284402635920669,-.0346380903052428,-.0309981645843277,-.0160386688935279,-1.67081451128867e-010,-2.76903513083773e-007,-1.37470057631949e-005,-.00018052844143149,-.00114244071703594,-.00446944662526241,-.012236450951948,-.025039904355241,-.0395129130085492,-.0481237399429412,-.0430666817373268,-.0222830047517444,-1.7257158376693e-010,-2.86002290982243e-007,-1.41987188917692e-005,-.000186460428984299,-.00117998019867963,-.00461630825846974,-.0126385287307287,-.0258626910573308,-.0408112685822381,-.0497050388456102,-.0444818106664909,-.0230152024363774,-1.35382659409353e-010,-2.24369214821818e-007,-1.11389156998738e-005,-.000146278478759263,-.000925696188564719,-.00362150056830361,-.0099149442407531,-.0202893188924589,-.0320164997847783,-.0389936755407666,-.0348960454086467,-.0180554599121578,-6.48385134009701e-011,-1.0745664478345e-007,-5.33473591100581e-006,-7.00568237223647e-005,-.000443341599207306,-.00173443714397409,-.00474854200551721,-.00971711798722703,-.0153335904273431,-.0186751535620109,-.0167126847540202,-.0086472609164468},
			{-4.13541370363107e-010,-6.55869442144476e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-4.13541370363107e-010,-6.55869442144475e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289},
			{4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906685,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906686,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 144; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,-144.1116465,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,-159.0586197,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,-81.71498915,-.2645751311,0.,0.,2.958039892,0.,-22.92947780,.2236067977,0.,-3.324307508,-.1934294858},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,0.,-.2645751311,0.,0.,2.958039892,0.,0.,.2236067977,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,0.,0.,0.,-.2645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,-254.3227936,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066828,0.,-305.0714286,.9128709293,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,-182.9477455,-.9354143468,0.,0.,8.844691708,0.,-63.03769304,.9035079028,0.,-11.53669556,-.8548456319},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066829,0.,0.,.9128709295,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,0.,-.9354143468,0.,0.,8.844691708,0.,0.,.9035079028,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,-5.850213669,0.,0.,0.,0.,0.,-.9354143468,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,-212.3677366,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,-284.1991313,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,-208.0853992,-1.636634177,0.,0.,13.68699853,0.,-93.25932554,2.074812507,0.,-23.01464213,-2.326595118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,0.,-1.636634177,0.,0.,13.68699853,0.,0.,2.074812507,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,0.,0.,0.,-1.636634177,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,-106.7569280,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,-158.7252693,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,-141.7297322,-1.218935425,0.,0.,11.40958244,0.,-83.44725167,2.747169230,0.,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,0.,-1.218935425,0.,0.,11.40958244,0.,0.,2.747169230,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,0.,0.,0.,-1.218935425,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,-33.04592764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,-53.17144007,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,-55.95917640,0.,0.,0.,4.138710979,0.,-42.40452863,1.621428571,0.,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,0.,0.,0.,0.,4.138710979,0.,0.,1.621428571,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-8.845830912,0.,0.,0.,0.,0.,0.,0.,0.,0.,-10.42167396,0.,0.,0.,0.,0.,-9.706337648,0.,0.,-6.398809524,-2.185359011},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689990,0.,29.50000000,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,16.25653401,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,-9.096005092,-.3240370350,0.,0.,2.070196679,0.,-10.27118455,.3129843186,0.,-3.053571429,-.2961272134},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689989,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,0.,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,0.,-.3240370349,0.,0.,2.070196679,0.,0.,.3129843186,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,-1.004158022,0.,0.,0.,0.,0.,-.3240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,64.91147819,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,44.99219432,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,-14.04517129,-1.020504077,0.,0.,5.704791219,0.,-27.63572997,1.293725044,0.,-11.16850681,-1.450721144},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,0.,-1.020504077,0.,0.,5.704791219,0.,0.,1.293725044,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,0.,0.,0.,-1.020504077,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,59.18483757,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,51.70172994,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,-2.439471337,-.9983468483,0.,0.,5.709015429,0.,-29.57691077,2.250018896,0.,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,0.,-.9983468480,0.,0.,5.709015429,0.,0.,2.250018896,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,0.,0.,0.,-.9983468483,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,27.51181564,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,29.34227987,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,6.515483910,0.,0.,0.,1.996693696,0.,-13.84473139,1.554721797,0.,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,0.,0.,0.,0.,1.996693696,0.,0.,1.554721797,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.221428571,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.471852928,0.,0.,0.,0.,0.,-1.980327354,0.,0.,-4.377552220,-2.322071589},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,-6.851601598,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,-.4433944513,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,4.225804810,-.1463850110,0.,0.,.1636634177,0.,.5511331920,.1855768722,0.,-.8658617402,-.2080969936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,0.,-.1463850109,0.,0.,.1636634177,0.,0.,.1855768722,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.1463850110,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,-14.87979615,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,-3.546268517,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,9.681316602,-.2746751327,0.,0.,-.1889822365,0.,3.496332169,.6190476190,0.,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,0.,-.2746751327,0.,0.,-.1889822365,0.,0.,.6190476190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,0.,0.,0.,-.2746751328,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,-11.83920042,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,-5.353808485,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,7.368172496,0.,0.,0.,-.6338656908,0.,5.580637822,.6016087651,0.,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,0.,0.,0.,0.,-.6338656908,0.,0.,.6016087651,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.517439175,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.767857143,0.,0.,0.,0.,0.,2.833614404,0.,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,1.836209865,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.3415650256,0.,-.6681531049,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,-.9660917834,-.2500000000e-1,0.,0.,-.2236067977,0.,.5443310539,.5634361697e-1,0.,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,0.,-.2500000000e-1,0.,0.,-.2236067977,0.,0.,.5634361697e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,0.,0.,0.,-.2500000000e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,3.381321241,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.5976143047,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,-2.391701925,0.,0.,0.,-.4500000000,0.,.7607257745,.1070899340,0.,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,0.,0.,0.,0.,-.4500000000,0.,0.,.1070899340,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1200198396,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.598750131,0.,0.,0.,0.,0.,.3055555556e-1,0.,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,-.5477225576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,.4065785564,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,.8748177653e-1,0.,0.,0.,-.5669467095e-1,0.,-.2415229459,.7142857143e-2,0.,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,0.,0.,0.,0.,-.5669467095e-1,0.,0.,.7142857143e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3593680677,0.,0.,0.,0.,0.,-.5197011503,0.,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3419382528,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2207201930,0.,0.,0.,0.,0.,-.8705321812e-1,0.,0.,.1974181422e-1,-.1984126984e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,-144.1116465,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,-159.0586197,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,-81.71498915,-.2645751311,0.,0.,2.958039892,0.,-22.92947780,.2236067977,0.,-3.324307508,-.1934294858},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,0.,-.2645751311,0.,0.,2.958039892,0.,0.,.2236067977,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,0.,0.,0.,-.2645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,-254.3227936,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066828,0.,-305.0714286,.9128709293,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,-182.9477455,-.9354143468,0.,0.,8.844691708,0.,-63.03769304,.9035079028,0.,-11.53669556,-.8548456319},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066829,0.,0.,.9128709295,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,0.,-.9354143468,0.,0.,8.844691708,0.,0.,.9035079028,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,-5.850213669,0.,0.,0.,0.,0.,-.9354143468,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,-212.3677366,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,-284.1991313,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,-208.0853992,-1.636634177,0.,0.,13.68699853,0.,-93.25932554,2.074812507,0.,-23.01464213,-2.326595118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,0.,-1.636634177,0.,0.,13.68699853,0.,0.,2.074812507,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,0.,0.,0.,-1.636634177,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,-106.7569280,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,-158.7252693,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,-141.7297322,-1.218935425,0.,0.,11.40958244,0.,-83.44725167,2.747169230,0.,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,0.,-1.218935425,0.,0.,11.40958244,0.,0.,2.747169230,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,0.,0.,0.,-1.218935425,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,-33.04592764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,-53.17144007,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,-55.95917640,0.,0.,0.,4.138710979,0.,-42.40452863,1.621428571,0.,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,0.,0.,0.,0.,4.138710979,0.,0.,1.621428571,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-8.845830912,0.,0.,0.,0.,0.,0.,0.,0.,0.,-10.42167396,0.,0.,0.,0.,0.,-9.706337648,0.,0.,-6.398809524,-2.185359011},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689990,0.,29.50000000,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,16.25653401,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,-9.096005092,-.3240370350,0.,0.,2.070196679,0.,-10.27118455,.3129843186,0.,-3.053571429,-.2961272134},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689989,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,0.,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,0.,-.3240370349,0.,0.,2.070196679,0.,0.,.3129843186,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,-1.004158022,0.,0.,0.,0.,0.,-.3240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,64.91147819,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,44.99219432,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,-14.04517129,-1.020504077,0.,0.,5.704791219,0.,-27.63572997,1.293725044,0.,-11.16850681,-1.450721144},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,0.,-1.020504077,0.,0.,5.704791219,0.,0.,1.293725044,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,0.,0.,0.,-1.020504077,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,59.18483757,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,51.70172994,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,-2.439471337,-.9983468483,0.,0.,5.709015429,0.,-29.57691077,2.250018896,0.,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,0.,-.9983468480,0.,0.,5.709015429,0.,0.,2.250018896,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,0.,0.,0.,-.9983468483,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,27.51181564,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,29.34227987,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,6.515483910,0.,0.,0.,1.996693696,0.,-13.84473139,1.554721797,0.,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,0.,0.,0.,0.,1.996693696,0.,0.,1.554721797,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.221428571,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.471852928,0.,0.,0.,0.,0.,-1.980327354,0.,0.,-4.377552220,-2.322071589},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,-6.851601598,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,-.4433944513,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,4.225804810,-.1463850110,0.,0.,.1636634177,0.,.5511331920,.1855768722,0.,-.8658617402,-.2080969936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,0.,-.1463850109,0.,0.,.1636634177,0.,0.,.1855768722,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.1463850110,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,-14.87979615,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,-3.546268517,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,9.681316602,-.2746751327,0.,0.,-.1889822365,0.,3.496332169,.6190476190,0.,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,0.,-.2746751327,0.,0.,-.1889822365,0.,0.,.6190476190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,0.,0.,0.,-.2746751328,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,-11.83920042,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,-5.353808485,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,7.368172496,0.,0.,0.,-.6338656908,0.,5.580637822,.6016087651,0.,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,0.,0.,0.,0.,-.6338656908,0.,0.,.6016087651,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.517439175,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.767857143,0.,0.,0.,0.,0.,2.833614404,0.,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,1.836209865,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.3415650256,0.,-.6681531049,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,-.9660917834,-.2500000000e-1,0.,0.,-.2236067977,0.,.5443310539,.5634361697e-1,0.,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,0.,-.2500000000e-1,0.,0.,-.2236067977,0.,0.,.5634361697e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,0.,0.,0.,-.2500000000e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,3.381321241,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.5976143047,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,-2.391701925,0.,0.,0.,-.4500000000,0.,.7607257745,.1070899340,0.,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,0.,0.,0.,0.,-.4500000000,0.,0.,.1070899340,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1200198396,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.598750131,0.,0.,0.,0.,0.,.3055555556e-1,0.,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,-.5477225576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,.4065785564,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,.8748177653e-1,0.,0.,0.,-.5669467095e-1,0.,-.2415229459,.7142857143e-2,0.,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,0.,0.,0.,0.,-.5669467095e-1,0.,0.,.7142857143e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3593680677,0.,0.,0.,0.,0.,-.5197011503,0.,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3419382528,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2207201930,0.,0.,0.,0.,0.,-.8705321812e-1,0.,0.,.1974181422e-1,-.1984126984e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,-.4040610177e-1,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,.4694764777,-11.64130577,0.,0.,-6.507687761,0.,-4.073608438,50.31152948,0.,31.49343956,-228.6336522},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,0.,-11.64130577,0.,0.,-6.507687761,0.,0.,50.31152948,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,0.,0.,0.,-11.64130577,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709293e-1,0.,.5714285714e-1,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,-.6224437520,10.85080642,0.,0.,7.410417377,0.,5.024466921,-50.14468862,0.,-35.90912479,239.3225832},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709295e-1,0.,0.,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,0.,10.85080642,0.,0.,7.410417377,0.,0.,-50.14468862,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,10.85080642,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.464751509,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,-.6998542123e-1,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,.6776309270,-6.197388082,0.,0.,-6.026182953,0.,-4.798824078,33.05867928,0.,29.90051531,-173.7315107},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,0.,-6.197388081,0.,0.,-6.026182953,0.,0.,33.05867928,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,0.,0.,0.,-6.197388082,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,.8081220354e-1,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,-.6455301571,1.974864371,0.,0.,3.507390157,0.,3.710995566,-14.50782199,0.,-18.91064403,91.34707467},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,0.,1.974864371,0.,0.,3.507390157,0.,0.,-14.50782199,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,0.,0.,0.,1.974864371,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,-.9035079028e-1,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,.5248906592,0.,0.,0.,-1.266180985,0.,-2.208209791,3.502380952,0.,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,0.,0.,0.,0.,-1.266180985,0.,0.,3.502380952,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1855768722,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3859879246,0.,0.,0.,0.,0.,.8871666102,0.,0.,-2.267857143,6.356265942},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,.9897433188e-1,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,-.8864412337,-15.76980237,0.,0.,1.794170455,0.,4.167058315,75.89869728,0.,4.660714286,-371.5212018},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,0.,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,0.,-15.76980236,0.,0.,1.794170455,0.,0.,75.89869727,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,-1.278019300,0.,0.,0.,0.,0.,-15.76980237,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.320391543,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,-.1212183053,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,.9389529552,8.919961562,0.,0.,-.8451542550e-1,0.,-3.528403268,-49.83237207,0.,-7.640075151,269.1248912},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,0.,8.919961562,0.,0.,-.8451542550e-1,0.,0.,-49.83237207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,0.,0.,0.,8.919961562,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,.1399708425,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,-.8470386588,-2.776822654,0.,0.,-.7563225568,0.,2.151452997,21.71637090,0.,7.490281115,-141.0307509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,0.,-2.776822653,0.,0.,-.7563225568,0.,0.,21.71637090,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,0.,0.,0.,-2.776822654,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,-.1564921593,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,.6060915266,0.,0.,0.,.5237229368,0.,-.8366600265,-5.171408842,0.,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,.5237229366,0.,0.,-5.171408842,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2642857143,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3365806955,0.,0.,0.,0.,0.,.1400231463,0.,0.,1.482553013,-9.676023611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,-.1564921593,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,.6060915266,-8.002380602,0.,0.,16.80277755,0.,5.405089377,48.86857636,0.,-68.44072363,-278.3505387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,0.,-8.002380594,0.,0.,16.80277755,0.,0.,48.86857636,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,0.,0.,0.,-8.002380602,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,.1807015806,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,-.3936679944,2.387560769,0.,0.,-7.275816105,0.,-5.503272838,-21.01190476,0.,34.70865684,144.9163216},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,0.,2.387560769,0.,0.,-7.275816105,0.,0.,-21.01190476,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,0.,0.,0.,2.387560770,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,-.2020305089,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,3.548977048,4.876757779,0.,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,0.,4.876757779,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2083333333,0.,0.,0.,0.,0.,-1.312336745,0.,0.,2.180166278,9.676640803},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,.3415650256,0.,.2138089935,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,.7763237544,-1.575000000,0.,0.,13.75181806,0.,-13.04353288,16.26921940,0.,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,0.,-1.575000000,0.,0.,13.75181806,0.,0.,16.26921940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,0.,0.,0.,-1.575000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.2390457218,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,-1.388730150,0.,0.,0.,-3.450000000,0.,3.286335346,-3.647357166,0.,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,0.,0.,0.,0.,-3.450000000,0.,0.,-3.647357166,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7535958771,0.,0.,0.,0.,0.,-.1750000000,0.,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,-.2710523708,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,-3.674234614,0.,0.,0.,3.571764270,0.,-16.97561276,2.292857143,0.,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,0.,0.,0.,0.,3.571764270,0.,0.,2.292857143,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,3.354434698,0.,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.025814759,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.869362509,0.,0.,0.,0.,0.,-5.484352741,0.,0.,-6.337122368,-3.339285714}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,-.4040610177e-1,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,.4694764777,-11.64130577,0.,0.,-6.507687761,0.,-4.073608438,50.31152948,0.,31.49343956,-228.6336522},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,0.,-11.64130577,0.,0.,-6.507687761,0.,0.,50.31152948,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,0.,0.,0.,-11.64130577,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709293e-1,0.,.5714285714e-1,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,-.6224437520,10.85080642,0.,0.,7.410417377,0.,5.024466921,-50.14468862,0.,-35.90912479,239.3225832},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709295e-1,0.,0.,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,0.,10.85080642,0.,0.,7.410417377,0.,0.,-50.14468862,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,10.85080642,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.464751509,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,-.6998542123e-1,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,.6776309270,-6.197388082,0.,0.,-6.026182953,0.,-4.798824078,33.05867928,0.,29.90051531,-173.7315107},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,0.,-6.197388081,0.,0.,-6.026182953,0.,0.,33.05867928,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,0.,0.,0.,-6.197388082,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,.8081220354e-1,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,-.6455301571,1.974864371,0.,0.,3.507390157,0.,3.710995566,-14.50782199,0.,-18.91064403,91.34707467},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,0.,1.974864371,0.,0.,3.507390157,0.,0.,-14.50782199,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,0.,0.,0.,1.974864371,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,-.9035079028e-1,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,.5248906592,0.,0.,0.,-1.266180985,0.,-2.208209791,3.502380952,0.,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,0.,0.,0.,0.,-1.266180985,0.,0.,3.502380952,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1855768722,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3859879246,0.,0.,0.,0.,0.,.8871666102,0.,0.,-2.267857143,6.356265942},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,.9897433188e-1,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,-.8864412337,-15.76980237,0.,0.,1.794170455,0.,4.167058315,75.89869728,0.,4.660714286,-371.5212018},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,0.,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,0.,-15.76980236,0.,0.,1.794170455,0.,0.,75.89869727,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,-1.278019300,0.,0.,0.,0.,0.,-15.76980237,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.320391543,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,-.1212183053,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,.9389529552,8.919961562,0.,0.,-.8451542550e-1,0.,-3.528403268,-49.83237207,0.,-7.640075151,269.1248912},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,0.,8.919961562,0.,0.,-.8451542550e-1,0.,0.,-49.83237207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,0.,0.,0.,8.919961562,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,.1399708425,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,-.8470386588,-2.776822654,0.,0.,-.7563225568,0.,2.151452997,21.71637090,0.,7.490281115,-141.0307509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,0.,-2.776822653,0.,0.,-.7563225568,0.,0.,21.71637090,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,0.,0.,0.,-2.776822654,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,-.1564921593,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,.6060915266,0.,0.,0.,.5237229368,0.,-.8366600265,-5.171408842,0.,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,.5237229366,0.,0.,-5.171408842,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2642857143,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3365806955,0.,0.,0.,0.,0.,.1400231463,0.,0.,1.482553013,-9.676023611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,-.1564921593,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,.6060915266,-8.002380602,0.,0.,16.80277755,0.,5.405089377,48.86857636,0.,-68.44072363,-278.3505387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,0.,-8.002380594,0.,0.,16.80277755,0.,0.,48.86857636,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,0.,0.,0.,-8.002380602,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,.1807015806,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,-.3936679944,2.387560769,0.,0.,-7.275816105,0.,-5.503272838,-21.01190476,0.,34.70865684,144.9163216},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,0.,2.387560769,0.,0.,-7.275816105,0.,0.,-21.01190476,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,0.,0.,0.,2.387560770,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,-.2020305089,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,3.548977048,4.876757779,0.,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,0.,4.876757779,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2083333333,0.,0.,0.,0.,0.,-1.312336745,0.,0.,2.180166278,9.676640803},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,.3415650256,0.,.2138089935,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,.7763237544,-1.575000000,0.,0.,13.75181806,0.,-13.04353288,16.26921940,0.,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,0.,-1.575000000,0.,0.,13.75181806,0.,0.,16.26921940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,0.,0.,0.,-1.575000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.2390457218,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,-1.388730150,0.,0.,0.,-3.450000000,0.,3.286335346,-3.647357166,0.,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,0.,0.,0.,0.,-3.450000000,0.,0.,-3.647357166,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7535958771,0.,0.,0.,0.,0.,-.1750000000,0.,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,-.2710523708,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,-3.674234614,0.,0.,0.,3.571764270,0.,-16.97561276,2.292857143,0.,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,0.,0.,0.,0.,3.571764270,0.,0.,2.292857143,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,3.354434698,0.,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.025814759,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.869362509,0.,0.,0.,0.,0.,-5.484352741,0.,0.,-6.337122368,-3.339285714}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,-.5103103630e-1,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,.1035406358,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,-.1189992461,.6614378278e-1,0.,0.,-.7395099729e-1,0.,-.2507429436e-1,-.2795084971e-1,0.,.5467611033e-1,.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,0.,.6614378278e-1,0.,0.,-.7395099729e-1,0.,0.,-.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,0.,0.,0.,.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,.9923207755e-1,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709293e-1,0.,-.1638392857,.3651483717,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,.1276586028,-.1403121520,0.,0.,.8964214566e-1,0.,.8558232884e-1,.5646924394e-1,0.,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709295e-1,0.,0.,.3651483718,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,0.,-.1403121520,0.,0.,.8964214566e-1,0.,0.,.5646924394e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,.3952847074e-1,0.,0.,0.,0.,0.,-.1403121520,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3651483717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,-.1031197389,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,.1159133539,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,.1364673394e-1,.3218713882,0.,0.,-.3049687729e-2,0.,-.2191913380,-.1210307296,0.,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,0.,.3218713881,0.,0.,-.3049687729e-2,0.,0.,-.1210307296,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,0.,0.,0.,.3218713882,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,-.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,.3169353608,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,-.4018262214,.4937160930,0.,0.,-.5387858374,0.,.3603626151,.2888254471,0.,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,0.,.4937160930,0.,0.,-.5387858374,0.,0.,.2888254471,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,0.,0.,0.,.4937160930,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,-.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,.1157619501,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,-.1986565342,0.,0.,0.,-.1582726231,0.,.3313752322,.4377976190,0.,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,0.,0.,0.,-.1582726231,0.,0.,.4377976190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2412424530e-1,0.,0.,0.,0.,0.,.5544791314e-1,0.,0.,-.1417410714,.3972666214},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,.1718750000,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,-.2157331140,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,.1746928107e-1,.1890216038,0.,0.,-.1725163898e-1,0.,-.7854574205e-1,-.9780759959e-1,0.,.7031250000e-1,.4812067217e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,0.,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,0.,.1890216037,0.,0.,-.1725163898e-1,0.,0.,-.9780759957e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,-.1141088661,0.,0.,0.,0.,0.,.1890216038,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,-.1786086271,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,.1300571401,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,.5297911641e-1,-.3496171374,0.,0.,-.8451542550e-1,0.,.1102626021,.1796840339,0.,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,0.,-.3496171374,0.,0.,-.8451542550e-1,0.,0.,.1796840339,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,-.3496171374,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,-.3093592167,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,.4264736607,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,-.2555233287,-.6942056633,0.,0.,.3964594048,0.,-.9743899444e-1,-.3527181261,0.,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,0.,-.6942056631,0.,0.,.3964594048,0.,0.,-.3527181261,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,0.,0.,0.,-.6942056633,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,-.9882117688e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,.1613825392,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,-.1546796083,0.,0.,0.,.6546536711e-1,0.,-.8715208612e-2,-.6464261051,0.,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,0.,0.,0.,.6546536709e-1,0.,0.,-.6464261051,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2103629347e-1,0.,0.,0.,0.,0.,.8751446639e-2,0.,0.,.9265956332e-1,-.6047514756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,-.2305827461,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,-.1467113994e-1,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,-.1210078569,.1341862601,0.,0.,.9547032698e-1,0.,.1373682881,-.1546473936,0.,-.6352789941e-1,.1066497092},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,0.,.1341862600,0.,0.,.9547032696e-1,0.,0.,-.1546473936,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,0.,0.,0.,.1341862601,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,-.3993810316,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,.2343473623,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,.3481045690,.5968901920,0.,0.,.1653594569,0.,-.2206772153,.1994047619,0.,.1132938028,-.1630966411},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,0.,.5968901920,0.,0.,.1653594569,0.,0.,.1994047619,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,0.,0.,0.,.5968901921,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,-.1275775908,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,.1426360826,0.,0.,0.,.2218529919,0.,-.4548734170,.6095947223,0.,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,0.,0.,0.,.2218529919,0.,0.,.6095947223,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1302083333e-1,0.,0.,0.,0.,0.,-.8202104657e-1,0.,0.,.1362603923,.6047900501},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,-.4725540094,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,.4269562821,0.,-.2839650695,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,.2566181299,-.3937500000,0.,0.,-.4192627457,0.,.3827327723e-1,-.1056442818e-1,0.,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,.4269562817,0.,0.,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,0.,-.3937500000,0.,0.,-.4192627457,0.,0.,-.1056442818e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4806814824,0.,0.,0.,0.,0.,-.3937500000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,-.1509518412,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,-.5229125166e-1,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,.3423605578,0.,0.,0.,-.4312500000,0.,.2909776087,-.4559196456,0.,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,0.,0.,0.,0.,-.4312500000,0.,0.,-.4559196456,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4709974231e-1,0.,0.,0.,0.,0.,-.1093750000e-1,0.,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,-.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,-.3303450769,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,-.2679129406,0.,0.,0.,.4464705337,0.,.1746728448,.2866071429,0.,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,0.,0.,0.,0.,.4464705337,0.,0.,.2866071429,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1746928107e-1,0.,0.,0.,0.,0.,.2096521686,0.,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1793351568,0.,0.,0.,0.,0.,-.3427720463,0.,0.,-.3960701481,-.2087053571}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,-.5103103630e-1,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,.1035406358,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,-.1189992461,.6614378278e-1,0.,0.,-.7395099729e-1,0.,-.2507429436e-1,-.2795084971e-1,0.,.5467611033e-1,.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,0.,.6614378278e-1,0.,0.,-.7395099729e-1,0.,0.,-.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,0.,0.,0.,.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,.9923207755e-1,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709293e-1,0.,-.1638392857,.3651483717,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,.1276586028,-.1403121520,0.,0.,.8964214566e-1,0.,.8558232884e-1,.5646924394e-1,0.,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709295e-1,0.,0.,.3651483718,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,0.,-.1403121520,0.,0.,.8964214566e-1,0.,0.,.5646924394e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,.3952847074e-1,0.,0.,0.,0.,0.,-.1403121520,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3651483717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,-.1031197389,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,.1159133539,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,.1364673394e-1,.3218713882,0.,0.,-.3049687729e-2,0.,-.2191913380,-.1210307296,0.,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,0.,.3218713881,0.,0.,-.3049687729e-2,0.,0.,-.1210307296,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,0.,0.,0.,.3218713882,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,-.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,.3169353608,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,-.4018262214,.4937160930,0.,0.,-.5387858374,0.,.3603626151,.2888254471,0.,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,0.,.4937160930,0.,0.,-.5387858374,0.,0.,.2888254471,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,0.,0.,0.,.4937160930,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,-.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,.1157619501,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,-.1986565342,0.,0.,0.,-.1582726231,0.,.3313752322,.4377976190,0.,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,0.,0.,0.,-.1582726231,0.,0.,.4377976190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2412424530e-1,0.,0.,0.,0.,0.,.5544791314e-1,0.,0.,-.1417410714,.3972666214},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,.1718750000,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,-.2157331140,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,.1746928107e-1,.1890216038,0.,0.,-.1725163898e-1,0.,-.7854574205e-1,-.9780759959e-1,0.,.7031250000e-1,.4812067217e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,0.,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,0.,.1890216037,0.,0.,-.1725163898e-1,0.,0.,-.9780759957e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,-.1141088661,0.,0.,0.,0.,0.,.1890216038,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,-.1786086271,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,.1300571401,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,.5297911641e-1,-.3496171374,0.,0.,-.8451542550e-1,0.,.1102626021,.1796840339,0.,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,0.,-.3496171374,0.,0.,-.8451542550e-1,0.,0.,.1796840339,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,-.3496171374,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,-.3093592167,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,.4264736607,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,-.2555233287,-.6942056633,0.,0.,.3964594048,0.,-.9743899444e-1,-.3527181261,0.,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,0.,-.6942056631,0.,0.,.3964594048,0.,0.,-.3527181261,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,0.,0.,0.,-.6942056633,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,-.9882117688e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,.1613825392,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,-.1546796083,0.,0.,0.,.6546536711e-1,0.,-.8715208612e-2,-.6464261051,0.,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,0.,0.,0.,.6546536709e-1,0.,0.,-.6464261051,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2103629347e-1,0.,0.,0.,0.,0.,.8751446639e-2,0.,0.,.9265956332e-1,-.6047514756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,-.2305827461,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,-.1467113994e-1,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,-.1210078569,.1341862601,0.,0.,.9547032698e-1,0.,.1373682881,-.1546473936,0.,-.6352789941e-1,.1066497092},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,0.,.1341862600,0.,0.,.9547032696e-1,0.,0.,-.1546473936,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,0.,0.,0.,.1341862601,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,-.3993810316,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,.2343473623,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,.3481045690,.5968901920,0.,0.,.1653594569,0.,-.2206772153,.1994047619,0.,.1132938028,-.1630966411},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,0.,.5968901920,0.,0.,.1653594569,0.,0.,.1994047619,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,0.,0.,0.,.5968901921,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,-.1275775908,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,.1426360826,0.,0.,0.,.2218529919,0.,-.4548734170,.6095947223,0.,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,0.,0.,0.,.2218529919,0.,0.,.6095947223,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1302083333e-1,0.,0.,0.,0.,0.,-.8202104657e-1,0.,0.,.1362603923,.6047900501},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,-.4725540094,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,.4269562821,0.,-.2839650695,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,.2566181299,-.3937500000,0.,0.,-.4192627457,0.,.3827327723e-1,-.1056442818e-1,0.,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,.4269562817,0.,0.,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,0.,-.3937500000,0.,0.,-.4192627457,0.,0.,-.1056442818e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4806814824,0.,0.,0.,0.,0.,-.3937500000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,-.1509518412,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,-.5229125166e-1,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,.3423605578,0.,0.,0.,-.4312500000,0.,.2909776087,-.4559196456,0.,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,0.,0.,0.,0.,-.4312500000,0.,0.,-.4559196456,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4709974231e-1,0.,0.,0.,0.,0.,-.1093750000e-1,0.,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,-.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,-.3303450769,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,-.2679129406,0.,0.,0.,.4464705337,0.,.1746728448,.2866071429,0.,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,0.,0.,0.,0.,.4464705337,0.,0.,.2866071429,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1746928107e-1,0.,0.,0.,0.,0.,.2096521686,0.,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1793351568,0.,0.,0.,0.,0.,-.3427720463,0.,0.,-.3960701481,-.2087053571}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,.5103103630e-1,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,-.1035406358,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,.1189992461,-.6614378278e-1,0.,0.,.7395099729e-1,0.,.2507429436e-1,.2795084971e-1,0.,-.5467611033e-1,-.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,0.,-.6614378278e-1,0.,0.,.7395099729e-1,0.,0.,.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,-.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,-.9923207755e-1,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709293e-1,0.,.1049107143,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,.4870046022e-1,-.2338535867,0.,0.,.5976143047e-1,0.,.1108137724,.1129384878,0.,-.1024538982,-.5342785200e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709295e-1,0.,0.,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,0.,-.2338535867,0.,0.,.5976143047e-1,0.,0.,.1129384878,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,.1185854122,0.,0.,0.,0.,0.,-.2338535867,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4564354646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,.1031197389,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,.4738596230e-1,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,.1162325270,-.4091585443,0.,0.,-.1189378215,0.,.3433564566e-1,.2593515633,0.,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,0.,-.4091585442,0.,0.,-.1189378215,0.,0.,.2593515633,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,-.4091585443,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,.2133947249,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,.3504772316e-1,-.3047338563,0.,0.,.6338656908e-1,0.,-.1639473093e-1,.3433961537,0.,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,0.,-.3047338563,0.,0.,.6338656908e-1,0.,0.,.3433961537,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,0.,0.,0.,-.3047338563,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,.1552904208,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,.2642678667,0.,0.,0.,.5173388725,0.,.1832986642,.2026785714,0.,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,0.,0.,0.,0.,.5173388725,0.,0.,.2026785714,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5528644321,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6513546229,0.,0.,0.,0.,0.,-.6066461029,0.,0.,-.3999255952,-.1365849382},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,-.1093750000,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,.2095472184,.1581138830,0.,0.,0.,.2966830521,0.,0.,.2916059217e-1,0.,-.2131252291,-.8100925875e-1,0.,0.,-.1035098339,0.,.8090801999e-1,.3912303983e-1,0.,.2343750000e-1,-.1850795084e-1},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.1581138830,0.,0.,0.,.2966830520,0.,0.,.2916059217e-1,0.,0.,-.8100925872e-1,0.,0.,-.1035098339,0.,0.,.3912303982e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,.2966830519,0.,0.,0.,0.,0.,-.8100925875e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,.2296396634,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,-.1224809960,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,-.1149239294,-.2551260193,0.,0.,-.4278593414,0.,-.8679563432e-2,.1617156305,0.,.1877213121,-.9067007150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,0.,-.2551260193,0.,0.,-.4278593414,0.,0.,.1617156305,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,0.,0.,0.,-.2551260193,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,-.7954951286e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,-.6561133241e-2,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,-.1143502189,-.2495867120,0.,0.,-.4025587803,0.,-.7962157829e-1,.2812523620,0.,.4211073418,-.2128834533},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,0.,-.2495867119,0.,0.,-.4025587803,0.,0.,.2812523620,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,0.,0.,0.,-.2495867120,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,-.4150489429,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,-.4254630581,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,.6629126072e-1,0.,0.,0.,.2495867120,0.,.5017470100,.1943402247,0.,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,0.,0.,0.,.2495867119,0.,0.,.1943402247,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4513392857,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2169908080,0.,0.,0.,0.,0.,-.1237704596,0.,0.,-.2735970138,-.1451294743},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,.3294039230e-1,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,-.1189992461,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,.1967692977,-.3659625275e-1,0.,0.,-.1432054905,0.,-.1672490034,.2319710903e-1,0.,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,0.,-.3659625271e-1,0.,0.,-.1432054905,0.,0.,.2319710903e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,-.3659625275e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,-.3309157119,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,.1835250428,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,.3043636809,-.6866878320e-1,0.,0.,-.3307189139,0.,-.4995787124,.7738095238e-1,0.,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,0.,-.6866878320e-1,0.,0.,-.3307189139,0.,0.,.7738095238e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,0.,0.,0.,-.6866878321e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,.4847948451,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,-.4686614144,0.,0.,0.,-.7923321136e-1,0.,-.8518830777e-1,.7520109565e-1,0.,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,0.,0.,0.,0.,-.7923321136e-1,0.,0.,.7520109565e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1573399484,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1104910714,0.,0.,0.,0.,0.,.1771009002,0.,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,.4185478369,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,-.3415650256,0.,-.5445447805,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,.3860054223,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,-.1726550062,.7042952119e-2,0.,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,-.3415650253,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,0.,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,0.,.7042952119e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,0.,0.,0.,-.6250000000e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,-.3320940506,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,.1718141126,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,.1977011672,0.,0.,0.,-.5625000000e-1,0.,-.2605485777,.1338624175e-1,0.,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,0.,0.,0.,-.5625000000e-1,0.,0.,.1338624175e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9992188321e-1,0.,0.,0.,0.,0.,.1909722222e-2,0.,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,-.2117596647,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,.1366902758,0.,0.,0.,-.7086833870e-2,0.,-.5391137183e-1,.8928571429e-3,0.,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,0.,0.,0.,-.7086833870e-2,0.,0.,.8928571429e-3,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2246050424e-1,0.,0.,0.,0.,0.,-.3248132189e-1,0.,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1379501207e-1,0.,0.,0.,0.,0.,-.5440826132e-2,0.,0.,.1233863389e-2,-.1240079365e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,.5103103630e-1,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,-.1035406358,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,.1189992461,-.6614378278e-1,0.,0.,.7395099729e-1,0.,.2507429436e-1,.2795084971e-1,0.,-.5467611033e-1,-.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,0.,-.6614378278e-1,0.,0.,.7395099729e-1,0.,0.,.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,-.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,-.9923207755e-1,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709293e-1,0.,.1049107143,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,.4870046022e-1,-.2338535867,0.,0.,.5976143047e-1,0.,.1108137724,.1129384878,0.,-.1024538982,-.5342785200e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709295e-1,0.,0.,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,0.,-.2338535867,0.,0.,.5976143047e-1,0.,0.,.1129384878,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,.1185854122,0.,0.,0.,0.,0.,-.2338535867,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4564354646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,.1031197389,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,.4738596230e-1,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,.1162325270,-.4091585443,0.,0.,-.1189378215,0.,.3433564566e-1,.2593515633,0.,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,0.,-.4091585442,0.,0.,-.1189378215,0.,0.,.2593515633,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,-.4091585443,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,.2133947249,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,.3504772316e-1,-.3047338563,0.,0.,.6338656908e-1,0.,-.1639473093e-1,.3433961537,0.,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,0.,-.3047338563,0.,0.,.6338656908e-1,0.,0.,.3433961537,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,0.,0.,0.,-.3047338563,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,.1552904208,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,.2642678667,0.,0.,0.,.5173388725,0.,.1832986642,.2026785714,0.,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,0.,0.,0.,0.,.5173388725,0.,0.,.2026785714,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5528644321,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6513546229,0.,0.,0.,0.,0.,-.6066461029,0.,0.,-.3999255952,-.1365849382},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,-.1093750000,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,.2095472184,.1581138830,0.,0.,0.,.2966830521,0.,0.,.2916059217e-1,0.,-.2131252291,-.8100925875e-1,0.,0.,-.1035098339,0.,.8090801999e-1,.3912303983e-1,0.,.2343750000e-1,-.1850795084e-1},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.1581138830,0.,0.,0.,.2966830520,0.,0.,.2916059217e-1,0.,0.,-.8100925872e-1,0.,0.,-.1035098339,0.,0.,.3912303982e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,.2966830519,0.,0.,0.,0.,0.,-.8100925875e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,.2296396634,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,-.1224809960,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,-.1149239294,-.2551260193,0.,0.,-.4278593414,0.,-.8679563432e-2,.1617156305,0.,.1877213121,-.9067007150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,0.,-.2551260193,0.,0.,-.4278593414,0.,0.,.1617156305,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,0.,0.,0.,-.2551260193,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,-.7954951286e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,-.6561133241e-2,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,-.1143502189,-.2495867120,0.,0.,-.4025587803,0.,-.7962157829e-1,.2812523620,0.,.4211073418,-.2128834533},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,0.,-.2495867119,0.,0.,-.4025587803,0.,0.,.2812523620,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,0.,0.,0.,-.2495867120,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,-.4150489429,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,-.4254630581,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,.6629126072e-1,0.,0.,0.,.2495867120,0.,.5017470100,.1943402247,0.,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,0.,0.,0.,.2495867119,0.,0.,.1943402247,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4513392857,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2169908080,0.,0.,0.,0.,0.,-.1237704596,0.,0.,-.2735970138,-.1451294743},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,.3294039230e-1,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,-.1189992461,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,.1967692977,-.3659625275e-1,0.,0.,-.1432054905,0.,-.1672490034,.2319710903e-1,0.,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,0.,-.3659625271e-1,0.,0.,-.1432054905,0.,0.,.2319710903e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,-.3659625275e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,-.3309157119,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,.1835250428,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,.3043636809,-.6866878320e-1,0.,0.,-.3307189139,0.,-.4995787124,.7738095238e-1,0.,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,0.,-.6866878320e-1,0.,0.,-.3307189139,0.,0.,.7738095238e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,0.,0.,0.,-.6866878321e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,.4847948451,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,-.4686614144,0.,0.,0.,-.7923321136e-1,0.,-.8518830777e-1,.7520109565e-1,0.,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,0.,0.,0.,0.,-.7923321136e-1,0.,0.,.7520109565e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1573399484,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1104910714,0.,0.,0.,0.,0.,.1771009002,0.,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,.4185478369,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,-.3415650256,0.,-.5445447805,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,.3860054223,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,-.1726550062,.7042952119e-2,0.,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,-.3415650253,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,0.,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,0.,.7042952119e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,0.,0.,0.,-.6250000000e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,-.3320940506,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,.1718141126,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,.1977011672,0.,0.,0.,-.5625000000e-1,0.,-.2605485777,.1338624175e-1,0.,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,0.,0.,0.,-.5625000000e-1,0.,0.,.1338624175e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9992188321e-1,0.,0.,0.,0.,0.,.1909722222e-2,0.,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,-.2117596647,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,.1366902758,0.,0.,0.,-.7086833870e-2,0.,-.5391137183e-1,.8928571429e-3,0.,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,0.,0.,0.,-.7086833870e-2,0.,0.,.8928571429e-3,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2246050424e-1,0.,0.,0.,0.,0.,-.3248132189e-1,0.,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1379501207e-1,0.,0.,0.,0.,0.,-.5440826132e-2,0.,0.,.1233863389e-2,-.1240079365e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.2357022604e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-2,.5892556509e-1,.1178511302e-1,.3928371006e-2,.3535533905e-1,.5892556509e-2,.2357022604e-1,.2357022604,.7856742012e-1,.3928371006e-1,.2357022604e-1,.7856742012e-1,.1964185503e-1,.7856742012e-2,.3928371006e-1,.7856742012e-2,.2357022604e-1,.1767766952,.5892556509e-1,.2946278254e-1,.5892556509e-1,.1473139128e-1,.2946278254e-1,.1414213562,.4714045206e-1,.4714045206e-1,.1178511302},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.1360827635e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-2,.3402069088e-1,.6804138176e-2,.2268046058e-2,.2041241452e-1,.3402069088e-2,.1360827635e-1,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.1837117307,.6123724358e-1,.3061862179e-1,.6123724358e-1,.1530931089e-1,.3061862179e-1,.1632993162,.5443310539e-1,.5443310539e-1,.1458029609},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.5270462768e-2,.1756820922e-1,.4392052306e-2,.1756820922e-2,.8784104612e-2,.1756820922e-2,.5270462768e-2,.7905694150e-1,.2635231383e-1,.1317615692e-1,.2635231383e-1,.6588078457e-2,.1317615692e-1,.9035079028e-1,.3011693010e-1,.3011693010e-1,.9411540654e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.2227177016e-2,.4454354030e-2,.1113588508e-2,.2227177016e-2,.2672612419e-1,.8908708064e-2,.8908708064e-2,.3711961694e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2,.1122391716e-2,.1122391716e-2,.8417937869e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8460345889e-3},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,-.1904761905e-1,.8333333333e-1,.8333333333e-2,0.,-.1190476190e-2,.6666666667e-1,.8333333333e-2,.1587301587e-2,.5000000000e-1,.5952380952e-2,.3809523810e-1,0.,-.2777777778e-1,-.2222222222e-1,-.1666666667e-1,.5555555556e-1,.5555555556e-2,0.,.4444444444e-1,.5555555556e-2,.3333333333e-1,0.,-.2083333333e-1,-.1666666667e-1,.4166666667e-1,.4166666667e-2,.3333333333e-1,0.,-.1666666667e-1,.3333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,-.1099714799e-1,.4811252245e-1,.4811252245e-2,0.,-.6873217492e-3,.3849001795e-1,.4811252245e-2,.9164289989e-3,.2886751347e-1,.3436608746e-2,.2199429598e-1,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2165063510e-1,-.1732050808e-1,.4330127020e-1,.4330127020e-2,.3464101616e-1,0.,-.1924500898e-1,.3849001795e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,-.3726779962e-2,.1242259987e-1,.1242259987e-2,0.,.9938079897e-2,.1242259987e-2,.7453559923e-2,0.,-.9316949905e-2,-.7453559923e-2,.1863389981e-1,.1863389981e-2,.1490711985e-1,0.,-.1064794275e-1,.2129588550e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,-.1259881577e-2,.3149703940e-2,.3149703940e-3,.2519763153e-2,0.,-.3149703940e-2,.6299407883e-2,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3968253968e-3,.7936507936e-3,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1,0.,0.,.6804138176e-2,.8748177653e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,-.2187044413e-2,.2041241452e-1,-.2916059217e-2,-.1944039478e-2,.2624453295e-1,0.,.2624453295e-1,0.,0.,.4536092117e-2,.5832118436e-2,0.,-.6804138176e-2,-.3240065798e-2,.1360827635e-1,-.1944039478e-2,.1749635531e-1,0.,0.,.3402069088e-2,0.,-.5103103630e-2,.1020620726e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,-.1262690680e-2,.1178511302e-1,-.1683587573e-2,-.1122391716e-2,.1515228816e-1,0.,.1515228816e-1,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3535533905e-2,0.,-.5303300858e-2,.1060660172e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,.1304101327e-2,0.,-.1521451549e-2,-.7245007376e-3,.3042903097e-2,-.4347004424e-3,.3912303982e-2,0.,0.,.1521451549e-2,0.,-.2282177323e-2,.4564354646e-2,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2571722499e-3,0.,-.3857583749e-3,.7715167495e-3,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1,0.,0.,0.,-.1262690680e-2,-.2244783431e-2,0.,0.,.1683587573e-2,.1543288609e-2,0.,-.2525381361e-2,-.3741305720e-3,.5050762721e-2,-.1964185503e-2,.8979133727e-2,0.,0.,0.,-.8417937869e-3,0.,0.,.1122391716e-2,0.,-.1683587573e-2,.3367175148e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,-.1296026319e-2,0.,0.,.9720197392e-3,.8910180945e-3,0.,-.1458029609e-2,-.2160043865e-3,.2916059217e-2,-.1134023029e-2,.5184105275e-2,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1882308131e-3,0.,0.,.2509744174e-3,0.,-.3764616260e-3,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2,0.,0.,0.,0.,.2509744174e-3,0.,0.,0.,-.3137180218e-3,0.,0.,.4182906958e-3,0.,-.6274360436e-3,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1449001475e-3,0.,0.,0.,-.1811251844e-3,0.,0.,.2415002458e-3,0.,-.3622503687e-3,.7245007376e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,.3299144397e-1,0.,.1443375673e-1,.9622504490e-2,.6185895742e-2,0.,.4811252245e-2,.2749286996e-2,0.,.2061965247e-2,0.,0.,.4811252245e-1,.3849001795e-1,.2886751347e-1,0.,.9622504490e-2,.6415002993e-2,0.,.3207501497e-2,0.,0.,.3608439183e-1,.2886751347e-1,0.,.7216878367e-2,0.,0.,.2886751347e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,.1904761905e-1,0.,.8333333333e-2,.5555555556e-2,.3571428571e-2,0.,.2777777778e-2,.1587301587e-2,0.,.1190476190e-2,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.3750000000e-1,.3000000000e-1,0.,.7500000000e-2,0.,0.,.3333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,.6454972245e-2,0.,.2151657415e-2,.1434438276e-2,0.,.7172191382e-3,0.,0.,.1613743061e-1,.1290994449e-1,0.,.3227486122e-2,0.,0.,.1844277784e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,.2182178902e-2,0.,.5455447254e-3,0.,0.,.5455447254e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6873217492e-3,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,.1894036020e-2,0.,.6734350295e-2,.2946278254e-2,0.,.3788072042e-2,0.,0.,0.,-.7856742012e-2,-.1010152544e-1,0.,.7856742012e-2,.3367175148e-2,0.,.4489566864e-2,0.,0.,0.,-.5892556509e-2,0.,.5892556509e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,.1093522207e-2,0.,.3888078956e-2,.1701034544e-2,0.,.2187044413e-2,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6123724358e-2,0.,.6123724358e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,-.2258769757e-2,0.,.1756820922e-2,.7529232524e-3,0.,.1003897670e-2,0.,0.,0.,-.2635231383e-2,0.,.2635231383e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4454354030e-3,0.,.4454354030e-3,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.,0.,0.,0.,.2187044413e-2,.3888078956e-2,0.,0.,-.2430049348e-2,-.2187044413e-2,0.,.2430049348e-2,0.,0.,.2430049348e-2,0.,0.,0.,0.,.1458029609e-2,0.,0.,-.1620032899e-2,0.,.1620032899e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,.2244783431e-2,0.,0.,-.1402989645e-2,-.1262690680e-2,0.,.1402989645e-2,0.,0.,.1402989645e-2,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3260253318e-3,0.,0.,-.3622503687e-3,0.,.3622503687e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.,0.,0.,0.,0.,-.4347004424e-3,0.,0.,0.,.4890379978e-3,0.,0.,-.5433755531e-3,0.,.5433755531e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2509744174e-3,0.,0.,0.,.2823462197e-3,0.,0.,-.3137180218e-3,0.,.3137180218e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,.1956151991e-1,0.,0.,.2173502212e-2,.2445189989e-2,0.,0.,.5433755531e-3,0.,0.,0.,0.,0.,.1014301032e-1,.1304101327e-1,0.,0.,.1449001475e-2,0.,0.,0.,0.,0.,.7607257744e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,.1129384878e-1,0.,0.,.1254872087e-2,.1411731098e-2,0.,0.,.3137180218e-3,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.7905694150e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,.2916059217e-2,0.,0.,.3240065798e-3,0.,0.,0.,0.,0.,.3402069088e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5750546329e-3,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,-.2823462197e-2,-.5019488348e-2,0.,0.,.1882308131e-2,.1568590109e-2,0.,0.,.8365813917e-3,0.,0.,0.,0.,0.,0.,-.1882308131e-2,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,-.2898002950e-2,0.,0.,.1086751106e-2,.9056259216e-3,0.,0.,.4830004916e-3,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4208968934e-3,0.,0.,.2805979290e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.,0.,0.,0.,0.,.5611958579e-3,0.,0.,0.,-.4910463757e-3,0.,0.,.3273642505e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3240065798e-3,0.,0.,0.,-.2835057572e-3,0.,0.,.1890038382e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,.5939138708e-2,0.,0.,0.,.3711961694e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2227177016e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,.3428963332e-2,0.,0.,0.,.2143102083e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4980119205e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6640158940e-3,0.,0.,0.,.3320079470e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3833697554e-3,0.,0.,0.,.1916848777e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4347004424e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.2357022604e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-2,.5892556509e-1,.1178511302e-1,.3928371006e-2,.3535533905e-1,.5892556509e-2,.2357022604e-1,.2357022604,.7856742012e-1,.3928371006e-1,.2357022604e-1,.7856742012e-1,.1964185503e-1,.7856742012e-2,.3928371006e-1,.7856742012e-2,.2357022604e-1,.1767766952,.5892556509e-1,.2946278254e-1,.5892556509e-1,.1473139128e-1,.2946278254e-1,.1414213562,.4714045206e-1,.4714045206e-1,.1178511302},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.1360827635e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-2,.3402069088e-1,.6804138176e-2,.2268046058e-2,.2041241452e-1,.3402069088e-2,.1360827635e-1,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.1837117307,.6123724358e-1,.3061862179e-1,.6123724358e-1,.1530931089e-1,.3061862179e-1,.1632993162,.5443310539e-1,.5443310539e-1,.1458029609},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.5270462768e-2,.1756820922e-1,.4392052306e-2,.1756820922e-2,.8784104612e-2,.1756820922e-2,.5270462768e-2,.7905694150e-1,.2635231383e-1,.1317615692e-1,.2635231383e-1,.6588078457e-2,.1317615692e-1,.9035079028e-1,.3011693010e-1,.3011693010e-1,.9411540654e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.2227177016e-2,.4454354030e-2,.1113588508e-2,.2227177016e-2,.2672612419e-1,.8908708064e-2,.8908708064e-2,.3711961694e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2,.1122391716e-2,.1122391716e-2,.8417937869e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8460345889e-3},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,-.1904761905e-1,.8333333333e-1,.8333333333e-2,0.,-.1190476190e-2,.6666666667e-1,.8333333333e-2,.1587301587e-2,.5000000000e-1,.5952380952e-2,.3809523810e-1,0.,-.2777777778e-1,-.2222222222e-1,-.1666666667e-1,.5555555556e-1,.5555555556e-2,0.,.4444444444e-1,.5555555556e-2,.3333333333e-1,0.,-.2083333333e-1,-.1666666667e-1,.4166666667e-1,.4166666667e-2,.3333333333e-1,0.,-.1666666667e-1,.3333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,-.1099714799e-1,.4811252245e-1,.4811252245e-2,0.,-.6873217492e-3,.3849001795e-1,.4811252245e-2,.9164289989e-3,.2886751347e-1,.3436608746e-2,.2199429598e-1,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2165063510e-1,-.1732050808e-1,.4330127020e-1,.4330127020e-2,.3464101616e-1,0.,-.1924500898e-1,.3849001795e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,-.3726779962e-2,.1242259987e-1,.1242259987e-2,0.,.9938079897e-2,.1242259987e-2,.7453559923e-2,0.,-.9316949905e-2,-.7453559923e-2,.1863389981e-1,.1863389981e-2,.1490711985e-1,0.,-.1064794275e-1,.2129588550e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,-.1259881577e-2,.3149703940e-2,.3149703940e-3,.2519763153e-2,0.,-.3149703940e-2,.6299407883e-2,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3968253968e-3,.7936507936e-3,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1,0.,0.,.6804138176e-2,.8748177653e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,-.2187044413e-2,.2041241452e-1,-.2916059217e-2,-.1944039478e-2,.2624453295e-1,0.,.2624453295e-1,0.,0.,.4536092117e-2,.5832118436e-2,0.,-.6804138176e-2,-.3240065798e-2,.1360827635e-1,-.1944039478e-2,.1749635531e-1,0.,0.,.3402069088e-2,0.,-.5103103630e-2,.1020620726e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,-.1262690680e-2,.1178511302e-1,-.1683587573e-2,-.1122391716e-2,.1515228816e-1,0.,.1515228816e-1,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3535533905e-2,0.,-.5303300858e-2,.1060660172e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,.1304101327e-2,0.,-.1521451549e-2,-.7245007376e-3,.3042903097e-2,-.4347004424e-3,.3912303982e-2,0.,0.,.1521451549e-2,0.,-.2282177323e-2,.4564354646e-2,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2571722499e-3,0.,-.3857583749e-3,.7715167495e-3,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1,0.,0.,0.,-.1262690680e-2,-.2244783431e-2,0.,0.,.1683587573e-2,.1543288609e-2,0.,-.2525381361e-2,-.3741305720e-3,.5050762721e-2,-.1964185503e-2,.8979133727e-2,0.,0.,0.,-.8417937869e-3,0.,0.,.1122391716e-2,0.,-.1683587573e-2,.3367175148e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,-.1296026319e-2,0.,0.,.9720197392e-3,.8910180945e-3,0.,-.1458029609e-2,-.2160043865e-3,.2916059217e-2,-.1134023029e-2,.5184105275e-2,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1882308131e-3,0.,0.,.2509744174e-3,0.,-.3764616260e-3,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2,0.,0.,0.,0.,.2509744174e-3,0.,0.,0.,-.3137180218e-3,0.,0.,.4182906958e-3,0.,-.6274360436e-3,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1449001475e-3,0.,0.,0.,-.1811251844e-3,0.,0.,.2415002458e-3,0.,-.3622503687e-3,.7245007376e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,.3299144397e-1,0.,.1443375673e-1,.9622504490e-2,.6185895742e-2,0.,.4811252245e-2,.2749286996e-2,0.,.2061965247e-2,0.,0.,.4811252245e-1,.3849001795e-1,.2886751347e-1,0.,.9622504490e-2,.6415002993e-2,0.,.3207501497e-2,0.,0.,.3608439183e-1,.2886751347e-1,0.,.7216878367e-2,0.,0.,.2886751347e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,.1904761905e-1,0.,.8333333333e-2,.5555555556e-2,.3571428571e-2,0.,.2777777778e-2,.1587301587e-2,0.,.1190476190e-2,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.3750000000e-1,.3000000000e-1,0.,.7500000000e-2,0.,0.,.3333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,.6454972245e-2,0.,.2151657415e-2,.1434438276e-2,0.,.7172191382e-3,0.,0.,.1613743061e-1,.1290994449e-1,0.,.3227486122e-2,0.,0.,.1844277784e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,.2182178902e-2,0.,.5455447254e-3,0.,0.,.5455447254e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6873217492e-3,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,.1894036020e-2,0.,.6734350295e-2,.2946278254e-2,0.,.3788072042e-2,0.,0.,0.,-.7856742012e-2,-.1010152544e-1,0.,.7856742012e-2,.3367175148e-2,0.,.4489566864e-2,0.,0.,0.,-.5892556509e-2,0.,.5892556509e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,.1093522207e-2,0.,.3888078956e-2,.1701034544e-2,0.,.2187044413e-2,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6123724358e-2,0.,.6123724358e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,-.2258769757e-2,0.,.1756820922e-2,.7529232524e-3,0.,.1003897670e-2,0.,0.,0.,-.2635231383e-2,0.,.2635231383e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4454354030e-3,0.,.4454354030e-3,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.,0.,0.,0.,.2187044413e-2,.3888078956e-2,0.,0.,-.2430049348e-2,-.2187044413e-2,0.,.2430049348e-2,0.,0.,.2430049348e-2,0.,0.,0.,0.,.1458029609e-2,0.,0.,-.1620032899e-2,0.,.1620032899e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,.2244783431e-2,0.,0.,-.1402989645e-2,-.1262690680e-2,0.,.1402989645e-2,0.,0.,.1402989645e-2,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3260253318e-3,0.,0.,-.3622503687e-3,0.,.3622503687e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.,0.,0.,0.,0.,-.4347004424e-3,0.,0.,0.,.4890379978e-3,0.,0.,-.5433755531e-3,0.,.5433755531e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2509744174e-3,0.,0.,0.,.2823462197e-3,0.,0.,-.3137180218e-3,0.,.3137180218e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,.1956151991e-1,0.,0.,.2173502212e-2,.2445189989e-2,0.,0.,.5433755531e-3,0.,0.,0.,0.,0.,.1014301032e-1,.1304101327e-1,0.,0.,.1449001475e-2,0.,0.,0.,0.,0.,.7607257744e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,.1129384878e-1,0.,0.,.1254872087e-2,.1411731098e-2,0.,0.,.3137180218e-3,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.7905694150e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,.2916059217e-2,0.,0.,.3240065798e-3,0.,0.,0.,0.,0.,.3402069088e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5750546329e-3,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,-.2823462197e-2,-.5019488348e-2,0.,0.,.1882308131e-2,.1568590109e-2,0.,0.,.8365813917e-3,0.,0.,0.,0.,0.,0.,-.1882308131e-2,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,-.2898002950e-2,0.,0.,.1086751106e-2,.9056259216e-3,0.,0.,.4830004916e-3,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4208968934e-3,0.,0.,.2805979290e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.,0.,0.,0.,0.,.5611958579e-3,0.,0.,0.,-.4910463757e-3,0.,0.,.3273642505e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3240065798e-3,0.,0.,0.,-.2835057572e-3,0.,0.,.1890038382e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,.5939138708e-2,0.,0.,0.,.3711961694e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2227177016e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,.3428963332e-2,0.,0.,0.,.2143102083e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4980119205e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6640158940e-3,0.,0.,0.,.3320079470e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3833697554e-3,0.,0.,0.,.1916848777e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4347004424e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.,3.464101616,-4.472135954,5.291502622,-6.,2.449489743,-4.242640686,5.477225575,-6.480740698,-2.828427124,4.898979486,-6.324555320,3.162277660,-5.477225575,-3.464101616,-3.464101616,6.,-7.745966692,9.165151390,-10.39230485,4.242640686,-7.348469229,9.486832980,-11.22497216,-4.898979486,8.485281372,-10.95445115,5.477225575,-9.486832980,-6.,5.477225575,-9.486832980,12.24744872,-14.49137675,-6.324555320,10.95445115,-14.14213562,7.071067810,-12.24744872,-7.745966692,-7.483314774,12.96148140,-16.73320053,8.366600265,-14.49137675,-9.165151390,9.486832980,-16.43167672,-10.39230485,-11.48912529},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,140.7124728,0.,-6.928203232,26.83281572,-63.49803146,120.,0.,8.485281372,-32.86335345,77.76888838,0.,-9.797958972,37.94733192,0.,10.95445115,0.,0.,-12.,46.47580015,-109.9818167,207.8460970,0.,14.69693846,-56.92099788,134.6996659,0.,-16.97056274,65.72670690,0.,18.97366596,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-21.90890230,84.85281372,0.,24.49489743,0.,0.,-25.92296279,100.3992032,0.,28.98275350,0.,0.,32.86335345,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,-984.9873096,0.,0.,-26.83281572,158.7450787,-540.,0.,0.,32.86335345,-194.4222209,0.,0.,-37.94733192,0.,0.,0.,0.,0.,-46.47580015,274.9545417,-935.3074363,0.,0.,56.92099788,-336.7491648,0.,0.,-65.72670690,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-84.85281372,0.,0.,0.,0.,0.,-100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,2626.632826,0.,0.,0.,-105.8300524,840.,0.,0.,0.,129.6148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,1454.922679,0.,0.,0.,224.4994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,-2954.961929,0.,0.,0.,0.,-420.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1181.984772,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,18.,-19.59591794,33.94112549,-43.81780460,51.84592558,42.42640686,-73.48469229,94.86832980,-75.89466384,131.4534138,121.2435566,3.464101616,-6.,7.745966692,-9.165151390,10.39230485,-25.45584412,44.09081537,-56.92099788,67.34983297,63.68673332,-110.3086578,142.4078650,-120.4989626,208.7103256,198.,-10.95445115,18.97366596,-24.49489743,28.98275350,56.92099788,-98.59006035,127.2792206,-127.2792206,220.4540769,224.6330341,22.44994432,-38.88444419,50.19960159,-100.3992032,173.8965210,210.7984820,-37.94733192,65.72670690,155.8845727,57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,-360.,0.,-67.88225098,262.9068276,-622.1511070,0.,146.9693846,-569.2099788,0.,-262.9068276,0.,0.,12.,-46.47580015,109.9818167,-207.8460970,0.,-88.18163075,341.5259873,-808.1979956,0.,220.6173157,-854.4471897,0.,-417.4206511,0.,0.,-37.94733192,146.9693846,-347.7930420,0.,197.1801207,-763.6753235,0.,-440.9081537,0.,0.,77.76888838,-301.1976095,0.,-347.7930420,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,1620.,0.,0.,-262.9068276,1555.377768,0.,0.,569.2099788,0.,0.,0.,0.,0.,46.47580015,-274.9545417,935.3074363,0.,0.,-341.5259873,2020.494989,0.,0.,854.4471897,0.,0.,0.,0.,0.,-146.9693846,869.4826050,0.,0.,763.6753235,0.,0.,0.,0.,0.,301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,-2520.,0.,0.,0.,-1036.918512,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,-1454.922679,0.,0.,0.,-1346.996659,0.,0.,0.,0.,0.,0.,0.,0.,0.,-579.6550700,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1260.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-64.80740698,-127.2792206,220.4540769,-284.6049894,398.4469852,-690.1304224,-969.9484525,0.,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-56.12486080,-161.6663230,280.0142853,-361.4968880,575.1086854,-996.1174629,-1488.,5.477225575,-9.486832980,12.24744872,-14.49137675,-94.86832980,164.3167672,-212.1320343,487.9036789,-845.0739613,-1471.733671,-22.44994432,38.88444419,-50.19960159,250.9980080,-434.7413025,-1081.487864,56.92099788,-98.59006035,-519.6152424,-114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,777.6888838,0.,-440.9081537,1707.629936,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,673.4983297,0.,-560.0285706,2168.981328,0.,1992.234926,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-328.6335345,1272.792206,0.,1690.147923,0.,0.,-77.76888838,301.1976095,0.,869.4826050,0.,0.,197.1801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,-1944.222209,0.,0.,-1707.629936,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,-1683.745824,0.,0.,-2168.981328,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-1272.792206,0.,0.,0.,0.,0.,-301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1296.148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1122.497216,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,221.3594362,-708.3501958,1226.898529,2909.845357,0.,0.,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,230.0434742,-920.1738966,1593.787941,4176.,0.,0.,0.,0.,44.27188724,-76.68115805,98.99494934,-622.2539673,1077.775487,3578.636612,7.483314774,-12.96148140,16.73320053,-234.2648074,405.7585490,2071.324214,-37.94733192,65.72670690,727.4613394,114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,-1328.156617,0.,-2453.797058,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,-1380.260845,0.,-3187.575881,0.,0.,0.,0.,0.,0.,153.3623161,-593.9696960,0.,-2155.550974,0.,0.,25.92296279,-100.3992032,0.,-811.5170980,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1328.156617,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,593.9696960,0.,0.,0.,0.,0.,100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,-690.1304224,-3637.306697,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,-796.8939703,-4860.,0.,0.,0.,0.,0.,0.,0.,254.5584412,-440.9081537,-3601.874512,0.,0.,0.,75.29940238,-130.4223908,-1695.553007,9.486832980,-16.43167672,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1593.787941,0.,0.,0.,0.,0.,0.,0.,0.,0.,881.8163075,0.,0.,0.,0.,0.,260.8447815,0.,0.,32.86335345,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1980.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1278.084504,0.,0.,0.,0.,0.,504.0833264,0.,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,20.78460970,-8.485281372,14.69693846,-18.97366596,22.44994432,9.797958972,-16.97056274,21.90890230,-10.95445115,18.97366596,12.,-32.86335345,56.92099788,-73.48469229,86.94826050,37.94733192,-65.72670690,84.85281372,-42.42640686,73.48469229,46.47580015,89.79977729,-155.5377768,200.7984064,-100.3992032,173.8965210,109.9818167,-189.7366596,328.6335345,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,-415.6921939,0.,-29.39387692,113.8419958,-269.3993319,0.,33.94112549,-131.4534138,0.,-37.94733192,0.,0.,-113.8419958,440.9081537,-1043.379126,0.,131.4534138,-509.1168823,0.,-146.9693846,0.,0.,311.0755535,-1204.790438,0.,-347.7930420,0.,0.,-657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,1870.614873,0.,0.,-113.8419958,673.4983297,0.,0.,131.4534138,0.,0.,0.,0.,0.,-440.9081537,2608.447815,0.,0.,509.1168823,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,-2909.845357,0.,0.,0.,-448.9988864,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1454.922679,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-112.2497216,-117.5755077,203.6467529,-262.9068276,230.0434742,-398.4469852,-384.,32.86335345,-56.92099788,73.48469229,-86.94826050,-303.5786554,525.8136552,-678.8225098,721.2489166,-1249.239769,-1301.322404,-179.5995546,311.0755535,-401.5968127,1104.391235,-1912.861731,-2419.599967,569.2099788,-985.9006035,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,1346.996659,0.,-407.2935059,1577.440966,0.,796.8939703,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-1051.627310,4072.935059,0.,2498.479538,0.,0.,-622.1511070,2409.580876,0.,3825.723462,0.,0.,1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,-3367.491648,0.,0.,-1577.440966,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-4072.935059,0.,0.,0.,0.,0.,-2409.580876,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2244.994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,460.0869483,-920.1738966,1593.787941,2592.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-2206.173157,3821.203999,7529.079625,89.79977729,-155.5377768,200.7984064,-1907.584860,3304.033899,10558.25440,-569.2099788,985.9006035,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,-2760.521690,0.,-3187.575881,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-7642.407998,0.,0.,311.0755535,-1204.790438,0.,-6608.067798,0.,0.,-1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2760.521690,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-1593.787941,-5760.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-13942.74005,0.,0.,0.,903.5928286,-1565.068689,-14297.63617,189.7366596,-328.6335345,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3187.575881,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,3130.137378,0.,0.,657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,6048.999917,0.,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-86.94826050,-37.94733192,65.72670690,-84.85281372,42.42640686,-73.48469229,-46.47580015,-224.4994432,388.8444419,-501.9960159,250.9980080,-434.7413025,-274.9545417,853.8149682,-1478.850905,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-131.4534138,509.1168823,0.,146.9693846,0.,0.,-777.6888838,3011.976095,0.,869.4826050,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-509.1168823,0.,0.,0.,0.,0.,-3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-678.8225098,1175.755077,1254.846604,224.4994432,-388.8444419,501.9960159,-2509.980080,4347.413025,5774.045376,-1707.629936,2957.701810,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-2351.510153,0.,0.,777.6888838,-3011.976095,0.,-8694.826050,0.,0.,-5915.403621,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-6274.233021,0.,0.,0.,2258.982072,-3912.671722,-20621.59063,853.8149682,-1478.850905,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,7825.343445,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,15122.49979,0.,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,334.6640106,-167.3320053,289.8275350,183.3030278,-1328.156617,2300.434742,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,-2007.984064,0.,-579.6550700,0.,0.,-4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2007.984064,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-2608.447815,-3666.060556,1328.156617,-2300.434742,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5216.895630,0.,0.,4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-1150.217371,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2300.434742,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.,3.464101616,-4.472135954,5.291502622,-6.,2.449489743,-4.242640686,5.477225575,-6.480740698,-2.828427124,4.898979486,-6.324555320,3.162277660,-5.477225575,-3.464101616,-3.464101616,6.,-7.745966692,9.165151390,-10.39230485,4.242640686,-7.348469229,9.486832980,-11.22497216,-4.898979486,8.485281372,-10.95445115,5.477225575,-9.486832980,-6.,5.477225575,-9.486832980,12.24744872,-14.49137675,-6.324555320,10.95445115,-14.14213562,7.071067810,-12.24744872,-7.745966692,-7.483314774,12.96148140,-16.73320053,8.366600265,-14.49137675,-9.165151390,9.486832980,-16.43167672,-10.39230485,-11.48912529},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,140.7124728,0.,-6.928203232,26.83281572,-63.49803146,120.,0.,8.485281372,-32.86335345,77.76888838,0.,-9.797958972,37.94733192,0.,10.95445115,0.,0.,-12.,46.47580015,-109.9818167,207.8460970,0.,14.69693846,-56.92099788,134.6996659,0.,-16.97056274,65.72670690,0.,18.97366596,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-21.90890230,84.85281372,0.,24.49489743,0.,0.,-25.92296279,100.3992032,0.,28.98275350,0.,0.,32.86335345,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,-984.9873096,0.,0.,-26.83281572,158.7450787,-540.,0.,0.,32.86335345,-194.4222209,0.,0.,-37.94733192,0.,0.,0.,0.,0.,-46.47580015,274.9545417,-935.3074363,0.,0.,56.92099788,-336.7491648,0.,0.,-65.72670690,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-84.85281372,0.,0.,0.,0.,0.,-100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,2626.632826,0.,0.,0.,-105.8300524,840.,0.,0.,0.,129.6148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,1454.922679,0.,0.,0.,224.4994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,-2954.961929,0.,0.,0.,0.,-420.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1181.984772,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,18.,-19.59591794,33.94112549,-43.81780460,51.84592558,42.42640686,-73.48469229,94.86832980,-75.89466384,131.4534138,121.2435566,3.464101616,-6.,7.745966692,-9.165151390,10.39230485,-25.45584412,44.09081537,-56.92099788,67.34983297,63.68673332,-110.3086578,142.4078650,-120.4989626,208.7103256,198.,-10.95445115,18.97366596,-24.49489743,28.98275350,56.92099788,-98.59006035,127.2792206,-127.2792206,220.4540769,224.6330341,22.44994432,-38.88444419,50.19960159,-100.3992032,173.8965210,210.7984820,-37.94733192,65.72670690,155.8845727,57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,-360.,0.,-67.88225098,262.9068276,-622.1511070,0.,146.9693846,-569.2099788,0.,-262.9068276,0.,0.,12.,-46.47580015,109.9818167,-207.8460970,0.,-88.18163075,341.5259873,-808.1979956,0.,220.6173157,-854.4471897,0.,-417.4206511,0.,0.,-37.94733192,146.9693846,-347.7930420,0.,197.1801207,-763.6753235,0.,-440.9081537,0.,0.,77.76888838,-301.1976095,0.,-347.7930420,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,1620.,0.,0.,-262.9068276,1555.377768,0.,0.,569.2099788,0.,0.,0.,0.,0.,46.47580015,-274.9545417,935.3074363,0.,0.,-341.5259873,2020.494989,0.,0.,854.4471897,0.,0.,0.,0.,0.,-146.9693846,869.4826050,0.,0.,763.6753235,0.,0.,0.,0.,0.,301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,-2520.,0.,0.,0.,-1036.918512,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,-1454.922679,0.,0.,0.,-1346.996659,0.,0.,0.,0.,0.,0.,0.,0.,0.,-579.6550700,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1260.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-64.80740698,-127.2792206,220.4540769,-284.6049894,398.4469852,-690.1304224,-969.9484525,0.,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-56.12486080,-161.6663230,280.0142853,-361.4968880,575.1086854,-996.1174629,-1488.,5.477225575,-9.486832980,12.24744872,-14.49137675,-94.86832980,164.3167672,-212.1320343,487.9036789,-845.0739613,-1471.733671,-22.44994432,38.88444419,-50.19960159,250.9980080,-434.7413025,-1081.487864,56.92099788,-98.59006035,-519.6152424,-114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,777.6888838,0.,-440.9081537,1707.629936,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,673.4983297,0.,-560.0285706,2168.981328,0.,1992.234926,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-328.6335345,1272.792206,0.,1690.147923,0.,0.,-77.76888838,301.1976095,0.,869.4826050,0.,0.,197.1801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,-1944.222209,0.,0.,-1707.629936,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,-1683.745824,0.,0.,-2168.981328,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-1272.792206,0.,0.,0.,0.,0.,-301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1296.148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1122.497216,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,221.3594362,-708.3501958,1226.898529,2909.845357,0.,0.,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,230.0434742,-920.1738966,1593.787941,4176.,0.,0.,0.,0.,44.27188724,-76.68115805,98.99494934,-622.2539673,1077.775487,3578.636612,7.483314774,-12.96148140,16.73320053,-234.2648074,405.7585490,2071.324214,-37.94733192,65.72670690,727.4613394,114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,-1328.156617,0.,-2453.797058,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,-1380.260845,0.,-3187.575881,0.,0.,0.,0.,0.,0.,153.3623161,-593.9696960,0.,-2155.550974,0.,0.,25.92296279,-100.3992032,0.,-811.5170980,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1328.156617,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,593.9696960,0.,0.,0.,0.,0.,100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,-690.1304224,-3637.306697,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,-796.8939703,-4860.,0.,0.,0.,0.,0.,0.,0.,254.5584412,-440.9081537,-3601.874512,0.,0.,0.,75.29940238,-130.4223908,-1695.553007,9.486832980,-16.43167672,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1593.787941,0.,0.,0.,0.,0.,0.,0.,0.,0.,881.8163075,0.,0.,0.,0.,0.,260.8447815,0.,0.,32.86335345,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1980.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1278.084504,0.,0.,0.,0.,0.,504.0833264,0.,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,20.78460970,-8.485281372,14.69693846,-18.97366596,22.44994432,9.797958972,-16.97056274,21.90890230,-10.95445115,18.97366596,12.,-32.86335345,56.92099788,-73.48469229,86.94826050,37.94733192,-65.72670690,84.85281372,-42.42640686,73.48469229,46.47580015,89.79977729,-155.5377768,200.7984064,-100.3992032,173.8965210,109.9818167,-189.7366596,328.6335345,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,-415.6921939,0.,-29.39387692,113.8419958,-269.3993319,0.,33.94112549,-131.4534138,0.,-37.94733192,0.,0.,-113.8419958,440.9081537,-1043.379126,0.,131.4534138,-509.1168823,0.,-146.9693846,0.,0.,311.0755535,-1204.790438,0.,-347.7930420,0.,0.,-657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,1870.614873,0.,0.,-113.8419958,673.4983297,0.,0.,131.4534138,0.,0.,0.,0.,0.,-440.9081537,2608.447815,0.,0.,509.1168823,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,-2909.845357,0.,0.,0.,-448.9988864,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1454.922679,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-112.2497216,-117.5755077,203.6467529,-262.9068276,230.0434742,-398.4469852,-384.,32.86335345,-56.92099788,73.48469229,-86.94826050,-303.5786554,525.8136552,-678.8225098,721.2489166,-1249.239769,-1301.322404,-179.5995546,311.0755535,-401.5968127,1104.391235,-1912.861731,-2419.599967,569.2099788,-985.9006035,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,1346.996659,0.,-407.2935059,1577.440966,0.,796.8939703,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-1051.627310,4072.935059,0.,2498.479538,0.,0.,-622.1511070,2409.580876,0.,3825.723462,0.,0.,1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,-3367.491648,0.,0.,-1577.440966,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-4072.935059,0.,0.,0.,0.,0.,-2409.580876,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2244.994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,460.0869483,-920.1738966,1593.787941,2592.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-2206.173157,3821.203999,7529.079625,89.79977729,-155.5377768,200.7984064,-1907.584860,3304.033899,10558.25440,-569.2099788,985.9006035,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,-2760.521690,0.,-3187.575881,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-7642.407998,0.,0.,311.0755535,-1204.790438,0.,-6608.067798,0.,0.,-1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2760.521690,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-1593.787941,-5760.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-13942.74005,0.,0.,0.,903.5928286,-1565.068689,-14297.63617,189.7366596,-328.6335345,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3187.575881,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,3130.137378,0.,0.,657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,6048.999917,0.,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-86.94826050,-37.94733192,65.72670690,-84.85281372,42.42640686,-73.48469229,-46.47580015,-224.4994432,388.8444419,-501.9960159,250.9980080,-434.7413025,-274.9545417,853.8149682,-1478.850905,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-131.4534138,509.1168823,0.,146.9693846,0.,0.,-777.6888838,3011.976095,0.,869.4826050,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-509.1168823,0.,0.,0.,0.,0.,-3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-678.8225098,1175.755077,1254.846604,224.4994432,-388.8444419,501.9960159,-2509.980080,4347.413025,5774.045376,-1707.629936,2957.701810,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-2351.510153,0.,0.,777.6888838,-3011.976095,0.,-8694.826050,0.,0.,-5915.403621,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-6274.233021,0.,0.,0.,2258.982072,-3912.671722,-20621.59063,853.8149682,-1478.850905,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,7825.343445,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,15122.49979,0.,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,334.6640106,-167.3320053,289.8275350,183.3030278,-1328.156617,2300.434742,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,-2007.984064,0.,-579.6550700,0.,0.,-4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2007.984064,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-2608.447815,-3666.060556,1328.156617,-2300.434742,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5216.895630,0.,0.,4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-1150.217371,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2300.434742,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[56],
		T o_output[1]
	) {
		static const T matrix[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[56],
		T o_output[1]
	) {
		static const T matrix[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{6.633249580,0.,14.83239697,0.,19.89974874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[56],
		T o_output[56]
	) {
		static const T matrix[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{6.633249580,0.,14.83239697,0.,19.89974874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,0.,0.,0.,0.,0.,-1.360827635,-2.357022604,-3.042903097,-3.600411499,.8380524812,1.451549477,1.873942317,.8198497638,1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-2.721655270,-3.220305944,-1.873942317,-3.245763304,-4.190262406,.5237828008,.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,1.217161239,2.052800957,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,-1.704828602,-2.952849755,-3.812112642,-4.510552507,-5.114485804,-3.104623248,-5.377365205,-6.942148628,-8.214061030,-.5462559154,-.9461429991,-1.221465360,2.694797933,4.667526936,1.022426944,1.298058152,2.248302670,2.902546266,3.434339057,3.719687555,6.442687835,8.317474224,3.264192666,5.653747546,-1.221656347,-.8344196900,-1.445257298,-1.865819148,-3.243190051,-5.617369949,-4.760960496,.4850599026,.8401483961,2.375793728,-.2614753427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,1.704828601,2.952849754,3.812112640,4.510552505,5.114485802,3.104623246,5.377365203,6.942148625,8.214061027,.5462559149,.9461429982,1.221465359,-2.694797932,-4.667526936,-1.022426944,1.298058156,2.248302677,2.902546276,3.434339069,3.719687568,6.442687856,8.317474253,3.264192678,5.653747565,-1.221656351,.8344196802,1.445257281,1.865819127,3.243190014,5.617369884,4.760960437,.4850600816,.8401487061,2.375794604,.2614758325},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.968566503,-3.409657202,-4.401848518,-5.208337406,-5.905699510,2.347500666,4.065990423,5.249171065,6.210902963,-2.609637841,-4.520025331,-5.835327609,2.775506203,4.807317761,-2.855122567,-.3360178568e-9,-.582e-9,-.7513587691e-9,-.8890196848e-9,-.1008053570e-8,.4007561299e-9,.6941299786e-9,.8961179490e-9,.1060301056e-8,-.4456004436e-9,-.7718026080e-9,-.9963928825e-9,.4740581912e-9,.8210928729e-9,-.4878407244e-9,-2.709993221,-4.693845947,-6.059729062,-7.169968120,3.014474194,5.221222463,6.740569213,-3.208782240,-5.557773872,3.304507380,-.7333648479e-8,-.1270225177e-7,-.1639853652e-7,.7812670753e-8,.1353194269e-7,-.8054303853e-8,3.483595229,6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,-.8327988349,-1.442449894,-1.862194806,-2.203378609,-2.498396504,-1.088657954,-1.885610890,-2.434313191,-2.880318211,.4375794197,.7579097871,.9784573278,.9743405328,1.687607307,-.5933186699,-.4672884697,-.8093673712,-1.044888783,-1.236329082,-1.022115558,-1.770356077,-2.285519867,-.3251247057,-.5631325093,.8276510476,.6684136118,1.157726336,1.494618273,2.033596264,3.522292053,1.746351412,-.1476726503,-.2557765332,-.5739768229,-.2050386800},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,.8327988342,1.442449893,1.862194804,2.203378607,2.498396503,1.088657953,1.885610888,2.434313189,2.880318209,-.4375794194,-.7579097864,-.9784573273,-.9743405322,-1.687607306,.5933186693,-.4672884647,-.8093673627,-1.044888772,-1.236329069,-1.022115547,-1.770356058,-2.285519843,-.3251247023,-.5631325034,.8276510391,-.6684136118,-1.157726336,-1.494618273,-2.033596264,-3.522292053,-1.746351412,-.1476726019,-.2557764494,-.5739766349,.2050384112},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.961633263,-1.665597669,-2.150277344,-2.544242465,-2.884899788,-.2081759125,-.3605712573,-.4654954917,-.5507816934,1.215013765,2.104465573,2.716853372,-1.352704319,-2.342952608,.5364099445,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.1039230485e-8,.5714598801e-10,.9897975472e-10,.1277823139e-9,.1511940727e-9,.2193666394e-9,.3799541647e-9,.4905187176e-9,-.3033617129e-9,-.5254378997e-9,.1563891368e-9,-1.872739842,-3.243680556,-4.187573592,-4.954803895,-.4572029592,-.7918987547,-1.022336896,1.670078386,2.892660618,-1.283464460,0.,0.,0.,0.,0.,-0.,1.663589468,2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,-1.425047564,-2.468254783,-3.186503223,-3.770321460,-4.275142691,-2.360183599,-4.087957910,-5.277530967,-6.244458853,-.4945548803e-2,-.8565941795e-2,-.1105858331e-1,2.175490654,3.768060344,.1537427033,.6224210815,1.078064937,1.391775849,1.646771393,1.648153002,2.854684738,3.685382148,1.157573982,2.004976951,-.8770258842,-.7554460893e-2,-.1308471009e-1,-.1689228809e-1,-.2731590581e-1,-.4731253672e-1,-.3482606398e-1,-.2198338910,-.3807634685,-1.005417380,.1997404621},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,1.425047564,2.468254784,3.186503224,3.770321460,4.275142693,2.360183599,4.087957910,5.277530967,6.244458853,.4945548803e-2,.8565941795e-2,.1105858331e-1,-2.175490654,-3.768060344,-.1537427033,.6224210804,1.078064935,1.391775847,1.646771390,1.648152999,2.854684733,3.685382143,1.157573980,2.004976947,-.8770258827,.7554451164e-2,.1308469324e-1,.1689226633e-1,.2731587063e-1,.4731247580e-1,.3482601912e-1,-.2198339029,-.3807634890,-1.005417434,-.1997405113},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.645503189,-2.850095128,-3.679456987,-4.353592220,-4.936509566,1.377214242,2.385405038,3.079544664,3.643766386,-.7456423653,-1.291490461,-1.667307016,-.7214030405e-1,-.1249506719,.8704037330,.2424871131e-9,.42e-9,.5422176684e-9,.6415605973e-9,.7274613394e-9,-.2092513633e-9,-.3624339929e-9,-.4679002728e-9,-.5536270690e-9,.1249332326e-9,.2163907064e-9,.2793592008e-9,-.1348127464e-10,-.2335025263e-10,-.9727958508e-10,-2.424562843,-4.199466029,-5.421487332,-6.414790321,1.641768329,2.843626160,3.671105586,-.5644916125,-.9777281536,-.5417963103,-.7483314774e-9,-.1296148140e-8,-.1673320053e-8,.3917700593e-9,.6785656478e-9,-.9473764217e-11,2.788419758,4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,.9641937137,1.670032500,2.156002687,2.551016781,2.892581141,.2394478538,.4147358487,.5354216783,.6335194733,-.9155822076,-1.585834902,-2.047304055,.5490545322,.9509903459,.3754286701,-.9430172903,-1.633353859,-2.108650765,-2.494989233,-.7446746083,-1.289814257,-1.665143045,.9321620643,1.614552057,.2692350424e-1,-1.398574920,-2.422402819,-3.127308592,-1.821631846,-3.155158911,1.076451798,-.1570858835,-.2720807313,-.2832576764,.8954348170},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,-.9641937137,-1.670032500,-2.156002687,-2.551016781,-2.892581141,-.2394478538,-.4147358487,-.5354216783,-.6335194733,.9155822076,1.585834902,2.047304055,-.5490545322,-.9509903459,-.3754286701,-.9430172892,-1.633353857,-2.108650763,-2.494989230,-.7446746070,-1.289814255,-1.665143042,.9321620636,1.614552056,.2692350421e-1,1.398574927,2.422402832,3.127308609,1.821631855,3.155158927,-1.076451804,-.1570858218,-.2720806245,-.2832575652,-.8954346999},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.113355001,1.928387427,2.489537463,2.945660451,3.340065000,-1.123448021,-1.945869052,-2.512106144,-2.972364074,-1.252806288,-2.169924144,-2.801360023,.9835692099,1.703591844,1.344816036,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.1039230485e-8,.6764731475e-9,.1171685862e-8,.1512639943e-8,.1789779717e-8,.2094450706e-9,.3627695035e-9,.4683334153e-9,-.5392014460e-9,-.9339242999e-9,-.350864418e-9,-.6338914369,-1.097932175,-1.417424343,-1.677119101,-1.926686150,-3.337118302,-4.308201201,-1.956429599,-3.388635469,.2565449507,.7483314774e-8,.1296148140e-7,.1673320053e-7,.3070569490e-7,.5318382366e-7,.4977473171e-7,.1905991076,.3301273382,.9829492714,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,1.926395974,3.336615703,4.307552349,5.096764673,5.779187924,-.6138328492,-1.063189683,-1.372571978,-1.624049066,-.8603944811,-1.490246955,-1.923900547,1.645753812,2.850529219,-1.343411045,.5526193998,.9571648777,1.235694544,1.462093502,.2281916739e-1,.3952395730e-1,.5102520946e-1,-.4132700278,-.7158046857,.4485373028,-1.314274316,-2.276389891,-2.938806712,-.4873895555,-.8441834734,1.215532077,-2.142609188,-3.711107975,-1.473102430,-1.425504294},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,-1.926395976,-3.336615704,-4.307552352,-5.096764678,-5.779187926,.6138328496,1.063189683,1.372571979,1.624049067,.8603944811,1.490246955,1.923900547,-1.645753812,-2.850529220,1.343411045,.5526194031,.9571648834,1.235694551,1.462093511,.2281916752e-1,.3952395754e-1,.5102520976e-1,-.4132700304,-.7158046901,.4485373055,1.314274282,2.276389831,2.938806635,.4873895427,.8441834514,-1.215532046,-2.142609204,-3.711108002,-1.473102441,1.425504133},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224410470,3.852791952,4.973933020,5.885236917,6.673231410,.7950455559,1.377059297,1.777775908,2.103492822,-1.500452022,-2.598859138,-3.355112719,-2.072134551,-3.589042323,-.2431355237,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-0.,-.2398353968,-.4154070926,-.5362882506,-.6345448156,-1.087945327,-1.884376583,-2.432719707,-2.347136393,-4.065359486,-2.942787847,0.,0.,0.,0.,0.,0.,.2728458417e-1,.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,2.679844379,4.641626620,5.992314198,7.090201776,8.039533138,-2.043694514,-3.539782735,-4.569839860,-5.407107442,.8110805448,1.404832712,1.813631233,.6037611818,1.045745042,-1.765469345,2.576012736,4.461784940,5.760139589,6.815489076,-1.403212584,-2.430435489,-3.137678723,-.8058635216e-2,-.1395796564e-1,1.237350351,1.238945990,2.145917403,2.770367454,-.4443853780,-.7696980531,-.2975289326,-.7339728352,-1.271278242,.1365904598,-2.538475875},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,-2.679844380,-4.641626622,-5.992314202,-7.090201781,-8.039533141,2.043694516,3.539782737,4.569839863,5.407107446,-.8110805448,-1.404832712,-1.813631233,-.6037611823,-1.045745043,1.765469346,2.576012735,4.461784937,5.760139587,6.815489073,-1.403212583,-2.430435488,-3.137678722,-.8058635216e-2,-.1395796564e-1,1.237350350,-1.238945998,-2.145917416,-2.770367471,.4443853807,.7696980577,.2975289345,-.7339727817,-1.271278149,.1365904499,2.538476058},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.094417748,5.359688761,6.919328434,8.187059814,9.283253244,3.470033583,6.010274468,7.759230975,9.180845900,2.032368460,3.520165433,4.544514032,-.5354861533,-.9274892243,-2.789841949,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6238553746e-1,-.1080549205,-.1394983026,-.1650566176,-.3561120901,-.6168042332,-.7962908411,-1.076648709,-1.864810267,-2.258512675,.1496662955e-7,.2592296279e-7,.3346640106e-7,.1111356701e-6,.1924926272e-6,.4338629615e-6,.1846114076e-2,.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,3.291626742,5.701264756,7.360301149,8.708825766,9.874880227,-3.696869036,-6.403165001,-8.266450470,-9.780996101,3.754901608,6.503680358,8.396215242,-3.509035267,-6.077827368,3.000239739,4.769620097,8.261224341,10.66519477,12.61922863,-4.867652070,-8.431020700,-10.88440092,4.583518828,7.938887491,-3.968421708,5.735706993,9.934535927,12.82543073,-5.454863311,-9.448100406,4.799476921,6.099313091,10.56432016,-5.461692032,5.821024942},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,-3.291626740,-5.701264753,-7.360301145,-8.708825761,-9.874880222,3.696869033,6.403164996,8.266450463,9.780996094,-3.754901605,-6.503680355,-8.396215237,3.509035265,6.077827365,-3.000239738,4.769620101,8.261224347,10.66519478,12.61922864,-4.867652073,-8.431020705,-10.88440092,4.583518832,7.938887498,-3.968421711,-5.735706944,-9.934535843,-12.82543062,5.454863264,9.448100324,-4.799476880,6.099312923,10.56431987,-5.461691879,-5.821025177},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.800843169,6.583253480,8.498943694,10.05608579,11.40252950,6.399791554,11.08476413,14.31036896,16.93225689,8.680560534,15.03517189,19.41032344,10.21325528,17.68987705,10.69428583,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3017302684e-2,-.5226121551e-2,-.6746893911e-2,-.7983032535e-2,-.2009496124e-1,-.3480549385e-1,-.4493369933e-1,-.7471563177e-1,-.1294112704,-.2049323002,-.7483314774e-8,-.1296148140e-7,-.1673320053e-7,-.6443340372e-7,-.1116019290e-6,-.3031212923e-6,.4333424029e-5,.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,-1.719209449,-2.977758114,-3.844269195,-4.548600653,-5.157628347,-1.016533920,-1.760688397,-2.273038946,-2.689495952,1.730662758,2.997595827,3.869879572,-.8438528152e-1,-.1461595950,-1.608041156,.6684577348,1.157802759,1.494716935,1.768572929,.8304449710,1.438372883,1.856931406,-.4995431840,-.8652341756,-.5315629344,.4697515910,.8136336225,1.050396490,.8765560645,1.518239640,-.5415238988e-1,-.9469058756,-1.640089087,-2.346450809,.7389290768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,.6685019987,1.157879426,1.494815912,1.768690039,2.005505995,1.636951328,2.835282871,3.660334447,4.330966124,1.284112192,2.224147558,2.871362151,-.5257834218,-.9106836003,-1.564527457,-.1324976951,-.2294927397,-.2962738530,-.3505559505,-.4894439054,-.8477417116,-1.094429844,-.7451837832,-1.290696174,-.4186690470,-.2123656599,-.3678281127,-.4748640516,-1.044420364,-1.808989135,-2.287254493,-.6336343026e-1,-.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,1.050707449,1.819878685,2.349453280,2.779910610,3.152122347,-.6204174076,-1.074594472,-1.387295498,-1.641470170,-.2618481958e-1,-.4535343788e-1,-.5855103656e-1,.6101686966,1.056843184,-.8964075294,-1.444837340,-2.502531681,-3.230754509,-3.822680287,.4587476822,.7945742934,1.025791002,.5055009012,.8755532444,-1.086251548,-2.213877417,-3.834548168,-4.950380399,.1678643553,.2907495923,1.305496370,.1849334761e-1,.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,-1.050707450,-1.819878688,-2.349453283,-2.779910614,-3.152122352,.6204174085,1.074594474,1.387295500,1.641470172,.2618481961e-1,.4535343793e-1,.5855103662e-1,-.6101686971,-1.056843185,.8964075306,-1.444837346,-2.502531692,-3.230754521,-3.822680303,.4587476840,.7945742967,1.025791006,.5055009034,.8755532483,-1.086251552,2.213877420,3.834548174,4.950380406,-.1678643556,-.2907495928,-1.305496372,.1849327171e-1,.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,1.719209448,2.977758112,3.844269192,4.548600649,5.157628344,1.016533919,1.760688396,2.273038944,2.689495949,-1.730662756,-2.997595824,-3.869879569,.8438528146e-1,.1461595949,1.608041155,.6684577365,1.157802762,1.494716939,1.768572933,.8304449729,1.438372886,1.856931411,-.4995431851,-.8652341775,-.5315629356,-.4697515776,-.8136335992,-1.050396460,-.8765560394,-1.518239596,.5415238836e-1,-.9469058829,-1.640089099,-2.346450828,-.7389299868},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,-.6685019987,-1.157879426,-1.494815912,-1.768690039,-2.005505995,-1.636951328,-2.835282871,-3.660334447,-4.330966124,-1.284112192,-2.224147558,-2.871362151,.5257834218,.9106836003,1.564527457,-.1324976936,-.2294927372,-.2962738499,-.3505559467,-.4894439003,-.8477417029,-1.094429832,-.7451837755,-1.290696161,-.4186690423,.2123656749,.3678281386,.4748640850,1.044420437,1.808989263,2.287254654,-.6336326377e-1,-.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,-2.155052650,-3.732660680,-4.818844218,-5.701733371,-6.465157947,-1.817524160,-3.148044191,-4.064107574,-4.808716931,2.003812619,3.470705265,4.480661231,.8618882312,1.492834207,-2.361581599,1.978516938,3.426891859,4.424098369,5.234663784,3.116327372,5.397637342,6.968319843,-.7599463103,-1.316265621,-2.583756308,-1.441495348,-2.496743182,-3.223281588,-3.286953601,-5.693170641,-.9156869635,.8622047985,1.493382518,2.564268432,-.3927122230},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,1.100336031,1.905837910,2.460426162,2.911215495,3.301008092,2.979568311,5.160763700,6.662517288,7.883196767,3.041169362,5.267459849,6.800261424,.1577963754,.2733113395,-2.799087386,.4783243874,.8284821415,1.069565846,1.265527375,1.930559283,3.343826766,4.316861791,3.440637089,5.959358251,2.951743964,.1430932586,.2478447941,.3199662533,.7646780256,1.324461192,1.899532545,.2128667917e-1,.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,1.054716618,1.826822770,2.358418055,2.790517875,3.164149855,-1.162044153,-2.012719514,-2.598409719,-3.074479842,1.144752980,1.982770324,2.559745481,-1.019684603,-1.766145540,.8049318120,-1.868089212,-3.235625429,-4.177174467,-4.942499484,1.853832431,3.210931959,4.145295333,-1.671751149,-2.895557928,1.349636370,-2.753730610,-4.769601326,-6.157528834,2.522275578,4.368709454,-2.092761921,.4701430699,.8143116839,-.4012398833,3.559792983},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,-1.054716618,-1.826822769,-2.358418053,-2.790517873,-3.164149853,1.162044152,2.012719513,2.598409717,3.074479840,-1.144752979,-1.982770321,-2.559745479,1.019684602,1.766145538,-.8049318114,-1.868089209,-3.235625422,-4.177174460,-4.942499475,1.853832428,3.210931954,4.145295326,-1.671751146,-2.895557923,1.349636368,2.753730616,4.769601337,6.157528849,-2.522275585,-4.368709465,2.092761926,.4701430553,.8143116587,-.4012398709,-3.559792582},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,2.155052650,3.732660680,4.818844218,5.701733371,6.465157947,1.817524160,3.148044191,4.064107574,4.808716931,-2.003812619,-3.470705265,-4.480661231,-.8618882312,-1.492834207,2.361581599,1.978516937,3.426891859,4.424098367,5.234663782,3.116327372,5.397637341,6.968319842,-.7599463103,-1.316265621,-2.583756308,1.441495367,2.496743214,3.223281630,3.286953644,5.693170715,.9156869755,.8622049336,1.493382752,2.564268834,.3927112853},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,-1.100336031,-1.905837911,-2.460426163,-2.911215496,-3.301008094,-2.979568312,-5.160763703,-6.662517291,-7.883196771,-3.041169365,-5.267459851,-6.800261430,-.1577963755,-.2733113396,2.799087387,.4783243870,.8284821409,1.069565845,1.265527374,1.930559282,3.343826764,4.316861789,3.440637085,5.959358244,2.951743961,-.1430932534,-.2478447850,-.3199662416,-.7646779976,-1.324461144,-1.899532475,.2128671353e-1,.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,-2.784908111,-4.823602340,-6.227243845,-7.368174283,-8.354724330,-.8097724007e-1,-.1402566941,-.1810706135,-.2142456391,2.270993172,3.933475556,5.078095108,-2.198518194,-3.807945213,.1780549380,3.578005823,6.197287874,8.000664244,9.466513600,1.789933617,3.100255967,4.002413242,-3.541924988,-6.134794038,1.442211311,-4.013547435,-6.951668076,-8.974564896,-3.781587883,-6.549902349,3.879680769,4.218875403,7.307306550,5.787244192,-4.275985292},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,.7243065273,1.254535705,1.619598631,1.916334944,2.172919582,2.659465646,4.606329621,5.946745969,7.036284721,4.948721068,8.571436320,11.06567671,5.758806642,9.974545694,3.789231007,.2491916055,.4316125216,.5572093695,.6592990173,1.322791395,2.291141904,2.957851479,3.610515302,6.253595946,6.564693789,.7674948247e-1,.1329340031,.1716170600,.5316983539,.9209285636,1.896167979,.2260074784e-1,.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,2.060601584,3.569066636,4.607645215,5.451839342,6.181804751,-2.578488405,-4.466072926,-5.765675354,-6.822039081,3.067086901,5.312350344,6.858214803,-3.560288453,-6.166600490,4.079521103,.1446504328,.2505418989,.3234482008,.3827090723,-.1721034134,-.2980918563,-.3848349315,.1998400143,.3461330583,-.2290664311,-2.797494156,-4.845402011,-6.255387098,3.249889483,5.628973706,-3.727231982,-3.881338999,-6.722676349,4.454821378,-1.567068824},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,-2.060601584,-3.569066636,-4.607645215,-5.451839342,-6.181804751,2.578488405,4.466072926,5.765675354,6.822039081,-3.067086901,-5.312350344,-6.858214803,3.560288453,6.166600490,-4.079521103,.1446504323,.2505418980,.3234481995,.3827090709,-.1721034127,-.2980918550,-.3848349300,.1998400136,.3461330570,-.2290664303,2.797494172,4.845402039,6.255387135,-3.249889503,-5.628973739,3.727232004,-3.881339015,-6.722676376,4.454821399,1.567068815},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,2.784908111,4.823602342,6.227243845,7.368174283,8.354724334,.8097724011e-1,.1402566942,.1810706136,.2142456393,-2.270993172,-3.933475557,-5.078095109,2.198518194,3.807945213,-.1780549380,3.578005828,6.197287884,8.000664256,9.466513613,1.789933619,3.100255970,4.002413246,-3.541924993,-6.134794047,1.442211314,4.013547495,6.951668180,8.974565029,3.781587939,6.549902446,-3.879680827,4.218875359,7.307306473,5.787244129,4.275985427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,-.7243065266,-1.254535704,-1.619598630,-1.916334942,-2.172919580,-2.659465643,-4.606329616,-5.946745963,-7.036284714,-4.948721064,-8.571436312,-11.06567670,-5.758806636,-9.974545684,-3.789231004,.2491916053,.4316125213,.5572093687,.6592990165,1.322791393,2.291141901,2.957851475,3.610515299,6.253595941,6.564693781,-.7674948090e-1,-.1329340004,-.1716170565,-.5316983431,-.9209285448,-1.896167939,.2260067231e-1,.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,-2.517563337,-4.360547610,-5.629442758,-6.660846498,-7.552690011,.1640586792,.2841579679,.3668463591,.4340584657,1.848945781,3.202468033,4.134368453,-2.129273724,-3.688010274,.6507752466,2.539766587,4.399004769,5.679090736,6.719590780,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207893,-1.820466373,-3.153140251,-4.070686560,-1.433342530,-2.482622088,1.817251596,.7671298784,1.328707925,.9100500654,.2540057989},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,.5106664481,.8845002336,1.141884891,1.351096424,1.531999344,1.778426716,3.080325431,3.976683031,4.705274817,3.022001204,5.234259622,6.757400118,2.956809489,5.121344262,1.089458308,.3202872177e-1,.5547537340e-1,.7161839911e-1,.8474003263e-1,.1620213035,.2806291295,.3622906482,.4123799546,.7142630336,.6762823204,-.2859462130e-1,-.4952733691e-1,-.6393951701e-1,-.1892065416,-.3277153433,-.6347740071,-.1157780971e-1,-.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,2.006896889,3.476047377,4.487557867,5.309750075,6.020690668,-1.942485396,-3.364483399,-4.343529391,-5.139333284,1.514566645,2.623306380,3.386673975,-.8275357615,-1.433333984,.1195045576e-1,.2438275696,.4223217388,.5452150204,.6451071121,-.1988874943,-.3444832452,-.4447259571,.1234754608,.2138657718,-.3202669808e-1,-2.331238944,-4.037824296,-5.212808751,1.622549074,2.810337434,-.7363933866,-3.287037546,-5.693316036,1.939519592,-1.523064210},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,-2.006896889,-3.476047376,-4.487557867,-5.309750075,-6.020690666,1.942485396,3.364483399,4.343529391,5.139333284,-1.514566645,-2.623306379,-3.386673973,.8275357615,1.433333984,-.1195045576e-1,.2438275663,.4223217331,.5452150131,.6451071034,-.1988874916,-.3444832405,-.4447259510,.1234754592,.2138657688,-.3202669764e-1,2.331238936,4.037824281,5.212808732,-1.622549068,-2.810337424,.7363933838,-3.287037517,-5.693315987,1.939519574,1.523064511},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,2.517563337,4.360547611,5.629442758,6.660846498,7.552690013,-.1640586792,-.2841579679,-.3668463591,-.4340584657,-1.848945782,-3.202468033,-4.134368454,2.129273724,3.688010274,-.6507752466,2.539766588,4.399004769,5.679090738,6.719590782,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207894,1.820466382,3.153140266,4.070686580,1.433342537,2.482622099,-1.817251606,.7671297908,1.328707774,.9100499616,-.2540054079},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,-.5106664484,-.8845002342,-1.141884892,-1.351096425,-1.531999345,-1.778426717,-3.080325433,-3.976683034,-4.705274821,-3.022001206,-5.234259626,-6.757400124,-2.956809490,-5.121344265,-1.089458309,.3202872292e-1,.5547537539e-1,.7161840168e-1,.8474003567e-1,.1620213093,.2806291396,.3622906613,.4123799696,.7142630596,.6762823444,.2859462354e-1,.4952734080e-1,.6393952203e-1,.1892065565,.3277153691,.6347740570,-.1157777851e-1,-.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,-3.069477201,-5.316490463,-6.863559675,-8.121073328,-9.208431602,2.156281693,3.734789450,4.821592446,5.704985119,-.5614333340,-.9724310592,-1.255403100,-1.126462044,-1.951089492,2.330362820,4.159176446,7.203904921,9.300201262,11.00414654,-1.935519673,-3.352418413,-4.327953560,-.5514408001,-.9551234835,2.488724225,-4.693254379,-8.128955037,-10.49443583,1.219702470,2.112586649,1.756836938,4.705330903,8.149872191,-.3189574095,-4.265567617},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,.2458344589,.4257977730,.5497025612,.6504168418,.7375033768,1.054396484,1.826268282,2.357702214,2.789670881,2.510226735,4.347840243,5.613037617,4.324472530,7.490206138,5.881975759,.1420874210e-1,.2461026322e-1,.3177171321e-1,.3759279804e-1,.8700209506e-1,.1506920490,.1945425987,.2894494276,.5013411150,.6894088761,-.1216128451e-2,-.2106396265e-2,-.2719345885e-2,-.9658561257e-2,-.1672911883e-1,-.4108123658e-1,-.3746153966e-3,-.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,2.823642741,4.890692689,6.313857110,7.470656481,8.470928224,-3.210678176,-5.561057729,-7.179294657,-8.494655996,3.324354634,5.757951125,7.433482942,-3.198010479,-5.539116631,2.859723954,2.797956434,4.846202701,6.256420786,7.402696906,-2.907006833,-5.035083533,-6.500264888,2.811341931,4.869387063,-2.534878984,1.242384808,2.151873609,2.778056884,-1.210043957,-2.095857614,1.103053355,-1.284643645,-2.225068063,1.185627028,-3.765218080},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,-2.823642742,-4.890692690,-6.313857114,-7.470656486,-8.470928225,3.210678177,5.561057732,7.179294660,8.494656000,-3.324354634,-5.757951127,-7.433482942,3.198010480,5.539116634,-2.859723955,2.797956438,4.846202708,6.256420796,7.402696917,-2.907006837,-5.035083540,-6.500264898,2.811341934,4.869387068,-2.534878988,-1.242384788,-2.151873575,-2.778056840,1.210043938,2.095857580,-1.103053339,-1.284643653,-2.225068076,1.185627035,3.765217145},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,3.069477200,5.316490462,6.863559672,8.121073323,9.208431600,-2.156281693,-3.734789450,-4.821592446,-5.704985119,.5614333335,.9724310584,1.255403098,1.126462044,1.951089492,-2.330362819,4.159176444,7.203904918,9.300201260,11.00414653,-1.935519673,-3.352418412,-4.327953559,-.5514408000,-.9551234833,2.488724224,4.693254424,8.128955115,10.49443593,-1.219702482,-2.112586669,-1.756836954,4.705330679,8.149871802,-.3189573943,4.265566870},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,-.2458344586,-.4257977724,-.5497025604,-.6504168409,-.7375033757,-1.054396483,-1.826268280,-2.357702211,-2.789670877,-2.510226732,-4.347840236,-5.613037612,-4.324472524,-7.490206128,-5.881975750,.1420874260e-1,.2461026409e-1,.3177171433e-1,.3759279937e-1,.8700209816e-1,.1506920544,.1945426056,.2894494379,.5013411328,.6894089005,.1216122464e-2,.2106385895e-2,.2719332498e-2,.9658513710e-2,.1672903648e-1,.4108103434e-1,-.3745694178e-3,-.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,0.,0.,0.,0.,0.,-1.360827635,-2.357022604,-3.042903097,-3.600411499,.8380524812,1.451549477,1.873942317,.8198497638,1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-2.721655270,-3.220305944,-1.873942317,-3.245763304,-4.190262406,.5237828008,.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,1.217161239,2.052800957,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,-1.704828602,-2.952849755,-3.812112642,-4.510552507,-5.114485804,-3.104623248,-5.377365205,-6.942148628,-8.214061030,-.5462559154,-.9461429991,-1.221465360,2.694797933,4.667526936,1.022426944,1.298058152,2.248302670,2.902546266,3.434339057,3.719687555,6.442687835,8.317474224,3.264192666,5.653747546,-1.221656347,-.8344196900,-1.445257298,-1.865819148,-3.243190051,-5.617369949,-4.760960496,.4850599026,.8401483961,2.375793728,-.2614753427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,1.704828601,2.952849754,3.812112640,4.510552505,5.114485802,3.104623246,5.377365203,6.942148625,8.214061027,.5462559149,.9461429982,1.221465359,-2.694797932,-4.667526936,-1.022426944,1.298058156,2.248302677,2.902546276,3.434339069,3.719687568,6.442687856,8.317474253,3.264192678,5.653747565,-1.221656351,.8344196802,1.445257281,1.865819127,3.243190014,5.617369884,4.760960437,.4850600816,.8401487061,2.375794604,.2614758325},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.968566503,-3.409657202,-4.401848518,-5.208337406,-5.905699510,2.347500666,4.065990423,5.249171065,6.210902963,-2.609637841,-4.520025331,-5.835327609,2.775506203,4.807317761,-2.855122567,-.3360178568e-9,-.582e-9,-.7513587691e-9,-.8890196848e-9,-.1008053570e-8,.4007561299e-9,.6941299786e-9,.8961179490e-9,.1060301056e-8,-.4456004436e-9,-.7718026080e-9,-.9963928825e-9,.4740581912e-9,.8210928729e-9,-.4878407244e-9,-2.709993221,-4.693845947,-6.059729062,-7.169968120,3.014474194,5.221222463,6.740569213,-3.208782240,-5.557773872,3.304507380,-.7333648479e-8,-.1270225177e-7,-.1639853652e-7,.7812670753e-8,.1353194269e-7,-.8054303853e-8,3.483595229,6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,-.8327988349,-1.442449894,-1.862194806,-2.203378609,-2.498396504,-1.088657954,-1.885610890,-2.434313191,-2.880318211,.4375794197,.7579097871,.9784573278,.9743405328,1.687607307,-.5933186699,-.4672884697,-.8093673712,-1.044888783,-1.236329082,-1.022115558,-1.770356077,-2.285519867,-.3251247057,-.5631325093,.8276510476,.6684136118,1.157726336,1.494618273,2.033596264,3.522292053,1.746351412,-.1476726503,-.2557765332,-.5739768229,-.2050386800},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,.8327988342,1.442449893,1.862194804,2.203378607,2.498396503,1.088657953,1.885610888,2.434313189,2.880318209,-.4375794194,-.7579097864,-.9784573273,-.9743405322,-1.687607306,.5933186693,-.4672884647,-.8093673627,-1.044888772,-1.236329069,-1.022115547,-1.770356058,-2.285519843,-.3251247023,-.5631325034,.8276510391,-.6684136118,-1.157726336,-1.494618273,-2.033596264,-3.522292053,-1.746351412,-.1476726019,-.2557764494,-.5739766349,.2050384112},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.961633263,-1.665597669,-2.150277344,-2.544242465,-2.884899788,-.2081759125,-.3605712573,-.4654954917,-.5507816934,1.215013765,2.104465573,2.716853372,-1.352704319,-2.342952608,.5364099445,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.1039230485e-8,.5714598801e-10,.9897975472e-10,.1277823139e-9,.1511940727e-9,.2193666394e-9,.3799541647e-9,.4905187176e-9,-.3033617129e-9,-.5254378997e-9,.1563891368e-9,-1.872739842,-3.243680556,-4.187573592,-4.954803895,-.4572029592,-.7918987547,-1.022336896,1.670078386,2.892660618,-1.283464460,0.,0.,0.,0.,0.,-0.,1.663589468,2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,-1.425047564,-2.468254783,-3.186503223,-3.770321460,-4.275142691,-2.360183599,-4.087957910,-5.277530967,-6.244458853,-.4945548803e-2,-.8565941795e-2,-.1105858331e-1,2.175490654,3.768060344,.1537427033,.6224210815,1.078064937,1.391775849,1.646771393,1.648153002,2.854684738,3.685382148,1.157573982,2.004976951,-.8770258842,-.7554460893e-2,-.1308471009e-1,-.1689228809e-1,-.2731590581e-1,-.4731253672e-1,-.3482606398e-1,-.2198338910,-.3807634685,-1.005417380,.1997404621},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,1.425047564,2.468254784,3.186503224,3.770321460,4.275142693,2.360183599,4.087957910,5.277530967,6.244458853,.4945548803e-2,.8565941795e-2,.1105858331e-1,-2.175490654,-3.768060344,-.1537427033,.6224210804,1.078064935,1.391775847,1.646771390,1.648152999,2.854684733,3.685382143,1.157573980,2.004976947,-.8770258827,.7554451164e-2,.1308469324e-1,.1689226633e-1,.2731587063e-1,.4731247580e-1,.3482601912e-1,-.2198339029,-.3807634890,-1.005417434,-.1997405113},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.645503189,-2.850095128,-3.679456987,-4.353592220,-4.936509566,1.377214242,2.385405038,3.079544664,3.643766386,-.7456423653,-1.291490461,-1.667307016,-.7214030405e-1,-.1249506719,.8704037330,.2424871131e-9,.42e-9,.5422176684e-9,.6415605973e-9,.7274613394e-9,-.2092513633e-9,-.3624339929e-9,-.4679002728e-9,-.5536270690e-9,.1249332326e-9,.2163907064e-9,.2793592008e-9,-.1348127464e-10,-.2335025263e-10,-.9727958508e-10,-2.424562843,-4.199466029,-5.421487332,-6.414790321,1.641768329,2.843626160,3.671105586,-.5644916125,-.9777281536,-.5417963103,-.7483314774e-9,-.1296148140e-8,-.1673320053e-8,.3917700593e-9,.6785656478e-9,-.9473764217e-11,2.788419758,4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,.9641937137,1.670032500,2.156002687,2.551016781,2.892581141,.2394478538,.4147358487,.5354216783,.6335194733,-.9155822076,-1.585834902,-2.047304055,.5490545322,.9509903459,.3754286701,-.9430172903,-1.633353859,-2.108650765,-2.494989233,-.7446746083,-1.289814257,-1.665143045,.9321620643,1.614552057,.2692350424e-1,-1.398574920,-2.422402819,-3.127308592,-1.821631846,-3.155158911,1.076451798,-.1570858835,-.2720807313,-.2832576764,.8954348170},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,-.9641937137,-1.670032500,-2.156002687,-2.551016781,-2.892581141,-.2394478538,-.4147358487,-.5354216783,-.6335194733,.9155822076,1.585834902,2.047304055,-.5490545322,-.9509903459,-.3754286701,-.9430172892,-1.633353857,-2.108650763,-2.494989230,-.7446746070,-1.289814255,-1.665143042,.9321620636,1.614552056,.2692350421e-1,1.398574927,2.422402832,3.127308609,1.821631855,3.155158927,-1.076451804,-.1570858218,-.2720806245,-.2832575652,-.8954346999},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.113355001,1.928387427,2.489537463,2.945660451,3.340065000,-1.123448021,-1.945869052,-2.512106144,-2.972364074,-1.252806288,-2.169924144,-2.801360023,.9835692099,1.703591844,1.344816036,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.1039230485e-8,.6764731475e-9,.1171685862e-8,.1512639943e-8,.1789779717e-8,.2094450706e-9,.3627695035e-9,.4683334153e-9,-.5392014460e-9,-.9339242999e-9,-.350864418e-9,-.6338914369,-1.097932175,-1.417424343,-1.677119101,-1.926686150,-3.337118302,-4.308201201,-1.956429599,-3.388635469,.2565449507,.7483314774e-8,.1296148140e-7,.1673320053e-7,.3070569490e-7,.5318382366e-7,.4977473171e-7,.1905991076,.3301273382,.9829492714,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,1.926395974,3.336615703,4.307552349,5.096764673,5.779187924,-.6138328492,-1.063189683,-1.372571978,-1.624049066,-.8603944811,-1.490246955,-1.923900547,1.645753812,2.850529219,-1.343411045,.5526193998,.9571648777,1.235694544,1.462093502,.2281916739e-1,.3952395730e-1,.5102520946e-1,-.4132700278,-.7158046857,.4485373028,-1.314274316,-2.276389891,-2.938806712,-.4873895555,-.8441834734,1.215532077,-2.142609188,-3.711107975,-1.473102430,-1.425504294},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,-1.926395976,-3.336615704,-4.307552352,-5.096764678,-5.779187926,.6138328496,1.063189683,1.372571979,1.624049067,.8603944811,1.490246955,1.923900547,-1.645753812,-2.850529220,1.343411045,.5526194031,.9571648834,1.235694551,1.462093511,.2281916752e-1,.3952395754e-1,.5102520976e-1,-.4132700304,-.7158046901,.4485373055,1.314274282,2.276389831,2.938806635,.4873895427,.8441834514,-1.215532046,-2.142609204,-3.711108002,-1.473102441,1.425504133},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224410470,3.852791952,4.973933020,5.885236917,6.673231410,.7950455559,1.377059297,1.777775908,2.103492822,-1.500452022,-2.598859138,-3.355112719,-2.072134551,-3.589042323,-.2431355237,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-0.,-.2398353968,-.4154070926,-.5362882506,-.6345448156,-1.087945327,-1.884376583,-2.432719707,-2.347136393,-4.065359486,-2.942787847,0.,0.,0.,0.,0.,0.,.2728458417e-1,.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,2.679844379,4.641626620,5.992314198,7.090201776,8.039533138,-2.043694514,-3.539782735,-4.569839860,-5.407107442,.8110805448,1.404832712,1.813631233,.6037611818,1.045745042,-1.765469345,2.576012736,4.461784940,5.760139589,6.815489076,-1.403212584,-2.430435489,-3.137678723,-.8058635216e-2,-.1395796564e-1,1.237350351,1.238945990,2.145917403,2.770367454,-.4443853780,-.7696980531,-.2975289326,-.7339728352,-1.271278242,.1365904598,-2.538475875},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,-2.679844380,-4.641626622,-5.992314202,-7.090201781,-8.039533141,2.043694516,3.539782737,4.569839863,5.407107446,-.8110805448,-1.404832712,-1.813631233,-.6037611823,-1.045745043,1.765469346,2.576012735,4.461784937,5.760139587,6.815489073,-1.403212583,-2.430435488,-3.137678722,-.8058635216e-2,-.1395796564e-1,1.237350350,-1.238945998,-2.145917416,-2.770367471,.4443853807,.7696980577,.2975289345,-.7339727817,-1.271278149,.1365904499,2.538476058},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.094417748,5.359688761,6.919328434,8.187059814,9.283253244,3.470033583,6.010274468,7.759230975,9.180845900,2.032368460,3.520165433,4.544514032,-.5354861533,-.9274892243,-2.789841949,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6238553746e-1,-.1080549205,-.1394983026,-.1650566176,-.3561120901,-.6168042332,-.7962908411,-1.076648709,-1.864810267,-2.258512675,.1496662955e-7,.2592296279e-7,.3346640106e-7,.1111356701e-6,.1924926272e-6,.4338629615e-6,.1846114076e-2,.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,3.291626742,5.701264756,7.360301149,8.708825766,9.874880227,-3.696869036,-6.403165001,-8.266450470,-9.780996101,3.754901608,6.503680358,8.396215242,-3.509035267,-6.077827368,3.000239739,4.769620097,8.261224341,10.66519477,12.61922863,-4.867652070,-8.431020700,-10.88440092,4.583518828,7.938887491,-3.968421708,5.735706993,9.934535927,12.82543073,-5.454863311,-9.448100406,4.799476921,6.099313091,10.56432016,-5.461692032,5.821024942},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,-3.291626740,-5.701264753,-7.360301145,-8.708825761,-9.874880222,3.696869033,6.403164996,8.266450463,9.780996094,-3.754901605,-6.503680355,-8.396215237,3.509035265,6.077827365,-3.000239738,4.769620101,8.261224347,10.66519478,12.61922864,-4.867652073,-8.431020705,-10.88440092,4.583518832,7.938887498,-3.968421711,-5.735706944,-9.934535843,-12.82543062,5.454863264,9.448100324,-4.799476880,6.099312923,10.56431987,-5.461691879,-5.821025177},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.800843169,6.583253480,8.498943694,10.05608579,11.40252950,6.399791554,11.08476413,14.31036896,16.93225689,8.680560534,15.03517189,19.41032344,10.21325528,17.68987705,10.69428583,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3017302684e-2,-.5226121551e-2,-.6746893911e-2,-.7983032535e-2,-.2009496124e-1,-.3480549385e-1,-.4493369933e-1,-.7471563177e-1,-.1294112704,-.2049323002,-.7483314774e-8,-.1296148140e-7,-.1673320053e-7,-.6443340372e-7,-.1116019290e-6,-.3031212923e-6,.4333424029e-5,.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,-1.719209449,-2.977758114,-3.844269195,-4.548600653,-5.157628347,-1.016533920,-1.760688397,-2.273038946,-2.689495952,1.730662758,2.997595827,3.869879572,-.8438528152e-1,-.1461595950,-1.608041156,.6684577348,1.157802759,1.494716935,1.768572929,.8304449710,1.438372883,1.856931406,-.4995431840,-.8652341756,-.5315629344,.4697515910,.8136336225,1.050396490,.8765560645,1.518239640,-.5415238988e-1,-.9469058756,-1.640089087,-2.346450809,.7389290768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,.6685019987,1.157879426,1.494815912,1.768690039,2.005505995,1.636951328,2.835282871,3.660334447,4.330966124,1.284112192,2.224147558,2.871362151,-.5257834218,-.9106836003,-1.564527457,-.1324976951,-.2294927397,-.2962738530,-.3505559505,-.4894439054,-.8477417116,-1.094429844,-.7451837832,-1.290696174,-.4186690470,-.2123656599,-.3678281127,-.4748640516,-1.044420364,-1.808989135,-2.287254493,-.6336343026e-1,-.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,1.050707449,1.819878685,2.349453280,2.779910610,3.152122347,-.6204174076,-1.074594472,-1.387295498,-1.641470170,-.2618481958e-1,-.4535343788e-1,-.5855103656e-1,.6101686966,1.056843184,-.8964075294,-1.444837340,-2.502531681,-3.230754509,-3.822680287,.4587476822,.7945742934,1.025791002,.5055009012,.8755532444,-1.086251548,-2.213877417,-3.834548168,-4.950380399,.1678643553,.2907495923,1.305496370,.1849334761e-1,.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,-1.050707450,-1.819878688,-2.349453283,-2.779910614,-3.152122352,.6204174085,1.074594474,1.387295500,1.641470172,.2618481961e-1,.4535343793e-1,.5855103662e-1,-.6101686971,-1.056843185,.8964075306,-1.444837346,-2.502531692,-3.230754521,-3.822680303,.4587476840,.7945742967,1.025791006,.5055009034,.8755532483,-1.086251552,2.213877420,3.834548174,4.950380406,-.1678643556,-.2907495928,-1.305496372,.1849327171e-1,.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,1.719209448,2.977758112,3.844269192,4.548600649,5.157628344,1.016533919,1.760688396,2.273038944,2.689495949,-1.730662756,-2.997595824,-3.869879569,.8438528146e-1,.1461595949,1.608041155,.6684577365,1.157802762,1.494716939,1.768572933,.8304449729,1.438372886,1.856931411,-.4995431851,-.8652341775,-.5315629356,-.4697515776,-.8136335992,-1.050396460,-.8765560394,-1.518239596,.5415238836e-1,-.9469058829,-1.640089099,-2.346450828,-.7389299868},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,-.6685019987,-1.157879426,-1.494815912,-1.768690039,-2.005505995,-1.636951328,-2.835282871,-3.660334447,-4.330966124,-1.284112192,-2.224147558,-2.871362151,.5257834218,.9106836003,1.564527457,-.1324976936,-.2294927372,-.2962738499,-.3505559467,-.4894439003,-.8477417029,-1.094429832,-.7451837755,-1.290696161,-.4186690423,.2123656749,.3678281386,.4748640850,1.044420437,1.808989263,2.287254654,-.6336326377e-1,-.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,-2.155052650,-3.732660680,-4.818844218,-5.701733371,-6.465157947,-1.817524160,-3.148044191,-4.064107574,-4.808716931,2.003812619,3.470705265,4.480661231,.8618882312,1.492834207,-2.361581599,1.978516938,3.426891859,4.424098369,5.234663784,3.116327372,5.397637342,6.968319843,-.7599463103,-1.316265621,-2.583756308,-1.441495348,-2.496743182,-3.223281588,-3.286953601,-5.693170641,-.9156869635,.8622047985,1.493382518,2.564268432,-.3927122230},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,1.100336031,1.905837910,2.460426162,2.911215495,3.301008092,2.979568311,5.160763700,6.662517288,7.883196767,3.041169362,5.267459849,6.800261424,.1577963754,.2733113395,-2.799087386,.4783243874,.8284821415,1.069565846,1.265527375,1.930559283,3.343826766,4.316861791,3.440637089,5.959358251,2.951743964,.1430932586,.2478447941,.3199662533,.7646780256,1.324461192,1.899532545,.2128667917e-1,.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,1.054716618,1.826822770,2.358418055,2.790517875,3.164149855,-1.162044153,-2.012719514,-2.598409719,-3.074479842,1.144752980,1.982770324,2.559745481,-1.019684603,-1.766145540,.8049318120,-1.868089212,-3.235625429,-4.177174467,-4.942499484,1.853832431,3.210931959,4.145295333,-1.671751149,-2.895557928,1.349636370,-2.753730610,-4.769601326,-6.157528834,2.522275578,4.368709454,-2.092761921,.4701430699,.8143116839,-.4012398833,3.559792983},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,-1.054716618,-1.826822769,-2.358418053,-2.790517873,-3.164149853,1.162044152,2.012719513,2.598409717,3.074479840,-1.144752979,-1.982770321,-2.559745479,1.019684602,1.766145538,-.8049318114,-1.868089209,-3.235625422,-4.177174460,-4.942499475,1.853832428,3.210931954,4.145295326,-1.671751146,-2.895557923,1.349636368,2.753730616,4.769601337,6.157528849,-2.522275585,-4.368709465,2.092761926,.4701430553,.8143116587,-.4012398709,-3.559792582},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,2.155052650,3.732660680,4.818844218,5.701733371,6.465157947,1.817524160,3.148044191,4.064107574,4.808716931,-2.003812619,-3.470705265,-4.480661231,-.8618882312,-1.492834207,2.361581599,1.978516937,3.426891859,4.424098367,5.234663782,3.116327372,5.397637341,6.968319842,-.7599463103,-1.316265621,-2.583756308,1.441495367,2.496743214,3.223281630,3.286953644,5.693170715,.9156869755,.8622049336,1.493382752,2.564268834,.3927112853},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,-1.100336031,-1.905837911,-2.460426163,-2.911215496,-3.301008094,-2.979568312,-5.160763703,-6.662517291,-7.883196771,-3.041169365,-5.267459851,-6.800261430,-.1577963755,-.2733113396,2.799087387,.4783243870,.8284821409,1.069565845,1.265527374,1.930559282,3.343826764,4.316861789,3.440637085,5.959358244,2.951743961,-.1430932534,-.2478447850,-.3199662416,-.7646779976,-1.324461144,-1.899532475,.2128671353e-1,.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,-2.784908111,-4.823602340,-6.227243845,-7.368174283,-8.354724330,-.8097724007e-1,-.1402566941,-.1810706135,-.2142456391,2.270993172,3.933475556,5.078095108,-2.198518194,-3.807945213,.1780549380,3.578005823,6.197287874,8.000664244,9.466513600,1.789933617,3.100255967,4.002413242,-3.541924988,-6.134794038,1.442211311,-4.013547435,-6.951668076,-8.974564896,-3.781587883,-6.549902349,3.879680769,4.218875403,7.307306550,5.787244192,-4.275985292},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,.7243065273,1.254535705,1.619598631,1.916334944,2.172919582,2.659465646,4.606329621,5.946745969,7.036284721,4.948721068,8.571436320,11.06567671,5.758806642,9.974545694,3.789231007,.2491916055,.4316125216,.5572093695,.6592990173,1.322791395,2.291141904,2.957851479,3.610515302,6.253595946,6.564693789,.7674948247e-1,.1329340031,.1716170600,.5316983539,.9209285636,1.896167979,.2260074784e-1,.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,2.060601584,3.569066636,4.607645215,5.451839342,6.181804751,-2.578488405,-4.466072926,-5.765675354,-6.822039081,3.067086901,5.312350344,6.858214803,-3.560288453,-6.166600490,4.079521103,.1446504328,.2505418989,.3234482008,.3827090723,-.1721034134,-.2980918563,-.3848349315,.1998400143,.3461330583,-.2290664311,-2.797494156,-4.845402011,-6.255387098,3.249889483,5.628973706,-3.727231982,-3.881338999,-6.722676349,4.454821378,-1.567068824},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,-2.060601584,-3.569066636,-4.607645215,-5.451839342,-6.181804751,2.578488405,4.466072926,5.765675354,6.822039081,-3.067086901,-5.312350344,-6.858214803,3.560288453,6.166600490,-4.079521103,.1446504323,.2505418980,.3234481995,.3827090709,-.1721034127,-.2980918550,-.3848349300,.1998400136,.3461330570,-.2290664303,2.797494172,4.845402039,6.255387135,-3.249889503,-5.628973739,3.727232004,-3.881339015,-6.722676376,4.454821399,1.567068815},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,2.784908111,4.823602342,6.227243845,7.368174283,8.354724334,.8097724011e-1,.1402566942,.1810706136,.2142456393,-2.270993172,-3.933475557,-5.078095109,2.198518194,3.807945213,-.1780549380,3.578005828,6.197287884,8.000664256,9.466513613,1.789933619,3.100255970,4.002413246,-3.541924993,-6.134794047,1.442211314,4.013547495,6.951668180,8.974565029,3.781587939,6.549902446,-3.879680827,4.218875359,7.307306473,5.787244129,4.275985427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,-.7243065266,-1.254535704,-1.619598630,-1.916334942,-2.172919580,-2.659465643,-4.606329616,-5.946745963,-7.036284714,-4.948721064,-8.571436312,-11.06567670,-5.758806636,-9.974545684,-3.789231004,.2491916053,.4316125213,.5572093687,.6592990165,1.322791393,2.291141901,2.957851475,3.610515299,6.253595941,6.564693781,-.7674948090e-1,-.1329340004,-.1716170565,-.5316983431,-.9209285448,-1.896167939,.2260067231e-1,.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,-2.517563337,-4.360547610,-5.629442758,-6.660846498,-7.552690011,.1640586792,.2841579679,.3668463591,.4340584657,1.848945781,3.202468033,4.134368453,-2.129273724,-3.688010274,.6507752466,2.539766587,4.399004769,5.679090736,6.719590780,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207893,-1.820466373,-3.153140251,-4.070686560,-1.433342530,-2.482622088,1.817251596,.7671298784,1.328707925,.9100500654,.2540057989},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,.5106664481,.8845002336,1.141884891,1.351096424,1.531999344,1.778426716,3.080325431,3.976683031,4.705274817,3.022001204,5.234259622,6.757400118,2.956809489,5.121344262,1.089458308,.3202872177e-1,.5547537340e-1,.7161839911e-1,.8474003263e-1,.1620213035,.2806291295,.3622906482,.4123799546,.7142630336,.6762823204,-.2859462130e-1,-.4952733691e-1,-.6393951701e-1,-.1892065416,-.3277153433,-.6347740071,-.1157780971e-1,-.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,2.006896889,3.476047377,4.487557867,5.309750075,6.020690668,-1.942485396,-3.364483399,-4.343529391,-5.139333284,1.514566645,2.623306380,3.386673975,-.8275357615,-1.433333984,.1195045576e-1,.2438275696,.4223217388,.5452150204,.6451071121,-.1988874943,-.3444832452,-.4447259571,.1234754608,.2138657718,-.3202669808e-1,-2.331238944,-4.037824296,-5.212808751,1.622549074,2.810337434,-.7363933866,-3.287037546,-5.693316036,1.939519592,-1.523064210},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,-2.006896889,-3.476047376,-4.487557867,-5.309750075,-6.020690666,1.942485396,3.364483399,4.343529391,5.139333284,-1.514566645,-2.623306379,-3.386673973,.8275357615,1.433333984,-.1195045576e-1,.2438275663,.4223217331,.5452150131,.6451071034,-.1988874916,-.3444832405,-.4447259510,.1234754592,.2138657688,-.3202669764e-1,2.331238936,4.037824281,5.212808732,-1.622549068,-2.810337424,.7363933838,-3.287037517,-5.693315987,1.939519574,1.523064511},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,2.517563337,4.360547611,5.629442758,6.660846498,7.552690013,-.1640586792,-.2841579679,-.3668463591,-.4340584657,-1.848945782,-3.202468033,-4.134368454,2.129273724,3.688010274,-.6507752466,2.539766588,4.399004769,5.679090738,6.719590782,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207894,1.820466382,3.153140266,4.070686580,1.433342537,2.482622099,-1.817251606,.7671297908,1.328707774,.9100499616,-.2540054079},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,-.5106664484,-.8845002342,-1.141884892,-1.351096425,-1.531999345,-1.778426717,-3.080325433,-3.976683034,-4.705274821,-3.022001206,-5.234259626,-6.757400124,-2.956809490,-5.121344265,-1.089458309,.3202872292e-1,.5547537539e-1,.7161840168e-1,.8474003567e-1,.1620213093,.2806291396,.3622906613,.4123799696,.7142630596,.6762823444,.2859462354e-1,.4952734080e-1,.6393952203e-1,.1892065565,.3277153691,.6347740570,-.1157777851e-1,-.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,-3.069477201,-5.316490463,-6.863559675,-8.121073328,-9.208431602,2.156281693,3.734789450,4.821592446,5.704985119,-.5614333340,-.9724310592,-1.255403100,-1.126462044,-1.951089492,2.330362820,4.159176446,7.203904921,9.300201262,11.00414654,-1.935519673,-3.352418413,-4.327953560,-.5514408001,-.9551234835,2.488724225,-4.693254379,-8.128955037,-10.49443583,1.219702470,2.112586649,1.756836938,4.705330903,8.149872191,-.3189574095,-4.265567617},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,.2458344589,.4257977730,.5497025612,.6504168418,.7375033768,1.054396484,1.826268282,2.357702214,2.789670881,2.510226735,4.347840243,5.613037617,4.324472530,7.490206138,5.881975759,.1420874210e-1,.2461026322e-1,.3177171321e-1,.3759279804e-1,.8700209506e-1,.1506920490,.1945425987,.2894494276,.5013411150,.6894088761,-.1216128451e-2,-.2106396265e-2,-.2719345885e-2,-.9658561257e-2,-.1672911883e-1,-.4108123658e-1,-.3746153966e-3,-.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,2.823642741,4.890692689,6.313857110,7.470656481,8.470928224,-3.210678176,-5.561057729,-7.179294657,-8.494655996,3.324354634,5.757951125,7.433482942,-3.198010479,-5.539116631,2.859723954,2.797956434,4.846202701,6.256420786,7.402696906,-2.907006833,-5.035083533,-6.500264888,2.811341931,4.869387063,-2.534878984,1.242384808,2.151873609,2.778056884,-1.210043957,-2.095857614,1.103053355,-1.284643645,-2.225068063,1.185627028,-3.765218080},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,-2.823642742,-4.890692690,-6.313857114,-7.470656486,-8.470928225,3.210678177,5.561057732,7.179294660,8.494656000,-3.324354634,-5.757951127,-7.433482942,3.198010480,5.539116634,-2.859723955,2.797956438,4.846202708,6.256420796,7.402696917,-2.907006837,-5.035083540,-6.500264898,2.811341934,4.869387068,-2.534878988,-1.242384788,-2.151873575,-2.778056840,1.210043938,2.095857580,-1.103053339,-1.284643653,-2.225068076,1.185627035,3.765217145},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,3.069477200,5.316490462,6.863559672,8.121073323,9.208431600,-2.156281693,-3.734789450,-4.821592446,-5.704985119,.5614333335,.9724310584,1.255403098,1.126462044,1.951089492,-2.330362819,4.159176444,7.203904918,9.300201260,11.00414653,-1.935519673,-3.352418412,-4.327953559,-.5514408000,-.9551234833,2.488724224,4.693254424,8.128955115,10.49443593,-1.219702482,-2.112586669,-1.756836954,4.705330679,8.149871802,-.3189573943,4.265566870},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,-.2458344586,-.4257977724,-.5497025604,-.6504168409,-.7375033757,-1.054396483,-1.826268280,-2.357702211,-2.789670877,-2.510226732,-4.347840236,-5.613037612,-4.324472524,-7.490206128,-5.881975750,.1420874260e-1,.2461026409e-1,.3177171433e-1,.3759279937e-1,.8700209816e-1,.1506920544,.1945426056,.2894494379,.5013411328,.6894089005,.1216122464e-2,.2106385895e-2,.2719332498e-2,.9658513710e-2,.1672903648e-1,.4108103434e-1,-.3745694178e-3,-.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,0.,0.,0.,0.,0.,-1.360827635,2.357022604,-3.042903097,3.600411499,.8380524812,-1.451549477,1.873942317,.8198497638,-1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-2.721655270,3.220305944,-1.873942317,3.245763304,-4.190262406,.5237828008,-.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,-1.217161239,2.052800957,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,-1.704828602,2.952849755,-3.812112642,4.510552507,-5.114485804,-3.104623248,5.377365205,-6.942148628,8.214061030,-.5462559154,.9461429991,-1.221465360,2.694797933,-4.667526936,1.022426944,1.298058152,-2.248302670,2.902546266,-3.434339057,3.719687555,-6.442687835,8.317474224,3.264192666,-5.653747546,-1.221656347,-.8344196900,1.445257298,-1.865819148,-3.243190051,5.617369949,-4.760960496,.4850599026,-.8401483961,2.375793728,-.2614753427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,1.704828601,-2.952849754,3.812112640,-4.510552505,5.114485802,3.104623246,-5.377365203,6.942148625,-8.214061027,.5462559149,-.9461429982,1.221465359,-2.694797932,4.667526936,-1.022426944,1.298058156,-2.248302677,2.902546276,-3.434339069,3.719687568,-6.442687856,8.317474253,3.264192678,-5.653747565,-1.221656351,.8344196802,-1.445257281,1.865819127,3.243190014,-5.617369884,4.760960437,.4850600816,-.8401487061,2.375794604,.2614758325},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.968566503,3.409657202,-4.401848518,5.208337406,-5.905699510,2.347500666,-4.065990423,5.249171065,-6.210902963,-2.609637841,4.520025331,-5.835327609,2.775506203,-4.807317761,-2.855122567,-.3360178568e-9,.582e-9,-.7513587691e-9,.8890196848e-9,-.1008053570e-8,.4007561299e-9,-.6941299786e-9,.8961179490e-9,-.1060301056e-8,-.4456004436e-9,.7718026080e-9,-.9963928825e-9,.4740581912e-9,-.8210928729e-9,-.4878407244e-9,-2.709993221,4.693845947,-6.059729062,7.169968120,3.014474194,-5.221222463,6.740569213,-3.208782240,5.557773872,3.304507380,-.7333648479e-8,.1270225177e-7,-.1639853652e-7,.7812670753e-8,-.1353194269e-7,-.8054303853e-8,3.483595229,-6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,-.8327988349,1.442449894,-1.862194806,2.203378609,-2.498396504,-1.088657954,1.885610890,-2.434313191,2.880318211,.4375794197,-.7579097871,.9784573278,.9743405328,-1.687607307,-.5933186699,-.4672884697,.8093673712,-1.044888783,1.236329082,-1.022115558,1.770356077,-2.285519867,-.3251247057,.5631325093,.8276510476,.6684136118,-1.157726336,1.494618273,2.033596264,-3.522292053,1.746351412,-.1476726503,.2557765332,-.5739768229,-.2050386800},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,.8327988342,-1.442449893,1.862194804,-2.203378607,2.498396503,1.088657953,-1.885610888,2.434313189,-2.880318209,-.4375794194,.7579097864,-.9784573273,-.9743405322,1.687607306,.5933186693,-.4672884647,.8093673627,-1.044888772,1.236329069,-1.022115547,1.770356058,-2.285519843,-.3251247023,.5631325034,.8276510391,-.6684136118,1.157726336,-1.494618273,-2.033596264,3.522292053,-1.746351412,-.1476726019,.2557764494,-.5739766349,.2050384112},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.961633263,1.665597669,-2.150277344,2.544242465,-2.884899788,-.2081759125,.3605712573,-.4654954917,.5507816934,1.215013765,-2.104465573,2.716853372,-1.352704319,2.342952608,.5364099445,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.1039230485e-8,.5714598801e-10,-.9897975472e-10,.1277823139e-9,-.1511940727e-9,.2193666394e-9,-.3799541647e-9,.4905187176e-9,-.3033617129e-9,.5254378997e-9,.1563891368e-9,-1.872739842,3.243680556,-4.187573592,4.954803895,-.4572029592,.7918987547,-1.022336896,1.670078386,-2.892660618,-1.283464460,0.,-0.,0.,0.,-0.,-0.,1.663589468,-2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,-1.425047564,2.468254783,-3.186503223,3.770321460,-4.275142691,-2.360183599,4.087957910,-5.277530967,6.244458853,-.4945548803e-2,.8565941795e-2,-.1105858331e-1,2.175490654,-3.768060344,.1537427033,.6224210815,-1.078064937,1.391775849,-1.646771393,1.648153002,-2.854684738,3.685382148,1.157573982,-2.004976951,-.8770258842,-.7554460893e-2,.1308471009e-1,-.1689228809e-1,-.2731590581e-1,.4731253672e-1,-.3482606398e-1,-.2198338910,.3807634685,-1.005417380,.1997404621},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,1.425047564,-2.468254784,3.186503224,-3.770321460,4.275142693,2.360183599,-4.087957910,5.277530967,-6.244458853,.4945548803e-2,-.8565941795e-2,.1105858331e-1,-2.175490654,3.768060344,-.1537427033,.6224210804,-1.078064935,1.391775847,-1.646771390,1.648152999,-2.854684733,3.685382143,1.157573980,-2.004976947,-.8770258827,.7554451164e-2,-.1308469324e-1,.1689226633e-1,.2731587063e-1,-.4731247580e-1,.3482601912e-1,-.2198339029,.3807634890,-1.005417434,-.1997405113},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.645503189,2.850095128,-3.679456987,4.353592220,-4.936509566,1.377214242,-2.385405038,3.079544664,-3.643766386,-.7456423653,1.291490461,-1.667307016,-.7214030405e-1,.1249506719,.8704037330,.2424871131e-9,-.42e-9,.5422176684e-9,-.6415605973e-9,.7274613394e-9,-.2092513633e-9,.3624339929e-9,-.4679002728e-9,.5536270690e-9,.1249332326e-9,-.2163907064e-9,.2793592008e-9,-.1348127464e-10,.2335025263e-10,-.9727958508e-10,-2.424562843,4.199466029,-5.421487332,6.414790321,1.641768329,-2.843626160,3.671105586,-.5644916125,.9777281536,-.5417963103,-.7483314774e-9,.1296148140e-8,-.1673320053e-8,.3917700593e-9,-.6785656478e-9,-.9473764217e-11,2.788419758,-4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,.9641937137,-1.670032500,2.156002687,-2.551016781,2.892581141,.2394478538,-.4147358487,.5354216783,-.6335194733,-.9155822076,1.585834902,-2.047304055,.5490545322,-.9509903459,.3754286701,-.9430172903,1.633353859,-2.108650765,2.494989233,-.7446746083,1.289814257,-1.665143045,.9321620643,-1.614552057,.2692350424e-1,-1.398574920,2.422402819,-3.127308592,-1.821631846,3.155158911,1.076451798,-.1570858835,.2720807313,-.2832576764,.8954348170},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,-.9641937137,1.670032500,-2.156002687,2.551016781,-2.892581141,-.2394478538,.4147358487,-.5354216783,.6335194733,.9155822076,-1.585834902,2.047304055,-.5490545322,.9509903459,-.3754286701,-.9430172892,1.633353857,-2.108650763,2.494989230,-.7446746070,1.289814255,-1.665143042,.9321620636,-1.614552056,.2692350421e-1,1.398574927,-2.422402832,3.127308609,1.821631855,-3.155158927,-1.076451804,-.1570858218,.2720806245,-.2832575652,-.8954346999},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.113355001,-1.928387427,2.489537463,-2.945660451,3.340065000,-1.123448021,1.945869052,-2.512106144,2.972364074,-1.252806288,2.169924144,-2.801360023,.9835692099,-1.703591844,1.344816036,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.1039230485e-8,.6764731475e-9,-.1171685862e-8,.1512639943e-8,-.1789779717e-8,.2094450706e-9,-.3627695035e-9,.4683334153e-9,-.5392014460e-9,.9339242999e-9,-.350864418e-9,-.6338914369,1.097932175,-1.417424343,1.677119101,-1.926686150,3.337118302,-4.308201201,-1.956429599,3.388635469,.2565449507,.7483314774e-8,-.1296148140e-7,.1673320053e-7,.3070569490e-7,-.5318382366e-7,.4977473171e-7,.1905991076,-.3301273382,.9829492714,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,1.926395974,-3.336615703,4.307552349,-5.096764673,5.779187924,-.6138328492,1.063189683,-1.372571978,1.624049066,-.8603944811,1.490246955,-1.923900547,1.645753812,-2.850529219,-1.343411045,.5526193998,-.9571648777,1.235694544,-1.462093502,.2281916739e-1,-.3952395730e-1,.5102520946e-1,-.4132700278,.7158046857,.4485373028,-1.314274316,2.276389891,-2.938806712,-.4873895555,.8441834734,1.215532077,-2.142609188,3.711107975,-1.473102430,-1.425504294},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,-1.926395976,3.336615704,-4.307552352,5.096764678,-5.779187926,.6138328496,-1.063189683,1.372571979,-1.624049067,.8603944811,-1.490246955,1.923900547,-1.645753812,2.850529220,1.343411045,.5526194031,-.9571648834,1.235694551,-1.462093511,.2281916752e-1,-.3952395754e-1,.5102520976e-1,-.4132700304,.7158046901,.4485373055,1.314274282,-2.276389831,2.938806635,.4873895427,-.8441834514,-1.215532046,-2.142609204,3.711108002,-1.473102441,1.425504133},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.224410470,-3.852791952,4.973933020,-5.885236917,6.673231410,.7950455559,-1.377059297,1.777775908,-2.103492822,-1.500452022,2.598859138,-3.355112719,-2.072134551,3.589042323,-.2431355237,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,-0.,-.2398353968,.4154070926,-.5362882506,.6345448156,-1.087945327,1.884376583,-2.432719707,-2.347136393,4.065359486,-2.942787847,0.,-0.,0.,0.,-0.,0.,.2728458417e-1,-.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,2.679844379,-4.641626620,5.992314198,-7.090201776,8.039533138,-2.043694514,3.539782735,-4.569839860,5.407107442,.8110805448,-1.404832712,1.813631233,.6037611818,-1.045745042,-1.765469345,2.576012736,-4.461784940,5.760139589,-6.815489076,-1.403212584,2.430435489,-3.137678723,-.8058635216e-2,.1395796564e-1,1.237350351,1.238945990,-2.145917403,2.770367454,-.4443853780,.7696980531,-.2975289326,-.7339728352,1.271278242,.1365904598,-2.538475875},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,-2.679844380,4.641626622,-5.992314202,7.090201781,-8.039533141,2.043694516,-3.539782737,4.569839863,-5.407107446,-.8110805448,1.404832712,-1.813631233,-.6037611823,1.045745043,1.765469346,2.576012735,-4.461784937,5.760139587,-6.815489073,-1.403212583,2.430435488,-3.137678722,-.8058635216e-2,.1395796564e-1,1.237350350,-1.238945998,2.145917416,-2.770367471,.4443853807,-.7696980577,.2975289345,-.7339727817,1.271278149,.1365904499,2.538476058},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.094417748,-5.359688761,6.919328434,-8.187059814,9.283253244,3.470033583,-6.010274468,7.759230975,-9.180845900,2.032368460,-3.520165433,4.544514032,-.5354861533,.9274892243,-2.789841949,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.6238553746e-1,.1080549205,-.1394983026,.1650566176,-.3561120901,.6168042332,-.7962908411,-1.076648709,1.864810267,-2.258512675,.1496662955e-7,-.2592296279e-7,.3346640106e-7,.1111356701e-6,-.1924926272e-6,.4338629615e-6,.1846114076e-2,-.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,3.291626742,-5.701264756,7.360301149,-8.708825766,9.874880227,-3.696869036,6.403165001,-8.266450470,9.780996101,3.754901608,-6.503680358,8.396215242,-3.509035267,6.077827368,3.000239739,4.769620097,-8.261224341,10.66519477,-12.61922863,-4.867652070,8.431020700,-10.88440092,4.583518828,-7.938887491,-3.968421708,5.735706993,-9.934535927,12.82543073,-5.454863311,9.448100406,4.799476921,6.099313091,-10.56432016,-5.461692032,5.821024942},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,-3.291626740,5.701264753,-7.360301145,8.708825761,-9.874880222,3.696869033,-6.403164996,8.266450463,-9.780996094,-3.754901605,6.503680355,-8.396215237,3.509035265,-6.077827365,-3.000239738,4.769620101,-8.261224347,10.66519478,-12.61922864,-4.867652073,8.431020705,-10.88440092,4.583518832,-7.938887498,-3.968421711,-5.735706944,9.934535843,-12.82543062,5.454863264,-9.448100324,-4.799476880,6.099312923,-10.56431987,-5.461691879,-5.821025177},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.800843169,-6.583253480,8.498943694,-10.05608579,11.40252950,6.399791554,-11.08476413,14.31036896,-16.93225689,8.680560534,-15.03517189,19.41032344,10.21325528,-17.68987705,10.69428583,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.3017302684e-2,.5226121551e-2,-.6746893911e-2,.7983032535e-2,-.2009496124e-1,.3480549385e-1,-.4493369933e-1,-.7471563177e-1,.1294112704,-.2049323002,-.7483314774e-8,.1296148140e-7,-.1673320053e-7,-.6443340372e-7,.1116019290e-6,-.3031212923e-6,.4333424029e-5,-.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,-1.719209449,2.977758114,-3.844269195,4.548600653,-5.157628347,-1.016533920,1.760688397,-2.273038946,2.689495952,1.730662758,-2.997595827,3.869879572,-.8438528152e-1,.1461595950,-1.608041156,.6684577348,-1.157802759,1.494716935,-1.768572929,.8304449710,-1.438372883,1.856931406,-.4995431840,.8652341756,-.5315629344,.4697515910,-.8136336225,1.050396490,.8765560645,-1.518239640,-.5415238988e-1,-.9469058756,1.640089087,-2.346450809,.7389290768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,.6685019987,-1.157879426,1.494815912,-1.768690039,2.005505995,1.636951328,-2.835282871,3.660334447,-4.330966124,1.284112192,-2.224147558,2.871362151,-.5257834218,.9106836003,-1.564527457,-.1324976951,.2294927397,-.2962738530,.3505559505,-.4894439054,.8477417116,-1.094429844,-.7451837832,1.290696174,-.4186690470,-.2123656599,.3678281127,-.4748640516,-1.044420364,1.808989135,-2.287254493,-.6336343026e-1,.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,1.050707449,-1.819878685,2.349453280,-2.779910610,3.152122347,-.6204174076,1.074594472,-1.387295498,1.641470170,-.2618481958e-1,.4535343788e-1,-.5855103656e-1,.6101686966,-1.056843184,-.8964075294,-1.444837340,2.502531681,-3.230754509,3.822680287,.4587476822,-.7945742934,1.025791002,.5055009012,-.8755532444,-1.086251548,-2.213877417,3.834548168,-4.950380399,.1678643553,-.2907495923,1.305496370,.1849334761e-1,-.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,-1.050707450,1.819878688,-2.349453283,2.779910614,-3.152122352,.6204174085,-1.074594474,1.387295500,-1.641470172,.2618481961e-1,-.4535343793e-1,.5855103662e-1,-.6101686971,1.056843185,.8964075306,-1.444837346,2.502531692,-3.230754521,3.822680303,.4587476840,-.7945742967,1.025791006,.5055009034,-.8755532483,-1.086251552,2.213877420,-3.834548174,4.950380406,-.1678643556,.2907495928,-1.305496372,.1849327171e-1,-.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,1.719209448,-2.977758112,3.844269192,-4.548600649,5.157628344,1.016533919,-1.760688396,2.273038944,-2.689495949,-1.730662756,2.997595824,-3.869879569,.8438528146e-1,-.1461595949,1.608041155,.6684577365,-1.157802762,1.494716939,-1.768572933,.8304449729,-1.438372886,1.856931411,-.4995431851,.8652341775,-.5315629356,-.4697515776,.8136335992,-1.050396460,-.8765560394,1.518239596,.5415238836e-1,-.9469058829,1.640089099,-2.346450828,-.7389299868},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,-.6685019987,1.157879426,-1.494815912,1.768690039,-2.005505995,-1.636951328,2.835282871,-3.660334447,4.330966124,-1.284112192,2.224147558,-2.871362151,.5257834218,-.9106836003,1.564527457,-.1324976936,.2294927372,-.2962738499,.3505559467,-.4894439003,.8477417029,-1.094429832,-.7451837755,1.290696161,-.4186690423,.2123656749,-.3678281386,.4748640850,1.044420437,-1.808989263,2.287254654,-.6336326377e-1,.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,-2.155052650,3.732660680,-4.818844218,5.701733371,-6.465157947,-1.817524160,3.148044191,-4.064107574,4.808716931,2.003812619,-3.470705265,4.480661231,.8618882312,-1.492834207,-2.361581599,1.978516938,-3.426891859,4.424098369,-5.234663784,3.116327372,-5.397637342,6.968319843,-.7599463103,1.316265621,-2.583756308,-1.441495348,2.496743182,-3.223281588,-3.286953601,5.693170641,-.9156869635,.8622047985,-1.493382518,2.564268432,-.3927122230},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,1.100336031,-1.905837910,2.460426162,-2.911215495,3.301008092,2.979568311,-5.160763700,6.662517288,-7.883196767,3.041169362,-5.267459849,6.800261424,.1577963754,-.2733113395,-2.799087386,.4783243874,-.8284821415,1.069565846,-1.265527375,1.930559283,-3.343826766,4.316861791,3.440637089,-5.959358251,2.951743964,.1430932586,-.2478447941,.3199662533,.7646780256,-1.324461192,1.899532545,.2128667917e-1,-.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,1.054716618,-1.826822770,2.358418055,-2.790517875,3.164149855,-1.162044153,2.012719514,-2.598409719,3.074479842,1.144752980,-1.982770324,2.559745481,-1.019684603,1.766145540,.8049318120,-1.868089212,3.235625429,-4.177174467,4.942499484,1.853832431,-3.210931959,4.145295333,-1.671751149,2.895557928,1.349636370,-2.753730610,4.769601326,-6.157528834,2.522275578,-4.368709454,-2.092761921,.4701430699,-.8143116839,-.4012398833,3.559792983},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,-1.054716618,1.826822769,-2.358418053,2.790517873,-3.164149853,1.162044152,-2.012719513,2.598409717,-3.074479840,-1.144752979,1.982770321,-2.559745479,1.019684602,-1.766145538,-.8049318114,-1.868089209,3.235625422,-4.177174460,4.942499475,1.853832428,-3.210931954,4.145295326,-1.671751146,2.895557923,1.349636368,2.753730616,-4.769601337,6.157528849,-2.522275585,4.368709465,2.092761926,.4701430553,-.8143116587,-.4012398709,-3.559792582},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,2.155052650,-3.732660680,4.818844218,-5.701733371,6.465157947,1.817524160,-3.148044191,4.064107574,-4.808716931,-2.003812619,3.470705265,-4.480661231,-.8618882312,1.492834207,2.361581599,1.978516937,-3.426891859,4.424098367,-5.234663782,3.116327372,-5.397637341,6.968319842,-.7599463103,1.316265621,-2.583756308,1.441495367,-2.496743214,3.223281630,3.286953644,-5.693170715,.9156869755,.8622049336,-1.493382752,2.564268834,.3927112853},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,-1.100336031,1.905837911,-2.460426163,2.911215496,-3.301008094,-2.979568312,5.160763703,-6.662517291,7.883196771,-3.041169365,5.267459851,-6.800261430,-.1577963755,.2733113396,2.799087387,.4783243870,-.8284821409,1.069565845,-1.265527374,1.930559282,-3.343826764,4.316861789,3.440637085,-5.959358244,2.951743961,-.1430932534,.2478447850,-.3199662416,-.7646779976,1.324461144,-1.899532475,.2128671353e-1,-.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,-2.784908111,4.823602340,-6.227243845,7.368174283,-8.354724330,-.8097724007e-1,.1402566941,-.1810706135,.2142456391,2.270993172,-3.933475556,5.078095108,-2.198518194,3.807945213,.1780549380,3.578005823,-6.197287874,8.000664244,-9.466513600,1.789933617,-3.100255967,4.002413242,-3.541924988,6.134794038,1.442211311,-4.013547435,6.951668076,-8.974564896,-3.781587883,6.549902349,3.879680769,4.218875403,-7.307306550,5.787244192,-4.275985292},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,.7243065273,-1.254535705,1.619598631,-1.916334944,2.172919582,2.659465646,-4.606329621,5.946745969,-7.036284721,4.948721068,-8.571436320,11.06567671,5.758806642,-9.974545694,3.789231007,.2491916055,-.4316125216,.5572093695,-.6592990173,1.322791395,-2.291141904,2.957851479,3.610515302,-6.253595946,6.564693789,.7674948247e-1,-.1329340031,.1716170600,.5316983539,-.9209285636,1.896167979,.2260074784e-1,-.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,2.060601584,-3.569066636,4.607645215,-5.451839342,6.181804751,-2.578488405,4.466072926,-5.765675354,6.822039081,3.067086901,-5.312350344,6.858214803,-3.560288453,6.166600490,4.079521103,.1446504328,-.2505418989,.3234482008,-.3827090723,-.1721034134,.2980918563,-.3848349315,.1998400143,-.3461330583,-.2290664311,-2.797494156,4.845402011,-6.255387098,3.249889483,-5.628973706,-3.727231982,-3.881338999,6.722676349,4.454821378,-1.567068824},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,-2.060601584,3.569066636,-4.607645215,5.451839342,-6.181804751,2.578488405,-4.466072926,5.765675354,-6.822039081,-3.067086901,5.312350344,-6.858214803,3.560288453,-6.166600490,-4.079521103,.1446504323,-.2505418980,.3234481995,-.3827090709,-.1721034127,.2980918550,-.3848349300,.1998400136,-.3461330570,-.2290664303,2.797494172,-4.845402039,6.255387135,-3.249889503,5.628973739,3.727232004,-3.881339015,6.722676376,4.454821399,1.567068815},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,2.784908111,-4.823602342,6.227243845,-7.368174283,8.354724334,.8097724011e-1,-.1402566942,.1810706136,-.2142456393,-2.270993172,3.933475557,-5.078095109,2.198518194,-3.807945213,-.1780549380,3.578005828,-6.197287884,8.000664256,-9.466513613,1.789933619,-3.100255970,4.002413246,-3.541924993,6.134794047,1.442211314,4.013547495,-6.951668180,8.974565029,3.781587939,-6.549902446,-3.879680827,4.218875359,-7.307306473,5.787244129,4.275985427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,-.7243065266,1.254535704,-1.619598630,1.916334942,-2.172919580,-2.659465643,4.606329616,-5.946745963,7.036284714,-4.948721064,8.571436312,-11.06567670,-5.758806636,9.974545684,-3.789231004,.2491916053,-.4316125213,.5572093687,-.6592990165,1.322791393,-2.291141901,2.957851475,3.610515299,-6.253595941,6.564693781,-.7674948090e-1,.1329340004,-.1716170565,-.5316983431,.9209285448,-1.896167939,.2260067231e-1,-.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,-2.517563337,4.360547610,-5.629442758,6.660846498,-7.552690011,.1640586792,-.2841579679,.3668463591,-.4340584657,1.848945781,-3.202468033,4.134368453,-2.129273724,3.688010274,.6507752466,2.539766587,-4.399004769,5.679090736,-6.719590780,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207893,-1.820466373,3.153140251,-4.070686560,-1.433342530,2.482622088,1.817251596,.7671298784,-1.328707925,.9100500654,.2540057989},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,.5106664481,-.8845002336,1.141884891,-1.351096424,1.531999344,1.778426716,-3.080325431,3.976683031,-4.705274817,3.022001204,-5.234259622,6.757400118,2.956809489,-5.121344262,1.089458308,.3202872177e-1,-.5547537340e-1,.7161839911e-1,-.8474003263e-1,.1620213035,-.2806291295,.3622906482,.4123799546,-.7142630336,.6762823204,-.2859462130e-1,.4952733691e-1,-.6393951701e-1,-.1892065416,.3277153433,-.6347740071,-.1157780971e-1,.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,2.006896889,-3.476047377,4.487557867,-5.309750075,6.020690668,-1.942485396,3.364483399,-4.343529391,5.139333284,1.514566645,-2.623306380,3.386673975,-.8275357615,1.433333984,.1195045576e-1,.2438275696,-.4223217388,.5452150204,-.6451071121,-.1988874943,.3444832452,-.4447259571,.1234754608,-.2138657718,-.3202669808e-1,-2.331238944,4.037824296,-5.212808751,1.622549074,-2.810337434,-.7363933866,-3.287037546,5.693316036,1.939519592,-1.523064210},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,-2.006896889,3.476047376,-4.487557867,5.309750075,-6.020690666,1.942485396,-3.364483399,4.343529391,-5.139333284,-1.514566645,2.623306379,-3.386673973,.8275357615,-1.433333984,-.1195045576e-1,.2438275663,-.4223217331,.5452150131,-.6451071034,-.1988874916,.3444832405,-.4447259510,.1234754592,-.2138657688,-.3202669764e-1,2.331238936,-4.037824281,5.212808732,-1.622549068,2.810337424,.7363933838,-3.287037517,5.693315987,1.939519574,1.523064511},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,2.517563337,-4.360547611,5.629442758,-6.660846498,7.552690013,-.1640586792,.2841579679,-.3668463591,.4340584657,-1.848945782,3.202468033,-4.134368454,2.129273724,-3.688010274,-.6507752466,2.539766588,-4.399004769,5.679090738,-6.719590782,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207894,1.820466382,-3.153140266,4.070686580,1.433342537,-2.482622099,-1.817251606,.7671297908,-1.328707774,.9100499616,-.2540054079},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,-.5106664484,.8845002342,-1.141884892,1.351096425,-1.531999345,-1.778426717,3.080325433,-3.976683034,4.705274821,-3.022001206,5.234259626,-6.757400124,-2.956809490,5.121344265,-1.089458309,.3202872292e-1,-.5547537539e-1,.7161840168e-1,-.8474003567e-1,.1620213093,-.2806291396,.3622906613,.4123799696,-.7142630596,.6762823444,.2859462354e-1,-.4952734080e-1,.6393952203e-1,.1892065565,-.3277153691,.6347740570,-.1157777851e-1,.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,-3.069477201,5.316490463,-6.863559675,8.121073328,-9.208431602,2.156281693,-3.734789450,4.821592446,-5.704985119,-.5614333340,.9724310592,-1.255403100,-1.126462044,1.951089492,2.330362820,4.159176446,-7.203904921,9.300201262,-11.00414654,-1.935519673,3.352418413,-4.327953560,-.5514408001,.9551234835,2.488724225,-4.693254379,8.128955037,-10.49443583,1.219702470,-2.112586649,1.756836938,4.705330903,-8.149872191,-.3189574095,-4.265567617},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,.2458344589,-.4257977730,.5497025612,-.6504168418,.7375033768,1.054396484,-1.826268282,2.357702214,-2.789670881,2.510226735,-4.347840243,5.613037617,4.324472530,-7.490206138,5.881975759,.1420874210e-1,-.2461026322e-1,.3177171321e-1,-.3759279804e-1,.8700209506e-1,-.1506920490,.1945425987,.2894494276,-.5013411150,.6894088761,-.1216128451e-2,.2106396265e-2,-.2719345885e-2,-.9658561257e-2,.1672911883e-1,-.4108123658e-1,-.3746153966e-3,.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,2.823642741,-4.890692689,6.313857110,-7.470656481,8.470928224,-3.210678176,5.561057729,-7.179294657,8.494655996,3.324354634,-5.757951125,7.433482942,-3.198010479,5.539116631,2.859723954,2.797956434,-4.846202701,6.256420786,-7.402696906,-2.907006833,5.035083533,-6.500264888,2.811341931,-4.869387063,-2.534878984,1.242384808,-2.151873609,2.778056884,-1.210043957,2.095857614,1.103053355,-1.284643645,2.225068063,1.185627028,-3.765218080},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,-2.823642742,4.890692690,-6.313857114,7.470656486,-8.470928225,3.210678177,-5.561057732,7.179294660,-8.494656000,-3.324354634,5.757951127,-7.433482942,3.198010480,-5.539116634,-2.859723955,2.797956438,-4.846202708,6.256420796,-7.402696917,-2.907006837,5.035083540,-6.500264898,2.811341934,-4.869387068,-2.534878988,-1.242384788,2.151873575,-2.778056840,1.210043938,-2.095857580,-1.103053339,-1.284643653,2.225068076,1.185627035,3.765217145},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,3.069477200,-5.316490462,6.863559672,-8.121073323,9.208431600,-2.156281693,3.734789450,-4.821592446,5.704985119,.5614333335,-.9724310584,1.255403098,1.126462044,-1.951089492,-2.330362819,4.159176444,-7.203904918,9.300201260,-11.00414653,-1.935519673,3.352418412,-4.327953559,-.5514408000,.9551234833,2.488724224,4.693254424,-8.128955115,10.49443593,-1.219702482,2.112586669,-1.756836954,4.705330679,-8.149871802,-.3189573943,4.265566870},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,-.2458344586,.4257977724,-.5497025604,.6504168409,-.7375033757,-1.054396483,1.826268280,-2.357702211,2.789670877,-2.510226732,4.347840236,-5.613037612,-4.324472524,7.490206128,-5.881975750,.1420874260e-1,-.2461026409e-1,.3177171433e-1,-.3759279937e-1,.8700209816e-1,-.1506920544,.1945426056,.2894494379,-.5013411328,.6894089005,.1216122464e-2,-.2106385895e-2,.2719332498e-2,.9658513710e-2,-.1672903648e-1,.4108103434e-1,-.3745694178e-3,.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[56],
		T o_output[52]
	) {
		static const T matrix[52][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,0.,0.,0.,0.,0.,-1.360827635,2.357022604,-3.042903097,3.600411499,.8380524812,-1.451549477,1.873942317,.8198497638,-1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-2.721655270,3.220305944,-1.873942317,3.245763304,-4.190262406,.5237828008,-.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,-1.217161239,2.052800957,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,-1.704828602,2.952849755,-3.812112642,4.510552507,-5.114485804,-3.104623248,5.377365205,-6.942148628,8.214061030,-.5462559154,.9461429991,-1.221465360,2.694797933,-4.667526936,1.022426944,1.298058152,-2.248302670,2.902546266,-3.434339057,3.719687555,-6.442687835,8.317474224,3.264192666,-5.653747546,-1.221656347,-.8344196900,1.445257298,-1.865819148,-3.243190051,5.617369949,-4.760960496,.4850599026,-.8401483961,2.375793728,-.2614753427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,1.704828601,-2.952849754,3.812112640,-4.510552505,5.114485802,3.104623246,-5.377365203,6.942148625,-8.214061027,.5462559149,-.9461429982,1.221465359,-2.694797932,4.667526936,-1.022426944,1.298058156,-2.248302677,2.902546276,-3.434339069,3.719687568,-6.442687856,8.317474253,3.264192678,-5.653747565,-1.221656351,.8344196802,-1.445257281,1.865819127,3.243190014,-5.617369884,4.760960437,.4850600816,-.8401487061,2.375794604,.2614758325},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.968566503,3.409657202,-4.401848518,5.208337406,-5.905699510,2.347500666,-4.065990423,5.249171065,-6.210902963,-2.609637841,4.520025331,-5.835327609,2.775506203,-4.807317761,-2.855122567,-.3360178568e-9,.582e-9,-.7513587691e-9,.8890196848e-9,-.1008053570e-8,.4007561299e-9,-.6941299786e-9,.8961179490e-9,-.1060301056e-8,-.4456004436e-9,.7718026080e-9,-.9963928825e-9,.4740581912e-9,-.8210928729e-9,-.4878407244e-9,-2.709993221,4.693845947,-6.059729062,7.169968120,3.014474194,-5.221222463,6.740569213,-3.208782240,5.557773872,3.304507380,-.7333648479e-8,.1270225177e-7,-.1639853652e-7,.7812670753e-8,-.1353194269e-7,-.8054303853e-8,3.483595229,-6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,-.8327988349,1.442449894,-1.862194806,2.203378609,-2.498396504,-1.088657954,1.885610890,-2.434313191,2.880318211,.4375794197,-.7579097871,.9784573278,.9743405328,-1.687607307,-.5933186699,-.4672884697,.8093673712,-1.044888783,1.236329082,-1.022115558,1.770356077,-2.285519867,-.3251247057,.5631325093,.8276510476,.6684136118,-1.157726336,1.494618273,2.033596264,-3.522292053,1.746351412,-.1476726503,.2557765332,-.5739768229,-.2050386800},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,.8327988342,-1.442449893,1.862194804,-2.203378607,2.498396503,1.088657953,-1.885610888,2.434313189,-2.880318209,-.4375794194,.7579097864,-.9784573273,-.9743405322,1.687607306,.5933186693,-.4672884647,.8093673627,-1.044888772,1.236329069,-1.022115547,1.770356058,-2.285519843,-.3251247023,.5631325034,.8276510391,-.6684136118,1.157726336,-1.494618273,-2.033596264,3.522292053,-1.746351412,-.1476726019,.2557764494,-.5739766349,.2050384112},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.961633263,1.665597669,-2.150277344,2.544242465,-2.884899788,-.2081759125,.3605712573,-.4654954917,.5507816934,1.215013765,-2.104465573,2.716853372,-1.352704319,2.342952608,.5364099445,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.1039230485e-8,.5714598801e-10,-.9897975472e-10,.1277823139e-9,-.1511940727e-9,.2193666394e-9,-.3799541647e-9,.4905187176e-9,-.3033617129e-9,.5254378997e-9,.1563891368e-9,-1.872739842,3.243680556,-4.187573592,4.954803895,-.4572029592,.7918987547,-1.022336896,1.670078386,-2.892660618,-1.283464460,0.,-0.,0.,0.,-0.,-0.,1.663589468,-2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,-1.425047564,2.468254783,-3.186503223,3.770321460,-4.275142691,-2.360183599,4.087957910,-5.277530967,6.244458853,-.4945548803e-2,.8565941795e-2,-.1105858331e-1,2.175490654,-3.768060344,.1537427033,.6224210815,-1.078064937,1.391775849,-1.646771393,1.648153002,-2.854684738,3.685382148,1.157573982,-2.004976951,-.8770258842,-.7554460893e-2,.1308471009e-1,-.1689228809e-1,-.2731590581e-1,.4731253672e-1,-.3482606398e-1,-.2198338910,.3807634685,-1.005417380,.1997404621},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,1.425047564,-2.468254784,3.186503224,-3.770321460,4.275142693,2.360183599,-4.087957910,5.277530967,-6.244458853,.4945548803e-2,-.8565941795e-2,.1105858331e-1,-2.175490654,3.768060344,-.1537427033,.6224210804,-1.078064935,1.391775847,-1.646771390,1.648152999,-2.854684733,3.685382143,1.157573980,-2.004976947,-.8770258827,.7554451164e-2,-.1308469324e-1,.1689226633e-1,.2731587063e-1,-.4731247580e-1,.3482601912e-1,-.2198339029,.3807634890,-1.005417434,-.1997405113},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.645503189,2.850095128,-3.679456987,4.353592220,-4.936509566,1.377214242,-2.385405038,3.079544664,-3.643766386,-.7456423653,1.291490461,-1.667307016,-.7214030405e-1,.1249506719,.8704037330,.2424871131e-9,-.42e-9,.5422176684e-9,-.6415605973e-9,.7274613394e-9,-.2092513633e-9,.3624339929e-9,-.4679002728e-9,.5536270690e-9,.1249332326e-9,-.2163907064e-9,.2793592008e-9,-.1348127464e-10,.2335025263e-10,-.9727958508e-10,-2.424562843,4.199466029,-5.421487332,6.414790321,1.641768329,-2.843626160,3.671105586,-.5644916125,.9777281536,-.5417963103,-.7483314774e-9,.1296148140e-8,-.1673320053e-8,.3917700593e-9,-.6785656478e-9,-.9473764217e-11,2.788419758,-4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,.9641937137,-1.670032500,2.156002687,-2.551016781,2.892581141,.2394478538,-.4147358487,.5354216783,-.6335194733,-.9155822076,1.585834902,-2.047304055,.5490545322,-.9509903459,.3754286701,-.9430172903,1.633353859,-2.108650765,2.494989233,-.7446746083,1.289814257,-1.665143045,.9321620643,-1.614552057,.2692350424e-1,-1.398574920,2.422402819,-3.127308592,-1.821631846,3.155158911,1.076451798,-.1570858835,.2720807313,-.2832576764,.8954348170},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,-.9641937137,1.670032500,-2.156002687,2.551016781,-2.892581141,-.2394478538,.4147358487,-.5354216783,.6335194733,.9155822076,-1.585834902,2.047304055,-.5490545322,.9509903459,-.3754286701,-.9430172892,1.633353857,-2.108650763,2.494989230,-.7446746070,1.289814255,-1.665143042,.9321620636,-1.614552056,.2692350421e-1,1.398574927,-2.422402832,3.127308609,1.821631855,-3.155158927,-1.076451804,-.1570858218,.2720806245,-.2832575652,-.8954346999},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.113355001,-1.928387427,2.489537463,-2.945660451,3.340065000,-1.123448021,1.945869052,-2.512106144,2.972364074,-1.252806288,2.169924144,-2.801360023,.9835692099,-1.703591844,1.344816036,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.1039230485e-8,.6764731475e-9,-.1171685862e-8,.1512639943e-8,-.1789779717e-8,.2094450706e-9,-.3627695035e-9,.4683334153e-9,-.5392014460e-9,.9339242999e-9,-.350864418e-9,-.6338914369,1.097932175,-1.417424343,1.677119101,-1.926686150,3.337118302,-4.308201201,-1.956429599,3.388635469,.2565449507,.7483314774e-8,-.1296148140e-7,.1673320053e-7,.3070569490e-7,-.5318382366e-7,.4977473171e-7,.1905991076,-.3301273382,.9829492714,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,1.926395974,-3.336615703,4.307552349,-5.096764673,5.779187924,-.6138328492,1.063189683,-1.372571978,1.624049066,-.8603944811,1.490246955,-1.923900547,1.645753812,-2.850529219,-1.343411045,.5526193998,-.9571648777,1.235694544,-1.462093502,.2281916739e-1,-.3952395730e-1,.5102520946e-1,-.4132700278,.7158046857,.4485373028,-1.314274316,2.276389891,-2.938806712,-.4873895555,.8441834734,1.215532077,-2.142609188,3.711107975,-1.473102430,-1.425504294},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,-1.926395976,3.336615704,-4.307552352,5.096764678,-5.779187926,.6138328496,-1.063189683,1.372571979,-1.624049067,.8603944811,-1.490246955,1.923900547,-1.645753812,2.850529220,1.343411045,.5526194031,-.9571648834,1.235694551,-1.462093511,.2281916752e-1,-.3952395754e-1,.5102520976e-1,-.4132700304,.7158046901,.4485373055,1.314274282,-2.276389831,2.938806635,.4873895427,-.8441834514,-1.215532046,-2.142609204,3.711108002,-1.473102441,1.425504133},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.224410470,-3.852791952,4.973933020,-5.885236917,6.673231410,.7950455559,-1.377059297,1.777775908,-2.103492822,-1.500452022,2.598859138,-3.355112719,-2.072134551,3.589042323,-.2431355237,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,-0.,-.2398353968,.4154070926,-.5362882506,.6345448156,-1.087945327,1.884376583,-2.432719707,-2.347136393,4.065359486,-2.942787847,0.,-0.,0.,0.,-0.,0.,.2728458417e-1,-.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,2.679844379,-4.641626620,5.992314198,-7.090201776,8.039533138,-2.043694514,3.539782735,-4.569839860,5.407107442,.8110805448,-1.404832712,1.813631233,.6037611818,-1.045745042,-1.765469345,2.576012736,-4.461784940,5.760139589,-6.815489076,-1.403212584,2.430435489,-3.137678723,-.8058635216e-2,.1395796564e-1,1.237350351,1.238945990,-2.145917403,2.770367454,-.4443853780,.7696980531,-.2975289326,-.7339728352,1.271278242,.1365904598,-2.538475875},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,-2.679844380,4.641626622,-5.992314202,7.090201781,-8.039533141,2.043694516,-3.539782737,4.569839863,-5.407107446,-.8110805448,1.404832712,-1.813631233,-.6037611823,1.045745043,1.765469346,2.576012735,-4.461784937,5.760139587,-6.815489073,-1.403212583,2.430435488,-3.137678722,-.8058635216e-2,.1395796564e-1,1.237350350,-1.238945998,2.145917416,-2.770367471,.4443853807,-.7696980577,.2975289345,-.7339727817,1.271278149,.1365904499,2.538476058},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.094417748,-5.359688761,6.919328434,-8.187059814,9.283253244,3.470033583,-6.010274468,7.759230975,-9.180845900,2.032368460,-3.520165433,4.544514032,-.5354861533,.9274892243,-2.789841949,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.6238553746e-1,.1080549205,-.1394983026,.1650566176,-.3561120901,.6168042332,-.7962908411,-1.076648709,1.864810267,-2.258512675,.1496662955e-7,-.2592296279e-7,.3346640106e-7,.1111356701e-6,-.1924926272e-6,.4338629615e-6,.1846114076e-2,-.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,3.291626742,-5.701264756,7.360301149,-8.708825766,9.874880227,-3.696869036,6.403165001,-8.266450470,9.780996101,3.754901608,-6.503680358,8.396215242,-3.509035267,6.077827368,3.000239739,4.769620097,-8.261224341,10.66519477,-12.61922863,-4.867652070,8.431020700,-10.88440092,4.583518828,-7.938887491,-3.968421708,5.735706993,-9.934535927,12.82543073,-5.454863311,9.448100406,4.799476921,6.099313091,-10.56432016,-5.461692032,5.821024942},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,-3.291626740,5.701264753,-7.360301145,8.708825761,-9.874880222,3.696869033,-6.403164996,8.266450463,-9.780996094,-3.754901605,6.503680355,-8.396215237,3.509035265,-6.077827365,-3.000239738,4.769620101,-8.261224347,10.66519478,-12.61922864,-4.867652073,8.431020705,-10.88440092,4.583518832,-7.938887498,-3.968421711,-5.735706944,9.934535843,-12.82543062,5.454863264,-9.448100324,-4.799476880,6.099312923,-10.56431987,-5.461691879,-5.821025177},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.800843169,-6.583253480,8.498943694,-10.05608579,11.40252950,6.399791554,-11.08476413,14.31036896,-16.93225689,8.680560534,-15.03517189,19.41032344,10.21325528,-17.68987705,10.69428583,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.3017302684e-2,.5226121551e-2,-.6746893911e-2,.7983032535e-2,-.2009496124e-1,.3480549385e-1,-.4493369933e-1,-.7471563177e-1,.1294112704,-.2049323002,-.7483314774e-8,.1296148140e-7,-.1673320053e-7,-.6443340372e-7,.1116019290e-6,-.3031212923e-6,.4333424029e-5,-.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,-1.719209449,2.977758114,-3.844269195,4.548600653,-5.157628347,-1.016533920,1.760688397,-2.273038946,2.689495952,1.730662758,-2.997595827,3.869879572,-.8438528152e-1,.1461595950,-1.608041156,.6684577348,-1.157802759,1.494716935,-1.768572929,.8304449710,-1.438372883,1.856931406,-.4995431840,.8652341756,-.5315629344,.4697515910,-.8136336225,1.050396490,.8765560645,-1.518239640,-.5415238988e-1,-.9469058756,1.640089087,-2.346450809,.7389290768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,.6685019987,-1.157879426,1.494815912,-1.768690039,2.005505995,1.636951328,-2.835282871,3.660334447,-4.330966124,1.284112192,-2.224147558,2.871362151,-.5257834218,.9106836003,-1.564527457,-.1324976951,.2294927397,-.2962738530,.3505559505,-.4894439054,.8477417116,-1.094429844,-.7451837832,1.290696174,-.4186690470,-.2123656599,.3678281127,-.4748640516,-1.044420364,1.808989135,-2.287254493,-.6336343026e-1,.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,1.050707449,-1.819878685,2.349453280,-2.779910610,3.152122347,-.6204174076,1.074594472,-1.387295498,1.641470170,-.2618481958e-1,.4535343788e-1,-.5855103656e-1,.6101686966,-1.056843184,-.8964075294,-1.444837340,2.502531681,-3.230754509,3.822680287,.4587476822,-.7945742934,1.025791002,.5055009012,-.8755532444,-1.086251548,-2.213877417,3.834548168,-4.950380399,.1678643553,-.2907495923,1.305496370,.1849334761e-1,-.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,-1.050707450,1.819878688,-2.349453283,2.779910614,-3.152122352,.6204174085,-1.074594474,1.387295500,-1.641470172,.2618481961e-1,-.4535343793e-1,.5855103662e-1,-.6101686971,1.056843185,.8964075306,-1.444837346,2.502531692,-3.230754521,3.822680303,.4587476840,-.7945742967,1.025791006,.5055009034,-.8755532483,-1.086251552,2.213877420,-3.834548174,4.950380406,-.1678643556,.2907495928,-1.305496372,.1849327171e-1,-.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,1.719209448,-2.977758112,3.844269192,-4.548600649,5.157628344,1.016533919,-1.760688396,2.273038944,-2.689495949,-1.730662756,2.997595824,-3.869879569,.8438528146e-1,-.1461595949,1.608041155,.6684577365,-1.157802762,1.494716939,-1.768572933,.8304449729,-1.438372886,1.856931411,-.4995431851,.8652341775,-.5315629356,-.4697515776,.8136335992,-1.050396460,-.8765560394,1.518239596,.5415238836e-1,-.9469058829,1.640089099,-2.346450828,-.7389299868},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,-.6685019987,1.157879426,-1.494815912,1.768690039,-2.005505995,-1.636951328,2.835282871,-3.660334447,4.330966124,-1.284112192,2.224147558,-2.871362151,.5257834218,-.9106836003,1.564527457,-.1324976936,.2294927372,-.2962738499,.3505559467,-.4894439003,.8477417029,-1.094429832,-.7451837755,1.290696161,-.4186690423,.2123656749,-.3678281386,.4748640850,1.044420437,-1.808989263,2.287254654,-.6336326377e-1,.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,-2.155052650,3.732660680,-4.818844218,5.701733371,-6.465157947,-1.817524160,3.148044191,-4.064107574,4.808716931,2.003812619,-3.470705265,4.480661231,.8618882312,-1.492834207,-2.361581599,1.978516938,-3.426891859,4.424098369,-5.234663784,3.116327372,-5.397637342,6.968319843,-.7599463103,1.316265621,-2.583756308,-1.441495348,2.496743182,-3.223281588,-3.286953601,5.693170641,-.9156869635,.8622047985,-1.493382518,2.564268432,-.3927122230},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,1.100336031,-1.905837910,2.460426162,-2.911215495,3.301008092,2.979568311,-5.160763700,6.662517288,-7.883196767,3.041169362,-5.267459849,6.800261424,.1577963754,-.2733113395,-2.799087386,.4783243874,-.8284821415,1.069565846,-1.265527375,1.930559283,-3.343826766,4.316861791,3.440637089,-5.959358251,2.951743964,.1430932586,-.2478447941,.3199662533,.7646780256,-1.324461192,1.899532545,.2128667917e-1,-.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,1.054716618,-1.826822770,2.358418055,-2.790517875,3.164149855,-1.162044153,2.012719514,-2.598409719,3.074479842,1.144752980,-1.982770324,2.559745481,-1.019684603,1.766145540,.8049318120,-1.868089212,3.235625429,-4.177174467,4.942499484,1.853832431,-3.210931959,4.145295333,-1.671751149,2.895557928,1.349636370,-2.753730610,4.769601326,-6.157528834,2.522275578,-4.368709454,-2.092761921,.4701430699,-.8143116839,-.4012398833,3.559792983},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,-1.054716618,1.826822769,-2.358418053,2.790517873,-3.164149853,1.162044152,-2.012719513,2.598409717,-3.074479840,-1.144752979,1.982770321,-2.559745479,1.019684602,-1.766145538,-.8049318114,-1.868089209,3.235625422,-4.177174460,4.942499475,1.853832428,-3.210931954,4.145295326,-1.671751146,2.895557923,1.349636368,2.753730616,-4.769601337,6.157528849,-2.522275585,4.368709465,2.092761926,.4701430553,-.8143116587,-.4012398709,-3.559792582},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,2.155052650,-3.732660680,4.818844218,-5.701733371,6.465157947,1.817524160,-3.148044191,4.064107574,-4.808716931,-2.003812619,3.470705265,-4.480661231,-.8618882312,1.492834207,2.361581599,1.978516937,-3.426891859,4.424098367,-5.234663782,3.116327372,-5.397637341,6.968319842,-.7599463103,1.316265621,-2.583756308,1.441495367,-2.496743214,3.223281630,3.286953644,-5.693170715,.9156869755,.8622049336,-1.493382752,2.564268834,.3927112853},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,-1.100336031,1.905837911,-2.460426163,2.911215496,-3.301008094,-2.979568312,5.160763703,-6.662517291,7.883196771,-3.041169365,5.267459851,-6.800261430,-.1577963755,.2733113396,2.799087387,.4783243870,-.8284821409,1.069565845,-1.265527374,1.930559282,-3.343826764,4.316861789,3.440637085,-5.959358244,2.951743961,-.1430932534,.2478447850,-.3199662416,-.7646779976,1.324461144,-1.899532475,.2128671353e-1,-.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,-2.784908111,4.823602340,-6.227243845,7.368174283,-8.354724330,-.8097724007e-1,.1402566941,-.1810706135,.2142456391,2.270993172,-3.933475556,5.078095108,-2.198518194,3.807945213,.1780549380,3.578005823,-6.197287874,8.000664244,-9.466513600,1.789933617,-3.100255967,4.002413242,-3.541924988,6.134794038,1.442211311,-4.013547435,6.951668076,-8.974564896,-3.781587883,6.549902349,3.879680769,4.218875403,-7.307306550,5.787244192,-4.275985292},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,.7243065273,-1.254535705,1.619598631,-1.916334944,2.172919582,2.659465646,-4.606329621,5.946745969,-7.036284721,4.948721068,-8.571436320,11.06567671,5.758806642,-9.974545694,3.789231007,.2491916055,-.4316125216,.5572093695,-.6592990173,1.322791395,-2.291141904,2.957851479,3.610515302,-6.253595946,6.564693789,.7674948247e-1,-.1329340031,.1716170600,.5316983539,-.9209285636,1.896167979,.2260074784e-1,-.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,2.060601584,-3.569066636,4.607645215,-5.451839342,6.181804751,-2.578488405,4.466072926,-5.765675354,6.822039081,3.067086901,-5.312350344,6.858214803,-3.560288453,6.166600490,4.079521103,.1446504328,-.2505418989,.3234482008,-.3827090723,-.1721034134,.2980918563,-.3848349315,.1998400143,-.3461330583,-.2290664311,-2.797494156,4.845402011,-6.255387098,3.249889483,-5.628973706,-3.727231982,-3.881338999,6.722676349,4.454821378,-1.567068824},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,-2.060601584,3.569066636,-4.607645215,5.451839342,-6.181804751,2.578488405,-4.466072926,5.765675354,-6.822039081,-3.067086901,5.312350344,-6.858214803,3.560288453,-6.166600490,-4.079521103,.1446504323,-.2505418980,.3234481995,-.3827090709,-.1721034127,.2980918550,-.3848349300,.1998400136,-.3461330570,-.2290664303,2.797494172,-4.845402039,6.255387135,-3.249889503,5.628973739,3.727232004,-3.881339015,6.722676376,4.454821399,1.567068815},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,2.784908111,-4.823602342,6.227243845,-7.368174283,8.354724334,.8097724011e-1,-.1402566942,.1810706136,-.2142456393,-2.270993172,3.933475557,-5.078095109,2.198518194,-3.807945213,-.1780549380,3.578005828,-6.197287884,8.000664256,-9.466513613,1.789933619,-3.100255970,4.002413246,-3.541924993,6.134794047,1.442211314,4.013547495,-6.951668180,8.974565029,3.781587939,-6.549902446,-3.879680827,4.218875359,-7.307306473,5.787244129,4.275985427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,-.7243065266,1.254535704,-1.619598630,1.916334942,-2.172919580,-2.659465643,4.606329616,-5.946745963,7.036284714,-4.948721064,8.571436312,-11.06567670,-5.758806636,9.974545684,-3.789231004,.2491916053,-.4316125213,.5572093687,-.6592990165,1.322791393,-2.291141901,2.957851475,3.610515299,-6.253595941,6.564693781,-.7674948090e-1,.1329340004,-.1716170565,-.5316983431,.9209285448,-1.896167939,.2260067231e-1,-.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,-2.517563337,4.360547610,-5.629442758,6.660846498,-7.552690011,.1640586792,-.2841579679,.3668463591,-.4340584657,1.848945781,-3.202468033,4.134368453,-2.129273724,3.688010274,.6507752466,2.539766587,-4.399004769,5.679090736,-6.719590780,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207893,-1.820466373,3.153140251,-4.070686560,-1.433342530,2.482622088,1.817251596,.7671298784,-1.328707925,.9100500654,.2540057989},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,.5106664481,-.8845002336,1.141884891,-1.351096424,1.531999344,1.778426716,-3.080325431,3.976683031,-4.705274817,3.022001204,-5.234259622,6.757400118,2.956809489,-5.121344262,1.089458308,.3202872177e-1,-.5547537340e-1,.7161839911e-1,-.8474003263e-1,.1620213035,-.2806291295,.3622906482,.4123799546,-.7142630336,.6762823204,-.2859462130e-1,.4952733691e-1,-.6393951701e-1,-.1892065416,.3277153433,-.6347740071,-.1157780971e-1,.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,2.006896889,-3.476047377,4.487557867,-5.309750075,6.020690668,-1.942485396,3.364483399,-4.343529391,5.139333284,1.514566645,-2.623306380,3.386673975,-.8275357615,1.433333984,.1195045576e-1,.2438275696,-.4223217388,.5452150204,-.6451071121,-.1988874943,.3444832452,-.4447259571,.1234754608,-.2138657718,-.3202669808e-1,-2.331238944,4.037824296,-5.212808751,1.622549074,-2.810337434,-.7363933866,-3.287037546,5.693316036,1.939519592,-1.523064210},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,-2.006896889,3.476047376,-4.487557867,5.309750075,-6.020690666,1.942485396,-3.364483399,4.343529391,-5.139333284,-1.514566645,2.623306379,-3.386673973,.8275357615,-1.433333984,-.1195045576e-1,.2438275663,-.4223217331,.5452150131,-.6451071034,-.1988874916,.3444832405,-.4447259510,.1234754592,-.2138657688,-.3202669764e-1,2.331238936,-4.037824281,5.212808732,-1.622549068,2.810337424,.7363933838,-3.287037517,5.693315987,1.939519574,1.523064511},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,2.517563337,-4.360547611,5.629442758,-6.660846498,7.552690013,-.1640586792,.2841579679,-.3668463591,.4340584657,-1.848945782,3.202468033,-4.134368454,2.129273724,-3.688010274,-.6507752466,2.539766588,-4.399004769,5.679090738,-6.719590782,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207894,1.820466382,-3.153140266,4.070686580,1.433342537,-2.482622099,-1.817251606,.7671297908,-1.328707774,.9100499616,-.2540054079},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,-.5106664484,.8845002342,-1.141884892,1.351096425,-1.531999345,-1.778426717,3.080325433,-3.976683034,4.705274821,-3.022001206,5.234259626,-6.757400124,-2.956809490,5.121344265,-1.089458309,.3202872292e-1,-.5547537539e-1,.7161840168e-1,-.8474003567e-1,.1620213093,-.2806291396,.3622906613,.4123799696,-.7142630596,.6762823444,.2859462354e-1,-.4952734080e-1,.6393952203e-1,.1892065565,-.3277153691,.6347740570,-.1157777851e-1,.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,-3.069477201,5.316490463,-6.863559675,8.121073328,-9.208431602,2.156281693,-3.734789450,4.821592446,-5.704985119,-.5614333340,.9724310592,-1.255403100,-1.126462044,1.951089492,2.330362820,4.159176446,-7.203904921,9.300201262,-11.00414654,-1.935519673,3.352418413,-4.327953560,-.5514408001,.9551234835,2.488724225,-4.693254379,8.128955037,-10.49443583,1.219702470,-2.112586649,1.756836938,4.705330903,-8.149872191,-.3189574095,-4.265567617},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,.2458344589,-.4257977730,.5497025612,-.6504168418,.7375033768,1.054396484,-1.826268282,2.357702214,-2.789670881,2.510226735,-4.347840243,5.613037617,4.324472530,-7.490206138,5.881975759,.1420874210e-1,-.2461026322e-1,.3177171321e-1,-.3759279804e-1,.8700209506e-1,-.1506920490,.1945425987,.2894494276,-.5013411150,.6894088761,-.1216128451e-2,.2106396265e-2,-.2719345885e-2,-.9658561257e-2,.1672911883e-1,-.4108123658e-1,-.3746153966e-3,.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,2.823642741,-4.890692689,6.313857110,-7.470656481,8.470928224,-3.210678176,5.561057729,-7.179294657,8.494655996,3.324354634,-5.757951125,7.433482942,-3.198010479,5.539116631,2.859723954,2.797956434,-4.846202701,6.256420786,-7.402696906,-2.907006833,5.035083533,-6.500264888,2.811341931,-4.869387063,-2.534878984,1.242384808,-2.151873609,2.778056884,-1.210043957,2.095857614,1.103053355,-1.284643645,2.225068063,1.185627028,-3.765218080},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,-2.823642742,4.890692690,-6.313857114,7.470656486,-8.470928225,3.210678177,-5.561057732,7.179294660,-8.494656000,-3.324354634,5.757951127,-7.433482942,3.198010480,-5.539116634,-2.859723955,2.797956438,-4.846202708,6.256420796,-7.402696917,-2.907006837,5.035083540,-6.500264898,2.811341934,-4.869387068,-2.534878988,-1.242384788,2.151873575,-2.778056840,1.210043938,-2.095857580,-1.103053339,-1.284643653,2.225068076,1.185627035,3.765217145},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,3.069477200,-5.316490462,6.863559672,-8.121073323,9.208431600,-2.156281693,3.734789450,-4.821592446,5.704985119,.5614333335,-.9724310584,1.255403098,1.126462044,-1.951089492,-2.330362819,4.159176444,-7.203904918,9.300201260,-11.00414653,-1.935519673,3.352418412,-4.327953559,-.5514408000,.9551234833,2.488724224,4.693254424,-8.128955115,10.49443593,-1.219702482,2.112586669,-1.756836954,4.705330679,-8.149871802,-.3189573943,4.265566870},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,-.2458344586,.4257977724,-.5497025604,.6504168409,-.7375033757,-1.054396483,1.826268280,-2.357702211,2.789670877,-2.510226732,4.347840236,-5.613037612,-4.324472524,7.490206128,-5.881975750,.1420874260e-1,-.2461026409e-1,.3177171433e-1,-.3759279937e-1,.8700209816e-1,-.1506920544,.1945426056,.2894494379,-.5013411328,.6894089005,.1216122464e-2,-.2106385895e-2,.2719332498e-2,.9658513710e-2,-.1672903648e-1,.4108103434e-1,-.3745694178e-3,.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 52; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 56; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,-.1092092502e-1,-.1092092502e-1,.2184185003e-1,-.3473628658e-1,-.3473628658e-1,.6947257305e-1,-.3832164410e-1,-.3832164410e-1,.7664328825e-1,.4062391359e-1,.4062391359e-1,-.8124782718e-1,.5779239332e-1,.5779239332e-1,-.1155847866,.3805405532e-1,.3805405532e-1,-.7610811063e-1,.1179212887e-1,.1179212887e-1,-.2358425774e-1,.1252607563e-1,-.9077889505e-1,.7825281938e-1,.7825281938e-1,.1252607563e-1,-.9077889505e-1,-.6978997413e-3,-.4910403324e-1,.4980193299e-1,.4980193299e-1,-.6978997413e-3,-.4910403324e-1,.3188726319e-2,-.1156256937e-1,.8373843054e-2,.8373843054e-2,.3188726319e-2,-.1156256937e-1,.2855524789e-1,-.8634838574e-1,.5779313787e-1,.5779313787e-1,.2855524789e-1,-.8634838574e-1,.1765812736e-1,-.4036819305e-1,.2271006571e-1,.2271006571e-1,.1765812736e-1,-.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,-.1668198852e-1,-.1668198852e-1,.3336397701e-1,-.5306055418e-1,-.5306055418e-1,.1061211082,-.5853727826e-1,-.5853727826e-1,.1170745566,.6205405300e-1,.6205405300e-1,-.1241081060,.8827933896e-1,.8827933896e-1,-.1765586779,.5812852966e-1,.5812852966e-1,-.1162570593,.1801277438e-1,.1801277438e-1,-.3602554875e-1,.1913389657e-1,-.1386670527,.1195331560,.1195331560,.1913389657e-1,-.1386670527,-.1066059464e-2,-.7500764972e-1,.7607370920e-1,.7607370920e-1,-.1066059464e-2,-.7500764972e-1,.4870859907e-2,-.1766211645e-1,.1279125655e-1,.1279125655e-1,.4870859907e-2,-.1766211645e-1,.4361886163e-1,-.1318993379,.8828047629e-1,.8828047629e-1,.4361886163e-1,-.1318993379,.2697323507e-1,-.6166343343e-1,.3469019838e-1,.3469019838e-1,.2697323507e-1,-.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{.1104870784,.1372943814e-1,.1372943814e-1,-.2604624096e-1,.1058560268,.1058560268,.1503953414e-1,.6279059801e-1,.6279059801e-1,-.6414708202e-1,.6180728788e-1,.6180728788e-1,.8198437210e-1,-.4481847772e-2,-.4481847772e-2,-.4131214635e-1,-.2730516345e-1,-.2730516345e-1,-.8534649204e-1,-.1322296188e-1,-.1322296188e-1,-.3971075014e-1,.6854799631e-1,.2788193386e-1,-.3874809825e-1,-.3874809825e-1,.6854799631e-1,.2788193386e-1,.3642583502e-1,.8712204498e-3,-.5473886199e-1,-.5473886199e-1,.3642583502e-1,.8712204498e-3,.2214798741e-2,-.1009100723e-1,-.1047747017e-1,-.1047747017e-1,.2214798741e-2,-.1009100723e-1,.1265572386e-1,-.6148813860e-1,-.5522608679e-1,-.5522608679e-1,.1265572386e-1,-.6148813860e-1,-.1134779051e-1,-.5533443179e-1,-.2579313736e-1,-.2579313736e-1,-.1134779051e-1,-.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{.1687718001,.2097206318e-1,.2097206318e-1,-.3978629025e-1,.1616977519,.1616977519,.2297326787e-1,.9591422281e-1,.9591422281e-1,-.9798628638e-1,.9441219172e-1,.9441219172e-1,.1252331970,-.6846135558e-2,-.6846135558e-2,-.6310534594e-1,-.4170932612e-1,-.4170932612e-1,-.1303689201,-.2019840792e-1,-.2019840792e-1,-.6065917284e-1,.1047087940,.4259035749e-1,-.5918869776e-1,-.5918869776e-1,.1047087940,.4259035749e-1,.5564138210e-1,.1330811220e-2,-.8361499282e-1,-.8361499282e-1,.5564138210e-1,.1330811220e-2,.3383160961e-2,-.1541426816e-1,-.1600460005e-1,-.1600460005e-1,.3383160961e-2,-.1541426816e-1,.1933193753e-1,-.9392468318e-1,-.8435924102e-1,-.8435924102e-1,.1933193753e-1,-.9392468318e-1,-.1733403633e-1,-.8452474073e-1,-.3939966813e-1,-.3939966813e-1,-.1733403633e-1,-.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{-.6804239411e-1,.1145848830e-1,.1145848830e-1,.2895473345e-1,.3479953926e-2,.3479953926e-2,-.8777788353e-1,.3406513033e-1,.3406513033e-1,.3473009539e-1,-.1014545622,-.1014545622,.9142437836e-1,-.5109385389e-1,-.5109385389e-1,.7796646760e-1,.7600422558e-2,.7600422558e-2,-.4998669741e-1,.1339603089e-1,.1339603089e-1,-.5386293723e-1,-.6480346250e-1,.1027298408,-.1762397985e-1,-.1762397985e-1,-.6480346250e-1,.1027298408,-.2109538461e-1,.4916147521e-1,.5370754828e-1,.5370754828e-1,-.2109538461e-1,.4916147521e-1,-.5671605418e-2,-.1354331144e-2,.1246131332e-1,.1246131332e-1,-.5671605418e-2,-.1354331144e-2,-.4342611086e-1,.1515827781e-1,.4183488198e-1,.4183488198e-1,-.4342611086e-1,.1515827781e-1,.8837167105e-3,-.5277408321e-1,.2665741414e-1,.2665741414e-1,.8837167105e-3,-.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{-.6656449560e-1,-.1342806488e-1,-.1342806488e-1,-.3079509387e-1,-.1065239097,-.1065239097,.9772524850e-1,-.6549601941e-1,-.6549601941e-1,.3360109025e-2,.4114868587e-1,.4114868587e-1,-.7177662214e-1,.7522679396e-1,.7522679396e-1,.1076722274,.1444928058e-1,.1444928058e-1,.1317043874e-1,-.1246734703e-1,-.1246734703e-1,-.6337331851e-1,-.1727352568e-1,.1680736381e-1,.6614705937e-1,.6614705937e-1,-.1727352568e-1,.1680736381e-1,-.2317222740e-1,.2883874010e-1,-.4751234575e-1,-.4751234575e-1,-.2317222740e-1,.2883874010e-1,.4353025737e-2,.8373369061e-2,-.1446295644e-1,-.1446295644e-1,.4353025737e-2,.8373369061e-2,.4030485339e-1,.7811447251e-1,-.2079224911e-1,-.2079224911e-1,.4030485339e-1,.7811447251e-1,.9854506346e-2,-.3162709295e-1,-.2557166486e-1,-.2557166486e-1,.9854506346e-2,-.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,.1891559699e-1,-.1891559699e-1,.3728221333e-11,.6016501311e-1,-.6016501307e-1,.2502617804e-10,.6637503462e-1,-.6637503465e-1,-.1129442338e-10,-.7036268232e-1,.7036268232e-1,-.2527951366e-10,-.1000993614,.1000993615,-0.,-.6591155720e-1,.6591155723e-1,-0.,-.2042456634e-1,.2042456633e-1,-0.,.9759050577e-1,-.3794735317e-1,-.5964315250e-1,.5964315260e-1,-.9759050570e-1,.3794735317e-1,.5710338621e-1,-.2915609201e-1,-.2794729420e-1,.2794729418e-1,-.5710338621e-1,.2915609203e-1,.1151029307e-1,-.2993628541e-2,-.8516664535e-2,.8516664535e-2,-.1151029308e-1,.2993628539e-2,.8322014746e-1,-.1688050365e-1,-.6633964381e-1,.6633964379e-1,-.8322014748e-1,.1688050367e-1,.3641824967e-1,-.2916737971e-2,-.3350151169e-1,.3350151170e-1,-.3641824966e-1,.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,.2889405168e-1,-.2889405167e-1,.5694952154e-11,.9190357559e-1,-.9190357554e-1,.3822811841e-10,.1013895401,-.1013895402,-.1725251670e-10,-.1074807726,.1074807726,-.3861509495e-10,-.1529043002,.1529043003,-0.,-.1006815667,.1006815667,-0.,-.3119904043e-1,.3119904041e-1,-0.,.1490719599,-.5796553947e-1,-.9110642037e-1,.9110642046e-1,-.1490719598,.5796553947e-1,.8722686330e-1,-.4453666621e-1,-.4269019705e-1,.4269019702e-1,-.8722686330e-1,.4453666622e-1,.1758226310e-1,-.4572843131e-2,-.1300941997e-1,.1300941997e-1,-.1758226310e-1,.4572843127e-2,.1271208750,-.2578539525e-1,-.1013354798,.1013354798,-.1271208750,.2578539527e-1,.5562979528e-1,-.4455390845e-2,-.5117440439e-1,.5117440444e-1,-.5562979523e-1,.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,.3444674858e-1,-.3444674856e-1,-.4446512362e-11,.7864938975e-1,-.7864938968e-1,-.4128474942e-11,.1099312557,-.1099312557,.9746388013e-11,-.1747386758e-1,.1747386758e-1,-.4936608124e-10,.3189597418e-1,-.3189597420e-1,-0.,.5026526501e-1,-.5026526506e-1,-0.,.2293909754e-1,-.2293909753e-1,-0.,.5770330047e-1,-.9292114363e-1,.3521784310e-1,-.3521784315e-1,-.5770330042e-1,.9292114363e-1,.4815974411e-1,-.7895094357e-1,.3079119953e-1,-.3079119951e-1,-.4815974411e-1,.7895094362e-1,.3346867218e-3,-.1099182731e-1,.1065714059e-1,-.1065714059e-1,-.3346867220e-3,.1099182730e-1,-.5423095927e-2,-.5878737247e-1,.6421046842e-1,-.6421046842e-1,.5423095927e-2,.5878737252e-1,-.2558351144e-1,-.1251003735e-1,.3809354879e-1,-.3809354879e-1,.2558351144e-1,.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,.5261827761e-1,-.5261827757e-1,-.6792159825e-11,.1201389273,-.1201389272,-.6306349641e-11,.1679227669,-.1679227669,.1488785361e-10,-.2669177363e-1,.2669177363e-1,-.7540793467e-10,.4872190534e-1,-.4872190538e-1,-0.,.7678146058e-1,-.7678146066e-1,-0.,.3504005029e-1,-.3504005026e-1,-0.,.8814324742e-1,-.1419393914,.5379614395e-1,-.5379614402e-1,-.8814324734e-1,.1419393914,.7356522427e-1,-.1205995584,.4703433420e-1,-.4703433416e-1,-.7356522427e-1,.1205995584,.5112424122e-3,-.1679029356e-1,.1627905114e-1,-.1627905114e-1,-.5112424126e-3,.1679029354e-1,-.8283915862e-2,-.8979919475e-1,.9808311064e-1,-.9808311064e-1,.8283915862e-2,.8979919482e-1,-.3907945924e-1,-.1910939770e-1,.5818885694e-1,-.5818885694e-1,.3907945924e-1,.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,.6060877170e-2,-.6060877164e-2,.4944073794e-11,-.3161264212e-1,.3161264209e-1,-.1584800096e-10,.2303508888e-3,-.2303508888e-3,-.5819067272e-11,.6681522508e-1,-.6681522508e-1,-.1528439436e-10,.4470780632e-1,-.4470780632e-1,-0.,-.1994876348e-1,.1994876348e-1,-0.,-.2329919003e-1,.2329919001e-1,-0.,-.9824065005e-1,-.7289231590e-1,.1486374908e-2,-.1486374909e-2,.9824064997e-1,.7289231590e-1,-.5309591204e-1,-.8058321394e-1,-.3033302763e-1,.3033302759e-1,.5309591204e-1,.8058321395e-1,-.9386233204e-2,-.2045354015e-1,-.1267656516e-1,.1267656516e-1,.9386233207e-2,.2045354012e-1,-.6111843873e-1,-.9989476013e-1,-.5006525860e-1,.5006525859e-1,.6111843873e-1,.9989476021e-1,.6661205800e-2,-.2978294282e-1,-.3944227928e-1,.3944227931e-1,-.6661205793e-2,.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,-.2989961082e-1,.2989961082e-1,-.5259821245e-11,-.7039060152e-1,.7039060149e-1,.2191617071e-10,-.1013287354,.1013287354,.6279229509e-12,-.4006762239e-1,.4006762239e-1,.3934858683e-10,-.8551663720e-1,.8551663723e-1,-0.,-.1484968306e-1,.1484968307e-1,-0.,.2177358772e-1,-.2177358771e-1,-0.,.4790109961e-2,.2984596792e-1,-.3463607749e-1,.3463607753e-1,-.4790109958e-2,-.2984596792e-1,-.2283783488e-1,-.4181200571e-2,.2701903535e-1,-.2701903532e-1,.2283783488e-1,.4181200574e-2,.9086687153e-2,-.2380170174e-1,.1471501461e-1,-.1471501461e-1,-.9086687153e-2,.2380170172e-1,.7038491177e-1,-.9773979388e-1,.2735488204e-1,-.2735488204e-1,-.7038491177e-1,.9773979394e-1,.1336506945e-1,-.5130832062e-1,.3794325106e-1,-.3794325110e-1,-.1336506945e-1,.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{.9882264718e-1,-.1440235391e-1,-.1440235396e-1,.3006820721e-1,.3375895322e-1,.3375895287e-1,.1352948785,-.2899076627e-1,-.2899076622e-1,.1129298746,.6881731865e-1,.6881731855e-1,.4625865234e-1,-.2871520174e-1,-.2871520191e-1,.1246232363e-1,-.6335778756e-1,-.6335778749e-1,.1534390562e-2,-.2959552517e-1,-.2959552520e-1,.1872238370e-4,-.3794484055e-1,.7521199400e-2,.8201583980e-1,.8201584015e-1,-.3794484065e-1,.7521199317e-2,-.5242564114e-1,-.1267437352e-1,.4949958872e-1,.4949958862e-1,-.5242564114e-1,-.1267437351e-1,-.1478824220e-1,-.1029932873e-2,-.5978541514e-3,-.5978541492e-3,-.1478824222e-1,-.1029932873e-2,-.8395409436e-1,-.1058736005e-2,-.8059922862e-2,-.8059922754e-2,-.8395409436e-1,-.1058736043e-2,-.4934714167e-1,-.1685816455e-3,-.3319675284e-1,-.3319675290e-1,-.4934714164e-1,-.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{.1509540871,-.2199995900e-1,-.2199995907e-1,.4592994519e-1,.5156765287e-1,.5156765232e-1,.2066663408,-.4428412697e-1,-.4428412690e-1,.1725032329,.1051201906,.1051201905,.7066125868e-1,-.4386319521e-1,-.4386319547e-1,.1903651379e-1,-.9678061914e-1,-.9678061909e-1,.2343820299e-2,-.4520791146e-1,-.4520791150e-1,.2859891351e-4,-.5796170139e-1,.1148882186e-1,.1252812647,.1252812653,-.5796170154e-1,.1148882174e-1,-.8008148967e-1,-.1936042536e-1,.7561187074e-1,.7561187061e-1,-.8008148966e-1,-.1936042534e-1,-.2258941311e-1,-.1573248451e-2,-.9132373015e-3,-.9132372980e-3,-.2258941314e-1,-.1573248449e-2,-.1282419975,-.1617245962e-2,-.1231173554e-1,-.1231173538e-1,-.1282419975,-.1617246020e-2,-.7537900402e-1,-.2575127172e-3,-.5070887761e-1,-.5070887768e-1,-.7537900399e-1,-.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{.1521474186,-.4127107599e-1,-.4127107613e-1,-.3344651713e-1,.7384207728e-1,.7384207651e-1,.3303033204e-1,-.7676670968e-1,-.7676670952e-1,-.7646932810e-1,.5434312853e-1,.5434312844e-1,.1406012128,-.1185729266e-2,-.1185729273e-2,.5653180033e-1,.3451242438e-1,.3451242437e-1,.8758681134e-2,.3020381423e-1,.3020381426e-1,.1246893713e-3,-.4714000659e-1,.2778316412e-1,-.2604070049e-1,-.2604070059e-1,-.4714000670e-1,.2778316383e-1,-.8257470907e-1,-.5115488591e-1,-.4912182046e-1,-.4912182038e-1,-.8257470907e-1,-.5115488588e-1,-.7397967795e-2,-.5467224064e-2,.7113199607e-3,.7113199580e-3,-.7397967806e-2,-.5467224056e-2,-.3155548387e-1,-.5355748779e-2,.6574391344e-2,.6574391251e-2,-.3155548387e-1,-.5355748971e-2,.2296424900e-1,-.1032248755e-2,.3449059684e-1,.3449059690e-1,.2296424899e-1,-.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{-.4252649632e-1,-.3621722030e-1,-.3621722042e-1,.3560242460e-1,.2348842407e-1,.2348842382e-1,-.1206537327,-.5391680609e-1,-.5391680599e-1,.2629256125e-1,-.6802515129e-1,-.6802515124e-1,.1427717610,.2147433158e-1,.2147433171e-1,.1219618694,.1982044929e-3,.1982044929e-3,.2648049029e-1,-.2844076554e-1,-.2844076556e-1,.4636110040e-3,.2835644722e-1,.4230017603e-1,-.2869463558e-1,-.2869463573e-1,.2835644729e-1,.4230017559e-1,.2013663457e-1,-.9116808757e-1,.4429713195e-1,.4429713188e-1,.2013663457e-1,-.9116808747e-1,.1463911665e-1,-.1492260701e-1,-.8259580003e-3,-.8259579974e-3,.1463911667e-1,-.1492260700e-1,.8008479954e-1,-.1363156197e-1,-.4081583933e-2,-.4081583877e-2,.8008479954e-1,-.1363156247e-1,.6542647961e-2,-.3434213984e-2,-.3335556699e-1,-.3335556704e-1,.6542647959e-2,-.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,.9258142764e-2,-.9258142654e-2,.8136908247e-10,-.4828910902e-1,.4828910902e-1,-0.,.3518672753e-3,-.3518668221e-3,.3485534728e-10,.1020619419,-.1020619425,-.5460999099e-9,.6829230417e-1,-.6829230235e-1,-0.,-.3047223942e-1,.3047223960e-1,-.3681086362e-9,-.3559010093e-1,.3559010063e-1,.4643401911e-10,-.2666533469e-1,.1205488497e-1,.1256702126,-.1256702128,.2666533393e-1,-.1205488583e-1,.3819594183e-1,-.3791605562e-2,.7296682180e-1,-.7296682193e-1,-.3819594232e-1,.3791605423e-2,.1658837758e-1,-.3172129929e-3,.1156231241e-1,-.1156231248e-1,-.1658837783e-1,.3172129864e-3,.6017702823e-1,.9452189607e-3,.7706103991e-1,-.7706103958e-1,-.6017702852e-1,-.9452190353e-3,.5568378542e-1,.1442892931e-4,-.1474045160e-1,.1474045137e-1,-.5568378595e-1,-.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,.3598418982e-1,-.3598418942e-1,-.8668398183e-10,-.1469158467,.1469158467,-0.,.1272304336e-2,-.1272302698e-2,-.1824763744e-10,.1329348046,-.1329348053,-.2240768661e-8,.2532572946e-1,-.2532572881e-1,-0.,.1092978850e-1,-.1092978857e-1,-.2733414349e-8,.3384746400e-1,-.3384746369e-1,.3998097089e-9,-.4975749160e-1,.5928626767e-1,-.9528779263e-2,.9528779277e-2,.4975749017e-1,-.5928627186e-1,.8709586801e-1,-.2026201294e-1,-.6683385514e-1,.6683385532e-1,-.8709586914e-1,.2026201219e-1,.1562966645e-1,-.2197560437e-2,-.1343210582e-1,.1343210590e-1,-.1562966667e-1,.2197560392e-2,.4738032804e-1,.6254379410e-2,-.5363470751e-1,.5363470732e-1,-.4738032827e-1,-.6254379903e-2,-.1447133378e-1,.1145953765e-3,.1435673898e-1,-.1435673875e-1,.1447133392e-1,-.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{-.5705528195e-1,-.5381888614e-2,-.5381890600e-2,-.3865155911e-1,.1066851510e-1,.1066851160e-1,-.1201849450,.1023929481e-1,.1023929537e-1,-.1298773899,.1146344760e-1,.1146344311e-1,-.1390909759e-1,.1113342295,.1113342303,-.1417761192e-2,.1805227680e-1,.1805227549e-1,-.4540571627e-4,-.3784627926e-1,-.3784627821e-1,-.2688892560e-7,.5375088148e-1,.3596809691e-2,-.1049770374e-2,-.1049766066e-2,.5375088190e-1,.3596800240e-2,-.2284622311e-1,-.5640425826e-3,-.1245758952e-1,-.1245758914e-1,-.2284622668e-1,-.5640434935e-3,-.1743701782e-1,-.9341106461e-4,.1604194740e-1,.1604194746e-1,-.1743701764e-1,-.9341075247e-4,-.2535811540e-1,.3827141178e-3,.1086557567,.1086557557,-.2535811250e-1,.3827130864e-3,-.5582706905e-1,.4444677760e-5,.1524183760e-1,.1524183769e-1,-.5582706642e-1,.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,-.1092092502e-1,-.1092092502e-1,.2184185003e-1,-.3473628658e-1,-.3473628658e-1,.6947257305e-1,-.3832164410e-1,-.3832164410e-1,.7664328825e-1,.4062391359e-1,.4062391359e-1,-.8124782718e-1,.5779239332e-1,.5779239332e-1,-.1155847866,.3805405532e-1,.3805405532e-1,-.7610811063e-1,.1179212887e-1,.1179212887e-1,-.2358425774e-1,.1252607563e-1,-.9077889505e-1,.7825281938e-1,.7825281938e-1,.1252607563e-1,-.9077889505e-1,-.6978997413e-3,-.4910403324e-1,.4980193299e-1,.4980193299e-1,-.6978997413e-3,-.4910403324e-1,.3188726319e-2,-.1156256937e-1,.8373843054e-2,.8373843054e-2,.3188726319e-2,-.1156256937e-1,.2855524789e-1,-.8634838574e-1,.5779313787e-1,.5779313787e-1,.2855524789e-1,-.8634838574e-1,.1765812736e-1,-.4036819305e-1,.2271006571e-1,.2271006571e-1,.1765812736e-1,-.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,-.1668198852e-1,-.1668198852e-1,.3336397701e-1,-.5306055418e-1,-.5306055418e-1,.1061211082,-.5853727826e-1,-.5853727826e-1,.1170745566,.6205405300e-1,.6205405300e-1,-.1241081060,.8827933896e-1,.8827933896e-1,-.1765586779,.5812852966e-1,.5812852966e-1,-.1162570593,.1801277438e-1,.1801277438e-1,-.3602554875e-1,.1913389657e-1,-.1386670527,.1195331560,.1195331560,.1913389657e-1,-.1386670527,-.1066059464e-2,-.7500764972e-1,.7607370920e-1,.7607370920e-1,-.1066059464e-2,-.7500764972e-1,.4870859907e-2,-.1766211645e-1,.1279125655e-1,.1279125655e-1,.4870859907e-2,-.1766211645e-1,.4361886163e-1,-.1318993379,.8828047629e-1,.8828047629e-1,.4361886163e-1,-.1318993379,.2697323507e-1,-.6166343343e-1,.3469019838e-1,.3469019838e-1,.2697323507e-1,-.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{.1104870784,.1372943814e-1,.1372943814e-1,-.2604624096e-1,.1058560268,.1058560268,.1503953414e-1,.6279059801e-1,.6279059801e-1,-.6414708202e-1,.6180728788e-1,.6180728788e-1,.8198437210e-1,-.4481847772e-2,-.4481847772e-2,-.4131214635e-1,-.2730516345e-1,-.2730516345e-1,-.8534649204e-1,-.1322296188e-1,-.1322296188e-1,-.3971075014e-1,.6854799631e-1,.2788193386e-1,-.3874809825e-1,-.3874809825e-1,.6854799631e-1,.2788193386e-1,.3642583502e-1,.8712204498e-3,-.5473886199e-1,-.5473886199e-1,.3642583502e-1,.8712204498e-3,.2214798741e-2,-.1009100723e-1,-.1047747017e-1,-.1047747017e-1,.2214798741e-2,-.1009100723e-1,.1265572386e-1,-.6148813860e-1,-.5522608679e-1,-.5522608679e-1,.1265572386e-1,-.6148813860e-1,-.1134779051e-1,-.5533443179e-1,-.2579313736e-1,-.2579313736e-1,-.1134779051e-1,-.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{.1687718001,.2097206318e-1,.2097206318e-1,-.3978629025e-1,.1616977519,.1616977519,.2297326787e-1,.9591422281e-1,.9591422281e-1,-.9798628638e-1,.9441219172e-1,.9441219172e-1,.1252331970,-.6846135558e-2,-.6846135558e-2,-.6310534594e-1,-.4170932612e-1,-.4170932612e-1,-.1303689201,-.2019840792e-1,-.2019840792e-1,-.6065917284e-1,.1047087940,.4259035749e-1,-.5918869776e-1,-.5918869776e-1,.1047087940,.4259035749e-1,.5564138210e-1,.1330811220e-2,-.8361499282e-1,-.8361499282e-1,.5564138210e-1,.1330811220e-2,.3383160961e-2,-.1541426816e-1,-.1600460005e-1,-.1600460005e-1,.3383160961e-2,-.1541426816e-1,.1933193753e-1,-.9392468318e-1,-.8435924102e-1,-.8435924102e-1,.1933193753e-1,-.9392468318e-1,-.1733403633e-1,-.8452474073e-1,-.3939966813e-1,-.3939966813e-1,-.1733403633e-1,-.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{-.6804239411e-1,.1145848830e-1,.1145848830e-1,.2895473345e-1,.3479953926e-2,.3479953926e-2,-.8777788353e-1,.3406513033e-1,.3406513033e-1,.3473009539e-1,-.1014545622,-.1014545622,.9142437836e-1,-.5109385389e-1,-.5109385389e-1,.7796646760e-1,.7600422558e-2,.7600422558e-2,-.4998669741e-1,.1339603089e-1,.1339603089e-1,-.5386293723e-1,-.6480346250e-1,.1027298408,-.1762397985e-1,-.1762397985e-1,-.6480346250e-1,.1027298408,-.2109538461e-1,.4916147521e-1,.5370754828e-1,.5370754828e-1,-.2109538461e-1,.4916147521e-1,-.5671605418e-2,-.1354331144e-2,.1246131332e-1,.1246131332e-1,-.5671605418e-2,-.1354331144e-2,-.4342611086e-1,.1515827781e-1,.4183488198e-1,.4183488198e-1,-.4342611086e-1,.1515827781e-1,.8837167105e-3,-.5277408321e-1,.2665741414e-1,.2665741414e-1,.8837167105e-3,-.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{-.6656449560e-1,-.1342806488e-1,-.1342806488e-1,-.3079509387e-1,-.1065239097,-.1065239097,.9772524850e-1,-.6549601941e-1,-.6549601941e-1,.3360109025e-2,.4114868587e-1,.4114868587e-1,-.7177662214e-1,.7522679396e-1,.7522679396e-1,.1076722274,.1444928058e-1,.1444928058e-1,.1317043874e-1,-.1246734703e-1,-.1246734703e-1,-.6337331851e-1,-.1727352568e-1,.1680736381e-1,.6614705937e-1,.6614705937e-1,-.1727352568e-1,.1680736381e-1,-.2317222740e-1,.2883874010e-1,-.4751234575e-1,-.4751234575e-1,-.2317222740e-1,.2883874010e-1,.4353025737e-2,.8373369061e-2,-.1446295644e-1,-.1446295644e-1,.4353025737e-2,.8373369061e-2,.4030485339e-1,.7811447251e-1,-.2079224911e-1,-.2079224911e-1,.4030485339e-1,.7811447251e-1,.9854506346e-2,-.3162709295e-1,-.2557166486e-1,-.2557166486e-1,.9854506346e-2,-.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,.1891559699e-1,-.1891559699e-1,.3728221333e-11,.6016501311e-1,-.6016501307e-1,.2502617804e-10,.6637503462e-1,-.6637503465e-1,-.1129442338e-10,-.7036268232e-1,.7036268232e-1,-.2527951366e-10,-.1000993614,.1000993615,-0.,-.6591155720e-1,.6591155723e-1,-0.,-.2042456634e-1,.2042456633e-1,-0.,.9759050577e-1,-.3794735317e-1,-.5964315250e-1,.5964315260e-1,-.9759050570e-1,.3794735317e-1,.5710338621e-1,-.2915609201e-1,-.2794729420e-1,.2794729418e-1,-.5710338621e-1,.2915609203e-1,.1151029307e-1,-.2993628541e-2,-.8516664535e-2,.8516664535e-2,-.1151029308e-1,.2993628539e-2,.8322014746e-1,-.1688050365e-1,-.6633964381e-1,.6633964379e-1,-.8322014748e-1,.1688050367e-1,.3641824967e-1,-.2916737971e-2,-.3350151169e-1,.3350151170e-1,-.3641824966e-1,.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,.2889405168e-1,-.2889405167e-1,.5694952154e-11,.9190357559e-1,-.9190357554e-1,.3822811841e-10,.1013895401,-.1013895402,-.1725251670e-10,-.1074807726,.1074807726,-.3861509495e-10,-.1529043002,.1529043003,-0.,-.1006815667,.1006815667,-0.,-.3119904043e-1,.3119904041e-1,-0.,.1490719599,-.5796553947e-1,-.9110642037e-1,.9110642046e-1,-.1490719598,.5796553947e-1,.8722686330e-1,-.4453666621e-1,-.4269019705e-1,.4269019702e-1,-.8722686330e-1,.4453666622e-1,.1758226310e-1,-.4572843131e-2,-.1300941997e-1,.1300941997e-1,-.1758226310e-1,.4572843127e-2,.1271208750,-.2578539525e-1,-.1013354798,.1013354798,-.1271208750,.2578539527e-1,.5562979528e-1,-.4455390845e-2,-.5117440439e-1,.5117440444e-1,-.5562979523e-1,.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,.3444674858e-1,-.3444674856e-1,-.4446512362e-11,.7864938975e-1,-.7864938968e-1,-.4128474942e-11,.1099312557,-.1099312557,.9746388013e-11,-.1747386758e-1,.1747386758e-1,-.4936608124e-10,.3189597418e-1,-.3189597420e-1,-0.,.5026526501e-1,-.5026526506e-1,-0.,.2293909754e-1,-.2293909753e-1,-0.,.5770330047e-1,-.9292114363e-1,.3521784310e-1,-.3521784315e-1,-.5770330042e-1,.9292114363e-1,.4815974411e-1,-.7895094357e-1,.3079119953e-1,-.3079119951e-1,-.4815974411e-1,.7895094362e-1,.3346867218e-3,-.1099182731e-1,.1065714059e-1,-.1065714059e-1,-.3346867220e-3,.1099182730e-1,-.5423095927e-2,-.5878737247e-1,.6421046842e-1,-.6421046842e-1,.5423095927e-2,.5878737252e-1,-.2558351144e-1,-.1251003735e-1,.3809354879e-1,-.3809354879e-1,.2558351144e-1,.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,.5261827761e-1,-.5261827757e-1,-.6792159825e-11,.1201389273,-.1201389272,-.6306349641e-11,.1679227669,-.1679227669,.1488785361e-10,-.2669177363e-1,.2669177363e-1,-.7540793467e-10,.4872190534e-1,-.4872190538e-1,-0.,.7678146058e-1,-.7678146066e-1,-0.,.3504005029e-1,-.3504005026e-1,-0.,.8814324742e-1,-.1419393914,.5379614395e-1,-.5379614402e-1,-.8814324734e-1,.1419393914,.7356522427e-1,-.1205995584,.4703433420e-1,-.4703433416e-1,-.7356522427e-1,.1205995584,.5112424122e-3,-.1679029356e-1,.1627905114e-1,-.1627905114e-1,-.5112424126e-3,.1679029354e-1,-.8283915862e-2,-.8979919475e-1,.9808311064e-1,-.9808311064e-1,.8283915862e-2,.8979919482e-1,-.3907945924e-1,-.1910939770e-1,.5818885694e-1,-.5818885694e-1,.3907945924e-1,.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,.6060877170e-2,-.6060877164e-2,.4944073794e-11,-.3161264212e-1,.3161264209e-1,-.1584800096e-10,.2303508888e-3,-.2303508888e-3,-.5819067272e-11,.6681522508e-1,-.6681522508e-1,-.1528439436e-10,.4470780632e-1,-.4470780632e-1,-0.,-.1994876348e-1,.1994876348e-1,-0.,-.2329919003e-1,.2329919001e-1,-0.,-.9824065005e-1,-.7289231590e-1,.1486374908e-2,-.1486374909e-2,.9824064997e-1,.7289231590e-1,-.5309591204e-1,-.8058321394e-1,-.3033302763e-1,.3033302759e-1,.5309591204e-1,.8058321395e-1,-.9386233204e-2,-.2045354015e-1,-.1267656516e-1,.1267656516e-1,.9386233207e-2,.2045354012e-1,-.6111843873e-1,-.9989476013e-1,-.5006525860e-1,.5006525859e-1,.6111843873e-1,.9989476021e-1,.6661205800e-2,-.2978294282e-1,-.3944227928e-1,.3944227931e-1,-.6661205793e-2,.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,-.2989961082e-1,.2989961082e-1,-.5259821245e-11,-.7039060152e-1,.7039060149e-1,.2191617071e-10,-.1013287354,.1013287354,.6279229509e-12,-.4006762239e-1,.4006762239e-1,.3934858683e-10,-.8551663720e-1,.8551663723e-1,-0.,-.1484968306e-1,.1484968307e-1,-0.,.2177358772e-1,-.2177358771e-1,-0.,.4790109961e-2,.2984596792e-1,-.3463607749e-1,.3463607753e-1,-.4790109958e-2,-.2984596792e-1,-.2283783488e-1,-.4181200571e-2,.2701903535e-1,-.2701903532e-1,.2283783488e-1,.4181200574e-2,.9086687153e-2,-.2380170174e-1,.1471501461e-1,-.1471501461e-1,-.9086687153e-2,.2380170172e-1,.7038491177e-1,-.9773979388e-1,.2735488204e-1,-.2735488204e-1,-.7038491177e-1,.9773979394e-1,.1336506945e-1,-.5130832062e-1,.3794325106e-1,-.3794325110e-1,-.1336506945e-1,.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{.9882264718e-1,-.1440235391e-1,-.1440235396e-1,.3006820721e-1,.3375895322e-1,.3375895287e-1,.1352948785,-.2899076627e-1,-.2899076622e-1,.1129298746,.6881731865e-1,.6881731855e-1,.4625865234e-1,-.2871520174e-1,-.2871520191e-1,.1246232363e-1,-.6335778756e-1,-.6335778749e-1,.1534390562e-2,-.2959552517e-1,-.2959552520e-1,.1872238370e-4,-.3794484055e-1,.7521199400e-2,.8201583980e-1,.8201584015e-1,-.3794484065e-1,.7521199317e-2,-.5242564114e-1,-.1267437352e-1,.4949958872e-1,.4949958862e-1,-.5242564114e-1,-.1267437351e-1,-.1478824220e-1,-.1029932873e-2,-.5978541514e-3,-.5978541492e-3,-.1478824222e-1,-.1029932873e-2,-.8395409436e-1,-.1058736005e-2,-.8059922862e-2,-.8059922754e-2,-.8395409436e-1,-.1058736043e-2,-.4934714167e-1,-.1685816455e-3,-.3319675284e-1,-.3319675290e-1,-.4934714164e-1,-.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{.1509540871,-.2199995900e-1,-.2199995907e-1,.4592994519e-1,.5156765287e-1,.5156765232e-1,.2066663408,-.4428412697e-1,-.4428412690e-1,.1725032329,.1051201906,.1051201905,.7066125868e-1,-.4386319521e-1,-.4386319547e-1,.1903651379e-1,-.9678061914e-1,-.9678061909e-1,.2343820299e-2,-.4520791146e-1,-.4520791150e-1,.2859891351e-4,-.5796170139e-1,.1148882186e-1,.1252812647,.1252812653,-.5796170154e-1,.1148882174e-1,-.8008148967e-1,-.1936042536e-1,.7561187074e-1,.7561187061e-1,-.8008148966e-1,-.1936042534e-1,-.2258941311e-1,-.1573248451e-2,-.9132373015e-3,-.9132372980e-3,-.2258941314e-1,-.1573248449e-2,-.1282419975,-.1617245962e-2,-.1231173554e-1,-.1231173538e-1,-.1282419975,-.1617246020e-2,-.7537900402e-1,-.2575127172e-3,-.5070887761e-1,-.5070887768e-1,-.7537900399e-1,-.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{.1521474186,-.4127107599e-1,-.4127107613e-1,-.3344651713e-1,.7384207728e-1,.7384207651e-1,.3303033204e-1,-.7676670968e-1,-.7676670952e-1,-.7646932810e-1,.5434312853e-1,.5434312844e-1,.1406012128,-.1185729266e-2,-.1185729273e-2,.5653180033e-1,.3451242438e-1,.3451242437e-1,.8758681134e-2,.3020381423e-1,.3020381426e-1,.1246893713e-3,-.4714000659e-1,.2778316412e-1,-.2604070049e-1,-.2604070059e-1,-.4714000670e-1,.2778316383e-1,-.8257470907e-1,-.5115488591e-1,-.4912182046e-1,-.4912182038e-1,-.8257470907e-1,-.5115488588e-1,-.7397967795e-2,-.5467224064e-2,.7113199607e-3,.7113199580e-3,-.7397967806e-2,-.5467224056e-2,-.3155548387e-1,-.5355748779e-2,.6574391344e-2,.6574391251e-2,-.3155548387e-1,-.5355748971e-2,.2296424900e-1,-.1032248755e-2,.3449059684e-1,.3449059690e-1,.2296424899e-1,-.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{-.4252649632e-1,-.3621722030e-1,-.3621722042e-1,.3560242460e-1,.2348842407e-1,.2348842382e-1,-.1206537327,-.5391680609e-1,-.5391680599e-1,.2629256125e-1,-.6802515129e-1,-.6802515124e-1,.1427717610,.2147433158e-1,.2147433171e-1,.1219618694,.1982044929e-3,.1982044929e-3,.2648049029e-1,-.2844076554e-1,-.2844076556e-1,.4636110040e-3,.2835644722e-1,.4230017603e-1,-.2869463558e-1,-.2869463573e-1,.2835644729e-1,.4230017559e-1,.2013663457e-1,-.9116808757e-1,.4429713195e-1,.4429713188e-1,.2013663457e-1,-.9116808747e-1,.1463911665e-1,-.1492260701e-1,-.8259580003e-3,-.8259579974e-3,.1463911667e-1,-.1492260700e-1,.8008479954e-1,-.1363156197e-1,-.4081583933e-2,-.4081583877e-2,.8008479954e-1,-.1363156247e-1,.6542647961e-2,-.3434213984e-2,-.3335556699e-1,-.3335556704e-1,.6542647959e-2,-.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,.9258142764e-2,-.9258142654e-2,.8136908247e-10,-.4828910902e-1,.4828910902e-1,-0.,.3518672753e-3,-.3518668221e-3,.3485534728e-10,.1020619419,-.1020619425,-.5460999099e-9,.6829230417e-1,-.6829230235e-1,-0.,-.3047223942e-1,.3047223960e-1,-.3681086362e-9,-.3559010093e-1,.3559010063e-1,.4643401911e-10,-.2666533469e-1,.1205488497e-1,.1256702126,-.1256702128,.2666533393e-1,-.1205488583e-1,.3819594183e-1,-.3791605562e-2,.7296682180e-1,-.7296682193e-1,-.3819594232e-1,.3791605423e-2,.1658837758e-1,-.3172129929e-3,.1156231241e-1,-.1156231248e-1,-.1658837783e-1,.3172129864e-3,.6017702823e-1,.9452189607e-3,.7706103991e-1,-.7706103958e-1,-.6017702852e-1,-.9452190353e-3,.5568378542e-1,.1442892931e-4,-.1474045160e-1,.1474045137e-1,-.5568378595e-1,-.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,.3598418982e-1,-.3598418942e-1,-.8668398183e-10,-.1469158467,.1469158467,-0.,.1272304336e-2,-.1272302698e-2,-.1824763744e-10,.1329348046,-.1329348053,-.2240768661e-8,.2532572946e-1,-.2532572881e-1,-0.,.1092978850e-1,-.1092978857e-1,-.2733414349e-8,.3384746400e-1,-.3384746369e-1,.3998097089e-9,-.4975749160e-1,.5928626767e-1,-.9528779263e-2,.9528779277e-2,.4975749017e-1,-.5928627186e-1,.8709586801e-1,-.2026201294e-1,-.6683385514e-1,.6683385532e-1,-.8709586914e-1,.2026201219e-1,.1562966645e-1,-.2197560437e-2,-.1343210582e-1,.1343210590e-1,-.1562966667e-1,.2197560392e-2,.4738032804e-1,.6254379410e-2,-.5363470751e-1,.5363470732e-1,-.4738032827e-1,-.6254379903e-2,-.1447133378e-1,.1145953765e-3,.1435673898e-1,-.1435673875e-1,.1447133392e-1,-.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{-.5705528195e-1,-.5381888614e-2,-.5381890600e-2,-.3865155911e-1,.1066851510e-1,.1066851160e-1,-.1201849450,.1023929481e-1,.1023929537e-1,-.1298773899,.1146344760e-1,.1146344311e-1,-.1390909759e-1,.1113342295,.1113342303,-.1417761192e-2,.1805227680e-1,.1805227549e-1,-.4540571627e-4,-.3784627926e-1,-.3784627821e-1,-.2688892560e-7,.5375088148e-1,.3596809691e-2,-.1049770374e-2,-.1049766066e-2,.5375088190e-1,.3596800240e-2,-.2284622311e-1,-.5640425826e-3,-.1245758952e-1,-.1245758914e-1,-.2284622668e-1,-.5640434935e-3,-.1743701782e-1,-.9341106461e-4,.1604194740e-1,.1604194746e-1,-.1743701764e-1,-.9341075247e-4,-.2535811540e-1,.3827141178e-3,.1086557567,.1086557557,-.2535811250e-1,.3827130864e-3,-.5582706905e-1,.4444677760e-5,.1524183760e-1,.1524183769e-1,-.5582706642e-1,.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{.1148215401,.1569113387e-1,.1569113387e-1,.1569113387e-1,.1021689440,.1021689440,.1021689440,.6587041484e-1,.6587041484e-1,.6587041484e-1,.1032031823,.1032031823,.1032031823,.7348534591e-1,.7348534591e-1,.7348534591e-1,.3478299668e-1,.3478299668e-1,.3478299668e-1,.8775204783e-2,.8775204783e-2,.8775204783e-2,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{.1753927996,.2396860291e-1,.2396860291e-1,.2396860291e-1,.1560656399,.1560656399,.1560656399,.1006187207,.1006187207,.1006187207,.1576454650,.1576454650,.1576454650,.1122507200,.1122507200,.1122507200,.5313190506e-1,.5313190506e-1,.5313190506e-1,.1340434672e-1,.1340434672e-1,.1340434672e-1,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{.2198665100,.3004623384e-1,.3004623384e-1,.3004623384e-1,.1956386332,.1956386332,.1956386332,.1261322415,.1261322415,.1261322415,.1976190488,.1976190488,.1976190488,.1407137244,.1407137244,.1407137244,.6660436781e-1,.6660436781e-1,.6660436781e-1,.1680323787e-1,.1680323787e-1,.1680323787e-1,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,.1092092502e-1,.1092092502e-1,-.2184185003e-1,.3473628658e-1,.3473628658e-1,-.6947257305e-1,.3832164410e-1,.3832164410e-1,-.7664328825e-1,-.4062391359e-1,-.4062391359e-1,.8124782718e-1,-.5779239332e-1,-.5779239332e-1,.1155847866,-.3805405532e-1,-.3805405532e-1,.7610811063e-1,-.1179212887e-1,-.1179212887e-1,.2358425774e-1,-.1252607563e-1,.9077889505e-1,-.7825281938e-1,-.7825281938e-1,-.1252607563e-1,.9077889505e-1,.6978997413e-3,.4910403324e-1,-.4980193299e-1,-.4980193299e-1,.6978997413e-3,.4910403324e-1,-.3188726319e-2,.1156256937e-1,-.8373843054e-2,-.8373843054e-2,-.3188726319e-2,.1156256937e-1,-.2855524789e-1,.8634838574e-1,-.5779313787e-1,-.5779313787e-1,-.2855524789e-1,.8634838574e-1,-.1765812736e-1,.4036819305e-1,-.2271006571e-1,-.2271006571e-1,-.1765812736e-1,.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,.1668198852e-1,.1668198852e-1,-.3336397701e-1,.5306055418e-1,.5306055418e-1,-.1061211082,.5853727826e-1,.5853727826e-1,-.1170745566,-.6205405300e-1,-.6205405300e-1,.1241081060,-.8827933896e-1,-.8827933896e-1,.1765586779,-.5812852966e-1,-.5812852966e-1,.1162570593,-.1801277438e-1,-.1801277438e-1,.3602554875e-1,-.1913389657e-1,.1386670527,-.1195331560,-.1195331560,-.1913389657e-1,.1386670527,.1066059464e-2,.7500764972e-1,-.7607370920e-1,-.7607370920e-1,.1066059464e-2,.7500764972e-1,-.4870859907e-2,.1766211645e-1,-.1279125655e-1,-.1279125655e-1,-.4870859907e-2,.1766211645e-1,-.4361886163e-1,.1318993379,-.8828047629e-1,-.8828047629e-1,-.4361886163e-1,.1318993379,-.2697323507e-1,.6166343343e-1,-.3469019838e-1,-.3469019838e-1,-.2697323507e-1,.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{-.1104870784,-.1372943814e-1,-.1372943814e-1,.2604624096e-1,-.1058560268,-.1058560268,-.1503953414e-1,-.6279059801e-1,-.6279059801e-1,.6414708202e-1,-.6180728788e-1,-.6180728788e-1,-.8198437210e-1,.4481847772e-2,.4481847772e-2,.4131214635e-1,.2730516345e-1,.2730516345e-1,.8534649204e-1,.1322296188e-1,.1322296188e-1,.3971075014e-1,-.6854799631e-1,-.2788193386e-1,.3874809825e-1,.3874809825e-1,-.6854799631e-1,-.2788193386e-1,-.3642583502e-1,-.8712204498e-3,.5473886199e-1,.5473886199e-1,-.3642583502e-1,-.8712204498e-3,-.2214798741e-2,.1009100723e-1,.1047747017e-1,.1047747017e-1,-.2214798741e-2,.1009100723e-1,-.1265572386e-1,.6148813860e-1,.5522608679e-1,.5522608679e-1,-.1265572386e-1,.6148813860e-1,.1134779051e-1,.5533443179e-1,.2579313736e-1,.2579313736e-1,.1134779051e-1,.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{-.1687718001,-.2097206318e-1,-.2097206318e-1,.3978629025e-1,-.1616977519,-.1616977519,-.2297326787e-1,-.9591422281e-1,-.9591422281e-1,.9798628638e-1,-.9441219172e-1,-.9441219172e-1,-.1252331970,.6846135558e-2,.6846135558e-2,.6310534594e-1,.4170932612e-1,.4170932612e-1,.1303689201,.2019840792e-1,.2019840792e-1,.6065917284e-1,-.1047087940,-.4259035749e-1,.5918869776e-1,.5918869776e-1,-.1047087940,-.4259035749e-1,-.5564138210e-1,-.1330811220e-2,.8361499282e-1,.8361499282e-1,-.5564138210e-1,-.1330811220e-2,-.3383160961e-2,.1541426816e-1,.1600460005e-1,.1600460005e-1,-.3383160961e-2,.1541426816e-1,-.1933193753e-1,.9392468318e-1,.8435924102e-1,.8435924102e-1,-.1933193753e-1,.9392468318e-1,.1733403633e-1,.8452474073e-1,.3939966813e-1,.3939966813e-1,.1733403633e-1,.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{.6804239411e-1,-.1145848830e-1,-.1145848830e-1,-.2895473345e-1,-.3479953926e-2,-.3479953926e-2,.8777788353e-1,-.3406513033e-1,-.3406513033e-1,-.3473009539e-1,.1014545622,.1014545622,-.9142437836e-1,.5109385389e-1,.5109385389e-1,-.7796646760e-1,-.7600422558e-2,-.7600422558e-2,.4998669741e-1,-.1339603089e-1,-.1339603089e-1,.5386293723e-1,.6480346250e-1,-.1027298408,.1762397985e-1,.1762397985e-1,.6480346250e-1,-.1027298408,.2109538461e-1,-.4916147521e-1,-.5370754828e-1,-.5370754828e-1,.2109538461e-1,-.4916147521e-1,.5671605418e-2,.1354331144e-2,-.1246131332e-1,-.1246131332e-1,.5671605418e-2,.1354331144e-2,.4342611086e-1,-.1515827781e-1,-.4183488198e-1,-.4183488198e-1,.4342611086e-1,-.1515827781e-1,-.8837167105e-3,.5277408321e-1,-.2665741414e-1,-.2665741414e-1,-.8837167105e-3,.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{.6656449560e-1,.1342806488e-1,.1342806488e-1,.3079509387e-1,.1065239097,.1065239097,-.9772524850e-1,.6549601941e-1,.6549601941e-1,-.3360109025e-2,-.4114868587e-1,-.4114868587e-1,.7177662214e-1,-.7522679396e-1,-.7522679396e-1,-.1076722274,-.1444928058e-1,-.1444928058e-1,-.1317043874e-1,.1246734703e-1,.1246734703e-1,.6337331851e-1,.1727352568e-1,-.1680736381e-1,-.6614705937e-1,-.6614705937e-1,.1727352568e-1,-.1680736381e-1,.2317222740e-1,-.2883874010e-1,.4751234575e-1,.4751234575e-1,.2317222740e-1,-.2883874010e-1,-.4353025737e-2,-.8373369061e-2,.1446295644e-1,.1446295644e-1,-.4353025737e-2,-.8373369061e-2,-.4030485339e-1,-.7811447251e-1,.2079224911e-1,.2079224911e-1,-.4030485339e-1,-.7811447251e-1,-.9854506346e-2,.3162709295e-1,.2557166486e-1,.2557166486e-1,-.9854506346e-2,.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,-.1891559699e-1,.1891559699e-1,-.3728221333e-11,-.6016501311e-1,.6016501307e-1,-.2502617804e-10,-.6637503462e-1,.6637503465e-1,.1129442338e-10,.7036268232e-1,-.7036268232e-1,.2527951366e-10,.1000993614,-.1000993615,0.,.6591155720e-1,-.6591155723e-1,0.,.2042456634e-1,-.2042456633e-1,0.,-.9759050577e-1,.3794735317e-1,.5964315250e-1,-.5964315260e-1,.9759050570e-1,-.3794735317e-1,-.5710338621e-1,.2915609201e-1,.2794729420e-1,-.2794729418e-1,.5710338621e-1,-.2915609203e-1,-.1151029307e-1,.2993628541e-2,.8516664535e-2,-.8516664535e-2,.1151029308e-1,-.2993628539e-2,-.8322014746e-1,.1688050365e-1,.6633964381e-1,-.6633964379e-1,.8322014748e-1,-.1688050367e-1,-.3641824967e-1,.2916737971e-2,.3350151169e-1,-.3350151170e-1,.3641824966e-1,-.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,-.2889405168e-1,.2889405167e-1,-.5694952154e-11,-.9190357559e-1,.9190357554e-1,-.3822811841e-10,-.1013895401,.1013895402,.1725251670e-10,.1074807726,-.1074807726,.3861509495e-10,.1529043002,-.1529043003,0.,.1006815667,-.1006815667,0.,.3119904043e-1,-.3119904041e-1,0.,-.1490719599,.5796553947e-1,.9110642037e-1,-.9110642046e-1,.1490719598,-.5796553947e-1,-.8722686330e-1,.4453666621e-1,.4269019705e-1,-.4269019702e-1,.8722686330e-1,-.4453666622e-1,-.1758226310e-1,.4572843131e-2,.1300941997e-1,-.1300941997e-1,.1758226310e-1,-.4572843127e-2,-.1271208750,.2578539525e-1,.1013354798,-.1013354798,.1271208750,-.2578539527e-1,-.5562979528e-1,.4455390845e-2,.5117440439e-1,-.5117440444e-1,.5562979523e-1,-.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,-.3444674858e-1,.3444674856e-1,.4446512362e-11,-.7864938975e-1,.7864938968e-1,.4128474942e-11,-.1099312557,.1099312557,-.9746388013e-11,.1747386758e-1,-.1747386758e-1,.4936608124e-10,-.3189597418e-1,.3189597420e-1,0.,-.5026526501e-1,.5026526506e-1,0.,-.2293909754e-1,.2293909753e-1,0.,-.5770330047e-1,.9292114363e-1,-.3521784310e-1,.3521784315e-1,.5770330042e-1,-.9292114363e-1,-.4815974411e-1,.7895094357e-1,-.3079119953e-1,.3079119951e-1,.4815974411e-1,-.7895094362e-1,-.3346867218e-3,.1099182731e-1,-.1065714059e-1,.1065714059e-1,.3346867220e-3,-.1099182730e-1,.5423095927e-2,.5878737247e-1,-.6421046842e-1,.6421046842e-1,-.5423095927e-2,-.5878737252e-1,.2558351144e-1,.1251003735e-1,-.3809354879e-1,.3809354879e-1,-.2558351144e-1,-.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,-.5261827761e-1,.5261827757e-1,.6792159825e-11,-.1201389273,.1201389272,.6306349641e-11,-.1679227669,.1679227669,-.1488785361e-10,.2669177363e-1,-.2669177363e-1,.7540793467e-10,-.4872190534e-1,.4872190538e-1,0.,-.7678146058e-1,.7678146066e-1,0.,-.3504005029e-1,.3504005026e-1,0.,-.8814324742e-1,.1419393914,-.5379614395e-1,.5379614402e-1,.8814324734e-1,-.1419393914,-.7356522427e-1,.1205995584,-.4703433420e-1,.4703433416e-1,.7356522427e-1,-.1205995584,-.5112424122e-3,.1679029356e-1,-.1627905114e-1,.1627905114e-1,.5112424126e-3,-.1679029354e-1,.8283915862e-2,.8979919475e-1,-.9808311064e-1,.9808311064e-1,-.8283915862e-2,-.8979919482e-1,.3907945924e-1,.1910939770e-1,-.5818885694e-1,.5818885694e-1,-.3907945924e-1,-.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,-.6060877170e-2,.6060877164e-2,-.4944073794e-11,.3161264212e-1,-.3161264209e-1,.1584800096e-10,-.2303508888e-3,.2303508888e-3,.5819067272e-11,-.6681522508e-1,.6681522508e-1,.1528439436e-10,-.4470780632e-1,.4470780632e-1,0.,.1994876348e-1,-.1994876348e-1,0.,.2329919003e-1,-.2329919001e-1,0.,.9824065005e-1,.7289231590e-1,-.1486374908e-2,.1486374909e-2,-.9824064997e-1,-.7289231590e-1,.5309591204e-1,.8058321394e-1,.3033302763e-1,-.3033302759e-1,-.5309591204e-1,-.8058321395e-1,.9386233204e-2,.2045354015e-1,.1267656516e-1,-.1267656516e-1,-.9386233207e-2,-.2045354012e-1,.6111843873e-1,.9989476013e-1,.5006525860e-1,-.5006525859e-1,-.6111843873e-1,-.9989476021e-1,-.6661205800e-2,.2978294282e-1,.3944227928e-1,-.3944227931e-1,.6661205793e-2,-.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,.2989961082e-1,-.2989961082e-1,.5259821245e-11,.7039060152e-1,-.7039060149e-1,-.2191617071e-10,.1013287354,-.1013287354,-.6279229509e-12,.4006762239e-1,-.4006762239e-1,-.3934858683e-10,.8551663720e-1,-.8551663723e-1,0.,.1484968306e-1,-.1484968307e-1,0.,-.2177358772e-1,.2177358771e-1,0.,-.4790109961e-2,-.2984596792e-1,.3463607749e-1,-.3463607753e-1,.4790109958e-2,.2984596792e-1,.2283783488e-1,.4181200571e-2,-.2701903535e-1,.2701903532e-1,-.2283783488e-1,-.4181200574e-2,-.9086687153e-2,.2380170174e-1,-.1471501461e-1,.1471501461e-1,.9086687153e-2,-.2380170172e-1,-.7038491177e-1,.9773979388e-1,-.2735488204e-1,.2735488204e-1,.7038491177e-1,-.9773979394e-1,-.1336506945e-1,.5130832062e-1,-.3794325106e-1,.3794325110e-1,.1336506945e-1,-.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{-.9882264718e-1,.1440235391e-1,.1440235396e-1,-.3006820721e-1,-.3375895322e-1,-.3375895287e-1,-.1352948785,.2899076627e-1,.2899076622e-1,-.1129298746,-.6881731865e-1,-.6881731855e-1,-.4625865234e-1,.2871520174e-1,.2871520191e-1,-.1246232363e-1,.6335778756e-1,.6335778749e-1,-.1534390562e-2,.2959552517e-1,.2959552520e-1,-.1872238370e-4,.3794484055e-1,-.7521199400e-2,-.8201583980e-1,-.8201584015e-1,.3794484065e-1,-.7521199317e-2,.5242564114e-1,.1267437352e-1,-.4949958872e-1,-.4949958862e-1,.5242564114e-1,.1267437351e-1,.1478824220e-1,.1029932873e-2,.5978541514e-3,.5978541492e-3,.1478824222e-1,.1029932873e-2,.8395409436e-1,.1058736005e-2,.8059922862e-2,.8059922754e-2,.8395409436e-1,.1058736043e-2,.4934714167e-1,.1685816455e-3,.3319675284e-1,.3319675290e-1,.4934714164e-1,.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{-.1509540871,.2199995900e-1,.2199995907e-1,-.4592994519e-1,-.5156765287e-1,-.5156765232e-1,-.2066663408,.4428412697e-1,.4428412690e-1,-.1725032329,-.1051201906,-.1051201905,-.7066125868e-1,.4386319521e-1,.4386319547e-1,-.1903651379e-1,.9678061914e-1,.9678061909e-1,-.2343820299e-2,.4520791146e-1,.4520791150e-1,-.2859891351e-4,.5796170139e-1,-.1148882186e-1,-.1252812647,-.1252812653,.5796170154e-1,-.1148882174e-1,.8008148967e-1,.1936042536e-1,-.7561187074e-1,-.7561187061e-1,.8008148966e-1,.1936042534e-1,.2258941311e-1,.1573248451e-2,.9132373015e-3,.9132372980e-3,.2258941314e-1,.1573248449e-2,.1282419975,.1617245962e-2,.1231173554e-1,.1231173538e-1,.1282419975,.1617246020e-2,.7537900402e-1,.2575127172e-3,.5070887761e-1,.5070887768e-1,.7537900399e-1,.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{-.1521474186,.4127107599e-1,.4127107613e-1,.3344651713e-1,-.7384207728e-1,-.7384207651e-1,-.3303033204e-1,.7676670968e-1,.7676670952e-1,.7646932810e-1,-.5434312853e-1,-.5434312844e-1,-.1406012128,.1185729266e-2,.1185729273e-2,-.5653180033e-1,-.3451242438e-1,-.3451242437e-1,-.8758681134e-2,-.3020381423e-1,-.3020381426e-1,-.1246893713e-3,.4714000659e-1,-.2778316412e-1,.2604070049e-1,.2604070059e-1,.4714000670e-1,-.2778316383e-1,.8257470907e-1,.5115488591e-1,.4912182046e-1,.4912182038e-1,.8257470907e-1,.5115488588e-1,.7397967795e-2,.5467224064e-2,-.7113199607e-3,-.7113199580e-3,.7397967806e-2,.5467224056e-2,.3155548387e-1,.5355748779e-2,-.6574391344e-2,-.6574391251e-2,.3155548387e-1,.5355748971e-2,-.2296424900e-1,.1032248755e-2,-.3449059684e-1,-.3449059690e-1,-.2296424899e-1,.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{.4252649632e-1,.3621722030e-1,.3621722042e-1,-.3560242460e-1,-.2348842407e-1,-.2348842382e-1,.1206537327,.5391680609e-1,.5391680599e-1,-.2629256125e-1,.6802515129e-1,.6802515124e-1,-.1427717610,-.2147433158e-1,-.2147433171e-1,-.1219618694,-.1982044929e-3,-.1982044929e-3,-.2648049029e-1,.2844076554e-1,.2844076556e-1,-.4636110040e-3,-.2835644722e-1,-.4230017603e-1,.2869463558e-1,.2869463573e-1,-.2835644729e-1,-.4230017559e-1,-.2013663457e-1,.9116808757e-1,-.4429713195e-1,-.4429713188e-1,-.2013663457e-1,.9116808747e-1,-.1463911665e-1,.1492260701e-1,.8259580003e-3,.8259579974e-3,-.1463911667e-1,.1492260700e-1,-.8008479954e-1,.1363156197e-1,.4081583933e-2,.4081583877e-2,-.8008479954e-1,.1363156247e-1,-.6542647961e-2,.3434213984e-2,.3335556699e-1,.3335556704e-1,-.6542647959e-2,.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,-.9258142764e-2,.9258142654e-2,-.8136908247e-10,.4828910902e-1,-.4828910902e-1,0.,-.3518672753e-3,.3518668221e-3,-.3485534728e-10,-.1020619419,.1020619425,.5460999099e-9,-.6829230417e-1,.6829230235e-1,0.,.3047223942e-1,-.3047223960e-1,.3681086362e-9,.3559010093e-1,-.3559010063e-1,-.4643401911e-10,.2666533469e-1,-.1205488497e-1,-.1256702126,.1256702128,-.2666533393e-1,.1205488583e-1,-.3819594183e-1,.3791605562e-2,-.7296682180e-1,.7296682193e-1,.3819594232e-1,-.3791605423e-2,-.1658837758e-1,.3172129929e-3,-.1156231241e-1,.1156231248e-1,.1658837783e-1,-.3172129864e-3,-.6017702823e-1,-.9452189607e-3,-.7706103991e-1,.7706103958e-1,.6017702852e-1,.9452190353e-3,-.5568378542e-1,-.1442892931e-4,.1474045160e-1,-.1474045137e-1,.5568378595e-1,.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,-.3598418982e-1,.3598418942e-1,.8668398183e-10,.1469158467,-.1469158467,0.,-.1272304336e-2,.1272302698e-2,.1824763744e-10,-.1329348046,.1329348053,.2240768661e-8,-.2532572946e-1,.2532572881e-1,0.,-.1092978850e-1,.1092978857e-1,.2733414349e-8,-.3384746400e-1,.3384746369e-1,-.3998097089e-9,.4975749160e-1,-.5928626767e-1,.9528779263e-2,-.9528779277e-2,-.4975749017e-1,.5928627186e-1,-.8709586801e-1,.2026201294e-1,.6683385514e-1,-.6683385532e-1,.8709586914e-1,-.2026201219e-1,-.1562966645e-1,.2197560437e-2,.1343210582e-1,-.1343210590e-1,.1562966667e-1,-.2197560392e-2,-.4738032804e-1,-.6254379410e-2,.5363470751e-1,-.5363470732e-1,.4738032827e-1,.6254379903e-2,.1447133378e-1,-.1145953765e-3,-.1435673898e-1,.1435673875e-1,-.1447133392e-1,.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{.5705528195e-1,.5381888614e-2,.5381890600e-2,.3865155911e-1,-.1066851510e-1,-.1066851160e-1,.1201849450,-.1023929481e-1,-.1023929537e-1,.1298773899,-.1146344760e-1,-.1146344311e-1,.1390909759e-1,-.1113342295,-.1113342303,.1417761192e-2,-.1805227680e-1,-.1805227549e-1,.4540571627e-4,.3784627926e-1,.3784627821e-1,.2688892560e-7,-.5375088148e-1,-.3596809691e-2,.1049770374e-2,.1049766066e-2,-.5375088190e-1,-.3596800240e-2,.2284622311e-1,.5640425826e-3,.1245758952e-1,.1245758914e-1,.2284622668e-1,.5640434935e-3,.1743701782e-1,.9341106461e-4,-.1604194740e-1,-.1604194746e-1,.1743701764e-1,.9341075247e-4,.2535811540e-1,-.3827141178e-3,-.1086557567,-.1086557557,.2535811250e-1,-.3827130864e-3,.5582706905e-1,-.4444677760e-5,-.1524183760e-1,-.1524183769e-1,.5582706642e-1,-.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[52],
		T o_output[56]
	) {
		static const T matrix[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{.1148215401,.1569113387e-1,.1569113387e-1,.1569113387e-1,.1021689440,.1021689440,.1021689440,.6587041484e-1,.6587041484e-1,.6587041484e-1,.1032031823,.1032031823,.1032031823,.7348534591e-1,.7348534591e-1,.7348534591e-1,.3478299668e-1,.3478299668e-1,.3478299668e-1,.8775204783e-2,.8775204783e-2,.8775204783e-2,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{.1753927996,.2396860291e-1,.2396860291e-1,.2396860291e-1,.1560656399,.1560656399,.1560656399,.1006187207,.1006187207,.1006187207,.1576454650,.1576454650,.1576454650,.1122507200,.1122507200,.1122507200,.5313190506e-1,.5313190506e-1,.5313190506e-1,.1340434672e-1,.1340434672e-1,.1340434672e-1,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{.2198665100,.3004623384e-1,.3004623384e-1,.3004623384e-1,.1956386332,.1956386332,.1956386332,.1261322415,.1261322415,.1261322415,.1976190488,.1976190488,.1976190488,.1407137244,.1407137244,.1407137244,.6660436781e-1,.6660436781e-1,.6660436781e-1,.1680323787e-1,.1680323787e-1,.1680323787e-1,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,.1092092502e-1,.1092092502e-1,-.2184185003e-1,.3473628658e-1,.3473628658e-1,-.6947257305e-1,.3832164410e-1,.3832164410e-1,-.7664328825e-1,-.4062391359e-1,-.4062391359e-1,.8124782718e-1,-.5779239332e-1,-.5779239332e-1,.1155847866,-.3805405532e-1,-.3805405532e-1,.7610811063e-1,-.1179212887e-1,-.1179212887e-1,.2358425774e-1,-.1252607563e-1,.9077889505e-1,-.7825281938e-1,-.7825281938e-1,-.1252607563e-1,.9077889505e-1,.6978997413e-3,.4910403324e-1,-.4980193299e-1,-.4980193299e-1,.6978997413e-3,.4910403324e-1,-.3188726319e-2,.1156256937e-1,-.8373843054e-2,-.8373843054e-2,-.3188726319e-2,.1156256937e-1,-.2855524789e-1,.8634838574e-1,-.5779313787e-1,-.5779313787e-1,-.2855524789e-1,.8634838574e-1,-.1765812736e-1,.4036819305e-1,-.2271006571e-1,-.2271006571e-1,-.1765812736e-1,.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,.1668198852e-1,.1668198852e-1,-.3336397701e-1,.5306055418e-1,.5306055418e-1,-.1061211082,.5853727826e-1,.5853727826e-1,-.1170745566,-.6205405300e-1,-.6205405300e-1,.1241081060,-.8827933896e-1,-.8827933896e-1,.1765586779,-.5812852966e-1,-.5812852966e-1,.1162570593,-.1801277438e-1,-.1801277438e-1,.3602554875e-1,-.1913389657e-1,.1386670527,-.1195331560,-.1195331560,-.1913389657e-1,.1386670527,.1066059464e-2,.7500764972e-1,-.7607370920e-1,-.7607370920e-1,.1066059464e-2,.7500764972e-1,-.4870859907e-2,.1766211645e-1,-.1279125655e-1,-.1279125655e-1,-.4870859907e-2,.1766211645e-1,-.4361886163e-1,.1318993379,-.8828047629e-1,-.8828047629e-1,-.4361886163e-1,.1318993379,-.2697323507e-1,.6166343343e-1,-.3469019838e-1,-.3469019838e-1,-.2697323507e-1,.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{-.1104870784,-.1372943814e-1,-.1372943814e-1,.2604624096e-1,-.1058560268,-.1058560268,-.1503953414e-1,-.6279059801e-1,-.6279059801e-1,.6414708202e-1,-.6180728788e-1,-.6180728788e-1,-.8198437210e-1,.4481847772e-2,.4481847772e-2,.4131214635e-1,.2730516345e-1,.2730516345e-1,.8534649204e-1,.1322296188e-1,.1322296188e-1,.3971075014e-1,-.6854799631e-1,-.2788193386e-1,.3874809825e-1,.3874809825e-1,-.6854799631e-1,-.2788193386e-1,-.3642583502e-1,-.8712204498e-3,.5473886199e-1,.5473886199e-1,-.3642583502e-1,-.8712204498e-3,-.2214798741e-2,.1009100723e-1,.1047747017e-1,.1047747017e-1,-.2214798741e-2,.1009100723e-1,-.1265572386e-1,.6148813860e-1,.5522608679e-1,.5522608679e-1,-.1265572386e-1,.6148813860e-1,.1134779051e-1,.5533443179e-1,.2579313736e-1,.2579313736e-1,.1134779051e-1,.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{-.1687718001,-.2097206318e-1,-.2097206318e-1,.3978629025e-1,-.1616977519,-.1616977519,-.2297326787e-1,-.9591422281e-1,-.9591422281e-1,.9798628638e-1,-.9441219172e-1,-.9441219172e-1,-.1252331970,.6846135558e-2,.6846135558e-2,.6310534594e-1,.4170932612e-1,.4170932612e-1,.1303689201,.2019840792e-1,.2019840792e-1,.6065917284e-1,-.1047087940,-.4259035749e-1,.5918869776e-1,.5918869776e-1,-.1047087940,-.4259035749e-1,-.5564138210e-1,-.1330811220e-2,.8361499282e-1,.8361499282e-1,-.5564138210e-1,-.1330811220e-2,-.3383160961e-2,.1541426816e-1,.1600460005e-1,.1600460005e-1,-.3383160961e-2,.1541426816e-1,-.1933193753e-1,.9392468318e-1,.8435924102e-1,.8435924102e-1,-.1933193753e-1,.9392468318e-1,.1733403633e-1,.8452474073e-1,.3939966813e-1,.3939966813e-1,.1733403633e-1,.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{.6804239411e-1,-.1145848830e-1,-.1145848830e-1,-.2895473345e-1,-.3479953926e-2,-.3479953926e-2,.8777788353e-1,-.3406513033e-1,-.3406513033e-1,-.3473009539e-1,.1014545622,.1014545622,-.9142437836e-1,.5109385389e-1,.5109385389e-1,-.7796646760e-1,-.7600422558e-2,-.7600422558e-2,.4998669741e-1,-.1339603089e-1,-.1339603089e-1,.5386293723e-1,.6480346250e-1,-.1027298408,.1762397985e-1,.1762397985e-1,.6480346250e-1,-.1027298408,.2109538461e-1,-.4916147521e-1,-.5370754828e-1,-.5370754828e-1,.2109538461e-1,-.4916147521e-1,.5671605418e-2,.1354331144e-2,-.1246131332e-1,-.1246131332e-1,.5671605418e-2,.1354331144e-2,.4342611086e-1,-.1515827781e-1,-.4183488198e-1,-.4183488198e-1,.4342611086e-1,-.1515827781e-1,-.8837167105e-3,.5277408321e-1,-.2665741414e-1,-.2665741414e-1,-.8837167105e-3,.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{.6656449560e-1,.1342806488e-1,.1342806488e-1,.3079509387e-1,.1065239097,.1065239097,-.9772524850e-1,.6549601941e-1,.6549601941e-1,-.3360109025e-2,-.4114868587e-1,-.4114868587e-1,.7177662214e-1,-.7522679396e-1,-.7522679396e-1,-.1076722274,-.1444928058e-1,-.1444928058e-1,-.1317043874e-1,.1246734703e-1,.1246734703e-1,.6337331851e-1,.1727352568e-1,-.1680736381e-1,-.6614705937e-1,-.6614705937e-1,.1727352568e-1,-.1680736381e-1,.2317222740e-1,-.2883874010e-1,.4751234575e-1,.4751234575e-1,.2317222740e-1,-.2883874010e-1,-.4353025737e-2,-.8373369061e-2,.1446295644e-1,.1446295644e-1,-.4353025737e-2,-.8373369061e-2,-.4030485339e-1,-.7811447251e-1,.2079224911e-1,.2079224911e-1,-.4030485339e-1,-.7811447251e-1,-.9854506346e-2,.3162709295e-1,.2557166486e-1,.2557166486e-1,-.9854506346e-2,.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,-.1891559699e-1,.1891559699e-1,-.3728221333e-11,-.6016501311e-1,.6016501307e-1,-.2502617804e-10,-.6637503462e-1,.6637503465e-1,.1129442338e-10,.7036268232e-1,-.7036268232e-1,.2527951366e-10,.1000993614,-.1000993615,0.,.6591155720e-1,-.6591155723e-1,0.,.2042456634e-1,-.2042456633e-1,0.,-.9759050577e-1,.3794735317e-1,.5964315250e-1,-.5964315260e-1,.9759050570e-1,-.3794735317e-1,-.5710338621e-1,.2915609201e-1,.2794729420e-1,-.2794729418e-1,.5710338621e-1,-.2915609203e-1,-.1151029307e-1,.2993628541e-2,.8516664535e-2,-.8516664535e-2,.1151029308e-1,-.2993628539e-2,-.8322014746e-1,.1688050365e-1,.6633964381e-1,-.6633964379e-1,.8322014748e-1,-.1688050367e-1,-.3641824967e-1,.2916737971e-2,.3350151169e-1,-.3350151170e-1,.3641824966e-1,-.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,-.2889405168e-1,.2889405167e-1,-.5694952154e-11,-.9190357559e-1,.9190357554e-1,-.3822811841e-10,-.1013895401,.1013895402,.1725251670e-10,.1074807726,-.1074807726,.3861509495e-10,.1529043002,-.1529043003,0.,.1006815667,-.1006815667,0.,.3119904043e-1,-.3119904041e-1,0.,-.1490719599,.5796553947e-1,.9110642037e-1,-.9110642046e-1,.1490719598,-.5796553947e-1,-.8722686330e-1,.4453666621e-1,.4269019705e-1,-.4269019702e-1,.8722686330e-1,-.4453666622e-1,-.1758226310e-1,.4572843131e-2,.1300941997e-1,-.1300941997e-1,.1758226310e-1,-.4572843127e-2,-.1271208750,.2578539525e-1,.1013354798,-.1013354798,.1271208750,-.2578539527e-1,-.5562979528e-1,.4455390845e-2,.5117440439e-1,-.5117440444e-1,.5562979523e-1,-.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,-.3444674858e-1,.3444674856e-1,.4446512362e-11,-.7864938975e-1,.7864938968e-1,.4128474942e-11,-.1099312557,.1099312557,-.9746388013e-11,.1747386758e-1,-.1747386758e-1,.4936608124e-10,-.3189597418e-1,.3189597420e-1,0.,-.5026526501e-1,.5026526506e-1,0.,-.2293909754e-1,.2293909753e-1,0.,-.5770330047e-1,.9292114363e-1,-.3521784310e-1,.3521784315e-1,.5770330042e-1,-.9292114363e-1,-.4815974411e-1,.7895094357e-1,-.3079119953e-1,.3079119951e-1,.4815974411e-1,-.7895094362e-1,-.3346867218e-3,.1099182731e-1,-.1065714059e-1,.1065714059e-1,.3346867220e-3,-.1099182730e-1,.5423095927e-2,.5878737247e-1,-.6421046842e-1,.6421046842e-1,-.5423095927e-2,-.5878737252e-1,.2558351144e-1,.1251003735e-1,-.3809354879e-1,.3809354879e-1,-.2558351144e-1,-.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,-.5261827761e-1,.5261827757e-1,.6792159825e-11,-.1201389273,.1201389272,.6306349641e-11,-.1679227669,.1679227669,-.1488785361e-10,.2669177363e-1,-.2669177363e-1,.7540793467e-10,-.4872190534e-1,.4872190538e-1,0.,-.7678146058e-1,.7678146066e-1,0.,-.3504005029e-1,.3504005026e-1,0.,-.8814324742e-1,.1419393914,-.5379614395e-1,.5379614402e-1,.8814324734e-1,-.1419393914,-.7356522427e-1,.1205995584,-.4703433420e-1,.4703433416e-1,.7356522427e-1,-.1205995584,-.5112424122e-3,.1679029356e-1,-.1627905114e-1,.1627905114e-1,.5112424126e-3,-.1679029354e-1,.8283915862e-2,.8979919475e-1,-.9808311064e-1,.9808311064e-1,-.8283915862e-2,-.8979919482e-1,.3907945924e-1,.1910939770e-1,-.5818885694e-1,.5818885694e-1,-.3907945924e-1,-.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,-.6060877170e-2,.6060877164e-2,-.4944073794e-11,.3161264212e-1,-.3161264209e-1,.1584800096e-10,-.2303508888e-3,.2303508888e-3,.5819067272e-11,-.6681522508e-1,.6681522508e-1,.1528439436e-10,-.4470780632e-1,.4470780632e-1,0.,.1994876348e-1,-.1994876348e-1,0.,.2329919003e-1,-.2329919001e-1,0.,.9824065005e-1,.7289231590e-1,-.1486374908e-2,.1486374909e-2,-.9824064997e-1,-.7289231590e-1,.5309591204e-1,.8058321394e-1,.3033302763e-1,-.3033302759e-1,-.5309591204e-1,-.8058321395e-1,.9386233204e-2,.2045354015e-1,.1267656516e-1,-.1267656516e-1,-.9386233207e-2,-.2045354012e-1,.6111843873e-1,.9989476013e-1,.5006525860e-1,-.5006525859e-1,-.6111843873e-1,-.9989476021e-1,-.6661205800e-2,.2978294282e-1,.3944227928e-1,-.3944227931e-1,.6661205793e-2,-.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,.2989961082e-1,-.2989961082e-1,.5259821245e-11,.7039060152e-1,-.7039060149e-1,-.2191617071e-10,.1013287354,-.1013287354,-.6279229509e-12,.4006762239e-1,-.4006762239e-1,-.3934858683e-10,.8551663720e-1,-.8551663723e-1,0.,.1484968306e-1,-.1484968307e-1,0.,-.2177358772e-1,.2177358771e-1,0.,-.4790109961e-2,-.2984596792e-1,.3463607749e-1,-.3463607753e-1,.4790109958e-2,.2984596792e-1,.2283783488e-1,.4181200571e-2,-.2701903535e-1,.2701903532e-1,-.2283783488e-1,-.4181200574e-2,-.9086687153e-2,.2380170174e-1,-.1471501461e-1,.1471501461e-1,.9086687153e-2,-.2380170172e-1,-.7038491177e-1,.9773979388e-1,-.2735488204e-1,.2735488204e-1,.7038491177e-1,-.9773979394e-1,-.1336506945e-1,.5130832062e-1,-.3794325106e-1,.3794325110e-1,.1336506945e-1,-.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{-.9882264718e-1,.1440235391e-1,.1440235396e-1,-.3006820721e-1,-.3375895322e-1,-.3375895287e-1,-.1352948785,.2899076627e-1,.2899076622e-1,-.1129298746,-.6881731865e-1,-.6881731855e-1,-.4625865234e-1,.2871520174e-1,.2871520191e-1,-.1246232363e-1,.6335778756e-1,.6335778749e-1,-.1534390562e-2,.2959552517e-1,.2959552520e-1,-.1872238370e-4,.3794484055e-1,-.7521199400e-2,-.8201583980e-1,-.8201584015e-1,.3794484065e-1,-.7521199317e-2,.5242564114e-1,.1267437352e-1,-.4949958872e-1,-.4949958862e-1,.5242564114e-1,.1267437351e-1,.1478824220e-1,.1029932873e-2,.5978541514e-3,.5978541492e-3,.1478824222e-1,.1029932873e-2,.8395409436e-1,.1058736005e-2,.8059922862e-2,.8059922754e-2,.8395409436e-1,.1058736043e-2,.4934714167e-1,.1685816455e-3,.3319675284e-1,.3319675290e-1,.4934714164e-1,.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{-.1509540871,.2199995900e-1,.2199995907e-1,-.4592994519e-1,-.5156765287e-1,-.5156765232e-1,-.2066663408,.4428412697e-1,.4428412690e-1,-.1725032329,-.1051201906,-.1051201905,-.7066125868e-1,.4386319521e-1,.4386319547e-1,-.1903651379e-1,.9678061914e-1,.9678061909e-1,-.2343820299e-2,.4520791146e-1,.4520791150e-1,-.2859891351e-4,.5796170139e-1,-.1148882186e-1,-.1252812647,-.1252812653,.5796170154e-1,-.1148882174e-1,.8008148967e-1,.1936042536e-1,-.7561187074e-1,-.7561187061e-1,.8008148966e-1,.1936042534e-1,.2258941311e-1,.1573248451e-2,.9132373015e-3,.9132372980e-3,.2258941314e-1,.1573248449e-2,.1282419975,.1617245962e-2,.1231173554e-1,.1231173538e-1,.1282419975,.1617246020e-2,.7537900402e-1,.2575127172e-3,.5070887761e-1,.5070887768e-1,.7537900399e-1,.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{-.1521474186,.4127107599e-1,.4127107613e-1,.3344651713e-1,-.7384207728e-1,-.7384207651e-1,-.3303033204e-1,.7676670968e-1,.7676670952e-1,.7646932810e-1,-.5434312853e-1,-.5434312844e-1,-.1406012128,.1185729266e-2,.1185729273e-2,-.5653180033e-1,-.3451242438e-1,-.3451242437e-1,-.8758681134e-2,-.3020381423e-1,-.3020381426e-1,-.1246893713e-3,.4714000659e-1,-.2778316412e-1,.2604070049e-1,.2604070059e-1,.4714000670e-1,-.2778316383e-1,.8257470907e-1,.5115488591e-1,.4912182046e-1,.4912182038e-1,.8257470907e-1,.5115488588e-1,.7397967795e-2,.5467224064e-2,-.7113199607e-3,-.7113199580e-3,.7397967806e-2,.5467224056e-2,.3155548387e-1,.5355748779e-2,-.6574391344e-2,-.6574391251e-2,.3155548387e-1,.5355748971e-2,-.2296424900e-1,.1032248755e-2,-.3449059684e-1,-.3449059690e-1,-.2296424899e-1,.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{.4252649632e-1,.3621722030e-1,.3621722042e-1,-.3560242460e-1,-.2348842407e-1,-.2348842382e-1,.1206537327,.5391680609e-1,.5391680599e-1,-.2629256125e-1,.6802515129e-1,.6802515124e-1,-.1427717610,-.2147433158e-1,-.2147433171e-1,-.1219618694,-.1982044929e-3,-.1982044929e-3,-.2648049029e-1,.2844076554e-1,.2844076556e-1,-.4636110040e-3,-.2835644722e-1,-.4230017603e-1,.2869463558e-1,.2869463573e-1,-.2835644729e-1,-.4230017559e-1,-.2013663457e-1,.9116808757e-1,-.4429713195e-1,-.4429713188e-1,-.2013663457e-1,.9116808747e-1,-.1463911665e-1,.1492260701e-1,.8259580003e-3,.8259579974e-3,-.1463911667e-1,.1492260700e-1,-.8008479954e-1,.1363156197e-1,.4081583933e-2,.4081583877e-2,-.8008479954e-1,.1363156247e-1,-.6542647961e-2,.3434213984e-2,.3335556699e-1,.3335556704e-1,-.6542647959e-2,.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,-.9258142764e-2,.9258142654e-2,-.8136908247e-10,.4828910902e-1,-.4828910902e-1,0.,-.3518672753e-3,.3518668221e-3,-.3485534728e-10,-.1020619419,.1020619425,.5460999099e-9,-.6829230417e-1,.6829230235e-1,0.,.3047223942e-1,-.3047223960e-1,.3681086362e-9,.3559010093e-1,-.3559010063e-1,-.4643401911e-10,.2666533469e-1,-.1205488497e-1,-.1256702126,.1256702128,-.2666533393e-1,.1205488583e-1,-.3819594183e-1,.3791605562e-2,-.7296682180e-1,.7296682193e-1,.3819594232e-1,-.3791605423e-2,-.1658837758e-1,.3172129929e-3,-.1156231241e-1,.1156231248e-1,.1658837783e-1,-.3172129864e-3,-.6017702823e-1,-.9452189607e-3,-.7706103991e-1,.7706103958e-1,.6017702852e-1,.9452190353e-3,-.5568378542e-1,-.1442892931e-4,.1474045160e-1,-.1474045137e-1,.5568378595e-1,.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,-.3598418982e-1,.3598418942e-1,.8668398183e-10,.1469158467,-.1469158467,0.,-.1272304336e-2,.1272302698e-2,.1824763744e-10,-.1329348046,.1329348053,.2240768661e-8,-.2532572946e-1,.2532572881e-1,0.,-.1092978850e-1,.1092978857e-1,.2733414349e-8,-.3384746400e-1,.3384746369e-1,-.3998097089e-9,.4975749160e-1,-.5928626767e-1,.9528779263e-2,-.9528779277e-2,-.4975749017e-1,.5928627186e-1,-.8709586801e-1,.2026201294e-1,.6683385514e-1,-.6683385532e-1,.8709586914e-1,-.2026201219e-1,-.1562966645e-1,.2197560437e-2,.1343210582e-1,-.1343210590e-1,.1562966667e-1,-.2197560392e-2,-.4738032804e-1,-.6254379410e-2,.5363470751e-1,-.5363470732e-1,.4738032827e-1,.6254379903e-2,.1447133378e-1,-.1145953765e-3,-.1435673898e-1,.1435673875e-1,-.1447133392e-1,.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{.5705528195e-1,.5381888614e-2,.5381890600e-2,.3865155911e-1,-.1066851510e-1,-.1066851160e-1,.1201849450,-.1023929481e-1,-.1023929537e-1,.1298773899,-.1146344760e-1,-.1146344311e-1,.1390909759e-1,-.1113342295,-.1113342303,.1417761192e-2,-.1805227680e-1,-.1805227549e-1,.4540571627e-4,.3784627926e-1,.3784627821e-1,.2688892560e-7,-.5375088148e-1,-.3596809691e-2,.1049770374e-2,.1049766066e-2,-.5375088190e-1,-.3596800240e-2,.2284622311e-1,.5640425826e-3,.1245758952e-1,.1245758914e-1,.2284622668e-1,.5640434935e-3,.1743701782e-1,.9341106461e-4,-.1604194740e-1,-.1604194746e-1,.1743701764e-1,.9341075247e-4,.2535811540e-1,-.3827141178e-3,-.1086557567,-.1086557557,.2535811250e-1,-.3827130864e-3,.5582706905e-1,-.4444677760e-5,-.1524183760e-1,-.1524183769e-1,.5582706642e-1,-.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 56; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 52; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 362880;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.,0.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.464101616,0.,0.,0.,0.,0.,7.838367178,0.,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,0.,7.668115805,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.098386677,0.,0.,0.,0.,0.,7.668115806,0.,0.,0.,0.,10.73312629,0.,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,0.,0.,0.,0.,0.,7.314285714,0.,0.,0.,0.,11.33763824,0.,0.,0.,12.92995257,0.,0.,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780458,0.,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,9.142857143,0.,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.533303756,0.,0.,0.,0.,6.337932448,0.,0.,0.,13.55261854,0.,0.,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.291502622,0.,0.,0.,0.,0.,-2.993325910,0.,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.732863826,0.,0.,0.,0.,0.,3.346640106,0.,0.,0.,0.,-5.074680381,0.,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357805,0.,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,0.,0.,0.,0.,0.,5.935526613,0.,0.,0.,0.,-.3207134902,0.,0.,0.,-5.966396197,0.,0.,2.530240385,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.792842914,0.,0.,0.,9.200874127,0.,0.,19.80226324,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,0.,-3.833259389,0.,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.857142857,0.,0.,0.,0.,9.972922524,0.,0.,0.,-7.273098319,0.,0.,1.490788040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.320493800,0.,0.,21.25401924,0.,0.,0.,0.,0.,0.},
			{5.416025602,0.,0.,0.,0.,0.,-4.376809636,0.,0.,0.,0.,2.010178182,0.,0.,0.,-.5158119621,0.,0.,.5766953058e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,15.73213272,0.,0.,0.,-4.036867140,0.,0.,.4513354669,0.,0.,0.,0.,0.,0.,0.,0.,21.79908256,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-2.309401077,0.,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.,0.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-3.577708763,0.,0.,0.,0.,0.,8.854377448,0.,0.,0.,0.,-3.098386677,0.,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,0.,0.,0.,0.,0.,-4.618802155,0.,0.,0.,0.,12.72792206,0.,0.,0.,-3.265986323,0.,0.,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.732050808,0.,0.,0.,0.,0.,3.919183589,0.,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,0.,3.834057902,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.549193338,0.,0.,0.,0.,0.,3.834057903,0.,0.,0.,0.,5.366563145,0.,0.,0.,4.647580015,0.,0.,0.,0.,0.,6.640783086,0.,0.,0.,0.,-6.196773354,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,0.,0.,0.,0.,0.,3.657142857,0.,0.,0.,0.,5.668819119,0.,0.,0.,6.464976283,0.,0.,4.969293464,0.,0.,-6.730254569,0.,0.,0.,0.,9.455027815,0.,0.,0.,-6.998542123,0.,0.,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.581988897,0.,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,0.,2.529822128,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.236067977,0.,0.,0.,0.,0.,2.529822127,0.,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.190890229,0.,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,1.649572198,0.,0.,0.,-2.571428571,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,4.571428571,0.,0.,0.,7.423074892,0.,0.,0.,0.,0.,-7.377111136,0.,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,0.,0.,0.,0.,0.,3.836097790,0.,0.,0.,0.,3.659207005,0.,0.,0.,.5216405308,0.,0.,-3.207665141,0.,0.,.7666518779,0.,0.,0.,0.,3.168966225,0.,0.,0.,6.776309270,0.,0.,8.333758491,0.,0.,8.636804253,0.,0.,0.,-8.748177653,0.,0.,15.06237033,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.645751311,0.,0.,0.,0.,0.,-1.496662955,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178903,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.366431913,0.,0.,0.,0.,0.,1.673320053,0.,0.,0.,0.,-2.537340190,0.,0.,0.,.7606388294,0.,0.,0.,0.,0.,2.898275350,0.,0.,0.,0.,5.747048932,0.,0.,0.,-2.195775165,0.,0.,0.,0.,0.,2.182178902,0.,0.,0.,8.504200642,0.,0.,0.,0.,0.,10.06230590,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.160246899,0.,0.,0.,0.,0.,2.967763307,0.,0.,0.,0.,-.1603567451,0.,0.,0.,-2.983198099,0.,0.,1.265120192,0.,0.,1.814229470,0.,0.,0.,0.,5.416047582,0.,0.,0.,5.345224839,0.,0.,-3.286878676,0.,0.,.8964214566,0.,0.,0.,4.600437064,0.,0.,9.901131622,0.,0.,-9.525793445,0.,0.,13.38877363,0.,0.,0.,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225576,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.683281572,0.,0.,0.,0.,0.,-1.897366596,0.,0.,0.,0.,.6639400021,0.,0.,0.,-.9583148473e-1,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,-1.916629694,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,7.423074892,0.,0.,0.,-1.071428571,0.,0.,0.,0.,0.,8.874119673,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,.9897433188,0.,0.,0.,0.,-2.727411870,0.,0.,0.,1.516350793,0.,0.,-.2869022921,0.,0.,3.428571429,0.,0.,0.,0.,4.986461263,0.,0.,0.,-3.636549159,0.,0.,.7453940198,0.,0.,3.049339172,0.,0.,0.,8.346248498,0.,0.,-2.245365598,0.,0.,2.160246900,0.,0.,10.62700962,0.,0.,12.04989627,0.,0.,0.},
			{2.708012801,0.,0.,0.,0.,0.,-2.188404818,0.,0.,0.,0.,1.005089091,0.,0.,0.,-.2579059811,0.,0.,.2883476528e-1,0.,0.,5.685642496,0.,0.,0.,0.,-2.611298058,0.,0.,0.,.6700593942,0.,0.,-.7491491772e-1,0.,0.,7.866066360,0.,0.,0.,-2.018433570,0.,0.,.2256677334,0.,0.,9.552971226,0.,0.,-1.068054652,0.,0.,10.89954128,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,3.43216369764802,-5.83506573884863,7.25392389703222,-8.09901689119009,8.47552946493044,-4.00974905232829,6.81702604882233,-8.4746582765388,9.46196865399936,4.32546942586304,-7.35378632026677,9.14193750783762,-4.4144312457315,7.50503145913621,4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,7.27823584467215,-12.3738225688191,15.3826488429769,-7.46210387346585,12.6864189798666,7.31197439196321,9.14177937313724,-15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,3.29802783242322,-5.60701962552938,6.97042595115334,-7.78249101017965,8.14428871473084,-3.07100849759875,5.22106113152625,-6.4906175528327,7.24678421547914,2.20598407348367,-3.75042195543293,4.66237685613718,-.923194488569861,1.5695348492102,-.507379758856801,4.96464287844228,-8.44045334766827,10.4928391542811,-11.7152706263276,-3.80884940873901,6.47547397319737,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,6.45780747384475,-10.9790017210618,13.6486625036235,-4.10479708970297,6.97861844197595,1.32544341327556,7.79428043908965,-13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,3.06556135691961,-5.21180037431139,6.47910494479513,-7.23393037100344,7.57022621762746,-1.5947619510804,2.71127534923244,-3.37055072311448,3.76322492874006,-.444911232504622,.756399320675517,-.940325842964601,2.09440936915946,-3.56073236366527,-2.63186985254377,4.2894282422279,-7.29251224169724,9.06576398575723,-10.1219390639593,-.964148801340397,1.63916180436132,-2.03774139204767,-2.01722273728899,3.42950637731078,3.46190525493913,5.18623489409548,-8.8171847889118,10.9611768422514,.205483038001612,-.349344361529711,-3.63947886955732,5.81833621896754,-9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,2.74931534527504,-4.67414645388234,5.81071476784009,-6.48767173775818,6.78927500842464,.106757773657142,-.181500267083405,.225633983149896,-.251920680005671,-2.26282036321899,3.84705006415361,-4.78250111498198,2.15044627554921,-3.65600143048661,-.119501749587716,3.45007465931297,-5.86551639682902,7.29177895707792,-8.14128212331119,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429708,1.33170919606025,3.7410698100085,-6.36024100270088,7.90680110023492,3.5848146864382,-6.09458965574695,-3.59724489879787,3.76406389114812,-6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,2.36915365304765,-4.02782866109789,5.00723794478697,-5.59058866190052,5.85048772793476,1.68415540504794,-2.86325431258853,3.55948498424847,-3.97417031416821,-2.32831086437971,3.95839130925952,-4.9209161654707,-.381749900992996,.649017927110068,2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,2.3938769796025,-4.06986110773884,5.05949102727518,4.93737708601548,-8.39409842544749,.508166439915615,2.07554270972349,-3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,1.94896143440088,-3.31345445440235,4.11915606854152,-4.59904391748787,4.81284738096772,2.83314651369184,-4.81666890668704,5.98789306701478,-6.6854915743852,-.637278791777431,1.08344588754871,-1.34689725343212,-2.63766838605638,4.48433402043889,.808065843843811,1.73374759666501,-2.94756663562365,3.66429874441148,-4.09119503456583,4.12740556141282,-7.01705539432372,8.72332687917667,1.98940127224618,-3.38220674597771,-2.96832239843811,1.33269841309056,-2.26573774926528,2.81667592802333,4.37531622836173,-7.43853151653138,4.52786743647191,.950542947793588,-1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,1.51514018159912,-2.57590935107797,3.20226904625418,-3.57533818439405,3.74155092353066,3.36446686788906,-5.7199734894835,7.11084574520705,-7.93926992086969,1.91964817870762,-3.26361861024795,4.05720211125474,-1.95173675811295,3.31817281869637,-2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898136,3.55510980002441,-6.0440879938959,7.51377213010338,4.5904502888745,-7.8042837062112,1.24891795372526,.626153181688298,-1.06453109468611,1.32338312764,2.84473124098147,-4.83636425064192,5.48872263996487,.347192252628366,-.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,1.09494796295235,-1.86153514438243,2.31418717000874,-2.5837934399814,2.70391057656363,3.24473228828999,-5.51641118734948,6.85778510905627,-7.65672734172539,4.05017972204568,-6.88576274670089,8.56010903534823,1.63294619261043,-2.776192869346,-2.28560008620039,.547225517982272,-.930345156393644,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070527,4.87179403320197,-8.28259983241364,5.49730994822616,.236321146004053,-.40177262626832,.499467904136592,1.3620925765208,-2.31571114636374,3.77635721863399,.0946962283839075,-.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,.714786270724957,-1.21521735159799,1.51071034695561,-1.68671036412375,1.76512329607374,2.59853682213183,-4.41780594598826,5.49204234459095,-6.13187350041214,4.7549644652192,-8.08397637227704,10.0496810203089,5.36731919542456,-9.12504854432544,3.25729414076559,.233201739437629,-.396469282990197,.492874996717866,-.550295671800538,1.22672710731592,-2.08557456680069,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984458,.0657431866024185,-.111770838907699,.138949104550329,.451524561772771,-.76764272750461,1.59193561065049,.0171974308138372,-.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,.398540259080393,-.677563431168936,.842320170000572,-.940451730878483,.984172086870924,1.67165845871656,-2.84200809304711,3.53307252095548,-3.9446807593569,3.84755914106515,-6.54128488547504,8.13186772631248,6.3113399865174,-10.7299904588883,7.98777931164405,.07249762486996,-.123254146474207,.153224700235958,-.171075607231601,.434860760346183,-.739312383715058,.919084035475439,1.40707187748205,-2.3921810358159,3.23180916099,.0113956405123911,-.0193738753136992,.0240848387004708,.0887334889355571,-.150856948291044,.36808995603673,.00166206328996398,-.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,.166073783576779,-.282344179950942,.350999163642368,-.391891091702269,.410109589767539,.764836159193308,-1.30030781284792,1.61649145672597,-1.80481513164277,2.02218309922343,-3.43793955015111,4.27391103771759,4.02885967082133,-6.84951625494143,6.68010475273478,.0125887190510937,-.0214022435166571,.0266064261611111,-.0297061146457814,.0823390061849577,-.139985605707145,.174024591276674,.299004079114616,-.508340689012672,.796298871799331,.000824565926335038,-.00140185515920491,.00174273111828065,.00697737063218123,-.0118623177482726,.0321260785532874,5.01144631634162e-005,-8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,.0319379183519753,-.054298066631689,.0675012177634823,-.0753652106918314,.07886883956794,.154660030302817,-.262939511058368,.326876043550518,-.364957592022403,.43925822557731,-.746788570738901,.928378137184539,.961797763577864,-1.63516477461472,1.79684767953686,.000465578153648566,-.000791535419924754,.000984005657525639,-.00109864378994572,.00319092434145458,-.00542493160168594,.0067440612905754,.0123137463285649,-.0209347588552144,.0353622090543065,5.86464816381893e-006,-9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,-8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,-1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,3.43216369764802,-5.37468949190064,5.572814208608,-4.46269573688388,2.39877236199698,-4.00974905232829,6.27917489029831,-6.51064123734671,5.21370528571982,4.32546942586304,-6.77358573667404,7.02327732701019,-4.4144312457315,6.91289790468793,4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,7.27823584467215,-11.3975501068473,11.8176927763525,-7.46210387346585,11.6854832203471,7.31197439196321,9.14177937313724,-14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,3.29802783242322,-5.16463580891203,5.35501741292457,-4.28828460541245,2.30502351007762,-3.07100849759875,4.80912874954551,-4.98640545923911,3.99310107128126,2.20598407348367,-3.45452037324197,3.58186277642043,-.923194488569861,1.44570135775136,-.507379758856801,4.96464287844228,-7.77451667983119,8.06110512843119,-6.45531290809181,-3.80884940873901,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,6.45780747384475,-10.1127781284249,10.4855608369157,-4.10479708970297,6.42801793274623,1.32544341327556,7.79428043908965,-12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,3.06556135691961,-4.80059864950596,4.9775609184691,-3.98601838486204,2.14255044478909,-1.5947619510804,2.49736057507455,-2.5894196337366,2.07360079279396,-.444911232504622,.696720767591157,-.722403666316151,2.09440936915946,-3.27979695088037,-2.63186985254377,4.2894282422279,-6.71714607327472,6.9647571533504,-5.57736018045176,-.964148801340397,1.50983487071459,-1.56549122161564,-2.01722273728899,3.15892446006539,3.46190525493913,5.18623489409548,-8.12152468439221,8.42090472917074,.205483038001612,-.321781716436834,-3.63947886955732,5.81833621896754,-9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,2.74931534527504,-4.30536466145163,4.46407134676943,-3.5748172149004,1.9215230524055,.106757773657142,-.16718022074295,.173342908623153,-.13881256945113,-2.26282036321899,3.54352469706742,-3.67414802589247,2.15044627554921,-3.3675494587808,-.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155853,-4.48598459696891,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761006,1.33170919606025,3.7410698100085,-5.8584293670558,6.07438596565011,3.5848146864382,-5.61373743608464,-3.59724489879787,3.76406389114812,-5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,2.36915365304765,-3.71004018615404,3.84680169804461,-3.08050921776368,1.65582437345554,1.68415540504794,-2.63734867013603,2.73456804580791,-2.18983527963514,-2.32831086437971,3.64608131714141,-3.78048514370595,-.381749900992996,.597811573855878,2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,2.3938769796025,-3.74875634795766,3.88694503628021,4.93737708601548,-7.73181907803741,.508166439915615,2.07554270972349,-3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,1.94896143440088,-3.05202882623936,3.1645343667906,-2.53415123836086,1.36214797290775,2.83314651369184,-4.43664234772504,4.60018826185061,-3.68381930162858,-.637278791777431,.997963945475626,-1.03475143356998,-2.63766838605638,4.13052808940267,.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633232,4.12740556141282,-6.46342227832349,6.70168045166518,1.98940127224618,-3.11535668501805,-2.96832239843811,1.33269841309056,-2.0869750946082,2.16390630168515,4.37531622836173,-6.85164468857581,4.52786743647191,.950542947793588,-1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,1.51514018159912,-2.37267471198348,2.46013753301879,-1.97007200846434,1.0589461087365,3.36446686788906,-5.26867781509235,5.46289467143737,-4.37467244553793,1.91964817870762,-3.0061249410246,3.11693834864261,-1.95173675811295,3.05637492003247,-2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002441,-5.56722026499826,5.77244215424071,4.5904502888745,-7.18853968487789,1.24891795372526,.626153181688298,-.980541495928357,1.01668674800429,2.84473124098147,-4.45478376427994,5.48872263996487,.347192252628366,-.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,1.09494796295235,-1.7146633520688,1.77787020176478,-1.42371402906152,.765269708188713,3.24473228828999,-5.08117621439183,5.26848128513793,-4.21898669004741,4.05017972204568,-6.34248838569497,6.576288631721,1.63294619261043,-2.55715621899877,-2.28560008620039,.547225517982272,-.856942588087487,.888531670338269,-.711533948435091,2.39319484717323,-3.74768814487269,3.88583745543936,4.87179403320197,-7.62911752450547,5.49730994822616,.236321146004053,-.370073484889949,.383715334269613,1.3620925765208,-2.13300568008583,3.77635721863399,.0946962283839075,-.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,.714786270724957,-1.11933887677121,1.16060055303996,-.929406031924799,.499571029238757,2.59853682213183,-4.06924896099705,4.21925182103647,-3.37876634868698,4.7549644652192,-7.44616510000515,7.72064967553202,5.36731919542456,-8.40509854032022,3.25729414076559,.233201739437629,-.365188565907959,.378650344801755,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064311,5.89679444984458,.0657431866024185,-.102952319693431,.106747403902078,.451524561772771,-.707077089737652,1.59193561065049,.0171974308138372,-.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,.398540259080393,-.624104888716883,.647110981340288,-.518204861963161,.278543636855169,1.67165845871656,-2.61777873930361,2.71427671758158,-2.17358603453451,3.84755914106515,-6.02518921139496,6.24729300320555,6.3113399865174,-9.88341340931671,7.98777931164405,.07249762486996,-.113529614838378,.117714605047621,-.0942655625163295,.434860760346183,-.68098195938567,.706084685269141,1.40707187748205,-2.20344223207683,3.23180916099,.0113956405123911,-.0178453112143479,.0185031347484812,.0887334889355571,-.138954604977606,.36808995603673,.00166206328996398,-.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,.166073783576779,-.260067729310804,.269654486884814,-.215938641412748,.11607057156663,.764836159193308,-1.1977158528687,1.24186670359493,-.994483762912882,2.02218309922343,-3.16669227065688,3.28342459824587,4.02885967082133,-6.30910167728468,6.68010475273478,.0125887190510937,-.019713644795161,.02044033985683,-.0163685732441573,.0823390061849577,-.12894099187304,.133694084564441,.299004079114616,-.468233518150739,.796298871799331,.000824565926335038,-.00129125129528224,.00133885010038651,.00697737063218123,-.0109264021014796,.0321260785532874,5.01144631634162e-005,-7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,.0319379183519753,-.0500140463222003,.0518576912013895,-.0415275099413171,.0223217196472694,.154660030302817,-.24219405407586,.251121942525088,-.201097826062499,.43925822557731,-.687868288629403,.713224862474738,.961797763577864,-1.50615319946369,1.79684767953686,.000465578153648566,-.000729084691473618,.000755960606624651,-.000605371370823784,.00319092434145458,-.00499691420436583,.0051811131626592,.0123137463285649,-.0192830438371028,.0353622090543065,5.86464816381893e-006,-9.18390427744748e-006,9.52244633459444e-006,5.19109369737154e-005,-8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,-1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,3.43216369764802,-4.57682649415414,2.98636125886026,.126713592860271,-3.19855198842569,-4.00974905232829,5.34704266035319,-3.48892427303922,-.148037726033603,4.32546942586304,-5.76805910494959,3.7636358461794,-4.4144312457315,5.8866906319401,4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,7.27823584467215,-9.70560427278211,6.33286855712841,-7.46210387346585,9.9507942329456,7.31197439196321,9.14177937313724,-12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,3.29802783242322,-4.39795490297745,2.86964825021048,.121761370614664,-3.07354613898791,-3.07100849759875,4.09522222779324,-2.67211637191124,-.11337994190694,2.20598407348367,-2.94170303097044,1.91944964063812,-.923194488569861,1.23108958885198,-.507379758856801,4.96464287844228,-6.62040364808343,4.31978730278896,.183291880031408,-3.80884940873901,5.07914086475798,-3.31411940634736,2.02329330850018,-2.6980829707225,.0162698697481229,6.45780747384475,-8.61155438605811,5.61900531555837,-4.10479708970297,5.47379022059323,1.32544341327556,7.79428043908965,-10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,3.06556135691961,-4.08795840577754,2.66737675689462,.113178836410129,-2.85690259486619,-1.5947619510804,2.12663188500128,-1.38761892782618,-.0588777326765052,-.444911232504622,.593293821521023,-.387121881618236,2.09440936915946,-2.79291698671264,-2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.15836332764069,-.964148801340397,1.28570259705192,-.838915879897342,-2.01722273728899,2.68998779997356,3.46190525493913,5.18623489409548,-6.91589893787678,4.51259681489189,.205483038001612,-.274013798814207,-3.63947886955732,5.81833621896754,-7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,2.74931534527504,-3.6662410134057,2.39220781955879,.101503207887319,-2.56218200503226,.106757773657142,-.142362617365695,.0928910470423959,.0039414381903617,-2.26282036321899,3.01749482284837,-1.96890348630228,2.15044627554921,-2.86764279249033,-.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,3.7410698100085,-4.98875605403389,3.25514367429686,3.5848146864382,-4.78038820054912,-3.59724489879787,3.76406389114812,-5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,2.36915365304765,-3.15929138677771,2.06142518510909,.0874678476499899,-2.20789618310877,1.68415540504794,-2.24583899888368,1.46540110005889,.0621780897499632,-2.32831086437971,3.10482709694771,-2.02588745115913,-.381749900992996,.50906751987416,2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217327,3.12121330296807,-1.52344991306454,2.03153653003234,-2.69130383032704,2.3938769796025,-3.19226020392739,2.0829372085896,4.93737708601548,-6.58404442855622,.508166439915615,2.07554270972349,-2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,1.94896143440088,-2.59896063091726,1.69581157410879,.0719545824309833,-1.81630452989169,2.83314651369184,-3.77803076303363,2.46515019062261,.104598208499715,-.637278791777431,.849817990926997,-.55450289173293,-2.63766838605638,3.51735861632833,.808065843843811,1.73374759666501,-2.31197070864004,1.50855178059943,.0640090061407144,4.12740556141282,-5.50393885468462,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843811,1.33269841309056,-1.77716739172311,1.15959472274786,4.37531622836173,-5.83453034569908,4.52786743647191,.950542947793588,-1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,1.51514018159912,-2.02045541424857,1.31833919902212,.0559381407795183,-1.41201150863478,3.36446686788906,-4.4865520602846,2.92745754610215,.124214263271286,1.91964817870762,-2.55987109498181,1.67030580640284,-1.95173675811295,2.60266155489591,-2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981012,.0386847883193903,3.55510980002441,-4.74077641992772,3.0933379389912,4.5904502888745,-6.12141388513832,1.24891795372526,.626153181688298,-.834981872710091,.544822382907868,2.84473124098147,-3.79347911889773,5.48872263996487,.347192252628366,-.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,1.09494796295235,-1.46012465838812,.952725588021822,.0404248755605117,-1.02041985541771,3.24473228828999,-4.32688473530227,2.82327524550586,.119793728554495,4.05017972204568,-5.4009573811209,3.52410341735766,1.63294619261043,-2.17755097302669,-2.28560008620039,.547225517982272,-.729731000653747,.476146603536493,.0202032646552823,2.39319484717323,-3.19135057367102,2.08234367785444,4.87179403320197,-6.49658873611034,5.49730994822616,.236321146004053,-.315136741011863,.205625482170815,1.3620925765208,-1.81636481949506,3.77635721863399,.0946962283839075,-.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,.714786270724957,-.953175031760129,.621942953572119,.0263895153231826,-.666134033494214,2.59853682213183,-3.46517626442722,2.26101386266496,.0959365479342438,4.7549644652192,-6.34079527029531,4.13734393811778,5.36731919542456,-7.15737676442222,3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831033,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.40719665133859,5.89679444984458,.0657431866024185,-.0876692328212795,.0572038290823807,.451524561772771,-.602112766253273,1.59193561065049,.0171974308138372,-.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,.398540259080393,-.531457639388288,.346774016236295,.0147138868003722,-.371413443660288,1.67165845871656,-2.22917418911978,1.45452737733324,.0617167093760001,3.84755914106515,-5.13076069911761,3.34780114662786,6.3113399865174,-8.41623844737545,7.98777931164405,.07249762486996,-.0966763475074086,.0630809359431453,.00267657237098111,.434860760346183,-.579891411113904,.37837686142582,1.40707187748205,-1.8763451919045,3.23180916099,.0113956405123911,-.0151962068293123,.00991546509750554,.0887334889355571,-.118327043544488,.36808995603673,.00166206328996398,-.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,.166073783576779,-.221461142188379,.144502522920426,.00613135259583728,-.154769899538562,.764836159193308,-1.01991708658492,.665491881382992,.0282373296477263,2.02218309922343,-2.69660249304008,1.75952250506474,4.02885967082133,-5.37252686958436,6.68010475273478,.0125887190510937,-.0167871896471047,.0109535751204636,.000464768572193082,.0823390061849577,-.109799933312678,.0716440239592634,.299004079114616,-.398725093710571,.796298871799331,.000824565926335038,-.00109956736063812,.0007174633715654,.00697737063218123,-.00930439734224517,.0321260785532874,5.01144631634162e-005,-6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,.0319379183519753,-.0425895510116904,.0277895142706456,.00117913035023069,-.029764050100784,.154660030302817,-.206240781926363,.134571297792104,.00570996311627599,.43925822557731,-.585755477154869,.382203141612095,.961797763577864,-1.28256746328293,1.79684767953686,.000465578153648566,-.000620853379055138,.000405104384309318,1.7188888943926e-005,.00319092434145458,-.00425513126926133,.00277645639124963,.0123137463285649,-.0164205106275597,.0353622090543065,5.86464816381893e-006,-7.82057018924326e-006,5.1028912428631e-006,5.19109369737154e-005,-6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,-9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,3.43216369764802,-3.49141841400203,.133641330308316,3.40070220695544,-4.09773034500049,-4.00974905232829,4.07897551472765,-.156131305216531,-3.97299303311745,4.32546942586304,-4.40014665135235,.168424801181435,-4.4144312457315,4.49064435720703,4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,7.27823584467215,-7.40389121595287,.283399396550352,-7.46210387346585,7.59093364435532,7.31197439196321,9.14177937313724,-9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,3.29802783242322,-3.35496675520006,.128418358139773,3.26779592010949,-3.93758279561022,-3.07100849759875,3.12402803860063,-.119578696522851,-3.04285759691612,2.20598407348367,-2.24406936639925,.0858964409190727,-.923194488569861,.939133013954983,-.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.9191336688493,-3.80884940873901,3.87460743053655,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,6.45780747384475,-6.5692985292465,.251453618731751,-4.10479708970297,4.17566451229335,1.32544341327556,7.79428043908965,-7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,3.06556135691961,-3.11848685368264,.119366596109985,3.03746038662953,-3.6600363220783,-1.5947619510804,1.62229477839744,-.0620967201893262,-1.58014330533833,-.444911232504622,.452592418790165,-.0173239199033356,2.09440936915946,-2.13056837715884,-2.63186985254377,4.2894282422279,-4.36348323472718,.167021432333127,4.25010849784967,-.964148801340397,.980794384058671,-.0375419530619817,-2.01722273728899,2.05204915464121,3.46190525493913,5.18623489409548,-5.27577283777776,.201941221845807,.205483038001612,-.209030607582029,-3.63947886955732,5.81833621896754,-5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,2.74931534527504,-2.7967810011421,.107052632842487,2.72411329584901,-3.28246374904242,.106757773657142,-.108600897231078,.00415692611322449,.105779161098541,-2.26282036321899,2.30188690834251,-.0881095280677934,2.15044627554921,-2.18757273479106,-.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,3.7410698100085,-3.80565764777524,.14566949316231,3.5848146864382,-3.64670485341608,-3.59724489879787,3.76406389114812,-3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,2.36915365304765,-2.41005599340123,.0922499256416775,2.34743641804055,-2.82858093939833,1.68415540504794,-1.71323156899255,.0655775157362847,1.66871731961812,-2.32831086437971,2.36850807266163,-.090659592246103,-.381749900992996,.388340636274758,2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622404,2.53843733881177,3.58714638367868,-3.64907680538233,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,2.3938769796025,-2.43520615704218,.0932126006787416,4.93737708601548,-5.02261861703304,.508166439915615,2.07554270972349,-2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,1.94896143440088,-1.98260935074574,.0758885128327144,1.93109596018978,-2.32690486658693,2.83314651369184,-2.88205947758122,.110316844590442,2.80717601371158,-.637278791777431,.648281114842891,-.0248143133692,-2.63766838605638,2.68320650944122,.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336551,1.7178549162266,4.12740556141282,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843811,1.33269841309056,-1.35570683401507,.0518925099589829,4.37531622836173,-4.45085403872016,4.52786743647191,.950542947793588,-.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,1.51514018159912,-1.54129837497396,.0589964137232872,1.50125140095791,-1.80895681150723,3.36446686788906,-3.42255283189436,.131005356341758,3.33362593315312,1.91964817870762,-1.95279001506978,.074747115516283,-1.95173675811295,1.98543258954328,-2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063146,1.03821099255249,3.55510980002441,-3.61648712494208,.138428596350645,4.5904502888745,-4.66970228947897,1.24891795372526,.626153181688298,-.636963426396326,.0243811052053838,2.84473124098147,-2.89384420968738,5.48872263996487,.347192252628366,-.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,1.09494796295235,-1.11385173231848,.042635000914324,1.08491094310714,-1.30728073869583,3.24473228828999,-3.30075109017009,.126343140340609,3.21498892012256,4.05017972204568,-4.12010419835037,.157705591526293,1.63294619261043,-1.66113825253464,-2.28560008620039,.547225517982272,-.556673113172381,.0213078258129724,.542209285918407,2.39319484717323,-2.43451224831939,.093186039847633,4.87179403320197,-4.95590330339709,5.49730994822616,.236321146004053,-.240401121148629,.00920185489986045,1.3620925765208,-1.3856084746561,3.77635721863399,.0946962283839075,-.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,.714786270724957,-.727126724577601,.027832293713515,.70823406529868,-.853397929051741,2.59853682213183,-2.6433993582315,.101181630171365,2.57471691018525,4.7549644652192,-4.83705672356223,.185148446522551,5.36731919542456,-5.45998347614275,3.25729414076559,.233201739437629,-.237227859491313,.00908039168484126,.231064057614915,1.22672710731592,-1.24790598293054,.0477662072687216,3.30495822058058,-3.36201679531997,5.89679444984458,.0657431866024185,-.0668782122732,.00255990322490874,.451524561772771,-.459319924425253,1.59193561065049,.0171974308138372,-.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,.398540259080393,-.405420872037066,.0155183304460168,.394886974518159,-.47582535601586,1.67165845871656,-1.70051886866409,.0650909105855692,1.6563349285852,3.84755914106515,-3.91398546692074,.149815966677054,6.3113399865174,-6.42030234909073,7.98777931164405,.07249762486996,-.0737492627952914,.00282290703160039,.0718330635538729,.434860760346183,-.442368430061906,.0169325753770764,1.40707187748205,-1.43136432250755,3.23180916099,.0113956405123911,-.0115923809684749,.000443722587923622,.0887334889355571,-.0902654315677738,.36808995603673,.00166206328996398,-.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,.166073783576779,-.168940970519646,.00646656841622833,.164551441038204,-.198278882483944,.764836159193308,-.778040701653472,.0297811324981325,.757825163693696,2.02218309922343,-2.05709515348012,.0787396124921655,4.02885967082133,-4.09841606758444,6.68010475273478,.0125887190510937,-.0128060574566477,.0004901785898768,.0124733225023729,.0823390061849577,-.0837605510092491,.00320610999159926,.299004079114616,-.304166246215339,.796298871799331,.000824565926335038,-.000838801675193727,3.21068856418541e-005,.00697737063218123,-.00709783170793108,.0321260785532874,5.01144631634162e-005,-5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,.0319379183519753,-.0324893117176766,.00124359624768518,.0316451541922531,-.0381313330936754,.154660030302817,-.157330164177211,.00602214055814436,.153242313889452,.43925822557731,-.446841815317497,.0171038035473819,.961797763577864,-.978402756631806,1.79684767953686,.000465578153648566,-.000473616144898038,1.81286469184544e-005,.00046131035548159,.00319092434145458,-.00324601417302372,.000124247970525588,.0123137463285649,-.0125263374623454,.0353622090543065,5.86464816381893e-006,-5.96589859842733e-006,2.283572264044e-007,5.19109369737154e-005,-5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,-6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,3.43216369764802,-2.18664125535227,-2.27972534004502,3.88041880521056,-.538361161121001,-4.00974905232829,2.55462252942968,2.66337136913609,-4.53343925473611,4.32546942586304,-2.75576887572707,-2.87308040304256,-4.4144312457315,2.81244670509951,4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,7.27823584467215,-4.63698476184224,-4.83437858766286,-7.46210387346585,4.75412760808657,7.31197439196321,9.14177937313724,-5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,3.29802783242322,-2.10118291403716,-2.19062908534965,3.72876422817853,-.517320923384123,-3.07100849759875,1.95654825118803,2.0398374063157,-3.47209520918979,2.20598407348367,-1.40543872850124,-1.4652674627761,-.923194488569861,.588169835027997,-.507379758856801,4.96464287844228,-3.16298810157525,-3.29763472137029,5.61304625700717,-3.80884940873901,2.42662879416799,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,6.45780747384475,-4.11428751282488,-4.28943041158218,-4.10479708970297,2.61517790433631,1.32544341327556,7.79428043908965,-4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,3.06556135691961,-1.95307786118934,-2.03621928395244,3.46593664692318,-.48085677636235,-1.5947619510804,1.01602737631215,1.05927908845788,-1.80304461367463,-.444911232504622,.283454211851926,.295520697836744,2.09440936915946,-1.33435416739534,-2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340397,.614261944257266,.640410728572749,-2.01722273728899,1.28517834515454,3.46190525493913,5.18623489409548,-3.30416500488362,-3.44482144542742,.205483038001612,-.130913828127032,-3.63947886955732,5.81833621896754,-3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,2.74931534527504,-1.75159662753591,-1.82616110784378,3.10838756746075,-.431251167473262,.106757773657142,-.0680156813349527,-.070911070513513,.120700790879792,-2.26282036321899,1.44164929037179,1.50301948784968,2.15044627554921,-1.37005543987348,-.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,3.7410698100085,-2.38344621859102,-2.48490817973373,3.5848146864382,-2.28389563582347,-3.59724489879787,3.76406389114812,-2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,2.36915365304765,-1.50939453196029,-1.57364860569271,2.6785751489707,-.371619894588022,1.68415540504794,-1.07298020016939,-1.11865636189275,1.9041132311976,-2.32831086437971,1.48337347495979,1.54651984745118,-.381749900992996,.243213947948102,2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001398,-2.69130383032704,2.3938769796025,-1.52514583383108,-1.59007043171884,4.93737708601548,-3.14561698874583,.508166439915615,2.07554270972349,-1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,1.94896143440088,-1.24168887412682,-1.29454686902578,2.20350404785819,-.305709527061058,2.83314651369184,-1.80500570449659,-1.88184377891507,3.20316744314248,-.637278791777431,.406012131085817,.423295838593301,-2.63766838605638,1.68046603255685,.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141282,-2.62958182530229,-2.74152163963163,1.98940127224618,-1.26745321062054,-2.96832239843811,1.33269841309056,-.849065950254408,-.885210208841134,4.37531622836173,-2.78752641722509,4.52786743647191,.950542947793588,-.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,1.51514018159912,-.965300119862265,-1.00639240141108,1.71302390303713,-.237661341149246,3.36446686788906,-2.14351141382237,-2.23475948529486,3.80388048506198,1.91964817870762,-1.22301331540983,-1.27507630270228,-1.95173675811295,1.24345704101385,-2.75500656898115,1.04781596872157,-.66756653456346,-.695984465482169,1.18466517033957,3.55510980002441,-2.2649705675878,-2.36138908644034,4.5904502888745,-2.92458893939423,1.24891795372526,.626153181688298,-.398923973340749,-.415905941872819,2.84473124098147,-1.81238636712449,5.48872263996487,.347192252628366,-.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,1.09494796295235,-.697594462028796,-.727290664744143,1.23795280192461,-.171750973622283,3.24473228828999,-2.06722817250133,-2.15522890943138,3.66850806244401,4.05017972204568,-2.58038102231816,-2.69022638759908,1.63294619261043,-1.04035466481729,-2.28560008620039,.547225517982272,-.348638934287124,-.363480297113412,.61869548740089,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320197,-3.10383383936015,5.49730994822616,.236321146004053,-.150560874408055,-.156970165907753,1.3620925765208,-.867793055767215,3.77635721863399,.0946962283839075,-.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,.714786270724957,-.455392366453167,-.474778162593069,.80814038343456,-.112119700737043,2.59853682213183,-1.65553520251249,-1.7260104019959,2.93791673259187,4.7549644652192,-3.02940138707022,-3.15836129611741,5.36731919542456,-3.41953432868287,3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066826,3.30495822058058,-2.10559828531327,5.89679444984458,.0657431866024185,-.0418851712113071,-.0436681993244439,.451524561772771,-.287667582930082,1.59193561065049,.0171974308138372,-.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,.398540259080393,-.253911132799744,-.264719986484413,.450591303972135,-.0625140918479549,1.67165845871656,-1.06501836010644,-1.11035559079056,1.88998413076671,3.84755914106515,-2.45129087378787,-2.55564094422791,6.3113399865174,-4.0209726602962,7.98777931164405,.07249762486996,-.0461884430503283,-.0481546590241492,.0819661216087146,.434860760346183,-.277051027559316,-.288844933452007,1.40707187748205,-.896449496423069,3.23180916099,.0113956405123911,-.00726019498888195,-.00756925739444607,.0887334889355571,-.056532358254944,.36808995603673,.00166206328996398,-.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,.166073783576779,-.105806079951921,-.110310185087204,.187763722716786,-.026049944826181,.764836159193308,-.487279293067771,-.508022497640507,.864726999689721,2.02218309922343,-1.28833860494792,-1.34318245298509,4.02885967082133,-2.56679795782664,6.68010475273478,.0125887190510937,-.00802030872060961,-.00836172873999097,.0142328590555998,.0823390061849577,-.0524584150880736,-.0546915402140795,.299004079114616,-.190496349504476,.796298871799331,.000824565926335038,-.000525333297401842,-.000547696439558021,.00697737063218123,-.00444530268095806,.0321260785532874,5.01144631634162e-005,-3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,.0319379183519753,-.0203477386368136,-.0212139303918346,.0361091456847505,-.00500970708930343,.154660030302817,-.0985343453312713,-.102728896816887,.174859285048361,.43925822557731,-.279852665057626,-.291765835225991,.961797763577864,-.612764091356553,1.79684767953686,.000465578153648566,-.000296621166194705,-.000309248161967433,.000526384631617522,.00319092434145458,-.00203294697585556,-.00211948838134787,.0123137463285649,-.00784512281993512,.0353622090543065,5.86464816381893e-006,-3.73638402898161e-006,-3.89543978958814e-006,5.19109369737154e-005,-3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,-4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,3.43216369764802,-.744472777196595,-3.65673100407672,1.66121346489508,3.26669916121878,-4.00974905232829,.869757178740834,4.27210791134071,-1.94077255839989,4.32546942586304,-.938240282077712,-4.60848594454081,-4.4144312457315,.957537046434959,4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,7.27823584467215,-1.5787266955735,-7.75445258971519,-7.46210387346585,1.61860962555059,7.31197439196321,9.14177937313724,-1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,3.29802783242322,-.715377282662343,-3.51381859652975,1.5962898991603,3.13902998310834,-3.07100849759875,.666134376965478,3.2719453330298,-1.4864094837286,2.20598407348367,-.47850138692202,-2.35032214902362,-.923194488569861,.200250694746914,-.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390441,-3.80884940873901,.82617991094677,4.05806335333585,2.02329330850018,-.438873818984882,.0162698697481229,6.45780747384475,-1.40076706380652,-6.88034339064389,-4.10479708970297,.890374108024764,1.32544341327556,7.79428043908965,-1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,3.06556135691961,-.66495283386875,-3.26614178293108,1.48377299342298,2.91777071127945,-1.5947619510804,.345920813804923,1.69910761471712,-.771886268054593,-.444911232504622,.0965059740773432,.474021882875009,2.09440936915946,-.454299648071273,-2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340397,.209134120399515,1.02723329226541,-2.01722273728899,.437557047580714,3.46190525493913,5.18623489409548,-1.1249494591216,-5.52557150711589,.205483038001612,-.0445714544901277,-3.63947886955732,5.81833621896754,-1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,2.74931534527504,-.596355713420295,-2.92920371774241,1.33070566358573,2.61677091290561,.106757773657142,-.0231568955595295,-.113742960840148,.0516722006304461,-2.26282036321899,.490829782123398,2.41087725104767,2.15044627554921,-.466454648534285,-.119501749587716,3.45007465931297,-.74835785526233,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239337,-1.88618542328239,-3.42380239520456,.742659122293478,1.33170919606025,3.7410698100085,-.811477795596299,-3.98584891938638,3.5848146864382,-.777584399208989,-3.59724489879787,3.76406389114812,-.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,2.36915365304765,-.513894602666682,-2.52416795342767,1.1467022833279,2.25493680750505,1.68415540504794,-.365311203903165,-1.79435010419287,.815153904032968,-2.32831086437971,.505035367744937,2.48065281165978,-.381749900992996,.0828056101286271,2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.330452474967379,-2.69130383032704,2.3938769796025,-.519257354978935,-2.55050893377231,4.93737708601548,-1.07096955656365,.508166439915615,2.07554270972349,-.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,1.94896143440088,-.422750445356608,-2.07648245560297,.943323589025389,1.85500204648398,2.83314651369184,-.614539585033937,-3.01851998146924,1.37128107917545,-.637278791777431,.138232541833329,.678976098432196,-2.63766838605638,.57213830168982,.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.8391571928101,4.12740556141282,-.89527812555293,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843811,1.33269841309056,-.289076447479325,-1.41989719504135,4.37531622836173,-.949052583847877,4.52786743647191,.950542947793588,-.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,1.51514018159912,-.328650005712205,-1.6142761726,.733348258592925,1.44209530674498,3.36446686788906,-.729788615894315,-3.58460475610227,1.62844728825442,1.91964817870762,-.416392088662432,-2.04525122687055,-1.95173675811295,.423352442647985,-2.75500656898115,1.04781596872157,-.227282418061714,-1.11637482313293,.507157044419472,3.55510980002441,-.771141093489059,-3.78772150021559,4.5904502888745,-.995717447724869,1.24891795372526,.626153181688298,-.135819278812355,-.66712253696142,2.84473124098147,-.617052435457252,5.48872263996487,.347192252628366,-.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,1.09494796295235,-.237505848402132,-1.1665906747753,.529969564290411,1.04216054572391,3.24473228828999,-.703816913228947,-3.45703591374062,1.57049413873188,4.05017972204568,-.878527020016876,-4.31518396831288,1.63294619261043,-.354203381414383,-2.28560008620039,.547225517982272,-.118699029853337,-.583030617100297,.264864522602584,2.39319484717323,-.519109393293246,-2.54978216897294,4.87179403320197,-1.05674389554722,5.49730994822616,.236321146004053,-.0512605677924316,-.251783696179213,1.3620925765208,-.295452354005303,3.77635721863399,.0946962283839075,-.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,.714786270724957,-.155044737648518,-.76155491046056,.345966184032581,.680326440323354,2.59853682213183,-.563650249872674,-2.76855972053751,1.25772682793123,4.7549644652192,-1.03140232004452,-5.06608294899263,5.36731919542456,-1.16422856843223,3.25729414076559,.233201739437629,-.0505839353643847,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929027,5.89679444984458,.0657431866024185,-.0142603957822122,-.0700447793236106,.451524561772771,-.0979404755348809,1.59193561065049,.0171974308138372,-.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,.398540259080393,-.0864476172000634,-.4246168452719,.192898854195334,.379326641949509,1.67165845871656,-.362600483446047,-1.78103547961334,.809105213656782,3.84755914106515,-.834576462901105,-4.09930587333892,6.3113399865174,-1.3689967095037,7.98777931164405,.07249762486996,-.0157255052185161,-.0772411621484507,.0350898270822693,.434860760346183,-.0943259199045095,-.463313805821701,1.40707187748205,-.305208842367266,3.23180916099,.0113956405123911,-.00247183552086188,-.0121412600482659,.0887334889355571,-.0192472366688018,.36808995603673,.00166206328996398,-.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,.166073783576779,-.0360231684064709,-.176940031673225,.080381948458016,.158067370120622,.764836159193308,-.165901090401825,-.814879575736018,.370191004490028,2.02218309922343,-.43863300248611,-2.1544950322241,4.02885967082133,-.873902474623302,6.68010475273478,.0125887190510937,-.00273062693415816,-.0134124020091826,.0060931096085216,.0823390061849577,-.0178602054051424,-.0877264674162827,.299004079114616,-.0648571621038786,.796298871799331,.000824565926335038,-.000178857111523577,-.000878517475844129,.00697737063218123,-.00151346583454046,.0321260785532874,5.01144631634162e-005,-1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,.0319379183519753,-.00692767387221854,-.0340276241262534,.0154583827232305,.0303981920101794,.154660030302817,-.0335474040556334,-.16477947382797,.0748575381590809,.43925822557731,-.0952797768044285,-.467998998326769,.961797763577864,-.208624155307574,1.79684767953686,.000465578153648566,-.000100988848916915,-.000496041045803229,.000225346098359972,.00319092434145458,-.000692145397560786,-.00339970729838099,.0123137463285649,-.00267098242974684,.0353622090543065,5.86464816381893e-006,-1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,-1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,-1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,3.43216369764802,.744472777196595,-3.65673100407672,-1.66121346489509,3.26669916121882,-4.00974905232829,-.869757178740834,4.27210791134071,1.9407725583999,4.32546942586304,.938240282077712,-4.60848594454081,-4.4144312457315,-.957537046434959,4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,7.27823584467215,1.5787266955735,-7.75445258971519,-7.46210387346585,-1.61860962555059,7.31197439196321,9.14177937313724,1.98294908983919,-8.99870040642758,10.9691712594618},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,3.29802783242322,.715377282662343,-3.51381859652975,-1.5962898991603,3.13902998310838,-3.07100849759875,-.666134376965478,3.2719453330298,1.48640948372861,2.20598407348367,.47850138692202,-2.35032214902362,-.923194488569861,-.200250694746914,-.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873901,-.82617991094677,4.05806335333585,2.02329330850018,.438873818984882,.0162698697481229,6.45780747384475,1.40076706380652,-6.8803433906439,-4.10479708970297,-.890374108024764,1.32544341327556,7.79428043908965,1.69066225204035,-4.03553981216457,8.98680735137072},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,3.06556135691961,.66495283386875,-3.26614178293108,-1.48377299342298,2.91777071127949,-1.5947619510804,-.345920813804923,1.69910761471712,.771886268054594,-.444911232504622,-.0965059740773432,.474021882875009,2.09440936915946,.454299648071273,-2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340397,-.209134120399515,1.02723329226541,-2.01722273728899,-.437557047580714,3.46190525493913,5.18623489409548,1.1249494591216,-5.52557150711589,.205483038001612,.0445714544901277,-3.63947886955732,5.81833621896754,1.26205895360836,1.69243379052583,6.23568135536064},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,2.74931534527504,.596355713420295,-2.92920371774242,-1.33070566358573,2.61677091290564,.106757773657142,.0231568955595295,-.113742960840148,-.0516722006304461,-2.26282036321899,-.490829782123398,2.41087725104768,2.15044627554921,.466454648534285,-.119501749587716,3.45007465931297,.74835785526233,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239337,-1.88618542328239,-3.42380239520456,-.742659122293478,1.33170919606025,3.7410698100085,.811477795596299,-3.98584891938638,3.5848146864382,.777584399208989,-3.59724489879787,3.76406389114812,.816465456274434,5.2355975909823,3.61790059797515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,2.36915365304765,.513894602666682,-2.52416795342767,-1.14670228332791,2.25493680750508,1.68415540504794,.365311203903165,-1.79435010419287,-.81515390403297,-2.32831086437971,-.505035367744937,2.48065281165978,-.381749900992996,-.0828056101286271,2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.330452474967379,-2.69130383032704,2.3938769796025,.519257354978935,-2.55050893377231,4.93737708601548,1.07096955656365,.508166439915615,2.07554270972349,.450207269195575,5.63164748396115,1.71909544841975},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,1.94896143440088,.422750445356608,-2.07648245560297,-.94332358902539,1.85500204648401,2.83314651369184,.614539585033937,-3.01851998146924,-1.37128107917545,-.637278791777431,-.138232541833329,.678976098432196,-2.63766838605638,-.57213830168982,.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810101,4.12740556141282,.89527812555293,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843811,1.33269841309056,.289076447479325,-1.41989719504135,4.37531622836173,.949052583847877,4.52786743647191,.950542947793588,.206182866184563,3.96849611187714,.647664476709578},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,1.51514018159912,.328650005712205,-1.6142761726,-.733348258592926,1.44209530674499,3.36446686788906,.729788615894315,-3.58460475610227,-1.62844728825442,1.91964817870762,.416392088662432,-2.04525122687055,-1.95173675811295,-.423352442647985,-2.75500656898115,1.04781596872157,.227282418061714,-1.11637482313293,-.507157044419472,3.55510980002441,.771141093489059,-3.78772150021559,4.5904502888745,.995717447724869,1.24891795372526,.626153181688298,.135819278812355,-.66712253696142,2.84473124098147,.617052435457252,5.48872263996487,.347192252628366,.0753096889837071,1.97345041935925,.183906859961575},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,1.09494796295235,.237505848402132,-1.1665906747753,-.529969564290411,1.04216054572392,3.24473228828999,.703816913228947,-3.45703591374062,-1.57049413873189,4.05017972204568,.878527020016876,-4.31518396831288,1.63294619261043,.354203381414383,-2.28560008620039,.547225517982272,.118699029853337,-.583030617100298,-.264864522602585,2.39319484717323,.519109393293246,-2.54978216897294,4.87179403320197,1.05674389554722,5.49730994822616,.236321146004053,.0512605677924316,-.251783696179213,1.3620925765208,.295452354005303,3.77635721863399,.0946962283839075,.0205406182123418,.676667877319621,.0362494356624604},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,.714786270724957,.155044737648518,-.761554910460561,-.345966184032582,.680326440323361,2.59853682213183,.563650249872674,-2.76855972053751,-1.25772682793124,4.7549644652192,1.03140232004452,-5.06608294899263,5.36731919542456,1.16422856843223,3.25729414076559,.233201739437629,.0505839353643847,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929027,5.89679444984458,.0657431866024185,.0142603957822122,-.0700447793236106,.451524561772771,.0979404755348809,1.59193561065049,.0171974308138372,.00373030548955022,.145628931559543,.00429748986299419},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,.398540259080393,.0864476172000634,-.4246168452719,-.192898854195334,.379326641949514,1.67165845871656,.362600483446047,-1.78103547961334,-.809105213656783,3.84755914106515,.834576462901105,-4.09930587333892,6.3113399865174,1.3689967095037,7.98777931164405,.07249762486996,.0157255052185161,-.0772411621484508,-.0350898270822694,.434860760346183,.0943259199045095,-.463313805821702,1.40707187748205,.305208842367266,3.23180916099,.0113956405123911,.00247183552086188,-.0121412600482659,.0887334889355571,.0192472366688018,.36808995603673,.00166206328996398,.00036051918927006,.0159028321410879,.000231576691000447},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,.166073783576779,.0360231684064709,-.176940031673225,-.0803819484580161,.158067370120624,.764836159193308,.165901090401825,-.814879575736019,-.370191004490028,2.02218309922343,.43863300248611,-2.15449503222411,4.02885967082133,.873902474623302,6.68010475273478,.0125887190510937,.00273062693415816,-.0134124020091826,-.00609310960852161,.0823390061849577,.0178602054051424,-.0877264674162827,.299004079114616,.0648571621038786,.796298871799331,.000824565926335038,.000178857111523577,-.000878517475844129,.00697737063218123,.00151346583454046,.0321260785532874,5.01144631634162e-005,1.08703595942912e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,.0319379183519753,.00692767387221854,-.0340276241262535,-.0154583827232306,.0303981920101797,.154660030302817,.0335474040556334,-.16477947382797,-.074857538159081,.43925822557731,.0952797768044285,-.467998998326769,.961797763577864,.208624155307574,1.79684767953686,.000465578153648566,.000100988848916915,-.00049604104580323,-.000225346098359973,.00319092434145458,.000692145397560786,-.00339970729838099,.0123137463285649,.00267098242974684,.0353622090543065,5.86464816381893e-006,1.2721045063612e-006,-6.24837352240853e-006,5.19109369737154e-005,1.12600338566422e-005,.000252650701908931,6.85464954370287e-008,1.48684632597818e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,3.43216369764802,2.18664125535227,-2.27972534004502,-3.88041880521058,-.538361161120983,-4.00974905232829,-2.55462252942968,2.6633713691361,4.53343925473613,4.32546942586304,2.75576887572707,-2.87308040304257,-4.4144312457315,-2.81244670509951,4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,7.27823584467215,4.63698476184224,-4.83437858766287,-7.46210387346585,-4.75412760808657,7.31197439196321,9.14177937313724,5.82425364622913,-8.99870040642758,10.9691712594618},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,3.29802783242322,2.10118291403716,-2.19062908534965,-3.72876422817854,-.517320923384106,-3.07100849759875,-1.95654825118803,2.0398374063157,3.47209520918981,2.20598407348367,1.40543872850124,-1.46526746277611,-.923194488569861,-.588169835027997,-.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873901,-2.42662879416799,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,6.45780747384475,4.11428751282488,-4.28943041158219,-4.10479708970297,-2.61517790433631,1.32544341327556,7.79428043908965,4.96575824182481,-4.03553981216457,8.98680735137072},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,3.06556135691961,1.95307786118934,-2.03621928395244,-3.46593664692319,-.480856776362333,-1.5947619510804,-1.01602737631215,1.05927908845788,1.80304461367463,-.444911232504622,-.283454211851926,.295520697836744,2.09440936915946,1.33435416739534,-2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611566,-4.8496457306699,-.964148801340397,-.614261944257266,.640410728572749,-2.01722273728899,-1.28517834515454,3.46190525493913,5.18623489409548,3.30416500488362,-3.44482144542743,.205483038001612,.130913828127032,-3.63947886955732,5.81833621896754,3.70687855778774,1.69243379052583,6.23568135536064},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,2.74931534527504,1.75159662753591,-1.82616110784379,-3.10838756746077,-.431251167473247,.106757773657142,.0680156813349527,-.0709110705135131,-.120700790879792,-2.26282036321899,-1.44164929037179,1.50301948784969,2.15044627554921,1.37005543987348,-.119501749587716,3.45007465931297,2.19805238043167,-2.2916222307293,-3.90066901628344,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,3.7410698100085,2.38344621859102,-2.48490817973374,3.5848146864382,2.28389563582347,-3.59724489879787,3.76406389114812,2.39809581349975,5.2355975909823,3.61790059797515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,2.36915365304765,1.50939453196029,-1.57364860569271,-2.67857514897071,-.371619894588009,1.68415540504794,1.07298020016939,-1.11865636189276,-1.90411323119761,-2.32831086437971,-1.48337347495979,1.54651984745118,-.381749900992996,-.243213947948102,2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188741,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001398,-2.69130383032704,2.3938769796025,1.52514583383108,-1.59007043171884,4.93737708601548,3.14561698874583,.508166439915615,2.07554270972349,1.32233416511154,5.63164748396115,1.71909544841975},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,1.94896143440088,1.24168887412682,-1.29454686902578,-2.2035040478582,-.305709527061048,2.83314651369184,1.80500570449659,-1.88184377891507,-3.20316744314249,-.637278791777431,-.406012131085817,.423295838593301,-2.63766838605638,-1.68046603255685,.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141282,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843811,1.33269841309056,.849065950254408,-.885210208841136,4.37531622836173,2.78752641722509,4.52786743647191,.950542947793588,.60559361625507,3.96849611187714,.647664476709578},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,1.51514018159912,.965300119862265,-1.00639240141108,-1.71302390303713,-.237661341149238,3.36446686788906,2.14351141382237,-2.23475948529486,-3.80388048506199,1.91964817870762,1.22301331540983,-1.27507630270228,-1.95173675811295,-1.24345704101385,-2.75500656898115,1.04781596872157,.66756653456346,-.69598446548217,-1.18466517033958,3.55510980002441,2.2649705675878,-2.36138908644035,4.5904502888745,2.92458893939423,1.24891795372526,.626153181688298,.398923973340749,-.41590594187282,2.84473124098147,1.81238636712449,5.48872263996487,.347192252628366,.221197171882668,1.97345041935925,.183906859961575},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,1.09494796295235,.697594462028796,-.727290664744144,-1.23795280192461,-.171750973622277,3.24473228828999,2.06722817250133,-2.15522890943138,-3.66850806244403,4.05017972204568,2.58038102231816,-2.69022638759909,1.63294619261043,1.04035466481729,-2.28560008620039,.547225517982272,.348638934287124,-.363480297113413,-.618695487400893,2.39319484717323,1.52471124557473,-1.58961734284588,4.87179403320197,3.10383383936015,5.49730994822616,.236321146004053,.150560874408055,-.156970165907753,1.3620925765208,.867793055767215,3.77635721863399,.0946962283839075,.0603312365062958,.676667877319621,.0362494356624604},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,.714786270724957,.455392366453167,-.47477816259307,-.808140383434563,-.112119700737039,2.59853682213183,1.65553520251249,-1.7260104019959,-2.93791673259189,4.7549644652192,3.02940138707022,-3.15836129611742,5.36731919542456,3.41953432868287,3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066827,3.30495822058058,2.10559828531327,5.89679444984458,.0657431866024185,.0418851712113071,-.043668199324444,.451524561772771,.287667582930082,1.59193561065049,.0171974308138372,.0109565320967587,.145628931559543,.00429748986299419},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,.398540259080393,.253911132799744,-.264719986484414,-.450591303972137,-.0625140918479527,1.67165845871656,1.06501836010644,-1.11035559079056,-1.88998413076672,3.84755914106515,2.45129087378787,-2.55564094422791,6.3113399865174,4.0209726602962,7.98777931164405,.07249762486996,.0461884430503283,-.0481546590241493,-.081966121608715,.434860760346183,.277051027559316,-.288844933452007,1.40707187748205,.896449496423069,3.23180916099,.0113956405123911,.00726019498888195,-.00756925739444608,.0887334889355571,.056532358254944,.36808995603673,.00166206328996398,.00105890525046813,.0159028321410879,.000231576691000447},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,.166073783576779,.105806079951921,-.110310185087204,-.187763722716787,-.0260499448261801,.764836159193308,.487279293067771,-.508022497640508,-.864726999689725,2.02218309922343,1.28833860494792,-1.34318245298509,4.02885967082133,2.56679795782664,6.68010475273478,.0125887190510937,.00802030872060961,-.00836172873999098,-.0142328590555999,.0823390061849577,.0524584150880736,-.0546915402140795,.299004079114616,.190496349504476,.796298871799331,.000824565926335038,.000525333297401842,-.000547696439558022,.00697737063218123,.00444530268095806,.0321260785532874,5.01144631634162e-005,3.19280670528998e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,.0319379183519753,.0203477386368136,-.0212139303918347,-.0361091456847507,-.00500970708930326,.154660030302817,.0985343453312713,-.102728896816887,-.174859285048362,.43925822557731,.279852665057626,-.291765835225991,.961797763577864,.612764091356553,1.79684767953686,.000465578153648566,.000296621166194705,-.000309248161967434,-.000526384631617524,.00319092434145458,.00203294697585556,-.00211948838134787,.0123137463285649,.00784512281993512,.0353622090543065,5.86464816381893e-006,3.73638402898161e-006,-3.89543978958815e-006,5.19109369737154e-005,3.30726056368548e-005,.000252650701908931,6.85464954370287e-008,4.367116725202e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,3.43216369764802,3.49141841400203,.13364133030831,-3.40070220695543,-4.09773034500053,-4.00974905232829,-4.07897551472765,-.156131305216524,3.97299303311744,4.32546942586304,4.40014665135235,.168424801181427,-4.4144312457315,-4.49064435720703,4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,7.27823584467215,7.40389121595287,.283399396550339,-7.46210387346585,-7.59093364435532,7.31197439196321,9.14177937313724,9.29960796197769,-8.99870040642758,10.9691712594618},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,3.29802783242322,3.35496675520006,.128418358139767,-3.26779592010948,-3.93758279561025,-3.07100849759875,-3.12402803860063,-.119578696522845,3.04285759691611,2.20598407348367,2.24406936639925,.0858964409190686,-.923194488569861,-.939133013954983,-.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873901,-3.87460743053655,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,6.45780747384475,6.5692985292465,.251453618731739,-4.10479708970297,-4.17566451229335,1.32544341327556,7.79428043908965,7.92884508263629,-4.03553981216457,8.98680735137072},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,3.06556135691961,3.11848685368264,.119366596109979,-3.03746038662953,-3.66003632207834,-1.5947619510804,-1.62229477839744,-.0620967201893232,1.58014330533832,-.444911232504622,-.452592418790165,-.0173239199033348,2.09440936915946,2.13056837715884,-2.63186985254377,4.2894282422279,4.36348323472718,.167021432333119,-4.25010849784966,-.964148801340397,-.980794384058671,-.0375419530619799,-2.01722273728899,-2.05204915464121,3.46190525493913,5.18623489409548,5.27577283777776,.201941221845797,.205483038001612,.209030607582029,-3.63947886955732,5.81833621896754,5.91878710028472,1.69243379052583,6.23568135536064},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,2.74931534527504,2.7967810011421,.107052632842481,-2.72411329584901,-3.28246374904245,.106757773657142,.108600897231078,.00415692611322429,-.105779161098541,-2.26282036321899,-2.30188690834251,-.0881095280677891,2.15044627554921,2.18757273479106,-.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.41844898660569,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,3.7410698100085,3.80565764777524,.145669493162303,3.5848146864382,3.64670485341608,-3.59724489879787,3.76406389114812,3.82904871171717,5.2355975909823,3.61790059797515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,2.36915365304765,2.41005599340123,.0922499256416731,-2.34743641804054,-2.82858093939836,1.68415540504794,1.71323156899255,.0655775157362815,-1.66871731961812,-2.32831086437971,-2.36850807266163,-.0906595922460986,-.381749900992996,-.388340636274758,2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622356,-2.53843733881177,3.58714638367868,3.64907680538233,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,2.3938769796025,2.43520615704218,.0932126006787372,4.93737708601548,5.02261861703304,.508166439915615,2.07554270972349,2.11137599376842,5.63164748396115,1.71909544841975},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,1.94896143440088,1.98260935074574,.0758885128327108,-1.93109596018978,-2.32690486658695,2.83314651369184,2.88205947758122,.110316844590437,-2.80717601371157,-.637278791777431,-.648281114842891,-.0248143133691988,-2.63766838605638,-2.68320650944122,.808065843843811,1.73374759666501,1.76367994615688,.0675085327336519,-1.71785491622659,4.12740556141282,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843811,1.33269841309056,1.35570683401507,.0518925099589804,4.37531622836173,4.45085403872016,4.52786743647191,.950542947793588,.966953631748983,3.96849611187714,.647664476709578},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,1.51514018159912,1.54129837497396,.0589964137232844,-1.50125140095791,-1.80895681150725,3.36446686788906,3.42255283189436,.131005356341752,-3.33362593315312,1.91964817870762,1.95279001506978,.0747471155162795,-1.95173675811295,-1.98543258954328,-2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063126,-1.03821099255249,3.55510980002441,3.61648712494208,.138428596350638,4.5904502888745,4.66970228947897,1.24891795372526,.626153181688298,.636963426396326,.0243811052053826,2.84473124098147,2.89384420968738,5.48872263996487,.347192252628366,.353186366143038,1.97345041935925,.183906859961575},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,1.09494796295235,1.11385173231848,.042635000914322,-1.08491094310714,-1.30728073869584,3.24473228828999,3.30075109017009,.126343140340603,-3.21498892012255,4.05017972204568,4.12010419835037,.157705591526285,1.63294619261043,1.66113825253464,-2.28560008620039,.547225517982272,.556673113172381,.0213078258129714,-.542209285918406,2.39319484717323,2.43451224831939,.0931860398476285,4.87179403320197,4.95590330339709,5.49730994822616,.236321146004053,.240401121148629,.00920185489986001,1.3620925765208,1.3856084746561,3.77635721863399,.0946962283839075,.0963311149288904,.676667877319621,.0362494356624604},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,.714786270724957,.727126724577601,.0278322937135137,-.708234065298678,-.853397929051749,2.59853682213183,2.6433993582315,.10118163017136,-2.57471691018525,4.7549644652192,4.83705672356223,.185148446522542,5.36731919542456,5.45998347614275,3.25729414076559,.233201739437629,.237227859491313,.00908039168484082,-.231064057614914,1.22672710731592,1.24790598293054,.0477662072687193,3.30495822058058,3.36201679531997,5.89679444984458,.0657431866024185,.0668782122732,.00255990322490862,.451524561772771,.459319924425253,1.59193561065049,.0171974308138372,.0174943364955697,.145628931559543,.00429748986299419},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,.398540259080393,.405420872037066,.015518330446016,-.394886974518158,-.475825356015865,1.67165845871656,1.70051886866409,.0650909105855661,-1.6563349285852,3.84755914106515,3.91398546692074,.149815966677047,6.3113399865174,6.42030234909073,7.98777931164405,.07249762486996,.0737492627952914,.00282290703160026,-.0718330635538727,.434860760346183,.442368430061906,.0169325753770755,1.40707187748205,1.43136432250755,3.23180916099,.0113956405123911,.0115923809684749,.0004437225879236,.0887334889355571,.0902654315677738,.36808995603673,.00166206328996398,.00169075804324028,.0159028321410879,.000231576691000447},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,.166073783576779,.168940970519646,.00646656841622802,-.164551441038203,-.198278882483946,.764836159193308,.778040701653472,.0297811324981311,-.757825163693694,2.02218309922343,2.05709515348012,.0787396124921617,4.02885967082133,4.09841606758444,6.68010475273478,.0125887190510937,.0128060574566477,.000490178589876777,-.0124733225023729,.0823390061849577,.0837605510092491,.00320610999159911,.299004079114616,.304166246215339,.796298871799331,.000824565926335038,.000838801675193727,3.21068856418526e-005,.00697737063218123,.00709783170793108,.0321260785532874,5.01144631634162e-005,5.09796661702641e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,.0319379183519753,.0324893117176766,.00124359624768512,-.031645154192253,-.0381313330936757,.154660030302817,.157330164177211,.00602214055814407,-.153242313889452,.43925822557731,.446841815317497,.0171038035473811,.961797763577864,.978402756631806,1.79684767953686,.000465578153648566,.000473616144898038,1.81286469184535e-005,-.000461310355481589,.00319092434145458,.00324601417302372,.000124247970525582,.0123137463285649,.0125263374623454,.0353622090543065,5.86464816381893e-006,5.96589859842733e-006,2.28357226404389e-007,5.19109369737154e-005,5.28071552829729e-005,.000252650701908931,6.85464954370287e-008,6.97299189482736e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,3.43216369764802,4.57682649415414,2.98636125886026,-.126713592860266,-3.19855198842569,-4.00974905232829,-5.34704266035319,-3.48892427303921,.148037726033597,4.32546942586304,5.76805910494959,3.76363584617939,-4.4144312457315,-5.8866906319401,4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,7.27823584467215,9.70560427278211,6.3328685571284,-7.46210387346585,-9.9507942329456,7.31197439196321,9.14177937313724,12.1906592274739,-8.99870040642758,10.9691712594618},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,3.29802783242322,4.39795490297745,2.86964825021048,-.12176137061466,-3.07354613898791,-3.07100849759875,-4.09522222779324,-2.67211637191123,.113379941906936,2.20598407348367,2.94170303097044,1.91944964063812,-.923194488569861,-1.23108958885198,-.507379758856801,4.96464287844228,6.62040364808343,4.31978730278895,-.183291880031401,-3.80884940873901,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,6.45780747384475,8.61155438605811,5.61900531555836,-4.10479708970297,-5.47379022059323,1.32544341327556,7.79428043908965,10.3937551846321,-4.03553981216457,8.98680735137072},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,3.06556135691961,4.08795840577754,2.66737675689461,-.113178836410125,-2.85690259486619,-1.5947619510804,-2.12663188500128,-1.38761892782617,.0588777326765029,-.444911232504622,-.593293821521023,-.387121881618236,2.09440936915946,2.79291698671264,-2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640684,-.964148801340397,-1.28570259705192,-.838915879897341,-2.01722273728899,-2.68998779997356,3.46190525493913,5.18623489409548,6.91589893787678,4.51259681489188,.205483038001612,.274013798814207,-3.63947886955732,5.81833621896754,7.75881272356294,1.69243379052583,6.23568135536064},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,2.74931534527504,3.6662410134057,2.39220781955879,-.101503207887315,-2.56218200503226,.106757773657142,.142362617365695,.0928910470423957,-.00394143819036155,-2.26282036321899,-3.01749482284837,-1.96890348630228,2.15044627554921,2.86764279249033,-.119501749587716,3.45007465931297,4.60071095140733,3.00194577372104,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,3.7410698100085,4.98875605403389,3.25514367429685,3.5848146864382,4.78038820054912,-3.59724489879787,3.76406389114812,5.01941890462394,5.2355975909823,3.61790059797515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,2.36915365304765,3.15929138677771,2.06142518510908,-.0874678476499865,-2.20789618310877,1.68415540504794,2.24583899888368,1.46540110005888,-.0621780897499608,-2.32831086437971,-3.10482709694771,-2.02588745115913,-.381749900992996,-.50906751987416,2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273613,3.58714638367868,4.78349754217327,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,2.3938769796025,3.19226020392739,2.0829372085896,4.93737708601548,6.58404442855622,.508166439915615,2.07554270972349,2.76775809758181,5.63164748396115,1.71909544841975},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,1.94896143440088,2.59896063091726,1.69581157410878,-.0719545824309805,-1.81630452989169,2.83314651369184,3.77803076303363,2.46515019062261,-.104598208499711,-.637278791777431,-.849817990926997,-.55450289173293,-2.63766838605638,-3.51735861632833,.808065843843811,1.73374759666501,2.31197070864004,1.50855178059943,-.0640090061407119,4.12740556141282,5.50393885468462,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843811,1.33269841309056,1.77716739172311,1.15959472274786,4.37531622836173,5.83453034569908,4.52786743647191,.950542947793588,1.26755904782392,3.96849611187714,.647664476709578},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,1.51514018159912,2.02045541424857,1.31833919902212,-.0559381407795162,-1.41201150863478,3.36446686788906,4.4865520602846,2.92745754610214,-.124214263271281,1.91964817870762,2.55987109498181,1.67030580640284,-1.95173675811295,-2.60266155489591,-2.75500656898115,1.04781596872157,1.39727364720762,.91171555098101,-.0386847883193888,3.55510980002441,4.74077641992772,3.0933379389912,4.5904502888745,6.12141388513832,1.24891795372526,.626153181688298,.834981872710091,.544822382907867,2.84473124098147,3.79347911889773,5.48872263996487,.347192252628366,.462984531288131,1.97345041935925,.183906859961575},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,1.09494796295235,1.46012465838812,.95272558802182,-.0404248755605102,-1.02041985541771,3.24473228828999,4.32688473530227,2.82327524550585,-.11979372855449,4.05017972204568,5.4009573811209,3.52410341735765,1.63294619261043,2.17755097302669,-2.28560008620039,.547225517982272,.729731000653747,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367102,2.08234367785444,4.87179403320197,6.49658873611034,5.49730994822616,.236321146004053,.315136741011863,.205625482170815,1.3620925765208,1.81636481949506,3.77635721863399,.0946962283839075,.126278419466942,.676667877319621,.0362494356624604},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,.714786270724957,.953175031760129,.621942953572118,-.0263895153231816,-.666134033494215,2.59853682213183,3.46517626442722,2.26101386266496,-.0959365479342401,4.7549644652192,6.34079527029531,4.13734393811777,5.36731919542456,7.15737676442222,3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831,1.22672710731592,1.63585353791591,1.06738798948578,3.30495822058058,4.40719665133859,5.89679444984458,.0657431866024185,.0876692328212795,.0572038290823806,.451524561772771,.602112766253273,1.59193561065049,.0171974308138372,.0229329553998636,.145628931559543,.00429748986299419},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,.398540259080393,.531457639388288,.346774016236294,-.0147138868003716,-.371413443660288,1.67165845871656,2.22917418911978,1.45452737733324,-.0617167093759977,3.84755914106515,5.13076069911761,3.34780114662785,6.3113399865174,8.41623844737545,7.98777931164405,.07249762486996,.0966763475074086,.0630809359431452,-.00267657237098101,.434860760346183,.579891411113904,.378376861425819,1.40707187748205,1.8763451919045,3.23180916099,.0113956405123911,.0151962068293123,.00991546509750553,.0887334889355571,.118327043544488,.36808995603673,.00166206328996398,.00221637892968443,.0159028321410879,.000231576691000447},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,.166073783576779,.221461142188379,.144502522920425,-.00613135259583705,-.154769899538562,.764836159193308,1.01991708658492,.665491881382991,-.0282373296477252,2.02218309922343,2.69660249304008,1.75952250506473,4.02885967082133,5.37252686958436,6.68010475273478,.0125887190510937,.0167871896471047,.0109535751204636,-.000464768572193064,.0823390061849577,.109799933312678,.0716440239592633,.299004079114616,.398725093710571,.796298871799331,.000824565926335038,.00109956736063812,.000717463371565399,.00697737063218123,.00930439734224517,.0321260785532874,5.01144631634162e-005,6.68281652681524e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,.0319379183519753,.0425895510116904,.0277895142706455,-.00117913035023065,-.0297640501007841,.154660030302817,.206240781926363,.134571297792104,-.00570996311627577,.43925822557731,.585755477154869,.382203141612094,.961797763577864,1.28256746328293,1.79684767953686,.000465578153648566,.000620853379055138,.000405104384309317,-1.71888889439253e-005,.00319092434145458,.00425513126926133,.00277645639124963,.0123137463285649,.0164205106275597,.0353622090543065,5.86464816381893e-006,7.82057018924326e-006,5.10289124286309e-006,5.19109369737154e-005,6.92237820625419e-005,.000252650701908931,6.85464954370287e-008,9.14074747380003e-008,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,3.43216369764802,5.37468949190064,5.57281420860799,4.46269573688388,2.39877236199692,-4.00974905232829,-6.27917489029831,-6.5106412373467,-5.21370528571982,4.32546942586304,6.77358573667404,7.02327732701018,-4.4144312457315,-6.91289790468793,4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,7.27823584467215,11.3975501068473,11.8176927763525,-7.46210387346585,-11.6854832203471,7.31197439196321,9.14177937313724,14.3158164577429,-8.99870040642758,10.9691712594618},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,3.29802783242322,5.16463580891203,5.35501741292456,4.28828460541245,2.30502351007756,-3.07100849759875,-4.80912874954551,-4.9864054592391,-3.99310107128126,2.20598407348367,3.45452037324197,3.58186277642042,-.923194488569861,-1.44570135775136,-.507379758856801,4.96464287844228,7.77451667983119,8.06110512843118,6.45531290809181,-3.80884940873901,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,6.45780747384475,10.1127781284249,10.4855608369157,-4.10479708970297,-6.42801793274623,1.32544341327556,7.79428043908965,12.2056640870224,-4.03553981216457,8.98680735137072},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,3.06556135691961,4.80059864950596,4.97756091846909,3.98601838486204,2.14255044478903,-1.5947619510804,-2.49736057507455,-2.58941963373659,-2.07360079279396,-.444911232504622,-.696720767591157,-.72240366631615,2.09440936915946,3.27979695088037,-2.63186985254377,4.2894282422279,6.71714607327472,6.96475715335038,5.57736018045176,-.964148801340397,-1.50983487071459,-1.56549122161564,-2.01722273728899,-3.15892446006539,3.46190525493913,5.18623489409548,8.12152468439221,8.42090472917073,.205483038001612,.321781716436834,-3.63947886955732,5.81833621896754,9.11138083740394,1.69243379052583,6.23568135536064},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,2.74931534527504,4.30536466145163,4.46407134676942,3.5748172149004,1.92152305240544,.106757773657142,.16718022074295,.173342908623152,.138812569451129,-2.26282036321899,-3.54352469706742,-3.67414802589246,2.15044627554921,3.3675494587808,-.119501749587716,3.45007465931297,5.40273764744738,5.60189629155852,4.48598459696891,1.77035093065653,2.7723288820058,2.87452397106014,-3.42380239520456,-5.36159588761006,1.33170919606025,3.7410698100085,5.8584293670558,6.0743859656501,3.5848146864382,5.61373743608464,-3.59724489879787,3.76406389114812,5.8944375707901,5.2355975909823,3.61790059797515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,2.36915365304765,3.71004018615404,3.8468016980446,3.08050921776368,1.6558243734555,1.68415540504794,2.63734867013603,2.7345680458079,2.18983527963514,-2.32831086437971,-3.64608131714141,-3.78048514370594,-.381749900992996,-.597811573855878,2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635523,3.33115715545233,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,2.3938769796025,3.74875634795766,3.8869450362802,4.93737708601548,7.73181907803741,.508166439915615,2.07554270972349,3.25025219596949,5.63164748396115,1.71909544841975},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,1.94896143440088,3.05202882623936,3.16453436679059,2.53415123836086,1.36214797290772,2.83314651369184,4.43664234772504,4.6001882618506,3.68381930162858,-.637278791777431,-.997963945475626,-1.03475143356997,-2.63766838605638,-4.13052808940267,.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141282,6.46342227832349,6.70168045166517,1.98940127224618,3.11535668501805,-2.96832239843811,1.33269841309056,2.0869750946082,2.16390630168514,4.37531622836173,6.85164468857581,4.52786743647191,.950542947793588,1.48852841666699,3.96849611187714,.647664476709578},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,1.51514018159912,2.37267471198348,2.46013753301878,1.97007200846434,1.05894610873647,3.36446686788906,5.26867781509235,5.46289467143736,4.37467244553793,1.91964817870762,3.0061249410246,3.1169383486426,-1.95173675811295,-3.05637492003247,-2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591415,1.36243031247302,3.55510980002441,5.56722026499826,5.7724421542407,4.5904502888745,7.18853968487789,1.24891795372526,.626153181688298,.980541495928357,1.01668674800429,2.84473124098147,4.45478376427994,5.48872263996487,.347192252628366,.543695090562255,1.97345041935925,.183906859961575},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,1.09494796295235,1.7146633520688,1.77787020176478,1.42371402906152,.765269708188692,3.24473228828999,5.08117621439183,5.26848128513792,4.21898669004741,4.05017972204568,6.34248838569497,6.57628863172099,1.63294619261043,2.55715621899877,-2.28560008620039,.547225517982272,.856942588087487,.888531670338268,.711533948435091,2.39319484717323,3.74768814487269,3.88583745543936,4.87179403320197,7.62911752450547,5.49730994822616,.236321146004053,.370073484889949,.383715334269613,1.3620925765208,2.13300568008583,3.77635721863399,.0946962283839075,.148292117918319,.676667877319621,.0362494356624604},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,.714786270724957,1.11933887677121,1.16060055303996,.929406031924798,.499571029238743,2.59853682213183,4.06924896099705,4.21925182103647,3.37876634868698,4.7549644652192,7.44616510000515,7.72064967553201,5.36731919542456,8.40509854032022,3.25729414076559,.233201739437629,.365188565907959,.378650344801754,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064311,5.89679444984458,.0657431866024185,.102952319693431,.106747403902078,.451524561772771,.707077089737652,1.59193561065049,.0171974308138372,.0269307815280536,.145628931559543,.00429748986299419},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,.398540259080393,.624104888716883,.647110981340287,.518204861963161,.278543636855161,1.67165845871656,2.61777873930361,2.71427671758157,2.17358603453451,3.84755914106515,6.02518921139496,6.24729300320554,6.3113399865174,9.88341340931671,7.98777931164405,.07249762486996,.113529614838378,.11771460504762,.0942655625163294,.434860760346183,.68098195938567,.70608468526914,1.40707187748205,2.20344223207683,3.23180916099,.0113956405123911,.0178453112143479,.0185031347484812,.0887334889355571,.138954604977606,.36808995603673,.00166206328996398,.00260275292468705,.0159028321410879,.000231576691000447},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,.166073783576779,.260067729310804,.269654486884813,.215938641412748,.116070571566627,.764836159193308,1.1977158528687,1.24186670359493,.994483762912882,2.02218309922343,3.16669227065688,3.28342459824587,4.02885967082133,6.30910167728468,6.68010475273478,.0125887190510937,.019713644795161,.02044033985683,.0163685732441573,.0823390061849577,.12894099187304,.133694084564441,.299004079114616,.468233518150739,.796298871799331,.000824565926335038,.00129125129528224,.00133885010038651,.00697737063218123,.0109264021014796,.0321260785532874,5.01144631634162e-005,7.84780978891182e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,.0319379183519753,.0500140463222003,.0518576912013894,.0415275099413171,.0223217196472688,.154660030302817,.24219405407586,.251121942525088,.201097826062499,.43925822557731,.687868288629403,.713224862474736,.961797763577864,1.50615319946369,1.79684767953686,.000465578153648566,.000729084691473618,.00075596060662465,.000605371370823784,.00319092434145458,.00499691420436583,.00518111316265919,.0123137463285649,.0192830438371028,.0353622090543065,5.86464816381893e-006,9.18390427744748e-006,9.52244633459443e-006,5.19109369737154e-005,8.12913346215728e-005,.000252650701908931,6.85464954370287e-008,1.07342236937103e-007,7.43273979308028e-007,7.65346329155389e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,3.43216369764802,5.83506573884863,7.25392389703223,8.09901689119005,8.47552946493044,-4.00974905232829,-6.81702604882233,-8.4746582765388,-9.46196865399931,4.32546942586304,7.35378632026677,9.14193750783762,-4.4144312457315,-7.50503145913621,4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,7.27823584467215,12.3738225688191,15.3826488429769,-7.46210387346585,-12.6864189798666,7.31197439196321,9.14177937313724,15.5420569434015,-8.99870040642758,10.9691712594618},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,3.29802783242322,5.60701962552938,6.97042595115334,7.78249101017962,8.14428871473084,-3.07100849759875,-5.22106113152625,-6.4906175528327,-7.24678421547911,2.20598407348367,3.75042195543293,4.66237685613718,-.923194488569861,-1.5695348492102,-.507379758856801,4.96464287844228,8.44045334766827,10.4928391542811,11.7152706263275,-3.80884940873901,-6.47547397319737,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,6.45780747384475,10.9790017210618,13.6486625036235,-4.10479708970297,-6.97861844197595,1.32544341327556,7.79428043908965,13.25115663731,-4.03553981216457,8.98680735137072},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,3.06556135691961,5.21180037431139,6.47910494479514,7.2339303710034,7.57022621762745,-1.5947619510804,-2.71127534923244,-3.37055072311448,-3.76322492874004,-.444911232504622,-.756399320675517,-.940325842964601,2.09440936915946,3.56073236366527,-2.63186985254377,4.2894282422279,7.29251224169724,9.06576398575724,10.1219390639593,-.964148801340397,-1.63916180436132,-2.03774139204767,-2.01722273728899,-3.42950637731078,3.46190525493913,5.18623489409548,8.8171847889118,10.9611768422514,.205483038001612,.349344361529711,-3.63947886955732,5.81833621896754,9.8918284001439,1.69243379052583,6.23568135536064},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,2.74931534527504,4.67414645388234,5.81071476784009,6.48767173775814,6.78927500842464,.106757773657142,.181500267083405,.225633983149896,.251920680005669,-2.26282036321899,-3.84705006415361,-4.78250111498198,2.15044627554921,3.65600143048661,-.119501749587716,3.45007465931297,5.86551639682902,7.29177895707792,8.14128212331115,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429708,1.33170919606025,3.7410698100085,6.36024100270088,7.90680110023492,3.5848146864382,6.09458965574695,-3.59724489879787,3.76406389114812,6.39933353748714,5.2355975909823,3.61790059797515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,2.36915365304765,4.02782866109789,5.00723794478697,5.59058866190049,5.85048772793475,1.68415540504794,2.86325431258853,3.55948498424847,3.97417031416819,-2.32831086437971,-3.95839130925952,-4.9209161654707,-.381749900992996,-.649017927110068,2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830962,-1.52344991306454,-2.59003683420207,-2.69130383032704,2.3938769796025,4.06986110773884,5.05949102727518,4.93737708601548,8.39409842544749,.508166439915615,2.07554270972349,3.52865691309219,5.63164748396115,1.71909544841975},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,1.94896143440088,3.31345445440235,4.11915606854152,4.59904391748785,4.81284738096772,2.83314651369184,4.81666890668704,5.98789306701478,6.68549157438517,-.637278791777431,-1.08344588754871,-1.34689725343212,-2.63766838605638,-4.48433402043889,.808065843843811,1.73374759666501,2.94756663562365,3.66429874441148,4.09119503456581,4.12740556141282,7.01705539432372,8.72332687917667,1.98940127224618,3.38220674597771,-2.96832239843811,1.33269841309056,2.26573774926528,2.81667592802333,4.37531622836173,7.43853151653138,4.52786743647191,.950542947793588,1.61603031737647,3.96849611187714,.647664476709578},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,1.51514018159912,2.57590935107797,3.20226904625419,3.57533818439403,3.74155092353066,3.36446686788906,5.7199734894835,7.11084574520705,7.93926992086965,1.91964817870762,3.26361861024795,4.05720211125474,-1.95173675811295,-3.31817281869637,-2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898135,3.55510980002441,6.0440879938959,7.51377213010338,4.5904502888745,7.8042837062112,1.24891795372526,.626153181688298,1.06453109468611,1.32338312764,2.84473124098147,4.83636425064192,5.48872263996487,.347192252628366,.590266023758358,1.97345041935925,.183906859961575},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,1.09494796295235,1.86153514438243,2.31418717000874,2.58379343998139,2.70391057656362,3.24473228828999,5.51641118734948,6.85778510905627,7.65672734172536,4.05017972204568,6.88576274670089,8.56010903534823,1.63294619261043,2.776192869346,-2.28560008620039,.547225517982272,.930345156393644,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070527,4.87179403320197,8.28259983241364,5.49730994822616,.236321146004053,.40177262626832,.499467904136592,1.3620925765208,2.31571114636374,3.77635721863399,.0946962283839075,.160994278443515,.676667877319621,.0362494356624604},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,.714786270724957,1.21521735159799,1.51071034695561,1.68671036412374,1.76512329607374,2.59853682213183,4.41780594598826,5.49204234459095,6.13187350041211,4.7549644652192,8.08397637227704,10.049681020309,5.36731919542456,9.12504854432544,3.25729414076559,.233201739437629,.396469282990197,.492874996717866,.550295671800536,1.22672710731592,2.08557456680069,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984458,.0657431866024185,.111770838907699,.138949104550329,.451524561772771,.76764272750461,1.59193561065049,.0171974308138372,.0292375737894382,.145628931559543,.00429748986299419},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,.398540259080393,.677563431168936,.842320170000573,.940451730878478,.984172086870924,1.67165845871656,2.84200809304711,3.53307252095549,3.94468075935688,3.84755914106515,6.54128488547504,8.13186772631248,6.3113399865174,10.7299904588883,7.98777931164405,.07249762486996,.123254146474207,.153224700235958,.1710756072316,.434860760346183,.739312383715058,.919084035475439,1.40707187748205,2.3921810358159,3.23180916099,.0113956405123911,.0193738753136992,.0240848387004708,.0887334889355571,.150856948291044,.36808995603673,.00166206328996398,.00282569522209903,.0159028321410879,.000231576691000447},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,.166073783576779,.282344179950942,.350999163642368,.391891091702267,.410109589767539,.764836159193308,1.30030781284792,1.61649145672597,1.80481513164276,2.02218309922343,3.43793955015111,4.2739110377176,4.02885967082133,6.84951625494143,6.68010475273478,.0125887190510937,.0214022435166571,.0266064261611111,.0297061146457812,.0823390061849577,.139985605707145,.174024591276674,.299004079114616,.508340689012672,.796298871799331,.000824565926335038,.00140185515920491,.00174273111828065,.00697737063218123,.0118623177482726,.0321260785532874,5.01144631634162e-005,8.52002447644409e-005,.000520025886875943,2.90963997114409e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,.0319379183519753,.054298066631689,.0675012177634823,.075365210691831,.07886883956794,.154660030302817,.262939511058368,.326876043550518,.364957592022401,.43925822557731,.746788570738901,.928378137184539,.961797763577864,1.63516477461472,1.79684767953686,.000465578153648566,.000791535419924754,.000984005657525639,.00109864378994572,.00319092434145458,.00542493160168594,.00674406129057541,.0123137463285649,.0209347588552144,.0353622090543065,5.86464816381893e-006,9.97056393284754e-006,1.23950123652113e-005,5.19109369737154e-005,8.82544530675537e-005,.000252650701908931,6.85464954370287e-008,1.16536780408791e-007,7.43273979308028e-007,7.65346329155389e-010}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.40022577929605,5.78076767221695,-7.18642267926874,8.02365168049826,-8.3966606253625,4.1644090826311,-7.0799655598807,8.80153432008931,-9.82692624602176,-4.80864541143983,8.17524004082037,-10.1631364184159,5.37622900930936,-9.14019623375092,-5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375092,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,-6.67435503772231,11.3471570254948,-14.1063386773585,7.46215578440288,-12.6865072343198,-8.17438210188192,7.8090615264033,-13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-3.13195404884644,5.32467544557844,-6.61942678751097,7.39059991847739,-7.73417912496331,3.83584465679206,-6.52136894437417,8.10710900955867,-9.05159934712192,-4.42925189189755,7.53022822838542,-9.36128313844683,4.95205415939119,-8.41905110415162,-5.42470353822284,3.97724364248013,-6.7617631798718,8.40595765692319,-9.38526430991047,-4.59252537503604,7.80781158453306,-9.70636382862251,5.13459946240545,-8.72939873011817,-5.62467189980939,-3.67768288095471,6.25247606762144,-7.77283197754114,4.11177446033512,-6.99048075972417,-4.5042232468252,2.21014640326355,-3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.66702109783921,4.53423694314245,-5.63678477479456,6.29347864012496,-6.58605413075654,3.26642040979696,-5.55328344227955,6.90362324406996,-7.70790568809696,-3.77173740709444,6.41237937831348,-7.97161747704437,4.21693061735794,-7.16925809522303,-4.61941604516583,2.13132645129358,-3.62349557083132,4.50458697357607,-5.02937810056321,-2.46104380066281,4.18405228663708,-5.2014490017318,2.75153061636385,-4.67791266762517,-3.01415077313091,.104423971232213,-.177532539414032,.220701460427408,-.116749549066047,.198487413238653,.127892723210502,-2.94703586087065,5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-2.03452907455009,3.45892910228435,-4.30000442088448,4.80096137363443,-5.0241517123509,2.49177904847469,-4.23630567890486,5.26640836144105,-5.87995282040647,-2.87725860980962,4.89166444644702,-6.08112457054741,3.21687291987534,-5.46904711383883,-3.52390772571971,.0953780462518065,-.162153445774676,.201582776983911,-.225067472325601,-.110133081351697,.187238671144273,-.232767741029294,.123132528211026,-.209339198368165,-.134884926556002,2.8024998850757,-4.76456617608066,5.92312100550576,-3.1332901246654,5.32694692824228,3.43234736114769,-3.77550789188145,6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-1.2742056900953,2.16629351671545,-2.69305077477823,3.00679522191912,-3.14657715137113,1.56057688324205,-2.65315687476095,3.2983001248078,-3.68255702755718,-1.80199896789671,3.06360167061167,-3.80854893001879,2.01469609360342,-3.42521079645606,-2.20698899398908,-1.62700976129611,2.76610026595298,-3.43870690112598,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865961,-2.10046056933792,3.57102008849817,2.30093927043685,3.19783168117628,-5.43667485806875,6.75866004623775,-3.57528450949468,6.07838727908375,3.91652795089531,-.49602746920004,.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,-.433821252801768,.737545103324376,-.916887022287339,1.02370573309382,-1.07129645743706,.531320354197217,-.903304582796468,1.12295267819227,-1.25377834648449,-.613515899259544,1.04304628770802,-1.29667406213985,.68593162794344,-1.16616120174253,-.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.3691862575809,5.72799698473071,3.69075862820297,1.36899683098764,-2.32744915738298,2.89339311993396,-1.53058498738027,2.60216726588949,1.67667184761831,3.00982432351404,-5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,.433821252801768,-.737545103324376,.916887022287339,-1.02370573309382,1.07129645743706,-.531320354197217,.903304582796468,-1.12295267819227,1.25377834648449,.613515899259544,-1.04304628770802,1.29667406213985,-.68593162794344,1.16616120174253,.751400451068813,-2.60976045384023,4.43688738516868,-5.51576363976884,6.15835836236801,3.01349180095816,-5.12327625260686,6.36905524216974,-3.36918625758091,5.72799698473071,3.69075862820298,-1.36899683098765,2.32744915738298,-2.89339311993396,1.53058498738027,-2.60216726588949,-1.67667184761832,3.00982432351407,-5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,1.2742056900953,-2.16629351671545,2.69305077477823,-3.00679522191912,3.14657715137113,-1.56057688324205,2.65315687476095,-3.2983001248078,3.68255702755718,1.80199896789671,-3.06360167061167,3.80854893001879,-2.01469609360342,3.42521079645606,2.20698899398908,-1.62700976129612,2.76610026595298,-3.43870690112599,3.83932140376655,1.87870904723461,-3.19401746644684,3.97067670865962,-2.10046056933793,3.57102008849817,2.30093927043686,-3.1978316811763,5.43667485806878,-6.75866004623778,3.5752845094947,-6.07838727908378,-3.91652795089533,-.496027469200023,.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.03452907455009,-3.45892910228435,4.30000442088448,-4.80096137363443,5.0241517123509,-2.49177904847469,4.23630567890486,-5.26640836144105,5.87995282040647,2.87725860980962,-4.89166444644702,6.08112457054741,-3.21687291987534,5.46904711383883,3.52390772571971,.0953780462518019,-.162153445774668,.201582776983901,-.225067472325591,-.110133081351691,.187238671144264,-.232767741029283,.12313252821102,-.209339198368155,-.134884926555995,-2.8024998850757,4.76456617608065,-5.92312100550575,3.13329012466539,-5.32694692824227,-3.43234736114768,-3.77550789188148,6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,2.66702109783921,-4.53423694314245,5.63678477479456,-6.29347864012496,6.58605413075654,-3.26642040979696,5.55328344227955,-6.90362324406996,7.70790568809696,3.77173740709444,-6.41237937831348,7.97161747704437,-4.21693061735794,7.16925809522303,4.61941604516583,2.13132645129358,-3.62349557083131,4.50458697357606,-5.0293781005632,-2.46104380066281,4.18405228663707,-5.20144900173179,2.75153061636384,-4.67791266762516,-3.0141507731309,-.104423971232209,.177532539414025,-.2207014604274,.116749549066043,-.198487413238645,-.127892723210497,-2.94703586087065,5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.13195404884644,-5.32467544557844,6.61942678751097,-7.39059991847739,7.73417912496331,-3.83584465679206,6.52136894437417,-8.10710900955867,9.05159934712192,4.42925189189755,-7.53022822838542,9.36128313844683,-4.95205415939119,8.41905110415162,5.42470353822284,3.97724364248012,-6.76176317987178,8.40595765692318,-9.38526430991045,-4.59252537503603,7.80781158453304,-9.70636382862249,5.13459946240544,-8.72939873011815,-5.62467189980938,3.67768288095471,-6.25247606762144,7.77283197754114,-4.11177446033511,6.99048075972417,4.5042232468252,2.21014640326349,-3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-2.,3.40022577929605,-4.22702676016057,4.71948170681969,-4.93888416262808,2.449489743,-4.1644090826311,5.17702934684997,-5.78016101583118,-2.828427124,4.80864541143983,-5.97791857375092,3.162277660,-5.37622900930936,-3.464101616,3.40022577929605,-5.78076767221695,7.18642267926874,-8.02365168049826,8.3966606253625,-4.1644090826311,7.0799655598807,-8.80153432008931,9.82692624602176,4.80864541143983,-8.17524004082037,10.1631364184159,-5.37622900930936,9.14019623375092,5.88936380548032,5.17702934684997,-8.80153432008931,10.9417207971531,-12.2164476526894,-5.97791857375093,10.1631364184159,-12.6344108908165,6.68351614602547,-11.3627319492038,-7.32142511479571,6.67435503772228,-11.3471570254948,14.1063386773585,-7.46215578440284,12.6865072343197,8.17438210188188,7.8090615264033,-13.2762861543698,-8.55439830449838,8.51143945889045},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.40022577929605,5.32467544557844,-5.52095651740661,4.42116822694257,-2.37645064234971,4.1644090826311,-6.52136894437417,6.7617631798718,-5.41480311178232,-4.80864541143983,7.53022822838542,-7.80781158453306,5.37622900930936,-8.41905110415162,-5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375092,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,-6.67435503772231,10.4518866380257,-10.8371697481941,7.46215578440288,-11.6855645116818,-8.17438210188192,7.8090615264033,-12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-3.13195404884644,4.90456807960123,-5.08536292603976,4.0723459639997,-2.18895293851099,3.83584465679206,-6.00684460241421,6.22827216283404,-4.98758483419414,-4.42925189189755,6.93610669392436,-7.19178921991208,4.95205415939119,-7.75480303503604,-5.42470353822284,3.97724364248013,-6.22827216283404,6.45786210779303,-5.17143989055336,-4.59252537503604,7.19178921991208,-7.45689685002025,5.13459946240545,-8.04066478857437,-5.62467189980939,-3.67768288095471,5.75916689254577,-5.97146442400448,4.11177446033512,-6.43894433484766,-4.5042232468252,2.21014640326355,-3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.66702109783921,4.17649376078907,-4.33044993712881,3.46781352289888,-1.86400680793392,3.26642040979696,-5.11513931437816,5.30369635131817,-4.24718682732847,-3.77173740709444,5.90645411811629,-6.12418103253549,4.21693061735794,-6.60361645843633,-4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812298,-2.77127267544969,-2.46104380066281,3.85393804714674,-3.99600399916797,2.75153061636385,-4.3088337276685,-3.01415077313091,.104423971232213,-.163525539687803,.169553506762539,-.116749549066047,.182827111459215,.127892723210502,-2.94703586087065,4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-2.03452907455009,3.18602578468041,-3.30347079372947,2.6454111829756,-1.42195202316674,2.49177904847469,-3.9020687402541,4.04590891241332,-3.23995377923585,-2.87725860980962,4.50572087357673,-4.67181319960176,3.21687291987534,-5.03754908153942,-3.52390772571971,.0953780462518065,-.149359829020983,.154865612009825,-.12401599635522,-.110133081351697,.172465874990319,-.178823405501172,.123132528211026,-.192822710157554,-.134884926556002,2.8024998850757,-4.38865042934355,4.55042724011644,-3.1332901246654,4.90666034634693,3.43234736114769,-3.77550789188145,5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-1.2742056900953,1.99537683408524,-2.06893149627983,1.65679518870216,-.890554665266828,1.56057688324205,-2.4438275442558,2.53391323933003,-2.02915141041228,-1.80199896789671,2.82188898006397,-2.92591098178889,2.01469609360342,-3.15496779285464,-2.20698899398908,-1.62700976129611,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294204,3.05046536107998,-2.10046056933792,3.2892729926303,2.30093927043685,3.19783168117628,-5.00773093883056,5.19232863088549,-3.57528450949468,5.59881339795158,3.91652795089531,-.49602746920004,.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,-.433821252801768,.679354114255872,-.704396833771809,.564079229896515,-.303201864171254,.531320354197217,-.832035467367304,.862706409586763,-.690853143909356,-.613515899259544,.960751801781151,-.996167555656712,.68593162794344,-1.07415316937022,-.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.3933599570385,3.01349180095816,-4.71905892262039,4.89301542900031,-3.3691862575809,5.27606827091984,3.69075862820297,1.36899683098764,-2.14381758303679,2.2228441487308,-1.53058498738027,2.39686092429589,1.67667184761831,3.00982432351404,-4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,.433821252801768,-.679354114255872,.704396833771809,-.564079229896515,.303201864171254,-.531320354197217,.832035467367304,-.862706409586763,.690853143909356,.613515899259544,-.960751801781151,.996167555656712,-.68593162794344,1.07415316937022,.751400451068813,-2.60976045384023,4.08682490876582,-4.2374756641165,3.39335995703851,3.01349180095816,-4.71905892262039,4.89301542900031,-3.36918625758091,5.27606827091984,3.69075862820298,-1.36899683098765,2.1438175830368,-2.2228441487308,1.53058498738027,-2.3968609242959,-1.67667184761832,3.00982432351407,-4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,1.2742056900953,-1.99537683408524,2.06893149627983,-1.65679518870216,.890554665266828,-1.56057688324205,2.4438275442558,-2.53391323933003,2.02915141041228,1.80199896789671,-2.82188898006397,2.92591098178889,-2.01469609360342,3.15496779285464,2.20698899398908,-1.62700976129612,2.54785990395621,-2.64178049699053,2.11553124179228,1.87870904723461,-2.94201520294205,3.05046536107998,-2.10046056933793,3.28927299263031,2.30093927043686,-3.1978316811763,5.00773093883059,-5.19232863088552,3.5752845094947,-5.5988133979516,-3.91652795089533,-.496027469200023,.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.03452907455009,-3.18602578468041,3.30347079372947,-2.6454111829756,1.42195202316674,-2.49177904847469,3.9020687402541,-4.04590891241332,3.23995377923585,2.87725860980962,-4.50572087357673,4.67181319960176,-3.21687291987534,5.03754908153942,3.52390772571971,.0953780462518019,-.149359829020976,.154865612009818,-.124015996355214,-.110133081351691,.172465874990311,-.178823405501163,.12313252821102,-.192822710157544,-.134884926555995,-2.8024998850757,4.38865042934354,-4.55042724011643,3.13329012466539,-4.90666034634692,-3.43234736114768,-3.77550789188148,5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,2.66702109783921,-4.17649376078907,4.33044993712881,-3.46781352289888,1.86400680793392,-3.26642040979696,5.11513931437816,-5.30369635131817,4.24718682732847,3.77173740709444,-5.90645411811629,6.12418103253549,-4.21693061735794,6.60361645843633,4.61941604516583,2.13132645129358,-3.33760825329423,3.46064097812297,-2.77127267544968,-2.46104380066281,3.85393804714673,-3.99600399916796,2.75153061636384,-4.30883372766849,-3.0141507731309,-.104423971232209,.163525539687797,-.169553506762532,.116749549066043,-.182827111459208,-.127892723210497,-2.94703586087065,4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.13195404884644,-4.90456807960123,5.08536292603976,-4.0723459639997,2.18895293851099,-3.83584465679206,6.00684460241421,-6.22827216283404,4.98758483419414,4.42925189189755,-6.93610669392436,7.19178921991208,-4.95205415939119,7.75480303503604,5.42470353822284,3.97724364248012,-6.22827216283403,6.45786210779302,-5.17143989055335,-4.59252537503603,7.19178921991206,-7.45689685002024,5.13459946240544,-8.04066478857435,-5.62467189980938,3.67768288095471,-5.75916689254577,5.97146442400447,-4.11177446033511,6.43894433484766,4.5042232468252,2.21014640326349,-3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-2.,3.13195404884644,-3.24740583490635,2.60051450395615,-1.3978193194228,2.449489743,-3.83584465679206,3.97724364248013,-3.18496680157705,-2.828427124,4.42925189189755,-4.59252537503604,3.162277660,-4.95205415939119,-3.464101616,3.40022577929605,-5.32467544557844,5.52095651740661,-4.42116822694257,2.37645064234971,-4.1644090826311,6.52136894437417,-6.7617631798718,5.41480311178232,4.80864541143983,-7.53022822838542,7.80781158453306,-5.37622900930936,8.41905110415162,5.88936380548032,5.17702934684997,-8.10710900955867,8.40595765692319,-6.73146995394289,-5.97791857375093,9.36128313844683,-9.70636382862251,6.68351614602547,-10.46623272812,-7.32142511479571,6.67435503772228,-10.4518866380257,10.8371697481941,-7.46215578440284,11.6855645116817,8.17438210188188,7.8090615264033,-12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.40022577929605,4.53423694314245,-2.95857174458962,-.12553446251004,3.1687879383249,4.1644090826311,-5.55328344227955,3.62349557083132,.153747689149879,-4.80864541143983,6.41237937831348,-4.18405228663708,5.37622900930936,-7.16925809522303,-5.88936380548032,5.17702934684997,-6.90362324406996,4.50458697357607,.191133071440505,-5.97791857375092,7.97161747704437,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,-6.67435503772231,8.90032284672419,-5.80742559866954,7.46215578440288,-9.95086345672774,-8.17438210188192,7.8090615264033,-10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-3.13195404884644,4.17649376078907,-2.72514572729006,-.115630018018827,2.91877623944935,3.83584465679206,-5.11513931437816,3.33760825329423,.141617271554667,-4.42925189189755,5.90645411811629,-3.85393804714674,4.95205415939119,-6.60361645843633,-5.42470353822284,3.97724364248013,-5.30369635131817,3.46064097812298,.146837644201687,-4.59252537503604,6.12418103253549,-3.99600399916797,5.13459946240545,-6.84704254830277,-5.62467189980939,-3.67768288095471,4.90422891550851,-3.19999004936569,4.11177446033512,-5.48309461793542,-4.5042232468252,2.21014640326355,-2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.66702109783921,3.55650076638925,-2.32060274065832,-.0984649496097572,2.4854891512059,3.26642040979696,-4.35580607412106,2.84214630515942,.120594442052505,-3.77173740709444,5.02965161709108,-3.28182786886397,4.21693061735794,-5.62332146066281,-4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898143,.0786873984258951,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636385,-3.66919510297595,-3.01415077313091,.104423971232213,-.139250467146397,.0908603812984511,-.116749549066047,.155686755269708,.127892723210502,-2.94703586087065,3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-2.03452907455009,2.71306598164557,-1.77026486598667,-.0751136925641364,1.89604797153805,2.49177904847469,-3.32281364706153,2.16812281562257,.0919951097438821,-2.87725860980962,3.83685470584401,-2.50353258258141,3.21687291987534,-4.28973397193188,-3.52390772571971,.0953780462518065,-.127187630774895,.0829894280920859,.00352130492348755,-.110133081351697,.146863625737388,-.0958279372639542,.123132528211026,-.164198525301527,-.134884926556002,2.8024998850757,-3.73716315870324,2.43848424018079,-3.1332901246654,4.17827543429579,3.43234736114769,-3.77550789188145,5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-1.2742056900953,1.69916672838959,-1.10869959708727,-.0470429720894781,1.18747632769106,1.56057688324205,-2.08104573641859,1.35787414544697,.0576156388045318,-1.80199896789671,2.40298463138773,-1.56793800686758,2.01469609360342,-2.68661849290085,-2.20698899398908,-1.62700976129611,2.16963467924844,-1.41567808207906,-.0600683040611749,1.87870904723461,-2.50527833232423,1.63468424297244,-2.10046056933792,2.8009863270915,2.30093927043685,3.19783168117628,-4.26434227893045,2.78246653954405,-3.57528450949468,4.76767960906116,3.91652795089531,-.49602746920004,.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,-.433821252801768,.578505216668687,-.377472375086663,-.016016441651465,.404293021256913,.531320354197217,-.70852129725097,.462307355479536,.0196160547715705,-.613515899259544,.818129923100605,-.533826552292297,.68593162794344,-.914697061432424,-.751400451068813,-2.60976045384023,3.48014309433045,-2.27077966086395,-.0963509182288022,3.01349180095816,-4.01852310483628,2.62207049601641,-3.3691862575809,4.49284541622379,3.69075862820297,1.36899683098764,-1.82557171487995,1.19117835294064,-1.53058498738027,2.04105122680137,1.67667184761831,3.00982432351404,-4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,.433821252801768,-.578505216668687,.377472375086663,.016016441651465,-.404293021256913,-.531320354197217,.70852129725097,-.462307355479536,-.0196160547715705,.613515899259544,-.818129923100605,.533826552292297,-.68593162794344,.914697061432424,.751400451068813,-2.60976045384023,3.48014309433046,-2.27077966086395,-.0963509182288023,3.01349180095816,-4.01852310483628,2.62207049601641,-3.36918625758091,4.49284541622379,3.69075862820298,-1.36899683098765,1.82557171487995,-1.19117835294064,1.53058498738027,-2.04105122680137,-1.67667184761832,3.00982432351407,-4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,1.2742056900953,-1.69916672838959,1.10869959708727,.0470429720894781,-1.18747632769106,-1.56057688324205,2.08104573641859,-1.35787414544697,-.0576156388045318,1.80199896789671,-2.40298463138773,1.56793800686758,-2.01469609360342,2.68661849290085,2.20698899398908,-1.62700976129612,2.16963467924844,-1.41567808207907,-.060068304061175,1.87870904723461,-2.50527833232424,1.63468424297245,-2.10046056933793,2.8009863270915,2.30093927043686,-3.1978316811763,4.26434227893047,-2.78246653954406,3.5752845094947,-4.76767960906118,-3.91652795089533,-.496027469200023,.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.03452907455009,-2.71306598164557,1.77026486598667,.0751136925641364,-1.89604797153805,-2.49177904847469,3.32281364706153,-2.16812281562257,-.0919951097438821,2.87725860980962,-3.83685470584401,2.50353258258141,-3.21687291987534,4.28973397193188,3.52390772571971,.0953780462518019,-.127187630774889,.0829894280920819,.00352130492348739,-.110133081351691,.146863625737381,-.0958279372639496,.12313252821102,-.164198525301519,-.134884926555995,-2.8024998850757,3.73716315870323,-2.43848424018078,3.13329012466539,-4.17827543429578,-3.43234736114768,-3.77550789188148,5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,2.66702109783921,-3.55650076638925,2.32060274065832,.0984649496097572,-2.4854891512059,-3.26642040979696,4.35580607412106,-2.84214630515942,-.120594442052505,3.77173740709444,-5.02965161709108,3.28182786886397,-4.21693061735794,5.62332146066281,4.61941604516583,2.13132645129358,-2.84214630515942,1.85448926898142,.0786873984258949,-2.46104380066281,3.28182786886397,-2.14137975655691,2.75153061636384,-3.66919510297594,-3.0141507731309,-.104423971232209,.139250467146391,-.0908603812984476,.116749549066043,-.155686755269702,-.127892723210497,-2.94703586087065,3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.13195404884644,-4.17649376078907,2.72514572729006,.115630018018827,-2.91877623944935,-3.83584465679206,5.11513931437816,-3.33760825329423,-.141617271554667,4.42925189189755,-5.90645411811629,3.85393804714674,-4.95205415939119,6.60361645843633,5.42470353822284,3.97724364248012,-5.30369635131817,3.46064097812297,.146837644201687,-4.59252537503603,6.12418103253548,-3.99600399916796,5.13459946240544,-6.84704254830276,-5.62467189980938,3.67768288095471,-4.90422891550851,3.19999004936568,-4.11177446033511,5.48309461793542,4.5042232468252,2.21014640326349,-2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-2.,2.66702109783921,-1.74022075994949,-.0738388981704632,1.86386913340851,2.449489743,-3.26642040979696,2.13132645129358,.090433811839997,-2.828427124,3.77173740709444,-2.46104380066281,3.162277660,-4.21693061735794,-3.464101616,3.40022577929605,-4.53423694314245,2.95857174458962,.12553446251004,-3.1687879383249,-4.1644090826311,5.55328344227955,-3.62349557083132,-.153747689149879,4.80864541143983,-6.41237937831348,4.18405228663708,-5.37622900930936,7.16925809522303,5.88936380548032,5.17702934684997,-6.90362324406997,4.50458697357607,.191133071440505,-5.97791857375093,7.97161747704438,-5.2014490017318,6.68351614602547,-8.91253928552127,-7.32142511479571,6.67435503772228,-8.90032284672415,5.80742559866952,-7.46215578440284,9.95086345672769,8.17438210188188,7.8090615264033,-10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.40022577929605,3.45892910228435,-.132397734060631,-3.36905705276318,4.05959901190681,4.1644090826311,-4.23630567890486,.162153445774676,4.1262353470069,-4.80864541143983,4.89166444644702,-.187238671144273,5.37622900930936,-5.46904711383883,-5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375092,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,-6.67435503772231,6.78958468653045,-.259885531387107,7.46215578440288,-7.59098645151066,-8.17438210188192,7.8090615264033,-7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-3.13195404884644,3.18602578468041,-.121951789723545,-3.10324447907128,3.73930391312628,3.83584465679206,-3.9020687402541,.149359829020983,3.80068276060982,-4.42925189189755,4.50572087357673,-.172465874990319,4.95205415939119,-5.03754908153942,-5.42470353822284,3.97724364248013,-4.04590891241332,.154865612009825,3.94078558921029,-4.59252537503604,4.67181319960176,-.178823405501172,5.13459946240545,-5.22324594675678,-5.62467189980939,-3.67768288095471,3.74117637274605,-.143201337715524,4.11177446033512,-4.18276234400124,-4.5042232468252,2.21014640326355,-2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.66702109783921,2.71306598164557,-.103848265663968,-2.64257341211137,3.18421096606244,3.26642040979696,-3.32281364706153,.127187630774895,3.23647823392353,-3.77173740709444,3.83685470584401,-.146863625737388,4.21693061735794,-4.28973397193188,-4.61941604516583,2.13132645129358,-2.16812281562257,.0829894280920859,2.11178929936587,-2.46104380066281,2.50353258258141,-.0958279372639542,2.75153061636385,-2.79903451954298,-3.01415077313091,.104423971232213,-.106226802736416,.0040660526897138,-.116749549066047,.118765176014247,.127892723210502,-2.94703586087065,2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-2.03452907455009,2.0696542765645,-.0792203391289715,-2.01587923055033,2.42906581999068,2.49177904847469,-2.53479846100042,.0970247040581402,2.46893774908671,-2.87725860980962,2.92693314656882,-.112034478016931,3.21687291987534,-3.2724107413517,-3.52390772571971,.0953780462518065,-.0970247040581402,.00371382314810282,.0945037478171115,-.110133081351697,.112034478016931,-.00428835358704871,.123132528211026,-.12525835434705,-.134884926556002,2.8024998850757,-2.85088374774363,.109123528450137,-3.1332901246654,3.18738492899079,3.43234736114769,-3.77550789188145,3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-1.2742056900953,1.29620426108275,-.0496149247273535,-1.2625254749334,1.52130020070251,1.56057688324205,-1.58751952117754,.0607656246043247,1.54627160050444,-1.80199896789671,1.83310964514541,-.0701660994486395,2.01469609360342,-2.04947888880939,-2.20698899398908,-1.62700976129611,1.65509933148238,-.0633523829764578,-1.61209551065421,1.87870904723461,-1.91114408921758,.0731530307047499,-2.10046056933792,2.13672404935301,2.30093927043685,3.19783168117628,-3.25304076422429,.124516928010565,-3.57528450949468,3.63701014237694,3.91652795089531,-.49602746920004,.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,-.433821252801768,.441310975771777,-.0168920991094272,-.429844559231871,.517948055079696,.531320354197217,-.540493354313145,.0206885117513158,.526449919441546,-.613515899259544,.624107966995901,-.0238890356585565,.68593162794344,-.697773919897946,-.751400451068813,-2.60976045384023,2.65481675969736,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.3691862575809,3.42735369966595,3.69075862820297,1.36899683098764,-1.39263192728723,.0533058950082319,-1.53058498738027,1.5570098290328,1.67667184761831,3.00982432351404,-3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,.433821252801768,-.441310975771777,.0168920991094272,.429844559231871,-.517948055079696,-.531320354197217,.540493354313145,-.0206885117513158,-.526449919441546,.613515899259544,-.624107966995901,.0238890356585565,-.68593162794344,.697773919897946,.751400451068813,-2.60976045384023,2.65481675969737,-.101618655082187,-2.58583765850746,3.01349180095816,-3.06551834185512,.117339115691434,-3.36918625758091,3.42735369966595,3.69075862820298,-1.36899683098765,1.39263192728723,-.0533058950082319,1.53058498738027,-1.5570098290328,-1.67667184761832,3.00982432351407,-3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,1.2742056900953,-1.29620426108275,.0496149247273535,1.2625254749334,-1.52130020070251,-1.56057688324205,1.58751952117754,-.0607656246043247,-1.54627160050444,1.80199896789671,-1.83310964514541,.0701660994486395,-2.01469609360342,2.04947888880939,2.20698899398908,-1.62700976129612,1.65509933148239,-.0633523829764579,-1.61209551065421,1.87870904723461,-1.91114408921759,.07315303070475,-2.10046056933793,2.13672404935301,2.30093927043686,-3.1978316811763,3.2530407642243,-.124516928010566,3.5752845094947,-3.63701014237696,-3.91652795089533,-.496027469200023,.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.03452907455009,-2.0696542765645,.0792203391289715,2.01587923055033,-2.42906581999068,-2.49177904847469,2.53479846100042,-.0970247040581402,-2.46893774908671,2.87725860980962,-2.92693314656882,.112034478016931,-3.21687291987534,3.2724107413517,3.52390772571971,.0953780462518019,-.0970247040581356,.00371382314810264,.0945037478171069,-.110133081351691,.112034478016925,-.00428835358704851,.12313252821102,-.125258354347044,-.134884926555995,-2.8024998850757,2.85088374774362,-.109123528450137,3.13329012466539,-3.18738492899078,-3.43234736114768,-3.77550789188148,3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,2.66702109783921,-2.71306598164557,.103848265663968,2.64257341211137,-3.18421096606244,-3.26642040979696,3.32281364706153,-.127187630774895,-3.23647823392353,3.77173740709444,-3.83685470584401,.146863625737388,-4.21693061735794,4.28973397193188,4.61941604516583,2.13132645129358,-2.16812281562256,.0829894280920858,2.11178929936586,-2.46104380066281,2.5035325825814,-.095827937263954,2.75153061636384,-2.79903451954297,-3.0141507731309,-.104423971232209,.106226802736412,-.00406605268971365,.116749549066043,-.118765176014243,-.127892723210497,-2.94703586087065,2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.13195404884644,-3.18602578468041,.121951789723545,3.10324447907128,-3.73930391312628,-3.83584465679206,3.9020687402541,-.149359829020983,-3.80068276060982,4.42925189189755,-4.50572087357673,.172465874990319,-4.95205415939119,5.03754908153942,5.42470353822284,3.97724364248012,-4.04590891241331,.154865612009825,3.94078558921028,-4.59252537503603,4.67181319960176,-.178823405501172,5.13459946240544,-5.22324594675677,-5.62467189980938,3.67768288095471,-3.74117637274605,.143201337715524,-4.11177446033511,4.18276234400124,4.5042232468252,2.21014640326349,-2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-2.,2.03452907455009,-.0778758486434948,-1.98166667284345,2.38784085258437,2.449489743,-2.49177904847469,.0953780462518065,2.42703609427915,-2.828427124,2.87725860980962,-.110133081351697,3.162277660,-3.21687291987534,-3.464101616,3.40022577929605,-3.45892910228435,.132397734060631,3.36905705276318,-4.05959901190681,-4.1644090826311,4.23630567890486,-.162153445774676,-4.1262353470069,4.80864541143983,-4.89166444644702,.187238671144273,-5.37622900930936,5.46904711383883,5.88936380548032,5.17702934684997,-5.26640836144105,.201582776983911,5.12957326201498,-5.97791857375093,6.08112457054741,-.232767741029294,6.68351614602547,-6.79890396036006,-7.32142511479571,6.67435503772228,-6.78958468653042,.259885531387105,-7.46215578440284,7.59098645151062,8.17438210188188,7.8090615264033,-7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.40022577929605,2.16629351671545,2.25851140965318,-3.84430965952581,.533351454031698,4.1644090826311,-2.65315687476095,-2.76610026595298,4.70829853978447,-4.80864541143983,3.06360167061167,3.19401746644684,5.37622900930936,-3.42521079645606,-5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375092,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,-6.67435503772231,4.25225058180805,4.43326649059391,7.46215578440288,-4.75416068069224,-8.17438210188192,7.8090615264033,-4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-3.13195404884644,1.99537683408524,2.08031890026244,-3.54100050546174,.491270978557942,3.83584465679206,-2.4438275442558,-2.54785990395621,4.33682220887952,-4.42925189189755,2.82188898006397,2.94201520294204,4.95205415939119,-3.15496779285464,-5.42470353822284,3.97724364248013,-2.53391323933003,-2.64178049699053,4.49668850050967,-4.59252537503604,2.92591098178889,3.05046536107998,5.13459946240545,-3.27126792601698,-5.62467189980939,-3.67768288095471,2.34306222576705,2.44280506311383,4.11177446033512,-2.61962320701725,-4.5042232468252,2.21014640326355,-1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.66702109783921,1.69916672838959,1.77149929746803,-3.01534534295104,.418342684514395,3.26642040979696,-2.08104573641859,-2.16963467924844,3.69302874444133,-3.77173740709444,2.40298463138773,2.50527833232423,4.21693061735794,-2.68661849290085,-4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636385,-1.75300798410243,-3.01415077313091,.104423971232213,-.0665288091384509,-.0693609030179036,-.116749549066047,.0743814698720828,.127892723210502,-2.94703586087065,1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-2.03452907455009,1.29620426108275,1.35138294525071,-2.30024718402619,.319131466736219,2.49177904847469,-1.58751952117754,-1.65509933148238,2.81721594171208,-2.87725860980962,1.83310964514541,1.91114408921758,3.21687291987534,-2.04947888880939,-3.52390772571971,.0953780462518065,-.0607656246043247,-.0633523829764578,.107834822891095,-.110133081351697,.0701660994486395,.0731530307047499,.123132528211026,-.0784480840492681,-.134884926556002,2.8024998850757,-1.78548064936278,-1.86148755350589,-3.1332901246654,1.99622805289337,3.43234736114769,-3.77550789188145,2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-1.2742056900953,.811800069931489,.846357940948566,-1.44062234704609,.199868920965739,1.56057688324205,-.994247972331933,-1.03657254753862,1.76439483124641,-1.80199896789671,1.14805866852966,1.19693087876444,2.01469609360342,-1.28356861276539,-2.20698899398908,-1.62700976129611,1.03657254753862,1.08069885633786,-1.83950412433748,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933792,1.338209404774,2.30093927043685,3.19783168117628,-2.03734766130249,-2.1240764020783,-3.57528450949468,2.27782393297861,3.91652795089531,-.49602746920004,.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,-.433821252801768,.27638875426455,.2881544676147,-.490480144821059,.0680481859118126,.531320354197217,-.338505709325782,-.352915706379786,.6007130419195,-.613515899259544,.390872724665213,.407511956177088,.68593162794344,-.437008991543009,-.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.3691862575809,2.14651815042224,3.69075862820297,1.36899683098764,-.872191775558762,-.909320487484575,-1.53058498738027,.975140050100603,1.67667184761831,3.00982432351404,-1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,.433821252801768,-.27638875426455,-.2881544676147,.490480144821059,-.0680481859118126,-.531320354197217,.338505709325782,.352915706379786,-.6007130419195,.613515899259544,-.390872724665213,-.407511956177088,-.68593162794344,.437008991543009,.751400451068813,-2.60976045384023,1.66268580954775,1.73346540682961,-2.95060621796617,3.01349180095816,-1.91990419952445,-2.00163343782271,-3.36918625758091,2.14651815042224,3.69075862820298,-1.36899683098765,.872191775558764,.909320487484576,1.53058498738027,-.975140050100604,-1.67667184761832,3.00982432351407,-1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,1.2742056900953,-.811800069931489,-.846357940948566,1.44062234704609,-.199868920965739,-1.56057688324205,.994247972331933,1.03657254753862,-1.76439483124641,1.80199896789671,-1.14805866852966,-1.19693087876444,-2.01469609360342,1.28356861276539,2.20698899398908,-1.62700976129612,1.03657254753862,1.08069885633786,-1.83950412433749,1.87870904723461,-1.19693087876444,-1.24788355079949,-2.10046056933793,1.338209404774,2.30093927043686,-3.1978316811763,2.0373476613025,2.12407640207831,3.5752845094947,-2.27782393297862,-3.91652795089533,-.496027469200023,.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.03452907455009,-1.29620426108275,-1.35138294525071,2.30024718402619,-.319131466736219,-2.49177904847469,1.58751952117754,1.65509933148238,-2.81721594171208,2.87725860980962,-1.83310964514541,-1.91114408921758,-3.21687291987534,2.04947888880939,3.52390772571971,.0953780462518019,-.0607656246043218,-.0633523829764547,.10783482289109,-.110133081351691,.0701660994486362,.0731530307047464,.12313252821102,-.0784480840492644,-.134884926555995,-2.8024998850757,1.78548064936277,1.86148755350588,3.13329012466539,-1.99622805289336,-3.43234736114768,-3.77550789188148,2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,2.66702109783921,-1.69916672838959,-1.77149929746803,3.01534534295104,-.418342684514395,-3.26642040979696,2.08104573641859,2.16963467924844,-3.69302874444133,3.77173740709444,-2.40298463138773,-2.50527833232423,-4.21693061735794,2.68661849290085,4.61941604516583,2.13132645129358,-1.35787414544697,-1.41567808207906,2.4096867091571,-2.46104380066281,1.56793800686758,1.63468424297244,2.75153061636384,-1.75300798410242,-3.0141507731309,-.104423971232209,.0665288091384484,.0693609030179009,.116749549066043,-.0743814698720799,-.127892723210497,-2.94703586087065,1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.13195404884644,-1.99537683408524,-2.08031890026244,3.54100050546174,-.491270978557942,-3.83584465679206,2.4438275442558,2.54785990395621,-4.33682220887952,4.42925189189755,-2.82188898006397,-2.94201520294204,-4.95205415939119,3.15496779285464,5.42470353822284,3.97724364248012,-2.53391323933002,-2.64178049699052,4.49668850050967,-4.59252537503603,2.92591098178889,3.05046536107998,5.13459946240544,-3.27126792601697,-5.62467189980938,3.67768288095471,-2.34306222576705,-2.44280506311383,-4.11177446033511,2.61962320701725,4.5042232468252,2.21014640326349,-1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-2.,1.2742056900953,1.32844790694992,-2.26120846666111,.313715316952933,2.449489743,-1.56057688324205,-1.62700976129611,2.76940347258375,-2.828427124,1.80199896789671,1.87870904723461,3.162277660,-2.01469609360342,-3.464101616,3.40022577929605,-2.16629351671545,-2.25851140965318,3.84430965952581,-.533351454031698,-4.1644090826311,2.65315687476095,2.76610026595298,-4.70829853978447,4.80864541143983,-3.06360167061167,-3.19401746644684,-5.37622900930936,3.42521079645606,5.88936380548032,5.17702934684997,-3.2983001248078,-3.43870690112598,5.8531712973623,-5.97791857375093,3.80854893001879,3.97067670865961,6.68351614602547,-4.25808715199512,-7.32142511479571,6.67435503772228,-4.25225058180803,-4.43326649059389,-7.46215578440284,4.75416068069222,8.17438210188188,7.8090615264033,-4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.40022577929605,.737545103324376,3.62270337995047,-1.64575508217185,-3.2363009692086,4.1644090826311,-.903304582796468,-4.43688738516868,2.01563009655897,-4.80864541143983,1.04304628770802,5.12327625260686,5.37622900930936,-1.16616120174253,-5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375092,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394474,-7.32142511479571,-6.67435503772231,1.44773853151531,7.11105971439921,7.46215578440288,-1.61862088558446,-8.17438210188192,7.8090615264033,-1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-3.13195404884644,.679354114255872,3.33687856485652,-1.51590795070228,-2.98096261298772,3.83584465679206,-.832035467367304,-4.08682490876582,1.85660048821863,-4.42925189189755,.960751801781151,4.71905892262039,4.95205415939119,-1.07415316937022,-5.42470353822284,3.97724364248013,-.862706409586763,-4.2374756641165,1.92503950203906,-4.59252537503604,.996167555656712,4.89301542900031,5.13459946240545,-1.1137491858232,-5.62467189980939,-3.67768288095471,.797728497114731,3.91831456811703,4.11177446033512,-.891887573859297,-4.5042232468252,2.21014640326355,-.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.66702109783921,.578505216668687,2.84152493765918,-1.29087413922765,-2.53844406932994,3.26642040979696,-.70852129725097,-3.48014309433045,1.58099148171138,-3.77173740709444,.818129923100605,4.01852310483628,4.21693061735794,-.914697061432424,-4.61941604516583,2.13132645129358,-.462307355479536,-2.27077966086395,1.03159071439798,-2.46104380066281,.533826552292297,2.62207049601641,2.75153061636385,-.596836229618421,-3.01415077313091,.104423971232213,-.0226506690028153,-.111256457118347,-.116749549066047,.0253242178213241,.127892723210502,-2.94703586087065,.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-2.03452907455009,.441310975771777,2.16764880728185,-.984739479553152,-1.93644447258226,2.49177904847469,-.540493354313145,-2.65481675969736,1.20605462730079,-2.87725860980962,.624107966995901,3.06551834185512,3.21687291987534,-.697773919897946,-3.52390772571971,.0953780462518065,-.0206885117513158,-.101618655082187,.0461642592626141,-.110133081351697,.0238890356585565,.117339115691434,.123132528211026,-.0267087538273405,-.134884926556002,2.8024998850757,-.607892005333882,-2.98586813553316,-3.1332901246654,.6796439236741,3.43234736114769,-3.77550789188145,.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-1.2742056900953,.27638875426455,1.35757727865237,-.616732719037493,-1.21277626178114,1.56057688324205,-.338505709325782,-1.66268580954775,.755340234698916,-1.80199896789671,.390872724665213,1.91990419952445,2.01469609360342,-.437008991543009,-2.20698899398908,-1.62700976129611,.352915706379786,1.73346540682961,-.78749464258244,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933792,.455612217872568,2.30093927043685,3.19783168117628,-.693643672830323,-3.40706658739604,-3.57528450949468,.775517202558345,3.91652795089531,-.49602746920004,.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,-.433821252801768,.0941004396444027,.462206283002972,-.209975330432464,-.412906739739006,.531320354197217,-.115249030860378,-.566084774633034,.257166209078973,-.613515899259544,.133078117935277,.653658394065701,.68593162794344,-.148785859041837,-.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.3691862575809,.730812301668709,3.69075862820297,1.36899683098764,-.296949960039812,-1.45857062726747,-1.53058498738027,.332000148390628,1.67667184761831,3.00982432351404,-.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,.433821252801768,-.0941004396444027,-.462206283002972,.209975330432464,.412906739739006,-.531320354197217,.115249030860378,.566084774633034,-.257166209078973,.613515899259544,-.133078117935277,-.653658394065701,-.68593162794344,.148785859041837,.751400451068813,-2.60976045384023,.566084774633034,2.7805177168944,-1.26315921681103,3.01349180095816,-.653658394065701,-3.21066530353978,-3.36918625758091,.730812301668709,3.69075862820298,-1.36899683098765,.296949960039813,1.45857062726747,1.53058498738027,-.332000148390629,-1.67667184761832,3.00982432351407,-.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,1.2742056900953,-.27638875426455,-1.35757727865237,.616732719037493,1.21277626178114,-1.56057688324205,.338505709325782,1.66268580954775,-.755340234698916,1.80199896789671,-.390872724665213,-1.91990419952445,-2.01469609360342,.437008991543009,2.20698899398908,-1.62700976129612,.352915706379786,1.73346540682961,-.787494642582441,1.87870904723461,-.407511956177088,-2.00163343782271,-2.10046056933793,.455612217872569,2.30093927043686,-3.1978316811763,.693643672830326,3.40706658739606,3.5752845094947,-.775517202558348,-3.91652795089533,-.496027469200023,.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.03452907455009,-.441310975771777,-2.16764880728185,.984739479553152,1.93644447258226,-2.49177904847469,.540493354313145,2.65481675969736,-1.20605462730079,2.87725860980962,-.624107966995901,-3.06551834185512,-3.21687291987534,.697773919897946,3.52390772571971,.0953780462518019,-.0206885117513148,-.101618655082182,.0461642592626119,-.110133081351691,.0238890356585554,.117339115691428,.12313252821102,-.0267087538273392,-.134884926555995,-2.8024998850757,.607892005333881,2.98586813553315,3.13329012466539,-.679643923674099,-3.43234736114768,-3.77550789188148,.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,2.66702109783921,-.578505216668687,-2.84152493765918,1.29087413922765,2.53844406932994,-3.26642040979696,.70852129725097,3.48014309433045,-1.58099148171138,3.77173740709444,-.818129923100605,-4.01852310483628,-4.21693061735794,.914697061432424,4.61941604516583,2.13132645129358,-.462307355479535,-2.27077966086395,1.03159071439797,-2.46104380066281,.533826552292296,2.62207049601641,2.75153061636384,-.59683622961842,-3.0141507731309,-.104423971232209,.0226506690028144,.111256457118342,.116749549066043,-.0253242178213231,-.127892723210497,-2.94703586087065,.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.13195404884644,-.679354114255872,-3.33687856485652,1.51590795070228,2.98096261298772,-3.83584465679206,.832035467367304,4.08682490876582,-1.85660048821863,4.42925189189755,-.960751801781151,-4.71905892262039,-4.95205415939119,1.07415316937022,5.42470353822284,3.97724364248012,-.862706409586761,-4.23747566411649,1.92503950203906,-4.59252537503603,.996167555656711,4.8930154290003,5.13459946240544,-1.1137491858232,-5.62467189980938,3.67768288095471,-.797728497114731,-3.91831456811703,-4.11177446033511,.891887573859297,4.5042232468252,2.21014640326349,-.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-2.,.433821252801768,2.13086048714476,-.96802694253214,-1.90358004395732,2.449489743,-.531320354197217,-2.60976045384023,1.18558603318945,-2.828427124,.613515899259544,3.01349180095816,3.162277660,-.68593162794344,-3.464101616,3.40022577929605,-.737545103324376,-3.62270337995047,1.64575508217185,3.2363009692086,-4.1644090826311,.903304582796468,4.43688738516868,-2.01563009655897,4.80864541143983,-1.04304628770802,-5.12327625260686,-5.37622900930936,1.16616120174253,5.88936380548032,5.17702934684997,-1.12295267819227,-5.51576363976884,2.50575194575887,-5.97791857375093,1.29667406213985,6.36905524216974,6.68351614602547,-1.44972567394475,-7.32142511479571,6.67435503772228,-1.4477385315153,-7.11105971439917,-7.46215578440284,1.61862088558445,8.17438210188188,7.8090615264033,-1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.40022577929605,-.737545103324376,3.62270337995047,1.64575508217186,-3.23630096920864,4.1644090826311,.903304582796468,-4.43688738516868,-2.01563009655898,-4.80864541143983,-1.04304628770802,5.12327625260686,5.37622900930936,1.16616120174253,-5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575887,-5.97791857375092,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394474,-7.32142511479571,-6.67435503772231,-1.44773853151531,7.11105971439921,7.46215578440288,1.61862088558446,-8.17438210188192,7.8090615264033,1.69386842694778,-8.55439830449838,-8.51143945889148},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-3.13195404884644,-.679354114255872,3.33687856485653,1.51590795070228,-2.98096261298775,3.83584465679206,.832035467367304,-4.08682490876582,-1.85660048821863,-4.42925189189755,-.960751801781151,4.71905892262039,4.95205415939119,1.07415316937022,-5.42470353822284,3.97724364248013,.862706409586763,-4.2374756641165,-1.92503950203907,-4.59252537503604,-.996167555656712,4.89301542900031,5.13459946240545,1.1137491858232,-5.62467189980939,-3.67768288095471,-.797728497114731,3.91831456811704,4.11177446033512,.891887573859297,-4.5042232468252,2.21014640326355,.479404240671233,-2.42109408164425,.161112776439225},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.66702109783921,-.578505216668687,2.84152493765918,1.29087413922765,-2.53844406932997,3.26642040979696,.70852129725097,-3.48014309433046,-1.58099148171138,-3.77173740709444,-.818129923100605,4.01852310483628,4.21693061735794,.914697061432424,-4.61941604516583,2.13132645129358,.462307355479536,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292297,2.62207049601641,2.75153061636385,.596836229618421,-3.01415077313091,.104423971232213,.0226506690028153,-.111256457118347,-.116749549066047,-.0253242178213241,.127892723210502,-2.94703586087065,-.639243394476215,3.22831603852647,4.81780334972074},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-2.03452907455009,-.441310975771777,2.16764880728185,.984739479553153,-1.93644447258228,2.49177904847469,.540493354313145,-2.65481675969737,-1.20605462730079,-2.87725860980962,-.624107966995901,3.06551834185512,3.21687291987534,.697773919897946,-3.52390772571971,.0953780462518065,.0206885117513158,-.101618655082187,-.0461642592626142,-.110133081351697,-.0238890356585565,.117339115691434,.123132528211026,.0267087538273405,-.134884926556002,2.8024998850757,.607892005333882,-2.98586813553316,-3.1332901246654,-.6796439236741,3.43234736114769,-3.77550789188145,-.818947781641524,4.13586167809416,1.3916398503358},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-1.2742056900953,-.27638875426455,1.35757727865237,.616732719037494,-1.21277626178116,1.56057688324205,.338505709325782,-1.66268580954775,-.755340234698916,-1.80199896789671,-.390872724665213,1.91990419952445,2.01469609360342,.437008991543009,-2.20698899398908,-1.62700976129611,-.352915706379786,1.73346540682961,.787494642582441,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933792,-.455612217872568,2.30093927043685,3.19783168117628,.693643672830323,-3.40706658739604,-3.57528450949468,-.775517202558345,3.91652795089531,-.49602746920004,-.107593629034159,.543370868210304,-3.52996795992507},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,-.433821252801768,-.0941004396444027,.462206283002972,.209975330432464,-.41290673973901,.531320354197217,.115249030860378,-.566084774633034,-.257166209078973,-.613515899259544,-.133078117935277,.653658394065701,.68593162794344,.148785859041837,-.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353978,-3.3691862575809,-.730812301668709,3.69075862820297,1.36899683098764,.296949960039812,-1.45857062726747,-1.53058498738027,-.332000148390628,1.67667184761831,3.00982432351404,.652862879236147,-3.29709735322896,-2.5031295519909},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,.433821252801768,.0941004396444027,-.462206283002972,-.209975330432464,.41290673973901,-.531320354197217,-.115249030860378,.566084774633034,.257166209078973,.613515899259544,.133078117935277,-.653658394065701,-.68593162794344,-.148785859041837,.751400451068813,-2.60976045384023,-.566084774633034,2.7805177168944,1.26315921681103,3.01349180095816,.653658394065701,-3.21066530353979,-3.36918625758091,-.730812301668709,3.69075862820298,-1.36899683098765,-.296949960039813,1.45857062726747,1.53058498738027,.332000148390629,-1.67667184761832,3.00982432351407,.652862879236155,-3.297097353229,2.50312955199098},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,1.2742056900953,.27638875426455,-1.35757727865237,-.616732719037494,1.21277626178116,-1.56057688324205,-.338505709325782,1.66268580954775,.755340234698916,1.80199896789671,.390872724665213,-1.91990419952445,-2.01469609360342,-.437008991543009,2.20698899398908,-1.62700976129612,-.352915706379786,1.73346540682961,.787494642582442,1.87870904723461,.407511956177088,-2.00163343782271,-2.10046056933793,-.455612217872569,2.30093927043686,-3.1978316811763,-.693643672830326,3.40706658739606,3.5752845094947,.775517202558348,-3.91652795089533,-.496027469200023,-.107593629034155,.543370868210286,3.52996795992515},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.03452907455009,.441310975771777,-2.16764880728185,-.984739479553153,1.93644447258228,-2.49177904847469,-.540493354313145,2.65481675969737,1.20605462730079,2.87725860980962,.624107966995901,-3.06551834185512,-3.21687291987534,-.697773919897946,3.52390772571971,.0953780462518019,.0206885117513148,-.101618655082182,-.046164259262612,-.110133081351691,-.0238890356585554,.117339115691428,.12313252821102,.0267087538273392,-.134884926555995,-2.8024998850757,-.607892005333881,2.98586813553315,3.13329012466539,.679643923674099,-3.43234736114768,-3.77550789188148,-.818947781641531,4.1358616780942,-1.39163985033612},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,2.66702109783921,.578505216668687,-2.84152493765918,-1.29087413922765,2.53844406932997,-3.26642040979696,-.70852129725097,3.48014309433046,1.58099148171138,3.77173740709444,.818129923100605,-4.01852310483628,-4.21693061735794,-.914697061432424,4.61941604516583,2.13132645129358,.462307355479535,-2.27077966086395,-1.03159071439798,-2.46104380066281,-.533826552292296,2.62207049601641,2.75153061636384,.59683622961842,-3.0141507731309,-.104423971232209,-.0226506690028144,.111256457118342,.116749549066043,.0253242178213231,-.127892723210497,-2.94703586087065,-.639243394476216,3.22831603852648,-4.81780334971999},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.13195404884644,.679354114255872,-3.33687856485653,-1.51590795070228,2.98096261298775,-3.83584465679206,-.832035467367304,4.08682490876582,1.85660048821863,4.42925189189755,.960751801781151,-4.71905892262039,-4.95205415939119,-1.07415316937022,5.42470353822284,3.97724364248012,.862706409586761,-4.23747566411649,-1.92503950203906,-4.59252537503603,-.996167555656711,4.89301542900031,5.13459946240544,1.1137491858232,-5.62467189980938,3.67768288095471,.797728497114731,-3.91831456811703,-4.11177446033511,-.891887573859297,4.5042232468252,2.21014640326349,.47940424067122,-2.42109408164419,-.161112776439403},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-2.,-.433821252801768,2.13086048714477,.968026942532141,-1.90358004395733,2.449489743,.531320354197217,-2.60976045384023,-1.18558603318945,-2.828427124,-.613515899259544,3.01349180095816,3.162277660,.68593162794344,-3.464101616,3.40022577929605,.737545103324376,-3.62270337995047,-1.64575508217186,3.23630096920864,-4.1644090826311,-.903304582796468,4.43688738516868,2.01563009655898,4.80864541143983,1.04304628770802,-5.12327625260686,-5.37622900930936,-1.16616120174253,5.88936380548032,5.17702934684997,1.12295267819227,-5.51576363976884,-2.50575194575888,-5.97791857375093,-1.29667406213985,6.36905524216974,6.68351614602547,1.44972567394475,-7.32142511479571,6.67435503772228,1.4477385315153,-7.11105971439918,-7.46215578440284,-1.61862088558445,8.17438210188188,7.8090615264033,1.69386842694778,-8.55439830449838,8.51143945889045},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.40022577929605,-2.16629351671545,2.25851140965319,3.84430965952583,.533351454031679,4.1644090826311,2.65315687476095,-2.76610026595298,-4.70829853978449,-4.80864541143983,-3.06360167061167,3.19401746644684,5.37622900930936,3.42521079645606,-5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375092,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,-6.67435503772231,-4.25225058180805,4.43326649059392,7.46215578440288,4.75416068069224,-8.17438210188192,7.8090615264033,4.9751753146033,-8.55439830449838,-8.51143945889148},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-3.13195404884644,-1.99537683408524,2.08031890026245,3.54100050546176,.491270978557926,3.83584465679206,2.4438275442558,-2.54785990395621,-4.33682220887954,-4.42925189189755,-2.82188898006397,2.94201520294205,4.95205415939119,3.15496779285464,-5.42470353822284,3.97724364248013,2.53391323933003,-2.64178049699053,-4.4966885005097,-4.59252537503604,-2.92591098178889,3.05046536107998,5.13459946240545,3.27126792601698,-5.62467189980939,-3.67768288095471,-2.34306222576705,2.44280506311384,4.11177446033512,2.61962320701725,-4.5042232468252,2.21014640326355,1.40809056120224,-2.42109408164425,.161112776439225},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.66702109783921,-1.69916672838959,1.77149929746803,3.01534534295106,.418342684514381,3.26642040979696,2.08104573641859,-2.16963467924844,-3.69302874444135,-3.77173740709444,-2.40298463138773,2.50527833232424,4.21693061735794,2.68661849290085,-4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207907,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297245,2.75153061636385,1.75300798410243,-3.01415077313091,.104423971232213,.0665288091384509,-.0693609030179037,-.116749549066047,-.0743814698720828,.127892723210502,-2.94703586087065,-1.87756493103306,3.22831603852647,4.81780334972074},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-2.03452907455009,-1.29620426108275,1.35138294525072,2.3002471840262,.319131466736208,2.49177904847469,1.58751952117754,-1.65509933148239,-2.8172159417121,-2.87725860980962,-1.83310964514541,1.91114408921759,3.21687291987534,2.04947888880939,-3.52390772571971,.0953780462518065,.0607656246043247,-.0633523829764579,-.107834822891096,-.110133081351697,-.0701660994486395,.07315303070475,.123132528211026,.0784480840492681,-.134884926556002,2.8024998850757,1.78548064936278,-1.86148755350589,-3.1332901246654,-1.99622805289337,3.43234736114769,-3.77550789188145,-2.40538681892419,4.13586167809416,1.3916398503358},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-1.2742056900953,-.811800069931489,.846357940948568,1.4406223470461,.199868920965732,1.56057688324205,.994247972331933,-1.03657254753862,-1.76439483124642,-1.80199896789671,-1.14805866852966,1.19693087876444,2.01469609360342,1.28356861276539,-2.20698899398908,-1.62700976129611,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.24788355079949,-2.10046056933792,-1.338209404774,2.30093927043685,3.19783168117628,2.03734766130249,-2.12407640207831,-3.57528450949468,-2.27782393297861,3.91652795089531,-.49602746920004,-.316020511784317,.543370868210304,-3.52996795992507},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,-.433821252801768,-.27638875426455,.2881544676147,.490480144821061,.0680481859118103,.531320354197217,.338505709325782,-.352915706379786,-.600713041919503,-.613515899259544,-.390872724665213,.407511956177088,.68593162794344,.437008991543009,-.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.3691862575809,-2.14651815042224,3.69075862820297,1.36899683098764,.872191775558762,-.909320487484577,-1.53058498738027,-.975140050100603,1.67667184761831,3.00982432351404,1.91756763921113,-3.29709735322896,-2.5031295519909},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,.433821252801768,.27638875426455,-.2881544676147,-.490480144821061,-.0680481859118103,-.531320354197217,-.338505709325782,.352915706379786,.600713041919503,.613515899259544,.390872724665213,-.407511956177088,-.68593162794344,-.437008991543009,.751400451068813,-2.60976045384023,-1.66268580954775,1.73346540682961,2.95060621796619,3.01349180095816,1.91990419952445,-2.00163343782271,-3.36918625758091,-2.14651815042224,3.69075862820298,-1.36899683098765,-.872191775558764,.909320487484578,1.53058498738027,.975140050100604,-1.67667184761832,3.00982432351407,1.91756763921115,-3.297097353229,2.50312955199098},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,1.2742056900953,.811800069931489,-.846357940948568,-1.4406223470461,-.199868920965732,-1.56057688324205,-.994247972331933,1.03657254753862,1.76439483124642,1.80199896789671,1.14805866852966,-1.19693087876444,-2.01469609360342,-1.28356861276539,2.20698899398908,-1.62700976129612,-1.03657254753862,1.08069885633786,1.83950412433749,1.87870904723461,1.19693087876444,-1.2478835507995,-2.10046056933793,-1.338209404774,2.30093927043686,-3.1978316811763,-2.0373476613025,2.12407640207832,3.5752845094947,2.27782393297862,-3.91652795089533,-.496027469200023,-.316020511784306,.543370868210286,3.52996795992515},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.03452907455009,1.29620426108275,-1.35138294525072,-2.3002471840262,-.319131466736208,-2.49177904847469,-1.58751952117754,1.65509933148239,2.8172159417121,2.87725860980962,1.83310964514541,-1.91114408921759,-3.21687291987534,-2.04947888880939,3.52390772571971,.0953780462518019,.0607656246043218,-.0633523829764548,-.10783482289109,-.110133081351691,-.0701660994486362,.0731530307047465,.12313252821102,.0784480840492644,-.134884926555995,-2.8024998850757,-1.78548064936277,1.86148755350589,3.13329012466539,1.99622805289336,-3.43234736114768,-3.77550789188148,-2.40538681892421,4.1358616780942,-1.39163985033612},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,2.66702109783921,1.69916672838959,-1.77149929746803,-3.01534534295106,-.418342684514381,-3.26642040979696,-2.08104573641859,2.16963467924844,3.69302874444135,3.77173740709444,2.40298463138773,-2.50527833232424,-4.21693061735794,-2.68661849290085,4.61941604516583,2.13132645129358,1.35787414544697,-1.41567808207906,-2.40968670915711,-2.46104380066281,-1.56793800686758,1.63468424297244,2.75153061636384,1.75300798410242,-3.0141507731309,-.104423971232209,-.0665288091384484,.069360903017901,.116749549066043,.0743814698720799,-.127892723210497,-2.94703586087065,-1.87756493103306,3.22831603852648,-4.81780334971999},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.13195404884644,1.99537683408524,-2.08031890026245,-3.54100050546176,-.491270978557926,-3.83584465679206,-2.4438275442558,2.54785990395621,4.33682220887954,4.42925189189755,2.82188898006397,-2.94201520294205,-4.95205415939119,-3.15496779285464,5.42470353822284,3.97724364248012,2.53391323933002,-2.64178049699052,-4.49668850050969,-4.59252537503603,-2.92591098178889,3.05046536107998,5.13459946240544,3.27126792601697,-5.62467189980938,3.67768288095471,2.34306222576705,-2.44280506311384,-4.11177446033511,-2.61962320701725,4.5042232468252,2.21014640326349,1.4080905612022,-2.42109408164419,-.161112776439403},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-2.,-1.2742056900953,1.32844790694992,2.26120846666112,.31371531695288,2.449489743,1.56057688324205,-1.62700976129612,-2.76940347258376,-2.828427124,-1.80199896789671,1.87870904723461,3.162277660,2.01469609360342,-3.464101616,3.40022577929605,2.16629351671545,-2.25851140965319,-3.84430965952583,-.533351454031679,-4.1644090826311,-2.65315687476095,2.76610026595298,4.70829853978449,4.80864541143983,3.06360167061167,-3.19401746644684,-5.37622900930936,-3.42521079645606,5.88936380548032,5.17702934684997,3.2983001248078,-3.43870690112599,-5.85317129736233,-5.97791857375093,-3.80854893001879,3.97067670865962,6.68351614602547,4.25808715199512,-7.32142511479571,6.67435503772228,4.25225058180803,-4.4332664905939,-7.46215578440284,-4.75416068069222,8.17438210188188,7.8090615264033,4.9751753146033,-8.55439830449838,8.51143945889045},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.40022577929605,-3.45892910228435,-.132397734060625,3.36905705276318,4.05959901190685,4.1644090826311,4.23630567890486,.162153445774668,-4.1262353470069,-4.80864541143983,-4.89166444644702,-.187238671144264,5.37622900930936,5.46904711383883,-5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375092,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,-6.67435503772231,-6.78958468653045,-.259885531387094,7.46215578440288,7.59098645151066,-8.17438210188192,7.8090615264033,7.94388135857975,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-3.13195404884644,-3.18602578468041,-.121951789723539,3.10324447907128,3.73930391312631,3.83584465679206,3.9020687402541,.149359829020976,-3.80068276060981,-4.42925189189755,-4.50572087357673,-.172465874990311,4.95205415939119,5.03754908153942,-5.42470353822284,3.97724364248013,4.04590891241332,.154865612009818,-3.94078558921028,-4.59252537503604,-4.67181319960176,-.178823405501163,5.13459946240545,5.22324594675678,-5.62467189980939,-3.67768288095471,-3.74117637274605,-.143201337715517,4.11177446033512,4.18276234400124,-4.5042232468252,2.21014640326355,2.24830355776488,-2.42109408164425,.161112776439225},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.66702109783921,-2.71306598164557,-.103848265663963,2.64257341211137,3.18421096606247,3.26642040979696,3.32281364706153,.127187630774889,-3.23647823392352,-3.77173740709444,-3.83685470584401,-.146863625737381,4.21693061735794,4.28973397193188,-4.61941604516583,2.13132645129358,2.16812281562257,.0829894280920819,-2.11178929936586,-2.46104380066281,-2.50353258258141,-.0958279372639496,2.75153061636385,2.79903451954298,-3.01415077313091,.104423971232213,.106226802736416,.00406605268971361,-.116749549066047,-.118765176014247,.127892723210502,-2.94703586087065,-2.99791507072668,3.22831603852647,4.81780334972074},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-2.03452907455009,-2.0696542765645,-.0792203391289677,2.01587923055033,2.4290658199907,2.49177904847469,2.53479846100042,.0970247040581356,-2.46893774908671,-2.87725860980962,-2.92693314656882,-.112034478016925,3.21687291987534,3.2724107413517,-3.52390772571971,.0953780462518065,.0970247040581402,.00371382314810264,-.0945037478171113,-.110133081351697,-.112034478016931,-.00428835358704851,.123132528211026,.12525835434705,-.134884926556002,2.8024998850757,2.85088374774363,.109123528450132,-3.1332901246654,-3.18738492899079,3.43234736114769,-3.77550789188145,-3.84069028782535,4.13586167809416,1.3916398503358},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-1.2742056900953,-1.29620426108275,-.0496149247273511,1.2625254749334,1.52130020070252,1.56057688324205,1.58751952117754,.0607656246043218,-1.54627160050443,-1.80199896789671,-1.83310964514541,-.0701660994486362,2.01469609360342,2.04947888880939,-2.20698899398908,-1.62700976129611,-1.65509933148238,-.0633523829764547,1.6120955106542,1.87870904723461,1.91114408921758,.0731530307047464,-2.10046056933792,-2.13672404935301,2.30093927043685,3.19783168117628,3.25304076422429,.124516928010559,-3.57528450949468,-3.63701014237694,3.91652795089531,-.49602746920004,-.504591153828001,.543370868210304,-3.52996795992507},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,-.433821252801768,-.441310975771777,-.0168920991094264,.42984455923187,.5179480550797,.531320354197217,.540493354313145,.0206885117513148,-.526449919441544,-.613515899259544,-.624107966995901,-.0238890356585554,.68593162794344,.697773919897946,-.751400451068813,-2.60976045384023,-2.65481675969736,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.3691862575809,-3.42735369966595,3.69075862820297,1.36899683098764,1.39263192728723,.0533058950082293,-1.53058498738027,-1.5570098290328,1.67667184761831,3.00982432351404,3.06178754711072,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,.433821252801768,.441310975771777,.0168920991094264,-.42984455923187,-.5179480550797,-.531320354197217,-.540493354313145,-.0206885117513148,.526449919441544,.613515899259544,.624107966995901,.0238890356585554,-.68593162794344,-.697773919897946,.751400451068813,-2.60976045384023,-2.65481675969737,-.101618655082182,2.58583765850746,3.01349180095816,3.06551834185512,.117339115691428,-3.36918625758091,-3.42735369966595,3.69075862820298,-1.36899683098765,-1.39263192728723,-.0533058950082294,1.53058498738027,1.5570098290328,-1.67667184761832,3.00982432351407,3.06178754711075,-3.297097353229,2.50312955199098},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,1.2742056900953,1.29620426108275,.0496149247273511,-1.2625254749334,-1.52130020070252,-1.56057688324205,-1.58751952117754,-.0607656246043218,1.54627160050443,1.80199896789671,1.83310964514541,.0701660994486362,-2.01469609360342,-2.04947888880939,2.20698899398908,-1.62700976129612,-1.65509933148239,-.0633523829764548,1.61209551065421,1.87870904723461,1.91114408921759,.0731530307047465,-2.10046056933793,-2.13672404935301,2.30093927043686,-3.1978316811763,-3.2530407642243,-.12451692801056,3.5752845094947,3.63701014237696,-3.91652795089533,-.496027469200023,-.504591153827984,.543370868210286,3.52996795992515},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.03452907455009,2.0696542765645,.0792203391289677,-2.01587923055033,-2.4290658199907,-2.49177904847469,-2.53479846100042,-.0970247040581356,2.46893774908671,2.87725860980962,2.92693314656882,.112034478016925,-3.21687291987534,-3.2724107413517,3.52390772571971,.0953780462518019,.0970247040581356,.00371382314810246,-.0945037478171067,-.110133081351691,-.112034478016925,-.0042883535870483,.12313252821102,.125258354347044,-.134884926555995,-2.8024998850757,-2.85088374774362,-.109123528450131,3.13329012466539,3.18738492899078,-3.43234736114768,-3.77550789188148,-3.84069028782539,4.1358616780942,-1.39163985033612},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,2.66702109783921,2.71306598164557,.103848265663963,-2.64257341211137,-3.18421096606247,-3.26642040979696,-3.32281364706153,-.127187630774889,3.23647823392352,3.77173740709444,3.83685470584401,.146863625737381,-4.21693061735794,-4.28973397193188,4.61941604516583,2.13132645129358,2.16812281562256,.0829894280920818,-2.11178929936586,-2.46104380066281,-2.5035325825814,-.0958279372639495,2.75153061636384,2.79903451954297,-3.0141507731309,-.104423971232209,-.106226802736412,-.00406605268971345,.116749549066043,.118765176014243,-.127892723210497,-2.94703586087065,-2.99791507072669,3.22831603852648,-4.81780334971999},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.13195404884644,3.18602578468041,.121951789723539,-3.10324447907128,-3.73930391312631,-3.83584465679206,-3.9020687402541,-.149359829020976,3.80068276060981,4.42925189189755,4.50572087357673,.172465874990311,-4.95205415939119,-5.03754908153942,5.42470353822284,3.97724364248012,4.04590891241331,.154865612009818,-3.94078558921027,-4.59252537503603,-4.67181319960176,-.178823405501163,5.13459946240544,5.22324594675677,-5.62467189980938,3.67768288095471,3.74117637274605,.143201337715517,-4.11177446033511,-4.18276234400124,4.5042232468252,2.21014640326349,2.24830355776482,-2.42109408164419,-.161112776439403},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-2.,-2.03452907455009,-.0778758486434911,1.98166667284344,2.38784085258436,2.449489743,2.49177904847469,.0953780462518019,-2.42703609427915,-2.828427124,-2.87725860980962,-.110133081351691,3.162277660,3.21687291987534,-3.464101616,3.40022577929605,3.45892910228435,.132397734060625,-3.36905705276318,-4.05959901190685,-4.1644090826311,-4.23630567890486,-.162153445774668,4.1262353470069,4.80864541143983,4.89166444644702,.187238671144264,-5.37622900930936,-5.46904711383883,5.88936380548032,5.17702934684997,5.26640836144105,.201582776983901,-5.12957326201497,-5.97791857375093,-6.08112457054741,-.232767741029283,6.68351614602547,6.79890396036006,-7.32142511479571,6.67435503772228,6.78958468653042,.259885531387093,-7.46215578440284,-7.59098645151062,8.17438210188188,7.8090615264033,7.94388135857975,-8.55439830449838,8.51143945889045},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.40022577929605,-4.53423694314245,-2.95857174458961,.125534462510035,3.16878793832491,4.1644090826311,5.55328344227955,3.62349557083131,-.153747689149873,-4.80864541143983,-6.41237937831348,-4.18405228663707,5.37622900930936,7.16925809522303,-5.88936380548032,5.17702934684997,6.90362324406996,4.50458697357606,-.191133071440498,-5.97791857375092,-7.97161747704437,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,-6.67435503772231,-8.90032284672419,-5.80742559866953,7.46215578440288,9.95086345672774,-8.17438210188192,7.8090615264033,10.4134659204853,-8.55439830449838,-8.51143945889148},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-3.13195404884644,-4.17649376078907,-2.72514572729005,.115630018018823,2.91877623944935,3.83584465679206,5.11513931437816,3.33760825329423,-.141617271554661,-4.42925189189755,-5.90645411811629,-3.85393804714673,4.95205415939119,6.60361645843633,-5.42470353822284,3.97724364248013,5.30369635131817,3.46064097812297,-.146837644201682,-4.59252537503604,-6.12418103253549,-3.99600399916796,5.13459946240545,6.84704254830277,-5.62467189980939,-3.67768288095471,-4.90422891550851,-3.19999004936568,4.11177446033512,5.48309461793542,-4.5042232468252,2.21014640326355,2.9472535428042,-2.42109408164425,.161112776439225},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.66702109783921,-3.55650076638925,-2.32060274065832,.0984649496097535,2.4854891512059,3.26642040979696,4.35580607412106,2.84214630515942,-.120594442052501,-3.77173740709444,-5.02965161709108,-3.28182786886397,4.21693061735794,5.62332146066281,-4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258921,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636385,3.66919510297595,-3.01415077313091,.104423971232213,.139250467146397,.0908603812984509,-.116749549066047,-.155686755269708,.127892723210502,-2.94703586087065,-3.92990340770938,3.22831603852647,4.81780334972074},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-2.03452907455009,-2.71306598164557,-1.77026486598667,.0751136925641335,1.89604797153805,2.49177904847469,3.32281364706153,2.16812281562256,-.0919951097438786,-2.87725860980962,-3.83685470584401,-2.5035325825814,3.21687291987534,4.28973397193188,-3.52390772571971,.0953780462518065,.127187630774895,.0829894280920858,-.00352130492348742,-.110133081351697,-.146863625737388,-.095827937263954,.123132528211026,.164198525301527,-.134884926556002,2.8024998850757,3.73716315870324,2.43848424018078,-3.1332901246654,-4.17827543429579,3.43234736114769,-3.77550789188145,-5.03467960032056,4.13586167809416,1.3916398503358},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-1.2742056900953,-1.69916672838959,-1.10869959708726,.0470429720894763,1.18747632769106,1.56057688324205,2.08104573641859,1.35787414544697,-.0576156388045296,-1.80199896789671,-2.40298463138773,-1.56793800686758,2.01469609360342,2.68661849290085,-2.20698899398908,-1.62700976129611,-2.16963467924844,-1.41567808207906,.0600683040611726,1.87870904723461,2.50527833232423,1.63468424297244,-2.10046056933792,-2.8009863270915,2.30093927043685,3.19783168117628,4.26434227893045,2.78246653954404,-3.57528450949468,-4.76767960906116,3.91652795089531,-.49602746920004,-.661457862596488,.543370868210304,-3.52996795992507},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,-.433821252801768,-.578505216668687,-.377472375086662,.0160164416514644,.404293021256913,.531320354197217,.70852129725097,.462307355479535,-.0196160547715697,-.613515899259544,-.818129923100605,-.533826552292296,.68593162794344,.914697061432424,-.751400451068813,-2.60976045384023,-3.48014309433045,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.3691862575809,-4.49284541622379,3.69075862820297,1.36899683098764,1.82557171487995,1.19117835294064,-1.53058498738027,-2.04105122680137,1.67667184761831,3.00982432351404,4.01363248497761,-3.29709735322896,-2.5031295519909},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,.433821252801768,.578505216668687,.377472375086662,-.0160164416514644,-.404293021256913,-.531320354197217,-.70852129725097,-.462307355479535,.0196160547715697,.613515899259544,.818129923100605,.533826552292296,-.68593162794344,-.914697061432424,.751400451068813,-2.60976045384023,-3.48014309433046,-2.27077966086395,.0963509182287986,3.01349180095816,4.01852310483628,2.62207049601641,-3.36918625758091,-4.49284541622379,3.69075862820298,-1.36899683098765,-1.82557171487995,-1.19117835294064,1.53058498738027,2.04105122680137,-1.67667184761832,3.00982432351407,4.01363248497766,-3.297097353229,2.50312955199098},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,1.2742056900953,1.69916672838959,1.10869959708726,-.0470429720894763,-1.18747632769106,-1.56057688324205,-2.08104573641859,-1.35787414544697,.0576156388045296,1.80199896789671,2.40298463138773,1.56793800686758,-2.01469609360342,-2.68661849290085,2.20698899398908,-1.62700976129612,-2.16963467924844,-1.41567808207906,.0600683040611727,1.87870904723461,2.50527833232424,1.63468424297244,-2.10046056933793,-2.8009863270915,2.30093927043686,-3.1978316811763,-4.26434227893047,-2.78246653954406,3.5752845094947,4.76767960906118,-3.91652795089533,-.496027469200023,-.661457862596465,.543370868210286,3.52996795992515},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.03452907455009,2.71306598164557,1.77026486598667,-.0751136925641335,-1.89604797153805,-2.49177904847469,-3.32281364706153,-2.16812281562256,.0919951097438786,2.87725860980962,3.83685470584401,2.5035325825814,-3.21687291987534,-4.28973397193188,3.52390772571971,.0953780462518019,.127187630774889,.0829894280920818,-.00352130492348725,-.110133081351691,-.146863625737381,-.0958279372639495,.12313252821102,.164198525301519,-.134884926555995,-2.8024998850757,-3.73716315870323,-2.43848424018078,3.13329012466539,4.17827543429578,-3.43234736114768,-3.77550789188148,-5.0346796003206,4.1358616780942,-1.39163985033612},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,2.66702109783921,3.55650076638925,2.32060274065832,-.0984649496097535,-2.4854891512059,-3.26642040979696,-4.35580607412106,-2.84214630515942,.120594442052501,3.77173740709444,5.02965161709108,3.28182786886397,-4.21693061735794,-5.62332146066281,4.61941604516583,2.13132645129358,2.84214630515942,1.85448926898142,-.0786873984258919,-2.46104380066281,-3.28182786886397,-2.14137975655691,2.75153061636384,3.66919510297594,-3.0141507731309,-.104423971232209,-.139250467146391,-.0908603812984474,.116749549066043,.155686755269702,-.127892723210497,-2.94703586087065,-3.92990340770939,3.22831603852648,-4.81780334971999},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.13195404884644,4.17649376078907,2.72514572729005,-.115630018018823,-2.91877623944935,-3.83584465679206,-5.11513931437816,-3.33760825329423,.141617271554661,4.42925189189755,5.90645411811629,3.85393804714673,-4.95205415939119,-6.60361645843633,5.42470353822284,3.97724364248012,5.30369635131817,3.46064097812296,-.146837644201681,-4.59252537503603,-6.12418103253548,-3.99600399916795,5.13459946240544,6.84704254830276,-5.62467189980938,3.67768288095471,4.90422891550851,3.19999004936568,-4.11177446033511,-5.48309461793542,4.5042232468252,2.21014640326349,2.94725354280412,-2.42109408164419,-.161112776439403},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-2.,-2.66702109783921,-1.74022075994948,.0738388981704604,1.86386913340857,2.449489743,3.26642040979696,2.13132645129358,-.0904338118399935,-2.828427124,-3.77173740709444,-2.46104380066281,3.162277660,4.21693061735794,-3.464101616,3.40022577929605,4.53423694314245,2.95857174458961,-.125534462510035,-3.16878793832491,-4.1644090826311,-5.55328344227955,-3.62349557083131,.153747689149873,4.80864541143983,6.41237937831348,4.18405228663707,-5.37622900930936,-7.16925809522303,5.88936380548032,5.17702934684997,6.90362324406997,4.50458697357606,-.191133071440498,-5.97791857375093,-7.97161747704438,-5.20144900173179,6.68351614602547,8.91253928552127,-7.32142511479571,6.67435503772228,8.90032284672415,5.8074255986695,-7.46215578440284,-9.95086345672769,8.17438210188188,7.8090615264033,10.4134659204853,-8.55439830449838,8.51143945889045},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.40022577929605,-5.32467544557844,-5.5209565174066,-4.42116822694256,-2.37645064234965,4.1644090826311,6.52136894437417,6.76176317987178,5.41480311178231,-4.80864541143983,-7.53022822838542,-7.80781158453304,5.37622900930936,8.41905110415162,-5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394288,-5.97791857375092,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,-6.67435503772231,-10.4518866380257,-10.8371697481941,7.46215578440288,11.6855645116818,-8.17438210188192,7.8090615264033,12.2288109301468,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-3.13195404884644,-4.90456807960123,-5.08536292603975,-4.0723459639997,-2.18895293851093,3.83584465679206,6.00684460241421,6.22827216283403,4.98758483419414,-4.42925189189755,-6.93610669392436,-7.19178921991206,4.95205415939119,7.75480303503604,-5.42470353822284,3.97724364248013,6.22827216283404,6.45786210779302,5.17143989055336,-4.59252537503604,-7.19178921991208,-7.45689685002024,5.13459946240545,8.04066478857437,-5.62467189980939,-3.67768288095471,-5.75916689254577,-5.97146442400447,4.11177446033512,6.43894433484766,-4.5042232468252,2.21014640326355,3.46103848741249,-2.42109408164425,.161112776439225},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.66702109783921,-4.17649376078907,-4.3304499371288,-3.46781352289888,-1.86400680793386,3.26642040979696,5.11513931437816,5.30369635131817,4.24718682732847,-3.77173740709444,-5.90645411811629,-6.12418103253548,4.21693061735794,6.60361645843633,-4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812297,2.77127267544969,-2.46104380066281,-3.85393804714674,-3.99600399916796,2.75153061636385,4.3088337276685,-3.01415077313091,.104423971232213,.163525539687803,.169553506762539,-.116749549066047,-.182827111459215,.127892723210502,-2.94703586087065,-4.61499044732823,3.22831603852647,4.81780334972074},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-2.03452907455009,-3.18602578468041,-3.30347079372946,-2.6454111829756,-1.4219520231667,2.49177904847469,3.9020687402541,4.04590891241331,3.23995377923585,-2.87725860980962,-4.50572087357673,-4.67181319960176,3.21687291987534,5.03754908153942,-3.52390772571971,.0953780462518065,.149359829020983,.154865612009825,.12401599635522,-.110133081351697,-.172465874990319,-.178823405501172,.123132528211026,.192822710157554,-.134884926556002,2.8024998850757,4.38865042934355,4.55042724011643,-3.1332901246654,-4.90666034634693,3.43234736114769,-3.77550789188145,-5.91235861300231,4.13586167809416,1.3916398503358},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-1.2742056900953,-1.99537683408524,-2.06893149627982,-1.65679518870216,-.890554665266804,1.56057688324205,2.4438275442558,2.53391323933002,2.02915141041227,-1.80199896789671,-2.82188898006397,-2.92591098178889,2.01469609360342,3.15496779285464,-2.20698899398908,-1.62700976129611,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294204,3.05046536107998,-2.10046056933792,-3.2892729926303,2.30093927043685,3.19783168117628,5.00773093883056,5.19232863088549,-3.57528450949468,-5.59881339795158,3.91652795089531,-.49602746920004,-.776767620090749,.543370868210304,-3.52996795992507},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,-.433821252801768,-.679354114255872,-.704396833771808,-.564079229896515,-.303201864171246,.531320354197217,.832035467367304,.862706409586761,.690853143909355,-.613515899259544,-.960751801781151,-.996167555656711,.68593162794344,1.07415316937022,-.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.8930154290003,-3.3691862575809,-5.27606827091984,3.69075862820297,1.36899683098764,2.14381758303679,2.22284414873079,-1.53058498738027,-2.39686092429589,1.67667184761831,3.00982432351404,4.71331573720647,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,.433821252801768,.679354114255872,.704396833771808,.564079229896515,.303201864171246,-.531320354197217,-.832035467367304,-.862706409586761,-.690853143909355,.613515899259544,.960751801781151,.996167555656711,-.68593162794344,-1.07415316937022,.751400451068813,-2.60976045384023,-4.08682490876582,-4.23747566411649,-3.3933599570385,3.01349180095816,4.71905892262039,4.89301542900031,-3.36918625758091,-5.27606827091984,3.69075862820298,-1.36899683098765,-2.1438175830368,-2.22284414873079,1.53058498738027,2.3968609242959,-1.67667184761832,3.00982432351407,4.71331573720652,-3.297097353229,2.50312955199098},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,1.2742056900953,1.99537683408524,2.06893149627982,1.65679518870216,.890554665266804,-1.56057688324205,-2.4438275442558,-2.53391323933002,-2.02915141041227,1.80199896789671,2.82188898006397,2.92591098178889,-2.01469609360342,-3.15496779285464,2.20698899398908,-1.62700976129612,-2.54785990395621,-2.64178049699052,-2.11553124179228,1.87870904723461,2.94201520294205,3.05046536107998,-2.10046056933793,-3.28927299263031,2.30093927043686,-3.1978316811763,-5.00773093883059,-5.19232863088551,3.5752845094947,5.5988133979516,-3.91652795089533,-.496027469200023,-.776767620090723,.543370868210286,3.52996795992515},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.03452907455009,3.18602578468041,3.30347079372946,2.6454111829756,1.4219520231667,-2.49177904847469,-3.9020687402541,-4.04590891241331,-3.23995377923585,2.87725860980962,4.50572087357673,4.67181319960176,-3.21687291987534,-5.03754908153942,3.52390772571971,.0953780462518019,.149359829020976,.154865612009818,.124015996355214,-.110133081351691,-.172465874990311,-.178823405501163,.12313252821102,.192822710157544,-.134884926555995,-2.8024998850757,-4.38865042934354,-4.55042724011642,3.13329012466539,4.90666034634692,-3.43234736114768,-3.77550789188148,-5.91235861300236,4.1358616780942,-1.39163985033612},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,2.66702109783921,4.17649376078907,4.3304499371288,3.46781352289888,1.86400680793386,-3.26642040979696,-5.11513931437816,-5.30369635131817,-4.24718682732847,3.77173740709444,5.90645411811629,6.12418103253548,-4.21693061735794,-6.60361645843633,4.61941604516583,2.13132645129358,3.33760825329423,3.46064097812296,2.77127267544968,-2.46104380066281,-3.85393804714673,-3.99600399916795,2.75153061636384,4.30883372766849,-3.0141507731309,-.104423971232209,-.163525539687797,-.169553506762532,.116749549066043,.182827111459208,-.127892723210497,-2.94703586087065,-4.61499044732824,3.22831603852648,-4.81780334971999},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.13195404884644,4.90456807960123,5.08536292603975,4.0723459639997,2.18895293851093,-3.83584465679206,-6.00684460241421,-6.22827216283403,-4.98758483419414,4.42925189189755,6.93610669392436,7.19178921991206,-4.95205415939119,-7.75480303503604,5.42470353822284,3.97724364248012,6.22827216283403,6.45786210779301,5.17143989055335,-4.59252537503603,-7.19178921991206,-7.45689685002023,5.13459946240544,8.04066478857435,-5.62467189980938,3.67768288095471,5.75916689254577,5.97146442400446,-4.11177446033511,-6.43894433484766,4.5042232468252,2.21014640326349,3.4610384874124,-2.42109408164419,-.161112776439403},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-2.,-3.13195404884644,-3.24740583490634,-2.60051450395615,-1.39781931942298,2.449489743,3.83584465679206,3.97724364248012,3.18496680157705,-2.828427124,-4.42925189189755,-4.59252537503603,3.162277660,4.95205415939119,-3.464101616,3.40022577929605,5.32467544557844,5.5209565174066,4.42116822694256,2.37645064234965,-4.1644090826311,-6.52136894437417,-6.76176317987178,-5.41480311178231,4.80864541143983,7.53022822838542,7.80781158453304,-5.37622900930936,-8.41905110415162,5.88936380548032,5.17702934684997,8.10710900955867,8.40595765692318,6.73146995394289,-5.97791857375093,-9.36128313844683,-9.70636382862249,6.68351614602547,10.46623272812,-7.32142511479571,6.67435503772228,10.4518866380257,10.8371697481941,-7.46215578440284,-11.6855645116817,8.17438210188188,7.8090615264033,12.2288109301468,-8.55439830449838,8.51143945889045},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.40022577929605,-5.78076767221695,-7.18642267926874,-8.02365168049822,-8.3966606253625,4.1644090826311,7.0799655598807,8.80153432008931,9.82692624602171,-4.80864541143983,-8.17524004082037,-10.1631364184159,5.37622900930936,9.14019623375092,-5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375092,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,-6.67435503772231,-11.3471570254948,-14.1063386773585,7.46215578440288,12.6865072343198,-8.17438210188192,7.8090615264033,13.2762861543698,-8.55439830449838,-8.51143945889148},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-3.13195404884644,-5.32467544557844,-6.61942678751097,-7.39059991847735,-7.7341791249633,3.83584465679206,6.52136894437417,8.10710900955867,9.05159934712187,-4.42925189189755,-7.53022822838542,-9.36128313844683,4.95205415939119,8.41905110415162,-5.42470353822284,3.97724364248013,6.7617631798718,8.40595765692319,9.38526430991042,-4.59252537503604,-7.80781158453306,-9.70636382862251,5.13459946240545,8.72939873011817,-5.62467189980939,-3.67768288095471,-6.25247606762144,-7.77283197754114,4.11177446033512,6.99048075972417,-4.5042232468252,2.21014640326355,3.75749838742693,-2.42109408164425,.161112776439225},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.66702109783921,-4.53423694314245,-5.63678477479456,-6.29347864012493,-6.58605413075653,3.26642040979696,5.55328344227955,6.90362324406997,7.70790568809692,-3.77173740709444,-6.41237937831348,-7.97161747704438,4.21693061735794,7.16925809522303,-4.61941604516583,2.13132645129358,3.62349557083132,4.50458697357607,5.02937810056318,-2.46104380066281,-4.18405228663708,-5.2014490017318,2.75153061636385,4.67791266762517,-3.01415077313091,.104423971232213,.177532539414032,.220701460427408,-.116749549066047,-.198487413238653,.127892723210502,-2.94703586087065,-5.01029365229357,3.22831603852647,4.81780334972074},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-2.03452907455009,-3.45892910228435,-4.30000442088448,-4.80096137363441,-5.0241517123509,2.49177904847469,4.23630567890486,5.26640836144105,5.87995282040644,-2.87725860980962,-4.89166444644702,-6.08112457054741,3.21687291987534,5.46904711383883,-3.52390772571971,.0953780462518065,.162153445774676,.201582776983911,.2250674723256,-.110133081351697,-.187238671144273,-.232767741029294,.123132528211026,.209339198368165,-.134884926556002,2.8024998850757,4.76456617608066,5.92312100550577,-3.1332901246654,-5.32694692824228,3.43234736114769,-3.77550789188145,-6.41878963063904,4.13586167809416,1.3916398503358},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-1.2742056900953,-2.16629351671545,-2.69305077477823,-3.00679522191911,-3.14657715137113,1.56057688324205,2.65315687476095,3.2983001248078,3.68255702755716,-1.80199896789671,-3.06360167061167,-3.80854893001879,2.01469609360342,3.42521079645606,-2.20698899398908,-1.62700976129611,-2.76610026595298,-3.43870690112598,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865961,-2.10046056933792,-3.57102008849817,2.30093927043685,3.19783168117628,5.43667485806875,6.75866004623775,-3.57528450949468,-6.07838727908375,3.91652795089531,-.49602746920004,-.84330269383352,.543370868210304,-3.52996795992507},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,-.433821252801768,-.737545103324376,-.916887022287339,-1.02370573309381,-1.07129645743706,.531320354197217,.903304582796468,1.12295267819227,1.25377834648448,-.613515899259544,-1.04304628770802,-1.29667406213985,.68593162794344,1.16616120174253,-.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.3691862575809,-5.72799698473071,3.69075862820297,1.36899683098764,2.32744915738298,2.89339311993396,-1.53058498738027,-2.60216726588949,1.67667184761831,3.00982432351404,5.11704112693288,-3.29709735322896,-2.5031295519909},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,.433821252801768,.737545103324376,.916887022287339,1.02370573309381,1.07129645743706,-.531320354197217,-.903304582796468,-1.12295267819227,-1.25377834648448,.613515899259544,1.04304628770802,1.29667406213985,-.68593162794344,-1.16616120174253,.751400451068813,-2.60976045384023,-4.43688738516868,-5.51576363976884,-6.15835836236798,3.01349180095816,5.12327625260686,6.36905524216974,-3.36918625758091,-5.72799698473071,3.69075862820298,-1.36899683098765,-2.32744915738298,-2.89339311993396,1.53058498738027,2.60216726588949,-1.67667184761832,3.00982432351407,5.11704112693294,-3.297097353229,2.50312955199098},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,1.2742056900953,2.16629351671545,2.69305077477823,3.00679522191911,3.14657715137113,-1.56057688324205,-2.65315687476095,-3.2983001248078,-3.68255702755716,1.80199896789671,3.06360167061167,3.80854893001879,-2.01469609360342,-3.42521079645606,2.20698899398908,-1.62700976129612,-2.76610026595298,-3.43870690112599,-3.83932140376653,1.87870904723461,3.19401746644684,3.97067670865962,-2.10046056933793,-3.57102008849817,2.30093927043686,-3.1978316811763,-5.43667485806878,-6.75866004623778,3.5752845094947,6.07838727908378,-3.91652795089533,-.496027469200023,-.843302693833492,.543370868210286,3.52996795992515},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.03452907455009,3.45892910228435,4.30000442088448,4.80096137363441,5.0241517123509,-2.49177904847469,-4.23630567890486,-5.26640836144105,-5.87995282040644,2.87725860980962,4.89166444644702,6.08112457054741,-3.21687291987534,-5.46904711383883,3.52390772571971,.0953780462518019,.162153445774668,.201582776983901,.225067472325589,-.110133081351691,-.187238671144264,-.232767741029283,.12313252821102,.209339198368155,-.134884926555995,-2.8024998850757,-4.76456617608065,-5.92312100550575,3.13329012466539,5.32694692824227,-3.43234736114768,-3.77550789188148,-6.41878963063909,4.1358616780942,-1.39163985033612},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,2.66702109783921,4.53423694314245,5.63678477479456,6.29347864012493,6.58605413075653,-3.26642040979696,-5.55328344227955,-6.90362324406997,-7.70790568809692,3.77173740709444,6.41237937831348,7.97161747704438,-4.21693061735794,-7.16925809522303,4.61941604516583,2.13132645129358,3.62349557083131,4.50458697357606,5.02937810056317,-2.46104380066281,-4.18405228663707,-5.20144900173179,2.75153061636384,4.67791266762516,-3.0141507731309,-.104423971232209,-.177532539414025,-.2207014604274,.116749549066043,.198487413238645,-.127892723210497,-2.94703586087065,-5.01029365229357,3.22831603852648,-4.81780334971999},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.13195404884644,5.32467544557844,6.61942678751097,7.39059991847735,7.7341791249633,-3.83584465679206,-6.52136894437417,-8.10710900955867,-9.05159934712187,4.42925189189755,7.53022822838542,9.36128313844683,-4.95205415939119,-8.41905110415162,5.42470353822284,3.97724364248012,6.76176317987178,8.40595765692318,9.3852643099104,-4.59252537503603,-7.80781158453304,-9.70636382862249,5.13459946240544,8.72939873011815,-5.62467189980938,3.67768288095471,6.25247606762144,7.77283197754114,-4.11177446033511,-6.99048075972417,4.5042232468252,2.21014640326349,3.75749838742683,-2.42109408164419,-.161112776439403},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-2.,-3.40022577929605,-4.22702676016057,-4.71948170681967,-4.93888416262814,2.449489743,4.1644090826311,5.17702934684997,5.78016101583115,-2.828427124,-4.80864541143983,-5.97791857375093,3.162277660,5.37622900930936,-3.464101616,3.40022577929605,5.78076767221695,7.18642267926874,8.02365168049822,8.3966606253625,-4.1644090826311,-7.0799655598807,-8.80153432008931,-9.82692624602171,4.80864541143983,8.17524004082037,10.1631364184159,-5.37622900930936,-9.14019623375092,5.88936380548032,5.17702934684997,8.80153432008931,10.9417207971531,12.2164476526893,-5.97791857375093,-10.1631364184159,-12.6344108908165,6.68351614602547,11.3627319492038,-7.32142511479571,6.67435503772228,11.3471570254948,14.1063386773585,-7.46215578440284,-12.6865072343197,8.17438210188188,7.8090615264033,13.2762861543698,-8.55439830449838,8.51143945889045}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[144][56] = 
		{
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.94468190274016,-6.70640454840984,8.33713798160189,-9.30842703960244,9.74116348802451,7.0795491341205,-12.0360327285782,14.9627218217782,-16.7059013133841,10.545740291788,-17.9289490028036,22.2885632192208,14.1178337840267,-24.0019111852095,17.5926352342147,-.0319379183519753,.054298066631689,-.0675012177634823,.0753652106918314,-.07886883956794,-.154660030302817,.262939511058368,-.326876043550518,.364957592022403,-.43925822557731,.746788570738901,-.928378137184539,-.961797763577864,1.63516477461472,-1.79684767953686,.000465578153648587,-.000791535419924791,.000984005657525685,-.00109864378994577,.00319092434145473,-.0054249316016862,.00674406129057572,.0123137463285655,-.0209347588552154,.0353622090543082,-5.86464816610367e-006,9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,-1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.71235176911142,-6.31141709357397,7.8461051355817,-8.76018813180053,9.16743767928438,5.99558490979395,-10.1931711799913,12.6717489298481,-14.148026649931,7.62488741580146,-12.9631693790598,16.1153015818976,8.14599447640013,-13.8491102054891,7.34718311089616,-.166073783576779,.282344179950942,-.350999163642368,.391891091702269,-.410109589767539,-.764836159193308,1.30030781284792,-1.61649145672597,1.80481513164277,-2.02218309922343,3.43793955015111,-4.27391103771759,-4.02885967082133,6.84951625494143,-6.68010475273478,.0125887190510936,-.021402243516657,.0266064261611109,-.0297061146457811,.0823390061849571,-.139985605707144,.174024591276672,.299004079114613,-.508340689012668,.796298871799325,-.000824565926326729,.00140185515919078,-.00174273111826309,-.00697737063211093,.0118623177481531,-.0321260785529637,5.01144631645353e-005,-8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,3.30970802258292,-5.62687727016469,6.99511218988806,-7.81005323374722,8.17313226782894,4.2909622271481,-7.29512018698114,9.06900608163536,-10.1255588665535,3.64701232940425,-6.20033267056672,7.70800935885959,1.49376547749553,-2.53956994188053,-1.29952592474851,-.398540259080393,.677563431168936,-.842320170000572,.940451730878483,-.984172086870924,-1.67165845871656,2.84200809304711,-3.53307252095548,3.9446807593569,-3.84755914106515,6.54128488547504,-8.13186772631248,-6.3113399865174,10.7299904588883,-7.98777931164405,.0724976248699598,-.123254146474207,.153224700235957,-.171075607231601,.434860760346183,-.739312383715057,.919084035475437,1.40707187748205,-2.39218103581589,3.23180916098999,-.0113956405123911,.0193738753136992,-.0240848387004708,-.0887334889355571,.150856948291044,-.36808995603673,.00166206328995187,-.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.76195386285985,-4.69563336286119,5.83742644431873,-6.51749536542353,6.820485095594,2.32621648423849,-3.95483062658776,4.91648966501871,-5.48926807103564,.160883133644535,-.273519489264329,.340028655734571,-2.12140648579363,3.60663050994097,-2.7058281817046,-.714786270724957,1.21521735159799,-1.51071034695561,1.68671036412375,-1.76512329607374,-2.59853682213183,4.41780594598826,-5.49204234459095,6.13187350041214,-4.7549644652192,8.08397637227704,-10.0496810203089,-5.36731919542456,9.12504854432544,-3.25729414076559,.233201739437629,-.396469282990198,.492874996717867,-.550295671800539,1.22672710731592,-2.0855745668007,2.59270415505294,3.30495822058058,-5.61880207113837,5.89679444984459,-.0657431866024218,.111770838907705,-.138949104550336,-.451524561772794,.767642727504649,-1.59193561065057,.017197430813833,-.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,2.10349449699454,-3.57617810764411,4.44576376432321,-4.96370189948081,5.19445782869084,.504794589391635,-.858207787533309,1.06689011900532,-1.19118441501636,-1.67800190212113,2.8527926629451,-3.54647947347266,-1.79846700964276,3.0575969440733,.309494957156941,-1.09494796295235,1.86153514438243,-2.31418717000874,2.5837934399814,-2.70391057656363,-3.24473228828999,5.51641118734948,-6.85778510905627,7.65672734172539,-4.05017972204568,6.88576274670089,-8.56010903534823,-1.63294619261043,2.776192869346,2.28560008620039,.547225517982271,-.930345156393643,1.15656845454869,-1.29131041119439,2.39319484717323,-4.06870140628397,5.05804933070526,4.87179403320196,-8.28259983241363,5.49730994822615,-.236321146004056,.401772626268325,-.499467904136599,-1.36209257652082,2.31571114636377,-3.77635721863404,.0946962283839159,-.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,1.37570022553441,-2.33884568572274,2.90756083364644,-3.24629602423868,3.39721202820789,-.821946062819915,1.3974010951554,-1.73719400169233,1.93957970348913,-1.61733246050546,2.74964776323587,-3.41825379680027,.359308464240942,-.610864951290378,1.71165917287845,-1.51514018159912,2.57590935107797,-3.20226904625418,3.57533818439405,-3.74155092353066,-3.36446686788906,5.7199734894835,-7.11084574520705,7.93926992086969,-1.91964817870762,3.26361861024795,-4.05720211125474,1.95173675811295,-3.31817281869637,2.75500656898115,1.04781596872157,-1.78140543388112,2.21457307046673,-2.47257414898137,3.55510980002442,-6.04408799389592,7.51377213010339,4.59045028887451,-7.80428370621122,1.24891795372526,-.626153181688296,1.06453109468611,-1.32338312764,-2.84473124098146,4.83636425064191,-5.48872263996485,.347192252628353,-.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,.624299774465593,-1.06138009357331,1.31946592651414,-1.47318568258101,1.5416721344202,-1.43546196207946,2.44044738286343,-3.03386806383218,3.38731823500443,-.337640235075449,.574026515774929,-.713607154795179,1.50389998701775,-2.55679975214599,-.0107903422167659,-1.94896143440088,3.31345445440235,-4.11915606854152,4.59904391748787,-4.81284738096772,-2.83314651369184,4.81666890668704,-5.98789306701478,6.6854915743852,.637278791777431,-1.08344588754871,1.34689725343212,2.63766838605638,-4.48433402043889,-.808065843843811,1.73374759666501,-2.94756663562366,3.66429874441148,-4.09119503456583,4.12740556141283,-7.01705539432373,8.72332687917668,1.98940127224618,-3.38220674597772,-2.96832239843812,-1.33269841309056,2.26573774926528,-2.81667592802333,-4.37531622836174,7.4385315165314,-4.52786743647192,.95054294779362,-1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.103494496994541,.175952328348057,-.21873700416264,.24422019266112,-.255573666062748,-1.29720437850508,2.20539388307836,-2.74165881101347,3.06106616679169,.991754894783212,-1.68609528016685,2.09608724079431,.589436063368126,-1.00210784875001,-1.41318416081787,-2.36915365304765,4.02782866109789,-5.00723794478697,5.59058866190052,-5.85048772793476,-1.68415540504794,2.86325431258853,-3.55948498424847,3.97417031416821,2.32831086437971,-3.95839130925952,4.9209161654707,.381749900992996,-.649017927110068,-2.46800676549626,2.56192161158569,-4.3555559528497,5.41465560654381,-6.04547109188663,3.58714638367868,-6.09855380269565,7.58148187830962,-1.52344991306454,2.59003683420207,-2.69130383032704,-2.39387697960251,4.06986110773887,-5.05949102727521,-4.93737708601551,8.39409842544753,-.508166439915618,2.07554270972356,-3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-.761953862859853,1.29540758356514,-1.61039968415816,1.79801365860384,-1.88160093296591,-.55104212557113,.936833819859258,-1.1646349055287,1.3003166154948,1.37646295656649,-2.34014241482376,2.90917287715098,-1.03415883911242,1.75818677195786,-.174795669385988,-2.74931534527504,4.67414645388234,-5.81071476784009,6.48767173775818,-6.78927500842464,-.106757773657142,.181500267083405,-.225633983149896,.251920680005671,2.26282036321899,-3.84705006415361,4.78250111498198,-2.15044627554921,3.65600143048661,.119501749587716,3.45007465931297,-5.86551639682903,7.29177895707793,-8.1412821233112,1.77035093065653,-3.00979643579225,3.74166037942858,-3.42380239520456,5.82085058429709,1.33170919606025,-3.74106981000851,6.36024100270089,-7.90680110023493,-3.58481468643821,6.09458965574696,3.59724489879787,3.76406389114815,-6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.30970802258292,2.22665149086864,-2.76808542972748,3.09057152692753,-3.23424810520086,.519224820053661,-.882740808667662,1.09738860459098,-1.22523601982935,.518731848959601,-.881902702778381,1.09634670392581,-1.30425364587756,2.21737843427209,1.47252576661999,-3.06556135691961,5.21180037431139,-6.47910494479513,7.23393037100344,-7.57022621762746,1.5947619510804,-2.71127534923244,3.37055072311448,-3.76322492874006,.444911232504622,-.756399320675517,.940325842964601,-2.09440936915946,3.56073236366527,2.63186985254377,4.2894282422279,-7.29251224169723,9.06576398575723,-10.1219390639593,-.964148801340396,1.63916180436132,-2.03774139204767,-2.01722273728898,3.42950637731078,3.46190525493913,-5.18623489409549,8.81718478891181,-10.9611768422514,-.205483038001613,.349344361529711,3.63947886955733,5.81833621896758,-9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.71235176911142,2.91119131427792,-3.61907837542113,4.04070642498084,-4.22855351665629,1.5663330178964,-2.66294295160588,3.31046579140125,-3.69614001190532,-1.07607439139602,1.82944794322174,-2.27429762516448,.363617646552178,-.61819104768005,.420994417051474,-3.29802783242322,5.60701962552938,-6.97042595115334,7.78249101017965,-8.14428871473084,3.07100849759875,-5.22106113152625,6.4906175528327,-7.24678421547914,-2.20598407348367,3.75042195543293,-4.66237685613718,.923194488569861,-1.5695348492102,.507379758856801,4.96464287844228,-8.44045334766828,10.4928391542811,-11.7152706263276,-3.80884940873902,6.47547397319738,-8.05005418818497,2.02329330850018,-3.43982703367553,.0162698697481229,-6.45780747384476,10.9790017210618,-13.6486625036235,4.10479708970298,-6.97861844197597,-1.32544341327556,7.79428043908953,-13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.40432270571992,2.98895928687546,-3.33717751886116,3.49231848196783,-3.47478060833048,-1.94468190274016,3.30617876911379,-4.11011122144132,4.58894533278275,-4.80227932539642,2.27090372268066,-3.8607926877578,4.79958540336237,-5.3587442878893,-2.44801058376252,4.16189434787989,-5.17390312560614,2.49586968540814,-4.24326022217388,-2.4264683064468,-3.43216369764802,5.83506573884863,-7.25392389703222,8.09901689119009,-8.47552946493044,4.00974905232829,-6.81702604882233,8.4746582765388,-9.46196865399936,-4.32546942586304,7.35378632026677,-9.14193750783762,4.4144312457315,-7.50503145913621,-4.2977488551218,5.37669458746301,-9.14098776917085,11.3637159548613,-12.6876058781097,-5.80779112496784,9.87390054991614,-12.2748442514899,5.9385809984625,-10.0962581027492,-5.79724384742923,-7.27823584467216,12.3738225688191,-15.3826488429769,7.46210387346587,-12.6864189798666,-7.31197439196322,9.14177937313726,-15.5420569434015,-8.9987004064276,-10.9691712594614},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.94468190274016,-6.17728122834917,6.40499151390394,-5.12910125078457,2.75697628631382,7.0795491341205,-11.0864112806424,11.4950845847679,-9.20523510120587,10.545740291788,-16.5143870041822,17.1231492859125,14.1178337840267,-22.1082033358775,17.5926352342147,-.0319379183519753,.0500140463222003,-.0518576912013895,.0415275099413171,-.0223217196472694,-.154660030302817,.24219405407586,-.251121942525088,.201097826062499,-.43925822557731,.687868288629403,-.713224862474738,-.961797763577864,1.50615319946369,-1.79684767953686,.000465578153648587,-.000729084691473652,.000755960606624687,-.000605371370823812,.00319092434145473,-.00499691420436607,.00518111316265945,.0123137463285655,-.0192830438371037,.0353622090543082,-5.86464816610367e-006,9.18390428102533e-006,-9.52244633830418e-006,-5.19109369939388e-005,8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,-1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.71235176911142,-5.81345757700538,6.02775639811867,-4.82701230968077,2.59459851167866,5.99558490979395,-9.38894821107085,9.73504871109304,-7.79580275781954,7.62488741580146,-11.9403985081938,12.380551947663,8.14599447640013,-12.7564401885048,7.34718311089616,-.166073783576779,.260067729310804,-.269654486884814,.215938641412748,-.11607057156663,-.764836159193308,1.1977158528687,-1.24186670359493,.994483762912882,-2.02218309922343,3.16669227065688,-3.28342459824587,-4.02885967082133,6.30910167728468,-6.68010475273478,.0125887190510936,-.0197136447951608,.0204403398568299,-.0163685732441572,.0823390061849571,-.128940991873039,.13369408456444,.299004079114613,-.468233518150735,.796298871799325,-.000824565926326729,.00129125129526923,-.00133885010037302,-.00697737063211093,.0109264021013695,-.0321260785529637,5.01144631645353e-005,-7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,3.30970802258292,-5.18292672091406,5.37398257218606,-4.30347185829346,2.3131869078075,4.2909622271481,-6.71954825634174,6.96724788777658,-5.57935475310453,3.64701232940425,-5.71113751622596,5.92166456181189,1.49376547749553,-2.33920241715482,-1.29952592474851,-.398540259080393,.624104888716883,-.647110981340288,.518204861963161,-.278543636855169,-1.67165845871656,2.61777873930361,-2.71427671758158,2.17358603453451,-3.84755914106515,6.02518921139496,-6.24729300320555,-6.3113399865174,9.88341340931671,-7.98777931164405,.0724976248699598,-.113529614838378,.11771460504762,-.0942655625163293,.434860760346183,-.680981959385669,.70608468526914,1.40707187748205,-2.20344223207683,3.23180916098999,-.0113956405123911,.0178453112143479,-.0185031347484812,-.0887334889355571,.138954604977606,-.36808995603673,.00166206328995187,-.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.76195386285985,-4.32515629175549,4.48459254499661,-3.59125053981239,1.93035623442995,2.32621648423849,-3.64280156596183,3.77708449255997,-3.0246798529178,.160883133644535,-.251939290930609,.261226413581035,-2.12140648579363,3.32207381553389,-2.7058281817046,-.714786270724957,1.11933887677121,-1.16060055303996,.929406031924799,-.499571029238757,-2.59853682213183,4.06924896099705,-4.21925182103647,3.37876634868698,-4.7549644652192,7.44616510000515,-7.72064967553202,-5.36731919542456,8.40509854032022,-3.25729414076559,.233201739437629,-.365188565907959,.378650344801756,-.303222252967675,1.22672710731592,-1.9210264648999,1.99184038309353,3.30495822058058,-5.17548864064312,5.89679444984459,-.0657431866024218,.102952319693436,-.106747403902083,-.451524561772794,.707077089737688,-1.59193561065057,.017197430813833,-.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,2.10349449699454,-3.29402405329413,3.41545015161673,-2.73508397421313,1.47015262309925,.504794589391635,-.790496728565174,.819636447612693,-.656362825532386,-1.67800190212113,2.62771242593193,-2.72457658514876,-1.79846700964276,2.81635801570608,.309494957156941,-1.09494796295235,1.7146633520688,-1.77787020176478,1.42371402906152,-.765269708188713,-3.24473228828999,5.08117621439183,-5.26848128513793,4.21898669004741,-4.05017972204568,6.34248838569497,-6.576288631721,-1.63294619261043,2.55715621899877,2.28560008620039,.547225517982271,-.856942588087486,.888531670338268,-.71153394843509,2.39319484717323,-3.74768814487268,3.88583745543936,4.87179403320196,-7.62911752450546,5.49730994822615,-.236321146004056,.370073484889954,-.383715334269619,-1.36209257652082,2.13300568008586,-3.77635721863404,.0946962283839159,-.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,1.37570022553441,-2.15431494568072,2.23372846974121,-1.78876419479899,.961490176493142,-.821946062819915,1.28714864891722,-1.33459622035742,1.06874132878065,-1.61733246050546,2.53270547426737,-2.62606743575437,.359308464240942,-.562668799566708,1.71165917287845,-1.51514018159912,2.37267471198348,-2.46013753301879,1.97007200846434,-1.0589461087365,-3.36446686788906,5.26867781509235,-5.46289467143737,4.37467244553793,-1.91964817870762,3.0061249410246,-3.11693834864261,1.95173675811295,-3.05637492003247,2.75500656898115,1.04781596872157,-1.64085573236143,1.70134184591416,-1.36243031247303,3.55510980002442,-5.56722026499827,5.77244215424072,4.59045028887451,-7.1885396848779,1.24891795372526,-.626153181688296,.980541495928355,-1.01668674800428,-2.84473124098146,4.45478376427993,-5.48872263996485,.347192252628353,-.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,.624299774465593,-.977639103165718,1.01367736516514,-.811750309157165,.436329142929648,-1.43546196207946,2.24790045069837,-2.33076377601413,1.86646982589538,-.337640235075449,.528736850703693,-.548227434979559,1.50389998701775,-2.35507282621716,-.0107903422167659,-1.94896143440088,3.05202882623936,-3.1645343667906,2.53415123836086,-1.36214797290775,-2.83314651369184,4.43664234772504,-4.60018826185061,3.68381930162858,.637278791777431,-.997963945475626,1.03475143356998,2.63766838605638,-4.13052808940267,-.808065843843811,1.73374759666501,-2.71500890173164,2.81509103173736,-2.25431788633233,4.12740556141283,-6.4634222783235,6.70168045166519,1.98940127224618,-3.11535668501805,-2.96832239843812,-1.33269841309056,2.08697509460821,-2.16390630168515,-4.37531622836174,6.85164468857583,-4.52786743647192,.95054294779362,-1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.103494496994541,.162070004447689,-.168044316710384,.134569470256975,-.0723333036764564,-1.29720437850508,2.0313922514988,-2.1062745341762,1.68669940023467,.991754894783212,-1.55306537925041,1.61031531675698,.589436063368126,-.923043332411645,-1.41318416081787,-2.36915365304765,3.71004018615404,-3.84680169804461,3.08050921776368,-1.65582437345554,-1.68415540504794,2.63734867013603,-2.73456804580791,2.18983527963514,2.32831086437971,-3.64608131714141,3.78048514370595,.381749900992996,-.597811573855878,-2.46800676549626,2.56192161158569,-4.01191038094212,4.15979959635524,-3.33115715545234,3.58714638367868,-5.61738881893157,5.82446004858606,-1.52344991306454,2.38568756196537,-2.69130383032704,-2.39387697960251,3.74875634795768,-3.88694503628023,-4.93737708601551,7.73181907803745,-.508166439915618,2.07554270972356,-3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-.761953862859853,1.19320224290905,-1.23718671009026,.990736035856232,-.532536915007162,-.55104212557113,.8629193076149,-.894728707041788,.716496519828253,1.37646295656649,-2.15550936517572,2.2349669193134,-1.03415883911242,1.61946898132209,-.174795669385988,-2.74931534527504,4.30536466145163,-4.46407134676943,3.5748172149004,-1.9215230524055,-.106757773657142,.16718022074295,-.173342908623153,.13881256945113,2.26282036321899,-3.54352469706742,3.67414802589247,-2.15044627554921,3.3675494587808,.119501749587716,3.45007465931297,-5.40273764744738,5.60189629155854,-4.48598459696892,1.77035093065653,-2.7723288820058,2.87452397106015,-3.42380239520456,5.36159588761007,1.33170919606025,-3.74106981000851,5.85842936705581,-6.07438596565012,-3.58481468643821,5.61373743608465,3.59724489879787,3.76406389114815,-5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.30970802258292,2.05097267206762,-2.12657673727971,1.7029573543373,-.91536758838471,.519224820053661,-.813094138225446,.843066855241088,-.675125837596018,.518731848959601,-.812322157391327,.842266416897198,-1.30425364587756,2.04243124304558,1.47252576661999,-3.06556135691961,4.80059864950596,-4.9775609184691,3.98601838486204,-2.14255044478909,1.5947619510804,-2.49736057507455,2.5894196337366,-2.07360079279396,.444911232504622,-.696720767591157,.722403666316151,-2.09440936915946,3.27979695088037,2.63186985254377,4.2894282422279,-6.71714607327471,6.96475715335039,-5.57736018045175,-.964148801340396,1.50983487071459,-1.56549122161564,-2.01722273728898,3.15892446006539,3.46190525493913,-5.18623489409549,8.12152468439221,-8.42090472917075,-.205483038001613,.321781716436834,3.63947886955733,5.81833621896758,-9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.71235176911142,2.68150352815894,-2.78035056321232,2.22649780572461,-1.19677919225587,1.5663330178964,-2.45284151714649,2.54325949118096,-2.03663586527377,-1.07607439139602,1.68510777367065,-1.74722512946481,.363617646552178,-.56941688005877,.420994417051474,-3.29802783242322,5.16463580891203,-5.35501741292457,4.28828460541245,-2.30502351007762,3.07100849759875,-4.80912874954551,4.98640545923911,-3.99310107128126,-2.20598407348367,3.45452037324197,-3.58186277642043,.923194488569861,-1.44570135775136,.507379758856801,4.96464287844228,-7.7745166798312,8.0611051284312,-6.45531290809182,-3.80884940873902,5.96457066234997,-6.18443989718937,2.02329330850018,-3.16843083510823,.0162698697481229,-6.45780747384476,10.1127781284249,-10.4855608369157,4.10479708970298,-6.42801793274624,-1.32544341327556,7.79428043908953,-12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,-2.21462594594878,2.29626268751802,-1.83884144047735,.98840751937666,.0657740155503869,-1.94468190274016,3.04532717950273,-3.15758567899759,2.52858674682842,-1.35915696689103,2.27090372268066,-3.55618305225695,3.68727300023486,-2.95275903358443,-2.44801058376252,3.83352833011347,-3.97484192853687,2.49586968540814,-3.90847458250196,-2.4264683064468,-3.43216369764802,5.37468949190064,-5.572814208608,4.46269573688388,-2.39877236199698,4.00974905232829,-6.27917489029831,6.51064123734671,-5.21370528571982,-4.32546942586304,6.77358573667404,-7.02327732701019,4.4144312457315,-6.91289790468793,-4.2977488551218,5.37669458746301,-8.4197801888431,8.73015469072479,-6.99108613109499,-5.80779112496784,9.09486746248342,-9.43012739369097,5.9385809984625,-9.29968140223042,-5.79724384742923,-7.27823584467216,11.3975501068473,-11.8176927763526,7.46210387346587,-11.6854832203471,-7.31197439196322,9.14177937313726,-14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.94468190274016,-5.26027492943627,3.43230866927273,.14563548266565,-3.67618541981627,7.0795491341205,-9.44065344626821,6.15998918791292,.261373053770353,10.545740291788,-14.0628559267203,9.17595809638574,14.1178337840267,-18.8262802750321,17.5926352342147,-.0319379183519753,.0425895510116904,-.0277895142706456,-.00117913035023069,.029764050100784,-.154660030302817,.206240781926363,-.134571297792104,-.00570996311627599,-.43925822557731,.585755477154869,-.382203141612095,-.961797763577864,1.28256746328293,-1.79684767953686,.000465578153648587,-.000620853379055167,.000405104384309337,1.71888889439268e-005,.00319092434145473,-.00425513126926153,.00277645639124976,.0123137463285655,-.0164205106275604,.0353622090543082,-5.86464816610367e-006,7.82057019228999e-006,-5.10289124485107e-006,-5.19109369939388e-005,6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,-9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.71235176911142,-4.95046024541044,3.23015580842145,.137057982126179,-3.45966893740063,5.99558490979395,-7.99517571934639,5.21682066468669,.2213536917852,7.62488741580146,-10.1678678043472,6.63449368950757,8.14599447640013,-10.8627695634925,7.34718311089616,-.166073783576779,.221461142188379,-.144502522920426,-.00613135259583728,.154769899538562,-.764836159193308,1.01991708658492,-.665491881382992,-.0282373296477263,-2.02218309922343,2.69660249304008,-1.75952250506474,-4.02885967082133,5.37252686958436,-6.68010475273478,.0125887190510936,-.0167871896471046,.0109535751204635,.000464768572193078,.0823390061849571,-.109799933312678,.0716440239592628,.299004079114613,-.398725093710568,.796298871799325,-.000824565926326729,.00109956736062704,-.000717463371558171,-.00697737063211093,.00930439734215142,-.0321260785529637,5.01144631645353e-005,-6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,3.30970802258292,-4.41353056195817,2.87981130513508,.122192596826733,-3.08443131194341,4.2909622271481,-5.72204339147722,3.73361077438991,.158419961451721,3.64701232940425,-4.86332941380344,3.17330328508799,1.49376547749553,-1.99195202144368,-1.29952592474851,-.398540259080393,.531457639388288,-.346774016236295,-.0147138868003722,.371413443660288,-1.67165845871656,2.22917418911978,-1.45452737733324,-.0617167093760001,-3.84755914106515,5.13076069911761,-3.34780114662786,-6.3113399865174,8.41623844737545,-7.98777931164405,.0724976248699598,-.0966763475074085,.0630809359431452,.0026765723709811,.434860760346183,-.579891411113903,.378376861425819,1.40707187748205,-1.87634519190449,3.23180916098999,-.0113956405123911,.0151962068293123,-.00991546509750554,-.0887334889355571,.118327043544488,-.36808995603673,.00166206328995187,-.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.76195386285985,-3.68309461175287,2.40320472508569,.101969815015613,-2.57396027644144,2.32621648423849,-3.10203421893763,2.02406510925841,.0858826310401591,.160883133644535,-.214539355880419,.139986084607739,-2.12140648579363,2.82891792677208,-2.7058281817046,-.714786270724957,.953175031760129,-.621942953572119,-.0263895153231826,.666134033494214,-2.59853682213183,3.46517626442722,-2.26101386266496,-.0959365479342438,-4.7549644652192,6.34079527029531,-4.13734393811778,-5.36731919542456,7.15737676442222,-3.25729414076559,.233201739437629,-.31097697947545,.202911254178075,.00860967974831034,1.22672710731592,-1.63585353791591,1.06738798948579,3.30495822058058,-4.4071966513386,5.89679444984459,-.0657431866024218,.0876692328212839,-.0572038290823836,-.451524561772794,.602112766253303,-1.59193561065057,.017197430813833,-.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,2.10349449699454,-2.80503210133656,1.8302723960547,.0776598579828548,-1.96031923262139,.504794589391635,-.673148909586561,.439227012039901,.0186367381391773,-1.67800190212113,2.23763323781713,-1.46004687328675,-1.79846700964276,2.39827472875065,.309494957156941,-1.09494796295235,1.46012465838812,-.952725588021822,-.0404248755605117,1.02041985541771,-3.24473228828999,4.32688473530227,-2.82327524550586,-.119793728554495,-4.05017972204568,5.4009573811209,-3.52410341735766,-1.63294619261043,2.17755097302669,2.28560008620039,.547225517982271,-.729731000653746,.476146603536492,.0202032646552823,2.39319484717323,-3.19135057367101,2.08234367785444,4.87179403320196,-6.49658873611033,5.49730994822615,-.236321146004056,.315136741011867,-.205625482170818,-1.36209257652082,1.81636481949508,-3.77635721863404,.0946962283839159,-.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,1.37570022553441,-1.83451076290121,1.19701104597108,.0507900944331592,-1.28206259359835,-.821946062819915,1.09607374475429,-.715183801128781,-.0303457958132313,-1.61733246050546,2.15672989741719,-1.40725776236676,.359308464240942,-.479141627283131,1.71165917287845,-1.51514018159912,2.02045541424857,-1.31833919902212,-.0559381407795183,1.41201150863478,-3.36446686788906,4.4865520602846,-2.92745754610215,-.124214263271286,-1.91964817870762,2.55987109498181,-1.67030580640284,1.95173675811295,-2.60266155489591,2.75500656898115,1.04781596872157,-1.39727364720762,.911715550981014,.0386847883193904,3.55510980002442,-4.74077641992773,3.09333793899121,4.59045028887451,-6.12141388513833,1.24891795372526,-.626153181688296,.834981872710088,-.544822382907866,-2.84473124098146,3.79347911889772,-5.48872263996485,.347192252628353,-.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,.624299774465593,-.832510334938,.543209713978403,.023048803737304,-.581806539810157,-1.43546196207946,1.91420366785489,-1.24901035342108,-.0529964648160466,-.337640235075449,.450246815259374,-.293784273363787,1.50389998701775,-2.00546649679692,-.0107903422167659,-1.94896143440088,2.59896063091726,-1.69581157410879,-.0719545824309833,1.81630452989169,-2.83314651369184,3.77803076303363,-2.46515019062261,-.104598208499715,.637278791777431,-.849817990926997,.55450289173293,2.63766838605638,-3.51735861632833,-.808065843843811,1.73374759666501,-2.31197070864005,1.50855178059943,.0640090061407145,4.12740556141283,-5.50393885468463,3.59129842139725,1.98940127224618,-2.65288758284874,-2.96832239843812,-1.33269841309056,1.77716739172311,-1.15959472274786,-4.37531622836174,5.8345303456991,-4.52786743647192,.95054294779362,-1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.103494496994541,.138011003497348,-.0900516361052146,-.0038209598123916,.0964500992128821,-1.29720437850508,1.72983572180117,-1.12871099482768,-.0478920709992736,.991754894783212,-1.32251561427285,.862936228716219,.589436063368126,-.786019208253834,-1.41318416081787,-2.36915365304765,3.15929138677771,-2.06142518510909,-.0874678476499899,2.20789618310877,-1.68415540504794,2.24583899888368,-1.46540110005889,-.0621780897499632,2.32831086437971,-3.10482709694771,2.02588745115913,.381749900992996,-.50906751987416,-2.46800676549626,2.56192161158569,-3.41634949355459,2.22915458763892,.094584734527365,3.58714638367868,-4.78349754217328,3.12121330296808,-1.52344991306454,2.03153653003234,-2.69130383032704,-2.39387697960251,3.19226020392741,-2.08293720858962,-4.93737708601551,6.58404442855625,-.508166439915618,2.07554270972356,-2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-.761953862859853,1.01607351391366,-.662983965136209,-.0281309168451499,.710091143032931,-.55104212557113,.734820486906378,-.47946747332299,-.0203441716962566,1.37646295656649,-1.83552787296833,1.19767470667912,-1.03415883911242,1.37906172093203,-.174795669385988,-2.74931534527504,3.6662410134057,-2.39220781955879,-.101503207887319,2.56218200503226,-.106757773657142,.142362617365695,-.0928910470423959,-.0039414381903617,2.26282036321899,-3.01749482284837,1.96890348630228,-2.15044627554921,2.86764279249033,.119501749587716,3.45007465931297,-4.60071095140733,3.00194577372105,.127374855762558,1.77035093065653,-2.36078164083594,1.54040072098213,-3.42380239520456,4.56567661189371,1.33170919606025,-3.74106981000851,4.9887560540339,-3.25514367429686,-3.58481468643821,4.78038820054912,3.59724489879787,3.76406389114815,-5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.30970802258292,1.74650946411896,-1.13959054518559,-.0483536986562693,1.2205621785349,.519224820053661,-.692391774386145,.451782905525936,.0191694943053244,.518731848959601,-.691734392719743,.451353966399165,-1.30425364587756,1.73923599488787,1.47252576661999,-3.06556135691961,4.08795840577754,-2.66737675689462,-.113178836410129,2.85690259486619,1.5947619510804,-2.12663188500128,1.38761892782618,.0588777326765052,.444911232504622,-.593293821521023,.387121881618236,-2.09440936915946,2.79291698671264,2.63186985254377,4.2894282422279,-5.71999780817022,3.7322760389191,.158363327640689,-.964148801340396,1.28570259705192,-.838915879897341,-2.01722273728898,2.68998779997356,3.46190525493913,-5.18623489409549,6.91589893787678,-4.51259681489189,-.205483038001613,.274013798814207,3.63947886955733,5.81833621896758,-7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.71235176911142,2.28343914757123,-1.48993504847196,-.0632190839557153,1.59579980399212,1.5663330178964,-2.0887216012301,1.36288261753995,.0578281520973969,-1.07607439139602,1.43495655249725,-.936303497985116,.363617646552178,-.484887967351202,.420994417051474,-3.29802783242322,4.39795490297745,-2.86964825021048,-.121761370614664,3.07354613898791,3.07100849759875,-4.09522222779324,2.67211637191124,.11337994190694,-2.20598407348367,2.94170303097044,-1.91944964063812,.923194488569861,-1.23108958885198,.507379758856801,4.96464287844228,-6.62040364808344,4.31978730278896,.183291880031408,-3.80884940873902,5.07914086475798,-3.31411940634737,2.02329330850018,-2.6980829707225,.0162698697481229,-6.45780747384476,8.61155438605813,-5.61900531555838,4.10479708970298,-5.47379022059324,-1.32544341327556,7.79428043908953,-10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.88586870354722,1.2305219003314,.0522119856161066,-1.31795450312975,1.96685998123043,-1.94468190274016,2.59325383159706,-1.69208790932325,-.0717965844951867,1.81231628640776,2.27090372268066,-3.02827406795474,1.97593690127584,.0838405143563207,-2.44801058376252,3.26444793765179,-2.13003942014441,2.49586968540814,-3.32826855353781,-2.4264683064468,-3.43216369764802,4.57682649415414,-2.98636125886026,-.126713592860271,3.19855198842569,4.00974905232829,-5.34704266035319,3.48892427303922,.148037726033603,-4.32546942586304,5.76805910494959,-3.7636358461794,4.4144312457315,-5.8866906319401,-4.2977488551218,5.37669458746301,-7.16987894860209,4.67831777200948,.198504602127597,-5.80779112496784,7.74475072947788,-5.0534193426251,5.9385809984625,-7.91916040788233,-5.79724384742923,-7.27823584467216,9.70560427278213,-6.33286855712843,7.46210387346587,-9.95079423294563,-7.31197439196322,9.14177937313726,-12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.94468190274016,-4.01278501048821,.153597725402263,3.90852233081443,-4.7096362989066,7.0795491341205,-7.20177426970699,.275662948451089,7.014653288031,10.545740291788,-10.727807619258,.410629237576691,14.1178337840267,-14.3615716486884,17.5926352342147,-.0319379183519753,.0324893117176766,-.00124359624768518,-.0316451541922531,.0381313330936754,-.154660030302817,.157330164177211,-.00602214055814436,-.153242313889452,-.43925822557731,.446841815317497,-.0171038035473819,-.961797763577864,.978402756631806,-1.79684767953686,.000465578153648587,-.000473616144898061,1.81286469184553e-005,.000461310355481612,.00319092434145473,-.00324601417302388,.000124247970525594,.0123137463285655,-.012526337462346,.0353622090543082,-5.86464816610367e-006,5.96589860075151e-006,-2.28357226493363e-007,-5.19109369939388e-005,5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,-6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.71235176911142,-3.77644380460732,.144551272241366,3.67832188935975,-4.43225260672405,5.99558490979395,-6.09909590528773,.233455631511481,5.94062539921619,7.62488741580146,-7.7565275696118,.2968972892872,8.14599447640013,-8.28663129998069,7.34718311089616,-.166073783576779,.168940970519646,-.00646656841622833,-.164551441038204,.198278882483944,-.764836159193308,.778040701653472,-.0297811324981325,-.757825163693696,-2.02218309922343,2.05709515348012,-.0787396124921655,-4.02885967082133,4.09841606758444,-6.68010475273478,.0125887190510936,-.0128060574566476,.000490178589876797,.0124733225023729,.0823390061849571,-.0837605510092484,.00320610999159924,.299004079114613,-.304166246215336,.796298871799325,-.000824565926326729,.000838801675185275,-3.21068856415306e-005,-.00697737063211093,.00709783170785957,-.0321260785529637,5.01144631645353e-005,-5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,3.30970802258292,-3.36684860010831,.128873160510414,3.27936904259757,-3.95152801322486,4.2909622271481,-4.36504370184005,.167081162489149,4.25162841944462,3.64701232940425,-3.7099763100915,.142007090144466,1.49376547749553,-1.51955464695027,-1.29952592474851,-.398540259080393,.405420872037066,-.0155183304460168,-.394886974518159,.47582535601586,-1.67165845871656,1.70051886866409,-.0650909105855692,-1.6563349285852,-3.84755914106515,3.91398546692074,-.149815966677054,-6.3113399865174,6.42030234909073,-7.98777931164405,.0724976248699598,-.0737492627952912,.00282290703160039,.0718330635538727,.434860760346183,-.442368430061905,.0169325753770763,1.40707187748205,-1.43136432250754,3.23180916098999,-.0113956405123911,.0115923809684749,-.000443722587923622,-.0887334889355571,.0902654315677738,-.36808995603673,.00166206328995187,-.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.76195386285985,-2.80963771827715,.107544750492195,2.73663596098029,-3.29755313334498,2.32621648423849,-2.36637753401768,.0905780414306531,2.30489284002443,.160883133644535,-.163660706519194,.00626445528521578,-2.12140648579363,2.15803158670053,-2.7058281817046,-.714786270724957,.727126724577601,-.027832293713515,-.70823406529868,.853397929051741,-2.59853682213183,2.6433993582315,-.101181630171365,-2.57471691018525,-4.7549644652192,4.83705672356223,-.185148446522551,-5.36731919542456,5.45998347614275,-3.25729414076559,.233201739437629,-.237227859491313,.00908039168484127,.231064057614915,1.22672710731592,-1.24790598293055,.0477662072687217,3.30495822058058,-3.36201679531997,5.89679444984459,-.0657431866024218,.0668782122732034,-.00255990322490887,-.451524561772794,.459319924425276,-1.59193561065057,.017197430813833,-.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,2.10349449699454,-2.13981035614575,.0819057095351855,2.08421247060184,-2.51140504655499,.504794589391635,-.513509634087683,.0196556535222271,.500167307151007,-1.67800190212113,1.70697182868444,-.0653379111049031,-1.79846700964276,1.82951670999345,.309494957156941,-1.09494796295235,1.11385173231848,-.042635000914324,-1.08491094310714,1.30728073869583,-3.24473228828999,3.30075109017009,-.126343140340609,-3.21498892012256,-4.05017972204568,4.12010419835037,-.157705591526293,-1.63294619261043,1.66113825253464,2.28560008620039,.547225517982271,-.55667311317238,.0213078258129724,.542209285918406,2.39319484717323,-2.43451224831938,.0931860398476328,4.87179403320196,-4.95590330339708,5.49730994822615,-.236321146004056,.240401121148633,-.00920185489986058,-1.36209257652082,1.38560847465611,-3.77635721863404,.0946962283839159,-.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,1.37570022553441,-1.39945105335743,.0535669112712696,1.36308964438237,-1.64247659972029,-.821946062819915,.836136580756822,-.0320048735946587,-.814411559679094,-1.61733246050546,1.64525495722615,-.0629755689776038,.359308464240942,-.365511758540104,1.71165917287845,-1.51514018159912,1.54129837497396,-.0589964137232872,-1.50125140095791,1.80895681150723,-3.36446686788906,3.42255283189436,-.131005356341758,-3.33362593315312,-1.91964817870762,1.95279001506978,-.074747115516283,1.95173675811295,-1.98543258954328,2.75500656898115,1.04781596872157,-1.06590602625894,.0407997789063147,1.0382109925525,3.55510980002442,-3.61648712494209,.138428596350645,4.59045028887451,-4.66970228947898,1.24891795372526,-.626153181688296,.636963426396325,-.0243811052053837,-2.84473124098146,2.89384420968737,-5.48872263996485,.347192252628353,-.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,.624299774465593,-.635078021192655,.0243089373722252,.618577028461073,-.745364252864075,-1.43546196207946,1.46024454775272,-.0558939092532152,-1.42230356501297,-.337640235075449,.343469437534987,-.0131470099270517,1.50389998701775,-1.52986412408779,-.0107903422167659,-1.94896143440088,1.98260935074574,-.0758885128327144,-1.93109596018978,2.32690486658693,-2.83314651369184,2.88205947758122,-.110316844590442,-2.80717601371158,.637278791777431,-.648281114842891,.0248143133692,2.63766838605638,-2.68320650944122,-.808065843843811,1.73374759666501,-1.76367994615688,.0675085327336552,1.7178549162266,4.12740556141283,-4.19866330771593,.160712605397532,1.98940127224618,-2.0237473648752,-2.96832239843812,-1.33269841309056,1.35570683401508,-.051892509958983,-4.37531622836174,4.45085403872017,-4.52786743647192,.95054294779362,-.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.103494496994541,.105281281595665,-.00402986089169074,-.102545797758389,.123564193970617,-1.29720437850508,1.31960001105773,-.0505104459264124,-1.28531334221177,.991754894783212,-1.00887708423625,.0386168770555543,.589436063368126,-.599612404132423,-1.41318416081787,-2.36915365304765,2.41005599340123,-.0922499256416775,-2.34743641804055,2.82858093939833,-1.68415540504794,1.71323156899255,-.0655775157362847,-1.66871731961812,2.32831086437971,-2.36850807266163,.090659592246103,.381749900992996,-.388340636274758,-2.46800676549626,2.56192161158569,-2.60615200198177,.0997559098622405,2.53843733881177,3.58714638367868,-3.64907680538234,.139676034420518,-1.52344991306454,1.54975157103559,-2.69130383032704,-2.39387697960251,2.4352061570422,-.0932126006787421,-4.93737708601551,5.02261861703307,-.508166439915618,2.07554270972356,-2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-.761953862859853,.775108643727059,-.0296689018487,-.754969288136848,.909712280760612,-.55104212557113,.560555612551141,-.0214564365862776,-.545990907719199,1.37646295656649,-1.40022695273267,.0535966104577403,-1.03415883911242,1.05201311272283,-.174795669385988,-2.74931534527504,2.7967810011421,-.107052632842487,-2.72411329584901,3.28246374904242,-.106757773657142,.108600897231078,-.00415692611322449,-.105779161098541,2.26282036321899,-2.30188690834251,.0881095280677934,-2.15044627554921,2.18757273479106,.119501749587716,3.45007465931297,-3.50963860084301,.134338746031892,3.4184489866057,1.77035093065653,-1.8009152199194,.0689337905617311,-3.42380239520456,3.48291275963918,1.33170919606025,-3.74106981000851,3.80565764777524,-.14566949316231,-3.58481468643821,3.64670485341609,3.59724489879787,3.76406389114815,-3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.30970802258292,1.33231952555822,-.050997311866919,-1.29770236975413,1.56368716064049,.519224820053661,-.528188995996034,.020217536751761,.514465260741381,.518731848959601,-.527687514356292,.0201983414868368,-1.30425364587756,1.32677098129081,1.47252576661999,-3.06556135691961,3.11848685368264,-.119366596109985,-3.03746038662953,3.6600363220783,1.5947619510804,-1.62229477839744,.0620967201893262,1.58014330533833,.444911232504622,-.452592418790165,.0173239199033356,-2.09440936915946,2.13056837715884,2.63186985254377,4.2894282422279,-4.36348323472717,.167021432333127,4.25010849784966,-.964148801340396,.98079438405867,-.0375419530619817,-2.01722273728898,2.05204915464121,3.46190525493913,-5.18623489409549,5.27577283777777,-.201941221845807,-.205483038001613,.20903060758203,3.63947886955733,5.81833621896758,-5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.71235176911142,1.74191473005723,-.0666754235978707,-1.69665521651631,2.04441175413968,1.5663330178964,-1.593375031711,.0609897565211624,1.55197496987264,-1.07607439139602,1.09465231795021,-.0419001032349368,.363617646552178,-.369895336889078,.420994417051474,-3.29802783242322,3.35496675520006,-.128418358139773,-3.26779592010949,3.93758279561022,3.07100849759875,-3.12402803860063,.119578696522851,3.04285759691612,-2.20598407348367,2.24406936639925,-.0858964409190727,.923194488569861,-.939133013954983,.507379758856801,4.96464287844228,-5.05035513899607,.19331288874743,4.91913366884931,-3.80884940873902,3.87460743053656,-.148308690032331,2.02329330850018,-2.058224531456,.0162698697481229,-6.45780747384476,6.56929852924652,-.251453618731752,4.10479708970298,-4.17566451229335,-1.32544341327556,7.79428043908953,-7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.43862930490481,.0550665406758484,1.40124994253785,-1.68845845881123,.568134589816676,-1.94468190274016,1.97825593593812,-.0757218767587681,-1.92685565797098,2.32179544632223,2.27090372268066,-2.31010982325998,.088424277306817,2.25008711195035,-2.44801058376252,2.49027435399315,-.0953204508907588,2.49586968540814,-2.53895972010479,-2.4264683064468,-3.43216369764802,3.49141841400203,-.133641330308316,-3.40070220695544,4.09773034500049,4.00974905232829,-4.07897551472765,.156131305216531,3.97299303311745,-4.32546942586304,4.40014665135235,-.168424801181435,4.4144312457315,-4.49064435720703,-4.2977488551218,5.37669458746301,-5.46952072998373,.209357327015084,5.32740823859776,-5.80779112496784,5.9080599501188,-.22614333130344,5.9385809984625,-6.04110785209613,-5.79724384742923,-7.27823584467216,7.40389121595289,-.283399396550353,7.46210387346587,-7.59093364435534,-7.31197439196322,9.14177937313726,-9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.94468190274016,-2.51316806304373,-2.62015220863919,4.45987405838045,-.618753566698315,7.0795491341205,-4.51040089229094,-4.70240611527619,8.00416822009146,10.545740291788,-6.71872114372664,-7.00473331197222,14.1178337840267,-8.99451206786853,17.5926352342147,-.0319379183519753,.0203477386368136,.0212139303918346,-.0361091456847505,.00500970708930343,-.154660030302817,.0985343453312713,.102728896816887,-.174859285048361,-.43925822557731,.279852665057626,.291765835225991,-.961797763577864,.612764091356553,-1.79684767953686,.000465578153648587,-.000296621166194719,-.000309248161967448,.000526384631617547,.00319092434145473,-.00203294697585566,-.00211948838134797,.0123137463285655,-.00784512281993549,.0353622090543082,-5.86464816610367e-006,3.73638403043722e-006,3.89543979110572e-006,-5.19109369939388e-005,3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,-4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.71235176911142,-2.36514987391856,-2.46583296876894,4.19720062576954,-.582310805943786,5.99558490979395,-3.81980420145785,-3.98241111267819,6.77863367944468,7.62488741580146,-4.85783746627756,-5.06463286632359,8.14599447640013,-5.18983625559255,7.34718311089616,-.166073783576779,.105806079951921,.110310185087204,-.187763722716786,.026049944826181,-.764836159193308,.487279293067771,.508022497640507,-.864726999689721,-2.02218309922343,1.28833860494792,1.34318245298509,-4.02885967082133,2.56679795782664,-6.68010475273478,.0125887190510936,-.00802030872060955,-.00836172873999091,.0142328590555997,.0823390061849571,-.0524584150880732,-.054691540214079,.299004079114613,-.190496349504475,.796298871799325,-.000824565926326729,.000525333297396549,.000547696439552502,-.00697737063211093,.00444530268091327,-.0321260785529637,5.01144631645353e-005,-3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,3.30970802258292,-2.1086243974646,-2.19838734760781,3.74196990142034,-.519153050663133,4.2909622271481,-2.73378424126438,-2.8501598950859,4.85138005845884,3.64701232940425,-2.32352193122762,-2.42243294886035,1.49376547749553,-.951682235351175,-1.29952592474851,-.398540259080393,.253911132799744,.264719986484413,-.450591303972135,.0625140918479549,-1.67165845871656,1.06501836010644,1.11035559079056,-1.88998413076671,-3.84755914106515,2.45129087378787,2.55564094422791,-6.3113399865174,4.0209726602962,-7.98777931164405,.0724976248699598,-.0461884430503282,-.0481546590241491,.0819661216087145,.434860760346183,-.277051027559315,-.288844933452006,1.40707187748205,-.896449496423068,3.23180916098999,-.0113956405123911,.00726019498888195,.00756925739444607,-.0887334889355571,.056532358254944,-.36808995603673,.00166206328995187,-.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.76195386285985,-1.75964866391836,-1.83455591410421,3.12267672961303,-.433233615748229,2.32621648423849,-1.48203913941402,-1.54512870999352,2.63003020438924,.160883133644535,-.102499102175716,-.1068624311232,-2.12140648579363,1.35155410732446,-2.7058281817046,-.714786270724957,.455392366453167,.474778162593069,-.80814038343456,.112119700737043,-2.59853682213183,1.65553520251249,1.7260104019959,-2.93791673259187,-4.7549644652192,3.02940138707022,3.15836129611741,-5.36731919542456,3.41953432868287,-3.25729414076559,.233201739437629,-.148573491622283,-.154898181376293,.263658873906486,1.22672710731592,-.781551330007755,-.814821529066827,3.30495822058058,-2.10559828531328,5.89679444984459,-.0657431866024218,.0418851712113093,.0436681993244461,-.451524561772794,.287667582930097,-1.59193561065057,.017197430813833,-.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,2.10349449699454,-1.3401423285773,-1.39719143090653,2.37821978308955,-.329949221416697,.504794589391635,-.321606068872707,-.335296657900582,.570722899656038,-1.67800190212113,1.06905978594731,1.11456905784922,-1.79846700964276,1.14580844833274,.309494957156941,-1.09494796295235,.697594462028796,.727290664744143,-1.23795280192461,.171750973622283,-3.24473228828999,2.06722817250133,2.15522890943138,-3.66850806244401,-4.05017972204568,2.58038102231816,2.69022638759908,-1.63294619261043,1.04035466481729,2.28560008620039,.547225517982271,-.348638934287124,-.363480297113411,.618695487400889,2.39319484717323,-1.52471124557473,-1.58961734284587,4.87179403320196,-3.10383383936015,5.49730994822615,-.236321146004056,.150560874408057,.156970165907755,-1.36209257652082,.867793055767227,-3.77635721863404,.0946962283839159,-.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,1.37570022553441,-.876462527620663,-.913773042600856,1.555372498783,-.215789116142874,-.821946062819915,.523664174783431,.545956263457973,-.929295698075521,-1.61733246050546,1.03040711208251,1.07427096146664,.359308464240942,-.228916444770656,1.71165917287845,-1.51514018159912,.965300119862265,1.00639240141108,-1.71302390303713,.237661341149246,-3.36446686788906,2.14351141382237,2.23475948529486,-3.80388048506198,-1.91964817870762,1.22301331540983,1.27507630270228,1.95173675811295,-1.24345704101385,2.75500656898115,1.04781596872157,-.667566534563461,-.69598446548217,1.18466517033957,3.55510980002442,-2.2649705675878,-2.36138908644035,4.59045028887451,-2.92458893939423,1.24891795372526,-.626153181688296,.398923973340748,.415905941872818,-2.84473124098146,1.81238636712449,-5.48872263996485,.347192252628353,-.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,.624299774465593,-.397743162474635,-.414674864349061,.70583596787811,-.0979262008100592,-1.43546196207946,.914536899448643,.953468219635059,-1.62293937090584,-.337640235075449,.215111554391374,.224268731891381,1.50389998701775,-.958138960199622,-.0107903422167659,-1.94896143440088,1.24168887412682,1.29454686902578,-2.20350404785819,.305709527061058,-2.83314651369184,1.80500570449659,1.88184377891507,-3.20316744314248,.637278791777431,-.406012131085817,-.423295838593301,2.63766838605638,-1.68046603255685,-.808065843843811,1.73374759666501,-1.10457552610647,-1.15159668335474,1.96018237289792,4.12740556141283,-2.62958182530229,-2.74152163963164,1.98940127224618,-1.26745321062054,-2.96832239843812,-1.33269841309056,.849065950254409,.885210208841136,-4.37531622836174,2.78752641722509,-4.52786743647192,.95054294779362,-.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.103494496994541,.0659366384819972,.068743523956616,-.117011316428444,.0162339044637634,-1.29720437850508,.826452599656954,.86163422086386,-1.46662476164646,.991754894783212,-.631849865121666,-.658747357376999,.589436063368126,-.3755313928685,-1.41318416081787,-2.36915365304765,1.50939453196029,1.57364860569271,-2.6785751489707,.371619894588022,-1.68415540504794,1.07298020016939,1.11865636189275,-1.9041132311976,2.32831086437971,-1.48337347495979,-1.54651984745118,.381749900992996,-.243213947948102,-2.46800676549626,2.56192161158569,-1.63220754705252,-1.70168970188741,2.8965194203795,3.58714638367868,-2.28538116617536,-2.38266855269129,-1.52344991306454,.970594274001399,-2.69130383032704,-2.39387697960251,1.52514583383108,1.59007043171885,-4.93737708601551,3.14561698874585,-.508166439915618,2.07554270972356,-1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-.761953862859853,.485442973823058,.506108007154288,-.861468262951918,.119518298795295,-.55104212557113,.351070505731392,.366015379224059,-.62301055983504,1.37646295656649,-.876948465821911,-.914279667219301,-1.03415883911242,.658865538494566,-.174795669385988,-2.74931534527504,1.75159662753591,1.82616110784378,-3.10838756746075,.431251167473262,-.106757773657142,.0680156813349527,.070911070513513,-.120700790879792,2.26282036321899,-1.44164929037179,-1.50301948784968,-2.15044627554921,1.37005543987348,.119501749587716,3.45007465931297,-2.19805238043167,-2.2916222307293,3.90066901628343,1.77035093065653,-1.1278956144227,-1.17590949421397,-3.42380239520456,2.18131424709139,1.33170919606025,-3.74106981000851,2.38344621859103,2.48490817973374,-3.58481468643821,2.28389563582348,3.59724489879787,3.76406389114815,-2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.30970802258292,.834418707369306,.869939440657895,-1.48076143475923,.2054377337102,.519224820053661,-.33079960987665,-.344881562761671,.587037779528388,.518731848959601,-.33048553682317,-.344554119658887,-1.30425364587756,.830943708281939,1.47252576661999,-3.06556135691961,1.95307786118934,2.03621928395244,-3.46593664692318,.48085677636235,1.5947619510804,-1.01602737631215,-1.05927908845788,1.80304461367463,.444911232504622,-.283454211851926,-.295520697836744,-2.09440936915946,1.33435416739534,2.63186985254377,4.2894282422279,-2.73280693595118,-2.84914098611565,4.84964573066987,-.964148801340396,.614261944257265,.640410728572748,-2.01722273728898,1.28517834515454,3.46190525493913,-5.18623489409549,3.30416500488362,3.44482144542742,-.205483038001613,.130913828127032,3.63947886955733,5.81833621896758,-3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.71235176911142,1.09094418382326,1.13738506181903,-1.93599215910843,.268595488990853,1.5663330178964,-.997915221393877,-1.04039590973615,1.77090274061412,-1.07607439139602,.685570056312228,.714754386796489,.363617646552178,-.23166183708041,.420994417051474,-3.29802783242322,2.10118291403716,2.19062908534965,-3.72876422817853,.517320923384123,3.07100849759875,-1.95654825118803,-2.0398374063157,3.47209520918979,-2.20598407348367,1.40543872850124,1.4652674627761,.923194488569861,-.588169835027997,.507379758856801,4.96464287844228,-3.16298810157525,-3.2976347213703,5.61304625700718,-3.80884940873902,2.426628794168,2.52992901249613,2.02329330850018,-1.28904592334466,.0162698697481229,-6.45780747384476,4.11428751282488,4.28943041158219,4.10479708970298,-2.61517790433632,-1.32544341327556,7.79428043908953,-4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.900999483948357,-.939354523617303,1.59891584058814,-.221830227920984,-1.44110338496998,-1.94468190274016,1.23896237294843,1.29170430168928,-2.19866559171934,.305038249745382,2.27090372268066,-1.44679922167927,-1.50838864882935,2.5674933620227,-2.44801058376252,1.55963450778316,1.62602726880111,2.49586968540814,-1.59012567711558,-2.4264683064468,-3.43216369764802,2.18664125535227,2.27972534004502,-3.88041880521056,.538361161121001,4.00974905232829,-2.55462252942968,-2.66337136913609,4.53343925473611,-4.32546942586304,2.75576887572707,2.87308040304256,4.4144312457315,-2.81244670509951,-4.2977488551218,5.37669458746301,-3.42550741762225,-3.57132933666014,6.07891366371537,-5.80779112496784,3.70016024840062,3.85767398203284,5.9385809984625,-3.78348685005768,-5.79724384742923,-7.27823584467216,4.63698476184225,4.83437858766287,7.46210387346587,-4.75412760808659,-7.31197439196322,9.14177937313726,-5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.94468190274016,-.855643422475599,-4.20278340045202,1.90927918078571,3.75450887490789,7.0795491341205,-1.53562943639462,-7.54276575929575,3.42659715096905,10.545740291788,-2.28748313278936,-11.235750652608,14.1178337840267,-3.06230816888874,17.5926352342147,-.0319379183519753,.00692767387221854,.0340276241262534,-.0154583827232305,-.0303981920101794,-.154660030302817,.0335474040556334,.16477947382797,-.0748575381590809,-.43925822557731,.0952797768044285,.467998998326769,-.961797763577864,.208624155307574,-1.79684767953686,.000465578153648587,-.00010098884891692,-.000496041045803253,.000225346098359983,.00319092434145473,-.000692145397560818,-.00339970729838115,.0123137463285655,-.00267098242974697,.0353622090543082,-5.86464816610367e-006,1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,-1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.71235176911142,-.805248547658387,-3.95525184959074,1.79682826632835,3.53337937191509,5.99558490979395,-1.30050607764117,-6.38787749160276,2.90194386409118,7.62488741580146,-1.65391910576877,-8.12378566015575,8.14599447640013,-1.7669527641717,7.34718311089616,-.166073783576779,.0360231684064709,.176940031673225,-.080381948458016,-.158067370120622,-.764836159193308,.165901090401825,.814879575736018,-.370191004490028,-2.02218309922343,.43863300248611,2.1544950322241,-4.02885967082133,.873902474623302,-6.68010475273478,.0125887190510936,-.00273062693415814,-.0134124020091825,.00609310960852155,.0823390061849571,-.0178602054051423,-.087726467416282,.299004079114613,-.0648571621038782,.796298871799325,-.000824565926326729,.000178857111521775,.000878517475835277,-.00697737063211093,.00151346583452521,-.0321260785529637,5.01144631645353e-005,-1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,3.30970802258292,-.717910840382491,-3.52626302465398,1.60194326888752,3.15014707155715,4.2909622271481,-.930755303993616,-4.57172093140434,2.0768835223697,3.64701232940425,-.791075728944641,-3.88563723611535,1.49376547749553,-.324013605353116,-1.29952592474851,-.398540259080393,.0864476172000634,.4246168452719,-.192898854195334,-.379326641949509,-1.67165845871656,.362600483446047,1.78103547961334,-.809105213656782,-3.84755914106515,.834576462901105,4.09930587333892,-6.3113399865174,1.3689967095037,-7.98777931164405,.0724976248699598,-.0157255052185161,-.0772411621484506,.0350898270822692,.434860760346183,-.0943259199045093,-.4633138058217,1.40707187748205,-.305208842367266,3.23180916098999,-.0113956405123911,.00247183552086188,.0121412600482659,-.0887334889355571,.0192472366688018,-.36808995603673,.00166206328995187,-.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.76195386285985,-.599097142483271,-2.94266917684245,1.33682287663953,2.62880012783543,2.32621648423849,-.504581074436857,-2.47842139571551,1.12592011530959,.160883133644535,-.0348972612997825,-.171409756339991,-2.12140648579363,.460155609590019,-2.7058281817046,-.714786270724957,.155044737648518,.76155491046056,-.345966184032581,-.680326440323354,-2.59853682213183,.563650249872674,2.76855972053751,-1.25772682793123,-4.7549644652192,1.03140232004452,5.06608294899263,-5.36731919542456,1.16422856843223,-3.25729414076559,.233201739437629,-.0505839353643848,-.248460186130407,.112872783443993,1.22672710731592,-.266090145216267,-1.30699216076136,3.30495822058058,-.716880557929028,5.89679444984459,-.0657431866024218,.0142603957822129,.0700447793236141,-.451524561772794,.0979404755348858,-1.59193561065057,.017197430813833,-.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,2.10349449699454,-.456270308973898,-2.24112665428606,1.0181196732794,2.00208507352643,.504794589391635,-.109495310522883,-.537823422397082,.244327381456737,-1.67800190212113,.363976443728615,1.7877939760679,-1.79846700964276,.390106605542927,.309494957156941,-1.09494796295235,.237505848402132,1.1665906747753,-.529969564290411,-1.04216054572391,-3.24473228828999,.703816913228947,3.45703591374062,-1.57049413873188,-4.05017972204568,.878527020016876,4.31518396831288,-1.63294619261043,.354203381414383,2.28560008620039,.547225517982271,-.118699029853337,-.583030617100296,.264864522602584,2.39319484717323,-.519109393293245,-2.54978216897294,4.87179403320196,-1.05674389554722,5.49730994822615,-.236321146004056,.0512605677924324,.251783696179217,-1.36209257652082,.295452354005307,-3.77635721863404,.0946962283839159,-.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,1.37570022553441,-.298403997660505,-1.4657126263737,.665857441582424,1.30937774789745,-.821946062819915,.178288835246813,.875726194023542,-.397832967008406,-1.61733246050546,.350816597143005,1.72315491808185,.359308464240942,-.0779378240336578,1.71165917287845,-1.51514018159912,.328650005712205,1.6142761726,-.733348258592925,-1.44209530674498,-3.36446686788906,.729788615894315,3.58460475610227,-1.62844728825442,-1.91964817870762,.416392088662432,2.04525122687055,1.95173675811295,-.423352442647985,2.75500656898115,1.04781596872157,-.227282418061715,-1.11637482313293,.507157044419473,3.55510980002442,-.771141093489061,-3.7877215002156,4.59045028887451,-.995717447724871,1.24891795372526,-.626153181688296,.135819278812355,.667122536961418,-2.84473124098146,.61705243545725,-5.48872263996485,.347192252628353,-.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,.624299774465593,-.135417255141262,-.66514786077106,.302169500949717,.594202296059882,-1.43546196207946,.311366953182089,1.52938458808924,-.694782927048218,-.337640235075449,.0732377548959319,.359732118052426,1.50389998701775,-.326211888161397,-.0107903422167659,-1.94896143440088,.422750445356608,2.07648245560297,-.943323589025389,-1.85500204648398,-2.83314651369184,.614539585033937,3.01851998146924,-1.37128107917545,.637278791777431,-.138232541833329,-.678976098432196,2.63766838605638,-.57213830168982,-.808065843843811,1.73374759666501,-.376068277103551,-1.84718712480164,.839157192810101,4.12740556141283,-.895278125552931,-4.39746271267498,1.98940127224618,-.431522276170264,-2.96832239843812,-1.33269841309056,.289076447479325,1.41989719504135,-4.37531622836174,.949052583847879,-4.52786743647192,.95054294779362,-.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.103494496994541,.0224490561721302,.110266167141295,-.0500927307472635,-.0985050295691045,-1.29720437850508,.281377414142329,1.38208077712737,-.627864394102026,.991754894783212,-.215122175485818,-1.05664565957165,.589436063368126,-.127854945702229,-1.41318416081787,-2.36915365304765,.513894602666682,2.52416795342767,-1.1467022833279,-2.25493680750505,-1.68415540504794,.365311203903165,1.79435010419287,-.815153904032968,2.32831086437971,-.505035367744937,-2.48065281165978,.381749900992996,-.0828056101286271,-2.46800676549626,2.56192161158569,-.555708021396345,-2.72954876752254,1.24000457270319,3.58714638367868,-.778090168915826,-3.82185424534203,-1.52344991306454,.33045247496738,-2.69130383032704,-2.39387697960251,.519257354978938,2.55050893377232,-4.93737708601551,1.07096955656365,-.508166439915618,2.07554270972356,-.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-.761953862859853,.165275889681504,.81180868969769,-.368795934107388,-.725220083878106,-.55104212557113,.119526892559043,.58709694613961,-.266711811977634,1.37646295656649,-.298569442157331,-1.4665252637196,-1.03415883911242,.224320041543878,-.174795669385988,-2.74931534527504,.596355713420295,2.92920371774241,-1.33070566358573,-2.61677091290561,-.106757773657142,.0231568955595295,.113742960840148,-.0516722006304461,2.26282036321899,-.490829782123398,-2.41087725104767,-2.15044627554921,.466454648534285,.119501749587716,3.45007465931297,-.748357855262331,-3.67581388579636,1.66988261247679,1.77035093065653,-.384007929239338,-1.88618542328239,-3.42380239520456,.742659122293479,1.33170919606025,-3.74106981000851,.8114777955963,3.98584891938639,-3.58481468643821,.777584399208991,3.59724489879787,3.76406389114815,-.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.30970802258292,.284089587580723,1.39540253750922,-.633916326355378,-1.24656702759982,.519224820053661,-.11262538089301,-.55319782656806,.251311807489747,.518731848959601,-.112518450303553,-.552672600425686,-1.30425364587756,.282906475254915,1.47252576661999,-3.06556135691961,.66495283386875,3.26614178293108,-1.48377299342298,-2.91777071127945,1.5947619510804,-.345920813804923,-1.69910761471712,.771886268054593,.444911232504622,-.0965059740773432,-.474021882875009,-2.09440936915946,.454299648071273,2.63186985254377,4.2894282422279,-.93042256665094,-4.57008657837224,2.07614105388364,-.964148801340396,.209134120399514,1.02723329226541,-2.01722273728898,.437557047580714,3.46190525493913,-5.18623489409549,1.1249494591216,5.5255715071159,-.205483038001613,.0445714544901277,3.63947886955733,5.81833621896758,-1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.71235176911142,.37142729485662,1.82439136244598,-.828801323796214,-1.62979932795776,1.5663330178964,-.33975427586002,-1.66881856898237,.758126281054385,-1.07607439139602,.233411970315802,1.14648220142474,.363617646552178,-.0788725314678718,.420994417051474,-3.29802783242322,.715377282662343,3.51381859652975,-1.5962898991603,-3.13902998310834,3.07100849759875,-.666134376965478,-3.2719453330298,1.4864094837286,-2.20598407348367,.47850138692202,2.35032214902362,.923194488569861,-.200250694746914,.507379758856801,4.96464287844228,-1.07688379630437,-5.28948067292902,2.40295403390442,-3.80884940873902,.826179910946771,4.05806335333585,2.02329330850018,-.438873818984883,.0162698697481229,-6.45780747384476,1.40076706380653,6.88034339064391,4.10479708970298,-.890374108024765,-1.32544341327556,7.79428043908953,-1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,-.306757949629772,-1.50674590047908,.684498415493822,1.3460343572585,-1.02189836036615,-1.94468190274016,.421822169673831,2.07192291330725,-.941252238253572,-1.85092883095057,2.27090372268066,-.492583148686599,-2.41948950668889,1.09914799358608,-2.44801058376252,.53099950921484,2.60818451365804,2.49586968540814,-.541380656765823,-2.4264683064468,-3.43216369764802,.744472777196595,3.65673100407672,-1.66121346489508,-3.26669916121878,4.00974905232829,-.869757178740834,-4.27210791134071,1.94077255839989,-4.32546942586304,.938240282077712,4.60848594454081,4.4144312457315,-.957537046434959,-4.2977488551218,5.37669458746301,-1.16626219059145,-5.72849302577651,2.60239261198785,-5.80779112496784,1.2597716106639,6.18779631297208,5.9385809984625,-1.28814132444211,-5.79724384742923,-7.27823584467216,1.57872669557351,7.75445258971521,7.46210387346587,-1.6186096255506,-7.31197439196322,9.14177937313726,-1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.94468190274016,.855643422475599,-4.20278340045202,-1.90927918078571,3.75450887490794,7.0795491341205,1.53562943639462,-7.54276575929575,-3.42659715096906,10.545740291788,2.28748313278936,-11.235750652608,14.1178337840267,3.06230816888874,17.5926352342147,-.0319379183519753,-.00692767387221854,.0340276241262535,.0154583827232306,-.0303981920101797,-.154660030302817,-.0335474040556334,.16477947382797,.074857538159081,-.43925822557731,-.0952797768044285,.467998998326769,-.961797763577864,-.208624155307574,-1.79684767953686,.000465578153648587,.00010098884891692,-.000496041045803253,-.000225346098359983,.00319092434145473,.000692145397560818,-.00339970729838115,.0123137463285655,.00267098242974697,.0353622090543082,-5.86464816610367e-006,-1.27210450685679e-006,6.24837352484276e-006,-5.19109369939388e-005,-1.12600338610288e-005,-.000252650702007359,6.85464900528793e-008,1.48684620919026e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.71235176911142,.805248547658387,-3.95525184959075,-1.79682826632836,3.53337937191513,5.99558490979395,1.30050607764117,-6.38787749160276,-2.90194386409118,7.62488741580146,1.65391910576877,-8.12378566015576,8.14599447640013,1.7669527641717,7.34718311089616,-.166073783576779,-.0360231684064709,.176940031673225,.0803819484580161,-.158067370120624,-.764836159193308,-.165901090401825,.814879575736019,.370191004490028,-2.02218309922343,-.43863300248611,2.15449503222411,-4.02885967082133,-.873902474623302,-6.68010475273478,.0125887190510936,.00273062693415814,-.0134124020091825,-.00609310960852156,.0823390061849571,.0178602054051423,-.087726467416282,.299004079114613,.0648571621038782,.796298871799325,-.000824565926326729,-.000178857111521775,.000878517475835277,-.00697737063211093,-.00151346583452521,-.0321260785529637,5.01144631645353e-005,1.08703595945339e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,3.30970802258292,.717910840382491,-3.52626302465399,-1.60194326888752,3.15014707155719,4.2909622271481,.930755303993616,-4.57172093140434,-2.0768835223697,3.64701232940425,.791075728944641,-3.88563723611535,1.49376547749553,.324013605353116,-1.29952592474851,-.398540259080393,-.0864476172000634,.4246168452719,.192898854195334,-.379326641949514,-1.67165845871656,-.362600483446047,1.78103547961334,.809105213656783,-3.84755914106515,-.834576462901105,4.09930587333892,-6.3113399865174,-1.3689967095037,-7.98777931164405,.0724976248699598,.0157255052185161,-.0772411621484506,-.0350898270822693,.434860760346183,.0943259199045093,-.463313805821701,1.40707187748205,.305208842367266,3.23180916098999,-.0113956405123911,-.00247183552086188,.0121412600482659,-.0887334889355571,-.0192472366688018,-.36808995603673,.00166206328995187,.000360519189267433,.015902832140972,-.000231576690964732},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.76195386285985,.599097142483271,-2.94266917684246,-1.33682287663953,2.62880012783546,2.32621648423849,.504581074436857,-2.47842139571551,-1.12592011530959,.160883133644535,.0348972612997825,-.171409756339991,-2.12140648579363,-.460155609590019,-2.7058281817046,-.714786270724957,-.155044737648518,.761554910460561,.345966184032582,-.680326440323361,-2.59853682213183,-.563650249872674,2.76855972053751,1.25772682793124,-4.7549644652192,-1.03140232004452,5.06608294899263,-5.36731919542456,-1.16422856843223,-3.25729414076559,.233201739437629,.0505839353643848,-.248460186130407,-.112872783443993,1.22672710731592,.266090145216267,-1.30699216076136,3.30495822058058,.716880557929028,5.89679444984459,-.0657431866024218,-.0142603957822129,.0700447793236141,-.451524561772794,-.0979404755348858,-1.59193561065057,.017197430813833,.0037303054895493,.145628931559507,-.00429748986298271},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,2.10349449699454,.456270308973898,-2.24112665428606,-1.0181196732794,2.00208507352645,.504794589391635,.109495310522883,-.537823422397082,-.244327381456737,-1.67800190212113,-.363976443728615,1.7877939760679,-1.79846700964276,-.390106605542927,.309494957156941,-1.09494796295235,-.237505848402132,1.1665906747753,.529969564290411,-1.04216054572392,-3.24473228828999,-.703816913228947,3.45703591374062,1.57049413873189,-4.05017972204568,-.878527020016876,4.31518396831288,-1.63294619261043,-.354203381414383,2.28560008620039,.547225517982271,.118699029853337,-.583030617100297,-.264864522602584,2.39319484717323,.519109393293245,-2.54978216897294,4.87179403320196,1.05674389554722,5.49730994822615,-.236321146004056,-.0512605677924324,.251783696179217,-1.36209257652082,-.295452354005307,-3.77635721863404,.0946962283839159,.0205406182123436,.676667877319681,-.0362494356624792},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,1.37570022553441,.298403997660505,-1.46571262637371,-.665857441582425,1.30937774789746,-.821946062819915,-.178288835246813,.875726194023542,.397832967008407,-1.61733246050546,-.350816597143005,1.72315491808185,.359308464240942,.0779378240336578,1.71165917287845,-1.51514018159912,-.328650005712205,1.6142761726,.733348258592926,-1.44209530674499,-3.36446686788906,-.729788615894315,3.58460475610227,1.62844728825442,-1.91964817870762,-.416392088662432,2.04525122687055,1.95173675811295,.423352442647985,2.75500656898115,1.04781596872157,.227282418061715,-1.11637482313293,-.507157044419473,3.55510980002442,.771141093489061,-3.7877215002156,4.59045028887451,.995717447724871,1.24891795372526,-.626153181688296,-.135819278812355,.667122536961419,-2.84473124098146,-.61705243545725,-5.48872263996485,.347192252628353,.0753096889837044,1.97345041935918,-.183906859961415},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,.624299774465593,.135417255141262,-.665147860771061,-.302169500949717,.594202296059888,-1.43546196207946,-.311366953182089,1.52938458808924,.694782927048219,-.337640235075449,-.0732377548959319,.359732118052426,1.50389998701775,.326211888161397,-.0107903422167659,-1.94896143440088,-.422750445356608,2.07648245560297,.94332358902539,-1.85500204648401,-2.83314651369184,-.614539585033937,3.01851998146924,1.37128107917545,.637278791777431,.138232541833329,-.678976098432196,2.63766838605638,.57213830168982,-.808065843843811,1.73374759666501,.376068277103551,-1.84718712480164,-.839157192810102,4.12740556141283,.895278125552931,-4.39746271267498,1.98940127224618,.431522276170264,-2.96832239843812,-1.33269841309056,-.289076447479325,1.41989719504135,-4.37531622836174,-.949052583847879,-4.52786743647192,.95054294779362,.20618286618457,3.96849611187727,-.647664476709676},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.103494496994541,-.0224490561721302,.110266167141295,.0500927307472635,-.0985050295691056,-1.29720437850508,-.281377414142329,1.38208077712737,.627864394102026,.991754894783212,.215122175485818,-1.05664565957165,.589436063368126,.127854945702229,-1.41318416081787,-2.36915365304765,-.513894602666682,2.52416795342767,1.14670228332791,-2.25493680750508,-1.68415540504794,-.365311203903165,1.79435010419287,.81515390403297,2.32831086437971,.505035367744937,-2.48065281165978,.381749900992996,.0828056101286271,-2.46800676549626,2.56192161158569,.555708021396345,-2.72954876752254,-1.24000457270319,3.58714638367868,.778090168915826,-3.82185424534203,-1.52344991306454,-.33045247496738,-2.69130383032704,-2.39387697960251,-.519257354978938,2.55050893377233,-4.93737708601551,-1.07096955656365,-.508166439915618,2.07554270972356,.45020726919559,5.63164748396134,-1.71909544841957},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-.761953862859853,-.165275889681504,.811808689697691,.368795934107389,-.725220083878115,-.55104212557113,-.119526892559043,.587096946139611,.266711811977634,1.37646295656649,.298569442157331,-1.4665252637196,-1.03415883911242,-.224320041543878,-.174795669385988,-2.74931534527504,-.596355713420295,2.92920371774242,1.33070566358573,-2.61677091290564,-.106757773657142,-.0231568955595295,.113742960840148,.0516722006304461,2.26282036321899,.490829782123398,-2.41087725104768,-2.15044627554921,-.466454648534285,.119501749587716,3.45007465931297,.748357855262331,-3.67581388579636,-1.6698826124768,1.77035093065653,.384007929239338,-1.88618542328239,-3.42380239520456,-.742659122293479,1.33170919606025,-3.74106981000851,-.8114777955963,3.98584891938639,-3.58481468643821,-.777584399208991,3.59724489879787,3.76406389114815,.816465456274442,5.23559759098235,-3.6179005979755},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.30970802258292,-.284089587580723,1.39540253750922,.633916326355378,-1.24656702759984,.519224820053661,.11262538089301,-.55319782656806,-.251311807489747,.518731848959601,.112518450303553,-.552672600425687,-1.30425364587756,-.282906475254915,1.47252576661999,-3.06556135691961,-.66495283386875,3.26614178293108,1.48377299342298,-2.91777071127949,1.5947619510804,.345920813804923,-1.69910761471712,-.771886268054594,.444911232504622,.0965059740773432,-.474021882875009,-2.09440936915946,-.454299648071273,2.63186985254377,4.2894282422279,.93042256665094,-4.57008657837224,-2.07614105388364,-.964148801340396,-.209134120399514,1.02723329226541,-2.01722273728898,-.437557047580714,3.46190525493913,-5.18623489409549,-1.1249494591216,5.5255715071159,-.205483038001613,-.0445714544901277,3.63947886955733,5.81833621896758,1.26205895360837,1.69243379052584,-6.23568135536041},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.71235176911142,-.37142729485662,1.82439136244598,.828801323796215,-1.62979932795778,1.5663330178964,.33975427586002,-1.66881856898237,-.758126281054386,-1.07607439139602,-.233411970315802,1.14648220142474,.363617646552178,.0788725314678718,.420994417051474,-3.29802783242322,-.715377282662343,3.51381859652975,1.5962898991603,-3.13902998310838,3.07100849759875,.666134376965478,-3.2719453330298,-1.48640948372861,-2.20598407348367,-.47850138692202,2.35032214902362,.923194488569861,.200250694746914,.507379758856801,4.96464287844228,1.07688379630437,-5.28948067292902,-2.40295403390442,-3.80884940873902,-.826179910946771,4.05806335333585,2.02329330850018,.438873818984883,.0162698697481229,-6.45780747384476,-1.40076706380653,6.88034339064391,4.10479708970298,.890374108024765,-1.32544341327556,7.79428043908953,1.69066225204032,-4.0355398121645,-8.98680735137036},
			{1.414213562,.306757949629772,-1.50674590047908,-.684498415493823,1.34603435725852,1.02189836036618,-1.94468190274016,-.421822169673831,2.07192291330725,.941252238253573,-1.85092883095059,2.27090372268066,.492583148686599,-2.41948950668889,-1.09914799358608,-2.44801058376252,-.53099950921484,2.60818451365804,2.49586968540814,.541380656765823,-2.4264683064468,-3.43216369764802,-.744472777196595,3.65673100407672,1.66121346489509,-3.26669916121882,4.00974905232829,.869757178740834,-4.27210791134071,-1.9407725583999,-4.32546942586304,-.938240282077712,4.60848594454081,4.4144312457315,.957537046434959,-4.2977488551218,5.37669458746301,1.16626219059145,-5.72849302577652,-2.60239261198785,-5.80779112496784,-1.2597716106639,6.18779631297208,5.9385809984625,1.28814132444211,-5.79724384742923,-7.27823584467216,-1.57872669557351,7.75445258971521,7.46210387346587,1.6186096255506,-7.31197439196322,9.14177937313726,1.98294908983919,-8.9987004064276,-10.9691712594614},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.94468190274016,2.51316806304373,-2.6201522086392,-4.45987405838047,-.618753566698294,7.0795491341205,4.51040089229094,-4.7024061152762,-8.00416822009149,10.545740291788,6.71872114372664,-7.00473331197223,14.1178337840267,8.99451206786853,17.5926352342147,-.0319379183519753,-.0203477386368136,.0212139303918347,.0361091456847507,.00500970708930326,-.154660030302817,-.0985343453312713,.102728896816887,.174859285048362,-.43925822557731,-.279852665057626,.291765835225991,-.961797763577864,-.612764091356553,-1.79684767953686,.000465578153648587,.000296621166194719,-.000309248161967448,-.000526384631617549,.00319092434145473,.00203294697585566,-.00211948838134797,.0123137463285655,.00784512281993549,.0353622090543082,-5.86464816610367e-006,-3.73638403043722e-006,3.89543979110573e-006,-5.19109369939388e-005,-3.30726056497391e-005,-.000252650702007359,6.85464900528793e-008,4.36711638217631e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.71235176911142,2.36514987391856,-2.46583296876895,-4.19720062576956,-.582310805943766,5.99558490979395,3.81980420145785,-3.98241111267819,-6.77863367944472,7.62488741580146,4.85783746627756,-5.06463286632359,8.14599447640013,5.18983625559255,7.34718311089616,-.166073783576779,-.105806079951921,.110310185087204,.187763722716787,.0260499448261801,-.764836159193308,-.487279293067771,.508022497640508,.864726999689725,-2.02218309922343,-1.28833860494792,1.34318245298509,-4.02885967082133,-2.56679795782664,-6.68010475273478,.0125887190510936,.00802030872060955,-.00836172873999092,-.0142328590555998,.0823390061849571,.0524584150880732,-.0546915402140791,.299004079114613,.190496349504475,.796298871799325,-.000824565926326729,-.000525333297396549,.000547696439552503,-.00697737063211093,-.00444530268091327,-.0321260785529637,5.01144631645353e-005,3.19280670536128e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,3.30970802258292,2.1086243974646,-2.19838734760782,-3.74196990142036,-.519153050663115,4.2909622271481,2.73378424126438,-2.85015989508591,-4.85138005845886,3.64701232940425,2.32352193122762,-2.42243294886036,1.49376547749553,.951682235351175,-1.29952592474851,-.398540259080393,-.253911132799744,.264719986484414,.450591303972137,.0625140918479527,-1.67165845871656,-1.06501836010644,1.11035559079056,1.88998413076672,-3.84755914106515,-2.45129087378787,2.55564094422791,-6.3113399865174,-4.0209726602962,-7.98777931164405,.0724976248699598,.0461884430503282,-.0481546590241492,-.0819661216087149,.434860760346183,.277051027559315,-.288844933452006,1.40707187748205,.896449496423068,3.23180916098999,-.0113956405123911,-.00726019498888195,.00756925739444608,-.0887334889355571,-.056532358254944,-.36808995603673,.00166206328995187,.00105890525046041,.015902832140972,-.000231576690964732},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.76195386285985,1.75964866391836,-1.83455591410421,-3.12267672961304,-.433233615748214,2.32621648423849,1.48203913941402,-1.54512870999353,-2.63003020438926,.160883133644535,.102499102175716,-.106862431123201,-2.12140648579363,-1.35155410732446,-2.7058281817046,-.714786270724957,-.455392366453167,.47477816259307,.808140383434563,.112119700737039,-2.59853682213183,-1.65553520251249,1.7260104019959,2.93791673259189,-4.7549644652192,-3.02940138707022,3.15836129611742,-5.36731919542456,-3.41953432868287,-3.25729414076559,.233201739437629,.148573491622283,-.154898181376293,-.263658873906487,1.22672710731592,.781551330007755,-.814821529066828,3.30495822058058,2.10559828531328,5.89679444984459,-.0657431866024218,-.0418851712113093,.0436681993244462,-.451524561772794,-.287667582930097,-1.59193561065057,.017197430813833,.010956532096756,.145628931559507,-.00429748986298271},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,2.10349449699454,1.3401423285773,-1.39719143090654,-2.37821978308956,-.329949221416686,.504794589391635,.321606068872707,-.335296657900582,-.570722899656041,-1.67800190212113,-1.06905978594731,1.11456905784922,-1.79846700964276,-1.14580844833274,.309494957156941,-1.09494796295235,-.697594462028796,.727290664744144,1.23795280192461,.171750973622277,-3.24473228828999,-2.06722817250133,2.15522890943138,3.66850806244403,-4.05017972204568,-2.58038102231816,2.69022638759909,-1.63294619261043,-1.04035466481729,2.28560008620039,.547225517982271,.348638934287124,-.363480297113412,-.618695487400892,2.39319484717323,1.52471124557473,-1.58961734284587,4.87179403320196,3.10383383936015,5.49730994822615,-.236321146004056,-.150560874408057,.156970165907756,-1.36209257652082,-.867793055767227,-3.77635721863404,.0946962283839159,.0603312365063012,.676667877319681,-.0362494356624792},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,1.37570022553441,.876462527620663,-.913773042600857,-1.55537249878301,-.215789116142867,-.821946062819915,-.523664174783431,.545956263457974,.929295698075525,-1.61733246050546,-1.03040711208251,1.07427096146664,.359308464240942,.228916444770656,1.71165917287845,-1.51514018159912,-.965300119862265,1.00639240141108,1.71302390303713,.237661341149238,-3.36446686788906,-2.14351141382237,2.23475948529486,3.80388048506199,-1.91964817870762,-1.22301331540983,1.27507630270228,1.95173675811295,1.24345704101385,2.75500656898115,1.04781596872157,.667566534563461,-.695984465482171,-1.18466517033958,3.55510980002442,2.2649705675878,-2.36138908644035,4.59045028887451,2.92458893939423,1.24891795372526,-.626153181688296,-.398923973340748,.415905941872819,-2.84473124098146,-1.81238636712449,-5.48872263996485,.347192252628353,.22119717188266,1.97345041935918,-.183906859961415},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,.624299774465593,.397743162474635,-.414674864349062,-.705835967878113,-.0979262008100559,-1.43546196207946,-.914536899448643,.953468219635061,1.62293937090585,-.337640235075449,-.215111554391374,.224268731891381,1.50389998701775,.958138960199622,-.0107903422167659,-1.94896143440088,-1.24168887412682,1.29454686902578,2.2035040478582,.305709527061048,-2.83314651369184,-1.80500570449659,1.88184377891507,3.20316744314249,.637278791777431,.406012131085817,-.423295838593301,2.63766838605638,1.68046603255685,-.808065843843811,1.73374759666501,1.10457552610647,-1.15159668335474,-1.96018237289793,4.12740556141283,2.62958182530229,-2.74152163963164,1.98940127224618,1.26745321062054,-2.96832239843812,-1.33269841309056,-.849065950254409,.885210208841138,-4.37531622836174,-2.78752641722509,-4.52786743647192,.95054294779362,.60559361625509,3.96849611187727,-.647664476709676},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.103494496994541,-.0659366384819972,.0687435239566161,.117011316428444,.0162339044637628,-1.29720437850508,-.826452599656954,.861634220863862,1.46662476164646,.991754894783212,.631849865121666,-.658747357377,.589436063368126,.3755313928685,-1.41318416081787,-2.36915365304765,-1.50939453196029,1.57364860569271,2.67857514897071,.371619894588009,-1.68415540504794,-1.07298020016939,1.11865636189276,1.90411323119761,2.32831086437971,1.48337347495979,-1.54651984745118,.381749900992996,.243213947948102,-2.46800676549626,2.56192161158569,1.63220754705252,-1.70168970188742,-2.89651942037951,3.58714638367868,2.28538116617536,-2.38266855269129,-1.52344991306454,-.970594274001399,-2.69130383032704,-2.39387697960251,-1.52514583383108,1.59007043171885,-4.93737708601551,-3.14561698874585,-.508166439915618,2.07554270972356,1.32233416511158,5.63164748396134,-1.71909544841957},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-.761953862859853,-.485442973823058,.506108007154288,.861468262951922,.119518298795291,-.55104212557113,-.351070505731392,.366015379224059,.623010559835043,1.37646295656649,.876948465821911,-.914279667219303,-1.03415883911242,-.658865538494566,-.174795669385988,-2.74931534527504,-1.75159662753591,1.82616110784379,3.10838756746077,.431251167473247,-.106757773657142,-.0680156813349527,.0709110705135131,.120700790879792,2.26282036321899,1.44164929037179,-1.50301948784969,-2.15044627554921,-1.37005543987348,.119501749587716,3.45007465931297,2.19805238043167,-2.29162223072931,-3.90066901628345,1.77035093065653,1.1278956144227,-1.17590949421397,-3.42380239520456,-2.18131424709139,1.33170919606025,-3.74106981000851,-2.38344621859103,2.48490817973374,-3.58481468643821,-2.28389563582348,3.59724489879787,3.76406389114815,2.39809581349978,5.23559759098235,-3.6179005979755},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.30970802258292,-.834418707369306,.869939440657896,1.48076143475924,.205437733710193,.519224820053661,.33079960987665,-.344881562761672,-.587037779528391,.518731848959601,.33048553682317,-.344554119658887,-1.30425364587756,-.830943708281939,1.47252576661999,-3.06556135691961,-1.95307786118934,2.03621928395244,3.46593664692319,.480856776362333,1.5947619510804,1.01602737631215,-1.05927908845788,-1.80304461367463,.444911232504622,.283454211851926,-.295520697836744,-2.09440936915946,-1.33435416739534,2.63186985254377,4.2894282422279,2.73280693595118,-2.84914098611565,-4.84964573066989,-.964148801340396,-.614261944257265,.640410728572749,-2.01722273728898,-1.28517834515454,3.46190525493913,-5.18623489409549,-3.30416500488362,3.44482144542743,-.205483038001613,-.130913828127032,3.63947886955733,5.81833621896758,3.70687855778777,1.69243379052584,-6.23568135536041},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.71235176911142,-1.09094418382326,1.13738506181903,1.93599215910844,.268595488990843,1.5663330178964,.997915221393877,-1.04039590973615,-1.77090274061413,-1.07607439139602,-.685570056312228,.71475438679649,.363617646552178,.23166183708041,.420994417051474,-3.29802783242322,-2.10118291403716,2.19062908534965,3.72876422817854,.517320923384106,3.07100849759875,1.95654825118803,-2.0398374063157,-3.47209520918981,-2.20598407348367,-1.40543872850124,1.46526746277611,.923194488569861,.588169835027997,.507379758856801,4.96464287844228,3.16298810157525,-3.2976347213703,-5.6130462570072,-3.80884940873902,-2.426628794168,2.52992901249613,2.02329330850018,1.28904592334466,.0162698697481229,-6.45780747384476,-4.11428751282488,4.2894304115822,4.10479708970298,2.61517790433632,-1.32544341327556,7.79428043908953,4.96575824182473,-4.0355398121645,-8.98680735137036},
			{1.414213562,.900999483948357,-.939354523617304,-1.59891584058815,-.221830227920976,1.44110338497001,-1.94468190274016,-1.23896237294843,1.29170430168928,2.19866559171935,.305038249745371,2.27090372268066,1.44679922167927,-1.50838864882935,-2.56749336202272,-2.44801058376252,-1.55963450778316,1.62602726880112,2.49586968540814,1.59012567711558,-2.4264683064468,-3.43216369764802,-2.18664125535227,2.27972534004502,3.88041880521058,.538361161120983,4.00974905232829,2.55462252942968,-2.6633713691361,-4.53343925473613,-4.32546942586304,-2.75576887572707,2.87308040304257,4.4144312457315,2.81244670509951,-4.2977488551218,5.37669458746301,3.42550741762225,-3.57132933666014,-6.0789136637154,-5.80779112496784,-3.70016024840062,3.85767398203284,5.9385809984625,3.78348685005768,-5.79724384742923,-7.27823584467216,-4.63698476184225,4.83437858766288,7.46210387346587,4.75412760808659,-7.31197439196322,9.14177937313726,5.82425364622914,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.94468190274016,4.01278501048821,.153597725402256,-3.90852233081442,-4.70963629890664,7.0795491341205,7.20177426970699,.275662948451076,-7.01465328803099,10.545740291788,10.727807619258,.410629237576671,14.1178337840267,14.3615716486884,17.5926352342147,-.0319379183519753,-.0324893117176766,-.00124359624768512,.031645154192253,.0381313330936757,-.154660030302817,-.157330164177211,-.00602214055814407,.153242313889452,-.43925822557731,-.446841815317497,-.0171038035473811,-.961797763577864,-.978402756631806,-1.79684767953686,.000465578153648587,.000473616144898061,1.81286469184544e-005,-.000461310355481611,.00319092434145473,.00324601417302388,.000124247970525588,.0123137463285655,.012526337462346,.0353622090543082,-5.86464816610367e-006,-5.96589860075151e-006,-2.28357226493352e-007,-5.19109369939388e-005,-5.28071553035454e-005,-.000252650702007359,6.85464900528793e-008,6.97299134711694e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.71235176911142,3.77644380460732,.144551272241359,-3.67832188935975,-4.43225260672409,5.99558490979395,6.09909590528773,.23345563151147,-5.94062539921618,7.62488741580146,7.7565275696118,.296897289287186,8.14599447640013,8.28663129998069,7.34718311089616,-.166073783576779,-.168940970519646,-.00646656841622802,.164551441038203,.198278882483946,-.764836159193308,-.778040701653472,-.0297811324981311,.757825163693694,-2.02218309922343,-2.05709515348012,-.0787396124921617,-4.02885967082133,-4.09841606758444,-6.68010475273478,.0125887190510936,.0128060574566476,.000490178589876773,-.0124733225023728,.0823390061849571,.0837605510092484,.00320610999159909,.299004079114613,.304166246215336,.796298871799325,-.000824565926326729,-.000838801675185275,-3.21068856415291e-005,-.00697737063211093,-.00709783170785957,-.0321260785529637,5.01144631645353e-005,5.09796661714025e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,3.30970802258292,3.36684860010831,.128873160510408,-3.27936904259757,-3.95152801322489,4.2909622271481,4.36504370184005,.167081162489141,-4.25162841944461,3.64701232940425,3.7099763100915,.142007090144459,1.49376547749553,1.51955464695027,-1.29952592474851,-.398540259080393,-.405420872037066,-.015518330446016,.394886974518158,.475825356015865,-1.67165845871656,-1.70051886866409,-.0650909105855661,1.6563349285852,-3.84755914106515,-3.91398546692074,-.149815966677047,-6.3113399865174,-6.42030234909073,-7.98777931164405,.0724976248699598,.0737492627952912,.00282290703160025,-.0718330635538726,.434860760346183,.442368430061905,.0169325753770755,1.40707187748205,1.43136432250754,3.23180916098999,-.0113956405123911,-.0115923809684749,-.0004437225879236,-.0887334889355571,-.0902654315677738,-.36808995603673,.00166206328995187,.00169075804322796,.015902832140972,-.000231576690964732},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.76195386285985,2.80963771827715,.10754475049219,-2.73663596098029,-3.29755313334501,2.32621648423849,2.36637753401768,.0905780414306488,-2.30489284002442,.160883133644535,.163660706519194,.00626445528521548,-2.12140648579363,-2.15803158670053,-2.7058281817046,-.714786270724957,-.727126724577601,-.0278322937135137,.708234065298678,.853397929051749,-2.59853682213183,-2.6433993582315,-.10118163017136,2.57471691018525,-4.7549644652192,-4.83705672356223,-.185148446522542,-5.36731919542456,-5.45998347614275,-3.25729414076559,.233201739437629,.237227859491313,.00908039168484083,-.231064057614915,1.22672710731592,1.24790598293055,.0477662072687194,3.30495822058058,3.36201679531997,5.89679444984459,-.0657431866024218,-.0668782122732034,-.00255990322490875,-.451524561772794,-.459319924425276,-1.59193561065057,.017197430813833,.0174943364955654,.145628931559507,-.00429748986298271},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,2.10349449699454,2.13981035614575,.0819057095351816,-2.08421247060183,-2.51140504655501,.504794589391635,.513509634087683,.0196556535222261,-.500167307151006,-1.67800190212113,-1.70697182868444,-.0653379111049,-1.79846700964276,-1.82951670999345,.309494957156941,-1.09494796295235,-1.11385173231848,-.042635000914322,1.08491094310714,1.30728073869584,-3.24473228828999,-3.30075109017009,-.126343140340603,3.21498892012255,-4.05017972204568,-4.12010419835037,-.157705591526285,-1.63294619261043,-1.66113825253464,2.28560008620039,.547225517982271,.55667311317238,.0213078258129713,-.542209285918405,2.39319484717323,2.43451224831938,.0931860398476283,4.87179403320196,4.95590330339708,5.49730994822615,-.236321146004056,-.240401121148633,-.00920185489986014,-1.36209257652082,-1.38560847465611,-3.77635721863404,.0946962283839159,.096331114928899,.676667877319681,-.0362494356624792},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,1.37570022553441,1.39945105335743,.053566911271267,-1.36308964438237,-1.64247659972031,-.821946062819915,-.836136580756822,-.0320048735946572,.814411559679092,-1.61733246050546,-1.64525495722615,-.0629755689776007,.359308464240942,.365511758540104,1.71165917287845,-1.51514018159912,-1.54129837497396,-.0589964137232844,1.50125140095791,1.80895681150725,-3.36446686788906,-3.42255283189436,-.131005356341752,3.33362593315312,-1.91964817870762,-1.95279001506978,-.0747471155162795,1.95173675811295,1.98543258954328,2.75500656898115,1.04781596872157,1.06590602625894,.0407997789063127,-1.03821099255249,3.55510980002442,3.61648712494209,.138428596350638,4.59045028887451,4.66970228947898,1.24891795372526,-.626153181688296,-.636963426396325,-.0243811052053825,-2.84473124098146,-2.89384420968737,-5.48872263996485,.347192252628353,.353186366143025,1.97345041935918,-.183906859961415},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,.624299774465593,.635078021192655,.0243089373722241,-.618577028461072,-.745364252864082,-1.43546196207946,-1.46024454775272,-.0558939092532125,1.42230356501297,-.337640235075449,-.343469437534987,-.0131470099270511,1.50389998701775,1.52986412408779,-.0107903422167659,-1.94896143440088,-1.98260935074574,-.0758885128327108,1.93109596018978,2.32690486658695,-2.83314651369184,-2.88205947758122,-.110316844590437,2.80717601371157,.637278791777431,.648281114842891,.0248143133691988,2.63766838605638,2.68320650944122,-.808065843843811,1.73374759666501,1.76367994615688,.067508532733652,-1.71785491622659,4.12740556141283,4.19866330771593,.160712605397524,1.98940127224618,2.0237473648752,-2.96832239843812,-1.33269841309056,-1.35570683401508,-.0518925099589806,-4.37531622836174,-4.45085403872017,-4.52786743647192,.95054294779362,.966953631749015,3.96849611187727,-.647664476709676},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.103494496994541,-.105281281595665,-.00402986089169054,.102545797758388,.123564193970618,-1.29720437850508,-1.31960001105773,-.05051044592641,1.28531334221177,.991754894783212,1.00887708423625,.0386168770555524,.589436063368126,.599612404132423,-1.41318416081787,-2.36915365304765,-2.41005599340123,-.0922499256416731,2.34743641804054,2.82858093939836,-1.68415540504794,-1.71323156899255,-.0655775157362815,1.66871731961812,2.32831086437971,2.36850807266163,.0906595922460986,.381749900992996,.388340636274758,-2.46800676549626,2.56192161158569,2.60615200198177,.0997559098622357,-2.53843733881177,3.58714638367868,3.64907680538234,.139676034420511,-1.52344991306454,-1.54975157103559,-2.69130383032704,-2.39387697960251,-2.4352061570422,-.0932126006787377,-4.93737708601551,-5.02261861703307,-.508166439915618,2.07554270972356,2.11137599376849,5.63164748396134,-1.71909544841957},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-.761953862859853,-.775108643727059,-.0296689018486986,.754969288136846,.90971228076062,-.55104212557113,-.560555612551141,-.0214564365862766,.545990907719198,1.37646295656649,1.40022695273267,.0535966104577377,-1.03415883911242,-1.05201311272283,-.174795669385988,-2.74931534527504,-2.7967810011421,-.107052632842481,2.72411329584901,3.28246374904245,-.106757773657142,-.108600897231078,-.00415692611322429,.105779161098541,2.26282036321899,2.30188690834251,.0881095280677891,-2.15044627554921,-2.18757273479106,.119501749587716,3.45007465931297,3.50963860084301,.134338746031885,-3.4184489866057,1.77035093065653,1.8009152199194,.0689337905617278,-3.42380239520456,-3.48291275963918,1.33170919606025,-3.74106981000851,-3.80565764777524,-.145669493162303,-3.58481468643821,-3.64670485341609,3.59724489879787,3.76406389114815,3.82904871171721,5.23559759098235,-3.6179005979755},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.30970802258292,-1.33231952555822,-.0509973118669166,1.29770236975413,1.5636871606405,.519224820053661,.528188995996034,.02021753675176,-.51446526074138,.518731848959601,.527687514356292,.0201983414868359,-1.30425364587756,-1.32677098129081,1.47252576661999,-3.06556135691961,-3.11848685368264,-.119366596109979,3.03746038662953,3.66003632207834,1.5947619510804,1.62229477839744,.0620967201893232,-1.58014330533832,.444911232504622,.452592418790165,.0173239199033348,-2.09440936915946,-2.13056837715884,2.63186985254377,4.2894282422279,4.36348323472717,.167021432333119,-4.25010849784966,-.964148801340396,-.98079438405867,-.0375419530619799,-2.01722273728898,-2.05204915464121,3.46190525493913,-5.18623489409549,-5.27577283777777,-.201941221845797,-.205483038001613,-.20903060758203,3.63947886955733,5.81833621896758,5.91878710028477,1.69243379052584,-6.23568135536041},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.71235176911142,-1.74191473005723,-.0666754235978675,1.6966552165163,2.0444117541397,1.5663330178964,1.593375031711,.0609897565211595,-1.55197496987263,-1.07607439139602,-1.09465231795021,-.0419001032349348,.363617646552178,.369895336889078,.420994417051474,-3.29802783242322,-3.35496675520006,-.128418358139767,3.26779592010948,3.93758279561025,3.07100849759875,3.12402803860063,.119578696522845,-3.04285759691611,-2.20598407348367,-2.24406936639925,-.0858964409190686,.923194488569861,.939133013954983,.507379758856801,4.96464287844228,5.05035513899607,.193312888747421,-4.91913366884929,-3.80884940873902,-3.87460743053656,-.148308690032324,2.02329330850018,2.058224531456,.0162698697481229,-6.45780747384476,-6.56929852924652,-.25145361873174,4.10479708970298,4.17566451229335,-1.32544341327556,7.79428043908953,7.92884508263617,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.43862930490481,.0550665406758457,-1.40124994253785,-1.68845845881124,-.568134589816805,-1.94468190274016,-1.97825593593812,-.0757218767587645,1.92685565797098,2.32179544632225,2.27090372268066,2.31010982325998,.0884242773068128,-2.25008711195034,-2.44801058376252,-2.49027435399315,-.0953204508907542,2.49586968540814,2.53895972010479,-2.4264683064468,-3.43216369764802,-3.49141841400203,-.13364133030831,3.40070220695543,4.09773034500053,4.00974905232829,4.07897551472765,.156131305216524,-3.97299303311744,-4.32546942586304,-4.40014665135235,-.168424801181427,4.4144312457315,4.49064435720703,-4.2977488551218,5.37669458746301,5.46952072998373,.209357327015074,-5.32740823859775,-5.80779112496784,-5.9080599501188,-.22614333130343,5.9385809984625,6.04110785209613,-5.79724384742923,-7.27823584467216,-7.40389121595289,-.28339939655034,7.46210387346587,7.59093364435534,-7.31197439196322,9.14177937313726,9.29960796197771,-8.9987004064276,-10.9691712594614},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.94468190274016,5.26027492943627,3.43230866927273,-.145635482665644,-3.67618541981628,7.0795491341205,9.44065344626821,6.15998918791291,-.261373053770343,10.545740291788,14.0628559267203,9.17595809638572,14.1178337840267,18.8262802750321,17.5926352342147,-.0319379183519753,-.0425895510116904,-.0277895142706455,.00117913035023065,.0297640501007841,-.154660030302817,-.206240781926363,-.134571297792104,.00570996311627577,-.43925822557731,-.585755477154869,-.382203141612094,-.961797763577864,-1.28256746328293,-1.79684767953686,.000465578153648587,.000620853379055167,.000405104384309337,-1.71888889439261e-005,.00319092434145473,.00425513126926153,.00277645639124976,.0123137463285655,.0164205106275604,.0353622090543082,-5.86464816610367e-006,-7.82057019228999e-006,-5.10289124485106e-006,-5.19109369939388e-005,-6.922378208951e-005,-.000252650702007359,6.85464900528793e-008,9.14074675581803e-008,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.71235176911142,4.95046024541044,3.23015580842144,-.137057982126173,-3.45966893740063,5.99558490979395,7.99517571934639,5.21682066468668,-.221353691785192,7.62488741580146,10.1678678043472,6.63449368950756,8.14599447640013,10.8627695634925,7.34718311089616,-.166073783576779,-.221461142188379,-.144502522920425,.00613135259583705,.154769899538562,-.764836159193308,-1.01991708658492,-.665491881382991,.0282373296477252,-2.02218309922343,-2.69660249304008,-1.75952250506473,-4.02885967082133,-5.37252686958436,-6.68010475273478,.0125887190510936,.0167871896471046,.0109535751204635,-.000464768572193061,.0823390061849571,.109799933312678,.0716440239592627,.299004079114613,.398725093710568,.796298871799325,-.000824565926326729,-.00109956736062704,-.00071746337155817,-.00697737063211093,-.00930439734215142,-.0321260785529637,5.01144631645353e-005,6.68281652696447e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,3.30970802258292,4.41353056195817,2.87981130513507,-.122192596826728,-3.08443131194341,4.2909622271481,5.72204339147722,3.7336107743899,-.158419961451715,3.64701232940425,4.86332941380344,3.17330328508798,1.49376547749553,1.99195202144368,-1.29952592474851,-.398540259080393,-.531457639388288,-.346774016236294,.0147138868003716,.371413443660288,-1.67165845871656,-2.22917418911978,-1.45452737733324,.0617167093759977,-3.84755914106515,-5.13076069911761,-3.34780114662785,-6.3113399865174,-8.41623844737545,-7.98777931164405,.0724976248699598,.0966763475074085,.0630809359431451,-.002676572370981,.434860760346183,.579891411113903,.378376861425818,1.40707187748205,1.87634519190449,3.23180916098999,-.0113956405123911,-.0151962068293123,-.00991546509750553,-.0887334889355571,-.118327043544488,-.36808995603673,.00166206328995187,.00221637892966828,.015902832140972,-.000231576690964732},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.76195386285985,3.68309461175287,2.40320472508569,-.101969815015609,-2.57396027644145,2.32621648423849,3.10203421893763,2.02406510925841,-.0858826310401558,.160883133644535,.214539355880419,.139986084607739,-2.12140648579363,-2.82891792677208,-2.7058281817046,-.714786270724957,-.953175031760129,-.621942953572118,.0263895153231816,.666134033494215,-2.59853682213183,-3.46517626442722,-2.26101386266496,.0959365479342401,-4.7549644652192,-6.34079527029531,-4.13734393811777,-5.36731919542456,-7.15737676442222,-3.25729414076559,.233201739437629,.31097697947545,.202911254178074,-.00860967974831002,1.22672710731592,1.63585353791591,1.06738798948579,3.30495822058058,4.4071966513386,5.89679444984459,-.0657431866024218,-.0876692328212839,-.0572038290823835,-.451524561772794,-.602112766253303,-1.59193561065057,.017197430813833,.0229329553998579,.145628931559507,-.00429748986298271},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,2.10349449699454,2.80503210133656,1.8302723960547,-.0776598579828518,-1.9603192326214,.504794589391635,.673148909586561,.4392270120399,-.0186367381391766,-1.67800190212113,-2.23763323781713,-1.46004687328674,-1.79846700964276,-2.39827472875065,.309494957156941,-1.09494796295235,-1.46012465838812,-.95272558802182,.0404248755605102,1.02041985541771,-3.24473228828999,-4.32688473530227,-2.82327524550585,.11979372855449,-4.05017972204568,-5.4009573811209,-3.52410341735765,-1.63294619261043,-2.17755097302669,2.28560008620039,.547225517982271,.729731000653746,.476146603536492,-.0202032646552815,2.39319484717323,3.19135057367101,2.08234367785444,4.87179403320196,6.49658873611033,5.49730994822615,-.236321146004056,-.315136741011867,-.205625482170818,-1.36209257652082,-1.81636481949508,-3.77635721863404,.0946962283839159,.126278419466953,.676667877319681,-.0362494356624792},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,1.37570022553441,1.83451076290121,1.19701104597108,-.0507900944331572,-1.28206259359836,-.821946062819915,-1.09607374475429,-.71518380112878,.0303457958132302,-1.61733246050546,-2.15672989741719,-1.40725776236676,.359308464240942,.479141627283131,1.71165917287845,-1.51514018159912,-2.02045541424857,-1.31833919902212,.0559381407795162,1.41201150863478,-3.36446686788906,-4.4865520602846,-2.92745754610214,.124214263271281,-1.91964817870762,-2.55987109498181,-1.67030580640284,1.95173675811295,2.60266155489591,2.75500656898115,1.04781596872157,1.39727364720762,.911715550981012,-.0386847883193889,3.55510980002442,4.74077641992773,3.0933379389912,4.59045028887451,6.12141388513833,1.24891795372526,-.626153181688296,-.834981872710088,-.544822382907866,-2.84473124098146,-3.79347911889772,-5.48872263996485,.347192252628353,.462984531288115,1.97345041935918,-.183906859961415},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,.624299774465593,.832510334938,.543209713978403,-.0230488037373032,-.581806539810158,-1.43546196207946,-1.91420366785489,-1.24901035342107,.0529964648160446,-.337640235075449,-.450246815259374,-.293784273363787,1.50389998701775,2.00546649679692,-.0107903422167659,-1.94896143440088,-2.59896063091726,-1.69581157410878,.0719545824309805,1.81630452989169,-2.83314651369184,-3.77803076303363,-2.46515019062261,.104598208499711,.637278791777431,.849817990926997,.55450289173293,2.63766838605638,3.51735861632833,-.808065843843811,1.73374759666501,2.31197070864005,1.50855178059943,-.064009006140712,4.12740556141283,5.50393885468463,3.59129842139725,1.98940127224618,2.65288758284874,-2.96832239843812,-1.33269841309056,-1.77716739172311,-1.15959472274786,-4.37531622836174,-5.8345303456991,-4.52786743647192,.95054294779362,1.26755904782396,3.96849611187727,-.647664476709676},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.103494496994541,-.138011003497348,-.0900516361052144,.00382095981239145,.0964500992128822,-1.29720437850508,-1.72983572180117,-1.12871099482768,.0478920709992718,.991754894783212,1.32251561427285,.862936228716218,.589436063368126,.786019208253834,-1.41318416081787,-2.36915365304765,-3.15929138677771,-2.06142518510908,.0874678476499865,2.20789618310877,-1.68415540504794,-2.24583899888368,-1.46540110005888,.0621780897499608,2.32831086437971,3.10482709694771,2.02588745115913,.381749900992996,.50906751987416,-2.46800676549626,2.56192161158569,3.41634949355459,2.22915458763891,-.0945847345273614,3.58714638367868,4.78349754217328,3.12121330296807,-1.52344991306454,-2.03153653003234,-2.69130383032704,-2.39387697960251,-3.19226020392741,-2.08293720858961,-4.93737708601551,-6.58404442855625,-.508166439915618,2.07554270972356,2.7677580975819,5.63164748396134,-1.71909544841957},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-.761953862859853,-1.01607351391366,-.662983965136208,.0281309168451488,.710091143032932,-.55104212557113,-.734820486906378,-.479467473322989,.0203441716962559,1.37646295656649,1.83552787296833,1.19767470667912,-1.03415883911242,-1.37906172093203,-.174795669385988,-2.74931534527504,-3.6662410134057,-2.39220781955879,.101503207887315,2.56218200503226,-.106757773657142,-.142362617365695,-.0928910470423957,.00394143819036155,2.26282036321899,3.01749482284837,1.96890348630228,-2.15044627554921,-2.86764279249033,.119501749587716,3.45007465931297,4.60071095140733,3.00194577372105,-.127374855762553,1.77035093065653,2.36078164083594,1.54040072098212,-3.42380239520456,-4.56567661189371,1.33170919606025,-3.74106981000851,-4.9887560540339,-3.25514367429686,-3.58481468643821,-4.78038820054912,3.59724489879787,3.76406389114815,5.01941890462399,5.23559759098235,-3.6179005979755},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.30970802258292,-1.74650946411896,-1.13959054518559,.0483536986562675,1.2205621785349,.519224820053661,.692391774386145,.451782905525935,-.0191694943053237,.518731848959601,.691734392719743,.451353966399164,-1.30425364587756,-1.73923599488787,1.47252576661999,-3.06556135691961,-4.08795840577754,-2.66737675689461,.113178836410125,2.85690259486619,1.5947619510804,2.12663188500128,1.38761892782617,-.0588777326765029,.444911232504622,.593293821521023,.387121881618236,-2.09440936915946,-2.79291698671264,2.63186985254377,4.2894282422279,5.71999780817022,3.73227603891909,-.158363327640683,-.964148801340396,-1.28570259705192,-.83891587989734,-2.01722273728898,-2.68998779997356,3.46190525493913,-5.18623489409549,-6.91589893787678,-4.51259681489188,-.205483038001613,-.274013798814207,3.63947886955733,5.81833621896758,7.758812723563,1.69243379052584,-6.23568135536041},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.71235176911142,-2.28343914757123,-1.48993504847196,.0632190839557129,1.59579980399212,1.5663330178964,2.0887216012301,1.36288261753995,-.0578281520973947,-1.07607439139602,-1.43495655249725,-.936303497985114,.363617646552178,.484887967351202,.420994417051474,-3.29802783242322,-4.39795490297745,-2.86964825021048,.12176137061466,3.07354613898791,3.07100849759875,4.09522222779324,2.67211637191123,-.113379941906936,-2.20598407348367,-2.94170303097044,-1.91944964063812,.923194488569861,1.23108958885198,.507379758856801,4.96464287844228,6.62040364808344,4.31978730278895,-.183291880031401,-3.80884940873902,-5.07914086475798,-3.31411940634736,2.02329330850018,2.6980829707225,.0162698697481229,-6.45780747384476,-8.61155438605813,-5.61900531555837,4.10479708970298,5.47379022059324,-1.32544341327556,7.79428043908953,10.3937551846319,-4.0355398121645,-8.98680735137036},
			{1.414213562,1.88586870354722,1.2305219003314,-.0522119856161046,-1.31795450312975,-1.96685998123013,-1.94468190274016,-2.59325383159706,-1.69208790932324,.071796584495184,1.81231628640776,2.27090372268066,3.02827406795474,1.97593690127583,-.0838405143563175,-2.44801058376252,-3.26444793765179,-2.13003942014441,2.49586968540814,3.32826855353781,-2.4264683064468,-3.43216369764802,-4.57682649415414,-2.98636125886026,.126713592860266,3.19855198842569,4.00974905232829,5.34704266035319,3.48892427303921,-.148037726033597,-4.32546942586304,-5.76805910494959,-3.76363584617939,4.4144312457315,5.8866906319401,-4.2977488551218,5.37669458746301,7.16987894860209,4.67831777200947,-.198504602127589,-5.80779112496784,-7.74475072947788,-5.05341934262509,5.9385809984625,7.91916040788233,-5.79724384742923,-7.27823584467216,-9.70560427278213,-6.33286855712842,7.46210387346587,9.95079423294563,-7.31197439196322,9.14177937313726,12.190659227474,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.94468190274016,6.17728122834917,6.40499151390393,5.12910125078457,2.75697628631375,7.0795491341205,11.0864112806424,11.4950845847679,9.20523510120587,10.545740291788,16.5143870041822,17.1231492859125,14.1178337840267,22.1082033358775,17.5926352342147,-.0319379183519753,-.0500140463222003,-.0518576912013894,-.0415275099413171,-.0223217196472688,-.154660030302817,-.24219405407586,-.251121942525088,-.201097826062499,-.43925822557731,-.687868288629403,-.713224862474736,-.961797763577864,-1.50615319946369,-1.79684767953686,.000465578153648587,.000729084691473652,.000755960606624685,.000605371370823812,.00319092434145473,.00499691420436607,.00518111316265944,.0123137463285655,.0192830438371037,.0353622090543082,-5.86464816610367e-006,-9.18390428102533e-006,-9.52244633830417e-006,-5.19109369939388e-005,-8.12913346532421e-005,-.000252650702007359,6.85464900528793e-008,1.07342228505649e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.71235176911142,5.81345757700538,6.02775639811866,4.82701230968076,2.59459851167859,5.99558490979395,9.38894821107085,9.73504871109302,7.79580275781954,7.62488741580146,11.9403985081938,12.380551947663,8.14599447640013,12.7564401885048,7.34718311089616,-.166073783576779,-.260067729310804,-.269654486884813,-.215938641412748,-.116070571566627,-.764836159193308,-1.1977158528687,-1.24186670359493,-.994483762912882,-2.02218309922343,-3.16669227065688,-3.28342459824587,-4.02885967082133,-6.30910167728468,-6.68010475273478,.0125887190510936,.0197136447951608,.0204403398568298,.0163685732441572,.0823390061849571,.128940991873039,.13369408456444,.299004079114613,.468233518150735,.796298871799325,-.000824565926326729,-.00129125129526923,-.00133885010037302,-.00697737063211093,-.0109264021013695,-.0321260785529637,5.01144631645353e-005,7.84780978908706e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,3.30970802258292,5.18292672091406,5.37398257218605,4.30347185829345,2.31318690780744,4.2909622271481,6.71954825634174,6.96724788777657,5.57935475310453,3.64701232940425,5.71113751622596,5.92166456181188,1.49376547749553,2.33920241715482,-1.29952592474851,-.398540259080393,-.624104888716883,-.647110981340287,-.518204861963161,-.278543636855161,-1.67165845871656,-2.61777873930361,-2.71427671758157,-2.17358603453451,-3.84755914106515,-6.02518921139496,-6.24729300320554,-6.3113399865174,-9.88341340931671,-7.98777931164405,.0724976248699598,.113529614838378,.11771460504762,.0942655625163293,.434860760346183,.680981959385669,.706084685269139,1.40707187748205,2.20344223207683,3.23180916098999,-.0113956405123911,-.0178453112143479,-.0185031347484812,-.0887334889355571,-.138954604977606,-.36808995603673,.00166206328995187,.00260275292466808,.015902832140972,-.000231576690964732},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.76195386285985,4.32515629175549,4.4845925449966,3.59125053981238,1.9303562344299,2.32621648423849,3.64280156596183,3.77708449255997,3.02467985291779,.160883133644535,.251939290930609,.261226413581034,-2.12140648579363,-3.32207381553389,-2.7058281817046,-.714786270724957,-1.11933887677121,-1.16060055303996,-.929406031924798,-.499571029238743,-2.59853682213183,-4.06924896099705,-4.21925182103647,-3.37876634868698,-4.7549644652192,-7.44616510000515,-7.72064967553201,-5.36731919542456,-8.40509854032022,-3.25729414076559,.233201739437629,.365188565907959,.378650344801755,.303222252967675,1.22672710731592,1.9210264648999,1.99184038309353,3.30495822058058,5.17548864064312,5.89679444984459,-.0657431866024218,-.102952319693436,-.106747403902083,-.451524561772794,-.707077089737688,-1.59193561065057,.017197430813833,.026930781528047,.145628931559507,-.00429748986298271},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,2.10349449699454,3.29402405329413,3.41545015161673,2.73508397421313,1.47015262309921,.504794589391635,.790496728565174,.819636447612691,.656362825532386,-1.67800190212113,-2.62771242593193,-2.72457658514876,-1.79846700964276,-2.81635801570608,.309494957156941,-1.09494796295235,-1.7146633520688,-1.77787020176478,-1.42371402906152,-.765269708188692,-3.24473228828999,-5.08117621439183,-5.26848128513792,-4.21898669004741,-4.05017972204568,-6.34248838569497,-6.57628863172099,-1.63294619261043,-2.55715621899877,2.28560008620039,.547225517982271,.856942588087486,.888531670338266,.711533948435089,2.39319484717323,3.74768814487268,3.88583745543935,4.87179403320196,7.62911752450546,5.49730994822615,-.236321146004056,-.370073484889954,-.383715334269618,-1.36209257652082,-2.13300568008586,-3.77635721863404,.0946962283839159,.148292117918333,.676667877319681,-.0362494356624792},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,1.37570022553441,2.15431494568072,2.2337284697412,1.78876419479899,.961490176493115,-.821946062819915,-1.28714864891722,-1.33459622035742,-1.06874132878065,-1.61733246050546,-2.53270547426737,-2.62606743575436,.359308464240942,.562668799566708,1.71165917287845,-1.51514018159912,-2.37267471198348,-2.46013753301878,-1.97007200846434,-1.05894610873647,-3.36446686788906,-5.26867781509235,-5.46289467143736,-4.37467244553793,-1.91964817870762,-3.0061249410246,-3.1169383486426,1.95173675811295,3.05637492003247,2.75500656898115,1.04781596872157,1.64085573236143,1.70134184591416,1.36243031247303,3.55510980002442,5.56722026499827,5.77244215424071,4.59045028887451,7.1885396848779,1.24891795372526,-.626153181688296,-.980541495928355,-1.01668674800428,-2.84473124098146,-4.45478376427993,-5.48872263996485,.347192252628353,.543695090562235,1.97345041935918,-.183906859961415},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,.624299774465593,.977639103165718,1.01367736516514,.811750309157165,.436329142929636,-1.43546196207946,-2.24790045069837,-2.33076377601413,-1.86646982589538,-.337640235075449,-.528736850703693,-.548227434979558,1.50389998701775,2.35507282621716,-.0107903422167659,-1.94896143440088,-3.05202882623936,-3.16453436679059,-2.53415123836086,-1.36214797290772,-2.83314651369184,-4.43664234772504,-4.6001882618506,-3.68381930162858,.637278791777431,.997963945475626,1.03475143356997,2.63766838605638,4.13052808940267,-.808065843843811,1.73374759666501,2.71500890173164,2.81509103173735,2.25431788633232,4.12740556141283,6.4634222783235,6.70168045166518,1.98940127224618,3.11535668501805,-2.96832239843812,-1.33269841309056,-2.08697509460821,-2.16390630168515,-4.37531622836174,-6.85164468857583,-4.52786743647192,.95054294779362,1.48852841666704,3.96849611187727,-.647664476709676},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.103494496994541,-.162070004447689,-.168044316710384,-.134569470256975,-.0723333036764544,-1.29720437850508,-2.0313922514988,-2.1062745341762,-1.68669940023467,.991754894783212,1.55306537925041,1.61031531675698,.589436063368126,.923043332411645,-1.41318416081787,-2.36915365304765,-3.71004018615404,-3.8468016980446,-3.08050921776368,-1.6558243734555,-1.68415540504794,-2.63734867013603,-2.7345680458079,-2.18983527963514,2.32831086437971,3.64608131714141,3.78048514370594,.381749900992996,.597811573855878,-2.46800676549626,2.56192161158569,4.01191038094212,4.15979959635524,3.33115715545234,3.58714638367868,5.61738881893157,5.82446004858605,-1.52344991306454,-2.38568756196537,-2.69130383032704,-2.39387697960251,-3.74875634795768,-3.88694503628022,-4.93737708601551,-7.73181907803745,-.508166439915618,2.07554270972356,3.2502521959696,5.63164748396134,-1.71909544841957},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-.761953862859853,-1.19320224290905,-1.23718671009026,-.990736035856232,-.532536915007148,-.55104212557113,-.8629193076149,-.894728707041787,-.716496519828253,1.37646295656649,2.15550936517572,2.2349669193134,-1.03415883911242,-1.61946898132209,-.174795669385988,-2.74931534527504,-4.30536466145163,-4.46407134676942,-3.5748172149004,-1.92152305240544,-.106757773657142,-.16718022074295,-.173342908623152,-.138812569451129,2.26282036321899,3.54352469706742,3.67414802589246,-2.15044627554921,-3.3675494587808,.119501749587716,3.45007465931297,5.40273764744738,5.60189629155853,4.48598459696892,1.77035093065653,2.7723288820058,2.87452397106015,-3.42380239520456,-5.36159588761007,1.33170919606025,-3.74106981000851,-5.85842936705581,-6.07438596565011,-3.58481468643821,-5.61373743608465,3.59724489879787,3.76406389114815,5.89443757079016,5.23559759098235,-3.6179005979755},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.30970802258292,-2.05097267206762,-2.12657673727971,-1.7029573543373,-.915367588384684,.519224820053661,.813094138225446,.843066855241086,.675125837596017,.518731848959601,.812322157391327,.842266416897197,-1.30425364587756,-2.04243124304558,1.47252576661999,-3.06556135691961,-4.80059864950596,-4.97756091846909,-3.98601838486204,-2.14255044478903,1.5947619510804,2.49736057507455,2.58941963373659,2.07360079279396,.444911232504622,.696720767591157,.72240366631615,-2.09440936915946,-3.27979695088037,2.63186985254377,4.2894282422279,6.71714607327471,6.96475715335038,5.57736018045175,-.964148801340396,-1.50983487071459,-1.56549122161564,-2.01722273728898,-3.15892446006539,3.46190525493913,-5.18623489409549,-8.12152468439221,-8.42090472917074,-.205483038001613,-.321781716436834,3.63947886955733,5.81833621896758,9.11138083740401,1.69243379052584,-6.23568135536041},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.71235176911142,-2.68150352815894,-2.78035056321231,-2.22649780572461,-1.19677919225583,1.5663330178964,2.45284151714649,2.54325949118096,2.03663586527377,-1.07607439139602,-1.68510777367065,-1.74722512946481,.363617646552178,.56941688005877,.420994417051474,-3.29802783242322,-5.16463580891203,-5.35501741292456,-4.28828460541245,-2.30502351007756,3.07100849759875,4.80912874954551,4.9864054592391,3.99310107128126,-2.20598407348367,-3.45452037324197,-3.58186277642042,.923194488569861,1.44570135775136,.507379758856801,4.96464287844228,7.7745166798312,8.06110512843118,6.45531290809181,-3.80884940873902,-5.96457066234997,-6.18443989718936,2.02329330850018,3.16843083510823,.0162698697481229,-6.45780747384476,-10.1127781284249,-10.4855608369157,4.10479708970298,6.42801793274624,-1.32544341327556,7.79428043908953,12.2056640870222,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.21462594594878,2.29626268751801,1.83884144047735,.988407519376633,-.0657740155504593,-1.94468190274016,-3.04532717950273,-3.15758567899759,-2.52858674682842,-1.359156966891,2.27090372268066,3.55618305225695,3.68727300023485,2.95275903358443,-2.44801058376252,-3.83352833011347,-3.97484192853686,2.49586968540814,3.90847458250196,-2.4264683064468,-3.43216369764802,-5.37468949190064,-5.57281420860799,-4.46269573688388,-2.39877236199692,4.00974905232829,6.27917489029831,6.5106412373467,5.21370528571982,-4.32546942586304,-6.77358573667404,-7.02327732701018,4.4144312457315,6.91289790468793,-4.2977488551218,5.37669458746301,8.4197801888431,8.73015469072478,6.99108613109499,-5.80779112496784,-9.09486746248342,-9.43012739369096,5.9385809984625,9.29968140223042,-5.79724384742923,-7.27823584467216,-11.3975501068473,-11.8176927763526,7.46210387346587,11.6854832203471,-7.31197439196322,9.14177937313726,14.315816457743,-8.9987004064276,-10.9691712594614},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.94468190274016,6.70640454840984,8.33713798160189,9.3084270396024,9.7411634880245,7.0795491341205,12.0360327285782,14.9627218217782,16.7059013133841,10.545740291788,17.9289490028036,22.2885632192208,14.1178337840267,24.0019111852095,17.5926352342147,-.0319379183519753,-.054298066631689,-.0675012177634823,-.075365210691831,-.07886883956794,-.154660030302817,-.262939511058368,-.326876043550518,-.364957592022401,-.43925822557731,-.746788570738901,-.928378137184539,-.961797763577864,-1.63516477461472,-1.79684767953686,.000465578153648587,.000791535419924791,.000984005657525685,.00109864378994577,.00319092434145473,.0054249316016862,.00674406129057573,.0123137463285655,.0209347588552154,.0353622090543082,-5.86464816610367e-006,-9.97056393673186e-006,-1.23950123700402e-005,-5.19109369939388e-005,-8.82544531019357e-005,-.000252650702007359,6.85464900528793e-008,1.16536771255129e-007,7.43273920925785e-007,-7.65372854723001e-010},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.71235176911142,6.31141709357397,7.84610513558171,8.76018813180049,9.16743767928437,5.99558490979395,10.1931711799913,12.6717489298481,14.148026649931,7.62488741580146,12.9631693790598,16.1153015818976,8.14599447640013,13.8491102054891,7.34718311089616,-.166073783576779,-.282344179950942,-.350999163642368,-.391891091702267,-.410109589767539,-.764836159193308,-1.30030781284792,-1.61649145672597,-1.80481513164276,-2.02218309922343,-3.43793955015111,-4.2739110377176,-4.02885967082133,-6.84951625494143,-6.68010475273478,.0125887190510936,.021402243516657,.0266064261611109,.029706114645781,.0823390061849571,.139985605707144,.174024591276672,.299004079114613,.508340689012668,.796298871799325,-.000824565926326729,-.00140185515919078,-.00174273111826309,-.00697737063211093,-.0118623177481531,-.0321260785529637,5.01144631645353e-005,8.52002447663434e-005,.000520025886887555,-2.90963995416334e-006},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,3.30970802258292,5.62687727016469,6.99511218988806,7.81005323374718,8.17313226782894,4.2909622271481,7.29512018698114,9.06900608163536,10.1255588665535,3.64701232940425,6.20033267056672,7.70800935885959,1.49376547749553,2.53956994188053,-1.29952592474851,-.398540259080393,-.677563431168936,-.842320170000573,-.940451730878478,-.984172086870924,-1.67165845871656,-2.84200809304711,-3.53307252095549,-3.94468075935688,-3.84755914106515,-6.54128488547504,-8.13186772631248,-6.3113399865174,-10.7299904588883,-7.98777931164405,.0724976248699598,.123254146474207,.153224700235957,.1710756072316,.434860760346183,.739312383715057,.919084035475437,1.40707187748205,2.39218103581589,3.23180916098999,-.0113956405123911,-.0193738753136992,-.0240848387004708,-.0887334889355571,-.150856948291044,-.36808995603673,.00166206328995187,.00282569522207844,.015902832140972,-.000231576690964732},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.76195386285985,4.69563336286119,5.83742644431873,6.5174953654235,6.82048509559399,2.32621648423849,3.95483062658776,4.91648966501871,5.48926807103562,.160883133644535,.273519489264329,.340028655734571,-2.12140648579363,-3.60663050994097,-2.7058281817046,-.714786270724957,-1.21521735159799,-1.51071034695561,-1.68671036412374,-1.76512329607374,-2.59853682213183,-4.41780594598826,-5.49204234459095,-6.13187350041211,-4.7549644652192,-8.08397637227704,-10.049681020309,-5.36731919542456,-9.12504854432544,-3.25729414076559,.233201739437629,.396469282990198,.492874996717867,.550295671800536,1.22672710731592,2.0855745668007,2.59270415505294,3.30495822058058,5.61880207113837,5.89679444984459,-.0657431866024218,-.111770838907705,-.138949104550336,-.451524561772794,-.767642727504649,-1.59193561065057,.017197430813833,.0292375737894311,.145628931559507,-.00429748986298271},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,2.10349449699454,3.57617810764411,4.44576376432321,4.96370189948079,5.19445782869083,.504794589391635,.858207787533309,1.06689011900532,1.19118441501636,-1.67800190212113,-2.8527926629451,-3.54647947347266,-1.79846700964276,-3.0575969440733,.309494957156941,-1.09494796295235,-1.86153514438243,-2.31418717000874,-2.58379343998139,-2.70391057656362,-3.24473228828999,-5.51641118734948,-6.85778510905627,-7.65672734172536,-4.05017972204568,-6.88576274670089,-8.56010903534823,-1.63294619261043,-2.776192869346,2.28560008620039,.547225517982271,.930345156393643,1.15656845454869,1.29131041119438,2.39319484717323,4.06870140628397,5.05804933070526,4.87179403320196,8.28259983241363,5.49730994822615,-.236321146004056,-.401772626268325,-.499467904136599,-1.36209257652082,-2.31571114636377,-3.77635721863404,.0946962283839159,.16099427844353,.676667877319681,-.0362494356624792},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,1.37570022553441,2.33884568572274,2.90756083364644,3.24629602423866,3.39721202820788,-.821946062819915,-1.3974010951554,-1.73719400169233,-1.93957970348912,-1.61733246050546,-2.74964776323587,-3.41825379680027,.359308464240942,.610864951290378,1.71165917287845,-1.51514018159912,-2.57590935107797,-3.20226904625419,-3.57533818439403,-3.74155092353066,-3.36446686788906,-5.7199734894835,-7.11084574520705,-7.93926992086965,-1.91964817870762,-3.26361861024795,-4.05720211125474,1.95173675811295,3.31817281869637,2.75500656898115,1.04781596872157,1.78140543388112,2.21457307046673,2.47257414898136,3.55510980002442,6.04408799389592,7.5137721301034,4.59045028887451,7.80428370621122,1.24891795372526,-.626153181688296,-1.06453109468611,-1.32338312764,-2.84473124098146,-4.83636425064191,-5.48872263996485,.347192252628353,.590266023758336,1.97345041935918,-.183906859961415},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,.624299774465593,1.06138009357331,1.31946592651414,1.47318568258101,1.5416721344202,-1.43546196207946,-2.44044738286343,-3.03386806383218,-3.38731823500441,-.337640235075449,-.574026515774929,-.71360715479518,1.50389998701775,2.55679975214599,-.0107903422167659,-1.94896143440088,-3.31345445440235,-4.11915606854152,-4.59904391748785,-4.81284738096772,-2.83314651369184,-4.81666890668704,-5.98789306701478,-6.68549157438517,.637278791777431,1.08344588754871,1.34689725343212,2.63766838605638,4.48433402043889,-.808065843843811,1.73374759666501,2.94756663562366,3.66429874441148,4.09119503456581,4.12740556141283,7.01705539432373,8.72332687917668,1.98940127224618,3.38220674597772,-2.96832239843812,-1.33269841309056,-2.26573774926528,-2.81667592802333,-4.37531622836174,-7.4385315165314,-4.52786743647192,.95054294779362,1.61603031737652,3.96849611187727,-.647664476709676},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.103494496994541,-.175952328348057,-.218737004162641,-.244220192661119,-.255573666062748,-1.29720437850508,-2.20539388307836,-2.74165881101347,-3.06106616679168,.991754894783212,1.68609528016685,2.09608724079431,.589436063368126,1.00210784875001,-1.41318416081787,-2.36915365304765,-4.02782866109789,-5.00723794478697,-5.59058866190049,-5.85048772793475,-1.68415540504794,-2.86325431258853,-3.55948498424847,-3.97417031416819,2.32831086437971,3.95839130925952,4.9209161654707,.381749900992996,.649017927110068,-2.46800676549626,2.56192161158569,4.3555559528497,5.41465560654381,6.0454710918866,3.58714638367868,6.09855380269565,7.58148187830963,-1.52344991306454,-2.59003683420207,-2.69130383032704,-2.39387697960251,-4.06986110773887,-5.05949102727521,-4.93737708601551,-8.39409842544753,-.508166439915618,2.07554270972356,3.5286569130923,5.63164748396134,-1.71909544841957},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-.761953862859853,-1.29540758356514,-1.61039968415816,-1.79801365860383,-1.88160093296591,-.55104212557113,-.936833819859258,-1.1646349055287,-1.3003166154948,1.37646295656649,2.34014241482376,2.90917287715098,-1.03415883911242,-1.75818677195786,-.174795669385988,-2.74931534527504,-4.67414645388234,-5.81071476784009,-6.48767173775814,-6.78927500842464,-.106757773657142,-.181500267083405,-.225633983149896,-.251920680005669,2.26282036321899,3.84705006415361,4.78250111498198,-2.15044627554921,-3.65600143048661,.119501749587716,3.45007465931297,5.86551639682903,7.29177895707793,8.14128212331116,1.77035093065653,3.00979643579225,3.74166037942858,-3.42380239520456,-5.82085058429709,1.33170919606025,-3.74106981000851,-6.36024100270089,-7.90680110023493,-3.58481468643821,-6.09458965574696,3.59724489879787,3.76406389114815,6.39933353748721,5.23559759098235,-3.6179005979755},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.30970802258292,-2.22665149086864,-2.76808542972749,-3.09057152692752,-3.23424810520085,.519224820053661,.882740808667662,1.09738860459098,1.22523601982934,.518731848959601,.881902702778381,1.09634670392581,-1.30425364587756,-2.21737843427209,1.47252576661999,-3.06556135691961,-5.21180037431139,-6.47910494479514,-7.2339303710034,-7.57022621762745,1.5947619510804,2.71127534923244,3.37055072311448,3.76322492874004,.444911232504622,.756399320675517,.940325842964601,-2.09440936915946,-3.56073236366527,2.63186985254377,4.2894282422279,7.29251224169723,9.06576398575723,10.1219390639593,-.964148801340396,-1.63916180436132,-2.03774139204767,-2.01722273728898,-3.42950637731078,3.46190525493913,-5.18623489409549,-8.81718478891181,-10.9611768422514,-.205483038001613,-.349344361529711,3.63947886955733,5.81833621896758,9.89182840014397,1.69243379052584,-6.23568135536041},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.71235176911142,-2.91119131427792,-3.61907837542113,-4.04070642498082,-4.22855351665629,1.5663330178964,2.66294295160588,3.31046579140125,3.6961400119053,-1.07607439139602,-1.82944794322174,-2.27429762516448,.363617646552178,.61819104768005,.420994417051474,-3.29802783242322,-5.60701962552938,-6.97042595115334,-7.78249101017962,-8.14428871473084,3.07100849759875,5.22106113152625,6.4906175528327,7.24678421547911,-2.20598407348367,-3.75042195543293,-4.66237685613718,.923194488569861,1.5695348492102,.507379758856801,4.96464287844228,8.44045334766828,10.4928391542811,11.7152706263275,-3.80884940873902,-6.47547397319738,-8.05005418818497,2.02329330850018,3.43982703367553,.0162698697481229,-6.45780747384476,-10.9790017210618,-13.6486625036235,4.10479708970298,6.97861844197597,-1.32544341327556,7.79428043908953,13.2511566373098,-4.0355398121645,-8.98680735137036},
			{1.414213562,2.40432270571992,2.98895928687546,3.33717751886114,3.49231848196783,3.47478060833006,-1.94468190274016,-3.30617876911379,-4.11011122144132,-4.58894533278273,-4.80227932539642,2.27090372268066,3.8607926877578,4.79958540336237,5.35874428788927,-2.44801058376252,-4.16189434787989,-5.17390312560615,2.49586968540814,4.24326022217388,-2.4264683064468,-3.43216369764802,-5.83506573884863,-7.25392389703223,-8.09901689119005,-8.47552946493044,4.00974905232829,6.81702604882233,8.4746582765388,9.46196865399931,-4.32546942586304,-7.35378632026677,-9.14193750783762,4.4144312457315,7.50503145913621,-4.2977488551218,5.37669458746301,9.14098776917085,11.3637159548613,12.6876058781097,-5.80779112496784,-9.87390054991614,-12.2748442514899,5.9385809984625,10.0962581027492,-5.79724384742923,-7.27823584467216,-12.3738225688191,-15.3826488429769,7.46210387346587,12.6864189798666,-7.31197439196322,9.14177937313726,15.5420569434015,-8.9987004064276,-10.9691712594614}
		};
		for (int j = 0; j < 144; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[56][144] = 
		{
			{-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.0058767981695121,-.0133218071347991,-.0199415226492611,-.0253092817487325,-.0290869878703852,-.0310371252151211,-.0310371252151211,-.0290869878703852,-.0253092817487325,-.0199415226492611,-.0133218071347991,-.0058767981695121,-.00550754478349754,-.0124847659007004,-.0186885487426998,-.0237190386071866,-.0272593823528392,-.0290869878703852,-.0290869878703852,-.0272593823528392,-.0237190386071866,-.0186885487426998,-.0124847659007004,-.00550754478349754,-.00479224604797332,-.0108632925195224,-.0162613519045569,-.0206385010917392,-.0237190386071866,-.0253092817487325,-.0253092817487325,-.0237190386071866,-.0206385010917392,-.0162613519045569,-.0108632925195224,-.00479224604797332,-.00377587495588562,-.00855933392240393,-.0128125373343938,-.0162613519045569,-.0186885487426998,-.0199415226492611,-.0199415226492611,-.0186885487426998,-.0162613519045569,-.0128125373343938,-.00855933392240393,-.00377587495588562,-.00252244920371164,-.00571800848521633,-.00855933392240393,-.0108632925195224,-.0124847659007004,-.0133218071347991,-.0133218071347991,-.0124847659007004,-.0108632925195224,-.00855933392240393,-.00571800848521633,-.00252244920371164,-.00111275630348505,-.00252244920371164,-.00377587495588562,-.00479224604797332,-.00550754478349754,-.0058767981695121,-.0058767981695121,-.00550754478349754,-.00479224604797332,-.00377587495588562,-.00252244920371164,-.00111275630348505},
			{.0018918113337895,.00428844840289335,.00641941367947712,.00814735927306806,.00936344787276585,.00999122031360907,.00999122031360907,.00936344787276585,.00814735927306806,.00641941367947712,.00428844840289335,.0018918113337895,.00395009749714678,.00895426991072093,.0134037202629815,.0170116664902499,.0195508565501554,.0208616438908705,.0208616438908705,.0195508565501554,.0170116664902499,.0134037202629815,.00895426991072092,.00395009749714678,.0050351690836213,.0114139620739561,.0170856536889082,.0216846842981607,.0249213768851943,.0265922308066324,.0265922308066324,.0249213768851943,.0216846842981607,.0170856536889082,.0114139620739561,.0050351690836213,.00487498195709245,.0110508422349653,.0165420966158804,.0209948152570779,.0241285368263833,.0257462347794558,.0257462347794558,.0241285368263833,.0209948152570779,.0165420966158804,.0110508422349653,.00487498195709245,.0035088724497807,.00795407987279395,.0119065275703486,.0151114669740713,.0173670300462221,.0185314027207393,.0185314027207393,.0173670300462221,.0151114669740713,.0119065275703486,.00795407987279395,.0035088724497807,.00127473997181245,.00288964152956687,.00432552816798997,.00548985215628924,.00630927675725883,.00673228227015115,.00673228227015115,.00630927675725883,.00548985215628924,.00432552816798997,.00288964152956687,.00127473997181245,-.00127473997181245,-.00288964152956687,-.00432552816798997,-.00548985215628924,-.00630927675725883,-.00673228227015115,-.00673228227015115,-.00630927675725883,-.00548985215628924,-.00432552816798997,-.00288964152956687,-.00127473997181245,-.0035088724497807,-.00795407987279395,-.0119065275703486,-.0151114669740713,-.0173670300462221,-.0185314027207393,-.0185314027207393,-.0173670300462221,-.0151114669740713,-.0119065275703486,-.00795407987279395,-.0035088724497807,-.00487498195709245,-.0110508422349653,-.0165420966158804,-.0209948152570779,-.0241285368263833,-.0257462347794558,-.0257462347794558,-.0241285368263833,-.0209948152570779,-.0165420966158804,-.0110508422349653,-.00487498195709245,-.0050351690836213,-.0114139620739561,-.0170856536889082,-.0216846842981607,-.0249213768851943,-.0265922308066324,-.0265922308066324,-.0249213768851943,-.0216846842981607,-.0170856536889082,-.0114139620739561,-.0050351690836213,-.00395009749714678,-.00895426991072092,-.0134037202629815,-.0170116664902499,-.0195508565501554,-.0208616438908705,-.0208616438908705,-.0195508565501554,-.0170116664902499,-.0134037202629815,-.00895426991072093,-.00395009749714678,-.0018918113337895,-.00428844840289335,-.00641941367947713,-.00814735927306806,-.00936344787276585,-.00999122031360907,-.00999122031360907,-.00936344787276585,-.00814735927306806,-.00641941367947713,-.00428844840289335,-.0018918113337895},
			{-.00235182533596431,-.00533123014211866,-.00798036224002773,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002773,-.00533123014211866,-.00235182533596431,-.00409570813081086,-.0092843470585992,-.013897815459963,-.0176387597554457,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554457,-.013897815459963,-.0092843470585992,-.00409570813081086,-.00328542799229539,-.00744756529085685,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480003,-.0173513258480003,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085686,-.00328542799229539,-.000186600113929722,-.000422994061970596,-.000633183289769704,-.000803620393549264,-.000923570130184837,-.000985490897277736,-.000985490897277736,-.000923570130184837,-.000803620393549264,-.000633183289769704,-.000422994061970597,-.000186600113929722,.00365824316969289,.00829268056399677,.0124133817294243,.0157547535948168,.0181063347139945,.019320274076138,.019320274076138,.0181063347139945,.0157547535948168,.0124133817294243,.00829268056399677,.00365824316969289,.00626131850458325,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686322,.0330678918746395,.0330678918746395,.0309901565686322,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458325,.00626131850458326,.0141934562191255,.0212463013314188,.0269652742156715,.0309901565686323,.0330678918746395,.0330678918746395,.0309901565686323,.0269652742156715,.0212463013314188,.0141934562191255,.00626131850458326,.00365824316969289,.00829268056399679,.0124133817294243,.0157547535948168,.0181063347139946,.019320274076138,.019320274076138,.0181063347139946,.0157547535948168,.0124133817294243,.00829268056399678,.00365824316969289,-.000186600113929713,-.000422994061970576,-.000633183289769673,-.000803620393549226,-.000923570130184793,-.000985490897277688,-.000985490897277688,-.000923570130184793,-.000803620393549226,-.000633183289769673,-.000422994061970576,-.000186600113929713,-.00328542799229539,-.00744756529085684,-.011148321727427,-.0141491710832533,-.0162611002461658,-.0173513258480002,-.0173513258480002,-.0162611002461658,-.0141491710832533,-.011148321727427,-.00744756529085684,-.00328542799229539,-.00409570813081086,-.00928434705859918,-.013897815459963,-.0176387597554456,-.0202715508147907,-.0216306571084982,-.0216306571084982,-.0202715508147907,-.0176387597554456,-.013897815459963,-.00928434705859918,-.00409570813081086,-.00235182533596432,-.00533123014211866,-.00798036224002774,-.010128476142081,-.0116402695902245,-.0124206915621332,-.0124206915621332,-.0116402695902245,-.010128476142081,-.00798036224002774,-.00533123014211866,-.00235182533596432},
			{.002625816508753,.00595232643558409,.00891008638917558,.0113084587769704,.0129963784252772,.0138677207253597,.0138677207253597,.0129963784252772,.0113084587769704,.00891008638917558,.00595232643558409,.002625816508753,.00327983286928533,.00743488199844395,.0111293360027156,.01412507487634,.0162334073987287,.0173217763333754,.0173217763333754,.0162334073987287,.01412507487634,.0111293360027156,.00743488199844395,.00327983286928533,-.000139403223161069,-.000316005892895126,-.000473031819685113,-.000600360153589864,-.000689970923659479,-.000736230029999612,-.000736230029999612,-.000689970923659479,-.000600360153589864,-.000473031819685113,-.000316005892895126,-.000139403223161069,-.00474831713981733,-.010763712369717,-.0161122895594359,-.0204493148918112,-.0235016141556679,-.0250772800730455,-.0250772800730455,-.0235016141556679,-.0204493148918112,-.0161122895594359,-.010763712369717,-.00474831713981733,-.0062268534463654,-.0141153291769463,-.0211293523195189,-.0268168454550162,-.0308195730805799,-.0328858716152058,-.0328858716152058,-.0308195730805799,-.0268168454550162,-.0211293523195189,-.0141153291769463,-.0062268534463654,-.00284444948144652,-.00644793411369711,-.00965196559807224,-.0122500333122624,-.0140784939652993,-.0150223867108016,-.0150223867108016,-.0140784939652993,-.0122500333122624,-.00965196559807224,-.00644793411369711,-.00284444948144652,.00284444948144652,.00644793411369712,.00965196559807225,.0122500333122624,.0140784939652993,.0150223867108016,.0150223867108016,.0140784939652993,.0122500333122624,.00965196559807225,.00644793411369712,.00284444948144652,.00622685344636543,.0141153291769464,.021129352319519,.0268168454550163,.0308195730805801,.0328858716152059,.0328858716152059,.0308195730805801,.0268168454550163,.021129352319519,.0141153291769464,.00622685344636543,.00474831713981732,.0107637123697169,.0161122895594359,.0204493148918111,.0235016141556678,.0250772800730455,.0250772800730455,.0235016141556678,.0204493148918111,.0161122895594359,.0107637123697169,.00474831713981732,.000139403223161063,.000316005892895114,.000473031819685095,.000600360153589841,.000689970923659453,.000736230029999584,.000736230029999584,.000689970923659453,.000600360153589841,.000473031819685095,.000316005892895114,.000139403223161063,-.00327983286928533,-.00743488199844395,-.0111293360027156,-.01412507487634,-.0162334073987287,-.0173217763333754,-.0173217763333754,-.0162334073987287,-.01412507487634,-.0111293360027156,-.00743488199844395,-.00327983286928533,-.00262581650875299,-.00595232643558406,-.00891008638917554,-.0113084587769704,-.0129963784252772,-.0138677207253597,-.0138677207253597,-.0129963784252772,-.0113084587769704,-.00891008638917554,-.00595232643558406,-.00262581650875299},
			{-.00274788724207344,-.00622904221162262,-.00932430450984375,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984375,-.00622904221162262,-.00274788724207344,-.00176296411460538,-.00399637136462942,-.00598220115906353,-.00759246007816475,-.00872572348723494,-.00931073969132328,-.00931073969132328,-.00872572348723494,-.00759246007816475,-.00598220115906353,-.00399637136462941,-.00176296411460538,.00351886839094271,.00797673915025254,.0119404464291104,.0151545159411986,.0174165045748593,.0185841942695623,.0185841942695623,.0174165045748593,.0151545159411986,.0119404464291104,.00797673915025254,.00351886839094271,.00572156044449334,.0129699068358449,.0194147601979758,.024640728021481,.028318644685133,.0302172684545957,.0302172684545957,.028318644685133,.024640728021481,.0194147601979758,.0129699068358449,.00572156044449334,.000863900578693703,.0019583311458107,.00293144199610321,.00372051285723621,.00427584288738108,.00456251680948201,.00456251680948201,.00427584288738108,.00372051285723621,.00293144199610321,.0019583311458107,.000863900578693703,-.00559347785892409,-.012679563105626,-.0189801422806282,-.0240891218318903,-.027684704824447,-.0295408260906547,-.0295408260906547,-.027684704824447,-.0240891218318903,-.0189801422806282,-.012679563105626,-.00559347785892409,-.00559347785892416,-.0126795631056261,-.0189801422806285,-.0240891218318906,-.0276847048244473,-.029540826090655,-.029540826090655,-.0276847048244473,-.0240891218318906,-.0189801422806285,-.0126795631056261,-.00559347785892416,.000863900578693674,.00195833114581063,.00293144199610311,.00372051285723608,.00427584288738094,.00456251680948185,.00456251680948185,.00427584288738094,.00372051285723608,.00293144199610311,.00195833114581063,.000863900578693674,.00572156044449339,.012969906835845,.0194147601979759,.0246407280214812,.0283186446851333,.030217268454596,.030217268454596,.0283186446851333,.0246407280214812,.0194147601979759,.012969906835845,.00572156044449339,.00351886839094272,.00797673915025255,.0119404464291104,.0151545159411987,.0174165045748594,.0185841942695623,.0185841942695623,.0174165045748594,.0151545159411987,.0119404464291104,.00797673915025255,.00351886839094272,-.00176296411460533,-.0039963713646293,-.00598220115906337,-.00759246007816454,-.00872572348723469,-.00931073969132302,-.00931073969132302,-.00872572348723469,-.00759246007816454,-.00598220115906337,-.00399637136462931,-.00176296411460533,-.00274788724207344,-.00622904221162261,-.00932430450984374,-.0118341740548762,-.0136005628530905,-.0145124127031825,-.0145124127031825,-.0136005628530905,-.0118341740548762,-.00932430450984374,-.00622904221162261,-.00274788724207344},
			{.00273408778348797,.00619776093897419,.00927747931558112,.0117747446895576,.0135322629603741,.0144395336435885,.0144395336435885,.0135322629603741,.0117747446895576,.00927747931558112,.00619776093897419,.00273408778348797,-.000117317226705007,-.00026594029991732,-.00039808822185238,-.000505243613786289,-.000580657128546422,-.000619587290580092,-.000619587290580092,-.000580657128546422,-.000505243613786289,-.00039808822185238,-.00026594029991732,-.000117317226705007,-.00525141148444389,-.0119041506894287,-.0178194210583757,-.0226159634897508,-.0259916603811327,-.0277342714684962,-.0277342714684962,-.0259916603811327,-.0226159634897508,-.0178194210583757,-.0119041506894287,-.00525141148444389,-.00192519773143665,-.00436413028570642,-.00653270251221464,-.00829114262588265,-.00952869257156061,-.0101675438446148,-.0101675438446148,-.00952869257156061,-.00829114262588265,-.00653270251221464,-.00436413028570642,-.00192519773143665,.00561226510581592,.0127221509353366,.0190438923531344,.0241700318293333,.0277776915847029,.0296400471490286,.0296400471490286,.0277776915847029,.0241700318293333,.0190438923531344,.0127221509353366,.00561226510581592,.00424652299470629,.00962622140444603,.0144095699795955,.0182882657912371,.0210180032179986,.0224271554193144,.0224271554193144,.0210180032179986,.0182882657912371,.0144095699795955,.00962622140444603,.00424652299470629,-.00424652299470642,-.00962622140444635,-.014409569979596,-.0182882657912377,-.0210180032179993,-.0224271554193151,-.0224271554193151,-.0210180032179993,-.0182882657912377,-.014409569979596,-.00962622140444634,-.00424652299470642,-.00561226510581605,-.0127221509353369,-.0190438923531349,-.0241700318293339,-.0277776915847036,-.0296400471490293,-.0296400471490293,-.0277776915847035,-.0241700318293339,-.0190438923531349,-.0127221509353369,-.00561226510581605,.00192519773143709,.00436413028570742,.00653270251221612,.00829114262588453,.00952869257156278,.0101675438446171,.0101675438446171,.00952869257156278,.00829114262588453,.00653270251221612,.00436413028570742,.00192519773143709,.00525141148444307,.0119041506894268,.0178194210583729,.0226159634897473,.0259916603811286,.0277342714684918,.0277342714684918,.0259916603811286,.0226159634897473,.0178194210583729,.0119041506894268,.00525141148444307,.000117317226705136,.000265940299917613,.000398088221852818,.000505243613786846,.000580657128547061,.000619587290580775,.000619587290580775,.000580657128547061,.000505243613786846,.000398088221852818,.000265940299917613,.000117317226705136,-.00273408778348764,-.00619776093897344,-.00927747931557999,-.0117747446895562,-.0135322629603725,-.0144395336435867,-.0144395336435867,-.0135322629603725,-.0117747446895562,-.00927747931557999,-.00619776093897344,-.00273408778348764},
			{.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671589,-.00310382375628885,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.00519219719506989,.0103637745886009,.0118657346942092,.008761335787885,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00808117200072262,.0161302511982261,.0184679124032298,.013636204249949,.00212863407481836,-.0137012335210775,-.030191889883639,-.0432638467988466,-.0494289337434856,-.046669484180745,-.0349701315186074,-.0163922196731834,.00757341249662506,.0151167486581678,.017307528978308,.0127794086877325,.00199488686808421,-.0128403520131025,-.0282948607109257,-.0405454750832129,-.0463231947636175,-.0437371280643321,-.0327728739133335,-.0153622570227696,.00658980680744916,.0131534434785487,.0150596936760282,.0111196682345142,.00173579863360809,-.0111726964751614,-.0246200330183752,-.035279584709493,-.0403069163778037,-.0380567180759071,-.0285164590876305,-.0133670661609339,.00519219719506989,.0103637745886009,.0118657346942092,.00876133578788499,.00136765903158772,-.008803117450139,-.0193984543273762,-.0277972581176119,-.0317583601877639,-.0299853987561402,-.0224685007033512,-.0105320907660995,.00346861425053362,.00692345357404236,.00792682845182869,.00585295454425802,.000913655901077986,-.00588086651748639,-.0129590138028949,-.0185697811562436,-.0212159701533398,-.020031554566579,-.0150099387214664,-.00703589612387871,.0015301487006993,.00305422071337079,.00349685072485819,.00258197947199539,.000403051273201007,-.00259429259374503,-.00571675507871576,-.00819189579934559,-.00935923882548017,-.00883674429588778,-.00662150258671588,-.00310382375628885},
			{-.00260142552913706,-.00519252000263166,-.00594504099050645,-.00438965658114597,-.00068523266475808,.00441059027814434,.00971912899628547,.0139271476391209,.0159117625644931,.0150234628799625,.0112573018965134,.00527685077526242,-.00543177022292261,-.0108419692266112,-.012413231232108,-.00916559234130155,-.00143076414931682,.00920930185008349,.0202935178745168,.0290798506392448,.0332237218109789,.0313689542147427,.0235052191758168,.0110180516762225,-.00692384973169649,-.0138202027405243,-.0158230823854093,-.0116833336957715,-.0018237877459474,.0117390499831386,.0258680434782914,.0370679369308766,.0423501083267716,.0399858455548738,.0299619827063265,.0140446541387724,-.0067035767727117,-.0133805315937793,-.0153196923188487,-.0113116441612351,-.00176576639381999,.0113655879099196,.0250450864961337,.0358886704147581,.0410027966380503,.0387137497036944,.0290087825585011,.0135978423729273,-.00482504264831925,-.00963092357798896,-.0110266759528246,-.00814179763298104,-.00127094809920464,.00818062479896092,.0180267362591577,.0258316375293235,.0295126391755971,.0278650487239994,.0208796912105726,.00978733765555977,-.00175289208073003,-.00349882289141084,-.00400588644770123,-.00295783760558683,-.000461723350547117,.0029719431455217,.00654894174688664,.00938438810964969,.0107216609806272,.0101231070474515,.00758539313302242,.00355564663741111,.00175289208073003,.00349882289141084,.00400588644770123,.00295783760558683,.000461723350547117,-.0029719431455217,-.00654894174688664,-.00938438810964969,-.0107216609806272,-.0101231070474515,-.00758539313302242,-.00355564663741111,.00482504264831925,.00963092357798896,.0110266759528246,.00814179763298104,.00127094809920464,-.00818062479896092,-.0180267362591577,-.0258316375293235,-.0295126391755971,-.0278650487239994,-.0208796912105726,-.00978733765555977,.0067035767727117,.0133805315937793,.0153196923188487,.0113116441612351,.00176576639381999,-.0113655879099196,-.0250450864961337,-.0358886704147581,-.0410027966380503,-.0387137497036944,-.0290087825585011,-.0135978423729273,.00692384973169649,.0138202027405243,.0158230823854093,.0116833336957715,.0018237877459474,-.0117390499831386,-.0258680434782914,-.0370679369308766,-.0423501083267716,-.0399858455548738,-.0299619827063265,-.0140446541387724,.00543177022292261,.0108419692266112,.012413231232108,.00916559234130155,.00143076414931682,-.00920930185008349,-.0202935178745168,-.0290798506392448,-.0332237218109789,-.0313689542147427,-.0235052191758168,-.0110180516762225,.00260142552913706,.00519252000263166,.00594504099050645,.00438965658114597,.00068523266475808,-.00441059027814434,-.00971912899628547,-.0139271476391209,-.0159117625644931,-.0150234628799625,-.0112573018965134,-.00527685077526242},
			{.00323398975244044,.00645513634342751,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132602,-.00655997303832753,.00563199907811109,.0112416317670242,.0128708144833851,.00950345936923257,.00148350575212864,-.00954878012159535,-.0210415885190764,-.0301518078398598,-.0344484326345773,-.0325252935908768,-.0243716812928383,-.0114242050632393,.00451778467430605,.00901762784526014,.0103244989234578,.00762332921148292,.00119001431965068,-.00765968386950273,-.016878796465716,-.0241866828229713,-.02763327885035,-.0260906067053993,-.0195500756844557,-.0091640814984193,.000256593398763391,.000512167786738112,.000586393212665884,.000432976800198454,6.75883958332245e-005,-.00043504160981969,-.000958652982468414,-.00137371379952092,-.00156946765956192,-.00148184960837375,-.00111037172587836,-.000520485810578236,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163794,-.00132505164225417,.0085288693781531,.0187941242444344,.0269312757552921,.0307689755634805,.0290512481165313,.0217685278774507,.0102039790939625,-.00860992505308021,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.032167352592636,.0460946108327734,.052663080917845,.0497230798980886,.0372582357416281,.0174647665990918,-.00860992505308022,-.0171856574630098,-.0196762724100466,-.0145284244154259,-.0022679111208102,.0145977085676037,.0321673525926361,.0460946108327734,.052663080917845,.0497230798980887,.0372582357416281,.0174647665990918,-.00503044198980496,-.0100409065574155,-.0114960753228541,-.00848838936163795,-.00132505164225417,.00852886937815311,.0187941242444344,.0269312757552922,.0307689755634806,.0290512481165313,.0217685278774507,.0102039790939625,.000256593398763379,.000512167786738088,.000586393212665856,.000432976800198433,6.75883958332213e-005,-.00043504160981967,-.000958652982468369,-.00137371379952085,-.00156946765956185,-.00148184960837368,-.00111037172587831,-.000520485810578211,.00451778467430605,.00901762784526012,.0103244989234578,.00762332921148291,.00119001431965068,-.00765968386950272,-.016878796465716,-.0241866828229712,-.0276332788503499,-.0260906067053993,-.0195500756844557,-.00916408149841928,.00563199907811108,.0112416317670242,.0128708144833851,.00950345936923255,.00148350575212863,-.00954878012159534,-.0210415885190763,-.0301518078398597,-.0344484326345772,-.0325252935908767,-.0243716812928383,-.0114242050632393,.00323398975244044,.00645513634342752,.00739064079513124,.0054570481611549,.000851854258768723,-.00548307210872332,-.0120824383495077,-.0173136813801404,-.0197808764850192,-.018676577305707,-.0139946343132603,-.00655997303832753},
			{-.00361075440083214,-.00720716939267161,-.00825166151372371,-.00609280244273312,-.000951096555391269,.00612185821915875,.0134900605081838,.0193307511845923,.0220853782133051,.0208524265261428,.0156250301648324,.0073242197194988,-.00451009083357084,-.00900227071838211,-.0103069161796764,-.00761034659166953,-.00118798771118922,.00764663933727674,.016850051675698,.0241454926160516,.0275862190496304,.0260461740935889,.0195167816743332,.00914847495923774,.000191693059983865,.000382624848254669,.000438076387901722,.000323463690539438,5.049321798266e-005,-.000325006246491721,-.000716180246871757,-.00102625945578221,-.00117250116198259,-.00110704440267768,-.000829524667738866,-.000388838988800048,.00652940026539944,.013032870287285,.0149216465305082,.0110177379767064,.00171988715149418,-.0110702801253114,-.0243943494584096,-.0349561886252798,-.0399374364354906,-.0377078649444119,-.0282550583007305,-.0132445347624081,.00856253222944282,.0170910600271185,.0195679655313662,.0144484535618117,.00225542753807154,-.0145173563434682,-.031990289301271,-.0458408857714791,-.0523732001011417,-.0494493821432545,-.0370531499848233,-.0173686328234554,.00391139611196793,.00780725887484712,.00893871838932497,.00660010655391095,.00103028756760808,-.00663158159716375,-.0146132814266631,-.0209402846694332,-.0239242698021651,-.0225886590405194,-.0169260147469522,-.00793405514577347,-.00391139611196793,-.00780725887484712,-.00893871838932498,-.00660010655391096,-.00103028756760808,.00663158159716376,.0146132814266631,.0209402846694332,.0239242698021651,.0225886590405194,.0169260147469522,.00793405514577348,-.00856253222944286,-.0170910600271186,-.0195679655313663,-.0144484535618117,-.00225542753807155,.0145173563434683,.0319902893012711,.0458408857714793,.0523732001011419,.0494493821432547,.0370531499848235,.0173686328234555,-.00652940026539942,-.013032870287285,-.0149216465305081,-.0110177379767064,-.00171988715149417,.0110702801253114,.0243943494584096,.0349561886252797,.0399374364354905,.0377078649444118,.0282550583007304,.0132445347624081,-.000191693059983858,-.000382624848254655,-.000438076387901706,-.000323463690539425,-5.04932179826581e-005,.000325006246491708,.00071618024687173,.00102625945578217,.00117250116198255,.00110704440267764,.000829524667738834,.000388838988800033,.00451009083357083,.00900227071838211,.0103069161796764,.00761034659166953,.00118798771118921,-.00764663933727674,-.016850051675698,-.0241454926160516,-.0275862190496304,-.0260461740935889,-.0195167816743332,-.00914847495923773,.00361075440083212,.00720716939267157,.00825166151372367,.00609280244273308,.000951096555391264,-.00612185821915872,-.0134900605081837,-.0193307511845922,-.022085378213305,-.0208524265261427,-.0156250301648324,-.00732421971949876},
			{.00377861359217487,.00754222115521882,.00863527033203834,.00637604876123444,.00099531177496977,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.023112098204709,-.0218218282260771,-.016351417129166,-.0076647129967619,.00242424800551054,.00483886858146159,.00554013697585819,.00409068646883364,.000638562934915723,-.00411019441654455,-.00905717992717652,-.0129785994288253,-.0148280464803178,-.014000246985618,-.010490591164109,-.00491745576570479,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.00816499167123309,-.00127456872700186,.00820392944654196,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972578,.0209391724674719,.00981521944459481,-.00786770495273284,-.0157041648451191,-.0179800758935133,-.0132759990387789,-.00207240544459477,.0133393105384579,.0293943603166261,.0421210168157691,.0481232507843108,.0454366930684684,.0340463829902474,.0159592133291381,-.00118794775089702,-.00237117779829748,-.00271481846955098,-.00200454812347154,-.000312913283053141,.00201410755078885,.00443826559803358,.00635986828336871,.00726614786377036,.00686050349665649,.00514067626358997,.0024096876655053,.00769157887618114,.015352612142481,.0175775747521701,.0129788031427645,.0020260126728559,-.0130406973541611,-.0287363395359262,-.0411780976955557,-.0470459659340941,-.0444195493741242,-.0332842222466921,-.0156019511230183,.00769157887618123,.0153526121424812,.0175775747521703,.0129788031427646,.00202601267285592,-.0130406973541613,-.0287363395359265,-.0411780976955562,-.0470459659340946,-.0444195493741247,-.0332842222466924,-.0156019511230185,-.00118794775089698,-.0023711777982974,-.00271481846955089,-.00200454812347147,-.00031291328305313,.00201410755078878,.00443826559803343,.00635986828336849,.00726614786377011,.00686050349665626,.00514067626358979,.00240968766550522,-.00786770495273291,-.0157041648451192,-.0179800758935135,-.013275999038779,-.00207240544459478,.0133393105384581,.0293943603166264,.0421210168157695,.0481232507843112,.0454366930684688,.0340463829902477,.0159592133291382,-.00483878804323051,-.00965835978064839,-.0110580883208755,-.0081649916712331,-.00127456872700187,.00820392944654197,.0180780901283157,.025905225699403,.0295967136408214,.0279444295972579,.0209391724674719,.00981521944459482,.00242424800551048,.00483886858146146,.00554013697585804,.00409068646883352,.000638562934915706,-.00411019441654444,-.00905717992717627,-.0129785994288249,-.0148280464803174,-.0140002469856176,-.0104905911641087,-.00491745576570466,.00377861359217486,.00754222115521882,.00863527033203834,.00637604876123444,.000995311774969769,-.00640645530223534,-.0141171955599465,-.0202294122126437,-.0231120982047089,-.0218218282260771,-.016351417129166,-.0076647129967619},
			{-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291795,-.0362874566153928,-.0189020194535391,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133056,.0295240165836947,.0169054630708873,-.00973006417443097,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00943679454938384,-.0147547632995149,-.00732147801570537,.00986165087005545,.0266803890281736,.031503454522165,.0180388899836877,-.0103824164082817,-.0416308186818771,-.0605059396217619,-.0564780510756009,-.029419235176602,-.00884385801490132,-.0138276859776402,-.00686145191476565,.00924201958947946,.0250039959133057,.0295240165836947,.0169054630708873,-.00973006417443096,-.0390150540567466,-.0567042056783988,-.0529293990726515,-.027570753760433,-.00769525174770046,-.0120317993014813,-.00597031292789838,.00804170162822512,.0217565730846378,.0256895508534857,.0147098465423953,-.0084663608594086,-.0339479288802396,-.0493396815183418,-.0460551322773697,-.0239899702938366,-.00606319209458482,-.0094800160930631,-.00470408965599446,.0063361646035602,.0171422958282622,.0202411417787383,.0115900854439309,-.00667075930920755,-.0267480286237165,-.0388753970291794,-.0362874566153928,-.0189020194535391,-.00405047684301532,-.00633306434256954,-.00314253711936504,.00423283438820543,.0114518014942175,.0135219657850183,.00774268272666788,-.00445635824921431,-.0178688500787296,-.0259704612641834,-.0242416041615842,-.012627373648802,-.001786832271014,-.00279377569118188,-.00138630263926177,.00186727769984554,.00505186161140503,.00596509639942324,.00341561645614388,-.00196588328720522,-.0078826862130179,-.0114566408051395,-.0106939706844047,-.00557045493859299},
			{.00303781657536634,.00474973406301401,.00235687098578853,-.00317458288582604,-.00858873504163554,-.010141337275906,-.00580693356275668,.00334222351587521,.013401456434816,.0194775827034515,.0181809574007362,.00947040224160834,.0063429536736534,.00991743325392784,.00492114092696054,-.0066285213996303,-.0179332580268801,-.0211750877428169,-.012124863256468,.00697855463035482,.0279822086740939,.0406691456503908,.0379617951494089,.0197741770113756,.00808533061765968,.0126417014630968,.00627295317887047,-.0084493423379166,-.0228594323180056,-.0269917760821381,-.0154555012012238,.00889552790747716,.0356687783299041,.0518407520280328,.048389706186527,.0252060423353226,.00782810670776326,.0122395227481066,.0060733876175387,-.00818053788513823,-.0221321902500056,-.0261330690598512,-.0149638052352352,.00861252866141587,.0345340241612611,.0501915082853594,.0468502528218308,.0244041460284054,.00563444710207634,.00880965807613161,.00437145053573306,-.00588811697401006,-.0159301169327542,-.0188098349613213,-.0107705186185175,.00619905156756988,.0248566019374114,.0361264107612158,.0337214707333665,.0175654056595871,.00204694101677262,.00320046494945897,.00158810638245006,-.00213909686741309,-.00578725989629442,-.00683343405369158,-.00391282692565031,.00225205644650943,.00903016695720288,.0131243812633644,.012250689435808,.00638134472992193,-.00204694101677262,-.00320046494945897,-.00158810638245006,.00213909686741309,.00578725989629442,.00683343405369158,.00391282692565031,-.00225205644650943,-.00903016695720288,-.0131243812633644,-.012250689435808,-.00638134472992193,-.00563444710207634,-.00880965807613161,-.00437145053573306,.00588811697401006,.0159301169327542,.0188098349613213,.0107705186185175,-.00619905156756988,-.0248566019374114,-.0361264107612158,-.0337214707333665,-.0175654056595871,-.00782810670776326,-.0122395227481066,-.0060733876175387,.00818053788513823,.0221321902500056,.0261330690598512,.0149638052352352,-.00861252866141587,-.0345340241612611,-.0501915082853594,-.0468502528218308,-.0244041460284054,-.00808533061765968,-.0126417014630968,-.00627295317887047,.0084493423379166,.0228594323180056,.0269917760821381,.0154555012012238,-.00889552790747716,-.0356687783299041,-.0518407520280328,-.048389706186527,-.0252060423353226,-.0063429536736534,-.00991743325392784,-.00492114092696054,.0066285213996303,.0179332580268801,.0211750877428169,.012124863256468,-.00697855463035482,-.0279822086740939,-.0406691456503908,-.0379617951494089,-.0197741770113756,-.00303781657536634,-.00474973406301401,-.00235687098578853,.00317458288582604,.00858873504163554,.010141337275906,.00580693356275668,-.00334222351587521,-.013401456434816,-.0194775827034515,-.0181809574007362,-.00947040224160835},
			{-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354764,-.0226018501305136,-.0117732310473289,-.0065767710679063,-.01028301505074,-.0051025466895586,.00687286554608826,.0185943234985903,.0219556552975931,.0125718165337835,-.00723580189137435,-.0290137040080097,-.042168313727554,-.0393611634058515,-.0205031034359056,-.00527564637794148,-.00824866040593754,-.00409307723852653,.00551316259147853,.0149156895386415,.0176120275664321,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.031574092667901,-.016446843331074,-.000299637130226379,-.000468493290714356,-.000232471592992847,.00031312716945722,.000847155796376566,.00100029778711831,.000572770891524226,-.00032966251846565,-.00132186188578481,-.00192118478526154,-.00179329125546946,-.000934119647899669,.00587430233536655,.0091846802486317,.00455754071797623,-.00613877079058787,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431711,.0259147334538654,.0376642916790926,.0351569747115048,.0183131550653389,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939193,.0644648579968841,.0601734237215266,.0313441429038559,.010054246316555,.0157201710580289,.00780052410656228,-.0105069010898283,-.0284260933844622,-.0335647332291363,-.019219178951509,.0110617404441344,.0443547332939194,.0644648579968841,.0601734237215266,.0313441429038559,.00587430233536656,.00918468024863171,.00455754071797623,-.00613877079058788,-.016608253020293,-.0196105590201443,-.0112290135176797,.00646294169431712,.0259147334538654,.0376642916790926,.0351569747115049,.0183131550653389,-.000299637130226364,-.000468493290714334,-.000232471592992836,.000313127169457205,.000847155796376526,.00100029778711826,.000572770891524199,-.000329662518465634,-.00132186188578475,-.00192118478526145,-.00179329125546937,-.000934119647899624,-.00527564637794147,-.00824866040593752,-.00409307723852653,.00551316259147852,.0149156895386415,.017612027566432,.0100846536508247,-.00580429691798344,-.0232737373522796,-.0338258864849756,-.0315740926679009,-.016446843331074,-.00657677106790629,-.01028301505074,-.00510254668955859,.00687286554608825,.0185943234985903,.021955655297593,.0125718165337835,-.00723580189137434,-.0290137040080097,-.0421683137275539,-.0393611634058515,-.0205031034359056,-.00377649391322152,-.00590468230499736,-.00292996917728826,.00394651640344462,.0106771771113592,.0126073110552328,.00721895108218894,-.00415492063170615,-.0166601627842796,-.0242137636354765,-.0226018501305136,-.0117732310473289},
			{.00421646110756717,.00659258663290769,.00327131497264031,-.00440629147242691,-.0119210842286909,-.0140760816664588,-.00805996969011775,.00463897510526447,.0186010966891058,.0270347033472955,.0252349995059512,.0131448300890066,.00526666188843086,.0082346128376237,.00408610667854559,-.0055037736089871,-.014890287939178,-.0175820340707359,-.0100674793638363,.00579441213021712,.0232341018964256,.0337682805916803,.0315203216066774,.0164188341582059,-.000223849711801565,-.000349996971430556,-.000173672398524789,.00023392770644707,.000632884117955103,.000747291803232149,.000427899569386332,-.000246280758644836,-.000987522480778167,-.00143525824110462,-.00133971290677435,-.000697852144734339,-.00762471196281051,-.0119215078435287,-.00591558507707979,.00796798605379006,.0215571378958162,.0254540633801011,.0145750063260015,-.00838875257661241,-.033636739632742,-.0488874012505932,-.0456329603679329,-.0237700623038804,-.00999890330935153,-.0156336403014995,-.00775758658064257,.010449066471062,.0282696236261719,.0333799781302062,.0191133881119304,-.0110008517447921,-.0441105852745744,-.0641100149794998,-.0598422026515798,-.0311716109136999,-.00456753568689175,-.00714150420140017,-.00354369398872959,.00477317186942322,.012913667706615,.0152480963782088,.00873106575768223,-.00502522940526383,-.0201498770593039,-.0292856898648238,-.0273361375479713,-.0142393061380111,.00456753568689175,.00714150420140018,.0035436939887296,-.00477317186942323,-.012913667706615,-.0152480963782088,-.00873106575768224,.00502522940526384,.0201498770593039,.0292856898648239,.0273361375479713,.0142393061380111,.00999890330935158,.0156336403014996,.0077575865806426,-.010449066471062,-.028269623626172,-.0333799781302063,-.0191133881119305,.0110008517447921,.0441105852745746,.0641100149795001,.0598422026515801,.0311716109137,.0076247119628105,.0119215078435287,.00591558507707978,-.00796798605379004,-.0215571378958161,-.025454063380101,-.0145750063260015,.0083887525766124,.033636739632742,.0488874012505931,.0456329603679328,.0237700623038804,.000223849711801557,.000349996971430543,.000173672398524782,-.000233927706447061,-.000632884117955079,-.000747291803232121,-.000427899569386316,.000246280758644826,.000987522480778129,.00143525824110457,.0013397129067743,.000697852144734312,-.00526666188843086,-.00823461283762369,-.00408610667854559,.0055037736089871,.014890287939178,.0175820340707359,.0100674793638363,-.00579441213021711,-.0232341018964256,-.0337682805916803,-.0315203216066774,-.0164188341582059,-.00421646110756715,-.00659258663290766,-.00327131497264029,.00440629147242689,.0119210842286909,.0140760816664587,.00805996969011771,-.00463897510526445,-.0186010966891057,-.0270347033472953,-.025234999505951,-.0131448300890066},
			{.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854622,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.0101727663375369,.0101365563802125,-.00731452673971809,-.0246336831441267,-.0203980242942646,.00741004225619188,.0354948866564383,.0345124826157485,-.00287922324282485,-.0514257400177226,-.0718259832229383,-.0438230751763429,.00953358692267161,.00949965212419449,-.00685493740419405,-.0230858894904042,-.0191163672896495,.00694445145067388,.0332646574217356,.0323439801935032,-.00269831471051887,-.0481945368893999,-.0673129836704801,-.0410695657945873,.00829540131749408,.00826587384013647,-.00596464449690627,-.020087582947275,-.0166336070239364,.00604253280328308,.028944371645266,.0281432684346891,-.00234786797311534,-.0419352157851109,-.0585706322241284,-.0357356085348261,.00653605799247159,.0065127928969091,-.00469962342316584,-.0158272761071956,-.0131058421372763,.00476099266474076,.0228056587485581,.0221744587810238,-.00184991667595225,-.0330413191366965,-.0461485869363911,-.0281565654077896,.00436637189291035,.00435082979406928,-.00313955348098353,-.0105733109491058,-.0087552743269776,.0031805508117616,.0152351750051339,.0148135058889446,-.00123582504736155,-.0220730427039245,-.030829269435596,-.0188098141018749,.00192618659675946,.00191933033641165,-.0013849864334046,-.00466432322601885,-.00386231234379255,.00140307204566677,.00672086359428575,.00653484796854621,-.000545173361442836,-.00973732885076348,-.0136000613395187,-.00829778422433439},
			{-.00327473465962885,-.00326307824304169,.00235463328641995,.00792987603467538,.00656636699674288,-.002385380868931,-.0114262268213642,-.0111099792585027,.000926856258488525,.016554558282902,.0231216395735111,.0141071699093245,-.00683763806148434,-.00681329949179824,.00491646861608254,.0165575620113583,.0137105584327533,-.00498066949411644,-.0238579340109859,-.0231976098634047,.00193527362955397,.0345658777187952,.0482779275954164,.0294557367047126,-.00871590227761381,-.00868487802727276,.0062669974080828,.0211058396406552,.0174767787351192,-.00634883393906744,-.030411586503362,-.0295698746885344,.00246688340095815,.0440609476003131,.0615396274713053,.0375470769767045,-.00843861757917165,-.00840858032359331,.00606762132158154,.0204343857631653,.0169207785452416,-.00614685433349348,-.0294440828159479,-.0286291489294765,.00238840282624089,.0426592078630735,.0595818270621095,.0363525672650008,-.00607387534936655,-.00605225539271983,.00436730012098293,.0147080858685528,.012179091982045,-.00442432977464276,-.0211930078739329,-.0206064417927042,.00171910397817624,.0307048765549653,.0428852933869165,.0261655372015035,-.00220658111756891,-.00219872679304704,.00158659857626521,.00534330763908673,.00442454822513244,-.00160731690676576,-.00769921809541031,-.00748612421965592,.000624534116884189,.0111547894756412,.0155798190061378,.00950569068328115,.00220658111756891,.00219872679304704,-.00158659857626521,-.00534330763908673,-.00442454822513244,.00160731690676576,.00769921809541031,.00748612421965592,-.000624534116884189,-.0111547894756412,-.0155798190061378,-.00950569068328115,.00607387534936655,.00605225539271983,-.00436730012098293,-.0147080858685528,-.012179091982045,.00442432977464276,.0211930078739329,.0206064417927042,-.00171910397817624,-.0307048765549653,-.0428852933869165,-.0261655372015035,.00843861757917165,.00840858032359331,-.00606762132158154,-.0204343857631653,-.0169207785452416,.00614685433349347,.0294440828159479,.0286291489294765,-.00238840282624089,-.0426592078630735,-.0595818270621095,-.0363525672650008,.00871590227761381,.00868487802727276,-.0062669974080828,-.0211058396406552,-.0174767787351192,.00634883393906744,.030411586503362,.0295698746885344,-.00246688340095815,-.0440609476003131,-.0615396274713053,-.0375470769767045,.00683763806148434,.00681329949179824,-.00491646861608254,-.0165575620113583,-.0137105584327533,.00498066949411644,.0238579340109859,.0231976098634047,-.00193527362955397,-.0345658777187952,-.0482779275954164,-.0294557367047126,.00327473465962885,.00326307824304169,-.00235463328641995,-.00792987603467538,-.00656636699674288,.002385380868931,.0114262268213642,.0111099792585027,-.000926856258488525,-.016554558282902,-.0231216395735111,-.0141071699093245},
			{.00407102114440158,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589114,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092,.00708969078054082,.0070644550293116,-.00509770214609633,-.0171679158335966,-.0142159644664865,.00516426963168224,.0247373981014314,.024052732727339,-.00200661273467494,-.0358400638320779,-.0500575747208443,-.0305415500311089,.00568709190283437,.00566684870169624,-.00408919112208648,-.0137714772263088,-.0114035292808877,.004142589136187,.0198434403974258,.0192942267538859,-.00160963170168385,-.0287495947462082,-.0401543645107642,-.0244993198234636,.00032300570868889,.000321855970010235,-.000232250876004936,-.000782168784534644,-.000647678131435131,.000235283684983902,.00112703375255846,.00109584045637403,-9.1421105446067e-005,-.00163287025850404,-.00228061884480666,-.00139147042059047,-.00633243679598169,-.00630989648997895,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615883,-.0214836463932346,.00179228526457281,.0320119658255772,.0447108961296189,.0272793893570287,-.010838372915793,-.0107997937321446,.00779311800518549,.0262454710298765,.0217326719902005,-.00789488312515758,-.0378173257324053,-.0367706427561445,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548538,-.010838372915793,-.0107997937321446,.0077931180051855,.0262454710298765,.0217326719902006,-.00789488312515758,-.0378173257324053,-.0367706427561446,.00306761152061517,.0547905387078465,.0765255747928819,.0466904296548539,-.0063324367959817,-.00630989648997896,.00455321362301119,.0153342007853678,.0126975490560225,-.00461267099685,-.0220951822615884,-.0214836463932346,.00179228526457282,.0320119658255772,.044710896129619,.0272793893570288,.000323005708688875,.00032185597001022,-.000232250876004925,-.000782168784534606,-.0006476781314351,.00023528368498389,.00112703375255841,.00109584045637398,-9.14211054460626e-005,-.00163287025850396,-.00228061884480655,-.00139147042059041,.00568709190283437,.00566684870169623,-.00408919112208647,-.0137714772263087,-.0114035292808876,.00414258913618699,.0198434403974258,.0192942267538859,-.00160963170168384,-.0287495947462082,-.0401543645107641,-.0244993198234635,.00708969078054081,.00706445502931158,-.00509770214609633,-.0171679158335965,-.0142159644664865,.00516426963168223,.0247373981014314,.0240527327273389,-.00200661273467494,-.0358400638320778,-.0500575747208442,-.0305415500311089,.00407102114440159,.00405653034642052,-.00292718735795645,-.00985810954628785,-.00816304881589115,.00296541154145592,.0142046351308886,.0138114886170208,-.00115223119376499,-.0205799748104051,-.0287439116082964,-.0175374779815092},
			{-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419407,-.00398827433495215,-.0145295292180157,-.0111084414324411,-.004451735343494,-.00147019434909273,.00789381658375618,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237627,-.0329362316722039,-.0251811462743288,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083196,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0103716459698384,-.00342525647171086,.0183910014051509,.0185076837908503,-.0121232889269308,-.0330054338592608,-.00788558538064342,.0369901615094009,.0379654642459228,-.0210632671661822,-.0767347805129326,-.0586669948054619,-.0097199704685807,-.0032100393562569,.0172354504835146,.0173448014338362,-.0113615534790256,-.0309316229408192,-.00739011505504753,.034665980553616,.0355800026697282,-.0197398113493751,-.0719133494016041,-.0549808061948678,-.00845757808525572,-.00279313178985934,.0149969764589386,.0150921253283751,-.00988595860751609,-.0269143427103298,-.00643031533265706,.0301636963179966,.0309590087568212,-.0171760805668307,-.0625735201462144,-.0478401105318693,-.00666383924361929,-.00220074601100209,.011816318957425,.0118912880282093,-.00778927942083195,-.0212061717150426,-.00506653172227985,.0237663810171494,.0243930183579287,-.0135332643197644,-.0493025160345462,-.0376938649300943,-.004451735343494,-.00147019434909273,.00789381658375617,.00794389919377696,-.00520357848236969,-.0141666478846187,-.00338466423219967,.0158769793947679,.0162956004772274,-.00904081099237626,-.0329362316722039,-.0251811462743288,-.00196384393296446,-.000648563319686994,.00348228782930674,.00350438133268179,-.00229551292783852,-.00624949224335884,-.00149311488771257,.00700398996175381,.00718866097419408,-.00398827433495215,-.0145295292180157,-.0111084414324411},
			{.00333875638300519,.00110263085932663,-.0059202824228546,-.00595784387271232,.00390263111617153,.0106248423145106,.00253846386580488,-.0119075736105008,-.0122215351790305,.00678051660291391,.0247018398989999,.01888560445932,.00697131526529464,.00230229057164388,-.0123615354026379,-.0124399636192375,.00814868434650274,.0221846450958567,.00530030642192468,-.0248629849443481,-.0255185359412845,.0141577022933105,.0515773820591734,.0394330965061518,.00888629992584825,.00293471802056214,-.0157571859758933,-.0158571580226064,.0103870862740315,.0282786536784102,.00675627349671033,-.0316927197895115,-.0325283472936162,.0180467507280219,.0657454252068613,.0502651665035225,.00860359425560065,.00284135391717673,-.015255892314898,-.0153526838855177,.0100566351063325,.0273790063776703,.00654133175006641,-.0306844585711385,-.0314935017110425,.0174726176464323,.0636538230041544,.0486660478934077,.00619262083804183,.00204512520616675,-.0109807545364729,-.0110504223380172,.00723847804416364,.0197066249734953,.00470826332583461,-.0220857948325456,-.0226681209230377,.0125762899688114,.045816199491271,.0350284280425796,.00224972022366444,.000742974526710104,-.00398920363411266,-.00401451328347947,.00262967019464013,.00715922933157688,.00171046726410221,-.00802355910202843,-.00823511262949449,.0045688464741287,.0166445893043808,.0127254945897002,-.00224972022366444,-.000742974526710104,.00398920363411266,.00401451328347947,-.00262967019464013,-.00715922933157688,-.00171046726410221,.00802355910202843,.00823511262949449,-.0045688464741287,-.0166445893043808,-.0127254945897002,-.00619262083804183,-.00204512520616675,.0109807545364729,.0110504223380172,-.00723847804416364,-.0197066249734953,-.00470826332583461,.0220857948325456,.0226681209230377,-.0125762899688114,-.045816199491271,-.0350284280425796,-.00860359425560065,-.00284135391717673,.015255892314898,.0153526838855177,-.0100566351063325,-.0273790063776703,-.00654133175006641,.0306844585711385,.0314935017110425,-.0174726176464323,-.0636538230041544,-.0486660478934077,-.00888629992584825,-.00293471802056214,.0157571859758933,.0158571580226064,-.0103870862740315,-.0282786536784102,-.00675627349671033,.0316927197895115,.0325283472936162,-.0180467507280219,-.0657454252068613,-.0502651665035225,-.00697131526529464,-.00230229057164388,.0123615354026379,.0124399636192375,-.00814868434650274,-.0221846450958567,-.00530030642192469,.0248629849443481,.0255185359412845,-.0141577022933105,-.0515773820591734,-.0394330965061518,-.00333875638300519,-.00110263085932663,.0059202824228546,.00595784387271232,-.00390263111617154,-.0106248423145106,-.00253846386580488,.0119075736105008,.0122215351790305,-.00678051660291391,-.0247018398989999,-.01888560445932},
			{.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508331,-.0313789443876932,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.0100832468852973,-.00396574221639922,-.0207637705507871,.00312820705548716,.0290658155367351,.000236810911249813,-.0375650336509844,-.00636557044983238,.0484244880818703,.0183243368217685,-.069210025250605,-.0731066150578859,.00944969219326573,-.00371656507959313,-.0194591328278033,.00293165427043734,.0272395403279159,.000221931511225863,-.0352047320936947,-.00596560632400807,.0453818608425999,.017172974596499,-.06486139263931,-.0685131503222357,.00822240251290047,-.00323387190024903,-.0169318554922051,.00255090228837061,.0237017735881517,.000193107900053695,-.0306324768799845,-.0051908163172188,.0394878392862589,.0149426146998584,-.0564374444077397,-.0596149258467637,.00647854125495116,-.00254801105712353,-.0133408361068458,.00200989013693136,.0186749454025678,.000152152305265038,-.024135739511287,-.00408991381840209,.0311129983582702,.0117734865981066,-.0444678196361585,-.0469713998923258,.00432794820291702,-.00170218255031759,-.00891226055092382,.00134269429858841,.0124756782760082,.000101644377989913,-.0161237269831564,-.00273224086470254,.0207848402955105,.00786520269904917,-.0297064435508332,-.0313789443876932,.00190923632351697,-.000750902876363605,-.00393156543717474,.000592317792647409,.00550353585771022,4.48395242828537e-005,-.00711284049239881,-.0012053040514606,.00916904967669009,.00346966506547543,-.0131047365658235,-.0138425456150847},
			{-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0142624380568961,-.0310672248312831,-.0432268243272434,-.0492027502280183,-.0487278197520516,-.0427730023800195,-.0332521174866264,-.0225204586912224,-.0127920758181991,-.00561973086192384,-.00156440510363151,-.00013271878100494,-.0133662947160114,-.0291151962551547,-.0405107788226866,-.046111222903296,-.045666133453227,-.0400854716016497,-.0311628068416243,-.0211054440205679,-.0119883188788287,-.0052666296341122,-.00146610976236707,-.000124379740278982,-.0116303317624909,-.0253338265369722,-.0352493946658487,-.0401224746074978,-.0397351916634755,-.0348793247110304,-.027115501335187,-.0183643501186492,-.0104313221276118,-.00458262003167689,-.0012756970648066,-.000108225777952072,-.00916369443284368,-.0199608618172178,-.0277734709771828,-.0316130360423954,-.0313078906148575,-.0274818878947751,-.0213646672943921,-.0144695178419413,-.00821897865513019,-.00361070780521928,-.00100513883262257,-8.527254244871e-005,-.00612174767311959,-.0133347265427199,-.018553890308656,-.0211188872840585,-.0209150368256344,-.0183591000884744,-.0142725298464402,-.00966626919188733,-.0054906363177203,-.00241211034118437,-.000671476592204514,-5.69657786105805e-005,-.00270055519912364,-.00588249745280389,-.00818488569155241,-.00931641157068581,-.00922648473202207,-.0080989556973843,-.00629620114051677,-.00426418972426832,-.00242214596975712,-.00106408128373998,-.000296215998926277,-2.51299527215787e-005},
			{.00459124870047564,.0100009097368206,.0139152296561375,.0158389513870659,.0156860656096203,.0137691389655327,.0107042527083733,.00724960390977444,.00411792158131722,.00180905830503188,.000503600637591741,4.27237565125139e-005,.00958651619965776,.0208818753803107,.0290549659195549,.0330716922483351,.0327524671142169,.0287499289103516,.022350453806482,.0151371554572844,.00859821031784408,.00377730937239729,.00105151691519805,8.92071004289135e-005,.0122198831859258,.0266180197827826,.037036216505918,.0421563170205638,.0417494023742321,.0366473873898747,.0284900091941002,.019295254668408,.0109600947313715,.00481491694443133,.00134036323562505,.000113711834820181,.0118311240516121,.0257712033140444,.0358579591324206,.040815170541386,.0404212013367775,.0354815000912993,.0275836379021353,.0186814021145725,.0106114140709319,.00466173684322829,.00129772138355337,.000110094245863453,.00851570438619621,.018549374357127,.0258095324275119,.0293775912826531,.029094023527884,.0255385679871328,.0198539128865321,.0134463384234132,.00763779206044802,.00335538472168774,.000934062700194595,7.92426863503829e-005,.00309367437106509,.00673881119463904,.00937635753650586,.0106725993661888,.010569581899212,.00927791873371822,.0072127376288433,.00488492679861124,.00277473717702488,.00121897934073137,.000339336090769426,2.87881139056347e-005,-.00309367437106509,-.00673881119463904,-.00937635753650586,-.0106725993661888,-.010569581899212,-.00927791873371822,-.0072127376288433,-.00488492679861124,-.00277473717702488,-.00121897934073137,-.000339336090769426,-2.87881139056347e-005,-.00851570438619621,-.018549374357127,-.0258095324275119,-.0293775912826531,-.029094023527884,-.0255385679871328,-.0198539128865321,-.0134463384234132,-.00763779206044802,-.00335538472168774,-.000934062700194595,-7.92426863503829e-005,-.0118311240516121,-.0257712033140444,-.0358579591324206,-.040815170541386,-.0404212013367775,-.0354815000912993,-.0275836379021353,-.0186814021145725,-.0106114140709319,-.00466173684322829,-.00129772138355337,-.000110094245863453,-.0122198831859258,-.0266180197827826,-.037036216505918,-.0421563170205638,-.0417494023742321,-.0366473873898747,-.0284900091941002,-.019295254668408,-.0109600947313715,-.00481491694443133,-.00134036323562505,-.000113711834820181,-.00958651619965776,-.0208818753803107,-.0290549659195549,-.0330716922483351,-.0327524671142169,-.0287499289103516,-.022350453806482,-.0151371554572845,-.00859821031784408,-.00377730937239729,-.00105151691519805,-8.92071004289135e-005,-.00459124870047563,-.0100009097368206,-.0139152296561375,-.0158389513870659,-.0156860656096203,-.0137691389655327,-.0107042527083733,-.00724960390977444,-.00411792158131722,-.00180905830503188,-.000503600637591741,-4.2723756512514e-005},
			{-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042617,-.0195002989275631,-.0171172512269981,-.0133071053509708,-.00901242203528271,-.00511923791436272,-.00224895003013865,-.000626056476974539,-5.31124913051166e-005,-.00993989955437502,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.033959806304287,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562087,-.00391655059726168,-.00109027850150277,-9.24955009028306e-005,-.00797342564302374,-.0173681530562416,-.0241659853793842,-.0275068307964137,-.0272413205925748,-.0239122759128171,-.0185896187731037,-.0125900776644864,-.00715141863890626,-.00314171433962441,-.000874581731371057,-7.41965242936831e-005,-.000452860977896186,-.000986446620240561,-.00137253826155738,-.00156228587949368,-.00154720588553185,-.00135812850566637,-.00105582133868279,-.000715069674966715,-.000406174031483403,-.000178437711946007,-4.96729957550138e-005,-4.2140871505504e-006,.00887821311745346,.0193390107581648,.0269082296615077,.0306281787690543,.0303325396949503,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930627,.00349822155622026,.000973825222354376,8.26160028005004e-005,.0151956328492542,.0330999609111817,.0460551659608014,.0524220981473091,.0519160928547411,.0455716503331479,.0354278116245859,.0239939777830022,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.0151956328492542,.0330999609111817,.0460551659608015,.0524220981473091,.0519160928547411,.045571650333148,.0354278116245859,.0239939777830023,.013629064451515,.00598743121959625,.00166676451020866,.000141402603138845,.00887821311745347,.0193390107581649,.0269082296615078,.0306281787690543,.0303325396949504,.0266257304177768,.020699082756761,.0140187414638274,.00796292855930629,.00349822155622027,.000973825222354377,8.26160028005006e-005,-.000452860977896164,-.000986446620240513,-.00137253826155732,-.00156228587949361,-.00154720588553178,-.0013581285056663,-.00105582133868274,-.000715069674966681,-.000406174031483384,-.000178437711945999,-4.96729957550114e-005,-4.2140871505502e-006,-.00797342564302373,-.0173681530562415,-.0241659853793842,-.0275068307964136,-.0272413205925747,-.023912275912817,-.0185896187731037,-.0125900776644864,-.00715141863890625,-.0031417143396244,-.000874581731371055,-7.41965242936829e-005,-.009939899554375,-.0216516343856671,-.0301260058170532,-.0342907988882749,-.0339598063042869,-.0298097243683288,-.0231743483455502,-.0156951494839949,-.00891516220562085,-.00391655059726168,-.00109027850150277,-9.24955009028305e-005,-.00570765954562574,-.0124327370718105,-.0172988654193792,-.0196903605042618,-.0195002989275631,-.0171172512269982,-.0133071053509708,-.00901242203528272,-.00511923791436273,-.00224895003013865,-.00062605647697454,-5.31124913051166e-005},
			{.00637261042846754,.0138811695555554,.019314209141412,.0219843169843424,.0217721129493893,.0191114366236973,.0148574030483831,.0100623826962154,.00571563679614984,.00251095607588359,.000698992993903286,5.93001760646835e-005,.00795984680451423,.017338574885438,.0241248304194484,.027459986337079,.0271949283002686,.0238715530231142,.0185579604316863,.0125686366127877,.00713923968808145,.00313636396281389,.000873092308301808,7.40701667321458e-005,-.000338318549950809,-.000736944021350783,-.00102538124737378,-.00116713587426844,-.00115587007319604,-.00101461616061174,-.000788771746176309,-.000534206627103106,-.000303440163905314,-.000133305342938937,-3.71091719441577e-005,-3.14821528841767e-006,-.0115237204206783,-.0251015998648952,-.0349262753137827,-.0397546853696583,-.0393709525182014,-.0345595976658438,-.0268669426491322,-.0181960102941585,-.0103357017040835,-.00454061269426197,-.0012640031786301,-.000107233708624919,-.0151119893856147,-.0329177641310666,-.0458016580195613,-.0521335438038479,-.0516303237875308,-.0453208038751262,-.0352328013276903,-.0238619043492618,-.013554044071107,-.00595447375803616,-.00165758990339312,-.000140624260860408,-.00690321215069344,-.0150369553289623,-.0209223652885514,-.0238147939270004,-.0235849211787824,-.0207027093526006,-.0160944728071212,-.010900205382576,-.00619153701972338,-.00272002544129556,-.000757193144462012,-6.42376778783298e-005,.00690321215069345,.0150369553289623,.0209223652885514,.0238147939270005,.0235849211787825,.0207027093526006,.0160944728071212,.010900205382576,.00619153701972339,.00272002544129557,.000757193144462013,6.42376778783299e-005,.0151119893856147,.0329177641310668,.0458016580195615,.0521335438038482,.051630323787531,.0453208038751264,.0352328013276905,.0238619043492619,.0135540440711071,.00595447375803619,.00165758990339312,.000140624260860409,.0115237204206782,.0251015998648951,.0349262753137827,.0397546853696582,.0393709525182013,.0345595976658438,.0268669426491322,.0181960102941585,.0103357017040834,.00454061269426196,.0012640031786301,.000107233708624919,.000338318549950797,.000736944021350755,.00102538124737374,.0011671358742684,.00115587007319599,.0010146161606117,.000788771746176279,.000534206627103086,.000303440163905302,.000133305342938932,3.71091719441563e-005,3.14821528841755e-006,-.00795984680451422,-.017338574885438,-.0241248304194484,-.027459986337079,-.0271949283002686,-.0238715530231142,-.0185579604316863,-.0125686366127877,-.00713923968808145,-.00313636396281389,-.000873092308301808,-7.40701667321458e-005,-.00637261042846751,-.0138811695555553,-.0193142091414119,-.0219843169843423,-.0217721129493892,-.0191114366236972,-.014857403048383,-.0100623826962154,-.00571563679614981,-.00251095607588358,-.000698992993903282,-5.93001760646832e-005},
			{-.00666886465162736,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907023,-6.20569627520987e-005,-.00427854858305904,-.00931976919031684,-.0129674931619453,-.0147601943251846,-.0146177212706554,-.0128313523954431,-.00997520897819624,-.00675584891158071,-.00383745876041692,-.0016858472177435,-.000469301476561824,-3.98139329439154e-005,.00853996360068238,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172028,.0033649434139527,.000936723622457736,7.94684298987088e-005,.0138856906557926,.0302465729786341,.0420849723059915,.0479030419872811,.0474406569696583,.0416431382277769,.0323737509132328,.0219255727222366,.0124541685613893,.00547128366175477,.00152308078343358,.000129212966948341,.00209660569165992,.00456694151066636,.00635442590918465,.00723289845409743,.00716308276514704,.00628771321435967,.0048881249127316,.00331055053017243,.00188045962840229,.000826111192469611,.000229970544394388,1.95099148220722e-005,-.0135748462316425,-.029569474604983,-.0411428600764948,-.0468306867209362,-.0463786526327775,-.0407109168753721,-.0316490336334332,-.0214347478726881,-.0121753701241565,-.00534880376058466,-.00148898516796897,-.000126320411489674,-.0135748462316426,-.0295694746049834,-.0411428600764953,-.0468306867209367,-.046378652632778,-.0407109168753726,-.0316490336334336,-.0214347478726884,-.0121753701241566,-.00534880376058472,-.00148898516796898,-.000126320411489675,.00209660569165985,.0045669415106662,.00635442590918444,.00723289845409719,.00716308276514679,.00628771321435945,.00488812491273143,.00331055053017232,.00188045962840222,.000826111192469583,.00022997054439438,1.95099148220715e-005,.0138856906557927,.0302465729786343,.0420849723059919,.0479030419872815,.0474406569696587,.0416431382277773,.0323737509132331,.0219255727222367,.0124541685613895,.00547128366175482,.0015230807834336,.000129212966948343,.00853996360068239,.0186022171086724,.0258830576409941,.0294612810463758,.0291769054745815,.0256113212874322,.0199104719577799,.013484643840449,.00765955031172029,.00336494341395271,.000936723622457737,7.94684298987089e-005,-.00427854858305892,-.00931976919031658,-.012967493161945,-.0147601943251842,-.014617721270655,-.0128313523954428,-.00997520897819597,-.00675584891158053,-.00383745876041681,-.00168584721774346,-.000469301476561811,-3.98139329439144e-005,-.00666886465162735,-.0145264867531766,-.0202121011574647,-.0230063387794926,-.0227842696598568,-.0199999020138189,-.0155481040488096,-.0105301695478151,-.00598134918480345,-.00262768710000618,-.000731488252907022,-6.20569627520987e-005},
			{.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831813,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932927,-.0625431691900999,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0166626077730442,.028928716294081,.0224873968275873,-.00191057605792923,-.0346390454417612,-.0621778248073664,-.073838479750627,-.0667363764908783,-.0465043628013169,-.02357169801793,-.00720471085575246,-.000642693442774579,.0156156559869597,.0271110553609456,.02107445951347,-.0017905299616912,-.0324625907722716,-.0582710422902724,-.0691990301932926,-.0625431691900998,-.0435823816589457,-.0220906314479756,-.00675202151676104,-.000602311465536526,.0135875546421786,.0235899757545631,.0183373897601727,-.00155798281630011,-.0282464743276406,-.050703023416817,-.0602117262779296,-.0544203029046721,-.0379220695385274,-.0192215851918263,-.00587509492913837,-.000524085568766528,.0107058165986337,.0185868583895589,.0144482754137867,-.00122755556348272,-.0222557760887318,-.0399495924021276,-.0474416269589545,-.0428784867831153,-.0298793073670401,-.0151449448571057,-.00462906611728874,-.000412934050972505,.00715195255929032,.0124168323082742,.00965207832311933,-.000820060671975437,-.0148678293981901,-.0266880706381115,-.0316930765831814,-.0286446998658781,-.0199606808901245,-.0101174839054231,-.00309241812239643,-.000275857961460674,.00315502103307612,.00547757647607656,.00425792954719506,-.000361762560228417,-.00655881230735289,-.0117732078753915,-.0139811222730974,-.0126363576678934,-.00880547899622987,-.00446323913068385,-.00136419308410438,-.000121692315956342},
			{-.00536389192389492,-.00931249836832228,-.00723896090430979,.000615037191458855,.0111507213412938,.0200157824556806,.0237694861813762,.0214832345431866,.0149703083366938,.0075880103734666,.00231928224558504,.000206890674868773,-.0111997933843804,-.0194444741052602,-.015114932887575,.00128419617057589,.0232826792338887,.0417929054334019,.0496306297482205,.0448569418485149,.0312579676567552,.0158437473362906,.00484265572818485,.000431987229526937,-.0142763193649554,-.0247857717265967,-.0192669276724185,.00163695829281661,.0296783121801492,.0532732028778971,.0632639099894119,.057178914430488,.0398443715568573,.0201959437239261,.00617290849726889,.000550651912820983,-.013822137481785,-.0239972457633677,-.0186539763038293,.00158488066824849,.0287341366282638,.0515783876396646,.0612512538529679,.0553598442367079,.0385767765104859,.0195534369602114,.00597652572281162,.000533133663449908,-.0099487788537054,-.0172725304977166,-.0134265981099973,.00114075173240722,.0206820089326614,.0371246468163075,.0440868989979581,.0398464310178064,.0277664593408513,.0140740041403415,.0043017321169096,.000383734348187365,-.0036142966884799,-.00627494597051114,-.00487775533058912,.000414424249392327,.00751357704253213,.0134870309233169,.016016350890574,.0144758292241828,.0100872904626957,.00511295178092569,.00156277834430738,.000139407037215731,.0036142966884799,.00627494597051114,.00487775533058912,-.000414424249392327,-.00751357704253213,-.0134870309233169,-.016016350890574,-.0144758292241828,-.0100872904626957,-.00511295178092569,-.00156277834430738,-.000139407037215731,.0099487788537054,.0172725304977166,.0134265981099973,-.00114075173240722,-.0206820089326614,-.0371246468163075,-.0440868989979581,-.0398464310178064,-.0277664593408513,-.0140740041403415,-.0043017321169096,-.000383734348187365,.013822137481785,.0239972457633677,.0186539763038293,-.00158488066824849,-.0287341366282638,-.0515783876396646,-.0612512538529679,-.0553598442367079,-.0385767765104859,-.0195534369602114,-.00597652572281162,-.000533133663449908,.0142763193649554,.0247857717265967,.0192669276724185,-.00163695829281661,-.0296783121801492,-.0532732028778971,-.0632639099894119,-.057178914430488,-.0398443715568573,-.0201959437239261,-.00617290849726889,-.000550651912820983,.0111997933843804,.0194444741052602,.015114932887575,-.00128419617057589,-.0232826792338887,-.0417929054334019,-.0496306297482205,-.0448569418485149,-.0312579676567552,-.0158437473362906,-.00484265572818485,-.000431987229526937,.00536389192389492,.00931249836832228,.00723896090430979,-.000615037191458855,-.0111507213412938,-.0200157824556806,-.0237694861813762,-.0214832345431866,-.0149703083366938,-.0075880103734666,-.00231928224558504,-.000206890674868773},
			{.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065728,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700374,-.043333498156298,-.0514601409111785,-.0465104827417968,-.032410215795647,-.0164277881345211,-.00502116832732015,-.000447911376826381,.00931524214857705,.0161726184158759,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248514,-.0412794520598759,-.0373090164275984,-.0259982954865174,-.0131777737243128,-.00402779847824231,-.000359298203996196,.000529071174374757,.000918544690683033,.000714019894739991,-6.06646169956767e-005,-.00109985907972585,-.00197427048868781,-.00234451964109179,-.00211901363606869,-.00147660667381522,-.000748448629543353,-.000228764001733872,-2.0406804214751e-005,-.0103722927557633,-.0180078123747257,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831512,.0459636534116219,.0415426711023014,.0289484618473594,.0146731265551326,.004484854425811,.000400069702861968,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191293,.000684745031277692,-.0177528462581856,-.0308215292474391,-.0239587526793251,.00203558551476311,.0369054866221305,.0662461350293133,.078669749465725,.0711029538598146,.0495471545865855,.0251139999606578,.00767611684191294,.000684745031277693,-.0103722927557633,-.0180078123747258,-.013998160815385,.0011893128899699,.021562430382798,.0387050220831513,.045963653411622,.0415426711023014,.0289484618473595,.0146731265551327,.00448485442581101,.000400069702861968,.000529071174374731,.000918544690682989,.000714019894739956,-6.06646169956738e-005,-.0010998590797258,-.00197427048868771,-.00234451964109168,-.00211901363606859,-.00147660667381515,-.000748448629543317,-.000228764001733861,-2.04068042147501e-005,.00931524214857704,.0161726184158758,.0125715944102697,-.00106810883778206,-.0193649817891992,-.0347605550248513,-.0412794520598758,-.0373090164275984,-.0259982954865174,-.0131777737243127,-.0040277984782423,-.000359298203996196,.0116126462360064,.0201612468444716,.0156721077327354,-.00133153490557513,-.0241409379700373,-.0433334981562979,-.0514601409111784,-.0465104827417967,-.0324102157956469,-.016427788134521,-.00502116832732014,-.00044791137682638,.00666817916784111,.0115769311726008,.00899919106943596,-.000764590011454859,-.0138621375690256,-.0248828323711065,-.0295492889927298,-.0267071110065729,-.0186104976765478,-.00943311562119823,-.00288324033626761,-.000257198338026687},
			{-.00744503202371008,-.0129256609858319,-.0100476103015583,.000853665892443807,.015477097348583,.0277817195939975,.0329918253981113,.0298185294215917,.0207786485173192,.0105320877132218,.00321914215189193,.000287162329453971,-.00929937817926298,-.0161450762523781,-.0125501848339245,.00106628983559657,.0193320029925161,.0347013573819497,.0412091526569903,.0372454787243047,.0259540200768997,.0131553318173118,.0040209390890345,.000358686314837344,.000395252850754388,.000686216571838048,.000533422368407895,-4.53206751167283e-005,-.000821670992004966,-.00147491694238737,-.00175151873070067,-.00158305010934183,-.00110312756683095,-.000559143020444886,-.000170902570797115,-1.52452976675568e-005,.013463002097378,.0233737343785841,.0181692970739038,-.00154370131167062,-.0279875483949175,-.0502382458492594,-.0596597856283979,-.0539214502859268,-.0375744506777747,-.0190453873833324,-.00582123990933241,-.000519281452573281,.0176551267617537,.0306518739553514,.0238268731365797,-.0020243807542107,-.0367023425450716,-.0658814870815153,-.0782367164641202,-.0707115718388728,-.0492744251929199,-.0249757614273064,-.00763386410894635,-.000680975892590434,.00806492662705201,.0140018883901234,.0108842029962012,-.000924744549741412,-.0167657646224416,-.0300949048150996,-.0357388188903564,-.032301305238965,-.0225087380642283,-.0114090193791377,-.00348717711010207,-.000311072284138478,-.00806492662705202,-.0140018883901234,-.0108842029962012,.000924744549741413,.0167657646224416,.0300949048150996,.0357388188903565,.032301305238965,.0225087380642284,.0114090193791378,.00348717711010208,.000311072284138478,-.0176551267617538,-.0306518739553516,-.0238268731365798,.00202438075421071,.0367023425450718,.0658814870815156,.0782367164641205,.0707115718388731,.0492744251929201,.0249757614273065,.00763386410894639,.000680975892590437,-.013463002097378,-.023373734378584,-.0181692970739037,.00154370131167062,.0279875483949174,.0502382458492593,.0596597856283978,.0539214502859267,.0375744506777747,.0190453873833324,.0058212399093324,.00051928145257328,-.000395252850754372,-.000686216571838022,-.000533422368407874,4.53206751167266e-005,.000821670992004935,.00147491694238732,.0017515187307006,.00158305010934177,.00110312756683091,.000559143020444865,.000170902570797108,1.52452976675562e-005,.00929937817926298,.0161450762523781,.0125501848339245,-.00106628983559657,-.0193320029925161,-.0347013573819496,-.0412091526569903,-.0372454787243047,-.0259540200768997,-.0131553318173118,-.00402093908903449,-.000358686314837344,.00744503202371004,.0129256609858318,.0100476103015582,-.000853665892443803,-.0154770973485829,-.0277817195939974,-.0329918253981111,-.0298185294215915,-.020778648517319,-.0105320877132217,-.00321914215189191,-.00028716232945397},
			{-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487695,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0179745913044958,-.0207802379693662,.00627359802990123,.0404962585820704,.0478877813534532,.0139860783272433,-.0421296772065364,-.0833025022297376,-.0850965784390184,-.0542536076498451,-.0190488437902153,-.00182534802689418,-.0168452044326026,-.0194745655588036,.00587941275277989,.0379517811011308,.044878876690839,.0131072993339146,-.0394825680985669,-.0780684053415047,-.0797497554206825,-.0508447227859509,-.0178519590468035,-.00171065701315776,-.0146574140643079,-.0169452838794544,.00511581723553786,.0330227497269562,.0390501807817485,.0114049737045745,-.0343547239962773,-.0679291810918004,-.0693921638888432,-.0442412057295224,-.0155334152610008,-.00148848345914621,-.0115487731909908,-.0133514165133201,.00403082103572803,.0260190675562714,.0307681613507652,.00898613179551763,-.0270685479533684,-.0535223131472601,-.0546750169210118,-.0348582395519807,-.0122389862866471,-.00117279608755002,-.00771508433934048,-.00891932872404869,.00269276430778718,.0173818722826965,.0205544741300114,.00600312808471561,-.0180829709744956,-.0357552402440313,-.0365252966548462,-.023286824809605,-.00817617679972923,-.000783478953016023,-.00340344167005923,-.00393468349964643,.00118788931522614,.00766786023698012,.00906742566682238,.00264822720991415,-.00797714376487696,-.0157731100811164,-.0161128137008811,-.0102727781621573,-.00360684855771114,-.000345624856323613},
			{.00578623505122275,.00668940613303525,-.00201954593586876,-.013036228022183,-.015415647248866,-.00450228521335004,.0135620449344063,.0268160677531348,.0273936022556361,.0174648825623257,.00613204972269062,.000587601273095628,.0120816448143678,.0139674638522171,-.00421680703729971,-.0272196126304688,-.0321878342299675,-.00940076065336832,.0283175170734496,.0559918847131752,.057197775359769,.0364666326159833,.0128037050133204,.00122691003928565,.0154004108740453,.0178042547598082,-.00537514237082621,-.0346967010520833,-.0410296677235249,-.0119831015407985,.0360961942322037,.0713725691695229,.0729097118112787,.0464838301496112,.0163208173179129,.00156393595415252,.0149104675326585,.0172378363610222,-.0052041394518047,-.033592872213516,-.0397243640751441,-.0116018752957915,.0349478424019757,.0691019469563708,.0705901874741346,.0450050096655201,.0158015924844782,.00151418143699325,.0107321276671759,.0124073011209858,-.00374579059122209,-.0241791877090991,-.0285924600162962,-.00835069769478169,.0251544564601237,.0497376031410035,.0508087960597447,.0323933175358136,.011373533896014,.00108986444975558,.00389887985834265,.00450745443375195,-.00136081007816862,-.00878406881409194,-.0103873686481536,-.00303372901021968,.00913837467102563,.0180691979356344,.018458352129919,.0117681840173745,.00413189663788404,.000395937383830426,-.00389887985834265,-.00450745443375195,.00136081007816862,.00878406881409194,.0103873686481536,.00303372901021968,-.00913837467102562,-.0180691979356344,-.018458352129919,-.0117681840173745,-.00413189663788404,-.000395937383830426,-.0107321276671759,-.0124073011209858,.00374579059122209,.0241791877090991,.0285924600162962,.00835069769478169,-.0251544564601236,-.0497376031410035,-.0508087960597447,-.0323933175358136,-.011373533896014,-.00108986444975558,-.0149104675326585,-.0172378363610222,.0052041394518047,.033592872213516,.0397243640751441,.0116018752957915,-.0349478424019757,-.0691019469563708,-.0705901874741346,-.0450050096655201,-.0158015924844782,-.00151418143699325,-.0154004108740453,-.0178042547598082,.00537514237082622,.0346967010520833,.0410296677235249,.0119831015407985,-.0360961942322037,-.0713725691695229,-.0729097118112787,-.0464838301496112,-.0163208173179129,-.00156393595415252,-.0120816448143678,-.0139674638522171,.00421680703729971,.0272196126304688,.0321878342299675,.00940076065336832,-.0283175170734496,-.0559918847131752,-.057197775359769,-.0364666326159833,-.0128037050133204,-.00122691003928565,-.00578623505122275,-.00668940613303525,.00201954593586876,.013036228022183,.015415647248866,.00450228521335004,-.0135620449344063,-.0268160677531348,-.0273936022556361,-.0174648825623257,-.00613204972269062,-.000587601273095628},
			{-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137491,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099975,.00974729658452533,-.0293613727260863,-.0580558878956585,-.0593062307364416,-.0378108853797437,-.0132756821226898,-.00127213706171361,-.010048707428968,-.0116172060950429,.00350725922283749,.0226394607568784,.0267716965627408,.00781892655072194,-.0235526245439282,-.0465703202274322,-.047573299745961,-.0303305160603543,-.0106492690078639,-.00102046204932158,-.000570729279560872,-.000659814181276736,.000199199304351608,.0012858373297019,.00152053298395657,.000444086002978713,-.00133770164299672,-.00264502131246258,-.00270198682589509,-.00172266072051549,-.000604838947861804,-5.79584562835985e-005,.0111889882847748,.0129354729270487,-.00390524678257324,-.0252084820831874,-.0298096249016592,-.00870618218253045,.0262252674745031,.0518549048349062,.052971697831619,.0337722827805856,.0118576988147661,.00113625936426423,.0191506731907347,.0221398940001207,-.00668408107674887,-.043145938642774,-.0510210905490773,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0191506731907347,.0221398940001207,-.00668408107674888,-.043145938642774,-.0510210905490774,-.0149011908380949,.0448862322456101,.088753005236516,.0906644682890648,.057803434410253,.0202952142781976,.00194478099280396,.0111889882847749,.0129354729270488,-.00390524678257325,-.0252084820831874,-.0298096249016593,-.00870618218253046,.0262252674745032,.0518549048349063,.0529716978316191,.0337722827805856,.0118576988147661,.00113625936426423,-.000570729279560845,-.000659814181276704,.000199199304351599,.00128583732970183,.0015205329839565,.000444086002978692,-.00133770164299666,-.00264502131246246,-.00270198682589496,-.00172266072051541,-.000604838947861776,-5.79584562835958e-005,-.010048707428968,-.0116172060950428,.00350725922283748,.0226394607568783,.0267716965627407,.00781892655072193,-.0235526245439282,-.0465703202274321,-.0475732997459609,-.0303305160603542,-.0106492690078639,-.00102046204932158,-.0125270049495772,-.0144823400702597,.00437224926235588,.0282229967348447,.0333743596099974,.00974729658452531,-.0293613727260862,-.0580558878956584,-.0593062307364415,-.0378108853797436,-.0132756821226898,-.00127213706171361,-.00719321950764429,-.00831600622248073,.00251061996166316,.0162061252066575,.0191641254687835,.00559706364137492,-.0168598000810742,-.0333366792003028,-.0340546473459024,-.0217116540952635,-.00762312268627624,-.000730482758292872},
			{.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503008,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0183442742643311,.00869643683821103,-.0295328180767883,-.0384851709140763,.00785168169205949,.0578877519827171,.0428339111870372,-.0335857944986678,-.0960555020299504,-.0889948538176096,-.037951617011481,-.00399677353202701,.0171916593215167,.00815001876222021,-.0276771999733829,-.0360670549159644,.00735834161689026,.0542505250758479,.040142553361595,-.0314755181231815,-.09002010344269,-.0834030927664127,-.0355670254903839,-.00374564662290655,.0149588727246486,.00709152566878545,-.0240825916820202,-.0313828045304254,.00640266850645313,.0472046755160671,.0349289929115195,-.027387598878123,-.0783286386075327,-.0725710198301417,-.0309477170036224,-.00325917644455892,.0117862965139631,.00558750821717743,-.0189749970879345,-.0247269327337649,.00504474841699291,.0371931971558938,.0275210221363089,-.0215790565990407,-.0617161852471821,-.0571796801659621,-.0243841214340887,-.00256794884708116,.00787376028170992,.00373269925985822,-.012676125892433,-.0165186698482411,.00337011203390995,.024846678358121,.0183852434691605,-.0144157512552827,-.0412291042875283,-.0381985209754617,-.0162896569439489,-.00171550186385153,.00347344016787763,.00164664748226028,-.00559196156252404,-.00728706606728562,.00148669531329552,.0109608938499066,.00811048862007311,-.00635938200665632,-.0181878571094644,-.0168509419086443,-.00718603903670932,-.000756778574503009},
			{-.00590524039988117,-.00279948660884738,.00950696592933597,.0123888349467185,-.00252754986470693,-.018634756913296,-.0137887462421145,.010811667317975,.0309214103019088,.0286485035369933,.0122170775898605,.00128660900891531,-.0123301276949928,-.00584532127908038,.019850521903057,.025867857452222,-.00527751801398726,-.0389093274357997,-.0287908688565067,.0225747352519459,.0645638300414449,.0598180062049106,.0255092285058617,.00268643650371509,-.0157171507315389,-.00745100114858395,.025303358778414,.0329736256376032,-.00672722522962469,-.0495975207474624,-.0366995733299215,.0287758996059307,.0822991840529711,.0762497066729088,.0325164831532697,.00342438687611848,-.0152171307379511,-.00721395757687207,.0244983664798579,.031924614123907,-.00651320761449607,-.0480196423883553,-.0355320258059113,.027860433095352,.0796809462956736,.0738239248952259,.0314820150123508,.00331544461723926,-.0109528550630685,-.00519240013976796,.0176332228431619,.0229784232957589,-.00468802037820009,-.0345631638655983,-.0255749349500402,.0200531421424059,.0573520640056069,.0531363476787172,.0226598531261651,.00238636211961347,-.00397906802090637,-.001886349561729,.00640598206739025,.00834784252877614,-.00170311319385134,-.0125564685414928,-.00929113050528727,.0072851157217069,.0208354591112203,.0193039294849313,.00823210901756414,.000866942650275162,.00397906802090637,.001886349561729,-.00640598206739025,-.00834784252877614,.00170311319385134,.0125564685414928,.00929113050528727,-.0072851157217069,-.0208354591112203,-.0193039294849313,-.00823210901756414,-.000866942650275162,.0109528550630685,.00519240013976796,-.0176332228431619,-.0229784232957589,.00468802037820009,.0345631638655983,.0255749349500402,-.0200531421424059,-.0573520640056069,-.0531363476787172,-.0226598531261651,-.00238636211961347,.0152171307379511,.00721395757687207,-.0244983664798579,-.031924614123907,.00651320761449607,.0480196423883553,.0355320258059113,-.027860433095352,-.0796809462956736,-.0738239248952259,-.0314820150123508,-.00331544461723926,.0157171507315389,.00745100114858395,-.025303358778414,-.0329736256376032,.00672722522962468,.0495975207474624,.0366995733299215,-.0287758996059307,-.0822991840529711,-.0762497066729088,-.0325164831532697,-.00342438687611848,.0123301276949928,.00584532127908038,-.019850521903057,-.025867857452222,.00527751801398726,.0389093274357997,.0287908688565067,-.0225747352519459,-.0645638300414449,-.0598180062049106,-.0255092285058617,-.00268643650371509,.00590524039988117,.00279948660884738,-.00950696592933597,-.0123888349467185,.00252754986470693,.018634756913296,.0137887462421145,-.010811667317975,-.0309214103019089,-.0286485035369933,-.0122170775898605,-.00128660900891531},
			{-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599055,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0178593978095905,.00477948696692302,.0371114332777958,.00213864689849874,-.0507609916515148,-.0177342669014864,.0604629216565472,.0470093228739552,-.0582937240306522,-.11263396580014,-.0629261870345667,-.00746684336239069,-.0167372488224798,.00447918028715571,.0347796325246627,.00200427056193216,-.0475715562643991,-.0166199801907683,.0566638906355675,.044055613873787,-.0546309889220372,-.105556902397195,-.0589723831159563,-.00699768360653812,-.0145634793136514,.00389744157752458,.0302625876081274,.00174396360938696,-.0413931454878217,-.014461441080834,.0493046024382294,.0383338401732651,-.047535726181104,-.091847577865363,-.0513132767932828,-.00608885137146429,-.0114747607406823,.00307084650856481,.0238442987914085,.00137409232554846,-.0326142146768307,-.0113943634480598,.038847757751324,.0302037469746293,-.0374540364163844,-.0723679388639163,-.0404304192248481,-.00479748768603915,-.00766564079353282,.00205145944204955,.0159290318673553,.000917953622114331,-.0217877182910991,-.00761193189450264,.0259519970207346,.0201774207025857,-.0250209303639761,-.0483449404162599,-.0270092839331132,-.00320493107818702,-.00338163008424685,.000904983308336037,.00702695245269255,.000404947175075573,-.00961146049296167,-.00335793687533659,.0114484954663725,.00890109186017875,-.0110377635912774,-.0213269456965256,-.0119149082982386,-.00141382718599056},
			{-.00423058507690693,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040535,-.00720462946394862,-.00435422423400533,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415894,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0223429825183335,-.0467666388316428,-.0604843092288456,-.0617437945066473,-.0526925952374313,-.038049798606524,-.0229959578102642,-.0112551190451086,-.00417346338855663,-.00102227348573477,-.000118584980154754,-1.93472111511475e-006,-.0209391191031562,-.0438281782544715,-.0566839343751764,-.0578642831589129,-.0493817925438394,-.035659038099294,-.0215510663844978,-.0105479328022725,-.00391123463011005,-.000958041580001008,-.0001111340001841,-1.81315792676906e-006,-.0182196268418573,-.0381359430175126,-.0493220429740223,-.0503490925971212,-.0429682752411084,-.03102777936875,-.0187520967637296,-.00917800785523133,-.0034032585182064,-.000833614824033867,-9.67003436401486e-005,-1.57767194829646e-006,-.0143554884304917,-.0300478211505446,-.0388614993834644,-.0396707255608121,-.0338552805420153,-.0244472036457214,-.014775028625764,-.00723147552495668,-.00268147304597067,-.000656816303965784,-7.61914981245143e-005,-1.2430688947399e-006,-.00959009257018474,-.0200732554493939,-.0259611770305218,-.0265017754217613,-.0226168044340364,-.0163317986134587,-.00987036372424938,-.00483094114415895,-.00179134098152265,-.000438781946510387,-5.08992448158815e-005,-8.30424254068003e-007,-.00423058507690692,-.00885514027394948,-.0114525451470328,-.0116910253775248,-.00997720456040534,-.00720462946394862,-.00435422423400534,-.00213112724015143,-.000790234334926295,-.000193564800479007,-2.24537546398003e-005,-3.66334363412092e-007},
			{.00719247221789664,.015054738114976,.0194706196180468,.0198760629317139,.0169623740708349,.0122486834132115,.00740267274248333,.00362315688940088,.00134348757825736,.000329082012180178,3.81739176729443e-005,6.2280977297559e-007,.0150178646226051,.0314342568299276,.0406546067448804,.0415011714047538,.0354173960992067,.0255752213886969,.0154567628103504,.00756514283587924,.00280519881915107,.000687122446765829,7.9707047918831e-005,1.30042530201362e-006,.0191431952513503,.0400690864645718,.051822219359513,.0529013310031521,.0451463737261948,.0326006039436125,.0197026565272781,.00964324889396088,.00357577259239878,.000875871469784373,.000101602166457287,1.65764548368426e-006,.0185341802631805,.0387943424159014,.0501735652087417,.051218346368036,.0437101026178578,.0315634596129842,.0190758430317083,.00933646191122561,.00346201420073717,.000848006797988197,9.83698302986383e-005,1.60490972398582e-006,.0133403723495074,.0279230570510056,.0361134958483479,.0368654994163948,.0314612805140234,.0227184746180725,.0137302457033551,.00672011799573213,.00249185870976258,.000610371016113917,7.08037876780903e-005,1.15516807332712e-006,.00484643033229705,.0101441809206661,.0131196893984132,.0133928851388839,.0114295838362937,.00825340564775263,.00498806761185259,.00244135492157639,.000905268557607982,.000221741982078836,2.57223423195746e-005,4.19661568867713e-007,-.00484643033229705,-.0101441809206661,-.0131196893984132,-.0133928851388839,-.0114295838362937,-.00825340564775263,-.00498806761185259,-.00244135492157639,-.000905268557607982,-.000221741982078836,-2.57223423195746e-005,-4.19661568867713e-007,-.0133403723495074,-.0279230570510056,-.0361134958483479,-.0368654994163948,-.0314612805140234,-.0227184746180725,-.0137302457033551,-.00672011799573213,-.00249185870976258,-.000610371016113917,-7.08037876780903e-005,-1.15516807332712e-006,-.0185341802631805,-.0387943424159014,-.0501735652087417,-.051218346368036,-.0437101026178578,-.0315634596129842,-.0190758430317083,-.00933646191122561,-.00346201420073717,-.000848006797988197,-9.83698302986383e-005,-1.60490972398582e-006,-.0191431952513503,-.0400690864645718,-.051822219359513,-.0529013310031521,-.0451463737261948,-.0326006039436125,-.0197026565272781,-.00964324889396088,-.00357577259239877,-.000875871469784373,-.000101602166457287,-1.65764548368426e-006,-.0150178646226051,-.0314342568299276,-.0406546067448804,-.0415011714047538,-.0354173960992067,-.025575221388697,-.0154567628103504,-.00756514283587924,-.00280519881915107,-.000687122446765829,-7.97070479188311e-005,-1.30042530201362e-006,-.00719247221789664,-.015054738114976,-.0194706196180468,-.0198760629317139,-.0169623740708349,-.0122486834132115,-.00740267274248333,-.00362315688940088,-.00134348757825736,-.000329082012180178,-3.81739176729443e-005,-6.2280977297559e-007},
			{-.00894139816848509,-.0187154574574963,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007,-.0155714612898606,-.0325930034464412,-.0421532388985272,-.0430310100838354,-.0367229713548178,-.0265179890644552,-.0160265383805396,-.00784401323233701,-.00290860557878677,-.000712451526903725,-8.26452523302774e-005,-1.34836228448754e-006,-.0124908594969934,-.0261449210871181,-.0338137940058043,-.0345179101026881,-.029457831026034,-.0212717656603367,-.0128559057758339,-.0062921819188208,-.0023331775316716,-.000571502684000991,-6.62950134052773e-005,-1.08160714868461e-006,-.000709434451366774,-.00148493446363173,-.00192049797733382,-.00196048915784558,-.00167309545011213,-.00120815732532064,-.00073016772489491,-.000357372575406282,-.000132515822671862,-3.24592309421519e-005,-3.76531066375748e-006,-6.1431270947059e-008,.0139082644774524,.0291117258431486,.0376508270012981,.038434842937163,.0328005694836934,.0236855872710412,.014314734520114,.0070061896291659,.002597935729815,.000636354166066229,7.38178649914994e-005,1.20434292678016e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102608,.0245005889388461,.0119915442295211,.0044465341160397,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0238048893142939,.0498265914174762,.0644418123347119,.0657837060414473,.0561402845963243,.0405394062102609,.0245005889388461,.0119915442295211,.00444653411603971,.00108916108925411,.000126344024330921,2.06131038958397e-006,.0139082644774524,.0291117258431486,.0376508270012982,.038434842937163,.0328005694836935,.0236855872710413,.0143147345201141,.00700618962916591,.002597935729815,.00063635416606623,7.38178649914996e-005,1.20434292678016e-006,-.00070943445136674,-.00148493446363166,-.00192049797733373,-.00196048915784549,-.00167309545011205,-.00120815732532058,-.000730167724894874,-.000357372575406265,-.000132515822671856,-3.24592309421504e-005,-3.7653106637573e-006,-6.14312709470561e-008,-.0124908594969934,-.0261449210871181,-.0338137940058042,-.0345179101026881,-.029457831026034,-.0212717656603366,-.0128559057758339,-.00629218191882079,-.0023331775316716,-.00057150268400099,-6.62950134052772e-005,-1.08160714868461e-006,-.0155714612898606,-.0325930034464411,-.0421532388985272,-.0430310100838353,-.0367229713548177,-.0265179890644552,-.0160265383805396,-.007844013232337,-.00290860557878677,-.000712451526903724,-8.26452523302773e-005,-1.34836228448754e-006,-.00894139816848509,-.0187154574574964,-.0242051074120083,-.0247091385701999,-.0210869553409934,-.015227080775471,-.00920271118139843,-.00450416593816153,-.00167017084080247,-.000409101795856463,-4.74563108795132e-005,-7.74252578903543e-007},
			{.00998308444276627,.0208958362733153,.0270250386669949,.0275877902097846,.0235436172110032,.0170010584848057,.0102748408129391,.00502890801389247,.00186474824444619,.000456762767608005,5.29850421517036e-005,8.64454163607971e-007,.0124695874152247,.0261003959766295,.0337562087138414,.0344591256926668,.0294076639906539,.0212355395912979,.0128340120159257,.00628146625843626,.00232920410268166,.00057052940815652,6.61821122117002e-005,1.07976515889088e-006,-.00052999672436047,-.00110934900341893,-.0014347451482894,-.00146462133295129,-.00124991830660814,-.000902577290542385,-.000545485917243634,-.000266982092533976,-9.89985076235179e-005,-2.42492960998119e-005,-2.8129481365704e-006,-4.5893418782406e-008,-.0180526136574343,-.0377863636688457,-.0488699244135907,-.0498875594186185,-.0425743995300703,-.0307433581628203,-.0185802026068238,-.00909387614760269,-.00337206199330092,-.000825973357633232,-9.58139241505331e-005,-1.5632099607735e-006,-.0236738567072675,-.0495523238883786,-.0640870961854267,-.0654216035175219,-.0558312637160867,-.0403162594435231,-.0243657268943327,-.011925537482675,-.00442205843166394,-.00108316586637731,-.000125648571112395,-2.04996402831038e-006,-.0108143045303454,-.0226356832070683,-.0292752204756728,-.029884828317171,-.0255039259384566,-.0184166151099846,-.0111303533681376,-.00544762924015274,-.00202001250249669,-.000494794138560608,-5.7396727901756e-005,-9.36431083136384e-007,.0108143045303454,.0226356832070683,.0292752204756728,.0298848283171711,.0255039259384567,.0184166151099846,.0111303533681376,.00544762924015275,.00202001250249669,.000494794138560608,5.73967279017561e-005,9.36431083136386e-007,.0236738567072676,.0495523238883789,.064087096185427,.0654216035175222,.055831263716087,.0403162594435233,.0243657268943328,.0119255374826751,.00442205843166396,.00108316586637732,.000125648571112395,2.04996402831039e-006,.0180526136574342,.0377863636688456,.0488699244135906,.0498875594186183,.0425743995300703,.0307433581628202,.0185802026068238,.00909387614760267,.00337206199330092,.00082597335763323,9.58139241505329e-005,1.56320996077349e-006,.00052999672436045,.00110934900341889,.00143474514828935,.00146462133295124,.00124991830660809,.00090257729054235,.000545485917243613,.000266982092533966,9.89985076235141e-005,2.4249296099811e-005,2.81294813657029e-006,4.58934187824042e-008,-.0124695874152247,-.0261003959766295,-.0337562087138414,-.0344591256926668,-.0294076639906538,-.0212355395912979,-.0128340120159257,-.00628146625843626,-.00232920410268166,-.00057052940815652,-6.61821122117001e-005,-1.07976515889088e-006,-.00998308444276622,-.0208958362733152,-.0270250386669947,-.0275877902097844,-.0235436172110031,-.0170010584848056,-.0102748408129391,-.00502890801389244,-.00186474824444618,-.000456762767608003,-5.29850421517033e-005,-8.64454163607967e-007},
			{.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.00942311384888909,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848839,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0241344144556512,.035879133509499,.0135952558126032,-.03168284598481,-.0737790147824256,-.090582360824203,-.0780224365848942,-.0492222895642595,-.0219539557572737,-.00613187847306387,-.000775628511111868,-1.32599621605223e-005,.0226179910563484,.0336247611193798,.0127410331393487,-.0296921364485932,-.0691433015522925,-.0848908528835268,-.0731200989406408,-.0461295345359716,-.0205745358306472,-.00574659777709663,-.000726893903293736,-1.242680679514e-005,.0196804533623871,.0292577064595496,.0110862767547695,-.0258358359568878,-.060163235458501,-.0738655553848838,-.0636235416960474,-.0401384079957445,-.0179023942426647,-.00500025175811553,-.000632487718627892,-1.08128609196926e-005,.0155064932450492,.0230525394525572,.00873502619300072,-.0203564017743282,-.0474034203917545,-.0581996620975496,-.0501298421011681,-.0316255901727637,-.0141055366094685,-.00393976544559436,-.000498345559214822,-8.51959818828757e-006,.0103590140021362,.0154001020862714,.00583537858704515,-.0135989644906143,-.0316675529294226,-.0388799134053468,-.0334889215791372,-.0211272739908538,-.0094231138488891,-.00263193520102131,-.000332916575284157,-5.69146328124328e-006,.00456978801071811,.00679361972704794,.00257422599289674,-.00599906370187122,-.0139698627374975,-.0171515321922409,-.0147733435144177,-.00932011129270131,-.00415692388111507,-.00116105508923281,-.000146863222116425,-2.51073902021028e-006},
			{-.00776915549838746,-.0115499204629525,-.00437647479049313,.0101990855232791,.0237503437017621,.0291595409512613,.0251163517269084,.0158452413384258,.00706723987011998,.0019739247223913,.000249684056884166,4.2685397699263e-006,-.0162219779196966,-.024116206035927,-.00913806879424977,.0212956659438816,.0495906602970113,.0608850510918503,.0524428817548842,.0330848256515858,.0147563797828581,.00412155005372477,.00052133970783773,8.91270073194743e-006,-.0206780723147828,-.0307408045330656,-.0116482495710665,.0271454764985451,.0632129611343108,.0776098633347217,.0668486732438716,.0421730581025391,.0188098818630402,.00525371878089153,.00066454906008391,1.13609740543078e-005,-.0200202272789457,-.0297628272171986,-.0112776761907307,.0262818797044071,.0612019258671093,.075140810003757,.0647219726899549,.0408313790283736,.0182114707916447,.00508657878993488,.00064340732629019,1.09995399578346e-005,-.0144099864483064,-.0214224309688108,-.00811734846026063,.018916944603062,.0440513941259184,.054084203879524,.0465850230556935,.0293892576875637,.0131080953105517,.00366117379238187,.000463106173741249,7.91715396241739e-006,-.00523501095632017,-.0077825653226702,-.00294895545378721,.00687234596731837,.0160034523084138,.019648276623219,.0169238955894816,.0106768376600795,.00476204629431086,.00133006960033359,.000168242066165155,2.87622669770101e-006,.00523501095632017,.0077825653226702,.00294895545378721,-.00687234596731837,-.0160034523084138,-.019648276623219,-.0169238955894816,-.0106768376600795,-.00476204629431086,-.00133006960033359,-.000168242066165155,-2.87622669770101e-006,.0144099864483064,.0214224309688108,.00811734846026063,-.018916944603062,-.0440513941259184,-.054084203879524,-.0465850230556935,-.0293892576875637,-.0131080953105517,-.00366117379238187,-.000463106173741249,-7.91715396241739e-006,.0200202272789457,.0297628272171986,.0112776761907307,-.0262818797044071,-.0612019258671093,-.075140810003757,-.0647219726899549,-.0408313790283736,-.0182114707916447,-.00508657878993488,-.00064340732629019,-1.09995399578346e-005,.0206780723147828,.0307408045330656,.0116482495710665,-.0271454764985451,-.0632129611343108,-.0776098633347217,-.0668486732438716,-.0421730581025391,-.0188098818630402,-.00525371878089153,-.00066454906008391,-1.13609740543078e-005,.0162219779196966,.024116206035927,.00913806879424977,-.0212956659438816,-.0495906602970113,-.0608850510918503,-.0524428817548842,-.0330848256515858,-.0147563797828581,-.00412155005372477,-.00052133970783773,-8.91270073194743e-006,.00776915549838746,.0115499204629525,.00437647479049313,-.0101990855232791,-.0237503437017621,-.0291595409512613,-.0251163517269084,-.0158452413384258,-.00706723987011998,-.0019739247223913,-.000249684056884166,-4.2685397699263e-006},
			{.00965830811000427,.014358406200055,.0054406610822788,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.0312236592031229,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597266e-006,.0168199612662992,.0250051907000728,.00947492124135092,-.0220806783296889,-.0514186981079021,-.0631294288607147,-.0543760596998002,-.034304416435326,-.0153003374562057,-.00427348086675615,-.000540557614849618,-9.24124553932794e-006,.0134923607368192,.0200582538732903,.00760043696398981,-.0177123164923541,-.0412462081514687,-.0506401301294754,-.0436184959823004,-.0275177542970243,-.012273373825647,-.00342803081073059,-.000433615644121537,-7.41299082085032e-006,.00076631600405656,.00113923436056569,.000431676605520518,-.00100599382582219,-.00234263151050095,-.00287617140711524,-.00247736865298144,-.00156290629377859,-.000697082072575792,-.000194699424652437,-2.46277589356807e-005,-4.21030360420172e-007,-.0150234114472303,-.022334371777439,-.00846289940774852,.0197222282696639,.0459266371382019,.0563865379466385,.0485681212230874,.0306403418180061,.0136660995377323,.00381702789659693,.000482820342463611,8.25418274299939e-006,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962322,1.41275647220829e-005,-.0257135350858614,-.038226713974856,-.0144847967196809,.03375586233291,.0786063937326629,.0965091868084627,.0831274636596914,.0524429159879888,.0233904084424938,.00653308877864433,.000826378073962323,1.41275647220829e-005,-.0150234114472303,-.022334371777439,-.00846289940774854,.0197222282696639,.045926637138202,.0563865379466386,.0485681212230875,.0306403418180062,.0136660995377323,.00381702789659694,.000482820342463612,8.25418274299941e-006,.000766316004056523,.00113923436056563,.000431676605520497,-.00100599382582214,-.00234263151050084,-.0028761714071151,-.00247736865298132,-.00156290629377852,-.000697082072575759,-.000194699424652428,-2.46277589356795e-005,-4.21030360420152e-007,.0134923607368192,.0200582538732903,.0076004369639898,-.0177123164923541,-.0412462081514686,-.0506401301294753,-.0436184959823003,-.0275177542970243,-.012273373825647,-.00342803081073058,-.000433615644121536,-7.41299082085031e-006,.0168199612662992,.0250051907000728,.0094749212413509,-.0220806783296889,-.051418698107902,-.0631294288607146,-.0543760596998001,-.0343044164353259,-.0153003374562056,-.00427348086675614,-.000540557614849618,-9.24124553932793e-006,.00965830811000427,.014358406200055,.00544066108227881,-.0126791014087128,-.0295254918295468,-.0362499927967748,-.031223659203123,-.01969817993161,-.00878571424746134,-.00245390546743038,-.00031039738515256,-5.30648051597267e-006},
			{-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417216e-006,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.0255511390040629,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0246779149055168,-.0190593019026422,.0284444259234539,.0612736165177699,.0313337181406889,-.0436605177951375,-.100744600559549,-.100201142004577,-.0591467378195483,-.0198408192849353,-.00281660053439872,-5.1170066408645e-005,-.0231273420636732,-.0178617600508334,.0266571941209454,.0574236475857209,.0293649451479564,-.0409172222852388,-.0944145746157812,-.0939052628709929,-.0554304057996932,-.0185941728133459,-.00263962673771932,-4.79549278690905e-005,-.0201236518197464,-.01554194334832,.0231950602669512,.0499656850775758,.025551139004063,-.0356030508145879,-.0821523727649525,-.0817092084708625,-.0482313178690934,-.0161792331579558,-.00229680216852035,-4.17267262628691e-005,-.0158556952557302,-.0122457056810621,.0182756991785256,.0393686336322138,.0201320852255895,-.0280521214015666,-.0647289566905582,-.0643797815975754,-.0380021024396871,-.0127478348722089,-.00180968124339263,-3.28770474449435e-005,-.0105922961801923,-.00818066564835097,.0122089643801069,.0262999648338464,.0134491112496156,-.018740040949054,-.0432417670523435,-.0430085028564841,-.0253870623784663,-.00851610985483727,-.00120894602302765,-2.19633020470182e-005,-.00467269839390473,-.00360882877384071,.00538587736593387,.0116019984098135,.00593295726124454,-.00826700441100803,-.0190757256045137,-.0189728231540152,-.0111992795125622,-.0037568070382537,-.000533315916022303,-9.68891772417217e-006},
			{.00794411476554508,.00613541631233657,-.0091565995280034,-.0197247070338324,-.0100866971092938,.0140548407519696,.0324308869658652,.0322559411834705,.0190400394460011,.00638699606694611,.00090669726269933,1.64722539026166e-005,.0165872924470305,.0128107344443715,-.0191189577031385,-.041185140655471,-.0210609992068756,.0293464735505198,.0677156136707053,.0673503272995239,.0397555563854127,.0133360323644821,.00189318169502602,3.43940263715078e-005,.0211437368750014,.0163297776977338,-.0243708376331069,-.052498488222838,-.0268463480086837,.0374077997926938,.0863167465308033,.0858511178610883,.050676204462949,.0169993722707645,.0024132290272223,4.38418895665644e-005,.0204710773480144,.0158102678018768,-.0235955122395174,-.0508283194885088,-.0259922675847922,.0362177209971177,.0835706954313348,.0831198801109617,.0490640092334553,.0164585601249111,.00233645539418102,4.24471188657934e-005,.0147344954209059,.011379778141117,-.0169833742082372,-.0365847692343567,-.0187084900904949,.0260684787183142,.0601517940761608,.0598273101146777,.0353148692221493,.011846400395665,.00168171370375142,3.05522209664918e-005,.0053529019781526,.00413416511271193,-.00616989824205258,-.0132908985350789,-.00679661642648238,.00947043025991277,.0218525744046085,.0217346924690766,.0128295559445646,.0043036845375764,.00061095058597832,1.10993311529656e-005,-.0053529019781526,-.00413416511271193,.00616989824205258,.0132908985350789,.00679661642648238,-.00947043025991277,-.0218525744046085,-.0217346924690766,-.0128295559445646,-.0043036845375764,-.00061095058597832,-1.10993311529656e-005,-.0147344954209059,-.011379778141117,.0169833742082372,.0365847692343567,.0187084900904949,-.0260684787183142,-.0601517940761608,-.0598273101146777,-.0353148692221493,-.011846400395665,-.00168171370375142,-3.05522209664918e-005,-.0204710773480144,-.0158102678018768,.0235955122395174,.0508283194885088,.0259922675847922,-.0362177209971177,-.0835706954313348,-.0831198801109616,-.0490640092334554,-.0164585601249111,-.00233645539418102,-4.24471188657934e-005,-.0211437368750014,-.0163297776977338,.0243708376331069,.052498488222838,.0268463480086837,-.0374077997926938,-.0863167465308033,-.0858511178610884,-.050676204462949,-.0169993722707645,-.0024132290272223,-4.38418895665644e-005,-.0165872924470305,-.0128107344443715,.0191189577031385,.041185140655471,.0210609992068756,-.0293464735505198,-.0677156136707053,-.0673503272995239,-.0397555563854127,-.0133360323644821,-.00189318169502602,-3.43940263715078e-005,-.00794411476554508,-.00613541631233657,.0091565995280034,.0197247070338324,.0100866971092938,-.0140548407519696,-.0324308869658652,-.0322559411834705,-.0190400394460011,-.00638699606694611,-.00090669726269933,-1.64722539026166e-005},
			{.00456148901693467,-2.90196057170823e-005,-.00924310280718151,-.00451266930256049,.0104810735535174,.0123349344803261,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.0154783325752891,-9.84711586118446e-005,-.0313642801168654,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0240905849924321,-.000153261199441414,-.0488155847670616,-.0238327535093197,.0553536777827474,.065144470669814,-.0274094549326145,-.1130665070651,-.105531184311138,-.0455710490124453,-.00750108772197701,-.000146948502626933,.0225769155038872,-.000143631429079044,-.0457483839809208,-.0223352841939668,.0518756728623682,.0610512866466247,-.0256872528506734,-.105962266052499,-.0989004057795262,-.0427077102237528,-.00702977631051804,-.000137715374212508,.0196447125447109,-.00012497713144907,-.0398067598089521,-.0194344633800554,.04513825997485,.0531221803284506,-.0223510912386548,-.0922002944480663,-.0860556014287929,-.0371609970655195,-.00611677599847365,-.000119829431036591,.0154783325752891,-9.84711586118446e-005,-.0313642801168653,-.0153126744376703,.0355650406271124,.0418556785891793,-.0176107246581052,-.0726458489916148,-.0678043629220731,-.0292796481547049,-.00481948987430778,-9.44152163932478e-005,.0103402014461478,-6.57830300343596e-005,-.0209527074731265,-.0102295345829156,.0237589987639806,.0279614193694405,-.0117647323890797,-.0485305965061746,-.0452962725882545,-.0195600823744624,-.00321962949985789,-6.30734836804337e-005,.00456148901693467,-2.90196057170823e-005,-.00924310280718152,-.00451266930256049,.0104810735535174,.0123349344803262,-.00518990832620142,-.0214088462493807,-.0199820526703932,-.0086287584807825,-.00142031126556745,-2.78243131496524e-005},
			{-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.0125539393028941,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855103,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0302448825382416,-.0608321599696493,-.0731299878104439,-.0669515498630353,-.0492363193955515,-.0292481479535268,-.0137419094370502,-.00486056030603143,-.00117656404705173,-.000160687486930479,-7.76736168420821e-006,-2.43706852358544e-008,-.0283445237093819,-.0570099288159738,-.0685350545084463,-.0627448227009383,-.046142682839166,-.0274104163597274,-.0128784721632876,-.00455515959306478,-.00110263769366926,-.000150591105035476,-7.27931963825655e-006,-2.28394163742625e-008,-.0246632459775862,-.0496056985105608,-.0596339852015085,-.0545957664330434,-.0401498486478859,-.0238504568981079,-.0112058657268433,-.00396355298338396,-.000959431350547685,-.000131032911457853,-6.33390959850106e-006,-1.98731208115696e-008,-.019432502397701,-.039084995386339,-.0469864170136372,-.0430167368512078,-.0316346044160782,-.018792094977277,-.00882925194855104,-.00312293657221845,-.000755948833210804,-.000103242588927529,-4.9905723509205e-006,-1.56582985131636e-008,-.0129817593993147,-.0261104821110847,-.0313889764798934,-.0287370569429125,-.0211332958986597,-.012553939302894,-.00589832550261674,-.00208625787714592,-.000505006800981864,-6.89705536523236e-005,-3.3339200581901e-006,-1.04604265441689e-008,-.00572678909872913,-.0115184097714994,-.0138469711844149,-.0126771001809598,-.00932276780442022,-.00553806002210423,-.00260199447163722,-.000920334332232941,-.000222779313164568,-3.04257537549408e-005,-1.47072954119649e-006,-4.61452526260544e-009},
			{.00973618795942079,.0195825969134841,.023541414184445,.0215525014129859,.0158497577055902,.00941532722371888,.00442368433837659,.00156467226043237,.000378749981717022,5.17272161167757e-005,2.50040624924348e-006,7.84521387564148e-009,.0203291369483494,.040888415067351,.0491544159693757,.0450015709053905,.0330942558153189,.0196591805070174,.00923664221622893,.0032670319014163,.000790829047220676,.000108006302341081,5.22084221044592e-006,1.63807876277374e-008,.0259134469295932,.052120253632755,.0626568827202963,.0573632723498304,.0421850786840353,.0250594568791694,.0117739006080788,.00416446886110988,.00100806574315481,.00013767508138138,6.65498087262162e-006,2.08805062376629e-008,.0250890454977502,.0504621179227659,.0606635383392967,.0555383370570115,.0408430171912504,.0242622239912202,.01139932980914,.00403198189010626,.000975995488498737,.000133295133992354,6.44326161446257e-006,2.02162210390667e-008,.0180583766900454,.0363211878310999,.0436638781965725,.0399749050398971,.0293976345040205,.0174632542402221,.00820491125204314,.00290210513530733,.000702493611431951,9.59420214212558e-005,4.63767524981805e-006,1.45510571458537e-008,.00656043641434894,.0131951419194234,.0158626714587474,.0145225026138949,.0106798808777874,.00634423409130352,.00298076618282515,.00105430718024807,.000255209244351746,3.4854823431899e-005,1.68482328777663e-006,5.28626059836008e-009,-.00656043641434894,-.0131951419194234,-.0158626714587474,-.0145225026138949,-.0106798808777874,-.00634423409130352,-.00298076618282515,-.00105430718024807,-.000255209244351746,-3.4854823431899e-005,-1.68482328777663e-006,-5.28626059836008e-009,-.0180583766900454,-.0363211878310999,-.0436638781965725,-.0399749050398971,-.0293976345040205,-.0174632542402221,-.00820491125204314,-.00290210513530733,-.000702493611431951,-9.59420214212558e-005,-4.63767524981805e-006,-1.45510571458537e-008,-.0250890454977502,-.0504621179227659,-.0606635383392967,-.0555383370570115,-.0408430171912504,-.0242622239912202,-.01139932980914,-.00403198189010626,-.000975995488498737,-.000133295133992354,-6.44326161446257e-006,-2.02162210390667e-008,-.0259134469295932,-.052120253632755,-.0626568827202963,-.0573632723498304,-.0421850786840353,-.0250594568791694,-.0117739006080788,-.00416446886110988,-.00100806574315481,-.00013767508138138,-6.65498087262162e-006,-2.08805062376629e-008,-.0203291369483494,-.040888415067351,-.0491544159693757,-.0450015709053905,-.0330942558153189,-.0196591805070174,-.00923664221622893,-.0032670319014163,-.000790829047220676,-.000108006302341081,-5.22084221044592e-006,-1.63807876277374e-008,-.00973618795942079,-.0195825969134841,-.023541414184445,-.0215525014129859,-.0158497577055902,-.00941532722371888,-.00442368433837659,-.00156467226043237,-.000378749981717022,-5.17272161167757e-005,-2.50040624924348e-006,-7.84521387564149e-009},
			{-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642852e-009,-.0210785206129219,-.042395665985065,-.0509663726925087,-.0466604432029348,-.0343141942102539,-.0203838678741335,-.0095771283278592,-.00338746300209602,-.000819981016188288,-.000111987689197342,-5.41329572570134e-006,-1.69846250996501e-008,-.0169084220471927,-.0340082601910801,-.0408833691663904,-.0374293092514581,-.0275255976721928,-.0163511969032008,-.00768242376878656,-.00271729952782253,-.000657758926587005,-8.98324482923012e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.000960335605521408,-.00193154293476799,-.00232202360306468,-.00212584582191778,-.0015633517683723,-.000928687285851265,-.000436333151684879,-.000154332526131126,-3.73582653241091e-005,-5.10214959062846e-006,-2.46629292637403e-007,-7.7381807429253e-010,.0188271115999142,.0378673603101058,.0455226248627925,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606287,.00302564611398512,.000732398368229125,.000100026219157067,4.83509846932233e-006,1.51704874410297e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019803,.0146410459204576,.00517858795138701,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.03222381257348,.0648124230214286,.0779149007348609,.071332206086851,.0524578637726292,.0311618614019804,.0146410459204576,.00517858795138702,.00125354691938223,.000171201308360324,8.27558205213341e-006,2.59652651100395e-008,.0188271115999142,.0378673603101059,.0455226248627926,.0416766141996025,.0306490752231225,.0182066551230808,.00855418969606288,.00302564611398512,.000732398368229127,.000100026219157067,4.83509846932234e-006,1.51704874410297e-008,-.000960335605521362,-.0019315429347679,-.00232202360306457,-.00212584582191767,-.00156335176837222,-.000928687285851221,-.000436333151684858,-.000154332526131119,-3.73582653241074e-005,-5.10214959062822e-006,-2.46629292637392e-007,-7.73818074292493e-010,-.0169084220471927,-.03400826019108,-.0408833691663903,-.0374293092514581,-.0275255976721927,-.0163511969032008,-.00768242376878655,-.00271729952782252,-.000657758926587003,-8.9832448292301e-005,-4.34234880508224e-006,-1.36244480707139e-008,-.0210785206129218,-.0423956659850649,-.0509663726925086,-.0466604432029347,-.0343141942102538,-.0203838678741334,-.00957712832785918,-.00338746300209602,-.000819981016188286,-.000111987689197342,-5.41329572570133e-006,-1.69846250996501e-008,-.012103645386543,-.0243443131722905,-.0292657588754282,-.0267932208563554,-.019703794496435,-.0117047639578373,-.00549935013137315,-.0019451389255596,-.000470847059236027,-6.43052376679644e-005,-3.1084065641785e-006,-9.75286088642853e-009},
			{.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700912,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0310089505146145,.0386669429547266,-.00289747232263218,-.06415515117807,-.101550028365995,-.0960231475730699,-.0624320697068259,-.0280149839425014,-.00808064824961023,-.00125121192864273,-6.57264138306186e-005,-2.15717135955328e-007,.0290605834541834,.0362374057814175,-.00271541715668044,-.0601241284817868,-.0951693954528975,-.0899897818942762,-.0585093123702308,-.0262547350141685,-.00757292165405677,-.00117259526903547,-6.1596664916753e-005,-2.0216310864738e-007,.0252863066365607,.03153103088061,-.00236274921934216,-.0523154379003901,-.0828091603746397,-.078302255105216,-.0509103478951312,-.0228448710011823,-.00658937971362595,-.00102030310506869,-5.35967269662744e-005,-1.75906941576636e-007,.0199234202501376,.0248437222637393,-.00186164180950711,-.0412200354065318,-.0652464484599557,-.0616953973317352,-.0401129461401394,-.0179997803577472,-.00519186068211652,-.000803910505277683,-4.22295802518279e-005,-1.38599439306033e-007,.0133097139424165,.0165966903495362,-.00124365794811817,-.0275367819917078,-.043587474131337,-.0412152170530784,-.0267971980618703,-.0120246385700913,-.00346838944519944,-.000537047290385572,-2.8211201992603e-005,-9.25904722473474e-008,.00587146336394566,.00732148412592855,-.000548628776790219,-.0121476094320865,-.0192282312449617,-.0181817308780028,-.0118213484795678,-.00530456365436809,-.00153004952980131,-.000236913693551003,-1.24451238898953e-005,-4.08454733139085e-008},
			{-.00998215054668693,-.0124473495345263,.000932730855103969,.0206523073750313,.0326901637864827,.0309109950234972,.020097626924235,.00901835704324359,.00260125692755759,.000402780024162244,2.11581155396364e-005,6.94419156685537e-008,-.0208427062366112,-.0259900357702638,.00194753977310106,.0431219679279274,.0682569830460539,.0645420829652721,.0419637964861058,.0188303077287876,.00543141818323182,.000841003717818616,4.41780941541987e-005,1.44994552227784e-007,-.0265680910755723,-.0331293657149398,.00248251899142349,.0549673520444605,.0870068273056559,.0822714631661617,.0534910368290498,.0240028969866065,.00692340098850254,.0010720231392338,5.63135907453441e-005,1.84823814399061e-007,-.0257228630214447,-.0320753995403257,.00240354099144272,.0532186397309965,.0842388222152567,.0796541073115629,.0517892914968888,.0232392771293986,.00670314230567841,.00103791816611484,5.45220496597903e-005,1.78943893536226e-007,-.018514580398419,-.0230869543217517,.00172999999610706,.0383052533138549,.0606326926078268,.0573327460732011,.0372764493748974,.0167269663744151,.00482472993138678,.000747063782032829,3.92434104660595e-005,1.28798691689977e-007,-.00672617088052745,-.00838727081777638,.000628492536515794,.0139159340296408,.0220272802651019,.0208284410901999,.013542179348353,.00607674771591152,.00175277847364417,.000271401163217019,1.42567575958641e-005,4.67913390880377e-008,.00672617088052745,.00838727081777638,-.000628492536515794,-.0139159340296408,-.0220272802651019,-.0208284410901999,-.013542179348353,-.00607674771591152,-.00175277847364417,-.000271401163217019,-1.42567575958641e-005,-4.67913390880377e-008,.018514580398419,.0230869543217517,-.00172999999610706,-.0383052533138549,-.0606326926078268,-.0573327460732011,-.0372764493748974,-.0167269663744151,-.00482472993138678,-.000747063782032829,-3.92434104660595e-005,-1.28798691689977e-007,.0257228630214447,.0320753995403257,-.00240354099144272,-.0532186397309965,-.0842388222152567,-.0796541073115629,-.0517892914968887,-.0232392771293986,-.00670314230567841,-.00103791816611484,-5.45220496597903e-005,-1.78943893536226e-007,.0265680910755723,.0331293657149398,-.00248251899142349,-.0549673520444605,-.0870068273056559,-.0822714631661617,-.0534910368290498,-.0240028969866065,-.00692340098850254,-.0010720231392338,-5.63135907453441e-005,-1.84823814399061e-007,.0208427062366112,.0259900357702638,-.00194753977310106,-.0431219679279274,-.0682569830460539,-.0645420829652721,-.0419637964861058,-.0188303077287876,-.00543141818323182,-.000841003717818616,-4.41780941541987e-005,-1.44994552227784e-007,.00998215054668692,.0124473495345263,-.000932730855103969,-.0206523073750313,-.0326901637864827,-.0309109950234972,-.020097626924235,-.00901835704324359,-.00260125692755759,-.000402780024162244,-2.11581155396364e-005,-6.94419156685537e-008},
			{-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243526,-.0333379180302253,-.0122284658175383,-.00222781405170755,-.000129893528533068,-4.50638134603823e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.0303850838757094,-.0124855976387746,.0513195122862854,.0643776067923383,-.0104517673025705,-.0993711220746294,-.120458590085119,-.0776706287552144,-.0284898603415706,-.00519035765787613,-.000302625737553551,-1.0498960140229e-006,-.028475915859076,-.0117010974617138,.0480949837022282,.060332606673922,-.00979505758496742,-.0931273950927363,-.112889886689072,-.0727903960443939,-.0266997737851217,-.00486423498282151,-.000283611033447067,-9.83928518294983e-007,-.0247775734201869,-.0101814039235535,.0418485921830267,.0524968397466978,-.00852291317573672,-.0810323671681161,-.0982281823586563,-.0633366593439912,-.0232321098480975,-.00423248685016351,-.00024677672313637,-8.56139666338399e-007,-.0195225825235887,-.00802206474915023,.0329730672384375,.0413629643549972,-.00671531763797095,-.0638464892543683,-.077395302748921,-.0499038036470548,-.0183048910405827,-.00333483317397825,-.000194438691014249,-6.74563929419808e-007,-.0130419368534069,-.00535908923760191,.0220274474582062,.0276322647649607,-.00448612515681085,-.0426522403048412,-.0517034388243525,-.0333379180302253,-.0122284658175383,-.00222781405170754,-.000129893528533068,-4.50638134603823e-007,-.00575333585387101,-.00236411513116128,.00971721490872359,.0121897308185766,-.00197901469695356,-.0188156610595505,-.0228085177608314,-.014706729625562,-.00539448025371328,-.000982780594900549,-5.73013888424015e-005,-1.98795053685637e-007},
			{-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.00895405768790681,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.0195503931804071,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0379888820906312,-.0734216553279688,-.0820431132524648,-.0673630603261267,-.0426889454411409,-.0208611494544935,-.00761967619488918,-.0019476747493676,-.000307771494551743,-2.34363985853461e-005,-4.72075244159458e-007,-2.84846595678631e-010,-.0356019491148407,-.0688083958532271,-.0768881468075028,-.0631304769703178,-.0400066961627314,-.019550393180407,-.00714091358400119,-.0018252976529769,-.000288433470137093,-2.19638332046732e-005,-4.4241361935433e-007,-2.66948997885548e-010,-.0309781048820443,-.0598718260236474,-.0669022099971477,-.0549313334090679,-.0348107803231637,-.0170112633011824,-.00621347919028936,-.00158823501355147,-.000250972840295259,-1.91112550167223e-005,-3.84954639910157e-007,-2.32278688674602e-010,-.0244080644511932,-.0471738149883679,-.0527131488435288,-.0432811345737181,-.027427880851956,-.0134034025849059,-.00489568361654803,-.00125139167557263,-.000197744868020521,-1.50580142319076e-005,-3.03310925490448e-007,-1.83015495150399e-010,-.0163056519239563,-.031514166502778,-.0352146831870669,-.0289136861566447,-.0183230210276619,-.0089540576879068,-.00327053024384099,-.000835984234769021,-.000132102199014542,-1.00594104551974e-005,-2.02624931021046e-007,-1.22262335326923e-010,-.00719309508160455,-.0139021976630664,-.0155346480848767,-.0127550186066989,-.00808303973670455,-.00395000387691696,-.0014427656815517,-.000368787100047267,-5.82757244194926e-005,-4.43762053836147e-006,-8.93862080176376e-008,-5.39349550086673e-011},
			{.0122290736621916,.0236353054365663,.0264106554398282,.0216849715365018,.0137420800410902,.00671545250392889,.00245286453144479,.000626979702197685,9.90753102188358e-005,7.54445587508782e-006,1.51966644376378e-007,9.16955121940153e-011,.0255342762759209,.0493504606770767,.0551453847819794,.0452781682034031,.0286934299414573,.0140218486117464,.00512157521848587,.00130913210415982,.000206869008490129,1.5752805648851e-005,3.17305986489221e-007,1.91460008035021e-010,.0325484114177003,.0629067799068687,.0702935400351982,.057715849511425,.0365753684430921,.0178735787347226,.00652844574550074,.00166874399986403,.000263694867445998,2.0080020819912e-005,4.04468318661127e-007,2.44053093348749e-010,.0315129275220274,.0609054853832662,.068057245681468,.0558796974508159,.0354117723303635,.0173049548869921,.00632075203208146,.00161565515581309,.000255305770199198,1.94412019873441e-005,3.91600703559772e-007,2.36288872704961e-010,.0226821030656547,.0438380247522649,.0489856570714547,.0402206064787555,.0254883799410055,.0124556111144829,.00454949635967349,.00116290232753188,.00018376178439362,1.39932206199794e-005,2.81862975520663e-007,1.7007396600062e-010,.0082401922088581,.0159259372453132,.0177960230838209,.0146117636086239,.00925968589413191,.00452500499468858,.00165278873606634,.000422471349823863,6.67589076578628e-005,5.0836038966856e-006,1.02398136897977e-007,6.17862534841391e-011,-.0082401922088581,-.0159259372453132,-.0177960230838209,-.0146117636086239,-.00925968589413191,-.00452500499468858,-.00165278873606634,-.000422471349823863,-6.67589076578628e-005,-5.0836038966856e-006,-1.02398136897977e-007,-6.17862534841391e-011,-.0226821030656547,-.0438380247522649,-.0489856570714547,-.0402206064787555,-.0254883799410055,-.0124556111144829,-.00454949635967349,-.00116290232753188,-.00018376178439362,-1.39932206199794e-005,-2.81862975520663e-007,-1.7007396600062e-010,-.0315129275220274,-.0609054853832662,-.068057245681468,-.0558796974508159,-.0354117723303635,-.0173049548869921,-.00632075203208146,-.00161565515581309,-.000255305770199198,-1.94412019873441e-005,-3.91600703559772e-007,-2.36288872704961e-010,-.0325484114177003,-.0629067799068687,-.0702935400351982,-.057715849511425,-.0365753684430921,-.0178735787347226,-.00652844574550074,-.00166874399986403,-.000263694867445998,-2.0080020819912e-005,-4.04468318661127e-007,-2.44053093348749e-010,-.0255342762759209,-.0493504606770767,-.0551453847819794,-.0452781682034031,-.0286934299414573,-.0140218486117464,-.00512157521848587,-.00130913210415982,-.000206869008490129,-1.5752805648851e-005,-3.17305986489221e-007,-1.91460008035021e-010,-.0122290736621916,-.0236353054365663,-.0264106554398282,-.0216849715365018,-.0137420800410902,-.00671545250392889,-.00245286453144479,-.000626979702197685,-9.90753102188358e-005,-7.54445587508782e-006,-1.51966644376378e-007,-9.16955121940154e-011},
			{.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594673,-.00597366639905114,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.024026051223952,.0244245520757115,-.0153331658656943,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597428,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0373943140340059,.0380145435446571,-.0238646465111032,-.0936981641555206,-.11582950861974,-.0870948447848803,-.0433104513950125,-.0139174385385296,-.00260622847687174,-.000224242431046333,-4.89861273612144e-006,-3.08869273786433e-009,.0350447392015092,.0356259981978926,-.0223651732816189,-.0878108827856643,-.108551661563476,-.0816224658782725,-.0405891514003328,-.013042972348477,-.0024424728633465,-.000210152738910646,-4.59082109730512e-006,-2.89462273257952e-009,.0304932632493343,.0309990305626775,-.019460470587951,-.0764063430332044,-.0944533892283943,-.0710216539142851,-.035317588514516,-.0113490012606163,-.00212525388113525,-.000182858909387888,-3.99458290860084e-006,-2.51868026423359e-009,.024026051223952,.0244245520757115,-.0153331658656942,-.0602015827738834,-.074421092597455,-.0559589139739466,-.0278272018254637,-.00894203035597429,-.00167451604620155,-.000144076988014896,-3.1473854666067e-006,-1.98450197181026e-009,.0160504504217561,.0163166663764535,-.0102432237508319,-.0402172837565659,-.0497165366848783,-.0373829958998318,-.0185897848594672,-.00597366639905115,-.00111864977434158,-9.62497137583584e-005,-2.10259080524915e-006,-1.3257338966573e-009,.00708051518115929,.00719795401054857,-.00451870815871105,-.0177415013722836,-.0219320133393134,-.0164911801868931,-.00820072025726624,-.0026352304461394,-.000493482519274953,-4.2459715555679e-005,-9.27539459961379e-007,-5.84835860352911e-010},
			{-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667211,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122411e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.045582652632387,-.0846551874810735,-.0879280076624583,-.0647472687189144,-.0353576784818276,-.0142140084005323,-.00403612324963366,-.000745564123550642,-7.69094460836833e-005,-3.26541333662429e-006,-2.74086344158761e-008,-3.18041491377665e-012,-.0427185847603009,-.0793360981199224,-.0824032791251887,-.0606790418511473,-.0331360703672659,-.0133209080116225,-.00378252390296373,-.000698718533627157,-7.20770403137805e-005,-3.06023955040496e-006,-2.56864838889783e-008,-2.9805817832234e-012,-.037170459259082,-.0690322308074595,-.0717010581395492,-.052798281255428,-.0288324849829098,-.0115908397087194,-.00329126424530519,-.000607971657616563,-6.27159515122412e-005,-2.66278740669025e-006,-2.23504221467474e-008,-2.5934752839618e-012,-.0292871035439614,-.0543914208171836,-.0564942256781907,-.0416004741639854,-.0227175017461715,-.00913257811383601,-.0025932312557919,-.000479029025826727,-4.94147396187547e-005,-2.09804592275085e-006,-1.76101974716146e-008,-2.04343397133398e-012,-.0195650628997481,-.0363358420848101,-.0377406074727477,-.0277909316793992,-.0151762822814421,-.00610096061448541,-.0017323916158866,-.000320012288242734,-3.30112019225251e-005,-1.40158620956029e-006,-1.17643802055039e-008,-1.36510304341361e-012,-.00863095559575242,-.0160292374818946,-.0166489373902397,-.0122597253338026,-.00669488358667212,-.00269138517084634,-.000764229340216611,-.000141170609267692,-1.45626016852417e-005,-6.18297441736797e-007,-5.18975296351133e-009,-6.0220321353943e-013}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.0058767981695121,.0133218071347991,.0199415226492611,.0253092817487325,.0290869878703852,.0310371252151211,.0310371252151211,.0290869878703852,.0253092817487325,.0199415226492611,.0133218071347991,.0058767981695121,.00550754478349754,.0124847659007004,.0186885487426998,.0237190386071866,.0272593823528392,.0290869878703852,.0290869878703852,.0272593823528392,.0237190386071866,.0186885487426998,.0124847659007004,.00550754478349754,.00479224604797332,.0108632925195224,.0162613519045569,.0206385010917392,.0237190386071866,.0253092817487325,.0253092817487325,.0237190386071866,.0206385010917392,.0162613519045569,.0108632925195224,.00479224604797332,.00377587495588562,.00855933392240393,.0128125373343938,.0162613519045569,.0186885487426998,.0199415226492611,.0199415226492611,.0186885487426998,.0162613519045569,.0128125373343938,.00855933392240393,.00377587495588562,.00252244920371164,.00571800848521633,.00855933392240393,.0108632925195224,.0124847659007004,.0133218071347991,.0133218071347991,.0124847659007004,.0108632925195224,.00855933392240393,.00571800848521633,.00252244920371164,.00111275630348505,.00252244920371164,.00377587495588562,.00479224604797332,.00550754478349754,.0058767981695121,.0058767981695121,.00550754478349754,.00479224604797332,.00377587495588562,.00252244920371164,.00111275630348505},
			{-.00189181133459202,-.00428844840471255,-.0064194136822003,-.00814735927652425,-.00936344787673791,-.00999122031784744,-.00999122031784744,-.00936344787673791,-.00814735927652425,-.0064194136822003,-.00428844840471255,-.00189181133459202,-.00395009749828707,-.0089542699133058,-.0134037202668508,-.0170116664951607,-.0195508565557993,-.0208616438968927,-.0208616438968927,-.0195508565557993,-.0170116664951607,-.0134037202668508,-.0089542699133058,-.00395009749828707,-.00503516908507482,-.0114139620772511,-.0170856536938404,-.0216846843044205,-.0249213768923885,-.026592230814309,-.026592230814309,-.0249213768923885,-.0216846843044205,-.0170856536938404,-.0114139620772511,-.00503516908507482,-.00487498195849973,-.0110508422381554,-.0165420966206557,-.0209948152631386,-.0241285368333486,-.0257462347868881,-.0257462347868881,-.0241285368333486,-.0209948152631386,-.0165420966206557,-.0110508422381554,-.00487498195849973,-.00350887245079362,-.00795407987509009,-.0119065275737857,-.0151114669784336,-.0173670300512356,-.0185314027260889,-.0185314027260889,-.0173670300512356,-.0151114669784336,-.0119065275737857,-.00795407987509009,-.00350887245079362,-.00127473997218044,-.00288964153040104,-.00432552816923864,-.00548985215787403,-.00630927675908016,-.00673228227209459,-.00673228227209459,-.00630927675908016,-.00548985215787403,-.00432552816923864,-.00288964153040104,-.00127473997218044,.00127473997218044,.00288964153040104,.00432552816923864,.00548985215787403,.00630927675908016,.00673228227209459,.00673228227209459,.00630927675908016,.00548985215787403,.00432552816923864,.00288964153040104,.00127473997218044,.00350887245079362,.00795407987509009,.0119065275737857,.0151114669784336,.0173670300512356,.0185314027260889,.0185314027260889,.0173670300512356,.0151114669784336,.0119065275737857,.00795407987509009,.00350887245079362,.00487498195849973,.0110508422381554,.0165420966206557,.0209948152631386,.0241285368333486,.0257462347868881,.0257462347868881,.0241285368333486,.0209948152631386,.0165420966206557,.0110508422381554,.00487498195849973,.00503516908507482,.0114139620772511,.0170856536938404,.0216846843044205,.0249213768923885,.026592230814309,.026592230814309,.0249213768923885,.0216846843044205,.0170856536938404,.0114139620772511,.00503516908507482,.00395009749828707,.0089542699133058,.0134037202668508,.0170116664951607,.0195508565557993,.0208616438968927,.0208616438968927,.0195508565557993,.0170116664951607,.0134037202668508,.0089542699133058,.00395009749828707,.00189181133459202,.00428844840471255,.0064194136822003,.00814735927652425,.00936344787673791,.00999122031784744,.00999122031784744,.00936344787673791,.00814735927652425,.0064194136822003,.00428844840471255,.00189181133459202},
			{.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433,.00409570813119402,.00928434705946777,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946777,.00409570813119402,.00328542799260275,.00744756529155359,.0111483217284699,.014149171084577,.0162611002476871,.0173513258496235,.0173513258496235,.0162611002476871,.014149171084577,.0111483217284699,.00744756529155359,.00328542799260275,.000186600113947178,.000422994062010168,.000633183289828939,.000803620393624444,.000923570130271238,.00098549089736993,.00098549089736993,.000923570130271238,.000803620393624444,.000633183289828939,.000422994062010168,.000186600113947178,-.00365824317003512,-.00829268056477257,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477257,-.00365824317003512,-.00626131850516901,-.0141934562204533,-.0212463013334064,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181941,-.0212463013334064,-.0141934562204533,-.00626131850516901,-.00626131850516901,-.0141934562204533,-.0212463013334065,-.0269652742181941,-.0309901565715314,-.033067891877733,-.033067891877733,-.0309901565715314,-.0269652742181942,-.0212463013334065,-.0141934562204533,-.00626131850516901,-.00365824317003513,-.00829268056477258,-.0124133817305856,-.0157547535962907,-.0181063347156884,-.0193202740779454,-.0193202740779454,-.0181063347156884,-.0157547535962907,-.0124133817305856,-.00829268056477258,-.00365824317003513,.000186600113947169,.000422994062010148,.000633183289828909,.000803620393624405,.000923570130271194,.000985490897369883,.000985490897369883,.000923570130271194,.000803620393624405,.000633183289828908,.000422994062010148,.000186600113947169,.00328542799260274,.00744756529155357,.0111483217284699,.014149171084577,.016261100247687,.0173513258496235,.0173513258496235,.016261100247687,.014149171084577,.0111483217284699,.00744756529155357,.00328542799260274,.00409570813119402,.00928434705946775,.0138978154612632,.0176387597570958,.0202715508166871,.0216306571105218,.0216306571105218,.0202715508166871,.0176387597570958,.0138978154612632,.00928434705946775,.00409570813119401,.00235182533618433,.00533123014261741,.00798036224077431,.0101284761430285,.0116402695913134,.0124206915632952,.0124206915632952,.0116402695913134,.0101284761430285,.00798036224077431,.00533123014261741,.00235182533618433},
			{-.002625816509223,-.00595232643664948,-.00891008639077038,-.0113084587789945,-.0129963784276034,-.0138677207278419,-.0138677207278419,-.0129963784276034,-.0113084587789945,-.00891008639077038,-.00595232643664948,-.002625816509223,-.00327983286987238,-.00743488199977471,-.0111293360047077,-.0141250748788682,-.0162334074016343,-.0173217763364758,-.0173217763364758,-.0162334074016343,-.0141250748788682,-.0111293360047077,-.00743488199977471,-.00327983286987238,.00013940322318602,.000316005892951688,.00047303181976978,.000600360153697322,.000689970923782976,.000736230030131388,.000736230030131388,.000689970923782976,.000600360153697322,.00047303181976978,.000316005892951688,.00013940322318602,.00474831714066722,.0107637123716435,.0161122895623198,.0204493148954714,.0235016141598744,.0250772800775341,.0250772800775341,.0235016141598744,.0204493148954714,.0161122895623198,.0107637123716435,.00474831714066722,.00622685344747993,.0141153291794728,.0211293523233008,.0268168454598161,.0308195730860963,.032885871621092,.032885871621092,.0308195730860963,.0268168454598161,.0211293523233008,.0141153291794728,.00622685344747993,.00284444948195564,.00644793411485122,.00965196559979983,.012250033314455,.0140784939678192,.0150223867134904,.0150223867134904,.0140784939678192,.012250033314455,.00965196559979983,.00644793411485122,.00284444948195564,-.00284444948195564,-.00644793411485122,-.00965196559979984,-.012250033314455,-.0140784939678192,-.0150223867134905,-.0150223867134905,-.0140784939678192,-.012250033314455,-.00965196559979984,-.00644793411485122,-.00284444948195564,-.00622685344747996,-.0141153291794729,-.0211293523233009,-.0268168454598162,-.0308195730860964,-.0328858716210921,-.0328858716210921,-.0308195730860964,-.0268168454598162,-.0211293523233009,-.0141153291794729,-.00622685344747996,-.00474831714066721,-.0107637123716435,-.0161122895623198,-.0204493148954713,-.0235016141598743,-.025077280077534,-.025077280077534,-.0235016141598743,-.0204493148954713,-.0161122895623198,-.0107637123716435,-.00474831714066721,-.000139403223186015,-.000316005892951675,-.000473031819769762,-.000600360153697299,-.000689970923782949,-.00073623003013136,-.00073623003013136,-.000689970923782949,-.000600360153697299,-.000473031819769762,-.000316005892951675,-.000139403223186015,.00327983286987238,.00743488199977471,.0111293360047077,.0141250748788682,.0162334074016343,.0173217763364758,.0173217763364758,.0162334074016343,.0141250748788682,.0111293360047077,.00743488199977471,.00327983286987238,.00262581650922298,.00595232643664945,.00891008639077034,.0113084587789945,.0129963784276034,.0138677207278418,.0138677207278418,.0129963784276034,.0113084587789945,.00891008639077034,.00595232643664945,.00262581650922298},
			{.00274788724207344,.00622904221162261,.00932430450984374,.0118341740548762,.0136005628530905,.0145124127031825,.0145124127031825,.0136005628530905,.0118341740548762,.00932430450984374,.00622904221162261,.00274788724207344,.00176296411460539,.00399637136462944,.00598220115906356,.00759246007816479,.00872572348723498,.00931073969132332,.00931073969132332,.00872572348723498,.00759246007816479,.00598220115906356,.00399637136462943,.00176296411460539,-.00351886839094271,-.00797673915025254,-.0119404464291104,-.0151545159411986,-.0174165045748593,-.0185841942695622,-.0185841942695622,-.0174165045748593,-.0151545159411986,-.0119404464291104,-.00797673915025254,-.00351886839094271,-.00572156044449335,-.0129699068358449,-.0194147601979758,-.0246407280214811,-.0283186446851331,-.0302172684545958,-.0302172684545958,-.0283186446851331,-.0246407280214811,-.0194147601979758,-.0129699068358449,-.00572156044449335,-.000863900578693703,-.0019583311458107,-.00293144199610321,-.00372051285723621,-.00427584288738108,-.00456251680948201,-.00456251680948201,-.00427584288738108,-.00372051285723621,-.00293144199610321,-.0019583311458107,-.000863900578693703,.00559347785892407,.0126795631056259,.0189801422806282,.0240891218318902,.0276847048244469,.0295408260906545,.0295408260906545,.0276847048244469,.0240891218318902,.0189801422806282,.0126795631056259,.00559347785892407,.00559347785892411,.012679563105626,.0189801422806283,.0240891218318904,.0276847048244471,.0295408260906548,.0295408260906548,.0276847048244471,.0240891218318904,.0189801422806283,.012679563105626,.00559347785892411,-.000863900578693557,-.00195833114581037,-.00293144199610271,-.00372051285723558,-.00427584288738036,-.00456251680948123,-.00456251680948123,-.00427584288738036,-.00372051285723558,-.00293144199610271,-.00195833114581037,-.000863900578693557,-.00572156044449333,-.0129699068358448,-.0194147601979757,-.024640728021481,-.028318644685133,-.0302172684545956,-.0302172684545956,-.028318644685133,-.024640728021481,-.0194147601979757,-.0129699068358448,-.00572156044449333,-.00351886839094282,-.00797673915025279,-.0119404464291107,-.0151545159411991,-.0174165045748599,-.0185841942695628,-.0185841942695628,-.0174165045748599,-.0151545159411991,-.0119404464291107,-.0079767391502528,-.00351886839094282,.00176296411460562,.00399637136462996,.00598220115906434,.00759246007816578,.00872572348723612,.00931073969132453,.00931073969132453,.00872572348723612,.00759246007816578,.00598220115906434,.00399637136462996,.00176296411460562,.00274788724207347,.00622904221162268,.00932430450984385,.0118341740548764,.0136005628530906,.0145124127031827,.0145124127031827,.0136005628530906,.0118341740548764,.00932430450984385,.00622904221162268,.00274788724207347},
			{-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.00700310156746219,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746218,-.00308935672586509,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00719757841895054,-.0163158149674573,-.0244232775945836,-.0309974130231087,-.035624139221637,-.0380125599333229,-.0380125599333229,-.035624139221637,-.0309974130231087,-.0244232775945836,-.0163158149674573,-.00719757841895054,-.00674533722814519,-.0152906530087609,-.0228887042283994,-.0290497708910623,-.0333857887368975,-.035624139221637,-.035624139221637,-.0333857887368975,-.0290497708910623,-.0228887042283994,-.0152906530087609,-.00674533722814519,-.00586927877022147,-.0133047618008894,-.0199160073487628,-.0252768983675548,-.0290497708910623,-.0309974130231087,-.0309974130231087,-.0290497708910623,-.0252768983675548,-.0199160073487628,-.0133047618008894,-.00586927877022147,-.0046244834876462,-.0104830003249202,-.0156920893912011,-.0199160073487628,-.0228887042283994,-.0244232775945836,-.0244232775945836,-.0228887042283994,-.0199160073487628,-.0156920893912011,-.0104830003249202,-.0046244834876462,-.00308935672586509,-.00700310156746218,-.0104830003249202,-.0133047618008894,-.0152906530087609,-.0163158149674573,-.0163158149674573,-.0152906530087609,-.0133047618008894,-.0104830003249202,-.00700310156746219,-.00308935672586509,-.00136284257592261,-.00308935672586509,-.0046244834876462,-.00586927877022147,-.00674533722814519,-.00719757841895054,-.00719757841895054,-.00674533722814519,-.00586927877022147,-.0046244834876462,-.00308935672586509,-.00136284257592261},
			{.00231698622849408,.00525225518720616,.00786214398058468,.00997843648419155,.0114678347596974,.012236695836953,.012236695836953,.0114678347596974,.00997843648419155,.00786214398058468,.00525225518720616,.00231698622849407,.00483786165004334,.0109666961477544,.0164161376459761,.0208349512830896,.0239448112857507,.0255501913584168,.0255501913584168,.0239448112857507,.0208349512830896,.0164161376459761,.0109666961477544,.00483786165004334,.00616679751037299,.0139791915092038,.0209255667251747,.0265582058759676,.03052232852132,.0325686982919874,.0325686982919874,.03052232852132,.0265582058759676,.0209255667251747,.0139791915092038,.00616679751037299,.00597060914873778,.0135344623487989,.0202598479878244,.0257132923061589,.029551301725675,.0315325689967173,.0315325689967173,.029551301725675,.0257132923061589,.0202598479878244,.0135344623487989,.00597060914873778,.0042974735362733,.0097417185286608,.0145824585745998,.0185076916715506,.0212701809756486,.0226962404368325,.0226962404368325,.0212701809756486,.0185076916715506,.0145824585745998,.0097417185286608,.0042974735362733,.00156123124248536,.00353907364270423,.00529766843861862,.00672366827160686,.0077272543489116,.00824532818128076,.00824532818128076,.0077272543489116,.00672366827160686,.00529766843861862,.00353907364270423,.00156123124248536,-.00156123124248536,-.00353907364270423,-.00529766843861862,-.00672366827160686,-.0077272543489116,-.00824532818128076,-.00824532818128076,-.0077272543489116,-.00672366827160686,-.00529766843861862,-.00353907364270423,-.00156123124248536,-.0042974735362733,-.0097417185286608,-.0145824585745998,-.0185076916715506,-.0212701809756486,-.0226962404368325,-.0226962404368325,-.0212701809756486,-.0185076916715506,-.0145824585745998,-.0097417185286608,-.0042974735362733,-.00597060914873778,-.0135344623487989,-.0202598479878244,-.0257132923061589,-.029551301725675,-.0315325689967173,-.0315325689967173,-.029551301725675,-.0257132923061589,-.0202598479878244,-.0135344623487989,-.00597060914873778,-.00616679751037299,-.0139791915092038,-.0209255667251747,-.0265582058759676,-.03052232852132,-.0325686982919874,-.0325686982919874,-.03052232852132,-.0265582058759676,-.0209255667251747,-.0139791915092038,-.00616679751037299,-.00483786165004334,-.0109666961477544,-.0164161376459761,-.0208349512830896,-.0239448112857507,-.0255501913584168,-.0255501913584168,-.0239448112857507,-.0208349512830896,-.0164161376459761,-.0109666961477544,-.00483786165004334,-.00231698622849407,-.00525225518720616,-.00786214398058468,-.00997843648419155,-.0114678347596974,-.012236695836953,-.012236695836953,-.0114678347596974,-.00997843648419155,-.00786214398058468,-.00525225518720616,-.00231698622849408},
			{-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.0264920363666131,-.0264920363666131,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947058,-.004023811085128,-.00912136739713697,-.01365384986449,-.0173291247239877,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239877,-.01365384986449,-.00912136739713697,-.004023811085128,-.000228537532606818,-.000518059808186956,-.000775487987034864,-.000984229955847933,-.00113113778066232,-.00120697492261531,-.00120697492261531,-.00113113778066232,-.000984229955847932,-.000775487987034864,-.000518059808186956,-.000228537532606818,.004480414561763,.0101564179939682,.0152032256144154,.019295553671226,.0221756405874862,.0236624065959092,.0236624065959092,.0221756405874862,.019295553671226,.0152032256144154,.0101564179939682,.004480414561763,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936553,.0404997309936553,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.00766851772899668,.0173833627170427,.0260212985997005,.0330255813114713,.0379550353327309,.0404997309936554,.0404997309936554,.0379550353327309,.0330255813114713,.0260212985997005,.0173833627170427,.00766851772899668,.004480414561763,.0101564179939682,.0152032256144154,.0192955536712261,.0221756405874863,.0236624065959092,.0236624065959092,.0221756405874863,.0192955536712261,.0152032256144154,.0101564179939682,.004480414561763,-.000228537532606807,-.000518059808186931,-.000775487987034827,-.000984229955847885,-.00113113778066226,-.00120697492261525,-.00120697492261525,-.00113113778066226,-.000984229955847886,-.000775487987034827,-.000518059808186931,-.000228537532606807,-.00402381108512799,-.00912136739713696,-.01365384986449,-.0173291247239876,-.0199156991358028,-.0212509473507201,-.0212509473507201,-.0199156991358028,-.0173291247239876,-.01365384986449,-.00912136739713696,-.00402381108512799,-.00501619752947057,-.011370956447737,-.0170212782133727,-.0216029805548362,-.0248274779032066,-.026492036366613,-.026492036366613,-.0248274779032066,-.0216029805548362,-.0170212782133727,-.011370956447737,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.00977390772832783,-.0124047992138418,-.0142563604866286,-.0152121782945392,-.0152121782945392,-.0142563604866286,-.0124047992138418,-.00977390772832783,-.00652939677677674,-.0028803860195172},
			{.00321595530276235,.0072900812758542,.0109125826103317,.0138499768923832,.0159172478252584,.0169844198386609,.0169844198386609,.0159172478252584,.0138499768923832,.0109125826103317,.0072900812758542,.00321595530276235,.00401695848624302,.00910583359827495,.0136305971932344,.0172996130152496,.019881782459596,.0212147567306737,.0212147567306737,.019881782459596,.0172996130152496,.0136305971932344,.00910583359827495,.00401695848624302,-.000170733382645958,-.00038702659670719,-.000579343295245751,-.000735288019200337,-.000845038350279965,-.000901693953533159,-.000901693953533159,-.000845038350279965,-.000735288019200337,-.000579343295245751,-.00038702659670719,-.000170733382645958,-.00581547706554894,-.0131828015237968,-.0197334440070673,-.0250451935407354,-.0287834814106213,-.0307132701622272,-.0307132701622272,-.0287834814106213,-.0250451935407354,-.0197334440070673,-.0131828015237968,-.00581547706554894,-.00762630682441431,-.0172876770198974,-.0258780658927918,-.0328437939425454,-.0377461140742206,-.0402768026076231,-.0402768026076231,-.0377461140742206,-.0328437939425454,-.0258780658927918,-.0172876770198974,-.00762630682441431,-.00348372491482343,-.00789707423793068,-.0118211953667475,-.015003165475677,-.0172425632833398,-.0183985910826998,-.0183985910826998,-.0172425632833398,-.015003165475677,-.0118211953667475,-.00789707423793069,-.00348372491482343,.00348372491482344,.00789707423793069,.0118211953667475,.015003165475677,.0172425632833398,.0183985910826999,.0183985910826999,.0172425632833398,.015003165475677,.0118211953667475,.00789707423793069,.00348372491482344,.00762630682441434,.0172876770198975,.0258780658927919,.0328437939425455,.0377461140742208,.0402768026076233,.0402768026076233,.0377461140742208,.0328437939425455,.0258780658927919,.0172876770198975,.00762630682441434,.00581547706554893,.0131828015237968,.0197334440070673,.0250451935407354,.0287834814106213,.0307132701622271,.0307132701622271,.0287834814106213,.0250451935407354,.0197334440070673,.0131828015237968,.00581547706554893,.000170733382645952,.000387026596707175,.000579343295245729,.000735288019200309,.000845038350279932,.000901693953533124,.000901693953533124,.000845038350279932,.000735288019200309,.000579343295245729,.000387026596707175,.000170733382645952,-.00401695848624301,-.00910583359827495,-.0136305971932344,-.0172996130152496,-.019881782459596,-.0212147567306737,-.0212147567306737,-.019881782459596,-.0172996130152496,-.0136305971932344,-.00910583359827495,-.00401695848624301,-.00321595530276233,-.00729008127585417,-.0109125826103316,-.0138499768923831,-.0159172478252584,-.0169844198386609,-.0169844198386609,-.0159172478252584,-.0138499768923831,-.0109125826103316,-.00729008127585417,-.00321595530276233},
			{.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00831104767246079,.0188398803203813,.0282015717775152,.0357927294935366,.0411352127240282,.0438931234047165,.0438931234047165,.0411352127240282,.0357927294935366,.0282015717775152,.0188398803203813,.00831104767246079,.00778884452614458,.0176561252551656,.0264295990860242,.0335437860758849,.0385505882151287,.0411352127240282,.0411352127240282,.0385505882151287,.0335437860758849,.0264295990860242,.0176561252551656,.00778884452614458,.00677725935348477,.0153630156090817,.0229970243998789,.0291872481432894,.0335437860758849,.0357927294935366,.0357927294935366,.0335437860758849,.0291872481432894,.0229970243998789,.0153630156090817,.00677725935348477,.00533989357102959,.0121047261147503,.018119664061931,.0229970243998789,.0264295990860242,.0282015717775152,.0282015717775152,.0264295990860242,.0229970243998789,.018119664061931,.0121047261147503,.00533989357102959,.0035672818733451,.00808648514742401,.0121047261147503,.0153630156090817,.0176561252551656,.0188398803203813,.0188398803203813,.0176561252551656,.0153630156090817,.0121047261147503,.00808648514742401,.0035672818733451,.00157367505558954,.0035672818733451,.00533989357102959,.00677725935348477,.00778884452614458,.00831104767246079,.00831104767246079,.00778884452614458,.00677725935348477,.00533989357102959,.0035672818733451,.00157367505558954},
			{-.00267542524640207,-.00606478189450901,-.00907842189039497,-.0115221059845388,-.0132419149757324,-.0141297192758912,-.0141297192758912,-.0132419149757324,-.0115221059845388,-.00907842189039497,-.00606478189450901,-.00267542524640207,-.00558628145387762,-.0126632499505153,-.0189557229845952,-.0240581294721656,-.0276490864927876,-.0295028197276516,-.0295028197276516,-.0276490864927876,-.0240581294721656,-.0189557229845952,-.0126632499505153,-.00558628145387762,-.00712080440781243,-.0161417799674716,-.0241627631719636,-.0306667746341718,-.0352441491885744,-.0376070934653196,-.0376070934653196,-.0352441491885744,-.0306667746341718,-.0241627631719636,-.0161417799674716,-.00712080440781243,-.00689426560092869,-.0156282509663382,-.0233940573872652,-.029691152479886,-.0341229040244673,-.0364106744098191,-.0364106744098191,-.0341229040244673,-.029691152479886,-.0233940573872652,-.0156282509663382,-.00689426560092869,-.00496229500775375,-.0112487676337471,-.0168383727729163,-.0213708415448263,-.0245606894326591,-.0262073610608292,-.0262073610608292,-.0245606894326591,-.0213708415448263,-.0168383727729163,-.0112487676337471,-.00496229500775375,-.00180275455686753,-.00408657024203424,-.006117220600383,-.0077638233758434,-.00892266476002541,-.00952088489339305,-.00952088489339305,-.00892266476002541,-.0077638233758434,-.006117220600383,-.00408657024203424,-.00180275455686753,.00180275455686753,.00408657024203424,.006117220600383,.0077638233758434,.00892266476002541,.00952088489339305,.00952088489339305,.00892266476002541,.0077638233758434,.006117220600383,.00408657024203424,.00180275455686753,.00496229500775375,.0112487676337471,.0168383727729163,.0213708415448263,.0245606894326591,.0262073610608292,.0262073610608292,.0245606894326591,.0213708415448263,.0168383727729163,.0112487676337471,.00496229500775375,.00689426560092869,.0156282509663382,.0233940573872652,.029691152479886,.0341229040244673,.0364106744098191,.0364106744098191,.0341229040244673,.029691152479886,.0233940573872652,.0156282509663382,.00689426560092869,.00712080440781243,.0161417799674716,.0241627631719636,.0306667746341718,.0352441491885744,.0376070934653196,.0376070934653196,.0352441491885744,.0306667746341718,.0241627631719636,.0161417799674716,.00712080440781243,.00558628145387762,.0126632499505153,.0189557229845952,.0240581294721656,.0276490864927876,.0295028197276516,.0295028197276516,.0276490864927876,.0240581294721656,.0189557229845952,.0126632499505153,.00558628145387762,.00267542524640207,.00606478189450901,.00907842189039497,.0115221059845388,.0132419149757324,.0141297192758912,.0141297192758912,.0132419149757324,.0115221059845388,.00907842189039497,.00606478189450901,.00267542524640207},
			{.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085,.00579220598764257,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001756,.0305903686539505,.0305903686539505,.0286683021001756,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764257,.00464629682613013,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613013,.000263892411929397,.00059820393939992,.000895456396096566,.0011364908598568,.00130612540425466,.00139369459289309,.00139369459289309,.00130612540425466,.0011364908598568,.000895456396096566,.00059820393939992,.000263892411929396,-.0051735371064033,-.011727621325126,-.0175551728012975,-.0222805862115142,-.0256062241241532,-.0273229936344479,-.0273229936344479,-.0256062241241532,-.0222805862115142,-.0175551728012975,-.011727621325126,-.0051735371064033,-.00885484154985533,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985533,-.00885484154985534,-.0200725782873315,-.0300468075010849,-.0381346565189728,-.0438266997309877,-.0467650611805397,-.0467650611805397,-.0438266997309877,-.0381346565189728,-.0300468075010849,-.0200725782873315,-.00885484154985534,-.0051735371064033,-.011727621325126,-.0175551728012976,-.0222805862115143,-.0256062241241533,-.027322993634448,-.027322993634448,-.0256062241241533,-.0222805862115143,-.0175551728012976,-.011727621325126,-.0051735371064033,.000263892411929384,.000598203939399891,.000895456396096523,.00113649085985675,.0013061254042546,.00139369459289302,.00139369459289302,.0013061254042546,.00113649085985675,.000895456396096523,.000598203939399891,.000263892411929384,.00464629682613012,.0105324478437675,.0157661077887853,.020009949647553,.0229966685133031,.0245384803458705,.0245384803458705,.0229966685133031,.020009949647553,.0157661077887853,.0105324478437675,.00464629682613012,.00579220598764256,.0131300495315137,.0196544791160234,.0249449732761729,.0286683021001755,.0305903686539505,.0305903686539505,.0286683021001755,.0249449732761729,.0196544791160234,.0131300495315137,.00579220598764256,.00332598328733085,.00753949797310551,.0112859365154748,.0143238283300821,.0164618271285175,.0175655104658559,.0175655104658559,.0164618271285175,.0143238283300821,.0112859365154748,.00753949797310551,.00332598328733085},
			{-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00929203378188851,-.0210636265466019,-.0315303157900712,-.0400174881323313,-.045990565969605,-.0490740038492001,-.0490740038492001,-.045990565969605,-.0400174881323313,-.0315303157900712,-.0210636265466019,-.00929203378188851,-.00870819291515191,-.0197401481490573,-.0295491900934304,-.0375030929520919,-.0431008679198909,-.045990565969605,-.045990565969605,-.0431008679198909,-.0375030929520919,-.0295491900934304,-.0197401481490573,-.00870819291515191,-.00757720630936466,-.0171763736242654,-.0257114549245893,-.0326323354691463,-.0375030929520919,-.0400174881323313,-.0400174881323313,-.0375030929520919,-.0326323354691463,-.0257114549245893,-.0171763736242654,-.00757720630936466,-.00597018250997528,-.0135334952236491,-.0202584002902347,-.0257114549245893,-.0295491900934304,-.0315303157900712,-.0315303157900712,-.0295491900934304,-.0257114549245893,-.0202584002902347,-.0135334952236491,-.00597018250997528,-.00398834238269105,-.00904096524624502,-.0135334952236491,-.0171763736242654,-.0197401481490573,-.0210636265466019,-.0210636265466019,-.0197401481490573,-.0171763736242654,-.0135334952236491,-.00904096524624502,-.00398834238269105,-.00175942219976748,-.00398834238269105,-.00597018250997528,-.00757720630936466,-.00870819291515191,-.00929203378188851,-.00929203378188851,-.00870819291515191,-.00757720630936466,-.00597018250997528,-.00398834238269105,-.00175942219976748},
			{.00299121635954409,.0067806322917519,.0101499842366785,.0128821061114311,.014804911017555,.0157975064003936,.0157975064003936,.014804911017555,.0128821061114311,.0101499842366785,.0067806322917519,.00299121635954408,.0062456525355466,.0141579438513248,.0211931425760292,.0268978064529921,.0309126184537943,.0329851552162446,.0329851552162446,.0309126184537943,.0268978064529921,.0211931425760292,.0141579438513248,.0062456525355466,.00796130135439456,.0180470586407878,.0270147904857234,.0342864963629789,.0394041566935386,.042046008708203,.042046008708203,.0394041566935386,.0342864963629789,.0270147904857234,.0180470586407878,.00796130135439456,.00770802326855251,.017472915763368,.0261553512911662,.0331957176344168,.0381505664904682,.0407083715394964,.0407083715394964,.0381505664904682,.0331957176344168,.0261553512911662,.017472915763368,.0077080232685525,.00554801448032919,.0125765045448471,.0188258730735172,.0238933272129638,.0274596855701536,.0293007204185775,.0293007204185775,.0274596855701536,.0238933272129638,.0188258730735172,.0125765045448471,.0055480144803292,.00201554086775423,.00456892442756064,.00683926054723933,.00868021841599365,.0099758424709522,.0106446729127462,.0106446729127462,.0099758424709522,.00868021841599365,.00683926054723933,.00456892442756064,.00201554086775423,-.00201554086775423,-.00456892442756064,-.00683926054723933,-.00868021841599365,-.0099758424709522,-.0106446729127462,-.0106446729127462,-.0099758424709522,-.00868021841599365,-.00683926054723933,-.00456892442756064,-.00201554086775423,-.00554801448032919,-.0125765045448471,-.0188258730735172,-.0238933272129638,-.0274596855701536,-.0293007204185775,-.0293007204185775,-.0274596855701536,-.0238933272129638,-.0188258730735172,-.0125765045448471,-.0055480144803292,-.00770802326855251,-.017472915763368,-.0261553512911662,-.0331957176344168,-.0381505664904682,-.0407083715394964,-.0407083715394964,-.0381505664904682,-.0331957176344168,-.0261553512911662,-.017472915763368,-.0077080232685525,-.00796130135439456,-.0180470586407878,-.0270147904857234,-.0342864963629789,-.0394041566935386,-.042046008708203,-.042046008708203,-.0394041566935386,-.0342864963629789,-.0270147904857234,-.0180470586407878,-.00796130135439456,-.0062456525355466,-.0141579438513248,-.0211931425760292,-.0268978064529921,-.0309126184537943,-.0329851552162446,-.0329851552162446,-.0309126184537943,-.0268978064529921,-.0211931425760292,-.0141579438513248,-.0062456525355466,-.00299121635954408,-.0067806322917519,-.0101499842366785,-.0128821061114311,-.014804911017555,-.0157975064003935,-.0157975064003935,-.014804911017555,-.0128821061114311,-.0101499842366785,-.0067806322917519,-.00299121635954409},
			{.00192735045455837,.0043690101814277,.00654000726824865,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824865,.0043690101814277,.00192735045455837,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0101789130179564,.023074046811849,.034539730417403,.0438369619027918,.0503801408431869,.0537578778068477,.0537578778068477,.0503801408431869,.0438369619027918,.034539730417403,.023074046811849,.0101789130179564,.0095393473923531,.021624248865999,.0323695159501406,.0410825799845608,.0472146352298162,.0503801408431869,.0503801408431869,.0472146352298162,.0410825799845608,.0323695159501406,.021624248865999,.0095393473923531,.008300413639527,.0188157745859791,.0281654877054601,.0357469324918558,.0410825799845608,.0438369619027918,.0438369619027918,.0410825799845608,.0357469324918558,.0281654877054601,.0188157745859791,.008300413639527,.00654000726824865,.0148252012362415,.0221919656425669,.0281654877054601,.0323695159501406,.034539730417403,.034539730417403,.0323695159501406,.0281654877054601,.0221919656425669,.0148252012362415,.00654000726824864,.0043690101814277,.0099038812169698,.0148252012362415,.0188157745859791,.021624248865999,.023074046811849,.023074046811849,.021624248865999,.0188157745859791,.0148252012362415,.0099038812169698,.0043690101814277,.00192735045455837,.0043690101814277,.00654000726824864,.008300413639527,.0095393473923531,.0101789130179564,.0101789130179564,.0095393473923531,.008300413639527,.00654000726824864,.0043690101814277,.00192735045455837},
			{.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00999122031784744,.0208616438968927,.026592230814309,.0257462347868881,.0185314027260889,.00673228227209459,-.00673228227209459,-.0185314027260889,-.0257462347868881,-.026592230814309,-.0208616438968927,-.00999122031784744,.00936344787673791,.0195508565557993,.0249213768923885,.0241285368333486,.0173670300512356,.00630927675908016,-.00630927675908016,-.0173670300512356,-.0241285368333486,-.0249213768923885,-.0195508565557993,-.00936344787673791,.00814735927652425,.0170116664951607,.0216846843044205,.0209948152631386,.0151114669784336,.00548985215787403,-.00548985215787403,-.0151114669784336,-.0209948152631386,-.0216846843044205,-.0170116664951607,-.00814735927652425,.0064194136822003,.0134037202668508,.0170856536938404,.0165420966206557,.0119065275737857,.00432552816923864,-.00432552816923864,-.0119065275737857,-.0165420966206557,-.0170856536938404,-.0134037202668508,-.0064194136822003,.00428844840471255,.0089542699133058,.0114139620772511,.0110508422381554,.00795407987509009,.00288964153040104,-.00288964153040104,-.00795407987509009,-.0110508422381554,-.0114139620772511,-.0089542699133058,-.00428844840471255,.00189181133459202,.00395009749828707,.00503516908507482,.00487498195849973,.00350887245079362,.00127473997218044,-.00127473997218044,-.00350887245079362,-.00487498195849973,-.00503516908507482,-.00395009749828707,-.00189181133459202},
			{-.003216292833121,-.00671561166886112,-.00856035585883146,-.00828801966032105,-.00596547927875537,-.00216720185657465,.00216720185657465,.00596547927875537,.00828801966032105,.00856035585883146,.00671561166886112,.003216292833121,-.00671561166886112,-.0140221809477405,-.0178740023617151,-.0173053650368621,-.0124559063286174,-.0045251122431745,.0045251122431745,.0124559063286174,.0173053650368621,.0178740023617151,.0140221809477405,.00671561166886112,-.00856035585883146,-.0178740023617151,-.0227838994245812,-.0220590603339104,-.0158774801127414,-.00576813744045716,.00576813744045716,.0158774801127414,.0220590603339104,.0227838994245812,.0178740023617151,.00856035585883146,-.00828801966032105,-.0173053650368621,-.0220590603339104,-.0213572810231997,-.0153723594557108,-.00558463191230799,.00558463191230799,.0153723594557108,.0213572810231997,.0220590603339104,.0173053650368621,.00828801966032105,-.00596547927875537,-.0124559063286174,-.0158774801127414,-.0153723594557108,-.0110645842501621,-.00401965817140192,.00401965817140192,.0110645842501621,.0153723594557108,.0158774801127414,.0124559063286174,.00596547927875537,-.00216720185657465,-.0045251122431745,-.00576813744045716,-.00558463191230799,-.00401965817140192,-.00146030356402064,.00146030356402064,.00401965817140193,.00558463191230799,.00576813744045716,.0045251122431745,.00216720185657466,.00216720185657465,.0045251122431745,.00576813744045716,.00558463191230799,.00401965817140192,.00146030356402064,-.00146030356402064,-.00401965817140193,-.00558463191230799,-.00576813744045716,-.0045251122431745,-.00216720185657466,.00596547927875537,.0124559063286174,.0158774801127414,.0153723594557108,.0110645842501621,.00401965817140193,-.00401965817140193,-.0110645842501621,-.0153723594557108,-.0158774801127414,-.0124559063286174,-.00596547927875537,.00828801966032105,.0173053650368621,.0220590603339104,.0213572810231997,.0153723594557108,.00558463191230799,-.00558463191230799,-.0153723594557108,-.0213572810231997,-.0220590603339104,-.0173053650368621,-.00828801966032105,.00856035585883146,.0178740023617151,.0227838994245812,.0220590603339104,.0158774801127414,.00576813744045716,-.00576813744045716,-.0158774801127414,-.0220590603339104,-.0227838994245812,-.0178740023617151,-.00856035585883146,.00671561166886112,.0140221809477405,.0178740023617151,.0173053650368621,.0124559063286174,.0045251122431745,-.0045251122431745,-.0124559063286174,-.0173053650368621,-.0178740023617151,-.0140221809477405,-.00671561166886112,.003216292833121,.00671561166886112,.00856035585883146,.00828801966032105,.00596547927875537,.00216720185657466,-.00216720185657466,-.00596547927875537,-.00828801966032105,-.00856035585883146,-.00671561166886112,-.003216292833121},
			{.0039983685679321,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316192,.00269417998711382,-.00269417998711382,-.00741604887316192,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.0039983685679321,.00696316618552943,.0145390741807499,.0185328835230693,.0179432847809911,.0129150626978197,.00469191938293559,-.00469191938293559,-.0129150626978197,-.0179432847809911,-.0185328835230693,-.0145390741807499,-.00696316618552943,.00558559847779338,.011662716133544,.0148664046264906,.0143934499750413,.0103599932305885,.00376368695863054,-.00376368695863054,-.0103599932305885,-.0143934499750413,-.0148664046264906,-.011662716133544,-.00558559847779338,.000317241258906341,.000662398982523078,.000844356596319848,.000817494527800616,.000588409157550378,.000213763447844004,-.000213763447844004,-.000588409157550378,-.000817494527800616,-.000844356596319848,-.000662398982523078,-.000317241258906341,-.00621942636635253,-.0129861472342896,-.0165533754841949,-.0160267521257476,-.0115355973598394,-.00419077275215303,.00419077275215303,.0115355973598394,.0160267521257476,.0165533754841949,.0129861472342895,.00621942636635253,-.0106449482959891,-.0222266263366319,-.0283321669513853,-.0274308171979002,-.019743916918636,-.00717277714038947,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.0106449482959891,-.0222266263366319,-.0283321669513854,-.0274308171979002,-.019743916918636,-.00717277714038948,.00717277714038947,.019743916918636,.0274308171979002,.0283321669513854,.0222266263366319,.0106449482959891,-.00621942636635254,-.0129861472342896,-.0165533754841949,-.0160267521257477,-.0115355973598394,-.00419077275215304,.00419077275215304,.0115355973598394,.0160267521257477,.0165533754841949,.0129861472342896,.00621942636635254,.000317241258906326,.000662398982523046,.000844356596319808,.000817494527800577,.000588409157550351,.000213763447843993,-.000213763447843993,-.000588409157550351,-.000817494527800577,-.000844356596319808,-.000662398982523046,-.000317241258906326,.00558559847779337,.0116627161335439,.0148664046264906,.0143934499750412,.0103599932305885,.00376368695863053,-.00376368695863053,-.0103599932305885,-.0143934499750412,-.0148664046264906,-.0116627161335439,-.00558559847779337,.00696316618552942,.0145390741807499,.0185328835230693,.017943284780991,.0129150626978197,.00469191938293558,-.00469191938293558,-.0129150626978197,-.017943284780991,-.0185328835230693,-.0145390741807499,-.00696316618552942,.00399836856793211,.00834858391459226,.0106418972314321,.0103033395961257,.00741604887316193,.00269417998711382,-.00269417998711382,-.00741604887316193,-.0103033395961257,-.0106418972314321,-.00834858391459226,-.00399836856793211},
			{-.00446418449222142,-.00932120643965728,-.0118816941913244,-.0115036940846361,-.008280029669763,-.0030080559891824,.0030080559891824,.008280029669763,.0115036940846361,.0118816941913244,.00932120643965728,.00446418449222142,-.00557608613676323,-.0116428543884434,-.0148410869615597,-.0143689377575399,-.0103423500381766,-.00375727735471369,.00375727735471369,.0103423500381766,.0143689377575399,.0148410869615597,.0116428543884434,.00557608613676323,.000237001216546111,.000494857967838362,.000630792921502108,.00061072509381806,.00043958243844784,.000159696116976629,-.000159696116976629,-.00043958243844784,-.00061072509381806,-.000630792921502108,-.000494857967838362,-.000237001216546111,.00807267517325051,.0168557262678721,.0214859080939843,.0208023628502637,.014972945241251,.00543952852887958,-.00543952852887958,-.014972945241251,-.0208023628502637,-.0214859080939843,-.0168557262678721,-.00807267517325051,.0105863538057353,.0221042811824758,.0281762142089068,.0272798258819948,.0196352376920871,.00713329501153745,-.00713329501153745,-.0196352376920871,-.0272798258819948,-.0281762142089068,-.0221042811824758,-.0105863538057353,.00483588522718639,.0100973166766822,.0128709979423768,.0124615244685555,.00896944855895662,.0032585153113594,-.00325851531135941,-.00896944855895662,-.0124615244685555,-.0128709979423768,-.0100973166766822,-.00483588522718639,-.0048358852271864,-.0100973166766822,-.0128709979423768,-.0124615244685555,-.00896944855895663,-.00325851531135941,.00325851531135941,.00896944855895663,.0124615244685555,.0128709979423768,.0100973166766822,.0048358852271864,-.0105863538057353,-.0221042811824759,-.0281762142089069,-.0272798258819949,-.0196352376920872,-.00713329501153749,.00713329501153749,.0196352376920872,.0272798258819949,.0281762142089069,.0221042811824759,.0105863538057353,-.00807267517325049,-.0168557262678721,-.0214859080939843,-.0208023628502637,-.014972945241251,-.00543952852887957,.00543952852887957,.014972945241251,.0208023628502637,.0214859080939843,.0168557262678721,.00807267517325049,-.000237001216546102,-.000494857967838343,-.000630792921502084,-.000610725093818036,-.000439582438447823,-.000159696116976623,.000159696116976623,.000439582438447823,.000610725093818036,.000630792921502084,.000494857967838343,.000237001216546102,.00557608613676323,.0116428543884434,.0148410869615597,.0143689377575399,.0103423500381766,.00375727735471369,-.00375727735471369,-.0103423500381766,-.0143689377575399,-.0148410869615597,-.0116428543884434,-.00557608613676323,.0044641844922214,.00932120643965724,.0118816941913244,.011503694084636,.00828002966976296,.00300805598918238,-.00300805598918238,-.00828002966976296,-.011503694084636,-.0118816941913244,-.00932120643965724,-.0044641844922214},
			{.00467171851954842,.00975453698756343,.0124340584252153,.012038485593966,.00866495728795332,.00314789653003547,-.00314789653003547,-.00866495728795332,-.012038485593966,-.0124340584252153,-.00975453698756343,-.00467171851954842,.00299723801522752,.00625822573807254,.00797732835137032,.0077235403881935,.00555918325881647,.0020195983787005,-.0020195983787005,-.00555918325881647,-.0077235403881935,-.00797732835137032,-.00625822573807254,-.00299723801522752,-.00598247350841671,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016646,.00403110922016646,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841671,-.00972729866058341,-.0203105761138428,-.0258897875287452,-.0250661387888921,-.0180418890967918,-.00655444662861003,.00655444662861003,.0180418890967918,.0250661387888921,.0258897875287452,.0203105761138428,.00972729866058341,-.00146872850921155,-.00306670158055196,-.00390910882534958,-.00378474579014224,-.00272415166852724,-.000989658379109305,.000989658379109305,.00272415166852724,.00378474579014224,.00390910882534958,.00306670158055196,.00146872850921155,.00950954380591769,.0198559045031346,.0253102199512128,.02450500937368,.0176380042079596,.00640771909172347,-.00640771909172347,-.0176380042079596,-.02450500937368,-.0253102199512128,-.0198559045031346,-.00950954380591768,.00950954380591779,.0198559045031348,.0253102199512131,.0245050093736803,.0176380042079598,.00640771909172355,-.00640771909172354,-.0176380042079598,-.0245050093736803,-.0253102199512131,-.0198559045031348,-.00950954380591779,-.0014687285092115,-.00306670158055185,-.00390910882534945,-.00378474579014211,-.00272415166852715,-.000989658379109271,.000989658379109271,.00272415166852715,.00378474579014211,.00390910882534945,.00306670158055185,.0014687285092115,-.0097272986605835,-.020310576113843,-.0258897875287454,-.0250661387888923,-.0180418890967919,-.00655444662861008,.00655444662861008,.0180418890967919,.0250661387888923,.0258897875287454,.020310576113843,.0097272986605835,-.00598247350841672,-.0124913902391127,-.0159227112720281,-.0154161516465507,-.0110961046154283,-.00403110922016647,.00403110922016647,.0110961046154283,.0154161516465507,.0159227112720281,.0124913902391127,.00598247350841672,.00299723801522744,.00625822573807237,.0079773283513701,.00772354038819329,.00555918325881632,.00201959837870045,-.00201959837870045,-.00555918325881632,-.00772354038819329,-.0079773283513701,-.00625822573807237,-.00299723801522744,.00467171851954842,.00975453698756342,.0124340584252153,.012038485593966,.00866495728795331,.00314789653003546,-.00314789653003546,-.00866495728795331,-.012038485593966,-.0124340584252153,-.00975453698756342,-.00467171851954842},
			{-.00231698622849408,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849407,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.012236695836953,-.0255501913584168,-.0325686982919874,-.0315325689967173,-.0226962404368325,-.00824532818128076,.00824532818128076,.0226962404368325,.0315325689967173,.0325686982919874,.0255501913584168,.012236695836953,-.0114678347596974,-.0239448112857507,-.03052232852132,-.029551301725675,-.0212701809756486,-.0077272543489116,.0077272543489116,.0212701809756486,.029551301725675,.03052232852132,.0239448112857507,.0114678347596974,-.00997843648419155,-.0208349512830896,-.0265582058759676,-.0257132923061589,-.0185076916715506,-.00672366827160686,.00672366827160686,.0185076916715506,.0257132923061589,.0265582058759676,.0208349512830896,.00997843648419155,-.00786214398058468,-.0164161376459761,-.0209255667251747,-.0202598479878244,-.0145824585745998,-.00529766843861862,.00529766843861862,.0145824585745998,.0202598479878244,.0209255667251747,.0164161376459761,.00786214398058468,-.00525225518720616,-.0109666961477544,-.0139791915092038,-.0135344623487989,-.0097417185286608,-.00353907364270423,.00353907364270423,.0097417185286608,.0135344623487989,.0139791915092038,.0109666961477544,.00525225518720616,-.00231698622849407,-.00483786165004334,-.00616679751037299,-.00597060914873778,-.0042974735362733,-.00156123124248536,.00156123124248536,.0042974735362733,.00597060914873778,.00616679751037299,.00483786165004334,.00231698622849408},
			{.00393913815260715,.00822491095042321,.0104842519363188,.0101507095738694,.00730619015269516,.00265426935934509,-.00265426935934509,-.00730619015269516,-.0101507095738694,-.0104842519363188,-.00822491095042321,-.00393913815260715,.00822491095042321,.0171735942029902,.0218910927256895,.0211946570783323,.0152553073958586,.00554210801278983,-.00554210801278983,-.0152553073958586,-.0211946570783323,-.0218910927256895,-.0171735942029902,-.00822491095042321,.0104842519363188,.0218910927256895,.0279044639730276,.0270167210140658,.0194458623404232,.00706449674830704,-.00706449674830704,-.0194458623404232,-.0270167210140658,-.0279044639730276,-.0218910927256895,-.0104842519363188,.0101507095738694,.0211946570783323,.0270167210140658,.026157220402348,.0188272184062363,.00683974929381445,-.00683974929381445,-.0188272184062363,-.026157220402348,-.0270167210140658,-.0211946570783323,-.0101507095738694,.00730619015269516,.0152553073958586,.0194458623404232,.0188272184062363,.0135512928156657,.00492305573060758,-.00492305573060758,-.0135512928156657,-.0188272184062363,-.0194458623404232,-.0152553073958586,-.00730619015269516,.00265426935934509,.00554210801278983,.00706449674830704,.00683974929381445,.00492305573060758,.00178849930086745,-.00178849930086745,-.00492305573060758,-.00683974929381445,-.00706449674830704,-.00554210801278983,-.00265426935934509,-.00265426935934509,-.00554210801278983,-.00706449674830704,-.00683974929381445,-.00492305573060758,-.00178849930086745,.00178849930086745,.00492305573060758,.00683974929381445,.00706449674830704,.00554210801278983,.00265426935934509,-.00730619015269516,-.0152553073958586,-.0194458623404232,-.0188272184062363,-.0135512928156657,-.00492305573060758,.00492305573060758,.0135512928156657,.0188272184062363,.0194458623404232,.0152553073958586,.00730619015269516,-.0101507095738694,-.0211946570783323,-.0270167210140658,-.026157220402348,-.0188272184062363,-.00683974929381445,.00683974929381445,.0188272184062363,.026157220402348,.0270167210140658,.0211946570783323,.0101507095738694,-.0104842519363188,-.0218910927256895,-.0279044639730276,-.0270167210140658,-.0194458623404232,-.00706449674830704,.00706449674830704,.0194458623404232,.0270167210140658,.0279044639730276,.0218910927256895,.0104842519363188,-.00822491095042321,-.0171735942029902,-.0218910927256895,-.0211946570783323,-.0152553073958586,-.00554210801278983,.00554210801278983,.0152553073958586,.0211946570783323,.0218910927256895,.0171735942029902,.00822491095042321,-.00393913815260715,-.00822491095042321,-.0104842519363188,-.0101507095738694,-.00730619015269516,-.00265426935934509,.00265426935934509,.00730619015269516,.0101507095738694,.0104842519363188,.00822491095042321,.00393913815260715},
			{-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969,-.00852810207437714,-.0178066565376609,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123493,.00574640420123493,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376609,.00852810207437714,-.00684093308933222,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010758,.00460955630010758,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933222,-.000388539604839473,-.000811269756660397,-.00103412141096871,-.0010012222303148,-.000720651097989894,-.000261805686438008,.000261805686438008,.000720651097989894,.0010012222303148,.00103412141096871,.000811269756660397,.000388539604839473,.00761721054519024,.0159047172243371,.0202736617284932,.0196286824710797,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710797,-.0202736617284932,-.0159047172243371,-.00761721054519024,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632954,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.0130373458317453,.0272219466141356,.0346996761691302,.0335957526812189,.0241812609872885,.00878482201632955,-.00878482201632955,-.0241812609872885,-.0335957526812189,-.0346996761691302,-.0272219466141356,-.0130373458317453,.00761721054519025,.0159047172243371,.0202736617284932,.0196286824710798,.014128163704906,.00513262743536863,-.00513262743536863,-.014128163704906,-.0196286824710798,-.0202736617284932,-.0159047172243371,-.00761721054519026,-.000388539604839454,-.000811269756660358,-.00103412141096866,-.00100122223031475,-.00072065109798986,-.000261805686437995,.000261805686437995,.00072065109798986,.00100122223031475,.00103412141096866,.000811269756660358,.000388539604839454,-.00684093308933221,-.0142838517710583,-.0182075528223322,-.0176283040385686,-.0126883485768188,-.00460955630010757,.00460955630010757,.0126883485768188,.0176283040385686,.0182075528223322,.0142838517710583,.00684093308933221,-.00852810207437712,-.0178066565376608,-.0226980540469838,-.0219759460114443,-.0158176568028604,-.00574640420123492,.00574640420123492,.0158176568028604,.0219759460114443,.0226980540469838,.0178066565376608,.00852810207437712,-.00489698139750969,-.0102248853327823,-.0130336090560768,-.0126189623285648,-.00908276782339724,-.00329968312182452,.00329968312182452,.00908276782339724,.0126189623285648,.0130336090560768,.0102248853327823,.00489698139750969},
			{.00546748706207169,.0114160997827322,.0145520440250068,.0140890903329313,.0101409238735274,.00368410114579698,-.00368410114579698,-.0101409238735274,-.0140890903329313,-.0145520440250068,-.0114160997827322,-.00546748706207169,.0068292828987853,.0142595262013292,.0181765451429699,.0175982828267856,.0126667401680366,.00460170617081502,-.00460170617081502,-.0126667401680366,-.0175982828267856,-.0181765451429699,-.0142595262013292,-.0068292828987853,-.000290266024493157,-.000606074758208074,-.000772560395559055,-.000747982426521798,-.00053837633707014,-.000195587000258209,.000195587000258209,.00053837633707014,.000747982426521798,.000772560395559055,.000606074758208074,.000290266024493157,-.00988696751735083,-.0206439643012052,-.0263147557466346,-.0254775872149812,-.0183380378947805,-.00666203466887181,.00666203466887181,.0183380378947805,.0254775872149812,.0263147557466346,.0206439643012052,.00988696751735083,-.0129655825309695,-.0270721050154096,-.0345086738493417,-.0334108268431253,-.0240481566631596,-.00873646648194734,.00873646648194734,.0240481566631596,.0334108268431253,.0345086738493417,.0270721050154096,.0129655825309695,-.00592272563093564,-.0123666368152112,-.0157636887204181,-.0152621881833591,-.0109852861223506,-.00399084991614105,.00399084991614105,.0109852861223506,.0152621881833591,.0157636887204181,.0123666368152112,.00592272563093564,.00592272563093565,.0123666368152113,.0157636887204182,.0152621881833591,.0109852861223506,.00399084991614106,-.00399084991614106,-.0109852861223506,-.0152621881833591,-.0157636887204182,-.0123666368152113,-.00592272563093565,.0129655825309696,.0270721050154097,.0345086738493419,.0334108268431254,.0240481566631597,.00873646648194738,-.00873646648194738,-.0240481566631597,-.0334108268431254,-.0345086738493419,-.0270721050154097,-.0129655825309696,.00988696751735081,.0206439643012052,.0263147557466346,.0254775872149811,.0183380378947805,.00666203466887179,-.00666203466887179,-.0183380378947805,-.0254775872149811,-.0263147557466346,-.0206439643012052,-.00988696751735081,.000290266024493146,.000606074758208051,.000772560395559025,.00074798242652177,.000538376337070119,.000195587000258202,-.000195587000258202,-.000538376337070119,-.00074798242652177,-.000772560395559025,-.000606074758208051,-.000290266024493146,-.0068292828987853,-.0142595262013292,-.0181765451429698,-.0175982828267856,-.0126667401680366,-.00460170617081502,.00460170617081502,.0126667401680366,.0175982828267856,.0181765451429698,.0142595262013292,.0068292828987853,-.00546748706207167,-.0114160997827322,-.0145520440250067,-.0140890903329312,-.0101409238735273,-.00368410114579696,.00368410114579696,.0101409238735273,.0140890903329312,.0145520440250067,.0114160997827322,.00546748706207167},
			{.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0141297192758912,.0295028197276516,.0376070934653196,.0364106744098191,.0262073610608292,.00952088489339305,-.00952088489339305,-.0262073610608292,-.0364106744098191,-.0376070934653196,-.0295028197276516,-.0141297192758912,.0132419149757324,.0276490864927876,.0352441491885744,.0341229040244673,.0245606894326591,.00892266476002541,-.00892266476002541,-.0245606894326591,-.0341229040244673,-.0352441491885744,-.0276490864927876,-.0132419149757324,.0115221059845388,.0240581294721656,.0306667746341718,.029691152479886,.0213708415448263,.0077638233758434,-.0077638233758434,-.0213708415448263,-.029691152479886,-.0306667746341718,-.0240581294721656,-.0115221059845388,.00907842189039497,.0189557229845952,.0241627631719636,.0233940573872652,.0168383727729163,.006117220600383,-.006117220600383,-.0168383727729163,-.0233940573872652,-.0241627631719636,-.0189557229845952,-.00907842189039497,.00606478189450901,.0126632499505153,.0161417799674716,.0156282509663382,.0112487676337471,.00408657024203424,-.00408657024203424,-.0112487676337471,-.0156282509663382,-.0161417799674716,-.0126632499505153,-.00606478189450901,.00267542524640207,.00558628145387762,.00712080440781243,.00689426560092869,.00496229500775375,.00180275455686753,-.00180275455686753,-.00496229500775375,-.00689426560092869,-.00712080440781243,-.00558628145387762,-.00267542524640207},
			{-.00454852494396312,-.00949730909922885,-.0121061713511056,-.0117210298057487,-.00843646169984582,-.00306488625715946,.00306488625715946,.00843646169984582,.0117210298057487,.0121061713511056,.00949730909922885,.00454852494396312,-.00949730909922885,-.0198303584651126,-.0252776565471576,-.024473481930491,-.0176153116569324,-.00639947510386962,.00639947510386962,.0176153116569324,.024473481930491,.0252776565471576,.0198303584651126,.00949730909922885,-.0121061713511056,-.0252776565471576,-.0322212995614867,-.0311962222891923,-.0224541477058241,-.00815737819577448,.00815737819577448,.0224541477058241,.0311962222891923,.0322212995614867,.0252776565471576,.0121061713511056,-.0117210298057487,-.024473481930491,-.0311962222891923,-.0302037564704542,-.0217397992222051,-.00789786218916395,.00789786218916395,.0217397992222051,.0302037564704542,.0311962222891923,.024473481930491,.0117210298057487,-.00843646169984582,-.0176153116569324,-.0224541477058241,-.0217397992222051,-.0156476851044708,-.00568465510060072,.00568465510060072,.0156476851044708,.0217397992222051,.0224541477058241,.0176153116569324,.00843646169984582,-.00306488625715946,-.00639947510386962,-.00815737819577448,-.00789786218916395,-.00568465510060072,-.00206518110487492,.00206518110487492,.00568465510060072,.00789786218916395,.00815737819577448,.00639947510386962,.00306488625715946,.00306488625715946,.00639947510386962,.00815737819577448,.00789786218916395,.00568465510060072,.00206518110487492,-.00206518110487492,-.00568465510060072,-.00789786218916395,-.00815737819577448,-.00639947510386962,-.00306488625715946,.00843646169984582,.0176153116569324,.0224541477058241,.0217397992222051,.0156476851044708,.00568465510060072,-.00568465510060072,-.0156476851044708,-.0217397992222051,-.0224541477058241,-.0176153116569324,-.00843646169984582,.0117210298057487,.024473481930491,.0311962222891923,.0302037564704542,.0217397992222051,.00789786218916395,-.00789786218916395,-.0217397992222051,-.0302037564704542,-.0311962222891923,-.024473481930491,-.0117210298057487,.0121061713511056,.0252776565471576,.0322212995614867,.0311962222891923,.0224541477058241,.00815737819577448,-.00815737819577448,-.0224541477058241,-.0311962222891923,-.0322212995614867,-.0252776565471576,-.0121061713511056,.00949730909922885,.0198303584651126,.0252776565471576,.024473481930491,.0176153116569324,.00639947510386962,-.00639947510386962,-.0176153116569324,-.024473481930491,-.0252776565471576,-.0198303584651126,-.00949730909922885,.00454852494396312,.00949730909922885,.0121061713511056,.0117210298057487,.00843646169984582,.00306488625715946,-.00306488625715946,-.00843646169984582,-.0117210298057487,-.0121061713511056,-.00949730909922885,-.00454852494396312},
			{.00565454705638537,.0118066805991472,.015049915394736,.0145711225952196,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952196,-.015049915394736,-.0118066805991472,-.00565454705638537,.00984740405706795,.0205613558916723,.0262094552293619,.02537563669192,.0182646568269614,.00663537602520149,-.00663537602520149,-.0182646568269614,-.02537563669192,-.0262094552293619,-.0205613558916723,-.00984740405706795,.00789922912161445,.0164935713308931,.0210242710474368,.0203554121649402,.0146512429334382,.00532265714165691,-.00532265714165691,-.0146512429334382,-.0203554121649402,-.0210242710474368,-.0164935713308931,-.00789922912161445,.000448646890909458,.000936773624827609,.0011941005500474,.00115611184843243,.000832136210868989,.000302307167093963,-.000302307167093963,-.000832136210868989,-.00115611184843243,-.0011941005500474,-.000936773624827609,-.000448646890909458,-.00879559711786466,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317994,.00592664766317994,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786466,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.0150542302516124,-.0314331964124508,-.0400677347558359,-.0387930337099593,-.0279221150819346,-.0101438387122661,.0101438387122661,.0279221150819346,.0387930337099593,.0400677347558359,.0314331964124508,.0150542302516124,-.00879559711786467,-.0183651855425165,-.0234100081138356,-.0226652502180242,-.0163137982370799,-.00592664766317995,.00592664766317995,.0163137982370799,.0226652502180242,.0234100081138356,.0183651855425165,.00879559711786467,.000448646890909437,.000936773624827564,.00119410055004734,.00115611184843238,.000832136210868949,.000302307167093948,-.000302307167093948,-.000832136210868949,-.00115611184843238,-.00119410055004734,-.000936773624827564,-.000448646890909437,.00789922912161443,.0164935713308931,.0210242710474368,.0203554121649401,.0146512429334382,.0053226571416569,-.0053226571416569,-.0146512429334382,-.0203554121649401,-.0210242710474368,-.0164935713308931,-.00789922912161443,.00984740405706793,.0205613558916723,.0262094552293618,.0253756366919199,.0182646568269614,.00663537602520148,-.00663537602520148,-.0182646568269614,-.0253756366919199,-.0262094552293618,-.0205613558916723,-.00984740405706793,.00565454705638537,.0118066805991472,.015049915394736,.0145711225952197,.0104878768961101,.00381014587741865,-.00381014587741865,-.0104878768961101,-.0145711225952197,-.015049915394736,-.0118066805991472,-.00565454705638537},
			{-.00299121635954409,-.0062456525355466,-.00796130135439456,-.00770802326855251,-.00554801448032919,-.00201554086775423,.00201554086775423,.00554801448032919,.00770802326855251,.00796130135439456,.0062456525355466,.00299121635954408,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.0157975064003936,-.0329851552162446,-.042046008708203,-.0407083715394964,-.0293007204185775,-.0106446729127462,.0106446729127462,.0293007204185775,.0407083715394964,.042046008708203,.0329851552162446,.0157975064003935,-.014804911017555,-.0309126184537943,-.0394041566935386,-.0381505664904682,-.0274596855701536,-.0099758424709522,.0099758424709522,.0274596855701536,.0381505664904682,.0394041566935386,.0309126184537943,.014804911017555,-.0128821061114311,-.0268978064529921,-.0342864963629789,-.0331957176344168,-.0238933272129638,-.00868021841599365,.00868021841599365,.0238933272129638,.0331957176344168,.0342864963629789,.0268978064529921,.0128821061114311,-.0101499842366785,-.0211931425760292,-.0270147904857234,-.0261553512911662,-.0188258730735172,-.00683926054723933,.00683926054723933,.0188258730735172,.0261553512911662,.0270147904857234,.0211931425760292,.0101499842366785,-.0067806322917519,-.0141579438513248,-.0180470586407878,-.017472915763368,-.0125765045448471,-.00456892442756064,.00456892442756064,.0125765045448471,.017472915763368,.0180470586407878,.0141579438513248,.0067806322917519,-.00299121635954408,-.0062456525355466,-.00796130135439456,-.0077080232685525,-.0055480144803292,-.00201554086775423,.00201554086775423,.0055480144803292,.0077080232685525,.00796130135439456,.0062456525355466,.00299121635954409},
			{.00508540548709832,.0106183143768374,.0135351110470164,.013104509708737,.00943225092720051,.00342664700787828,-.00342664700787828,-.00943225092720051,-.013104509708737,-.0135351110470164,-.0106183143768374,-.00508540548709832,.0106183143768374,.0221710147777587,.0282612791816195,.0273621846271071,.0196945171590198,.0071548306777916,-.0071548306777916,-.0196945171590198,-.0273621846271071,-.0282612791816195,-.0221710147777587,-.0106183143768374,.0135351110470164,.0282612791816195,.03602450807902,.0348784368472584,.0251045003288081,.00912022608388363,-.00912022608388363,-.0251045003288081,-.0348784368472584,-.03602450807902,-.0282612791816195,-.0135351110470164,.013104509708737,.0273621846271071,.0348784368472584,.0337688263290031,.0243058344441419,.00883007836780732,-.00883007836780732,-.0243058344441419,-.0337688263290031,-.0348784368472584,-.0273621846271071,-.013104509708737,.00943225092720051,.0196945171590198,.0251045003288081,.0243058344441419,.0174946437957377,.00635563761813038,-.00635563761813038,-.0174946437957377,-.0243058344441419,-.0251045003288081,-.0196945171590198,-.00943225092720051,.00342664700787828,.0071548306777916,.00912022608388363,.00883007836780732,.00635563761813038,.00230894266866042,-.00230894266866042,-.00635563761813038,-.00883007836780732,-.00912022608388363,-.0071548306777916,-.00342664700787828,-.00342664700787828,-.0071548306777916,-.00912022608388363,-.00883007836780732,-.00635563761813038,-.00230894266866042,.00230894266866042,.00635563761813038,.00883007836780732,.00912022608388363,.0071548306777916,.00342664700787828,-.00943225092720051,-.0196945171590198,-.0251045003288081,-.0243058344441419,-.0174946437957377,-.00635563761813038,.00635563761813038,.0174946437957377,.0243058344441419,.0251045003288081,.0196945171590198,.00943225092720051,-.013104509708737,-.0273621846271071,-.0348784368472584,-.0337688263290031,-.0243058344441419,-.00883007836780732,.00883007836780732,.0243058344441419,.0337688263290031,.0348784368472584,.0273621846271071,.013104509708737,-.0135351110470164,-.0282612791816195,-.03602450807902,-.0348784368472584,-.0251045003288081,-.00912022608388363,.00912022608388363,.0251045003288081,.0348784368472584,.03602450807902,.0282612791816195,.0135351110470164,-.0106183143768374,-.0221710147777587,-.0282612791816195,-.0273621846271071,-.0196945171590198,-.0071548306777916,.0071548306777916,.0196945171590198,.0273621846271071,.0282612791816195,.0221710147777587,.0106183143768374,-.00508540548709832,-.0106183143768374,-.0135351110470164,-.013104509708737,-.00943225092720051,-.00342664700787828,.00342664700787828,.00943225092720051,.013104509708737,.0135351110470164,.0106183143768374,.00508540548709832},
			{.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600146,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600146,-.00872116867787892,-.00684176956018095,-.00327671334903246,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0173053012158188,.0361334271498335,.0460590948455173,.0445937867433876,.032097331049117,.0116606549432606,-.0116606549432606,-.032097331049117,-.0445937867433876,-.0460590948455173,-.0361334271498335,-.0173053012158188,.0162179674524962,.0338630768777066,.0431650909614457,.0417918516972545,.0300805784178282,.0109279879030203,-.0109279879030203,-.0300805784178282,-.0417918516972545,-.0431650909614457,-.0338630768777066,-.0162179674524962,.0141116402109451,.0294650706837015,.037558974951999,.0363640867222272,.0261738285770315,.00950870286111265,-.00950870286111265,-.0261738285770315,-.0363640867222272,-.037558974951999,-.0294650706837015,-.0141116402109451,.0111187506496062,.0232159245068477,.0295932202708924,.0286517518035574,.0206227106943835,.00749203455682688,-.00749203455682688,-.0206227106943835,-.0286517518035574,-.0295932202708924,-.0232159245068477,-.0111187506496062,.00742781052075099,.0155092504306706,.0197695622285424,.019140620218149,.013776870467688,.00500500594506989,-.00500500594506989,-.013776870467688,-.019140620218149,-.0197695622285424,-.0155092504306706,-.00742781052075099,.00327671334903246,.00684176956018095,.00872116867787892,.00844371643600145,.00607754536054053,.00220791439770589,-.00220791439770589,-.00607754536054053,-.00844371643600145,-.00872116867787892,-.00684176956018095,-.00327671334903246},
			{-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.00288038601951719,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847933,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847885,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0152121782945392,-.0264920363666131,-.0212509473507201,-.00120697492261531,.0236624065959092,.0404997309936553,.0404997309936554,.0236624065959092,-.00120697492261525,-.0212509473507201,-.026492036366613,-.0152121782945392,-.0142563604866286,-.0248274779032066,-.0199156991358028,-.00113113778066232,.0221756405874862,.0379550353327309,.0379550353327309,.0221756405874863,-.00113113778066226,-.0199156991358028,-.0248274779032066,-.0142563604866286,-.0124047992138418,-.0216029805548362,-.0173291247239877,-.000984229955847932,.019295553671226,.0330255813114713,.0330255813114713,.0192955536712261,-.000984229955847886,-.0173291247239876,-.0216029805548362,-.0124047992138418,-.00977390772832783,-.0170212782133727,-.01365384986449,-.000775487987034864,.0152032256144154,.0260212985997005,.0260212985997005,.0152032256144154,-.000775487987034827,-.01365384986449,-.0170212782133727,-.00977390772832783,-.00652939677677674,-.011370956447737,-.00912136739713697,-.000518059808186956,.0101564179939682,.0173833627170427,.0173833627170427,.0101564179939682,-.000518059808186931,-.00912136739713696,-.011370956447737,-.00652939677677674,-.00288038601951719,-.00501619752947058,-.004023811085128,-.000228537532606818,.004480414561763,.00766851772899668,.00766851772899668,.004480414561763,-.000228537532606807,-.00402381108512799,-.00501619752947057,-.0028803860195172},
			{.00489698139750969,.00852810207437714,.00684093308933222,.000388539604839473,-.00761721054519024,-.0130373458317453,-.0130373458317453,-.00761721054519025,.000388539604839454,.00684093308933221,.00852810207437712,.00489698139750969,.0102248853327823,.0178066565376609,.0142838517710583,.000811269756660397,-.0159047172243371,-.0272219466141356,-.0272219466141356,-.0159047172243371,.000811269756660358,.0142838517710583,.0178066565376608,.0102248853327823,.0130336090560768,.0226980540469838,.0182075528223322,.00103412141096871,-.0202736617284932,-.0346996761691302,-.0346996761691302,-.0202736617284932,.00103412141096866,.0182075528223322,.0226980540469838,.0130336090560768,.0126189623285648,.0219759460114443,.0176283040385686,.0010012222303148,-.0196286824710797,-.0335957526812189,-.0335957526812189,-.0196286824710798,.00100122223031475,.0176283040385686,.0219759460114443,.0126189623285648,.00908276782339724,.0158176568028604,.0126883485768188,.000720651097989894,-.014128163704906,-.0241812609872885,-.0241812609872885,-.014128163704906,.00072065109798986,.0126883485768188,.0158176568028604,.00908276782339724,.00329968312182452,.00574640420123493,.00460955630010758,.000261805686438008,-.00513262743536863,-.00878482201632954,-.00878482201632955,-.00513262743536863,.000261805686437995,.00460955630010757,.00574640420123492,.00329968312182452,-.00329968312182452,-.00574640420123493,-.00460955630010758,-.000261805686438008,.00513262743536863,.00878482201632955,.00878482201632955,.00513262743536863,-.000261805686437995,-.00460955630010757,-.00574640420123492,-.00329968312182452,-.00908276782339724,-.0158176568028604,-.0126883485768188,-.000720651097989894,.014128163704906,.0241812609872885,.0241812609872885,.014128163704906,-.00072065109798986,-.0126883485768188,-.0158176568028604,-.00908276782339724,-.0126189623285648,-.0219759460114443,-.0176283040385686,-.0010012222303148,.0196286824710797,.0335957526812189,.0335957526812189,.0196286824710798,-.00100122223031475,-.0176283040385686,-.0219759460114443,-.0126189623285648,-.0130336090560768,-.0226980540469838,-.0182075528223322,-.00103412141096871,.0202736617284932,.0346996761691302,.0346996761691302,.0202736617284932,-.00103412141096866,-.0182075528223322,-.0226980540469838,-.0130336090560768,-.0102248853327823,-.0178066565376609,-.0142838517710583,-.000811269756660397,.0159047172243371,.0272219466141356,.0272219466141356,.0159047172243371,-.000811269756660358,-.0142838517710583,-.0178066565376608,-.0102248853327823,-.00489698139750969,-.00852810207437714,-.00684093308933222,-.000388539604839473,.00761721054519024,.0130373458317453,.0130373458317453,.00761721054519026,-.000388539604839454,-.00684093308933221,-.00852810207437712,-.00489698139750969},
			{-.00608773439400276,-.0106018005990698,-.00850437857006723,-.000483017133170316,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170293,-.00850437857006722,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.0184630551643588,-.0148103908586545,-.000841175222238799,.0164910055329813,.0282254167678824,.0282254167678824,.0164910055329813,-.000841175222238758,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00850437857006723,-.0148103908586545,-.0118803564975286,-.000674760147281664,.0132284844204532,.0226414020193,.0226414020193,.0132284844204532,-.000674760147281631,-.0118803564975285,-.0148103908586545,-.00850437857006724,-.000483017133170316,-.000841175222238799,-.000674760147281664,-3.83238715483232e-005,.000751328808837938,.00128594758620117,.00128594758620117,.000751328808837939,-3.83238715483214e-005,-.000674760147281662,-.000841175222238797,-.000483017133170317,.0094694161276367,.0164910055329813,.0132284844204532,.000751328808837938,-.0147295916665949,-.0252106436310925,-.0252106436310925,-.0147295916665949,.000751328808837902,.0132284844204532,.0164910055329813,.0094694161276367,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710571,-.0431496382710572,-.0252106436310926,.0012859475862011,.0226414020192999,.0282254167678823,.0162075148308275,.0162075148308275,.0282254167678824,.0226414020193,.00128594758620117,-.0252106436310925,-.0431496382710572,-.0431496382710572,-.0252106436310926,.00128594758620111,.0226414020192999,.0282254167678823,.0162075148308275,.00946941612763672,.0164910055329813,.0132284844204532,.000751328808837939,-.0147295916665949,-.0252106436310926,-.0252106436310926,-.0147295916665949,.000751328808837903,.0132284844204532,.0164910055329813,.00946941612763672,-.000483017133170293,-.000841175222238758,-.000674760147281631,-3.83238715483214e-005,.000751328808837902,.0012859475862011,.00128594758620111,.000751328808837903,-3.83238715483196e-005,-.00067476014728163,-.000841175222238757,-.000483017133170294,-.00850437857006722,-.0148103908586545,-.0118803564975285,-.000674760147281662,.0132284844204532,.0226414020192999,.0226414020192999,.0132284844204532,-.00067476014728163,-.0118803564975285,-.0148103908586545,-.00850437857006722,-.0106018005990698,-.0184630551643587,-.0148103908586545,-.000841175222238797,.0164910055329813,.0282254167678823,.0282254167678823,.0164910055329813,-.000841175222238757,-.0148103908586545,-.0184630551643587,-.0106018005990698,-.00608773439400276,-.0106018005990698,-.00850437857006724,-.000483017133170317,.0094694161276367,.0162075148308275,.0162075148308275,.00946941612763672,-.000483017133170294,-.00850437857006722,-.0106018005990698,-.00608773439400276},
			{.00679696456586262,.0118369262425784,.00949515140679809,.000539289352389854,-.0105726172847424,-.0180957145755819,-.0180957145755819,-.0105726172847425,.000539289352389828,.00949515140679807,.0118369262425784,.00679696456586262,.00848989551256602,.0147851685874852,.0118601241096038,.000673611022753388,-.0132059561546973,-.0226028434433086,-.0226028434433086,-.0132059561546973,.000673611022753355,.0118601241096038,.0147851685874851,.00848989551256602,-.00036084728884685,-.00062841621455069,-.000504092615039049,-2.86305892620399e-005,.000561294714169291,.000960692009068383,.000960692009068383,.000561294714169292,-2.86305892620386e-005,-.000504092615039048,-.000628416214550689,-.00036084728884685,-.0122910885963404,-.0214049533061549,-.017170274472633,-.000975207851248453,.0191186778278397,.0327228469278241,.0327228469278241,.0191186778278397,-.000975207851248406,-.017170274472633,-.0214049533061549,-.0122910885963405,-.0161183015228526,-.0280700516286174,-.0225167737593592,-.00127886916367651,.0250718731324701,.0429121236361326,.0429121236361326,.0250718731324701,-.00127886916367645,-.0225167737593592,-.0280700516286173,-.0161183015228526,-.00736289922404356,-.0128225024865171,-.0102857447979673,-.000584192122199518,.0114529235583944,.0196024153893991,.0196024153893991,.0114529235583944,-.00058419212219949,-.0102857447979673,-.012822502486517,-.00736289922404357,.00736289922404357,.0128225024865171,.0102857447979674,.000584192122199519,-.0114529235583944,-.0196024153893992,-.0196024153893992,-.0114529235583944,.000584192122199491,.0102857447979673,.012822502486517,.00736289922404358,.0161183015228527,.0280700516286175,.0225167737593594,.00127886916367651,-.0250718731324702,-.0429121236361328,-.0429121236361328,-.0250718731324703,.00127886916367645,.0225167737593593,.0280700516286175,.0161183015228527,.0122910885963404,.0214049533061549,.017170274472633,.000975207851248451,-.0191186778278397,-.032722846927824,-.032722846927824,-.0191186778278397,.000975207851248404,.0171702744726329,.0214049533061548,.0122910885963404,.000360847288846836,.000628416214550666,.00050409261503903,2.86305892620388e-005,-.000561294714169269,-.000960692009068346,-.000960692009068346,-.00056129471416927,2.86305892620375e-005,.000504092615039029,.000628416214550665,.000360847288846836,-.00848989551256602,-.0147851685874852,-.0118601241096038,-.000673611022753387,.0132059561546973,.0226028434433086,.0226028434433086,.0132059561546973,-.000673611022753355,-.0118601241096038,-.0147851685874851,-.00848989551256602,-.00679696456586259,-.0118369262425784,-.00949515140679805,-.000539289352389851,.0105726172847424,.0180957145755818,.0180957145755819,.0105726172847424,-.000539289352389826,-.00949515140679803,-.0118369262425784,-.00679696456586259},
			{.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929397,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0175655104658559,.0305903686539505,.0245384803458705,.00139369459289309,-.0273229936344479,-.0467650611805397,-.0467650611805397,-.027322993634448,.00139369459289302,.0245384803458705,.0305903686539505,.0175655104658559,.0164618271285175,.0286683021001756,.0229966685133031,.00130612540425466,-.0256062241241532,-.0438266997309877,-.0438266997309877,-.0256062241241533,.0013061254042546,.0229966685133031,.0286683021001755,.0164618271285175,.0143238283300821,.0249449732761729,.020009949647553,.0011364908598568,-.0222805862115142,-.0381346565189728,-.0381346565189728,-.0222805862115143,.00113649085985675,.020009949647553,.0249449732761729,.0143238283300821,.0112859365154748,.0196544791160234,.0157661077887853,.000895456396096566,-.0175551728012975,-.0300468075010849,-.0300468075010849,-.0175551728012976,.000895456396096523,.0157661077887853,.0196544791160234,.0112859365154748,.00753949797310551,.0131300495315137,.0105324478437675,.00059820393939992,-.011727621325126,-.0200725782873315,-.0200725782873315,-.011727621325126,.000598203939399891,.0105324478437675,.0131300495315137,.00753949797310551,.00332598328733085,.00579220598764257,.00464629682613013,.000263892411929396,-.0051735371064033,-.00885484154985533,-.00885484154985534,-.0051735371064033,.000263892411929384,.00464629682613012,.00579220598764256,.00332598328733085},
			{-.00565454705638537,-.00984740405706795,-.00789922912161445,-.000448646890909458,.00879559711786466,.0150542302516124,.0150542302516124,.00879559711786467,-.000448646890909437,-.00789922912161443,-.00984740405706793,-.00565454705638537,-.0118066805991472,-.0205613558916723,-.0164935713308931,-.000936773624827609,.0183651855425165,.0314331964124508,.0314331964124508,.0183651855425165,-.000936773624827564,-.0164935713308931,-.0205613558916723,-.0118066805991472,-.015049915394736,-.0262094552293619,-.0210242710474368,-.0011941005500474,.0234100081138356,.0400677347558359,.0400677347558359,.0234100081138356,-.00119410055004734,-.0210242710474368,-.0262094552293618,-.015049915394736,-.0145711225952196,-.02537563669192,-.0203554121649402,-.00115611184843243,.0226652502180242,.0387930337099593,.0387930337099593,.0226652502180242,-.00115611184843238,-.0203554121649401,-.0253756366919199,-.0145711225952197,-.0104878768961101,-.0182646568269614,-.0146512429334382,-.000832136210868989,.0163137982370799,.0279221150819346,.0279221150819346,.0163137982370799,-.000832136210868949,-.0146512429334382,-.0182646568269614,-.0104878768961101,-.00381014587741865,-.00663537602520149,-.00532265714165691,-.000302307167093963,.00592664766317994,.0101438387122661,.0101438387122661,.00592664766317995,-.000302307167093948,-.0053226571416569,-.00663537602520148,-.00381014587741865,.00381014587741865,.00663537602520149,.00532265714165691,.000302307167093963,-.00592664766317994,-.0101438387122661,-.0101438387122661,-.00592664766317995,.000302307167093948,.0053226571416569,.00663537602520148,.00381014587741865,.0104878768961101,.0182646568269614,.0146512429334382,.000832136210868989,-.0163137982370799,-.0279221150819346,-.0279221150819346,-.0163137982370799,.000832136210868949,.0146512429334382,.0182646568269614,.0104878768961101,.0145711225952196,.02537563669192,.0203554121649402,.00115611184843243,-.0226652502180242,-.0387930337099593,-.0387930337099593,-.0226652502180242,.00115611184843238,.0203554121649401,.0253756366919199,.0145711225952197,.015049915394736,.0262094552293619,.0210242710474368,.0011941005500474,-.0234100081138356,-.0400677347558359,-.0400677347558359,-.0234100081138356,.00119410055004734,.0210242710474368,.0262094552293618,.015049915394736,.0118066805991472,.0205613558916723,.0164935713308931,.000936773624827609,-.0183651855425165,-.0314331964124508,-.0314331964124508,-.0183651855425165,.000936773624827564,.0164935713308931,.0205613558916723,.0118066805991472,.00565454705638537,.00984740405706795,.00789922912161445,.000448646890909458,-.00879559711786466,-.0150542302516124,-.0150542302516124,-.00879559711786467,.000448646890909437,.00789922912161443,.00984740405706793,.00565454705638537},
			{.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997765,.0122419048552221,.00702951017978813,.0122419048552221,.0213192997308994,.0171015662920701,.000971305481648202,-.0190421729606395,-.0325919039263692,-.0325919039263693,-.0190421729606395,.000971305481648156,.01710156629207,.0213192997308993,.0122419048552221,.00982001050997767,.0171015662920701,.0137182540390003,.000779145905068479,-.015274938076857,-.0261440390921353,-.0261440390921353,-.015274938076857,.000779145905068442,.0137182540390002,.01710156629207,.00982001050997767,.000557740143521656,.000971305481648202,.000779145905068479,4.42525950940344e-005,-.000867559779759337,-.00148488436959081,-.00148488436959081,-.000867559779759338,4.42525950940323e-005,.000779145905068478,.0009713054816482,.000557740143521656,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759337,.017008267421531,.0291107437634532,.0291107437634533,.0170082674215311,-.000867559779759295,-.0152749380768569,-.0190421729606395,-.0109343398968667,-.0187148260943523,-.0325919039263692,-.0261440390921353,-.00148488436959081,.0291107437634532,.0498249105249045,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0187148260943523,-.0325919039263693,-.0261440390921353,-.00148488436959081,.0291107437634533,.0498249105249046,.0498249105249046,.0291107437634533,-.00148488436959074,-.0261440390921352,-.0325919039263692,-.0187148260943523,-.0109343398968667,-.0190421729606395,-.015274938076857,-.000867559779759338,.0170082674215311,.0291107437634533,.0291107437634533,.0170082674215311,-.000867559779759297,-.015274938076857,-.0190421729606395,-.0109343398968667,.000557740143521629,.000971305481648156,.000779145905068442,4.42525950940323e-005,-.000867559779759295,-.00148488436959074,-.00148488436959074,-.000867559779759297,4.42525950940302e-005,.000779145905068441,.000971305481648154,.000557740143521629,.00982001050997765,.01710156629207,.0137182540390002,.000779145905068478,-.0152749380768569,-.0261440390921352,-.0261440390921352,-.015274938076857,.000779145905068441,.0137182540390002,.01710156629207,.00982001050997766,.0122419048552221,.0213192997308993,.01710156629207,.0009713054816482,-.0190421729606395,-.0325919039263692,-.0325919039263692,-.0190421729606395,.000971305481648154,.01710156629207,.0213192997308993,.0122419048552221,.00702951017978813,.0122419048552221,.00982001050997767,.000557740143521656,-.0109343398968667,-.0187148260943523,-.0187148260943523,-.0109343398968667,.000557740143521629,.00982001050997766,.0122419048552221,.00702951017978813},
			{-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471107,-.00842941499026791,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325645,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.052284927874802,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0196388377264335,-.0342010718763043,-.0274348550531775,-.00155819792446331,.0305480355512773,.0522849278748019,.0522849278748019,.0305480355512773,-.00155819792446323,-.0274348550531775,-.0342010718763042,-.0196388377264335,-.0184048822427321,-.032052136140997,-.0257110570204733,-.00146029259521891,.0286286288883224,.0489997399036621,.0489997399036621,.0286286288883225,-.00146029259521884,-.0257110570204733,-.032052136140997,-.0184048822427321,-.0160145269186782,-.0278893279653464,-.0223718038144274,-.00127063540895594,.0249104526684997,.0426358421285364,.0426358421285364,.0249104526684998,-.00127063540895588,-.0223718038144274,-.0278893279653464,-.0160145269186782,-.0126180606165174,-.021974375678262,-.0176270443744447,-.00100115068606864,.0196272798660954,.0335933520325644,.0335933520325644,.0196272798660954,-.00100115068606859,-.0176270443744446,-.0219743756782619,-.0126180606165174,-.0084294149902679,-.0146798416471108,-.011775634671588,-.00066881233631236,.0131118792459179,.0224418247623542,.0224418247623542,.0131118792459179,-.000668812336312328,-.011775634671588,-.0146798416471108,-.00842941499026791,-.00371856236046697,-.00647588316266141,-.00519471777234038,-.000295040685848126,.00578419032579968,.00990001381564839,.0099000138156484,.00578419032579969,-.000295040685848112,-.00519471777234037,-.0064758831626614,-.00371856236046697},
			{.00632197580064374,.011009732437834,.00883160664375298,.000501602473032872,-.00983377653008651,-.0168311410974181,-.0168311410974182,-.00983377653008653,.000501602473032848,.00883160664375296,.011009732437834,.00632197580064374,.0132002702054535,.0229882947439242,.0184403733456156,.0010473447524243,-.0205329016482428,-.0351433819676135,-.0351433819676135,-.0205329016482428,.00104734475242425,.0184403733456155,.0229882947439242,.0132002702054535,.0168263169407732,.0293030617759155,.0235058496220271,.00133504500106871,-.0261731847507426,-.0447970893136278,-.0447970893136278,-.0261731847507427,.00133504500106864,.023505849622027,.0293030617759154,.0168263169407732,.0162910103172928,.0283708243105143,.0227580426576453,.00129257234147053,-.0253405201097541,-.0433719302187032,-.0433719302187032,-.0253405201097541,.00129257234147047,.0227580426576453,.0283708243105143,.0162910103172928,.0117258028408243,.0204205071273958,.0163805875786201,.000930356567112207,-.01823938091645,-.0312178737024467,-.0312178737024467,-.01823938091645,.000930356567112162,.0163805875786201,.0204205071273958,.0117258028408243,.00425987259346659,.00741857592503808,.00595091159541767,.00033798968788685,-.00662619352742934,-.0113411564578213,-.0113411564578213,-.00662619352742935,.000337989687886833,.00595091159541766,.00741857592503807,.00425987259346659,-.00425987259346659,-.00741857592503808,-.00595091159541767,-.00033798968788685,.00662619352742934,.0113411564578213,.0113411564578213,.00662619352742935,-.000337989687886833,-.00595091159541766,-.00741857592503807,-.00425987259346659,-.0117258028408243,-.0204205071273958,-.0163805875786201,-.000930356567112207,.01823938091645,.0312178737024467,.0312178737024467,.01823938091645,-.000930356567112162,-.0163805875786201,-.0204205071273958,-.0117258028408243,-.0162910103172928,-.0283708243105143,-.0227580426576453,-.00129257234147053,.0253405201097541,.0433719302187032,.0433719302187032,.0253405201097541,-.00129257234147047,-.0227580426576453,-.0283708243105143,-.0162910103172928,-.0168263169407732,-.0293030617759155,-.0235058496220271,-.00133504500106871,.0261731847507426,.0447970893136278,.0447970893136278,.0261731847507427,-.00133504500106864,-.023505849622027,-.0293030617759154,-.0168263169407732,-.0132002702054535,-.0229882947439242,-.0184403733456156,-.0010473447524243,.0205329016482428,.0351433819676135,.0351433819676135,.0205329016482428,-.00104734475242425,-.0184403733456155,-.0229882947439242,-.0132002702054535,-.00632197580064374,-.011009732437834,-.00883160664375298,-.000501602473032872,.00983377653008651,.0168311410974181,.0168311410974182,.00983377653008653,-.000501602473032848,-.00883160664375296,-.011009732437834,-.00632197580064374},
			{.00407348097349134,.0070939745774067,.00569052820876414,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876412,.00709397457740669,.00407348097349134,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886327,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0215132688564257,.0374653971228924,.030053377955339,.00170692030493147,-.0334636963248448,-.0572752688411622,-.0572752688411622,-.0334636963248449,.00170692030493139,.0300533779553389,.0374653971228923,.0215132688564257,.0201615383493805,.0351113559686839,.0281650518207517,.00159967039025467,-.0313610916717506,-.0536765257255297,-.0536765257255297,-.0313610916717506,.00159967039025459,.0281650518207517,.0351113559686839,.0201615383493805,.0175430352859562,.0305512280869836,.024507083207637,.0013919113519926,-.0272880336941418,-.0467052249938773,-.0467052249938773,-.0272880336941418,.00139191135199254,.0245070832076369,.0305512280869835,.0175430352859562,.0138223928661745,.0240717224972153,.0193094596561158,.00109670562867373,-.0215006078547757,-.0367996733886327,-.0367996733886328,-.0215006078547757,.00109670562867367,.0193094596561158,.0240717224972153,.0138223928661745,.0092339614754254,.016080960824834,.0128995614798497,.000732647206826071,-.0143633440715662,-.0245837873130079,-.0245837873130079,-.0143633440715662,.000732647206826036,.0128995614798497,.0160809608248339,.00923396147542541,.00407348097349134,.0070939745774067,.00569052820876413,.000323200878109585,-.00633626303801956,-.0108449217751671,-.0108449217751671,-.00633626303801957,.00032320087810957,.00569052820876413,.00709397457740669,.00407348097349134},
			{.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934987,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.0196119187341802,.0244966910215155,-.00104118649372649,-.0354646295960857,-.046507645660954,-.0212448630312338,.0212448630312338,.0465076456609542,.0354646295960857,.00104118649372645,-.0244966910215155,-.0196119187341801,.018379654635609,.0229575049128665,-.000975766238139751,-.0332363014853734,-.0435854582486035,-.0199099971087667,.0199099971087667,.0435854582486037,.0332363014853733,.000975766238139714,-.0229575049128665,-.0183796546356089,.0159925757761478,.0199758724649255,-.000849037471739172,-.028919698468867,-.0379247465525524,-.0173241632542941,.0173241632542941,.0379247465525526,.028919698468867,.000849037471739139,-.0199758724649255,-.0159925757761477,.0126007650168123,.0157392579193999,-.000668968015009197,-.0227862184218481,-.0298814166223063,-.0136499406559534,.0136499406559534,.0298814166223065,.0227862184218481,.000668968015009171,-.0157392579193999,-.0126007650168123,.0084178607750957,.0105145109596169,-.000446899819640007,-.0152221880187527,-.0199620899646645,-.00911875587528429,.0091187558752843,.0199620899646646,.0152221880187526,.00044689981963999,-.0105145109596169,-.00841786077509566,.00371346531996135,.00463838412728406,-.000197145928884917,-.00671513449934986,-.0088061005970828,-.00402265903520803,.00402265903520804,.00880610059708284,.00671513449934985,.000197145928884909,-.00463838412728406,-.00371346531996133},
			{-.00631331025337702,-.00788577663899888,.00033517033471411,.0114164867138149,.0149713651270644,.00683897447387038,-.00683897447387039,-.0149713651270645,-.0114164867138149,-.000335170334714097,.00788577663899888,.00631331025337699,-.0131821765636861,-.0164654825796768,.000699834849514613,.0238375966899433,.0312601742324971,.0142797621867037,-.0142797621867037,-.0312601742324972,-.0238375966899433,-.000699834849514586,.0164654825796768,.0131821765636861,-.0168032530681212,-.0209884664598731,.00089207590457249,.0303856626242094,.0398471842676737,.0182023398500645,-.0182023398500645,-.0398471842676739,-.0303856626242093,-.000892075904572456,.0209884664598731,.0168032530681211,-.0162686801907029,-.0203207466521331,.000863695710496402,.0294189836701143,.0385794997386939,.0176232569099964,-.0176232569099964,-.0385794997386941,-.0294189836701142,-.000863695710496369,.0203207466521331,.0162686801907028,-.0117097302549807,-.0146262916897378,.000621663446188857,.0211749422273102,.0277684194425538,.0126847157981637,-.0126847157981637,-.0277684194425539,-.0211749422273102,-.000621663446188833,.0146262916897378,.0117097302549806,-.00425403357597065,-.00531359259224779,.000225844414470279,.00769265501789862,.0100880025489925,.00460823654619044,-.00460823654619044,-.0100880025489926,-.0076926550178986,-.00022584441447027,.00531359259224779,.00425403357597063,.00425403357597065,.00531359259224779,-.000225844414470279,-.00769265501789862,-.0100880025489925,-.00460823654619044,.00460823654619044,.0100880025489926,.0076926550178986,.00022584441447027,-.00531359259224779,-.00425403357597063,.0117097302549807,.0146262916897378,-.000621663446188857,-.0211749422273102,-.0277684194425538,-.0126847157981637,.0126847157981637,.0277684194425539,.0211749422273102,.000621663446188833,-.0146262916897378,-.0117097302549806,.0162686801907029,.0203207466521331,-.000863695710496402,-.0294189836701143,-.0385794997386939,-.0176232569099964,.0176232569099964,.0385794997386941,.0294189836701142,.000863695710496369,-.0203207466521331,-.0162686801907028,.0168032530681212,.0209884664598731,-.00089207590457249,-.0303856626242094,-.0398471842676737,-.0182023398500645,.0182023398500645,.0398471842676739,.0303856626242093,.000892075904572456,-.0209884664598731,-.0168032530681211,.0131821765636861,.0164654825796768,-.000699834849514612,-.0238375966899433,-.0312601742324971,-.0142797621867037,.0142797621867037,.0312601742324972,.0238375966899433,.000699834849514586,-.0164654825796768,-.0131821765636861,.00631331025337702,.00788577663899888,-.00033517033471411,-.0114164867138149,-.0149713651270644,-.00683897447387039,.00683897447387039,.0149713651270645,.0114164867138149,.000335170334714097,-.00788577663899888,-.00631331025337699},
			{.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.00980328691616649,-.00784845864116279,.0136681051009101,.0170724421228125,-.000725632541047572,-.0247163110990939,-.0324125037180548,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990939,.000725632541047544,-.0170724421228125,-.01366810510091,.0109640564380927,.0136948916904454,-.000582076013801829,-.0198265251716479,-.0260001307746002,-.0118769550522377,.0118769550522377,.0260001307746003,.0198265251716478,.000582076013801807,-.0136948916904454,-.0109640564380926,.000622717705357654,.000777819010395326,-3.30597568249531e-005,-.00112607303052629,-.00147671091136544,-.000674566957815848,.000674566957815848,.00147671091136544,.00112607303052629,3.30597568249518e-005,-.000777819010395325,-.000622717705357651,-.0122082068670625,-.0152489246770109,.000648127308443884,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.0289505053954479,-.0220763475742017,-.000648127308443859,.0152489246770109,.0122082068670624,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533856,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0208951313564363,-.0260995154849644,.00110931157975103,.0377850889533857,.0495506522505917,.0226349195967991,-.0226349195967991,-.0495506522505919,-.0377850889533856,-.00110931157975099,.0260995154849644,.0208951313564362,-.0122082068670625,-.015248924677011,.000648127308443885,.0220763475742018,.0289505053954478,.0132246969948783,-.0132246969948783,-.028950505395448,-.0220763475742018,-.00064812730844386,.0152489246770109,.0122082068670624,.000622717705357624,.000777819010395288,-3.30597568249515e-005,-.00112607303052624,-.00147671091136537,-.000674566957815815,.000674566957815816,.00147671091136537,.00112607303052624,3.30597568249502e-005,-.000777819010395288,-.000622717705357621,.0109640564380927,.0136948916904453,-.000582076013801828,-.0198265251716478,-.0260001307746001,-.0118769550522376,.0118769550522377,.0260001307746002,.0198265251716478,.000582076013801806,-.0136948916904453,-.0109640564380926,.0136681051009101,.0170724421228125,-.000725632541047571,-.0247163110990939,-.0324125037180547,-.0148061505200542,.0148061505200542,.0324125037180549,.0247163110990938,.000725632541047543,-.0170724421228125,-.01366810510091,.00784845864116283,.0098032869161665,-.000416670558577616,-.0141925266151514,-.018611811440545,-.00850194369545339,.0085019436954534,.0186118114405451,.0141925266151514,.0004166705585776,-.0098032869161665,-.0078484586411628},
			{-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.0219267917271965,-.0273881331711885,.0011640818884958,.0396506612828389,.0519971285804239,.023752478952853,-.0237524789528531,-.0519971285804241,-.0396506612828388,-.00116408188849575,.0273881331711885,.0219267917271964,-.020549078582017,-.0256672707868813,.00109093981920452,.0371593147172276,.0487300237322494,.0222600534812618,-.0222600534812618,-.0487300237322496,-.0371593147172275,-.00109093981920448,.0256672707868813,.0205490785820169,-.017880243283583,-.0223337043684609,.000949252751030661,.0323332058293213,.0424011556561903,.0193690033429938,-.0193690033429938,-.0424011556561905,-.0323332058293212,-.000949252751030624,.0223337043684609,.0178802432835829,-.0140880835716219,-.0175970253098102,.000747928978091183,.0254757666681284,.0334084394123555,.0152610975962314,-.0152610975962314,-.0334084394123556,-.0254757666681284,-.000747928978091154,.0175970253098102,.0140880835716218,-.00941145445816961,-.011755580626746,.000499649187873189,.0170189235863855,.0223182950647207,.0101950790026495,-.0101950790026495,-.0223182950647208,-.0170189235863855,-.00049964918787317,.011755580626746,.00941145445816956,-.00415178044334086,-.00518587110665708,.000220415849214713,.00750774860854079,.00984551977489349,.00449746952605954,-.00449746952605955,-.00984551977489354,-.00750774860854077,-.000220415849214705,.00518587110665707,.00415178044334084},
			{.007058495447099,.00881656631296387,-.000374731826353174,-.0127640201823163,-.0167384950754978,-.00764620591247183,.00764620591247184,.0167384950754978,.0127640201823163,.00037473182635316,-.00881656631296387,-.00705849544709897,.0147381214487064,.0184089691712473,-.000782439148523992,-.0266512433181539,-.0349499372975651,-.0159652594812031,.0159652594812031,.0349499372975653,.0266512433181539,.000782439148523962,-.0184089691712473,-.0147381214487063,.0187866080578478,.0234658188815225,-.000997371182181883,-.0339722035956245,-.0445505063817578,-.0203508346337808,.0203508346337809,.044550506381758,.0339722035956245,.000997371182181844,-.0234658188815225,-.0187866080578477,.0181889374112355,.0227192854412644,-.000965641160587157,-.0328914236683848,-.043133191990884,-.0197034002242685,.0197034002242685,.0431331919908843,.0328914236683847,.00096564116058712,-.0227192854412643,-.0181889374112354,.0130918764284278,.0163526912438262,-.000695040862629039,-.0236743051276634,-.031046036760756,-.0141819434046015,.0141819434046015,.0310460367607562,.0236743051276633,.000695040862629013,-.0163526912438262,-.0130918764284278,.00475615412876842,.00594077712243872,-.000252501731629378,-.00860064977654093,-.0112787297320448,-.00515216508851936,.00515216508851937,.0112787297320449,.00860064977654092,.000252501731629368,-.00594077712243872,-.00475615412876839,-.00475615412876842,-.00594077712243872,.000252501731629378,.00860064977654093,.0112787297320448,.00515216508851936,-.00515216508851937,-.0112787297320449,-.00860064977654092,-.000252501731629368,.00594077712243872,.00475615412876839,-.0130918764284278,-.0163526912438262,.000695040862629039,.0236743051276634,.031046036760756,.0141819434046015,-.0141819434046015,-.0310460367607562,-.0236743051276633,-.000695040862629013,.0163526912438262,.0130918764284278,-.0181889374112355,-.0227192854412644,.000965641160587157,.0328914236683848,.043133191990884,.0197034002242685,-.0197034002242685,-.0431331919908843,-.0328914236683847,-.00096564116058712,.0227192854412643,.0181889374112354,-.0187866080578478,-.0234658188815225,.000997371182181883,.0339722035956245,.0445505063817578,.0203508346337808,-.0203508346337809,-.044550506381758,-.0339722035956245,-.000997371182181844,.0234658188815225,.0187866080578477,-.0147381214487064,-.0184089691712473,.000782439148523992,.0266512433181539,.0349499372975651,.0159652594812031,-.0159652594812031,-.0349499372975653,-.0266512433181539,-.000782439148523962,.0184089691712473,.0147381214487063,-.007058495447099,-.00881656631296387,.000374731826353174,.0127640201823163,.0167384950754978,.00764620591247183,-.00764620591247184,-.0167384950754978,-.0127640201823163,-.00037473182635316,.00881656631296387,.00705849544709897},
			{.00454804760548224,.00568083717114684,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0240195968866161,.030002196693142,-.00127518781828895,-.0434351232114027,-.0569600005008582,-.0260195370395991,.0260195370395991,.0569600005008585,.0434351232114027,.0012751878182889,-.030002196693142,-.024019596886616,.0225103877517677,.0281170864005526,-.00119506469577804,-.0407059897861686,-.0533810664545187,-.0243846668471451,.0243846668471451,.0533810664545189,.0407059897861686,.00119506469577799,-.0281170864005526,-.0225103877517676,.0195868251611838,.0244653473517476,-.00103985428907903,-.0354192523801375,-.0464481388367057,-.0212176800957699,.0212176800957699,.046448138836706,.0354192523801374,.00103985428907899,-.0244653473517476,-.0195868251611837,.0154327223291678,.0192765754153156,-.00081931514546592,-.0279073041491499,-.0365971117562267,-.0167176948123295,.0167176948123296,.0365971117562268,.0279073041491498,.000819315145465889,-.0192765754153156,-.0154327223291677,.0103097318118634,.0128775933723276,-.000547338262102131,-.0186432967063791,-.0244484673052389,-.0111681494911592,.0111681494911592,.024448467305239,.0186432967063791,.00054733826210211,-.0128775933723276,-.0103097318118633,.00454804760548224,.00568083717114683,-.000241453465305272,-.00822432653836584,-.0107852265426879,-.00492673102247789,.0049267310224779,.0107852265426879,.00822432653836582,.000241453465305263,-.00568083717114683,-.00454804760548222},
			{-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023209,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698057,.00556381945057915,-.00278749101749906,-.00434479121890393,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0229461392419872,-.0147215720619734,.0293841911838684,.0477776964900954,.00721397248999972,-.0467081472022112,-.0467081472022117,.00721397248999947,.0477776964900958,.0293841911838684,-.014721572061973,-.0229461392419872,-.0215043780368769,-.0137965802255102,.0275379116661827,.044775708724637,.00676070122021755,-.0437733617950215,-.043773361795022,.00676070122021732,.0447757087246374,.0275379116661828,-.0137965802255098,-.0215043780368769,-.0187114721191434,-.0120047334448111,.0239613936044832,.0389604118742327,.00588264734609042,-.0380882459090025,-.0380882459090029,.00588264734609022,.0389604118742331,.0239613936044832,-.0120047334448108,-.0187114721191434,-.0147430199232581,-.00945869054146635,.0188795034966012,.030697431224158,.0046350167679315,-.0300102399588261,-.0300102399588264,.00463501676793134,.0306974312241583,.0188795034966013,-.00945869054146609,-.0147430199232581,-.0098489805145056,-.00631881794371566,.0126123320072455,.0205072233196368,.00309639341663969,-.0200481495737406,-.0200481495737408,.00309639341663958,.020507223319637,.0126123320072455,-.00631881794371548,-.00984898051450559,-.00434479121890393,-.00278749101749914,.00556381945057914,.00904658138698048,.00136594675023208,-.00884406503749014,-.00884406503749024,.00136594675023204,.00904658138698056,.00556381945057915,-.00278749101749906,-.00434479121890393},
			{.00738663555257316,.00473904940765641,-.00945912116166398,-.015380209620101,-.00232226367616612,.0150359089640387,.0150359089640388,-.00232226367616604,-.0153802096201012,-.00945912116166399,.00473904940765628,.00738663555257316,.0154232771965445,.00989512371934247,-.0197506221436933,-.0321138405466809,-.00484888094803859,.031394941608239,.0313949416082394,-.00484888094803843,-.0321138405466812,-.0197506221436933,.00989512371934219,.0154232771965445,.0196599725865644,.0126132636134246,-.0251760170659189,-.0409353483537532,-.00618084375318826,.0400189715524956,.040018971552496,-.00618084375318805,-.0409353483537535,-.0251760170659189,.0126132636134242,.0196599725865644,.0190345170231114,.0122119896103414,-.0243750759725306,-.0396330453491579,-.0059842085292456,.0387458218422929,.0387458218422933,-.0059842085292454,-.0396330453491583,-.0243750759725306,.0122119896103411,.0190345170231114,.0137005004254646,.00878984051179792,-.0175444818655976,-.0285267314116404,-.00430726198103432,.0278881333311886,.0278881333311889,-.00430726198103417,-.0285267314116406,-.0175444818655976,.00878984051179767,.0137005004254646,.00497726143544053,.00319326541691349,-.006373743314669,-.0103634900715324,-.00156478729132365,.0101314934655784,.0101314934655785,-.0015647872913236,-.0103634900715325,-.006373743314669,.0031932654169134,.00497726143544053,-.00497726143544053,-.00319326541691349,.006373743314669,.0103634900715324,.00156478729132365,-.0101314934655784,-.0101314934655785,.0015647872913236,.0103634900715325,.006373743314669,-.0031932654169134,-.00497726143544053,-.0137005004254646,-.00878984051179792,.0175444818655976,.0285267314116404,.00430726198103432,-.0278881333311886,-.0278881333311889,.00430726198103417,.0285267314116406,.0175444818655976,-.00878984051179767,-.0137005004254646,-.0190345170231114,-.0122119896103414,.0243750759725306,.0396330453491579,.0059842085292456,-.0387458218422929,-.0387458218422933,.0059842085292454,.0396330453491583,.0243750759725306,-.0122119896103411,-.0190345170231114,-.0196599725865644,-.0126132636134246,.0251760170659189,.0409353483537532,.00618084375318826,-.0400189715524956,-.040018971552496,.00618084375318805,.0409353483537535,.0251760170659189,-.0126132636134242,-.0196599725865644,-.0154232771965445,-.00989512371934246,.0197506221436933,.0321138405466809,.00484888094803859,-.031394941608239,-.0313949416082394,.00484888094803843,.0321138405466812,.0197506221436933,-.00989512371934219,-.0154232771965445,-.00738663555257316,-.00473904940765642,.00945912116166398,.015380209620101,.00232226367616612,-.0150359089640386,-.0150359089640388,.00232226367616604,.0153802096201012,.00945912116166399,-.00473904940765628,-.00738663555257316},
			{.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.00740597712796921,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0251362361485768,.0161266742054342,-.0321887687006243,-.0523378442423354,-.0079025109263769,.0511662116993059,.0511662116993065,-.00790251092637663,-.0523378442423359,-.0321887687006243,.0161266742054337,.0251362361485767,.0235568658789501,.0151133964164499,-.0301662708214208,-.0490493314083496,-.00740597712796946,.047951315360425,.0479513153604255,-.0074059771279692,-.04904933140835,-.0301662708214208,.0151133964164494,.0235568658789501,.020497390733761,.013150526613093,-.026248391580802,-.0426789928793145,-.00644411730055037,.041723582932936,.0417235829329365,-.00644411730055015,-.0426789928793148,-.026248391580802,.0131505266130926,.020497390733761,.0161501691603129,.0103614763511745,-.0206814598854213,-.0336273510880302,-.0050774064779557,.0328745707731171,.0328745707731175,-.00507740647795553,-.0336273510880305,-.0206814598854213,.0103614763511742,.0161501691603129,.0107890175957071,.00692191825117444,-.0138161174904002,-.0224645376147099,-.00339192904343299,.0219616475221864,.0219616475221867,-.00339192904343287,-.0224645376147101,-.0138161174904002,.00692191825117425,.0107890175957071,.0047594803179262,.00305354341917725,-.00609485883977799,-.00991003339090553,-.0014963196953581,.00968818784507958,.00968818784507969,-.00149631969535805,-.00991003339090562,-.00609485883977799,.00305354341917717,.00475948031792619},
			{.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648773,.00909571150529476,.000203199397318671,-.00473557895480586,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440274,.00068950902635523,-.0160690655456805,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.036404249458381,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0250100059159633,-.00107315666733768,-.0480371673090711,-.0176107025324564,.0513380676165945,.0388449726674057,-.038844972667407,-.0513380676165957,.0176107025324604,.0480371673090636,.00107315666733886,-.0250100059159602,.0234385669959365,-.0010057276487278,-.0450188763669993,-.0165041796687072,.0481123731564347,.0364042494583809,-.0364042494583821,-.0481123731564359,.0165041796687109,.0450188763669923,.00100572764872891,-.0234385669959336,.0203944560547184,-.000875107609545862,-.0391719978383809,-.0143606802852317,.0418637231617973,.0316762055424581,-.0316762055424591,-.0418637231617983,.0143606802852349,.0391719978383748,.000875107609546825,-.020394456054716,.0160690655456825,-.000689509026354471,-.0308641426440322,-.0113149726653577,.0329849891396142,.0249581073275306,-.0249581073275314,-.032984989139615,.0113149726653602,.0308641426440273,.00068950902635523,-.0160690655456805,.0107348368427603,-.000460622111378126,-.020618593821368,-.0075588953880111,.0220354118083187,.0166731045625194,-.0166731045625199,-.0220354118083192,.00755889538801282,.0206185938213648,.000460622111378633,-.010734836842759,.00473557895480644,-.000203199397318448,-.00909571150529619,-.00333454028648697,.0097207283117994,.00735519358459587,-.00735519358459611,-.00972072831179962,.00333454028648772,.00909571150529476,.000203199397318671,-.00473557895480586}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[56][144] = 
		{
			{-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.0041555238362304,-.00941994016019063,-.0141007858887576,-.0178963647467683,-.0205676063620141,-.0219465617023582,-.0219465617023582,-.0205676063620141,-.0178963647467683,-.0141007858887576,-.00941994016019063,-.0041555238362304,-.00389442226307229,-.00882806262758282,-.0132147995430121,-.0167718930379425,-.0192752941075644,-.0205676063620141,-.0205676063620141,-.0192752941075644,-.0167718930379425,-.0132147995430121,-.00882806262758282,-.00389442226307229,-.00338862967674239,-.00768150780454087,-.0114985121999394,-.0145936240716447,-.0167718930379425,-.0178963647467683,-.0178963647467683,-.0167718930379425,-.0145936240716447,-.0114985121999394,-.00768150780454087,-.00338862967674239,-.00266994678551479,-.00605236305737515,-.0090598320309655,-.0114985121999394,-.0132147995430121,-.0141007858887576,-.0141007858887576,-.0132147995430121,-.0114985121999394,-.0090598320309655,-.00605236305737515,-.00266994678551479,-.00178364093667255,-.00404324257371201,-.00605236305737515,-.00768150780454087,-.00882806262758282,-.00941994016019063,-.00941994016019063,-.00882806262758282,-.00768150780454087,-.00605236305737515,-.00404324257371201,-.00178364093667255,-.000786837527794772,-.00178364093667255,-.00266994678551479,-.00338862967674239,-.00389442226307229,-.0041555238362304,-.0041555238362304,-.00389442226307229,-.00338862967674239,-.00266994678551479,-.00178364093667255,-.000786837527794772},
			{.00133771262320103,.00303239094725451,.00453921094519749,.0057610529922694,.00662095748786621,.00706485963794559,.00706485963794559,.00662095748786621,.0057610529922694,.00453921094519749,.00303239094725451,.00133771262320103,.00279314072693881,.00633162497525767,.00947786149229762,.0120290647360828,.0138245432463938,.0147514098638258,.0147514098638258,.0138245432463938,.0120290647360828,.00947786149229762,.00633162497525767,.00279314072693881,.00356040220390621,.00807088998373582,.0120813815859818,.0153333873170859,.0176220745942872,.0188035467326598,.0188035467326598,.0176220745942872,.0153333873170859,.0120813815859818,.00807088998373582,.00356040220390621,.00344713280046435,.00781412548316908,.0116970286936326,.014845576239943,.0170614520122337,.0182053372049096,.0182053372049096,.0170614520122337,.014845576239943,.0116970286936326,.00781412548316909,.00344713280046435,.00248114750387688,.00562438381687355,.00841918638645813,.0106854207724132,.0122803447163296,.0131036805304146,.0131036805304146,.0122803447163296,.0106854207724132,.00841918638645813,.00562438381687355,.00248114750387688,.000901377278433765,.00204328512101712,.0030586103001915,.0038819116879217,.0044613323800127,.00476044244669653,.00476044244669653,.0044613323800127,.0038819116879217,.0030586103001915,.00204328512101712,.000901377278433765,-.000901377278433765,-.00204328512101712,-.0030586103001915,-.0038819116879217,-.0044613323800127,-.00476044244669653,-.00476044244669653,-.0044613323800127,-.0038819116879217,-.0030586103001915,-.00204328512101712,-.000901377278433765,-.00248114750387688,-.00562438381687355,-.00841918638645813,-.0106854207724132,-.0122803447163296,-.0131036805304146,-.0131036805304146,-.0122803447163296,-.0106854207724132,-.00841918638645813,-.00562438381687355,-.00248114750387688,-.00344713280046435,-.00781412548316908,-.0116970286936326,-.014845576239943,-.0170614520122337,-.0182053372049096,-.0182053372049096,-.0170614520122337,-.014845576239943,-.0116970286936326,-.00781412548316909,-.00344713280046435,-.00356040220390621,-.00807088998373582,-.0120813815859818,-.0153333873170859,-.0176220745942872,-.0188035467326598,-.0188035467326598,-.0176220745942872,-.0153333873170859,-.0120813815859818,-.00807088998373582,-.00356040220390621,-.00279314072693881,-.00633162497525767,-.00947786149229762,-.0120290647360828,-.0138245432463938,-.0147514098638258,-.0147514098638258,-.0138245432463938,-.0120290647360828,-.00947786149229762,-.00633162497525767,-.00279314072693881,-.00133771262320103,-.00303239094725451,-.00453921094519749,-.0057610529922694,-.00662095748786621,-.0070648596379456,-.0070648596379456,-.00662095748786621,-.0057610529922694,-.00453921094519749,-.00303239094725451,-.00133771262320103},
			{-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542,-.00289610299382129,-.00656502476575684,-.0098272395580117,-.0124724866380865,-.0143341510500878,-.0152951843269753,-.0152951843269753,-.0143341510500878,-.0124724866380865,-.0098272395580117,-.00656502476575684,-.00289610299382129,-.00232314841306506,-.00526622392188377,-.00788305389439265,-.0100049748237765,-.0114983342566515,-.0122692401729353,-.0122692401729353,-.0114983342566515,-.0100049748237765,-.00788305389439265,-.00526622392188377,-.00232314841306506,-.000131946205964698,-.00029910196969996,-.000447728198048283,-.0005682454299284,-.00065306270212733,-.000696847296446543,-.000696847296446543,-.00065306270212733,-.0005682454299284,-.000447728198048283,-.00029910196969996,-.000131946205964698,.00258676855320165,.00586381066256298,.00877758640064877,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064877,.00586381066256298,.00258676855320165,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902698,.0233825305902698,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00442742077492767,.0100362891436658,.0150234037505425,.0190673282594864,.0219133498654938,.0233825305902699,.0233825305902699,.0219133498654938,.0190673282594864,.0150234037505425,.0100362891436658,.00442742077492767,.00258676855320165,.00586381066256299,.00877758640064879,.0111402931057571,.0128031120620766,.013661496817224,.013661496817224,.0128031120620766,.0111402931057571,.00877758640064879,.00586381066256299,.00258676855320165,-.000131946205964692,-.000299101969699946,-.000447728198048262,-.000568245429928373,-.000653062702127299,-.00069684729644651,-.00069684729644651,-.000653062702127299,-.000568245429928373,-.000447728198048261,-.000299101969699946,-.000131946205964692,-.00232314841306506,-.00526622392188376,-.00788305389439264,-.0100049748237765,-.0114983342566515,-.0122692401729352,-.0122692401729352,-.0114983342566515,-.0100049748237765,-.00788305389439264,-.00526622392188376,-.00232314841306506,-.00289610299382128,-.00656502476575683,-.00982723955801168,-.0124724866380864,-.0143341510500878,-.0152951843269752,-.0152951843269752,-.0143341510500878,-.0124724866380864,-.00982723955801168,-.00656502476575683,-.00289610299382128,-.00166299164366542,-.00376974898655276,-.00564296825773739,-.00716191416504105,-.00823091356425874,-.00878275523292796,-.00878275523292796,-.00823091356425874,-.00716191416504105,-.00564296825773739,-.00376974898655276,-.00166299164366542},
			{.00185673265998067,.00420893038754785,.00630038250840617,.0079962878880739,.00918982731780452,.00980595936709009,.00980595936709009,.00918982731780452,.0079962878880739,.00630038250840617,.00420893038754785,.00185673265998067,.00231919206364203,.00525725547980847,.00786962895969996,.00998793623246273,.0114787524564332,.0122483455107577,.0122483455107577,.0114787524564332,.00998793623246273,.00786962895969996,.00525725547980847,.00231919206364203,-9.85729644424585e-005,-.000223449909820004,-.000334484007504598,-.000424518735869586,-.000487883119069876,-.000520593246863243,-.000520593246863243,-.000487883119069876,-.000424518735869586,-.000334484007504598,-.000223449909820004,-9.85729644424585e-005,-.00335756724967493,-.00761109400937633,-.0113931092109241,-.0144598492344335,-.0166181507426867,-.0177323147980429,-.0177323147980429,-.0166181507426867,-.0144598492344335,-.0113931092109241,-.00761109400937633,-.00335756724967493,-.0044030502985414,-.00998104498233226,-.0149407083111532,-.0189623732762762,-.0217927291243017,-.023253822830477,-.023253822830477,-.0217927291243018,-.0189623732762762,-.0149407083111532,-.00998104498233226,-.0044030502985414,-.00201132951760402,-.00455937793764214,-.0068249703279767,-.00866208162714706,-.00995499855438335,-.0106224315156169,-.0106224315156169,-.00995499855438335,-.00866208162714707,-.0068249703279767,-.00455937793764214,-.00201132951760402,.00201132951760402,.00455937793764215,.00682497032797671,.00866208162714707,.00995499855438336,.0106224315156169,.0106224315156169,.00995499855438336,.00866208162714707,.00682497032797671,.00455937793764215,.00201132951760402,.00440305029854142,.0099810449823323,.0149407083111532,.0189623732762763,.0217927291243019,.0232538228304771,.0232538228304771,.0217927291243019,.0189623732762763,.0149407083111532,.0099810449823323,.00440305029854142,.00335756724967493,.00761109400937631,.011393109210924,.0144598492344335,.0166181507426867,.0177323147980428,.0177323147980428,.0166181507426867,.0144598492344335,.011393109210924,.00761109400937631,.00335756724967493,9.85729644424547e-005,.000223449909819995,.000334484007504586,.00042451873586957,.000487883119069857,.000520593246863223,.000520593246863223,.000487883119069857,.00042451873586957,.000334484007504586,.000223449909819995,9.85729644424547e-005,-.00231919206364203,-.00525725547980847,-.00786962895969996,-.00998793623246273,-.0114787524564332,-.0122483455107577,-.0122483455107577,-.0114787524564332,-.00998793623246273,-.00786962895969996,-.00525725547980847,-.00231919206364203,-.00185673265998066,-.00420893038754783,-.00630038250840614,-.00799628788807386,-.00918982731780447,-.00980595936709004,-.00980595936709004,-.00918982731780447,-.00799628788807386,-.00630038250840614,-.00420893038754783,-.00185673265998067},
			{-.00194304970229352,-.00440459798697359,-.0065932789470194,-.00836802472173726,-.00961705021883699,-.0102618254310909,-.0102618254310909,-.00961705021883699,-.00836802472173726,-.0065932789470194,-.00440459798697359,-.00194304970229352,-.00124660388009712,-.00282586129132368,-.00423005500487988,-.00536868000574208,-.00617001824695479,-.00658368717186054,-.00658368717186054,-.00617001824695479,-.00536868000574208,-.00423005500487988,-.00282586129132368,-.00124660388009713,.00248821570068215,.00564040634341175,.00844317063819117,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819117,.00564040634341175,.00248821570068215,.00404575418820262,.00917110907256417,.0137283085874776,.0174236258727659,.0200243056855872,.021366835427542,.021366835427542,.0200243056855872,.0174236258727659,.0137283085874776,.00917110907256417,.00404575418820262,.000610869957304142,.00138474923264624,.00207284251355276,.00263079987014941,.00302347750015778,.00322618657441121,.00322618657441121,.00302347750015778,.00263079987014941,.00207284251355276,.00138474923264624,.000610869957304142,-.00395518612341859,-.00896580505210553,-.013420987310977,-.0170335813956648,-.0195760425113499,-.0208885184450436,-.0208885184450436,-.0195760425113499,-.0170335813956648,-.013420987310977,-.00896580505210553,-.00395518612341859,-.00395518612341863,-.00896580505210563,-.0134209873109772,-.017033581395665,-.0195760425113501,-.0208885184450439,-.0208885184450439,-.0195760425113501,-.017033581395665,-.0134209873109772,-.00896580505210563,-.00395518612341863,.000610869957304121,.0013847492326462,.00207284251355269,.00263079987014932,.00302347750015768,.0032261865744111,.0032261865744111,.00302347750015768,.00263079987014932,.00207284251355269,.0013847492326462,.000610869957304121,.00404575418820265,.00917110907256426,.0137283085874777,.0174236258727661,.0200243056855874,.0213668354275422,.0213668354275422,.0200243056855874,.0174236258727661,.0137283085874777,.00917110907256426,.00404575418820265,.00248821570068215,.00564040634341176,.00844317063819118,.0107158609847942,.0123153284862006,.0131410097874288,.0131410097874288,.0123153284862006,.0107158609847942,.00844317063819118,.00564040634341176,.00248821570068215,-.00124660388009709,-.0028258612913236,-.00423005500487976,-.00536868000574194,-.00617001824695462,-.00658368717186035,-.00658368717186035,-.00617001824695462,-.00536868000574194,-.00423005500487976,-.00282586129132361,-.00124660388009709,-.00194304970229352,-.00440459798697359,-.00659327894701939,-.00836802472173725,-.00961705021883698,-.0102618254310909,-.0102618254310909,-.00961705021883698,-.00836802472173725,-.00659327894701939,-.00440459798697359,-.00194304970229352},
			{.00193329201257368,.00438247878927793,.00656016853809603,.00832600181892304,.00956875490660448,.0102102921592464,.0102102921592464,.00956875490660448,.00832600181892304,.00656016853809603,.00438247878927793,.00193329201257368,-8.29558065749951e-005,-.000188048189511932,-.000281490881256575,-.000357261185553734,-.000410586593247804,-.000438114374821766,-.000438114374821766,-.000410586593247804,-.000357261185553734,-.000281490881256575,-.000188048189511932,-8.29558065749951e-005,-.00371330867243083,-.00841750567898221,-.01260023347052,-.0159919011508891,-.0183788793146453,-.019611091431816,-.019611091431816,-.0183788793146453,-.0159919011508891,-.01260023347052,-.00841750567898221,-.00371330867243082,-.00136132037138296,-.00308590611981871,-.00461931824708002,-.00586272317609315,-.00673780313496994,-.00718953920243542,-.00718953920243542,-.00673780313496994,-.00586272317609315,-.00461931824708002,-.00308590611981871,-.00136132037138295,.00396847071518603,.00899591920002855,.0134660654266406,.0170907934125251,.0196417940904338,.0209586783392964,.0209586783392964,.0196417940904338,.0170907934125251,.0134660654266406,.00899591920002855,.00396847071518603,.0030027452068136,.00680676643408263,.0101891046492422,.0129317567605373,.0148619726063683,.0158583936839055,.0158583936839055,.0148619726063683,.0129317567605373,.0101891046492422,.00680676643408263,.0030027452068136,-.0030027452068137,-.00680676643408285,-.0101891046492425,-.0129317567605377,-.0148619726063688,-.0158583936839061,-.0158583936839061,-.0148619726063688,-.0129317567605377,-.0101891046492425,-.00680676643408285,-.0030027452068137,-.00396847071518612,-.00899591920002876,-.0134660654266409,-.0170907934125255,-.0196417940904342,-.0209586783392969,-.0209586783392969,-.0196417940904342,-.0170907934125255,-.0134660654266409,-.00899591920002876,-.00396847071518612,.00136132037138326,.00308590611981942,.00461931824708107,.00586272317609448,.00673780313497147,.00718953920243705,.00718953920243705,.00673780313497147,.00586272317609448,.00461931824708107,.00308590611981942,.00136132037138326,.00371330867243024,.00841750567898089,.012600233470518,.0159919011508866,.0183788793146425,.0196110914318129,.0196110914318129,.0183788793146425,.0159919011508866,.012600233470518,.00841750567898089,.00371330867243024,8.29558065750864e-005,.000188048189512139,.000281490881256885,.000357261185554128,.000410586593248256,.000438114374822248,.000438114374822248,.000410586593248256,.000357261185554128,.000281490881256885,.000188048189512139,8.29558065750864e-005,-.00193329201257344,-.0043824787892774,-.00656016853809523,-.00832600181892204,-.00956875490660332,-.0102102921592451,-.0102102921592451,-.00956875490660332,-.00832600181892204,-.00656016853809523,-.0043824787892774,-.00193329201257344},
			{-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0115910496926654,-.0247276171423163,-.0330003087473892,-.0349515342460601,-.0305921594597511,-.0213488900791909,-.00968823513593025,.00150517158936591,.00964225249732755,.0130587860981281,.0114058100075172,.00571425152315335,-.0108627561179969,-.0231739213892035,-.0309268198520728,-.0327554451522206,-.0286699803853337,-.0200074878867027,-.00907949998368246,.00141059803249448,.00903640654503392,.0122382711093729,.0106891554885031,.00535521133449932,-.00945194312945921,-.0201641816016119,-.0269101634282779,-.0285012939069832,-.0249464335921091,-.0174089923049226,-.00790028944380993,.00122739498492244,.007862792815244,.0106488115237211,.00930088908208952,.00465969708148589,-.0074473128027459,-.0158876292146258,-.0212028788026431,-.0224565518540571,-.0196556297185417,-.013716778603044,-.00622474404621712,.000967080975841837,.00619519994950027,.00839034146824932,.00732829529222188,.00367143784686028,-.00497512986223131,-.0106136294579435,-.0141644480754732,-.0150019563688357,-.0131308181841942,-.0091634065399345,-.00415840059486461,.00064605228349379,.00413866384931332,.00560511415306926,.00489562097272715,.00245268065851968,-.00219473482625876,-.00468210938194629,-.00624852181688227,-.00661798124198739,-.00579254507201903,-.00404235628360899,-.00183444188590311,.000285000288521492,.00182573519401407,.00247264686099665,.00215966017823466,.00108197852277371},
			{.00373129696748193,.00796011451098884,.0106231924821284,.0112513152129516,.0098479805408067,.00687246202231334,.00311875829553411,-.000484532664068795,-.00310395593642735,-.00420377879992814,-.00367166610627629,-.00183948563288991,.0077909390577761,.016620699876881,.0221812002498267,.0234927189949964,.0205625595883267,.0143496841069869,.00651195978990583,-.00101170303252743,-.00648105249983563,-.00877747998297588,-.00766642996357517,-.00384084155948903,.0099310701835657,.0211863211550268,.0282742625507885,.0299460487897454,.0262109895753331,.0182914689643539,.00830076184995504,-.00128961268294456,-.00826136448532482,-.0111886088569481,-.00977235907777577,-.00489590109849087,.00961512655393934,.0205123068664937,.0273747549478629,.0289933555580273,.0253771222247394,.0177095505014919,.00803668428539622,-.00124858539139083,-.00799854029488872,-.0108326583272072,-.0094614646283383,-.00474014459543961,.00692069282783456,.0147641712479721,.0197035649160317,.0208685872972819,.0182657260709564,.0127468274548747,.00578457527121418,-.000898696019720848,-.00575712031884831,-.00779703734224594,-.00681009137288203,-.00341182037704094,.00251422184947983,.00536368292374106,.00715811764181876,.00758135918698483,.00663576447137033,.00463080112004091,.00210148115205368,-.000326487712290173,-.00209150702911081,-.00283258947258012,-.00247404139334002,-.0012394818772994,-.00251422184947983,-.00536368292374106,-.00715811764181876,-.00758135918698483,-.00663576447137033,-.00463080112004091,-.00210148115205368,.000326487712290173,.00209150702911081,.00283258947258012,.00247404139334002,.0012394818772994,-.00692069282783456,-.0147641712479721,-.0197035649160317,-.0208685872972819,-.0182657260709564,-.0127468274548747,-.00578457527121418,.000898696019720848,.00575712031884831,.00779703734224594,.00681009137288203,.00341182037704094,-.00961512655393934,-.0205123068664937,-.0273747549478629,-.0289933555580273,-.0253771222247394,-.0177095505014919,-.00803668428539622,.00124858539139083,.00799854029488872,.0108326583272072,.0094614646283383,.00474014459543961,-.0099310701835657,-.0211863211550268,-.0282742625507885,-.0299460487897454,-.026210989575333,-.0182914689643539,-.00830076184995504,.00128961268294456,.00826136448532482,.0111886088569481,.00977235907777577,.00489590109849087,-.0077909390577761,-.016620699876881,-.0221812002498267,-.0234927189949964,-.0205625595883267,-.0143496841069869,-.00651195978990583,.00101170303252743,.00648105249983563,.00877747998297588,.00766642996357517,.00384084155948903,-.00373129696748193,-.00796011451098884,-.0106231924821284,-.0112513152129516,-.00984798054080671,-.00687246202231334,-.00311875829553411,.000484532664068795,.00310395593642735,.00420377879992814,.00367166610627629,.00183948563288991},
			{-.00463860142102606,-.00989570082574293,-.0132063344657042,-.0139871919040607,-.0122426215144303,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173,-.0080781328720134,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782068,.00104899697753615,.00671996056648802,.00910104020299664,.00794903405616202,.00398242474081937,-.006479984172588,-.0138239910928359,-.0184488449315262,-.0195396788666562,-.0171025674430431,-.011935111442324,-.00541621440729949,.000841467195356086,.00539050778207919,.00730052320305626,.00637642580128229,.00319455617998525,-.000368039046268343,-.000785151377213537,-.00104782590705614,-.00110978122522199,-.00097136234330704,-.000677870024886959,-.000307620872482971,4.77922130358021e-005,.00030616083159755,.000414642617227202,.000362157315203368,.000181438932321165,.00721530281433017,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.013289452703208,.00603082137473739,-.00093695300190748,-.00600219768054942,-.00812895281996605,-.00709999311788209,-.00355705964429505,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450082,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.0123494548973161,.0263455511549027,.0351595269866947,.0372384216450083,.0325938119046078,.0227457531570724,.0103221387206607,-.00160365533307635,-.0102731474268141,-.0139132256532883,-.0121520949344494,-.00608813639214713,.00721530281433019,.0153926736826547,.0205423345505516,.0217569512768401,.0190432877175959,.0132894527032081,.0060308213747374,-.000936953001907482,-.00600219768054943,-.00812895281996606,-.0070999931178821,-.00355705964429506,-.000368039046268326,-.000785151377213499,-.00104782590705609,-.00110978122522194,-.000971362343306994,-.000677870024886926,-.000307620872482956,4.77922130357998e-005,.000306160831597535,.000414642617227182,.000362157315203351,.000181438932321156,-.00647998417258799,-.0138239910928359,-.0184488449315262,-.0195396788666561,-.0171025674430431,-.011935111442324,-.00541621440729948,.000841467195356084,.00539050778207918,.00730052320305625,.00637642580128228,.00319455617998524,-.00807813287201338,-.0172333811156298,-.0229988556642598,-.0243587203235838,-.0213205477942233,-.0148786499327011,-.00675200717782067,.00104899697753615,.00671996056648801,.00910104020299662,.007949034056162,.00398242474081937,-.00463860142102606,-.00989570082574293,-.0132063344657043,-.0139871919040607,-.0122426215144304,-.00854357409245923,-.00387711747083593,.000602351923116915,.00385871576103221,.00522597222492994,.00456447068312552,.00228677608484173},
			{.00517900543192415,.0110485647887122,.0147448922047198,.0156167207038177,.0136689052516612,.00953891326647014,.00432880746137177,-.000672526824057758,-.00430826192482315,-.00583480581394941,-.00509623835206271,-.00255318892270115,.00646894868289895,.0138004486724991,.0184174263304009,.0195064025624373,.0170734415684041,.0119147858063732,.00540699053010265,-.000840034166769796,-.00538132768356915,-.00728809032569326,-.00636556667272444,-.00318911581302657,-.000274950685842771,-.000586562517874416,-.000782798604414518,-.000829083522805726,-.000725675020631865,-.000506415909248477,-.000229814120882911,3.57040968488888e-005,.000228723369108405,.000309766784644738,.000270556624922728,.00013554746265675,-.00936530034663045,-.0199793433325387,-.0266634870132945,-.0282400321341923,-.0247177580278924,-.0172494099292266,-.00782787014830752,.00121614386801799,.00779071723872097,.0105511974509746,.00921563096089514,.00461698320596323,-.0122814780526446,-.026200533625503,-.0349659934482049,-.0370334449537273,-.0324144011931623,-.0226205505030159,-.0102653211180761,.00159482810706601,.0102165994939113,.013836641124904,.0120852044460302,.0060546246051647,-.00561022419736452,-.0119684998091909,-.0159725939896759,-.0169170134165105,-.0148070172936393,-.0103331503949069,-.00468923631858351,.000728523325820062,.00466698010205544,.00632062838987608,.00552056569433968,.00276577471540888,.00561022419736453,.0119684998091909,.0159725939896759,.0169170134165105,.0148070172936393,.010333150394907,.00468923631858351,-.000728523325820062,-.00466698010205545,-.00632062838987609,-.00552056569433968,-.00276577471540888,.0122814780526447,.0262005336255031,.034965993448205,.0370334449537275,.0324144011931624,.022620550503016,.0102653211180761,-.00159482810706601,-.0102165994939113,-.0138366411249041,-.0120852044460302,-.00605462460516473,.00936530034663043,.0199793433325386,.0266634870132944,.0282400321341923,.0247177580278923,.0172494099292265,.0078278701483075,-.00121614386801798,-.00779071723872095,-.0105511974509746,-.00921563096089512,-.00461698320596322,.00027495068584276,.000586562517874393,.000782798604414488,.000829083522805695,.000725675020631837,.000506415909248458,.000229814120882903,-3.57040968488874e-005,-.000228723369108396,-.000309766784644726,-.000270556624922718,-.000135547462656745,-.00646894868289895,-.0138004486724991,-.0184174263304009,-.0195064025624373,-.0170734415684041,-.0119147858063732,-.00540699053010265,.000840034166769796,.00538132768356914,.00728809032569326,.00636556667272443,.00318911581302657,-.00517900543192412,-.0110485647887121,-.0147448922047197,-.0156167207038176,-.0136689052516611,-.00953891326647009,-.00432880746137175,.000672526824057755,.00430826192482312,.00583480581394939,.00509623835206268,.00255318892270114},
			{-.00541977053728884,-.0115621979370935,-.015430362720618,-.0163427213723106,-.0143043545587521,-.0099823647144083,-.00453004798877423,.000703791705661616,.0045085473174344,.00610605821077421,.00533315572547085,.0026718832952154,-.00347716631903209,-.00741796815275411,-.00989966958442869,-.0104850122207482,-.00917725566884735,-.00640439334662279,-.00290634634470048,.000451532181612414,.00289255214258342,.00391746842536516,.0034215967881247,.00171420220442671,.00694040822993805,.0148062308480902,.019759695669824,.0209280369217824,.0183177607650484,.012783140124006,.00580105414555624,-.000901256190189091,-.00577352098058376,-.00781924924071369,-.00682949169783763,-.00342153983899534,.0112848679704134,.0240744282936295,.0321285937918819,.0340282769713043,.0297840566287605,.0207849515140105,.0094323169405852,-.00146541194362748,-.00938754894982324,-.0127138335939062,-.0111045214577846,-.00556330752592008,.0017039064892699,.00363500704682817,.00485110854611964,.00513794242893165,.00449710599180968,.00313832770190446,.00142418910757755,-.000221263104428595,-.00141742957169544,-.00191966655001643,-.00167667590101747,-.000840005910576199,-.0110322454417378,-.0235354992633652,-.0314093645879802,-.0332665215482455,-.0291173121245713,-.0203196605576931,-.00922116553296151,.00143260730012424,.00917739971635521,.012429222307352,.0108559361577392,.00543876758281373,-.011032245441738,-.0235354992633655,-.0314093645879806,-.0332665215482459,-.0291173121245716,-.0203196605576934,-.00922116553296161,.00143260730012425,.00917739971635532,.0124292223073521,.0108559361577394,.00543876758281379,.00170390648926984,.00363500704682805,.00485110854611947,.00513794242893148,.00449710599180952,.00313832770190435,.0014241891075775,-.000221263104428587,-.00141742957169539,-.00191966655001636,-.00167667590101741,-.00084000591057617,.0112848679704135,.0240744282936297,.0321285937918822,.0340282769713046,.0297840566287608,.0207849515140107,.00943231694058528,-.0014654119436275,-.00938754894982332,-.0127138335939063,-.0111045214577847,-.00556330752592013,.00694040822993806,.0148062308480902,.0197596956698241,.0209280369217824,.0183177607650485,.012783140124006,.00580105414555625,-.000901256190189092,-.00577352098058377,-.00781924924071369,-.00682949169783764,-.00342153983899535,-.00347716631903199,-.00741796815275391,-.00989966958442842,-.0104850122207479,-.0091772556688471,-.00640439334662262,-.0029063463447004,.000451532181612402,.00289255214258334,.00391746842536505,.0034215967881246,.00171420220442666,-.00541977053728884,-.0115621979370935,-.0154303627206179,-.0163427213723106,-.0143043545587521,-.00998236471440829,-.00453004798877423,.000703791705661616,.00450854731743439,.0061060582107742,.00533315572547085,.0026718832952154},
			{-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204719,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359087,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0208025406961852,-.0399360129142935,-.0427841602198989,-.029437434204069,-.00734147704933528,.0127554214359088,.0222763063293018,.0188658840114939,.00697324020575009,-.0051770667545793,-.0104331931866418,-.00667282142029397,-.0194954669515699,-.0374267370182747,-.0400959283675706,-.0275878092991634,-.00688019436093554,.0119539675796769,.0208766323397023,.0176804950717236,.00653509472530391,-.0048517791789962,-.00977765052548706,-.00625355197583752,-.0169634706797109,-.032565896350363,-.0348884233924082,-.024004810724789,-.00598662117723957,.0104014322430853,.0181652556189287,.0153842203675868,.00568634175509704,-.0042216487582364,-.00850776687819747,-.00544136469517216,-.0133657461372436,-.0256591066515262,-.0274890568679043,-.0189137124281916,-.00471693914444836,.00819542801409716,.0143126486144801,.0121214336284679,.00448034495907358,-.00332629369594071,-.00670338366693099,-.00428732424684867,-.00892890153799983,-.0171414026940184,-.0183638892752913,-.0126351850660089,-.00315112113824737,.00547490346204718,.00956147370408142,.00809764649549972,.00299306590032929,-.00222210930781969,-.00447815274350306,-.00286411964349082,-.00393890646241235,-.00756177919074763,-.00810106840506984,-.00557390087666116,-.00139008940377084,.00241520555870888,.0042179601155764,.00357220560398291,.00132036472426753,-.000980263997257417,-.00197549773687001,-.00126348121601029},
			{.00669658564383886,.0128558782631327,.0137727308070988,.00947626072033451,.00236330891169302,-.00410612209904174,-.00717100835620119,-.0060731527881528,-.00224476908541414,.00166655945599322,.00335856916390389,.00214806069987574,.013982454653438,.0268430427690799,.028757428666868,.0197864095008088,.00493458330070284,-.00857357302734623,-.0149730481312158,-.0126807283562296,-.0046870724296822,.00347977211971043,.00701268430396727,.00448514555420918,.0178233634574797,.0342166893750909,.0366569472911615,.025221635027061,.00629008810395184,-.010928689703139,-.0190860678989135,-.0161640596018723,-.00597458726213122,.00443564772967482,.00893903282793333,.00571719210637408,.0172563371411996,.033128131461881,.0354907558571953,.0244192426596455,.00608997741804402,-.0105810081513981,-.0184788703405621,-.0156498218041611,-.00578451381080864,.00429453356800305,.00865464953138992,.00553530733545099,.0124206174529098,.0238446806208565,.025545230022447,.0175762717825614,.00438339139919734,-.00761590675004047,-.0133005618506412,-.0112642937052734,-.00416352743964371,.00309108231662293,.00622936896392408,.00398415575306398,.00451229213042641,.00866254557198487,.0092803389877543,.00638529228900029,.00159244438452158,-.00276678645200671,-.00483196755688163,-.00409221071607914,-.00151256990016366,.00112296079197981,.00226307046811526,.00144740587326696,-.00451229213042641,-.00866254557198487,-.0092803389877543,-.00638529228900029,-.00159244438452158,.00276678645200671,.00483196755688163,.00409221071607914,.00151256990016366,-.00112296079197981,-.00226307046811526,-.00144740587326696,-.0124206174529098,-.0238446806208565,-.025545230022447,-.0175762717825614,-.00438339139919734,.00761590675004047,.0133005618506412,.0112642937052734,.00416352743964371,-.00309108231662293,-.00622936896392408,-.00398415575306398,-.0172563371411996,-.033128131461881,-.0354907558571953,-.0244192426596455,-.00608997741804402,.0105810081513981,.0184788703405621,.0156498218041611,.00578451381080864,-.00429453356800305,-.00865464953138992,-.00553530733545099,-.0178233634574797,-.0342166893750909,-.0366569472911615,-.025221635027061,-.00629008810395184,.010928689703139,.0190860678989135,.0161640596018723,.00597458726213122,-.00443564772967482,-.00893903282793333,-.00571719210637408,-.013982454653438,-.0268430427690799,-.028757428666868,-.0197864095008088,-.00493458330070284,.00857357302734624,.0149730481312158,.0126807283562296,.0046870724296822,-.00347977211971043,-.00701268430396727,-.00448514555420918,-.00669658564383886,-.0128558782631327,-.0137727308070988,-.00947626072033451,-.00236330891169302,.00410612209904174,.00717100835620119,.00607315278815281,.00224476908541414,-.00166655945599322,-.00335856916390389,-.00214806069987574},
			{-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203508,.00754990434136365,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315,-.0144978834787229,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606307,.00888961672521672,.0155249927504182,.0131481722408983,.00485984983510451,-.00360804536645972,-.00727118967533986,-.00465047942165492,-.0116296744515833,-.0223262550411753,-.0239185137194828,-.0164570175096932,-.00410425771180905,.00713092698429712,.012453584125954,.0105469852216292,.00389839465524694,-.00289424267204597,-.00583268371028218,-.00373044532996883,-.000660522337643703,-.00126804840771953,-.00135848278992925,-.000934697503477068,-.000233106602371594,.000405010181569894,.000707317348663856,.000599029608497395,.000221414344955362,-.000164382239881848,-.000331274782891918,-.000211875446734245,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627044,.00456998989967055,-.00794011160648144,-.0138667592696614,-.0117438083374094,-.00434076645531976,.00322266794806499,.006494549688551,.0041537590171735,.0221636560036439,.0425490359719281,.0455835382498503,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183392,.00551580349401449,.0111158395594773,.0071094257520315,.0221636560036439,.0425490359719281,.0455835382498504,.0313635326981254,.00782183168183649,-.0135900117690351,-.0237338504813011,-.0201002834000981,-.00742950101183393,.0055158034940145,.0111158395594773,.00710942575203151,.0129493561350382,.0248597352310675,.0266326760619006,.0183244837627045,.00456998989967055,-.00794011160648145,-.0138667592696614,-.0117438083374094,-.00434076645531976,.003222667948065,.00649454968855101,.0041537590171735,-.000660522337643672,-.00126804840771947,-.00135848278992918,-.000934697503477023,-.000233106602371583,.000405010181569874,.000707317348663822,.000599029608497366,.000221414344955352,-.00016438223988184,-.000331274782891902,-.000211875446734235,-.0116296744515833,-.0223262550411753,-.0239185137194827,-.0164570175096932,-.00410425771180904,.0071309269842971,.012453584125954,.0105469852216291,.00389839465524693,-.00289424267204597,-.00583268371028217,-.00373044532996882,-.0144978834787228,-.0278325455670121,-.0298175005958216,-.0205157868568154,-.00511648458606306,.0088896167252167,.0155249927504182,.0131481722408983,.0048598498351045,-.00360804536645972,-.00727118967533985,-.00465047942165491,-.00832493151223852,-.0159819214988645,-.0171217164692106,-.0117805140835438,-.00293797255474641,.00510456926461644,.00891471514203509,.00754990434136366,.00279060851167585,-.00207180107447468,-.00417524089971735,-.00267038445585315},
			{.00929479849593368,.0178438392786048,.0191164220692825,.0131529616098434,.00328025075555347,-.00569925922554386,-.0099532928015145,-.0084294794994272,-.00311571858085951,.0023131690011613,.00466166271491865,.00298148824255666,.0116098689755076,.0222882331591425,.0238777802016866,.016428991010076,.00409726811134705,-.00711878292950237,-.0124323755217502,-.0105290235783908,-.00389175564205705,.00288931374181348,.00582275057946566,.00372409233650869,-.000493455983937409,-.000947320081473204,-.0010148808353067,-.000698283842916624,-.000174146794559474,.000302570687259703,.000528415101729981,.000447516651629383,.000165411867579778,-.000122804930734209,-.000247485231958592,-.0001582856492233,-.0168079722487348,-.032267375730295,-.034568612948002,-.0237847666975931,-.00593174383418366,.0103060858116714,.01799874082957,.0152431983931265,.0056342169725212,-.0041829503237926,-.00842977904035172,-.00539148553491988,-.0220416574634008,-.0423148273072351,-.045332626345935,-.0311908939779877,-.00777877686962253,.0135152063489619,.0236032088979544,.0199896425729312,.00738860576070614,-.00548544207826128,-.011054653074738,-.00707029233633635,-.0100687099322352,-.0193295682367182,-.0207081099006063,-.0142481147124679,-.00355337379041754,.00617379580587153,.010782032352062,.00913134200774618,.00337514219752845,-.00250577005054177,-.00504980605001451,-.00322973545836477,.0100687099322352,.0193295682367182,.0207081099006063,.0142481147124679,.00355337379041755,-.00617379580587153,-.010782032352062,-.00913134200774619,-.00337514219752846,.00250577005054177,.00504980605001452,.00322973545836478,.0220416574634009,.0423148273072353,.0453326263459352,.0311908939779879,.00777877686962257,-.013515206348962,-.0236032088979545,-.0199896425729313,-.00738860576070618,.0054854420782613,.0110546530747381,.00707029233633639,.0168079722487348,.032267375730295,.0345686129480019,.023784766697593,.00593174383418365,-.0103060858116714,-.01799874082957,-.0152431983931265,-.00563421697252118,.00418295032379259,.0084297790403517,.00539148553491986,.000493455983937391,.000947320081473168,.00101488083530666,.000698283842916597,.000174146794559467,-.000302570687259691,-.000528415101729961,-.000447516651629366,-.000165411867579771,.000122804930734204,.000247485231958582,.000158285649223294,-.0116098689755076,-.0222882331591425,-.0238777802016866,-.016428991010076,-.00409726811134704,.00711878292950237,.0124323755217502,.0105290235783908,.00389175564205705,-.00288931374181348,-.00582275057946566,-.00372409233650869,-.00929479849593364,-.0178438392786047,-.0191164220692824,-.0131529616098433,-.00328025075555346,.00569925922554383,.00995329280151445,.00842947949942716,.00311571858085949,-.00231316900116129,-.00466166271491862,-.00298148824255664},
			{-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0309875936214648,-.0507886397889319,-.0363634894844747,-.00203591827901426,.0244040104867404,.0250986750455939,.00523969112684981,-.0144235812974772,-.0174186443922174,-.00517215145746049,.00716762775243705,.00719323205880091,-.0290405684660784,-.0475974672027388,-.0340786838416493,-.00190799662907995,.0228706477193558,.0235216648305512,.00491046871109679,-.0135173129385977,-.0163241890040815,-.00484717272183615,.00671726843415898,.00674126396027401,-.025268891118137,-.0414156912131383,-.0296526754443501,-.00166019336468258,.0199002959496719,.0204667614621517,.0042727159196164,-.0117617363191147,-.0142040661159181,-.00421764057001676,.00584485544325101,.0058657345227164,-.019909698329518,-.0326319787562901,-.0233637408147432,-.00130808862585082,.0156797101690669,.0161260359462774,.00336653019752945,-.0092672298459836,-.0111915742601573,-.003323135590667,.00460524002065306,.00462169092748591,-.0133005471007852,-.021799585471186,-.0156079981732476,-.000873860271119,.0104747304644562,.0107728955558519,.00224898904631168,-.00619091384612107,-.00747645986973424,-.00221999955571561,.00307650125036322,.00308749117384471,-.00586741949230167,-.00961669559518964,-.00688533125920179,-.000385495780696793,.0046208353113631,.00475236822169549,.000992121757722516,-.00273106724863572,-.00329817458189373,-.000979333298653398,.00135717149585569,.00136201960403993},
			{.00997527550886585,.0163494681388942,.0117058404244761,.00065538634573533,-.00785594167460167,-.0080795624708938,-.00168671898857876,.00464312263238137,.00560726911956695,.00166497716447436,-.00230734475380579,-.00231558708502107,.0208283511742426,.0341376499933582,.0244417665390733,.00136844510727013,-.0164031872460608,-.0168701069287199,-.00352186517506783,.0096948288442119,.011707964381237,.00347646829883996,-.0048177302741764,-.00483494024185036,.0265497927509654,.043515087908134,.031155794841906,.00174434998167423,-.0209090589166153,-.0215042390488418,-.00448930353212625,.0123579487601599,.0149240823364804,.00443143636610296,-.00614113614848276,-.00616307360628593,.0257051468334037,.0421307139622166,.0301646151679837,.0016888558350856,-.0202438653529731,-.0208201106304678,-.00434648238332583,.0119647972554527,.014449292746328,.00429045618329429,-.0059457641685332,-.00596700371564648,.0185018287934532,.0303244817750634,.0217116264332397,.00121559008085388,-.0145709547315911,-.0149857195853513,-.00312847358668079,.00861191853147071,.0104001872586715,.00308814753183857,-.00427959083079407,-.00429487844875196,.00672153834378315,.0110165956718055,.00788762728301515,.000441612309247667,-.00529348920192007,-.00544416932653508,-.0011365446845897,.00312862805450344,.00377828906656089,.00112189461259403,-.00155473462575028,-.0015602884718828,-.00672153834378315,-.0110165956718055,-.00788762728301515,-.000441612309247667,.00529348920192007,.00544416932653508,.00113654468458969,-.00312862805450344,-.00377828906656089,-.00112189461259403,.00155473462575028,.0015602884718828,-.0185018287934532,-.0303244817750634,-.0217116264332397,-.00121559008085388,.0145709547315911,.0149857195853513,.00312847358668079,-.00861191853147071,-.0104001872586715,-.00308814753183857,.00427959083079407,.00429487844875196,-.0257051468334037,-.0421307139622166,-.0301646151679837,-.0016888558350856,.0202438653529731,.0208201106304678,.00434648238332583,-.0119647972554527,-.014449292746328,-.00429045618329429,.0059457641685332,.00596700371564648,-.0265497927509654,-.043515087908134,-.031155794841906,-.00174434998167423,.0209090589166153,.0215042390488418,.00448930353212624,-.0123579487601599,-.0149240823364804,-.00443143636610296,.00614113614848276,.00616307360628593,-.0208283511742426,-.0341376499933582,-.0244417665390733,-.00136844510727013,.0164031872460608,.0168701069287199,.00352186517506784,-.0096948288442119,-.011707964381237,-.00347646829883996,.0048177302741764,.00483494024185036,-.00997527550886584,-.0163494681388943,-.0117058404244761,-.00065538634573533,.00785594167460167,.0080795624708938,.00168671898857876,-.00464312263238137,-.00560726911956695,-.00166497716447436,.00230734475380579,.00231558708502107},
			{-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962,-.0215961371406428,-.0353960505441994,-.0253427521805069,-.00141888947227826,.017007850422056,.0174919819510481,.00365169007740165,-.0100522048780116,-.0121395497079787,-.0036046197569249,.0049953240579315,.00501316842875873,-.0173236351861994,-.0283934234472885,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064034,-.00806351288610232,-.00973790493624807,-.00289149477275801,.00400706714598473,.00402138125078621,-.000983918170479597,-.00161264105089006,-.00115461363289064,-6.46444836215388e-005,.000774876218005063,.000796933209270456,.000166370689198568,-.000457977598884836,-.000553076851722798,-.000164226169402932,.000227586539019653,.000228399527036137,.019289441206072,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433103,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543011,-.0076638869878077,.0330151194341706,.0541117528845208,.0387427614754033,.00216912890884522,-.0260007708483173,-.026740887478779,-.00558252539594694,.0153673197416279,.0185583505455568,.00551056658950743,-.00763660738543012,-.0076638869878077,.0192894412060721,.0316153778545216,.0226358781203494,.00126733706473451,-.0151912320532778,-.0156236532128436,-.00326165094211563,.00897852304433104,.0108429173622702,.00321960822987157,-.00446177059782641,-.00447770900105517,-.00098391817047955,-.00161264105088998,-.00115461363289059,-6.46444836215357e-005,.000774876218005026,.000796933209270418,.00016637068919856,-.000457977598884814,-.000553076851722771,-.000164226169402924,.000227586539019642,.000228399527036126,-.0173236351861993,-.0283934234472884,-.0203290334067721,-.00113818149177376,.0136430785790229,.0140314312707926,.00292925287064033,-.0080635128861023,-.00973790493624805,-.002891494772758,.00400706714598472,.0040213812507862,-.0215961371406427,-.0353960505441994,-.0253427521805068,-.00141888947227826,.0170078504220559,.017491981951048,.00365169007740164,-.0100522048780115,-.0121395497079786,-.00360461975692489,.00499532405793149,.00501316842875872,-.0124008696089065,-.0203250148214152,-.0145522397489249,-.000814750490820839,.00976619726195273,.0100441938279818,.00209686261052553,-.00577214717439624,-.00697073611169898,-.0020698340311606,.0028684001167997,.00287864665831962},
			{-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0414838298697222,-.0542596836678707,-.0148939790510807,.0268456372262698,.0261559940473834,-.0055759508977495,-.0233383661040445,-.00857245981277814,.0130869087160192,.0130044018098243,-.00242202207908883,-.00733386119927366,-.0388773009057509,-.0508504170331271,-.0139581544681687,.0251588611690376,.0245125499324097,-.00522560047055151,-.0218719603403289,-.00803383151195207,.0122646267154354,.0121873039169134,-.00226984059728378,-.00687305703307963,-.0338280665787232,-.0442461604297742,-.0121453230462166,.0218913250365373,.0213289542177341,-.0045469195780892,-.0190313142466737,-.00699042837174835,.0106717441650267,.0106044637542081,-.00197504242987831,-.00598041081807679,-.0266535875082312,-.0348621434267833,-.00956946297461998,.0172484686990498,.0168053691859232,-.00358257893886599,-.014995027826669,-.00550785230048017,.00840841040400752,.0083553992656627,-.00155616242845937,-.00471204591914336,-.0178057592933257,-.0232894327682907,-.00639281876182168,.0115227296039941,.0112267197977613,-.00239331903125935,-.0100173327885387,-.00367948563229073,.00561718499046395,.00558177123728939,-.00103958439417987,-.00314785225026289,-.00785485426736497,-.0102739286402175,-.00282013582822094,.00508315092384475,.0049525687986252,-.00105579166247069,-.00441905834541759,-.00162317275800423,.00247797180485658,.00246234933879579,-.00045860352150052,-.00138864736255758},
			{.0133541389835151,.0174668385049753,.00479454927113329,-.00864193040388231,-.00841992604968446,.0017949650137812,.00751289805161027,.00275957692744254,-.00421283180475702,-.00418627184884443,.000779677757981102,.00236085727977567,.0278834099500396,.0364707166195125,.0100110073002612,-.0180443298147953,-.0175807852593257,.00374788261429831,.0156869129896357,.00576198996068088,-.00879638263519743,-.00874092551138884,.00162796527590073,.00492946429917962,.0355428401014885,.0464890360080293,.0127609798215341,-.0230010149581752,-.0224101370833209,.00477740680633519,.0199960277841051,.00734477914307506,-.0112127039710897,-.0111420130589112,.00207515901375732,.00628356293888253,.0344120924880564,.0450100499065603,.0123550064261315,-.0222692686290634,-.0216971887384138,.00462542003968309,.0193598810769079,.00711111488148247,-.010855986887727,-.010787544911762,.00200914062311667,.00608365984231783,.0247688390097475,.0323969453570202,.00889277992146099,-.0160287820256653,-.0156170152981078,.00332924492618789,.0139346881567349,.00511837691184837,-.00781382857224871,-.00776456599733339,.00144612190203756,.00437884418905171,.00899828352070364,.0117695019702978,.00323066232490896,-.00582310398568677,-.00567351305179212,.00120948300176336,.00506233960976318,.00185945762740475,-.0028386895666607,-.00282079294195926,.00052536232622418,.00159079242634532,-.00899828352070364,-.0117695019702978,-.00323066232490896,.00582310398568677,.00567351305179212,-.00120948300176336,-.00506233960976318,-.00185945762740475,.0028386895666607,.00282079294195926,-.00052536232622418,-.00159079242634532,-.0247688390097475,-.0323969453570202,-.00889277992146099,.0160287820256653,.0156170152981078,-.00332924492618789,-.0139346881567349,-.00511837691184837,.00781382857224871,.00776456599733339,-.00144612190203756,-.00437884418905171,-.0344120924880564,-.0450100499065603,-.0123550064261315,.0222692686290634,.0216971887384138,-.00462542003968309,-.0193598810769079,-.00711111488148247,.010855986887727,.010787544911762,-.00200914062311667,-.00608365984231783,-.0355428401014885,-.0464890360080293,-.0127609798215341,.0230010149581752,.0224101370833208,-.00477740680633519,-.0199960277841051,-.00734477914307506,.0112127039710897,.0111420130589112,-.00207515901375732,-.00628356293888253,-.0278834099500396,-.0364707166195125,-.0100110073002612,.0180443298147953,.0175807852593257,-.00374788261429831,-.0156869129896357,-.00576198996068088,.00879638263519743,.00874092551138884,-.00162796527590073,-.00492946429917962,-.0133541389835151,-.0174668385049753,-.00479454927113329,.00864193040388231,.00841992604968446,-.0017949650137812,-.00751289805161027,-.00275957692744254,.00421283180475702,.00418627184884443,-.000779677757981102,-.00236085727977568},
			{-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292215,-.00120362482446452,.00306032152371413,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0516941832706635,-.048938878193706,.0129572628308372,.0342412839072111,-.00450113803238466,-.0265624900371196,.000167450601247885,.0205526352721649,.00221197642247413,-.0146822029633365,-.00280420321439346,.00712993225085284,-.0484461132060871,-.0458639305845591,.0121431267935323,.0320898215531325,-.00421832068670833,-.0248935047998746,.000156929276588187,.0192612636873552,.00207299261526764,-.0137596847821795,-.00262800837119474,.0066819414317466,-.0421541183373008,-.0399072996640798,.0105660241857151,.0279221189410846,-.00367046141876753,-.0216604321320916,.000136547905664685,.0167596848347518,.001803760306727,-.0119726298397675,-.00228669275075799,.0058141165760511,-.0332137953944511,-.0314434968176033,.00832511221392646,.0220002121279829,-.00289200579622364,-.0170665450818856,.000107587926854457,.013205180535928,.00142120694563906,-.00943339568033769,-.00180171589750568,.00458102045478133,-.0221882643688671,-.0210056276852712,.00556153816537164,.0146971015227122,-.00193198604377585,-.0114011966907983,7.18734289651178e-005,.00882163671119439,.000949428243842857,-.00630191987292214,-.00120362482446452,.00306032152371413,-.00978815787589279,-.0092664480938018,.002453423696891,.00648349720513435,-.000852278668404253,-.00502953774700035,3.17063316939494e-005,.00389158752651689,.000418831927908926,-.00278003658203833,-.000530968516029268,.00135003395160268},
			{1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394712,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,9.38463500544579e-005,.00110620145747811,.00397374980154702,.00904536355795187,.015924369058549,.0235127977673766,.0302450800394711,.0344557717846379,.0347915983448399,.030565980615856,.0219678453543708,.0100850666679019,8.79497578075961e-005,.00103669615509983,.00372406952887607,.00847702157560611,.0149238025892903,.0220354320420534,.0283447088011335,.0322908326405259,.0326055584089547,.0286454464212646,.0205875527108991,.00945139763454551,7.65271815013742e-005,.000902054045409203,.00324040170051977,.00737605861435863,.0129855565030636,.0191735548724589,.0246634070303329,.0280970234814972,.0283708738774972,.0249250860049403,.0179137205405699,.00822388645802562,6.02967930241717e-005,.000710740484695346,.00255315597436325,.0058116955424022,.0102314941881775,.0151071011240822,.0194326292933208,.0221380217619632,.0223537921630579,.0196388096682036,.0141144607515462,.00647971047522447,4.02808883575737e-005,.000474805651832001,.00170561957949523,.00388246617391187,.00683508449545443,.010092202640724,.012981844171125,.0147891643705445,.0149333084120673,.0131195816567463,.0094290755651378,.00432872929307012,1.77695599831757e-005,.000209456341570312,.000752419091586911,.00171271584051363,.00301523747077974,.00445208652288796,.0057268264950684,.00652410992157336,.00658769779901337,.00578758817666173,.00415955383985738,.0019095808945752},
			{-3.02102579557316e-005,-.000356099425944935,-.00127919739538736,-.00291181027531203,-.00512624408687041,-.00756904967962207,-.00973625153619673,-.0110917233656258,-.0111998299356331,-.00983955325421882,-.00707171109480605,-.00324650309107673,-6.30789456599155e-005,-.00074353474146506,-.00267096107256627,-.00607985282341967,-.0107035852745446,-.0158041574533295,-.020329269696504,-.023159491603162,-.0233852178602818,-.0205449634342814,-.0147657156892056,-.00677869061452103,-8.04064095237273e-005,-.000947779933413657,-.00340466042315561,-.00774995730904432,-.0136438054243522,-.0201454787025301,-.0259136161429872,-.0295212855370935,-.0298090176429547,-.0261885598477368,-.0188217822951288,-.00864076226835519,-7.78483878402044e-005,-.000917627590643462,-.00329634573482353,-.0075034028495138,-.0132097461207719,-.0195045774155398,-.0250892093278478,-.0285821055764826,-.0288606838727134,-.0253554060687339,-.0182229926275056,-.00836586804816125,-5.60330408925769e-005,-.000660482069535263,-.00237261529082107,-.00540073456066038,-.00950799708383307,-.0140388364388575,-.0180584946102595,-.0205725813339951,-.0207730940163711,-.0182500954035624,-.0131163883981527,-.00602151231964795,-2.03562705656154e-005,-.000239946850947697,-.000861948558184859,-.00196203547440233,-.00345416486580917,-.00510017568961999,-.00656047925364063,-.0074738230372111,-.00754666738671032,-.00663008599864202,-.0047650590941222,-.00218755812714027,2.03562705656154e-005,.000239946850947697,.000861948558184859,.00196203547440233,.00345416486580917,.00510017568962,.00656047925364063,.0074738230372111,.00754666738671032,.00663008599864202,.0047650590941222,.00218755812714027,5.60330408925769e-005,.000660482069535263,.00237261529082107,.00540073456066038,.00950799708383307,.0140388364388575,.0180584946102595,.0205725813339951,.0207730940163711,.0182500954035624,.0131163883981527,.00602151231964795,7.78483878402044e-005,.000917627590643462,.00329634573482353,.00750340284951379,.0132097461207719,.0195045774155398,.0250892093278478,.0285821055764826,.0288606838727134,.0253554060687339,.0182229926275056,.00836586804816125,8.04064095237273e-005,.000947779933413657,.00340466042315561,.00774995730904432,.0136438054243522,.0201454787025301,.0259136161429872,.0295212855370935,.0298090176429547,.0261885598477368,.0188217822951288,.00864076226835519,6.30789456599155e-005,.00074353474146506,.00267096107256627,.00607985282341966,.0107035852745446,.0158041574533295,.020329269696504,.023159491603162,.0233852178602818,.0205449634342814,.0147657156892056,.00677869061452103,3.02102579557315e-005,.000356099425944935,.00127919739538736,.00291181027531203,.00512624408687041,.00756904967962207,.00973625153619672,.0110917233656258,.0111998299356331,.00983955325421882,.00707171109480605,.00324650309107673},
			{3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491522,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500883,.00403592477071172,6.54041959386344e-005,.00077094332204201,.00276941948707301,.00630397165299627,.0110981466354321,.0163867388699784,.021078658252914,.0240132093332518,.0242472564339873,.0213023030101423,.015310017502792,.00702857038146807,5.24648654853783e-005,.000618422673152875,.0022215275148126,.00505681661629785,.00890252929508159,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327876,2.97980960162188e-006,3.51241121514965e-005,.000126174516176939,.000287208412095994,.000505630616352145,.000746578428543733,.000960341876387736,.00109403977390252,.00110470293989661,.000970531112496787,.000697523094674574,.000320221068507963,-5.84183358335144e-005,-.000688598418638334,-.0024736161852904,-.00563064078419707,-.00991274715595561,-.0146364617863459,-.0188272345384989,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218605,-9.99867395890061e-005,-.00117858048818797,-.00423375321862369,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.032565920170009,-.0234052068248199,-.0107449350355781,-9.99867395890061e-005,-.00117858048818797,-.0042337532186237,-.00963720389798855,-.0169663044034469,-.0250512458503864,-.0322240229907758,-.0367102213220828,-.0370680210958888,-.0325659201700091,-.0234052068248199,-.0107449350355781,-5.84183358335145e-005,-.000688598418638335,-.0024736161852904,-.00563064078419708,-.00991274715595562,-.0146364617863459,-.018827234538499,-.021448344515795,-.0216573929099447,-.0190269916694853,-.0136747456529279,-.00627784470218606,2.97980960162174e-006,3.51241121514949e-005,.000126174516176933,.00028720841209598,.000505630616352121,.000746578428543697,.00096034187638769,.00109403977390247,.00110470293989656,.000970531112496741,.000697523094674541,.000320221068507948,5.24648654853782e-005,.000618422673152874,.0022215275148126,.00505681661629784,.00890252929508158,.0131448454983542,.0169085324569847,.0192625225256701,.0194502665913391,.0170879321413032,.0122811388066968,.00563806334327875,6.54041959386343e-005,.000770943322042009,.00276941948707301,.00630397165299626,.0110981466354321,.0163867388699784,.0210786582529139,.0240132093332518,.0242472564339873,.0213023030101423,.0153100175027919,.00702857038146806,3.7556202779616e-005,.000442688780416571,.00159024781737124,.00361984784491523,.00637274473810929,.00940954443465593,.0121037244217697,.0137887936112712,.013923187441041,.0122321450488033,.00879127269500884,.00403592477071172},
			{-4.19315566304071e-005,-.000494262686103036,-.00177551406892167,-.00404156553827383,-.00711517904100237,-.0105057704487168,-.0135138264378992,-.0153952087107654,-.0155452596229099,-.0136572082602461,-.00981546912576032,-.00450611604885183,-5.23755171918012e-005,-.000617369491942089,-.00221774422687797,-.00504820479710383,-.008887368181188,-.0131224596692141,-.0168797370239278,-.0192297182193646,-.0194171425546437,-.0170588311884377,-.0122602238803855,-.00562846165395503,2.22612437958035e-006,2.62401471318847e-005,9.4261111981906e-005,.000214564597630484,.000377741128665099,.00055774585065616,.000717441967632789,.000817323567112939,.000825289691448544,.000725054033484014,.000521098114970246,.000239227340925692,7.58256825576655e-005,.000893785219253475,.00321069802758278,.00730844476493686,.0128665222724588,.0189977973412616,.0244373258701412,.0278394675137098,.0281108076152006,.0246966061215671,.0177495114871256,.00814850085580817,9.94363684763003e-005,.00117209306138154,.00421044877365996,.00958415647735547,.0168729143812158,.0249133527446604,.0320466477561979,.0365081520733029,.0368639823593503,.0323866629825667,.0232763742438573,.0106857901742116,4.54228976457494e-005,.000535416407238526,.00192334843497638,.00437807781360722,.00770760914409425,.0113805108641198,.0146390261754792,.0166770577030509,.0168396022821627,.0147943463773532,.0106327330839207,.00488130812483214,-4.54228976457494e-005,-.000535416407238526,-.00192334843497639,-.00437807781360723,-.00770760914409426,-.0113805108641198,-.0146390261754792,-.0166770577030509,-.0168396022821627,-.0147943463773532,-.0106327330839207,-.00488130812483215,-9.94363684763008e-005,-.00117209306138155,-.00421044877365998,-.00958415647735551,-.0168729143812159,-.0249133527446605,-.032046647756198,-.036508152073303,-.0368639823593504,-.0323866629825669,-.0232763742438574,-.0106857901742116,-7.58256825576654e-005,-.000893785219253473,-.00321069802758277,-.00730844476493685,-.0128665222724588,-.0189977973412615,-.0244373258701411,-.0278394675137098,-.0281108076152005,-.0246966061215671,-.0177495114871256,-.00814850085580816,-2.22612437958026e-006,-2.62401471318837e-005,-9.42611119819025e-005,-.000214564597630476,-.000377741128665085,-.000557745850656139,-.000717441967632762,-.000817323567112908,-.000825289691448512,-.000725054033483986,-.000521098114970227,-.000239227340925682,5.23755171918011e-005,.000617369491942089,.00221774422687796,.00504820479710383,.008887368181188,.0131224596692141,.0168797370239278,.0192297182193646,.0194171425546437,.0170588311884377,.0122602238803855,.00562846165395503,4.19315566304069e-005,.000494262686103033,.00177551406892166,.00404156553827381,.00711517904100233,.0105057704487168,.0135138264378992,.0153952087107653,.0155452596229098,.013657208260246,.00981546912576027,.00450611604885181},
			{4.38808991888882e-005,.000517240304071817,.0018580553675488,.00422945256989752,.00744595429549842,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731,2.81527019748646e-005,.000331846256550951,.0011920739999021,.00271349311244975,.00477710657881669,.00705353791336682,.00907313629206733,.0103362898376332,.0104370335006432,.00916940235127242,.00659007199462349,.00302539071720239,-5.61926656806376e-005,-.00066236362564373,-.0023793743066966,-.00541611996712565,-.00953508310299627,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136763,-.0183020855787247,-.0131537538647564,-.00603866617409735,-9.13673651610601e-005,-.00107698075043354,-.00386878177964248,-.00880642704521119,-.0155037211557789,-.0228916988068627,-.0294461454354727,-.0335456102525707,-.0338725658341033,-.0297585693083877,-.0213875568642763,-.00981866602574447,-1.37955930732719e-005,-.000162613731440516,-.000584148826303059,-.00132968575520406,-.00234091272972111,-.0034564262736338,-.0044460846527431,-.00506506439824835,-.00511443154534631,-.00449325765165264,-.00322931531199247,-.00148252410228482,8.93220195809498e-005,.00105287150952592,.00378217541095215,.00860928677962801,.0151566555762327,.0223792463037912,.0287869653955147,.0327946597841923,.033114296153308,.029092395362165,.0209087760126605,.00959886582549863,8.93220195809508e-005,.00105287150952593,.00378217541095219,.00860928677962811,.0151566555762329,.0223792463037915,.028786965395515,.0327946597841926,.0331142961533084,.0290923953621653,.0209087760126607,.00959886582549874,-1.37955930732714e-005,-.00016261373144051,-.00058414882630304,-.00132968575520402,-.00234091272972103,-.00345642627363368,-.00444608465274295,-.00506506439824818,-.00511443154534613,-.00449325765165249,-.00322931531199236,-.00148252410228477,-9.13673651610609e-005,-.00107698075043355,-.00386878177964252,-.00880642704521126,-.0155037211557791,-.0228916988068629,-.029446145435473,-.033545610252571,-.0338725658341036,-.029758569308388,-.0213875568642765,-.00981866602574456,-5.61926656806377e-005,-.000662363625643731,-.0023793743066966,-.00541611996712565,-.00953508310299628,-.0140788297402289,-.0181099389603954,-.0206311877184246,-.0208322716136764,-.0183020855787247,-.0131537538647564,-.00603866617409736,2.81527019748638e-005,.000331846256550942,.00119207399990207,.00271349311244967,.00477710657881656,.00705353791336663,.00907313629206708,.0103362898376329,.010437033500643,.00916940235127217,.00659007199462331,.0030253907172023,4.38808991888881e-005,.000517240304071817,.0018580553675488,.00422945256989751,.00744595429549841,.0109941698092706,.0141420663393061,.0161109115834517,.0162679381638635,.0142921137927641,.0102717772916352,.00471559941873731},
			{8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302079,-.00878002632372056,-.00505719415206449,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.00045445289149014,.00509449990124688,.0166677075081626,.0328835502828953,.047189744356069,.0522116897303996,.0439663615491188,.0244935039192365,.00135098128617801,-.0159009907838248,-.0204556914571699,-.0117822429454629,.000425898521554925,.00477440019995963,.0156204352934354,.0308173976031705,.0442246990395499,.048931103488301,.0412038491393894,.0229545180639013,.00126609587749552,-.0149018932278846,-.019170411085791,-.0110419362381424,.000370584459499054,.00415431946341246,.0135917132307096,.0268149525202462,.0384809652079676,.0425761199468399,.0358524516752331,.019973273536417,.00110166021408729,-.0129664926452581,-.0166806318196771,-.0096078520246925,.000291988467548458,.00327324404123221,.0107090932063305,.0211278608508173,.0303196687633597,.0335462961243491,.0282486276857305,.0157372101887599,.000868012862992909,-.0102164735188442,-.0131428936047439,-.00757015551303623,.000195061035141679,.0021866698240338,.00715414147618305,.0141143328107841,.0202548615148722,.0224103893627198,.0188713157200156,.0105131429862114,.000579870461985248,-.00682505003302078,-.00878002632372056,-.00505719415206449,8.60494618083241e-005,.000964630180363513,.0031559866545311,.00622641390818722,.00893525419410878,.00988614636529727,.00832491512281191,.00463778065783548,.000255804759449434,-.00301081085584189,-.00387323146967982,-.00223093676668575},
			{-.000146293799182738,-.00163998020354913,-.00536553359148041,-.0105856065426876,-.0151909408292571,-.0168075648663272,-.0141532955069821,-.00788475067656189,-.000434896968818276,.0051187183448384,.00658493074687408,.00379284435342441,-.000305461099423673,-.00342427470479066,-.0112032211823346,-.0221027208990368,-.0317186477684445,-.0350941548540275,-.0295520468412377,-.0164633403725859,-.000908063820704495,.0106878715433549,.0137493194981995,.00791944985099954,-.00038936970167901,-.00436490545862284,-.014280688761482,-.0281742253235845,-.0404315981398701,-.0447343397636272,-.0376698430153202,-.0209857357994469,-.00115750430951867,.0136237752115456,.0175261872670667,.0100948822346398,-.000376982428752617,-.00422604186707806,-.0138263678718428,-.0272779002703818,-.0391453212702289,-.0433111769611684,-.0364714276673539,-.0203181028639926,-.00112067986803374,.0131903531422229,.0169686152112542,.00977372714511676,-.000271341159812294,-.00304178395110495,-.00995182376735971,-.0196338516919752,-.0281756815823874,-.0311741452469417,-.0262510895163342,-.0146243887667217,-.000806633285738924,.00949403857306351,.0122135234447537,.00703484899288287,-9.85756613730145e-005,-.00110505116489299,-.00361540287663619,-.0071327914908855,-.0102359570090312,-.0113252703260409,-.00953677102517345,-.00531290127842359,-.000293042197071052,.00344909387167086,.00443705684789683,.00255569369797207,9.85756613730145e-005,.00110505116489299,.00361540287663619,.0071327914908855,.0102359570090312,.0113252703260409,.00953677102517345,.00531290127842359,.000293042197071052,-.00344909387167086,-.00443705684789683,-.00255569369797207,.000271341159812294,.00304178395110495,.00995182376735971,.0196338516919752,.0281756815823874,.0311741452469417,.0262510895163341,.0146243887667217,.000806633285738924,-.00949403857306351,-.0122135234447537,-.00703484899288287,.000376982428752617,.00422604186707806,.0138263678718428,.0272779002703818,.0391453212702289,.0433111769611684,.0364714276673539,.0203181028639926,.00112067986803374,-.0131903531422229,-.0169686152112542,-.00977372714511676,.00038936970167901,.00436490545862284,.014280688761482,.0281742253235845,.0404315981398701,.0447343397636272,.0376698430153202,.0209857357994469,.00115750430951867,-.0136237752115456,-.0175261872670667,-.0100948822346398,.000305461099423673,.00342427470479066,.0112032211823346,.0221027208990368,.0317186477684445,.0350941548540275,.0295520468412377,.0164633403725859,.000908063820704495,-.0106878715433549,-.0137493194981995,-.00791944985099954,.000146293799182738,.00163998020354913,.00536553359148041,.0105856065426876,.0151909408292571,.0168075648663272,.0141532955069821,.00788475067656189,.000434896968818276,-.0051187183448384,-.00658493074687408,-.00379284435342441},
			{.000181866688959546,.00203875879391254,.00667022002460173,.013159609110584,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.0081861265388698,-.00471511470855014,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053052,.0306414104040703,.0170702209457237,.000941537361279075,-.0110818536551902,-.0142561543633291,-.00821138090239868,.000254062196557116,.00284808361770293,.00931809316299003,.0183835710409382,.026381458519847,.0291889804801732,.0245794241800656,.0136931099430282,.000755267002369574,-.00888945965934217,-.0114357681533554,-.0065868708927751,1.44297896450536e-005,.000161760576944934,.000529233101409994,.00104411859237783,.00149836891174606,.00165782573712333,.00139602005068532,.000777717813756163,4.28963620630315e-005,-.000504888309558718,-.000649509179715462,-.000374109815194419,-.000282891999889151,-.00317127097766542,-.0103754672901093,-.0204696536806732,-.0293751044500413,-.0325012110209999,-.0273685835856313,-.0152469407451353,-.000840971209593475,.00989819443838388,.0127334462466717,.00733431854530109,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.0468430913146029,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000484187855082692,-.00542783427302108,-.0177582796779235,-.035035129002638,-.0502773808452297,-.0556279133309325,-.046843091314603,-.0260961198579412,-.00143937632141918,.0169413964912067,.0217941123411145,.0125531579766626,-.000282891999889152,-.00317127097766543,-.0103754672901093,-.0204696536806732,-.0293751044500414,-.0325012110209999,-.0273685835856313,-.0152469407451354,-.000840971209593476,.00989819443838389,.0127334462466717,.00733431854530111,1.44297896450529e-005,.000161760576944927,.000529233101409968,.00104411859237778,.00149836891174599,.00165782573712325,.00139602005068526,.000777717813756126,4.28963620630294e-005,-.000504888309558694,-.000649509179715431,-.000374109815194401,.000254062196557116,.00284808361770292,.00931809316299001,.0183835710409382,.026381458519847,.0291889804801731,.0245794241800655,.0136931099430282,.000755267002369572,-.00888945965934215,-.0114357681533554,-.00658687089277509,.000316721171978464,.00355050217433172,.0116162003917936,.0229174833727234,.0328878777485841,.0363878146053051,.0306414104040702,.0170702209457237,.000941537361279073,-.0110818536551902,-.0142561543633291,-.00821138090239866,.000181866688959546,.00203875879391254,.00667022002460174,.0131596091105841,.018884779301865,.0208945026295579,.0175948195077333,.00980201147846774,.000540646782019285,-.00636338903147506,-.00818612653886981,-.00471511470855015},
			{-.000203054430513827,-.00227627724582951,-.00744731064410978,-.0146927232745112,-.0210848843647913,-.0233287434691147,-.0196446423233178,-.010943960491264,-.000603632941579964,.00710473338089701,.00913982253690271,.00526443263155787,-.000253629525609746,-.00284323329737286,-.00930222433932735,-.0183522636004527,-.0263365305817064,-.0291392712986882,-.0245375651278731,-.0136697904136697,-.00075398077366707,.0088743208036425,.0114162929039563,.00657565337317555,1.07800533648885e-005,.000120846366765943,.000395373821517966,.000780028983233564,.00111938546758022,.00123851077219281,.0010429237719346,.000581009130510076,3.20465567117658e-005,-.000377186574041089,-.000485228391442131,-.000279485971128269,.000367187436559508,.00411623821592978,.0134671225727665,.0265691488808889,.0381282231590321,.0421858389935269,.0355238043246551,.0197901852642516,.00109156166590775,-.0128476331738681,-.0165277260852754,-.00951978008079133,.000481522671607105,.00539795707957059,.0176605302753814,.0348422802025231,.0500006319693174,.0553217127646963,.046585246282749,.0259524753061578,.00143145335939787,-.0168481435739603,-.021674147935839,-.0124840598593624,.000219961321613734,.00246580658242685,.00806739497188598,.0159160813255256,.0228404719819027,.0252711611958884,.0212803112797474,.011855185859552,.000653893142166497,-.00769629374853215,-.00990083023278439,-.00570276430932191,-.000219961321613735,-.00246580658242685,-.00806739497188599,-.0159160813255256,-.0228404719819027,-.0252711611958885,-.0212803112797474,-.0118551858595521,-.000653893142166498,.00769629374853216,.0099008302327844,.00570276430932192,-.000481522671607107,-.00539795707957062,-.0176605302753815,-.0348422802025233,-.0500006319693176,-.0553217127646966,-.0465852462827492,-.0259524753061579,-.00143145335939788,.0168481435739604,.0216741479358391,.0124840598593625,-.000367187436559507,-.00411623821592978,-.0134671225727665,-.0265691488808889,-.038128223159032,-.0421858389935268,-.035523804324655,-.0197901852642516,-.00109156166590775,.0128476331738681,.0165277260852754,.00951978008079131,-1.07800533648881e-005,-.000120846366765938,-.00039537382151795,-.000780028983233534,-.00111938546758017,-.00123851077219277,-.00104292377193456,-.000581009130510054,-3.20465567117645e-005,.000377186574041075,.000485228391442112,.000279485971128258,.000253629525609746,.00284323329737286,.00930222433932735,.0183522636004527,.0263365305817064,.0291392712986881,.0245375651278731,.0136697904136697,.00075398077366707,-.00887432080364249,-.0114162929039563,-.00657565337317555,.000203054430513826,.0022762772458295,.00744731064410975,.0146927232745111,.0210848843647912,.0233287434691146,.0196446423233177,.0109439604912639,.000603632941579961,-.00710473338089698,-.00913982253690267,-.00526443263155784},
			{.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048,.000554003280665635,.00578143006001033,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204796,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00129071596800793,.0134695666195525,.0383630938779612,.060172367677722,.0589037642240113,.0297901804457638,-.00988965082866861,-.0338617749353493,-.0286351790597393,-.0044361037099508,.014693847184694,.0127099554020962,.00120961717444335,.0126232413010787,.0359526482741082,.0563915928631674,.0552026988204797,.0279183916447378,-.00926826024324192,-.0317341580441981,-.0268359617781597,-.00415737262741865,.0137705973690585,.011911358286295,.00105251674778132,.0109837832674493,.0312832565832275,.0490676696533039,.0480331845966229,.0242924583066762,-.00806453424679342,-.0276126476408772,-.0233506102683525,-.0036174290590239,.0119821251418304,.0103643568808608,.000829292066562059,.00865427019924752,.0246484975705927,.0386610752312962,.0378459905760725,.0191403538171554,-.00635415473006169,-.0217563755385589,-.0183982591115506,-.00285022088847831,.00944087715622579,.00816621583878259,.000554003280665635,.00578143006001032,.0164662717372871,.0258272849528053,.025282772842752,.0127865914017056,-.00424485257757836,-.014534208042919,-.0122908397623899,-.00190407190241768,.00630691782521586,.00545538845447363,.000244393679684404,.00255042707419764,.00726395110101834,.0113934798333501,.0111532731001901,.00564069245136809,-.00187257941849324,-.00641163817773756,-.00542199597145488,-.000839964590203203,.00278224138477972,.00240659668458048},
			{-.000415496844730152,-.0043360139403437,-.0123495368891892,-.0193702019109772,-.0189618233479971,-.00958981393734538,.00318359640435585,.0109005087031771,.00921800523314784,.00142803462579379,-.00473012443753221,-.00409148604117949,-.000867556408455861,-.00905358663684252,-.0257858032029976,-.0404449348150074,-.0395922413616564,-.0200235083437195,.00664734160455473,.0227602358497139,.0192471726671977,.00298173285024315,-.00987648840327513,-.00854301297387297,-.00110586971823095,-.0115405605969585,-.0328690315056423,-.051554951622511,-.0504680276371612,-.0255238637098843,.00847333235691014,.0290123562694813,.0245342725922579,.00380079961925163,-.0125895092713119,-.0108897349592236,-.00107068796176225,-.0111734131963732,-.0318233475134598,-.0499148002350218,-.0488624552731521,-.0247118563437564,.00820376469397058,.0280893672084472,.0237537477354436,.00367988229564073,-.0121889909806472,-.0105432927002232,-.000770650542793007,-.00804230294180483,-.02290553448866,-.0359272442282916,-.0351697964516905,-.0177868867353227,.0059048349660612,.0202179223629944,.0170972675881759,.00264867392725914,-.00877328675655798,-.00758876024801777,-.000279970008957894,-.00292169213103891,-.0083213627187413,-.0130520259571515,-.0127768523875183,-.00646180669720086,.00214517035484276,.00734497880785625,.00621127462321504,.000962238033926173,-.00318725159515452,-.00275692438613841,.000279970008957894,.00292169213103891,.0083213627187413,.0130520259571515,.0127768523875183,.00646180669720086,-.00214517035484276,-.00734497880785625,-.00621127462321504,-.000962238033926173,.00318725159515452,.00275692438613841,.000770650542793007,.00804230294180483,.02290553448866,.0359272442282916,.0351697964516905,.0177868867353227,-.0059048349660612,-.0202179223629944,-.0170972675881759,-.00264867392725914,.00877328675655798,.00758876024801777,.00107068796176225,.0111734131963732,.0318233475134598,.0499148002350218,.0488624552731521,.0247118563437564,-.00820376469397058,-.0280893672084472,-.0237537477354436,-.00367988229564073,.0121889909806472,.0105432927002232,.00110586971823095,.0115405605969585,.0328690315056423,.051554951622511,.0504680276371612,.0255238637098843,-.00847333235691015,-.0290123562694813,-.0245342725922579,-.00380079961925163,.0125895092713119,.0108897349592236,.000867556408455861,.00905358663684252,.0257858032029976,.0404449348150074,.0395922413616564,.0200235083437195,-.00664734160455473,-.0227602358497139,-.0192471726671977,-.00298173285024315,.00987648840327513,.00854301297387297,.000415496844730152,.0043360139403437,.0123495368891892,.0193702019109772,.0189618233479971,.00958981393734538,-.00318359640435585,-.0109005087031771,-.00921800523314784,-.00142803462579379,.00473012443753221,.00409148604117949},
			{.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703812,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.00588030439408339,.00508637429395636,.000899536743208372,.00938732485666896,.026736333462767,.0419358379330301,.041051712031237,.020761625765838,-.00689237951523805,-.0235992360051222,-.0199566723826478,-.0030916471033844,.0102405608742263,.00885793015048143,.000721575635236956,.00753017033241687,.0214469135896479,.0336394028639592,.0329301892448019,.0166542205347873,-.00552881598728651,-.0189304481891057,-.0160085162284352,-.00248000678059721,.00821460521072995,.00710550916733107,4.09828174776266e-005,.000427685721688117,.00121810507752847,.00191059320784475,.00187031250699067,.000945897903253332,-.000314016224231211,-.00107517918429856,-.000909224295609892,-.000140855178957249,.000466559082044881,.000403566543921213,-.000803456701900855,-.00838465924372367,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752584,.00615620046137393,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733643,-.00791180949317304,-.00137516782835013,-.0143508836460861,-.0408732004597185,-.0641094603592521,-.0627578518723987,-.0317393592123806,.0105367330925572,.0360773591216924,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.00137516782835013,-.0143508836460862,-.0408732004597186,-.0641094603592521,-.0627578518723987,-.0317393592123807,.0105367330925572,.0360773591216925,.0305087858041877,.00472635905679869,-.0156552691869696,-.0135415708815498,-.000803456701900856,-.00838465924372368,-.0238806101775214,-.0374566467590265,-.0366669548576305,-.0185440644752585,.00615620046137394,.0210785879189631,.0178250886298303,.00276142648299933,-.00914676062733645,-.00791180949317305,4.09828174776246e-005,.000427685721688097,.00121810507752841,.00191059320784466,.00187031250699058,.000945897903253286,-.000314016224231196,-.00107517918429851,-.000909224295609848,-.000140855178957242,.000466559082044858,.000403566543921193,.000721575635236955,.00753017033241686,.0214469135896478,.0336394028639591,.0329301892448018,.0166542205347873,-.0055288159872865,-.0189304481891056,-.0160085162284351,-.00248000678059721,.00821460521072994,.00710550916733106,.00089953674320837,.00938732485666894,.026736333462767,.04193583793303,.0410517120312369,.020761625765838,-.00689237951523804,-.0235992360051222,-.0199566723826478,-.00309164710338439,.0102405608742263,.00885793015048141,.000516529312084902,.00539036174691256,.0153524578461789,.024080272076484,.0235725919319012,.0119216789703813,-.00395772165674538,-.0135510830785834,-.0114594610338502,-.00177527640041103,.0058803043940834,.00508637429395636},
			{.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00282614566821761,.026835845753926,.062928864644162,.0679214968761902,.0237487430487256,-.0302881490742531,-.0409328219869992,-.00555197736985201,.0272131253366938,.020882855935959,-.00614930946231859,-.0129713607321759,.00264857212778643,.0251496849184886,.0589748924848905,.0636538256066845,.0222565523129904,-.0283850717047081,-.0383609141756602,-.00520313325716331,.0255032591162163,.0195707357913519,-.00576293353530576,-.0121563388897685,.0023045857657278,.0218833405621194,.0513154602550304,.0553867115372414,.0193659568929926,-.0246985277552192,-.0333787461712128,-.00452737031997128,.0221909939028245,.0170289638920514,-.00501446589087264,-.0105775203457034,.00181581404406023,.0172421776245331,.0404321396036532,.0436399331104409,.0152586972574031,-.0194603013836524,-.0262995619308917,-.00356717581611422,.0174845818192747,.0134173491179298,-.00395096495149618,-.00833417019261825,.00121304300143431,.0115185268917511,.0270104332212114,.0291533792327723,.0101934754715906,-.0130003303347394,-.0175692547623,-.00238302907325655,.0116804634694043,.00896337458042357,-.00263941695957371,-.00556758929031759,.000535123262049546,.0050812969342646,.0119154152965853,.0128607571012423,.00449676214242187,-.00573498150372383,-.00775052237147804,-.00105125233790734,.00515273383268983,.00395411394219069,-.00116435560128201,-.00245609309749454},
			{-.00090976995509462,-.00863877841154346,-.020257551125304,-.0218647389120542,-.00764500327777482,.0097501159735198,.013176762981398,.00178724764942572,-.00876022920331713,-.00672244007828754,.00197953596529396,.0041756355320037,-.00189959746932745,-.0180377484624032,-.0422977178317532,-.0456535220494547,-.0159627483828599,.0203582186081548,.0275130492859463,.00373176877615995,-.0182913374223476,-.0140364386501337,.00413326631535552,.00871871690750996,-.00242140718192188,-.02299262574193,-.0539166846610791,-.0581943111398842,-.0203476337495614,.025950517172659,.0350707432565425,.00475686657924678,-.0233158742926257,-.0178921765820591,.00526865343968954,.0111137038650945,-.00234437337189749,-.0222611463044402,-.0522013979161087,-.0563429374666252,-.0197003011718718,.0251249364009332,.0339550147687404,.0046055332722702,-.0225741111376221,-.0173229610688503,.00510103832266685,.0107601363368395,-.0016874128374327,-.0160229358089307,-.0375730717775044,-.0405540333803145,-.0141797127954073,.0180842099346574,.0244398475527877,.00331493100033043,-.0162481989361725,-.0124685714486963,.00367158135008905,.00774483808976782,-.000613021027013781,-.00582098011077765,-.013649939444666,-.0147328944291874,-.00515135472942478,.00656982138629644,.00887876405495684,.00120428288870561,-.00590281606138012,-.00452971336078232,.00133385056701396,.0028136259808645,.000613021027013781,.00582098011077765,.013649939444666,.0147328944291874,.00515135472942478,-.00656982138629644,-.00887876405495684,-.00120428288870561,.00590281606138012,.00452971336078232,-.00133385056701396,-.0028136259808645,.0016874128374327,.0160229358089307,.0375730717775044,.0405540333803145,.0141797127954073,-.0180842099346574,-.0244398475527877,-.00331493100033043,.0162481989361725,.0124685714486963,-.00367158135008905,-.00774483808976782,.00234437337189749,.0222611463044402,.0522013979161087,.0563429374666252,.0197003011718718,-.0251249364009332,-.0339550147687404,-.0046055332722702,.0225741111376221,.0173229610688503,-.00510103832266685,-.0107601363368395,.00242140718192188,.02299262574193,.0539166846610791,.0581943111398842,.0203476337495614,-.025950517172659,-.0350707432565425,-.00475686657924678,.0233158742926257,.0178921765820591,-.00526865343968954,-.0111137038650945,.00189959746932745,.0180377484624032,.0422977178317532,.0456535220494547,.0159627483828599,-.0203582186081548,-.0275130492859463,-.00373176877615995,.0182913374223476,.0140364386501337,-.00413326631535552,-.00871871690750996,.000909769955094619,.00863877841154346,.020257551125304,.0218647389120542,.00764500327777482,-.00975011597351981,-.013176762981398,-.00178724764942572,.00876022920331713,.00672244007828754,-.00197953596529396,-.00417563553200371},
			{.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00527985557699714,.0444955335780938,.0796442410302246,.041219887573566,-.0332406109919311,-.0427537419249747,.0125400203887214,.0358934414260092,-.001512251724891,-.0262417461372046,-.00337960764579559,.0126285013024013,.00494810953208644,.0416997720150249,.0746400015057946,.0386299427398909,-.0311520233277094,-.0400674213273932,.0117521006991787,.0336381700350283,-.0014172333060487,-.0245929140118533,-.00316725875604049,.0118350221439043,.00430546809553527,.0362839659950049,.0649460451612915,.0336128343401532,-.0271061183425881,-.0348636187369765,.0102257830566852,.0292693738768018,-.00123316849466542,-.0213988809196043,-.00275590736947309,.0102979349830697,.0033923360763522,.0285886236076473,.0511718603246688,.0264840031398202,-.0213572743086375,-.0274695129471004,.00805703166355282,.0230617323671449,-.000971630001628045,-.0168604653724913,-.00217141639076207,.00811388113437024,.00226622849921943,.0190984478291357,.0341850352134156,.0176924695366315,-.0142675910094163,-.0183508330835358,.00538244866195297,.0154062433542828,-.000649091231183039,-.0112635264541151,-.0014506008831851,.00542042658867729,.000999726790903561,.00842511245712309,.015080427927989,.00780487748658528,-.00629402241595725,-.00809530878079113,.00237441993597328,.00679632889347279,-.000286340893593815,-.0049688057316852,-.00063991983435387,.00239117356466622},
			{-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.36805442056342e-006,-8.38522436364201e-005,-.000722856514180969,-.00295108426386073,-.00795857100195779,-.0162605977118075,-.0269052706245537,-.0372592914205353,-.0436594558033572,-.042768865222384,-.033069007459881,-.0157988744548141,-1.28209626571885e-006,-7.85836051712954e-005,-.000677437698056106,-.00276566053049211,-.00745851481395462,-.0152389051863002,-.0252147476572524,-.0349182003841082,-.0409162270209603,-.0400815943915947,-.0309912020589656,-.0148061931138207,-1.11558253342258e-006,-6.83774687490572e-005,-.000589454695127486,-.0024064671769895,-.0064898315939296,-.0132597347865977,-.0219399532025914,-.0303831588068934,-.0356021848114063,-.0348759510581064,-.0269661839217411,-.0128832216939646,-8.78982445184583e-007,-5.38755250068191e-005,-.000464439262650617,-.00189608777487504,-.00511342538303042,-.0104475229362594,-.0172867834834987,-.0239392984565476,-.0280514390660412,-.027479229748374,-.0212470181010361,-.0101508632191237,-5.87198621468207e-007,-3.59912011760789e-005,-.000310265689921581,-.00126666935578621,-.00341599124344908,-.00697940112403575,-.0115483255515964,-.0159924957882962,-.0187395851191542,-.0183573243307094,-.0141939350525009,-.00678121949037336,-2.59037512618698e-007,-1.58772021730903e-005,-.000136870983053836,-.000558780057100242,-.00150693452348,-.00307890148348425,-.00509444235123848,-.00705494900380919,-.00826680332565275,-.00809817233744839,-.00626152973771969,-.0029914753970567},
			{4.40393013976499e-007,2.6993036058117e-005,.000232696122440507,.000949989177276297,.00256195880647401,.00523448009649574,.00866112710437402,.0119942097336745,.0140544988860133,.0137678071694569,.0106453074128955,.0050858458801123,9.19539549723029e-007,5.63613941066353e-005,.00048586894174179,.00198357510812152,.00534936380130629,.0109295818012743,.0180844124790659,.0250438809603261,.0293457597352286,.0287471481233613,.0222273761718653,.0106192339163871,1.17213236262568e-006,7.1843580904145e-005,.000619334655895749,.00252845304873323,.00681880668738833,.0139318820415032,.0230521081253869,.0319233070161965,.0374068898959917,.0366438427349157,.0283331227625236,.0135362831793791,1.13484254932205e-006,6.95579740866872e-005,.000599631357507937,.00244801371855121,.00660187553145923,.0134886579681297,.022318736335937,.0309077099756012,.0362168400475543,.0354780682047664,.0274317426011938,.0131056445512863,8.16827178275341e-007,5.0065838414078e-005,.000431597484647737,.00176201019189673,.00475184100640955,.00970874984676124,.0160643874648916,.0222464848021472,.0260678446360387,.0255360978134559,.0197445829974339,.00943306775437879,2.96745541228046e-007,1.81884426869712e-005,.000156795259243056,.000640121536048445,.00172629862078525,.00352709643428847,.0058360391029489,.00808193623891563,.00947019990385576,.00927702134312668,.00717301912047857,.00342694375341951,-2.96745541228046e-007,-1.81884426869712e-005,-.000156795259243056,-.000640121536048445,-.00172629862078525,-.00352709643428847,-.0058360391029489,-.00808193623891563,-.00947019990385576,-.00927702134312668,-.00717301912047857,-.00342694375341951,-8.16827178275341e-007,-5.0065838414078e-005,-.000431597484647737,-.00176201019189673,-.00475184100640955,-.00970874984676124,-.0160643874648916,-.0222464848021472,-.0260678446360387,-.0255360978134559,-.0197445829974339,-.00943306775437879,-1.13484254932205e-006,-6.95579740866872e-005,-.000599631357507937,-.00244801371855121,-.00660187553145923,-.0134886579681297,-.022318736335937,-.0309077099756012,-.0362168400475543,-.0354780682047664,-.0274317426011938,-.0131056445512863,-1.17213236262568e-006,-7.1843580904145e-005,-.000619334655895749,-.00252845304873323,-.00681880668738833,-.0139318820415032,-.0230521081253869,-.0319233070161965,-.0374068898959917,-.0366438427349157,-.0283331227625236,-.0135362831793791,-9.19539549723029e-007,-5.63613941066352e-005,-.00048586894174179,-.00198357510812152,-.00534936380130629,-.0109295818012743,-.0180844124790659,-.0250438809603261,-.0293457597352286,-.0287471481233613,-.0222273761718653,-.0106192339163871,-4.40393013976499e-007,-2.6993036058117e-005,-.000232696122440507,-.000949989177276297,-.00256195880647401,-.00523448009649574,-.00866112710437402,-.0119942097336745,-.0140544988860133,-.0137678071694569,-.0106453074128955,-.0050858458801123},
			{-5.47479249038328e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.014910729120128,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278,-9.53436115108903e-007,-5.84390183710293e-005,-.000503779306073239,-.00205669472909975,-.00554655494976578,-.0113324739708158,-.0187510498958539,-.0259670620771616,-.0304275190396141,-.0298068410819887,-.0230467337622952,-.0110106858739491,-7.64811749616544e-007,-4.68776535500915e-005,-.000404113423430025,-.00164980565479233,-.00444924450443136,-.00909049815478573,-.0150414097502034,-.0208298320830514,-.0244078483124377,-.0239099630454571,-.0184872509991657,-.00883237145550259,-4.34384682750355e-008,-2.66247670431932e-006,-2.2952142317351e-005,-9.37028368505082e-005,-.00025270057154655,-.000516306549812956,-.000854296237699806,-.00118305713865876,-.00138627517832104,-.00135799714338606,-.00105000722912861,-.000501645911501147,8.51599050625011e-007,5.21971129219724e-005,.00044997034618045,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801575,.023193505114819,.0271775380818778,.0266231550969231,.0205850987611647,.00983462812913714,1.45756655498971e-006,8.93387163903609e-005,.000770153192219288,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,1.45756655498971e-006,8.93387163903609e-005,.000770153192219289,.0031441744270585,.00847930224382976,.0173245325862927,.028665689044114,.0396971759462765,.0465161046457617,.0455672425058471,.0352327206840039,.0168325986639731,8.51599050625012e-007,5.21971129219725e-005,.000449970346180451,.00183701797212365,.00495412419836905,.0101220458527282,.0167482393801576,.0231935051148191,.0271775380818778,.0266231550969231,.0205850987611648,.00983462812913715,-4.34384682750334e-008,-2.6624767043192e-006,-2.29521423173499e-005,-9.37028368505037e-005,-.000252700571546538,-.000516306549812931,-.000854296237699765,-.0011830571386587,-.00138627517832097,-.00135799714338599,-.00105000722912856,-.000501645911501123,-7.64811749616543e-007,-4.68776535500914e-005,-.000404113423430024,-.00164980565479233,-.00444924450443135,-.00909049815478571,-.0150414097502034,-.0208298320830514,-.0244078483124376,-.0239099630454571,-.0184872509991656,-.00883237145550258,-9.53436115108901e-007,-5.84390183710292e-005,-.000503779306073238,-.00205669472909974,-.00554655494976577,-.0113324739708158,-.0187510498958538,-.0259670620771615,-.030427519039614,-.0298068410819887,-.0230467337622952,-.0110106858739491,-5.47479249038329e-007,-3.35566792418533e-005,-.000289278654122016,-.00118098912758303,-.00318492627930372,-.00650729948338485,-.0107671720768514,-.0149107291201281,-.0174719994448759,-.0171155955948952,-.0132338268846953,-.00632252327989278},
			{6.11261401273432e-007,3.74660826168086e-005,.000322980050454363,.00131857612920148,.00355597495952486,.00726540961635832,.0120215637451267,.0166478513879527,.019507513540437,.0191095881083022,.0147755875313232,.00705910670850027,7.63509266142174e-007,4.6797820350488e-005,.000403425213480253,.00164699601620824,.00444166738830656,.00907501692868488,.0150157940511236,.0207943586321327,.0243662814574726,.0238692440950028,.0184557669915978,.00881732982223001,-3.24515476414362e-008,-1.98905470301973e-006,-1.71468417157011e-005,-7.00025160864056e-005,-.000188784848135954,-.000385716791226497,-.000638218522855875,-.000883825710764993,-.00103564367667356,-.00101451802389758,-.000784428203227012,-.000374764277900816,-1.10535636397296e-006,-6.77506755168088e-005,-.000584051362415962,-.00238440790267356,-.00643034149293689,-.0131381872625712,-.0217388370391121,-.0301046466206003,-.0352758315710584,-.0345562549580405,-.0267189939936707,-.0127651255386803,-1.44954346598925e-006,-8.88469567034131e-005,-.000765913929467261,-.00312686750465765,-.00843262842551854,-.0172291707200675,-.0285079004521123,-.0394786651862746,-.0462600594955417,-.045316420311225,-.0350387842542686,-.0167399446189638,-6.62156769174321e-007,-4.05856155279577e-005,-.000349872290759864,-.00142836453897385,-.00385205557811836,-.00787034834568896,-.0130225134342083,-.0180339989827199,-.0211317647632424,-.0207007069245407,-.0160058450967311,-.007646868069241,6.62156769174322e-007,4.05856155279578e-005,.000349872290759864,.00142836453897385,.00385205557811836,.00787034834568897,.0130225134342083,.0180339989827199,.0211317647632424,.0207007069245407,.0160058450967311,.00764686806924101,1.44954346598925e-006,8.88469567034135e-005,.000765913929467265,.00312686750465767,.00843262842551858,.0172291707200676,.0285079004521125,.0394786651862747,.0462600594955419,.0453164203112252,.0350387842542687,.0167399446189638,1.10535636397296e-006,6.77506755168087e-005,.000584051362415961,.00238440790267356,.00643034149293688,.0131381872625712,.0217388370391121,.0301046466206002,.0352758315710583,.0345562549580404,.0267189939936707,.0127651255386803,3.24515476414349e-008,1.98905470301965e-006,1.71468417157005e-005,7.00025160864029e-005,.000188784848135946,.000385716791226483,.000638218522855851,.000883825710764959,.00103564367667352,.00101451802389754,.000784428203226982,.000374764277900801,-7.63509266142173e-007,-4.67978203504879e-005,-.000403425213480253,-.00164699601620824,-.00444166738830656,-.00907501692868487,-.0150157940511236,-.0207943586321327,-.0243662814574726,-.0238692440950028,-.0184557669915978,-.00881732982223001,-6.11261401273429e-007,-3.74660826168084e-005,-.000322980050454361,-.00131857612920147,-.00355597495952484,-.00726540961635829,-.0120215637451267,-.0166478513879526,-.0195075135404369,-.0191095881083021,-.0147755875313231,-.00705910670850024},
			{-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906563,.00323132809181624,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190404,.00732492904924932,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-9.37620916445639e-006,-.000548452180033515,-.00433589285085915,-.0155237909939332,-.034805414745598,-.0551701940084311,-.0640514016115786,-.052169641675679,-.0224031552490585,.00961329757959373,.0253703786143574,.0170656081260598,-8.78707935565717e-006,-.000513991608357755,-.00406345825800857,-.0145483938094544,-.0326185066919698,-.0517037178155989,-.0600268977504876,-.0488916974141509,-.0209955110362555,.00900927093453281,.0237762966095638,.01599333485698,-7.64584728235915e-006,-.000447236354976977,-.00353571192673622,-.0126589043718029,-.0283821404873109,-.0449886377882491,-.0522308351225414,-.0425418317820523,-.0182686948075582,.00783918147347644,.020688322644971,.0139161820330408,-6.02426565351236e-006,-.000352383524387965,-.00278583486359916,-.00997412059146167,-.0223626692760875,-.0354471512988985,-.0411533757428001,-.0335192800192893,-.0143941497389828,.00617659625654242,.016300606974774,.0109647465288904,-4.02447228210328e-006,-.000235407568014932,-.00186105922877672,-.00666314770420021,-.0149392387108601,-.0236802435494798,-.0274922504281194,-.0223923414258862,-.00961592001096531,.00412623577077897,.0108895166190405,.00732492904924932,-1.77536058744882e-006,-.000103847980292826,-.000820989927144264,-.00293938906598824,-.00659031389822105,-.010446331382599,-.0121279647240725,-.00987818467653581,-.00424197862548232,.00182025265636417,.00480381457906564,.00323132809181624},
			{3.01831341787578e-006,.000176553289823538,.00139577555712289,.00499729323775215,.0112042776028942,.0177599426294488,.0206189091483633,.0167940290914578,.0072118425373148,-.00309463500286609,-.00816702708367357,-.00549362253845184,6.30223112790928e-006,.000368642842811124,.00291437599275742,.0104343362129758,.0233945045787828,.037082717323626,.0430522325112947,.0350658921887862,.0150583098027748,-.00646159041306787,-.0170527328289948,-.0114706705943021,8.03342179680519e-006,.000469907146940446,.00371494017739558,.0133005950221825,.0298208553755469,.0472691501765359,.0548784606654209,.0446983134887451,.0191947505157252,-.00823655626282739,-.0217370313501955,-.0146216051595062,7.77784929616953e-006,.000454957683604922,.0035967543563515,.01287745449555,.0288721450037781,.0457653457929115,.0531325763115239,.0432762968137264,.0185840953662182,-.00797452131259557,-.0210454969581169,-.0141564384735731,5.59827325600047e-006,.000327465515948182,.00258884081637867,.00926882308497601,.0207813434103976,.0329405857131029,.0382433073283765,.0311490395153758,.0133762998116843,-.00573983214301874,-.0151479462115431,-.0101893991370919,2.03379940271082e-006,.000118965105897601,.000940501234094081,.00336727522791983,.00754966431306182,.0119670013385727,.0138934296425726,.0113161496526603,.00485948243743287,-.00208522639934011,-.00550310471613919,-.00370171174777645,-2.03379940271082e-006,-.000118965105897601,-.000940501234094081,-.00336727522791983,-.00754966431306182,-.0119670013385727,-.0138934296425726,-.0113161496526603,-.00485948243743287,.00208522639934011,.00550310471613919,.00370171174777645,-5.59827325600047e-006,-.000327465515948182,-.00258884081637867,-.00926882308497601,-.0207813434103976,-.0329405857131029,-.0382433073283765,-.0311490395153758,-.0133762998116843,.00573983214301874,.0151479462115431,.0101893991370919,-7.77784929616953e-006,-.000454957683604922,-.0035967543563515,-.01287745449555,-.0288721450037781,-.0457653457929115,-.0531325763115239,-.0432762968137265,-.0185840953662182,.00797452131259557,.0210454969581169,.0141564384735731,-8.03342179680519e-006,-.000469907146940446,-.00371494017739558,-.0133005950221825,-.0298208553755469,-.0472691501765359,-.0548784606654209,-.0446983134887451,-.0191947505157252,.00823655626282739,.0217370313501955,.0146216051595062,-6.30223112790928e-006,-.000368642842811124,-.00291437599275742,-.0104343362129758,-.0233945045787828,-.037082717323626,-.0430522325112947,-.0350658921887862,-.0150583098027748,.00646159041306787,.0170527328289948,.0114706705943021,-3.01831341787578e-006,-.000176553289823538,-.00139577555712289,-.00499729323775215,-.0112042776028942,-.0177599426294488,-.0206189091483634,-.0167940290914578,-.0072118425373148,.00309463500286609,.00816702708367357,.00549362253845184},
			{-3.75224835608883e-006,-.000219484095846042,-.00173517319595299,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131,-6.5345473857451e-006,-.000382231954981349,-.00302180729935703,-.0108189723668713,-.0242568854796668,-.0384496805377896,-.0446392472280685,-.0363585101022895,-.0156133973760028,.00669978105920017,.0176813399042196,.0118935086675575,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786293,-.0194579906613178,-.0308429342861059,-.0358079794052745,-.029165473474441,-.0125244990989618,.00537432051580024,.0141833273288231,.00954053976870304,-2.97713422859992e-007,-1.7414455344253e-005,-.000137673283428536,-.000492911460431877,-.00110514163839842,-.00175176417356002,-.0020337603052895,-.0016564906264595,-.000711345055883006,.00030524145496262,.000805560341504197,.000541867242857217,5.83658858918833e-006,.000341405538160759,.00269904630894985,.00966339165295146,.02166599347167,.0343428478572752,.039871303339182,.032475036548949,.0139457213459092,-.00598417355813986,-.0157927857332021,-.0106231561080896,9.98969681400165e-006,.000584337539768473,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,9.98969681400166e-006,.000584337539768474,.0046195913762544,.016539516420047,.0370827415105201,.058779993241099,.0682423004210598,.0555831140383219,.0238689991541032,-.0102422979818929,-.0270303686669682,-.018182215022694,5.83658858918834e-006,.000341405538160759,.00269904630894985,.00966339165295147,.02166599347167,.0343428478572752,.039871303339182,.0324750365489491,.0139457213459093,-.00598417355813987,-.0157927857332022,-.0106231561080896,-2.97713422859977e-007,-1.74144553442522e-005,-.00013767328342853,-.000492911460431854,-.00110514163839836,-.00175176417355993,-.0020337603052894,-.00165649062645942,-.000711345055882972,.000305241454962606,.000805560341504158,.000541867242857191,-5.24177607691426e-006,-.000306612562306019,-.00242398383174452,-.00867858585786291,-.0194579906613177,-.0308429342861059,-.0358079794052744,-.029165473474441,-.0125244990989617,.00537432051580023,.0141833273288231,.00954053976870302,-6.53454738574509e-006,-.000382231954981348,-.00302180729935702,-.0108189723668713,-.0242568854796668,-.0384496805377895,-.0446392472280684,-.0363585101022894,-.0156133973760028,.00669978105920015,.0176813399042196,.0118935086675575,-3.75224835608883e-006,-.000219484095846043,-.001735173195953,-.00621243812030823,-.0139287166029995,-.0220784611501613,-.0256326157178006,-.0208776754850327,-.00896547858308749,.00384712834440214,.0101529263884114,.00682945515757131},
			{-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784,-1.55303998105985e-005,-.000854853930745827,-.00602179902609634,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.61827009417749e-005,-.00199163733724155,-.0140295778569733,-.0418230593862318,-.0708529069753802,-.0712371902047935,-.0308726481949129,.0221562845712239,.0433269897361088,.0201132464521264,-.0134769616164845,-.0174499209706319,-3.39092546786013e-005,-.00186649796555022,-.0131480656834027,-.0391952158145448,-.0664010481476665,-.0667611859360495,-.0289328453375767,.020764151837813,.0406046505976176,.0188494827253538,-.0126301716525392,-.0163535003997299,-2.95052510894069e-005,-.00162408438797623,-.0114404454773706,-.0341046919218024,-.0577771353798894,-.0580904998573378,-.0251751586552827,.0180673836520465,.0353310747356644,.0164013844004648,-.0109898135315149,-.0142295706602257,-2.32475831875794e-005,-.00127963787865143,-.00901408048120715,-.0268715443273595,-.0455233801269445,-.0457702842029491,-.019835845264483,.0142355339786843,.0278378278000431,.0129228708166516,-.00865902152520925,-.0112116696327964,-1.55303998105985e-005,-.000854853930745827,-.00602179902609635,-.0179513639574835,-.0304116040104777,-.0305765467051345,-.0132512100312938,.00950995776302659,.0185968834740743,.00863304150216106,-.00578460415304274,-.00748988445537439,-6.85109942331355e-006,-.000377111300634594,-.00265646373165771,-.00791908648564711,-.013415811906918,-.013488574927447,-.00584565487758069,.0041952343108092,.00820385174873033,.00380839040708625,-.00255182729745069,-.00330409671989784},
			{1.16476424390402e-005,.000641131783107114,.00451628823154087,.0134633410098786,.0228083947504036,.0229321000995076,.00993827320683666,-.00713237192763805,-.0139474741042205,-.00647469362056888,.0043383944810003,.00561733742272306,2.43202492860194e-005,.0013386816149243,.00942999892153632,.0281114235173858,.0476238731611908,.0478821696313965,.0207510904569588,-.0148923753616752,-.0291222922492883,-.0135191446446746,.00905855719998496,.011728986973914,3.10008974207295e-005,.00170641061015528,.0120203714117433,.0358334878300008,.0607059076280363,.0610351568179937,.0264513089096609,-.0189832347320423,-.0371220370342043,-.0172327845581125,.0115468965483827,.0149508797278819,3.00146455997538e-005,.00165212345360114,.0116379594759614,.0346934936503285,.0587746308933795,.0590934054635625,.0256097961229407,-.0183793086724707,-.0359410493961229,-.0166845467145346,.0111795475780318,.0144752376147941,2.16036826314166e-005,.00118915116425069,.00837667005463561,.0249713835102858,.0423042966933997,.0425337415030114,.0184331980818002,-.0132289002122403,-.0258693384205835,-.0120090590732415,.00804671829410513,.0104188616322334,7.84841232696798e-006,.000432007302429133,.0030431645214108,.00907186601040717,.0153687484359428,.0154521035519588,.00669660545930529,-.00480593356555747,-.00939808448469602,-.00436277688733732,.00292329618651467,.00378507328877716,-7.84841232696798e-006,-.000432007302429133,-.0030431645214108,-.00907186601040717,-.0153687484359428,-.0154521035519588,-.00669660545930529,.00480593356555747,.00939808448469602,.00436277688733732,-.00292329618651467,-.00378507328877716,-2.16036826314166e-005,-.00118915116425069,-.00837667005463561,-.0249713835102858,-.0423042966933997,-.0425337415030114,-.0184331980818002,.0132289002122403,.0258693384205835,.0120090590732415,-.00804671829410513,-.0104188616322334,-3.00146455997538e-005,-.00165212345360114,-.0116379594759614,-.0346934936503285,-.0587746308933795,-.0590934054635624,-.0256097961229407,.0183793086724707,.0359410493961229,.0166845467145346,-.0111795475780318,-.0144752376147941,-3.10008974207295e-005,-.00170641061015528,-.0120203714117433,-.0358334878300008,-.0607059076280363,-.0610351568179937,-.0264513089096609,.0189832347320423,.0371220370342043,.0172327845581125,-.0115468965483827,-.0149508797278819,-2.43202492860194e-005,-.0013386816149243,-.00942999892153632,-.0281114235173858,-.0476238731611908,-.0478821696313965,-.0207510904569588,.0148923753616752,.0291222922492883,.0135191446446746,-.00905855719998497,-.011728986973914,-1.16476424390402e-005,-.000641131783107114,-.00451628823154087,-.0134633410098786,-.0228083947504036,-.0229321000995076,-.00993827320683666,.00713237192763805,.0139474741042205,.00647469362056888,-.0043383944810003,-.00561733742272306},
			{-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.00322545981703339,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072408,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-.000103908282720131,-.00530406999588436,-.0322235977909857,-.0746218160727394,-.079950093891901,-.0193814114565919,.0460640969795862,.0391409609341219,-.0168523016252334,-.0345178310254824,-.000108372033446398,.0170346160153937,-9.73794750050022e-005,-.0049708025007032,-.0301989115161326,-.0699331476072539,-.0749266368953818,-.0181636306855565,.0431697787993811,.0366816400692728,-.0157934309174492,-.0323489925497693,-.000101562757520101,.015964290055286,-8.47322032940573e-005,-.00432521378866087,-.0262767930276145,-.0608504993454398,-.0651954534488237,-.0158046081859416,.0375630539515719,.0319175697275982,-.0137422408483843,-.0281476298054,-8.83721771641939e-005,.0138909094584902,-6.67616397764769e-005,-.00340789397288197,-.02070383776641,-.0479449248288804,-.0513683724605766,-.0124526628306406,.0295964341693814,.0251482814072409,-.0108276959358351,-.0221778951635208,-6.9629624024101e-005,.010944833928335,-4.4599688035262e-005,-.00227662185285837,-.0138310668911987,-.032029301518079,-.0343163138935974,-.00831892205335797,.0197717092529484,.0168001491446457,-.00723337327387024,-.0148158015423745,-4.65156266365657e-005,.00731162656333502,-1.96747605151687e-005,-.00100431172754334,-.0061014536365919,-.0141294449489894,-.015138340364311,-.00366981937216172,.00872211581883144,.00741123818576233,-.00319093906593472,-.00653586067588674,-2.05199599953224e-005,.0032254598170334},
			{3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.72326768036953e-008,5.49235412022583e-006,.000113623211690342,.000831956416390095,.00343693515366776,.00971699735195307,.0206815637605425,.0348153353344463,.0473418949315913,.0517107103024967,.043014832840112,.0213863615446234,1.61499062071363e-008,5.14725627994169e-006,.000106483991585053,.000779682590591159,.00322098423849286,.00910645500038528,.0193820912882233,.0326278039463224,.0443672896278861,.0484616018046971,.0403121072713726,.0200426049296945,1.40524184983822e-008,4.47875042965911e-006,9.26542602749084e-005,.00067842041423425,.0028026551928824,.0079237436466072,.0168648198114988,.0283902302500237,.0386050366790666,.0421675953362895,.0350765258115668,.017439548481424,1.10720890677357e-008,3.5288675522334e-006,7.30035347571707e-005,.000534536546334465,.00220824962801375,.00624322392727203,.0132880177946399,.0223690433086643,.030417426340031,.0332244141027671,.0276372652876165,.0137408542244633,7.39663854831856e-009,2.35743748167825e-006,4.87695462026149e-005,.000357093733613827,.00147520709262294,.00417074596164615,.00887697561402263,.0149434968427065,.0203201678410383,.0221953581292932,.0184628989656689,.0091794901054094,3.26296210727695e-009,1.0399628321353e-006,2.15142568085066e-005,.000157528763088314,.000650774647452401,.00183988793598998,.00391599979728114,.00659219233567233,.00896406350610278,.00979128722597762,.00814474566006138,.00404945140720489},
			{-5.54740393494954e-009,-1.76805421500992e-006,-3.65766652977238e-005,-.000267816680517056,-.00110639036597811,-.00312801719432003,-.00665764172873835,-.0112074711567434,-.0152399199046756,-.0166462936129339,-.0138469870744205,-.00688452453082954,-1.15829660203181e-008,-3.69169293144801e-006,-7.63719888164115e-005,-.000559200582196592,-.00231014041245373,-.00653129234821247,-.0139011398527498,-.0234011727115081,-.0318209159596439,-.0347574208663796,-.0289124755737218,-.0143748705956402,-1.47647475649067e-008,-4.70578210489157e-006,-9.73509836708664e-005,-.000712810123054713,-.00294472417250793,-.00832540496318534,-.0177197118967874,-.0298293552102391,-.0405619588803169,-.0443051066712203,-.0368545847906068,-.0183235740526672,-1.42950269960305e-008,-4.55607398170153e-006,-9.42538931700292e-005,-.000690133028507038,-.00285104173686768,-.00806054341115132,-.0171559831153849,-.0288803744276712,-.0392715347591976,-.0428955993418035,-.0356821057856366,-.0177406342096373,-1.0289151187989e-008,-3.27933161891947e-006,-6.78412539656127e-005,-.000496737996514833,-.00205209822143352,-.0058017483868841,-.0123483854981035,-.0207872667141207,-.0282665264384568,-.0308750243738453,-.0256829582228968,-.0127692006181209,-3.73795071866204e-009,-1.19134997219017e-006,-2.46460820122577e-005,-.000180460187350232,-.000745507756803769,-.00210771998156325,-.00448605094857911,-.00755181619294033,-.0102689600807936,-.0112166025591737,-.00933037433240469,-.00463892907735314,3.73795071866204e-009,1.19134997219017e-006,2.46460820122577e-005,.000180460187350232,.000745507756803769,.00210771998156325,.0044860509485791,.00755181619294033,.0102689600807936,.0112166025591737,.00933037433240469,.00463892907735314,1.0289151187989e-008,3.27933161891947e-006,6.78412539656127e-005,.000496737996514833,.00205209822143352,.0058017483868841,.0123483854981035,.0207872667141207,.0282665264384568,.0308750243738453,.0256829582228968,.0127692006181209,1.42950269960305e-008,4.55607398170153e-006,9.42538931700292e-005,.000690133028507038,.00285104173686769,.00806054341115132,.0171559831153849,.0288803744276712,.0392715347591976,.0428955993418035,.0356821057856366,.0177406342096373,1.47647475649067e-008,4.70578210489157e-006,9.73509836708664e-005,.000712810123054713,.00294472417250793,.00832540496318534,.0177197118967874,.0298293552102391,.0405619588803169,.0443051066712203,.0368545847906068,.0183235740526672,1.15829660203181e-008,3.69169293144801e-006,7.63719888164114e-005,.000559200582196592,.00231014041245373,.00653129234821247,.0139011398527498,.0234011727115081,.0318209159596439,.0347574208663796,.0289124755737218,.0143748705956402,5.54740393494954e-009,1.76805421500992e-006,3.65766652977238e-005,.000267816680517056,.00110639036597811,.00312801719432003,.00665764172873835,.0112074711567434,.0152399199046756,.0166462936129339,.0138469870744205,.00688452453082954},
			{6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978,1.20099435917211e-008,3.82777811718284e-006,7.91872544617432e-005,.000579814137143969,.00239529806043259,.00677205238670959,.014413571204413,.024263799423424,.0329939158106694,.0360386677528614,.0299782629188682,.0149047648667097,9.63393962702057e-009,3.07050428713007e-006,6.35211333748341e-005,.000465105797498382,.00192142092314525,.00543229394429081,.0115620422138201,.0194635367753551,.0264665183938168,.0289089075829344,.0240474714038134,.0119560598918884,5.47172008094484e-010,1.743932453074e-007,3.60776457511335e-006,2.64162827510142e-005,.000109129575813761,.000308534130494273,.000656681077600405,.00110545663708768,.00150319999683172,.00164191863623543,.00136580710768774,.000679059819058227,-1.07271545504665e-008,-3.41893091623005e-006,-7.07292178810733e-005,-.00051788385284143,-.00213945488523406,-.00604872554323789,-.0128740493036516,-.0216721689330841,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859674,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175488,-.0370933112099723,-.0504394866543156,-.0550941546796317,-.0458293038356739,-.0227856763924034,-1.83602150466119e-008,-5.85172018881405e-006,-.000121057606121532,-.000886391527464572,-.00366181465836281,-.0103527828567504,-.0220347635175489,-.0370933112099723,-.0504394866543156,-.0550941546796318,-.0458293038356739,-.0227856763924034,-1.07271545504666e-008,-3.41893091623005e-006,-7.07292178810734e-005,-.000517883852841431,-.00213945488523406,-.0060487255432379,-.0128740493036517,-.0216721689330842,-.0294698165252092,-.0321893567463841,-.0267762672679743,-.0133127782859675,5.47172008094458e-010,1.74393245307392e-007,3.60776457511317e-006,2.6416282751013e-005,.000109129575813756,.000308534130494258,.000656681077600374,.00110545663708762,.00150319999683165,.00164191863623535,.00136580710768767,.000679059819058194,9.63393962702055e-009,3.07050428713007e-006,6.3521133374834e-005,.000465105797498381,.00192142092314524,.0054322939442908,.01156204221382,.0194635367753551,.0264665183938168,.0289089075829343,.0240474714038134,.0119560598918883,1.20099435917211e-008,3.82777811718283e-006,7.91872544617431e-005,.000579814137143968,.00239529806043259,.00677205238670958,.014413571204413,.024263799423424,.0329939158106693,.0360386677528614,.0299782629188681,.0149047648667096,6.89631407326868e-009,2.19797536077311e-006,4.54706696328262e-005,.000332939148575391,.00137542092497599,.00388862777103897,.00827651796895816,.0139326867072111,.018945668162356,.0206940165628452,.017214028931997,.00855856973215978},
			{2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387107,.0553680555817044,.0585549188607202,.0369926009098689,.00167071599568083,-.0222958057593638,-.0178801188985916,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.52535049751819e-007,4.64755929344966e-005,.000884740439678186,.00571388117519031,.0198095851258014,.0441461398648543,.0678986188177073,.0718067137062254,.0453645424580294,.00204882232817407,-.0273416575782538,-.0219266391921467,1.42950905123719e-007,4.35554194721634e-005,.000829150066540988,.00535486425639097,.0185649011716725,.0413723315504668,.0636323850317286,.067294924903922,.0425141789736188,.00192008988574559,-.0256237153674089,-.0205489356311119,1.24384991327898e-007,3.7898609096873e-005,.000721463244650099,.00465939508054745,.0161537632045298,.0359990522387106,.0553680555817044,.0585549188607202,.036992600909869,.00167071599568083,-.0222958057593638,-.0178801188985916,9.80046034659848e-008,2.9860822570306e-005,.000568450569898921,.00367119989626912,.0127277667541895,.0283641362365462,.0436252338327776,.046136206166545,.0291469665643977,.00131637954799016,-.0175671644872398,-.0140879855670184,6.54713508421429e-008,1.99483322394548e-005,.000379749780949684,.00245252169714351,.00850270347647212,.0189484804713464,.0291435594739958,.0308209985411929,.0194714452835291,.000879398968822909,-.0117356322945065,-.00941138898681875,2.88821111799103e-008,8.80003149748458e-006,.000167523279310062,.00108190839835928,.00375089293222898,.00835895567487687,.0128564252009364,.0135964127071225,.00858965700690012,.00038793912852479,-.00517707107515956,-.00415175156122967},
			{-4.91028494999023e-008,-1.49610469790016e-005,-.000284808486486742,-.00183936641356981,-.00637694142212144,-.0142111682876331,-.0218573742001049,-.0231154364976192,-.0146033865958862,-.000659540312839943,.00880160526598481,.00705844634424949,-1.02526631182363e-007,-3.12386299642582e-005,-.000594679432029528,-.00384059262983655,-.0133150382903681,-.0296728850674858,-.0456381445486889,-.0482649755879333,-.0304918359479907,-.00137711858055358,.0183777305412829,.0147380189220751,-1.30690172571732e-007,-3.98197218991008e-005,-.000758034831541094,-.00489558378913558,-.0169726112318295,-.037823874884499,-.0581747095182799,-.0615231176135874,-.0388677873847604,-.00175540601372305,.0234259991596233,.0187864773676752,-1.26532440654065e-007,-3.85529110484099e-005,-.000733918973770133,-.00473983738085433,-.0164326504524067,-.0366205592199579,-.0563239594442265,-.0595658424432908,-.0376312610492393,-.00169956013435736,.0226807325302158,.0181888108787948,-9.10744283614461e-008,-2.7749281664479e-005,-.000528253866393645,-.00341159925277751,-.0118277513551488,-.0263584301385007,-.0405403735431023,-.0428737881159049,-.0270859043804411,-.00122329472902273,.0163249419621616,.0130917853539995,-3.30864731915692e-008,-1.00810499763263e-005,-.0001919096029833,-.00123940154495863,-.00429690951861456,-.00957576685179106,-.0147279319403104,-.0155756392506594,-.00984005132149962,-.000444411334612696,.00593069607246235,.00475612104229521,3.30864731915692e-008,1.00810499763263e-005,.0001919096029833,.00123940154495863,.00429690951861456,.00957576685179106,.0147279319403104,.0155756392506594,.00984005132149962,.000444411334612696,-.00593069607246235,-.00475612104229521,9.10744283614461e-008,2.7749281664479e-005,.000528253866393645,.00341159925277751,.0118277513551488,.0263584301385007,.0405403735431023,.0428737881159049,.0270859043804411,.00122329472902273,-.0163249419621616,-.0130917853539995,1.26532440654065e-007,3.85529110484099e-005,.000733918973770133,.00473983738085433,.0164326504524067,.0366205592199579,.0563239594442265,.0595658424432908,.0376312610492393,.00169956013435736,-.0226807325302158,-.0181888108787948,1.30690172571732e-007,3.98197218991008e-005,.000758034831541094,.00489558378913558,.0169726112318295,.037823874884499,.0581747095182799,.0615231176135874,.0388677873847604,.00175540601372305,-.0234259991596233,-.0187864773676752,1.02526631182363e-007,3.12386299642582e-005,.000594679432029528,.00384059262983655,.0133150382903681,.0296728850674858,.0456381445486889,.0482649755879333,.0304918359479907,.00137711858055358,-.0183777305412829,-.0147380189220751,4.91028494999023e-008,1.49610469790016e-005,.000284808486486742,.00183936641356981,.00637694142212144,.0142111682876331,.0218573742001049,.0231154364976192,.0146033865958862,.000659540312839943,-.00880160526598481,-.00705844634424949},
			{1.40569330619306e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201856,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016855,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,7.42388591541396e-007,.000213988711239994,.00367013709763599,.020145373447898,.054921428306325,.0851770859238296,.0702659942916243,.0073905155369812,-.0455218423314331,-.0362883751543923,.00882865075987345,.0214854988611044,6.9574252794381e-007,.000200543285020632,.00343953354254533,.0188795911045881,.051470582661825,.0798252044262831,.0658510126016854,.00692615164227047,-.042661595317046,-.0340082891258741,.00827392536468535,.0201355132097623,6.05382164106225e-007,.000174497494413,.00299282015382307,.0164275824191952,.0447857813316544,.0694578138677587,.0572985363352896,.00602660970361757,-.0371208713855467,-.0295914233235353,.00719933975824334,.0175203901913839,4.76988729148252e-007,.000137488916976098,.00235808315206798,.0129435125870923,.0352873179751444,.0547267434201857,.0451462855186283,.00474844664088364,-.0292480325931122,-.0233154794469702,.00567245638473819,.0138045504923454,3.18649281047824e-007,9.18485948812865e-005,.00157530242360068,.00864683110537086,.023573467916033,.0365598522130086,.0301596883603081,.00317216952046941,-.0195389618000005,-.0155757574740372,.00378944834189189,.00922204199128376,1.40569330619305e-007,4.05182006321504e-005,.000694930823256063,.0038144735693841,.0103992282500556,.0161280575817525,.0133046815312258,.00139937471265289,-.00861944132492815,-.0068711085580182,.00167168184119089,.00406822279778917},
			{-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-2.0141694363462e-010,-3.33807606470966e-007,-1.65720363705403e-005,-.000217627310910816,-.00137721402318709,-.00538792470927308,-.0147510602465105,-.0301856428110978,-.0476328767706495,-.0580132417457791,-.051916950382044,-.0268621961430683,-1.88761431858901e-010,-3.12833670424248e-007,-1.55307754038588e-005,-.000203953262708863,-.00129068034844448,-.00504938842044631,-.0138242155963766,-.0282890061570013,-.0446399883770301,-.0543681300147957,-.0486548833232205,-.0251743796492038,-1.64245823029177e-010,-2.72204036407589e-007,-1.35136980227764e-005,-.000177464597313201,-.00112305174849653,-.00439359327137406,-.0120287796399897,-.0246149388313989,-.0388423183634187,-.0473070063778303,-.0423357741945106,-.0219048280361805,-1.29411487552265e-010,-2.14473212283634e-007,-1.06476239773168e-005,-.000139826737159004,-.000884867539951324,-.00346177108471815,-.00947763686126094,-.019394440549112,-.0306043837625979,-.0372738250147888,-.0333569244815142,-.0172591078936303,-8.64525196255565e-011,-1.43277462803437e-007,-7.11307734943365e-006,-9.34103607574579e-005,-.000591130121526225,-.00231261411410569,-.00633147491192501,-.012956332423902,-.0204450635558571,-.0249005412854803,-.0222838808434348,-.0115298370501388,-3.81377694440583e-011,-6.32055938519091e-008,-3.13787157581305e-006,-4.12071599264428e-005,-.000260771859326364,-.00102018939735754,-.00279307452781809,-.00571557221193217,-.00901916015333691,-.0109846550070609,-.00983033824354329,-.00508628631126404},
			{6.48385134009702e-011,1.0745664478345e-007,5.33473591100581e-006,7.00568237223647e-005,.000443341599207306,.00173443714397409,.00474854200551721,.00971711798722703,.0153335904273431,.0186751535620109,.0167126847540202,.0086472609164468,1.35382659409353e-010,2.24369214821818e-007,1.11389156998738e-005,.000146278478759263,.000925696188564719,.00362150056830361,.0099149442407531,.0202893188924589,.0320164997847783,.0389936755407666,.0348960454086467,.0180554599121578,1.7257158376693e-010,2.86002290982243e-007,1.41987188917692e-005,.000186460428984299,.00117998019867963,.00461630825846974,.0126385287307287,.0258626910573308,.0408112685822381,.0497050388456101,.0444818106664909,.0230152024363774,1.67081451128867e-010,2.76903513083773e-007,1.37470057631949e-005,.00018052844143149,.00114244071703594,.00446944662526241,.012236450951948,.025039904355241,.0395129130085492,.0481237399429412,.0430666817373268,.0222830047517444,1.20260445247909e-010,1.9930722141311e-007,9.89470119356512e-006,.000129939203901917,.000822296121872419,.00321697972775721,.00880744708519725,.0180230063024996,.0284402635920669,.0346380903052428,.0309981645843277,.0160386688935279,4.36894754025765e-011,7.24064170021469e-008,3.59465078913488e-006,4.72056763219235e-005,.000298732356396325,.00116869812344953,.00319966171749153,.00654758668912568,.0103320771354786,.012583688604042,.0112613382258831,.00582669579070132,-4.36894754025765e-011,-7.24064170021469e-008,-3.59465078913488e-006,-4.72056763219235e-005,-.000298732356396325,-.00116869812344953,-.00319966171749153,-.00654758668912568,-.0103320771354786,-.012583688604042,-.0112613382258831,-.00582669579070132,-1.20260445247909e-010,-1.9930722141311e-007,-9.89470119356512e-006,-.000129939203901917,-.000822296121872419,-.00321697972775721,-.00880744708519725,-.0180230063024996,-.0284402635920669,-.0346380903052428,-.0309981645843277,-.0160386688935279,-1.67081451128867e-010,-2.76903513083773e-007,-1.37470057631949e-005,-.00018052844143149,-.00114244071703594,-.00446944662526241,-.012236450951948,-.025039904355241,-.0395129130085492,-.0481237399429412,-.0430666817373268,-.0222830047517444,-1.7257158376693e-010,-2.86002290982243e-007,-1.41987188917692e-005,-.000186460428984299,-.00117998019867963,-.00461630825846974,-.0126385287307287,-.0258626910573308,-.0408112685822381,-.0497050388456102,-.0444818106664909,-.0230152024363774,-1.35382659409353e-010,-2.24369214821818e-007,-1.11389156998738e-005,-.000146278478759263,-.000925696188564719,-.00362150056830361,-.0099149442407531,-.0202893188924589,-.0320164997847783,-.0389936755407666,-.0348960454086467,-.0180554599121578,-6.48385134009701e-011,-1.0745664478345e-007,-5.33473591100581e-006,-7.00568237223647e-005,-.000443341599207306,-.00173443714397409,-.00474854200551721,-.00971711798722703,-.0153335904273431,-.0186751535620109,-.0167126847540202,-.0086472609164468},
			{-4.13541370363107e-010,-6.55869442144476e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.18403540897137e-009,-3.46384228510943e-006,-.000158563343663295,-.00184288182980323,-.00984111516993843,-.0306251138857421,-.061585355370028,-.0819038310281344,-.0662546072765787,-.0168748533830729,.0268803415312295,.0264417730382408,-2.04680720295228e-009,-3.24620073004761e-006,-.000148600426806764,-.00172708912499165,-.00922277419687066,-.0287008642053535,-.0577157991349224,-.0767576160208537,-.0620916706961018,-.0158145656940172,.0251913849189151,.0247803727408388,-1.78097575505911e-009,-2.82459666349179e-006,-.000129300774861722,-.00150278143148985,-.00802495575319428,-.0249733063403579,-.0502198931071257,-.0667886320470723,-.0540274432986983,-.0137606307214519,.0219196247268608,.0215619932297993,-1.40325469169329e-009,-2.22553760708242e-006,-.0001018778152644,-.00118406165178689,-.00632297030395405,-.0196768031174225,-.0395689275492526,-.0526236592528572,-.0425689474288071,-.0108421855635509,.0172707664047356,.0169889837500738,-9.37435354989436e-010,-1.48675621687759e-006,-6.80588253032585e-005,-.000791004841418217,-.00422402002043004,-.0131449629383957,-.0264337699088145,-.0351549002362628,-.028437914072674,-.00724305297733525,.0115376254441925,.0113493823373164,-4.13541370363107e-010,-6.55869442144475e-007,-3.00235528043745e-005,-.000348944835868313,-.00186338931894618,-.00579878490605646,-.0116610253429981,-.0155082753613942,-.0125451359320819,-.00319520918207054,.00508972209287048,.00500668030021289},
			{4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906685,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.24897089581519e-012,1.93808311505349e-008,2.30899591395707e-006,5.43831908772756e-005,.000527193447711421,.00285397012027109,.0100508217305121,.0250016542281072,.0457832327865349,.0621744904907474,.0598602571462754,.0322318027893339,2.10766263673014e-012,1.8163086841527e-008,2.16391613838377e-006,5.096616398704e-005,.000494068613399084,.00267464830248915,.00941930438906686,.0234307400647474,.0429065619801723,.0582679174767988,.0560990929882761,.030206600974669,1.83392751913618e-012,1.58041349742831e-008,1.88287503233516e-006,4.43468746144535e-005,.000429900881983536,.00232727526714407,.0081959613598453,.0203876456552513,.0373340227205607,.0507003044420999,.0488131585372653,.0262834838088483,1.44497609703163e-012,1.2452289980826e-008,1.4835424993805e-006,3.49414974841139e-005,.000338724672636802,.00183369140663752,.00645770791571392,.0160636995405713,.0294159773896932,.0399474500854692,.0384605425083467,.0207091095227104,9.65307073969268e-013,8.31867297341522e-009,9.91071113306241e-007,2.33424447406317e-005,.000226282859139277,.00122498585958624,.00431403062339345,.0107312521172389,.0196511562511815,.026686639477119,.0256933203450701,.0138345886544025,4.25836734304683e-013,3.66970949273344e-009,4.37202313890297e-007,1.02973144060421e-005,9.98226951438086e-005,.000540391748990981,.00190309670559272,.00473399758463879,.00866893492130421,.0117725765312941,.011334382523687,.00610300723148638}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 144; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,-144.1116465,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,-159.0586197,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,-81.71498915,-.2645751311,0.,0.,2.958039892,0.,-22.92947780,.2236067977,0.,-3.324307508,-.1934294858},
			{0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,33.98823325,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,34.53410150,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,14.75000000,0.,0.,-.2645751311,0.,0.,2.958039892,0.,0.,.2236067977,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,-8.500000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,-7.678758582,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,-2.459674775,0.,0.,0.,0.,0.,-.2645751311,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,-254.3227936,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066828,0.,-305.0714286,.9128709293,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,-182.9477455,-.9354143468,0.,0.,8.844691708,0.,-63.03769304,.9035079028,0.,-11.53669556,-.8548456319},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,55.33985905,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,63.17066829,0.,0.,.9128709295,0.,0.,0.,-5.850213671,0.,0.,33.43604921,0.,0.,-.9354143468,0.,0.,8.844691708,0.,0.,.9035079028,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,-12.37436867,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,-13.26806944,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,-5.850213669,0.,0.,0.,0.,0.,-.9354143468,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,-212.3677366,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,-284.1991313,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,-208.0853992,-1.636634177,0.,0.,13.68699853,0.,-93.25932554,2.074812507,0.,-23.01464213,-2.326595118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,40.02082790,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,52.88300766,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,36.47616523,0.,0.,-1.636634177,0.,0.,13.68699853,0.,0.,2.074812507,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,-9.500000000,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,-6.067673910,0.,0.,0.,0.,0.,-1.636634177,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,-106.7569280,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,-158.7252693,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,-141.7297322,-1.218935425,0.,0.,11.40958244,0.,-83.44725167,2.747169230,0.,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,23.88339730,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,21.21428571,0.,0.,-1.218935425,0.,0.,11.40958244,0.,0.,2.747169230,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.627379873,0.,0.,0.,0.,0.,-1.218935425,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,-33.04592764,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,-53.17144007,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,-55.95917640,0.,0.,0.,4.138710979,0.,-42.40452863,1.621428571,0.,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.494338457,0.,0.,0.,0.,0.,4.138710979,0.,0.,1.621428571,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-8.845830912,0.,0.,0.,0.,0.,0.,0.,0.,0.,-10.42167396,0.,0.,0.,0.,0.,-9.706337648,0.,0.,-6.398809524,-2.185359011},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689990,0.,29.50000000,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,16.25653401,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,-9.096005092,-.3240370350,0.,0.,2.070196679,0.,-10.27118455,.3129843186,0.,-3.053571429,-.2961272134},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,-7.850689989,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,-3.162277660,0.,0.,.3162277660,0.,0.,0.,-1.004158022,0.,0.,3.091022771,0.,0.,-.3240370349,0.,0.,2.070196679,0.,0.,.3129843186,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.245365598,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,-1.004158022,0.,0.,0.,0.,0.,-.3240370350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,64.91147819,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,44.99219432,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,-14.04517129,-1.020504077,0.,0.,5.704791219,0.,-27.63572997,1.293725044,0.,-11.16850681,-1.450721144},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,-14.98165545,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,-8.714212528,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,4.964285714,0.,0.,-1.020504077,0.,0.,5.704791219,0.,0.,1.293725044,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,1.558845727,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,-1.565247584,0.,0.,0.,0.,0.,-1.020504077,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,59.18483757,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,51.70172994,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,-2.439471337,-.9983468483,0.,0.,5.709015429,0.,-29.57691077,2.250018896,0.,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,-8.720665110,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,1.773290113,0.,0.,-.9983468480,0.,0.,5.709015429,0.,0.,2.250018896,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5486726408,0.,0.,0.,0.,0.,-.9983468483,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,27.51181564,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,29.34227987,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,6.515483910,0.,0.,0.,1.996693696,0.,-13.84473139,1.554721797,0.,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4979550017,0.,0.,0.,0.,0.,1.996693696,0.,0.,1.554721797,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.221428571,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.471852928,0.,0.,0.,0.,0.,-1.980327354,0.,0.,-4.377552220,-2.322071589},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,-6.851601598,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,-.4433944513,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,4.225804810,-.1463850110,0.,0.,.1636634177,0.,.5511331920,.1855768722,0.,-.8658617402,-.2080969936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,2.020725943,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,-.2500000000,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,-1.152673615,0.,0.,-.1463850109,0.,0.,.1636634177,0.,0.,.1855768722,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,-.1463850110,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,-14.87979615,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,-3.546268517,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,9.681316602,-.2746751327,0.,0.,-.1889822365,0.,3.496332169,.6190476190,0.,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,-2.395787117,0.,0.,-.2746751327,0.,0.,-.1889822365,0.,0.,.6190476190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000,0.,0.,0.,0.,0.,-.2746751328,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,-11.83920042,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,-5.353808485,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,7.368172496,0.,0.,0.,-.6338656908,0.,5.580637822,.6016087651,0.,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.357142857,0.,0.,0.,0.,0.,-.6338656908,0.,0.,.6016087651,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.517439175,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.767857143,0.,0.,0.,0.,0.,2.833614404,0.,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,1.836209865,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,.3415650256,0.,-.6681531049,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,-.9660917834,-.2500000000e-1,0.,0.,-.2236067977,0.,.5443310539,.5634361697e-1,0.,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,.1889822365,0.,0.,-.2500000000e-1,0.,0.,-.2236067977,0.,0.,.5634361697e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1479019946,0.,0.,0.,0.,0.,-.2500000000e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,3.381321241,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.5976143047,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,-2.391701925,0.,0.,0.,-.4500000000,0.,.7607257745,.1070899340,0.,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5070925528,0.,0.,0.,0.,0.,-.4500000000,0.,0.,.1070899340,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1200198396,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.598750131,0.,0.,0.,0.,0.,.3055555556e-1,0.,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,-.5477225576,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,.4065785564,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,.8748177653e-1,0.,0.,0.,-.5669467095e-1,0.,-.2415229459,.7142857143e-2,0.,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1916629694,0.,0.,0.,0.,0.,-.5669467095e-1,0.,0.,.7142857143e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3593680677,0.,0.,0.,0.,0.,-.5197011503,0.,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3419382528,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2207201930,0.,0.,0.,0.,0.,-.8705321812e-1,0.,0.,.1974181422e-1,-.1984126984e-2}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[56][56] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,-.4040610177e-1,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,.4694764777,-11.64130577,0.,0.,-6.507687761,0.,-4.073608438,50.31152948,0.,31.49343956,-228.6336522},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,-.6454972245e-1,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,.7500000000,0.,0.,-11.64130577,0.,0.,-6.507687761,0.,0.,50.31152948,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.1154700539,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,1.341640786,0.,0.,0.,0.,0.,-11.64130577,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709293e-1,0.,.5714285714e-1,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,-.6224437520,10.85080642,0.,0.,7.410417377,0.,5.024466921,-50.14468862,0.,-35.90912479,239.3225832},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,.9128709295e-1,0.,0.,-2.464751509,0.,0.,0.,-1.581138830,0.,0.,-.9596449171,0.,0.,10.85080642,0.,0.,7.410417377,0.,0.,-50.14468862,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,.1632993162,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,10.85080642,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.464751509,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,-.6998542123e-1,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,.6776309270,-6.197388082,0.,0.,-6.026182953,0.,-4.798824078,33.05867928,0.,29.90051531,-173.7315107},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,-.1118033988,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,.9691236663,0.,0.,-6.197388081,0.,0.,-6.026182953,0.,0.,33.05867928,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.2000000000,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,1.290994449,0.,0.,0.,0.,0.,-6.197388082,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,.8081220354e-1,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,-.6455301571,1.974864371,0.,0.,3.507390157,0.,3.710995566,-14.50782199,0.,-18.91064403,91.34707467},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,-.7857142857,0.,0.,1.974864371,0.,0.,3.507390157,0.,0.,-14.50782199,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7267220925,0.,0.,0.,0.,0.,1.974864371,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,-.9035079028e-1,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,.5248906592,0.,0.,0.,-1.266180985,0.,-2.208209791,3.502380952,0.,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5111012519,0.,0.,0.,0.,0.,-1.266180985,0.,0.,3.502380952,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1855768722,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3859879246,0.,0.,0.,0.,0.,.8871666102,0.,0.,-2.267857143,6.356265942},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,.9897433188e-1,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,-.8864412337,-15.76980237,0.,0.,1.794170455,0.,4.167058315,75.89869728,0.,4.660714286,-371.5212018},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,.1581138830,0.,0.,3.320391543,0.,0.,0.,-1.278019301,0.,0.,-1.195584279,0.,0.,-15.76980236,0.,0.,1.794170455,0.,0.,75.89869727,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,.2828427124,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,-1.278019300,0.,0.,0.,0.,0.,-15.76980237,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.320391543,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,-.1212183053,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,.9389529552,8.919961562,0.,0.,-.8451542550e-1,0.,-3.528403268,-49.83237207,0.,-7.640075151,269.1248912},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,-.1936491673,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,1.107142857,0.,0.,8.919961562,0.,0.,-.8451542550e-1,0.,0.,-49.83237207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,.4472135954,0.,0.,0.,0.,0.,8.919961562,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,.1399708425,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,-.8470386588,-2.776822654,0.,0.,-.7563225568,0.,2.151452997,21.71637090,0.,7.490281115,-141.0307509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-.7010681843,0.,0.,-2.776822653,0.,0.,-.7563225568,0.,0.,21.71637090,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,0.,0.,0.,-2.776822654,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,-.1564921593,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,.6060915266,0.,0.,0.,.5237229368,0.,-.8366600265,-5.171408842,0.,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2766416676,0.,0.,0.,0.,0.,.5237229366,0.,0.,-5.171408842,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2642857143,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3365806955,0.,0.,0.,0.,0.,.1400231463,0.,0.,1.482553013,-9.676023611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,-.1564921593,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,.6060915266,-8.002380602,0.,0.,16.80277755,0.,5.405089377,48.86857636,0.,-68.44072363,-278.3505387},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,-.2500000000,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,-.4610694458e-1,0.,0.,-8.002380594,0.,0.,16.80277755,0.,0.,48.86857636,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,-.4472135954,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,-4.041451885,0.,0.,0.,0.,0.,-8.002380602,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,.1807015806,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,-.3936679944,2.387560769,0.,0.,-7.275816105,0.,-5.503272838,-21.01190476,0.,34.70865684,144.9163216},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,.7985957060,0.,0.,2.387560769,0.,0.,-7.275816105,0.,0.,-21.01190476,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.375000000,0.,0.,0.,0.,0.,2.387560770,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,-.2020305089,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,3.548977048,4.876757779,0.,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6428571429,0.,0.,0.,0.,0.,1.774823935,0.,0.,4.876757779,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2083333333,0.,0.,0.,0.,0.,-1.312336745,0.,0.,2.180166278,9.676640803},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,.3415650256,0.,.2138089935,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,.7763237544,-1.575000000,0.,0.,13.75181806,0.,-13.04353288,16.26921940,0.,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,.3415650253,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,3.968626966,0.,0.,-1.575000000,0.,0.,13.75181806,0.,0.,16.26921940,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.922725929,0.,0.,0.,0.,0.,-1.575000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,-.2390457218,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,-1.388730150,0.,0.,0.,-3.450000000,0.,3.286335346,-3.647357166,0.,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8451542550,0.,0.,0.,0.,0.,-3.450000000,0.,0.,-3.647357166,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7535958771,0.,0.,0.,0.,0.,-.1750000000,0.,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,-.2710523708,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,-3.674234614,0.,0.,0.,3.571764270,0.,-16.97561276,2.292857143,0.,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.491618602,0.,0.,0.,0.,0.,3.571764270,0.,0.,2.292857143,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,3.354434698,0.,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.025814759,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.869362509,0.,0.,0.,0.,0.,-5.484352741,0.,0.,-6.337122368,-3.339285714}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,-.5103103630e-1,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,.1035406358,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,-.1189992461,.6614378278e-1,0.,0.,-.7395099729e-1,0.,-.2507429436e-1,-.2795084971e-1,0.,.5467611033e-1,.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,-.5590169942e-1,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,.3227486122e-1,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,.1250000000,0.,0.,.6614378278e-1,0.,0.,-.7395099729e-1,0.,0.,-.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,-.2453738645,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,.5590169942e-1,0.,0.,0.,0.,0.,.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,.9923207755e-1,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709293e-1,0.,-.1638392857,.3651483717,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,.1276586028,-.1403121520,0.,0.,.8964214566e-1,0.,.8558232884e-1,.5646924394e-1,0.,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,.9128709295e-1,0.,0.,.3651483718,0.,0.,0.,.3952847075e-1,0.,0.,-.2525381361,0.,0.,-.1403121520,0.,0.,.8964214566e-1,0.,0.,.5646924394e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,-.2651650429,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.3163924252,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,.3952847074e-1,0.,0.,0.,0.,0.,-.1403121520,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3651483717,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,-.1031197389,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,.1159133539,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,.1364673394e-1,.3218713882,0.,0.,-.3049687729e-2,0.,-.2191913380,-.1210307296,0.,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,.2420614591,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,-.3913118960,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,.3608439183,0.,0.,.3218713881,0.,0.,-.3049687729e-2,0.,0.,-.1210307296,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,.4250000000,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,-.6132223630,0.,0.,0.,0.,0.,.3218713882,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,-.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,.3169353608,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,-.4018262214,.4937160930,0.,0.,-.5387858374,0.,.3603626151,.2888254471,0.,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,-.2259240285,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.3750000000,0.,0.,.4937160930,0.,0.,-.5387858374,0.,0.,.2888254471,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1816805231,0.,0.,0.,0.,0.,.4937160930,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,-.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,.1157619501,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,-.1986565342,0.,0.,0.,-.1582726231,0.,.3313752322,.4377976190,0.,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,0.,0.,0.,-.1582726231,0.,0.,.4377976190,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2412424530e-1,0.,0.,0.,0.,0.,.5544791314e-1,0.,0.,-.1417410714,.3972666214},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,.1718750000,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,-.2157331140,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,.1746928107e-1,.1890216038,0.,0.,-.1725163898e-1,0.,-.7854574205e-1,-.9780759959e-1,0.,.7031250000e-1,.4812067217e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,.1581138830,0.,0.,-.3162277660,0.,0.,0.,-.1141088661,0.,0.,.2916059217e-1,0.,0.,.1890216037,0.,0.,-.1725163898e-1,0.,0.,-.9780759957e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.4592793268,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,-.1141088661,0.,0.,0.,0.,0.,.1890216038,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,-.1786086271,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,.1300571401,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,.5297911641e-1,-.3496171374,0.,0.,-.8451542550e-1,0.,.1102626021,.1796840339,0.,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,.4192627457,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,-.4357106264,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,.8928571429e-2,0.,0.,-.3496171374,0.,0.,-.8451542550e-1,0.,0.,.1796840339,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.3897114318,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,.2795084971,0.,0.,0.,0.,0.,-.3496171374,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,-.3093592167,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,.4264736607,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,-.2555233287,-.6942056633,0.,0.,.3964594048,0.,-.9743899444e-1,-.3527181261,0.,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,.1340277411,0.,0.,-.6942056631,0.,0.,.3964594048,0.,0.,-.3527181261,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,0.,0.,0.,-.6942056633,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,-.9882117688e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,.1613825392,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,-.1546796083,0.,0.,0.,.6546536711e-1,0.,-.8715208612e-2,-.6464261051,0.,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,0.,0.,0.,.6546536709e-1,0.,0.,-.6464261051,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2103629347e-1,0.,0.,0.,0.,0.,.8751446639e-2,0.,0.,.9265956332e-1,-.6047514756},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,-.2305827461,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,-.1467113994e-1,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,-.1210078569,.1341862601,0.,0.,.9547032698e-1,0.,.1373682881,-.1546473936,0.,-.6352789941e-1,.1066497092},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,.5412658775,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,.6250000000e-1,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,-.1959545145,0.,0.,.1341862600,0.,0.,.9547032696e-1,0.,0.,-.1546473936,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,-.3913118960,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,.3608439183,0.,0.,0.,0.,0.,.1341862601,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,-.3993810316,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,.2343473623,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,.3481045690,.5968901920,0.,0.,.1653594569,0.,-.2206772153,.1994047619,0.,.1132938028,-.1630966411},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,-.7216878367e-1,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,-.5190872089,0.,0.,.5968901920,0.,0.,.1653594569,0.,0.,.1994047619,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3437500000,0.,0.,0.,0.,0.,.5968901921,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,-.1275775908,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,.1426360826,0.,0.,0.,.2218529919,0.,-.4548734170,.6095947223,0.,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,0.,0.,0.,.2218529919,0.,0.,.6095947223,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1302083333e-1,0.,0.,0.,0.,0.,-.8202104657e-1,0.,0.,.1362603923,.6047900501},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,-.4725540094,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,.4269562821,0.,-.2839650695,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,.2566181299,-.3937500000,0.,0.,-.4192627457,0.,.3827327723e-1,-.1056442818e-1,0.,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,.4269562817,0.,0.,0.,0.,0.,0.,-.4806814823,0.,0.,.4724555913e-1,0.,0.,-.3937500000,0.,0.,-.4192627457,0.,0.,-.1056442818e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4806814824,0.,0.,0.,0.,0.,-.3937500000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,-.1509518412,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,-.5229125166e-1,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,.3423605578,0.,0.,0.,-.4312500000,0.,.2909776087,-.4559196456,0.,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1056442819,0.,0.,0.,0.,0.,-.4312500000,0.,0.,-.4559196456,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4709974231e-1,0.,0.,0.,0.,0.,-.1093750000e-1,0.,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,-.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,-.3303450769,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,-.2679129406,0.,0.,0.,.4464705337,0.,.1746728448,.2866071429,0.,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3114523254,0.,0.,0.,0.,0.,.4464705337,0.,0.,.2866071429,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1746928107e-1,0.,0.,0.,0.,0.,.2096521686,0.,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1793351568,0.,0.,0.,0.,0.,-.3427720463,0.,0.,-.3960701481,-.2087053571}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[56][56] = 
		{
			{1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,.5103103630e-1,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,-.1035406358,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,.1189992461,-.6614378278e-1,0.,0.,.7395099729e-1,0.,.2507429436e-1,.2795084971e-1,0.,-.5467611033e-1,-.1208934286e-1},
			{0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,.5590169942e-1,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,-.3227486122e-1,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,-.1250000000,0.,0.,-.6614378278e-1,0.,0.,.7395099729e-1,0.,0.,.2795084971e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.2453738645,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,-.6614378278e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,-.9923207755e-1,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709293e-1,0.,.1049107143,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,.4870046022e-1,-.2338535867,0.,0.,.5976143047e-1,0.,.1108137724,.1129384878,0.,-.1024538982,-.5342785200e-1},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,-.9128709295e-1,0.,0.,.4564354646,0.,0.,0.,.1185854122,0.,0.,-.1515228816,0.,0.,-.2338535867,0.,0.,.5976143047e-1,0.,0.,.1129384878,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,.2651650429,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,.5103103630e-1,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,.1185854122,0.,0.,0.,0.,0.,-.2338535867,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4564354646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,.1031197389,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,.4738596230e-1,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,.1162325270,-.4091585443,0.,0.,-.1189378215,0.,.3433564566e-1,.2593515633,0.,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,-.2420614591,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,-.1677050983,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,.3092947872e-1,0.,0.,-.4091585442,0.,0.,-.1189378215,0.,0.,.2593515633,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,.3750000000,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,-.4091585443,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,.1786086271,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,.2133947249,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,.3504772316e-1,-.3047338563,0.,0.,.6338656908e-1,0.,-.1639473093e-1,.3433961537,0.,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,-.2581988897,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.2857142857,0.,0.,-.3047338563,0.,0.,.6338656908e-1,0.,0.,.3433961537,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6568449682,0.,0.,0.,0.,0.,-.3047338563,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,.5705443310e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,.1552904208,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,.2642678667,0.,0.,0.,.5173388725,0.,.1832986642,.2026785714,0.,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6867923071,0.,0.,0.,0.,0.,.5173388725,0.,0.,.2026785714,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5528644321,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6513546229,0.,0.,0.,0.,0.,-.6066461029,0.,0.,-.3999255952,-.1365849382},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,-.1093750000,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,.2095472184,.1581138830,0.,0.,0.,.2966830521,0.,0.,.2916059217e-1,0.,-.2131252291,-.8100925875e-1,0.,0.,-.1035098339,0.,.8090801999e-1,.3912303983e-1,0.,.2343750000e-1,-.1850795084e-1},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,-.1825741858,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.1581138830,0.,0.,0.,.2966830520,0.,0.,.2916059217e-1,0.,0.,-.8100925872e-1,0.,0.,-.1035098339,0.,0.,.3912303982e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,.2551551816,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,-.4065863991,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,.2966830519,0.,0.,0.,0.,0.,-.8100925875e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,.2296396634,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,-.1224809960,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,-.1149239294,-.2551260193,0.,0.,-.4278593414,0.,-.8679563432e-2,.1617156305,0.,.1877213121,-.9067007150e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,-.5590169942e-1,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,.4841229182e-1,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,.2767857143,0.,0.,-.2551260193,0.,0.,-.4278593414,0.,0.,.1617156305,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,.6149186937,0.,0.,0.,0.,0.,-.2551260193,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,-.7954951286e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,-.6561133241e-2,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,-.1143502189,-.2495867120,0.,0.,-.4025587803,0.,-.7962157829e-1,.2812523620,0.,.4211073418,-.2128834533},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,.3354101966,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.3917733971,0.,0.,-.2495867119,0.,0.,-.4025587803,0.,0.,.2812523620,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1371681602,0.,0.,0.,0.,0.,-.2495867120,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,-.4150489429,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,-.4254630581,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,.6629126072e-1,0.,0.,0.,.2495867120,0.,.5017470100,.1943402247,0.,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,0.,0.,0.,.2495867119,0.,0.,.1943402247,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4513392857,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2169908080,0.,0.,0.,0.,0.,-.1237704596,0.,0.,-.2735970138,-.1451294743},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,.3294039230e-1,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,-.1189992461,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,.1967692977,-.3659625275e-1,0.,0.,-.1432054905,0.,-.1672490034,.2319710903e-1,0.,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,-.5625000000,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,.3803822929,0.,0.,-.3659625271e-1,0.,0.,-.1432054905,0.,0.,.2319710903e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,-.5031152948,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,.2165063510,0.,0.,0.,0.,0.,-.3659625275e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,-.3309157119,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,.1835250428,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,.3043636809,-.6866878320e-1,0.,0.,-.3307189139,0.,-.4995787124,.7738095238e-1,0.,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,.3992978531,0.,0.,-.6866878320e-1,0.,0.,-.3307189139,0.,0.,.7738095238e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1562500000,0.,0.,0.,0.,0.,-.6866878321e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,.4847948451,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,.1073287078,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,-.4686614144,0.,0.,0.,-.7923321136e-1,0.,-.8518830777e-1,.7520109565e-1,0.,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1696428571,0.,0.,0.,0.,0.,-.7923321136e-1,0.,0.,.7520109565e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1573399484,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1104910714,0.,0.,0.,0.,0.,.1771009002,0.,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,.4185478369,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,-.3415650256,0.,-.5445447805,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,.3860054223,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,-.1726550062,.7042952119e-2,0.,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,-.3415650253,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,.1889822365,0.,0.,-.6250000000e-2,0.,0.,-.5590169942e-1,0.,0.,.7042952119e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,0.,0.,0.,-.6250000000e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,-.3320940506,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,.1718141126,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,.1977011672,0.,0.,0.,-.5625000000e-1,0.,-.2605485777,.1338624175e-1,0.,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813079e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,0.,0.,0.,-.5625000000e-1,0.,0.,.1338624175e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9992188321e-1,0.,0.,0.,0.,0.,.1909722222e-2,0.,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,.1711632992,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,-.2117596647,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,.1366902758,0.,0.,0.,-.7086833870e-2,0.,-.5391137183e-1,.8928571429e-3,0.,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,0.,0.,0.,-.7086833870e-2,0.,0.,.8928571429e-3,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2246050424e-1,0.,0.,0.,0.,0.,-.3248132189e-1,0.,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1379501207e-1,0.,0.,0.,0.,0.,-.5440826132e-2,0.,0.,.1233863389e-2,-.1240079365e-3}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[56][56] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.2357022604e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-2,.5892556509e-1,.1178511302e-1,.3928371006e-2,.3535533905e-1,.5892556509e-2,.2357022604e-1,.2357022604,.7856742012e-1,.3928371006e-1,.2357022604e-1,.7856742012e-1,.1964185503e-1,.7856742012e-2,.3928371006e-1,.7856742012e-2,.2357022604e-1,.1767766952,.5892556509e-1,.2946278254e-1,.5892556509e-1,.1473139128e-1,.2946278254e-1,.1414213562,.4714045206e-1,.4714045206e-1,.1178511302},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.1360827635e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-2,.3402069088e-1,.6804138176e-2,.2268046058e-2,.2041241452e-1,.3402069088e-2,.1360827635e-1,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.1837117307,.6123724358e-1,.3061862179e-1,.6123724358e-1,.1530931089e-1,.3061862179e-1,.1632993162,.5443310539e-1,.5443310539e-1,.1458029609},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.5270462768e-2,.1756820922e-1,.4392052306e-2,.1756820922e-2,.8784104612e-2,.1756820922e-2,.5270462768e-2,.7905694150e-1,.2635231383e-1,.1317615692e-1,.2635231383e-1,.6588078457e-2,.1317615692e-1,.9035079028e-1,.3011693010e-1,.3011693010e-1,.9411540654e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.2227177016e-2,.4454354030e-2,.1113588508e-2,.2227177016e-2,.2672612419e-1,.8908708064e-2,.8908708064e-2,.3711961694e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2,.1122391716e-2,.1122391716e-2,.8417937869e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8460345889e-3},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,-.1904761905e-1,.8333333333e-1,.8333333333e-2,0.,-.1190476190e-2,.6666666667e-1,.8333333333e-2,.1587301587e-2,.5000000000e-1,.5952380952e-2,.3809523810e-1,0.,-.2777777778e-1,-.2222222222e-1,-.1666666667e-1,.5555555556e-1,.5555555556e-2,0.,.4444444444e-1,.5555555556e-2,.3333333333e-1,0.,-.2083333333e-1,-.1666666667e-1,.4166666667e-1,.4166666667e-2,.3333333333e-1,0.,-.1666666667e-1,.3333333333e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,-.1099714799e-1,.4811252245e-1,.4811252245e-2,0.,-.6873217492e-3,.3849001795e-1,.4811252245e-2,.9164289989e-3,.2886751347e-1,.3436608746e-2,.2199429598e-1,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2165063510e-1,-.1732050808e-1,.4330127020e-1,.4330127020e-2,.3464101616e-1,0.,-.1924500898e-1,.3849001795e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,-.3726779962e-2,.1242259987e-1,.1242259987e-2,0.,.9938079897e-2,.1242259987e-2,.7453559923e-2,0.,-.9316949905e-2,-.7453559923e-2,.1863389981e-1,.1863389981e-2,.1490711985e-1,0.,-.1064794275e-1,.2129588550e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,-.1259881577e-2,.3149703940e-2,.3149703940e-3,.2519763153e-2,0.,-.3149703940e-2,.6299407883e-2,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3968253968e-3,.7936507936e-3,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1,0.,0.,.6804138176e-2,.8748177653e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,-.2187044413e-2,.2041241452e-1,-.2916059217e-2,-.1944039478e-2,.2624453295e-1,0.,.2624453295e-1,0.,0.,.4536092117e-2,.5832118436e-2,0.,-.6804138176e-2,-.3240065798e-2,.1360827635e-1,-.1944039478e-2,.1749635531e-1,0.,0.,.3402069088e-2,0.,-.5103103630e-2,.1020620726e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,-.1262690680e-2,.1178511302e-1,-.1683587573e-2,-.1122391716e-2,.1515228816e-1,0.,.1515228816e-1,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3535533905e-2,0.,-.5303300858e-2,.1060660172e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,.1304101327e-2,0.,-.1521451549e-2,-.7245007376e-3,.3042903097e-2,-.4347004424e-3,.3912303982e-2,0.,0.,.1521451549e-2,0.,-.2282177323e-2,.4564354646e-2,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2571722499e-3,0.,-.3857583749e-3,.7715167495e-3,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1,0.,0.,0.,-.1262690680e-2,-.2244783431e-2,0.,0.,.1683587573e-2,.1543288609e-2,0.,-.2525381361e-2,-.3741305720e-3,.5050762721e-2,-.1964185503e-2,.8979133727e-2,0.,0.,0.,-.8417937869e-3,0.,0.,.1122391716e-2,0.,-.1683587573e-2,.3367175148e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,-.1296026319e-2,0.,0.,.9720197392e-3,.8910180945e-3,0.,-.1458029609e-2,-.2160043865e-3,.2916059217e-2,-.1134023029e-2,.5184105275e-2,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1882308131e-3,0.,0.,.2509744174e-3,0.,-.3764616260e-3,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2,0.,0.,0.,0.,.2509744174e-3,0.,0.,0.,-.3137180218e-3,0.,0.,.4182906958e-3,0.,-.6274360436e-3,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1449001475e-3,0.,0.,0.,-.1811251844e-3,0.,0.,.2415002458e-3,0.,-.3622503687e-3,.7245007376e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,.3299144397e-1,0.,.1443375673e-1,.9622504490e-2,.6185895742e-2,0.,.4811252245e-2,.2749286996e-2,0.,.2061965247e-2,0.,0.,.4811252245e-1,.3849001795e-1,.2886751347e-1,0.,.9622504490e-2,.6415002993e-2,0.,.3207501497e-2,0.,0.,.3608439183e-1,.2886751347e-1,0.,.7216878367e-2,0.,0.,.2886751347e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,.1904761905e-1,0.,.8333333333e-2,.5555555556e-2,.3571428571e-2,0.,.2777777778e-2,.1587301587e-2,0.,.1190476190e-2,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.3750000000e-1,.3000000000e-1,0.,.7500000000e-2,0.,0.,.3333333333e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,.6454972245e-2,0.,.2151657415e-2,.1434438276e-2,0.,.7172191382e-3,0.,0.,.1613743061e-1,.1290994449e-1,0.,.3227486122e-2,0.,0.,.1844277784e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,.2182178902e-2,0.,.5455447254e-3,0.,0.,.5455447254e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6873217492e-3,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,.1894036020e-2,0.,.6734350295e-2,.2946278254e-2,0.,.3788072042e-2,0.,0.,0.,-.7856742012e-2,-.1010152544e-1,0.,.7856742012e-2,.3367175148e-2,0.,.4489566864e-2,0.,0.,0.,-.5892556509e-2,0.,.5892556509e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,.1093522207e-2,0.,.3888078956e-2,.1701034544e-2,0.,.2187044413e-2,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6123724358e-2,0.,.6123724358e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,-.2258769757e-2,0.,.1756820922e-2,.7529232524e-3,0.,.1003897670e-2,0.,0.,0.,-.2635231383e-2,0.,.2635231383e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4454354030e-3,0.,.4454354030e-3,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.,0.,0.,0.,.2187044413e-2,.3888078956e-2,0.,0.,-.2430049348e-2,-.2187044413e-2,0.,.2430049348e-2,0.,0.,.2430049348e-2,0.,0.,0.,0.,.1458029609e-2,0.,0.,-.1620032899e-2,0.,.1620032899e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,.2244783431e-2,0.,0.,-.1402989645e-2,-.1262690680e-2,0.,.1402989645e-2,0.,0.,.1402989645e-2,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3260253318e-3,0.,0.,-.3622503687e-3,0.,.3622503687e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.,0.,0.,0.,0.,-.4347004424e-3,0.,0.,0.,.4890379978e-3,0.,0.,-.5433755531e-3,0.,.5433755531e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2509744174e-3,0.,0.,0.,.2823462197e-3,0.,0.,-.3137180218e-3,0.,.3137180218e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,.1956151991e-1,0.,0.,.2173502212e-2,.2445189989e-2,0.,0.,.5433755531e-3,0.,0.,0.,0.,0.,.1014301032e-1,.1304101327e-1,0.,0.,.1449001475e-2,0.,0.,0.,0.,0.,.7607257744e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,.1129384878e-1,0.,0.,.1254872087e-2,.1411731098e-2,0.,0.,.3137180218e-3,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.7905694150e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,.2916059217e-2,0.,0.,.3240065798e-3,0.,0.,0.,0.,0.,.3402069088e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5750546329e-3,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,-.2823462197e-2,-.5019488348e-2,0.,0.,.1882308131e-2,.1568590109e-2,0.,0.,.8365813917e-3,0.,0.,0.,0.,0.,0.,-.1882308131e-2,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,-.2898002950e-2,0.,0.,.1086751106e-2,.9056259216e-3,0.,0.,.4830004916e-3,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4208968934e-3,0.,0.,.2805979290e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.,0.,0.,0.,0.,.5611958579e-3,0.,0.,0.,-.4910463757e-3,0.,0.,.3273642505e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3240065798e-3,0.,0.,0.,-.2835057572e-3,0.,0.,.1890038382e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,.5939138708e-2,0.,0.,0.,.3711961694e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2227177016e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,.3428963332e-2,0.,0.,0.,.2143102083e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4980119205e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6640158940e-3,0.,0.,0.,.3320079470e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3833697554e-3,0.,0.,0.,.1916848777e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7529232524e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4347004424e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[56][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.,3.464101616,-4.472135954,5.291502622,-6.,2.449489743,-4.242640686,5.477225575,-6.480740698,-2.828427124,4.898979486,-6.324555320,3.162277660,-5.477225575,-3.464101616,-3.464101616,6.,-7.745966692,9.165151390,-10.39230485,4.242640686,-7.348469229,9.486832980,-11.22497216,-4.898979486,8.485281372,-10.95445115,5.477225575,-9.486832980,-6.,5.477225575,-9.486832980,12.24744872,-14.49137675,-6.324555320,10.95445115,-14.14213562,7.071067810,-12.24744872,-7.745966692,-7.483314774,12.96148140,-16.73320053,8.366600265,-14.49137675,-9.165151390,9.486832980,-16.43167672,-10.39230485,-11.48912529},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,140.7124728,0.,-6.928203232,26.83281572,-63.49803146,120.,0.,8.485281372,-32.86335345,77.76888838,0.,-9.797958972,37.94733192,0.,10.95445115,0.,0.,-12.,46.47580015,-109.9818167,207.8460970,0.,14.69693846,-56.92099788,134.6996659,0.,-16.97056274,65.72670690,0.,18.97366596,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-21.90890230,84.85281372,0.,24.49489743,0.,0.,-25.92296279,100.3992032,0.,28.98275350,0.,0.,32.86335345,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,-984.9873096,0.,0.,-26.83281572,158.7450787,-540.,0.,0.,32.86335345,-194.4222209,0.,0.,-37.94733192,0.,0.,0.,0.,0.,-46.47580015,274.9545417,-935.3074363,0.,0.,56.92099788,-336.7491648,0.,0.,-65.72670690,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-84.85281372,0.,0.,0.,0.,0.,-100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,2626.632826,0.,0.,0.,-105.8300524,840.,0.,0.,0.,129.6148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,1454.922679,0.,0.,0.,224.4994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,-2954.961929,0.,0.,0.,0.,-420.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1181.984772,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,18.,-19.59591794,33.94112549,-43.81780460,51.84592558,42.42640686,-73.48469229,94.86832980,-75.89466384,131.4534138,121.2435566,3.464101616,-6.,7.745966692,-9.165151390,10.39230485,-25.45584412,44.09081537,-56.92099788,67.34983297,63.68673332,-110.3086578,142.4078650,-120.4989626,208.7103256,198.,-10.95445115,18.97366596,-24.49489743,28.98275350,56.92099788,-98.59006035,127.2792206,-127.2792206,220.4540769,224.6330341,22.44994432,-38.88444419,50.19960159,-100.3992032,173.8965210,210.7984820,-37.94733192,65.72670690,155.8845727,57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,-360.,0.,-67.88225098,262.9068276,-622.1511070,0.,146.9693846,-569.2099788,0.,-262.9068276,0.,0.,12.,-46.47580015,109.9818167,-207.8460970,0.,-88.18163075,341.5259873,-808.1979956,0.,220.6173157,-854.4471897,0.,-417.4206511,0.,0.,-37.94733192,146.9693846,-347.7930420,0.,197.1801207,-763.6753235,0.,-440.9081537,0.,0.,77.76888838,-301.1976095,0.,-347.7930420,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,1620.,0.,0.,-262.9068276,1555.377768,0.,0.,569.2099788,0.,0.,0.,0.,0.,46.47580015,-274.9545417,935.3074363,0.,0.,-341.5259873,2020.494989,0.,0.,854.4471897,0.,0.,0.,0.,0.,-146.9693846,869.4826050,0.,0.,763.6753235,0.,0.,0.,0.,0.,301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,-2520.,0.,0.,0.,-1036.918512,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,-1454.922679,0.,0.,0.,-1346.996659,0.,0.,0.,0.,0.,0.,0.,0.,0.,-579.6550700,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1260.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,727.4613394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-64.80740698,-127.2792206,220.4540769,-284.6049894,398.4469852,-690.1304224,-969.9484525,0.,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-56.12486080,-161.6663230,280.0142853,-361.4968880,575.1086854,-996.1174629,-1488.,5.477225575,-9.486832980,12.24744872,-14.49137675,-94.86832980,164.3167672,-212.1320343,487.9036789,-845.0739613,-1471.733671,-22.44994432,38.88444419,-50.19960159,250.9980080,-434.7413025,-1081.487864,56.92099788,-98.59006035,-519.6152424,-114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,777.6888838,0.,-440.9081537,1707.629936,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,673.4983297,0.,-560.0285706,2168.981328,0.,1992.234926,0.,0.,18.97366596,-73.48469229,173.8965210,0.,-328.6335345,1272.792206,0.,1690.147923,0.,0.,-77.76888838,301.1976095,0.,869.4826050,0.,0.,197.1801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,-1944.222209,0.,0.,-1707.629936,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,-1683.745824,0.,0.,-2168.981328,0.,0.,0.,0.,0.,73.48469229,-434.7413025,0.,0.,-1272.792206,0.,0.,0.,0.,0.,-301.1976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1296.148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1122.497216,0.,0.,0.,0.,0.,0.,0.,0.,0.,289.8275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,221.3594362,-708.3501958,1226.898529,2909.845357,0.,0.,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,230.0434742,-920.1738966,1593.787941,4176.,0.,0.,0.,0.,44.27188724,-76.68115805,98.99494934,-622.2539673,1077.775487,3578.636612,7.483314774,-12.96148140,16.73320053,-234.2648074,405.7585490,2071.324214,-37.94733192,65.72670690,727.4613394,114.8912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,-1328.156617,0.,-2453.797058,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,-1380.260845,0.,-3187.575881,0.,0.,0.,0.,0.,0.,153.3623161,-593.9696960,0.,-2155.550974,0.,0.,25.92296279,-100.3992032,0.,-811.5170980,0.,0.,-131.4534138,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1328.156617,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,593.9696960,0.,0.,0.,0.,0.,100.3992032,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,-690.1304224,-3637.306697,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,-796.8939703,-4860.,0.,0.,0.,0.,0.,0.,0.,254.5584412,-440.9081537,-3601.874512,0.,0.,0.,75.29940238,-130.4223908,-1695.553007,9.486832980,-16.43167672,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1380.260845,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1593.787941,0.,0.,0.,0.,0.,0.,0.,0.,0.,881.8163075,0.,0.,0.,0.,0.,260.8447815,0.,0.,32.86335345,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1980.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1278.084504,0.,0.,0.,0.,0.,504.0833264,0.,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,20.78460970,-8.485281372,14.69693846,-18.97366596,22.44994432,9.797958972,-16.97056274,21.90890230,-10.95445115,18.97366596,12.,-32.86335345,56.92099788,-73.48469229,86.94826050,37.94733192,-65.72670690,84.85281372,-42.42640686,73.48469229,46.47580015,89.79977729,-155.5377768,200.7984064,-100.3992032,173.8965210,109.9818167,-189.7366596,328.6335345,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,-415.6921939,0.,-29.39387692,113.8419958,-269.3993319,0.,33.94112549,-131.4534138,0.,-37.94733192,0.,0.,-113.8419958,440.9081537,-1043.379126,0.,131.4534138,-509.1168823,0.,-146.9693846,0.,0.,311.0755535,-1204.790438,0.,-347.7930420,0.,0.,-657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,1870.614873,0.,0.,-113.8419958,673.4983297,0.,0.,131.4534138,0.,0.,0.,0.,0.,-440.9081537,2608.447815,0.,0.,509.1168823,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,-2909.845357,0.,0.,0.,-448.9988864,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1454.922679,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-112.2497216,-117.5755077,203.6467529,-262.9068276,230.0434742,-398.4469852,-384.,32.86335345,-56.92099788,73.48469229,-86.94826050,-303.5786554,525.8136552,-678.8225098,721.2489166,-1249.239769,-1301.322404,-179.5995546,311.0755535,-401.5968127,1104.391235,-1912.861731,-2419.599967,569.2099788,-985.9006035,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,1346.996659,0.,-407.2935059,1577.440966,0.,796.8939703,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-1051.627310,4072.935059,0.,2498.479538,0.,0.,-622.1511070,2409.580876,0.,3825.723462,0.,0.,1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,-3367.491648,0.,0.,-1577.440966,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-4072.935059,0.,0.,0.,0.,0.,-2409.580876,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2244.994432,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,460.0869483,-920.1738966,1593.787941,2592.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-2206.173157,3821.203999,7529.079625,89.79977729,-155.5377768,200.7984064,-1907.584860,3304.033899,10558.25440,-569.2099788,985.9006035,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,-2760.521690,0.,-3187.575881,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-7642.407998,0.,0.,311.0755535,-1204.790438,0.,-6608.067798,0.,0.,-1971.801207,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2760.521690,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,1204.790438,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-1593.787941,-5760.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-13942.74005,0.,0.,0.,903.5928286,-1565.068689,-14297.63617,189.7366596,-328.6335345,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3187.575881,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,3130.137378,0.,0.,657.2670690,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,6048.999917,0.,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-86.94826050,-37.94733192,65.72670690,-84.85281372,42.42640686,-73.48469229,-46.47580015,-224.4994432,388.8444419,-501.9960159,250.9980080,-434.7413025,-274.9545417,853.8149682,-1478.850905,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,1043.379126,0.,-131.4534138,509.1168823,0.,146.9693846,0.,0.,-777.6888838,3011.976095,0.,869.4826050,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,-2608.447815,0.,0.,-509.1168823,0.,0.,0.,0.,0.,-3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1738.965210,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,593.9696960,-678.8225098,1175.755077,1254.846604,224.4994432,-388.8444419,501.9960159,-2509.980080,4347.413025,5774.045376,-1707.629936,2957.701810,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-3563.818176,0.,-2351.510153,0.,0.,777.6888838,-3011.976095,0.,-8694.826050,0.,0.,-5915.403621,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3563.818176,0.,0.,0.,0.,0.,3011.976095,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-2645.448922,-6274.233021,0.,0.,0.,2258.982072,-3912.671722,-20621.59063,853.8149682,-1478.850905,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5290.897845,0.,0.,0.,0.,0.,7825.343445,0.,0.,2957.701810,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,0.,0.,0.,15122.49979,0.,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,334.6640106,-167.3320053,289.8275350,183.3030278,-1328.156617,2300.434742,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,-2007.984064,0.,-579.6550700,0.,0.,-4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2007.984064,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-2608.447815,-3666.060556,1328.156617,-2300.434742,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5216.895630,0.,0.,4600.869483,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-1150.217371,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2300.434742,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][56] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[56][56] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{6.633249580,0.,14.83239697,0.,19.89974874,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[52][56] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,0.,0.,0.,0.,0.,-1.360827635,-2.357022604,-3.042903097,-3.600411499,.8380524812,1.451549477,1.873942317,.8198497638,1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,-2.108185107,-2.721655270,-3.220305944,-1.873942317,-3.245763304,-4.190262406,.5237828008,.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,1.217161239,2.052800957,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,-1.704828602,-2.952849755,-3.812112642,-4.510552507,-5.114485804,-3.104623248,-5.377365205,-6.942148628,-8.214061030,-.5462559154,-.9461429991,-1.221465360,2.694797933,4.667526936,1.022426944,1.298058152,2.248302670,2.902546266,3.434339057,3.719687555,6.442687835,8.317474224,3.264192666,5.653747546,-1.221656347,-.8344196900,-1.445257298,-1.865819148,-3.243190051,-5.617369949,-4.760960496,.4850599026,.8401483961,2.375793728,-.2614753427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.984283251,1.704828602,2.200924260,2.604168704,2.952849756,-1.237409468,-2.143256068,-2.766931687,-3.273877723,-1.032732859,-1.788745783,-2.309260876,1.210247240,2.096209709,1.042423382,1.704828601,2.952849754,3.812112640,4.510552505,5.114485802,3.104623246,5.377365203,6.942148625,8.214061027,.5462559149,.9461429982,1.221465359,-2.694797932,-4.667526936,-1.022426944,1.298058156,2.248302677,2.902546276,3.434339069,3.719687568,6.442687856,8.317474253,3.264192678,5.653747565,-1.221656351,.8344196802,1.445257281,1.865819127,3.243190014,5.617369884,4.760960437,.4850600816,.8401487061,2.375794604,.2614758325},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.968566503,-3.409657202,-4.401848518,-5.208337406,-5.905699510,2.347500666,4.065990423,5.249171065,6.210902963,-2.609637841,-4.520025331,-5.835327609,2.775506203,4.807317761,-2.855122567,-.3360178568e-9,-.582e-9,-.7513587691e-9,-.8890196848e-9,-.1008053570e-8,.4007561299e-9,.6941299786e-9,.8961179490e-9,.1060301056e-8,-.4456004436e-9,-.7718026080e-9,-.9963928825e-9,.4740581912e-9,.8210928729e-9,-.4878407244e-9,-2.709993221,-4.693845947,-6.059729062,-7.169968120,3.014474194,5.221222463,6.740569213,-3.208782240,-5.557773872,3.304507380,-.7333648479e-8,-.1270225177e-7,-.1639853652e-7,.7812670753e-8,.1353194269e-7,-.8054303853e-8,3.483595229,6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,-.8327988349,-1.442449894,-1.862194806,-2.203378609,-2.498396504,-1.088657954,-1.885610890,-2.434313191,-2.880318211,.4375794197,.7579097871,.9784573278,.9743405328,1.687607307,-.5933186699,-.4672884697,-.8093673712,-1.044888783,-1.236329082,-1.022115558,-1.770356077,-2.285519867,-.3251247057,-.5631325093,.8276510476,.6684136118,1.157726336,1.494618273,2.033596264,3.522292053,1.746351412,-.1476726503,-.2557765332,-.5739768229,-.2050386800},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.480816631,.8327988360,1.075138674,1.272121234,1.442449896,-1.465249839,-2.537887167,-3.276398245,-3.876686683,-.4816921701e-1,-.8343153125e-1,-.1077096437,1.474494615,2.553899588,-.4344958536,.8327988342,1.442449893,1.862194804,2.203378607,2.498396503,1.088657953,1.885610888,2.434313189,2.880318209,-.4375794194,-.7579097864,-.9784573273,-.9743405322,-1.687607306,.5933186693,-.4672884647,-.8093673627,-1.044888772,-1.236329069,-1.022115547,-1.770356058,-2.285519843,-.3251247023,-.5631325034,.8276510391,-.6684136118,-1.157726336,-1.494618273,-2.033596264,-3.522292053,-1.746351412,-.1476726019,-.2557764494,-.5739766349,.2050384112},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.961633263,-1.665597669,-2.150277344,-2.544242465,-2.884899788,-.2081759125,-.3605712573,-.4654954917,-.5507816934,1.215013765,2.104465573,2.716853372,-1.352704319,-2.342952608,.5364099445,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.1039230485e-8,.5714598801e-10,.9897975472e-10,.1277823139e-9,.1511940727e-9,.2193666394e-9,.3799541647e-9,.4905187176e-9,-.3033617129e-9,-.5254378997e-9,.1563891368e-9,-1.872739842,-3.243680556,-4.187573592,-4.954803895,-.4572029592,-.7918987547,-1.022336896,1.670078386,2.892660618,-1.283464460,0.,0.,0.,0.,0.,-0.,1.663589468,2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,-1.425047564,-2.468254783,-3.186503223,-3.770321460,-4.275142691,-2.360183599,-4.087957910,-5.277530967,-6.244458853,-.4945548803e-2,-.8565941795e-2,-.1105858331e-1,2.175490654,3.768060344,.1537427033,.6224210815,1.078064937,1.391775849,1.646771393,1.648153002,2.854684738,3.685382148,1.157573982,2.004976951,-.8770258842,-.7554460893e-2,-.1308471009e-1,-.1689228809e-1,-.2731590581e-1,-.4731253672e-1,-.3482606398e-1,-.2198338910,-.3807634685,-1.005417380,.1997404621},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.822751595,1.425047563,1.839728492,2.176796108,2.468254782,-1.348091029,-2.334962155,-3.014423181,-3.566713607,-.7313658101,-1.266762742,-1.635383668,1.406175430,2.435567290,.5915326939,1.425047564,2.468254784,3.186503224,3.770321460,4.275142693,2.360183599,4.087957910,5.277530967,6.244458853,.4945548803e-2,.8565941795e-2,.1105858331e-1,-2.175490654,-3.768060344,-.1537427033,.6224210804,1.078064935,1.391775847,1.646771390,1.648152999,2.854684733,3.685382143,1.157573980,2.004976947,-.8770258827,.7554451164e-2,.1308469324e-1,.1689226633e-1,.2731587063e-1,.4731247580e-1,.3482601912e-1,-.2198339029,-.3807634890,-1.005417434,-.1997405113},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.645503189,-2.850095128,-3.679456987,-4.353592220,-4.936509566,1.377214242,2.385405038,3.079544664,3.643766386,-.7456423653,-1.291490461,-1.667307016,-.7214030405e-1,-.1249506719,.8704037330,.2424871131e-9,.42e-9,.5422176684e-9,.6415605973e-9,.7274613394e-9,-.2092513633e-9,-.3624339929e-9,-.4679002728e-9,-.5536270690e-9,.1249332326e-9,.2163907064e-9,.2793592008e-9,-.1348127464e-10,-.2335025263e-10,-.9727958508e-10,-2.424562843,-4.199466029,-5.421487332,-6.414790321,1.641768329,2.843626160,3.671105586,-.5644916125,-.9777281536,-.5417963103,-.7483314774e-9,-.1296148140e-8,-.1673320053e-8,.3917700593e-9,.6785656478e-9,-.9473764217e-11,2.788419758,4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,.9641937137,1.670032500,2.156002687,2.551016781,2.892581141,.2394478538,.4147358487,.5354216783,.6335194733,-.9155822076,-1.585834902,-2.047304055,.5490545322,.9509903459,.3754286701,-.9430172903,-1.633353859,-2.108650765,-2.494989233,-.7446746083,-1.289814257,-1.665143045,.9321620643,1.614552057,.2692350424e-1,-1.398574920,-2.422402819,-3.127308592,-1.821631846,-3.155158911,1.076451798,-.1570858835,-.2720807313,-.2832576764,.8954348170},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.556677500,-.9641937137,-1.244768731,-1.472830225,-1.670032500,-.8469574563,-1.466973346,-1.893854446,-2.240838800,1.390251875,2.407986883,3.108697698,-.5638685582,-.9766489917,-.8112194737,-.9641937137,-1.670032500,-2.156002687,-2.551016781,-2.892581141,-.2394478538,-.4147358487,-.5354216783,-.6335194733,.9155822076,1.585834902,2.047304055,-.5490545322,-.9509903459,-.3754286701,-.9430172892,-1.633353857,-2.108650763,-2.494989230,-.7446746070,-1.289814255,-1.665143042,.9321620636,1.614552056,.2692350421e-1,1.398574927,2.422402832,3.127308609,1.821631855,3.155158927,-1.076451804,-.1570858218,-.2720806245,-.2832575652,-.8954346999},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.113355001,1.928387427,2.489537463,2.945660451,3.340065000,-1.123448021,-1.945869052,-2.512106144,-2.972364074,-1.252806288,-2.169924144,-2.801360023,.9835692099,1.703591844,1.344816036,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.1039230485e-8,.6764731475e-9,.1171685862e-8,.1512639943e-8,.1789779717e-8,.2094450706e-9,.3627695035e-9,.4683334153e-9,-.5392014460e-9,-.9339242999e-9,-.350864418e-9,-.6338914369,-1.097932175,-1.417424343,-1.677119101,-1.926686150,-3.337118302,-4.308201201,-1.956429599,-3.388635469,.2565449507,.7483314774e-8,.1296148140e-7,.1673320053e-7,.3070569490e-7,.5318382366e-7,.4977473171e-7,.1905991076,.3301273382,.9829492714,0.},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,1.926395974,3.336615703,4.307552349,5.096764673,5.779187924,-.6138328492,-1.063189683,-1.372571978,-1.624049066,-.8603944811,-1.490246955,-1.923900547,1.645753812,2.850529219,-1.343411045,.5526193998,.9571648777,1.235694544,1.462093502,.2281916739e-1,.3952395730e-1,.5102520946e-1,-.4132700278,-.7158046857,.4485373028,-1.314274316,-2.276389891,-2.938806712,-.4873895555,-.8441834734,1.215532077,-2.142609188,-3.711107975,-1.473102430,-1.425504294},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.112205235,-1.926395976,-2.486966510,-2.942618459,-3.336615704,.8625243345e-1,.1493935970,.1928663044,.2282024888,.9832929299,1.703113314,2.198709833,-1.447727447,-2.507537494,1.036438633,-1.926395976,-3.336615704,-4.307552352,-5.096764678,-5.779187926,.6138328496,1.063189683,1.372571979,1.624049067,.8603944811,1.490246955,1.923900547,-1.645753812,-2.850529220,1.343411045,.5526194031,.9571648834,1.235694551,1.462093511,.2281916752e-1,.3952395754e-1,.5102520976e-1,-.4132700304,-.7158046901,.4485373055,1.314274282,2.276389831,2.938806635,.4873895427,.8441834514,-1.215532046,-2.142609204,-3.711108002,-1.473102441,1.425504133},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.224410470,3.852791952,4.973933020,5.885236917,6.673231410,.7950455559,1.377059297,1.777775908,2.103492822,-1.500452022,-2.598859138,-3.355112719,-2.072134551,-3.589042323,-.2431355237,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-0.,-.2398353968,-.4154070926,-.5362882506,-.6345448156,-1.087945327,-1.884376583,-2.432719707,-2.347136393,-4.065359486,-2.942787847,0.,0.,0.,0.,0.,0.,.2728458417e-1,.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,2.679844379,4.641626620,5.992314198,7.090201776,8.039533138,-2.043694514,-3.539782735,-4.569839860,-5.407107442,.8110805448,1.404832712,1.813631233,.6037611818,1.045745042,-1.765469345,2.576012736,4.461784940,5.760139589,6.815489076,-1.403212584,-2.430435489,-3.137678723,-.8058635216e-2,-.1395796564e-1,1.237350351,1.238945990,2.145917403,2.770367454,-.4443853780,-.7696980531,-.2975289326,-.7339728352,-1.271278242,.1365904598,-2.538475875},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.547208874,-2.679844380,-3.459664217,-4.093529907,-4.641626622,1.110178425,1.922885438,2.482434426,2.937256024,-.3090193970,-.5352372963,-.6909883782,-.5874815429,-1.017547881,1.298238850,-2.679844380,-4.641626622,-5.992314202,-7.090201781,-8.039533141,2.043694516,3.539782737,4.569839863,5.407107446,-.8110805448,-1.404832712,-1.813631233,-.6037611823,-1.045745043,1.765469346,2.576012735,4.461784937,5.760139587,6.815489073,-1.403212583,-2.430435488,-3.137678722,-.8058635216e-2,-.1395796564e-1,1.237350350,-1.238945998,-2.145917416,-2.770367471,.4443853807,.7696980577,.2975289345,-.7339727817,-1.271278149,.1365904499,2.538476058},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.094417748,5.359688761,6.919328434,8.187059814,9.283253244,3.470033583,6.010274468,7.759230975,9.180845900,2.032368460,3.520165433,4.544514032,-.5354861533,-.9274892243,-2.789841949,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6238553746e-1,-.1080549205,-.1394983026,-.1650566176,-.3561120901,-.6168042332,-.7962908411,-1.076648709,-1.864810267,-2.258512675,.1496662955e-7,.2592296279e-7,.3346640106e-7,.1111356701e-6,.1924926272e-6,.4338629615e-6,.1846114076e-2,.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,3.291626742,5.701264756,7.360301149,8.708825766,9.874880227,-3.696869036,-6.403165001,-8.266450470,-9.780996101,3.754901608,6.503680358,8.396215242,-3.509035267,-6.077827368,3.000239739,4.769620097,8.261224341,10.66519477,12.61922863,-4.867652070,-8.431020700,-10.88440092,4.583518828,7.938887491,-3.968421708,5.735706993,9.934535927,12.82543073,-5.454863311,-9.448100406,4.799476921,6.099313091,10.56432016,-5.461692032,5.821024942},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.900421584,-3.291626740,-4.249471847,-5.028042897,-5.701264753,2.131014888,3.691026055,4.765094150,5.638135432,-2.158906718,-3.739336127,-4.827462181,2.009239863,3.480105527,-1.706029997,-3.291626740,-5.701264753,-7.360301145,-8.708825761,-9.874880222,3.696869033,6.403164996,8.266450463,9.780996094,-3.754901605,-6.503680355,-8.396215237,3.509035265,6.077827365,-3.000239738,4.769620101,8.261224347,10.66519478,12.61922864,-4.867652073,-8.431020705,-10.88440092,4.583518832,7.938887498,-3.968421711,-5.735706944,-9.934535843,-12.82543062,5.454863264,9.448100324,-4.799476880,6.099312923,10.56431987,-5.461691879,-5.821025177},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.800843169,6.583253480,8.498943694,10.05608579,11.40252950,6.399791554,11.08476413,14.31036896,16.93225689,8.680560534,15.03517189,19.41032344,10.21325528,17.68987705,10.69428583,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3017302684e-2,-.5226121551e-2,-.6746893911e-2,-.7983032535e-2,-.2009496124e-1,-.3480549385e-1,-.4493369933e-1,-.7471563177e-1,-.1294112704,-.2049323002,-.7483314774e-8,-.1296148140e-7,-.1673320053e-7,-.6443340372e-7,-.1116019290e-6,-.3031212923e-6,.4333424029e-5,.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,-1.719209449,-2.977758114,-3.844269195,-4.548600653,-5.157628347,-1.016533920,-1.760688397,-2.273038946,-2.689495952,1.730662758,2.997595827,3.869879572,-.8438528152e-1,-.1461595950,-1.608041156,.6684577348,1.157802759,1.494716935,1.768572929,.8304449710,1.438372883,1.856931406,-.4995431840,-.8652341756,-.5315629344,.4697515910,.8136336225,1.050396490,.8765560645,1.518239640,-.5415238988e-1,-.9469058756,-1.640089087,-2.346450809,.7389290768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,.6685019987,1.157879426,1.494815912,1.768690039,2.005505995,1.636951328,2.835282871,3.660334447,4.330966124,1.284112192,2.224147558,2.871362151,-.5257834218,-.9106836003,-1.564527457,-.1324976951,-.2294927397,-.2962738530,-.3505559505,-.4894439054,-.8477417116,-1.094429844,-.7451837832,-1.290696174,-.4186690470,-.2123656599,-.3678281127,-.4748640516,-1.044420364,-1.808989135,-2.287254493,-.6336343026e-1,-.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,1.050707449,1.819878685,2.349453280,2.779910610,3.152122347,-.6204174076,-1.074594472,-1.387295498,-1.641470170,-.2618481958e-1,-.4535343788e-1,-.5855103656e-1,.6101686966,1.056843184,-.8964075294,-1.444837340,-2.502531681,-3.230754509,-3.822680287,.4587476822,.7945742934,1.025791002,.5055009012,.8755532444,-1.086251548,-2.213877417,-3.834548168,-4.950380399,.1678643553,.2907495923,1.305496370,.1849334761e-1,.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.378545847,-2.387711448,-3.082522223,-3.647289482,-4.135637540,.6826083759,1.182312388,1.526358730,1.806012005,.3104739793,.5377567068,.6942409231,-1.165283944,-2.018330996,1.521997205,-1.050707450,-1.819878688,-2.349453283,-2.779910614,-3.152122352,.6204174085,1.074594474,1.387295500,1.641470172,.2618481961e-1,.4535343793e-1,.5855103662e-1,-.6101686971,-1.056843185,.8964075306,-1.444837346,-2.502531692,-3.230754521,-3.822680303,.4587476840,.7945742967,1.025791006,.5055009034,.8755532483,-1.086251552,2.213877420,3.834548174,4.950380406,-.1678643556,-.2907495928,-1.305496372,.1849327171e-1,.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.220666420,-.3822054518,-.4934251163,-.5838284711,-.6619992612,-1.207580205,-2.091590269,-2.700231427,-3.194956910,1.141614383,1.977334114,2.552727365,.3043001809,.5270633740,-1.381752244,1.719209448,2.977758112,3.844269192,4.548600649,5.157628344,1.016533919,1.760688396,2.273038944,2.689495949,-1.730662756,-2.997595824,-3.869879569,.8438528146e-1,.1461595949,1.608041155,.6684577365,1.157802762,1.494716939,1.768572933,.8304449729,1.438372886,1.856931411,-.4995431851,-.8652341775,-.5315629356,-.4697515776,-.8136335992,-1.050396460,-.8765560394,-1.518239596,.5415238836e-1,-.9469058829,-1.640089099,-2.346450828,-.7389299868},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.599212267,2.769916901,3.575947341,4.231117955,4.797636804,-.4911838891,-.8507554514,-1.098320565,-1.299550418,-1.809746875,-3.134573538,-4.046717036,-.2960880102,-.5128394771,1.681241028,-.6685019987,-1.157879426,-1.494815912,-1.768690039,-2.005505995,-1.636951328,-2.835282871,-3.660334447,-4.330966124,-1.284112192,-2.224147558,-2.871362151,.5257834218,.9106836003,1.564527457,-.1324976936,-.2294927372,-.2962738499,-.3505559467,-.4894439003,-.8477417029,-1.094429832,-.7451837755,-1.290696161,-.4186690423,.2123656749,.3678281386,.4748640850,1.044420437,1.808989263,2.287254654,-.6336326377e-1,-.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,-2.155052650,-3.732660680,-4.818844218,-5.701733371,-6.465157947,-1.817524160,-3.148044191,-4.064107574,-4.808716931,2.003812619,3.470705265,4.480661231,.8618882312,1.492834207,-2.361581599,1.978516938,3.426891859,4.424098369,5.234663784,3.116327372,5.397637342,6.968319843,-.7599463103,-1.316265621,-2.583756308,-1.441495348,-2.496743182,-3.223281588,-3.286953601,-5.693170641,-.9156869635,.8622047985,1.493382518,2.564268432,-.3927122230},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,1.100336031,1.905837910,2.460426162,2.911215495,3.301008092,2.979568311,5.160763700,6.662517288,7.883196767,3.041169362,5.267459849,6.800261424,.1577963754,.2733113395,-2.799087386,.4783243874,.8284821415,1.069565846,1.265527375,1.930559283,3.343826766,4.316861791,3.440637089,5.959358251,2.951743964,.1430932586,.2478447941,.3199662533,.7646780256,1.324461192,1.899532545,.2128667917e-1,.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,1.054716618,1.826822770,2.358418055,2.790517875,3.164149855,-1.162044153,-2.012719514,-2.598409719,-3.074479842,1.144752980,1.982770324,2.559745481,-1.019684603,-1.766145540,.8049318120,-1.868089212,-3.235625429,-4.177174467,-4.942499484,1.853832431,3.210931959,4.145295333,-1.671751149,-2.895557928,1.349636370,-2.753730610,-4.769601326,-6.157528834,2.522275578,4.368709454,-2.092761921,.4701430699,.8143116839,-.4012398833,3.559792983},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.879499530,-3.255388680,-4.202688712,-4.972688346,-5.638498590,2.065816712,3.578099503,4.619306596,5.465637273,-2.026895457,-3.510685915,-4.532276027,1.793091677,3.105725887,-1.397354961,-1.054716618,-1.826822769,-2.358418053,-2.790517873,-3.164149853,1.162044152,2.012719513,2.598409717,3.074479840,-1.144752979,-1.982770321,-2.559745479,1.019684602,1.766145538,-.8049318114,-1.868089209,-3.235625422,-4.177174460,-4.942499475,1.853832428,3.210931954,4.145295326,-1.671751146,-2.895557923,1.349636368,2.753730616,4.769601337,6.157528849,-2.522275585,-4.368709465,2.092761926,.4701430553,.8143116587,-.4012398709,-3.559792582},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,.26338380e-1,.4561941236e-1,.5889440808e-1,.6968480341e-1,.79015140e-1,-1.374692422,-2.381037118,-3.073905703,-3.637094276,.7961290470,1.378935959,1.780198668,.8745080341,1.514692347,-1.346759223,2.155052650,3.732660680,4.818844218,5.701733371,6.465157947,1.817524160,3.148044191,4.064107574,4.808716931,-2.003812619,-3.470705265,-4.480661231,-.8618882312,-1.492834207,2.361581599,1.978516937,3.426891859,4.424098367,5.234663782,3.116327372,5.397637341,6.968319842,-.7599463103,-1.316265621,-2.583756308,1.441495367,2.496743214,3.223281630,3.286953644,5.693170715,.9156869755,.8622049336,1.493382752,2.564268834,.3927112853},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,1.853161149,3.209769267,4.143794304,4.903003542,5.559483450,-.3287941509e-1,-.5694881744e-1,-.7352060719e-1,-.8699075556e-1,-1.855328981,-3.213524061,-4.148641723,-1.088359331,-1.885093659,1.207824361,-1.100336031,-1.905837911,-2.460426163,-2.911215496,-3.301008094,-2.979568312,-5.160763703,-6.662517291,-7.883196771,-3.041169365,-5.267459851,-6.800261430,-.1577963755,-.2733113396,2.799087387,.4783243870,.8284821409,1.069565845,1.265527374,1.930559282,3.343826764,4.316861789,3.440637085,5.959358244,2.951743961,-.1430932534,-.2478447850,-.3199662416,-.7646779976,-1.324461144,-1.899532475,.2128671353e-1,.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,-2.784908111,-4.823602340,-6.227243845,-7.368174283,-8.354724330,-.8097724007e-1,-.1402566941,-.1810706135,-.2142456391,2.270993172,3.933475556,5.078095108,-2.198518194,-3.807945213,.1780549380,3.578005823,6.197287874,8.000664244,9.466513600,1.789933617,3.100255967,4.002413242,-3.541924988,-6.134794038,1.442211311,-4.013547435,-6.951668076,-8.974564896,-3.781587883,-6.549902349,3.879680769,4.218875403,7.307306550,5.787244192,-4.275985292},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,.7243065273,1.254535705,1.619598631,1.916334944,2.172919582,2.659465646,4.606329621,5.946745969,7.036284721,4.948721068,8.571436320,11.06567671,5.758806642,9.974545694,3.789231007,.2491916055,.4316125216,.5572093695,.6592990173,1.322791395,2.291141904,2.957851479,3.610515302,6.253595946,6.564693789,.7674948247e-1,.1329340031,.1716170600,.5316983539,.9209285636,1.896167979,.2260074784e-1,.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,2.060601584,3.569066636,4.607645215,5.451839342,6.181804751,-2.578488405,-4.466072926,-5.765675354,-6.822039081,3.067086901,5.312350344,6.858214803,-3.560288453,-6.166600490,4.079521103,.1446504328,.2505418989,.3234482008,.3827090723,-.1721034134,-.2980918563,-.3848349315,.1998400143,.3461330583,-.2290664311,-2.797494156,-4.845402011,-6.255387098,3.249889483,5.628973706,-3.727231982,-3.881338999,-6.722676349,4.454821378,-1.567068824},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-2.026046016,-3.509214639,-4.530376616,-5.360413903,-6.078138048,2.535017263,4.390778697,5.668470925,6.707025247,-3.015006854,-5.222145057,-6.741760278,3.499303138,6.060970826,-4.008937548,-2.060601584,-3.569066636,-4.607645215,-5.451839342,-6.181804751,2.578488405,4.466072926,5.765675354,6.822039081,-3.067086901,-5.312350344,-6.858214803,3.560288453,6.166600490,-4.079521103,.1446504323,.2505418980,.3234481995,.3827090709,-.1721034127,-.2980918550,-.3848349300,.1998400136,.3461330570,-.2290664303,2.797494172,4.845402039,6.255387135,-3.249889503,-5.628973739,3.727232004,-3.881339015,-6.722676376,4.454821399,1.567068815},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.771510310,-1.336295057,-1.725149499,-2.041224415,-2.314530931,-.5358691510,-.9281525955,-1.198239849,-1.417776509,1.372241334,2.376791711,3.068424905,-1.053211817,-1.824216377,-.1417580100,2.784908111,4.823602342,6.227243845,7.368174283,8.354724334,.8097724011e-1,.1402566942,.1810706136,.2142456393,-2.270993172,-3.933475557,-5.078095109,2.198518194,3.807945213,-.1780549380,3.578005828,6.197287884,8.000664256,9.466513613,1.789933619,3.100255970,4.002413246,-3.541924993,-6.134794047,1.442211314,4.013547495,6.951668180,8.974565029,3.781587939,6.549902446,-3.879680827,4.218875359,7.307306473,5.787244129,4.275985427},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.797556326,4.845509695,6.255526114,7.401638317,8.392668978,2.441512801,4.228824217,5.459388590,6.459635693,.3276795614,.5675576492,.7327137743,-2.025931334,-3.509016004,-2.806015147,-.7243065266,-1.254535704,-1.619598630,-1.916334942,-2.172919580,-2.659465643,-4.606329616,-5.946745963,-7.036284714,-4.948721064,-8.571436312,-11.06567670,-5.758806636,-9.974545684,-3.789231004,.2491916053,.4316125213,.5572093687,.6592990165,1.322791393,2.291141901,2.957851475,3.610515299,6.253595941,6.564693781,-.7674948090e-1,-.1329340004,-.1716170565,-.5316983431,-.9209285448,-1.896167939,.2260067231e-1,.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,-2.517563337,-4.360547610,-5.629442758,-6.660846498,-7.552690011,.1640586792,.2841579679,.3668463591,.4340584657,1.848945781,3.202468033,4.134368453,-2.129273724,-3.688010274,.6507752466,2.539766587,4.399004769,5.679090736,6.719590780,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207893,-1.820466373,-3.153140251,-4.070686560,-1.433342530,-2.482622088,1.817251596,.7671298784,1.328707925,.9100500654,.2540057989},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,.5106664481,.8845002336,1.141884891,1.351096424,1.531999344,1.778426716,3.080325431,3.976683031,4.705274817,3.022001204,5.234259622,6.757400118,2.956809489,5.121344262,1.089458308,.3202872177e-1,.5547537340e-1,.7161839911e-1,.8474003263e-1,.1620213035,.2806291295,.3622906482,.4123799546,.7142630336,.6762823204,-.2859462130e-1,-.4952733691e-1,-.6393951701e-1,-.1892065416,-.3277153433,-.6347740071,-.1157780971e-1,-.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,2.006896889,3.476047377,4.487557867,5.309750075,6.020690668,-1.942485396,-3.364483399,-4.343529391,-5.139333284,1.514566645,2.623306380,3.386673975,-.8275357615,-1.433333984,.1195045576e-1,.2438275696,.4223217388,.5452150204,.6451071121,-.1988874943,-.3444832452,-.4447259571,.1234754608,.2138657718,-.3202669808e-1,-2.331238944,-4.037824296,-5.212808751,1.622549074,2.810337434,-.7363933866,-3.287037546,-5.693316036,1.939519592,-1.523064210},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.748349281,-3.028229785,-3.909427840,-4.625697402,-5.245047844,1.670691240,2.893722111,3.735779182,4.420233539,-1.265582533,-2.192053249,-2.829928575,.6290039809,1.089466853,.1184619536,-2.006896889,-3.476047376,-4.487557867,-5.309750075,-6.020690666,1.942485396,3.364483399,4.343529391,5.139333284,-1.514566645,-2.623306379,-3.386673973,.8275357615,1.433333984,-.1195045576e-1,.2438275663,.4223217331,.5452150131,.6451071034,-.1988874916,-.3444832405,-.4447259510,.1234754592,.2138657688,-.3202669764e-1,2.331238936,4.037824281,5.212808732,-1.622549068,-2.810337424,.7363933838,-3.287037517,-5.693315987,1.939519574,1.523064511},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-.863849047,-1.496230441,-1.931625192,-2.285529750,-2.591547142,-.3828590477,-.6631313225,-.8560988564,-1.012949827,1.313720149,2.275430046,2.937567557,-1.219296340,-2.111883210,.1856069443,2.517563337,4.360547611,5.629442758,6.660846498,7.552690013,-.1640586792,-.2841579679,-.3668463591,-.4340584657,-1.848945782,-3.202468033,-4.134368454,2.129273724,3.688010274,-.6507752466,2.539766588,4.399004769,5.679090738,6.719590782,.9546117338,1.653436024,2.134576728,-2.422713264,-4.196262466,1.409207894,1.820466382,3.153140266,4.070686580,1.433342537,2.482622099,-1.817251606,.7671297908,1.328707774,.9100499616,-.2540054079},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,2.612198328,4.524460225,5.841053031,6.911227151,7.836594984,1.860129886,3.221839470,4.159376872,4.921441084,-.4585659317,-.7942594926,-1.025384596,-2.363107229,-4.093021785,-2.140742745,-.5106664484,-.8845002342,-1.141884892,-1.351096425,-1.531999345,-1.778426717,-3.080325433,-3.976683034,-4.705274821,-3.022001206,-5.234259626,-6.757400124,-2.956809490,-5.121344265,-1.089458309,.3202872292e-1,.5547537539e-1,.7161840168e-1,.8474003567e-1,.1620213093,.2806291396,.3622906613,.4123799696,.7142630596,.6762823444,.2859462354e-1,.4952734080e-1,.6393952203e-1,.1892065565,.3277153691,.6347740570,-.1157777851e-1,-.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,-3.069477201,-5.316490463,-6.863559675,-8.121073328,-9.208431602,2.156281693,3.734789450,4.821592446,5.704985119,-.5614333340,-.9724310592,-1.255403100,-1.126462044,-1.951089492,2.330362820,4.159176446,7.203904921,9.300201262,11.00414654,-1.935519673,-3.352418413,-4.327953560,-.5514408001,-.9551234835,2.488724225,-4.693254379,-8.128955037,-10.49443583,1.219702470,2.112586649,1.756836938,4.705330903,8.149872191,-.3189574095,-4.265567617},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,.2458344589,.4257977730,.5497025612,.6504168418,.7375033768,1.054396484,1.826268282,2.357702214,2.789670881,2.510226735,4.347840243,5.613037617,4.324472530,7.490206138,5.881975759,.1420874210e-1,.2461026322e-1,.3177171321e-1,.3759279804e-1,.8700209506e-1,.1506920490,.1945425987,.2894494276,.5013411150,.6894088761,-.1216128451e-2,-.2106396265e-2,-.2719345885e-2,-.9658561257e-2,-.1672911883e-1,-.4108123658e-1,-.3746153966e-3,-.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,2.823642741,4.890692689,6.313857110,7.470656481,8.470928224,-3.210678176,-5.561057729,-7.179294657,-8.494655996,3.324354634,5.757951125,7.433482942,-3.198010479,-5.539116631,2.859723954,2.797956434,4.846202701,6.256420786,7.402696906,-2.907006833,-5.035083533,-6.500264888,2.811341931,4.869387063,-2.534878984,1.242384808,2.151873609,2.778056884,-1.210043957,-2.095857614,1.103053355,-1.284643645,-2.225068063,1.185627028,-3.765218080},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.914096079,-3.315311660,-4.280048947,-5.064222210,-5.742288236,2.173949813,3.765391528,4.861099562,5.751730568,-2.246794550,-3.891562317,-5.023985347,2.155283216,3.733060035,-1.918656241,-2.823642742,-4.890692690,-6.313857114,-7.470656486,-8.470928225,3.210678177,5.561057732,7.179294660,8.494656000,-3.324354634,-5.757951127,-7.433482942,3.198010480,5.539116634,-2.859723955,2.797956438,4.846202708,6.256420796,7.402696917,-2.907006837,-5.035083540,-6.500264898,2.811341934,4.869387068,-2.534878988,-1.242384788,-2.151873575,-2.778056840,1.210043938,2.095857580,-1.103053339,-1.284643653,-2.225068076,1.185627035,3.765217145},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,-1.488298306,-2.577808283,-3.327936182,-3.937667194,-4.464894918,.9564376257,1.656598562,2.138659547,2.530496102,-.7448321426e-1,-.1290087115,-.1665495303,-.8305776036,-1.438602609,1.441026240,3.069477200,5.316490462,6.863559672,8.121073323,9.208431600,-2.156281693,-3.734789450,-4.821592446,-5.704985119,.5614333335,.9724310584,1.255403098,1.126462044,1.951089492,-2.330362819,4.159176444,7.203904918,9.300201260,11.00414653,-1.935519673,-3.352418412,-4.327953559,-.5514408000,-.9551234833,2.488724224,4.693254424,8.128955115,10.49443593,-1.219702482,-2.112586669,-1.756836954,4.705330679,8.149871802,-.3189573943,4.265566870},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,4.690415760,3.402394385,5.893119942,7.607985127,9.001889402,10.20718315,4.663809445,8.077954914,10.42859495,12.33927995,4.448012921,7.704184374,9.946059258,2.665659156,4.617057095,-.9771433915e-1,-.2458344586,-.4257977724,-.5497025604,-.6504168409,-.7375033757,-1.054396483,-1.826268280,-2.357702211,-2.789670877,-2.510226732,-4.347840236,-5.613037612,-4.324472524,-7.490206128,-5.881975750,.1420874260e-1,.2461026409e-1,.3177171433e-1,.3759279937e-1,.8700209816e-1,.1506920544,.1945426056,.2894494379,.5013411328,.6894089005,.1216122464e-2,.2106385895e-2,.2719332498e-2,.9658513710e-2,.1672903648e-1,.4108103434e-1,-.3745694178e-3,-.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};
		for (int j = 0; j < 52; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[52][56] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,0.,0.,0.,0.,0.,-1.360827635,2.357022604,-3.042903097,3.600411499,.8380524812,-1.451549477,1.873942317,.8198497638,-1.420021446,-1.368533972,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.217161239,2.108185107,-2.721655270,3.220305944,-1.873942317,3.245763304,-4.190262406,.5237828008,-.9072184234,1.530067495,0.,0.,0.,0.,0.,0.,.7027283688,-1.217161239,2.052800957,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,-1.704828602,2.952849755,-3.812112642,4.510552507,-5.114485804,-3.104623248,5.377365205,-6.942148628,8.214061030,-.5462559154,.9461429991,-1.221465360,2.694797933,-4.667526936,1.022426944,1.298058152,-2.248302670,2.902546266,-3.434339057,3.719687555,-6.442687835,8.317474224,3.264192666,-5.653747546,-1.221656347,-.8344196900,1.445257298,-1.865819148,-3.243190051,5.617369949,-4.760960496,.4850599026,-.8401483961,2.375793728,-.2614753427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.984283251,-1.704828602,2.200924260,-2.604168704,2.952849756,-1.237409468,2.143256068,-2.766931687,3.273877723,-1.032732859,1.788745783,-2.309260876,1.210247240,-2.096209709,1.042423382,1.704828601,-2.952849754,3.812112640,-4.510552505,5.114485802,3.104623246,-5.377365203,6.942148625,-8.214061027,.5462559149,-.9461429982,1.221465359,-2.694797932,4.667526936,-1.022426944,1.298058156,-2.248302677,2.902546276,-3.434339069,3.719687568,-6.442687856,8.317474253,3.264192678,-5.653747565,-1.221656351,.8344196802,-1.445257281,1.865819127,3.243190014,-5.617369884,4.760960437,.4850600816,-.8401487061,2.375794604,.2614758325},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.968566503,3.409657202,-4.401848518,5.208337406,-5.905699510,2.347500666,-4.065990423,5.249171065,-6.210902963,-2.609637841,4.520025331,-5.835327609,2.775506203,-4.807317761,-2.855122567,-.3360178568e-9,.582e-9,-.7513587691e-9,.8890196848e-9,-.1008053570e-8,.4007561299e-9,-.6941299786e-9,.8961179490e-9,-.1060301056e-8,-.4456004436e-9,.7718026080e-9,-.9963928825e-9,.4740581912e-9,-.8210928729e-9,-.4878407244e-9,-2.709993221,4.693845947,-6.059729062,7.169968120,3.014474194,-5.221222463,6.740569213,-3.208782240,5.557773872,3.304507380,-.7333648479e-8,.1270225177e-7,-.1639853652e-7,.7812670753e-8,-.1353194269e-7,-.8054303853e-8,3.483595229,-6.033763930,-3.596173599,-.1217502607e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,-.8327988349,1.442449894,-1.862194806,2.203378609,-2.498396504,-1.088657954,1.885610890,-2.434313191,2.880318211,.4375794197,-.7579097871,.9784573278,.9743405328,-1.687607307,-.5933186699,-.4672884697,.8093673712,-1.044888783,1.236329082,-1.022115558,1.770356077,-2.285519867,-.3251247057,.5631325093,.8276510476,.6684136118,-1.157726336,1.494618273,2.033596264,-3.522292053,1.746351412,-.1476726503,.2557765332,-.5739768229,-.2050386800},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.480816631,-.8327988360,1.075138674,-1.272121234,1.442449896,-1.465249839,2.537887167,-3.276398245,3.876686683,-.4816921701e-1,.8343153125e-1,-.1077096437,1.474494615,-2.553899588,-.4344958536,.8327988342,-1.442449893,1.862194804,-2.203378607,2.498396503,1.088657953,-1.885610888,2.434313189,-2.880318209,-.4375794194,.7579097864,-.9784573273,-.9743405322,1.687607306,.5933186693,-.4672884647,.8093673627,-1.044888772,1.236329069,-1.022115547,1.770356058,-2.285519843,-.3251247023,.5631325034,.8276510391,-.6684136118,1.157726336,-1.494618273,-2.033596264,3.522292053,-1.746351412,-.1476726019,.2557764494,-.5739766349,.2050384112},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.961633263,1.665597669,-2.150277344,2.544242465,-2.884899788,-.2081759125,.3605712573,-.4654954917,.5507816934,1.215013765,-2.104465573,2.716853372,-1.352704319,2.342952608,.5364099445,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.1039230485e-8,.5714598801e-10,-.9897975472e-10,.1277823139e-9,-.1511940727e-9,.2193666394e-9,-.3799541647e-9,.4905187176e-9,-.3033617129e-9,.5254378997e-9,.1563891368e-9,-1.872739842,3.243680556,-4.187573592,4.954803895,-.4572029592,.7918987547,-1.022336896,1.670078386,-2.892660618,-1.283464460,0.,-0.,0.,0.,-0.,-0.,1.663589468,-2.881421481,1.646826248,-.1953151300e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,-1.425047564,2.468254783,-3.186503223,3.770321460,-4.275142691,-2.360183599,4.087957910,-5.277530967,6.244458853,-.4945548803e-2,.8565941795e-2,-.1105858331e-1,2.175490654,-3.768060344,.1537427033,.6224210815,-1.078064937,1.391775849,-1.646771393,1.648153002,-2.854684738,3.685382148,1.157573982,-2.004976951,-.8770258842,-.7554460893e-2,.1308471009e-1,-.1689228809e-1,-.2731590581e-1,.4731253672e-1,-.3482606398e-1,-.2198338910,.3807634685,-1.005417380,.1997404621},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.822751595,-1.425047563,1.839728492,-2.176796108,2.468254782,-1.348091029,2.334962155,-3.014423181,3.566713607,-.7313658101,1.266762742,-1.635383668,1.406175430,-2.435567290,.5915326939,1.425047564,-2.468254784,3.186503224,-3.770321460,4.275142693,2.360183599,-4.087957910,5.277530967,-6.244458853,.4945548803e-2,-.8565941795e-2,.1105858331e-1,-2.175490654,3.768060344,-.1537427033,.6224210804,-1.078064935,1.391775847,-1.646771390,1.648152999,-2.854684733,3.685382143,1.157573980,-2.004976947,-.8770258827,.7554451164e-2,-.1308469324e-1,.1689226633e-1,.2731587063e-1,-.4731247580e-1,.3482601912e-1,-.2198339029,.3807634890,-1.005417434,-.1997405113},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.645503189,2.850095128,-3.679456987,4.353592220,-4.936509566,1.377214242,-2.385405038,3.079544664,-3.643766386,-.7456423653,1.291490461,-1.667307016,-.7214030405e-1,.1249506719,.8704037330,.2424871131e-9,-.42e-9,.5422176684e-9,-.6415605973e-9,.7274613394e-9,-.2092513633e-9,.3624339929e-9,-.4679002728e-9,.5536270690e-9,.1249332326e-9,-.2163907064e-9,.2793592008e-9,-.1348127464e-10,.2335025263e-10,-.9727958508e-10,-2.424562843,4.199466029,-5.421487332,6.414790321,1.641768329,-2.843626160,3.671105586,-.5644916125,.9777281536,-.5417963103,-.7483314774e-9,.1296148140e-8,-.1673320053e-8,.3917700593e-9,-.6785656478e-9,-.9473764217e-11,2.788419758,-4.829684694,-1.069368685,-.6893475176e-8},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,.9641937137,-1.670032500,2.156002687,-2.551016781,2.892581141,.2394478538,-.4147358487,.5354216783,-.6335194733,-.9155822076,1.585834902,-2.047304055,.5490545322,-.9509903459,.3754286701,-.9430172903,1.633353859,-2.108650765,2.494989233,-.7446746083,1.289814257,-1.665143045,.9321620643,-1.614552057,.2692350424e-1,-1.398574920,2.422402819,-3.127308592,-1.821631846,3.155158911,1.076451798,-.1570858835,.2720807313,-.2832576764,.8954348170},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.556677500,.9641937137,-1.244768731,1.472830225,-1.670032500,-.8469574563,1.466973346,-1.893854446,2.240838800,1.390251875,-2.407986883,3.108697698,-.5638685582,.9766489917,-.8112194737,-.9641937137,1.670032500,-2.156002687,2.551016781,-2.892581141,-.2394478538,.4147358487,-.5354216783,.6335194733,.9155822076,-1.585834902,2.047304055,-.5490545322,.9509903459,-.3754286701,-.9430172892,1.633353857,-2.108650763,2.494989230,-.7446746070,1.289814255,-1.665143042,.9321620636,-1.614552056,.2692350421e-1,1.398574927,-2.422402832,3.127308609,1.821631855,-3.155158927,-1.076451804,-.1570858218,.2720806245,-.2832575652,-.8954346999},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.113355001,-1.928387427,2.489537463,-2.945660451,3.340065000,-1.123448021,1.945869052,-2.512106144,2.972364074,-1.252806288,2.169924144,-2.801360023,.9835692099,-1.703591844,1.344816036,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.1039230485e-8,.6764731475e-9,-.1171685862e-8,.1512639943e-8,-.1789779717e-8,.2094450706e-9,-.3627695035e-9,.4683334153e-9,-.5392014460e-9,.9339242999e-9,-.350864418e-9,-.6338914369,1.097932175,-1.417424343,1.677119101,-1.926686150,3.337118302,-4.308201201,-1.956429599,3.388635469,.2565449507,.7483314774e-8,-.1296148140e-7,.1673320053e-7,.3070569490e-7,-.5318382366e-7,.4977473171e-7,.1905991076,-.3301273382,.9829492714,0.},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,1.926395974,-3.336615703,4.307552349,-5.096764673,5.779187924,-.6138328492,1.063189683,-1.372571978,1.624049066,-.8603944811,1.490246955,-1.923900547,1.645753812,-2.850529219,-1.343411045,.5526193998,-.9571648777,1.235694544,-1.462093502,.2281916739e-1,-.3952395730e-1,.5102520946e-1,-.4132700278,.7158046857,.4485373028,-1.314274316,2.276389891,-2.938806712,-.4873895555,.8441834734,1.215532077,-2.142609188,3.711107975,-1.473102430,-1.425504294},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.112205235,1.926395976,-2.486966510,2.942618459,-3.336615704,.8625243345e-1,-.1493935970,.1928663044,-.2282024888,.9832929299,-1.703113314,2.198709833,-1.447727447,2.507537494,1.036438633,-1.926395976,3.336615704,-4.307552352,5.096764678,-5.779187926,.6138328496,-1.063189683,1.372571979,-1.624049067,.8603944811,-1.490246955,1.923900547,-1.645753812,2.850529220,1.343411045,.5526194031,-.9571648834,1.235694551,-1.462093511,.2281916752e-1,-.3952395754e-1,.5102520976e-1,-.4132700304,.7158046901,.4485373055,1.314274282,-2.276389831,2.938806635,.4873895427,-.8441834514,-1.215532046,-2.142609204,3.711108002,-1.473102441,1.425504133},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.224410470,-3.852791952,4.973933020,-5.885236917,6.673231410,.7950455559,-1.377059297,1.777775908,-2.103492822,-1.500452022,2.598859138,-3.355112719,-2.072134551,3.589042323,-.2431355237,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,-0.,-.2398353968,.4154070926,-.5362882506,.6345448156,-1.087945327,1.884376583,-2.432719707,-2.347136393,4.065359486,-2.942787847,0.,-0.,0.,0.,-0.,0.,.2728458417e-1,-.4725828605e-1,.2015923197,.1700390544e-6},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,2.679844379,-4.641626620,5.992314198,-7.090201776,8.039533138,-2.043694514,3.539782735,-4.569839860,5.407107442,.8110805448,-1.404832712,1.813631233,.6037611818,-1.045745042,-1.765469345,2.576012736,-4.461784940,5.760139589,-6.815489076,-1.403212584,2.430435489,-3.137678723,-.8058635216e-2,.1395796564e-1,1.237350351,1.238945990,-2.145917403,2.770367454,-.4443853780,.7696980531,-.2975289326,-.7339728352,1.271278242,.1365904598,-2.538475875},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.547208874,2.679844380,-3.459664217,4.093529907,-4.641626622,1.110178425,-1.922885438,2.482434426,-2.937256024,-.3090193970,.5352372963,-.6909883782,-.5874815429,1.017547881,1.298238850,-2.679844380,4.641626622,-5.992314202,7.090201781,-8.039533141,2.043694516,-3.539782737,4.569839863,-5.407107446,-.8110805448,1.404832712,-1.813631233,-.6037611823,1.045745043,1.765469346,2.576012735,-4.461784937,5.760139587,-6.815489073,-1.403212583,2.430435488,-3.137678722,-.8058635216e-2,.1395796564e-1,1.237350350,-1.238945998,2.145917416,-2.770367471,.4443853807,-.7696980577,.2975289345,-.7339727817,1.271278149,.1365904499,2.538476058},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.094417748,-5.359688761,6.919328434,-8.187059814,9.283253244,3.470033583,-6.010274468,7.759230975,-9.180845900,2.032368460,-3.520165433,4.544514032,-.5354861533,.9274892243,-2.789841949,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.6238553746e-1,.1080549205,-.1394983026,.1650566176,-.3561120901,.6168042332,-.7962908411,-1.076648709,1.864810267,-2.258512675,.1496662955e-7,-.2592296279e-7,.3346640106e-7,.1111356701e-6,-.1924926272e-6,.4338629615e-6,.1846114076e-2,-.3197563376e-2,.1686564735e-1,.4250976359e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,3.291626742,-5.701264756,7.360301149,-8.708825766,9.874880227,-3.696869036,6.403165001,-8.266450470,9.780996101,3.754901608,-6.503680358,8.396215242,-3.509035267,6.077827368,3.000239739,4.769620097,-8.261224341,10.66519477,-12.61922863,-4.867652070,8.431020700,-10.88440092,4.583518828,-7.938887491,-3.968421708,5.735706993,-9.934535927,12.82543073,-5.454863311,9.448100406,4.799476921,6.099313091,-10.56432016,-5.461692032,5.821024942},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.900421584,3.291626740,-4.249471847,5.028042897,-5.701264753,2.131014888,-3.691026055,4.765094150,-5.638135432,-2.158906718,3.739336127,-4.827462181,2.009239863,-3.480105527,-1.706029997,-3.291626740,5.701264753,-7.360301145,8.708825761,-9.874880222,3.696869033,-6.403164996,8.266450463,-9.780996094,-3.754901605,6.503680355,-8.396215237,3.509035265,-6.077827365,-3.000239738,4.769620101,-8.261224347,10.66519478,-12.61922864,-4.867652073,8.431020705,-10.88440092,4.583518832,-7.938887498,-3.968421711,-5.735706944,9.934535843,-12.82543062,5.454863264,-9.448100324,-4.799476880,6.099312923,-10.56431987,-5.461691879,-5.821025177},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.800843169,-6.583253480,8.498943694,-10.05608579,11.40252950,6.399791554,-11.08476413,14.31036896,-16.93225689,8.680560534,-15.03517189,19.41032344,10.21325528,-17.68987705,10.69428583,0.,-0.,0.,-0.,0.,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.3017302684e-2,.5226121551e-2,-.6746893911e-2,.7983032535e-2,-.2009496124e-1,.3480549385e-1,-.4493369933e-1,-.7471563177e-1,.1294112704,-.2049323002,-.7483314774e-8,.1296148140e-7,-.1673320053e-7,-.6443340372e-7,.1116019290e-6,-.3031212923e-6,.4333424029e-5,-.7505710589e-5,.4573704284e-4,-.2224294657e-7},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,-1.719209449,2.977758114,-3.844269195,4.548600653,-5.157628347,-1.016533920,1.760688397,-2.273038946,2.689495952,1.730662758,-2.997595827,3.869879572,-.8438528152e-1,.1461595950,-1.608041156,.6684577348,-1.157802759,1.494716935,-1.768572929,.8304449710,-1.438372883,1.856931406,-.4995431840,.8652341756,-.5315629344,.4697515910,-.8136336225,1.050396490,.8765560645,-1.518239640,-.5415238988e-1,-.9469058756,1.640089087,-2.346450809,.7389290768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,.6685019987,-1.157879426,1.494815912,-1.768690039,2.005505995,1.636951328,-2.835282871,3.660334447,-4.330966124,1.284112192,-2.224147558,2.871362151,-.5257834218,.9106836003,-1.564527457,-.1324976951,.2294927397,-.2962738530,.3505559505,-.4894439054,.8477417116,-1.094429844,-.7451837832,1.290696174,-.4186690470,-.2123656599,.3678281127,-.4748640516,-1.044420364,1.808989135,-2.287254493,-.6336343026e-1,.1097486806,-.3886022550,.1510539642e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,1.050707449,-1.819878685,2.349453280,-2.779910610,3.152122347,-.6204174076,1.074594472,-1.387295498,1.641470170,-.2618481958e-1,.4535343788e-1,-.5855103656e-1,.6101686966,-1.056843184,-.8964075294,-1.444837340,2.502531681,-3.230754509,3.822680287,.4587476822,-.7945742934,1.025791002,.5055009012,-.8755532444,-1.086251548,-2.213877417,3.834548168,-4.950380399,.1678643553,-.2907495923,1.305496370,.1849334761e-1,-.3203141766e-1,.2822662419e-2,2.197288848},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.378545847,2.387711448,-3.082522223,3.647289482,-4.135637540,.6826083759,-1.182312388,1.526358730,-1.806012005,.3104739793,-.5377567068,.6942409231,-1.165283944,2.018330996,1.521997205,-1.050707450,1.819878688,-2.349453283,2.779910614,-3.152122352,.6204174085,-1.074594474,1.387295500,-1.641470172,.2618481961e-1,-.4535343793e-1,.5855103662e-1,-.6101686971,1.056843185,.8964075306,-1.444837346,2.502531692,-3.230754521,3.822680303,.4587476840,-.7945742967,1.025791006,.5055009034,-.8755532483,-1.086251552,2.213877420,-3.834548174,4.950380406,-.1678643556,.2907495928,-1.305496372,.1849327171e-1,-.3203128620e-1,.2822650835e-2,-2.197289667},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.220666420,.3822054518,-.4934251163,.5838284711,-.6619992612,-1.207580205,2.091590269,-2.700231427,3.194956910,1.141614383,-1.977334114,2.552727365,.3043001809,-.5270633740,-1.381752244,1.719209448,-2.977758112,3.844269192,-4.548600649,5.157628344,1.016533919,-1.760688396,2.273038944,-2.689495949,-1.730662756,2.997595824,-3.869879569,.8438528146e-1,-.1461595949,1.608041155,.6684577365,-1.157802762,1.494716939,-1.768572933,.8304449729,-1.438372886,1.856931411,-.4995431851,.8652341775,-.5315629356,-.4697515776,.8136335992,-1.050396460,-.8765560394,1.518239596,.5415238836e-1,-.9469058829,1.640089099,-2.346450828,-.7389299868},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.599212267,-2.769916901,3.575947341,-4.231117955,4.797636804,-.4911838891,.8507554514,-1.098320565,1.299550418,-1.809746875,3.134573538,-4.046717036,-.2960880102,.5128394771,1.681241028,-.6685019987,1.157879426,-1.494815912,1.768690039,-2.005505995,-1.636951328,2.835282871,-3.660334447,4.330966124,-1.284112192,2.224147558,-2.871362151,.5257834218,-.9106836003,1.564527457,-.1324976936,.2294927372,-.2962738499,.3505559467,-.4894439003,.8477417029,-1.094429832,-.7451837755,1.290696161,-.4186690423,.2123656749,-.3678281386,.4748640850,1.044420437,-1.808989263,2.287254654,-.6336326377e-1,.1097483922,-.3886012338,-.1510564688e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,-2.155052650,3.732660680,-4.818844218,5.701733371,-6.465157947,-1.817524160,3.148044191,-4.064107574,4.808716931,2.003812619,-3.470705265,4.480661231,.8618882312,-1.492834207,-2.361581599,1.978516938,-3.426891859,4.424098369,-5.234663784,3.116327372,-5.397637342,6.968319843,-.7599463103,1.316265621,-2.583756308,-1.441495348,2.496743182,-3.223281588,-3.286953601,5.693170641,-.9156869635,.8622047985,-1.493382518,2.564268432,-.3927122230},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,1.100336031,-1.905837910,2.460426162,-2.911215495,3.301008092,2.979568311,-5.160763700,6.662517288,-7.883196767,3.041169362,-5.267459849,6.800261424,.1577963754,-.2733113395,-2.799087386,.4783243874,-.8284821415,1.069565846,-1.265527375,1.930559283,-3.343826766,4.316861791,3.440637089,-5.959358251,2.951743964,.1430932586,-.2478447941,.3199662533,.7646780256,-1.324461192,1.899532545,.2128667917e-1,-.3686960985e-1,.1414057120,-.7761524549e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,1.054716618,-1.826822770,2.358418055,-2.790517875,3.164149855,-1.162044153,2.012719514,-2.598409719,3.074479842,1.144752980,-1.982770324,2.559745481,-1.019684603,1.766145540,.8049318120,-1.868089212,3.235625429,-4.177174467,4.942499484,1.853832431,-3.210931959,4.145295333,-1.671751149,2.895557928,1.349636370,-2.753730610,4.769601326,-6.157528834,2.522275578,-4.368709454,-2.092761921,.4701430699,-.8143116839,-.4012398833,3.559792983},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.879499530,3.255388680,-4.202688712,4.972688346,-5.638498590,2.065816712,-3.578099503,4.619306596,-5.465637273,-2.026895457,3.510685915,-4.532276027,1.793091677,-3.105725887,-1.397354961,-1.054716618,1.826822769,-2.358418053,2.790517873,-3.164149853,1.162044152,-2.012719513,2.598409717,-3.074479840,-1.144752979,1.982770321,-2.559745479,1.019684602,-1.766145538,-.8049318114,-1.868089209,3.235625422,-4.177174460,4.942499475,1.853832428,-3.210931954,4.145295326,-1.671751146,2.895557923,1.349636368,2.753730616,-4.769601337,6.157528849,-2.522275585,4.368709465,2.092761926,.4701430553,-.8143116587,-.4012398709,-3.559792582},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,.26338380e-1,-.4561941236e-1,.5889440808e-1,-.6968480341e-1,.79015140e-1,-1.374692422,2.381037118,-3.073905703,3.637094276,.7961290470,-1.378935959,1.780198668,.8745080341,-1.514692347,-1.346759223,2.155052650,-3.732660680,4.818844218,-5.701733371,6.465157947,1.817524160,-3.148044191,4.064107574,-4.808716931,-2.003812619,3.470705265,-4.480661231,-.8618882312,1.492834207,2.361581599,1.978516937,-3.426891859,4.424098367,-5.234663782,3.116327372,-5.397637341,6.968319842,-.7599463103,1.316265621,-2.583756308,1.441495367,-2.496743214,3.223281630,3.286953644,-5.693170715,.9156869755,.8622049336,-1.493382752,2.564268834,.3927112853},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,1.853161149,-3.209769267,4.143794304,-4.903003542,5.559483450,-.3287941509e-1,.5694881744e-1,-.7352060719e-1,.8699075556e-1,-1.855328981,3.213524061,-4.148641723,-1.088359331,1.885093659,1.207824361,-1.100336031,1.905837911,-2.460426163,2.911215496,-3.301008094,-2.979568312,5.160763703,-6.662517291,7.883196771,-3.041169365,5.267459851,-6.800261430,-.1577963755,.2733113396,2.799087387,.4783243870,-.8284821409,1.069565845,-1.265527374,1.930559282,-3.343826764,4.316861789,3.440637085,-5.959358244,2.951743961,-.1430932534,.2478447850,-.3199662416,-.7646779976,1.324461144,-1.899532475,.2128671353e-1,-.3686966936e-1,.1414059402,.7761365057e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,-2.784908111,4.823602340,-6.227243845,7.368174283,-8.354724330,-.8097724007e-1,.1402566941,-.1810706135,.2142456391,2.270993172,-3.933475556,5.078095108,-2.198518194,3.807945213,.1780549380,3.578005823,-6.197287874,8.000664244,-9.466513600,1.789933617,-3.100255967,4.002413242,-3.541924988,6.134794038,1.442211311,-4.013547435,6.951668076,-8.974564896,-3.781587883,6.549902349,3.879680769,4.218875403,-7.307306550,5.787244192,-4.275985292},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,.7243065273,-1.254535705,1.619598631,-1.916334944,2.172919582,2.659465646,-4.606329621,5.946745969,-7.036284721,4.948721068,-8.571436320,11.06567671,5.758806642,-9.974545694,3.789231007,.2491916055,-.4316125216,.5572093695,-.6592990173,1.322791395,-2.291141904,2.957851479,3.610515302,-6.253595946,6.564693789,.7674948247e-1,-.1329340031,.1716170600,.5316983539,-.9209285636,1.896167979,.2260074784e-1,-.3914564355e-1,.1930005379,.6515299128e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,2.060601584,-3.569066636,4.607645215,-5.451839342,6.181804751,-2.578488405,4.466072926,-5.765675354,6.822039081,3.067086901,-5.312350344,6.858214803,-3.560288453,6.166600490,4.079521103,.1446504328,-.2505418989,.3234482008,-.3827090723,-.1721034134,.2980918563,-.3848349315,.1998400143,-.3461330583,-.2290664311,-2.797494156,4.845402011,-6.255387098,3.249889483,-5.628973706,-3.727231982,-3.881338999,6.722676349,4.454821378,-1.567068824},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-2.026046016,3.509214639,-4.530376616,5.360413903,-6.078138048,2.535017263,-4.390778697,5.668470925,-6.707025247,-3.015006854,5.222145057,-6.741760278,3.499303138,-6.060970826,-4.008937548,-2.060601584,3.569066636,-4.607645215,5.451839342,-6.181804751,2.578488405,-4.466072926,5.765675354,-6.822039081,-3.067086901,5.312350344,-6.858214803,3.560288453,-6.166600490,-4.079521103,.1446504323,-.2505418980,.3234481995,-.3827090709,-.1721034127,.2980918550,-.3848349300,.1998400136,-.3461330570,-.2290664303,2.797494172,-4.845402039,6.255387135,-3.249889503,5.628973739,3.727232004,-3.881339015,6.722676376,4.454821399,1.567068815},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.771510310,1.336295057,-1.725149499,2.041224415,-2.314530931,-.5358691510,.9281525955,-1.198239849,1.417776509,1.372241334,-2.376791711,3.068424905,-1.053211817,1.824216377,-.1417580100,2.784908111,-4.823602342,6.227243845,-7.368174283,8.354724334,.8097724011e-1,-.1402566942,.1810706136,-.2142456393,-2.270993172,3.933475557,-5.078095109,2.198518194,-3.807945213,-.1780549380,3.578005828,-6.197287884,8.000664256,-9.466513613,1.789933619,-3.100255970,4.002413246,-3.541924993,6.134794047,1.442211314,4.013547495,-6.951668180,8.974565029,3.781587939,-6.549902446,-3.879680827,4.218875359,-7.307306473,5.787244129,4.275985427},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.797556326,-4.845509695,6.255526114,-7.401638317,8.392668978,2.441512801,-4.228824217,5.459388590,-6.459635693,.3276795614,-.5675576492,.7327137743,-2.025931334,3.509016004,-2.806015147,-.7243065266,1.254535704,-1.619598630,1.916334942,-2.172919580,-2.659465643,4.606329616,-5.946745963,7.036284714,-4.948721064,8.571436312,-11.06567670,-5.758806636,9.974545684,-3.789231004,.2491916053,-.4316125213,.5572093687,-.6592990165,1.322791393,-2.291141901,2.957851475,3.610515299,-6.253595941,6.564693781,-.7674948090e-1,.1329340004,-.1716170565,-.5316983431,.9209285448,-1.896167939,.2260067231e-1,-.3914551274e-1,.1929998929,-.6515244319e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,-2.517563337,4.360547610,-5.629442758,6.660846498,-7.552690011,.1640586792,-.2841579679,.3668463591,-.4340584657,1.848945781,-3.202468033,4.134368453,-2.129273724,3.688010274,.6507752466,2.539766587,-4.399004769,5.679090736,-6.719590780,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207893,-1.820466373,3.153140251,-4.070686560,-1.433342530,2.482622088,1.817251596,.7671298784,-1.328707925,.9100500654,.2540057989},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,.5106664481,-.8845002336,1.141884891,-1.351096424,1.531999344,1.778426716,-3.080325431,3.976683031,-4.705274817,3.022001204,-5.234259622,6.757400118,2.956809489,-5.121344262,1.089458308,.3202872177e-1,-.5547537340e-1,.7161839911e-1,-.8474003263e-1,.1620213035,-.2806291295,.3622906482,.4123799546,-.7142630336,.6762823204,-.2859462130e-1,.4952733691e-1,-.6393951701e-1,-.1892065416,.3277153433,-.6347740071,-.1157780971e-1,.2005335466e-1,-.9455952384e-1,-.1841465513e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,2.006896889,-3.476047377,4.487557867,-5.309750075,6.020690668,-1.942485396,3.364483399,-4.343529391,5.139333284,1.514566645,-2.623306380,3.386673975,-.8275357615,1.433333984,.1195045576e-1,.2438275696,-.4223217388,.5452150204,-.6451071121,-.1988874943,.3444832452,-.4447259571,.1234754608,-.2138657718,-.3202669808e-1,-2.331238944,4.037824296,-5.212808751,1.622549074,-2.810337434,-.7363933866,-3.287037546,5.693316036,1.939519592,-1.523064210},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.748349281,3.028229785,-3.909427840,4.625697402,-5.245047844,1.670691240,-2.893722111,3.735779182,-4.420233539,-1.265582533,2.192053249,-2.829928575,.6290039809,-1.089466853,.1184619536,-2.006896889,3.476047376,-4.487557867,5.309750075,-6.020690666,1.942485396,-3.364483399,4.343529391,-5.139333284,-1.514566645,2.623306379,-3.386673973,.8275357615,-1.433333984,-.1195045576e-1,.2438275663,-.4223217331,.5452150131,-.6451071034,-.1988874916,.3444832405,-.4447259510,.1234754592,-.2138657688,-.3202669764e-1,2.331238936,-4.037824281,5.212808732,-1.622549068,2.810337424,.7363933838,-3.287037517,5.693315987,1.939519574,1.523064511},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-.863849047,1.496230441,-1.931625192,2.285529750,-2.591547142,-.3828590477,.6631313225,-.8560988564,1.012949827,1.313720149,-2.275430046,2.937567557,-1.219296340,2.111883210,.1856069443,2.517563337,-4.360547611,5.629442758,-6.660846498,7.552690013,-.1640586792,.2841579679,-.3668463591,.4340584657,-1.848945782,3.202468033,-4.134368454,2.129273724,-3.688010274,-.6507752466,2.539766588,-4.399004769,5.679090738,-6.719590782,.9546117338,-1.653436024,2.134576728,-2.422713264,4.196262466,1.409207894,1.820466382,-3.153140266,4.070686580,1.433342537,-2.482622099,-1.817251606,.7671297908,-1.328707774,.9100499616,-.2540054079},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,2.612198328,-4.524460225,5.841053031,-6.911227151,7.836594984,1.860129886,-3.221839470,4.159376872,-4.921441084,-.4585659317,.7942594926,-1.025384596,-2.363107229,4.093021785,-2.140742745,-.5106664484,.8845002342,-1.141884892,1.351096425,-1.531999345,-1.778426717,3.080325433,-3.976683034,4.705274821,-3.022001206,5.234259626,-6.757400124,-2.956809490,5.121344265,-1.089458309,.3202872292e-1,-.5547537539e-1,.7161840168e-1,-.8474003567e-1,.1620213093,-.2806291396,.3622906613,.4123799696,-.7142630596,.6762823444,.2859462354e-1,-.4952734080e-1,.6393952203e-1,.1892065565,-.3277153691,.6347740570,-.1157777851e-1,.2005330061e-1,-.9455926900e-1,.1841563171e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,-3.069477201,5.316490463,-6.863559675,8.121073328,-9.208431602,2.156281693,-3.734789450,4.821592446,-5.704985119,-.5614333340,.9724310592,-1.255403100,-1.126462044,1.951089492,2.330362820,4.159176446,-7.203904921,9.300201262,-11.00414654,-1.935519673,3.352418413,-4.327953560,-.5514408001,.9551234835,2.488724225,-4.693254379,8.128955037,-10.49443583,1.219702470,-2.112586649,1.756836938,4.705330903,-8.149872191,-.3189574095,-4.265567617},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,.2458344589,-.4257977730,.5497025612,-.6504168418,.7375033768,1.054396484,-1.826268282,2.357702214,-2.789670881,2.510226735,-4.347840243,5.613037617,4.324472530,-7.490206138,5.881975759,.1420874210e-1,-.2461026322e-1,.3177171321e-1,-.3759279804e-1,.8700209506e-1,-.1506920490,.1945425987,.2894494276,-.5013411150,.6894088761,-.1216128451e-2,.2106396265e-2,-.2719345885e-2,-.9658561257e-2,.1672911883e-1,-.4108123658e-1,-.3746153966e-3,.6488529003e-3,-.3654099802e-2,-.4318072850e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,2.823642741,-4.890692689,6.313857110,-7.470656481,8.470928224,-3.210678176,5.561057729,-7.179294657,8.494655996,3.324354634,-5.757951125,7.433482942,-3.198010479,5.539116631,2.859723954,2.797956434,-4.846202701,6.256420786,-7.402696906,-2.907006833,5.035083533,-6.500264888,2.811341931,-4.869387063,-2.534878984,1.242384808,-2.151873609,2.778056884,-1.210043957,2.095857614,1.103053355,-1.284643645,2.225068063,1.185627028,-3.765218080},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.914096079,3.315311660,-4.280048947,5.064222210,-5.742288236,2.173949813,-3.765391528,4.861099562,-5.751730568,-2.246794550,3.891562317,-5.023985347,2.155283216,-3.733060035,-1.918656241,-2.823642742,4.890692690,-6.313857114,7.470656486,-8.470928225,3.210678177,-5.561057732,7.179294660,-8.494656000,-3.324354634,5.757951127,-7.433482942,3.198010480,-5.539116634,-2.859723955,2.797956438,-4.846202708,6.256420796,-7.402696917,-2.907006837,5.035083540,-6.500264898,2.811341934,-4.869387068,-2.534878988,-1.242384788,2.151873575,-2.778056840,1.210043938,-2.095857580,-1.103053339,-1.284643653,2.225068076,1.185627035,3.765217145},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,-1.488298306,2.577808283,-3.327936182,3.937667194,-4.464894918,.9564376257,-1.656598562,2.138659547,-2.530496102,-.7448321426e-1,.1290087115,-.1665495303,-.8305776036,1.438602609,1.441026240,3.069477200,-5.316490462,6.863559672,-8.121073323,9.208431600,-2.156281693,3.734789450,-4.821592446,5.704985119,.5614333335,-.9724310584,1.255403098,1.126462044,-1.951089492,-2.330362819,4.159176444,-7.203904918,9.300201260,-11.00414653,-1.935519673,3.352418412,-4.327953559,-.5514408000,.9551234833,2.488724224,4.693254424,-8.128955115,10.49443593,-1.219702482,2.112586669,-1.756836954,4.705330679,-8.149871802,-.3189573943,4.265566870},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-4.690415760,3.402394385,-5.893119942,7.607985127,-9.001889402,10.20718315,4.663809445,-8.077954914,10.42859495,-12.33927995,4.448012921,-7.704184374,9.946059258,2.665659156,-4.617057095,-.9771433915e-1,-.2458344586,.4257977724,-.5497025604,.6504168409,-.7375033757,-1.054396483,1.826268280,-2.357702211,2.789670877,-2.510226732,4.347840236,-5.613037612,-4.324472524,7.490206128,-5.881975750,.1420874260e-1,-.2461026409e-1,.3177171433e-1,-.3759279937e-1,.8700209816e-1,-.1506920544,.1945426056,.2894494379,-.5013411328,.6894089005,.1216122464e-2,-.2106385895e-2,.2719332498e-2,.9658513710e-2,-.1672903648e-1,.4108103434e-1,-.3745694178e-3,.6487732625e-3,-.3653651312e-2,.4309580089e-4}
		};
		for (int j = 0; j < 52; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 56; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{-.1148215401,-.1569113387e-1,-.1569113387e-1,-.1569113387e-1,-.1021689440,-.1021689440,-.1021689440,-.6587041484e-1,-.6587041484e-1,-.6587041484e-1,-.1032031823,-.1032031823,-.1032031823,-.7348534591e-1,-.7348534591e-1,-.7348534591e-1,-.3478299668e-1,-.3478299668e-1,-.3478299668e-1,-.8775204783e-2,-.8775204783e-2,-.8775204783e-2,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.8027748855e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.3747304424e-1,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.5845080675e-2,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.4674800411e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1,-.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{-.1753927996,-.2396860291e-1,-.2396860291e-1,-.2396860291e-1,-.1560656399,-.1560656399,-.1560656399,-.1006187207,-.1006187207,-.1006187207,-.1576454650,-.1576454650,-.1576454650,-.1122507200,-.1122507200,-.1122507200,-.5313190506e-1,-.5313190506e-1,-.5313190506e-1,-.1340434672e-1,-.1340434672e-1,-.1340434672e-1,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.1226258893,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.5724102059e-1,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.8928508212e-2,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.7140875581e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1,-.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{-.2198665100,-.3004623384e-1,-.3004623384e-1,-.3004623384e-1,-.1956386332,-.1956386332,-.1956386332,-.1261322415,-.1261322415,-.1261322415,-.1976190488,-.1976190488,-.1976190488,-.1407137244,-.1407137244,-.1407137244,-.6660436781e-1,-.6660436781e-1,-.6660436781e-1,-.1680323787e-1,-.1680323787e-1,-.1680323787e-1,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.1537196874,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.7175541674e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.1119247737e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.8951561274e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1,-.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,-.1092092502e-1,-.1092092502e-1,.2184185003e-1,-.3473628658e-1,-.3473628658e-1,.6947257305e-1,-.3832164410e-1,-.3832164410e-1,.7664328825e-1,.4062391359e-1,.4062391359e-1,-.8124782718e-1,.5779239332e-1,.5779239332e-1,-.1155847866,.3805405532e-1,.3805405532e-1,-.7610811063e-1,.1179212887e-1,.1179212887e-1,-.2358425774e-1,.1252607563e-1,-.9077889505e-1,.7825281938e-1,.7825281938e-1,.1252607563e-1,-.9077889505e-1,-.6978997413e-3,-.4910403324e-1,.4980193299e-1,.4980193299e-1,-.6978997413e-3,-.4910403324e-1,.3188726319e-2,-.1156256937e-1,.8373843054e-2,.8373843054e-2,.3188726319e-2,-.1156256937e-1,.2855524789e-1,-.8634838574e-1,.5779313787e-1,.5779313787e-1,.2855524789e-1,-.8634838574e-1,.1765812736e-1,-.4036819305e-1,.2271006571e-1,.2271006571e-1,.1765812736e-1,-.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,-.1668198852e-1,-.1668198852e-1,.3336397701e-1,-.5306055418e-1,-.5306055418e-1,.1061211082,-.5853727826e-1,-.5853727826e-1,.1170745566,.6205405300e-1,.6205405300e-1,-.1241081060,.8827933896e-1,.8827933896e-1,-.1765586779,.5812852966e-1,.5812852966e-1,-.1162570593,.1801277438e-1,.1801277438e-1,-.3602554875e-1,.1913389657e-1,-.1386670527,.1195331560,.1195331560,.1913389657e-1,-.1386670527,-.1066059464e-2,-.7500764972e-1,.7607370920e-1,.7607370920e-1,-.1066059464e-2,-.7500764972e-1,.4870859907e-2,-.1766211645e-1,.1279125655e-1,.1279125655e-1,.4870859907e-2,-.1766211645e-1,.4361886163e-1,-.1318993379,.8828047629e-1,.8828047629e-1,.4361886163e-1,-.1318993379,.2697323507e-1,-.6166343343e-1,.3469019838e-1,.3469019838e-1,.2697323507e-1,-.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{.1104870784,.1372943814e-1,.1372943814e-1,-.2604624096e-1,.1058560268,.1058560268,.1503953414e-1,.6279059801e-1,.6279059801e-1,-.6414708202e-1,.6180728788e-1,.6180728788e-1,.8198437210e-1,-.4481847772e-2,-.4481847772e-2,-.4131214635e-1,-.2730516345e-1,-.2730516345e-1,-.8534649204e-1,-.1322296188e-1,-.1322296188e-1,-.3971075014e-1,.6854799631e-1,.2788193386e-1,-.3874809825e-1,-.3874809825e-1,.6854799631e-1,.2788193386e-1,.3642583502e-1,.8712204498e-3,-.5473886199e-1,-.5473886199e-1,.3642583502e-1,.8712204498e-3,.2214798741e-2,-.1009100723e-1,-.1047747017e-1,-.1047747017e-1,.2214798741e-2,-.1009100723e-1,.1265572386e-1,-.6148813860e-1,-.5522608679e-1,-.5522608679e-1,.1265572386e-1,-.6148813860e-1,-.1134779051e-1,-.5533443179e-1,-.2579313736e-1,-.2579313736e-1,-.1134779051e-1,-.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{.1687718001,.2097206318e-1,.2097206318e-1,-.3978629025e-1,.1616977519,.1616977519,.2297326787e-1,.9591422281e-1,.9591422281e-1,-.9798628638e-1,.9441219172e-1,.9441219172e-1,.1252331970,-.6846135558e-2,-.6846135558e-2,-.6310534594e-1,-.4170932612e-1,-.4170932612e-1,-.1303689201,-.2019840792e-1,-.2019840792e-1,-.6065917284e-1,.1047087940,.4259035749e-1,-.5918869776e-1,-.5918869776e-1,.1047087940,.4259035749e-1,.5564138210e-1,.1330811220e-2,-.8361499282e-1,-.8361499282e-1,.5564138210e-1,.1330811220e-2,.3383160961e-2,-.1541426816e-1,-.1600460005e-1,-.1600460005e-1,.3383160961e-2,-.1541426816e-1,.1933193753e-1,-.9392468318e-1,-.8435924102e-1,-.8435924102e-1,.1933193753e-1,-.9392468318e-1,-.1733403633e-1,-.8452474073e-1,-.3939966813e-1,-.3939966813e-1,-.1733403633e-1,-.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{-.6804239411e-1,.1145848830e-1,.1145848830e-1,.2895473345e-1,.3479953926e-2,.3479953926e-2,-.8777788353e-1,.3406513033e-1,.3406513033e-1,.3473009539e-1,-.1014545622,-.1014545622,.9142437836e-1,-.5109385389e-1,-.5109385389e-1,.7796646760e-1,.7600422558e-2,.7600422558e-2,-.4998669741e-1,.1339603089e-1,.1339603089e-1,-.5386293723e-1,-.6480346250e-1,.1027298408,-.1762397985e-1,-.1762397985e-1,-.6480346250e-1,.1027298408,-.2109538461e-1,.4916147521e-1,.5370754828e-1,.5370754828e-1,-.2109538461e-1,.4916147521e-1,-.5671605418e-2,-.1354331144e-2,.1246131332e-1,.1246131332e-1,-.5671605418e-2,-.1354331144e-2,-.4342611086e-1,.1515827781e-1,.4183488198e-1,.4183488198e-1,-.4342611086e-1,.1515827781e-1,.8837167105e-3,-.5277408321e-1,.2665741414e-1,.2665741414e-1,.8837167105e-3,-.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{-.6656449560e-1,-.1342806488e-1,-.1342806488e-1,-.3079509387e-1,-.1065239097,-.1065239097,.9772524850e-1,-.6549601941e-1,-.6549601941e-1,.3360109025e-2,.4114868587e-1,.4114868587e-1,-.7177662214e-1,.7522679396e-1,.7522679396e-1,.1076722274,.1444928058e-1,.1444928058e-1,.1317043874e-1,-.1246734703e-1,-.1246734703e-1,-.6337331851e-1,-.1727352568e-1,.1680736381e-1,.6614705937e-1,.6614705937e-1,-.1727352568e-1,.1680736381e-1,-.2317222740e-1,.2883874010e-1,-.4751234575e-1,-.4751234575e-1,-.2317222740e-1,.2883874010e-1,.4353025737e-2,.8373369061e-2,-.1446295644e-1,-.1446295644e-1,.4353025737e-2,.8373369061e-2,.4030485339e-1,.7811447251e-1,-.2079224911e-1,-.2079224911e-1,.4030485339e-1,.7811447251e-1,.9854506346e-2,-.3162709295e-1,-.2557166486e-1,-.2557166486e-1,.9854506346e-2,-.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,.1891559699e-1,-.1891559699e-1,.3728221333e-11,.6016501311e-1,-.6016501307e-1,.2502617804e-10,.6637503462e-1,-.6637503465e-1,-.1129442338e-10,-.7036268232e-1,.7036268232e-1,-.2527951366e-10,-.1000993614,.1000993615,-0.,-.6591155720e-1,.6591155723e-1,-0.,-.2042456634e-1,.2042456633e-1,-0.,.9759050577e-1,-.3794735317e-1,-.5964315250e-1,.5964315260e-1,-.9759050570e-1,.3794735317e-1,.5710338621e-1,-.2915609201e-1,-.2794729420e-1,.2794729418e-1,-.5710338621e-1,.2915609203e-1,.1151029307e-1,-.2993628541e-2,-.8516664535e-2,.8516664535e-2,-.1151029308e-1,.2993628539e-2,.8322014746e-1,-.1688050365e-1,-.6633964381e-1,.6633964379e-1,-.8322014748e-1,.1688050367e-1,.3641824967e-1,-.2916737971e-2,-.3350151169e-1,.3350151170e-1,-.3641824966e-1,.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,.2889405168e-1,-.2889405167e-1,.5694952154e-11,.9190357559e-1,-.9190357554e-1,.3822811841e-10,.1013895401,-.1013895402,-.1725251670e-10,-.1074807726,.1074807726,-.3861509495e-10,-.1529043002,.1529043003,-0.,-.1006815667,.1006815667,-0.,-.3119904043e-1,.3119904041e-1,-0.,.1490719599,-.5796553947e-1,-.9110642037e-1,.9110642046e-1,-.1490719598,.5796553947e-1,.8722686330e-1,-.4453666621e-1,-.4269019705e-1,.4269019702e-1,-.8722686330e-1,.4453666622e-1,.1758226310e-1,-.4572843131e-2,-.1300941997e-1,.1300941997e-1,-.1758226310e-1,.4572843127e-2,.1271208750,-.2578539525e-1,-.1013354798,.1013354798,-.1271208750,.2578539527e-1,.5562979528e-1,-.4455390845e-2,-.5117440439e-1,.5117440444e-1,-.5562979523e-1,.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,.3444674858e-1,-.3444674856e-1,-.4446512362e-11,.7864938975e-1,-.7864938968e-1,-.4128474942e-11,.1099312557,-.1099312557,.9746388013e-11,-.1747386758e-1,.1747386758e-1,-.4936608124e-10,.3189597418e-1,-.3189597420e-1,-0.,.5026526501e-1,-.5026526506e-1,-0.,.2293909754e-1,-.2293909753e-1,-0.,.5770330047e-1,-.9292114363e-1,.3521784310e-1,-.3521784315e-1,-.5770330042e-1,.9292114363e-1,.4815974411e-1,-.7895094357e-1,.3079119953e-1,-.3079119951e-1,-.4815974411e-1,.7895094362e-1,.3346867218e-3,-.1099182731e-1,.1065714059e-1,-.1065714059e-1,-.3346867220e-3,.1099182730e-1,-.5423095927e-2,-.5878737247e-1,.6421046842e-1,-.6421046842e-1,.5423095927e-2,.5878737252e-1,-.2558351144e-1,-.1251003735e-1,.3809354879e-1,-.3809354879e-1,.2558351144e-1,.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,.5261827761e-1,-.5261827757e-1,-.6792159825e-11,.1201389273,-.1201389272,-.6306349641e-11,.1679227669,-.1679227669,.1488785361e-10,-.2669177363e-1,.2669177363e-1,-.7540793467e-10,.4872190534e-1,-.4872190538e-1,-0.,.7678146058e-1,-.7678146066e-1,-0.,.3504005029e-1,-.3504005026e-1,-0.,.8814324742e-1,-.1419393914,.5379614395e-1,-.5379614402e-1,-.8814324734e-1,.1419393914,.7356522427e-1,-.1205995584,.4703433420e-1,-.4703433416e-1,-.7356522427e-1,.1205995584,.5112424122e-3,-.1679029356e-1,.1627905114e-1,-.1627905114e-1,-.5112424126e-3,.1679029354e-1,-.8283915862e-2,-.8979919475e-1,.9808311064e-1,-.9808311064e-1,.8283915862e-2,.8979919482e-1,-.3907945924e-1,-.1910939770e-1,.5818885694e-1,-.5818885694e-1,.3907945924e-1,.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,.6060877170e-2,-.6060877164e-2,.4944073794e-11,-.3161264212e-1,.3161264209e-1,-.1584800096e-10,.2303508888e-3,-.2303508888e-3,-.5819067272e-11,.6681522508e-1,-.6681522508e-1,-.1528439436e-10,.4470780632e-1,-.4470780632e-1,-0.,-.1994876348e-1,.1994876348e-1,-0.,-.2329919003e-1,.2329919001e-1,-0.,-.9824065005e-1,-.7289231590e-1,.1486374908e-2,-.1486374909e-2,.9824064997e-1,.7289231590e-1,-.5309591204e-1,-.8058321394e-1,-.3033302763e-1,.3033302759e-1,.5309591204e-1,.8058321395e-1,-.9386233204e-2,-.2045354015e-1,-.1267656516e-1,.1267656516e-1,.9386233207e-2,.2045354012e-1,-.6111843873e-1,-.9989476013e-1,-.5006525860e-1,.5006525859e-1,.6111843873e-1,.9989476021e-1,.6661205800e-2,-.2978294282e-1,-.3944227928e-1,.3944227931e-1,-.6661205793e-2,.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,-.2989961082e-1,.2989961082e-1,-.5259821245e-11,-.7039060152e-1,.7039060149e-1,.2191617071e-10,-.1013287354,.1013287354,.6279229509e-12,-.4006762239e-1,.4006762239e-1,.3934858683e-10,-.8551663720e-1,.8551663723e-1,-0.,-.1484968306e-1,.1484968307e-1,-0.,.2177358772e-1,-.2177358771e-1,-0.,.4790109961e-2,.2984596792e-1,-.3463607749e-1,.3463607753e-1,-.4790109958e-2,-.2984596792e-1,-.2283783488e-1,-.4181200571e-2,.2701903535e-1,-.2701903532e-1,.2283783488e-1,.4181200574e-2,.9086687153e-2,-.2380170174e-1,.1471501461e-1,-.1471501461e-1,-.9086687153e-2,.2380170172e-1,.7038491177e-1,-.9773979388e-1,.2735488204e-1,-.2735488204e-1,-.7038491177e-1,.9773979394e-1,.1336506945e-1,-.5130832062e-1,.3794325106e-1,-.3794325110e-1,-.1336506945e-1,.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{.9882264718e-1,-.1440235391e-1,-.1440235396e-1,.3006820721e-1,.3375895322e-1,.3375895287e-1,.1352948785,-.2899076627e-1,-.2899076622e-1,.1129298746,.6881731865e-1,.6881731855e-1,.4625865234e-1,-.2871520174e-1,-.2871520191e-1,.1246232363e-1,-.6335778756e-1,-.6335778749e-1,.1534390562e-2,-.2959552517e-1,-.2959552520e-1,.1872238370e-4,-.3794484055e-1,.7521199400e-2,.8201583980e-1,.8201584015e-1,-.3794484065e-1,.7521199317e-2,-.5242564114e-1,-.1267437352e-1,.4949958872e-1,.4949958862e-1,-.5242564114e-1,-.1267437351e-1,-.1478824220e-1,-.1029932873e-2,-.5978541514e-3,-.5978541492e-3,-.1478824222e-1,-.1029932873e-2,-.8395409436e-1,-.1058736005e-2,-.8059922862e-2,-.8059922754e-2,-.8395409436e-1,-.1058736043e-2,-.4934714167e-1,-.1685816455e-3,-.3319675284e-1,-.3319675290e-1,-.4934714164e-1,-.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{.1509540871,-.2199995900e-1,-.2199995907e-1,.4592994519e-1,.5156765287e-1,.5156765232e-1,.2066663408,-.4428412697e-1,-.4428412690e-1,.1725032329,.1051201906,.1051201905,.7066125868e-1,-.4386319521e-1,-.4386319547e-1,.1903651379e-1,-.9678061914e-1,-.9678061909e-1,.2343820299e-2,-.4520791146e-1,-.4520791150e-1,.2859891351e-4,-.5796170139e-1,.1148882186e-1,.1252812647,.1252812653,-.5796170154e-1,.1148882174e-1,-.8008148967e-1,-.1936042536e-1,.7561187074e-1,.7561187061e-1,-.8008148966e-1,-.1936042534e-1,-.2258941311e-1,-.1573248451e-2,-.9132373015e-3,-.9132372980e-3,-.2258941314e-1,-.1573248449e-2,-.1282419975,-.1617245962e-2,-.1231173554e-1,-.1231173538e-1,-.1282419975,-.1617246020e-2,-.7537900402e-1,-.2575127172e-3,-.5070887761e-1,-.5070887768e-1,-.7537900399e-1,-.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{.1521474186,-.4127107599e-1,-.4127107613e-1,-.3344651713e-1,.7384207728e-1,.7384207651e-1,.3303033204e-1,-.7676670968e-1,-.7676670952e-1,-.7646932810e-1,.5434312853e-1,.5434312844e-1,.1406012128,-.1185729266e-2,-.1185729273e-2,.5653180033e-1,.3451242438e-1,.3451242437e-1,.8758681134e-2,.3020381423e-1,.3020381426e-1,.1246893713e-3,-.4714000659e-1,.2778316412e-1,-.2604070049e-1,-.2604070059e-1,-.4714000670e-1,.2778316383e-1,-.8257470907e-1,-.5115488591e-1,-.4912182046e-1,-.4912182038e-1,-.8257470907e-1,-.5115488588e-1,-.7397967795e-2,-.5467224064e-2,.7113199607e-3,.7113199580e-3,-.7397967806e-2,-.5467224056e-2,-.3155548387e-1,-.5355748779e-2,.6574391344e-2,.6574391251e-2,-.3155548387e-1,-.5355748971e-2,.2296424900e-1,-.1032248755e-2,.3449059684e-1,.3449059690e-1,.2296424899e-1,-.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{-.4252649632e-1,-.3621722030e-1,-.3621722042e-1,.3560242460e-1,.2348842407e-1,.2348842382e-1,-.1206537327,-.5391680609e-1,-.5391680599e-1,.2629256125e-1,-.6802515129e-1,-.6802515124e-1,.1427717610,.2147433158e-1,.2147433171e-1,.1219618694,.1982044929e-3,.1982044929e-3,.2648049029e-1,-.2844076554e-1,-.2844076556e-1,.4636110040e-3,.2835644722e-1,.4230017603e-1,-.2869463558e-1,-.2869463573e-1,.2835644729e-1,.4230017559e-1,.2013663457e-1,-.9116808757e-1,.4429713195e-1,.4429713188e-1,.2013663457e-1,-.9116808747e-1,.1463911665e-1,-.1492260701e-1,-.8259580003e-3,-.8259579974e-3,.1463911667e-1,-.1492260700e-1,.8008479954e-1,-.1363156197e-1,-.4081583933e-2,-.4081583877e-2,.8008479954e-1,-.1363156247e-1,.6542647961e-2,-.3434213984e-2,-.3335556699e-1,-.3335556704e-1,.6542647959e-2,-.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,.9258142764e-2,-.9258142654e-2,.8136908247e-10,-.4828910902e-1,.4828910902e-1,-0.,.3518672753e-3,-.3518668221e-3,.3485534728e-10,.1020619419,-.1020619425,-.5460999099e-9,.6829230417e-1,-.6829230235e-1,-0.,-.3047223942e-1,.3047223960e-1,-.3681086362e-9,-.3559010093e-1,.3559010063e-1,.4643401911e-10,-.2666533469e-1,.1205488497e-1,.1256702126,-.1256702128,.2666533393e-1,-.1205488583e-1,.3819594183e-1,-.3791605562e-2,.7296682180e-1,-.7296682193e-1,-.3819594232e-1,.3791605423e-2,.1658837758e-1,-.3172129929e-3,.1156231241e-1,-.1156231248e-1,-.1658837783e-1,.3172129864e-3,.6017702823e-1,.9452189607e-3,.7706103991e-1,-.7706103958e-1,-.6017702852e-1,-.9452190353e-3,.5568378542e-1,.1442892931e-4,-.1474045160e-1,.1474045137e-1,-.5568378595e-1,-.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,.3598418982e-1,-.3598418942e-1,-.8668398183e-10,-.1469158467,.1469158467,-0.,.1272304336e-2,-.1272302698e-2,-.1824763744e-10,.1329348046,-.1329348053,-.2240768661e-8,.2532572946e-1,-.2532572881e-1,-0.,.1092978850e-1,-.1092978857e-1,-.2733414349e-8,.3384746400e-1,-.3384746369e-1,.3998097089e-9,-.4975749160e-1,.5928626767e-1,-.9528779263e-2,.9528779277e-2,.4975749017e-1,-.5928627186e-1,.8709586801e-1,-.2026201294e-1,-.6683385514e-1,.6683385532e-1,-.8709586914e-1,.2026201219e-1,.1562966645e-1,-.2197560437e-2,-.1343210582e-1,.1343210590e-1,-.1562966667e-1,.2197560392e-2,.4738032804e-1,.6254379410e-2,-.5363470751e-1,.5363470732e-1,-.4738032827e-1,-.6254379903e-2,-.1447133378e-1,.1145953765e-3,.1435673898e-1,-.1435673875e-1,.1447133392e-1,-.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{-.5705528195e-1,-.5381888614e-2,-.5381890600e-2,-.3865155911e-1,.1066851510e-1,.1066851160e-1,-.1201849450,.1023929481e-1,.1023929537e-1,-.1298773899,.1146344760e-1,.1146344311e-1,-.1390909759e-1,.1113342295,.1113342303,-.1417761192e-2,.1805227680e-1,.1805227549e-1,-.4540571627e-4,-.3784627926e-1,-.3784627821e-1,-.2688892560e-7,.5375088148e-1,.3596809691e-2,-.1049770374e-2,-.1049766066e-2,.5375088190e-1,.3596800240e-2,-.2284622311e-1,-.5640425826e-3,-.1245758952e-1,-.1245758914e-1,-.2284622668e-1,-.5640434935e-3,-.1743701782e-1,-.9341106461e-4,.1604194740e-1,.1604194746e-1,-.1743701764e-1,-.9341075247e-4,-.2535811540e-1,.3827141178e-3,.1086557567,.1086557557,-.2535811250e-1,.3827130864e-3,-.5582706905e-1,.4444677760e-5,.1524183760e-1,.1524183769e-1,-.5582706642e-1,.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 52; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[56][52] = 
		{
			{-.6629224703e-1,-.9059280363e-2,-.9059280363e-2,-.9059280363e-2,-.5898726732e-1,-.5898726732e-1,-.5898726732e-1,-.3803030173e-1,-.3803030173e-1,-.3803030173e-1,-.5958438509e-1,-.5958438509e-1,-.5958438509e-1,-.4242678423e-1,-.4242678423e-1,-.4242678423e-1,-.2008197249e-1,-.2008197249e-1,-.2008197249e-1,-.5066366842e-2,-.5066366842e-2,-.5066366842e-2,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.4634822961e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.2163507217e-1,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.3374658900e-2,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.2698997275e-1,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2,-.9687440041e-2},
			{.1148215401,.1569113387e-1,.1569113387e-1,.1569113387e-1,.1021689440,.1021689440,.1021689440,.6587041484e-1,.6587041484e-1,.6587041484e-1,.1032031823,.1032031823,.1032031823,.7348534591e-1,.7348534591e-1,.7348534591e-1,.3478299668e-1,.3478299668e-1,.3478299668e-1,.8775204783e-2,.8775204783e-2,.8775204783e-2,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.8027748855e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.3747304424e-1,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.5845080675e-2,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.4674800411e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1,.1677913835e-1},
			{-.1482339708,-.2025716672e-1,-.2025716672e-1,-.2025716672e-1,-.1318995396,-.1318995396,-.1318995396,-.8503833989e-1,-.8503833989e-1,-.8503833989e-1,-.1332347355,-.1332347355,-.1332347355,-.9486917362e-1,-.9486917362e-1,-.9486917362e-1,-.4490465562e-1,-.4490465562e-1,-.4490465562e-1,-.1132874066e-1,-.1132874066e-1,-.1132874066e-1,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.1036377921,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.4837749209e-1,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.7545966703e-2,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.6035141379e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1,-.2166177446e-1},
			{.1753927996,.2396860291e-1,.2396860291e-1,.2396860291e-1,.1560656399,.1560656399,.1560656399,.1006187207,.1006187207,.1006187207,.1576454650,.1576454650,.1576454650,.1122507200,.1122507200,.1122507200,.5313190506e-1,.5313190506e-1,.5313190506e-1,.1340434672e-1,.1340434672e-1,.1340434672e-1,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.1226258893,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.5724102059e-1,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.8928508212e-2,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.7140875581e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1,.2563055720e-1},
			{-.1988767411,-.2717784109e-1,-.2717784109e-1,-.2717784109e-1,-.1769618019,-.1769618019,-.1769618019,-.1140909052,-.1140909052,-.1140909052,-.1787531553,-.1787531553,-.1787531553,-.1272803527,-.1272803527,-.1272803527,-.6024591747e-1,-.6024591747e-1,-.6024591747e-1,-.1519910052e-1,-.1519910052e-1,-.1519910052e-1,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.1390446888,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.6490521652e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.1012397670e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.8096991825e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1,-.2906232012e-1},
			{.2198665100,.3004623384e-1,.3004623384e-1,.3004623384e-1,.1956386332,.1956386332,.1956386332,.1261322415,.1261322415,.1261322415,.1976190488,.1976190488,.1976190488,.1407137244,.1407137244,.1407137244,.6660436781e-1,.6660436781e-1,.6660436781e-1,.1680323787e-1,.1680323787e-1,.1680323787e-1,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.1537196874,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.7175541674e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.1119247737e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.8951561274e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1,.3212960380e-1},
			{-0.,-.6305198993e-2,-.6305198993e-2,.1261039799e-1,-.2005500436e-1,-.2005500436e-1,.4011000875e-1,-.2212501156e-1,-.2212501156e-1,.4425002309e-1,.2345422744e-1,.2345422744e-1,-.4690845492e-1,.3336645383e-1,.3336645383e-1,-.6673290766e-1,.2197051908e-1,.2197051908e-1,-.4394103815e-1,.6808188776e-2,.6808188776e-2,-.1361637755e-1,.7231933122e-2,-.5241121945e-1,.4517928632e-1,.4517928632e-1,.7231933122e-2,-.5241121945e-1,-.4029326034e-3,-.2835022679e-1,.2875315941e-1,.2875315941e-1,-.4029326034e-3,-.2835022679e-1,.1841011997e-2,-.6675652538e-2,.4834640540e-2,.4834640540e-2,.1841011997e-2,-.6675652538e-2,.1648638004e-1,-.4985326374e-1,.3336688370e-1,.3336688370e-1,.1648638004e-1,-.4985326374e-1,.1019492458e-1,-.2330658713e-1,.1311166255e-1,.1311166255e-1,.1019492458e-1,-.2330658713e-1},
			{-0.,.1092092502e-1,.1092092502e-1,-.2184185003e-1,.3473628658e-1,.3473628658e-1,-.6947257305e-1,.3832164410e-1,.3832164410e-1,-.7664328825e-1,-.4062391359e-1,-.4062391359e-1,.8124782718e-1,-.5779239332e-1,-.5779239332e-1,.1155847866,-.3805405532e-1,-.3805405532e-1,.7610811063e-1,-.1179212887e-1,-.1179212887e-1,.2358425774e-1,-.1252607563e-1,.9077889505e-1,-.7825281938e-1,-.7825281938e-1,-.1252607563e-1,.9077889505e-1,.6978997413e-3,.4910403324e-1,-.4980193299e-1,-.4980193299e-1,.6978997413e-3,.4910403324e-1,-.3188726319e-2,.1156256937e-1,-.8373843054e-2,-.8373843054e-2,-.3188726319e-2,.1156256937e-1,-.2855524789e-1,.8634838574e-1,-.5779313787e-1,-.5779313787e-1,-.2855524789e-1,.8634838574e-1,-.1765812736e-1,.4036819305e-1,-.2271006571e-1,-.2271006571e-1,-.1765812736e-1,.4036819305e-1},
			{-0.,-.1409885357e-1,-.1409885357e-1,.2819770712e-1,-.4484435312e-1,-.4484435312e-1,.8968870610e-1,-.4947302977e-1,-.4947302977e-1,.9894605961e-1,.5244524690e-1,.5244524690e-1,-.1048904938,.7460965892e-1,.7460965892e-1,-.1492193178,.4912757415e-1,.4912757415e-1,-.9825514828e-1,.1522357290e-1,.1522357290e-1,-.3044714580e-1,.1617109409e-1,-.1171950495,.1010239554,.1010239554,.1617109409e-1,-.1171950495,-.9009846914e-3,-.6339303429e-1,.6429401899e-1,.6429401899e-1,-.9009846914e-3,-.6339303429e-1,.4116627974e-2,-.1492721287e-1,.1081058489e-1,.1081058489e-1,.4116627974e-2,-.1492721287e-1,.3686466649e-1,-.1114752866,.7461062013e-1,.7461062013e-1,.3686466649e-1,-.1114752866,.2279654438e-1,-.5211511311e-1,.2931856875e-1,.2931856875e-1,.2279654438e-1,-.5211511311e-1},
			{-0.,.1668198852e-1,.1668198852e-1,-.3336397701e-1,.5306055418e-1,.5306055418e-1,-.1061211082,.5853727826e-1,.5853727826e-1,-.1170745566,-.6205405300e-1,-.6205405300e-1,.1241081060,-.8827933896e-1,-.8827933896e-1,.1765586779,-.5812852966e-1,-.5812852966e-1,.1162570593,-.1801277438e-1,-.1801277438e-1,.3602554875e-1,-.1913389657e-1,.1386670527,-.1195331560,-.1195331560,-.1913389657e-1,.1386670527,.1066059464e-2,.7500764972e-1,-.7607370920e-1,-.7607370920e-1,.1066059464e-2,.7500764972e-1,-.4870859907e-2,.1766211645e-1,-.1279125655e-1,-.1279125655e-1,-.4870859907e-2,.1766211645e-1,-.4361886163e-1,.1318993379,-.8828047629e-1,-.8828047629e-1,-.4361886163e-1,.1318993379,-.2697323507e-1,.6166343343e-1,-.3469019838e-1,-.3469019838e-1,-.2697323507e-1,.6166343343e-1},
			{-0.,-.1891559700e-1,-.1891559700e-1,.3783119399e-1,-.6016501319e-1,-.6016501319e-1,.1203300262,-.6637503460e-1,-.6637503460e-1,.1327500692,.7036268232e-1,.7036268232e-1,-.1407253646,.1000993615,.1000993615,-.2001987230,.6591155723e-1,.6591155723e-1,-.1318231145,.2042456633e-1,.2042456633e-1,-.4084913264e-1,.2169579941e-1,-.1572336584,.1355378589,.1355378589,.2169579941e-1,-.1572336584,-.1208797810e-2,-.8505068041e-1,.8625947822e-1,.8625947822e-1,-.1208797810e-2,-.8505068041e-1,.5523035994e-2,-.2002695761e-1,.1450392162e-1,.1450392162e-1,.5523035994e-2,-.2002695761e-1,.4945914013e-1,-.1495597912,.1001006511,.1001006511,.4945914013e-1,-.1495597912,.3058477373e-1,-.6991976135e-1,.3933498764e-1,.3933498764e-1,.3058477373e-1,-.6991976135e-1},
			{.6378974448e-1,.7926694808e-2,.7926694808e-2,-.1503780423e-1,.6111600559e-1,.6111600559e-1,.8683079085e-2,.3625216869e-1,.3625216869e-1,-.3703533510e-1,.3568445429e-1,.3568445429e-1,.4733369933e-1,-.2587596019e-2,-.2587596019e-2,-.2385157883e-1,-.1576464347e-1,-.1576464347e-1,-.4927482018e-1,-.7634280606e-2,-.7634280606e-2,-.2292701230e-1,.3957620414e-1,.1609764203e-1,-.2237122496e-1,-.2237122496e-1,.3957620414e-1,.1609764203e-1,.2103046567e-1,.5029993615e-3,-.3160349670e-1,-.3160349670e-1,.2103046567e-1,.5029993615e-3,.1278714650e-2,-.5826045744e-2,-.6049170223e-2,-.6049170223e-2,.1278714650e-2,-.5826045744e-2,.7306785581e-2,-.3550019339e-1,-.3188479608e-1,-.3188479608e-1,.7306785581e-2,-.3550019339e-1,-.6551649907e-2,-.3194734909e-1,-.1489167480e-1,-.1489167480e-1,-.6551649907e-2,-.3194734909e-1},
			{-.1104870784,-.1372943814e-1,-.1372943814e-1,.2604624096e-1,-.1058560268,-.1058560268,-.1503953414e-1,-.6279059801e-1,-.6279059801e-1,.6414708202e-1,-.6180728788e-1,-.6180728788e-1,-.8198437210e-1,.4481847772e-2,.4481847772e-2,.4131214635e-1,.2730516345e-1,.2730516345e-1,.8534649204e-1,.1322296188e-1,.1322296188e-1,.3971075014e-1,-.6854799631e-1,-.2788193386e-1,.3874809825e-1,.3874809825e-1,-.6854799631e-1,-.2788193386e-1,-.3642583502e-1,-.8712204498e-3,.5473886199e-1,.5473886199e-1,-.3642583502e-1,-.8712204498e-3,-.2214798741e-2,.1009100723e-1,.1047747017e-1,.1047747017e-1,-.2214798741e-2,.1009100723e-1,-.1265572386e-1,.6148813860e-1,.5522608679e-1,.5522608679e-1,-.1265572386e-1,.6148813860e-1,.1134779051e-1,.5533443179e-1,.2579313736e-1,.2579313736e-1,.1134779051e-1,.5533443179e-1},
			{.1426382049,.1772462843e-1,.1772462843e-1,-.3362555248e-1,.1366595430,.1366595430,.1941595509e-1,.8106231350e-1,.8106231350e-1,-.8281352685e-1,.7979286553e-1,.7979286553e-1,.1058413693,-.5786040596e-2,-.5786040596e-2,-.5333375162e-1,-.3525081444e-1,-.3525081444e-1,-.1101818475,-.1707077039e-1,-.1707077039e-1,-.5126635801e-1,.8849508273e-1,.3599542184e-1,-.5002357975e-1,-.5002357975e-1,.8849508273e-1,.3599542184e-1,.4702555082e-1,.1124740765e-2,-.7066756694e-1,-.7066756694e-1,.4702555082e-1,.1124740765e-2,.2859292880e-2,-.1302743432e-1,-.1352635582e-1,-.1352635582e-1,.2859292880e-2,-.1302743432e-1,.1633846925e-1,-.7938084562e-1,-.7129657147e-1,-.7129657147e-1,.1633846925e-1,-.7938084562e-1,-.1464993456e-1,-.7143644426e-1,-.3329879716e-1,-.3329879716e-1,-.1464993456e-1,-.7143644426e-1},
			{-.1687718001,-.2097206318e-1,-.2097206318e-1,.3978629025e-1,-.1616977519,-.1616977519,-.2297326787e-1,-.9591422281e-1,-.9591422281e-1,.9798628638e-1,-.9441219172e-1,-.9441219172e-1,-.1252331970,.6846135558e-2,.6846135558e-2,.6310534594e-1,.4170932612e-1,.4170932612e-1,.1303689201,.2019840792e-1,.2019840792e-1,.6065917284e-1,-.1047087940,-.4259035749e-1,.5918869776e-1,.5918869776e-1,-.1047087940,-.4259035749e-1,-.5564138210e-1,-.1330811220e-2,.8361499282e-1,.8361499282e-1,-.5564138210e-1,-.1330811220e-2,-.3383160961e-2,.1541426816e-1,.1600460005e-1,.1600460005e-1,-.3383160961e-2,.1541426816e-1,-.1933193753e-1,.9392468318e-1,.8435924102e-1,.8435924102e-1,-.1933193753e-1,.9392468318e-1,.1733403633e-1,.8452474073e-1,.3939966813e-1,.3939966813e-1,.1733403633e-1,.8452474073e-1},
			{-.3928429454e-1,.6615561300e-2,.6615561300e-2,.1671702314e-1,.2009152335e-2,.2009152335e-2,-.5067858467e-1,.1966751216e-1,.1966751216e-1,.2005142992e-1,-.5857481876e-1,-.5857481876e-1,.5278388944e-1,-.2949905028e-1,-.2949905028e-1,.4501396104e-1,.4388106009e-2,.4388106009e-2,-.2885983319e-1,.7734202039e-2,.7734202039e-2,-.3109778130e-1,-.3741429651e-1,.5931110120e-1,-.1017520951e-1,-.1017520951e-1,-.3741429651e-1,.5931110120e-1,-.1217942598e-1,.2838339094e-1,.3100806745e-1,.3100806745e-1,-.1217942598e-1,.2838339094e-1,-.3274502914e-2,-.7819234506e-3,.7194542597e-2,.7194542597e-2,-.3274502914e-2,-.7819234506e-3,-.2507207679e-1,.8751635773e-2,.2415338037e-1,.2415338037e-1,-.2507207679e-1,.8751635773e-2,.5102140805e-3,-.3046913114e-1,.1539066522e-1,.1539066522e-1,.5102140805e-3,-.3046913114e-1},
			{.6804239411e-1,-.1145848830e-1,-.1145848830e-1,-.2895473345e-1,-.3479953926e-2,-.3479953926e-2,.8777788353e-1,-.3406513033e-1,-.3406513033e-1,-.3473009539e-1,.1014545622,.1014545622,-.9142437836e-1,.5109385389e-1,.5109385389e-1,-.7796646760e-1,-.7600422558e-2,-.7600422558e-2,.4998669741e-1,-.1339603089e-1,-.1339603089e-1,.5386293723e-1,.6480346250e-1,-.1027298408,.1762397985e-1,.1762397985e-1,.6480346250e-1,-.1027298408,.2109538461e-1,-.4916147521e-1,-.5370754828e-1,-.5370754828e-1,.2109538461e-1,-.4916147521e-1,.5671605418e-2,.1354331144e-2,-.1246131332e-1,-.1246131332e-1,.5671605418e-2,.1354331144e-2,.4342611086e-1,-.1515827781e-1,-.4183488198e-1,-.4183488198e-1,.4342611086e-1,-.1515827781e-1,-.8837167105e-3,.5277408321e-1,-.2665741414e-1,-.2665741414e-1,-.8837167105e-3,.5277408321e-1},
			{-.8784235306e-1,.1479284478e-1,.1479284478e-1,.3738040014e-1,.4492601200e-2,.4492601200e-2,-.1133207603,.4397789414e-1,.4397789414e-1,.4483636035e-1,-.1309772765,-.1309772765,.1180283649,-.6596188172e-1,-.6596188172e-1,.1006542768,.9812103330e-2,.9812103330e-2,-.6453254885e-1,.1729420151e-1,.1729420151e-1,-.6953675294e-1,-.8366091034e-1,.1326236541,-.2275246015e-1,-.2275246015e-1,-.8366091034e-1,.1326236541,-.2723402442e-1,.6346719158e-1,.6933614668e-1,.6933614668e-1,-.2723402442e-1,.6346719158e-1,-.7322011109e-2,-.1748433989e-2,.1608748632e-1,.1608748632e-1,-.7322011109e-2,-.1748433989e-2,-.5606286805e-1,.1956925251e-1,.5400860040e-1,.5400860040e-1,-.5606286805e-1,.1956925251e-1,.1140873367e-2,-.6813104845e-1,.3441457367e-1,.3441457367e-1,.1140873367e-2,-.6813104845e-1},
			{-.3843102945e-1,-.7752696870e-2,-.7752696870e-2,-.1777955573e-1,-.6150160792e-1,-.6150160792e-1,.5642169852e-1,-.3781414444e-1,-.3781414444e-1,.1939959850e-2,.2375720486e-1,.2375720486e-1,-.4144025211e-1,.4343220974e-1,.4343220974e-1,.6216458946e-1,.8342296029e-2,.8342296029e-2,.7603956351e-2,-.7198026162e-2,-.7198026162e-2,-.3658860250e-1,-.9972874699e-2,.9703736018e-2,.3819002253e-1,.3819002253e-1,-.9972874699e-2,.9703736018e-2,-.1337849172e-1,.1665005436e-1,-.2743126561e-1,-.2743126561e-1,-.1337849172e-1,.1665005436e-1,.2513220581e-2,.4834366881e-2,-.8350191794e-2,-.8350191794e-2,.2513220581e-2,.4834366881e-2,.2327001795e-1,.4509941173e-1,-.1200441062e-1,-.1200441062e-1,.2327001795e-1,.4509941173e-1,.5689501892e-2,-.1825991063e-1,-.1476380759e-1,-.1476380759e-1,.5689501892e-2,-.1825991063e-1},
			{.6656449560e-1,.1342806488e-1,.1342806488e-1,.3079509387e-1,.1065239097,.1065239097,-.9772524850e-1,.6549601941e-1,.6549601941e-1,-.3360109025e-2,-.4114868587e-1,-.4114868587e-1,.7177662214e-1,-.7522679396e-1,-.7522679396e-1,-.1076722274,-.1444928058e-1,-.1444928058e-1,-.1317043874e-1,.1246734703e-1,.1246734703e-1,.6337331851e-1,.1727352568e-1,-.1680736381e-1,-.6614705937e-1,-.6614705937e-1,.1727352568e-1,-.1680736381e-1,.2317222740e-1,-.2883874010e-1,.4751234575e-1,.4751234575e-1,.2317222740e-1,-.2883874010e-1,-.4353025737e-2,-.8373369061e-2,.1446295644e-1,.1446295644e-1,-.4353025737e-2,-.8373369061e-2,-.4030485339e-1,-.7811447251e-1,.2079224911e-1,.2079224911e-1,-.4030485339e-1,-.7811447251e-1,-.9854506346e-2,.3162709295e-1,.2557166486e-1,.2557166486e-1,-.9854506346e-2,.3162709295e-1},
			{.6415098438e-1,-.6677637612e-2,-.6677637612e-2,.1828956849e-1,.1812295099e-1,.1812295099e-1,-.2237381796e-1,-.1590719212e-1,-.1590719212e-1,-.2340644828e-1,.3417872293e-1,.3417872293e-1,-.5666049222e-1,-.3109343555e-1,-.3109343555e-1,.7294130587e-2,-.1843512010e-1,-.1843512010e-1,.3961603168e-1,.6111788234e-2,.6111788234e-2,-.3831187635e-1,.4528437006e-1,-.5509956014e-1,-.4988063885e-1,-.4988063885e-1,.4528437006e-1,-.5509956014e-1,.2060313504e-1,-.1847766698e-1,.2137716413e-1,.2137716413e-1,.2060313504e-1,-.1847766698e-1,.3382692282e-3,.6695837349e-2,.9566303944e-2,.9566303944e-2,.3382692282e-3,.6695837349e-2,-.3542270067e-2,.4085563164e-1,-.2260821834e-2,-.2260821834e-2,-.3542270067e-2,.4085563164e-1,-.9871108351e-2,.6693485532e-3,.1314289991e-1,.1314289991e-1,-.9871108351e-2,.6693485532e-3},
			{-0.,.1092092502e-1,-.1092092502e-1,.2152489590e-11,.3473628652e-1,-.3473628651e-1,.1444887063e-10,.3832164411e-1,-.3832164413e-1,-.6520838383e-11,-.4062391359e-1,.4062391359e-1,-.1459513402e-10,-.5779239326e-1,.5779239332e-1,-0.,-.3805405529e-1,.3805405532e-1,-0.,-.1179212888e-1,.1179212887e-1,-0.,.5634390479e-1,-.2190891458e-1,-.3443499017e-1,.3443499021e-1,-.5634390474e-1,.2190891458e-1,.3296865543e-1,-.1683327758e-1,-.1613537783e-1,.1613537782e-1,-.3296865543e-1,.1683327758e-1,.6645470808e-2,-.1728372245e-2,-.4917098564e-2,.4917098564e-2,-.6645470808e-2,.1728372243e-2,.4804717454e-1,-.9745963331e-2,-.3830121122e-1,.3830121122e-1,-.4804717454e-1,.9745963337e-2,.2102608626e-1,-.1683979453e-2,-.1934210679e-1,.1934210681e-1,-.2102608624e-1,.1683979451e-2},
			{-0.,-.1891559699e-1,.1891559699e-1,-.3728221333e-11,-.6016501311e-1,.6016501307e-1,-.2502617804e-10,-.6637503462e-1,.6637503465e-1,.1129442338e-10,.7036268232e-1,-.7036268232e-1,.2527951366e-10,.1000993614,-.1000993615,0.,.6591155720e-1,-.6591155723e-1,0.,.2042456634e-1,-.2042456633e-1,0.,-.9759050577e-1,.3794735317e-1,.5964315250e-1,-.5964315260e-1,.9759050570e-1,-.3794735317e-1,-.5710338621e-1,.2915609201e-1,.2794729420e-1,-.2794729418e-1,.5710338621e-1,-.2915609203e-1,-.1151029307e-1,.2993628541e-2,.8516664535e-2,-.8516664535e-2,.1151029308e-1,-.2993628539e-2,-.8322014746e-1,.1688050365e-1,.6633964381e-1,-.6633964379e-1,.8322014748e-1,-.1688050367e-1,-.3641824967e-1,.2916737971e-2,.3350151169e-1,-.3350151170e-1,.3641824966e-1,-.2916737967e-2},
			{-0.,.2441993071e-1,-.2441993070e-1,.4813113043e-11,.7767269793e-1,-.7767269789e-1,.3230865693e-10,.8568980122e-1,-.8568980126e-1,-.1458103789e-10,-.9083783227e-1,.9083783227e-1,-.3263571179e-10,-.1292277199,.1292277200,-0.,-.8509145442e-1,.8509145450e-1,-0.,-.2636800176e-1,.2636800175e-1,-0.,.1259888012,-.4898982231e-1,-.7699897880e-1,.7699897888e-1,-.1259888011,.4898982231e-1,.7372015464e-1,-.3764035293e-1,-.3607980167e-1,.3607980165e-1,-.7372015464e-1,.3764035295e-1,.1485972446e-1,-.3864757828e-2,-.1099496664e-1,.1099496664e-1,-.1485972446e-1,.3864757825e-2,.1074367484,-.2179263651e-1,-.8564411187e-1,.8564411187e-1,-.1074367484,.2179263652e-1,.4701575817e-1,-.3765492529e-2,-.4325026560e-1,.4325026564e-1,-.4701575813e-1,.3765492524e-2},
			{-0.,-.2889405168e-1,.2889405167e-1,-.5694952154e-11,-.9190357559e-1,.9190357554e-1,-.3822811841e-10,-.1013895401,.1013895402,.1725251670e-10,.1074807726,-.1074807726,.3861509495e-10,.1529043002,-.1529043003,0.,.1006815667,-.1006815667,0.,.3119904043e-1,-.3119904041e-1,0.,-.1490719599,.5796553947e-1,.9110642037e-1,-.9110642046e-1,.1490719598,-.5796553947e-1,-.8722686330e-1,.4453666621e-1,.4269019705e-1,-.4269019702e-1,.8722686330e-1,-.4453666622e-1,-.1758226310e-1,.4572843131e-2,.1300941997e-1,-.1300941997e-1,.1758226310e-1,-.4572843127e-2,-.1271208750,.2578539525e-1,.1013354798,-.1013354798,.1271208750,-.2578539527e-1,-.5562979528e-1,.4455390845e-2,.5117440439e-1,-.5117440444e-1,.5562979523e-1,-.4455390839e-2},
			{-0.,.3276277505e-1,-.3276277505e-1,.6457468772e-11,.1042088596,-.1042088595,.4334661190e-10,.1149649323,-.1149649324,-.1956251514e-10,-.1218717408,.1218717408,-.4378540206e-10,-.1733771798,.1733771800,-0.,-.1141621659,.1141621660,-0.,-.3537638663e-1,.3537638661e-1,-0.,.1690317144,-.6572674372e-1,-.1033049705,.1033049707,-.1690317143,.6572674372e-1,.9890596622e-1,-.5049983272e-1,-.4840613350e-1,.4840613347e-1,-.9890596622e-1,.5049983276e-1,.1993641241e-1,-.5185116733e-2,-.1475129569e-1,.1475129569e-1,-.1993641243e-1,.5185116730e-2,.1441415236,-.2923788999e-1,-.1149036337,.1149036336,-.1441415237,.2923789002e-1,.6307825877e-1,-.5051938359e-2,-.5802632039e-1,.5802632041e-1,-.6307825875e-1,.5051938352e-2},
			{-0.,.1988783956e-1,-.1988783955e-1,-.2567195108e-11,.4540824633e-1,-.4540824629e-1,-.2383576118e-11,.6346884007e-1,-.6346884007e-1,.5627079741e-11,-.1008854215e-1,.1008854215e-1,-.2850152028e-10,.1841514927e-1,-.1841514929e-1,-0.,.2902066428e-1,-.2902066430e-1,-0.,.1324389414e-1,-.1324389413e-1,-0.,.3331501605e-1,-.5364804727e-1,.2033303119e-1,-.2033303122e-1,-.3331501602e-1,.5364804727e-1,.2780504122e-1,-.4558234851e-1,.1777730733e-1,-.1777730732e-1,-.2780504122e-1,.4558234853e-1,.1932314689e-3,-.6346134454e-2,.6152902984e-2,-.6152902984e-2,-.1932314690e-3,.6346134447e-2,-.3131025892e-2,-.3394090531e-1,.3707193121e-1,-.3707193121e-1,.3131025892e-2,.3394090534e-1,-.1477064721e-1,-.7222673430e-2,.2199332064e-1,-.2199332064e-1,.1477064721e-1,.7222673422e-2},
			{-0.,-.3444674858e-1,.3444674856e-1,.4446512362e-11,-.7864938975e-1,.7864938968e-1,.4128474942e-11,-.1099312557,.1099312557,-.9746388013e-11,.1747386758e-1,-.1747386758e-1,.4936608124e-10,-.3189597418e-1,.3189597420e-1,0.,-.5026526501e-1,.5026526506e-1,0.,-.2293909754e-1,.2293909753e-1,0.,-.5770330047e-1,.9292114363e-1,-.3521784310e-1,.3521784315e-1,.5770330042e-1,-.9292114363e-1,-.4815974411e-1,.7895094357e-1,-.3079119953e-1,.3079119951e-1,.4815974411e-1,-.7895094362e-1,-.3346867218e-3,.1099182731e-1,-.1065714059e-1,.1065714059e-1,.3346867220e-3,-.1099182730e-1,.5423095927e-2,.5878737247e-1,-.6421046842e-1,.6421046842e-1,-.5423095927e-2,-.5878737252e-1,.2558351144e-1,.1251003735e-1,-.3809354879e-1,.3809354879e-1,-.2558351144e-1,-.1251003734e-1},
			{-0.,.4447056119e-1,-.4447056116e-1,-.5740422774e-11,.1015359256,-.1015359255,-.5329838231e-11,.1419206409,-.1419206409,.1258253282e-10,-.2255866604e-1,.2255866604e-1,-.6373133682e-10,.4117752560e-1,-.4117752563e-1,-0.,.6489217809e-1,-.6489217815e-1,-0.,.2961424758e-1,-.2961424756e-1,-0.,.7449464057e-1,-.1199606806,.4546603994e-1,-.4546604000e-1,-.7449464051e-1,.1199606806,.6217396229e-1,-.1019252299,.3975126766e-1,-.3975126763e-1,-.6217396229e-1,.1019252299,.4320786998e-3,-.1419038804e-1,.1375830933e-1,-.1375830933e-1,-.4320787002e-3,.1419038802e-1,-.7001186736e-2,-.7589417150e-1,.8289535827e-1,-.8289535827e-1,.7001186736e-2,.7589417157e-1,-.3302817125e-1,-.1615038877e-1,.4917856001e-1,-.4917856001e-1,.3302817125e-1,.1615038875e-1},
			{-0.,-.5261827761e-1,.5261827757e-1,.6792159825e-11,-.1201389273,.1201389272,.6306349641e-11,-.1679227669,.1679227669,-.1488785361e-10,.2669177363e-1,-.2669177363e-1,.7540793467e-10,-.4872190534e-1,.4872190538e-1,0.,-.7678146058e-1,.7678146066e-1,0.,-.3504005029e-1,.3504005026e-1,0.,-.8814324742e-1,.1419393914,-.5379614395e-1,.5379614402e-1,.8814324734e-1,-.1419393914,-.7356522427e-1,.1205995584,-.4703433420e-1,.4703433416e-1,.7356522427e-1,-.1205995584,-.5112424122e-3,.1679029356e-1,-.1627905114e-1,.1627905114e-1,.5112424126e-3,-.1679029354e-1,.8283915862e-2,.8979919475e-1,-.9808311064e-1,.9808311064e-1,-.8283915862e-2,-.8979919482e-1,.3907945924e-1,.1910939770e-1,-.5818885694e-1,.5818885694e-1,-.3907945924e-1,-.1910939768e-1},
			{-0.,.3499249066e-2,-.3499249064e-2,.2854462337e-11,-.1825156744e-1,.1825156743e-1,-.9149847624e-11,.1329931477e-3,-.1329931477e-3,-.3359640058e-11,.3857578819e-1,-.3857578819e-1,-.8824449203e-11,.2581206402e-1,-.2581206402e-1,-0.,-.1151742398e-1,.1151742398e-1,-0.,-.1345179364e-1,.1345179363e-1,-0.,-.5671926576e-1,-.4208439823e-1,.8581589533e-3,-.8581589543e-3,.5671926571e-1,.4208439823e-1,-.3065493912e-1,-.4652474025e-1,-.1751278167e-1,.1751278166e-1,.3065493912e-1,.4652474030e-1,-.5419144268e-2,-.1180885691e-1,-.7318818308e-2,.7318818308e-2,.5419144270e-2,.1180885690e-1,-.3528674706e-1,-.5767426670e-1,-.2890519053e-1,.2890519053e-1,.3528674706e-1,.5767426674e-1,.3845848962e-2,-.1719519006e-1,-.2277201058e-1,.2277201058e-1,-.3845848960e-2,.1719519004e-1},
			{-0.,-.6060877170e-2,.6060877164e-2,-.4944073794e-11,.3161264212e-1,-.3161264209e-1,.1584800096e-10,-.2303508888e-3,.2303508888e-3,.5819067272e-11,-.6681522508e-1,.6681522508e-1,.1528439436e-10,-.4470780632e-1,.4470780632e-1,0.,.1994876348e-1,-.1994876348e-1,0.,.2329919003e-1,-.2329919001e-1,0.,.9824065005e-1,.7289231590e-1,-.1486374908e-2,.1486374909e-2,-.9824064997e-1,-.7289231590e-1,.5309591204e-1,.8058321394e-1,.3033302763e-1,-.3033302759e-1,-.5309591204e-1,-.8058321395e-1,.9386233204e-2,.2045354015e-1,.1267656516e-1,-.1267656516e-1,-.9386233207e-2,-.2045354012e-1,.6111843873e-1,.9989476013e-1,.5006525860e-1,-.5006525859e-1,-.6111843873e-1,-.9989476021e-1,-.6661205800e-2,.2978294282e-1,.3944227928e-1,-.3944227931e-1,.6661205793e-2,-.2978294278e-1},
			{-0.,.7824558781e-2,-.7824558776e-2,.6382771823e-11,-.4081174549e-1,.4081174547e-1,-.2045968127e-10,.2973817188e-3,-.2973817188e-3,-.7512383548e-11,.8625808467e-1,-.8625808467e-1,-.1973206828e-10,.5771752979e-1,-.5771752979e-1,-0.,-.2575374293e-1,.2575374293e-1,-0.,-.3007912500e-1,.3007912498e-1,-0.,-.1268281339,-.9410357522e-1,.1918901755e-2,-.1918901757e-2,.1268281337,.9410357522e-1,-.6854652772e-1,-.1040324818,-.3915977029e-1,.3915977026e-1,.6854652772e-1,.1040324819,-.1211757496e-1,-.2640540679e-1,-.1636537525e-1,.1636537525e-1,.1211757497e-1,.2640540677e-1,-.7890356511e-1,-.1289635809,-.6463397093e-1,.6463397093e-1,.7890356511e-1,.1289635810,.8599579709e-2,-.3844961385e-1,-.5091976363e-1,.5091976363e-1,-.8599579704e-2,.3844961381e-1},
			{-0.,-.1726254836e-1,.1726254835e-1,-.3036759212e-11,-.4064003275e-1,.4064003272e-1,.1265330706e-10,-.5850217264e-1,.5850217264e-1,.3625314848e-12,-.2313305258e-1,.2313305258e-1,.2271791720e-10,-.4937305351e-1,.4937305353e-1,-0.,-.8573468510e-2,.8573468521e-2,-0.,.1257098673e-1,-.1257098673e-1,-0.,.2765571274e-2,.1723157762e-1,-.1999714867e-1,.1999714868e-1,-.2765571273e-2,-.1723157762e-1,-.1318543011e-1,-.2414017276e-2,.1559944734e-1,-.1559944732e-1,.1318543011e-1,.2414017277e-2,.5246201273e-2,-.1374191891e-1,.8495717644e-2,-.8495717644e-2,-.5246201275e-2,.1374191889e-1,.4063674774e-1,-.5643009631e-1,.1579334851e-1,-.1579334851e-1,-.4063674774e-1,.5643009636e-1,.7716326443e-2,-.2962287271e-1,.2190654622e-1,-.2190654623e-1,-.7716326443e-2,.2962287268e-1},
			{-0.,.2989961082e-1,-.2989961082e-1,.5259821245e-11,.7039060152e-1,-.7039060149e-1,-.2191617071e-10,.1013287354,-.1013287354,-.6279229509e-12,.4006762239e-1,-.4006762239e-1,-.3934858683e-10,.8551663720e-1,-.8551663723e-1,0.,.1484968306e-1,-.1484968307e-1,0.,-.2177358772e-1,.2177358771e-1,0.,-.4790109961e-2,-.2984596792e-1,.3463607749e-1,-.3463607753e-1,.4790109958e-2,.2984596792e-1,.2283783488e-1,.4181200571e-2,-.2701903535e-1,.2701903532e-1,-.2283783488e-1,-.4181200574e-2,-.9086687153e-2,.2380170174e-1,-.1471501461e-1,.1471501461e-1,.9086687153e-2,-.2380170172e-1,-.7038491177e-1,.9773979388e-1,-.2735488204e-1,.2735488204e-1,.7038491177e-1,-.9773979394e-1,-.1336506945e-1,.5130832062e-1,-.3794325106e-1,.3794325110e-1,.1336506945e-1,-.5130832053e-1},
			{-0.,-.6549542859e-2,.6549542859e-2,.3125048446e-11,.2474749778e-1,-.2474749776e-1,-.6523037303e-11,-.4134369484e-2,.4134369484e-2,.2615992430e-11,-.1581775699e-1,.1581775699e-1,.1478280308e-10,.4030268983e-1,-.4030268983e-1,0.,.2506983936e-1,-.2506983937e-1,-0.,-.1074824591e-1,.1074824591e-1,-0.,.5270056993e-1,.5127448905e-1,.2937809615e-1,-.2937809618e-1,-.5270056989e-1,-.5127448905e-1,.3612819854e-1,.4282129605e-1,-.1231409337e-1,.1231409336e-1,-.3612819854e-1,-.4282129607e-1,-.4248825618e-3,-.9042030486e-2,-.9734733542e-2,.9734733542e-2,.4248825618e-3,.9042030479e-2,-.1241991071e-1,-.2079208603e-1,-.2280719716e-3,.2280719716e-3,.1241991071e-1,.2079208605e-1,-.1596311243e-1,-.4029185479e-1,-.1958926507e-1,.1958926508e-1,.1596311242e-1,.4029185473e-1},
			{.5705528195e-1,-.8315202908e-2,-.8315202935e-2,.1735988752e-1,.1949074073e-1,.1949074053e-1,.7811253455e-1,-.1673782671e-1,-.1673782668e-1,.6520009352e-1,.3973169745e-1,.3973169740e-1,.2670744539e-1,-.1657872946e-1,-.1657872956e-1,.7195125899e-2,-.3657963570e-1,-.3657963568e-1,.8858808038e-3,-.1708698443e-1,-.1708698444e-1,.1080937327e-4,-.2190746391e-1,.4342366498e-2,.4735186720e-1,.4735186739e-1,-.2190746397e-1,.4342366452e-2,-.3026795803e-1,-.7317552966e-2,.2857860087e-1,.2857860082e-1,-.3026795803e-1,-.7317552960e-2,-.8537995620e-2,-.5946320215e-3,-.3451712553e-3,-.3451712540e-3,-.8537995631e-2,-.5946320210e-3,-.4847091898e-1,-.6112615175e-3,-.4653398635e-2,-.4653398572e-2,-.4847091899e-1,-.6112615395e-3,-.2849058552e-1,-.9733065843e-4,-.1916615420e-1,-.1916615422e-1,-.2849058551e-1,-.9733066188e-4},
			{-.9882264718e-1,.1440235391e-1,.1440235396e-1,-.3006820721e-1,-.3375895322e-1,-.3375895287e-1,-.1352948785,.2899076627e-1,.2899076622e-1,-.1129298746,-.6881731865e-1,-.6881731855e-1,-.4625865234e-1,.2871520174e-1,.2871520191e-1,-.1246232363e-1,.6335778756e-1,.6335778749e-1,-.1534390562e-2,.2959552517e-1,.2959552520e-1,-.1872238370e-4,.3794484055e-1,-.7521199400e-2,-.8201583980e-1,-.8201584015e-1,.3794484065e-1,-.7521199317e-2,.5242564114e-1,.1267437352e-1,-.4949958872e-1,-.4949958862e-1,.5242564114e-1,.1267437351e-1,.1478824220e-1,.1029932873e-2,.5978541514e-3,.5978541492e-3,.1478824222e-1,.1029932873e-2,.8395409436e-1,.1058736005e-2,.8059922862e-2,.8059922754e-2,.8395409436e-1,.1058736043e-2,.4934714167e-1,.1685816455e-3,.3319675284e-1,.3319675290e-1,.4934714164e-1,.1685816515e-3},
			{.1275794890,-.1859335895e-1,-.1859335901e-1,.3881788860e-1,.4358262121e-1,.4358262077e-1,.1746649372,-.3742691832e-1,-.3742691825e-1,.1457918412,.8884277635e-1,.8884277625e-1,.5971966340e-1,-.3707116604e-1,-.3707116626e-1,.1608879062e-1,-.8179455202e-1,-.8179455197e-1,.1980889697e-2,-.3820765871e-1,-.3820765876e-1,.2417049342e-4,-.4898657853e-1,.9709826674e-2,.1058819939,.1058819943,-.4898657866e-1,.9709826569e-2,-.6768121175e-1,-.1636254586e-1,.6390369425e-1,.6390369415e-1,-.6768121170e-1,-.1636254585e-1,-.1909153859e-1,-.1329637622e-2,-.7718263907e-3,-.7718263880e-3,-.1909153862e-1,-.1329637620e-2,-.1083842698,-.1366822305e-2,-.1040531567e-1,-.1040531553e-1,-.1083842698,-.1366822354e-2,-.6370688596e-1,-.2176379685e-3,-.4285682365e-1,-.4285682373e-1,-.6370688593e-1,-.2176379762e-3},
			{-.1509540871,.2199995900e-1,.2199995907e-1,-.4592994519e-1,-.5156765287e-1,-.5156765232e-1,-.2066663408,.4428412697e-1,.4428412690e-1,-.1725032329,-.1051201906,-.1051201905,-.7066125868e-1,.4386319521e-1,.4386319547e-1,-.1903651379e-1,.9678061914e-1,.9678061909e-1,-.2343820299e-2,.4520791146e-1,.4520791150e-1,-.2859891351e-4,.5796170139e-1,-.1148882186e-1,-.1252812647,-.1252812653,.5796170154e-1,-.1148882174e-1,.8008148967e-1,.1936042536e-1,-.7561187074e-1,-.7561187061e-1,.8008148966e-1,.1936042534e-1,.2258941311e-1,.1573248451e-2,.9132373015e-3,.9132372980e-3,.2258941314e-1,.1573248449e-2,.1282419975,.1617245962e-2,.1231173554e-1,.1231173538e-1,.1282419975,.1617246020e-2,.7537900402e-1,.2575127172e-3,.5070887761e-1,.5070887768e-1,.7537900399e-1,.2575127264e-3},
			{.8784235306e-1,-.2382786683e-1,-.2382786691e-1,-.1931035567e-1,.4263274319e-1,.4263274275e-1,.1907007110e-1,-.4432128050e-1,-.4432128040e-1,-.4414958716e-1,.3137501988e-1,.3137501983e-1,.8117614807e-1,-.6845811107e-3,-.6845811148e-3,.3263865014e-1,.1992575751e-1,.1992575750e-1,.5056826910e-2,.1743818028e-1,.1743818029e-1,.7198944209e-4,-.2721629549e-1,.1604061729e-1,-.1503460544e-1,-.1503460550e-1,-.2721629555e-1,.1604061712e-1,-.4767453051e-1,-.2953428715e-1,-.2836049626e-1,-.2836049622e-1,-.4767453051e-1,-.2953428713e-1,-.4271218698e-2,-.3156503285e-2,.4106807708e-3,.4106807692e-3,-.4271218704e-2,-.3156503281e-2,-.1821856711e-1,-.3092142999e-2,.3795726612e-2,.3795726558e-2,-.1821856711e-1,-.3092143110e-2,.1325841534e-1,-.5959690968e-3,.1991315537e-1,.1991315540e-1,.1325841533e-1,-.5959691180e-3},
			{-.1521474186,.4127107599e-1,.4127107613e-1,.3344651713e-1,-.7384207728e-1,-.7384207651e-1,-.3303033204e-1,.7676670968e-1,.7676670952e-1,.7646932810e-1,-.5434312853e-1,-.5434312844e-1,-.1406012128,.1185729266e-2,.1185729273e-2,-.5653180033e-1,-.3451242438e-1,-.3451242437e-1,-.8758681134e-2,-.3020381423e-1,-.3020381426e-1,-.1246893713e-3,.4714000659e-1,-.2778316412e-1,.2604070049e-1,.2604070059e-1,.4714000670e-1,-.2778316383e-1,.8257470907e-1,.5115488591e-1,.4912182046e-1,.4912182038e-1,.8257470907e-1,.5115488588e-1,.7397967795e-2,.5467224064e-2,-.7113199607e-3,-.7113199580e-3,.7397967806e-2,.5467224056e-2,.3155548387e-1,.5355748779e-2,-.6574391344e-2,-.6574391251e-2,.3155548387e-1,.5355748971e-2,-.2296424900e-1,.1032248755e-2,-.3449059684e-1,-.3449059690e-1,-.2296424899e-1,.1032248792e-2},
			{.1964214726,-.5328072997e-1,-.5328073014e-1,-.4317926793e-1,.9532971186e-1,.9532971087e-1,.4264197529e-1,-.9910539600e-1,-.9910539584e-1,-.9872147804e-1,.7015667723e-1,.7015667712e-1,.1815153852,-.1530769900e-2,-.1530769908e-2,.7298224036e-1,.4455534829e-1,.4455534826e-1,.1130740872e-1,.3899295649e-1,.3899295653e-1,.1609732862e-3,-.6085748681e-1,.3586791064e-1,-.3361839976e-1,-.3361839990e-1,-.6085748695e-1,.3586791026e-1,-.1066034910,-.6604067373e-1,-.6341599750e-1,-.6341599740e-1,-.1066034910,-.6604067369e-1,-.9550735356e-2,-.7058155916e-2,.9183101206e-3,.9183101170e-3,-.9550735365e-2,-.7058155904e-2,-.4073795449e-1,-.6914241940e-2,.8487502725e-2,.8487502609e-2,-.4073795449e-1,-.6914242189e-2,.2964671797e-1,-.1332627413e-2,.4452716905e-1,.4452716912e-1,.2964671795e-1,-.1332627460e-2},
			{-.2455268409e-1,-.2091002188e-1,-.2091002195e-1,.2055506942e-1,.1356104795e-1,.1356104781e-1,-.6965946504e-1,-.3112888250e-1,-.3112888244e-1,.1518001731e-1,-.3927433939e-1,-.3927433937e-1,.8242931462e-1,.1239821111e-1,.1239821119e-1,.7041471809e-1,.1144334173e-3,.1144334173e-3,.1528851819e-1,-.1642028363e-1,-.1642028365e-1,.2676659379e-3,.1637160243e-1,.2442201801e-1,-.1656685557e-1,-.1656685566e-1,.1637160247e-1,.2442201776e-1,.1162589139e-1,-.5263591988e-1,.2557496105e-1,.2557496101e-1,.1162589139e-1,-.5263591982e-1,.8451897935e-2,-.8615571172e-2,-.4768670737e-3,-.4768670720e-3,.8451897947e-2,-.8615571165e-2,.4623698055e-1,-.7870185971e-2,-.2356503582e-2,-.2356503549e-2,.4623698055e-1,-.7870186258e-2,.3777399560e-2,-.1982744368e-2,-.1925784557e-1,-.1925784560e-1,.3777399559e-2,-.1982744438e-2},
			{.4252649632e-1,.3621722030e-1,.3621722042e-1,-.3560242460e-1,-.2348842407e-1,-.2348842382e-1,.1206537327,.5391680609e-1,.5391680599e-1,-.2629256125e-1,.6802515129e-1,.6802515124e-1,-.1427717610,-.2147433158e-1,-.2147433171e-1,-.1219618694,-.1982044929e-3,-.1982044929e-3,-.2648049029e-1,.2844076554e-1,.2844076556e-1,-.4636110040e-3,-.2835644722e-1,-.4230017603e-1,.2869463558e-1,.2869463573e-1,-.2835644729e-1,-.4230017559e-1,-.2013663457e-1,.9116808757e-1,-.4429713195e-1,-.4429713188e-1,-.2013663457e-1,.9116808747e-1,-.1463911665e-1,.1492260701e-1,.8259580003e-3,.8259579974e-3,-.1463911667e-1,.1492260700e-1,-.8008479954e-1,.1363156197e-1,.4081583933e-2,.4081583877e-2,-.8008479954e-1,.1363156247e-1,-.6542647961e-2,.3434213984e-2,.3335556699e-1,.3335556704e-1,-.6542647959e-2,.3434214107e-2},
			{-.7172298094e-1,.7825782221e-2,.7825782248e-2,-.2116827304e-1,-.3452157079e-1,-.3452157044e-1,.5353368348e-1,.2358452775e-1,.2358452771e-1,.1456970694e-1,-.1134355155e-2,-.1134355154e-2,-.1080888597e-1,-.1345623877e-1,-.1345623885e-1,.8828442067e-1,-.1757049739e-1,-.1757049738e-1,.3207110342e-1,.1421672136e-1,.1421672137e-1,.7341622498e-3,.1742099043e-1,.1372110241e-1,.3559988215e-1,.3559988227e-1,.1742099047e-1,.1372110226e-1,.3952709529e-1,-.4515668313e-1,-.2064715052e-1,-.2064715049e-1,.3952709529e-1,-.4515668309e-1,-.3441468367e-2,-.1566496243e-1,.5466084410e-3,.5466084391e-3,-.3441468372e-2,-.1566496242e-1,-.2689444059e-1,-.1290670793e-1,.6112228954e-3,.6112228873e-3,-.2689444061e-1,-.1290670839e-1,-.1704789669e-1,-.4722488405e-2,.1736405931e-1,.1736405934e-1,-.1704789669e-1,-.4722488575e-2},
			{-0.,.5345191218e-2,-.5345191154e-2,.4697846168e-10,-.2787973010e-1,.2787973010e-1,-0.,.2031506661e-3,-.2031504045e-3,.2012374414e-10,.5892548964e-1,-.5892548997e-1,-.3152909300e-9,.3942858020e-1,-.3942857915e-1,-0.,-.1759315563e-1,.1759315574e-1,-.2125276202e-9,-.2054795436e-1,.2054795418e-1,.2680869344e-10,-.1539523816e-1,.6959891086e-2,.7255573106e-1,-.7255573117e-1,.1539523772e-1,-.6959891579e-2,.2205243730e-1,-.2189084492e-2,.4212741421e-1,-.4212741429e-1,-.2205243759e-1,.2189084412e-2,.9577304262e-2,-.1831430069e-3,.6675504182e-2,-.6675504223e-2,-.9577304404e-2,.1831430031e-3,.3474322345e-1,.5457224215e-3,.4449121214e-1,-.4449121195e-1,-.3474322362e-1,-.5457224645e-3,.3214904850e-1,.8330546222e-5,-.8510403700e-2,.8510403566e-2,-.3214904881e-1,-.8330505214e-5},
			{-0.,-.9258142764e-2,.9258142654e-2,-.8136908247e-10,.4828910902e-1,-.4828910902e-1,0.,-.3518672753e-3,.3518668221e-3,-.3485534728e-10,-.1020619419,.1020619425,.5460999099e-9,-.6829230417e-1,.6829230235e-1,0.,.3047223942e-1,-.3047223960e-1,.3681086362e-9,.3559010093e-1,-.3559010063e-1,-.4643401911e-10,.2666533469e-1,-.1205488497e-1,-.1256702126,.1256702128,-.2666533393e-1,.1205488583e-1,-.3819594183e-1,.3791605562e-2,-.7296682180e-1,.7296682193e-1,.3819594232e-1,-.3791605423e-2,-.1658837758e-1,.3172129929e-3,-.1156231241e-1,.1156231248e-1,.1658837783e-1,-.3172129864e-3,-.6017702823e-1,-.9452189607e-3,-.7706103991e-1,.7706103958e-1,.6017702852e-1,.9452190353e-3,-.5568378542e-1,-.1442892931e-4,.1474045160e-1,-.1474045137e-1,.5568378595e-1,.1442885828e-4},
			{-0.,.1195221091e-1,-.1195221077e-1,.1050470338e-9,-.6234097168e-1,.6234097168e-1,-0.,.4542586991e-3,-.4542581141e-3,.4499805985e-10,.1317614004,-.1317614012,-.7050119522e-9,.8816498557e-1,-.8816498323e-1,-0.,-.3933949193e-1,.3933949216e-1,-.4752262059e-9,-.4594662274e-1,.4594662234e-1,.5994606091e-10,-.3442479906e-1,.1556278958e-1,.1622395468,-.1622395470,.3442479807e-1,-.1556279069e-1,.4931074886e-1,-.4894941733e-2,.9419976188e-1,-.9419976205e-1,-.4931074951e-1,.4894941553e-2,.2141550337e-1,-.4095202129e-3,.1492688113e-1,-.1492688123e-1,-.2141550369e-1,.4095202046e-3,.7768820939e-1,.1220272431e-2,.9948537474e-1,-.9948537432e-1,-.7768820976e-1,-.1220272527e-2,.7188745786e-1,.1862766764e-4,-.1902984119e-1,.1902984089e-1,-.7188745855e-1,-.1862757594e-4},
			{-0.,.2077548167e-1,-.2077548144e-1,-.5004702023e-10,-.8482190361e-1,.8482190361e-1,-0.,.7345652507e-3,-.7345643049e-3,-.1053527838e-10,.7674994521e-1,-.7674994560e-1,-.1293708389e-8,.1462181672e-1,-.1462181634e-1,-0.,.6310316331e-2,-.6310316369e-2,-.1578137510e-8,.1954184244e-1,-.1954184227e-1,.2308302430e-9,-.2872750116e-1,.3422894259e-1,-.5501443271e-2,.5501443279e-2,.2872750034e-1,-.3422894501e-1,.5028482283e-1,-.1169827862e-1,-.3858654425e-1,.3858654435e-1,-.5028482348e-1,.1169827819e-1,.9023792131e-2,-.1268762110e-2,-.7755029908e-2,.7755029953e-2,-.9023792256e-2,.1268762084e-2,.2735504514e-1,.3610967635e-2,-.3096601281e-1,.3096601270e-1,-.2735504527e-1,-.3610967920e-2,-.8355028450e-2,.6616167146e-4,.8288867110e-2,-.8288866978e-2,.8355028531e-2,-.6616134575e-4},
			{-0.,-.3598418982e-1,.3598418942e-1,.8668398183e-10,.1469158467,-.1469158467,0.,-.1272304336e-2,.1272302698e-2,.1824763744e-10,-.1329348046,.1329348053,.2240768661e-8,-.2532572946e-1,.2532572881e-1,0.,-.1092978850e-1,.1092978857e-1,.2733414349e-8,-.3384746400e-1,.3384746369e-1,-.3998097089e-9,.4975749160e-1,-.5928626767e-1,.9528779263e-2,-.9528779277e-2,-.4975749017e-1,.5928627186e-1,-.8709586801e-1,.2026201294e-1,.6683385514e-1,-.6683385532e-1,.8709586914e-1,-.2026201219e-1,-.1562966645e-1,.2197560437e-2,.1343210582e-1,-.1343210590e-1,.1562966667e-1,-.2197560392e-2,-.4738032804e-1,-.6254379410e-2,.5363470751e-1,-.5363470732e-1,.4738032827e-1,.6254379903e-2,.1447133378e-1,-.1145953765e-3,-.1435673898e-1,.1435673875e-1,-.1447133392e-1,.1145948124e-3},
			{-0.,.3049813486e-1,-.3049813448e-1,.5159489251e-10,-.7284083561e-1,.7284083561e-1,0.,.9365245507e-3,-.9365233445e-3,.2547635812e-12,-.4535362990e-1,.4535363013e-1,-.2097135016e-8,-.3646628667e-1,.3646628574e-1,-0.,.4224940278e-2,-.4224940304e-2,-.6160896976e-8,-.1719394538e-1,.1719394523e-1,.1085920617e-8,.1774744259e-2,.7496052877e-1,-.4278522505e-1,.4278522511e-1,-.1774744209e-2,-.7496053404e-1,.1400845960e-1,-.2905963061e-1,.3201571278e-1,-.3201571285e-1,-.1400845978e-1,.2905962954e-1,-.9257865704e-2,-.4524719828e-2,.8894085672e-2,-.8894085727e-2,.9257865841e-2,.4524719734e-2,-.3468187011e-1,.1211453038e-1,.1405391517e-1,-.1405391512e-1,.3468187029e-1,-.1211453133e-1,-.1203442885e-1,.2814087114e-3,-.7555975650e-2,.7555975540e-2,.1203442897e-1,-.2814073261e-3},
			{-.3294088239e-1,-.3107234840e-2,-.3107235986e-2,-.2231548805e-1,.6159470063e-2,.6159468046e-2,-.6938881037e-1,.5911659616e-2,.5911659936e-2,-.7498474602e-1,.6618424559e-2,.6618421963e-2,-.8030421240e-2,.6427884738e-1,.6427884785e-1,-.8185448057e-3,.1042248687e-1,.1042248611e-1,-.2621500251e-4,-.2185055951e-1,-.2185055891e-1,-.1552432843e-7,.3103308589e-1,.2076619043e-2,-.6060852082e-3,-.6060827207e-3,.3103308613e-1,.2076613587e-2,-.1319027306e-1,-.3256501369e-3,-.7192392665e-2,-.7192392441e-2,-.1319027512e-1,-.3256506628e-3,-.1006726693e-1,-.5393090330e-4,.9261822647e-2,.9261822685e-2,-.1006726683e-1,-.5393072308e-4,-.1464051475e-1,.2209600989e-3,.6273243036e-1,.6273242982e-1,-.1464051308e-1,.2209595034e-3,-.3223177334e-1,.2566135901e-5,.8799879043e-2,.8799879093e-2,-.3223177183e-1,.2565820943e-5},
			{.5705528195e-1,.5381888614e-2,.5381890600e-2,.3865155911e-1,-.1066851510e-1,-.1066851160e-1,.1201849450,-.1023929481e-1,-.1023929537e-1,.1298773899,-.1146344760e-1,-.1146344311e-1,.1390909759e-1,-.1113342295,-.1113342303,.1417761192e-2,-.1805227680e-1,-.1805227549e-1,.4540571627e-4,.3784627926e-1,.3784627821e-1,.2688892560e-7,-.5375088148e-1,-.3596809691e-2,.1049770374e-2,.1049766066e-2,-.5375088190e-1,-.3596800240e-2,.2284622311e-1,.5640425826e-3,.1245758952e-1,.1245758914e-1,.2284622668e-1,.5640434935e-3,.1743701782e-1,.9341106461e-4,-.1604194740e-1,-.1604194746e-1,.1743701764e-1,.9341075247e-4,.2535811540e-1,-.3827141178e-3,-.1086557567,-.1086557557,.2535811250e-1,-.3827130864e-3,.5582706905e-1,-.4444677760e-5,-.1524183760e-1,-.1524183769e-1,.5582706642e-1,-.4444132237e-5},
			{-.9622647658e-1,-.1521904615e-1,-.1521905176e-1,.2303665143e-1,.2394074360e-1,.2394073576e-1,-.6868961150e-1,.2703716563e-1,.2703716708e-1,.2875691115e-1,.1193436050e-1,.1193435581e-1,-.4141413255e-1,.4419346599e-1,.4419346630e-1,-.6047823382e-2,-.1939598041e-2,-.1939597903e-2,-.2394938610e-3,.1956630608e-1,.1956630553e-1,-.1638512340e-6,.7690057838e-1,.1273571901e-1,-.9250753169e-4,-.9250715204e-4,.7690057901e-1,.1273568554e-1,-.3922896378e-1,-.2163267888e-2,.6138290615e-2,.6138290424e-2,-.3922896992e-1,-.2163271380e-2,-.1380977785e-1,-.4605464129e-3,-.1063029164e-1,-.1063029169e-1,-.1380977770e-1,-.4605448736e-3,-.1736811689e-1,.1804648916e-2,-.3701532946e-1,-.3701532913e-1,-.1736811492e-1,.1804644052e-2,.2184875653e-2,.2503078297e-4,-.8121609812e-2,-.8121609862e-2,.2184875549e-2,.2502771078e-4},
			{-0.,.1674979297e-2,-.1674982434e-2,.7799173872e-9,.8552224190e-2,-.8552212976e-2,.8146652029e-8,-.5371317491e-2,.5371318813e-2,.1853757791e-9,-.3772692781e-1,.3772692287e-1,-0.,.4276550920e-1,-.4276550437e-1,-.5101217004e-8,.3604660856e-1,-.3604661115e-1,-.6036428485e-9,-.2085360269e-1,.2085360354e-1,.7968451864e-10,-.2421703160e-1,-.4950513841e-3,-.7201207141e-1,.7201209823e-1,.2421706142e-1,.4950595926e-3,.6007831857e-2,.1187381795e-3,-.5445880325e-1,.5445879711e-1,-.6007817507e-2,-.1187357396e-3,.1020354507e-1,-.1554709471e-4,.3739408882e-2,-.3739408860e-2,-.1020354540e-1,.1554696392e-4,-.4847648031e-2,.3514398769e-4,.2906736482e-1,-.2906737056e-1,.4847640569e-2,-.3514585146e-4,.2921937085e-1,.2957903456e-6,.2579194922e-1,-.2579194283e-1,-.2921936573e-1,-.2952085868e-6}
		};
		for (int j = 0; j < 56; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 52; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
