
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (3)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (20)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (20)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (4)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (13)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*3+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[20],
		T io_hv[20]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{260.000002,-329.0896534,214.6625259,-52.91502622,-416.4132569,254.5584411,-65.72670691,340.8225342,-80.49844713,-105.8300523,-416.4132569,254.5584411,-65.72670691,278.2804344,-65.72670689,-81.9756061,340.8225342,-80.49844713,-81.9756061,-105.8300523},
			{-329.0896534,1020.,-1068.943403,366.6060557,254.5584412,-489.8979486,227.6839914,-80.49844717,92.95160030,0.,254.5584412,-489.8979486,227.6839914,-65.72670689,75.89466382,0.,-80.49844717,92.95160030,0.,0.},
			{214.6625259,-1068.943403,1584.,-709.9295738,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,0.,0.,0.,0.},
			{-52.91502622,366.6060557,-709.9295738,400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-416.4132569,254.5584412,-65.72670689,0.,1220.000000,-415.6921939,53.66563145,-1366.103948,197.1801207,518.4592560,610.000000,-207.8460970,26.83281572,-743.6128021,107.3312629,301.1976094,-455.3679830,65.72670689,200.7984063,129.6148141},
			{254.5584411,-489.8979486,227.6839914,0.,-415.6921939,640.,-185.9032006,197.1801207,-227.6839914,0.,-207.8460970,320.,-92.95160030,107.3312629,-123.9354671,0.,65.72670689,-75.89466382,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,53.66563145,-185.9032006,144.,0.,0.,0.,26.83281572,-92.95160030,72.,0.,0.,0.,0.,0.,0.,0.},
			{340.8225342,-80.49844717,0.,0.,-1366.103948,197.1801207,0.,1872.,-124.7076582,-819.7560616,-455.3679830,65.72670689,0.,764.2407998,-50.91168823,-380.9881888,312.,-20.78460970,-190.4940944,-81.97560616},
			{-80.49844713,92.95160030,0.,0.,197.1801207,-227.6839914,0.,-124.7076582,144.,0.,65.72670689,-75.89466382,0.,-50.91168823,58.78775383,0.,-20.78460970,24.,0.,0.},
			{-105.8300523,0.,0.,0.,518.4592560,0.,0.,-819.7560616,0.,400.,129.6148141,0.,0.,-267.7312084,0.,154.9193338,-81.97560616,0.,61.96773354,20.},
			{-416.4132569,254.5584412,-65.72670689,0.,610.000000,-207.8460970,26.83281572,-455.3679830,65.72670689,129.6148141,1220.000000,-415.6921939,53.66563145,-743.6128021,107.3312629,200.7984063,-1366.103948,197.1801207,301.1976094,518.4592560},
			{254.5584411,-489.8979486,227.6839914,0.,-207.8460970,320.,-92.95160030,65.72670689,-75.89466382,0.,-415.6921939,640.,-185.9032006,107.3312629,-123.9354671,0.,197.1801207,-227.6839914,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,26.83281572,-92.95160030,72.,0.,0.,0.,53.66563145,-185.9032006,144.,0.,0.,0.,0.,0.,0.,0.},
			{278.2804344,-65.72670689,0.,0.,-743.6128021,107.3312629,0.,764.2407998,-50.91168823,-267.7312084,-743.6128021,107.3312629,0.,832.,-55.42562586,-311.0755536,764.2407998,-50.91168823,-311.0755536,-267.7312084},
			{-65.72670689,75.89466382,0.,0.,107.3312629,-123.9354671,0.,-50.91168823,58.78775383,0.,107.3312629,-123.9354671,0.,-55.42562586,64.,0.,-50.91168823,58.78775383,0.,0.},
			{-81.9756061,0.,0.,0.,301.1976094,0.,0.,-380.9881888,0.,154.9193338,200.7984063,0.,0.,-311.0755536,0.,144.,-190.4940944,0.,108.,61.96773354},
			{340.8225342,-80.49844717,0.,0.,-455.3679830,65.72670689,0.,312.,-20.78460970,-81.97560616,-1366.103948,197.1801207,0.,764.2407998,-50.91168823,-190.4940944,1872.,-124.7076582,-380.9881888,-819.7560616},
			{-80.49844713,92.95160030,0.,0.,65.72670689,-75.89466382,0.,-20.78460970,24.,0.,197.1801207,-227.6839914,0.,-50.91168823,58.78775383,0.,-124.7076582,144.,0.,0.},
			{-81.9756061,0.,0.,0.,200.7984063,0.,0.,-190.4940944,0.,61.96773354,301.1976094,0.,0.,-311.0755536,0.,108.,-380.9881888,0.,144.,154.9193338},
			{-105.8300523,0.,0.,0.,129.6148141,0.,0.,-81.97560616,0.,20.,518.4592560,0.,0.,-267.7312084,0.,61.96773354,-819.7560616,0.,154.9193338,400.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{260.000002,-329.0896534,214.6625259,-52.91502622,-416.4132569,254.5584411,-65.72670691,340.8225342,-80.49844713,-105.8300523,-416.4132569,254.5584411,-65.72670691,278.2804344,-65.72670689,-81.9756061,340.8225342,-80.49844713,-81.9756061,-105.8300523},
			{-329.0896534,1020.,-1068.943403,366.6060557,254.5584412,-489.8979486,227.6839914,-80.49844717,92.95160030,0.,254.5584412,-489.8979486,227.6839914,-65.72670689,75.89466382,0.,-80.49844717,92.95160030,0.,0.},
			{214.6625259,-1068.943403,1584.,-709.9295738,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,0.,0.,0.,0.},
			{-52.91502622,366.6060557,-709.9295738,400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-416.4132569,254.5584412,-65.72670689,0.,1220.000000,-415.6921939,53.66563145,-1366.103948,197.1801207,518.4592560,610.000000,-207.8460970,26.83281572,-743.6128021,107.3312629,301.1976094,-455.3679830,65.72670689,200.7984063,129.6148141},
			{254.5584411,-489.8979486,227.6839914,0.,-415.6921939,640.,-185.9032006,197.1801207,-227.6839914,0.,-207.8460970,320.,-92.95160030,107.3312629,-123.9354671,0.,65.72670689,-75.89466382,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,53.66563145,-185.9032006,144.,0.,0.,0.,26.83281572,-92.95160030,72.,0.,0.,0.,0.,0.,0.,0.},
			{340.8225342,-80.49844717,0.,0.,-1366.103948,197.1801207,0.,1872.,-124.7076582,-819.7560616,-455.3679830,65.72670689,0.,764.2407998,-50.91168823,-380.9881888,312.,-20.78460970,-190.4940944,-81.97560616},
			{-80.49844713,92.95160030,0.,0.,197.1801207,-227.6839914,0.,-124.7076582,144.,0.,65.72670689,-75.89466382,0.,-50.91168823,58.78775383,0.,-20.78460970,24.,0.,0.},
			{-105.8300523,0.,0.,0.,518.4592560,0.,0.,-819.7560616,0.,400.,129.6148141,0.,0.,-267.7312084,0.,154.9193338,-81.97560616,0.,61.96773354,20.},
			{-416.4132569,254.5584412,-65.72670689,0.,610.000000,-207.8460970,26.83281572,-455.3679830,65.72670689,129.6148141,1220.000000,-415.6921939,53.66563145,-743.6128021,107.3312629,200.7984063,-1366.103948,197.1801207,301.1976094,518.4592560},
			{254.5584411,-489.8979486,227.6839914,0.,-207.8460970,320.,-92.95160030,65.72670689,-75.89466382,0.,-415.6921939,640.,-185.9032006,107.3312629,-123.9354671,0.,197.1801207,-227.6839914,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,26.83281572,-92.95160030,72.,0.,0.,0.,53.66563145,-185.9032006,144.,0.,0.,0.,0.,0.,0.,0.},
			{278.2804344,-65.72670689,0.,0.,-743.6128021,107.3312629,0.,764.2407998,-50.91168823,-267.7312084,-743.6128021,107.3312629,0.,832.,-55.42562586,-311.0755536,764.2407998,-50.91168823,-311.0755536,-267.7312084},
			{-65.72670689,75.89466382,0.,0.,107.3312629,-123.9354671,0.,-50.91168823,58.78775383,0.,107.3312629,-123.9354671,0.,-55.42562586,64.,0.,-50.91168823,58.78775383,0.,0.},
			{-81.9756061,0.,0.,0.,301.1976094,0.,0.,-380.9881888,0.,154.9193338,200.7984063,0.,0.,-311.0755536,0.,144.,-190.4940944,0.,108.,61.96773354},
			{340.8225342,-80.49844717,0.,0.,-455.3679830,65.72670689,0.,312.,-20.78460970,-81.97560616,-1366.103948,197.1801207,0.,764.2407998,-50.91168823,-190.4940944,1872.,-124.7076582,-380.9881888,-819.7560616},
			{-80.49844713,92.95160030,0.,0.,65.72670689,-75.89466382,0.,-20.78460970,24.,0.,197.1801207,-227.6839914,0.,-50.91168823,58.78775383,0.,-124.7076582,144.,0.,0.},
			{-81.9756061,0.,0.,0.,200.7984063,0.,0.,-190.4940944,0.,61.96773354,301.1976094,0.,0.,-311.0755536,0.,108.,-380.9881888,0.,144.,154.9193338},
			{-105.8300523,0.,0.,0.,129.6148141,0.,0.,-81.97560616,0.,20.,518.4592560,0.,0.,-267.7312084,0.,61.96773354,-819.7560616,0.,154.9193338,400.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{3.162277659,2.738612787,2.357022604,2.091650066,3.872983346,3.354101966,2.886751347,3.674234614,3.181980514,3.346640106,1.936491673,1.677050983,1.443375673,3.,2.598076212,3.240370349,1.224744872,1.060660172,2.160246899,.8366600264},
			{2.738612786,3.162277659,3.061862179,2.898275350,3.354101966,3.872983346,3.750000000,3.181980514,3.674234614,2.898275350,1.677050983,1.936491673,1.875000000,2.598076212,3.,2.806243040,1.060660172,1.224744872,1.870828694,.7245688375},
			{3.415650256,2.958039892,2.545875386,2.259240285,5.019960159,4.347413025,3.741657387,5.291502622,4.582575695,5.163977793,1.673320053,1.449137675,1.247219129,3.240370349,2.806243040,4.,.8819171036,.7637626160,2.,.5163977793},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.581138830,1.369306394,1.178511302,1.,.8660254040,.6831300514,3.162277660,2.738612788,2.357022604,2.449489743,2.121320343,1.763834207,3.,2.598076212,2.645751311,2.732520206},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.369306394,1.581138830,1.530931089,.8660254040,1.,.5916079782,2.738612788,3.162277660,3.061862179,2.121320343,2.449489743,1.527525232,2.598076212,3.,2.291287848,2.366431913},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.296148140,1.122497216,.9660917834,.6831300514,.5916079782,.4000000000,3.888444420,3.367491648,2.898275350,2.509980079,2.173706512,1.549193338,4.098780308,3.549647869,3.098386677,4.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{3.162277659,2.738612787,2.357022604,2.091650066,3.872983346,3.354101966,2.886751347,3.674234614,3.181980514,3.346640106,1.936491673,1.677050983,1.443375673,3.,2.598076212,3.240370349,1.224744872,1.060660172,2.160246899,.8366600264},
			{2.738612786,3.162277659,3.061862179,2.898275350,3.354101966,3.872983346,3.750000000,3.181980514,3.674234614,2.898275350,1.677050983,1.936491673,1.875000000,2.598076212,3.,2.806243040,1.060660172,1.224744872,1.870828694,.7245688375},
			{3.415650256,2.958039892,2.545875386,2.259240285,5.019960159,4.347413025,3.741657387,5.291502622,4.582575695,5.163977793,1.673320053,1.449137675,1.247219129,3.240370349,2.806243040,4.,.8819171036,.7637626160,2.,.5163977793},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.581138830,1.369306394,1.178511302,1.,.8660254040,.6831300514,3.162277660,2.738612788,2.357022604,2.449489743,2.121320343,1.763834207,3.,2.598076212,2.645751311,2.732520206},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.369306394,1.581138830,1.530931089,.8660254040,1.,.5916079782,2.738612788,3.162277660,3.061862179,2.121320343,2.449489743,1.527525232,2.598076212,3.,2.291287848,2.366431913},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.296148140,1.122497216,.9660917834,.6831300514,.5916079782,.4000000000,3.888444420,3.367491648,2.898275350,2.509980079,2.173706512,1.549193338,4.098780308,3.549647869,3.098386677,4.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{2.581988897,2.236067977,1.924500898,1.707825129,3.162277660,2.738612788,2.357022604,3.,2.598076212,2.732520206,1.581138830,1.369306394,1.178511302,2.449489743,2.121320343,2.645751311,1.,.8660254040,1.763834207,.6831300514},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.738612788,3.162277660,3.061862179,2.598076212,3.,2.366431913,1.369306394,1.581138830,1.530931089,2.121320343,2.449489743,2.291287848,.8660254040,1.,1.527525232,.5916079782},
			{2.645751311,2.291287848,1.972026594,1.750000000,3.888444420,3.367491648,2.898275350,4.098780308,3.549647869,4.,1.296148140,1.122497216,.9660917834,2.509980079,2.173706512,3.098386677,.6831300514,.5916079782,1.549193338,.4000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.936491673,1.677050983,1.443375673,1.224744872,1.060660172,.8366600264,3.872983346,3.354101966,2.886751347,3.,2.598076212,2.160246899,3.674234614,3.181980514,3.240370349,3.346640106},
			{2.738612786,3.162277659,3.061862179,2.898275350,1.677050983,1.936491673,1.875000000,1.060660172,1.224744872,.7245688375,3.354101966,3.872983346,3.750000000,2.598076212,3.,1.870828694,3.181980514,3.674234614,2.806243040,2.898275350},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,1.673320053,1.449137675,1.247219129,.8819171036,.7637626160,.5163977793,5.019960159,4.347413025,3.741657387,3.240370349,2.806243040,2.,5.291502622,4.582575695,4.,5.163977793},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{2.581988897,2.236067977,1.924500898,1.707825129,3.162277660,2.738612788,2.357022604,3.,2.598076212,2.732520206,1.581138830,1.369306394,1.178511302,2.449489743,2.121320343,2.645751311,1.,.8660254040,1.763834207,.6831300514},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.738612788,3.162277660,3.061862179,2.598076212,3.,2.366431913,1.369306394,1.581138830,1.530931089,2.121320343,2.449489743,2.291287848,.8660254040,1.,1.527525232,.5916079782},
			{2.645751311,2.291287848,1.972026594,1.750000000,3.888444420,3.367491648,2.898275350,4.098780308,3.549647869,4.,1.296148140,1.122497216,.9660917834,2.509980079,2.173706512,3.098386677,.6831300514,.5916079782,1.549193338,.4000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.936491673,1.677050983,1.443375673,1.224744872,1.060660172,.8366600264,3.872983346,3.354101966,2.886751347,3.,2.598076212,2.160246899,3.674234614,3.181980514,3.240370349,3.346640106},
			{2.738612786,3.162277659,3.061862179,2.898275350,1.677050983,1.936491673,1.875000000,1.060660172,1.224744872,.7245688375,3.354101966,3.872983346,3.750000000,2.598076212,3.,1.870828694,3.181980514,3.674234614,2.806243040,2.898275350},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,1.673320053,1.449137675,1.247219129,.8819171036,.7637626160,.5163977793,5.019960159,4.347413025,3.741657387,3.240370349,2.806243040,2.,5.291502622,4.582575695,4.,5.163977793},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,3.223582651,.3876663793,.3474891228e-1,.866848376,.1042467368,.1300185243,4.743042334,.5703958130,1.742587995,6.030295213},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,2.320915249,.2791120656,.2501852413e-1,2.966411012,.3567390516,2.114760201,2.458656996,.2956768161,4.293419991,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,1.143186367,.1374790002,.1232308492e-1,2.966411012,.3567390516,4.293419991,.5965041289,.7173527742e-1,2.114760201,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,.2405189635,.2892468600e-1,.2592696785e-2,.866848376,.1042467368,1.742587995,.2640450499e-1,.3175392072e-2,.1300185243,.2504783060e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,3.223582651,.3876663793,.3474891228e-1,.866848376,.1042467368,.1300185243,4.743042334,.5703958130,1.742587995,6.030295213},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,2.320915249,.2791120656,.2501852413e-1,2.966411012,.3567390516,2.114760201,2.458656996,.2956768161,4.293419991,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,1.143186367,.1374790002,.1232308492e-1,2.966411012,.3567390516,4.293419991,.5965041289,.7173527742e-1,2.114760201,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,.2405189635,.2892468600e-1,.2592696785e-2,.866848376,.1042467368,1.742587995,.2640450499e-1,.3175392072e-2,.1300185243,.2504783060e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,.2405189635,.2892468600e-1,.2592696785e-2,0.,0.,0.,.2640450499e-1,.3175392072e-2,0.,.2504783060e-2},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,1.143186367,.1374790002,.1232308492e-1,0.,0.,0.,.5965041289,.7173527742e-1,0.,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,2.320915249,.2791120656,.2501852413e-1,0.,0.,0.,2.458656996,.2956768161,0.,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,3.223582651,.3876663793,.3474891228e-1,0.,0.,0.,4.743042334,.5703958130,0.,6.030295213}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,.2405189635,.2892468600e-1,.2592696785e-2,0.,0.,0.,.2640450499e-1,.3175392072e-2,0.,.2504783060e-2},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,1.143186367,.1374790002,.1232308492e-1,0.,0.,0.,.5965041289,.7173527742e-1,0.,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,2.320915249,.2791120656,.2501852413e-1,0.,0.,0.,2.458656996,.2956768161,0.,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,3.223582651,.3876663793,.3474891228e-1,0.,0.,0.,4.743042334,.5703958130,0.,6.030295213}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[20],
		T o_output[4]
	) {
		static const T matrix[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 4; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.6050149664e-1,-.1134259259,-.1134259259,-.6050149664e-1},
			{-.7275878641e-2,-.1364054309e-1,-.1364054309e-1,-.7275878641e-2},
			{-.6521815710e-3,-.1222685433e-2,-.1222685433e-2,-.6521815710e-3},
			{-.5357858524e-4,-.1004471126e-3,-.1004471126e-3,-.5357858524e-4},
			{-.1028964625e-1,-.9168839524e-1,-.1861472469,-.1379081491},
			{-.1237427527e-2,-.110263988e-1,-.2238597139e-1,-.1658476255e-1},
			{-.1109182092e-3,-.9883636770e-3,-.2006591748e-2,-.1486593859e-2},
			{-.1129611617e-2,-.4784216114e-1,-.1971947192,-.2029121820},
			{-.1358465079e-3,-.5753473514e-2,-.2371453477e-1,-.2440211387e-1},
			{-.1071571705e-3,-.2157101540e-1,-.1805083621,-.2579821713},
			{-.1379081491,-.1861472469,-.9168839524e-1,-.1028964625e-1},
			{-.1658476255e-1,-.2238597139e-1,-.110263988e-1,-.1237427527e-2},
			{-.1486593859e-2,-.2006591748e-2,-.9883636770e-3,-.1109182092e-3},
			{-.3708465643e-1,-.2379187449,-.2379187449,-.3708465643e-1},
			{-.4459781576e-2,-.2861198511e-1,-.2861198511e-1,-.4459781576e-2},
			{-.5562324896e-2,-.1696127377,-.3443504935,-.7454968932e-1},
			{-.2029121820,-.1971947192,-.4784216114e-1,-.1129611617e-2},
			{-.2440211387e-1,-.2371453477e-1,-.5753473514e-2,-.1358465079e-3},
			{-.7454968932e-1,-.3443504935,-.1696127377,-.5562324896e-2},
			{-.2579821713,-.1805083621,-.2157101540e-1,-.1071571705e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.6050149664e-1,-.1134259259,-.1134259259,-.6050149664e-1},
			{-.7275878641e-2,-.1364054309e-1,-.1364054309e-1,-.7275878641e-2},
			{-.6521815710e-3,-.1222685433e-2,-.1222685433e-2,-.6521815710e-3},
			{-.5357858524e-4,-.1004471126e-3,-.1004471126e-3,-.5357858524e-4},
			{-.1028964625e-1,-.9168839524e-1,-.1861472469,-.1379081491},
			{-.1237427527e-2,-.110263988e-1,-.2238597139e-1,-.1658476255e-1},
			{-.1109182092e-3,-.9883636770e-3,-.2006591748e-2,-.1486593859e-2},
			{-.1129611617e-2,-.4784216114e-1,-.1971947192,-.2029121820},
			{-.1358465079e-3,-.5753473514e-2,-.2371453477e-1,-.2440211387e-1},
			{-.1071571705e-3,-.2157101540e-1,-.1805083621,-.2579821713},
			{-.1379081491,-.1861472469,-.9168839524e-1,-.1028964625e-1},
			{-.1658476255e-1,-.2238597139e-1,-.110263988e-1,-.1237427527e-2},
			{-.1486593859e-2,-.2006591748e-2,-.9883636770e-3,-.1109182092e-3},
			{-.3708465643e-1,-.2379187449,-.2379187449,-.3708465643e-1},
			{-.4459781576e-2,-.2861198511e-1,-.2861198511e-1,-.4459781576e-2},
			{-.5562324896e-2,-.1696127377,-.3443504935,-.7454968932e-1},
			{-.2029121820,-.1971947192,-.4784216114e-1,-.1129611617e-2},
			{-.2440211387e-1,-.2371453477e-1,-.5753473514e-2,-.1358465079e-3},
			{-.7454968932e-1,-.3443504935,-.1696127377,-.5562324896e-2},
			{-.2579821713,-.1805083621,-.2157101540e-1,-.1071571705e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7275878632e-2,-.6483348608e-1,-.1316259805,-.9751578751e-1},
			{-.8749933967e-3,-.779684142e-2,-.1582927222e-1,-.1172719805e-1},
			{-.7843101787e-4,-.6988786584e-3,-.1418874633e-2,-.1051180599e-2},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7987560345e-3,-.3382951655e-1,-.1394377231,-.1434805799},
			{-.9605798645e-4,-.4068320131e-2,-.1676870832e-1,-.1725490019e-1},
			{0.,0.,0.,0.},
			{-.7577156190e-4,-.1525301126e-1,-.1276386869,-.1824209426}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7275878632e-2,-.6483348608e-1,-.1316259805,-.9751578751e-1},
			{-.8749933967e-3,-.779684142e-2,-.1582927222e-1,-.1172719805e-1},
			{-.7843101787e-4,-.6988786584e-3,-.1418874633e-2,-.1051180599e-2},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7987560345e-3,-.3382951655e-1,-.1394377231,-.1434805799},
			{-.9605798645e-4,-.4068320131e-2,-.1676870832e-1,-.1725490019e-1},
			{0.,0.,0.,0.},
			{-.7577156190e-4,-.1525301126e-1,-.1276386869,-.1824209426}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{-.9751578751e-1,-.1316259805,-.6483348608e-1,-.7275878632e-2},
			{-.1172719805e-1,-.1582927222e-1,-.779684142e-2,-.8749933967e-3},
			{-.1051180599e-2,-.1418874633e-2,-.6988786584e-3,-.7843101787e-4},
			{-.1434805799,-.1394377231,-.3382951655e-1,-.7987560345e-3},
			{-.1725490019e-1,-.1676870832e-1,-.4068320131e-2,-.9605798645e-4},
			{-.1824209426,-.1276386869,-.1525301126e-1,-.7577156190e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[4],
		T o_output[20]
	) {
		static const T matrix[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{-.9751578751e-1,-.1316259805,-.6483348608e-1,-.7275878632e-2},
			{-.1172719805e-1,-.1582927222e-1,-.779684142e-2,-.8749933967e-3},
			{-.1051180599e-2,-.1418874633e-2,-.6988786584e-3,-.7843101787e-4},
			{-.1434805799,-.1394377231,-.3382951655e-1,-.7987560345e-3},
			{-.1725490019e-1,-.1676870832e-1,-.4068320131e-2,-.9605798645e-4},
			{-.1824209426,-.1276386869,-.1525301126e-1,-.7577156190e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 4; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.,2.645751311},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,-4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,2.645751311,.5000000000,0.,5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,-2.160246899,.4082482906,0.,0.,1.673320053},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,2.645751311,.5000000000,0.,-5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,1.936491673,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.,2.645751311},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,-4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,2.645751311,.5000000000,0.,5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,-2.160246899,.4082482906,0.,0.,1.673320053},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,2.645751311,.5000000000,0.,-5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,1.936491673,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,.5000000000,0.,0.,1.936491673,0.,4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,.5000000000,0.,0.,1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,-1.224744872,0.,-5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,1.936491673,0.,0.,-1.936491673,.5000000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,-.5000000000,0.,0.,-1.936491673,0.,-4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,1.673320053,0.,0.,0.,0.,0.,0.,-.4082482906,0.,-2.160246899,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,1.224744872,0.,5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,-.5000000000,0.,0.,.5000000000,.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,.5000000000,0.,0.,1.936491673,0.,4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,.5000000000,0.,0.,1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,-1.224744872,0.,-5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,1.936491673,0.,0.,-1.936491673,.5000000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,-.5000000000,0.,0.,-1.936491673,0.,-4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,1.673320053,0.,0.,0.,0.,0.,0.,-.4082482906,0.,-2.160246899,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,1.224744872,0.,5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,-.5000000000,0.,0.,.5000000000,.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,.5000000000,0.,0.,0.,0.,-1.045825033,.7905694150,0.,1.045825033,.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,.5000000000,0.,0.,0.,0.,0.,.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.4841229182,.1250000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,.4183300132,0.,0.,0.,0.,0.,.5400617248,-.2041241452,0.,-.5400617248,-.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.1250000000,.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,.5000000000,0.,0.,0.,0.,-1.045825033,.7905694150,0.,1.045825033,.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,.5000000000,0.,0.,0.,0.,0.,.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.4841229182,.1250000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,.4183300132,0.,0.,0.,0.,0.,.5400617248,-.2041241452,0.,-.5400617248,-.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.1250000000,.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,.8100925875,-.5000000000,0.,0.,0.,0.,1.045825033,-.7905694150,0.,-1.045825033,-.8100925875},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,-.4183300132,0.,0.,0.,0.,0.,-.5400617248,.2041241452,0.,.5400617248,.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.4841229182,-.1250000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,.8100925875,-.5000000000,0.,0.,0.,0.,1.045825033,-.7905694150,0.,-1.045825033,-.8100925875},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,-.4183300132,0.,0.,0.,0.,0.,-.5400617248,.2041241452,0.,.5400617248,.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.4841229182,-.1250000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419},
			{0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.},
			{0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419},
			{0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.},
			{0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.},
			{0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.},
			{0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.},
			{0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.},
			{0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[20],
		T o_output[1]
	) {
		static const T matrix[1][20] = 
		{
			{1.414213562,1.224744872,1.054092553,.9354143468,1.154700539,1.,.8606629657,.9128709293,.7905694150,.7483314774,1.154700539,1.,.8606629657,1.118033988,.9682458365,.9660917834,.9128709293,.7905694150,.9660917834,.7483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[20],
		T o_output[1]
	) {
		static const T matrix[1][20] = 
		{
			{1.414213562,1.224744872,1.054092553,.9354143468,1.154700539,1.,.8606629657,.9128709293,.7905694150,.7483314774,1.154700539,1.,.8606629657,1.118033988,.9682458365,.9660917834,.9128709293,.7905694150,.9660917834,.7483314774}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.414213562,1.224744872,1.054092553,1.118033988,.9682458365,.9165151392,1.414213562,1.224744872,1.054092553,1.369306394,1.185854122,1.183215957,1.118033988,.9682458365,1.183215957,.9165151392},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.825741858,2.108185107,2.041241452,1.443375673,1.666666667,1.183215956,1.825741858,2.108185107,2.041241452,1.767766952,2.041241452,1.527525232,1.443375673,1.666666667,1.527525232,1.183215956},
			{2.645751311,3.436931772,3.549647869,3.500000000,2.160246900,2.806243040,2.898275350,1.707825129,2.218529918,1.400000000,2.160246900,2.806243040,2.898275350,2.091650066,2.717133141,1.807392228,1.707825129,2.218529918,1.807392228,1.400000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,1.732050808,1.500000000,1.290994449,1.643167672,1.423024947,1.496662955,.8660254040,.7500000000,.6454972245,1.341640786,1.161895004,1.449137675,.5477225575,.4743416490,.9660917834,.3741657387},
			{1.825741857,2.108185106,2.041241452,1.932183566,2.236067977,2.581988897,2.500000000,2.121320343,2.449489743,1.932183566,1.118033988,1.290994449,1.250000000,1.732050808,2.,1.870828694,.7071067810,.8164965809,1.247219129,.4830458916},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,1.643167673,1.423024947,1.224744872,1.732050808,1.500000000,1.690308509,.5477225576,.4743416490,.4082482906,1.060660172,.9185586536,1.309307341,.2886751347,.2500000000,.6546536709,.1690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8660254040,.7500000000,.6454972245,.5477225575,.4743416490,.3741657387,1.732050808,1.500000000,1.290994449,1.341640786,1.161895004,.9660917834,1.643167672,1.423024947,1.449137675,1.496662955},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.118033988,1.290994449,1.250000000,.7071067810,.8164965809,.4830458916,2.236067977,2.581988897,2.500000000,1.732050808,2.,1.247219129,2.121320343,2.449489743,1.870828694,1.932183566},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,1.341640786,1.161895004,1.,1.060660172,.9185586536,.8280786714,1.341640786,1.161895004,1.,1.732050808,1.500000000,1.603567452,1.060660172,.9185586536,1.603567452,.8280786714},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.5477225576,.4743416490,.4082482906,.2886751347,.2500000000,.1690308509,1.643167673,1.423024947,1.224744872,1.060660172,.9185586536,.6546536709,1.732050808,1.500000000,1.309307341,1.690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[20],
		T o_output[20]
	) {
		static const T matrix[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.414213562,1.224744872,1.054092553,1.118033988,.9682458365,.9165151392,1.414213562,1.224744872,1.054092553,1.369306394,1.185854122,1.183215957,1.118033988,.9682458365,1.183215957,.9165151392},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.825741858,2.108185107,2.041241452,1.443375673,1.666666667,1.183215956,1.825741858,2.108185107,2.041241452,1.767766952,2.041241452,1.527525232,1.443375673,1.666666667,1.527525232,1.183215956},
			{2.645751311,3.436931772,3.549647869,3.500000000,2.160246900,2.806243040,2.898275350,1.707825129,2.218529918,1.400000000,2.160246900,2.806243040,2.898275350,2.091650066,2.717133141,1.807392228,1.707825129,2.218529918,1.807392228,1.400000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,1.732050808,1.500000000,1.290994449,1.643167672,1.423024947,1.496662955,.8660254040,.7500000000,.6454972245,1.341640786,1.161895004,1.449137675,.5477225575,.4743416490,.9660917834,.3741657387},
			{1.825741857,2.108185106,2.041241452,1.932183566,2.236067977,2.581988897,2.500000000,2.121320343,2.449489743,1.932183566,1.118033988,1.290994449,1.250000000,1.732050808,2.,1.870828694,.7071067810,.8164965809,1.247219129,.4830458916},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,1.643167673,1.423024947,1.224744872,1.732050808,1.500000000,1.690308509,.5477225576,.4743416490,.4082482906,1.060660172,.9185586536,1.309307341,.2886751347,.2500000000,.6546536709,.1690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8660254040,.7500000000,.6454972245,.5477225575,.4743416490,.3741657387,1.732050808,1.500000000,1.290994449,1.341640786,1.161895004,.9660917834,1.643167672,1.423024947,1.449137675,1.496662955},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.118033988,1.290994449,1.250000000,.7071067810,.8164965809,.4830458916,2.236067977,2.581988897,2.500000000,1.732050808,2.,1.247219129,2.121320343,2.449489743,1.870828694,1.932183566},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,1.341640786,1.161895004,1.,1.060660172,.9185586536,.8280786714,1.341640786,1.161895004,1.,1.732050808,1.500000000,1.603567452,1.060660172,.9185586536,1.603567452,.8280786714},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.5477225576,.4743416490,.4082482906,.2886751347,.2500000000,.1690308509,1.643167673,1.423024947,1.224744872,1.060660172,.9185586536,.6546536709,1.732050808,1.500000000,1.309307341,1.690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[20],
		T o_output[13]
	) {
		static const T matrix[13][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,1.154700539,2.,2.581988897,.6085806194,1.054092553,.2771598064,1.154700539,2.,2.581988897,1.490711985,2.581988897,1.073435315,.6085806194,1.054092553,1.073435315,.2771598064},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,1.660371852,2.875848407,3.712704328,1.674178825,2.899762786,.9415775350,1.258317121,2.179469186,1.733484547,.8240209079},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,.9018648819,1.562075797,2.016631182,.9093644209,1.575065380,.5114370684,.3712464703,.6430177487,.5114370684,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.660371852,2.875848407,3.712704328,1.258317121,2.179469186,.8240209079,.9018648819,1.562075797,2.016631182,1.674178825,2.899762786,1.733484547,.3712464703,.6430177487,.9415775350,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,3.012867026,5.218438765,6.736975475,.7599890903,1.316339718,.1069282633,4.143232550,7.176289285,1.427907550,4.923365040},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,.2256172947,.3907806174,.5044956077,.5691146707e-1,.9857355252e-1,.8007278541e-2,.2323400915e-1,.4024248430e-1,.8007278541e-2,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,3.012867026,5.218438765,6.736975475,4.143232550,7.176289285,4.923365040,.2256172947,.3907806174,.5044956077,.7599890903,1.316339718,1.427907550,.2323400915e-1,.4024248430e-1,.1069282633,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,.1686682004,.2921418926,.3771535615,.2043791133,.3539950084,.1381330405,.1298511510e-1,.2249087909e-1,.2149722933e-1,.8638170614e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,1.083797870,1.877192976,2.423445711,2.679889337,4.641704490,3.696095549,.5361372318,.9286169253,1.811248012,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,2.211635545,3.830665132,4.945367419,.4170631113,.7223744989,.4386799219e-1,2.232577295,3.866937307,.5752122252,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,1.083797870,1.877192976,2.423445711,.2043791133,.3539950084,.2149722933e-1,.5361372318,.9286169253,.1381330405,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,2.211635545,3.830665132,4.945367419,2.679889337,4.641704490,1.811248012,2.232577295,3.866937307,3.696095549,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,.1686682004,.2921418926,.3771535615,.4170631113,.7223744989,.5752122252,.1298511510e-1,.2249087909e-1,.4386799219e-1,.8638170614e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 13; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[20],
		T o_output[13]
	) {
		static const T matrix[13][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,1.154700539,2.,2.581988897,.6085806194,1.054092553,.2771598064,1.154700539,2.,2.581988897,1.490711985,2.581988897,1.073435315,.6085806194,1.054092553,1.073435315,.2771598064},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,1.660371852,2.875848407,3.712704328,1.674178825,2.899762786,.9415775350,1.258317121,2.179469186,1.733484547,.8240209079},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,.9018648819,1.562075797,2.016631182,.9093644209,1.575065380,.5114370684,.3712464703,.6430177487,.5114370684,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.660371852,2.875848407,3.712704328,1.258317121,2.179469186,.8240209079,.9018648819,1.562075797,2.016631182,1.674178825,2.899762786,1.733484547,.3712464703,.6430177487,.9415775350,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,3.012867026,5.218438765,6.736975475,.7599890903,1.316339718,.1069282633,4.143232550,7.176289285,1.427907550,4.923365040},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,.2256172947,.3907806174,.5044956077,.5691146707e-1,.9857355252e-1,.8007278541e-2,.2323400915e-1,.4024248430e-1,.8007278541e-2,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,3.012867026,5.218438765,6.736975475,4.143232550,7.176289285,4.923365040,.2256172947,.3907806174,.5044956077,.7599890903,1.316339718,1.427907550,.2323400915e-1,.4024248430e-1,.1069282633,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,.1686682004,.2921418926,.3771535615,.2043791133,.3539950084,.1381330405,.1298511510e-1,.2249087909e-1,.2149722933e-1,.8638170614e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,1.083797870,1.877192976,2.423445711,2.679889337,4.641704490,3.696095549,.5361372318,.9286169253,1.811248012,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,2.211635545,3.830665132,4.945367419,.4170631113,.7223744989,.4386799219e-1,2.232577295,3.866937307,.5752122252,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,1.083797870,1.877192976,2.423445711,.2043791133,.3539950084,.2149722933e-1,.5361372318,.9286169253,.1381330405,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,2.211635545,3.830665132,4.945367419,2.679889337,4.641704490,1.811248012,2.232577295,3.866937307,3.696095549,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,.1686682004,.2921418926,.3771535615,.4170631113,.7223744989,.5752122252,.1298511510e-1,.2249087909e-1,.4386799219e-1,.8638170614e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 13; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[20],
		T o_output[13]
	) {
		static const T matrix[13][20] = 
		{
			{1.414213562,0.,0.,0.,1.154700539,0.,0.,.6085806194,0.,.2771598064,1.154700539,0.,0.,1.490711985,0.,1.073435315,.6085806194,0.,1.073435315,.2771598064},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,1.660371852,0.,0.,1.674178825,0.,.9415775350,1.258317121,0.,1.733484547,.8240209079},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,.9018648819,0.,0.,.9093644209,0.,.5114370684,.3712464703,0.,.5114370684,.1320524832},
			{1.414213562,0.,0.,0.,1.660371852,0.,0.,1.258317121,0.,.8240209079,.9018648819,0.,0.,1.674178825,0.,1.733484547,.3712464703,0.,.9415775350,.1320524832},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,3.012867026,0.,0.,.7599890903,0.,.1069282633,4.143232550,0.,1.427907550,4.923365040},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,.2256172947,0.,0.,.5691146707e-1,0.,.8007278541e-2,.2323400915e-1,0.,.8007278541e-2,.2067470429e-2},
			{1.414213562,0.,0.,0.,3.012867026,0.,0.,4.143232550,0.,4.923365040,.2256172947,0.,0.,.7599890903,0.,1.427907550,.2323400915e-1,0.,.1069282633,.2067470429e-2},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,.1686682004,0.,0.,.2043791133,0.,.1381330405,.1298511510e-1,0.,.2149722933e-1,.8638170614e-3},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,1.083797870,0.,0.,2.679889337,0.,3.696095549,.5361372318,0.,1.811248012,.2291748860},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,2.211635545,0.,0.,.4170631113,0.,.4386799219e-1,2.232577295,0.,.5752122252,1.947434359},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,1.083797870,0.,0.,.2043791133,0.,.2149722933e-1,.5361372318,0.,.1381330405,.2291748860},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,2.211635545,0.,0.,2.679889337,0.,1.811248012,2.232577295,0.,3.696095549,1.947434359},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,.1686682004,0.,0.,.4170631113,0.,.5752122252,.1298511510e-1,0.,.4386799219e-1,.8638170614e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 13; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[20],
		T o_output[13]
	) {
		static const T matrix[13][20] = 
		{
			{1.414213562,0.,0.,0.,1.154700539,0.,0.,.6085806194,0.,.2771598064,1.154700539,0.,0.,1.490711985,0.,1.073435315,.6085806194,0.,1.073435315,.2771598064},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,1.660371852,0.,0.,1.674178825,0.,.9415775350,1.258317121,0.,1.733484547,.8240209079},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,.9018648819,0.,0.,.9093644209,0.,.5114370684,.3712464703,0.,.5114370684,.1320524832},
			{1.414213562,0.,0.,0.,1.660371852,0.,0.,1.258317121,0.,.8240209079,.9018648819,0.,0.,1.674178825,0.,1.733484547,.3712464703,0.,.9415775350,.1320524832},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,3.012867026,0.,0.,.7599890903,0.,.1069282633,4.143232550,0.,1.427907550,4.923365040},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,.2256172947,0.,0.,.5691146707e-1,0.,.8007278541e-2,.2323400915e-1,0.,.8007278541e-2,.2067470429e-2},
			{1.414213562,0.,0.,0.,3.012867026,0.,0.,4.143232550,0.,4.923365040,.2256172947,0.,0.,.7599890903,0.,1.427907550,.2323400915e-1,0.,.1069282633,.2067470429e-2},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,.1686682004,0.,0.,.2043791133,0.,.1381330405,.1298511510e-1,0.,.2149722933e-1,.8638170614e-3},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,1.083797870,0.,0.,2.679889337,0.,3.696095549,.5361372318,0.,1.811248012,.2291748860},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,2.211635545,0.,0.,.4170631113,0.,.4386799219e-1,2.232577295,0.,.5752122252,1.947434359},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,1.083797870,0.,0.,.2043791133,0.,.2149722933e-1,.5361372318,0.,.1381330405,.2291748860},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,2.211635545,0.,0.,2.679889337,0.,1.811248012,2.232577295,0.,3.696095549,1.947434359},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,.1686682004,0.,0.,.4170631113,0.,.5752122252,.1298511510e-1,0.,.4386799219e-1,.8638170614e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 13; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 20; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[13],
		T o_output[20]
	) {
		static const T matrix[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{.3663702899,-.4301677717,-.4301677717,-.4301677717,-.1306735064,-.1306735064,-.1306735064,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663},
			{.4729820103,-.5553442052,-.5553442052,-.5553442052,-.1686987714,-.1686987714,-.1686987714,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233},
			{.5596398619,-.6570921251,-.6570921251,-.6570921251,-.1996070782,-.1996070782,-.1996070782,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{.2991400890,-.2743243432,-.2743243432,-.5050428582,-.2084706567e-1,-.2084706567e-1,-.2783892823,-.1447574103,-.2953969950,-.2252816005e-1,-.2252816005e-1,-.1447574103,-.2953969950},
			{.3861881943,-.3541512041,-.3541512041,-.6520075263,-.2691344605e-1,-.2691344605e-1,-.3593990180,-.1868810131,-.3813558807,-.2908372956e-1,-.2908372956e-1,-.1868810131,-.3813558807},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{.1576606701,-.1129237275,-.1129237275,-.3827480421,-.2146825292e-2,-.2146825292e-2,-.3828351954,-.7160914354e-1,-.2981940789,-.1734356272e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{.2991400890,-.5050428582,-.2743243432,-.2743243432,-.2783892823,-.2084706567e-1,-.2084706567e-1,-.2252816005e-1,-.1447574103,-.2953969950,-.1447574103,-.2953969950,-.2252816005e-1},
			{.3861881943,-.6520075263,-.3541512041,-.3541512041,-.3593990180,-.2691344605e-1,-.2691344605e-1,-.2908372956e-1,-.1868810131,-.3813558807,-.1868810131,-.3813558807,-.2908372956e-1},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{.3861881943,-.5092425882,-.2766055121,-.5092425882,-.7022308507e-1,-.5258626531e-2,-.7022308507e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{.1576606701,-.3827480421,-.1129237275,-.1129237275,-.3828351954,-.2146825292e-2,-.2146825292e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789,-.7160914354e-1,-.2981940789,-.1734356272e-2},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 13; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[13],
		T o_output[20]
	) {
		static const T matrix[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{.3663702899,-.4301677717,-.4301677717,-.4301677717,-.1306735064,-.1306735064,-.1306735064,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663},
			{.4729820103,-.5553442052,-.5553442052,-.5553442052,-.1686987714,-.1686987714,-.1686987714,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233},
			{.5596398619,-.6570921251,-.6570921251,-.6570921251,-.1996070782,-.1996070782,-.1996070782,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{.2991400890,-.2743243432,-.2743243432,-.5050428582,-.2084706567e-1,-.2084706567e-1,-.2783892823,-.1447574103,-.2953969950,-.2252816005e-1,-.2252816005e-1,-.1447574103,-.2953969950},
			{.3861881943,-.3541512041,-.3541512041,-.6520075263,-.2691344605e-1,-.2691344605e-1,-.3593990180,-.1868810131,-.3813558807,-.2908372956e-1,-.2908372956e-1,-.1868810131,-.3813558807},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{.1576606701,-.1129237275,-.1129237275,-.3827480421,-.2146825292e-2,-.2146825292e-2,-.3828351954,-.7160914354e-1,-.2981940789,-.1734356272e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{.2991400890,-.5050428582,-.2743243432,-.2743243432,-.2783892823,-.2084706567e-1,-.2084706567e-1,-.2252816005e-1,-.1447574103,-.2953969950,-.1447574103,-.2953969950,-.2252816005e-1},
			{.3861881943,-.6520075263,-.3541512041,-.3541512041,-.3593990180,-.2691344605e-1,-.2691344605e-1,-.2908372956e-1,-.1868810131,-.3813558807,-.1868810131,-.3813558807,-.2908372956e-1},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{.3861881943,-.5092425882,-.2766055121,-.5092425882,-.7022308507e-1,-.5258626531e-2,-.7022308507e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{.1576606701,-.3827480421,-.1129237275,-.1129237275,-.3828351954,-.2146825292e-2,-.2146825292e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789,-.7160914354e-1,-.2981940789,-.1734356272e-2},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 13; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[13],
		T o_output[20]
	) {
		static const T matrix[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 13; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[13],
		T o_output[20]
	) {
		static const T matrix[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 20; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 13; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 15600;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[20][20] = 
		{
			{260.000002,-329.0896534,214.6625259,-52.91502622,-416.4132569,254.5584411,-65.72670691,340.8225342,-80.49844713,-105.8300523,-416.4132569,254.5584411,-65.72670691,278.2804344,-65.72670689,-81.9756061,340.8225342,-80.49844713,-81.9756061,-105.8300523},
			{-329.0896534,1020.,-1068.943403,366.6060557,254.5584412,-489.8979486,227.6839914,-80.49844717,92.95160030,0.,254.5584412,-489.8979486,227.6839914,-65.72670689,75.89466382,0.,-80.49844717,92.95160030,0.,0.},
			{214.6625259,-1068.943403,1584.,-709.9295738,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,-65.72670689,227.6839914,-176.3632615,0.,0.,0.,0.,0.,0.,0.},
			{-52.91502622,366.6060557,-709.9295738,400.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-416.4132569,254.5584412,-65.72670689,0.,1220.000000,-415.6921939,53.66563145,-1366.103948,197.1801207,518.4592560,610.000000,-207.8460970,26.83281572,-743.6128021,107.3312629,301.1976094,-455.3679830,65.72670689,200.7984063,129.6148141},
			{254.5584411,-489.8979486,227.6839914,0.,-415.6921939,640.,-185.9032006,197.1801207,-227.6839914,0.,-207.8460970,320.,-92.95160030,107.3312629,-123.9354671,0.,65.72670689,-75.89466382,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,53.66563145,-185.9032006,144.,0.,0.,0.,26.83281572,-92.95160030,72.,0.,0.,0.,0.,0.,0.,0.},
			{340.8225342,-80.49844717,0.,0.,-1366.103948,197.1801207,0.,1872.,-124.7076582,-819.7560616,-455.3679830,65.72670689,0.,764.2407998,-50.91168823,-380.9881888,312.,-20.78460970,-190.4940944,-81.97560616},
			{-80.49844713,92.95160030,0.,0.,197.1801207,-227.6839914,0.,-124.7076582,144.,0.,65.72670689,-75.89466382,0.,-50.91168823,58.78775383,0.,-20.78460970,24.,0.,0.},
			{-105.8300523,0.,0.,0.,518.4592560,0.,0.,-819.7560616,0.,400.,129.6148141,0.,0.,-267.7312084,0.,154.9193338,-81.97560616,0.,61.96773354,20.},
			{-416.4132569,254.5584412,-65.72670689,0.,610.000000,-207.8460970,26.83281572,-455.3679830,65.72670689,129.6148141,1220.000000,-415.6921939,53.66563145,-743.6128021,107.3312629,200.7984063,-1366.103948,197.1801207,301.1976094,518.4592560},
			{254.5584411,-489.8979486,227.6839914,0.,-207.8460970,320.,-92.95160030,65.72670689,-75.89466382,0.,-415.6921939,640.,-185.9032006,107.3312629,-123.9354671,0.,197.1801207,-227.6839914,0.,0.},
			{-65.72670691,227.6839914,-176.3632615,0.,26.83281572,-92.95160030,72.,0.,0.,0.,53.66563145,-185.9032006,144.,0.,0.,0.,0.,0.,0.,0.},
			{278.2804344,-65.72670689,0.,0.,-743.6128021,107.3312629,0.,764.2407998,-50.91168823,-267.7312084,-743.6128021,107.3312629,0.,832.,-55.42562586,-311.0755536,764.2407998,-50.91168823,-311.0755536,-267.7312084},
			{-65.72670689,75.89466382,0.,0.,107.3312629,-123.9354671,0.,-50.91168823,58.78775383,0.,107.3312629,-123.9354671,0.,-55.42562586,64.,0.,-50.91168823,58.78775383,0.,0.},
			{-81.9756061,0.,0.,0.,301.1976094,0.,0.,-380.9881888,0.,154.9193338,200.7984063,0.,0.,-311.0755536,0.,144.,-190.4940944,0.,108.,61.96773354},
			{340.8225342,-80.49844717,0.,0.,-455.3679830,65.72670689,0.,312.,-20.78460970,-81.97560616,-1366.103948,197.1801207,0.,764.2407998,-50.91168823,-190.4940944,1872.,-124.7076582,-380.9881888,-819.7560616},
			{-80.49844713,92.95160030,0.,0.,65.72670689,-75.89466382,0.,-20.78460970,24.,0.,197.1801207,-227.6839914,0.,-50.91168823,58.78775383,0.,-124.7076582,144.,0.,0.},
			{-81.9756061,0.,0.,0.,200.7984063,0.,0.,-190.4940944,0.,61.96773354,301.1976094,0.,0.,-311.0755536,0.,108.,-380.9881888,0.,144.,154.9193338},
			{-105.8300523,0.,0.,0.,129.6148141,0.,0.,-81.97560616,0.,20.,518.4592560,0.,0.,-267.7312084,0.,61.96773354,-819.7560616,0.,154.9193338,400.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{3.162277659,2.738612787,2.357022604,2.091650066,3.872983346,3.354101966,2.886751347,3.674234614,3.181980514,3.346640106,1.936491673,1.677050983,1.443375673,3.,2.598076212,3.240370349,1.224744872,1.060660172,2.160246899,.8366600264},
			{2.738612786,3.162277659,3.061862179,2.898275350,3.354101966,3.872983346,3.750000000,3.181980514,3.674234614,2.898275350,1.677050983,1.936491673,1.875000000,2.598076212,3.,2.806243040,1.060660172,1.224744872,1.870828694,.7245688375},
			{3.415650256,2.958039892,2.545875386,2.259240285,5.019960159,4.347413025,3.741657387,5.291502622,4.582575695,5.163977793,1.673320053,1.449137675,1.247219129,3.240370349,2.806243040,4.,.8819171036,.7637626160,2.,.5163977793},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.581138830,1.369306394,1.178511302,1.,.8660254040,.6831300514,3.162277660,2.738612788,2.357022604,2.449489743,2.121320343,1.763834207,3.,2.598076212,2.645751311,2.732520206},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.369306394,1.581138830,1.530931089,.8660254040,1.,.5916079782,2.738612788,3.162277660,3.061862179,2.121320343,2.449489743,1.527525232,2.598076212,3.,2.291287848,2.366431913},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.296148140,1.122497216,.9660917834,.6831300514,.5916079782,.4000000000,3.888444420,3.367491648,2.898275350,2.509980079,2.173706512,1.549193338,4.098780308,3.549647869,3.098386677,4.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,2.,1.732050808,1.490711985,1.581138830,1.369306394,1.296148140,2.,1.732050808,1.490711985,1.936491673,1.677050983,1.673320053,1.581138830,1.369306394,1.673320053,1.296148140},
			{2.121320343,2.449489743,2.371708245,2.244994432,1.732050808,2.,1.936491673,1.369306394,1.581138830,1.122497216,1.732050808,2.,1.936491673,1.677050983,1.936491673,1.449137675,1.369306394,1.581138830,1.449137675,1.122497216},
			{1.825741857,2.371708244,2.449489743,2.415229459,1.490711985,1.936491673,2.,1.178511302,1.530931089,.9660917834,1.490711985,1.936491673,2.,1.443375673,1.875000000,1.247219129,1.178511302,1.530931089,1.247219129,.9660917834},
			{2.581988897,2.236067977,1.924500898,1.707825129,3.162277660,2.738612788,2.357022604,3.,2.598076212,2.732520206,1.581138830,1.369306394,1.178511302,2.449489743,2.121320343,2.645751311,1.,.8660254040,1.763834207,.6831300514},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.738612788,3.162277660,3.061862179,2.598076212,3.,2.366431913,1.369306394,1.581138830,1.530931089,2.121320343,2.449489743,2.291287848,.8660254040,1.,1.527525232,.5916079782},
			{2.645751311,2.291287848,1.972026594,1.750000000,3.888444420,3.367491648,2.898275350,4.098780308,3.549647869,4.,1.296148140,1.122497216,.9660917834,2.509980079,2.173706512,3.098386677,.6831300514,.5916079782,1.549193338,.4000000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.936491673,1.677050983,1.443375673,1.224744872,1.060660172,.8366600264,3.872983346,3.354101966,2.886751347,3.,2.598076212,2.160246899,3.674234614,3.181980514,3.240370349,3.346640106},
			{2.738612786,3.162277659,3.061862179,2.898275350,1.677050983,1.936491673,1.875000000,1.060660172,1.224744872,.7245688375,3.354101966,3.872983346,3.750000000,2.598076212,3.,1.870828694,3.181980514,3.674234614,2.806243040,2.898275350},
			{3.415650256,2.958039892,2.545875386,2.259240285,3.346640106,2.898275350,2.494438258,2.645751311,2.291287848,2.065591118,3.346640106,2.898275350,2.494438258,4.320493799,3.741657387,4.,2.645751311,2.291287848,4.,2.065591118},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,1.673320053,1.449137675,1.247219129,.8819171036,.7637626160,.5163977793,5.019960159,4.347413025,3.741657387,3.240370349,2.806243040,2.,5.291502622,4.582575695,4.,5.163977793},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,3.223582651,.3876663793,.3474891228e-1,.866848376,.1042467368,.1300185243,4.743042334,.5703958130,1.742587995,6.030295213},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,2.320915249,.2791120656,.2501852413e-1,2.966411012,.3567390516,2.114760201,2.458656996,.2956768161,4.293419991,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,1.143186367,.1374790002,.1232308492e-1,2.966411012,.3567390516,4.293419991,.5965041289,.7173527742e-1,2.114760201,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,.2405189635,.2892468600e-1,.2592696785e-2,.866848376,.1042467368,1.742587995,.2640450499e-1,.3175392072e-2,.1300185243,.2504783060e-2}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,.2405189635,.2892468600e-1,.2592696785e-2,0.,0.,0.,.2640450499e-1,.3175392072e-2,0.,.2504783060e-2},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,1.143186367,.1374790002,.1232308492e-1,0.,0.,0.,.5965041289,.7173527742e-1,0.,.2689510557},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,2.320915249,.2791120656,.2501852413e-1,0.,0.,0.,2.458656996,.2956768161,0.,2.250608683},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,0.,0.,0.,0.,0.,0.,3.223582651,.3876663793,.3474891228e-1,0.,0.,0.,4.743042334,.5703958130,0.,6.030295213}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[4][20] = 
		{
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,3.223582651,.3876663793,.3474891228e-1,4.743042334,.5703958130,6.030295213,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,2.320915249,.2791120656,.2501852413e-1,2.458656996,.2956768161,2.250608683,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,1.143186367,.1374790002,.1232308492e-1,.5965041289,.7173527742e-1,.2689510557,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.1700725905,.1524464801e-1,.1252391530e-2,.2405189635,.2892468600e-1,.2592696785e-2,.2640450499e-1,.3175392072e-2,.2504783060e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 4; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[20][4] = 
		{
			{-.6050149664e-1,-.1134259259,-.1134259259,-.6050149664e-1},
			{-.7275878641e-2,-.1364054309e-1,-.1364054309e-1,-.7275878641e-2},
			{-.6521815710e-3,-.1222685433e-2,-.1222685433e-2,-.6521815710e-3},
			{-.5357858524e-4,-.1004471126e-3,-.1004471126e-3,-.5357858524e-4},
			{-.1028964625e-1,-.9168839524e-1,-.1861472469,-.1379081491},
			{-.1237427527e-2,-.110263988e-1,-.2238597139e-1,-.1658476255e-1},
			{-.1109182092e-3,-.9883636770e-3,-.2006591748e-2,-.1486593859e-2},
			{-.1129611617e-2,-.4784216114e-1,-.1971947192,-.2029121820},
			{-.1358465079e-3,-.5753473514e-2,-.2371453477e-1,-.2440211387e-1},
			{-.1071571705e-3,-.2157101540e-1,-.1805083621,-.2579821713},
			{-.1379081491,-.1861472469,-.9168839524e-1,-.1028964625e-1},
			{-.1658476255e-1,-.2238597139e-1,-.110263988e-1,-.1237427527e-2},
			{-.1486593859e-2,-.2006591748e-2,-.9883636770e-3,-.1109182092e-3},
			{-.3708465643e-1,-.2379187449,-.2379187449,-.3708465643e-1},
			{-.4459781576e-2,-.2861198511e-1,-.2861198511e-1,-.4459781576e-2},
			{-.5562324896e-2,-.1696127377,-.3443504935,-.7454968932e-1},
			{-.2029121820,-.1971947192,-.4784216114e-1,-.1129611617e-2},
			{-.2440211387e-1,-.2371453477e-1,-.5753473514e-2,-.1358465079e-3},
			{-.7454968932e-1,-.3443504935,-.1696127377,-.5562324896e-2},
			{-.2579821713,-.1805083621,-.2157101540e-1,-.1071571705e-3}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7275878632e-2,-.6483348608e-1,-.1316259805,-.9751578751e-1},
			{-.8749933967e-3,-.779684142e-2,-.1582927222e-1,-.1172719805e-1},
			{-.7843101787e-4,-.6988786584e-3,-.1418874633e-2,-.1051180599e-2},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{-.7987560345e-3,-.3382951655e-1,-.1394377231,-.1434805799},
			{-.9605798645e-4,-.4068320131e-2,-.1676870832e-1,-.1725490019e-1},
			{0.,0.,0.,0.},
			{-.7577156190e-4,-.1525301126e-1,-.1276386869,-.1824209426}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[20][4] = 
		{
			{-.4278101850e-1,-.8020424136e-1,-.8020424136e-1,-.4278101850e-1},
			{-.5144823126e-2,-.9645320522e-2,-.9645320522e-2,-.5144823126e-2},
			{-.4611620101e-3,-.8645691572e-3,-.8645691572e-3,-.4611620101e-3},
			{-.3788578095e-4,-.7102683444e-4,-.7102683444e-4,-.3788578095e-4},
			{-.9751578751e-1,-.1316259805,-.6483348608e-1,-.7275878632e-2},
			{-.1172719805e-1,-.1582927222e-1,-.779684142e-2,-.8749933967e-3},
			{-.1051180599e-2,-.1418874633e-2,-.6988786584e-3,-.7843101787e-4},
			{-.1434805799,-.1394377231,-.3382951655e-1,-.7987560345e-3},
			{-.1725490019e-1,-.1676870832e-1,-.4068320131e-2,-.9605798645e-4},
			{-.1824209426,-.1276386869,-.1525301126e-1,-.7577156190e-4},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.},
			{0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 4; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[20][20] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.,2.645751311},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.936491673,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,-4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,2.645751311,.5000000000,0.,5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,4.183300132,-3.240370350},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,1.936491673,0.,0.,-1.581138830,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,-2.160246899,.4082482906,0.,0.,1.673320053},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,2.645751311,.5000000000,0.,-5.291502622,2.049390154},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,1.936491673,-.5000000000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[20][20] = 
		{
			{.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,.5000000000,0.,0.,1.936491673,0.,4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,.5000000000,0.,0.,1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,-1.224744872,0.,-5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,1.936491673,0.,0.,-1.936491673,.5000000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,-3.240370350,-.5000000000,0.,0.,-1.936491673,0.,-4.183300132,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-1.581138830,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,1.673320053,0.,0.,0.,0.,0.,0.,-.4082482906,0.,-2.160246899,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,.5000000000,0.,0.,.5000000000,-.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.049390154,0.,0.,0.,1.224744872,0.,5.291502622,.5000000000,0.,2.645751311,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,-.5000000000,0.,0.,.5000000000,.3872983346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-1.936491673,-.5000000000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,.5000000000,0.,0.,0.,0.,-1.045825033,.7905694150,0.,1.045825033,.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,.5000000000,0.,0.,0.,0.,0.,.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.4841229182,.1250000000},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,.4183300132,0.,0.,0.,0.,0.,.5400617248,-.2041241452,0.,-.5400617248,-.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.1250000000,.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[20][20] = 
		{
			{1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,.6614378278,1.224744872,0.,0.,2.371708244,0.,2.561737692,.9682458365,0.,2.561737692,.6614378278},
			{0.,1.,0.,0.,0.,1.224744872,0.,0.,.9682458365,0.,0.,1.224744872,0.,0.,2.371708244,0.,0.,.9682458365,0.,0.},
			{0.,0.,1.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,-.8100925875,-.5000000000,0.,0.,-1.936491673,0.,-3.137475099,-.7905694150,0.,-3.137475099,-.8100925875},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.7905694150,0.,0.,-.5000000000,0.,0.,-1.936491673,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,.6123724358,0.,1.984313483,.2500000000,0.,1.984313483,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.4841229182,-.1250000000},
			{0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,.8100925875,-.5000000000,0.,0.,0.,0.,1.045825033,-.7905694150,0.,-1.045825033,-.8100925875},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,.7905694150,0.,0.,-.5000000000,0.,0.,0.,0.,0.,-.7905694150,0.,0.},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,-.4183300132,0.,0.,0.,0.,0.,-.5400617248,.2041241452,0.,.5400617248,.4183300132},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,.5123475385,0.,0.,0.,-.6123724358,0.,-.6614378278,.2500000000,0.,-.6614378278,.5123475385},
			{0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,-.6123724358,0.,0.,.2500000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,.1250000000,0.,0.,.1250000000,-.9682458365e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.4841229182,-.1250000000}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[20][20] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419},
			{0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.},
			{0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.},
			{0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[20][20] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774},
			{0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.},
			{0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.},
			{0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][20] = 
		{
			{1.414213562,1.224744872,1.054092553,.9354143468,1.154700539,1.,.8606629657,.9128709293,.7905694150,.7483314774,1.154700539,1.,.8606629657,1.118033988,.9682458365,.9660917834,.9128709293,.7905694150,.9660917834,.7483314774}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[20][20] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.414213562,1.224744872,1.054092553,1.118033988,.9682458365,.9165151392,1.414213562,1.224744872,1.054092553,1.369306394,1.185854122,1.183215957,1.118033988,.9682458365,1.183215957,.9165151392},
			{2.236067977,2.581988897,2.500000000,2.366431913,1.825741858,2.108185107,2.041241452,1.443375673,1.666666667,1.183215956,1.825741858,2.108185107,2.041241452,1.767766952,2.041241452,1.527525232,1.443375673,1.666666667,1.527525232,1.183215956},
			{2.645751311,3.436931772,3.549647869,3.500000000,2.160246900,2.806243040,2.898275350,1.707825129,2.218529918,1.400000000,2.160246900,2.806243040,2.898275350,2.091650066,2.717133141,1.807392228,1.707825129,2.218529918,1.807392228,1.400000000},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,1.732050808,1.500000000,1.290994449,1.643167672,1.423024947,1.496662955,.8660254040,.7500000000,.6454972245,1.341640786,1.161895004,1.449137675,.5477225575,.4743416490,.9660917834,.3741657387},
			{1.825741857,2.108185106,2.041241452,1.932183566,2.236067977,2.581988897,2.500000000,2.121320343,2.449489743,1.932183566,1.118033988,1.290994449,1.250000000,1.732050808,2.,1.870828694,.7071067810,.8164965809,1.247219129,.4830458916},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,1.643167673,1.423024947,1.224744872,1.732050808,1.500000000,1.690308509,.5477225576,.4743416490,.4082482906,1.060660172,.9185586536,1.309307341,.2886751347,.2500000000,.6546536709,.1690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8660254040,.7500000000,.6454972245,.5477225575,.4743416490,.3741657387,1.732050808,1.500000000,1.290994449,1.341640786,1.161895004,.9660917834,1.643167672,1.423024947,1.449137675,1.496662955},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.118033988,1.290994449,1.250000000,.7071067810,.8164965809,.4830458916,2.236067977,2.581988897,2.500000000,1.732050808,2.,1.247219129,2.121320343,2.449489743,1.870828694,1.932183566},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,1.341640786,1.161895004,1.,1.060660172,.9185586536,.8280786714,1.341640786,1.161895004,1.,1.732050808,1.500000000,1.603567452,1.060660172,.9185586536,1.603567452,.8280786714},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.5477225576,.4743416490,.4082482906,.2886751347,.2500000000,.1690308509,1.643167673,1.423024947,1.224744872,1.060660172,.9185586536,.6546536709,1.732050808,1.500000000,1.309307341,1.690308509},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[13][20] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,1.154700539,2.,2.581988897,.6085806194,1.054092553,.2771598064,1.154700539,2.,2.581988897,1.490711985,2.581988897,1.073435315,.6085806194,1.054092553,1.073435315,.2771598064},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,1.660371852,2.875848407,3.712704328,1.674178825,2.899762786,.9415775350,1.258317121,2.179469186,1.733484547,.8240209079},
			{1.414213562,2.449489743,3.162277660,3.741657387,.9018648819,1.562075797,2.016631182,.3712464703,.6430177487,.1320524832,.9018648819,1.562075797,2.016631182,.9093644209,1.575065380,.5114370684,.3712464703,.6430177487,.5114370684,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.660371852,2.875848407,3.712704328,1.258317121,2.179469186,.8240209079,.9018648819,1.562075797,2.016631182,1.674178825,2.899762786,1.733484547,.3712464703,.6430177487,.9415775350,.1320524832},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,3.012867026,5.218438765,6.736975475,.7599890903,1.316339718,.1069282633,4.143232550,7.176289285,1.427907550,4.923365040},
			{1.414213562,2.449489743,3.162277660,3.741657387,.2256172947,.3907806174,.5044956077,.2323400915e-1,.4024248430e-1,.2067470429e-2,.2256172947,.3907806174,.5044956077,.5691146707e-1,.9857355252e-1,.8007278541e-2,.2323400915e-1,.4024248430e-1,.8007278541e-2,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,3.012867026,5.218438765,6.736975475,4.143232550,7.176289285,4.923365040,.2256172947,.3907806174,.5044956077,.7599890903,1.316339718,1.427907550,.2323400915e-1,.4024248430e-1,.1069282633,.2067470429e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,.1686682004,.2921418926,.3771535615,.2043791133,.3539950084,.1381330405,.1298511510e-1,.2249087909e-1,.2149722933e-1,.8638170614e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,1.083797870,1.877192976,2.423445711,2.679889337,4.641704490,3.696095549,.5361372318,.9286169253,1.811248012,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,2.211635545,3.830665132,4.945367419,.4170631113,.7223744989,.4386799219e-1,2.232577295,3.866937307,.5752122252,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,.1686682004,.2921418926,.3771535615,.1298511510e-1,.2249087909e-1,.8638170614e-3,1.083797870,1.877192976,2.423445711,.2043791133,.3539950084,.2149722933e-1,.5361372318,.9286169253,.1381330405,.2291748860},
			{1.414213562,2.449489743,3.162277660,3.741657387,1.083797870,1.877192976,2.423445711,.5361372318,.9286169253,.2291748860,2.211635545,3.830665132,4.945367419,2.679889337,4.641704490,1.811248012,2.232577295,3.866937307,3.696095549,1.947434359},
			{1.414213562,2.449489743,3.162277660,3.741657387,2.211635545,3.830665132,4.945367419,2.232577295,3.866937307,1.947434359,.1686682004,.2921418926,.3771535615,.4170631113,.7223744989,.5752122252,.1298511510e-1,.2249087909e-1,.4386799219e-1,.8638170614e-3}
		};
		for (int j = 0; j < 13; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[13][20] = 
		{
			{1.414213562,0.,0.,0.,1.154700539,0.,0.,.6085806194,0.,.2771598064,1.154700539,0.,0.,1.490711985,0.,1.073435315,.6085806194,0.,1.073435315,.2771598064},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,1.660371852,0.,0.,1.674178825,0.,.9415775350,1.258317121,0.,1.733484547,.8240209079},
			{1.414213562,0.,0.,0.,.9018648819,0.,0.,.3712464703,0.,.1320524832,.9018648819,0.,0.,.9093644209,0.,.5114370684,.3712464703,0.,.5114370684,.1320524832},
			{1.414213562,0.,0.,0.,1.660371852,0.,0.,1.258317121,0.,.8240209079,.9018648819,0.,0.,1.674178825,0.,1.733484547,.3712464703,0.,.9415775350,.1320524832},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,3.012867026,0.,0.,.7599890903,0.,.1069282633,4.143232550,0.,1.427907550,4.923365040},
			{1.414213562,0.,0.,0.,.2256172947,0.,0.,.2323400915e-1,0.,.2067470429e-2,.2256172947,0.,0.,.5691146707e-1,0.,.8007278541e-2,.2323400915e-1,0.,.8007278541e-2,.2067470429e-2},
			{1.414213562,0.,0.,0.,3.012867026,0.,0.,4.143232550,0.,4.923365040,.2256172947,0.,0.,.7599890903,0.,1.427907550,.2323400915e-1,0.,.1069282633,.2067470429e-2},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,.1686682004,0.,0.,.2043791133,0.,.1381330405,.1298511510e-1,0.,.2149722933e-1,.8638170614e-3},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,1.083797870,0.,0.,2.679889337,0.,3.696095549,.5361372318,0.,1.811248012,.2291748860},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,2.211635545,0.,0.,.4170631113,0.,.4386799219e-1,2.232577295,0.,.5752122252,1.947434359},
			{1.414213562,0.,0.,0.,.1686682004,0.,0.,.1298511510e-1,0.,.8638170614e-3,1.083797870,0.,0.,.2043791133,0.,.2149722933e-1,.5361372318,0.,.1381330405,.2291748860},
			{1.414213562,0.,0.,0.,1.083797870,0.,0.,.5361372318,0.,.2291748860,2.211635545,0.,0.,2.679889337,0.,1.811248012,2.232577295,0.,3.696095549,1.947434359},
			{1.414213562,0.,0.,0.,2.211635545,0.,0.,2.232577295,0.,1.947434359,.1686682004,0.,0.,.4170631113,0.,.5752122252,.1298511510e-1,0.,.4386799219e-1,.8638170614e-3}
		};
		for (int j = 0; j < 13; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 20; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{.3663702899,-.4301677717,-.4301677717,-.4301677717,-.1306735064,-.1306735064,-.1306735064,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663,-.1888893663},
			{.4729820103,-.5553442052,-.5553442052,-.5553442052,-.1686987714,-.1686987714,-.1686987714,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233,-.2438551233},
			{.5596398619,-.6570921251,-.6570921251,-.6570921251,-.1996070782,-.1996070782,-.1996070782,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731,-.2885332731},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{.2991400890,-.2743243432,-.2743243432,-.5050428582,-.2084706567e-1,-.2084706567e-1,-.2783892823,-.1447574103,-.2953969950,-.2252816005e-1,-.2252816005e-1,-.1447574103,-.2953969950},
			{.3861881943,-.3541512041,-.3541512041,-.6520075263,-.2691344605e-1,-.2691344605e-1,-.3593990180,-.1868810131,-.3813558807,-.2908372956e-1,-.2908372956e-1,-.1868810131,-.3813558807},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{.1576606701,-.1129237275,-.1129237275,-.3827480421,-.2146825292e-2,-.2146825292e-2,-.3828351954,-.7160914354e-1,-.2981940789,-.1734356272e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{.2991400890,-.5050428582,-.2743243432,-.2743243432,-.2783892823,-.2084706567e-1,-.2084706567e-1,-.2252816005e-1,-.1447574103,-.2953969950,-.1447574103,-.2953969950,-.2252816005e-1},
			{.3861881943,-.6520075263,-.3541512041,-.3541512041,-.3593990180,-.2691344605e-1,-.2691344605e-1,-.2908372956e-1,-.1868810131,-.3813558807,-.1868810131,-.3813558807,-.2908372956e-1},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{.3861881943,-.5092425882,-.2766055121,-.5092425882,-.7022308507e-1,-.5258626531e-2,-.7022308507e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1,-.2729788643e-1,-.3579392902,-.5570501439e-1},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{.1576606701,-.3827480421,-.1129237275,-.1129237275,-.3828351954,-.2146825292e-2,-.2146825292e-2,-.1734356272e-2,-.7160914354e-1,-.2981940789,-.7160914354e-1,-.2981940789,-.1734356272e-2},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 13; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[20][13] = 
		{
			{.2115239854,-.2483574787,-.2483574787,-.2483574787,-.7544438409e-1,-.7544438409e-1,-.7544438409e-1,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265,-.1090553265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1727086110,-.1583812334,-.1583812334,-.2915866302,-.1203605898e-1,-.1203605898e-1,-.1607281271,-.8357572982e-1,-.1705475346,-.1300663927e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9102543031e-1,-.6519654444e-1,-.6519654444e-1,-.2209796851,-.1239470160e-2,-.1239470160e-2,-.2210300030,-.4134355830e-1,-.1721624317,-.1001331061e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.4145480460e-1,-.2319043083e-1,-.2319043083e-1,-.1447106438,-.1102938321e-3,-.1102938321e-3,-.2626479148,-.1767253736e-1,-.1501739875,-.6661218231e-4,-.6661218231e-4,-.1767253736e-1,-.1501739875},
			{.1727086110,-.2915866302,-.1583812334,-.1583812334,-.1607281271,-.1203605898e-1,-.1203605898e-1,-.1300663927e-1,-.8357572982e-1,-.1705475346,-.8357572982e-1,-.1705475346,-.1300663927e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.2229658579,-.2940113453,-.1596982669,-.2940113453,-.4054331706e-1,-.3036069443e-2,-.4054331706e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1,-.1576044208e-1,-.2066563455,-.3216130505e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.1653553812,-.8981615242e-1,-.3044263349,-.5704327255e-2,-.4271661750e-3,-.7617492052e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1},
			{.9102543031e-1,-.2209796851,-.6519654444e-1,-.6519654444e-1,-.2210300030,-.1239470160e-2,-.1239470160e-2,-.1001331061e-2,-.4134355830e-1,-.1721624317,-.4134355830e-1,-.1721624317,-.1001331061e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.1605537679,-.3044263349,-.8981615242e-1,-.1653553812,-.7617492052e-1,-.4271661750e-3,-.5704327255e-2,-.1657732202e-2,-.1396721461,-.4435677800e-1,-.1065195826e-1,-.2850198285,-.3382825861e-2},
			{.4145480460e-1,-.1447106438,-.2319043083e-1,-.2319043083e-1,-.2626479148,-.1102938321e-3,-.1102938321e-3,-.6661218231e-4,-.1767253736e-1,-.1501739875,-.1767253736e-1,-.1501739875,-.6661218231e-4}
		};
		for (int j = 0; j < 20; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 13; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
