
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (4)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (35)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (35)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (100)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (27)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*4+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[35],
		T io_hv[35]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{777.0000,-1510.348301,1690.467392,-952.470474,210.,-1871.410147,1934.644152,-1117.354016,259.229627,2556.16901,-1341.640786,311.769144,-1746.195870,366.6060572,469.574280,-1871.410147,1934.644152,-1117.354016,259.229627,2087.10326,-1095.445113,254.558441,-1352.597504,283.9718296,354.964788,2556.16901,-1341.640786,311.769144,-1352.597504,283.9718296,336.749168,-1746.195870,366.6060572,354.964788,469.574280},
			{-1510.348301,6784.000001,-11711.90164,8798.545336,-2424.871131,1934.644153,-5565.240696,5464.415795,-1795.995545,-1341.640786,2478.709341,-1080.,366.6060559,-423.3202096,0.,1934.644153,-5565.240696,5464.415795,-1795.995545,-1095.445114,2023.857701,-881.8163075,283.9718295,-327.9024243,0.,-1341.640786,2478.709341,-1080.,283.9718295,-327.9024243,0.,366.6060559,-423.3202096,0.,0.},
			{1690.467392,-11711.90164,27216.,-25557.46466,8452.336953,-1117.354017,5464.415795,-7936.346767,3477.930418,311.7691454,-1080.,836.5644027,0.,0.,0.,-1117.354017,5464.415795,-7936.346767,3477.930418,254.5584412,-881.8163075,683.0519743,0.,0.,0.,311.7691454,-1080.,836.5644027,0.,0.,0.,0.,0.,0.,0.},
			{-952.470474,8798.545336,-25557.46466,28800.,-11112.15551,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{210.,-2424.871131,8452.336953,-11112.15551,4900.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,8128.00002,-5043.731953,1717.300206,-211.6601049,-15027.14343,5126.683137,-763.6753235,12443.02215,-1795.995545,-3834.057901,4064.00002,-2521.865976,858.6501032,-105.8300524,-8179.74082,2790.612835,-415.6921939,7228.74263,-1043.379125,-2318.620276,-5009.04781,1708.894378,-254.5584412,4819.16175,-695.5860830,-1649.727254,3110.755535,-448.9988863,-1159.310137,-766.811579},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,5126.683137,-8196.623692,2645.448922,-1795.995545,2073.837024,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,2790.612835,-4461.676814,1440.,-1043.379125,1204.790437,0.,1708.894378,-2732.207897,881.8163075,-695.5860830,803.1936248,0.,-448.9988863,518.4592560,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,1717.300206,-7250.224824,8640.,-2839.718295,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,858.6501032,-3625.112412,4320.,-1419.859148,-415.6921939,1440.,-1115.419204,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-15027.14343,5126.683137,-763.6753235,0.,33696.,-6734.213542,482.9906830,-31970.48640,2839.718295,10911.92009,-5009.04781,1708.894378,-254.5584412,0.,13756.33440,-2749.231165,197.1801207,-14858.53936,1319.781800,5499.090835,5616.,-1122.368924,80.49844717,-7429.269681,659.8909002,3130.137376,-3197.048640,283.9718295,1649.727251,727.4613394},
			{-1341.640786,2478.709341,-1080.,0.,0.,5126.683137,-8196.623692,2645.448922,0.,-6734.213542,9216.,-1673.128805,2839.718295,-3279.024246,0.,1708.894378,-2732.207897,881.8163075,0.,-2749.231165,3762.416245,-683.0519743,1319.781800,-1523.952755,0.,-1122.368924,1536.,-278.8548009,659.8909002,-761.9763776,0.,283.9718295,-327.9024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,80.49844717,-278.8548009,216.,0.,0.,0.,0.,0.,0.,0.},
			{-1746.195870,366.6060559,0.,0.,0.,12443.02215,-1795.995545,0.,0.,-31970.48640,2839.718295,0.,33600.,-1385.640646,-12423.76754,3110.755535,-448.9988863,0.,0.,-10441.51713,927.4481114,0.,13013.22404,-536.6563145,-5366.563145,-3197.048640,283.9718295,0.,5205.289617,-214.6625258,-2545.584412,1680.,-69.28203232,-1073.312629,-354.9647869},
			{366.6060572,-423.3202096,0.,0.,0.,-1795.995545,2073.837024,0.,0.,2839.718295,-3279.024246,0.,-1385.640646,1600.,0.,-448.9988863,518.4592560,0.,0.,927.4481114,-1070.924834,0.,-536.6563145,619.6773354,0.,283.9718295,-327.9024246,0.,-214.6625258,247.8709341,0.,-69.28203232,80.,0.,0.},
			{469.574280,0.,0.,0.,0.,-3834.057901,0.,0.,0.,10911.92009,0.,0.,-12423.76754,0.,4900.,-766.811579,0.,0.,0.,2969.848480,0.,0.,-4124.318126,0.,1852.025918,727.4613394,0.,0.,-1374.772709,0.,752.9940236,-354.9647869,0.,264.5751311,70.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,4064.00002,-2521.865976,858.6501032,-105.8300524,-5009.04781,1708.894378,-254.5584412,3110.755535,-448.9988865,-766.811579,8128.00002,-5043.731953,1717.300206,-211.6601049,-8179.74082,2790.612835,-415.6921939,4819.16175,-695.5860830,-1159.310137,-15027.14343,5126.683137,-763.6753235,7228.74263,-1043.379125,-1649.727254,12443.02215,-1795.995545,-2318.620276,-3834.057901},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,1708.894378,-2732.207897,881.8163075,-448.9988863,518.4592560,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,2790.612835,-4461.676814,1440.,-695.5860830,803.1936248,0.,5126.683137,-8196.623692,2645.448922,-1043.379125,1204.790437,0.,-1795.995545,2073.837024,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,858.6501032,-3625.112412,4320.,-1419.859148,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,1717.300206,-7250.224824,8640.,-2839.718295,-415.6921939,1440.,-1115.419204,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2087.10326,-1095.445114,254.5584412,0.,0.,-8179.74082,2790.612835,-415.6921939,0.,13756.33440,-2749.231165,197.1801207,-10441.51713,927.4481114,2969.848480,-8179.74082,2790.612835,-415.6921939,0.,14976.,-2992.983796,214.6625258,-12131.94659,1077.597327,3591.991091,13756.33440,-2749.231165,197.1801207,-12131.94659,1077.597327,3833.619699,-10441.51713,927.4481114,3591.991091,2969.848480},
			{-1095.445113,2023.857701,-881.8163075,0.,0.,2790.612835,-4461.676814,1440.,0.,-2749.231165,3762.416245,-683.0519743,927.4481114,-1070.924834,0.,2790.612835,-4461.676814,1440.,0.,-2992.983796,4096.,-743.6128024,1077.597327,-1244.302214,0.,-2749.231165,3762.416245,-683.0519743,1077.597327,-1244.302214,0.,927.4481114,-1070.924834,0.,0.},
			{254.558441,-881.8163075,683.0519743,0.,0.,-415.6921939,1440.,-1115.419204,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,-415.6921939,1440.,-1115.419204,0.,214.6625258,-743.6128024,576.,0.,0.,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,7228.74263,-1043.379125,0.,0.,-14858.53936,1319.781800,0.,13013.22404,-536.6563145,-4124.318126,4819.16175,-695.5860830,0.,0.,-12131.94659,1077.597327,0.,12096.,-498.8306327,-4156.921939,-7429.269681,659.8909002,0.,9072.,-374.1229745,-3549.242172,5205.289617,-214.6625258,-2494.153164,-1374.772709},
			{283.9718296,-327.9024243,0.,0.,0.,-1043.379125,1204.790437,0.,0.,1319.781800,-1523.952755,0.,-536.6563145,619.6773354,0.,-695.5860830,803.1936248,0.,0.,1077.597327,-1244.302214,0.,-498.8306327,576.,0.,659.8909002,-761.9763776,0.,-374.1229745,432.,0.,-214.6625258,247.8709341,0.,0.},
			{354.964788,0.,0.,0.,0.,-2318.620276,0.,0.,0.,5499.090835,0.,0.,-5366.563145,0.,1852.025918,-1159.310137,0.,0.,0.,3591.991091,0.,0.,-4156.921939,0.,1600.,1649.727251,0.,0.,-2494.153164,0.,1138.419957,-1073.312629,0.,640.,264.5751311},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-5009.04781,1708.894378,-254.5584412,0.,5616.,-1122.368924,80.49844717,-3197.048640,283.9718295,727.4613394,-15027.14343,5126.683137,-763.6753235,0.,13756.33440,-2749.231165,197.1801207,-7429.269681,659.8909002,1649.727251,33696.,-6734.213542,482.9906830,-14858.53936,1319.781800,3130.137376,-31970.48640,2839.718295,5499.090835,10911.92009},
			{-1341.640786,2478.709341,-1080.,0.,0.,1708.894378,-2732.207897,881.8163075,0.,-1122.368924,1536.,-278.8548009,283.9718295,-327.9024246,0.,5126.683137,-8196.623692,2645.448922,0.,-2749.231165,3762.416245,-683.0519743,659.8909002,-761.9763776,0.,-6734.213542,9216.,-1673.128805,1319.781800,-1523.952755,0.,2839.718295,-3279.024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,80.49844717,-278.8548009,216.,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,4819.16175,-695.5860830,0.,0.,-7429.269681,659.8909002,0.,5205.289617,-214.6625258,-1374.772709,7228.74263,-1043.379125,0.,0.,-12131.94659,1077.597327,0.,9072.,-374.1229745,-2494.153164,-14858.53936,1319.781800,0.,12096.,-498.8306327,-3549.242172,13013.22404,-536.6563145,-4156.921939,-4124.318126},
			{283.9718296,-327.9024243,0.,0.,0.,-695.5860830,803.1936248,0.,0.,659.8909002,-761.9763776,0.,-214.6625258,247.8709341,0.,-1043.379125,1204.790437,0.,0.,1077.597327,-1244.302214,0.,-374.1229745,432.,0.,1319.781800,-1523.952755,0.,-498.8306327,576.,0.,-536.6563145,619.6773354,0.,0.},
			{336.749168,0.,0.,0.,0.,-1649.727254,0.,0.,0.,3130.137376,0.,0.,-2545.584412,0.,752.9940236,-1649.727254,0.,0.,0.,3833.619699,0.,0.,-3549.242172,0.,1138.419957,3130.137376,0.,0.,-3549.242172,0.,1296.,-2545.584412,0.,1138.419957,752.9940236},
			{-1746.195870,366.6060560,0.,0.,0.,3110.755535,-448.9988865,0.,0.,-3197.048640,283.9718295,0.,1680.,-69.28203232,-354.9647869,12443.02215,-1795.995545,0.,0.,-10441.51713,927.4481114,0.,5205.289617,-214.6625258,-1073.312629,-31970.48640,2839.718295,0.,13013.22404,-536.6563145,-2545.584412,33600.,-1385.640646,-5366.563145,-12423.76754},
			{366.6060572,-423.3202096,0.,0.,0.,-448.9988863,518.4592560,0.,0.,283.9718295,-327.9024246,0.,-69.28203232,80.,0.,-1795.995545,2073.837024,0.,0.,927.4481114,-1070.924834,0.,-214.6625258,247.8709341,0.,2839.718295,-3279.024246,0.,-536.6563145,619.6773354,0.,-1385.640646,1600.,0.,0.},
			{354.964788,0.,0.,0.,0.,-1159.310137,0.,0.,0.,1649.727251,0.,0.,-1073.312629,0.,264.5751311,-2318.620276,0.,0.,0.,3591.991091,0.,0.,-2494.153164,0.,640.,5499.090835,0.,0.,-4156.921939,0.,1138.419957,-5366.563145,0.,1600.,1852.025918},
			{469.574280,0.,0.,0.,0.,-766.811579,0.,0.,0.,727.4613394,0.,0.,-354.9647869,0.,70.,-3834.057901,0.,0.,0.,2969.848480,0.,0.,-1374.772709,0.,264.5751311,10911.92009,0.,0.,-4124.318126,0.,752.9940236,-12423.76754,0.,1852.025918,4900.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{777.0000,-1510.348301,1690.467392,-952.470474,210.,-1871.410147,1934.644152,-1117.354016,259.229627,2556.16901,-1341.640786,311.769144,-1746.195870,366.6060572,469.574280,-1871.410147,1934.644152,-1117.354016,259.229627,2087.10326,-1095.445113,254.558441,-1352.597504,283.9718296,354.964788,2556.16901,-1341.640786,311.769144,-1352.597504,283.9718296,336.749168,-1746.195870,366.6060572,354.964788,469.574280},
			{-1510.348301,6784.000001,-11711.90164,8798.545336,-2424.871131,1934.644153,-5565.240696,5464.415795,-1795.995545,-1341.640786,2478.709341,-1080.,366.6060559,-423.3202096,0.,1934.644153,-5565.240696,5464.415795,-1795.995545,-1095.445114,2023.857701,-881.8163075,283.9718295,-327.9024243,0.,-1341.640786,2478.709341,-1080.,283.9718295,-327.9024243,0.,366.6060559,-423.3202096,0.,0.},
			{1690.467392,-11711.90164,27216.,-25557.46466,8452.336953,-1117.354017,5464.415795,-7936.346767,3477.930418,311.7691454,-1080.,836.5644027,0.,0.,0.,-1117.354017,5464.415795,-7936.346767,3477.930418,254.5584412,-881.8163075,683.0519743,0.,0.,0.,311.7691454,-1080.,836.5644027,0.,0.,0.,0.,0.,0.,0.},
			{-952.470474,8798.545336,-25557.46466,28800.,-11112.15551,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{210.,-2424.871131,8452.336953,-11112.15551,4900.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,8128.00002,-5043.731953,1717.300206,-211.6601049,-15027.14343,5126.683137,-763.6753235,12443.02215,-1795.995545,-3834.057901,4064.00002,-2521.865976,858.6501032,-105.8300524,-8179.74082,2790.612835,-415.6921939,7228.74263,-1043.379125,-2318.620276,-5009.04781,1708.894378,-254.5584412,4819.16175,-695.5860830,-1649.727254,3110.755535,-448.9988863,-1159.310137,-766.811579},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,5126.683137,-8196.623692,2645.448922,-1795.995545,2073.837024,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,2790.612835,-4461.676814,1440.,-1043.379125,1204.790437,0.,1708.894378,-2732.207897,881.8163075,-695.5860830,803.1936248,0.,-448.9988863,518.4592560,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,1717.300206,-7250.224824,8640.,-2839.718295,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,858.6501032,-3625.112412,4320.,-1419.859148,-415.6921939,1440.,-1115.419204,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-15027.14343,5126.683137,-763.6753235,0.,33696.,-6734.213542,482.9906830,-31970.48640,2839.718295,10911.92009,-5009.04781,1708.894378,-254.5584412,0.,13756.33440,-2749.231165,197.1801207,-14858.53936,1319.781800,5499.090835,5616.,-1122.368924,80.49844717,-7429.269681,659.8909002,3130.137376,-3197.048640,283.9718295,1649.727251,727.4613394},
			{-1341.640786,2478.709341,-1080.,0.,0.,5126.683137,-8196.623692,2645.448922,0.,-6734.213542,9216.,-1673.128805,2839.718295,-3279.024246,0.,1708.894378,-2732.207897,881.8163075,0.,-2749.231165,3762.416245,-683.0519743,1319.781800,-1523.952755,0.,-1122.368924,1536.,-278.8548009,659.8909002,-761.9763776,0.,283.9718295,-327.9024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,80.49844717,-278.8548009,216.,0.,0.,0.,0.,0.,0.,0.},
			{-1746.195870,366.6060559,0.,0.,0.,12443.02215,-1795.995545,0.,0.,-31970.48640,2839.718295,0.,33600.,-1385.640646,-12423.76754,3110.755535,-448.9988863,0.,0.,-10441.51713,927.4481114,0.,13013.22404,-536.6563145,-5366.563145,-3197.048640,283.9718295,0.,5205.289617,-214.6625258,-2545.584412,1680.,-69.28203232,-1073.312629,-354.9647869},
			{366.6060572,-423.3202096,0.,0.,0.,-1795.995545,2073.837024,0.,0.,2839.718295,-3279.024246,0.,-1385.640646,1600.,0.,-448.9988863,518.4592560,0.,0.,927.4481114,-1070.924834,0.,-536.6563145,619.6773354,0.,283.9718295,-327.9024246,0.,-214.6625258,247.8709341,0.,-69.28203232,80.,0.,0.},
			{469.574280,0.,0.,0.,0.,-3834.057901,0.,0.,0.,10911.92009,0.,0.,-12423.76754,0.,4900.,-766.811579,0.,0.,0.,2969.848480,0.,0.,-4124.318126,0.,1852.025918,727.4613394,0.,0.,-1374.772709,0.,752.9940236,-354.9647869,0.,264.5751311,70.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,4064.00002,-2521.865976,858.6501032,-105.8300524,-5009.04781,1708.894378,-254.5584412,3110.755535,-448.9988865,-766.811579,8128.00002,-5043.731953,1717.300206,-211.6601049,-8179.74082,2790.612835,-415.6921939,4819.16175,-695.5860830,-1159.310137,-15027.14343,5126.683137,-763.6753235,7228.74263,-1043.379125,-1649.727254,12443.02215,-1795.995545,-2318.620276,-3834.057901},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,1708.894378,-2732.207897,881.8163075,-448.9988863,518.4592560,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,2790.612835,-4461.676814,1440.,-695.5860830,803.1936248,0.,5126.683137,-8196.623692,2645.448922,-1043.379125,1204.790437,0.,-1795.995545,2073.837024,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,858.6501032,-3625.112412,4320.,-1419.859148,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,1717.300206,-7250.224824,8640.,-2839.718295,-415.6921939,1440.,-1115.419204,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2087.10326,-1095.445114,254.5584412,0.,0.,-8179.74082,2790.612835,-415.6921939,0.,13756.33440,-2749.231165,197.1801207,-10441.51713,927.4481114,2969.848480,-8179.74082,2790.612835,-415.6921939,0.,14976.,-2992.983796,214.6625258,-12131.94659,1077.597327,3591.991091,13756.33440,-2749.231165,197.1801207,-12131.94659,1077.597327,3833.619699,-10441.51713,927.4481114,3591.991091,2969.848480},
			{-1095.445113,2023.857701,-881.8163075,0.,0.,2790.612835,-4461.676814,1440.,0.,-2749.231165,3762.416245,-683.0519743,927.4481114,-1070.924834,0.,2790.612835,-4461.676814,1440.,0.,-2992.983796,4096.,-743.6128024,1077.597327,-1244.302214,0.,-2749.231165,3762.416245,-683.0519743,1077.597327,-1244.302214,0.,927.4481114,-1070.924834,0.,0.},
			{254.558441,-881.8163075,683.0519743,0.,0.,-415.6921939,1440.,-1115.419204,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,-415.6921939,1440.,-1115.419204,0.,214.6625258,-743.6128024,576.,0.,0.,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,7228.74263,-1043.379125,0.,0.,-14858.53936,1319.781800,0.,13013.22404,-536.6563145,-4124.318126,4819.16175,-695.5860830,0.,0.,-12131.94659,1077.597327,0.,12096.,-498.8306327,-4156.921939,-7429.269681,659.8909002,0.,9072.,-374.1229745,-3549.242172,5205.289617,-214.6625258,-2494.153164,-1374.772709},
			{283.9718296,-327.9024243,0.,0.,0.,-1043.379125,1204.790437,0.,0.,1319.781800,-1523.952755,0.,-536.6563145,619.6773354,0.,-695.5860830,803.1936248,0.,0.,1077.597327,-1244.302214,0.,-498.8306327,576.,0.,659.8909002,-761.9763776,0.,-374.1229745,432.,0.,-214.6625258,247.8709341,0.,0.},
			{354.964788,0.,0.,0.,0.,-2318.620276,0.,0.,0.,5499.090835,0.,0.,-5366.563145,0.,1852.025918,-1159.310137,0.,0.,0.,3591.991091,0.,0.,-4156.921939,0.,1600.,1649.727251,0.,0.,-2494.153164,0.,1138.419957,-1073.312629,0.,640.,264.5751311},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-5009.04781,1708.894378,-254.5584412,0.,5616.,-1122.368924,80.49844717,-3197.048640,283.9718295,727.4613394,-15027.14343,5126.683137,-763.6753235,0.,13756.33440,-2749.231165,197.1801207,-7429.269681,659.8909002,1649.727251,33696.,-6734.213542,482.9906830,-14858.53936,1319.781800,3130.137376,-31970.48640,2839.718295,5499.090835,10911.92009},
			{-1341.640786,2478.709341,-1080.,0.,0.,1708.894378,-2732.207897,881.8163075,0.,-1122.368924,1536.,-278.8548009,283.9718295,-327.9024246,0.,5126.683137,-8196.623692,2645.448922,0.,-2749.231165,3762.416245,-683.0519743,659.8909002,-761.9763776,0.,-6734.213542,9216.,-1673.128805,1319.781800,-1523.952755,0.,2839.718295,-3279.024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,80.49844717,-278.8548009,216.,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,4819.16175,-695.5860830,0.,0.,-7429.269681,659.8909002,0.,5205.289617,-214.6625258,-1374.772709,7228.74263,-1043.379125,0.,0.,-12131.94659,1077.597327,0.,9072.,-374.1229745,-2494.153164,-14858.53936,1319.781800,0.,12096.,-498.8306327,-3549.242172,13013.22404,-536.6563145,-4156.921939,-4124.318126},
			{283.9718296,-327.9024243,0.,0.,0.,-695.5860830,803.1936248,0.,0.,659.8909002,-761.9763776,0.,-214.6625258,247.8709341,0.,-1043.379125,1204.790437,0.,0.,1077.597327,-1244.302214,0.,-374.1229745,432.,0.,1319.781800,-1523.952755,0.,-498.8306327,576.,0.,-536.6563145,619.6773354,0.,0.},
			{336.749168,0.,0.,0.,0.,-1649.727254,0.,0.,0.,3130.137376,0.,0.,-2545.584412,0.,752.9940236,-1649.727254,0.,0.,0.,3833.619699,0.,0.,-3549.242172,0.,1138.419957,3130.137376,0.,0.,-3549.242172,0.,1296.,-2545.584412,0.,1138.419957,752.9940236},
			{-1746.195870,366.6060560,0.,0.,0.,3110.755535,-448.9988865,0.,0.,-3197.048640,283.9718295,0.,1680.,-69.28203232,-354.9647869,12443.02215,-1795.995545,0.,0.,-10441.51713,927.4481114,0.,5205.289617,-214.6625258,-1073.312629,-31970.48640,2839.718295,0.,13013.22404,-536.6563145,-2545.584412,33600.,-1385.640646,-5366.563145,-12423.76754},
			{366.6060572,-423.3202096,0.,0.,0.,-448.9988863,518.4592560,0.,0.,283.9718295,-327.9024246,0.,-69.28203232,80.,0.,-1795.995545,2073.837024,0.,0.,927.4481114,-1070.924834,0.,-214.6625258,247.8709341,0.,2839.718295,-3279.024246,0.,-536.6563145,619.6773354,0.,-1385.640646,1600.,0.,0.},
			{354.964788,0.,0.,0.,0.,-1159.310137,0.,0.,0.,1649.727251,0.,0.,-1073.312629,0.,264.5751311,-2318.620276,0.,0.,0.,3591.991091,0.,0.,-2494.153164,0.,640.,5499.090835,0.,0.,-4156.921939,0.,1138.419957,-5366.563145,0.,1600.,1852.025918},
			{469.574280,0.,0.,0.,0.,-766.811579,0.,0.,0.,727.4613394,0.,0.,-354.9647869,0.,70.,-3834.057901,0.,0.,0.,2969.848480,0.,0.,-1374.772709,0.,264.5751311,10911.92009,0.,0.,-4124.318126,0.,752.9940236,-12423.76754,0.,1852.025918,4900.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,3.872983346,3.354101966,2.886751347,2.561737690,3.674234614,3.181980514,2.738612787,3.346640106,2.898275348,3.030457633,1.936491673,1.677050983,1.443375673,1.280868845,3.,2.598076212,2.236067977,3.240370349,2.806243040,3.207134903,1.224744872,1.060660172,.9128709292,2.160246899,1.870828694,2.535462764,.8366600264,.7245688370,1.603567452,.6060915266},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,3.354101966,3.872983346,3.750000000,3.549647869,3.181980514,3.674234614,3.557562366,2.898275350,3.346640106,2.624453295,1.677050983,1.936491673,1.875000000,1.774823935,2.598076212,3.,2.904737510,2.806243040,3.240370350,2.777460300,1.060660172,1.224744872,1.185854122,1.870828694,2.160246900,2.195775165,.7245688375,.8366600264,1.388730150,.5248906592},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.886751347,3.750000000,3.872983346,3.818813079,2.738612788,3.557562368,3.674234614,2.494438258,3.240370349,2.258769757,1.443375673,1.875000000,1.936491673,1.909406540,2.236067977,2.904737510,3.,2.415229459,3.137475099,2.390457218,.9128709293,1.185854122,1.224744872,1.610152972,2.091650066,1.889822365,.6236095646,.8100925872,1.195228610,.4517539516},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,5.019960159,4.347413025,3.741657387,3.320391543,5.291502622,4.582575695,3.944053189,5.163977793,4.472135954,4.909902528,1.673320053,1.449137675,1.247219129,1.106797181,3.240370349,2.806243040,2.415229459,4.,3.464101616,4.330127020,.8819171036,.7637626160,.6573421980,2.,1.732050808,2.738612788,.5163977793,.4472135954,1.299038106,.3273268354},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,4.347413025,5.019960159,4.860555525,4.600869484,4.582575696,5.291502622,5.123475385,4.472135954,5.163977793,4.252100321,1.449137675,1.673320053,1.620185175,1.533623161,2.806243040,3.240370350,3.137475099,3.464101616,4.,3.750000000,.7637626162,.8819171036,.8539125641,1.732050808,2.,2.371708245,.4472135954,.5163977793,1.125000000,.2834733546},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,5.796550700,5.019960159,4.320493800,3.834057903,6.546536711,5.669467095,4.879500367,6.708203931,5.809475019,6.614378278,1.449137675,1.254990040,1.080123450,.9585144758,3.207134903,2.777460300,2.390457218,4.330127020,3.750000000,5.,.6546536711,.5669467095,.4879500367,1.732050808,1.500000000,2.635231383,.3354101966,.2904737510,1.,.1889822365},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.581138830,1.369306394,1.178511302,1.045825033,1.,.8660254040,.7453559923,.6831300514,.5916079782,.4948716594,3.162277660,2.738612788,2.357022604,2.091650066,2.449489743,2.121320343,1.825741858,1.763834207,1.527525232,1.309307342,3.,2.598076212,2.236067977,2.645751311,2.291287848,2.070196679,2.732520206,2.366431913,2.618614683,2.474358298},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.369306394,1.581138830,1.530931089,1.449137675,.8660254040,1.,.9682458365,.5916079782,.6831300507,.4285714286,2.738612788,3.162277660,3.061862179,2.898275350,2.121320343,2.449489743,2.371708245,1.527525232,1.763834207,1.133893419,2.598076212,3.,2.904737510,2.291287848,2.645751311,1.792842914,2.366431913,2.732520203,2.267786838,2.142857143},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,1.178511302,1.530931089,1.581138830,1.559023911,.7453559923,.9682458365,1.,.5091750773,.6614378278,.3688555568,2.357022604,3.061862179,3.162277659,3.118047822,1.825741858,2.371708244,2.449489743,1.314684396,1.707825129,.9759000734,2.236067977,2.904737510,3.,1.972026594,2.561737692,1.543033500,2.036700309,2.645751311,1.951800147,1.844277784},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,4.582575696,3.968626966,3.415650256,3.031088914,4.140393357,3.585685828,3.086067000,3.535533905,3.061862179,2.988071523,3.055050464,2.645751311,2.277100171,2.020725943,5.070925527,4.391550330,3.779644731,5.477225575,4.743416490,5.270462768,2.070196679,1.792842914,1.543033500,4.107919181,3.557562368,5.,1.414213562,1.224744872,3.162277660,.9960238407},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.296148140,1.122497216,.9660917834,.8573214100,.6831300514,.5916079782,.5091750773,.4000000000,.3464101616,.2535462764,3.888444420,3.367491648,2.898275350,2.571964230,2.509980079,2.173706512,1.870828694,1.549193338,1.341640786,1.006230590,4.098780308,3.549647869,3.055050464,3.098386677,2.683281572,2.121320343,4.,3.464101616,3.354101966,3.803194146},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,1.122497216,1.296148140,1.254990040,1.187939392,.5916079782,.6831300507,.6614378278,.3464101616,.4000000000,.2195775163,3.367491648,3.888444419,3.764970119,3.563818176,2.173706511,2.509980079,2.430277762,1.341640786,1.549193338,.8714212528,3.549647869,4.098780304,3.968626966,2.683281572,3.098386677,1.837117307,3.464101616,4.,2.904737510,3.293662744},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,2.898275350,2.509980080,2.160246900,1.917028952,1.963961013,1.700840129,1.463850110,1.341640786,1.161895004,.9449111824,4.347413025,3.764970119,3.240370350,2.875543427,4.810702356,4.166190450,3.585685828,3.897114318,3.375000000,3.,3.927922025,3.401680258,2.927700220,5.196152424,4.500000000,4.743416490,3.354101966,2.904737510,5.,2.834733546},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.095445115,.9486832980,.8164965809,.7245688375,.4948716594,.4285714286,.3688555568,.2535462764,.2195775163,.1428571429,4.381780461,3.794733192,3.265986323,2.898275350,2.424366106,2.099562637,1.807015806,1.309307341,1.133893419,.7559289460,4.948716594,4.285714286,3.688555568,3.273268354,2.834733546,1.992047682,5.070925527,4.391550326,3.779644731,5.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,3.872983346,3.354101966,2.886751347,2.561737690,3.674234614,3.181980514,2.738612787,3.346640106,2.898275348,3.030457633,1.936491673,1.677050983,1.443375673,1.280868845,3.,2.598076212,2.236067977,3.240370349,2.806243040,3.207134903,1.224744872,1.060660172,.9128709292,2.160246899,1.870828694,2.535462764,.8366600264,.7245688370,1.603567452,.6060915266},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,3.354101966,3.872983346,3.750000000,3.549647869,3.181980514,3.674234614,3.557562366,2.898275350,3.346640106,2.624453295,1.677050983,1.936491673,1.875000000,1.774823935,2.598076212,3.,2.904737510,2.806243040,3.240370350,2.777460300,1.060660172,1.224744872,1.185854122,1.870828694,2.160246900,2.195775165,.7245688375,.8366600264,1.388730150,.5248906592},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.886751347,3.750000000,3.872983346,3.818813079,2.738612788,3.557562368,3.674234614,2.494438258,3.240370349,2.258769757,1.443375673,1.875000000,1.936491673,1.909406540,2.236067977,2.904737510,3.,2.415229459,3.137475099,2.390457218,.9128709293,1.185854122,1.224744872,1.610152972,2.091650066,1.889822365,.6236095646,.8100925872,1.195228610,.4517539516},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,5.019960159,4.347413025,3.741657387,3.320391543,5.291502622,4.582575695,3.944053189,5.163977793,4.472135954,4.909902528,1.673320053,1.449137675,1.247219129,1.106797181,3.240370349,2.806243040,2.415229459,4.,3.464101616,4.330127020,.8819171036,.7637626160,.6573421980,2.,1.732050808,2.738612788,.5163977793,.4472135954,1.299038106,.3273268354},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,4.347413025,5.019960159,4.860555525,4.600869484,4.582575696,5.291502622,5.123475385,4.472135954,5.163977793,4.252100321,1.449137675,1.673320053,1.620185175,1.533623161,2.806243040,3.240370350,3.137475099,3.464101616,4.,3.750000000,.7637626162,.8819171036,.8539125641,1.732050808,2.,2.371708245,.4472135954,.5163977793,1.125000000,.2834733546},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,5.796550700,5.019960159,4.320493800,3.834057903,6.546536711,5.669467095,4.879500367,6.708203931,5.809475019,6.614378278,1.449137675,1.254990040,1.080123450,.9585144758,3.207134903,2.777460300,2.390457218,4.330127020,3.750000000,5.,.6546536711,.5669467095,.4879500367,1.732050808,1.500000000,2.635231383,.3354101966,.2904737510,1.,.1889822365},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.581138830,1.369306394,1.178511302,1.045825033,1.,.8660254040,.7453559923,.6831300514,.5916079782,.4948716594,3.162277660,2.738612788,2.357022604,2.091650066,2.449489743,2.121320343,1.825741858,1.763834207,1.527525232,1.309307342,3.,2.598076212,2.236067977,2.645751311,2.291287848,2.070196679,2.732520206,2.366431913,2.618614683,2.474358298},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.369306394,1.581138830,1.530931089,1.449137675,.8660254040,1.,.9682458365,.5916079782,.6831300507,.4285714286,2.738612788,3.162277660,3.061862179,2.898275350,2.121320343,2.449489743,2.371708245,1.527525232,1.763834207,1.133893419,2.598076212,3.,2.904737510,2.291287848,2.645751311,1.792842914,2.366431913,2.732520203,2.267786838,2.142857143},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,1.178511302,1.530931089,1.581138830,1.559023911,.7453559923,.9682458365,1.,.5091750773,.6614378278,.3688555568,2.357022604,3.061862179,3.162277659,3.118047822,1.825741858,2.371708244,2.449489743,1.314684396,1.707825129,.9759000734,2.236067977,2.904737510,3.,1.972026594,2.561737692,1.543033500,2.036700309,2.645751311,1.951800147,1.844277784},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,4.582575696,3.968626966,3.415650256,3.031088914,4.140393357,3.585685828,3.086067000,3.535533905,3.061862179,2.988071523,3.055050464,2.645751311,2.277100171,2.020725943,5.070925527,4.391550330,3.779644731,5.477225575,4.743416490,5.270462768,2.070196679,1.792842914,1.543033500,4.107919181,3.557562368,5.,1.414213562,1.224744872,3.162277660,.9960238407},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.296148140,1.122497216,.9660917834,.8573214100,.6831300514,.5916079782,.5091750773,.4000000000,.3464101616,.2535462764,3.888444420,3.367491648,2.898275350,2.571964230,2.509980079,2.173706512,1.870828694,1.549193338,1.341640786,1.006230590,4.098780308,3.549647869,3.055050464,3.098386677,2.683281572,2.121320343,4.,3.464101616,3.354101966,3.803194146},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,1.122497216,1.296148140,1.254990040,1.187939392,.5916079782,.6831300507,.6614378278,.3464101616,.4000000000,.2195775163,3.367491648,3.888444419,3.764970119,3.563818176,2.173706511,2.509980079,2.430277762,1.341640786,1.549193338,.8714212528,3.549647869,4.098780304,3.968626966,2.683281572,3.098386677,1.837117307,3.464101616,4.,2.904737510,3.293662744},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,2.898275350,2.509980080,2.160246900,1.917028952,1.963961013,1.700840129,1.463850110,1.341640786,1.161895004,.9449111824,4.347413025,3.764970119,3.240370350,2.875543427,4.810702356,4.166190450,3.585685828,3.897114318,3.375000000,3.,3.927922025,3.401680258,2.927700220,5.196152424,4.500000000,4.743416490,3.354101966,2.904737510,5.,2.834733546},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.095445115,.9486832980,.8164965809,.7245688375,.4948716594,.4285714286,.3688555568,.2535462764,.2195775163,.1428571429,4.381780461,3.794733192,3.265986323,2.898275350,2.424366106,2.099562637,1.807015806,1.309307341,1.133893419,.7559289460,4.948716594,4.285714286,3.688555568,3.273268354,2.834733546,1.992047682,5.070925527,4.391550326,3.779644731,5.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,3.162277660,2.738612788,2.357022604,2.091650066,3.,2.598076212,2.236067977,2.732520206,2.366431913,2.474358298,1.581138830,1.369306394,1.178511302,1.045825033,2.449489743,2.121320343,1.825741858,2.645751311,2.291287848,2.618614683,1.,.8660254040,.7453559923,1.763834207,1.527525232,2.070196679,.6831300514,.5916079782,1.309307342,.4948716594},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.738612788,3.162277660,3.061862179,2.898275350,2.598076212,3.,2.904737510,2.366431913,2.732520203,2.142857143,1.369306394,1.581138830,1.530931089,1.449137675,2.121320343,2.449489743,2.371708245,2.291287848,2.645751311,2.267786838,.8660254040,1.,.9682458365,1.527525232,1.763834207,1.792842914,.5916079782,.6831300507,1.133893419,.4285714286},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.357022604,3.061862179,3.162277659,3.118047822,2.236067977,2.904737510,3.,2.036700309,2.645751311,1.844277784,1.178511302,1.530931089,1.581138830,1.559023911,1.825741858,2.371708244,2.449489743,1.972026594,2.561737692,1.951800147,.7453559923,.9682458365,1.,1.314684396,1.707825129,1.543033500,.5091750773,.6614378278,.9759000734,.3688555568},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,3.888444420,3.367491648,2.898275350,2.571964230,4.098780308,3.549647869,3.055050464,4.,3.464101616,3.803194146,1.296148140,1.122497216,.9660917834,.8573214100,2.509980079,2.173706512,1.870828694,3.098386677,2.683281572,3.354101966,.6831300514,.5916079782,.5091750773,1.549193338,1.341640786,2.121320343,.4000000000,.3464101616,1.006230590,.2535462764},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,3.367491648,3.888444419,3.764970119,3.563818176,3.549647869,4.098780304,3.968626966,3.464101616,4.,3.293662744,1.122497216,1.296148140,1.254990040,1.187939392,2.173706511,2.509980079,2.430277762,2.683281572,3.098386677,2.904737510,.5916079782,.6831300507,.6614378278,1.341640786,1.549193338,1.837117307,.3464101616,.4000000000,.8714212528,.2195775163},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,4.381780461,3.794733192,3.265986323,2.898275350,4.948716594,4.285714286,3.688555568,5.070925527,4.391550326,5.,1.095445115,.9486832980,.8164965809,.7245688375,2.424366106,2.099562637,1.807015806,3.273268354,2.834733546,3.779644731,.4948716594,.4285714286,.3688555568,1.309307341,1.133893419,1.992047682,.2535462764,.2195775163,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.936491673,1.677050983,1.443375673,1.280868845,1.224744872,1.060660172,.9128709292,.8366600264,.7245688370,.6060915266,3.872983346,3.354101966,2.886751347,2.561737690,3.,2.598076212,2.236067977,2.160246899,1.870828694,1.603567452,3.674234614,3.181980514,2.738612787,3.240370349,2.806243040,2.535462764,3.346640106,2.898275348,3.207134903,3.030457633},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,1.677050983,1.936491673,1.875000000,1.774823935,1.060660172,1.224744872,1.185854122,.7245688375,.8366600264,.5248906592,3.354101966,3.872983346,3.750000000,3.549647869,2.598076212,3.,2.904737510,1.870828694,2.160246900,1.388730150,3.181980514,3.674234614,3.557562366,2.806243040,3.240370350,2.195775165,2.898275350,3.346640106,2.777460300,2.624453295},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,1.443375673,1.875000000,1.936491673,1.909406540,.9128709293,1.185854122,1.224744872,.6236095646,.8100925872,.4517539516,2.886751347,3.750000000,3.872983346,3.818813079,2.236067977,2.904737510,3.,1.610152972,2.091650066,1.195228610,2.738612788,3.557562368,3.674234614,2.415229459,3.137475099,1.889822365,2.494438258,3.240370349,2.390457218,2.258769757},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,4.347413025,3.764970119,3.240370350,2.875543427,3.927922025,3.401680258,2.927700220,3.354101966,2.904737510,2.834733546,2.898275350,2.509980080,2.160246900,1.917028952,4.810702356,4.166190450,3.585685828,5.196152424,4.500000000,5.,1.963961013,1.700840129,1.463850110,3.897114318,3.375000000,4.743416490,1.341640786,1.161895004,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.673320053,1.449137675,1.247219129,1.106797181,.8819171036,.7637626160,.6573421980,.5163977793,.4472135954,.3273268354,5.019960159,4.347413025,3.741657387,3.320391543,3.240370349,2.806243040,2.415229459,2.,1.732050808,1.299038106,5.291502622,4.582575695,3.944053189,4.,3.464101616,2.738612788,5.163977793,4.472135954,4.330127020,4.909902528},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,1.449137675,1.673320053,1.620185175,1.533623161,.7637626162,.8819171036,.8539125641,.4472135954,.5163977793,.2834733546,4.347413025,5.019960159,4.860555525,4.600869484,2.806243040,3.240370350,3.137475099,1.732050808,2.,1.125000000,4.582575696,5.291502622,5.123475385,3.464101616,4.,2.371708245,4.472135954,5.163977793,3.750000000,4.252100321},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,3.055050464,2.645751311,2.277100171,2.020725943,2.070196679,1.792842914,1.543033500,1.414213562,1.224744872,.9960238407,4.582575696,3.968626966,3.415650256,3.031088914,5.070925527,4.391550330,3.779644731,4.107919181,3.557562368,3.162277660,4.140393357,3.585685828,3.086067000,5.477225575,4.743416490,5.,3.535533905,3.061862179,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.449137675,1.254990040,1.080123450,.9585144758,.6546536711,.5669467095,.4879500367,.3354101966,.2904737510,.1889822365,5.796550700,5.019960159,4.320493800,3.834057903,3.207134903,2.777460300,2.390457218,1.732050808,1.500000000,1.,6.546536711,5.669467095,4.879500367,4.330127020,3.750000000,2.635231383,6.708203931,5.809475019,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,3.162277660,2.738612788,2.357022604,2.091650066,3.,2.598076212,2.236067977,2.732520206,2.366431913,2.474358298,1.581138830,1.369306394,1.178511302,1.045825033,2.449489743,2.121320343,1.825741858,2.645751311,2.291287848,2.618614683,1.,.8660254040,.7453559923,1.763834207,1.527525232,2.070196679,.6831300514,.5916079782,1.309307342,.4948716594},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.738612788,3.162277660,3.061862179,2.898275350,2.598076212,3.,2.904737510,2.366431913,2.732520203,2.142857143,1.369306394,1.581138830,1.530931089,1.449137675,2.121320343,2.449489743,2.371708245,2.291287848,2.645751311,2.267786838,.8660254040,1.,.9682458365,1.527525232,1.763834207,1.792842914,.5916079782,.6831300507,1.133893419,.4285714286},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.357022604,3.061862179,3.162277659,3.118047822,2.236067977,2.904737510,3.,2.036700309,2.645751311,1.844277784,1.178511302,1.530931089,1.581138830,1.559023911,1.825741858,2.371708244,2.449489743,1.972026594,2.561737692,1.951800147,.7453559923,.9682458365,1.,1.314684396,1.707825129,1.543033500,.5091750773,.6614378278,.9759000734,.3688555568},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,3.888444420,3.367491648,2.898275350,2.571964230,4.098780308,3.549647869,3.055050464,4.,3.464101616,3.803194146,1.296148140,1.122497216,.9660917834,.8573214100,2.509980079,2.173706512,1.870828694,3.098386677,2.683281572,3.354101966,.6831300514,.5916079782,.5091750773,1.549193338,1.341640786,2.121320343,.4000000000,.3464101616,1.006230590,.2535462764},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,3.367491648,3.888444419,3.764970119,3.563818176,3.549647869,4.098780304,3.968626966,3.464101616,4.,3.293662744,1.122497216,1.296148140,1.254990040,1.187939392,2.173706511,2.509980079,2.430277762,2.683281572,3.098386677,2.904737510,.5916079782,.6831300507,.6614378278,1.341640786,1.549193338,1.837117307,.3464101616,.4000000000,.8714212528,.2195775163},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,4.381780461,3.794733192,3.265986323,2.898275350,4.948716594,4.285714286,3.688555568,5.070925527,4.391550326,5.,1.095445115,.9486832980,.8164965809,.7245688375,2.424366106,2.099562637,1.807015806,3.273268354,2.834733546,3.779644731,.4948716594,.4285714286,.3688555568,1.309307341,1.133893419,1.992047682,.2535462764,.2195775163,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.936491673,1.677050983,1.443375673,1.280868845,1.224744872,1.060660172,.9128709292,.8366600264,.7245688370,.6060915266,3.872983346,3.354101966,2.886751347,2.561737690,3.,2.598076212,2.236067977,2.160246899,1.870828694,1.603567452,3.674234614,3.181980514,2.738612787,3.240370349,2.806243040,2.535462764,3.346640106,2.898275348,3.207134903,3.030457633},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,1.677050983,1.936491673,1.875000000,1.774823935,1.060660172,1.224744872,1.185854122,.7245688375,.8366600264,.5248906592,3.354101966,3.872983346,3.750000000,3.549647869,2.598076212,3.,2.904737510,1.870828694,2.160246900,1.388730150,3.181980514,3.674234614,3.557562366,2.806243040,3.240370350,2.195775165,2.898275350,3.346640106,2.777460300,2.624453295},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,1.443375673,1.875000000,1.936491673,1.909406540,.9128709293,1.185854122,1.224744872,.6236095646,.8100925872,.4517539516,2.886751347,3.750000000,3.872983346,3.818813079,2.236067977,2.904737510,3.,1.610152972,2.091650066,1.195228610,2.738612788,3.557562368,3.674234614,2.415229459,3.137475099,1.889822365,2.494438258,3.240370349,2.390457218,2.258769757},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,4.347413025,3.764970119,3.240370350,2.875543427,3.927922025,3.401680258,2.927700220,3.354101966,2.904737510,2.834733546,2.898275350,2.509980080,2.160246900,1.917028952,4.810702356,4.166190450,3.585685828,5.196152424,4.500000000,5.,1.963961013,1.700840129,1.463850110,3.897114318,3.375000000,4.743416490,1.341640786,1.161895004,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.673320053,1.449137675,1.247219129,1.106797181,.8819171036,.7637626160,.6573421980,.5163977793,.4472135954,.3273268354,5.019960159,4.347413025,3.741657387,3.320391543,3.240370349,2.806243040,2.415229459,2.,1.732050808,1.299038106,5.291502622,4.582575695,3.944053189,4.,3.464101616,2.738612788,5.163977793,4.472135954,4.330127020,4.909902528},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,1.449137675,1.673320053,1.620185175,1.533623161,.7637626162,.8819171036,.8539125641,.4472135954,.5163977793,.2834733546,4.347413025,5.019960159,4.860555525,4.600869484,2.806243040,3.240370350,3.137475099,1.732050808,2.,1.125000000,4.582575696,5.291502622,5.123475385,3.464101616,4.,2.371708245,4.472135954,5.163977793,3.750000000,4.252100321},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,3.055050464,2.645751311,2.277100171,2.020725943,2.070196679,1.792842914,1.543033500,1.414213562,1.224744872,.9960238407,4.582575696,3.968626966,3.415650256,3.031088914,5.070925527,4.391550330,3.779644731,4.107919181,3.557562368,3.162277660,4.140393357,3.585685828,3.086067000,5.477225575,4.743416490,5.,3.535533905,3.061862179,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.449137675,1.254990040,1.080123450,.9585144758,.6546536711,.5669467095,.4879500367,.3354101966,.2904737510,.1889822365,5.796550700,5.019960159,4.320493800,3.834057903,3.207134903,2.777460300,2.390457218,1.732050808,1.500000000,1.,6.546536711,5.669467095,4.879500367,4.330127020,3.750000000,2.635231383,6.708203931,5.809475019,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,.172756623071377,.00390388665157827,6.57541912233899e-005,.00486900211054309,.000110027806742895,.000110027806742895,5.33523806427656,.120563624637786,.00203068488848474,.368327957423731,.00832333122831677,.0131603422009587,7.19422103343416,.16257219522684,.629639404683915,9.00134910987265},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,.844111317574503,.0190748976597194,.000321283525938621,.123027723654541,.00278013241624376,.014376841680633,4.7630793830775,.107634206373357,.00181291129080671,1.70046113382978,.0384263561113877,.314193514488106,6.0685481480974,.13713467930593,2.74657165283107,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,1.80585734803576,.0408080585873606,.000687340880331092,.625327304106952,.014130902021659,.173615533771693,3.86201434387853,.0872722907753235,.00146995018268015,3.27577043476888,.0740245800161786,1.43801908896753,4.43071582310304,.100123584532725,4.76431769725595,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.72409573525399,.061558050796614,.00103683846501621,1.66715454931021,.0376737069316204,.818062705137864,2.81340752869054,.0635762837863472,.00107083209499401,4.21756481949244,.0953068814408436,3.27222005765247,2.75487590413568,.0622536090052649,5.23550287189747,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,3.27976309580122,.0741148046448401,.00124833517774672,3.01515438576158,.0681352803972318,2.22245957594899,1.80736427869175,.0408421116086564,.000687914444399609,4.06994367638345,.0919709965436096,4.74332322095564,1.41847476579742,.0320541383810507,4.04940821803081,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,3.27976309580122,.0741148046448401,.00124833517774672,4.06994367638345,.0919709965436096,4.04940821803081,.991945941760336,.0224156067156837,.000377551968601727,3.01515438576158,.0681352803972318,4.74332322095564,.576746649127296,.013033095369547,2.22245957594899,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,2.72409573525399,.061558050796614,.00103683846501621,4.21756481949244,.0953068814408436,5.23550287189747,.439603191878621,.00993398112262371,.000167320660844516,1.66715454931021,.0376737069316204,3.27222005765247,.170154848292167,.00384509276081317,.818062705137864,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,1.80585734803576,.0408080585873606,.000687340880331092,3.27577043476888,.0740245800161786,4.76431769725595,.140734880491406,.00318027182678497,5.35661561215366e-005,.625327304106952,.014130902021659,1.43801908896753,.0308216365902836,.000696495297778531,.173615533771693,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.844111317574503,.0190748976597194,.000321283525938621,1.70046113382978,.0384263561113877,2.74657165283107,.0249321870529784,.00056340781892775,9.4896263063424e-006,.123027723654541,.00278013241624376,.314193514488106,.00229823078236534,5.1934521012163e-005,.014376841680633,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.172756623071377,.00390388665157827,6.57541912233899e-005,.368327957423731,.00832333122831677,.629639404683915,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,.00486900211054309,.000110027806742895,.0131603422009587,1.66187980881431e-005,3.75545104141921e-007,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,.172756623071377,.0201880888841649,.00175840721715611,.00486900211054309,.000568984537846987,.000110027806742895,5.33523806427656,.623468196598569,.0543048419765405,.368327957423731,.0430422718809401,.0131603422009587,7.19422103343416,.840706255138427,.629639404683915,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,.844111317574503,.0986416266094925,.00859180624463178,.123027723654541,.014376841680633,.014376841680633,4.7630793830775,.556606561402958,.0484811118273528,1.70046113382978,.198713426363864,.314193514488106,6.0685481480974,.709161751356258,2.74657165283107,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,1.80585734803576,.211029875475197,.0183809600898976,.625327304106952,.0730748434797269,.173615533771693,3.86201434387853,.45130940535495,.039309600832947,3.27577043476888,.382801790716804,1.43801908896753,4.43071582310304,.517767036891475,4.76431769725595,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.72409573525399,.318333884134563,.0277272703988645,1.66715454931021,.194821267114416,.818062705137864,2.81340752869054,.328770730954697,.0286363325160947,4.21756481949244,.492858519092109,3.27222005765247,2.75487590413568,.321930810017305,5.23550287189747,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,3.27976309580122,.383268367486451,.0333831432664456,3.01515438576158,.352346576520278,2.22245957594899,1.80736427869175,.211205973165025,.0183962984155437,4.06994367638345,.475607726017603,4.74332322095564,1.41847476579742,.165760907653494,4.04940821803081,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,3.27976309580122,.383268367486451,.0333831432664456,4.06994367638345,.475607726017603,4.04940821803081,.991945941760336,.115917366756987,.0100965443280308,3.01515438576158,.352346576520278,4.74332322095564,.576746649127296,.0673977784805412,2.22245957594899,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,2.72409573525399,.318333884134563,.0277272703988645,4.21756481949244,.492858519092109,5.23550287189747,.439603191878621,.0513713926084575,.00447451108643047,1.66715454931021,.194821267114416,3.27222005765247,.170154848292167,.0198840492440458,.818062705137864,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,1.80585734803576,.211029875475197,.0183809600898976,3.27577043476888,.382801790716804,4.76431769725595,.140734880491406,.0164460743984419,.00143247318181469,.625327304106952,.0730748434797269,1.43801908896753,.0308216365902836,.00360177183250731,.173615533771693,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.844111317574503,.0986416266094925,.00859180624463178,1.70046113382978,.198713426363864,2.74657165283107,.0249321870529784,.00291353928576501,.000253772832951388,.123027723654541,.014376841680633,.314193514488106,.00229823078236534,.000268567922156809,.014376841680633,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.172756623071377,.0201880888841649,.00175840721715611,.368327957423731,.0430422718809401,.629639404683915,.000932318622844532,.00010894940458837,9.4896263063424e-006,.00486900211054309,.000568984537846987,.0131603422009587,1.66187980881431e-005,1.9420486861126e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,.172756623071377,.0479640552145288,.00992569280167208,.00486900211054309,.00135182710686955,.000110027806742895,5.33523806427656,1.48127260473935,.306534899222172,.368327957423731,.102262374457619,.0131603422009587,7.19422103343416,1.99739962866362,.629639404683915,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,.844111317574503,.234358608796285,.0484982253050728,.123027723654541,.0341573505118309,.014376841680633,4.7630793830775,1.32241877857201,.273661651286941,1.70046113382978,.472115107510773,.314193514488106,6.0685481480974,1.68486841886076,2.74657165283107,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,1.80585734803576,.501377255533426,.10375512649862,.625327304106952,.173615533771693,.173615533771693,3.86201434387853,1.07224756942001,.221891162762184,3.27577043476888,.909483127972704,1.43801908896753,4.43071582310304,1.23014153980692,4.76431769725595,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.72409573525399,.756316463776911,.156512305865719,1.66715454931021,.462867885437628,.818062705137864,2.81340752869054,.781112941542527,.161643694787028,4.21756481949244,1.17096241047498,3.27222005765247,2.75487590413568,.764862252961431,5.23550287189747,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,3.27976309580122,.910591648648909,.188438048697755,3.01515438576158,.837125829385702,2.22245957594899,1.80736427869175,.501795638836895,.103841706846431,4.06994367638345,1.12997695631596,4.74332322095564,1.41847476579742,.393824565203833,4.04940821803081,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,3.27976309580122,.910591648648909,.188438048697755,4.06994367638345,1.12997695631596,4.04940821803081,.991945941760336,.275403333686328,.0569920302764555,3.01515438576158,.837125829385702,4.74332322095564,.576746649127296,.160127627083755,2.22245957594899,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,2.72409573525399,.756316463776911,.156512305865719,4.21756481949244,1.17096241047498,5.23550287189747,.439603191878621,.122051191950714,.0252573022041014,1.66715454931021,.462867885437628,3.27222005765247,.170154848292167,.0472416998608471,.818062705137864,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,1.80585734803576,.501377255533426,.10375512649862,3.27577043476888,.909483127972704,4.76431769725595,.140734880491406,.0390735559485202,.00808589080538567,.625327304106952,.173615533771693,1.43801908896753,.0308216365902836,.00855730247849368,.173615533771693,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.844111317574503,.234358608796285,.0484982253050728,1.70046113382978,.472115107510773,2.74657165283107,.0249321870529784,.00692215890141759,.00143247318181469,.123027723654541,.0341573505118309,.314193514488106,.00229823078236534,.000638079548841519,.014376841680633,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.172756623071377,.0479640552145288,.00992569280167208,.368327957423731,.102262374457619,.629639404683915,.000932318622844532,.000258848437177504,5.35661561215366e-005,.00486900211054309,.00135182710686955,.0131603422009587,1.66187980881431e-005,4.61403409428573e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,.172756623071377,.0847706354072961,.031004156339822,.00486900211054309,.00238919003611327,.000110027806742895,5.33523806427656,2.61796921368607,.957500511953663,.368327957423731,.180736312270888,.0131603422009587,7.19422103343416,3.53016097005959,.629639404683915,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,.844111317574503,.414200344236389,.151490338216903,.123027723654541,.0603689636701225,.014376841680633,4.7630793830775,2.33721439175002,.854816765967696,1.70046113382978,.834406045737126,.314193514488106,6.0685481480974,2.9778000589335,2.74657165283107,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,1.80585734803576,.886123334239273,.324092255049372,.625327304106952,.306844345178412,.173615533771693,3.86201434387853,1.89506719911644,.693105099882267,3.27577043476888,1.60740084018392,1.43801908896753,4.43071582310304,2.17412559267414,4.76431769725595,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.72409573525399,1.3366973855027,.488885974725042,1.66715454931021,.818062705137864,.818062705137864,2.81340752869054,1.38052214482826,.504914516765946,4.21756481949244,2.06953367746009,3.27222005765247,2.75487590413568,1.35180102876198,5.23550287189747,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,3.27976309580122,1.60936001568865,.58861006873112,3.01515438576158,1.4795181132097,2.22245957594899,1.80736427869175,.886862775854878,.324362699711839,4.06994367638345,1.99709687085614,4.74332322095564,1.41847476579742,.696037031940087,4.04940821803081,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,3.27976309580122,1.60936001568865,.58861006873112,4.06994367638345,1.99709687085614,4.04940821803081,.991945941760336,.48674190465042,.178021811889788,3.01515438576158,1.4795181132097,4.74332322095564,.576746649127296,.283006110168114,2.22245957594899,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,2.72409573525399,1.3366973855027,.488885974725042,4.21756481949244,2.06953367746009,5.23550287189747,.439603191878621,.215710640970697,.0788943766349663,1.66715454931021,.818062705137864,3.27222005765247,.170154848292167,.083493960154389,.818062705137864,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,1.80585734803576,.886123334239273,.324092255049372,3.27577043476888,1.60740084018392,4.76431769725595,.140734880491406,.0690577589939741,.0252573022041014,.625327304106952,.306844345178412,1.43801908896753,.0308216365902836,.0151239916064188,.173615533771693,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.844111317574503,.414200344236389,.151490338216903,1.70046113382978,.834406045737126,2.74657165283107,.0249321870529784,.0122340741590526,.00447451108643047,.123027723654541,.0603689636701225,.314193514488106,.00229823078236534,.00112772801535997,.014376841680633,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.172756623071377,.0847706354072961,.031004156339822,.368327957423731,.180736312270888,.629639404683915,.000932318622844532,.00045748313806202,.000167320660844516,.00486900211054309,.00238919003611327,.0131603422009587,1.66187980881431e-005,8.15474421864669e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,.172756623071377,.127338292609009,.0699595626855254,.00486900211054309,.00358892414166019,.000110027806742895,5.33523806427656,3.93258500625844,2.16055926028883,.368327957423731,.271493227654192,.0131603422009587,7.19422103343416,5.3028347422728,.629639404683915,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,.844111317574503,.622191450845088,.341831517573806,.123027723654541,.0906832976229768,.014376841680633,4.7630793830775,3.51084887681561,1.92885774629333,1.70046113382978,1.25340385495866,.314193514488106,6.0685481480974,4.47310526119946,2.74657165283107,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,1.80585734803576,1.33109103029466,.731300416128337,.625327304106952,.460926532212655,.173615533771693,3.86201434387853,2.8466770404089,1.5639622362693,3.27577043476888,2.41455873892965,1.43801908896753,4.43071582310304,3.26586487830957,4.76431769725595,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.72409573525399,2.00792128060645,1.10315044924867,1.66715454931021,1.22885368994647,.818062705137864,2.81340752869054,2.07375268554642,1.13931817396847,4.21756481949244,3.10875203091765,3.27222005765247,2.75487590413568,2.03060923305097,5.23550287189747,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,3.27976309580122,2.41750171632384,1.3281736341858,3.01515438576158,2.22245957594899,2.22245957594899,1.80736427869175,1.33220178323833,.731910663029089,4.06994367638345,2.99994101126824,4.74332322095564,1.41847476579742,1.04555270600537,4.04940821803081,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,3.27976309580122,2.41750171632384,1.3281736341858,4.06994367638345,2.99994101126824,4.04940821803081,.991945941760336,.731159826532418,.401698661682273,3.01515438576158,2.22245957594899,4.74332322095564,.576746649127296,.425117904255122,2.22245957594899,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,.981387965414959,.723377579561695,.397423100941195,.200115539631623,2.72409573525399,2.00792128060645,1.10315044924867,4.21756481949244,3.10875203091765,5.23550287189747,.439603191878621,.324029949602564,.178021811889788,1.66715454931021,1.22885368994647,3.27222005765247,.170154848292167,.125420533633389,.818062705137864,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,.555278916264744,.409294114646486,.224865880330915,.113227330974451,1.80585734803576,1.33109103029466,.731300416128337,3.27577043476888,2.41455873892965,4.76431769725595,.140734880491406,.103735179988284,.0569920302764555,.625327304106952,.460926532212655,1.43801908896753,.0308216365902836,.0227185187340072,.173615533771693,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.844111317574503,.622191450845088,.341831517573806,1.70046113382978,1.25340385495866,2.74657165283107,.0249321870529784,.018377426423439,.0100965443280308,.123027723654541,.0906832976229768,.314193514488106,.00229823078236534,.0016940177375493,.014376841680633,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.172756623071377,.127338292609009,.0699595626855254,.368327957423731,.271493227654192,.629639404683915,.000932318622844532,.000687208741781061,.000377551968601727,.00486900211054309,.00358892414166019,.0131603422009587,1.66187980881431e-005,1.22496569770466e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,.172756623071377,.171884955945497,.127469057765715,.00486900211054309,.00484443489339947,.000110027806742895,5.33523806427656,5.30831839113911,3.93662342336207,.368327957423731,.366469508258039,.0131603422009587,7.19422103343416,7.1579216069788,.629639404683915,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,.844111317574503,.839852238686345,.62283038639938,.123027723654541,.12240697042152,.014376841680633,4.7630793830775,4.73904661479697,3.51445420820192,1.70046113382978,1.69188122423627,.314193514488106,6.0685481480974,6.03792845818671,2.74657165283107,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,1.80585734803576,1.79674564825646,1.33245794297741,.625327304106952,.622172129531742,.173615533771693,3.86201434387853,3.84252802245541,2.84960032604169,3.27577043476888,3.259242086299,1.43801908896753,4.43071582310304,4.40836004025762,4.76431769725595,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.72409573525399,2.71035093833706,2.00998323805485,1.66715454931021,1.65874269265149,.818062705137864,2.81340752869054,2.79921209632888,2.07588224620454,4.21756481949244,4.19628451843992,3.27222005765247,2.75487590413568,2.7409758008707,5.23550287189747,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,3.27976309580122,3.26321460335875,2.41998427663067,3.01515438576158,2.99994101126824,2.22245957594899,1.80736427869175,1.79824497510369,1.33356983668368,4.06994367638345,4.04940821803081,4.74332322095564,1.41847476579742,1.41131765730707,4.04940821803081,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,3.27976309580122,3.26321460335875,2.41998427663067,4.06994367638345,4.04940821803081,4.04940821803081,.991945941760336,.986940942882965,.731910663029089,3.01515438576158,2.99994101126824,4.74332322095564,.576746649127296,.573836595005234,2.22245957594899,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,.981387965414959,.976436238050563,.724120424237357,.49217250527751,2.72409573525399,2.71035093833706,2.00998323805485,4.21756481949244,4.19628451843992,5.23550287189747,.439603191878621,.437385113867296,.324362699711839,1.66715454931021,1.65874269265149,3.27222005765247,.170154848292167,.169296308726466,.818062705137864,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,.555278916264744,.552477180456442,.409714423434642,.278476020673683,1.80585734803576,1.79674564825646,1.33245794297741,3.27577043476888,3.259242086299,4.76431769725595,.140734880491406,.140024783409284,.103841706846431,.625327304106952,.622172129531742,1.43801908896753,.0308216365902836,.0306661218062029,.173615533771693,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.844111317574503,.839852238686345,.62283038639938,1.70046113382978,1.69188122423627,2.74657165283107,.0249321870529784,.0248063882942385,.0183962984155437,.123027723654541,.12240697042152,.314193514488106,.00229823078236534,.00228663474453522,.014376841680633,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.172756623071377,.171884955945497,.127469057765715,.368327957423731,.366469508258039,.629639404683915,.000932318622844532,.000927614481757561,.000687914444399609,.00486900211054309,.00484443489339947,.0131603422009587,1.66187980881431e-005,1.65349456687952e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,.172756623071377,.21445261314721,.198422869712097,.00486900211054309,.0060441689989464,.000110027806742895,5.33523806427656,6.62293418371148,6.1278880563708,.368327957423731,.457226423641343,.0131603422009587,7.19422103343416,8.93059537919201,.629639404683915,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,.844111317574503,1.04784334529504,.969519935107731,.123027723654541,.152721304374374,.014376841680633,4.7630793830775,5.91268109986256,5.47072443843503,1.70046113382978,2.1108790334578,.314193514488106,6.0685481480974,7.53323366045268,2.74657165283107,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,1.80585734803576,2.24171334431186,2.07415143290851,.625327304106952,.776254316565985,.173615533771693,3.86201434387853,4.79413786374787,4.43578923494485,3.27577043476888,4.06639998504474,1.43801908896753,4.43071582310304,5.50009932589306,4.76431769725595,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.72409573525399,3.3815748334408,3.12881140849955,1.66715454931021,2.06953367746009,.818062705137864,2.81340752869054,3.49244263704704,3.2313921487783,4.21756481949244,5.23550287189747,3.27222005765247,2.75487590413568,3.41978400515968,5.23550287189747,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,3.27976309580122,4.07135630399393,3.76703360991161,3.01515438576158,3.74288247400753,2.22245957594899,1.80736427869175,2.24358398248714,2.07588224620454,4.06994367638345,5.05225235844291,4.74332322095564,1.41847476579742,1.76083333137235,4.04940821803081,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,3.27976309580122,4.07135630399393,3.76703360991161,4.06994367638345,5.05225235844291,4.04940821803081,.991945941760336,1.23135886476496,1.13931817396847,3.01515438576158,3.74288247400753,4.74332322095564,.576746649127296,.715948389092241,2.22245957594899,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,2.72409573525399,3.3815748334408,3.12881140849955,4.21756481949244,5.23550287189747,5.23550287189747,.439603191878621,.545704422499163,.504914516765946,1.66715454931021,2.06953367746009,3.27222005765247,.170154848292167,.211222882205465,.818062705137864,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,.555278916264744,.689299272303129,.637776046087675,.540839285781897,1.80585734803576,2.24171334431186,2.07415143290851,3.27577043476888,4.06639998504474,4.76431769725595,.140734880491406,.174702204403594,.161643694787028,.625327304106952,.776254316565985,1.43801908896753,.0308216365902836,.0382606489337913,.173615533771693,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.844111317574503,1.04784334529504,.969519935107731,1.70046113382978,2.1108790334578,2.74657165283107,.0249321870529784,.0309497405586248,.0286363325160947,.123027723654541,.152721304374374,.314193514488106,.00229823078236534,.00285292446672455,.014376841680633,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.172756623071377,.21445261314721,.198422869712097,.368327957423731,.457226423641343,.629639404683915,.000932318622844532,.0011573400854766,.00107083209499401,.00486900211054309,.0060441689989464,.0131603422009587,1.66187980881431e-005,2.0629858427195e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,.172756623071377,.251259193339977,.272378587590661,.00486900211054309,.00708153192819011,.000110027806742895,5.33523806427656,7.7596307926582,8.41186046814944,.368327957423731,.535700361454612,.0131603422009587,7.19422103343416,10.463356720588,.629639404683915,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,.844111317574503,1.22768508073515,1.33087718642915,.123027723654541,.178932917532666,.014376841680633,4.7630793830775,6.92747671304058,7.50976033055426,1.70046113382978,2.47316997168416,.314193514488106,6.0685481480974,8.82616530052541,2.74657165283107,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,1.80585734803576,2.6264594230177,2.8472244079752,.625327304106952,.909483127972704,.173615533771693,3.86201434387853,5.6169574934443,6.08908644662382,3.27577043476888,4.76431769725595,1.43801908896753,4.43071582310304,6.44408337876027,4.76431769725595,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.72409573525399,3.9619557551666,4.29497483592082,1.66715454931021,2.42472849716033,.818062705137864,2.81340752869054,4.09185184033277,4.43578923494485,4.21756481949244,6.13407413888259,3.27222005765247,2.75487590413568,4.00672278096023,5.23550287189747,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,3.27976309580122,4.77012467103367,5.17107375557585,3.01515438576158,4.38527475783153,2.22245957594899,1.80736427869175,2.62865111950513,2.84960032604169,4.06994367638345,5.91937227298309,4.74332322095564,1.41847476579742,2.0630457981086,4.04940821803081,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,3.27976309580122,4.77012467103367,5.17107375557585,4.06994367638345,5.91937227298309,4.04940821803081,.991945941760336,1.44269743572906,1.5639622362693,3.01515438576158,4.38527475783153,4.74332322095564,.576746649127296,.838826872176601,2.22245957594899,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,2.72409573525399,3.9619557551666,4.29497483592082,4.21756481949244,6.13407413888259,5.23550287189747,.439603191878621,.639363871519146,.693105099882267,1.66715454931021,2.42472849716033,3.27222005765247,.170154848292167,.247475142499007,.818062705137864,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,.555278916264744,.807603957755601,.875486474339234,.869841424679861,1.80585734803576,2.6264594230177,2.8472244079752,3.27577043476888,4.76431769725595,4.76431769725595,.140734880491406,.204686407449048,.221891162762184,.625327304106952,.909483127972704,1.43801908896753,.0308216365902836,.0448273380617164,.173615533771693,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.844111317574503,1.22768508073515,1.33087718642915,1.70046113382978,2.47316997168416,2.74657165283107,.0249321870529784,.0362616558162599,.039309600832947,.123027723654541,.178932917532666,.314193514488106,.00229823078236534,.003342572933243,.014376841680633,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.172756623071377,.251259193339977,.272378587590661,.368327957423731,.535700361454612,.629639404683915,.000932318622844532,.00135597478636112,.00146995018268015,.00486900211054309,.00708153192819011,.0131603422009587,1.66187980881431e-005,2.4170568551556e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,.172756623071377,.279035159670341,.33592853869153,.00486900211054309,.00786437449721268,.000110027806742895,5.33523806427656,8.61743520079898,10.3744718692395,.368327957423731,.594920464031291,.0131603422009587,7.19422103343416,11.6200500941132,.629639404683915,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,.844111317574503,1.36340206292194,1.64139050859212,.123027723654541,.198713426363864,.014376841680633,4.7630793830775,7.69328893020963,9.26189843365359,1.70046113382978,2.74657165283107,.314193514488106,6.0685481480974,9.80187196802991,2.74657165283107,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,1.80585734803576,2.91680680307593,3.51152395332694,.625327304106952,1.01002381826467,.173615533771693,3.86201434387853,6.23789565750936,7.50976033055426,3.27577043476888,5.29099903451185,1.43801908896753,4.43071582310304,7.15645788167572,4.76431769725595,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.72409573525399,4.39993833480895,5.29705595843705,1.66715454931021,2.69277511548354,.818062705137864,2.81340752869054,4.5441940509206,5.47072443843503,4.21756481949244,6.81217803026546,3.27222005765247,2.75487590413568,4.44965422390436,5.23550287189747,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,3.27976309580122,5.29744795219613,6.37756170755723,3.01515438576158,4.87005401069696,2.22245957594899,1.80736427869175,2.919240785177,3.51445420820192,4.06994367638345,6.57374150328145,4.74332322095564,1.41847476579742,2.29110945565894,4.04940821803081,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,3.27976309580122,5.29744795219613,6.37756170755723,4.06994367638345,6.57374150328145,4.04940821803081,.991945941760336,1.6021834026584,1.92885774629333,3.01515438576158,4.87005401069696,4.74332322095564,.576746649127296,.931556720779814,2.22245957594899,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,2.72409573525399,4.39993833480895,5.29705595843705,4.21756481949244,6.81217803026546,5.23550287189747,.439603191878621,.710043670861402,.854816765967696,1.66715454931021,2.69277511548354,3.27222005765247,.170154848292167,.274832793115808,.818062705137864,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,1.80585734803576,2.91680680307593,3.51152395332694,3.27577043476888,5.29099903451185,4.76431769725595,.140734880491406,.227313888999126,.273661651286941,.625327304106952,1.01002381826467,1.43801908896753,.0308216365902836,.0497828687077028,.173615533771693,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.844111317574503,1.36340206292194,1.64139050859212,1.70046113382978,2.74657165283107,2.74657165283107,.0249321870529784,.0402702754319125,.0484811118273528,.123027723654541,.198713426363864,.314193514488106,.00229823078236534,.00371208455992771,.014376841680633,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.172756623071377,.279035159670341,.33592853869153,.368327957423731,.594920464031291,.629639404683915,.000932318622844532,.00150587381895025,.00181291129080671,.00486900211054309,.00786437449721268,.0131603422009587,1.66187980881431e-005,2.68425539597291e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,.172756623071377,.295319361902928,.376281515036571,.00486900211054309,.00832333122831677,.000110027806742895,5.33523806427656,9.12033977275976,11.620691733626,.368327957423731,.629639404683915,.0131603422009587,7.19422103343416,12.2981841540248,.629639404683915,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,.844111317574503,1.44296879187171,1.83856039664088,.123027723654541,.210310135628253,.014376841680633,4.7630793830775,8.14226128523923,10.3744718692395,1.70046113382978,2.90685872308354,.314193514488106,6.0685481480974,10.3738990400802,2.74657165283107,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,1.80585734803576,3.08702861996377,3.93334117545275,.625327304106952,1.06896775972274,.173615533771693,3.86201434387853,6.60193277208899,8.41186046814944,3.27577043476888,5.59977624521248,1.43801908896753,4.43071582310304,7.57410133403447,4.76431769725595,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.72409573525399,4.65671416814689,5.93335787735631,1.66715454931021,2.84992267566634,.818062705137864,2.81340752869054,4.80938849808895,6.1278880563708,4.21756481949244,7.20972966791673,3.27222005765247,2.75487590413568,4.7093314249164,5.23550287189747,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,3.27976309580122,5.60660151503774,7.14365796638211,3.01515438576158,5.15426530682,2.22245957594899,1.80736427869175,3.08960464673337,3.93662342336207,4.06994367638345,6.95737823275544,4.74332322095564,1.41847476579742,2.42481622493138,4.04940821803081,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,3.27976309580122,5.60660151503774,7.14365796638211,4.06994367638345,6.95737823275544,4.04940821803081,.991945941760336,1.6956851626997,2.16055926028883,3.01515438576158,5.15426530682,4.74332322095564,.576746649127296,.985921403890809,2.22245957594899,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,2.72409573525399,4.65671416814689,5.93335787735631,4.21756481949244,7.20972966791673,5.23550287189747,.439603191878621,.751481082347236,.957500511953663,1.66715454931021,2.84992267566634,3.27222005765247,.170154848292167,.290871749599041,.818062705137864,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,1.80585734803576,3.08702861996377,3.93334117545275,3.27577043476888,5.59977624521248,4.76431769725595,.140734880491406,.240579691570783,.306534899222172,.625327304106952,1.06896775972274,1.43801908896753,.0308216365902836,.0526881452424316,.173615533771693,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.844111317574503,1.44296879187171,1.83856039664088,1.70046113382978,2.90685872308354,2.74657165283107,.0249321870529784,.0426204068987497,.0543048419765405,.123027723654541,.210310135628253,.314193514488106,.00229823078236534,.00392871796107236,.014376841680633,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.172756623071377,.295319361902928,.376281515036571,.368327957423731,.629639404683915,.629639404683915,.000932318622844532,.00159375505167201,.00203068488848474,.00486900211054309,.00832333122831677,.0131603422009587,1.66187980881431e-005,2.84090575416998e-005,.000110027806742895,2.74870870898337e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,.172756623071377,.00390388665157827,6.57541912233899e-005,.00486900211054309,.000110027806742895,.000110027806742895,5.33523806427656,.120563624637786,.00203068488848474,.368327957423731,.00832333122831677,.0131603422009587,7.19422103343416,.16257219522684,.629639404683915,9.00134910987265},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,.844111317574503,.0190748976597194,.000321283525938621,.123027723654541,.00278013241624376,.014376841680633,4.7630793830775,.107634206373357,.00181291129080671,1.70046113382978,.0384263561113877,.314193514488106,6.0685481480974,.13713467930593,2.74657165283107,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,1.80585734803576,.0408080585873606,.000687340880331092,.625327304106952,.014130902021659,.173615533771693,3.86201434387853,.0872722907753235,.00146995018268015,3.27577043476888,.0740245800161786,1.43801908896753,4.43071582310304,.100123584532725,4.76431769725595,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.72409573525399,.061558050796614,.00103683846501621,1.66715454931021,.0376737069316204,.818062705137864,2.81340752869054,.0635762837863472,.00107083209499401,4.21756481949244,.0953068814408436,3.27222005765247,2.75487590413568,.0622536090052649,5.23550287189747,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,3.27976309580122,.0741148046448401,.00124833517774672,3.01515438576158,.0681352803972318,2.22245957594899,1.80736427869175,.0408421116086564,.000687914444399609,4.06994367638345,.0919709965436096,4.74332322095564,1.41847476579742,.0320541383810507,4.04940821803081,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,3.27976309580122,.0741148046448401,.00124833517774672,4.06994367638345,.0919709965436096,4.04940821803081,.991945941760336,.0224156067156837,.000377551968601727,3.01515438576158,.0681352803972318,4.74332322095564,.576746649127296,.013033095369547,2.22245957594899,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,2.72409573525399,.061558050796614,.00103683846501621,4.21756481949244,.0953068814408436,5.23550287189747,.439603191878621,.00993398112262371,.000167320660844516,1.66715454931021,.0376737069316204,3.27222005765247,.170154848292167,.00384509276081317,.818062705137864,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,1.80585734803576,.0408080585873606,.000687340880331092,3.27577043476888,.0740245800161786,4.76431769725595,.140734880491406,.00318027182678497,5.35661561215366e-005,.625327304106952,.014130902021659,1.43801908896753,.0308216365902836,.000696495297778531,.173615533771693,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.844111317574503,.0190748976597194,.000321283525938621,1.70046113382978,.0384263561113877,2.74657165283107,.0249321870529784,.00056340781892775,9.4896263063424e-006,.123027723654541,.00278013241624376,.314193514488106,.00229823078236534,5.1934521012163e-005,.014376841680633,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.172756623071377,.00390388665157827,6.57541912233899e-005,.368327957423731,.00832333122831677,.629639404683915,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,.00486900211054309,.000110027806742895,.0131603422009587,1.66187980881431e-005,3.75545104141921e-007,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,.172756623071377,.0201880888841649,.00175840721715611,.00486900211054309,.000568984537846987,.000110027806742895,5.33523806427656,.623468196598569,.0543048419765405,.368327957423731,.0430422718809401,.0131603422009587,7.19422103343416,.840706255138427,.629639404683915,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,.844111317574503,.0986416266094925,.00859180624463178,.123027723654541,.014376841680633,.014376841680633,4.7630793830775,.556606561402958,.0484811118273528,1.70046113382978,.198713426363864,.314193514488106,6.0685481480974,.709161751356258,2.74657165283107,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,1.80585734803576,.211029875475197,.0183809600898976,.625327304106952,.0730748434797269,.173615533771693,3.86201434387853,.45130940535495,.039309600832947,3.27577043476888,.382801790716804,1.43801908896753,4.43071582310304,.517767036891475,4.76431769725595,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.72409573525399,.318333884134563,.0277272703988645,1.66715454931021,.194821267114416,.818062705137864,2.81340752869054,.328770730954697,.0286363325160947,4.21756481949244,.492858519092109,3.27222005765247,2.75487590413568,.321930810017305,5.23550287189747,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,3.27976309580122,.383268367486451,.0333831432664456,3.01515438576158,.352346576520278,2.22245957594899,1.80736427869175,.211205973165025,.0183962984155437,4.06994367638345,.475607726017603,4.74332322095564,1.41847476579742,.165760907653494,4.04940821803081,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,3.27976309580122,.383268367486451,.0333831432664456,4.06994367638345,.475607726017603,4.04940821803081,.991945941760336,.115917366756987,.0100965443280308,3.01515438576158,.352346576520278,4.74332322095564,.576746649127296,.0673977784805412,2.22245957594899,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,2.72409573525399,.318333884134563,.0277272703988645,4.21756481949244,.492858519092109,5.23550287189747,.439603191878621,.0513713926084575,.00447451108643047,1.66715454931021,.194821267114416,3.27222005765247,.170154848292167,.0198840492440458,.818062705137864,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,1.80585734803576,.211029875475197,.0183809600898976,3.27577043476888,.382801790716804,4.76431769725595,.140734880491406,.0164460743984419,.00143247318181469,.625327304106952,.0730748434797269,1.43801908896753,.0308216365902836,.00360177183250731,.173615533771693,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.844111317574503,.0986416266094925,.00859180624463178,1.70046113382978,.198713426363864,2.74657165283107,.0249321870529784,.00291353928576501,.000253772832951388,.123027723654541,.014376841680633,.314193514488106,.00229823078236534,.000268567922156809,.014376841680633,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.172756623071377,.0201880888841649,.00175840721715611,.368327957423731,.0430422718809401,.629639404683915,.000932318622844532,.00010894940458837,9.4896263063424e-006,.00486900211054309,.000568984537846987,.0131603422009587,1.66187980881431e-005,1.9420486861126e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,.172756623071377,.0479640552145288,.00992569280167208,.00486900211054309,.00135182710686955,.000110027806742895,5.33523806427656,1.48127260473935,.306534899222172,.368327957423731,.102262374457619,.0131603422009587,7.19422103343416,1.99739962866362,.629639404683915,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,.844111317574503,.234358608796285,.0484982253050728,.123027723654541,.0341573505118309,.014376841680633,4.7630793830775,1.32241877857201,.273661651286941,1.70046113382978,.472115107510773,.314193514488106,6.0685481480974,1.68486841886076,2.74657165283107,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,1.80585734803576,.501377255533426,.10375512649862,.625327304106952,.173615533771693,.173615533771693,3.86201434387853,1.07224756942001,.221891162762184,3.27577043476888,.909483127972704,1.43801908896753,4.43071582310304,1.23014153980692,4.76431769725595,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.72409573525399,.756316463776911,.156512305865719,1.66715454931021,.462867885437628,.818062705137864,2.81340752869054,.781112941542527,.161643694787028,4.21756481949244,1.17096241047498,3.27222005765247,2.75487590413568,.764862252961431,5.23550287189747,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,3.27976309580122,.910591648648909,.188438048697755,3.01515438576158,.837125829385702,2.22245957594899,1.80736427869175,.501795638836895,.103841706846431,4.06994367638345,1.12997695631596,4.74332322095564,1.41847476579742,.393824565203833,4.04940821803081,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,3.27976309580122,.910591648648909,.188438048697755,4.06994367638345,1.12997695631596,4.04940821803081,.991945941760336,.275403333686328,.0569920302764555,3.01515438576158,.837125829385702,4.74332322095564,.576746649127296,.160127627083755,2.22245957594899,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,2.72409573525399,.756316463776911,.156512305865719,4.21756481949244,1.17096241047498,5.23550287189747,.439603191878621,.122051191950714,.0252573022041014,1.66715454931021,.462867885437628,3.27222005765247,.170154848292167,.0472416998608471,.818062705137864,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,1.80585734803576,.501377255533426,.10375512649862,3.27577043476888,.909483127972704,4.76431769725595,.140734880491406,.0390735559485202,.00808589080538567,.625327304106952,.173615533771693,1.43801908896753,.0308216365902836,.00855730247849368,.173615533771693,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.844111317574503,.234358608796285,.0484982253050728,1.70046113382978,.472115107510773,2.74657165283107,.0249321870529784,.00692215890141759,.00143247318181469,.123027723654541,.0341573505118309,.314193514488106,.00229823078236534,.000638079548841519,.014376841680633,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.172756623071377,.0479640552145288,.00992569280167208,.368327957423731,.102262374457619,.629639404683915,.000932318622844532,.000258848437177504,5.35661561215366e-005,.00486900211054309,.00135182710686955,.0131603422009587,1.66187980881431e-005,4.61403409428573e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,.172756623071377,.0847706354072961,.031004156339822,.00486900211054309,.00238919003611327,.000110027806742895,5.33523806427656,2.61796921368607,.957500511953663,.368327957423731,.180736312270888,.0131603422009587,7.19422103343416,3.53016097005959,.629639404683915,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,.844111317574503,.414200344236389,.151490338216903,.123027723654541,.0603689636701225,.014376841680633,4.7630793830775,2.33721439175002,.854816765967696,1.70046113382978,.834406045737126,.314193514488106,6.0685481480974,2.9778000589335,2.74657165283107,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,1.80585734803576,.886123334239273,.324092255049372,.625327304106952,.306844345178412,.173615533771693,3.86201434387853,1.89506719911644,.693105099882267,3.27577043476888,1.60740084018392,1.43801908896753,4.43071582310304,2.17412559267414,4.76431769725595,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.72409573525399,1.3366973855027,.488885974725042,1.66715454931021,.818062705137864,.818062705137864,2.81340752869054,1.38052214482826,.504914516765946,4.21756481949244,2.06953367746009,3.27222005765247,2.75487590413568,1.35180102876198,5.23550287189747,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,3.27976309580122,1.60936001568865,.58861006873112,3.01515438576158,1.4795181132097,2.22245957594899,1.80736427869175,.886862775854878,.324362699711839,4.06994367638345,1.99709687085614,4.74332322095564,1.41847476579742,.696037031940087,4.04940821803081,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,3.27976309580122,1.60936001568865,.58861006873112,4.06994367638345,1.99709687085614,4.04940821803081,.991945941760336,.48674190465042,.178021811889788,3.01515438576158,1.4795181132097,4.74332322095564,.576746649127296,.283006110168114,2.22245957594899,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,2.72409573525399,1.3366973855027,.488885974725042,4.21756481949244,2.06953367746009,5.23550287189747,.439603191878621,.215710640970697,.0788943766349663,1.66715454931021,.818062705137864,3.27222005765247,.170154848292167,.083493960154389,.818062705137864,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,1.80585734803576,.886123334239273,.324092255049372,3.27577043476888,1.60740084018392,4.76431769725595,.140734880491406,.0690577589939741,.0252573022041014,.625327304106952,.306844345178412,1.43801908896753,.0308216365902836,.0151239916064188,.173615533771693,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.844111317574503,.414200344236389,.151490338216903,1.70046113382978,.834406045737126,2.74657165283107,.0249321870529784,.0122340741590526,.00447451108643047,.123027723654541,.0603689636701225,.314193514488106,.00229823078236534,.00112772801535997,.014376841680633,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.172756623071377,.0847706354072961,.031004156339822,.368327957423731,.180736312270888,.629639404683915,.000932318622844532,.00045748313806202,.000167320660844516,.00486900211054309,.00238919003611327,.0131603422009587,1.66187980881431e-005,8.15474421864669e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,.172756623071377,.127338292609009,.0699595626855254,.00486900211054309,.00358892414166019,.000110027806742895,5.33523806427656,3.93258500625844,2.16055926028883,.368327957423731,.271493227654192,.0131603422009587,7.19422103343416,5.3028347422728,.629639404683915,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,.844111317574503,.622191450845088,.341831517573806,.123027723654541,.0906832976229768,.014376841680633,4.7630793830775,3.51084887681561,1.92885774629333,1.70046113382978,1.25340385495866,.314193514488106,6.0685481480974,4.47310526119946,2.74657165283107,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,1.80585734803576,1.33109103029466,.731300416128337,.625327304106952,.460926532212655,.173615533771693,3.86201434387853,2.8466770404089,1.5639622362693,3.27577043476888,2.41455873892965,1.43801908896753,4.43071582310304,3.26586487830957,4.76431769725595,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.72409573525399,2.00792128060645,1.10315044924867,1.66715454931021,1.22885368994647,.818062705137864,2.81340752869054,2.07375268554642,1.13931817396847,4.21756481949244,3.10875203091765,3.27222005765247,2.75487590413568,2.03060923305097,5.23550287189747,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,3.27976309580122,2.41750171632384,1.3281736341858,3.01515438576158,2.22245957594899,2.22245957594899,1.80736427869175,1.33220178323833,.731910663029089,4.06994367638345,2.99994101126824,4.74332322095564,1.41847476579742,1.04555270600537,4.04940821803081,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,3.27976309580122,2.41750171632384,1.3281736341858,4.06994367638345,2.99994101126824,4.04940821803081,.991945941760336,.731159826532418,.401698661682273,3.01515438576158,2.22245957594899,4.74332322095564,.576746649127296,.425117904255122,2.22245957594899,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,.981387965414959,.723377579561695,.397423100941195,.200115539631623,2.72409573525399,2.00792128060645,1.10315044924867,4.21756481949244,3.10875203091765,5.23550287189747,.439603191878621,.324029949602564,.178021811889788,1.66715454931021,1.22885368994647,3.27222005765247,.170154848292167,.125420533633389,.818062705137864,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,.555278916264744,.409294114646486,.224865880330915,.113227330974451,1.80585734803576,1.33109103029466,.731300416128337,3.27577043476888,2.41455873892965,4.76431769725595,.140734880491406,.103735179988284,.0569920302764555,.625327304106952,.460926532212655,1.43801908896753,.0308216365902836,.0227185187340072,.173615533771693,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.844111317574503,.622191450845088,.341831517573806,1.70046113382978,1.25340385495866,2.74657165283107,.0249321870529784,.018377426423439,.0100965443280308,.123027723654541,.0906832976229768,.314193514488106,.00229823078236534,.0016940177375493,.014376841680633,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.172756623071377,.127338292609009,.0699595626855254,.368327957423731,.271493227654192,.629639404683915,.000932318622844532,.000687208741781061,.000377551968601727,.00486900211054309,.00358892414166019,.0131603422009587,1.66187980881431e-005,1.22496569770466e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,.172756623071377,.171884955945497,.127469057765715,.00486900211054309,.00484443489339947,.000110027806742895,5.33523806427656,5.30831839113911,3.93662342336207,.368327957423731,.366469508258039,.0131603422009587,7.19422103343416,7.1579216069788,.629639404683915,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,.844111317574503,.839852238686345,.62283038639938,.123027723654541,.12240697042152,.014376841680633,4.7630793830775,4.73904661479697,3.51445420820192,1.70046113382978,1.69188122423627,.314193514488106,6.0685481480974,6.03792845818671,2.74657165283107,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,1.80585734803576,1.79674564825646,1.33245794297741,.625327304106952,.622172129531742,.173615533771693,3.86201434387853,3.84252802245541,2.84960032604169,3.27577043476888,3.259242086299,1.43801908896753,4.43071582310304,4.40836004025762,4.76431769725595,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.72409573525399,2.71035093833706,2.00998323805485,1.66715454931021,1.65874269265149,.818062705137864,2.81340752869054,2.79921209632888,2.07588224620454,4.21756481949244,4.19628451843992,3.27222005765247,2.75487590413568,2.7409758008707,5.23550287189747,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,3.27976309580122,3.26321460335875,2.41998427663067,3.01515438576158,2.99994101126824,2.22245957594899,1.80736427869175,1.79824497510369,1.33356983668368,4.06994367638345,4.04940821803081,4.74332322095564,1.41847476579742,1.41131765730707,4.04940821803081,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,3.27976309580122,3.26321460335875,2.41998427663067,4.06994367638345,4.04940821803081,4.04940821803081,.991945941760336,.986940942882965,.731910663029089,3.01515438576158,2.99994101126824,4.74332322095564,.576746649127296,.573836595005234,2.22245957594899,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,.981387965414959,.976436238050563,.724120424237357,.49217250527751,2.72409573525399,2.71035093833706,2.00998323805485,4.21756481949244,4.19628451843992,5.23550287189747,.439603191878621,.437385113867296,.324362699711839,1.66715454931021,1.65874269265149,3.27222005765247,.170154848292167,.169296308726466,.818062705137864,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,.555278916264744,.552477180456442,.409714423434642,.278476020673683,1.80585734803576,1.79674564825646,1.33245794297741,3.27577043476888,3.259242086299,4.76431769725595,.140734880491406,.140024783409284,.103841706846431,.625327304106952,.622172129531742,1.43801908896753,.0308216365902836,.0306661218062029,.173615533771693,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.844111317574503,.839852238686345,.62283038639938,1.70046113382978,1.69188122423627,2.74657165283107,.0249321870529784,.0248063882942385,.0183962984155437,.123027723654541,.12240697042152,.314193514488106,.00229823078236534,.00228663474453522,.014376841680633,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.172756623071377,.171884955945497,.127469057765715,.368327957423731,.366469508258039,.629639404683915,.000932318622844532,.000927614481757561,.000687914444399609,.00486900211054309,.00484443489339947,.0131603422009587,1.66187980881431e-005,1.65349456687952e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,.172756623071377,.21445261314721,.198422869712097,.00486900211054309,.0060441689989464,.000110027806742895,5.33523806427656,6.62293418371148,6.1278880563708,.368327957423731,.457226423641343,.0131603422009587,7.19422103343416,8.93059537919201,.629639404683915,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,.844111317574503,1.04784334529504,.969519935107731,.123027723654541,.152721304374374,.014376841680633,4.7630793830775,5.91268109986256,5.47072443843503,1.70046113382978,2.1108790334578,.314193514488106,6.0685481480974,7.53323366045268,2.74657165283107,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,1.80585734803576,2.24171334431186,2.07415143290851,.625327304106952,.776254316565985,.173615533771693,3.86201434387853,4.79413786374787,4.43578923494485,3.27577043476888,4.06639998504474,1.43801908896753,4.43071582310304,5.50009932589306,4.76431769725595,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.72409573525399,3.3815748334408,3.12881140849955,1.66715454931021,2.06953367746009,.818062705137864,2.81340752869054,3.49244263704704,3.2313921487783,4.21756481949244,5.23550287189747,3.27222005765247,2.75487590413568,3.41978400515968,5.23550287189747,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,3.27976309580122,4.07135630399393,3.76703360991161,3.01515438576158,3.74288247400753,2.22245957594899,1.80736427869175,2.24358398248714,2.07588224620454,4.06994367638345,5.05225235844291,4.74332322095564,1.41847476579742,1.76083333137235,4.04940821803081,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,3.27976309580122,4.07135630399393,3.76703360991161,4.06994367638345,5.05225235844291,4.04940821803081,.991945941760336,1.23135886476496,1.13931817396847,3.01515438576158,3.74288247400753,4.74332322095564,.576746649127296,.715948389092241,2.22245957594899,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,2.72409573525399,3.3815748334408,3.12881140849955,4.21756481949244,5.23550287189747,5.23550287189747,.439603191878621,.545704422499163,.504914516765946,1.66715454931021,2.06953367746009,3.27222005765247,.170154848292167,.211222882205465,.818062705137864,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,.555278916264744,.689299272303129,.637776046087675,.540839285781897,1.80585734803576,2.24171334431186,2.07415143290851,3.27577043476888,4.06639998504474,4.76431769725595,.140734880491406,.174702204403594,.161643694787028,.625327304106952,.776254316565985,1.43801908896753,.0308216365902836,.0382606489337913,.173615533771693,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.844111317574503,1.04784334529504,.969519935107731,1.70046113382978,2.1108790334578,2.74657165283107,.0249321870529784,.0309497405586248,.0286363325160947,.123027723654541,.152721304374374,.314193514488106,.00229823078236534,.00285292446672455,.014376841680633,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.172756623071377,.21445261314721,.198422869712097,.368327957423731,.457226423641343,.629639404683915,.000932318622844532,.0011573400854766,.00107083209499401,.00486900211054309,.0060441689989464,.0131603422009587,1.66187980881431e-005,2.0629858427195e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,.172756623071377,.251259193339977,.272378587590661,.00486900211054309,.00708153192819011,.000110027806742895,5.33523806427656,7.7596307926582,8.41186046814944,.368327957423731,.535700361454612,.0131603422009587,7.19422103343416,10.463356720588,.629639404683915,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,.844111317574503,1.22768508073515,1.33087718642915,.123027723654541,.178932917532666,.014376841680633,4.7630793830775,6.92747671304058,7.50976033055426,1.70046113382978,2.47316997168416,.314193514488106,6.0685481480974,8.82616530052541,2.74657165283107,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,1.80585734803576,2.6264594230177,2.8472244079752,.625327304106952,.909483127972704,.173615533771693,3.86201434387853,5.6169574934443,6.08908644662382,3.27577043476888,4.76431769725595,1.43801908896753,4.43071582310304,6.44408337876027,4.76431769725595,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.72409573525399,3.9619557551666,4.29497483592082,1.66715454931021,2.42472849716033,.818062705137864,2.81340752869054,4.09185184033277,4.43578923494485,4.21756481949244,6.13407413888259,3.27222005765247,2.75487590413568,4.00672278096023,5.23550287189747,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,3.27976309580122,4.77012467103367,5.17107375557585,3.01515438576158,4.38527475783153,2.22245957594899,1.80736427869175,2.62865111950513,2.84960032604169,4.06994367638345,5.91937227298309,4.74332322095564,1.41847476579742,2.0630457981086,4.04940821803081,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,3.27976309580122,4.77012467103367,5.17107375557585,4.06994367638345,5.91937227298309,4.04940821803081,.991945941760336,1.44269743572906,1.5639622362693,3.01515438576158,4.38527475783153,4.74332322095564,.576746649127296,.838826872176601,2.22245957594899,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,2.72409573525399,3.9619557551666,4.29497483592082,4.21756481949244,6.13407413888259,5.23550287189747,.439603191878621,.639363871519146,.693105099882267,1.66715454931021,2.42472849716033,3.27222005765247,.170154848292167,.247475142499007,.818062705137864,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,.555278916264744,.807603957755601,.875486474339234,.869841424679861,1.80585734803576,2.6264594230177,2.8472244079752,3.27577043476888,4.76431769725595,4.76431769725595,.140734880491406,.204686407449048,.221891162762184,.625327304106952,.909483127972704,1.43801908896753,.0308216365902836,.0448273380617164,.173615533771693,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.844111317574503,1.22768508073515,1.33087718642915,1.70046113382978,2.47316997168416,2.74657165283107,.0249321870529784,.0362616558162599,.039309600832947,.123027723654541,.178932917532666,.314193514488106,.00229823078236534,.003342572933243,.014376841680633,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.172756623071377,.251259193339977,.272378587590661,.368327957423731,.535700361454612,.629639404683915,.000932318622844532,.00135597478636112,.00146995018268015,.00486900211054309,.00708153192819011,.0131603422009587,1.66187980881431e-005,2.4170568551556e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,.172756623071377,.279035159670341,.33592853869153,.00486900211054309,.00786437449721268,.000110027806742895,5.33523806427656,8.61743520079898,10.3744718692395,.368327957423731,.594920464031291,.0131603422009587,7.19422103343416,11.6200500941132,.629639404683915,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,.844111317574503,1.36340206292194,1.64139050859212,.123027723654541,.198713426363864,.014376841680633,4.7630793830775,7.69328893020963,9.26189843365359,1.70046113382978,2.74657165283107,.314193514488106,6.0685481480974,9.80187196802991,2.74657165283107,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,1.80585734803576,2.91680680307593,3.51152395332694,.625327304106952,1.01002381826467,.173615533771693,3.86201434387853,6.23789565750936,7.50976033055426,3.27577043476888,5.29099903451185,1.43801908896753,4.43071582310304,7.15645788167572,4.76431769725595,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.72409573525399,4.39993833480895,5.29705595843705,1.66715454931021,2.69277511548354,.818062705137864,2.81340752869054,4.5441940509206,5.47072443843503,4.21756481949244,6.81217803026546,3.27222005765247,2.75487590413568,4.44965422390436,5.23550287189747,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,3.27976309580122,5.29744795219613,6.37756170755723,3.01515438576158,4.87005401069696,2.22245957594899,1.80736427869175,2.919240785177,3.51445420820192,4.06994367638345,6.57374150328145,4.74332322095564,1.41847476579742,2.29110945565894,4.04940821803081,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,3.27976309580122,5.29744795219613,6.37756170755723,4.06994367638345,6.57374150328145,4.04940821803081,.991945941760336,1.6021834026584,1.92885774629333,3.01515438576158,4.87005401069696,4.74332322095564,.576746649127296,.931556720779814,2.22245957594899,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,2.72409573525399,4.39993833480895,5.29705595843705,4.21756481949244,6.81217803026546,5.23550287189747,.439603191878621,.710043670861402,.854816765967696,1.66715454931021,2.69277511548354,3.27222005765247,.170154848292167,.274832793115808,.818062705137864,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,1.80585734803576,2.91680680307593,3.51152395332694,3.27577043476888,5.29099903451185,4.76431769725595,.140734880491406,.227313888999126,.273661651286941,.625327304106952,1.01002381826467,1.43801908896753,.0308216365902836,.0497828687077028,.173615533771693,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.844111317574503,1.36340206292194,1.64139050859212,1.70046113382978,2.74657165283107,2.74657165283107,.0249321870529784,.0402702754319125,.0484811118273528,.123027723654541,.198713426363864,.314193514488106,.00229823078236534,.00371208455992771,.014376841680633,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.172756623071377,.279035159670341,.33592853869153,.368327957423731,.594920464031291,.629639404683915,.000932318622844532,.00150587381895025,.00181291129080671,.00486900211054309,.00786437449721268,.0131603422009587,1.66187980881431e-005,2.68425539597291e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,.172756623071377,.295319361902928,.376281515036571,.00486900211054309,.00832333122831677,.000110027806742895,5.33523806427656,9.12033977275976,11.620691733626,.368327957423731,.629639404683915,.0131603422009587,7.19422103343416,12.2981841540248,.629639404683915,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,.844111317574503,1.44296879187171,1.83856039664088,.123027723654541,.210310135628253,.014376841680633,4.7630793830775,8.14226128523923,10.3744718692395,1.70046113382978,2.90685872308354,.314193514488106,6.0685481480974,10.3738990400802,2.74657165283107,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,1.80585734803576,3.08702861996377,3.93334117545275,.625327304106952,1.06896775972274,.173615533771693,3.86201434387853,6.60193277208899,8.41186046814944,3.27577043476888,5.59977624521248,1.43801908896753,4.43071582310304,7.57410133403447,4.76431769725595,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.72409573525399,4.65671416814689,5.93335787735631,1.66715454931021,2.84992267566634,.818062705137864,2.81340752869054,4.80938849808895,6.1278880563708,4.21756481949244,7.20972966791673,3.27222005765247,2.75487590413568,4.7093314249164,5.23550287189747,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,3.27976309580122,5.60660151503774,7.14365796638211,3.01515438576158,5.15426530682,2.22245957594899,1.80736427869175,3.08960464673337,3.93662342336207,4.06994367638345,6.95737823275544,4.74332322095564,1.41847476579742,2.42481622493138,4.04940821803081,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,3.27976309580122,5.60660151503774,7.14365796638211,4.06994367638345,6.95737823275544,4.04940821803081,.991945941760336,1.6956851626997,2.16055926028883,3.01515438576158,5.15426530682,4.74332322095564,.576746649127296,.985921403890809,2.22245957594899,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,2.72409573525399,4.65671416814689,5.93335787735631,4.21756481949244,7.20972966791673,5.23550287189747,.439603191878621,.751481082347236,.957500511953663,1.66715454931021,2.84992267566634,3.27222005765247,.170154848292167,.290871749599041,.818062705137864,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,1.80585734803576,3.08702861996377,3.93334117545275,3.27577043476888,5.59977624521248,4.76431769725595,.140734880491406,.240579691570783,.306534899222172,.625327304106952,1.06896775972274,1.43801908896753,.0308216365902836,.0526881452424316,.173615533771693,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.844111317574503,1.44296879187171,1.83856039664088,1.70046113382978,2.90685872308354,2.74657165283107,.0249321870529784,.0426204068987497,.0543048419765405,.123027723654541,.210310135628253,.314193514488106,.00229823078236534,.00392871796107236,.014376841680633,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.172756623071377,.295319361902928,.376281515036571,.368327957423731,.629639404683915,.629639404683915,.000932318622844532,.00159375505167201,.00203068488848474,.00486900211054309,.00832333122831677,.0131603422009587,1.66187980881431e-005,2.84090575416998e-005,.000110027806742895,2.74870870898337e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,0.,0.,0.,0.,0.,0.,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,0.,0.,0.,1.66187980881431e-005,3.75545104141921e-007,0.,2.74870870898337e-007},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,0.,0.,0.,0.,0.,0.,.0249321870529784,.00056340781892775,9.4896263063424e-006,0.,0.,0.,.00229823078236534,5.1934521012163e-005,0.,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,0.,0.,0.,0.,0.,0.,.140734880491406,.00318027182678497,5.35661561215366e-005,0.,0.,0.,.0308216365902836,.000696495297778531,0.,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,0.,0.,0.,0.,0.,0.,.439603191878621,.00993398112262371,.000167320660844516,0.,0.,0.,.170154848292167,.00384509276081317,0.,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,0.,0.,0.,0.,0.,0.,.991945941760336,.0224156067156837,.000377551968601727,0.,0.,0.,.576746649127296,.013033095369547,0.,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,0.,0.,0.,0.,0.,0.,1.80736427869175,.0408421116086564,.000687914444399609,0.,0.,0.,1.41847476579742,.0320541383810507,0.,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,0.,0.,0.,0.,0.,0.,2.81340752869054,.0635762837863472,.00107083209499401,0.,0.,0.,2.75487590413568,.0622536090052649,0.,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,0.,0.,0.,0.,0.,0.,3.86201434387853,.0872722907753235,.00146995018268015,0.,0.,0.,4.43071582310304,.100123584532725,0.,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,0.,0.,0.,0.,0.,0.,4.7630793830775,.107634206373357,.00181291129080671,0.,0.,0.,6.0685481480974,.13713467930593,0.,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,0.,0.,0.,0.,0.,0.,5.33523806427656,.120563624637786,.00203068488848474,0.,0.,0.,7.19422103343416,.16257219522684,0.,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,0.,0.,0.,0.,0.,0.,.000932318622844532,.00010894940458837,9.4896263063424e-006,0.,0.,0.,1.66187980881431e-005,1.9420486861126e-006,0.,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,0.,0.,0.,0.,0.,0.,.0249321870529784,.00291353928576501,.000253772832951388,0.,0.,0.,.00229823078236534,.000268567922156809,0.,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,0.,0.,0.,0.,0.,0.,.140734880491406,.0164460743984419,.00143247318181469,0.,0.,0.,.0308216365902836,.00360177183250731,0.,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,0.,0.,0.,0.,0.,0.,.439603191878621,.0513713926084575,.00447451108643047,0.,0.,0.,.170154848292167,.0198840492440458,0.,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,0.,0.,0.,0.,0.,0.,.991945941760336,.115917366756987,.0100965443280308,0.,0.,0.,.576746649127296,.0673977784805412,0.,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,0.,0.,0.,0.,0.,0.,1.80736427869175,.211205973165025,.0183962984155437,0.,0.,0.,1.41847476579742,.165760907653494,0.,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,0.,0.,0.,0.,0.,0.,2.81340752869054,.328770730954697,.0286363325160947,0.,0.,0.,2.75487590413568,.321930810017305,0.,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,0.,0.,0.,0.,0.,0.,3.86201434387853,.45130940535495,.039309600832947,0.,0.,0.,4.43071582310304,.517767036891475,0.,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,0.,0.,0.,0.,0.,0.,4.7630793830775,.556606561402958,.0484811118273528,0.,0.,0.,6.0685481480974,.709161751356258,0.,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,0.,0.,0.,0.,0.,0.,5.33523806427656,.623468196598569,.0543048419765405,0.,0.,0.,7.19422103343416,.840706255138427,0.,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,0.,0.,0.,0.,0.,0.,.000932318622844532,.000258848437177504,5.35661561215366e-005,0.,0.,0.,1.66187980881431e-005,4.61403409428573e-006,0.,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,0.,0.,0.,0.,0.,0.,.0249321870529784,.00692215890141759,.00143247318181469,0.,0.,0.,.00229823078236534,.000638079548841519,0.,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,0.,0.,0.,0.,0.,0.,.140734880491406,.0390735559485202,.00808589080538567,0.,0.,0.,.0308216365902836,.00855730247849368,0.,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,0.,0.,0.,0.,0.,0.,.439603191878621,.122051191950714,.0252573022041014,0.,0.,0.,.170154848292167,.0472416998608471,0.,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,0.,0.,0.,0.,0.,0.,.991945941760336,.275403333686328,.0569920302764555,0.,0.,0.,.576746649127296,.160127627083755,0.,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,0.,0.,0.,0.,0.,0.,1.80736427869175,.501795638836895,.103841706846431,0.,0.,0.,1.41847476579742,.393824565203833,0.,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,0.,0.,0.,0.,0.,0.,2.81340752869054,.781112941542527,.161643694787028,0.,0.,0.,2.75487590413568,.764862252961431,0.,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,0.,0.,0.,0.,0.,0.,3.86201434387853,1.07224756942001,.221891162762184,0.,0.,0.,4.43071582310304,1.23014153980692,0.,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,0.,0.,0.,0.,0.,0.,4.7630793830775,1.32241877857201,.273661651286941,0.,0.,0.,6.0685481480974,1.68486841886076,0.,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,0.,0.,0.,0.,0.,0.,5.33523806427656,1.48127260473935,.306534899222172,0.,0.,0.,7.19422103343416,1.99739962866362,0.,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,0.,0.,0.,0.,0.,0.,.000932318622844532,.00045748313806202,.000167320660844516,0.,0.,0.,1.66187980881431e-005,8.15474421864669e-006,0.,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,0.,0.,0.,0.,0.,0.,.0249321870529784,.0122340741590526,.00447451108643047,0.,0.,0.,.00229823078236534,.00112772801535997,0.,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,0.,0.,0.,0.,0.,0.,.140734880491406,.0690577589939741,.0252573022041014,0.,0.,0.,.0308216365902836,.0151239916064188,0.,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,0.,0.,0.,0.,0.,0.,.439603191878621,.215710640970697,.0788943766349663,0.,0.,0.,.170154848292167,.083493960154389,0.,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,0.,0.,0.,0.,0.,0.,.991945941760336,.48674190465042,.178021811889788,0.,0.,0.,.576746649127296,.283006110168114,0.,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,0.,0.,0.,0.,0.,0.,1.80736427869175,.886862775854878,.324362699711839,0.,0.,0.,1.41847476579742,.696037031940087,0.,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,0.,0.,0.,0.,0.,0.,2.81340752869054,1.38052214482826,.504914516765946,0.,0.,0.,2.75487590413568,1.35180102876198,0.,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,0.,0.,0.,0.,0.,0.,3.86201434387853,1.89506719911644,.693105099882267,0.,0.,0.,4.43071582310304,2.17412559267414,0.,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,0.,0.,0.,0.,0.,0.,4.7630793830775,2.33721439175002,.854816765967696,0.,0.,0.,6.0685481480974,2.9778000589335,0.,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,0.,0.,0.,0.,0.,0.,5.33523806427656,2.61796921368607,.957500511953663,0.,0.,0.,7.19422103343416,3.53016097005959,0.,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,0.,0.,0.,0.,0.,0.,.000932318622844532,.000687208741781061,.000377551968601727,0.,0.,0.,1.66187980881431e-005,1.22496569770466e-005,0.,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,0.,0.,0.,0.,0.,0.,.0249321870529784,.018377426423439,.0100965443280308,0.,0.,0.,.00229823078236534,.0016940177375493,0.,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.409294114646486,.224865880330915,.113227330974451,0.,0.,0.,0.,0.,0.,.140734880491406,.103735179988284,.0569920302764555,0.,0.,0.,.0308216365902836,.0227185187340072,0.,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.723377579561695,.397423100941195,.200115539631623,0.,0.,0.,0.,0.,0.,.439603191878621,.324029949602564,.178021811889788,0.,0.,0.,.170154848292167,.125420533633389,0.,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,0.,0.,0.,0.,0.,0.,.991945941760336,.731159826532418,.401698661682273,0.,0.,0.,.576746649127296,.425117904255122,0.,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,0.,0.,0.,0.,0.,0.,1.80736427869175,1.33220178323833,.731910663029089,0.,0.,0.,1.41847476579742,1.04555270600537,0.,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,0.,0.,0.,0.,0.,0.,2.81340752869054,2.07375268554642,1.13931817396847,0.,0.,0.,2.75487590413568,2.03060923305097,0.,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,0.,0.,0.,0.,0.,0.,3.86201434387853,2.8466770404089,1.5639622362693,0.,0.,0.,4.43071582310304,3.26586487830957,0.,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,0.,0.,0.,0.,0.,0.,4.7630793830775,3.51084887681561,1.92885774629333,0.,0.,0.,6.0685481480974,4.47310526119946,0.,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,0.,0.,0.,0.,0.,0.,5.33523806427656,3.93258500625844,2.16055926028883,0.,0.,0.,7.19422103343416,5.3028347422728,0.,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,0.,0.,0.,0.,0.,0.,.000932318622844532,.000927614481757561,.000687914444399609,0.,0.,0.,1.66187980881431e-005,1.65349456687952e-005,0.,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.232537853099397,.172448954960703,.117210661865929,0.,0.,0.,0.,0.,0.,.0249321870529784,.0248063882942385,.0183962984155437,0.,0.,0.,.00229823078236534,.00228663474453522,0.,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.552477180456442,.409714423434642,.278476020673683,0.,0.,0.,0.,0.,0.,.140734880491406,.140024783409284,.103841706846431,0.,0.,0.,.0308216365902836,.0306661218062029,0.,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.976436238050563,.724120424237357,.49217250527751,0.,0.,0.,0.,0.,0.,.439603191878621,.437385113867296,.324362699711839,0.,0.,0.,.170154848292167,.169296308726466,0.,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,0.,0.,0.,0.,0.,0.,.991945941760336,.986940942882965,.731910663029089,0.,0.,0.,.576746649127296,.573836595005234,0.,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,0.,0.,0.,0.,0.,0.,1.80736427869175,1.79824497510369,1.33356983668368,0.,0.,0.,1.41847476579742,1.41131765730707,0.,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,0.,0.,0.,0.,0.,0.,2.81340752869054,2.79921209632888,2.07588224620454,0.,0.,0.,2.75487590413568,2.7409758008707,0.,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,0.,0.,0.,0.,0.,0.,3.86201434387853,3.84252802245541,2.84960032604169,0.,0.,0.,4.43071582310304,4.40836004025762,0.,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,0.,0.,0.,0.,0.,0.,4.7630793830775,4.73904661479697,3.51445420820192,0.,0.,0.,6.0685481480974,6.03792845818671,0.,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,0.,0.,0.,0.,0.,0.,5.33523806427656,5.30831839113911,3.93662342336207,0.,0.,0.,7.19422103343416,7.1579216069788,0.,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,0.,0.,0.,0.,0.,0.,.000932318622844532,.0011573400854766,.00107083209499401,0.,0.,0.,1.66187980881431e-005,2.0629858427195e-005,0.,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.290126323030972,.268440177733536,.227639458852635,0.,0.,0.,0.,0.,0.,.0249321870529784,.0309497405586248,.0286363325160947,0.,0.,0.,.00229823078236534,.00285292446672455,0.,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.689299272303129,.637776046087675,.540839285781897,0.,0.,0.,0.,0.,0.,.140734880491406,.174702204403594,.161643694787028,0.,0.,0.,.0308216365902836,.0382606489337913,0.,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,0.,0.,0.,0.,0.,0.,.439603191878621,.545704422499163,.504914516765946,0.,0.,0.,.170154848292167,.211222882205465,0.,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,0.,0.,0.,0.,0.,0.,.991945941760336,1.23135886476496,1.13931817396847,0.,0.,0.,.576746649127296,.715948389092241,0.,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,0.,0.,0.,0.,0.,0.,1.80736427869175,2.24358398248714,2.07588224620454,0.,0.,0.,1.41847476579742,1.76083333137235,0.,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,0.,0.,0.,0.,0.,0.,2.81340752869054,3.49244263704704,3.2313921487783,0.,0.,0.,2.75487590413568,3.41978400515968,0.,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,0.,0.,0.,0.,0.,0.,3.86201434387853,4.79413786374787,4.43578923494485,0.,0.,0.,4.43071582310304,5.50009932589306,0.,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,0.,0.,0.,0.,0.,0.,4.7630793830775,5.91268109986256,5.47072443843503,0.,0.,0.,6.0685481480974,7.53323366045268,0.,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,0.,0.,0.,0.,0.,0.,5.33523806427656,6.62293418371148,6.1278880563708,0.,0.,0.,7.19422103343416,8.93059537919201,0.,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,0.,0.,0.,0.,0.,0.,.000932318622844532,.00135597478636112,.00146995018268015,0.,0.,0.,1.66187980881431e-005,2.4170568551556e-005,0.,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.339920809644861,.368492586412728,.366116582887396,0.,0.,0.,0.,0.,0.,.0249321870529784,.0362616558162599,.039309600832947,0.,0.,0.,.00229823078236534,.003342572933243,0.,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.807603957755601,.875486474339234,.869841424679861,0.,0.,0.,0.,0.,0.,.140734880491406,.204686407449048,.221891162762184,0.,0.,0.,.0308216365902836,.0448273380617164,0.,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,0.,0.,0.,0.,0.,0.,.439603191878621,.639363871519146,.693105099882267,0.,0.,0.,.170154848292167,.247475142499007,0.,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,0.,0.,0.,0.,0.,0.,.991945941760336,1.44269743572906,1.5639622362693,0.,0.,0.,.576746649127296,.838826872176601,0.,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,0.,0.,0.,0.,0.,0.,1.80736427869175,2.62865111950513,2.84960032604169,0.,0.,0.,1.41847476579742,2.0630457981086,0.,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,0.,0.,0.,0.,0.,0.,2.81340752869054,4.09185184033277,4.43578923494485,0.,0.,0.,2.75487590413568,4.00672278096023,0.,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,0.,0.,0.,0.,0.,0.,3.86201434387853,5.6169574934443,6.08908644662382,0.,0.,0.,4.43071582310304,6.44408337876027,0.,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,0.,0.,0.,0.,0.,0.,4.7630793830775,6.92747671304058,7.50976033055426,0.,0.,0.,6.0685481480974,8.82616530052541,0.,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,0.,0.,0.,0.,0.,0.,5.33523806427656,7.7596307926582,8.41186046814944,0.,0.,0.,7.19422103343416,10.463356720588,0.,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,0.,0.,0.,0.,0.,0.,.000932318622844532,.00150587381895025,.00181291129080671,0.,0.,0.,1.66187980881431e-005,2.68425539597291e-005,0.,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.377498057419076,.454467354307313,.50145308337853,0.,0.,0.,0.,0.,0.,.0249321870529784,.0402702754319125,.0484811118273528,0.,0.,0.,.00229823078236534,.00371208455992771,0.,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,0.,0.,0.,0.,0.,0.,.140734880491406,.227313888999126,.273661651286941,0.,0.,0.,.0308216365902836,.0497828687077028,0.,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,0.,0.,0.,0.,0.,0.,.439603191878621,.710043670861402,.854816765967696,0.,0.,0.,.170154848292167,.274832793115808,0.,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,0.,0.,0.,0.,0.,0.,.991945941760336,1.6021834026584,1.92885774629333,0.,0.,0.,.576746649127296,.931556720779814,0.,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,0.,0.,0.,0.,0.,0.,1.80736427869175,2.919240785177,3.51445420820192,0.,0.,0.,1.41847476579742,2.29110945565894,0.,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,0.,0.,0.,0.,0.,0.,2.81340752869054,4.5441940509206,5.47072443843503,0.,0.,0.,2.75487590413568,4.44965422390436,0.,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,0.,0.,0.,0.,0.,0.,3.86201434387853,6.23789565750936,7.50976033055426,0.,0.,0.,4.43071582310304,7.15645788167572,0.,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,0.,0.,0.,0.,0.,0.,4.7630793830775,7.69328893020963,9.26189843365359,0.,0.,0.,6.0685481480974,9.80187196802991,0.,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,0.,0.,0.,0.,0.,0.,5.33523806427656,8.61743520079898,10.3744718692395,0.,0.,0.,7.19422103343416,11.6200500941132,0.,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,0.,0.,0.,0.,0.,0.,.000932318622844532,.00159375505167201,.00203068488848474,0.,0.,0.,1.66187980881431e-005,2.84090575416998e-005,0.,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.399528452143108,.509059650839749,.594469094037471,0.,0.,0.,0.,0.,0.,.0249321870529784,.0426204068987497,.0543048419765405,0.,0.,0.,.00229823078236534,.00392871796107236,0.,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,0.,0.,0.,0.,0.,0.,.140734880491406,.240579691570783,.306534899222172,0.,0.,0.,.0308216365902836,.0526881452424316,0.,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,0.,0.,0.,0.,0.,0.,.439603191878621,.751481082347236,.957500511953663,0.,0.,0.,.170154848292167,.290871749599041,0.,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,0.,0.,0.,0.,0.,0.,.991945941760336,1.6956851626997,2.16055926028883,0.,0.,0.,.576746649127296,.985921403890809,0.,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,0.,0.,0.,0.,0.,0.,1.80736427869175,3.08960464673337,3.93662342336207,0.,0.,0.,1.41847476579742,2.42481622493138,0.,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,0.,0.,0.,0.,0.,0.,2.81340752869054,4.80938849808895,6.1278880563708,0.,0.,0.,2.75487590413568,4.7093314249164,0.,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,0.,0.,0.,0.,0.,0.,3.86201434387853,6.60193277208899,8.41186046814944,0.,0.,0.,4.43071582310304,7.57410133403447,0.,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,0.,0.,0.,0.,0.,0.,4.7630793830775,8.14226128523923,10.3744718692395,0.,0.,0.,6.0685481480974,10.3738990400802,0.,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,0.,0.,0.,0.,0.,0.,5.33523806427656,9.12033977275976,11.620691733626,0.,0.,0.,7.19422103343416,12.2981841540248,0.,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,0.,0.,0.,0.,0.,0.,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,0.,0.,0.,1.66187980881431e-005,3.75545104141921e-007,0.,2.74870870898337e-007},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,0.,0.,0.,0.,0.,0.,.0249321870529784,.00056340781892775,9.4896263063424e-006,0.,0.,0.,.00229823078236534,5.1934521012163e-005,0.,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,0.,0.,0.,0.,0.,0.,.140734880491406,.00318027182678497,5.35661561215366e-005,0.,0.,0.,.0308216365902836,.000696495297778531,0.,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,0.,0.,0.,0.,0.,0.,.439603191878621,.00993398112262371,.000167320660844516,0.,0.,0.,.170154848292167,.00384509276081317,0.,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,0.,0.,0.,0.,0.,0.,.991945941760336,.0224156067156837,.000377551968601727,0.,0.,0.,.576746649127296,.013033095369547,0.,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,0.,0.,0.,0.,0.,0.,1.80736427869175,.0408421116086564,.000687914444399609,0.,0.,0.,1.41847476579742,.0320541383810507,0.,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,0.,0.,0.,0.,0.,0.,2.81340752869054,.0635762837863472,.00107083209499401,0.,0.,0.,2.75487590413568,.0622536090052649,0.,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,0.,0.,0.,0.,0.,0.,3.86201434387853,.0872722907753235,.00146995018268015,0.,0.,0.,4.43071582310304,.100123584532725,0.,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,0.,0.,0.,0.,0.,0.,4.7630793830775,.107634206373357,.00181291129080671,0.,0.,0.,6.0685481480974,.13713467930593,0.,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,0.,0.,0.,0.,0.,0.,5.33523806427656,.120563624637786,.00203068488848474,0.,0.,0.,7.19422103343416,.16257219522684,0.,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,0.,0.,0.,0.,0.,0.,.000932318622844532,.00010894940458837,9.4896263063424e-006,0.,0.,0.,1.66187980881431e-005,1.9420486861126e-006,0.,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,0.,0.,0.,0.,0.,0.,.0249321870529784,.00291353928576501,.000253772832951388,0.,0.,0.,.00229823078236534,.000268567922156809,0.,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,0.,0.,0.,0.,0.,0.,.140734880491406,.0164460743984419,.00143247318181469,0.,0.,0.,.0308216365902836,.00360177183250731,0.,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,0.,0.,0.,0.,0.,0.,.439603191878621,.0513713926084575,.00447451108643047,0.,0.,0.,.170154848292167,.0198840492440458,0.,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,0.,0.,0.,0.,0.,0.,.991945941760336,.115917366756987,.0100965443280308,0.,0.,0.,.576746649127296,.0673977784805412,0.,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,0.,0.,0.,0.,0.,0.,1.80736427869175,.211205973165025,.0183962984155437,0.,0.,0.,1.41847476579742,.165760907653494,0.,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,0.,0.,0.,0.,0.,0.,2.81340752869054,.328770730954697,.0286363325160947,0.,0.,0.,2.75487590413568,.321930810017305,0.,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,0.,0.,0.,0.,0.,0.,3.86201434387853,.45130940535495,.039309600832947,0.,0.,0.,4.43071582310304,.517767036891475,0.,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,0.,0.,0.,0.,0.,0.,4.7630793830775,.556606561402958,.0484811118273528,0.,0.,0.,6.0685481480974,.709161751356258,0.,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,0.,0.,0.,0.,0.,0.,5.33523806427656,.623468196598569,.0543048419765405,0.,0.,0.,7.19422103343416,.840706255138427,0.,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,0.,0.,0.,0.,0.,0.,.000932318622844532,.000258848437177504,5.35661561215366e-005,0.,0.,0.,1.66187980881431e-005,4.61403409428573e-006,0.,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,0.,0.,0.,0.,0.,0.,.0249321870529784,.00692215890141759,.00143247318181469,0.,0.,0.,.00229823078236534,.000638079548841519,0.,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,0.,0.,0.,0.,0.,0.,.140734880491406,.0390735559485202,.00808589080538567,0.,0.,0.,.0308216365902836,.00855730247849368,0.,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,0.,0.,0.,0.,0.,0.,.439603191878621,.122051191950714,.0252573022041014,0.,0.,0.,.170154848292167,.0472416998608471,0.,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,0.,0.,0.,0.,0.,0.,.991945941760336,.275403333686328,.0569920302764555,0.,0.,0.,.576746649127296,.160127627083755,0.,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,0.,0.,0.,0.,0.,0.,1.80736427869175,.501795638836895,.103841706846431,0.,0.,0.,1.41847476579742,.393824565203833,0.,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,0.,0.,0.,0.,0.,0.,2.81340752869054,.781112941542527,.161643694787028,0.,0.,0.,2.75487590413568,.764862252961431,0.,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,0.,0.,0.,0.,0.,0.,3.86201434387853,1.07224756942001,.221891162762184,0.,0.,0.,4.43071582310304,1.23014153980692,0.,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,0.,0.,0.,0.,0.,0.,4.7630793830775,1.32241877857201,.273661651286941,0.,0.,0.,6.0685481480974,1.68486841886076,0.,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,0.,0.,0.,0.,0.,0.,5.33523806427656,1.48127260473935,.306534899222172,0.,0.,0.,7.19422103343416,1.99739962866362,0.,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,0.,0.,0.,0.,0.,0.,.000932318622844532,.00045748313806202,.000167320660844516,0.,0.,0.,1.66187980881431e-005,8.15474421864669e-006,0.,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,0.,0.,0.,0.,0.,0.,.0249321870529784,.0122340741590526,.00447451108643047,0.,0.,0.,.00229823078236534,.00112772801535997,0.,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,0.,0.,0.,0.,0.,0.,.140734880491406,.0690577589939741,.0252573022041014,0.,0.,0.,.0308216365902836,.0151239916064188,0.,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,0.,0.,0.,0.,0.,0.,.439603191878621,.215710640970697,.0788943766349663,0.,0.,0.,.170154848292167,.083493960154389,0.,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,0.,0.,0.,0.,0.,0.,.991945941760336,.48674190465042,.178021811889788,0.,0.,0.,.576746649127296,.283006110168114,0.,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,0.,0.,0.,0.,0.,0.,1.80736427869175,.886862775854878,.324362699711839,0.,0.,0.,1.41847476579742,.696037031940087,0.,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,0.,0.,0.,0.,0.,0.,2.81340752869054,1.38052214482826,.504914516765946,0.,0.,0.,2.75487590413568,1.35180102876198,0.,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,0.,0.,0.,0.,0.,0.,3.86201434387853,1.89506719911644,.693105099882267,0.,0.,0.,4.43071582310304,2.17412559267414,0.,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,0.,0.,0.,0.,0.,0.,4.7630793830775,2.33721439175002,.854816765967696,0.,0.,0.,6.0685481480974,2.9778000589335,0.,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,0.,0.,0.,0.,0.,0.,5.33523806427656,2.61796921368607,.957500511953663,0.,0.,0.,7.19422103343416,3.53016097005959,0.,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,0.,0.,0.,0.,0.,0.,.000932318622844532,.000687208741781061,.000377551968601727,0.,0.,0.,1.66187980881431e-005,1.22496569770466e-005,0.,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,0.,0.,0.,0.,0.,0.,.0249321870529784,.018377426423439,.0100965443280308,0.,0.,0.,.00229823078236534,.0016940177375493,0.,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.409294114646486,.224865880330915,.113227330974451,0.,0.,0.,0.,0.,0.,.140734880491406,.103735179988284,.0569920302764555,0.,0.,0.,.0308216365902836,.0227185187340072,0.,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.723377579561695,.397423100941195,.200115539631623,0.,0.,0.,0.,0.,0.,.439603191878621,.324029949602564,.178021811889788,0.,0.,0.,.170154848292167,.125420533633389,0.,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,0.,0.,0.,0.,0.,0.,.991945941760336,.731159826532418,.401698661682273,0.,0.,0.,.576746649127296,.425117904255122,0.,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,0.,0.,0.,0.,0.,0.,1.80736427869175,1.33220178323833,.731910663029089,0.,0.,0.,1.41847476579742,1.04555270600537,0.,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,0.,0.,0.,0.,0.,0.,2.81340752869054,2.07375268554642,1.13931817396847,0.,0.,0.,2.75487590413568,2.03060923305097,0.,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,0.,0.,0.,0.,0.,0.,3.86201434387853,2.8466770404089,1.5639622362693,0.,0.,0.,4.43071582310304,3.26586487830957,0.,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,0.,0.,0.,0.,0.,0.,4.7630793830775,3.51084887681561,1.92885774629333,0.,0.,0.,6.0685481480974,4.47310526119946,0.,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,0.,0.,0.,0.,0.,0.,5.33523806427656,3.93258500625844,2.16055926028883,0.,0.,0.,7.19422103343416,5.3028347422728,0.,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,0.,0.,0.,0.,0.,0.,.000932318622844532,.000927614481757561,.000687914444399609,0.,0.,0.,1.66187980881431e-005,1.65349456687952e-005,0.,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.232537853099397,.172448954960703,.117210661865929,0.,0.,0.,0.,0.,0.,.0249321870529784,.0248063882942385,.0183962984155437,0.,0.,0.,.00229823078236534,.00228663474453522,0.,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.552477180456442,.409714423434642,.278476020673683,0.,0.,0.,0.,0.,0.,.140734880491406,.140024783409284,.103841706846431,0.,0.,0.,.0308216365902836,.0306661218062029,0.,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.976436238050563,.724120424237357,.49217250527751,0.,0.,0.,0.,0.,0.,.439603191878621,.437385113867296,.324362699711839,0.,0.,0.,.170154848292167,.169296308726466,0.,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,0.,0.,0.,0.,0.,0.,.991945941760336,.986940942882965,.731910663029089,0.,0.,0.,.576746649127296,.573836595005234,0.,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,0.,0.,0.,0.,0.,0.,1.80736427869175,1.79824497510369,1.33356983668368,0.,0.,0.,1.41847476579742,1.41131765730707,0.,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,0.,0.,0.,0.,0.,0.,2.81340752869054,2.79921209632888,2.07588224620454,0.,0.,0.,2.75487590413568,2.7409758008707,0.,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,0.,0.,0.,0.,0.,0.,3.86201434387853,3.84252802245541,2.84960032604169,0.,0.,0.,4.43071582310304,4.40836004025762,0.,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,0.,0.,0.,0.,0.,0.,4.7630793830775,4.73904661479697,3.51445420820192,0.,0.,0.,6.0685481480974,6.03792845818671,0.,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,0.,0.,0.,0.,0.,0.,5.33523806427656,5.30831839113911,3.93662342336207,0.,0.,0.,7.19422103343416,7.1579216069788,0.,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,0.,0.,0.,0.,0.,0.,.000932318622844532,.0011573400854766,.00107083209499401,0.,0.,0.,1.66187980881431e-005,2.0629858427195e-005,0.,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.290126323030972,.268440177733536,.227639458852635,0.,0.,0.,0.,0.,0.,.0249321870529784,.0309497405586248,.0286363325160947,0.,0.,0.,.00229823078236534,.00285292446672455,0.,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.689299272303129,.637776046087675,.540839285781897,0.,0.,0.,0.,0.,0.,.140734880491406,.174702204403594,.161643694787028,0.,0.,0.,.0308216365902836,.0382606489337913,0.,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,0.,0.,0.,0.,0.,0.,.439603191878621,.545704422499163,.504914516765946,0.,0.,0.,.170154848292167,.211222882205465,0.,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,0.,0.,0.,0.,0.,0.,.991945941760336,1.23135886476496,1.13931817396847,0.,0.,0.,.576746649127296,.715948389092241,0.,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,0.,0.,0.,0.,0.,0.,1.80736427869175,2.24358398248714,2.07588224620454,0.,0.,0.,1.41847476579742,1.76083333137235,0.,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,0.,0.,0.,0.,0.,0.,2.81340752869054,3.49244263704704,3.2313921487783,0.,0.,0.,2.75487590413568,3.41978400515968,0.,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,0.,0.,0.,0.,0.,0.,3.86201434387853,4.79413786374787,4.43578923494485,0.,0.,0.,4.43071582310304,5.50009932589306,0.,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,0.,0.,0.,0.,0.,0.,4.7630793830775,5.91268109986256,5.47072443843503,0.,0.,0.,6.0685481480974,7.53323366045268,0.,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,0.,0.,0.,0.,0.,0.,5.33523806427656,6.62293418371148,6.1278880563708,0.,0.,0.,7.19422103343416,8.93059537919201,0.,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,0.,0.,0.,0.,0.,0.,.000932318622844532,.00135597478636112,.00146995018268015,0.,0.,0.,1.66187980881431e-005,2.4170568551556e-005,0.,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.339920809644861,.368492586412728,.366116582887396,0.,0.,0.,0.,0.,0.,.0249321870529784,.0362616558162599,.039309600832947,0.,0.,0.,.00229823078236534,.003342572933243,0.,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.807603957755601,.875486474339234,.869841424679861,0.,0.,0.,0.,0.,0.,.140734880491406,.204686407449048,.221891162762184,0.,0.,0.,.0308216365902836,.0448273380617164,0.,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,0.,0.,0.,0.,0.,0.,.439603191878621,.639363871519146,.693105099882267,0.,0.,0.,.170154848292167,.247475142499007,0.,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,0.,0.,0.,0.,0.,0.,.991945941760336,1.44269743572906,1.5639622362693,0.,0.,0.,.576746649127296,.838826872176601,0.,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,0.,0.,0.,0.,0.,0.,1.80736427869175,2.62865111950513,2.84960032604169,0.,0.,0.,1.41847476579742,2.0630457981086,0.,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,0.,0.,0.,0.,0.,0.,2.81340752869054,4.09185184033277,4.43578923494485,0.,0.,0.,2.75487590413568,4.00672278096023,0.,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,0.,0.,0.,0.,0.,0.,3.86201434387853,5.6169574934443,6.08908644662382,0.,0.,0.,4.43071582310304,6.44408337876027,0.,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,0.,0.,0.,0.,0.,0.,4.7630793830775,6.92747671304058,7.50976033055426,0.,0.,0.,6.0685481480974,8.82616530052541,0.,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,0.,0.,0.,0.,0.,0.,5.33523806427656,7.7596307926582,8.41186046814944,0.,0.,0.,7.19422103343416,10.463356720588,0.,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,0.,0.,0.,0.,0.,0.,.000932318622844532,.00150587381895025,.00181291129080671,0.,0.,0.,1.66187980881431e-005,2.68425539597291e-005,0.,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.377498057419076,.454467354307313,.50145308337853,0.,0.,0.,0.,0.,0.,.0249321870529784,.0402702754319125,.0484811118273528,0.,0.,0.,.00229823078236534,.00371208455992771,0.,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,0.,0.,0.,0.,0.,0.,.140734880491406,.227313888999126,.273661651286941,0.,0.,0.,.0308216365902836,.0497828687077028,0.,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,0.,0.,0.,0.,0.,0.,.439603191878621,.710043670861402,.854816765967696,0.,0.,0.,.170154848292167,.274832793115808,0.,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,0.,0.,0.,0.,0.,0.,.991945941760336,1.6021834026584,1.92885774629333,0.,0.,0.,.576746649127296,.931556720779814,0.,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,0.,0.,0.,0.,0.,0.,1.80736427869175,2.919240785177,3.51445420820192,0.,0.,0.,1.41847476579742,2.29110945565894,0.,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,0.,0.,0.,0.,0.,0.,2.81340752869054,4.5441940509206,5.47072443843503,0.,0.,0.,2.75487590413568,4.44965422390436,0.,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,0.,0.,0.,0.,0.,0.,3.86201434387853,6.23789565750936,7.50976033055426,0.,0.,0.,4.43071582310304,7.15645788167572,0.,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,0.,0.,0.,0.,0.,0.,4.7630793830775,7.69328893020963,9.26189843365359,0.,0.,0.,6.0685481480974,9.80187196802991,0.,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,0.,0.,0.,0.,0.,0.,5.33523806427656,8.61743520079898,10.3744718692395,0.,0.,0.,7.19422103343416,11.6200500941132,0.,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,0.,0.,0.,0.,0.,0.,.000932318622844532,.00159375505167201,.00203068488848474,0.,0.,0.,1.66187980881431e-005,2.84090575416998e-005,0.,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.399528452143108,.509059650839749,.594469094037471,0.,0.,0.,0.,0.,0.,.0249321870529784,.0426204068987497,.0543048419765405,0.,0.,0.,.00229823078236534,.00392871796107236,0.,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,0.,0.,0.,0.,0.,0.,.140734880491406,.240579691570783,.306534899222172,0.,0.,0.,.0308216365902836,.0526881452424316,0.,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,0.,0.,0.,0.,0.,0.,.439603191878621,.751481082347236,.957500511953663,0.,0.,0.,.170154848292167,.290871749599041,0.,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,0.,0.,0.,0.,0.,0.,.991945941760336,1.6956851626997,2.16055926028883,0.,0.,0.,.576746649127296,.985921403890809,0.,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,0.,0.,0.,0.,0.,0.,1.80736427869175,3.08960464673337,3.93662342336207,0.,0.,0.,1.41847476579742,2.42481622493138,0.,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,0.,0.,0.,0.,0.,0.,2.81340752869054,4.80938849808895,6.1278880563708,0.,0.,0.,2.75487590413568,4.7093314249164,0.,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,0.,0.,0.,0.,0.,0.,3.86201434387853,6.60193277208899,8.41186046814944,0.,0.,0.,4.43071582310304,7.57410133403447,0.,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,0.,0.,0.,0.,0.,0.,4.7630793830775,8.14226128523923,10.3744718692395,0.,0.,0.,6.0685481480974,10.3738990400802,0.,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,0.,0.,0.,0.,0.,0.,5.33523806427656,9.12033977275976,11.620691733626,0.,0.,0.,7.19422103343416,12.2981841540248,0.,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{-5.02239597033184e-005,-.000112582670872584,-.000165039185570316,-.000202840374019418,-.000222620323957012,-.000222620323957012,-.000202840374019418,-.000165039185570316,-.000112582670872584,-5.02239597033184e-005,-.000582196454262715,-.00130505902323575,-.00191313526892416,-.00235132688129378,-.00258061618439108,-.00258061618439108,-.00235132688129378,-.00191313526892416,-.00130505902323575,-.000582196454262715,-.00202770964317194,-.00454533988819102,-.00666318526175589,-.00818934597856097,-.00898792887538633,-.00898792887538633,-.00818934597856097,-.0066631852617559,-.00454533988819102,-.00202770964317194,-.00440456112754404,-.0098733205961796,-.0144736732344206,-.0177887771450126,-.0195234471932234,-.0195234471932234,-.0177887771450126,-.0144736732344206,-.00987332059617959,-.00440456112754404,-.00726150487522216,-.016277482266154,-.0238617754892464,-.0293271652321865,-.0321869995374054,-.0321869995374054,-.0293271652321865,-.0238617754892464,-.016277482266154,-.00726150487522216,-.0098017919040896,-.0219718221824466,-.03220932325003,-.0395866663566687,-.0434469544404215,-.0434469544404215,-.0395866663566687,-.03220932325003,-.0219718221824466,-.0098017919040896,-.0111426514503529,-.0249775101026156,-.0366154745926699,-.0450020189787675,-.0493903843956318,-.0493903843956318,-.0450020189787675,-.0366154745926699,-.0249775101026156,-.0111426514503529,-.0106221354093669,-.0238107146832105,-.0349050251580164,-.0428998018485296,-.0470831698638901,-.0470831698638901,-.0428998018485296,-.0349050251580163,-.0238107146832105,-.0106221354093669,-.00804698659228353,-.0180382281362673,-.0264429193024774,-.0324995038175014,-.0356686882642096,-.0356686882642096,-.0324995038175014,-.0264429193024774,-.0180382281362673,-.00804698659228353,-.00379931823220995,-.00851660037567366,-.0124848058669686,-.0153443722038775,-.0168406764553548,-.0168406764553548,-.0153443722038775,-.0124848058669686,-.00851660037567366,-.00379931823220995},
			{-8.45935382972169e-007,-1.89625958134751e-006,-2.77979847617661e-006,-3.41649385058338e-006,-3.74965275769605e-006,-3.74965275769605e-006,-3.41649385058338e-006,-2.77979847617661e-006,-1.89625958134751e-006,-8.45935382972169e-007,-5.07100227590783e-005,-.000113672236039353,-.000166636420263325,-.00020480344528281,-.000224774823832688,-.000224774823832688,-.00020480344528281,-.000166636420263325,-.000113672236039353,-5.07100227590783e-005,-.000419614707621413,-.000940613699128858,-.00137888111587085,-.00169470517141149,-.00185996410278327,-.00185996410278327,-.00169470517141149,-.00137888111587085,-.000940613699128858,-.000419614707621413,-.00161093167609136,-.00361108506296509,-.00529362585909322,-.006506097600196,-.00714053877317793,-.00714053877317793,-.006506097600196,-.00529362585909322,-.00361108506296509,-.00161093167609136,-.00398946534464676,-.00894283657654873,-.0131096415979721,-.0161123226329801,-.017683513460247,-.017683513460247,-.0161123226329801,-.0131096415979721,-.00894283657654873,-.00398946534464676,-.00726896179664528,-.016294197796655,-.023886279416173,-.0293572816296078,-.0322200527309955,-.0322200527309955,-.0293572816296078,-.023886279416173,-.016294197796655,-.00726896179664528,-.0103097688766848,-.0231105098656633,-.0338785684935517,-.0416382417346981,-.045698588896762,-.045698588896762,-.0416382417346981,-.0338785684935517,-.0231105098656633,-.0103097688766848,-.0115149706602782,-.0258121056087878,-.0378389396388598,-.0465057110062776,-.0510407087352223,-.0510407087352223,-.0465057110062776,-.0378389396388598,-.0258121056087878,-.0115149706602782,-.00968771265055556,-.021716100667658,-.0318344514317794,-.0391259325039141,-.0429412921921929,-.0429412921921929,-.0391259325039141,-.0318344514317794,-.021716100667658,-.00968771265055556,-.00484090582829009,-.0108514364620189,-.0159075301916269,-.0195510500287827,-.0214575678641308,-.0214575678641308,-.0195510500287827,-.0159075301916269,-.0108514364620189,-.00484090582829009},
			{-1.30587941023263e-008,-2.92727599954231e-008,-4.2912043492978e-008,-5.27407774218851e-008,-5.7883786756776e-008,-5.7883786756776e-008,-5.27407774218851e-008,-4.2912043492978e-008,-2.92727599954231e-008,-1.30587941023263e-008,-4.04816026703356e-006,-9.07440786578996e-006,-1.33025169157498e-005,-1.63493748303836e-005,-1.79436817686314e-005,-1.79436817686314e-005,-1.63493748303836e-005,-1.33025169157498e-005,-9.07440786578996e-006,-4.04816026703356e-006,-7.95857439793226e-005,-.000178400422298479,-.000261523911036573,-.00032142431960244,-.000352767965960796,-.000352767965960796,-.00032142431960244,-.000261523911036573,-.000178400422298479,-7.95857439793226e-005,-.000539996871821585,-.00121046389913573,-.00177446470693278,-.00218089469839968,-.00239356433166203,-.00239356433166203,-.00218089469839968,-.00177446470693278,-.00121046389913573,-.000539996871821585,-.00200882638285304,-.0045030109301789,-.00660113364491101,-.00811308182877234,-.00890422783798524,-.00890422783798524,-.00811308182877234,-.00660113364491101,-.0045030109301789,-.00200882638285304,-.00494059139703164,-.0110748929087561,-.0162351034290897,-.0199536518580156,-.0218994293529137,-.0218994293529137,-.0199536518580156,-.0162351034290897,-.0110748929087561,-.00494059139703164,-.00874276804581894,-.0195979007476136,-.0287293022381269,-.0353095683983229,-.0387527758890071,-.0387527758890071,-.0353095683983229,-.0287293022381268,-.0195979007476136,-.00874276804581894,-.0114407232759916,-.0256456717219007,-.0375949579236433,-.0462058468122232,-.0507116033273556,-.0507116033273556,-.0462058468122232,-.0375949579236432,-.0256456717219007,-.0114407232759916,-.0106892900748741,-.0239612494409041,-.0351256997397901,-.0431710205566739,-.0473808364254027,-.0473808364254027,-.0431710205566739,-.0351256997397901,-.0239612494409041,-.0106892900748741,-.00565310744519306,-.0126720780015972,-.0185764773269063,-.0228313027353587,-.0250576939422306,-.0250576939422306,-.0228313027353587,-.0185764773269063,-.0126720780015972,-.00565310744519306},
			{-1.93186678290668e-010,-4.33049730595595e-010,-6.3482394133133e-010,-7.80226376246059e-010,-8.56310039258135e-010,-8.56310039258135e-010,-7.80226376246059e-010,-6.3482394133133e-010,-4.33049730595595e-010,-1.93186678290668e-010,-3.09691872038552e-007,-6.94209264016414e-007,-1.01766755630007e-006,-1.25075791566716e-006,-1.37272539416124e-006,-1.37272539416124e-006,-1.25075791566716e-006,-1.01766755630007e-006,-6.94209264016414e-007,-3.09691872038552e-007,-1.44653210137677e-005,-3.24256487218327e-005,-4.75339820521491e-005,-5.84213419665859e-005,-6.41182907371273e-005,-6.41182907371273e-005,-5.84213419665859e-005,-4.75339820521491e-005,-3.24256487218327e-005,-1.44653210137677e-005,-.000173465688278746,-.000388842906981744,-.000570019490439081,-.000700578873068661,-.000768895721248536,-.000768895721248536,-.000700578873068661,-.000570019490439081,-.000388842906981744,-.000173465688278746,-.000969344894307206,-.00217289592143802,-.00318533012606401,-.00391490997676346,-.00429667186083074,-.00429667186083074,-.00391490997676346,-.00318533012606401,-.00217289592143802,-.000969344894307206,-.00321805680650706,-.00721363732442475,-.010574743162472,-.0129968216385873,-.01426420483388,-.01426420483388,-.0129968216385873,-.010574743162472,-.00721363732442475,-.00321805680650706,-.00710488740889465,-.0159264065491336,-.023347120347727,-.0286946315642816,-.0314927844397286,-.0314927844397286,-.0286946315642816,-.023347120347727,-.0159264065491336,-.00710488740889465,-.0108931213951503,-.0244181603372121,-.0357955026643469,-.0439942376324665,-.0482843293960987,-.0482843293960987,-.0439942376324665,-.0357955026643469,-.0244181603372121,-.0108931213951503,-.0113027648830266,-.0253364224225416,-.0371416177061049,-.0456486718663717,-.050100095546689,-.050100095546689,-.0456486718663717,-.0371416177061049,-.0253364224225416,-.0113027648830266,-.00632639147606921,-.0141813201023726,-.0207889322741007,-.0255505065864879,-.0280420605663306,-.0280420605663306,-.0255505065864879,-.0207889322741007,-.0141813201023726,-.00632639147606921},
			{-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262448,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-.000233400654593986,-.00270558184440079,-.0094231669672911,-.0204688649887572,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813583,-.0373959351122988,-.017656181783494,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000233400654593986,-.00270558184440079,-.00942316696729111,-.0204688649887571,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813582,-.0373959351122988,-.017656181783494,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262447},
			{-1.60504953819679e-006,-1.86057442617858e-005,-6.4801231236956e-005,-.000140760283448081,-.000232062049974506,-.000313244150044994,-.000356095132090808,-.000339460560939038,-.000257164353231333,-.000121418024586148,-1.86057442617858e-005,-.000215677903576663,-.000751176276838271,-.00163169408403043,-.0026900647312993,-.00363112814186739,-.00412785699311923,-.00393502894053338,-.00298105078728298,-.00140747849861324,-6.4801231236956e-005,-.000751176276838271,-.00261624296938716,-.00568296458123414,-.00936912301075426,-.0126467165764205,-.014376754391578,-.0137051609821032,-.0103825871557633,-.00490205274061561,-.000140760283448081,-.00163169408403043,-.00568296458123414,-.0123444522582422,-.0203514714995352,-.0274710121398013,-.0312289748912649,-.0297701495437364,-.0225529034413806,-.0106481670189776,-.000232062049974506,-.0026900647312993,-.00936912301075426,-.0203514714995352,-.033552107742962,-.0452896174679006,-.0514851189152542,-.0490800512895761,-.0371814611144508,-.0175549196574778,-.000313244150044994,-.00363112814186739,-.0126467165764205,-.0274710121398013,-.0452896174679006,-.0611332517796597,-.0694961210432554,-.0662496903395319,-.0501886249195276,-.0236961445777924,-.000356095132090808,-.00412785699311923,-.014376754391578,-.0312289748912649,-.0514851189152542,-.0694961210432554,-.0790030089920025,-.075312475042366,-.0570542977980853,-.0269377153005406,-.000339460560939038,-.00393502894053338,-.0137051609821032,-.0297701495437364,-.0490800512895761,-.0662496903395319,-.075312475042366,-.0717943401064784,-.0543890724391648,-.0256793511684561,-.000257164353231333,-.00298105078728298,-.0103825871557633,-.0225529034413806,-.0371814611144508,-.0501886249195276,-.0570542977980853,-.0543890724391648,-.0412034040065755,-.0194538467631362,-.000121418024586148,-.00140747849861324,-.00490205274061561,-.0106481670189776,-.0175549196574778,-.0236961445777924,-.0269377153005406,-.0256793511684561,-.0194538467631362,-.00918497301395746},
			{-2.70342721665942e-008,-3.13381451640588e-007,-1.09146420736547e-006,-2.3708625325406e-006,-3.90868223643537e-006,-5.27605373253693e-006,-5.99780411074261e-006,-5.71762364703134e-006,-4.33148694252304e-006,-2.04507577148045e-006,-1.62058306631703e-006,-1.878580901668e-005,-6.54283718476915e-005,-.000142122549078592,-.000234307926063146,-.00031627569935075,-.000359541389428153,-.000342745830361303,-.000259653167200857,-.000122593097538308,-1.34099819434037e-005,-.000155448594362042,-.000541405931790853,-.00117603401917912,-.00193884850644768,-.00261711448526345,-.00297512891523367,-.00283614921805116,-.00214857501357473,-.00101443194029589,-5.14819054148423e-005,-.000596778568799888,-.00207849712916292,-.00451488095923928,-.00744338179155761,-.0100472946912819,-.0114217383779821,-.0108881851148137,-.00824853725324792,-.00389447871149425,-.00012749471661469,-.00147791955052476,-.00514738917162678,-.0111810832137794,-.0184335028884922,-.0248820819487376,-.0282858858081064,-.0269645434541597,-.0204274669151108,-.00964465972428348,-.000232300357136792,-.00269282718941148,-.00937874426988752,-.0203723706574197,-.033586562784559,-.0453361258919233,-.0515379895702794,-.0491304521532631,-.0372196431803629,-.0175729470044231,-.00032947799961708,-.00381931102740118,-.0133021315121917,-.0288946948441916,-.0476367477719961,-.0643014769902532,-.0730977511924463,-.0696830745129148,-.0527896458390131,-.0249241951142798,-.000367993651863299,-.004265784708565,-.0148571375279271,-.0322724560897768,-.0532054364658069,-.0718182560454652,-.0816428060009037,-.0778289569952232,-.0589607032197774,-.0278378088689286,-.000309598422927644,-.00358886685037487,-.0124995263493677,-.0271512876888108,-.0447625091834733,-.0604217455831262,-.0686872826563242,-.065478635899881,-.049604498988284,-.0234203543456317,-.000154704919937309,-.00179334039722375,-.00624595630961966,-.0135673746279838,-.0223676216885663,-.0301924706996799,-.0343227218781454,-.0327193757277722,-.0247871419109467,-.0117030442522349},
			{-4.17330922690109e-010,-4.83770265983857e-009,-1.68490485682806e-008,-3.65992560176709e-008,-6.03387416602869e-008,-8.14470002666849e-008,-9.2588737297087e-008,-8.82635618042971e-008,-6.6865622688263e-008,-3.15700512824513e-008,-1.29370479862118e-007,-1.49966340979313e-006,-5.22312002294925e-006,-1.13455846575704e-005,-1.8704705399135e-005,-2.52481590386612e-005,-2.87020412883381e-005,-2.7361258713725e-005,-2.07280055781616e-005,-9.78655657087546e-006,-2.54338890004804e-006,-2.94829799993105e-005,-.000102685137321483,-.000223051148247843,-.000367729486213021,-.000496372027946554,-.000564274425659415,-.000537915000222613,-.000407507024506811,-.000192401074639172,-1.72571365319964e-005,-.000200044834358053,-.000696728461200002,-.00151342333799558,-.00249507967511344,-.00336793160368666,-.00382865585555338,-.00364980463714188,-.00276497407043368,-.00130545966206281,-6.41977629259773e-005,-.000744180868411951,-.00259187893037573,-.0056300413732684,-.00928187206303727,-.0125289426924045,-.0142428693476609,-.0135775302227536,-.0102858982161417,-.00485640185665492,-.00015789065592134,-.00183026946862863,-.00637457515237077,-.0138467585905541,-.0228282233120914,-.0308142042579516,-.0350295069642396,-.0333931441681891,-.025297567112414,-.0119440372937759,-.000279400029347063,-.00323880688356021,-.0112803159519127,-.0245029367570041,-.0403963504117509,-.054528176627927,-.0619874888523968,-.0590918151941233,-.0447660499753643,-.0211359206213315,-.000365620865418288,-.00423827935328899,-.0147613401837082,-.0320643665046187,-.0528623731064046,-.0713551790777345,-.0811163813127856,-.0773271236259448,-.0585805304731015,-.0276583134477209,-.000341606679368254,-.00395990675875596,-.0137918069785563,-.0299583607055857,-.0493903424240987,-.0666685303985192,-.0757886113280989,-.0722482342377723,-.0547329279680274,-.0258416997153403,-.000180661133614816,-.00209422498813209,-.00729389568128642,-.0158436931512049,-.0261204355502155,-.0352581287359617,-.0400813487105939,-.0382089949857042,-.0289459000949523,-.0136665675675336},
			{-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779044,-.0447078783667129,-.0376133897003122,-.0187952392881457,-4.81475219575139e-006,-.000288622746300124,-.00238829215039021,-.00916882894461163,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619087,-.0551388873133415,-.0275526505160409,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550778,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550777,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-4.81475219575139e-006,-.000288622746300124,-.00238829215039022,-.00916882894461162,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619086,-.0551388873133415,-.0275526505160409,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779045,-.044707878366713,-.0376133897003122,-.0187952392881457,-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211},
			{-3.31100861878508e-008,-1.98480079913645e-006,-1.64238066106426e-005,-6.30521996263144e-005,-.000156148500303414,-.000284508671025808,-.000403526490267504,-.000450698337824376,-.000379179080671673,-.000189474057272324,-3.8381232568618e-007,-2.30078232481278e-005,-.000190384868710001,-.000730901491494302,-.00181007438983512,-.00329802628977656,-.00467768159306037,-.00522449794427726,-.00439544626910526,-.00219638445415982,-1.33676519027876e-006,-8.01330628639868e-005,-.000663084138302037,-.00254562869914522,-.00630423848900844,-.0114865689446415,-.0162917173481382,-.0181962029908938,-.0153087307912168,-.00764970295714498,-2.90370172750447e-006,-.000174063863093159,-.0014403416335721,-.00552957729977356,-.0136939743226831,-.0249509564807709,-.0353886293208596,-.0395255250831765,-.0332534003485589,-.0166165724938767,-4.7871385229751e-006,-.000286967430772327,-.00237459476467777,-.00911624367502254,-.0225763381244699,-.0411349705180729,-.0583428625235449,-.0651630854416681,-.0548226535530947,-.0273946299483073,-6.46181975011305e-006,-.000387357040725739,-.00320529754367204,-.0123053726445113,-.030474202256271,-.0555251876749137,-.0787529041667412,-.0879590407639987,-.0740012230234841,-.0369780736440852,-7.34577982424567e-006,-.000440346472755273,-.00364377388066238,-.0139887155008144,-.0346429935762571,-.0631208884080524,-.0895260958832347,-.0999916079356561,-.0841243971631835,-.0420365775924714,-7.00263023825107e-006,-.000419776197925987,-.0034735592093105,-.0133352488781344,-.033024686304908,-.0601722693038774,-.0853439881325168,-.0953206159256855,-.0801946235041213,-.0400728876177358,-5.30496641835636e-006,-.00031800888486993,-.00263145622867657,-.010102353697401,-.0250184353398111,-.0455845671001885,-.0646538480039295,-.0722118188820862,-.0607528557338689,-.0303579249318969,-2.50469606272866e-006,-.000150145267478098,-.0012424203162536,-.00476974282864777,-.0118122475336369,-.02152237671891,-.030525780139787,-.0340942136430102,-.028683958870995,-.014333243427598},
			{-5.57681903477072e-010,-3.34305226934559e-008,-2.76630501080463e-007,-1.06200480743303e-006,-2.63005032304177e-006,-4.79205449098542e-006,-6.79670297198979e-006,-7.59123082633588e-006,-6.3866131386048e-006,-3.19136145764326e-006,-3.34305226934559e-008,-2.00400952691774e-006,-1.65827547682883e-005,-6.36624132755433e-005,-.000157659691772648,-.000287262120951629,-.00040743178419295,-.000455060156746716,-.000382848742505824,-.000191307770554455,-2.76630501080463e-007,-1.65827547682883e-005,-.000137218786643261,-.000526792998299499,-.0013046005871693,-.00237703326355233,-.00337141179786177,-.00376552650213992,-.00316799232989985,-.00158303131884398,-1.06200480743303e-006,-6.36624132755433e-005,-.000526792998299499,-.00202239700441925,-.00500845745477193,-.00912560525126807,-.0129430974646002,-.0144561327553171,-.0121621551894091,-.00607737347965237,-2.63005032304177e-006,-.000157659691772648,-.0013046005871693,-.00500845745477193,-.0124034232751763,-.0225995220276466,-.0320535250214294,-.0358005503901207,-.030119524846696,-.0150505891984046,-4.79205449098542e-006,-.000287262120951629,-.00237703326355233,-.00912560525126807,-.0225995220276466,-.0411772124958644,-.0584027754849978,-.0652300021690509,-.0548789515711727,-.0274227618111799,-6.79670297198979e-006,-.00040743178419295,-.00337141179786177,-.0129430974646002,-.0320535250214294,-.0584027754849978,-.0828342662751547,-.0925175100657338,-.0778363296880565,-.0388944589534263,-7.59123082633588e-006,-.000455060156746716,-.00376552650213992,-.0144561327553171,-.0358005503901207,-.0652300021690509,-.0925175100657338,-.103332715477079,-.0869353196354016,-.0434411827319366,-6.3866131386048e-006,-.000382848742505824,-.00316799232989985,-.0121621551894091,-.030119524846696,-.0548789515711727,-.0778363296880565,-.0869353196354016,-.0731399515169607,-.0365477002003157,-3.19136145764326e-006,-.000191307770554455,-.00158303131884398,-.00607737347965237,-.0150505891984046,-.0274227618111799,-.0388944589534263,-.0434411827319366,-.0365477002003157,-.0182627191053361},
			{-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596958,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596959,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861},
			{-5.90195050823199e-010,-1.82957487178445e-007,-3.59689507657609e-006,-2.44052765295024e-005,-9.07893469973411e-005,-.000223291106959738,-.000395131310801382,-.000517065986523608,-.000483104798924764,-.000255493425333229,-6.8415447117756e-009,-2.12084433297021e-006,-4.16952301711737e-005,-.000282906117811307,-.00105243067689039,-.00258839190514417,-.00458036462030578,-.00599383214211218,-.00560015384355916,-.0029616813806621,-2.38281529966161e-008,-7.38660717382125e-006,-.000145218713843646,-.000985322839048874,-.00366547033510042,-.00901501063419505,-.0159527758058908,-.0208756874850881,-.0195045604772907,-.0103151261942612,-5.17591642291992e-008,-1.60450796946238e-005,-.0003154419589321,-.00214030381004998,-.00796208086622,-.0195822737922466,-.0346523854772107,-.0453458619764426,-.0423675200132299,-.0224063657308857,-8.53318667863116e-008,-2.64524480537304e-005,-.000520048026649204,-.00352857551549082,-.0131265493548066,-.0322839830104985,-.0571290666185282,-.0747586849808788,-.0698484921013427,-.0369398742075263,-1.15183452383392e-007,-3.57062889348363e-005,-.000701976053853711,-.00476297455073299,-.0177186006765078,-.0435778655721675,-.0771144869130937,-.10091146198997,-.0942835398662266,-.049862523838675,-1.30940247999038e-007,-4.05908160548146e-005,-.000798004545609922,-.00541453703618982,-.0201424589971083,-.0495392038284753,-.0876635474261488,-.114715886574844,-.107181282005842,-.0566835869406137,-1.24823526157934e-007,-3.86946631537408e-005,-.000760726688663539,-.00516160321748368,-.0192015273831553,-.0472250373695059,-.0835684464667091,-.109357066963176,-.102174432709151,-.0540356789095071,-9.45622704554012e-008,-2.9313826607454e-005,-.000576301960778015,-.00391026382973348,-.0145464567574012,-.0357761625028231,-.0633087550044376,-.0828453806800583,-.0774040490351568,-.0409356844864075,-4.46467946852167e-008,-1.38402810304592e-005,-.000272096209150128,-.00184619875908625,-.00686798936951685,-.0168914195291221,-.029890705593759,-.0391147619872944,-.0365456822095551,-.0193274452036918},
			{-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807,-9.68328634942698e-010,-1.55229910445862e-006,-7.2505954733558e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039878,-.0236458445483912,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.0641631467467009,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.064163146746701,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039877,-.0236458445483913,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-9.68328634942698e-010,-1.55229910445862e-006,-7.25059547335579e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807},
			{-.00537304737262447,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.017656181783494,-.0373959351122988,-.0493631599813582,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887572,-.00942316696729111,-.00270558184440079,-.000233400654593986,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.017656181783494,-.0373959351122988,-.0493631599813583,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887571,-.0094231669672911,-.00270558184440079,-.000233400654593986,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.00537304737262448,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005},
			{-.000121418024586148,-.000257164353231333,-.000339460560939038,-.000356095132090808,-.000313244150044994,-.000232062049974506,-.000140760283448081,-6.4801231236956e-005,-1.86057442617858e-005,-1.60504953819679e-006,-.00140747849861324,-.00298105078728298,-.00393502894053338,-.00412785699311923,-.00363112814186739,-.0026900647312993,-.00163169408403043,-.000751176276838271,-.000215677903576663,-1.86057442617858e-005,-.00490205274061561,-.0103825871557633,-.0137051609821032,-.014376754391578,-.0126467165764205,-.00936912301075426,-.00568296458123414,-.00261624296938716,-.000751176276838271,-6.4801231236956e-005,-.0106481670189775,-.0225529034413806,-.0297701495437364,-.031228974891265,-.0274710121398013,-.0203514714995352,-.0123444522582422,-.00568296458123414,-.00163169408403043,-.000140760283448081,-.0175549196574778,-.0371814611144508,-.0490800512895761,-.0514851189152542,-.0452896174679006,-.033552107742962,-.0203514714995352,-.00936912301075425,-.0026900647312993,-.000232062049974506,-.0236961445777924,-.0501886249195276,-.0662496903395319,-.0694961210432554,-.0611332517796597,-.0452896174679006,-.0274710121398013,-.0126467165764205,-.00363112814186739,-.000313244150044994,-.0269377153005406,-.0570542977980853,-.075312475042366,-.0790030089920025,-.0694961210432554,-.0514851189152542,-.031228974891265,-.014376754391578,-.00412785699311923,-.000356095132090808,-.0256793511684561,-.0543890724391648,-.0717943401064784,-.075312475042366,-.0662496903395319,-.0490800512895761,-.0297701495437364,-.0137051609821032,-.00393502894053338,-.000339460560939038,-.0194538467631362,-.0412034040065755,-.0543890724391648,-.0570542977980853,-.0501886249195276,-.0371814611144508,-.0225529034413806,-.0103825871557633,-.00298105078728298,-.000257164353231333,-.00918497301395746,-.0194538467631362,-.0256793511684561,-.0269377153005406,-.0236961445777924,-.0175549196574778,-.0106481670189776,-.00490205274061561,-.00140747849861324,-.000121418024586148},
			{-2.04507577148045e-006,-4.33148694252304e-006,-5.71762364703134e-006,-5.99780411074261e-006,-5.27605373253693e-006,-3.90868223643537e-006,-2.3708625325406e-006,-1.09146420736547e-006,-3.13381451640588e-007,-2.70342721665942e-008,-.000122593097538308,-.000259653167200857,-.000342745830361303,-.000359541389428153,-.00031627569935075,-.000234307926063146,-.000142122549078592,-6.54283718476915e-005,-1.878580901668e-005,-1.62058306631703e-006,-.00101443194029589,-.00214857501357473,-.00283614921805116,-.00297512891523367,-.00261711448526345,-.00193884850644768,-.00117603401917912,-.000541405931790853,-.000155448594362042,-1.34099819434037e-005,-.00389447871149424,-.00824853725324792,-.0108881851148137,-.0114217383779821,-.0100472946912819,-.00744338179155761,-.00451488095923928,-.00207849712916292,-.000596778568799888,-5.14819054148423e-005,-.00964465972428348,-.0204274669151108,-.0269645434541598,-.0282858858081064,-.0248820819487376,-.0184335028884922,-.0111810832137794,-.00514738917162678,-.00147791955052476,-.00012749471661469,-.0175729470044231,-.0372196431803629,-.0491304521532631,-.0515379895702794,-.0453361258919233,-.033586562784559,-.0203723706574197,-.00937874426988752,-.00269282718941148,-.000232300357136792,-.0249241951142798,-.0527896458390131,-.0696830745129148,-.0730977511924463,-.0643014769902532,-.0476367477719961,-.0288946948441916,-.0133021315121917,-.00381931102740118,-.000329477999617081,-.0278378088689286,-.0589607032197774,-.0778289569952232,-.0816428060009037,-.0718182560454652,-.0532054364658069,-.0322724560897768,-.0148571375279271,-.004265784708565,-.000367993651863299,-.0234203543456317,-.049604498988284,-.065478635899881,-.0686872826563242,-.0604217455831262,-.0447625091834733,-.0271512876888108,-.0124995263493677,-.00358886685037487,-.000309598422927644,-.0117030442522349,-.0247871419109467,-.0327193757277722,-.0343227218781454,-.0301924706996799,-.0223676216885663,-.0135673746279838,-.00624595630961966,-.00179334039722375,-.000154704919937309},
			{-3.15700512824513e-008,-6.6865622688263e-008,-8.82635618042971e-008,-9.2588737297087e-008,-8.14470002666849e-008,-6.03387416602869e-008,-3.65992560176709e-008,-1.68490485682806e-008,-4.83770265983857e-009,-4.17330922690109e-010,-9.78655657087546e-006,-2.07280055781616e-005,-2.7361258713725e-005,-2.87020412883381e-005,-2.52481590386612e-005,-1.8704705399135e-005,-1.13455846575704e-005,-5.22312002294925e-006,-1.49966340979313e-006,-1.29370479862118e-007,-.000192401074639172,-.000407507024506811,-.000537915000222613,-.000564274425659415,-.000496372027946554,-.000367729486213021,-.000223051148247843,-.000102685137321483,-2.94829799993105e-005,-2.54338890004804e-006,-.00130545966206281,-.00276497407043368,-.00364980463714188,-.00382865585555338,-.00336793160368666,-.00249507967511344,-.00151342333799558,-.000696728461200002,-.000200044834358053,-1.72571365319964e-005,-.00485640185665492,-.0102858982161417,-.0135775302227536,-.0142428693476609,-.0125289426924045,-.00928187206303727,-.0056300413732684,-.00259187893037573,-.000744180868411951,-6.41977629259773e-005,-.0119440372937759,-.025297567112414,-.0333931441681891,-.0350295069642396,-.0308142042579516,-.0228282233120914,-.0138467585905541,-.00637457515237077,-.00183026946862863,-.00015789065592134,-.0211359206213315,-.0447660499753643,-.0590918151941233,-.0619874888523968,-.054528176627927,-.0403963504117509,-.0245029367570041,-.0112803159519127,-.00323880688356021,-.000279400029347063,-.0276583134477209,-.0585805304731015,-.0773271236259449,-.0811163813127856,-.0713551790777345,-.0528623731064046,-.0320643665046187,-.0147613401837082,-.00423827935328899,-.000365620865418288,-.0258416997153403,-.0547329279680274,-.0722482342377723,-.0757886113280989,-.0666685303985192,-.0493903424240987,-.0299583607055857,-.0137918069785563,-.00395990675875596,-.000341606679368254,-.0136665675675336,-.0289459000949523,-.0382089949857042,-.0400813487105939,-.0352581287359617,-.0261204355502155,-.0158436931512049,-.00729389568128642,-.00209422498813209,-.000180661133614816},
			{-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684},
			{-6.13522731444136e-006,-6.71979884668803e-005,-.000210744143550815,-.000390716517328129,-.000516288304945811,-.000516288304945811,-.000390716517328129,-.000210744143550815,-6.71979884668803e-005,-6.13522731444136e-006,-7.11195933109096e-005,-.000778959501602571,-.00244294742701878,-.0045291883068169,-.00598481725240988,-.00598481725240988,-.0045291883068169,-.00244294742701878,-.000778959501602571,-7.11195933109096e-005,-.000247699696758929,-.00271300809456183,-.00850844765415349,-.0157745357915389,-.0208442898723963,-.0208442898723963,-.0157745357915389,-.00850844765415349,-.00271300809456183,-.000247699696758929,-.000538049645975034,-.00589315636598079,-.0184819251214903,-.0342652151339464,-.0452776605429554,-.0452776605429554,-.0342652151339464,-.0184819251214903,-.00589315636598079,-.000538049645975034,-.000887046408080575,-.00971565212579472,-.0304699118679336,-.0564907648096207,-.0746462458462128,-.0746462458462128,-.0564907648096207,-.0304699118679336,-.00971565212579472,-.000887046408080575,-.00119736121515858,-.0131144717225921,-.0411291792262685,-.0762528884413686,-.100759688353677,-.100759688353677,-.0762528884413686,-.0411291792262685,-.0131144717225921,-.00119736121515858,-.00136115710384714,-.014908497220736,-.0467555435824129,-.0866840845325749,-.11454335070129,-.11454335070129,-.0866840845325749,-.0467555435824129,-.014908497220736,-.00136115710384714,-.00129757222819962,-.0142120640616299,-.0445714125837813,-.0826347380463795,-.109192590902875,-.109192590902875,-.0826347380463795,-.0445714125837813,-.0142120640616299,-.00129757222819962,-.000982998796422244,-.0107666005511246,-.0337658621019407,-.0626014076726695,-.082720778930882,-.082720778930882,-.0626014076726695,-.0337658621019407,-.0107666005511246,-.000982998796422244,-.000464114759811927,-.00508336149236723,-.0159422728047302,-.0295567373954438,-.039055932300929,-.039055932300929,-.0295567373954438,-.0159422728047302,-.00508336149236723,-.000464114759811927},
			{-1.03337249781483e-007,-1.13183342085305e-006,-3.54961912345676e-006,-6.58094122280528e-006,-8.69597991941751e-006,-8.69597991941751e-006,-6.58094122280528e-006,-3.54961912345676e-006,-1.13183342085305e-006,-1.03337249781483e-007,-6.19460350490148e-006,-6.78483246903409e-005,-.000212783706840731,-.000394497837425948,-.000521284897762289,-.000521284897762289,-.000394497837425948,-.000212783706840731,-6.78483246903409e-005,-6.19460350490148e-006,-5.12590331676487e-005,-.000561430529479367,-.00176074014710611,-.00326438612530762,-.0043135222202759,-.0043135222202759,-.00326438612530762,-.00176074014710611,-.000561430529479367,-5.12590331676487e-005,-.000196787192431025,-.00215537303015378,-.00675961072102864,-.0125322180484573,-.0165599285581662,-.0165599285581662,-.0125322180484573,-.00675961072102864,-.00215537303015378,-.000196787192431025,-.000487342632915848,-.00533777200871002,-.0167401467828093,-.0310359839203022,-.0410105916179671,-.0410105916179671,-.0310359839203022,-.0167401467828093,-.00533777200871002,-.000487342632915848,-.000887957326235526,-.00972562924066487,-.0305012017707475,-.0565487758254857,-.0747229009342903,-.0747229009342903,-.0565487758254857,-.0305012017707475,-.00972562924066487,-.000887957326235526,-.00125941435131387,-.0137941280277271,-.0432606952017087,-.080204687454718,-.105981550045219,-.105981550045219,-.080204687454718,-.0432606952017087,-.0137941280277271,-.00125941435131387,-.00140663864321037,-.0154066479494616,-.048317827678703,-.0895805360822827,-.118370688411972,-.118370688411972,-.0895805360822827,-.048317827678703,-.0154066479494616,-.00140663864321037,-.00118342559270231,-.0129618374762797,-.0406504926725559,-.0753654106685247,-.0995869854483788,-.0995869854483788,-.0753654106685247,-.0406504926725559,-.0129618374762797,-.00118342559270231,-.00059135237137032,-.00647697107125416,-.020312865791922,-.0376597519885992,-.0497631624376525,-.0497631624376525,-.0376597519885992,-.020312865791922,-.00647697107125416,-.00059135237137032},
			{-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474138,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512756,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512757,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474139,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916},
			{-1.72916292281226e-007,-9.79398733224426e-006,-7.29759010167018e-005,-.000239119650097383,-.000474634569865839,-.000640675640139909,-.000604924495042971,-.000382283481676095,-.000135370258905769,-1.308068949145e-005,-2.00444673906772e-006,-.000113531962266802,-.000845937099932474,-.00277187647596092,-.00550196689545652,-.00742671601811804,-.00701228852107834,-.00443143250491902,-.00156921288590388,-.000151631434204312,-6.9812104698908e-006,-.000395416105699515,-.00294628179627482,-.00965406199030398,-.019162588931357,-.0258662236377037,-.0244228300443723,-.0154340658681427,-.00546535121888794,-.000528111291459933,-1.51644829240855e-005,-.000858917061539686,-.00639987007723783,-.0209704117690521,-.041624694440043,-.0561862313643539,-.0530509129846532,-.0335256513631973,-.0118717557063213,-.00114715559656894,-2.50006671481639e-005,-.00141603902163566,-.0105510370774588,-.0345725131033195,-.0686238453397381,-.0926304758086336,-.0874614864268777,-.0552714955631914,-.0195721683597038,-.00189123858561784,-3.37466325590279e-005,-.00191141093432354,-.0142420987911849,-.0466669905017148,-.0926304758086336,-.12503532855168,-.118058075331442,-.0746070831112178,-.026419085942947,-.00255284922006524,-3.83630838022064e-005,-.00217288696066628,-.0161903807287985,-.0530509129846532,-.105302083100265,-.142139835110194,-.134208111862577,-.0848131373294056,-.0300331479365882,-.00290207232951189,-3.65709968299352e-005,-.00207138306607571,-.0154340658681427,-.0505727011047881,-.100383018401244,-.135499937544735,-.12793873555065,-.0808511899721722,-.0286301842585129,-.0027665053860117,-2.77050056147299e-005,-.00156921288590388,-.0116923496377051,-.0383122443879716,-.0760469314348448,-.102650374774569,-.0969222524410603,-.0612502492768556,-.0216893025727857,-.0020958150965657,-1.308068949145e-005,-.000740890898631601,-.00552044627469675,-.0180888096370989,-.0359049303476018,-.0484655263124446,-.0457610406806555,-.028918799122657,-.0102404250042771,-.000989521781403059},
			{-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.0551532224231419,-.0216893025727857,-.0022181246897901,-5.68214139596138e-007,-.000166430787646874,-.00294628179627482,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-5.68214139596138e-007,-.000166430787646874,-.00294628179627483,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.055153222423142,-.0216893025727857,-.0022181246897901,-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461163,-.00238829215039022,-.000288622746300124,-4.81475219575139e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461162,-.00238829215039021,-.000288622746300124,-4.81475219575139e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006},
			{-.000189474057272324,-.000379179080671673,-.000450698337824376,-.000403526490267504,-.000284508671025808,-.000156148500303414,-6.30521996263144e-005,-1.64238066106426e-005,-1.98480079913645e-006,-3.31100861878508e-008,-.00219638445415982,-.00439544626910526,-.00522449794427726,-.00467768159306037,-.00329802628977656,-.00181007438983512,-.000730901491494302,-.000190384868710001,-2.30078232481278e-005,-3.8381232568618e-007,-.00764970295714498,-.0153087307912168,-.0181962029908938,-.0162917173481382,-.0114865689446415,-.00630423848900844,-.00254562869914522,-.000663084138302037,-8.01330628639868e-005,-1.33676519027876e-006,-.0166165724938767,-.0332534003485589,-.0395255250831765,-.0353886293208596,-.0249509564807709,-.0136939743226831,-.00552957729977355,-.0014403416335721,-.000174063863093159,-2.90370172750447e-006,-.0273946299483073,-.0548226535530947,-.0651630854416681,-.0583428625235449,-.0411349705180729,-.0225763381244699,-.00911624367502254,-.00237459476467777,-.000286967430772327,-4.7871385229751e-006,-.0369780736440852,-.0740012230234841,-.0879590407639987,-.0787529041667412,-.0555251876749137,-.030474202256271,-.0123053726445112,-.00320529754367204,-.000387357040725739,-6.46181975011305e-006,-.0420365775924714,-.0841243971631835,-.0999916079356561,-.0895260958832346,-.0631208884080524,-.0346429935762571,-.0139887155008144,-.00364377388066238,-.000440346472755273,-7.34577982424566e-006,-.0400728876177358,-.0801946235041213,-.0953206159256855,-.0853439881325168,-.0601722693038774,-.033024686304908,-.0133352488781344,-.0034735592093105,-.000419776197925987,-7.00263023825107e-006,-.0303579249318969,-.0607528557338689,-.0722118188820862,-.0646538480039295,-.0455845671001885,-.0250184353398111,-.010102353697401,-.00263145622867657,-.00031800888486993,-5.30496641835636e-006,-.014333243427598,-.028683958870995,-.0340942136430102,-.030525780139787,-.02152237671891,-.0118122475336369,-.00476974282864777,-.0012424203162536,-.000150145267478098,-2.50469606272866e-006},
			{-3.19136145764326e-006,-6.3866131386048e-006,-7.59123082633588e-006,-6.79670297198979e-006,-4.79205449098542e-006,-2.63005032304177e-006,-1.06200480743303e-006,-2.76630501080463e-007,-3.34305226934559e-008,-5.57681903477072e-010,-.000191307770554455,-.000382848742505824,-.000455060156746716,-.00040743178419295,-.000287262120951629,-.000157659691772648,-6.36624132755433e-005,-1.65827547682883e-005,-2.00400952691774e-006,-3.34305226934559e-008,-.00158303131884398,-.00316799232989985,-.00376552650213992,-.00337141179786177,-.00237703326355233,-.0013046005871693,-.000526792998299499,-.000137218786643261,-1.65827547682883e-005,-2.76630501080463e-007,-.00607737347965237,-.0121621551894091,-.0144561327553171,-.0129430974646002,-.00912560525126807,-.00500845745477193,-.00202239700441925,-.000526792998299499,-6.36624132755433e-005,-1.06200480743303e-006,-.0150505891984046,-.030119524846696,-.0358005503901207,-.0320535250214294,-.0225995220276466,-.0124034232751763,-.00500845745477193,-.0013046005871693,-.000157659691772648,-2.63005032304177e-006,-.0274227618111799,-.0548789515711727,-.0652300021690509,-.0584027754849978,-.0411772124958644,-.0225995220276466,-.00912560525126807,-.00237703326355233,-.000287262120951629,-4.79205449098542e-006,-.0388944589534263,-.0778363296880565,-.0925175100657338,-.0828342662751547,-.0584027754849978,-.0320535250214294,-.0129430974646002,-.00337141179786177,-.00040743178419295,-6.7967029719898e-006,-.0434411827319366,-.0869353196354016,-.103332715477079,-.0925175100657338,-.0652300021690509,-.0358005503901207,-.0144561327553171,-.00376552650213992,-.000455060156746716,-7.59123082633588e-006,-.0365477002003157,-.0731399515169607,-.0869353196354016,-.0778363296880565,-.0548789515711727,-.030119524846696,-.0121621551894091,-.00316799232989985,-.000382848742505824,-6.3866131386048e-006,-.0182627191053361,-.0365477002003157,-.0434411827319366,-.0388944589534263,-.0274227618111799,-.0150505891984046,-.00607737347965237,-.00158303131884398,-.000191307770554455,-3.19136145764326e-006},
			{-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474138,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512757,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512756,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474139,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006},
			{-1.308068949145e-005,-.000135370258905769,-.000382283481676095,-.000604924495042971,-.000640675640139909,-.000474634569865839,-.000239119650097383,-7.29759010167018e-005,-9.79398733224426e-006,-1.72916292281226e-007,-.000151631434204312,-.00156921288590388,-.00443143250491902,-.00701228852107834,-.00742671601811804,-.00550196689545652,-.00277187647596092,-.000845937099932474,-.000113531962266802,-2.00444673906772e-006,-.000528111291459933,-.00546535121888794,-.0154340658681427,-.0244228300443723,-.0258662236377037,-.019162588931357,-.00965406199030398,-.00294628179627483,-.000395416105699515,-6.9812104698908e-006,-.00114715559656894,-.0118717557063213,-.0335256513631973,-.0530509129846532,-.0561862313643539,-.041624694440043,-.0209704117690521,-.00639987007723783,-.000858917061539686,-1.51644829240855e-005,-.00189123858561784,-.0195721683597038,-.0552714955631914,-.0874614864268777,-.0926304758086336,-.0686238453397381,-.0345725131033195,-.0105510370774588,-.00141603902163566,-2.50006671481639e-005,-.00255284922006524,-.026419085942947,-.0746070831112178,-.118058075331442,-.12503532855168,-.0926304758086336,-.0466669905017148,-.0142420987911849,-.00191141093432354,-3.37466325590279e-005,-.00290207232951189,-.0300331479365882,-.0848131373294056,-.134208111862577,-.142139835110194,-.105302083100265,-.0530509129846532,-.0161903807287985,-.00217288696066627,-3.83630838022064e-005,-.0027665053860117,-.0286301842585129,-.0808511899721722,-.12793873555065,-.135499937544735,-.100383018401244,-.0505727011047881,-.0154340658681427,-.00207138306607571,-3.65709968299352e-005,-.0020958150965657,-.0216893025727857,-.0612502492768556,-.0969222524410603,-.102650374774569,-.0760469314348448,-.0383122443879716,-.0116923496377051,-.00156921288590388,-2.77050056147299e-005,-.000989521781403059,-.0102404250042771,-.028918799122657,-.0457610406806555,-.0484655263124446,-.0359049303476018,-.0180888096370989,-.00552044627469675,-.000740890898631601,-1.308068949145e-005},
			{-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005},
			{-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596958,-1.81488157315799e-005,-5.85455199908462e-008,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596959,-1.81488157315799e-005,-5.85455199908462e-008,-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008},
			{-.000255493425333229,-.000483104798924764,-.000517065986523608,-.000395131310801382,-.000223291106959738,-9.07893469973411e-005,-2.44052765295024e-005,-3.59689507657609e-006,-1.82957487178445e-007,-5.90195050823199e-010,-.0029616813806621,-.00560015384355916,-.00599383214211218,-.00458036462030578,-.00258839190514417,-.00105243067689039,-.000282906117811307,-4.16952301711737e-005,-2.12084433297021e-006,-6.8415447117756e-009,-.0103151261942612,-.0195045604772907,-.0208756874850881,-.0159527758058908,-.00901501063419505,-.00366547033510042,-.000985322839048874,-.000145218713843646,-7.38660717382125e-006,-2.38281529966161e-008,-.0224063657308857,-.0423675200132299,-.0453458619764426,-.0346523854772107,-.0195822737922466,-.00796208086622,-.00214030381004998,-.0003154419589321,-1.60450796946238e-005,-5.17591642291991e-008,-.0369398742075263,-.0698484921013427,-.0747586849808788,-.0571290666185282,-.0322839830104985,-.0131265493548066,-.00352857551549082,-.000520048026649204,-2.64524480537304e-005,-8.53318667863116e-008,-.049862523838675,-.0942835398662266,-.10091146198997,-.0771144869130937,-.0435778655721675,-.0177186006765078,-.00476297455073299,-.000701976053853711,-3.57062889348363e-005,-1.15183452383392e-007,-.0566835869406137,-.107181282005842,-.114715886574844,-.0876635474261487,-.0495392038284753,-.0201424589971083,-.00541453703618982,-.000798004545609921,-4.05908160548146e-005,-1.30940247999038e-007,-.0540356789095071,-.102174432709151,-.109357066963176,-.0835684464667091,-.0472250373695059,-.0192015273831553,-.00516160321748368,-.000760726688663539,-3.86946631537408e-005,-1.24823526157934e-007,-.0409356844864075,-.0774040490351568,-.0828453806800583,-.0633087550044376,-.0357761625028231,-.0145464567574012,-.00391026382973348,-.000576301960778015,-2.9313826607454e-005,-9.45622704554012e-008,-.0193274452036918,-.0365456822095551,-.0391147619872944,-.029890705593759,-.0168914195291221,-.00686798936951685,-.00184619875908625,-.000272096209150128,-1.38402810304592e-005,-4.46467946852167e-008},
			{-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007,-.0022181246897901,-.0216893025727857,-.0551532224231419,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627483,-.000166430787646874,-5.68214139596138e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627482,-.000166430787646874,-5.68214139596138e-007,-.0022181246897901,-.0216893025727857,-.055153222423142,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007},
			{-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.2505954733558e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483912,-.00712261469039878,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.064163146746701,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.0641631467467009,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483913,-.00712261469039877,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.25059547335579e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{-5.02239597033184e-005,-.000112582670872584,-.000165039185570316,-.000202840374019418,-.000222620323957012,-.000222620323957012,-.000202840374019418,-.000165039185570316,-.000112582670872584,-5.02239597033184e-005,-.000582196454262715,-.00130505902323575,-.00191313526892416,-.00235132688129378,-.00258061618439108,-.00258061618439108,-.00235132688129378,-.00191313526892416,-.00130505902323575,-.000582196454262715,-.00202770964317194,-.00454533988819102,-.00666318526175589,-.00818934597856097,-.00898792887538633,-.00898792887538633,-.00818934597856097,-.0066631852617559,-.00454533988819102,-.00202770964317194,-.00440456112754404,-.0098733205961796,-.0144736732344206,-.0177887771450126,-.0195234471932234,-.0195234471932234,-.0177887771450126,-.0144736732344206,-.00987332059617959,-.00440456112754404,-.00726150487522216,-.016277482266154,-.0238617754892464,-.0293271652321865,-.0321869995374054,-.0321869995374054,-.0293271652321865,-.0238617754892464,-.016277482266154,-.00726150487522216,-.0098017919040896,-.0219718221824466,-.03220932325003,-.0395866663566687,-.0434469544404215,-.0434469544404215,-.0395866663566687,-.03220932325003,-.0219718221824466,-.0098017919040896,-.0111426514503529,-.0249775101026156,-.0366154745926699,-.0450020189787675,-.0493903843956318,-.0493903843956318,-.0450020189787675,-.0366154745926699,-.0249775101026156,-.0111426514503529,-.0106221354093669,-.0238107146832105,-.0349050251580164,-.0428998018485296,-.0470831698638901,-.0470831698638901,-.0428998018485296,-.0349050251580163,-.0238107146832105,-.0106221354093669,-.00804698659228353,-.0180382281362673,-.0264429193024774,-.0324995038175014,-.0356686882642096,-.0356686882642096,-.0324995038175014,-.0264429193024774,-.0180382281362673,-.00804698659228353,-.00379931823220995,-.00851660037567366,-.0124848058669686,-.0153443722038775,-.0168406764553548,-.0168406764553548,-.0153443722038775,-.0124848058669686,-.00851660037567366,-.00379931823220995},
			{-8.45935382972169e-007,-1.89625958134751e-006,-2.77979847617661e-006,-3.41649385058338e-006,-3.74965275769605e-006,-3.74965275769605e-006,-3.41649385058338e-006,-2.77979847617661e-006,-1.89625958134751e-006,-8.45935382972169e-007,-5.07100227590783e-005,-.000113672236039353,-.000166636420263325,-.00020480344528281,-.000224774823832688,-.000224774823832688,-.00020480344528281,-.000166636420263325,-.000113672236039353,-5.07100227590783e-005,-.000419614707621413,-.000940613699128858,-.00137888111587085,-.00169470517141149,-.00185996410278327,-.00185996410278327,-.00169470517141149,-.00137888111587085,-.000940613699128858,-.000419614707621413,-.00161093167609136,-.00361108506296509,-.00529362585909322,-.006506097600196,-.00714053877317793,-.00714053877317793,-.006506097600196,-.00529362585909322,-.00361108506296509,-.00161093167609136,-.00398946534464676,-.00894283657654873,-.0131096415979721,-.0161123226329801,-.017683513460247,-.017683513460247,-.0161123226329801,-.0131096415979721,-.00894283657654873,-.00398946534464676,-.00726896179664528,-.016294197796655,-.023886279416173,-.0293572816296078,-.0322200527309955,-.0322200527309955,-.0293572816296078,-.023886279416173,-.016294197796655,-.00726896179664528,-.0103097688766848,-.0231105098656633,-.0338785684935517,-.0416382417346981,-.045698588896762,-.045698588896762,-.0416382417346981,-.0338785684935517,-.0231105098656633,-.0103097688766848,-.0115149706602782,-.0258121056087878,-.0378389396388598,-.0465057110062776,-.0510407087352223,-.0510407087352223,-.0465057110062776,-.0378389396388598,-.0258121056087878,-.0115149706602782,-.00968771265055556,-.021716100667658,-.0318344514317794,-.0391259325039141,-.0429412921921929,-.0429412921921929,-.0391259325039141,-.0318344514317794,-.021716100667658,-.00968771265055556,-.00484090582829009,-.0108514364620189,-.0159075301916269,-.0195510500287827,-.0214575678641308,-.0214575678641308,-.0195510500287827,-.0159075301916269,-.0108514364620189,-.00484090582829009},
			{-1.30587941023263e-008,-2.92727599954231e-008,-4.2912043492978e-008,-5.27407774218851e-008,-5.7883786756776e-008,-5.7883786756776e-008,-5.27407774218851e-008,-4.2912043492978e-008,-2.92727599954231e-008,-1.30587941023263e-008,-4.04816026703356e-006,-9.07440786578996e-006,-1.33025169157498e-005,-1.63493748303836e-005,-1.79436817686314e-005,-1.79436817686314e-005,-1.63493748303836e-005,-1.33025169157498e-005,-9.07440786578996e-006,-4.04816026703356e-006,-7.95857439793226e-005,-.000178400422298479,-.000261523911036573,-.00032142431960244,-.000352767965960796,-.000352767965960796,-.00032142431960244,-.000261523911036573,-.000178400422298479,-7.95857439793226e-005,-.000539996871821585,-.00121046389913573,-.00177446470693278,-.00218089469839968,-.00239356433166203,-.00239356433166203,-.00218089469839968,-.00177446470693278,-.00121046389913573,-.000539996871821585,-.00200882638285304,-.0045030109301789,-.00660113364491101,-.00811308182877234,-.00890422783798524,-.00890422783798524,-.00811308182877234,-.00660113364491101,-.0045030109301789,-.00200882638285304,-.00494059139703164,-.0110748929087561,-.0162351034290897,-.0199536518580156,-.0218994293529137,-.0218994293529137,-.0199536518580156,-.0162351034290897,-.0110748929087561,-.00494059139703164,-.00874276804581894,-.0195979007476136,-.0287293022381269,-.0353095683983229,-.0387527758890071,-.0387527758890071,-.0353095683983229,-.0287293022381268,-.0195979007476136,-.00874276804581894,-.0114407232759916,-.0256456717219007,-.0375949579236433,-.0462058468122232,-.0507116033273556,-.0507116033273556,-.0462058468122232,-.0375949579236432,-.0256456717219007,-.0114407232759916,-.0106892900748741,-.0239612494409041,-.0351256997397901,-.0431710205566739,-.0473808364254027,-.0473808364254027,-.0431710205566739,-.0351256997397901,-.0239612494409041,-.0106892900748741,-.00565310744519306,-.0126720780015972,-.0185764773269063,-.0228313027353587,-.0250576939422306,-.0250576939422306,-.0228313027353587,-.0185764773269063,-.0126720780015972,-.00565310744519306},
			{-1.93186678290668e-010,-4.33049730595595e-010,-6.3482394133133e-010,-7.80226376246059e-010,-8.56310039258135e-010,-8.56310039258135e-010,-7.80226376246059e-010,-6.3482394133133e-010,-4.33049730595595e-010,-1.93186678290668e-010,-3.09691872038552e-007,-6.94209264016414e-007,-1.01766755630007e-006,-1.25075791566716e-006,-1.37272539416124e-006,-1.37272539416124e-006,-1.25075791566716e-006,-1.01766755630007e-006,-6.94209264016414e-007,-3.09691872038552e-007,-1.44653210137677e-005,-3.24256487218327e-005,-4.75339820521491e-005,-5.84213419665859e-005,-6.41182907371273e-005,-6.41182907371273e-005,-5.84213419665859e-005,-4.75339820521491e-005,-3.24256487218327e-005,-1.44653210137677e-005,-.000173465688278746,-.000388842906981744,-.000570019490439081,-.000700578873068661,-.000768895721248536,-.000768895721248536,-.000700578873068661,-.000570019490439081,-.000388842906981744,-.000173465688278746,-.000969344894307206,-.00217289592143802,-.00318533012606401,-.00391490997676346,-.00429667186083074,-.00429667186083074,-.00391490997676346,-.00318533012606401,-.00217289592143802,-.000969344894307206,-.00321805680650706,-.00721363732442475,-.010574743162472,-.0129968216385873,-.01426420483388,-.01426420483388,-.0129968216385873,-.010574743162472,-.00721363732442475,-.00321805680650706,-.00710488740889465,-.0159264065491336,-.023347120347727,-.0286946315642816,-.0314927844397286,-.0314927844397286,-.0286946315642816,-.023347120347727,-.0159264065491336,-.00710488740889465,-.0108931213951503,-.0244181603372121,-.0357955026643469,-.0439942376324665,-.0482843293960987,-.0482843293960987,-.0439942376324665,-.0357955026643469,-.0244181603372121,-.0108931213951503,-.0113027648830266,-.0253364224225416,-.0371416177061049,-.0456486718663717,-.050100095546689,-.050100095546689,-.0456486718663717,-.0371416177061049,-.0253364224225416,-.0113027648830266,-.00632639147606921,-.0141813201023726,-.0207889322741007,-.0255505065864879,-.0280420605663306,-.0280420605663306,-.0255505065864879,-.0207889322741007,-.0141813201023726,-.00632639147606921},
			{-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262448,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-.000233400654593986,-.00270558184440079,-.0094231669672911,-.0204688649887572,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813583,-.0373959351122988,-.017656181783494,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000233400654593986,-.00270558184440079,-.00942316696729111,-.0204688649887571,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813582,-.0373959351122988,-.017656181783494,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262447},
			{-1.60504953819679e-006,-1.86057442617858e-005,-6.4801231236956e-005,-.000140760283448081,-.000232062049974506,-.000313244150044994,-.000356095132090808,-.000339460560939038,-.000257164353231333,-.000121418024586148,-1.86057442617858e-005,-.000215677903576663,-.000751176276838271,-.00163169408403043,-.0026900647312993,-.00363112814186739,-.00412785699311923,-.00393502894053338,-.00298105078728298,-.00140747849861324,-6.4801231236956e-005,-.000751176276838271,-.00261624296938716,-.00568296458123414,-.00936912301075426,-.0126467165764205,-.014376754391578,-.0137051609821032,-.0103825871557633,-.00490205274061561,-.000140760283448081,-.00163169408403043,-.00568296458123414,-.0123444522582422,-.0203514714995352,-.0274710121398013,-.0312289748912649,-.0297701495437364,-.0225529034413806,-.0106481670189776,-.000232062049974506,-.0026900647312993,-.00936912301075426,-.0203514714995352,-.033552107742962,-.0452896174679006,-.0514851189152542,-.0490800512895761,-.0371814611144508,-.0175549196574778,-.000313244150044994,-.00363112814186739,-.0126467165764205,-.0274710121398013,-.0452896174679006,-.0611332517796597,-.0694961210432554,-.0662496903395319,-.0501886249195276,-.0236961445777924,-.000356095132090808,-.00412785699311923,-.014376754391578,-.0312289748912649,-.0514851189152542,-.0694961210432554,-.0790030089920025,-.075312475042366,-.0570542977980853,-.0269377153005406,-.000339460560939038,-.00393502894053338,-.0137051609821032,-.0297701495437364,-.0490800512895761,-.0662496903395319,-.075312475042366,-.0717943401064784,-.0543890724391648,-.0256793511684561,-.000257164353231333,-.00298105078728298,-.0103825871557633,-.0225529034413806,-.0371814611144508,-.0501886249195276,-.0570542977980853,-.0543890724391648,-.0412034040065755,-.0194538467631362,-.000121418024586148,-.00140747849861324,-.00490205274061561,-.0106481670189776,-.0175549196574778,-.0236961445777924,-.0269377153005406,-.0256793511684561,-.0194538467631362,-.00918497301395746},
			{-2.70342721665942e-008,-3.13381451640588e-007,-1.09146420736547e-006,-2.3708625325406e-006,-3.90868223643537e-006,-5.27605373253693e-006,-5.99780411074261e-006,-5.71762364703134e-006,-4.33148694252304e-006,-2.04507577148045e-006,-1.62058306631703e-006,-1.878580901668e-005,-6.54283718476915e-005,-.000142122549078592,-.000234307926063146,-.00031627569935075,-.000359541389428153,-.000342745830361303,-.000259653167200857,-.000122593097538308,-1.34099819434037e-005,-.000155448594362042,-.000541405931790853,-.00117603401917912,-.00193884850644768,-.00261711448526345,-.00297512891523367,-.00283614921805116,-.00214857501357473,-.00101443194029589,-5.14819054148423e-005,-.000596778568799888,-.00207849712916292,-.00451488095923928,-.00744338179155761,-.0100472946912819,-.0114217383779821,-.0108881851148137,-.00824853725324792,-.00389447871149425,-.00012749471661469,-.00147791955052476,-.00514738917162678,-.0111810832137794,-.0184335028884922,-.0248820819487376,-.0282858858081064,-.0269645434541597,-.0204274669151108,-.00964465972428348,-.000232300357136792,-.00269282718941148,-.00937874426988752,-.0203723706574197,-.033586562784559,-.0453361258919233,-.0515379895702794,-.0491304521532631,-.0372196431803629,-.0175729470044231,-.00032947799961708,-.00381931102740118,-.0133021315121917,-.0288946948441916,-.0476367477719961,-.0643014769902532,-.0730977511924463,-.0696830745129148,-.0527896458390131,-.0249241951142798,-.000367993651863299,-.004265784708565,-.0148571375279271,-.0322724560897768,-.0532054364658069,-.0718182560454652,-.0816428060009037,-.0778289569952232,-.0589607032197774,-.0278378088689286,-.000309598422927644,-.00358886685037487,-.0124995263493677,-.0271512876888108,-.0447625091834733,-.0604217455831262,-.0686872826563242,-.065478635899881,-.049604498988284,-.0234203543456317,-.000154704919937309,-.00179334039722375,-.00624595630961966,-.0135673746279838,-.0223676216885663,-.0301924706996799,-.0343227218781454,-.0327193757277722,-.0247871419109467,-.0117030442522349},
			{-4.17330922690109e-010,-4.83770265983857e-009,-1.68490485682806e-008,-3.65992560176709e-008,-6.03387416602869e-008,-8.14470002666849e-008,-9.2588737297087e-008,-8.82635618042971e-008,-6.6865622688263e-008,-3.15700512824513e-008,-1.29370479862118e-007,-1.49966340979313e-006,-5.22312002294925e-006,-1.13455846575704e-005,-1.8704705399135e-005,-2.52481590386612e-005,-2.87020412883381e-005,-2.7361258713725e-005,-2.07280055781616e-005,-9.78655657087546e-006,-2.54338890004804e-006,-2.94829799993105e-005,-.000102685137321483,-.000223051148247843,-.000367729486213021,-.000496372027946554,-.000564274425659415,-.000537915000222613,-.000407507024506811,-.000192401074639172,-1.72571365319964e-005,-.000200044834358053,-.000696728461200002,-.00151342333799558,-.00249507967511344,-.00336793160368666,-.00382865585555338,-.00364980463714188,-.00276497407043368,-.00130545966206281,-6.41977629259773e-005,-.000744180868411951,-.00259187893037573,-.0056300413732684,-.00928187206303727,-.0125289426924045,-.0142428693476609,-.0135775302227536,-.0102858982161417,-.00485640185665492,-.00015789065592134,-.00183026946862863,-.00637457515237077,-.0138467585905541,-.0228282233120914,-.0308142042579516,-.0350295069642396,-.0333931441681891,-.025297567112414,-.0119440372937759,-.000279400029347063,-.00323880688356021,-.0112803159519127,-.0245029367570041,-.0403963504117509,-.054528176627927,-.0619874888523968,-.0590918151941233,-.0447660499753643,-.0211359206213315,-.000365620865418288,-.00423827935328899,-.0147613401837082,-.0320643665046187,-.0528623731064046,-.0713551790777345,-.0811163813127856,-.0773271236259448,-.0585805304731015,-.0276583134477209,-.000341606679368254,-.00395990675875596,-.0137918069785563,-.0299583607055857,-.0493903424240987,-.0666685303985192,-.0757886113280989,-.0722482342377723,-.0547329279680274,-.0258416997153403,-.000180661133614816,-.00209422498813209,-.00729389568128642,-.0158436931512049,-.0261204355502155,-.0352581287359617,-.0400813487105939,-.0382089949857042,-.0289459000949523,-.0136665675675336},
			{-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779044,-.0447078783667129,-.0376133897003122,-.0187952392881457,-4.81475219575139e-006,-.000288622746300124,-.00238829215039021,-.00916882894461163,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619087,-.0551388873133415,-.0275526505160409,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550778,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550777,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-4.81475219575139e-006,-.000288622746300124,-.00238829215039022,-.00916882894461162,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619086,-.0551388873133415,-.0275526505160409,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779045,-.044707878366713,-.0376133897003122,-.0187952392881457,-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211},
			{-3.31100861878508e-008,-1.98480079913645e-006,-1.64238066106426e-005,-6.30521996263144e-005,-.000156148500303414,-.000284508671025808,-.000403526490267504,-.000450698337824376,-.000379179080671673,-.000189474057272324,-3.8381232568618e-007,-2.30078232481278e-005,-.000190384868710001,-.000730901491494302,-.00181007438983512,-.00329802628977656,-.00467768159306037,-.00522449794427726,-.00439544626910526,-.00219638445415982,-1.33676519027876e-006,-8.01330628639868e-005,-.000663084138302037,-.00254562869914522,-.00630423848900844,-.0114865689446415,-.0162917173481382,-.0181962029908938,-.0153087307912168,-.00764970295714498,-2.90370172750447e-006,-.000174063863093159,-.0014403416335721,-.00552957729977356,-.0136939743226831,-.0249509564807709,-.0353886293208596,-.0395255250831765,-.0332534003485589,-.0166165724938767,-4.7871385229751e-006,-.000286967430772327,-.00237459476467777,-.00911624367502254,-.0225763381244699,-.0411349705180729,-.0583428625235449,-.0651630854416681,-.0548226535530947,-.0273946299483073,-6.46181975011305e-006,-.000387357040725739,-.00320529754367204,-.0123053726445113,-.030474202256271,-.0555251876749137,-.0787529041667412,-.0879590407639987,-.0740012230234841,-.0369780736440852,-7.34577982424567e-006,-.000440346472755273,-.00364377388066238,-.0139887155008144,-.0346429935762571,-.0631208884080524,-.0895260958832347,-.0999916079356561,-.0841243971631835,-.0420365775924714,-7.00263023825107e-006,-.000419776197925987,-.0034735592093105,-.0133352488781344,-.033024686304908,-.0601722693038774,-.0853439881325168,-.0953206159256855,-.0801946235041213,-.0400728876177358,-5.30496641835636e-006,-.00031800888486993,-.00263145622867657,-.010102353697401,-.0250184353398111,-.0455845671001885,-.0646538480039295,-.0722118188820862,-.0607528557338689,-.0303579249318969,-2.50469606272866e-006,-.000150145267478098,-.0012424203162536,-.00476974282864777,-.0118122475336369,-.02152237671891,-.030525780139787,-.0340942136430102,-.028683958870995,-.014333243427598},
			{-5.57681903477072e-010,-3.34305226934559e-008,-2.76630501080463e-007,-1.06200480743303e-006,-2.63005032304177e-006,-4.79205449098542e-006,-6.79670297198979e-006,-7.59123082633588e-006,-6.3866131386048e-006,-3.19136145764326e-006,-3.34305226934559e-008,-2.00400952691774e-006,-1.65827547682883e-005,-6.36624132755433e-005,-.000157659691772648,-.000287262120951629,-.00040743178419295,-.000455060156746716,-.000382848742505824,-.000191307770554455,-2.76630501080463e-007,-1.65827547682883e-005,-.000137218786643261,-.000526792998299499,-.0013046005871693,-.00237703326355233,-.00337141179786177,-.00376552650213992,-.00316799232989985,-.00158303131884398,-1.06200480743303e-006,-6.36624132755433e-005,-.000526792998299499,-.00202239700441925,-.00500845745477193,-.00912560525126807,-.0129430974646002,-.0144561327553171,-.0121621551894091,-.00607737347965237,-2.63005032304177e-006,-.000157659691772648,-.0013046005871693,-.00500845745477193,-.0124034232751763,-.0225995220276466,-.0320535250214294,-.0358005503901207,-.030119524846696,-.0150505891984046,-4.79205449098542e-006,-.000287262120951629,-.00237703326355233,-.00912560525126807,-.0225995220276466,-.0411772124958644,-.0584027754849978,-.0652300021690509,-.0548789515711727,-.0274227618111799,-6.79670297198979e-006,-.00040743178419295,-.00337141179786177,-.0129430974646002,-.0320535250214294,-.0584027754849978,-.0828342662751547,-.0925175100657338,-.0778363296880565,-.0388944589534263,-7.59123082633588e-006,-.000455060156746716,-.00376552650213992,-.0144561327553171,-.0358005503901207,-.0652300021690509,-.0925175100657338,-.103332715477079,-.0869353196354016,-.0434411827319366,-6.3866131386048e-006,-.000382848742505824,-.00316799232989985,-.0121621551894091,-.030119524846696,-.0548789515711727,-.0778363296880565,-.0869353196354016,-.0731399515169607,-.0365477002003157,-3.19136145764326e-006,-.000191307770554455,-.00158303131884398,-.00607737347965237,-.0150505891984046,-.0274227618111799,-.0388944589534263,-.0434411827319366,-.0365477002003157,-.0182627191053361},
			{-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596958,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596959,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861},
			{-5.90195050823199e-010,-1.82957487178445e-007,-3.59689507657609e-006,-2.44052765295024e-005,-9.07893469973411e-005,-.000223291106959738,-.000395131310801382,-.000517065986523608,-.000483104798924764,-.000255493425333229,-6.8415447117756e-009,-2.12084433297021e-006,-4.16952301711737e-005,-.000282906117811307,-.00105243067689039,-.00258839190514417,-.00458036462030578,-.00599383214211218,-.00560015384355916,-.0029616813806621,-2.38281529966161e-008,-7.38660717382125e-006,-.000145218713843646,-.000985322839048874,-.00366547033510042,-.00901501063419505,-.0159527758058908,-.0208756874850881,-.0195045604772907,-.0103151261942612,-5.17591642291992e-008,-1.60450796946238e-005,-.0003154419589321,-.00214030381004998,-.00796208086622,-.0195822737922466,-.0346523854772107,-.0453458619764426,-.0423675200132299,-.0224063657308857,-8.53318667863116e-008,-2.64524480537304e-005,-.000520048026649204,-.00352857551549082,-.0131265493548066,-.0322839830104985,-.0571290666185282,-.0747586849808788,-.0698484921013427,-.0369398742075263,-1.15183452383392e-007,-3.57062889348363e-005,-.000701976053853711,-.00476297455073299,-.0177186006765078,-.0435778655721675,-.0771144869130937,-.10091146198997,-.0942835398662266,-.049862523838675,-1.30940247999038e-007,-4.05908160548146e-005,-.000798004545609922,-.00541453703618982,-.0201424589971083,-.0495392038284753,-.0876635474261488,-.114715886574844,-.107181282005842,-.0566835869406137,-1.24823526157934e-007,-3.86946631537408e-005,-.000760726688663539,-.00516160321748368,-.0192015273831553,-.0472250373695059,-.0835684464667091,-.109357066963176,-.102174432709151,-.0540356789095071,-9.45622704554012e-008,-2.9313826607454e-005,-.000576301960778015,-.00391026382973348,-.0145464567574012,-.0357761625028231,-.0633087550044376,-.0828453806800583,-.0774040490351568,-.0409356844864075,-4.46467946852167e-008,-1.38402810304592e-005,-.000272096209150128,-.00184619875908625,-.00686798936951685,-.0168914195291221,-.029890705593759,-.0391147619872944,-.0365456822095551,-.0193274452036918},
			{-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807,-9.68328634942698e-010,-1.55229910445862e-006,-7.2505954733558e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039878,-.0236458445483912,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.0641631467467009,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.064163146746701,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039877,-.0236458445483913,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-9.68328634942698e-010,-1.55229910445862e-006,-7.25059547335579e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807},
			{-.00537304737262447,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.017656181783494,-.0373959351122988,-.0493631599813582,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887572,-.00942316696729111,-.00270558184440079,-.000233400654593986,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.017656181783494,-.0373959351122988,-.0493631599813583,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887571,-.0094231669672911,-.00270558184440079,-.000233400654593986,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.00537304737262448,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005},
			{-.000121418024586148,-.000257164353231333,-.000339460560939038,-.000356095132090808,-.000313244150044994,-.000232062049974506,-.000140760283448081,-6.4801231236956e-005,-1.86057442617858e-005,-1.60504953819679e-006,-.00140747849861324,-.00298105078728298,-.00393502894053338,-.00412785699311923,-.00363112814186739,-.0026900647312993,-.00163169408403043,-.000751176276838271,-.000215677903576663,-1.86057442617858e-005,-.00490205274061561,-.0103825871557633,-.0137051609821032,-.014376754391578,-.0126467165764205,-.00936912301075426,-.00568296458123414,-.00261624296938716,-.000751176276838271,-6.4801231236956e-005,-.0106481670189775,-.0225529034413806,-.0297701495437364,-.031228974891265,-.0274710121398013,-.0203514714995352,-.0123444522582422,-.00568296458123414,-.00163169408403043,-.000140760283448081,-.0175549196574778,-.0371814611144508,-.0490800512895761,-.0514851189152542,-.0452896174679006,-.033552107742962,-.0203514714995352,-.00936912301075425,-.0026900647312993,-.000232062049974506,-.0236961445777924,-.0501886249195276,-.0662496903395319,-.0694961210432554,-.0611332517796597,-.0452896174679006,-.0274710121398013,-.0126467165764205,-.00363112814186739,-.000313244150044994,-.0269377153005406,-.0570542977980853,-.075312475042366,-.0790030089920025,-.0694961210432554,-.0514851189152542,-.031228974891265,-.014376754391578,-.00412785699311923,-.000356095132090808,-.0256793511684561,-.0543890724391648,-.0717943401064784,-.075312475042366,-.0662496903395319,-.0490800512895761,-.0297701495437364,-.0137051609821032,-.00393502894053338,-.000339460560939038,-.0194538467631362,-.0412034040065755,-.0543890724391648,-.0570542977980853,-.0501886249195276,-.0371814611144508,-.0225529034413806,-.0103825871557633,-.00298105078728298,-.000257164353231333,-.00918497301395746,-.0194538467631362,-.0256793511684561,-.0269377153005406,-.0236961445777924,-.0175549196574778,-.0106481670189776,-.00490205274061561,-.00140747849861324,-.000121418024586148},
			{-2.04507577148045e-006,-4.33148694252304e-006,-5.71762364703134e-006,-5.99780411074261e-006,-5.27605373253693e-006,-3.90868223643537e-006,-2.3708625325406e-006,-1.09146420736547e-006,-3.13381451640588e-007,-2.70342721665942e-008,-.000122593097538308,-.000259653167200857,-.000342745830361303,-.000359541389428153,-.00031627569935075,-.000234307926063146,-.000142122549078592,-6.54283718476915e-005,-1.878580901668e-005,-1.62058306631703e-006,-.00101443194029589,-.00214857501357473,-.00283614921805116,-.00297512891523367,-.00261711448526345,-.00193884850644768,-.00117603401917912,-.000541405931790853,-.000155448594362042,-1.34099819434037e-005,-.00389447871149424,-.00824853725324792,-.0108881851148137,-.0114217383779821,-.0100472946912819,-.00744338179155761,-.00451488095923928,-.00207849712916292,-.000596778568799888,-5.14819054148423e-005,-.00964465972428348,-.0204274669151108,-.0269645434541598,-.0282858858081064,-.0248820819487376,-.0184335028884922,-.0111810832137794,-.00514738917162678,-.00147791955052476,-.00012749471661469,-.0175729470044231,-.0372196431803629,-.0491304521532631,-.0515379895702794,-.0453361258919233,-.033586562784559,-.0203723706574197,-.00937874426988752,-.00269282718941148,-.000232300357136792,-.0249241951142798,-.0527896458390131,-.0696830745129148,-.0730977511924463,-.0643014769902532,-.0476367477719961,-.0288946948441916,-.0133021315121917,-.00381931102740118,-.000329477999617081,-.0278378088689286,-.0589607032197774,-.0778289569952232,-.0816428060009037,-.0718182560454652,-.0532054364658069,-.0322724560897768,-.0148571375279271,-.004265784708565,-.000367993651863299,-.0234203543456317,-.049604498988284,-.065478635899881,-.0686872826563242,-.0604217455831262,-.0447625091834733,-.0271512876888108,-.0124995263493677,-.00358886685037487,-.000309598422927644,-.0117030442522349,-.0247871419109467,-.0327193757277722,-.0343227218781454,-.0301924706996799,-.0223676216885663,-.0135673746279838,-.00624595630961966,-.00179334039722375,-.000154704919937309},
			{-3.15700512824513e-008,-6.6865622688263e-008,-8.82635618042971e-008,-9.2588737297087e-008,-8.14470002666849e-008,-6.03387416602869e-008,-3.65992560176709e-008,-1.68490485682806e-008,-4.83770265983857e-009,-4.17330922690109e-010,-9.78655657087546e-006,-2.07280055781616e-005,-2.7361258713725e-005,-2.87020412883381e-005,-2.52481590386612e-005,-1.8704705399135e-005,-1.13455846575704e-005,-5.22312002294925e-006,-1.49966340979313e-006,-1.29370479862118e-007,-.000192401074639172,-.000407507024506811,-.000537915000222613,-.000564274425659415,-.000496372027946554,-.000367729486213021,-.000223051148247843,-.000102685137321483,-2.94829799993105e-005,-2.54338890004804e-006,-.00130545966206281,-.00276497407043368,-.00364980463714188,-.00382865585555338,-.00336793160368666,-.00249507967511344,-.00151342333799558,-.000696728461200002,-.000200044834358053,-1.72571365319964e-005,-.00485640185665492,-.0102858982161417,-.0135775302227536,-.0142428693476609,-.0125289426924045,-.00928187206303727,-.0056300413732684,-.00259187893037573,-.000744180868411951,-6.41977629259773e-005,-.0119440372937759,-.025297567112414,-.0333931441681891,-.0350295069642396,-.0308142042579516,-.0228282233120914,-.0138467585905541,-.00637457515237077,-.00183026946862863,-.00015789065592134,-.0211359206213315,-.0447660499753643,-.0590918151941233,-.0619874888523968,-.054528176627927,-.0403963504117509,-.0245029367570041,-.0112803159519127,-.00323880688356021,-.000279400029347063,-.0276583134477209,-.0585805304731015,-.0773271236259449,-.0811163813127856,-.0713551790777345,-.0528623731064046,-.0320643665046187,-.0147613401837082,-.00423827935328899,-.000365620865418288,-.0258416997153403,-.0547329279680274,-.0722482342377723,-.0757886113280989,-.0666685303985192,-.0493903424240987,-.0299583607055857,-.0137918069785563,-.00395990675875596,-.000341606679368254,-.0136665675675336,-.0289459000949523,-.0382089949857042,-.0400813487105939,-.0352581287359617,-.0261204355502155,-.0158436931512049,-.00729389568128642,-.00209422498813209,-.000180661133614816},
			{-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684},
			{-6.13522731444136e-006,-6.71979884668803e-005,-.000210744143550815,-.000390716517328129,-.000516288304945811,-.000516288304945811,-.000390716517328129,-.000210744143550815,-6.71979884668803e-005,-6.13522731444136e-006,-7.11195933109096e-005,-.000778959501602571,-.00244294742701878,-.0045291883068169,-.00598481725240988,-.00598481725240988,-.0045291883068169,-.00244294742701878,-.000778959501602571,-7.11195933109096e-005,-.000247699696758929,-.00271300809456183,-.00850844765415349,-.0157745357915389,-.0208442898723963,-.0208442898723963,-.0157745357915389,-.00850844765415349,-.00271300809456183,-.000247699696758929,-.000538049645975034,-.00589315636598079,-.0184819251214903,-.0342652151339464,-.0452776605429554,-.0452776605429554,-.0342652151339464,-.0184819251214903,-.00589315636598079,-.000538049645975034,-.000887046408080575,-.00971565212579472,-.0304699118679336,-.0564907648096207,-.0746462458462128,-.0746462458462128,-.0564907648096207,-.0304699118679336,-.00971565212579472,-.000887046408080575,-.00119736121515858,-.0131144717225921,-.0411291792262685,-.0762528884413686,-.100759688353677,-.100759688353677,-.0762528884413686,-.0411291792262685,-.0131144717225921,-.00119736121515858,-.00136115710384714,-.014908497220736,-.0467555435824129,-.0866840845325749,-.11454335070129,-.11454335070129,-.0866840845325749,-.0467555435824129,-.014908497220736,-.00136115710384714,-.00129757222819962,-.0142120640616299,-.0445714125837813,-.0826347380463795,-.109192590902875,-.109192590902875,-.0826347380463795,-.0445714125837813,-.0142120640616299,-.00129757222819962,-.000982998796422244,-.0107666005511246,-.0337658621019407,-.0626014076726695,-.082720778930882,-.082720778930882,-.0626014076726695,-.0337658621019407,-.0107666005511246,-.000982998796422244,-.000464114759811927,-.00508336149236723,-.0159422728047302,-.0295567373954438,-.039055932300929,-.039055932300929,-.0295567373954438,-.0159422728047302,-.00508336149236723,-.000464114759811927},
			{-1.03337249781483e-007,-1.13183342085305e-006,-3.54961912345676e-006,-6.58094122280528e-006,-8.69597991941751e-006,-8.69597991941751e-006,-6.58094122280528e-006,-3.54961912345676e-006,-1.13183342085305e-006,-1.03337249781483e-007,-6.19460350490148e-006,-6.78483246903409e-005,-.000212783706840731,-.000394497837425948,-.000521284897762289,-.000521284897762289,-.000394497837425948,-.000212783706840731,-6.78483246903409e-005,-6.19460350490148e-006,-5.12590331676487e-005,-.000561430529479367,-.00176074014710611,-.00326438612530762,-.0043135222202759,-.0043135222202759,-.00326438612530762,-.00176074014710611,-.000561430529479367,-5.12590331676487e-005,-.000196787192431025,-.00215537303015378,-.00675961072102864,-.0125322180484573,-.0165599285581662,-.0165599285581662,-.0125322180484573,-.00675961072102864,-.00215537303015378,-.000196787192431025,-.000487342632915848,-.00533777200871002,-.0167401467828093,-.0310359839203022,-.0410105916179671,-.0410105916179671,-.0310359839203022,-.0167401467828093,-.00533777200871002,-.000487342632915848,-.000887957326235526,-.00972562924066487,-.0305012017707475,-.0565487758254857,-.0747229009342903,-.0747229009342903,-.0565487758254857,-.0305012017707475,-.00972562924066487,-.000887957326235526,-.00125941435131387,-.0137941280277271,-.0432606952017087,-.080204687454718,-.105981550045219,-.105981550045219,-.080204687454718,-.0432606952017087,-.0137941280277271,-.00125941435131387,-.00140663864321037,-.0154066479494616,-.048317827678703,-.0895805360822827,-.118370688411972,-.118370688411972,-.0895805360822827,-.048317827678703,-.0154066479494616,-.00140663864321037,-.00118342559270231,-.0129618374762797,-.0406504926725559,-.0753654106685247,-.0995869854483788,-.0995869854483788,-.0753654106685247,-.0406504926725559,-.0129618374762797,-.00118342559270231,-.00059135237137032,-.00647697107125416,-.020312865791922,-.0376597519885992,-.0497631624376525,-.0497631624376525,-.0376597519885992,-.020312865791922,-.00647697107125416,-.00059135237137032},
			{-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474138,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512756,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512757,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474139,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916},
			{-1.72916292281226e-007,-9.79398733224426e-006,-7.29759010167018e-005,-.000239119650097383,-.000474634569865839,-.000640675640139909,-.000604924495042971,-.000382283481676095,-.000135370258905769,-1.308068949145e-005,-2.00444673906772e-006,-.000113531962266802,-.000845937099932474,-.00277187647596092,-.00550196689545652,-.00742671601811804,-.00701228852107834,-.00443143250491902,-.00156921288590388,-.000151631434204312,-6.9812104698908e-006,-.000395416105699515,-.00294628179627482,-.00965406199030398,-.019162588931357,-.0258662236377037,-.0244228300443723,-.0154340658681427,-.00546535121888794,-.000528111291459933,-1.51644829240855e-005,-.000858917061539686,-.00639987007723783,-.0209704117690521,-.041624694440043,-.0561862313643539,-.0530509129846532,-.0335256513631973,-.0118717557063213,-.00114715559656894,-2.50006671481639e-005,-.00141603902163566,-.0105510370774588,-.0345725131033195,-.0686238453397381,-.0926304758086336,-.0874614864268777,-.0552714955631914,-.0195721683597038,-.00189123858561784,-3.37466325590279e-005,-.00191141093432354,-.0142420987911849,-.0466669905017148,-.0926304758086336,-.12503532855168,-.118058075331442,-.0746070831112178,-.026419085942947,-.00255284922006524,-3.83630838022064e-005,-.00217288696066628,-.0161903807287985,-.0530509129846532,-.105302083100265,-.142139835110194,-.134208111862577,-.0848131373294056,-.0300331479365882,-.00290207232951189,-3.65709968299352e-005,-.00207138306607571,-.0154340658681427,-.0505727011047881,-.100383018401244,-.135499937544735,-.12793873555065,-.0808511899721722,-.0286301842585129,-.0027665053860117,-2.77050056147299e-005,-.00156921288590388,-.0116923496377051,-.0383122443879716,-.0760469314348448,-.102650374774569,-.0969222524410603,-.0612502492768556,-.0216893025727857,-.0020958150965657,-1.308068949145e-005,-.000740890898631601,-.00552044627469675,-.0180888096370989,-.0359049303476018,-.0484655263124446,-.0457610406806555,-.028918799122657,-.0102404250042771,-.000989521781403059},
			{-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.0551532224231419,-.0216893025727857,-.0022181246897901,-5.68214139596138e-007,-.000166430787646874,-.00294628179627482,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-5.68214139596138e-007,-.000166430787646874,-.00294628179627483,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.055153222423142,-.0216893025727857,-.0022181246897901,-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461163,-.00238829215039022,-.000288622746300124,-4.81475219575139e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461162,-.00238829215039021,-.000288622746300124,-4.81475219575139e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006},
			{-.000189474057272324,-.000379179080671673,-.000450698337824376,-.000403526490267504,-.000284508671025808,-.000156148500303414,-6.30521996263144e-005,-1.64238066106426e-005,-1.98480079913645e-006,-3.31100861878508e-008,-.00219638445415982,-.00439544626910526,-.00522449794427726,-.00467768159306037,-.00329802628977656,-.00181007438983512,-.000730901491494302,-.000190384868710001,-2.30078232481278e-005,-3.8381232568618e-007,-.00764970295714498,-.0153087307912168,-.0181962029908938,-.0162917173481382,-.0114865689446415,-.00630423848900844,-.00254562869914522,-.000663084138302037,-8.01330628639868e-005,-1.33676519027876e-006,-.0166165724938767,-.0332534003485589,-.0395255250831765,-.0353886293208596,-.0249509564807709,-.0136939743226831,-.00552957729977355,-.0014403416335721,-.000174063863093159,-2.90370172750447e-006,-.0273946299483073,-.0548226535530947,-.0651630854416681,-.0583428625235449,-.0411349705180729,-.0225763381244699,-.00911624367502254,-.00237459476467777,-.000286967430772327,-4.7871385229751e-006,-.0369780736440852,-.0740012230234841,-.0879590407639987,-.0787529041667412,-.0555251876749137,-.030474202256271,-.0123053726445112,-.00320529754367204,-.000387357040725739,-6.46181975011305e-006,-.0420365775924714,-.0841243971631835,-.0999916079356561,-.0895260958832346,-.0631208884080524,-.0346429935762571,-.0139887155008144,-.00364377388066238,-.000440346472755273,-7.34577982424566e-006,-.0400728876177358,-.0801946235041213,-.0953206159256855,-.0853439881325168,-.0601722693038774,-.033024686304908,-.0133352488781344,-.0034735592093105,-.000419776197925987,-7.00263023825107e-006,-.0303579249318969,-.0607528557338689,-.0722118188820862,-.0646538480039295,-.0455845671001885,-.0250184353398111,-.010102353697401,-.00263145622867657,-.00031800888486993,-5.30496641835636e-006,-.014333243427598,-.028683958870995,-.0340942136430102,-.030525780139787,-.02152237671891,-.0118122475336369,-.00476974282864777,-.0012424203162536,-.000150145267478098,-2.50469606272866e-006},
			{-3.19136145764326e-006,-6.3866131386048e-006,-7.59123082633588e-006,-6.79670297198979e-006,-4.79205449098542e-006,-2.63005032304177e-006,-1.06200480743303e-006,-2.76630501080463e-007,-3.34305226934559e-008,-5.57681903477072e-010,-.000191307770554455,-.000382848742505824,-.000455060156746716,-.00040743178419295,-.000287262120951629,-.000157659691772648,-6.36624132755433e-005,-1.65827547682883e-005,-2.00400952691774e-006,-3.34305226934559e-008,-.00158303131884398,-.00316799232989985,-.00376552650213992,-.00337141179786177,-.00237703326355233,-.0013046005871693,-.000526792998299499,-.000137218786643261,-1.65827547682883e-005,-2.76630501080463e-007,-.00607737347965237,-.0121621551894091,-.0144561327553171,-.0129430974646002,-.00912560525126807,-.00500845745477193,-.00202239700441925,-.000526792998299499,-6.36624132755433e-005,-1.06200480743303e-006,-.0150505891984046,-.030119524846696,-.0358005503901207,-.0320535250214294,-.0225995220276466,-.0124034232751763,-.00500845745477193,-.0013046005871693,-.000157659691772648,-2.63005032304177e-006,-.0274227618111799,-.0548789515711727,-.0652300021690509,-.0584027754849978,-.0411772124958644,-.0225995220276466,-.00912560525126807,-.00237703326355233,-.000287262120951629,-4.79205449098542e-006,-.0388944589534263,-.0778363296880565,-.0925175100657338,-.0828342662751547,-.0584027754849978,-.0320535250214294,-.0129430974646002,-.00337141179786177,-.00040743178419295,-6.7967029719898e-006,-.0434411827319366,-.0869353196354016,-.103332715477079,-.0925175100657338,-.0652300021690509,-.0358005503901207,-.0144561327553171,-.00376552650213992,-.000455060156746716,-7.59123082633588e-006,-.0365477002003157,-.0731399515169607,-.0869353196354016,-.0778363296880565,-.0548789515711727,-.030119524846696,-.0121621551894091,-.00316799232989985,-.000382848742505824,-6.3866131386048e-006,-.0182627191053361,-.0365477002003157,-.0434411827319366,-.0388944589534263,-.0274227618111799,-.0150505891984046,-.00607737347965237,-.00158303131884398,-.000191307770554455,-3.19136145764326e-006},
			{-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474138,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512757,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512756,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474139,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006},
			{-1.308068949145e-005,-.000135370258905769,-.000382283481676095,-.000604924495042971,-.000640675640139909,-.000474634569865839,-.000239119650097383,-7.29759010167018e-005,-9.79398733224426e-006,-1.72916292281226e-007,-.000151631434204312,-.00156921288590388,-.00443143250491902,-.00701228852107834,-.00742671601811804,-.00550196689545652,-.00277187647596092,-.000845937099932474,-.000113531962266802,-2.00444673906772e-006,-.000528111291459933,-.00546535121888794,-.0154340658681427,-.0244228300443723,-.0258662236377037,-.019162588931357,-.00965406199030398,-.00294628179627483,-.000395416105699515,-6.9812104698908e-006,-.00114715559656894,-.0118717557063213,-.0335256513631973,-.0530509129846532,-.0561862313643539,-.041624694440043,-.0209704117690521,-.00639987007723783,-.000858917061539686,-1.51644829240855e-005,-.00189123858561784,-.0195721683597038,-.0552714955631914,-.0874614864268777,-.0926304758086336,-.0686238453397381,-.0345725131033195,-.0105510370774588,-.00141603902163566,-2.50006671481639e-005,-.00255284922006524,-.026419085942947,-.0746070831112178,-.118058075331442,-.12503532855168,-.0926304758086336,-.0466669905017148,-.0142420987911849,-.00191141093432354,-3.37466325590279e-005,-.00290207232951189,-.0300331479365882,-.0848131373294056,-.134208111862577,-.142139835110194,-.105302083100265,-.0530509129846532,-.0161903807287985,-.00217288696066627,-3.83630838022064e-005,-.0027665053860117,-.0286301842585129,-.0808511899721722,-.12793873555065,-.135499937544735,-.100383018401244,-.0505727011047881,-.0154340658681427,-.00207138306607571,-3.65709968299352e-005,-.0020958150965657,-.0216893025727857,-.0612502492768556,-.0969222524410603,-.102650374774569,-.0760469314348448,-.0383122443879716,-.0116923496377051,-.00156921288590388,-2.77050056147299e-005,-.000989521781403059,-.0102404250042771,-.028918799122657,-.0457610406806555,-.0484655263124446,-.0359049303476018,-.0180888096370989,-.00552044627469675,-.000740890898631601,-1.308068949145e-005},
			{-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005},
			{-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596958,-1.81488157315799e-005,-5.85455199908462e-008,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596959,-1.81488157315799e-005,-5.85455199908462e-008,-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008},
			{-.000255493425333229,-.000483104798924764,-.000517065986523608,-.000395131310801382,-.000223291106959738,-9.07893469973411e-005,-2.44052765295024e-005,-3.59689507657609e-006,-1.82957487178445e-007,-5.90195050823199e-010,-.0029616813806621,-.00560015384355916,-.00599383214211218,-.00458036462030578,-.00258839190514417,-.00105243067689039,-.000282906117811307,-4.16952301711737e-005,-2.12084433297021e-006,-6.8415447117756e-009,-.0103151261942612,-.0195045604772907,-.0208756874850881,-.0159527758058908,-.00901501063419505,-.00366547033510042,-.000985322839048874,-.000145218713843646,-7.38660717382125e-006,-2.38281529966161e-008,-.0224063657308857,-.0423675200132299,-.0453458619764426,-.0346523854772107,-.0195822737922466,-.00796208086622,-.00214030381004998,-.0003154419589321,-1.60450796946238e-005,-5.17591642291991e-008,-.0369398742075263,-.0698484921013427,-.0747586849808788,-.0571290666185282,-.0322839830104985,-.0131265493548066,-.00352857551549082,-.000520048026649204,-2.64524480537304e-005,-8.53318667863116e-008,-.049862523838675,-.0942835398662266,-.10091146198997,-.0771144869130937,-.0435778655721675,-.0177186006765078,-.00476297455073299,-.000701976053853711,-3.57062889348363e-005,-1.15183452383392e-007,-.0566835869406137,-.107181282005842,-.114715886574844,-.0876635474261487,-.0495392038284753,-.0201424589971083,-.00541453703618982,-.000798004545609921,-4.05908160548146e-005,-1.30940247999038e-007,-.0540356789095071,-.102174432709151,-.109357066963176,-.0835684464667091,-.0472250373695059,-.0192015273831553,-.00516160321748368,-.000760726688663539,-3.86946631537408e-005,-1.24823526157934e-007,-.0409356844864075,-.0774040490351568,-.0828453806800583,-.0633087550044376,-.0357761625028231,-.0145464567574012,-.00391026382973348,-.000576301960778015,-2.9313826607454e-005,-9.45622704554012e-008,-.0193274452036918,-.0365456822095551,-.0391147619872944,-.029890705593759,-.0168914195291221,-.00686798936951685,-.00184619875908625,-.000272096209150128,-1.38402810304592e-005,-4.46467946852167e-008},
			{-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007,-.0022181246897901,-.0216893025727857,-.0551532224231419,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627483,-.000166430787646874,-5.68214139596138e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627482,-.000166430787646874,-5.68214139596138e-007,-.0022181246897901,-.0216893025727857,-.055153222423142,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007},
			{-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.2505954733558e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483912,-.00712261469039878,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.064163146746701,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.0641631467467009,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483913,-.00712261469039877,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.25059547335579e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036796,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036795,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166},
			{-1.1349414128987e-006,-1.31562479400023e-005,-4.58213900489777e-005,-9.95325509741368e-005,-.000164092649236315,-.000221497062722267,-.000251797282715359,-.000240034864648708,-.000181842658097301,-8.58555085657902e-005,-1.31562479400023e-005,-.000152507308211391,-.000531161839358935,-.00115378195194428,-.00190216301383433,-.00256759533314912,-.00291883567237296,-.00278248564875054,-.0021079212273054,-.000995237591006242,-4.58213900489777e-005,-.000531161839358935,-.00184996314537334,-.00401846279369377,-.00662497041642304,-.00894257905329049,-.0101659005244202,-.00969101227025514,-.00734159778603745,-.00346627473553786,-9.95325509741368e-005,-.00115378195194428,-.00401846279369377,-.00872884590413951,-.0143906635082426,-.0194249389752361,-.0220822199209436,-.0210506746248672,-.0159473109630529,-.00752939110831236,-.000164092649236315,-.00190216301383433,-.00662497041642304,-.0143906635082426,-.0237249229144092,-.0320245956373459,-.0364054767247764,-.0347048370969986,-.0262912632953882,-.0124132027362624,-.000221497062722267,-.00256759533314912,-.00894257905329049,-.0194249389752361,-.0320245956373459,-.0432277369007862,-.0491411784688113,-.0468456053029506,-.0354887170283886,-.0167557045233543,-.000251797282715359,-.00291883567237296,-.0101659005244202,-.0220822199209436,-.0364054767247764,-.0491411784688113,-.0558635634071245,-.053253961824449,-.0403434808795061,-.0190478411637101,-.000240034864648708,-.00278248564875054,-.00969101227025514,-.0210506746248672,-.0347048370969986,-.0468456053029506,-.053253961824449,-.0507662647534973,-.038458881954326,-.0181580433524764,-.000181842658097301,-.0021079212273054,-.00734159778603745,-.0159473109630529,-.0262912632953882,-.0354887170283886,-.0403434808795061,-.038458881954326,-.0291352063887049,-.0137559469700066,-8.58555085657902e-005,-.000995237591006242,-.00346627473553786,-.00752939110831236,-.0124132027362624,-.0167557045233543,-.0190478411637101,-.0181580433524764,-.0137559469700066,-.0064947567048982},
			{-1.91161171784846e-008,-2.21594149611604e-007,-7.71781742654136e-007,-1.67645297446285e-006,-2.763855715616e-006,-3.73073337316569e-006,-4.24108796005352e-006,-4.04297045415503e-006,-3.06282379048705e-006,-1.44608694643564e-006,-1.14592527597117e-006,-1.32835729492742e-005,-4.62648454277032e-005,-.000100495818239503,-.000165680723448716,-.000223640691794434,-.000254234154648952,-.00024235790093583,-.000183602515332728,-8.66864106188711e-005,-9.48228917027153e-006,-.000109918755228315,-.000382831805844932,-.000831581630087009,-.00137097292696425,-.00185057940015954,-.00210373383142091,-.00200546034506998,-.00151927196238748,-.000717311704224692,-3.64032044368432e-005,-.000421986172976531,-.00146971941509561,-.00319250294337036,-.0052632657411594,-.00710451021065933,-.00807638866214051,-.00769910953153036,-.00583259662818021,-.00275381230681073,-9.01523787074882e-005,-.00104504693649993,-.00363975378962373,-.00790621976356029,-.0130344548969134,-.0175942888806334,-.020001141672057,-.0190668115330659,-.0144444003819494,-.00681980429507682,-.000164261157846817,-.00190411636669871,-.00663177367400153,-.0144054414445077,-.0237492863079749,-.0320574820593633,-.0364428619234804,-.0347404758894987,-.0263182620931215,-.0124259499955376,-.00023297612784248,-.00270066072764841,-.00940602739898747,-.0204316346700341,-.0336842673921381,-.0454680104321141,-.051687915571302,-.0492733745350088,-.0373279165590501,-.0176240673855734,-.000260210806734789,-.00301636529530396,-.01050558269779,-.0228200725526466,-.0376219249308874,-.0507831758761377,-.057730181773567,-.0550333832785173,-.0416915130812311,-.0196843034297875,-.000218919144354553,-.00253771208734517,-.00883849984558959,-.01919885964777,-.0316518737949094,-.0427246260442284,-.0485692433603775,-.0463003874798657,-.0350756776212305,-.016560691379957,-.000109392897999453,-.00126808315618723,-.00441655806269213,-.00959358260487664,-.0158162969791731,-.0213493007781521,-.0242698293952193,-.023136092459402,-.0175271561360877,-.00827530195346473},
			{-2.95097525510867e-010,-3.42077235703851e-009,-1.19140765023158e-008,-2.58795821233052e-008,-4.26659334075081e-008,-5.75917262110692e-008,-6.54701240215422e-008,-6.24117630999618e-008,-4.72811352436055e-008,-2.23233973501177e-008,-9.14787436199951e-008,-1.06042216684182e-006,-3.69330358815301e-006,-8.02253985001058e-006,-1.32262240313143e-005,-1.78531444734238e-005,-2.02954080342344e-005,-1.93473315833786e-005,-1.46569133086574e-005,-6.92014051755746e-006,-1.79844753889302e-006,-2.08476150925998e-005,-7.26093569462482e-005,-.000157720979519105,-.000260024013412071,-.000350988027044924,-.000399002272939181,-.000380363344459719,-.000288150980485938,-.000136048104620829,-1.2202638268856e-005,-.000141453058953237,-.000492661419690163,-.00107015190538498,-.0017642877583389,-.0023814872761676,-.0027072685190056,-.00258080160960999,-.00195513191552442,-.000923099379853643,-4.53946735139408e-005,-.00052621533862221,-.00183273516816672,-.00398104043444918,-.00656327467961113,-.00885930034123408,-.010071229501942,-.00960076369480724,-.00727322838114722,-.00343399468591358,-.000111645553517426,-.00129419595300744,-.0045075053186138,-.00979113689941693,-.0161419915106792,-.0217889327934133,-.0247696019225699,-.023612518692696,-.0178880812574289,-.00844570976740211,-.00019756565546715,-.00229018231092328,-.00797638790562856,-.0173261927444337,-.0285645333188729,-.0385572434695171,-.0438317737278189,-.0417842232474103,-.031654377512867,-.0149453528019069,-.000258532993348771,-.00299691607206422,-.0104378437460552,-.0226729309958482,-.0373793425030134,-.0504557310119577,-.0573579433067165,-.0546785334999813,-.0414226903539633,-.0195573810002261,-.000241552399543637,-.00280007692272149,-.00975228024192592,-.0211837600137409,-.0349242460624195,-.0471417699489713,-.0535906410209485,-.0510872163717606,-.0387020245305973,-.0182728411109243,-.000127746712709587,-.00148084069082919,-.00515756309886556,-.0112031828692115,-.0184699371099763,-.0249312619277241,-.0283417934798407,-.027017839463842,-.0204678422500889,-.00966372260509664},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106448,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{-2.3412366475277e-008,-1.40346610474412e-006,-1.16133850303456e-005,-4.45846379362571e-005,-.000110413663465783,-.000201178010641796,-.000285336317731836,-.000318691851029199,-.000268120099297756,-.000133978390791535,-2.71396298267277e-007,-1.62689878433847e-005,-.000134622431735668,-.00051682540115133,-.00127991587584218,-.00233205675464775,-.00330762037555709,-.00369427792566825,-.0031080498640454,-.00155307834203885,-9.45235731149608e-007,-5.66626321633216e-005,-.000468871290814305,-.00180003131602355,-.00445776978697114,-.00812223079546563,-.0115199838170822,-.0128666585301024,-.0108249073566845,-.00540915683648703,-2.05322718260318e-006,-.000123081737985171,-.00101847533659284,-.00391000160679661,-.00968310210751827,-.0176429905292981,-.0250235397762786,-.0279487668236465,-.0235137048901805,-.0117496910935978,-3.38501811296807e-006,-.000202916616332326,-.0016790920611167,-.00644615772325802,-.0159638817863846,-.0290868166049117,-.0412546337351168,-.0460772596109979,-.0387654701002611,-.0193709286096546,-4.56919656531553e-006,-.000273902790309784,-.00226648762944902,-.00870121244425688,-.0215485150723445,-.0392622367419453,-.0556867125891282,-.062196434207296,-.0523267666298084,-.0261474466358464,-5.19425072819777e-006,-.000311371977038976,-.00257653722080653,-.00989151559332482,-.024496295684836,-.0446332082396283,-.0633045095088931,-.0707047440517022,-.0594849317130106,-.0297243490813528,-4.95160732891562e-006,-.00029682659621248,-.00245617727240442,-.00942944491302682,-.0233519796389196,-.0425482196753798,-.0603473127579275,-.0674018539257125,-.0567061621094263,-.0283358105837029,-3.75117772937632e-006,-.000224866239028424,-.00186072054418367,-.00714344280726179,-.0176907052881247,-.0322331565225002,-.0457171743654439,-.0510614668268089,-.0429587562772,-.0214662945877596,-1.77108757123392e-006,-.000106168736824841,-.000878523830938623,-.00337271749954253,-.00835252033429224,-.0152186185292077,-.0215849861435475,-.0241082496725556,-.0202826218343075,-.0101351336267259},
			{-3.94340655797694e-010,-2.36389493013898e-008,-1.95607303248633e-007,-7.50950801186721e-007,-1.85972641877534e-006,-3.38849422728519e-006,-4.80599476247265e-006,-5.36781079627058e-006,-4.51601746031396e-006,-2.25663332851228e-006,-2.36389493013898e-008,-1.41704872641982e-006,-1.17257783505037e-005,-4.50161241457132e-005,-.000111482237201632,-.000203124993756515,-.000288097777549774,-.000321776122768307,-.00027071494206603,-.000135275021888423,-1.95607303248633e-007,-1.17257783505037e-005,-9.70283345672377e-005,-.000372498901477441,-.000922491922170731,-.00168081634020727,-.00238394814506931,-.00266262932510325,-.00224010885981014,-.00111937218068057,-7.50950801186721e-007,-4.50161241457132e-005,-.000372498901477441,-.00143005063645349,-.00354151423048786,-.00645277735730557,-.0091521519891917,-.0102220295037144,-.00859994241054315,-.0042973520003992,-1.85972641877534e-006,-.000111482237201632,-.000922491922170731,-.00354151423048786,-.00877054471011803,-.0159802752815396,-.0226652649095649,-.0253148119577435,-.0212977202708341,-.0106423736858525,-3.38849422728519e-006,-.000203124993756515,-.00168081634020727,-.00645277735730557,-.0159802752815396,-.0291166861938667,-.0412969985964523,-.0461245768827176,-.0388052788106219,-.0193908208406645,-4.80599476247265e-006,-.000288097777549774,-.00238394814506931,-.0091521519891917,-.0226652649095649,-.0412969985964523,-.0585726714132265,-.0654197587632339,-.0550385965596167,-.0275025356838052,-5.36781079627058e-006,-.000321776122768307,-.00266262932510325,-.0102220295037144,-.0253148119577435,-.0461245768827176,-.0654197587632339,-.0730672638515391,-.0614725540550301,-.0307175549006202,-4.51601746031396e-006,-.00027071494206603,-.00224010885981014,-.00859994241054315,-.0212977202708341,-.0388052788106219,-.0550385965596167,-.0614725540550301,-.0517177557069423,-.025843126655234,-2.25663332851228e-006,-.000135275021888423,-.00111937218068057,-.0042973520003992,-.0106423736858525,-.0193908208406645,-.0275025356838052,-.0307175549006202,-.025843126655234,-.0129136925256951},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166929,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166928,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349},
			{-4.17330922769922e-010,-1.2937047988686e-007,-2.54338890053445e-006,-1.72571365352968e-005,-6.4197762938255e-005,-.000157890655951536,-.000279400029400497,-.000365620865488212,-.000341606679433586,-.000180661133649367,-4.83770266076376e-009,-1.49966341007993e-006,-2.9482980004949e-005,-.000200044834396311,-.000744180868554274,-.00183026946897867,-.00323880688417962,-.00423827935409955,-.00395990675951328,-.0020942249885326,-1.68490485715029e-008,-5.22312002394816e-006,-.000102685137341121,-.00069672846133325,-.00259187893087141,-.00637457515358989,-.01128031595407,-.0147613401865313,-.0137918069811939,-.00729389568268136,-3.65992560246704e-008,-1.13455846597402e-005,-.000223051148290501,-.00151342333828502,-.00563004137434513,-.0138467585932022,-.0245029367616903,-.0320643665107509,-.0299583607113151,-.015843693154235,-6.03387416718265e-008,-1.87047054027123e-005,-.000367729486283348,-.00249507967559062,-.0092818720648124,-.0228282233164573,-.0403963504194766,-.0528623731165144,-.0493903424335445,-.026120435555211,-8.14470002822614e-008,-2.52481590434899e-005,-.000496372028041484,-.00336793160433076,-.0125289426948006,-.0308142042638448,-.0545281766383553,-.071355179091381,-.0666685304112694,-.0352581287427047,-9.25887373147944e-008,-2.87020412938273e-005,-.000564274425767331,-.0038286558562856,-.0142428693503848,-.0350295069709389,-.0619874888642517,-.0811163813282989,-.0757886113425933,-.0400813487182593,-8.82635618211773e-008,-2.73612587189578e-005,-.000537915000325487,-.0036498046378399,-.0135775302253503,-.0333931441745755,-.0590918152054245,-.0773271236407335,-.0722482342515896,-.0382089949930116,-6.68656227010509e-008,-2.07280055821258e-005,-.000407507024584746,-.00276497407096247,-.0102858982181088,-.0252975671172521,-.0447660499839257,-.0585805304843048,-.054732927978495,-.0289459001004882,-3.1570051288489e-008,-9.78655657274711e-006,-.000192401074675968,-.00130545966231248,-.00485640185758369,-.0119440372960602,-.0211359206253737,-.0276583134530104,-.0258416997202825,-.0136665675701473},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.05454958484093e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493428,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493427,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.05454958484094e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036796,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036795,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166},
			{-1.1349414128987e-006,-1.31562479400023e-005,-4.58213900489777e-005,-9.95325509741368e-005,-.000164092649236315,-.000221497062722267,-.000251797282715359,-.000240034864648708,-.000181842658097301,-8.58555085657902e-005,-1.31562479400023e-005,-.000152507308211391,-.000531161839358935,-.00115378195194428,-.00190216301383433,-.00256759533314912,-.00291883567237296,-.00278248564875054,-.0021079212273054,-.000995237591006242,-4.58213900489777e-005,-.000531161839358935,-.00184996314537334,-.00401846279369377,-.00662497041642304,-.00894257905329049,-.0101659005244202,-.00969101227025514,-.00734159778603745,-.00346627473553786,-9.95325509741368e-005,-.00115378195194428,-.00401846279369377,-.00872884590413951,-.0143906635082426,-.0194249389752361,-.0220822199209436,-.0210506746248672,-.0159473109630529,-.00752939110831236,-.000164092649236315,-.00190216301383433,-.00662497041642304,-.0143906635082426,-.0237249229144092,-.0320245956373459,-.0364054767247764,-.0347048370969986,-.0262912632953882,-.0124132027362624,-.000221497062722267,-.00256759533314912,-.00894257905329049,-.0194249389752361,-.0320245956373459,-.0432277369007862,-.0491411784688113,-.0468456053029506,-.0354887170283886,-.0167557045233543,-.000251797282715359,-.00291883567237296,-.0101659005244202,-.0220822199209436,-.0364054767247764,-.0491411784688113,-.0558635634071245,-.053253961824449,-.0403434808795061,-.0190478411637101,-.000240034864648708,-.00278248564875054,-.00969101227025514,-.0210506746248672,-.0347048370969986,-.0468456053029506,-.053253961824449,-.0507662647534973,-.038458881954326,-.0181580433524764,-.000181842658097301,-.0021079212273054,-.00734159778603745,-.0159473109630529,-.0262912632953882,-.0354887170283886,-.0403434808795061,-.038458881954326,-.0291352063887049,-.0137559469700066,-8.58555085657902e-005,-.000995237591006242,-.00346627473553786,-.00752939110831236,-.0124132027362624,-.0167557045233543,-.0190478411637101,-.0181580433524764,-.0137559469700066,-.0064947567048982},
			{-1.91161171784846e-008,-2.21594149611604e-007,-7.71781742654136e-007,-1.67645297446285e-006,-2.763855715616e-006,-3.73073337316569e-006,-4.24108796005352e-006,-4.04297045415503e-006,-3.06282379048705e-006,-1.44608694643564e-006,-1.14592527597117e-006,-1.32835729492742e-005,-4.62648454277032e-005,-.000100495818239503,-.000165680723448716,-.000223640691794434,-.000254234154648952,-.00024235790093583,-.000183602515332728,-8.66864106188711e-005,-9.48228917027153e-006,-.000109918755228315,-.000382831805844932,-.000831581630087009,-.00137097292696425,-.00185057940015954,-.00210373383142091,-.00200546034506998,-.00151927196238748,-.000717311704224692,-3.64032044368432e-005,-.000421986172976531,-.00146971941509561,-.00319250294337036,-.0052632657411594,-.00710451021065933,-.00807638866214051,-.00769910953153036,-.00583259662818021,-.00275381230681073,-9.01523787074882e-005,-.00104504693649993,-.00363975378962373,-.00790621976356029,-.0130344548969134,-.0175942888806334,-.020001141672057,-.0190668115330659,-.0144444003819494,-.00681980429507682,-.000164261157846817,-.00190411636669871,-.00663177367400153,-.0144054414445077,-.0237492863079749,-.0320574820593633,-.0364428619234804,-.0347404758894987,-.0263182620931215,-.0124259499955376,-.00023297612784248,-.00270066072764841,-.00940602739898747,-.0204316346700341,-.0336842673921381,-.0454680104321141,-.051687915571302,-.0492733745350088,-.0373279165590501,-.0176240673855734,-.000260210806734789,-.00301636529530396,-.01050558269779,-.0228200725526466,-.0376219249308874,-.0507831758761377,-.057730181773567,-.0550333832785173,-.0416915130812311,-.0196843034297875,-.000218919144354553,-.00253771208734517,-.00883849984558959,-.01919885964777,-.0316518737949094,-.0427246260442284,-.0485692433603775,-.0463003874798657,-.0350756776212305,-.016560691379957,-.000109392897999453,-.00126808315618723,-.00441655806269213,-.00959358260487664,-.0158162969791731,-.0213493007781521,-.0242698293952193,-.023136092459402,-.0175271561360877,-.00827530195346473},
			{-2.95097525510867e-010,-3.42077235703851e-009,-1.19140765023158e-008,-2.58795821233052e-008,-4.26659334075081e-008,-5.75917262110692e-008,-6.54701240215422e-008,-6.24117630999618e-008,-4.72811352436055e-008,-2.23233973501177e-008,-9.14787436199951e-008,-1.06042216684182e-006,-3.69330358815301e-006,-8.02253985001058e-006,-1.32262240313143e-005,-1.78531444734238e-005,-2.02954080342344e-005,-1.93473315833786e-005,-1.46569133086574e-005,-6.92014051755746e-006,-1.79844753889302e-006,-2.08476150925998e-005,-7.26093569462482e-005,-.000157720979519105,-.000260024013412071,-.000350988027044924,-.000399002272939181,-.000380363344459719,-.000288150980485938,-.000136048104620829,-1.2202638268856e-005,-.000141453058953237,-.000492661419690163,-.00107015190538498,-.0017642877583389,-.0023814872761676,-.0027072685190056,-.00258080160960999,-.00195513191552442,-.000923099379853643,-4.53946735139408e-005,-.00052621533862221,-.00183273516816672,-.00398104043444918,-.00656327467961113,-.00885930034123408,-.010071229501942,-.00960076369480724,-.00727322838114722,-.00343399468591358,-.000111645553517426,-.00129419595300744,-.0045075053186138,-.00979113689941693,-.0161419915106792,-.0217889327934133,-.0247696019225699,-.023612518692696,-.0178880812574289,-.00844570976740211,-.00019756565546715,-.00229018231092328,-.00797638790562856,-.0173261927444337,-.0285645333188729,-.0385572434695171,-.0438317737278189,-.0417842232474103,-.031654377512867,-.0149453528019069,-.000258532993348771,-.00299691607206422,-.0104378437460552,-.0226729309958482,-.0373793425030134,-.0504557310119577,-.0573579433067165,-.0546785334999813,-.0414226903539633,-.0195573810002261,-.000241552399543637,-.00280007692272149,-.00975228024192592,-.0211837600137409,-.0349242460624195,-.0471417699489713,-.0535906410209485,-.0510872163717606,-.0387020245305973,-.0182728411109243,-.000127746712709587,-.00148084069082919,-.00515756309886556,-.0112031828692115,-.0184699371099763,-.0249312619277241,-.0283417934798407,-.027017839463842,-.0204678422500889,-.00966372260509664},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106448,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{-2.3412366475277e-008,-1.40346610474412e-006,-1.16133850303456e-005,-4.45846379362571e-005,-.000110413663465783,-.000201178010641796,-.000285336317731836,-.000318691851029199,-.000268120099297756,-.000133978390791535,-2.71396298267277e-007,-1.62689878433847e-005,-.000134622431735668,-.00051682540115133,-.00127991587584218,-.00233205675464775,-.00330762037555709,-.00369427792566825,-.0031080498640454,-.00155307834203885,-9.45235731149608e-007,-5.66626321633216e-005,-.000468871290814305,-.00180003131602355,-.00445776978697114,-.00812223079546563,-.0115199838170822,-.0128666585301024,-.0108249073566845,-.00540915683648703,-2.05322718260318e-006,-.000123081737985171,-.00101847533659284,-.00391000160679661,-.00968310210751827,-.0176429905292981,-.0250235397762786,-.0279487668236465,-.0235137048901805,-.0117496910935978,-3.38501811296807e-006,-.000202916616332326,-.0016790920611167,-.00644615772325802,-.0159638817863846,-.0290868166049117,-.0412546337351168,-.0460772596109979,-.0387654701002611,-.0193709286096546,-4.56919656531553e-006,-.000273902790309784,-.00226648762944902,-.00870121244425688,-.0215485150723445,-.0392622367419453,-.0556867125891282,-.062196434207296,-.0523267666298084,-.0261474466358464,-5.19425072819777e-006,-.000311371977038976,-.00257653722080653,-.00989151559332482,-.024496295684836,-.0446332082396283,-.0633045095088931,-.0707047440517022,-.0594849317130106,-.0297243490813528,-4.95160732891562e-006,-.00029682659621248,-.00245617727240442,-.00942944491302682,-.0233519796389196,-.0425482196753798,-.0603473127579275,-.0674018539257125,-.0567061621094263,-.0283358105837029,-3.75117772937632e-006,-.000224866239028424,-.00186072054418367,-.00714344280726179,-.0176907052881247,-.0322331565225002,-.0457171743654439,-.0510614668268089,-.0429587562772,-.0214662945877596,-1.77108757123392e-006,-.000106168736824841,-.000878523830938623,-.00337271749954253,-.00835252033429224,-.0152186185292077,-.0215849861435475,-.0241082496725556,-.0202826218343075,-.0101351336267259},
			{-3.94340655797694e-010,-2.36389493013898e-008,-1.95607303248633e-007,-7.50950801186721e-007,-1.85972641877534e-006,-3.38849422728519e-006,-4.80599476247265e-006,-5.36781079627058e-006,-4.51601746031396e-006,-2.25663332851228e-006,-2.36389493013898e-008,-1.41704872641982e-006,-1.17257783505037e-005,-4.50161241457132e-005,-.000111482237201632,-.000203124993756515,-.000288097777549774,-.000321776122768307,-.00027071494206603,-.000135275021888423,-1.95607303248633e-007,-1.17257783505037e-005,-9.70283345672377e-005,-.000372498901477441,-.000922491922170731,-.00168081634020727,-.00238394814506931,-.00266262932510325,-.00224010885981014,-.00111937218068057,-7.50950801186721e-007,-4.50161241457132e-005,-.000372498901477441,-.00143005063645349,-.00354151423048786,-.00645277735730557,-.0091521519891917,-.0102220295037144,-.00859994241054315,-.0042973520003992,-1.85972641877534e-006,-.000111482237201632,-.000922491922170731,-.00354151423048786,-.00877054471011803,-.0159802752815396,-.0226652649095649,-.0253148119577435,-.0212977202708341,-.0106423736858525,-3.38849422728519e-006,-.000203124993756515,-.00168081634020727,-.00645277735730557,-.0159802752815396,-.0291166861938667,-.0412969985964523,-.0461245768827176,-.0388052788106219,-.0193908208406645,-4.80599476247265e-006,-.000288097777549774,-.00238394814506931,-.0091521519891917,-.0226652649095649,-.0412969985964523,-.0585726714132265,-.0654197587632339,-.0550385965596167,-.0275025356838052,-5.36781079627058e-006,-.000321776122768307,-.00266262932510325,-.0102220295037144,-.0253148119577435,-.0461245768827176,-.0654197587632339,-.0730672638515391,-.0614725540550301,-.0307175549006202,-4.51601746031396e-006,-.00027071494206603,-.00224010885981014,-.00859994241054315,-.0212977202708341,-.0388052788106219,-.0550385965596167,-.0614725540550301,-.0517177557069423,-.025843126655234,-2.25663332851228e-006,-.000135275021888423,-.00111937218068057,-.0042973520003992,-.0106423736858525,-.0193908208406645,-.0275025356838052,-.0307175549006202,-.025843126655234,-.0129136925256951},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166929,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166928,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349},
			{-4.17330922769922e-010,-1.2937047988686e-007,-2.54338890053445e-006,-1.72571365352968e-005,-6.4197762938255e-005,-.000157890655951536,-.000279400029400497,-.000365620865488212,-.000341606679433586,-.000180661133649367,-4.83770266076376e-009,-1.49966341007993e-006,-2.9482980004949e-005,-.000200044834396311,-.000744180868554274,-.00183026946897867,-.00323880688417962,-.00423827935409955,-.00395990675951328,-.0020942249885326,-1.68490485715029e-008,-5.22312002394816e-006,-.000102685137341121,-.00069672846133325,-.00259187893087141,-.00637457515358989,-.01128031595407,-.0147613401865313,-.0137918069811939,-.00729389568268136,-3.65992560246704e-008,-1.13455846597402e-005,-.000223051148290501,-.00151342333828502,-.00563004137434513,-.0138467585932022,-.0245029367616903,-.0320643665107509,-.0299583607113151,-.015843693154235,-6.03387416718265e-008,-1.87047054027123e-005,-.000367729486283348,-.00249507967559062,-.0092818720648124,-.0228282233164573,-.0403963504194766,-.0528623731165144,-.0493903424335445,-.026120435555211,-8.14470002822614e-008,-2.52481590434899e-005,-.000496372028041484,-.00336793160433076,-.0125289426948006,-.0308142042638448,-.0545281766383553,-.071355179091381,-.0666685304112694,-.0352581287427047,-9.25887373147944e-008,-2.87020412938273e-005,-.000564274425767331,-.0038286558562856,-.0142428693503848,-.0350295069709389,-.0619874888642517,-.0811163813282989,-.0757886113425933,-.0400813487182593,-8.82635618211773e-008,-2.73612587189578e-005,-.000537915000325487,-.0036498046378399,-.0135775302253503,-.0333931441745755,-.0590918152054245,-.0773271236407335,-.0722482342515896,-.0382089949930116,-6.68656227010509e-008,-2.07280055821258e-005,-.000407507024584746,-.00276497407096247,-.0102858982181088,-.0252975671172521,-.0447660499839257,-.0585805304843048,-.054732927978495,-.0289459001004882,-3.1570051288489e-008,-9.78655657274711e-006,-.000192401074675968,-.00130545966231248,-.00485640185758369,-.0119440372960602,-.0211359206253737,-.0276583134530104,-.0258416997202825,-.0136665675701473},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.05454958484093e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493428,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493427,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.05454958484094e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036796,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036795,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005},
			{-8.58555085657902e-005,-.000181842658097301,-.000240034864648708,-.000251797282715359,-.000221497062722267,-.000164092649236315,-9.95325509741368e-005,-4.58213900489777e-005,-1.31562479400023e-005,-1.1349414128987e-006,-.000995237591006242,-.0021079212273054,-.00278248564875054,-.00291883567237296,-.00256759533314912,-.00190216301383433,-.00115378195194428,-.000531161839358935,-.000152507308211391,-1.31562479400023e-005,-.00346627473553786,-.00734159778603745,-.00969101227025514,-.0101659005244202,-.00894257905329049,-.00662497041642304,-.00401846279369377,-.00184996314537334,-.000531161839358935,-4.58213900489777e-005,-.00752939110831236,-.0159473109630529,-.0210506746248672,-.0220822199209436,-.0194249389752361,-.0143906635082426,-.00872884590413951,-.00401846279369377,-.00115378195194428,-9.95325509741368e-005,-.0124132027362624,-.0262912632953882,-.0347048370969986,-.0364054767247764,-.0320245956373459,-.0237249229144092,-.0143906635082426,-.00662497041642304,-.00190216301383433,-.000164092649236315,-.0167557045233543,-.0354887170283886,-.0468456053029506,-.0491411784688114,-.0432277369007862,-.0320245956373459,-.0194249389752361,-.00894257905329049,-.00256759533314912,-.000221497062722267,-.0190478411637101,-.0403434808795061,-.053253961824449,-.0558635634071245,-.0491411784688113,-.0364054767247764,-.0220822199209436,-.0101659005244202,-.00291883567237296,-.000251797282715359,-.0181580433524764,-.038458881954326,-.0507662647534973,-.053253961824449,-.0468456053029506,-.0347048370969986,-.0210506746248672,-.00969101227025514,-.00278248564875054,-.000240034864648708,-.0137559469700066,-.0291352063887049,-.038458881954326,-.0403434808795061,-.0354887170283886,-.0262912632953882,-.0159473109630529,-.00734159778603745,-.0021079212273054,-.000181842658097301,-.0064947567048982,-.0137559469700066,-.0181580433524764,-.0190478411637101,-.0167557045233543,-.0124132027362624,-.00752939110831236,-.00346627473553786,-.000995237591006242,-8.58555085657902e-005},
			{-1.44608694643564e-006,-3.06282379048705e-006,-4.04297045415503e-006,-4.24108796005352e-006,-3.73073337316569e-006,-2.763855715616e-006,-1.67645297446285e-006,-7.71781742654136e-007,-2.21594149611604e-007,-1.91161171784847e-008,-8.66864106188711e-005,-.000183602515332728,-.00024235790093583,-.000254234154648952,-.000223640691794434,-.000165680723448716,-.000100495818239503,-4.62648454277032e-005,-1.32835729492742e-005,-1.14592527597117e-006,-.000717311704224692,-.00151927196238748,-.00200546034506998,-.00210373383142091,-.00185057940015954,-.00137097292696425,-.000831581630087009,-.000382831805844932,-.000109918755228315,-9.48228917027153e-006,-.00275381230681073,-.00583259662818021,-.00769910953153036,-.00807638866214051,-.00710451021065933,-.0052632657411594,-.00319250294337036,-.00146971941509561,-.000421986172976531,-3.64032044368432e-005,-.00681980429507682,-.0144444003819494,-.0190668115330659,-.020001141672057,-.0175942888806334,-.0130344548969134,-.00790621976356029,-.00363975378962373,-.00104504693649993,-9.01523787074882e-005,-.0124259499955376,-.0263182620931215,-.0347404758894987,-.0364428619234804,-.0320574820593634,-.0237492863079749,-.0144054414445077,-.00663177367400152,-.00190411636669871,-.000164261157846817,-.0176240673855734,-.0373279165590501,-.0492733745350088,-.051687915571302,-.0454680104321141,-.0336842673921381,-.0204316346700341,-.00940602739898747,-.00270066072764841,-.00023297612784248,-.0196843034297875,-.0416915130812311,-.0550333832785173,-.057730181773567,-.0507831758761377,-.0376219249308874,-.0228200725526466,-.01050558269779,-.00301636529530396,-.000260210806734789,-.016560691379957,-.0350756776212305,-.0463003874798657,-.0485692433603775,-.0427246260442284,-.0316518737949094,-.01919885964777,-.00883849984558959,-.00253771208734517,-.000218919144354553,-.00827530195346473,-.0175271561360877,-.023136092459402,-.0242698293952193,-.0213493007781521,-.0158162969791731,-.00959358260487664,-.00441655806269213,-.00126808315618723,-.000109392897999453},
			{-2.23233973501177e-008,-4.72811352436055e-008,-6.24117630999618e-008,-6.54701240215422e-008,-5.75917262110692e-008,-4.26659334075081e-008,-2.58795821233052e-008,-1.19140765023158e-008,-3.42077235703851e-009,-2.95097525510867e-010,-6.92014051755746e-006,-1.46569133086574e-005,-1.93473315833786e-005,-2.02954080342344e-005,-1.78531444734238e-005,-1.32262240313143e-005,-8.02253985001058e-006,-3.69330358815301e-006,-1.06042216684182e-006,-9.14787436199951e-008,-.000136048104620829,-.000288150980485938,-.000380363344459719,-.000399002272939181,-.000350988027044924,-.000260024013412071,-.000157720979519105,-7.26093569462482e-005,-2.08476150925998e-005,-1.79844753889302e-006,-.000923099379853643,-.00195513191552442,-.00258080160960999,-.0027072685190056,-.0023814872761676,-.0017642877583389,-.00107015190538498,-.000492661419690163,-.000141453058953237,-1.2202638268856e-005,-.00343399468591358,-.00727322838114722,-.00960076369480724,-.010071229501942,-.00885930034123408,-.00656327467961113,-.00398104043444918,-.00183273516816672,-.000526215338622209,-4.53946735139408e-005,-.00844570976740211,-.0178880812574289,-.023612518692696,-.0247696019225699,-.0217889327934133,-.0161419915106792,-.00979113689941693,-.0045075053186138,-.00129419595300744,-.000111645553517426,-.0149453528019069,-.031654377512867,-.0417842232474103,-.0438317737278189,-.0385572434695171,-.0285645333188729,-.0173261927444337,-.00797638790562856,-.00229018231092328,-.00019756565546715,-.0195573810002261,-.0414226903539633,-.0546785334999813,-.0573579433067165,-.0504557310119577,-.0373793425030134,-.0226729309958482,-.0104378437460552,-.00299691607206422,-.000258532993348771,-.0182728411109243,-.0387020245305973,-.0510872163717606,-.0535906410209485,-.0471417699489713,-.0349242460624195,-.0211837600137409,-.00975228024192592,-.00280007692272149,-.000241552399543637,-.00966372260509664,-.0204678422500889,-.027017839463842,-.0283417934798407,-.0249312619277241,-.0184699371099763,-.0112031828692115,-.00515756309886556,-.00148084069082919,-.000127746712709587},
			{-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910325,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106448,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252843,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252842,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106449,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910326,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006},
			{-.000133978390791535,-.000268120099297756,-.000318691851029199,-.000285336317731836,-.000201178010641796,-.000110413663465783,-4.45846379362571e-005,-1.16133850303456e-005,-1.40346610474412e-006,-2.3412366475277e-008,-.00155307834203885,-.0031080498640454,-.00369427792566825,-.00330762037555709,-.00233205675464775,-.00127991587584218,-.00051682540115133,-.000134622431735668,-1.62689878433847e-005,-2.71396298267277e-007,-.00540915683648703,-.0108249073566845,-.0128666585301024,-.0115199838170822,-.00812223079546563,-.00445776978697114,-.00180003131602355,-.000468871290814306,-5.66626321633216e-005,-9.45235731149608e-007,-.0117496910935979,-.0235137048901805,-.0279487668236465,-.0250235397762786,-.0176429905292981,-.00968310210751827,-.00391000160679661,-.00101847533659284,-.000123081737985171,-2.05322718260318e-006,-.0193709286096546,-.0387654701002611,-.0460772596109979,-.0412546337351168,-.0290868166049117,-.0159638817863846,-.00644615772325802,-.0016790920611167,-.000202916616332326,-3.38501811296807e-006,-.0261474466358464,-.0523267666298084,-.062196434207296,-.0556867125891282,-.0392622367419453,-.0215485150723445,-.00870121244425688,-.00226648762944902,-.000273902790309784,-4.56919656531553e-006,-.0297243490813528,-.0594849317130106,-.0707047440517022,-.0633045095088931,-.0446332082396283,-.024496295684836,-.00989151559332482,-.00257653722080653,-.000311371977038976,-5.19425072819777e-006,-.0283358105837029,-.0567061621094263,-.0674018539257125,-.0603473127579275,-.0425482196753798,-.0233519796389196,-.00942944491302682,-.00245617727240442,-.00029682659621248,-4.95160732891562e-006,-.0214662945877596,-.0429587562772,-.0510614668268089,-.0457171743654439,-.0322331565225002,-.0176907052881247,-.00714344280726179,-.00186072054418367,-.000224866239028424,-3.75117772937632e-006,-.0101351336267259,-.0202826218343075,-.0241082496725556,-.0215849861435475,-.0152186185292077,-.00835252033429224,-.00337271749954253,-.000878523830938623,-.000106168736824841,-1.77108757123392e-006},
			{-2.25663332851228e-006,-4.51601746031396e-006,-5.36781079627058e-006,-4.80599476247265e-006,-3.38849422728519e-006,-1.85972641877534e-006,-7.50950801186721e-007,-1.95607303248633e-007,-2.36389493013898e-008,-3.94340655797694e-010,-.000135275021888423,-.00027071494206603,-.000321776122768307,-.000288097777549774,-.000203124993756515,-.000111482237201632,-4.50161241457132e-005,-1.17257783505037e-005,-1.41704872641982e-006,-2.36389493013898e-008,-.00111937218068057,-.00224010885981014,-.00266262932510325,-.00238394814506931,-.00168081634020727,-.000922491922170731,-.000372498901477441,-9.70283345672377e-005,-1.17257783505037e-005,-1.95607303248633e-007,-.0042973520003992,-.00859994241054315,-.0102220295037144,-.0091521519891917,-.00645277735730557,-.00354151423048786,-.00143005063645349,-.000372498901477441,-4.50161241457133e-005,-7.50950801186721e-007,-.0106423736858525,-.0212977202708341,-.0253148119577435,-.0226652649095649,-.0159802752815396,-.00877054471011803,-.00354151423048786,-.000922491922170731,-.000111482237201632,-1.85972641877534e-006,-.0193908208406645,-.0388052788106218,-.0461245768827176,-.0412969985964523,-.0291166861938667,-.0159802752815396,-.00645277735730557,-.00168081634020727,-.000203124993756515,-3.38849422728519e-006,-.0275025356838052,-.0550385965596167,-.0654197587632339,-.0585726714132265,-.0412969985964523,-.0226652649095649,-.0091521519891917,-.00238394814506931,-.000288097777549774,-4.80599476247265e-006,-.0307175549006202,-.0614725540550301,-.0730672638515391,-.0654197587632339,-.0461245768827176,-.0253148119577435,-.0102220295037144,-.00266262932510325,-.000321776122768307,-5.36781079627058e-006,-.025843126655234,-.0517177557069423,-.0614725540550301,-.0550385965596167,-.0388052788106219,-.0212977202708341,-.00859994241054315,-.00224010885981014,-.00027071494206603,-4.51601746031396e-006,-.0129136925256951,-.025843126655234,-.0307175549006202,-.0275025356838052,-.0193908208406645,-.0106423736858525,-.0042973520003992,-.00111937218068057,-.000135275021888423,-2.25663332851228e-006},
			{-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166929,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166928,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008},
			{-.000180661133649367,-.000341606679433586,-.000365620865488212,-.000279400029400497,-.000157890655951536,-6.4197762938255e-005,-1.72571365352968e-005,-2.54338890053445e-006,-1.2937047988686e-007,-4.17330922769922e-010,-.0020942249885326,-.00395990675951328,-.00423827935409955,-.00323880688417962,-.00183026946897867,-.000744180868554274,-.000200044834396311,-2.9482980004949e-005,-1.49966341007993e-006,-4.83770266076376e-009,-.00729389568268136,-.0137918069811939,-.0147613401865313,-.01128031595407,-.00637457515358989,-.00259187893087141,-.00069672846133325,-.000102685137341121,-5.22312002394816e-006,-1.68490485715029e-008,-.015843693154235,-.0299583607113152,-.0320643665107509,-.0245029367616903,-.0138467585932022,-.00563004137434513,-.00151342333828502,-.000223051148290501,-1.13455846597402e-005,-3.65992560246704e-008,-.026120435555211,-.0493903424335445,-.0528623731165144,-.0403963504194766,-.0228282233164573,-.0092818720648124,-.00249507967559062,-.000367729486283348,-1.87047054027123e-005,-6.03387416718265e-008,-.0352581287427047,-.0666685304112694,-.071355179091381,-.0545281766383553,-.0308142042638448,-.0125289426948006,-.00336793160433076,-.000496372028041484,-2.52481590434899e-005,-8.14470002822614e-008,-.0400813487182593,-.0757886113425933,-.0811163813282989,-.0619874888642517,-.0350295069709389,-.0142428693503848,-.0038286558562856,-.000564274425767331,-2.87020412938273e-005,-9.25887373147944e-008,-.0382089949930116,-.0722482342515896,-.0773271236407335,-.0590918152054245,-.0333931441745755,-.0135775302253503,-.0036498046378399,-.000537915000325487,-2.73612587189578e-005,-8.82635618211772e-008,-.0289459001004882,-.054732927978495,-.0585805304843048,-.0447660499839257,-.0252975671172521,-.0102858982181088,-.00276497407096247,-.000407507024584746,-2.07280055821258e-005,-6.68656227010509e-008,-.0136665675701473,-.0258416997202825,-.0276583134530104,-.0211359206253737,-.0119440372960602,-.00485640185758369,-.00130545966231248,-.000192401074675968,-9.78655657274711e-006,-3.1570051288489e-008},
			{-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484094e-010,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493428,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493427,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484093e-010},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036796,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036795,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005},
			{-8.58555085657902e-005,-.000181842658097301,-.000240034864648708,-.000251797282715359,-.000221497062722267,-.000164092649236315,-9.95325509741368e-005,-4.58213900489777e-005,-1.31562479400023e-005,-1.1349414128987e-006,-.000995237591006242,-.0021079212273054,-.00278248564875054,-.00291883567237296,-.00256759533314912,-.00190216301383433,-.00115378195194428,-.000531161839358935,-.000152507308211391,-1.31562479400023e-005,-.00346627473553786,-.00734159778603745,-.00969101227025514,-.0101659005244202,-.00894257905329049,-.00662497041642304,-.00401846279369377,-.00184996314537334,-.000531161839358935,-4.58213900489777e-005,-.00752939110831236,-.0159473109630529,-.0210506746248672,-.0220822199209436,-.0194249389752361,-.0143906635082426,-.00872884590413951,-.00401846279369377,-.00115378195194428,-9.95325509741368e-005,-.0124132027362624,-.0262912632953882,-.0347048370969986,-.0364054767247764,-.0320245956373459,-.0237249229144092,-.0143906635082426,-.00662497041642304,-.00190216301383433,-.000164092649236315,-.0167557045233543,-.0354887170283886,-.0468456053029506,-.0491411784688114,-.0432277369007862,-.0320245956373459,-.0194249389752361,-.00894257905329049,-.00256759533314912,-.000221497062722267,-.0190478411637101,-.0403434808795061,-.053253961824449,-.0558635634071245,-.0491411784688113,-.0364054767247764,-.0220822199209436,-.0101659005244202,-.00291883567237296,-.000251797282715359,-.0181580433524764,-.038458881954326,-.0507662647534973,-.053253961824449,-.0468456053029506,-.0347048370969986,-.0210506746248672,-.00969101227025514,-.00278248564875054,-.000240034864648708,-.0137559469700066,-.0291352063887049,-.038458881954326,-.0403434808795061,-.0354887170283886,-.0262912632953882,-.0159473109630529,-.00734159778603745,-.0021079212273054,-.000181842658097301,-.0064947567048982,-.0137559469700066,-.0181580433524764,-.0190478411637101,-.0167557045233543,-.0124132027362624,-.00752939110831236,-.00346627473553786,-.000995237591006242,-8.58555085657902e-005},
			{-1.44608694643564e-006,-3.06282379048705e-006,-4.04297045415503e-006,-4.24108796005352e-006,-3.73073337316569e-006,-2.763855715616e-006,-1.67645297446285e-006,-7.71781742654136e-007,-2.21594149611604e-007,-1.91161171784847e-008,-8.66864106188711e-005,-.000183602515332728,-.00024235790093583,-.000254234154648952,-.000223640691794434,-.000165680723448716,-.000100495818239503,-4.62648454277032e-005,-1.32835729492742e-005,-1.14592527597117e-006,-.000717311704224692,-.00151927196238748,-.00200546034506998,-.00210373383142091,-.00185057940015954,-.00137097292696425,-.000831581630087009,-.000382831805844932,-.000109918755228315,-9.48228917027153e-006,-.00275381230681073,-.00583259662818021,-.00769910953153036,-.00807638866214051,-.00710451021065933,-.0052632657411594,-.00319250294337036,-.00146971941509561,-.000421986172976531,-3.64032044368432e-005,-.00681980429507682,-.0144444003819494,-.0190668115330659,-.020001141672057,-.0175942888806334,-.0130344548969134,-.00790621976356029,-.00363975378962373,-.00104504693649993,-9.01523787074882e-005,-.0124259499955376,-.0263182620931215,-.0347404758894987,-.0364428619234804,-.0320574820593634,-.0237492863079749,-.0144054414445077,-.00663177367400152,-.00190411636669871,-.000164261157846817,-.0176240673855734,-.0373279165590501,-.0492733745350088,-.051687915571302,-.0454680104321141,-.0336842673921381,-.0204316346700341,-.00940602739898747,-.00270066072764841,-.00023297612784248,-.0196843034297875,-.0416915130812311,-.0550333832785173,-.057730181773567,-.0507831758761377,-.0376219249308874,-.0228200725526466,-.01050558269779,-.00301636529530396,-.000260210806734789,-.016560691379957,-.0350756776212305,-.0463003874798657,-.0485692433603775,-.0427246260442284,-.0316518737949094,-.01919885964777,-.00883849984558959,-.00253771208734517,-.000218919144354553,-.00827530195346473,-.0175271561360877,-.023136092459402,-.0242698293952193,-.0213493007781521,-.0158162969791731,-.00959358260487664,-.00441655806269213,-.00126808315618723,-.000109392897999453},
			{-2.23233973501177e-008,-4.72811352436055e-008,-6.24117630999618e-008,-6.54701240215422e-008,-5.75917262110692e-008,-4.26659334075081e-008,-2.58795821233052e-008,-1.19140765023158e-008,-3.42077235703851e-009,-2.95097525510867e-010,-6.92014051755746e-006,-1.46569133086574e-005,-1.93473315833786e-005,-2.02954080342344e-005,-1.78531444734238e-005,-1.32262240313143e-005,-8.02253985001058e-006,-3.69330358815301e-006,-1.06042216684182e-006,-9.14787436199951e-008,-.000136048104620829,-.000288150980485938,-.000380363344459719,-.000399002272939181,-.000350988027044924,-.000260024013412071,-.000157720979519105,-7.26093569462482e-005,-2.08476150925998e-005,-1.79844753889302e-006,-.000923099379853643,-.00195513191552442,-.00258080160960999,-.0027072685190056,-.0023814872761676,-.0017642877583389,-.00107015190538498,-.000492661419690163,-.000141453058953237,-1.2202638268856e-005,-.00343399468591358,-.00727322838114722,-.00960076369480724,-.010071229501942,-.00885930034123408,-.00656327467961113,-.00398104043444918,-.00183273516816672,-.000526215338622209,-4.53946735139408e-005,-.00844570976740211,-.0178880812574289,-.023612518692696,-.0247696019225699,-.0217889327934133,-.0161419915106792,-.00979113689941693,-.0045075053186138,-.00129419595300744,-.000111645553517426,-.0149453528019069,-.031654377512867,-.0417842232474103,-.0438317737278189,-.0385572434695171,-.0285645333188729,-.0173261927444337,-.00797638790562856,-.00229018231092328,-.00019756565546715,-.0195573810002261,-.0414226903539633,-.0546785334999813,-.0573579433067165,-.0504557310119577,-.0373793425030134,-.0226729309958482,-.0104378437460552,-.00299691607206422,-.000258532993348771,-.0182728411109243,-.0387020245305973,-.0510872163717606,-.0535906410209485,-.0471417699489713,-.0349242460624195,-.0211837600137409,-.00975228024192592,-.00280007692272149,-.000241552399543637,-.00966372260509664,-.0204678422500889,-.027017839463842,-.0283417934798407,-.0249312619277241,-.0184699371099763,-.0112031828692115,-.00515756309886556,-.00148084069082919,-.000127746712709587},
			{-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910325,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106448,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252843,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252842,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106449,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910326,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006},
			{-.000133978390791535,-.000268120099297756,-.000318691851029199,-.000285336317731836,-.000201178010641796,-.000110413663465783,-4.45846379362571e-005,-1.16133850303456e-005,-1.40346610474412e-006,-2.3412366475277e-008,-.00155307834203885,-.0031080498640454,-.00369427792566825,-.00330762037555709,-.00233205675464775,-.00127991587584218,-.00051682540115133,-.000134622431735668,-1.62689878433847e-005,-2.71396298267277e-007,-.00540915683648703,-.0108249073566845,-.0128666585301024,-.0115199838170822,-.00812223079546563,-.00445776978697114,-.00180003131602355,-.000468871290814306,-5.66626321633216e-005,-9.45235731149608e-007,-.0117496910935979,-.0235137048901805,-.0279487668236465,-.0250235397762786,-.0176429905292981,-.00968310210751827,-.00391000160679661,-.00101847533659284,-.000123081737985171,-2.05322718260318e-006,-.0193709286096546,-.0387654701002611,-.0460772596109979,-.0412546337351168,-.0290868166049117,-.0159638817863846,-.00644615772325802,-.0016790920611167,-.000202916616332326,-3.38501811296807e-006,-.0261474466358464,-.0523267666298084,-.062196434207296,-.0556867125891282,-.0392622367419453,-.0215485150723445,-.00870121244425688,-.00226648762944902,-.000273902790309784,-4.56919656531553e-006,-.0297243490813528,-.0594849317130106,-.0707047440517022,-.0633045095088931,-.0446332082396283,-.024496295684836,-.00989151559332482,-.00257653722080653,-.000311371977038976,-5.19425072819777e-006,-.0283358105837029,-.0567061621094263,-.0674018539257125,-.0603473127579275,-.0425482196753798,-.0233519796389196,-.00942944491302682,-.00245617727240442,-.00029682659621248,-4.95160732891562e-006,-.0214662945877596,-.0429587562772,-.0510614668268089,-.0457171743654439,-.0322331565225002,-.0176907052881247,-.00714344280726179,-.00186072054418367,-.000224866239028424,-3.75117772937632e-006,-.0101351336267259,-.0202826218343075,-.0241082496725556,-.0215849861435475,-.0152186185292077,-.00835252033429224,-.00337271749954253,-.000878523830938623,-.000106168736824841,-1.77108757123392e-006},
			{-2.25663332851228e-006,-4.51601746031396e-006,-5.36781079627058e-006,-4.80599476247265e-006,-3.38849422728519e-006,-1.85972641877534e-006,-7.50950801186721e-007,-1.95607303248633e-007,-2.36389493013898e-008,-3.94340655797694e-010,-.000135275021888423,-.00027071494206603,-.000321776122768307,-.000288097777549774,-.000203124993756515,-.000111482237201632,-4.50161241457132e-005,-1.17257783505037e-005,-1.41704872641982e-006,-2.36389493013898e-008,-.00111937218068057,-.00224010885981014,-.00266262932510325,-.00238394814506931,-.00168081634020727,-.000922491922170731,-.000372498901477441,-9.70283345672377e-005,-1.17257783505037e-005,-1.95607303248633e-007,-.0042973520003992,-.00859994241054315,-.0102220295037144,-.0091521519891917,-.00645277735730557,-.00354151423048786,-.00143005063645349,-.000372498901477441,-4.50161241457133e-005,-7.50950801186721e-007,-.0106423736858525,-.0212977202708341,-.0253148119577435,-.0226652649095649,-.0159802752815396,-.00877054471011803,-.00354151423048786,-.000922491922170731,-.000111482237201632,-1.85972641877534e-006,-.0193908208406645,-.0388052788106218,-.0461245768827176,-.0412969985964523,-.0291166861938667,-.0159802752815396,-.00645277735730557,-.00168081634020727,-.000203124993756515,-3.38849422728519e-006,-.0275025356838052,-.0550385965596167,-.0654197587632339,-.0585726714132265,-.0412969985964523,-.0226652649095649,-.0091521519891917,-.00238394814506931,-.000288097777549774,-4.80599476247265e-006,-.0307175549006202,-.0614725540550301,-.0730672638515391,-.0654197587632339,-.0461245768827176,-.0253148119577435,-.0102220295037144,-.00266262932510325,-.000321776122768307,-5.36781079627058e-006,-.025843126655234,-.0517177557069423,-.0614725540550301,-.0550385965596167,-.0388052788106219,-.0212977202708341,-.00859994241054315,-.00224010885981014,-.00027071494206603,-4.51601746031396e-006,-.0129136925256951,-.025843126655234,-.0307175549006202,-.0275025356838052,-.0193908208406645,-.0106423736858525,-.0042973520003992,-.00111937218068057,-.000135275021888423,-2.25663332851228e-006},
			{-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166929,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166928,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008},
			{-.000180661133649367,-.000341606679433586,-.000365620865488212,-.000279400029400497,-.000157890655951536,-6.4197762938255e-005,-1.72571365352968e-005,-2.54338890053445e-006,-1.2937047988686e-007,-4.17330922769922e-010,-.0020942249885326,-.00395990675951328,-.00423827935409955,-.00323880688417962,-.00183026946897867,-.000744180868554274,-.000200044834396311,-2.9482980004949e-005,-1.49966341007993e-006,-4.83770266076376e-009,-.00729389568268136,-.0137918069811939,-.0147613401865313,-.01128031595407,-.00637457515358989,-.00259187893087141,-.00069672846133325,-.000102685137341121,-5.22312002394816e-006,-1.68490485715029e-008,-.015843693154235,-.0299583607113152,-.0320643665107509,-.0245029367616903,-.0138467585932022,-.00563004137434513,-.00151342333828502,-.000223051148290501,-1.13455846597402e-005,-3.65992560246704e-008,-.026120435555211,-.0493903424335445,-.0528623731165144,-.0403963504194766,-.0228282233164573,-.0092818720648124,-.00249507967559062,-.000367729486283348,-1.87047054027123e-005,-6.03387416718265e-008,-.0352581287427047,-.0666685304112694,-.071355179091381,-.0545281766383553,-.0308142042638448,-.0125289426948006,-.00336793160433076,-.000496372028041484,-2.52481590434899e-005,-8.14470002822614e-008,-.0400813487182593,-.0757886113425933,-.0811163813282989,-.0619874888642517,-.0350295069709389,-.0142428693503848,-.0038286558562856,-.000564274425767331,-2.87020412938273e-005,-9.25887373147944e-008,-.0382089949930116,-.0722482342515896,-.0773271236407335,-.0590918152054245,-.0333931441745755,-.0135775302253503,-.0036498046378399,-.000537915000325487,-2.73612587189578e-005,-8.82635618211772e-008,-.0289459001004882,-.054732927978495,-.0585805304843048,-.0447660499839257,-.0252975671172521,-.0102858982181088,-.00276497407096247,-.000407507024584746,-2.07280055821258e-005,-6.68656227010509e-008,-.0136665675701473,-.0258416997202825,-.0276583134530104,-.0211359206253737,-.0119440372960602,-.00485640185758369,-.00130545966231248,-.000192401074675968,-9.78655657274711e-006,-3.1570051288489e-008},
			{-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484094e-010,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493428,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493427,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484093e-010},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.,3.354101966},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370350,0.,-7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,7.245688375,2.049390154,0.,13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-3.354101966,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370350,0.,7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,.4082482906,0.,0.,0.,0.,-5.916079782,1.673320053,0.,0.,4.242640686},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,.4082482906,0.,0.,0.,0.,0.,1.673320053,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,-.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,7.245688375,2.049390154,0.,-13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,3.354101966,0.,0.,0.,1.936491673,0.,-10.60660172,-.5000000000,0.,10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.,3.354101966},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370350,0.,-7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,7.245688375,2.049390154,0.,13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-3.354101966,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370350,0.,7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,.4082482906,0.,0.,0.,0.,-5.916079782,1.673320053,0.,0.,4.242640686},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,.4082482906,0.,0.,0.,0.,0.,1.673320053,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,-.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,7.245688375,2.049390154,0.,-13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,3.354101966,0.,0.,0.,1.936491673,0.,-10.60660172,-.5000000000,0.,10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,-13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,.5000000000,0.,3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246899,0.,-5.916079782,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246900,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,.3872983346,0.,2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,-13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,.5000000000,0.,3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246899,0.,-5.916079782,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246900,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,.3872983346,0.,2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,-1.811422094,.7905694150,0.,0.,1.045825033,0.,0.,.8100925875,0.,1.811422094,.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,.7905694150,0.,0.,1.045825033,0.,0.,.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,.8385254914,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,-.8385254914,.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,1.403121520,-.2041241452,0.,0.,-.5400617248,0.,0.,-.4183300132,0.,-1.403121520,-.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,-.2041241452,0.,0.,-.5400617250,0.,0.,-.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.6495190530,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,.6495190530,.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,-1.811422094,.7905694150,0.,0.,1.045825033,0.,0.,.8100925875,0.,1.811422094,.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,.7905694150,0.,0.,1.045825033,0.,0.,.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,.8385254914,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,-.8385254914,.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,1.403121520,-.2041241452,0.,0.,-.5400617248,0.,0.,-.4183300132,0.,-1.403121520,-.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,-.2041241452,0.,0.,-.5400617250,0.,0.,-.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.6495190530,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,.6495190530,.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925875,0.,.6846531970,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,1.811422094,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925875,0.,-1.811422094,-.6846531970},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925872,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,-1.403121520,.2041241452,0.,0.,.5400617248,0.,0.,.4183300132,0.,1.403121520,.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,.2041241452,0.,0.,.5400617250,0.,0.,.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.6495190530,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-.8385254914,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,.8385254914,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925875,0.,.6846531970,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,1.811422094,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925875,0.,-1.811422094,-.6846531970},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925872,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,-1.403121520,.2041241452,0.,0.,.5400617248,0.,0.,.4183300132,0.,1.403121520,.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,.2041241452,0.,0.,.5400617250,0.,0.,.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.6495190530,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-.8385254914,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,.8385254914,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[35],
		T o_output[1]
	) {
		static const T matrix[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[35],
		T o_output[1]
	) {
		static const T matrix[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.414213562,1.224744872,1.054092553,.9354143465,1.118033988,.9682458365,.8333333333,.9165151392,.7937253933,.7745966692,1.414213562,1.224744872,1.054092553,.9354143465,1.369306394,1.185854122,1.020620726,1.183215957,1.024695077,1.024695077,1.118033988,.9682458365,.8333333333,1.183215957,1.024695077,1.080123450,.9165151392,.7937253933,1.024695077,.7745966692},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.825741858,2.108185107,2.041241452,1.932183566,1.443375673,1.666666667,1.613743061,1.183215956,1.366260101,1.,1.825741858,2.108185107,2.041241452,1.932183566,1.767766952,2.041241452,1.976423538,1.527525232,1.763834207,1.322875656,1.443375673,1.666666667,1.613743061,1.527525232,1.763834207,1.394433378,1.183215956,1.366260101,1.322875656,1.},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,2.160246900,2.806243040,2.898275350,2.857738034,1.707825129,2.218529918,2.291287848,1.400000000,1.818653348,1.183215956,2.160246900,2.806243040,2.898275350,2.857738034,2.091650066,2.717133141,2.806243040,1.807392228,2.347871376,1.565247584,1.707825129,2.218529918,2.291287848,1.807392228,2.347871376,1.649915823,1.400000000,1.818653348,1.565247584,1.183215956},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,2.449489743,3.394112549,3.651483716,3.703280398,1.936491673,2.683281572,2.886751347,1.587450787,2.199636334,1.341640786,2.449489743,3.394112549,3.651483716,3.703280398,2.371708244,3.286335344,3.535533905,2.049390154,2.839718296,1.774823935,1.936491673,2.683281572,2.886751347,2.049390154,2.839718296,1.870828693,1.587450787,2.199636334,1.774823935,1.341640786},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,1.732050808,1.500000000,1.290994449,1.145643924,1.643167672,1.423024947,1.224744872,1.496662955,1.296148140,1.355261854,.8660254040,.7500000000,.6454972245,.5728219619,1.341640786,1.161895004,1.,1.449137675,1.254990040,1.434274331,.5477225575,.4743416490,.4082482906,.9660917834,.8366600265,1.133893419,.3741657387,.3240370349,.7171371655,.2710523708},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,2.236067977,2.581988897,2.500000000,2.366431913,2.121320343,2.449489743,2.371708244,1.932183566,2.231093404,1.749635531,1.118033988,1.290994449,1.250000000,1.183215956,1.732050808,2.,1.936491673,1.870828694,2.160246900,1.851640200,.7071067810,.8164965809,.7905694148,1.247219129,1.440164600,1.463850110,.4830458916,.5577733510,.9258201003,.3499271062},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.645751311,3.436931771,3.549647869,3.500000000,2.509980080,3.260559769,3.367491647,2.286190427,2.969848480,2.070196679,1.322875656,1.718465886,1.774823935,1.750000000,2.049390152,2.662235902,2.749545417,2.213594362,2.875543427,2.190890230,.8366600265,1.086853256,1.122497216,1.475729575,1.917028952,1.732050808,.5715476066,.7424621200,1.095445115,.4140393357},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,1.643167673,1.423024947,1.224744872,1.086853256,1.732050808,1.500000000,1.290994449,1.690308509,1.463850109,1.607142857,.5477225576,.4743416490,.4082482906,.3622844188,1.060660172,.9185586536,.7905694150,1.309307341,1.133893419,1.417366774,.2886751347,.2500000000,.2151657415,.6546536709,.5669467095,.8964214566,.1690308509,.1463850109,.4252100321,.1071428571},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,2.121320343,2.449489743,2.371708244,2.244994432,2.236067977,2.581988897,2.500000000,2.182178903,2.519763153,2.074812507,.7071067810,.8164965809,.7905694148,.7483314772,1.369306394,1.581138830,1.530931089,1.690308509,1.951800147,1.829812638,.3726779962,.4303314828,.4166666667,.8451542550,.9759000734,1.157275125,.2182178903,.2519763153,.5489437912,.1383208338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,1.496662955,1.296148140,1.115546702,.9899494934,1.690308509,1.463850109,1.259881577,1.732050808,1.500000000,1.707825127,.3741657387,.3240370349,.2788866754,.2474873734,.8280786708,.7171371654,.6172133998,1.118033988,.9682458365,1.290994449,.1690308509,.1463850109,.1259881577,.4472135954,.3872983346,.6804138176,.8660254040e-1,.7500000000e-1,.2581988897,.4879500362e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.8660254040,.7500000000,.6454972245,.5728219619,.5477225575,.4743416490,.4082482906,.3741657387,.3240370349,.2710523708,1.732050808,1.500000000,1.290994449,1.145643924,1.341640786,1.161895004,1.,.9660917834,.8366600265,.7171371655,1.643167672,1.423024947,1.224744872,1.449137675,1.254990040,1.133893419,1.496662955,1.296148140,1.434274331,1.355261854},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.118033988,1.290994449,1.250000000,1.183215956,.7071067810,.8164965809,.7905694148,.4830458916,.5577733510,.3499271062,2.236067977,2.581988897,2.500000000,2.366431913,1.732050808,2.,1.936491673,1.247219129,1.440164600,.9258201003,2.121320343,2.449489743,2.371708244,1.870828694,2.160246900,1.463850110,1.932183566,2.231093404,1.851640200,1.749635531},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,1.322875656,1.718465886,1.774823935,1.750000000,.8366600265,1.086853256,1.122497216,.5715476066,.7424621200,.4140393357,2.645751311,3.436931771,3.549647869,3.500000000,2.049390152,2.662235902,2.749545417,1.475729575,1.917028952,1.095445115,2.509980080,3.260559769,3.367491647,2.213594362,2.875543427,1.732050808,2.286190427,2.969848480,2.190890230,2.070196679},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,1.341640786,1.161895004,1.,.8874119673,1.060660172,.9185586536,.7905694148,.8280786714,.7171371654,.6561133241,1.341640786,1.161895004,1.,.8874119673,1.732050808,1.500000000,1.290994449,1.603567452,1.388730150,1.388730150,1.060660172,.9185586536,.7905694148,1.603567452,1.388730150,1.646831374,.8280786714,.7171371654,1.388730150,.6561133241},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.069044968,1.234426800,.8470386591,1.732050808,2.,1.936491673,1.833030278,2.236067977,2.581988897,2.500000000,2.070196679,2.390457218,1.792842914,1.369306394,1.581138830,1.530931089,2.070196679,2.390457218,2.126050161,1.069044968,1.234426800,1.792842914,.8470386591},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,1.449137675,1.254990040,1.080123450,.9585144758,1.309307342,1.133893419,.9759000734,1.118033988,.9682458365,.9449111824,.9660917834,.8366600265,.7200822999,.6390096507,1.603567452,1.388730150,1.195228610,1.732050808,1.500000000,1.666666667,.6546536711,.5669467095,.4879500367,1.299038106,1.125000000,1.581138830,.4472135954,.3872983346,1.,.3149703940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.5477225576,.4743416490,.4082482906,.3622844188,.2886751347,.2500000000,.2151657415,.1690308509,.1463850109,.1071428571,1.643167673,1.423024947,1.224744872,1.086853256,1.060660172,.9185586536,.7905694150,.6546536709,.5669467095,.4252100321,1.732050808,1.500000000,1.290994449,1.309307341,1.133893419,.8964214566,1.690308509,1.463850109,1.417366774,1.607142857},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,.7071067810,.8164965809,.7905694148,.7483314772,.3726779962,.4303314828,.4166666667,.2182178903,.2519763153,.1383208338,2.121320343,2.449489743,2.371708244,2.244994432,1.369306394,1.581138830,1.530931089,.8451542550,.9759000734,.5489437912,2.236067977,2.581988897,2.500000000,1.690308509,1.951800147,1.157275125,2.182178903,2.519763153,1.829812638,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.9660917834,.8366600265,.7200822999,.6390096507,.6546536711,.5669467095,.4879500367,.4472135954,.3872983346,.3149703940,1.449137675,1.254990040,1.080123450,.9585144758,1.603567452,1.388730150,1.195228610,1.299038106,1.125000000,1.,1.309307342,1.133893419,.9759000734,1.732050808,1.500000000,1.581138830,1.118033988,.9682458365,1.666666667,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.3741657387,.3240370349,.2788866754,.2474873734,.1690308509,.1463850109,.1259881577,.8660254040e-1,.7500000000e-1,.4879500362e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8280786708,.7171371654,.6172133998,.4472135954,.3872983346,.2581988897,1.690308509,1.463850109,1.259881577,1.118033988,.9682458365,.6804138176,1.732050808,1.500000000,1.290994449,1.707825127},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.414213562,1.224744872,1.054092553,.9354143465,1.118033988,.9682458365,.8333333333,.9165151392,.7937253933,.7745966692,1.414213562,1.224744872,1.054092553,.9354143465,1.369306394,1.185854122,1.020620726,1.183215957,1.024695077,1.024695077,1.118033988,.9682458365,.8333333333,1.183215957,1.024695077,1.080123450,.9165151392,.7937253933,1.024695077,.7745966692},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.825741858,2.108185107,2.041241452,1.932183566,1.443375673,1.666666667,1.613743061,1.183215956,1.366260101,1.,1.825741858,2.108185107,2.041241452,1.932183566,1.767766952,2.041241452,1.976423538,1.527525232,1.763834207,1.322875656,1.443375673,1.666666667,1.613743061,1.527525232,1.763834207,1.394433378,1.183215956,1.366260101,1.322875656,1.},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,2.160246900,2.806243040,2.898275350,2.857738034,1.707825129,2.218529918,2.291287848,1.400000000,1.818653348,1.183215956,2.160246900,2.806243040,2.898275350,2.857738034,2.091650066,2.717133141,2.806243040,1.807392228,2.347871376,1.565247584,1.707825129,2.218529918,2.291287848,1.807392228,2.347871376,1.649915823,1.400000000,1.818653348,1.565247584,1.183215956},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,2.449489743,3.394112549,3.651483716,3.703280398,1.936491673,2.683281572,2.886751347,1.587450787,2.199636334,1.341640786,2.449489743,3.394112549,3.651483716,3.703280398,2.371708244,3.286335344,3.535533905,2.049390154,2.839718296,1.774823935,1.936491673,2.683281572,2.886751347,2.049390154,2.839718296,1.870828693,1.587450787,2.199636334,1.774823935,1.341640786},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,1.732050808,1.500000000,1.290994449,1.145643924,1.643167672,1.423024947,1.224744872,1.496662955,1.296148140,1.355261854,.8660254040,.7500000000,.6454972245,.5728219619,1.341640786,1.161895004,1.,1.449137675,1.254990040,1.434274331,.5477225575,.4743416490,.4082482906,.9660917834,.8366600265,1.133893419,.3741657387,.3240370349,.7171371655,.2710523708},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,2.236067977,2.581988897,2.500000000,2.366431913,2.121320343,2.449489743,2.371708244,1.932183566,2.231093404,1.749635531,1.118033988,1.290994449,1.250000000,1.183215956,1.732050808,2.,1.936491673,1.870828694,2.160246900,1.851640200,.7071067810,.8164965809,.7905694148,1.247219129,1.440164600,1.463850110,.4830458916,.5577733510,.9258201003,.3499271062},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.645751311,3.436931771,3.549647869,3.500000000,2.509980080,3.260559769,3.367491647,2.286190427,2.969848480,2.070196679,1.322875656,1.718465886,1.774823935,1.750000000,2.049390152,2.662235902,2.749545417,2.213594362,2.875543427,2.190890230,.8366600265,1.086853256,1.122497216,1.475729575,1.917028952,1.732050808,.5715476066,.7424621200,1.095445115,.4140393357},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,1.643167673,1.423024947,1.224744872,1.086853256,1.732050808,1.500000000,1.290994449,1.690308509,1.463850109,1.607142857,.5477225576,.4743416490,.4082482906,.3622844188,1.060660172,.9185586536,.7905694150,1.309307341,1.133893419,1.417366774,.2886751347,.2500000000,.2151657415,.6546536709,.5669467095,.8964214566,.1690308509,.1463850109,.4252100321,.1071428571},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,2.121320343,2.449489743,2.371708244,2.244994432,2.236067977,2.581988897,2.500000000,2.182178903,2.519763153,2.074812507,.7071067810,.8164965809,.7905694148,.7483314772,1.369306394,1.581138830,1.530931089,1.690308509,1.951800147,1.829812638,.3726779962,.4303314828,.4166666667,.8451542550,.9759000734,1.157275125,.2182178903,.2519763153,.5489437912,.1383208338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,1.496662955,1.296148140,1.115546702,.9899494934,1.690308509,1.463850109,1.259881577,1.732050808,1.500000000,1.707825127,.3741657387,.3240370349,.2788866754,.2474873734,.8280786708,.7171371654,.6172133998,1.118033988,.9682458365,1.290994449,.1690308509,.1463850109,.1259881577,.4472135954,.3872983346,.6804138176,.8660254040e-1,.7500000000e-1,.2581988897,.4879500362e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.8660254040,.7500000000,.6454972245,.5728219619,.5477225575,.4743416490,.4082482906,.3741657387,.3240370349,.2710523708,1.732050808,1.500000000,1.290994449,1.145643924,1.341640786,1.161895004,1.,.9660917834,.8366600265,.7171371655,1.643167672,1.423024947,1.224744872,1.449137675,1.254990040,1.133893419,1.496662955,1.296148140,1.434274331,1.355261854},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.118033988,1.290994449,1.250000000,1.183215956,.7071067810,.8164965809,.7905694148,.4830458916,.5577733510,.3499271062,2.236067977,2.581988897,2.500000000,2.366431913,1.732050808,2.,1.936491673,1.247219129,1.440164600,.9258201003,2.121320343,2.449489743,2.371708244,1.870828694,2.160246900,1.463850110,1.932183566,2.231093404,1.851640200,1.749635531},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,1.322875656,1.718465886,1.774823935,1.750000000,.8366600265,1.086853256,1.122497216,.5715476066,.7424621200,.4140393357,2.645751311,3.436931771,3.549647869,3.500000000,2.049390152,2.662235902,2.749545417,1.475729575,1.917028952,1.095445115,2.509980080,3.260559769,3.367491647,2.213594362,2.875543427,1.732050808,2.286190427,2.969848480,2.190890230,2.070196679},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,1.341640786,1.161895004,1.,.8874119673,1.060660172,.9185586536,.7905694148,.8280786714,.7171371654,.6561133241,1.341640786,1.161895004,1.,.8874119673,1.732050808,1.500000000,1.290994449,1.603567452,1.388730150,1.388730150,1.060660172,.9185586536,.7905694148,1.603567452,1.388730150,1.646831374,.8280786714,.7171371654,1.388730150,.6561133241},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.069044968,1.234426800,.8470386591,1.732050808,2.,1.936491673,1.833030278,2.236067977,2.581988897,2.500000000,2.070196679,2.390457218,1.792842914,1.369306394,1.581138830,1.530931089,2.070196679,2.390457218,2.126050161,1.069044968,1.234426800,1.792842914,.8470386591},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,1.449137675,1.254990040,1.080123450,.9585144758,1.309307342,1.133893419,.9759000734,1.118033988,.9682458365,.9449111824,.9660917834,.8366600265,.7200822999,.6390096507,1.603567452,1.388730150,1.195228610,1.732050808,1.500000000,1.666666667,.6546536711,.5669467095,.4879500367,1.299038106,1.125000000,1.581138830,.4472135954,.3872983346,1.,.3149703940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.5477225576,.4743416490,.4082482906,.3622844188,.2886751347,.2500000000,.2151657415,.1690308509,.1463850109,.1071428571,1.643167673,1.423024947,1.224744872,1.086853256,1.060660172,.9185586536,.7905694150,.6546536709,.5669467095,.4252100321,1.732050808,1.500000000,1.290994449,1.309307341,1.133893419,.8964214566,1.690308509,1.463850109,1.417366774,1.607142857},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,.7071067810,.8164965809,.7905694148,.7483314772,.3726779962,.4303314828,.4166666667,.2182178903,.2519763153,.1383208338,2.121320343,2.449489743,2.371708244,2.244994432,1.369306394,1.581138830,1.530931089,.8451542550,.9759000734,.5489437912,2.236067977,2.581988897,2.500000000,1.690308509,1.951800147,1.157275125,2.182178903,2.519763153,1.829812638,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.9660917834,.8366600265,.7200822999,.6390096507,.6546536711,.5669467095,.4879500367,.4472135954,.3872983346,.3149703940,1.449137675,1.254990040,1.080123450,.9585144758,1.603567452,1.388730150,1.195228610,1.299038106,1.125000000,1.,1.309307342,1.133893419,.9759000734,1.732050808,1.500000000,1.581138830,1.118033988,.9682458365,1.666666667,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.3741657387,.3240370349,.2788866754,.2474873734,.1690308509,.1463850109,.1259881577,.8660254040e-1,.7500000000e-1,.4879500362e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8280786708,.7171371654,.6172133998,.4472135954,.3872983346,.2581988897,1.690308509,1.463850109,1.259881577,1.118033988,.9682458365,.6804138176,1.732050808,1.500000000,1.290994449,1.707825127},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,-.2397923765,-.4153325792,-.5361920542,-.6344309944,-.4964996319,-.8599625886,-1.110206928,-.5734033834,-.9931637928,-.5309563362,.2624521439e-1,.4545804478e-1,.5868608345e-1,.7424497583e-1,.1285960703,.1087015236,-.2482151352e-2,-.4299212254e-2,-.8901689590e-2,.2178212518e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,1.851946997,3.207666290,4.141079375,4.899791194,3.834529752,6.641600355,8.574269190,4.428467199,7.670330186,4.100643259,1.565440216,2.711421991,3.500430739,4.428467199,7.670330186,6.483686290,1.143425313,1.980470737,4.100643259,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.2397923765,-.4153325792,-.5361920542,-.6344309944,.2624521439e-1,.4545804478e-1,.5868608345e-1,-.2482151352e-2,-.4299212254e-2,.2178212518e-3,1.851946997,3.207666290,4.141079375,4.899791194,-.4964996319,-.8599625886,-1.110206928,.7424497583e-1,.1285960703,-.8901689590e-2,1.565440216,2.711421991,3.500430739,-.5734033834,-.9931637928,.1087015236,1.143425313,1.980470737,-.5309563362,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,.6999612018,1.212368365,1.565160828,1.851923267,1.081581075,1.873353374,2.418488807,.9321846605,1.614591194,.6441723235,.2236285858,.3873360728,.5000487198,.4721128436,.8177234318,.5158410843,.6173679058e-1,.1069312580,.1652303366,.1581446975e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,1.382070207,2.393815818,3.090402932,3.656614062,2.135576909,3.698927711,4.775295141,1.840594370,3.188002963,1.271915321,.8718456225,1.510080914,1.949506078,1.840594370,3.188002963,2.011074701,.4752394227,.8231388257,1.271915321,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.6999612018,1.212368365,1.565160828,1.851923267,.2236285858,.3873360728,.5000487198,.6173679058e-1,.1069312580,.1581446975e-1,1.382070207,2.393815818,3.090402932,3.656614062,1.081581075,1.873353374,2.418488807,.4721128436,.8177234318,.1652303366,.8718456225,1.510080914,1.949506078,.9321846605,1.614591194,.5158410843,.4752394227,.8231388257,.6441723235,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,2.055529306,3.560281195,4.596303257,5.438419356,1.618556373,2.803421873,3.619202076,.7108692826,1.231261715,.2503276966,1.928531457,3.340314468,4.312327437,2.074742821,3.593559979,1.155190024,1.563483000,2.708031992,2.132347336,1.176124927},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,.7042861545,1.219859402,1.574831717,1.863366016,.5545660866,.9605366384,1.240047468,.2435651936,.4218672900,.8576979675e-1,.2264006569,.3921374405,.5062472589,.2435651936,.4218672900,.1356139561,.6288826255e-1,.1089256659,.8576979675e-1,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.055529306,3.560281195,4.596303257,5.438419356,1.928531457,3.340314468,4.312327437,1.563483000,2.708031992,1.176124927,.7042861545,1.219859402,1.574831717,1.863366016,1.618556373,2.803421873,3.619202076,2.074742821,3.593559979,2.132347336,.2264006569,.3921374405,.5062472589,.7108692826,1.231261715,1.155190024,.6288826255e-1,.1089256659,.2503276966,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,2.637214240,4.567789052,5.896990310,6.977413031,1.219036711,2.111433520,2.725848953,.3143011126,.5443854957,.6497290549e-1,3.174462530,5.498330390,7.098314011,2.004818052,3.472446724,.6552866779,3.301857911,5.718985660,2.643567355,3.186694352},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,.4134436882,.7161054738,.9244881913,1.093869180,.1911119037,.3310155272,.4273392081,.4927389259e-1,.8534488540e-1,.1018598993e-1,.7802110799e-1,.1351365231,.1744605012,.4927389259e-1,.8534488540e-1,.1610546421e-1,.1272246435e-1,.2203595465e-1,.1018598993e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.637214240,4.567789052,5.896990310,6.977413031,3.174462530,5.498330390,7.098314011,3.301857911,5.718985660,3.186694352,.4134436882,.7161054738,.9244881913,1.093869180,1.219036711,2.111433520,2.725848953,2.004818052,3.472446724,2.643567355,.7802110799e-1,.1351365231,.1744605012,.3143011126,.5443854957,.6552866779,.1272246435e-1,.2203595465e-1,.6497290549e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,3.239870679,5.611620623,7.244571074,8.571892094,.4061142418,.7034105006,.9080990514,.2839399656e-1,.4917984466e-1,.1591703120e-2,4.791094442,8.298418997,10.71321286,.8205190389,1.421180663,.7272675482e-1,6.122166972,10.60390425,1.329187787,7.258877438},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,.1121154690,.1941896887,.2506978100,.2966296492,.1405355127e-1,.2434146484e-1,.3142469598e-1,.9825720096e-3,.1701864642e-2,.5508076085e-4,.5737338284e-2,.9937361407e-2,.1282907841e-1,.9825720096e-3,.1701864642e-2,.8709032970e-4,.2536990019e-3,.4394195611e-3,.5508076085e-4,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.239870679,5.611620623,7.244571074,8.571892094,4.791094442,8.298418997,10.71321286,6.122166972,10.60390425,7.258877438,.1121154690,.1941896887,.2506978100,.2966296492,.4061142418,.7034105006,.9080990514,.8205190389,1.421180663,1.329187787,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2839399656e-1,.4917984466e-1,.7272675482e-1,.2536990019e-3,.4394195611e-3,.1591703120e-2,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,.1738221700,.3010688299,.3886781879,.4598902340,.2400807184,.4158320022,.5368368063,.1849554605,.3203522547,.1142442288,.1379080812e-1,.2388638034e-1,.3083718442e-1,.2602406990e-1,.4507501127e-1,.2541630033e-1,.9454479976e-3,.1637563968e-2,.2261780150e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,1.235370164,2.139723890,2.762371663,3.268482230,2.838211585,4.915926669,6.346434042,3.637057692,6.299568709,3.736911807,.6965841650,1.206519166,1.557609545,2.186525993,3.787174111,3.552117365,.3394012747,.5878602519,1.350584487,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,2.054909283,3.559207282,4.594916842,5.436778930,.3993492080,.6916931185,.8929719767,.4328832311e-1,.7497757498e-1,.3762235130e-2,1.927368199,3.338299645,4.309726311,.5117504685,.8863778119,.7032397725e-1,1.562068614,2.705582205,.5258003943,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,1.235370164,2.139723890,2.762371663,3.268482230,.2400807184,.4158320022,.5368368063,.2602406990e-1,.4507501127e-1,.2261780150e-2,.6965841650,1.206519166,1.557609545,.1849554605,.3203522547,.2541630033e-1,.3394012747,.5878602519,.1142442288,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,2.054909283,3.559207282,4.594916842,5.436778930,2.838211585,4.915926669,6.346434042,2.186525993,3.787174111,1.350584487,1.927368199,3.338299645,4.309726311,3.637057692,6.299568709,3.552117365,1.562068614,2.705582205,3.736911807,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,.1738221700,.3010688299,.3886781879,.4598902340,.3993492080,.6916931185,.8929719767,.5117504685,.8863778119,.5258003945,.1379080812e-1,.2388638034e-1,.3083718442e-1,.4328832311e-1,.7497757498e-1,.7032397725e-1,.9454479976e-3,.1637563968e-2,.3762235130e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,.7282240147e-1,.1261320992,.1628358399,.1926699641,.4836674308e-1,.8377365642e-1,.1081513254,.1791787294e-1,.3103466628e-1,.5322103273e-2,.2420523895e-2,.4192470366e-2,.5412455971e-2,.2196466622e-2,.3804391785e-2,.1031552789e-2,.6952117983e-4,.1204142157e-3,.7997598703e-4,.1852758219e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.5940552538,1.028933882,1.328347929,1.571722466,1.857843328,3.217879037,4.154263974,3.240776599,5.613189724,4.532588975,.1610768260,.2789932465,.3601787325,.6882539233,1.192090763,1.522004526,.3774006705e-1,.6536771361e-1,.2044304294,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,2.797223960,4.844934019,6.254782921,7.400758958,.2277441565,.3944644502,.5092514154,.1034248755e-1,.1791371391e-1,.3765823900e-3,3.571361895,6.185780254,7.985807971,.3972713534,.6880941682,.2287137395e-1,3.940069325,6.824400256,.5556284741,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,.5940552538,1.028933882,1.328347929,1.571722466,.4836674308e-1,.8377365642e-1,.1081513254,.2196466622e-2,.3804391785e-2,.7997598703e-4,.1610768260,.2789932465,.3601787325,.1791787294e-1,.3103466628e-1,.1031552789e-2,.3774006705e-1,.6536771361e-1,.5322103273e-2,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,2.797223960,4.844934019,6.254782921,7.400758958,1.857843328,3.217879037,4.154263974,.6882539233,1.192090763,.2044304294,3.571361895,6.185780254,7.985807971,3.240776599,5.613189724,1.522004526,3.940069325,6.824400256,4.532588975,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2277441565,.3944644502,.5092514154,.3972713534,.6880941682,.5556284741,.2420523895e-2,.4192470366e-2,.5412455971e-2,.1034248755e-1,.1791371391e-1,.2287137395e-1,.6952117983e-4,.1204142157e-3,.3765823900e-3,.1852758219e-5}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,-.2397923765,-.4153325792,-.5361920542,-.6344309944,-.4964996319,-.8599625886,-1.110206928,-.5734033834,-.9931637928,-.5309563362,.2624521439e-1,.4545804478e-1,.5868608345e-1,.7424497583e-1,.1285960703,.1087015236,-.2482151352e-2,-.4299212254e-2,-.8901689590e-2,.2178212518e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,1.851946997,3.207666290,4.141079375,4.899791194,3.834529752,6.641600355,8.574269190,4.428467199,7.670330186,4.100643259,1.565440216,2.711421991,3.500430739,4.428467199,7.670330186,6.483686290,1.143425313,1.980470737,4.100643259,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.2397923765,-.4153325792,-.5361920542,-.6344309944,.2624521439e-1,.4545804478e-1,.5868608345e-1,-.2482151352e-2,-.4299212254e-2,.2178212518e-3,1.851946997,3.207666290,4.141079375,4.899791194,-.4964996319,-.8599625886,-1.110206928,.7424497583e-1,.1285960703,-.8901689590e-2,1.565440216,2.711421991,3.500430739,-.5734033834,-.9931637928,.1087015236,1.143425313,1.980470737,-.5309563362,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,.6999612018,1.212368365,1.565160828,1.851923267,1.081581075,1.873353374,2.418488807,.9321846605,1.614591194,.6441723235,.2236285858,.3873360728,.5000487198,.4721128436,.8177234318,.5158410843,.6173679058e-1,.1069312580,.1652303366,.1581446975e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,1.382070207,2.393815818,3.090402932,3.656614062,2.135576909,3.698927711,4.775295141,1.840594370,3.188002963,1.271915321,.8718456225,1.510080914,1.949506078,1.840594370,3.188002963,2.011074701,.4752394227,.8231388257,1.271915321,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.6999612018,1.212368365,1.565160828,1.851923267,.2236285858,.3873360728,.5000487198,.6173679058e-1,.1069312580,.1581446975e-1,1.382070207,2.393815818,3.090402932,3.656614062,1.081581075,1.873353374,2.418488807,.4721128436,.8177234318,.1652303366,.8718456225,1.510080914,1.949506078,.9321846605,1.614591194,.5158410843,.4752394227,.8231388257,.6441723235,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,2.055529306,3.560281195,4.596303257,5.438419356,1.618556373,2.803421873,3.619202076,.7108692826,1.231261715,.2503276966,1.928531457,3.340314468,4.312327437,2.074742821,3.593559979,1.155190024,1.563483000,2.708031992,2.132347336,1.176124927},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,.7042861545,1.219859402,1.574831717,1.863366016,.5545660866,.9605366384,1.240047468,.2435651936,.4218672900,.8576979675e-1,.2264006569,.3921374405,.5062472589,.2435651936,.4218672900,.1356139561,.6288826255e-1,.1089256659,.8576979675e-1,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.055529306,3.560281195,4.596303257,5.438419356,1.928531457,3.340314468,4.312327437,1.563483000,2.708031992,1.176124927,.7042861545,1.219859402,1.574831717,1.863366016,1.618556373,2.803421873,3.619202076,2.074742821,3.593559979,2.132347336,.2264006569,.3921374405,.5062472589,.7108692826,1.231261715,1.155190024,.6288826255e-1,.1089256659,.2503276966,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,2.637214240,4.567789052,5.896990310,6.977413031,1.219036711,2.111433520,2.725848953,.3143011126,.5443854957,.6497290549e-1,3.174462530,5.498330390,7.098314011,2.004818052,3.472446724,.6552866779,3.301857911,5.718985660,2.643567355,3.186694352},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,.4134436882,.7161054738,.9244881913,1.093869180,.1911119037,.3310155272,.4273392081,.4927389259e-1,.8534488540e-1,.1018598993e-1,.7802110799e-1,.1351365231,.1744605012,.4927389259e-1,.8534488540e-1,.1610546421e-1,.1272246435e-1,.2203595465e-1,.1018598993e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.637214240,4.567789052,5.896990310,6.977413031,3.174462530,5.498330390,7.098314011,3.301857911,5.718985660,3.186694352,.4134436882,.7161054738,.9244881913,1.093869180,1.219036711,2.111433520,2.725848953,2.004818052,3.472446724,2.643567355,.7802110799e-1,.1351365231,.1744605012,.3143011126,.5443854957,.6552866779,.1272246435e-1,.2203595465e-1,.6497290549e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,3.239870679,5.611620623,7.244571074,8.571892094,.4061142418,.7034105006,.9080990514,.2839399656e-1,.4917984466e-1,.1591703120e-2,4.791094442,8.298418997,10.71321286,.8205190389,1.421180663,.7272675482e-1,6.122166972,10.60390425,1.329187787,7.258877438},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,.1121154690,.1941896887,.2506978100,.2966296492,.1405355127e-1,.2434146484e-1,.3142469598e-1,.9825720096e-3,.1701864642e-2,.5508076085e-4,.5737338284e-2,.9937361407e-2,.1282907841e-1,.9825720096e-3,.1701864642e-2,.8709032970e-4,.2536990019e-3,.4394195611e-3,.5508076085e-4,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.239870679,5.611620623,7.244571074,8.571892094,4.791094442,8.298418997,10.71321286,6.122166972,10.60390425,7.258877438,.1121154690,.1941896887,.2506978100,.2966296492,.4061142418,.7034105006,.9080990514,.8205190389,1.421180663,1.329187787,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2839399656e-1,.4917984466e-1,.7272675482e-1,.2536990019e-3,.4394195611e-3,.1591703120e-2,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,.1738221700,.3010688299,.3886781879,.4598902340,.2400807184,.4158320022,.5368368063,.1849554605,.3203522547,.1142442288,.1379080812e-1,.2388638034e-1,.3083718442e-1,.2602406990e-1,.4507501127e-1,.2541630033e-1,.9454479976e-3,.1637563968e-2,.2261780150e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,1.235370164,2.139723890,2.762371663,3.268482230,2.838211585,4.915926669,6.346434042,3.637057692,6.299568709,3.736911807,.6965841650,1.206519166,1.557609545,2.186525993,3.787174111,3.552117365,.3394012747,.5878602519,1.350584487,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,2.054909283,3.559207282,4.594916842,5.436778930,.3993492080,.6916931185,.8929719767,.4328832311e-1,.7497757498e-1,.3762235130e-2,1.927368199,3.338299645,4.309726311,.5117504685,.8863778119,.7032397725e-1,1.562068614,2.705582205,.5258003943,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,1.235370164,2.139723890,2.762371663,3.268482230,.2400807184,.4158320022,.5368368063,.2602406990e-1,.4507501127e-1,.2261780150e-2,.6965841650,1.206519166,1.557609545,.1849554605,.3203522547,.2541630033e-1,.3394012747,.5878602519,.1142442288,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,2.054909283,3.559207282,4.594916842,5.436778930,2.838211585,4.915926669,6.346434042,2.186525993,3.787174111,1.350584487,1.927368199,3.338299645,4.309726311,3.637057692,6.299568709,3.552117365,1.562068614,2.705582205,3.736911807,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,.1738221700,.3010688299,.3886781879,.4598902340,.3993492080,.6916931185,.8929719767,.5117504685,.8863778119,.5258003945,.1379080812e-1,.2388638034e-1,.3083718442e-1,.4328832311e-1,.7497757498e-1,.7032397725e-1,.9454479976e-3,.1637563968e-2,.3762235130e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,.7282240147e-1,.1261320992,.1628358399,.1926699641,.4836674308e-1,.8377365642e-1,.1081513254,.1791787294e-1,.3103466628e-1,.5322103273e-2,.2420523895e-2,.4192470366e-2,.5412455971e-2,.2196466622e-2,.3804391785e-2,.1031552789e-2,.6952117983e-4,.1204142157e-3,.7997598703e-4,.1852758219e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.5940552538,1.028933882,1.328347929,1.571722466,1.857843328,3.217879037,4.154263974,3.240776599,5.613189724,4.532588975,.1610768260,.2789932465,.3601787325,.6882539233,1.192090763,1.522004526,.3774006705e-1,.6536771361e-1,.2044304294,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,2.797223960,4.844934019,6.254782921,7.400758958,.2277441565,.3944644502,.5092514154,.1034248755e-1,.1791371391e-1,.3765823900e-3,3.571361895,6.185780254,7.985807971,.3972713534,.6880941682,.2287137395e-1,3.940069325,6.824400256,.5556284741,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,.5940552538,1.028933882,1.328347929,1.571722466,.4836674308e-1,.8377365642e-1,.1081513254,.2196466622e-2,.3804391785e-2,.7997598703e-4,.1610768260,.2789932465,.3601787325,.1791787294e-1,.3103466628e-1,.1031552789e-2,.3774006705e-1,.6536771361e-1,.5322103273e-2,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,2.797223960,4.844934019,6.254782921,7.400758958,1.857843328,3.217879037,4.154263974,.6882539233,1.192090763,.2044304294,3.571361895,6.185780254,7.985807971,3.240776599,5.613189724,1.522004526,3.940069325,6.824400256,4.532588975,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2277441565,.3944644502,.5092514154,.3972713534,.6880941682,.5556284741,.2420523895e-2,.4192470366e-2,.5412455971e-2,.1034248755e-1,.1791371391e-1,.2287137395e-1,.6952117983e-4,.1204142157e-3,.3765823900e-3,.1852758219e-5}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,-.2397923765,-0.,-0.,-0.,-.4964996319,-0.,-0.,-.5734033834,-0.,-.5309563362,.2624521439e-1,0.,0.,.7424497583e-1,0.,.1087015236,-.2482151352e-2,-0.,-.8901689590e-2,.2178212518e-3},
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,1.851946997,0.,0.,0.,3.834529752,0.,0.,4.428467199,0.,4.100643259,1.565440216,0.,0.,4.428467199,0.,6.483686290,1.143425313,0.,4.100643259,.7749487345},
			{1.414213562,0.,0.,0.,0.,-.2397923765,-0.,-0.,-0.,.2624521439e-1,0.,0.,-.2482151352e-2,-0.,.2178212518e-3,1.851946997,0.,0.,0.,-.4964996319,-0.,-0.,.7424497583e-1,0.,-.8901689590e-2,1.565440216,0.,0.,-.5734033834,-0.,.1087015236,1.143425313,0.,-.5309563362,.7749487345},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,.6999612018,0.,0.,0.,1.081581075,0.,0.,.9321846605,0.,.6441723235,.2236285858,0.,0.,.4721128436,0.,.5158410843,.6173679058e-1,0.,.1652303366,.1581446975e-1},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,1.382070207,0.,0.,0.,2.135576909,0.,0.,1.840594370,0.,1.271915321,.8718456225,0.,0.,1.840594370,0.,2.011074701,.4752394227,0.,1.271915321,.2403694019},
			{1.414213562,0.,0.,0.,0.,.6999612018,0.,0.,0.,.2236285858,0.,0.,.6173679058e-1,0.,.1581446975e-1,1.382070207,0.,0.,0.,1.081581075,0.,0.,.4721128436,0.,.1652303366,.8718456225,0.,0.,.9321846605,0.,.5158410843,.4752394227,0.,.6441723235,.2403694019},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,2.055529306,0.,0.,0.,1.618556373,0.,0.,.7108692826,0.,.2503276966,1.928531457,0.,0.,2.074742821,0.,1.155190024,1.563483000,0.,2.132347336,1.176124927},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,.7042861545,0.,0.,0.,.5545660866,0.,0.,.2435651936,0.,.8576979675e-1,.2264006569,0.,0.,.2435651936,0.,.1356139561,.6288826255e-1,0.,.8576979675e-1,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,2.055529306,0.,0.,0.,1.928531457,0.,0.,1.563483000,0.,1.176124927,.7042861545,0.,0.,0.,1.618556373,0.,0.,2.074742821,0.,2.132347336,.2264006569,0.,0.,.7108692826,0.,1.155190024,.6288826255e-1,0.,.2503276966,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,2.637214240,0.,0.,0.,1.219036711,0.,0.,.3143011126,0.,.6497290549e-1,3.174462530,0.,0.,2.004818052,0.,.6552866779,3.301857911,0.,2.643567355,3.186694352},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,.4134436882,0.,0.,0.,.1911119037,0.,0.,.4927389259e-1,0.,.1018598993e-1,.7802110799e-1,0.,0.,.4927389259e-1,0.,.1610546421e-1,.1272246435e-1,0.,.1018598993e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,2.637214240,0.,0.,0.,3.174462530,0.,0.,3.301857911,0.,3.186694352,.4134436882,0.,0.,0.,1.219036711,0.,0.,2.004818052,0.,2.643567355,.7802110799e-1,0.,0.,.3143011126,0.,.6552866779,.1272246435e-1,0.,.6497290549e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,3.239870679,0.,0.,0.,.4061142418,0.,0.,.2839399656e-1,0.,.1591703120e-2,4.791094442,0.,0.,.8205190389,0.,.7272675482e-1,6.122166972,0.,1.329187787,7.258877438},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,.1121154690,0.,0.,0.,.1405355127e-1,0.,0.,.9825720096e-3,0.,.5508076085e-4,.5737338284e-2,0.,0.,.9825720096e-3,0.,.8709032970e-4,.2536990019e-3,0.,.5508076085e-4,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,3.239870679,0.,0.,0.,4.791094442,0.,0.,6.122166972,0.,7.258877438,.1121154690,0.,0.,0.,.4061142418,0.,0.,.8205190389,0.,1.329187787,.5737338284e-2,0.,0.,.2839399656e-1,0.,.7272675482e-1,.2536990019e-3,0.,.1591703120e-2,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,.1738221700,0.,0.,0.,.2400807184,0.,0.,.1849554605,0.,.1142442288,.1379080812e-1,0.,0.,.2602406990e-1,0.,.2541630033e-1,.9454479976e-3,0.,.2261780150e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,1.235370164,0.,0.,0.,2.838211585,0.,0.,3.637057692,0.,3.736911807,.6965841650,0.,0.,2.186525993,0.,3.552117365,.3394012747,0.,1.350584487,.1534430404},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,2.054909283,0.,0.,0.,.3993492080,0.,0.,.4328832311e-1,0.,.3762235130e-2,1.927368199,0.,0.,.5117504685,0.,.7032397725e-1,1.562068614,0.,.5258003943,1.174706518},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,1.235370164,0.,0.,0.,.2400807184,0.,0.,.2602406990e-1,0.,.2261780150e-2,.6965841650,0.,0.,.1849554605,0.,.2541630033e-1,.3394012747,0.,.1142442288,.1534430404},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,2.054909283,0.,0.,0.,2.838211585,0.,0.,2.186525993,0.,1.350584487,1.927368199,0.,0.,3.637057692,0.,3.552117365,1.562068614,0.,3.736911807,1.174706518},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,.1738221700,0.,0.,0.,.3993492080,0.,0.,.5117504685,0.,.5258003945,.1379080812e-1,0.,0.,.4328832311e-1,0.,.7032397725e-1,.9454479976e-3,0.,.3762235130e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,.7282240147e-1,0.,0.,0.,.4836674308e-1,0.,0.,.1791787294e-1,0.,.5322103273e-2,.2420523895e-2,0.,0.,.2196466622e-2,0.,.1031552789e-2,.6952117983e-4,0.,.7997598703e-4,.1852758219e-5},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.5940552538,0.,0.,0.,1.857843328,0.,0.,3.240776599,0.,4.532588975,.1610768260,0.,0.,.6882539233,0.,1.522004526,.3774006705e-1,0.,.2044304294,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,2.797223960,0.,0.,0.,.2277441565,0.,0.,.1034248755e-1,0.,.3765823900e-3,3.571361895,0.,0.,.3972713534,0.,.2287137395e-1,3.940069325,0.,.5556284741,4.033366817},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,.5940552538,0.,0.,0.,.4836674308e-1,0.,0.,.2196466622e-2,0.,.7997598703e-4,.1610768260,0.,0.,.1791787294e-1,0.,.1031552789e-2,.3774006705e-1,0.,.5322103273e-2,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,2.797223960,0.,0.,0.,1.857843328,0.,0.,.6882539233,0.,.2044304294,3.571361895,0.,0.,3.240776599,0.,1.522004526,3.940069325,0.,4.532588975,4.033366817},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.7282240147e-1,0.,0.,0.,.2277441565,0.,0.,.3972713534,0.,.5556284741,.2420523895e-2,0.,0.,.1034248755e-1,0.,.2287137395e-1,.6952117983e-4,0.,.3765823900e-3,.1852758219e-5}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,-.2397923765,-0.,-0.,-0.,-.4964996319,-0.,-0.,-.5734033834,-0.,-.5309563362,.2624521439e-1,0.,0.,.7424497583e-1,0.,.1087015236,-.2482151352e-2,-0.,-.8901689590e-2,.2178212518e-3},
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,1.851946997,0.,0.,0.,3.834529752,0.,0.,4.428467199,0.,4.100643259,1.565440216,0.,0.,4.428467199,0.,6.483686290,1.143425313,0.,4.100643259,.7749487345},
			{1.414213562,0.,0.,0.,0.,-.2397923765,-0.,-0.,-0.,.2624521439e-1,0.,0.,-.2482151352e-2,-0.,.2178212518e-3,1.851946997,0.,0.,0.,-.4964996319,-0.,-0.,.7424497583e-1,0.,-.8901689590e-2,1.565440216,0.,0.,-.5734033834,-0.,.1087015236,1.143425313,0.,-.5309563362,.7749487345},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,.6999612018,0.,0.,0.,1.081581075,0.,0.,.9321846605,0.,.6441723235,.2236285858,0.,0.,.4721128436,0.,.5158410843,.6173679058e-1,0.,.1652303366,.1581446975e-1},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,1.382070207,0.,0.,0.,2.135576909,0.,0.,1.840594370,0.,1.271915321,.8718456225,0.,0.,1.840594370,0.,2.011074701,.4752394227,0.,1.271915321,.2403694019},
			{1.414213562,0.,0.,0.,0.,.6999612018,0.,0.,0.,.2236285858,0.,0.,.6173679058e-1,0.,.1581446975e-1,1.382070207,0.,0.,0.,1.081581075,0.,0.,.4721128436,0.,.1652303366,.8718456225,0.,0.,.9321846605,0.,.5158410843,.4752394227,0.,.6441723235,.2403694019},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,2.055529306,0.,0.,0.,1.618556373,0.,0.,.7108692826,0.,.2503276966,1.928531457,0.,0.,2.074742821,0.,1.155190024,1.563483000,0.,2.132347336,1.176124927},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,.7042861545,0.,0.,0.,.5545660866,0.,0.,.2435651936,0.,.8576979675e-1,.2264006569,0.,0.,.2435651936,0.,.1356139561,.6288826255e-1,0.,.8576979675e-1,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,2.055529306,0.,0.,0.,1.928531457,0.,0.,1.563483000,0.,1.176124927,.7042861545,0.,0.,0.,1.618556373,0.,0.,2.074742821,0.,2.132347336,.2264006569,0.,0.,.7108692826,0.,1.155190024,.6288826255e-1,0.,.2503276966,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,2.637214240,0.,0.,0.,1.219036711,0.,0.,.3143011126,0.,.6497290549e-1,3.174462530,0.,0.,2.004818052,0.,.6552866779,3.301857911,0.,2.643567355,3.186694352},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,.4134436882,0.,0.,0.,.1911119037,0.,0.,.4927389259e-1,0.,.1018598993e-1,.7802110799e-1,0.,0.,.4927389259e-1,0.,.1610546421e-1,.1272246435e-1,0.,.1018598993e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,2.637214240,0.,0.,0.,3.174462530,0.,0.,3.301857911,0.,3.186694352,.4134436882,0.,0.,0.,1.219036711,0.,0.,2.004818052,0.,2.643567355,.7802110799e-1,0.,0.,.3143011126,0.,.6552866779,.1272246435e-1,0.,.6497290549e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,3.239870679,0.,0.,0.,.4061142418,0.,0.,.2839399656e-1,0.,.1591703120e-2,4.791094442,0.,0.,.8205190389,0.,.7272675482e-1,6.122166972,0.,1.329187787,7.258877438},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,.1121154690,0.,0.,0.,.1405355127e-1,0.,0.,.9825720096e-3,0.,.5508076085e-4,.5737338284e-2,0.,0.,.9825720096e-3,0.,.8709032970e-4,.2536990019e-3,0.,.5508076085e-4,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,3.239870679,0.,0.,0.,4.791094442,0.,0.,6.122166972,0.,7.258877438,.1121154690,0.,0.,0.,.4061142418,0.,0.,.8205190389,0.,1.329187787,.5737338284e-2,0.,0.,.2839399656e-1,0.,.7272675482e-1,.2536990019e-3,0.,.1591703120e-2,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,.1738221700,0.,0.,0.,.2400807184,0.,0.,.1849554605,0.,.1142442288,.1379080812e-1,0.,0.,.2602406990e-1,0.,.2541630033e-1,.9454479976e-3,0.,.2261780150e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,1.235370164,0.,0.,0.,2.838211585,0.,0.,3.637057692,0.,3.736911807,.6965841650,0.,0.,2.186525993,0.,3.552117365,.3394012747,0.,1.350584487,.1534430404},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,2.054909283,0.,0.,0.,.3993492080,0.,0.,.4328832311e-1,0.,.3762235130e-2,1.927368199,0.,0.,.5117504685,0.,.7032397725e-1,1.562068614,0.,.5258003943,1.174706518},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,1.235370164,0.,0.,0.,.2400807184,0.,0.,.2602406990e-1,0.,.2261780150e-2,.6965841650,0.,0.,.1849554605,0.,.2541630033e-1,.3394012747,0.,.1142442288,.1534430404},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,2.054909283,0.,0.,0.,2.838211585,0.,0.,2.186525993,0.,1.350584487,1.927368199,0.,0.,3.637057692,0.,3.552117365,1.562068614,0.,3.736911807,1.174706518},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,.1738221700,0.,0.,0.,.3993492080,0.,0.,.5117504685,0.,.5258003945,.1379080812e-1,0.,0.,.4328832311e-1,0.,.7032397725e-1,.9454479976e-3,0.,.3762235130e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,.7282240147e-1,0.,0.,0.,.4836674308e-1,0.,0.,.1791787294e-1,0.,.5322103273e-2,.2420523895e-2,0.,0.,.2196466622e-2,0.,.1031552789e-2,.6952117983e-4,0.,.7997598703e-4,.1852758219e-5},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.5940552538,0.,0.,0.,1.857843328,0.,0.,3.240776599,0.,4.532588975,.1610768260,0.,0.,.6882539233,0.,1.522004526,.3774006705e-1,0.,.2044304294,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,2.797223960,0.,0.,0.,.2277441565,0.,0.,.1034248755e-1,0.,.3765823900e-3,3.571361895,0.,0.,.3972713534,0.,.2287137395e-1,3.940069325,0.,.5556284741,4.033366817},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,.5940552538,0.,0.,0.,.4836674308e-1,0.,0.,.2196466622e-2,0.,.7997598703e-4,.1610768260,0.,0.,.1791787294e-1,0.,.1031552789e-2,.3774006705e-1,0.,.5322103273e-2,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,2.797223960,0.,0.,0.,1.857843328,0.,0.,.6882539233,0.,.2044304294,3.571361895,0.,0.,3.240776599,0.,1.522004526,3.940069325,0.,4.532588975,4.033366817},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.7282240147e-1,0.,0.,0.,.2277441565,0.,0.,.3972713534,0.,.5556284741,.2420523895e-2,0.,0.,.1034248755e-1,0.,.2287137395e-1,.6952117983e-4,0.,.3765823900e-3,.1852758219e-5}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-.2973526952e-2,-.2973526952e-2,.3850159296e-3,-.1846817770,-.1846817770,-.9353365550e-1,-.7236569171e-1,-.7236569171e-1,-.2112064808,-.2591194752e-1,-.2591194752e-1,-.1652833479,-.2652578911e-2,-.2652578911e-2,-.7665322818e-1,-.1119869688,-.1862786300,-.1575707306e-1,-.1575707306e-1,-.1119869688,-.1862786300,-.2130681262e-1,-.1003272446,-.2611900580e-2,-.2611900580e-2,-.2130681262e-1,-.1003272446},
			{-.3838806789e-2,-.3838806789e-2,.4970534278e-3,-.2384231489,-.2384231489,-.1207514300,-.9342370632e-1,-.9342370632e-1,-.2726663942,-.3345218040e-1,-.3345218040e-1,-.2133798847,-.3424464647e-2,-.3424464647e-2,-.9895889209e-1,-.1445745550,-.2404846772,-.2034229384e-1,-.2034229384e-1,-.1445745550,-.2404846772,-.2750697681e-1,-.1295219158,-.3371949150e-2,-.3371949150e-2,-.2750697681e-1,-.1295219158},
			{-.4542137448e-2,-.4542137448e-2,.5881215471e-3,-.2821060742,-.2821060742,-.1428750188,-.1105404201,-.1105404201,-.3226232284,-.3958115364e-1,-.3958115364e-1,-.2524744844,-.4051881214e-2,-.4051881214e-2,-.1170897402,-.1710629204,-.2845453076,-.2406932666e-1,-.2406932666e-1,-.1710629204,-.2845453076,-.3254669388e-1,-.1532523976,-.3989744040e-2,-.3989744040e-2,-.3254669388e-1,-.1532523976},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-.2513505346e-2,-.2513505346e-2,-.4213989523e-4,-.1165020402,-.1165020402,-.2988279787e-1,-.2326276051e-1,-.2326276051e-1,-.1981573996,-.4889852992e-2,-.4889852992e-2,-.1989545587,-.1357416837e-3,-.1357416837e-3,-.1133541712,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.5777297190e-2,-.1280930321,-.8681624939e-4,-.8681624939e-4,-.5777297190e-2,-.1280930321},
			{-.3244921449e-2,-.3244921449e-2,-.5440237079e-4,-.1504034872,-.1504034872,-.3857852615e-1,-.3003209470e-1,-.3003209470e-1,-.2558201030,-.6312773071e-2,-.6312773071e-2,-.2568492309,-.1752417601e-3,-.1752417601e-3,-.1463396058,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.7458458603e-2,-.1653673934,-.1120792961e-3,-.1120792961e-3,-.7458458603e-2,-.1653673934},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-.1835908907e-2,-.1835908907e-2,.3985396969e-5,-.6350477757e-1,-.6350477757e-1,-.8249696816e-2,-.6461794816e-2,-.6461794816e-2,-.1606485206,-.7973608935e-3,-.7973608935e-3,-.2069388682,-.6002353001e-5,-.6002353001e-5,-.1448464796,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.1353612365e-2,-.1413173578,-.2493496593e-5,-.2493496593e-5,-.1353612365e-2,-.1413173578},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{.3850159296e-3,-.2973526952e-2,-.2973526952e-2,-.9353365550e-1,-.1846817770,-.1846817770,-.2112064808,-.7236569171e-1,-.7236569171e-1,-.1652833479,-.2591194752e-1,-.2591194752e-1,-.7665322818e-1,-.2652578911e-2,-.2652578911e-2,-.1575707306e-1,-.1119869688,-.1862786300,-.1119869688,-.1862786300,-.1575707306e-1,-.2611900580e-2,-.2130681262e-1,-.1003272446,-.2130681262e-1,-.1003272446,-.2611900580e-2},
			{.4970534278e-3,-.3838806789e-2,-.3838806789e-2,-.1207514300,-.2384231489,-.2384231489,-.2726663942,-.9342370632e-1,-.9342370632e-1,-.2133798847,-.3345218040e-1,-.3345218040e-1,-.9895889209e-1,-.3424464647e-2,-.3424464647e-2,-.2034229384e-1,-.1445745550,-.2404846772,-.1445745550,-.2404846772,-.2034229384e-1,-.3371949150e-2,-.2750697681e-1,-.1295219158,-.2750697681e-1,-.1295219158,-.3371949150e-2},
			{.5881215471e-3,-.4542137448e-2,-.4542137448e-2,-.1428750188,-.2821060742,-.2821060742,-.3226232284,-.1105404201,-.1105404201,-.2524744844,-.3958115364e-1,-.3958115364e-1,-.1170897402,-.4051881214e-2,-.4051881214e-2,-.2406932666e-1,-.1710629204,-.2845453076,-.1710629204,-.2845453076,-.2406932666e-1,-.3989744040e-2,-.3254669388e-1,-.1532523976,-.3254669388e-1,-.1532523976,-.3989744040e-2},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{.7971907625e-3,-.6156805564e-2,.7971907625e-3,-.1445283416,-.2853705526,-.1445283416,-.1663073324,-.5698189327e-1,-.1663073324,-.7640125172e-1,-.1197764475e-1,-.7640125172e-1,-.9608398221e-2,-.3324978618e-3,-.9608398221e-2,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2},
			{.1029168849e-2,-.7948401806e-2,.1029168849e-2,-.1865852866,-.3684117992,-.1865852866,-.2147018429,-.7356330789e-1,-.2147018429,-.9863359184e-1,-.1546307288e-1,-.9863359184e-1,-.1240438876e-1,-.4292528938e-3,-.1240438876e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{.9206691217e-3,-.7110444628e-2,-.1192093711e-3,-.1245649596,-.2459529459,-.6308698245e-1,-.7304211084e-1,-.2502642370e-1,-.2131806773,-.1969833902e-1,-.3088165462e-2,-.1256488891,-.6717834486e-3,-.2324701322e-4,-.1941294556e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-.4213989523e-4,-.2513505346e-2,-.2513505346e-2,-.2988279787e-1,-.1165020402,-.1165020402,-.1981573996,-.2326276051e-1,-.2326276051e-1,-.1989545587,-.4889852992e-2,-.4889852992e-2,-.1133541712,-.1357416837e-3,-.1357416837e-3,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.8681624939e-4,-.5777297190e-2,-.1280930321,-.5777297190e-2,-.1280930321,-.8681624939e-4},
			{-.5440237079e-4,-.3244921449e-2,-.3244921449e-2,-.3857852615e-1,-.1504034872,-.1504034872,-.2558201030,-.3003209470e-1,-.3003209470e-1,-.2568492309,-.6312773071e-2,-.6312773071e-2,-.1463396058,-.1752417601e-3,-.1752417601e-3,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1120792961e-3,-.7458458603e-2,-.1653673934,-.7458458603e-2,-.1653673934,-.1120792961e-3},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-.1192093711e-3,-.7110444628e-2,.9206691217e-3,-.6308698245e-1,-.2459529459,-.1245649596,-.2131806773,-.2502642370e-1,-.7304211084e-1,-.1256488891,-.3088165462e-2,-.1969833902e-1,-.1941294556e-1,-.2324701322e-4,-.6717834486e-3,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{.3985396969e-5,-.1835908907e-2,-.1835908907e-2,-.8249696816e-2,-.6350477757e-1,-.6350477757e-1,-.1606485206,-.6461794816e-2,-.6461794816e-2,-.2069388682,-.7973608935e-3,-.7973608935e-3,-.1448464796,-.6002353001e-5,-.6002353001e-5,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.2493496593e-5,-.1353612365e-2,-.1413173578,-.1353612365e-2,-.1413173578,-.2493496593e-5},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-.2973526952e-2,-.2973526952e-2,.3850159296e-3,-.1846817770,-.1846817770,-.9353365550e-1,-.7236569171e-1,-.7236569171e-1,-.2112064808,-.2591194752e-1,-.2591194752e-1,-.1652833479,-.2652578911e-2,-.2652578911e-2,-.7665322818e-1,-.1119869688,-.1862786300,-.1575707306e-1,-.1575707306e-1,-.1119869688,-.1862786300,-.2130681262e-1,-.1003272446,-.2611900580e-2,-.2611900580e-2,-.2130681262e-1,-.1003272446},
			{-.3838806789e-2,-.3838806789e-2,.4970534278e-3,-.2384231489,-.2384231489,-.1207514300,-.9342370632e-1,-.9342370632e-1,-.2726663942,-.3345218040e-1,-.3345218040e-1,-.2133798847,-.3424464647e-2,-.3424464647e-2,-.9895889209e-1,-.1445745550,-.2404846772,-.2034229384e-1,-.2034229384e-1,-.1445745550,-.2404846772,-.2750697681e-1,-.1295219158,-.3371949150e-2,-.3371949150e-2,-.2750697681e-1,-.1295219158},
			{-.4542137448e-2,-.4542137448e-2,.5881215471e-3,-.2821060742,-.2821060742,-.1428750188,-.1105404201,-.1105404201,-.3226232284,-.3958115364e-1,-.3958115364e-1,-.2524744844,-.4051881214e-2,-.4051881214e-2,-.1170897402,-.1710629204,-.2845453076,-.2406932666e-1,-.2406932666e-1,-.1710629204,-.2845453076,-.3254669388e-1,-.1532523976,-.3989744040e-2,-.3989744040e-2,-.3254669388e-1,-.1532523976},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-.2513505346e-2,-.2513505346e-2,-.4213989523e-4,-.1165020402,-.1165020402,-.2988279787e-1,-.2326276051e-1,-.2326276051e-1,-.1981573996,-.4889852992e-2,-.4889852992e-2,-.1989545587,-.1357416837e-3,-.1357416837e-3,-.1133541712,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.5777297190e-2,-.1280930321,-.8681624939e-4,-.8681624939e-4,-.5777297190e-2,-.1280930321},
			{-.3244921449e-2,-.3244921449e-2,-.5440237079e-4,-.1504034872,-.1504034872,-.3857852615e-1,-.3003209470e-1,-.3003209470e-1,-.2558201030,-.6312773071e-2,-.6312773071e-2,-.2568492309,-.1752417601e-3,-.1752417601e-3,-.1463396058,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.7458458603e-2,-.1653673934,-.1120792961e-3,-.1120792961e-3,-.7458458603e-2,-.1653673934},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-.1835908907e-2,-.1835908907e-2,.3985396969e-5,-.6350477757e-1,-.6350477757e-1,-.8249696816e-2,-.6461794816e-2,-.6461794816e-2,-.1606485206,-.7973608935e-3,-.7973608935e-3,-.2069388682,-.6002353001e-5,-.6002353001e-5,-.1448464796,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.1353612365e-2,-.1413173578,-.2493496593e-5,-.2493496593e-5,-.1353612365e-2,-.1413173578},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{.3850159296e-3,-.2973526952e-2,-.2973526952e-2,-.9353365550e-1,-.1846817770,-.1846817770,-.2112064808,-.7236569171e-1,-.7236569171e-1,-.1652833479,-.2591194752e-1,-.2591194752e-1,-.7665322818e-1,-.2652578911e-2,-.2652578911e-2,-.1575707306e-1,-.1119869688,-.1862786300,-.1119869688,-.1862786300,-.1575707306e-1,-.2611900580e-2,-.2130681262e-1,-.1003272446,-.2130681262e-1,-.1003272446,-.2611900580e-2},
			{.4970534278e-3,-.3838806789e-2,-.3838806789e-2,-.1207514300,-.2384231489,-.2384231489,-.2726663942,-.9342370632e-1,-.9342370632e-1,-.2133798847,-.3345218040e-1,-.3345218040e-1,-.9895889209e-1,-.3424464647e-2,-.3424464647e-2,-.2034229384e-1,-.1445745550,-.2404846772,-.1445745550,-.2404846772,-.2034229384e-1,-.3371949150e-2,-.2750697681e-1,-.1295219158,-.2750697681e-1,-.1295219158,-.3371949150e-2},
			{.5881215471e-3,-.4542137448e-2,-.4542137448e-2,-.1428750188,-.2821060742,-.2821060742,-.3226232284,-.1105404201,-.1105404201,-.2524744844,-.3958115364e-1,-.3958115364e-1,-.1170897402,-.4051881214e-2,-.4051881214e-2,-.2406932666e-1,-.1710629204,-.2845453076,-.1710629204,-.2845453076,-.2406932666e-1,-.3989744040e-2,-.3254669388e-1,-.1532523976,-.3254669388e-1,-.1532523976,-.3989744040e-2},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{.7971907625e-3,-.6156805564e-2,.7971907625e-3,-.1445283416,-.2853705526,-.1445283416,-.1663073324,-.5698189327e-1,-.1663073324,-.7640125172e-1,-.1197764475e-1,-.7640125172e-1,-.9608398221e-2,-.3324978618e-3,-.9608398221e-2,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2},
			{.1029168849e-2,-.7948401806e-2,.1029168849e-2,-.1865852866,-.3684117992,-.1865852866,-.2147018429,-.7356330789e-1,-.2147018429,-.9863359184e-1,-.1546307288e-1,-.9863359184e-1,-.1240438876e-1,-.4292528938e-3,-.1240438876e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{.9206691217e-3,-.7110444628e-2,-.1192093711e-3,-.1245649596,-.2459529459,-.6308698245e-1,-.7304211084e-1,-.2502642370e-1,-.2131806773,-.1969833902e-1,-.3088165462e-2,-.1256488891,-.6717834486e-3,-.2324701322e-4,-.1941294556e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-.4213989523e-4,-.2513505346e-2,-.2513505346e-2,-.2988279787e-1,-.1165020402,-.1165020402,-.1981573996,-.2326276051e-1,-.2326276051e-1,-.1989545587,-.4889852992e-2,-.4889852992e-2,-.1133541712,-.1357416837e-3,-.1357416837e-3,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.8681624939e-4,-.5777297190e-2,-.1280930321,-.5777297190e-2,-.1280930321,-.8681624939e-4},
			{-.5440237079e-4,-.3244921449e-2,-.3244921449e-2,-.3857852615e-1,-.1504034872,-.1504034872,-.2558201030,-.3003209470e-1,-.3003209470e-1,-.2568492309,-.6312773071e-2,-.6312773071e-2,-.1463396058,-.1752417601e-3,-.1752417601e-3,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1120792961e-3,-.7458458603e-2,-.1653673934,-.7458458603e-2,-.1653673934,-.1120792961e-3},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-.1192093711e-3,-.7110444628e-2,.9206691217e-3,-.6308698245e-1,-.2459529459,-.1245649596,-.2131806773,-.2502642370e-1,-.7304211084e-1,-.1256488891,-.3088165462e-2,-.1969833902e-1,-.1941294556e-1,-.2324701322e-4,-.6717834486e-3,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{.3985396969e-5,-.1835908907e-2,-.1835908907e-2,-.8249696816e-2,-.6350477757e-1,-.6350477757e-1,-.1606485206,-.6461794816e-2,-.6461794816e-2,-.2069388682,-.7973608935e-3,-.7973608935e-3,-.1448464796,-.6002353001e-5,-.6002353001e-5,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.2493496593e-5,-.1353612365e-2,-.1413173578,-.1353612365e-2,-.1413173578,-.2493496593e-5},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 148400;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[35][35] = 
		{
			{777.0000,-1510.348301,1690.467392,-952.470474,210.,-1871.410147,1934.644152,-1117.354016,259.229627,2556.16901,-1341.640786,311.769144,-1746.195870,366.6060572,469.574280,-1871.410147,1934.644152,-1117.354016,259.229627,2087.10326,-1095.445113,254.558441,-1352.597504,283.9718296,354.964788,2556.16901,-1341.640786,311.769144,-1352.597504,283.9718296,336.749168,-1746.195870,366.6060572,354.964788,469.574280},
			{-1510.348301,6784.000001,-11711.90164,8798.545336,-2424.871131,1934.644153,-5565.240696,5464.415795,-1795.995545,-1341.640786,2478.709341,-1080.,366.6060559,-423.3202096,0.,1934.644153,-5565.240696,5464.415795,-1795.995545,-1095.445114,2023.857701,-881.8163075,283.9718295,-327.9024243,0.,-1341.640786,2478.709341,-1080.,283.9718295,-327.9024243,0.,366.6060559,-423.3202096,0.,0.},
			{1690.467392,-11711.90164,27216.,-25557.46466,8452.336953,-1117.354017,5464.415795,-7936.346767,3477.930418,311.7691454,-1080.,836.5644027,0.,0.,0.,-1117.354017,5464.415795,-7936.346767,3477.930418,254.5584412,-881.8163075,683.0519743,0.,0.,0.,311.7691454,-1080.,836.5644027,0.,0.,0.,0.,0.,0.,0.},
			{-952.470474,8798.545336,-25557.46466,28800.,-11112.15551,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,259.2296280,-1795.995545,3477.930418,-1959.591794,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{210.,-2424.871131,8452.336953,-11112.15551,4900.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,8128.00002,-5043.731953,1717.300206,-211.6601049,-15027.14343,5126.683137,-763.6753235,12443.02215,-1795.995545,-3834.057901,4064.00002,-2521.865976,858.6501032,-105.8300524,-8179.74082,2790.612835,-415.6921939,7228.74263,-1043.379125,-2318.620276,-5009.04781,1708.894378,-254.5584412,4819.16175,-695.5860830,-1649.727254,3110.755535,-448.9988863,-1159.310137,-766.811579},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,5126.683137,-8196.623692,2645.448922,-1795.995545,2073.837024,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,2790.612835,-4461.676814,1440.,-1043.379125,1204.790437,0.,1708.894378,-2732.207897,881.8163075,-695.5860830,803.1936248,0.,-448.9988863,518.4592560,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,1717.300206,-7250.224824,8640.,-2839.718295,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,858.6501032,-3625.112412,4320.,-1419.859148,-415.6921939,1440.,-1115.419204,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-15027.14343,5126.683137,-763.6753235,0.,33696.,-6734.213542,482.9906830,-31970.48640,2839.718295,10911.92009,-5009.04781,1708.894378,-254.5584412,0.,13756.33440,-2749.231165,197.1801207,-14858.53936,1319.781800,5499.090835,5616.,-1122.368924,80.49844717,-7429.269681,659.8909002,3130.137376,-3197.048640,283.9718295,1649.727251,727.4613394},
			{-1341.640786,2478.709341,-1080.,0.,0.,5126.683137,-8196.623692,2645.448922,0.,-6734.213542,9216.,-1673.128805,2839.718295,-3279.024246,0.,1708.894378,-2732.207897,881.8163075,0.,-2749.231165,3762.416245,-683.0519743,1319.781800,-1523.952755,0.,-1122.368924,1536.,-278.8548009,659.8909002,-761.9763776,0.,283.9718295,-327.9024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,80.49844717,-278.8548009,216.,0.,0.,0.,0.,0.,0.,0.},
			{-1746.195870,366.6060559,0.,0.,0.,12443.02215,-1795.995545,0.,0.,-31970.48640,2839.718295,0.,33600.,-1385.640646,-12423.76754,3110.755535,-448.9988863,0.,0.,-10441.51713,927.4481114,0.,13013.22404,-536.6563145,-5366.563145,-3197.048640,283.9718295,0.,5205.289617,-214.6625258,-2545.584412,1680.,-69.28203232,-1073.312629,-354.9647869},
			{366.6060572,-423.3202096,0.,0.,0.,-1795.995545,2073.837024,0.,0.,2839.718295,-3279.024246,0.,-1385.640646,1600.,0.,-448.9988863,518.4592560,0.,0.,927.4481114,-1070.924834,0.,-536.6563145,619.6773354,0.,283.9718295,-327.9024246,0.,-214.6625258,247.8709341,0.,-69.28203232,80.,0.,0.},
			{469.574280,0.,0.,0.,0.,-3834.057901,0.,0.,0.,10911.92009,0.,0.,-12423.76754,0.,4900.,-766.811579,0.,0.,0.,2969.848480,0.,0.,-4124.318126,0.,1852.025918,727.4613394,0.,0.,-1374.772709,0.,752.9940236,-354.9647869,0.,264.5751311,70.},
			{-1871.410147,1934.644153,-1117.354017,259.2296280,0.,4064.00002,-2521.865976,858.6501032,-105.8300524,-5009.04781,1708.894378,-254.5584412,3110.755535,-448.9988865,-766.811579,8128.00002,-5043.731953,1717.300206,-211.6601049,-8179.74082,2790.612835,-415.6921939,4819.16175,-695.5860830,-1159.310137,-15027.14343,5126.683137,-763.6753235,7228.74263,-1043.379125,-1649.727254,12443.02215,-1795.995545,-2318.620276,-3834.057901},
			{1934.644152,-5565.240696,5464.415795,-1795.995545,0.,-2521.865976,5472.000002,-3625.112412,733.2121114,1708.894378,-2732.207897,881.8163075,-448.9988863,518.4592560,0.,-5043.731953,10944.00000,-7250.224824,1466.424223,2790.612835,-4461.676814,1440.,-695.5860830,803.1936248,0.,5126.683137,-8196.623692,2645.448922,-1043.379125,1204.790437,0.,-1795.995545,2073.837024,0.,0.},
			{-1117.354016,5464.415795,-7936.346767,3477.930418,0.,858.6501032,-3625.112412,4320.,-1419.859148,-254.5584412,881.8163075,-683.0519743,0.,0.,0.,1717.300206,-7250.224824,8640.,-2839.718295,-415.6921939,1440.,-1115.419204,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,0.,0.,0.,0.,0.,0.},
			{259.229627,-1795.995545,3477.930418,-1959.591794,0.,-105.8300524,733.2121114,-1419.859148,800.,0.,0.,0.,0.,0.,0.,-211.6601049,1466.424223,-2839.718295,1600.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2087.10326,-1095.445114,254.5584412,0.,0.,-8179.74082,2790.612835,-415.6921939,0.,13756.33440,-2749.231165,197.1801207,-10441.51713,927.4481114,2969.848480,-8179.74082,2790.612835,-415.6921939,0.,14976.,-2992.983796,214.6625258,-12131.94659,1077.597327,3591.991091,13756.33440,-2749.231165,197.1801207,-12131.94659,1077.597327,3833.619699,-10441.51713,927.4481114,3591.991091,2969.848480},
			{-1095.445113,2023.857701,-881.8163075,0.,0.,2790.612835,-4461.676814,1440.,0.,-2749.231165,3762.416245,-683.0519743,927.4481114,-1070.924834,0.,2790.612835,-4461.676814,1440.,0.,-2992.983796,4096.,-743.6128024,1077.597327,-1244.302214,0.,-2749.231165,3762.416245,-683.0519743,1077.597327,-1244.302214,0.,927.4481114,-1070.924834,0.,0.},
			{254.558441,-881.8163075,683.0519743,0.,0.,-415.6921939,1440.,-1115.419204,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,-415.6921939,1440.,-1115.419204,0.,214.6625258,-743.6128024,576.,0.,0.,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,7228.74263,-1043.379125,0.,0.,-14858.53936,1319.781800,0.,13013.22404,-536.6563145,-4124.318126,4819.16175,-695.5860830,0.,0.,-12131.94659,1077.597327,0.,12096.,-498.8306327,-4156.921939,-7429.269681,659.8909002,0.,9072.,-374.1229745,-3549.242172,5205.289617,-214.6625258,-2494.153164,-1374.772709},
			{283.9718296,-327.9024243,0.,0.,0.,-1043.379125,1204.790437,0.,0.,1319.781800,-1523.952755,0.,-536.6563145,619.6773354,0.,-695.5860830,803.1936248,0.,0.,1077.597327,-1244.302214,0.,-498.8306327,576.,0.,659.8909002,-761.9763776,0.,-374.1229745,432.,0.,-214.6625258,247.8709341,0.,0.},
			{354.964788,0.,0.,0.,0.,-2318.620276,0.,0.,0.,5499.090835,0.,0.,-5366.563145,0.,1852.025918,-1159.310137,0.,0.,0.,3591.991091,0.,0.,-4156.921939,0.,1600.,1649.727251,0.,0.,-2494.153164,0.,1138.419957,-1073.312629,0.,640.,264.5751311},
			{2556.16901,-1341.640786,311.7691454,0.,0.,-5009.04781,1708.894378,-254.5584412,0.,5616.,-1122.368924,80.49844717,-3197.048640,283.9718295,727.4613394,-15027.14343,5126.683137,-763.6753235,0.,13756.33440,-2749.231165,197.1801207,-7429.269681,659.8909002,1649.727251,33696.,-6734.213542,482.9906830,-14858.53936,1319.781800,3130.137376,-31970.48640,2839.718295,5499.090835,10911.92009},
			{-1341.640786,2478.709341,-1080.,0.,0.,1708.894378,-2732.207897,881.8163075,0.,-1122.368924,1536.,-278.8548009,283.9718295,-327.9024246,0.,5126.683137,-8196.623692,2645.448922,0.,-2749.231165,3762.416245,-683.0519743,659.8909002,-761.9763776,0.,-6734.213542,9216.,-1673.128805,1319.781800,-1523.952755,0.,2839.718295,-3279.024246,0.,0.},
			{311.769144,-1080.,836.5644027,0.,0.,-254.5584412,881.8163075,-683.0519743,0.,80.49844717,-278.8548009,216.,0.,0.,0.,-763.6753235,2645.448922,-2049.155923,0.,197.1801207,-683.0519743,529.0897845,0.,0.,0.,482.9906830,-1673.128805,1296.,0.,0.,0.,0.,0.,0.,0.},
			{-1352.597504,283.9718295,0.,0.,0.,4819.16175,-695.5860830,0.,0.,-7429.269681,659.8909002,0.,5205.289617,-214.6625258,-1374.772709,7228.74263,-1043.379125,0.,0.,-12131.94659,1077.597327,0.,9072.,-374.1229745,-2494.153164,-14858.53936,1319.781800,0.,12096.,-498.8306327,-3549.242172,13013.22404,-536.6563145,-4156.921939,-4124.318126},
			{283.9718296,-327.9024243,0.,0.,0.,-695.5860830,803.1936248,0.,0.,659.8909002,-761.9763776,0.,-214.6625258,247.8709341,0.,-1043.379125,1204.790437,0.,0.,1077.597327,-1244.302214,0.,-374.1229745,432.,0.,1319.781800,-1523.952755,0.,-498.8306327,576.,0.,-536.6563145,619.6773354,0.,0.},
			{336.749168,0.,0.,0.,0.,-1649.727254,0.,0.,0.,3130.137376,0.,0.,-2545.584412,0.,752.9940236,-1649.727254,0.,0.,0.,3833.619699,0.,0.,-3549.242172,0.,1138.419957,3130.137376,0.,0.,-3549.242172,0.,1296.,-2545.584412,0.,1138.419957,752.9940236},
			{-1746.195870,366.6060560,0.,0.,0.,3110.755535,-448.9988865,0.,0.,-3197.048640,283.9718295,0.,1680.,-69.28203232,-354.9647869,12443.02215,-1795.995545,0.,0.,-10441.51713,927.4481114,0.,5205.289617,-214.6625258,-1073.312629,-31970.48640,2839.718295,0.,13013.22404,-536.6563145,-2545.584412,33600.,-1385.640646,-5366.563145,-12423.76754},
			{366.6060572,-423.3202096,0.,0.,0.,-448.9988863,518.4592560,0.,0.,283.9718295,-327.9024246,0.,-69.28203232,80.,0.,-1795.995545,2073.837024,0.,0.,927.4481114,-1070.924834,0.,-214.6625258,247.8709341,0.,2839.718295,-3279.024246,0.,-536.6563145,619.6773354,0.,-1385.640646,1600.,0.,0.},
			{354.964788,0.,0.,0.,0.,-1159.310137,0.,0.,0.,1649.727251,0.,0.,-1073.312629,0.,264.5751311,-2318.620276,0.,0.,0.,3591.991091,0.,0.,-2494.153164,0.,640.,5499.090835,0.,0.,-4156.921939,0.,1138.419957,-5366.563145,0.,1600.,1852.025918},
			{469.574280,0.,0.,0.,0.,-766.811579,0.,0.,0.,727.4613394,0.,0.,-354.9647869,0.,70.,-3834.057901,0.,0.,0.,2969.848480,0.,0.,-1374.772709,0.,264.5751311,10911.92009,0.,0.,-4124.318126,0.,752.9940236,-12423.76754,0.,1852.025918,4900.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,3.872983346,3.354101966,2.886751347,2.561737690,3.674234614,3.181980514,2.738612787,3.346640106,2.898275348,3.030457633,1.936491673,1.677050983,1.443375673,1.280868845,3.,2.598076212,2.236067977,3.240370349,2.806243040,3.207134903,1.224744872,1.060660172,.9128709292,2.160246899,1.870828694,2.535462764,.8366600264,.7245688370,1.603567452,.6060915266},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,3.354101966,3.872983346,3.750000000,3.549647869,3.181980514,3.674234614,3.557562366,2.898275350,3.346640106,2.624453295,1.677050983,1.936491673,1.875000000,1.774823935,2.598076212,3.,2.904737510,2.806243040,3.240370350,2.777460300,1.060660172,1.224744872,1.185854122,1.870828694,2.160246900,2.195775165,.7245688375,.8366600264,1.388730150,.5248906592},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,2.886751347,3.750000000,3.872983346,3.818813079,2.738612788,3.557562368,3.674234614,2.494438258,3.240370349,2.258769757,1.443375673,1.875000000,1.936491673,1.909406540,2.236067977,2.904737510,3.,2.415229459,3.137475099,2.390457218,.9128709293,1.185854122,1.224744872,1.610152972,2.091650066,1.889822365,.6236095646,.8100925872,1.195228610,.4517539516},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,5.019960159,4.347413025,3.741657387,3.320391543,5.291502622,4.582575695,3.944053189,5.163977793,4.472135954,4.909902528,1.673320053,1.449137675,1.247219129,1.106797181,3.240370349,2.806243040,2.415229459,4.,3.464101616,4.330127020,.8819171036,.7637626160,.6573421980,2.,1.732050808,2.738612788,.5163977793,.4472135954,1.299038106,.3273268354},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,4.347413025,5.019960159,4.860555525,4.600869484,4.582575696,5.291502622,5.123475385,4.472135954,5.163977793,4.252100321,1.449137675,1.673320053,1.620185175,1.533623161,2.806243040,3.240370350,3.137475099,3.464101616,4.,3.750000000,.7637626162,.8819171036,.8539125641,1.732050808,2.,2.371708245,.4472135954,.5163977793,1.125000000,.2834733546},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,5.796550700,5.019960159,4.320493800,3.834057903,6.546536711,5.669467095,4.879500367,6.708203931,5.809475019,6.614378278,1.449137675,1.254990040,1.080123450,.9585144758,3.207134903,2.777460300,2.390457218,4.330127020,3.750000000,5.,.6546536711,.5669467095,.4879500367,1.732050808,1.500000000,2.635231383,.3354101966,.2904737510,1.,.1889822365},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,1.581138830,1.369306394,1.178511302,1.045825033,1.,.8660254040,.7453559923,.6831300514,.5916079782,.4948716594,3.162277660,2.738612788,2.357022604,2.091650066,2.449489743,2.121320343,1.825741858,1.763834207,1.527525232,1.309307342,3.,2.598076212,2.236067977,2.645751311,2.291287848,2.070196679,2.732520206,2.366431913,2.618614683,2.474358298},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.369306394,1.581138830,1.530931089,1.449137675,.8660254040,1.,.9682458365,.5916079782,.6831300507,.4285714286,2.738612788,3.162277660,3.061862179,2.898275350,2.121320343,2.449489743,2.371708245,1.527525232,1.763834207,1.133893419,2.598076212,3.,2.904737510,2.291287848,2.645751311,1.792842914,2.366431913,2.732520203,2.267786838,2.142857143},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,1.178511302,1.530931089,1.581138830,1.559023911,.7453559923,.9682458365,1.,.5091750773,.6614378278,.3688555568,2.357022604,3.061862179,3.162277659,3.118047822,1.825741858,2.371708244,2.449489743,1.314684396,1.707825129,.9759000734,2.236067977,2.904737510,3.,1.972026594,2.561737692,1.543033500,2.036700309,2.645751311,1.951800147,1.844277784},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,4.582575696,3.968626966,3.415650256,3.031088914,4.140393357,3.585685828,3.086067000,3.535533905,3.061862179,2.988071523,3.055050464,2.645751311,2.277100171,2.020725943,5.070925527,4.391550330,3.779644731,5.477225575,4.743416490,5.270462768,2.070196679,1.792842914,1.543033500,4.107919181,3.557562368,5.,1.414213562,1.224744872,3.162277660,.9960238407},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,1.296148140,1.122497216,.9660917834,.8573214100,.6831300514,.5916079782,.5091750773,.4000000000,.3464101616,.2535462764,3.888444420,3.367491648,2.898275350,2.571964230,2.509980079,2.173706512,1.870828694,1.549193338,1.341640786,1.006230590,4.098780308,3.549647869,3.055050464,3.098386677,2.683281572,2.121320343,4.,3.464101616,3.354101966,3.803194146},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,1.122497216,1.296148140,1.254990040,1.187939392,.5916079782,.6831300507,.6614378278,.3464101616,.4000000000,.2195775163,3.367491648,3.888444419,3.764970119,3.563818176,2.173706511,2.509980079,2.430277762,1.341640786,1.549193338,.8714212528,3.549647869,4.098780304,3.968626966,2.683281572,3.098386677,1.837117307,3.464101616,4.,2.904737510,3.293662744},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,2.898275350,2.509980080,2.160246900,1.917028952,1.963961013,1.700840129,1.463850110,1.341640786,1.161895004,.9449111824,4.347413025,3.764970119,3.240370350,2.875543427,4.810702356,4.166190450,3.585685828,3.897114318,3.375000000,3.,3.927922025,3.401680258,2.927700220,5.196152424,4.500000000,4.743416490,3.354101966,2.904737510,5.,2.834733546},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,1.095445115,.9486832980,.8164965809,.7245688375,.4948716594,.4285714286,.3688555568,.2535462764,.2195775163,.1428571429,4.381780461,3.794733192,3.265986323,2.898275350,2.424366106,2.099562637,1.807015806,1.309307341,1.133893419,.7559289460,4.948716594,4.285714286,3.688555568,3.273268354,2.834733546,1.992047682,5.070925527,4.391550326,3.779644731,5.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.121320343,1.825741858,1.620185175,1.469693846,2.,1.732050808,1.490711985,1.322875656,1.581138830,1.369306394,1.178511302,1.296148140,1.122497216,1.095445115,2.,1.732050808,1.490711985,1.322875656,1.936491673,1.677050983,1.443375673,1.673320053,1.449137675,1.449137675,1.581138830,1.369306394,1.178511302,1.673320053,1.449137675,1.527525232,1.296148140,1.122497216,1.449137675,1.095445115},
			{2.121320343,2.449489743,2.371708245,2.244994432,2.121320343,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.122497216,1.296148140,.9486832980,1.732050808,2.,1.936491673,1.833030278,1.677050983,1.936491673,1.875000000,1.449137675,1.673320053,1.254990040,1.369306394,1.581138830,1.530931089,1.449137675,1.673320053,1.322875656,1.122497216,1.296148140,1.254990040,.9486832980},
			{1.825741857,2.371708244,2.449489743,2.415229459,2.347382389,1.490711985,1.936491673,2.,1.972026594,1.178511302,1.530931089,1.581138830,.9660917834,1.254990040,.8164965809,1.490711985,1.936491673,2.,1.972026594,1.443375673,1.875000000,1.936491673,1.247219129,1.620185175,1.080123450,1.178511302,1.530931089,1.581138830,1.247219129,1.620185175,1.138550085,.9660917834,1.254990040,1.080123450,.8164965809},
			{1.620185174,2.244994432,2.415229459,2.449489743,2.430277761,1.322875656,1.833030278,1.972026594,2.,1.045825033,1.449137675,1.559023911,.8573214100,1.187939392,.7245688375,1.322875656,1.833030278,1.972026594,2.,1.280868845,1.774823935,1.909406540,1.106797181,1.533623161,.9585144758,1.045825033,1.449137675,1.559023911,1.106797181,1.533623161,1.010362971,.8573214100,1.187939392,.9585144758,.7245688375},
			{2.581988897,2.236067977,1.924500898,1.707825129,1.549193338,3.162277660,2.738612788,2.357022604,2.091650066,3.,2.598076212,2.236067977,2.732520206,2.366431913,2.474358298,1.581138830,1.369306394,1.178511302,1.045825033,2.449489743,2.121320343,1.825741858,2.645751311,2.291287848,2.618614683,1.,.8660254040,.7453559923,1.763834207,1.527525232,2.070196679,.6831300514,.5916079782,1.309307342,.4948716594},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,2.738612788,3.162277660,3.061862179,2.898275350,2.598076212,3.,2.904737510,2.366431913,2.732520203,2.142857143,1.369306394,1.581138830,1.530931089,1.449137675,2.121320343,2.449489743,2.371708245,2.291287848,2.645751311,2.267786838,.8660254040,1.,.9682458365,1.527525232,1.763834207,1.792842914,.5916079782,.6831300507,1.133893419,.4285714286},
			{1.924500898,2.500000000,2.581988897,2.545875387,2.474358298,2.357022604,3.061862179,3.162277659,3.118047822,2.236067977,2.904737510,3.,2.036700309,2.645751311,1.844277784,1.178511302,1.530931089,1.581138830,1.559023911,1.825741858,2.371708244,2.449489743,1.972026594,2.561737692,1.951800147,.7453559923,.9682458365,1.,1.314684396,1.707825129,1.543033500,.5091750773,.6614378278,.9759000734,.3688555568},
			{2.645751311,2.291287848,1.972026594,1.750000000,1.587450787,3.888444420,3.367491648,2.898275350,2.571964230,4.098780308,3.549647869,3.055050464,4.,3.464101616,3.803194146,1.296148140,1.122497216,.9660917834,.8573214100,2.509980079,2.173706512,1.870828694,3.098386677,2.683281572,3.354101966,.6831300514,.5916079782,.5091750773,1.549193338,1.341640786,2.121320343,.4000000000,.3464101616,1.006230590,.2535462764},
			{2.291287848,2.645751311,2.561737690,2.424871131,2.291287848,3.367491648,3.888444419,3.764970119,3.563818176,3.549647869,4.098780304,3.968626966,3.464101616,4.,3.293662744,1.122497216,1.296148140,1.254990040,1.187939392,2.173706511,2.509980079,2.430277762,2.683281572,3.098386677,2.904737510,.5916079782,.6831300507,.6614378278,1.341640786,1.549193338,1.837117307,.3464101616,.4000000000,.8714212528,.2195775163},
			{2.683281572,2.323790008,2.,1.774823935,1.609968943,4.381780461,3.794733192,3.265986323,2.898275350,4.948716594,4.285714286,3.688555568,5.070925527,4.391550326,5.,1.095445115,.9486832980,.8164965809,.7245688375,2.424366106,2.099562637,1.807015806,3.273268354,2.834733546,3.779644731,.4948716594,.4285714286,.3688555568,1.309307341,1.133893419,1.992047682,.2535462764,.2195775163,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,2.738612787,2.357022604,2.091650066,1.897366595,1.936491673,1.677050983,1.443375673,1.280868845,1.224744872,1.060660172,.9128709292,.8366600264,.7245688370,.6060915266,3.872983346,3.354101966,2.886751347,2.561737690,3.,2.598076212,2.236067977,2.160246899,1.870828694,1.603567452,3.674234614,3.181980514,2.738612787,3.240370349,2.806243040,2.535462764,3.346640106,2.898275348,3.207134903,3.030457633},
			{2.738612786,3.162277659,3.061862179,2.898275350,2.738612786,1.677050983,1.936491673,1.875000000,1.774823935,1.060660172,1.224744872,1.185854122,.7245688375,.8366600264,.5248906592,3.354101966,3.872983346,3.750000000,3.549647869,2.598076212,3.,2.904737510,1.870828694,2.160246900,1.388730150,3.181980514,3.674234614,3.557562366,2.806243040,3.240370350,2.195775165,2.898275350,3.346640106,2.777460300,2.624453295},
			{2.357022604,3.061862179,3.162277660,3.118047822,3.030457633,1.443375673,1.875000000,1.936491673,1.909406540,.9128709293,1.185854122,1.224744872,.6236095646,.8100925872,.4517539516,2.886751347,3.750000000,3.872983346,3.818813079,2.236067977,2.904737510,3.,1.610152972,2.091650066,1.195228610,2.738612788,3.557562368,3.674234614,2.415229459,3.137475099,1.889822365,2.494438258,3.240370349,2.390457218,2.258769757},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,3.346640106,2.898275350,2.494438258,2.213594362,2.645751311,2.291287848,1.972026594,2.065591118,1.788854382,1.636634177,3.346640106,2.898275350,2.494438258,2.213594362,4.320493799,3.741657387,3.220305944,4.,3.464101616,3.464101616,2.645751311,2.291287848,1.972026594,4.,3.464101616,4.107919181,2.065591118,1.788854382,3.464101616,1.636634177},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,2.898275350,3.346640106,3.240370350,3.067246323,2.291287848,2.645751311,2.561737692,1.788854382,2.065591118,1.417366774,2.898275350,3.346640106,3.240370350,3.067246323,3.741657387,4.320493800,4.183300132,3.464101616,4.,3.,2.291287848,2.645751311,2.561737692,3.464101616,4.,3.557562368,1.788854382,2.065591118,3.,1.417366774},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,4.347413025,3.764970119,3.240370350,2.875543427,3.927922025,3.401680258,2.927700220,3.354101966,2.904737510,2.834733546,2.898275350,2.509980080,2.160246900,1.917028952,4.810702356,4.166190450,3.585685828,5.196152424,4.500000000,5.,1.963961013,1.700840129,1.463850110,3.897114318,3.375000000,4.743416490,1.341640786,1.161895004,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,2.958039892,2.545875386,2.259240285,2.049390154,1.673320053,1.449137675,1.247219129,1.106797181,.8819171036,.7637626160,.6573421980,.5163977793,.4472135954,.3273268354,5.019960159,4.347413025,3.741657387,3.320391543,3.240370349,2.806243040,2.415229459,2.,1.732050808,1.299038106,5.291502622,4.582575695,3.944053189,4.,3.464101616,2.738612788,5.163977793,4.472135954,4.330127020,4.909902528},
			{2.958039892,3.415650256,3.307189139,3.130495168,2.958039892,1.449137675,1.673320053,1.620185175,1.533623161,.7637626162,.8819171036,.8539125641,.4472135954,.5163977793,.2834733546,4.347413025,5.019960159,4.860555525,4.600869484,2.806243040,3.240370350,3.137475099,1.732050808,2.,1.125000000,4.582575696,5.291502622,5.123475385,3.464101616,4.,2.371708245,4.472135954,5.163977793,3.750000000,4.252100321},
			{3.741657386,3.240370350,2.788866755,2.474873734,2.244994432,3.055050464,2.645751311,2.277100171,2.020725943,2.070196679,1.792842914,1.543033500,1.414213562,1.224744872,.9960238407,4.582575696,3.968626966,3.415650256,3.031088914,5.070925527,4.391550330,3.779644731,4.107919181,3.557562368,3.162277660,4.140393357,3.585685828,3.086067000,5.477225575,4.743416490,5.,3.535533905,3.061862179,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,3.074085231,2.645751311,2.347871376,2.129788722,1.449137675,1.254990040,1.080123450,.9585144758,.6546536711,.5669467095,.4879500367,.3354101966,.2904737510,.1889822365,5.796550700,5.019960159,4.320493800,3.834057903,3.207134903,2.777460300,2.390457218,1.732050808,1.500000000,1.,6.546536711,5.669467095,4.879500367,4.330127020,3.750000000,2.635231383,6.708203931,5.809475019,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,.172756623071377,.00390388665157827,6.57541912233899e-005,.00486900211054309,.000110027806742895,.000110027806742895,5.33523806427656,.120563624637786,.00203068488848474,.368327957423731,.00832333122831677,.0131603422009587,7.19422103343416,.16257219522684,.629639404683915,9.00134910987265},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,.844111317574503,.0190748976597194,.000321283525938621,.123027723654541,.00278013241624376,.014376841680633,4.7630793830775,.107634206373357,.00181291129080671,1.70046113382978,.0384263561113877,.314193514488106,6.0685481480974,.13713467930593,2.74657165283107,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,1.80585734803576,.0408080585873606,.000687340880331092,.625327304106952,.014130902021659,.173615533771693,3.86201434387853,.0872722907753235,.00146995018268015,3.27577043476888,.0740245800161786,1.43801908896753,4.43071582310304,.100123584532725,4.76431769725595,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.72409573525399,.061558050796614,.00103683846501621,1.66715454931021,.0376737069316204,.818062705137864,2.81340752869054,.0635762837863472,.00107083209499401,4.21756481949244,.0953068814408436,3.27222005765247,2.75487590413568,.0622536090052649,5.23550287189747,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,3.27976309580122,.0741148046448401,.00124833517774672,3.01515438576158,.0681352803972318,2.22245957594899,1.80736427869175,.0408421116086564,.000687914444399609,4.06994367638345,.0919709965436096,4.74332322095564,1.41847476579742,.0320541383810507,4.04940821803081,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,3.27976309580122,.0741148046448401,.00124833517774672,4.06994367638345,.0919709965436096,4.04940821803081,.991945941760336,.0224156067156837,.000377551968601727,3.01515438576158,.0681352803972318,4.74332322095564,.576746649127296,.013033095369547,2.22245957594899,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,2.72409573525399,.061558050796614,.00103683846501621,4.21756481949244,.0953068814408436,5.23550287189747,.439603191878621,.00993398112262371,.000167320660844516,1.66715454931021,.0376737069316204,3.27222005765247,.170154848292167,.00384509276081317,.818062705137864,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,1.80585734803576,.0408080585873606,.000687340880331092,3.27577043476888,.0740245800161786,4.76431769725595,.140734880491406,.00318027182678497,5.35661561215366e-005,.625327304106952,.014130902021659,1.43801908896753,.0308216365902836,.000696495297778531,.173615533771693,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.844111317574503,.0190748976597194,.000321283525938621,1.70046113382978,.0384263561113877,2.74657165283107,.0249321870529784,.00056340781892775,9.4896263063424e-006,.123027723654541,.00278013241624376,.314193514488106,.00229823078236534,5.1934521012163e-005,.014376841680633,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.172756623071377,.00390388665157827,6.57541912233899e-005,.368327957423731,.00832333122831677,.629639404683915,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,.00486900211054309,.000110027806742895,.0131603422009587,1.66187980881431e-005,3.75545104141921e-007,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,.172756623071377,.0201880888841649,.00175840721715611,.00486900211054309,.000568984537846987,.000110027806742895,5.33523806427656,.623468196598569,.0543048419765405,.368327957423731,.0430422718809401,.0131603422009587,7.19422103343416,.840706255138427,.629639404683915,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,.844111317574503,.0986416266094925,.00859180624463178,.123027723654541,.014376841680633,.014376841680633,4.7630793830775,.556606561402958,.0484811118273528,1.70046113382978,.198713426363864,.314193514488106,6.0685481480974,.709161751356258,2.74657165283107,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,1.80585734803576,.211029875475197,.0183809600898976,.625327304106952,.0730748434797269,.173615533771693,3.86201434387853,.45130940535495,.039309600832947,3.27577043476888,.382801790716804,1.43801908896753,4.43071582310304,.517767036891475,4.76431769725595,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.72409573525399,.318333884134563,.0277272703988645,1.66715454931021,.194821267114416,.818062705137864,2.81340752869054,.328770730954697,.0286363325160947,4.21756481949244,.492858519092109,3.27222005765247,2.75487590413568,.321930810017305,5.23550287189747,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,3.27976309580122,.383268367486451,.0333831432664456,3.01515438576158,.352346576520278,2.22245957594899,1.80736427869175,.211205973165025,.0183962984155437,4.06994367638345,.475607726017603,4.74332322095564,1.41847476579742,.165760907653494,4.04940821803081,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,3.27976309580122,.383268367486451,.0333831432664456,4.06994367638345,.475607726017603,4.04940821803081,.991945941760336,.115917366756987,.0100965443280308,3.01515438576158,.352346576520278,4.74332322095564,.576746649127296,.0673977784805412,2.22245957594899,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,2.72409573525399,.318333884134563,.0277272703988645,4.21756481949244,.492858519092109,5.23550287189747,.439603191878621,.0513713926084575,.00447451108643047,1.66715454931021,.194821267114416,3.27222005765247,.170154848292167,.0198840492440458,.818062705137864,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,1.80585734803576,.211029875475197,.0183809600898976,3.27577043476888,.382801790716804,4.76431769725595,.140734880491406,.0164460743984419,.00143247318181469,.625327304106952,.0730748434797269,1.43801908896753,.0308216365902836,.00360177183250731,.173615533771693,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.844111317574503,.0986416266094925,.00859180624463178,1.70046113382978,.198713426363864,2.74657165283107,.0249321870529784,.00291353928576501,.000253772832951388,.123027723654541,.014376841680633,.314193514488106,.00229823078236534,.000268567922156809,.014376841680633,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.172756623071377,.0201880888841649,.00175840721715611,.368327957423731,.0430422718809401,.629639404683915,.000932318622844532,.00010894940458837,9.4896263063424e-006,.00486900211054309,.000568984537846987,.0131603422009587,1.66187980881431e-005,1.9420486861126e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,.172756623071377,.0479640552145288,.00992569280167208,.00486900211054309,.00135182710686955,.000110027806742895,5.33523806427656,1.48127260473935,.306534899222172,.368327957423731,.102262374457619,.0131603422009587,7.19422103343416,1.99739962866362,.629639404683915,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,.844111317574503,.234358608796285,.0484982253050728,.123027723654541,.0341573505118309,.014376841680633,4.7630793830775,1.32241877857201,.273661651286941,1.70046113382978,.472115107510773,.314193514488106,6.0685481480974,1.68486841886076,2.74657165283107,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,1.80585734803576,.501377255533426,.10375512649862,.625327304106952,.173615533771693,.173615533771693,3.86201434387853,1.07224756942001,.221891162762184,3.27577043476888,.909483127972704,1.43801908896753,4.43071582310304,1.23014153980692,4.76431769725595,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.72409573525399,.756316463776911,.156512305865719,1.66715454931021,.462867885437628,.818062705137864,2.81340752869054,.781112941542527,.161643694787028,4.21756481949244,1.17096241047498,3.27222005765247,2.75487590413568,.764862252961431,5.23550287189747,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,3.27976309580122,.910591648648909,.188438048697755,3.01515438576158,.837125829385702,2.22245957594899,1.80736427869175,.501795638836895,.103841706846431,4.06994367638345,1.12997695631596,4.74332322095564,1.41847476579742,.393824565203833,4.04940821803081,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,3.27976309580122,.910591648648909,.188438048697755,4.06994367638345,1.12997695631596,4.04940821803081,.991945941760336,.275403333686328,.0569920302764555,3.01515438576158,.837125829385702,4.74332322095564,.576746649127296,.160127627083755,2.22245957594899,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,2.72409573525399,.756316463776911,.156512305865719,4.21756481949244,1.17096241047498,5.23550287189747,.439603191878621,.122051191950714,.0252573022041014,1.66715454931021,.462867885437628,3.27222005765247,.170154848292167,.0472416998608471,.818062705137864,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,1.80585734803576,.501377255533426,.10375512649862,3.27577043476888,.909483127972704,4.76431769725595,.140734880491406,.0390735559485202,.00808589080538567,.625327304106952,.173615533771693,1.43801908896753,.0308216365902836,.00855730247849368,.173615533771693,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.844111317574503,.234358608796285,.0484982253050728,1.70046113382978,.472115107510773,2.74657165283107,.0249321870529784,.00692215890141759,.00143247318181469,.123027723654541,.0341573505118309,.314193514488106,.00229823078236534,.000638079548841519,.014376841680633,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.172756623071377,.0479640552145288,.00992569280167208,.368327957423731,.102262374457619,.629639404683915,.000932318622844532,.000258848437177504,5.35661561215366e-005,.00486900211054309,.00135182710686955,.0131603422009587,1.66187980881431e-005,4.61403409428573e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,.172756623071377,.0847706354072961,.031004156339822,.00486900211054309,.00238919003611327,.000110027806742895,5.33523806427656,2.61796921368607,.957500511953663,.368327957423731,.180736312270888,.0131603422009587,7.19422103343416,3.53016097005959,.629639404683915,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,.844111317574503,.414200344236389,.151490338216903,.123027723654541,.0603689636701225,.014376841680633,4.7630793830775,2.33721439175002,.854816765967696,1.70046113382978,.834406045737126,.314193514488106,6.0685481480974,2.9778000589335,2.74657165283107,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,1.80585734803576,.886123334239273,.324092255049372,.625327304106952,.306844345178412,.173615533771693,3.86201434387853,1.89506719911644,.693105099882267,3.27577043476888,1.60740084018392,1.43801908896753,4.43071582310304,2.17412559267414,4.76431769725595,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.72409573525399,1.3366973855027,.488885974725042,1.66715454931021,.818062705137864,.818062705137864,2.81340752869054,1.38052214482826,.504914516765946,4.21756481949244,2.06953367746009,3.27222005765247,2.75487590413568,1.35180102876198,5.23550287189747,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,3.27976309580122,1.60936001568865,.58861006873112,3.01515438576158,1.4795181132097,2.22245957594899,1.80736427869175,.886862775854878,.324362699711839,4.06994367638345,1.99709687085614,4.74332322095564,1.41847476579742,.696037031940087,4.04940821803081,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,3.27976309580122,1.60936001568865,.58861006873112,4.06994367638345,1.99709687085614,4.04940821803081,.991945941760336,.48674190465042,.178021811889788,3.01515438576158,1.4795181132097,4.74332322095564,.576746649127296,.283006110168114,2.22245957594899,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,2.72409573525399,1.3366973855027,.488885974725042,4.21756481949244,2.06953367746009,5.23550287189747,.439603191878621,.215710640970697,.0788943766349663,1.66715454931021,.818062705137864,3.27222005765247,.170154848292167,.083493960154389,.818062705137864,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,1.80585734803576,.886123334239273,.324092255049372,3.27577043476888,1.60740084018392,4.76431769725595,.140734880491406,.0690577589939741,.0252573022041014,.625327304106952,.306844345178412,1.43801908896753,.0308216365902836,.0151239916064188,.173615533771693,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.844111317574503,.414200344236389,.151490338216903,1.70046113382978,.834406045737126,2.74657165283107,.0249321870529784,.0122340741590526,.00447451108643047,.123027723654541,.0603689636701225,.314193514488106,.00229823078236534,.00112772801535997,.014376841680633,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.172756623071377,.0847706354072961,.031004156339822,.368327957423731,.180736312270888,.629639404683915,.000932318622844532,.00045748313806202,.000167320660844516,.00486900211054309,.00238919003611327,.0131603422009587,1.66187980881431e-005,8.15474421864669e-006,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,.172756623071377,.127338292609009,.0699595626855254,.00486900211054309,.00358892414166019,.000110027806742895,5.33523806427656,3.93258500625844,2.16055926028883,.368327957423731,.271493227654192,.0131603422009587,7.19422103343416,5.3028347422728,.629639404683915,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,.844111317574503,.622191450845088,.341831517573806,.123027723654541,.0906832976229768,.014376841680633,4.7630793830775,3.51084887681561,1.92885774629333,1.70046113382978,1.25340385495866,.314193514488106,6.0685481480974,4.47310526119946,2.74657165283107,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,1.80585734803576,1.33109103029466,.731300416128337,.625327304106952,.460926532212655,.173615533771693,3.86201434387853,2.8466770404089,1.5639622362693,3.27577043476888,2.41455873892965,1.43801908896753,4.43071582310304,3.26586487830957,4.76431769725595,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.72409573525399,2.00792128060645,1.10315044924867,1.66715454931021,1.22885368994647,.818062705137864,2.81340752869054,2.07375268554642,1.13931817396847,4.21756481949244,3.10875203091765,3.27222005765247,2.75487590413568,2.03060923305097,5.23550287189747,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,3.27976309580122,2.41750171632384,1.3281736341858,3.01515438576158,2.22245957594899,2.22245957594899,1.80736427869175,1.33220178323833,.731910663029089,4.06994367638345,2.99994101126824,4.74332322095564,1.41847476579742,1.04555270600537,4.04940821803081,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,3.27976309580122,2.41750171632384,1.3281736341858,4.06994367638345,2.99994101126824,4.04940821803081,.991945941760336,.731159826532418,.401698661682273,3.01515438576158,2.22245957594899,4.74332322095564,.576746649127296,.425117904255122,2.22245957594899,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,.981387965414959,.723377579561695,.397423100941195,.200115539631623,2.72409573525399,2.00792128060645,1.10315044924867,4.21756481949244,3.10875203091765,5.23550287189747,.439603191878621,.324029949602564,.178021811889788,1.66715454931021,1.22885368994647,3.27222005765247,.170154848292167,.125420533633389,.818062705137864,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,.555278916264744,.409294114646486,.224865880330915,.113227330974451,1.80585734803576,1.33109103029466,.731300416128337,3.27577043476888,2.41455873892965,4.76431769725595,.140734880491406,.103735179988284,.0569920302764555,.625327304106952,.460926532212655,1.43801908896753,.0308216365902836,.0227185187340072,.173615533771693,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.844111317574503,.622191450845088,.341831517573806,1.70046113382978,1.25340385495866,2.74657165283107,.0249321870529784,.018377426423439,.0100965443280308,.123027723654541,.0906832976229768,.314193514488106,.00229823078236534,.0016940177375493,.014376841680633,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.172756623071377,.127338292609009,.0699595626855254,.368327957423731,.271493227654192,.629639404683915,.000932318622844532,.000687208741781061,.000377551968601727,.00486900211054309,.00358892414166019,.0131603422009587,1.66187980881431e-005,1.22496569770466e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,.172756623071377,.171884955945497,.127469057765715,.00486900211054309,.00484443489339947,.000110027806742895,5.33523806427656,5.30831839113911,3.93662342336207,.368327957423731,.366469508258039,.0131603422009587,7.19422103343416,7.1579216069788,.629639404683915,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,.844111317574503,.839852238686345,.62283038639938,.123027723654541,.12240697042152,.014376841680633,4.7630793830775,4.73904661479697,3.51445420820192,1.70046113382978,1.69188122423627,.314193514488106,6.0685481480974,6.03792845818671,2.74657165283107,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,1.80585734803576,1.79674564825646,1.33245794297741,.625327304106952,.622172129531742,.173615533771693,3.86201434387853,3.84252802245541,2.84960032604169,3.27577043476888,3.259242086299,1.43801908896753,4.43071582310304,4.40836004025762,4.76431769725595,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.72409573525399,2.71035093833706,2.00998323805485,1.66715454931021,1.65874269265149,.818062705137864,2.81340752869054,2.79921209632888,2.07588224620454,4.21756481949244,4.19628451843992,3.27222005765247,2.75487590413568,2.7409758008707,5.23550287189747,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,3.27976309580122,3.26321460335875,2.41998427663067,3.01515438576158,2.99994101126824,2.22245957594899,1.80736427869175,1.79824497510369,1.33356983668368,4.06994367638345,4.04940821803081,4.74332322095564,1.41847476579742,1.41131765730707,4.04940821803081,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,3.27976309580122,3.26321460335875,2.41998427663067,4.06994367638345,4.04940821803081,4.04940821803081,.991945941760336,.986940942882965,.731910663029089,3.01515438576158,2.99994101126824,4.74332322095564,.576746649127296,.573836595005234,2.22245957594899,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,.981387965414959,.976436238050563,.724120424237357,.49217250527751,2.72409573525399,2.71035093833706,2.00998323805485,4.21756481949244,4.19628451843992,5.23550287189747,.439603191878621,.437385113867296,.324362699711839,1.66715454931021,1.65874269265149,3.27222005765247,.170154848292167,.169296308726466,.818062705137864,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,.555278916264744,.552477180456442,.409714423434642,.278476020673683,1.80585734803576,1.79674564825646,1.33245794297741,3.27577043476888,3.259242086299,4.76431769725595,.140734880491406,.140024783409284,.103841706846431,.625327304106952,.622172129531742,1.43801908896753,.0308216365902836,.0306661218062029,.173615533771693,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.844111317574503,.839852238686345,.62283038639938,1.70046113382978,1.69188122423627,2.74657165283107,.0249321870529784,.0248063882942385,.0183962984155437,.123027723654541,.12240697042152,.314193514488106,.00229823078236534,.00228663474453522,.014376841680633,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.172756623071377,.171884955945497,.127469057765715,.368327957423731,.366469508258039,.629639404683915,.000932318622844532,.000927614481757561,.000687914444399609,.00486900211054309,.00484443489339947,.0131603422009587,1.66187980881431e-005,1.65349456687952e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,.172756623071377,.21445261314721,.198422869712097,.00486900211054309,.0060441689989464,.000110027806742895,5.33523806427656,6.62293418371148,6.1278880563708,.368327957423731,.457226423641343,.0131603422009587,7.19422103343416,8.93059537919201,.629639404683915,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,.844111317574503,1.04784334529504,.969519935107731,.123027723654541,.152721304374374,.014376841680633,4.7630793830775,5.91268109986256,5.47072443843503,1.70046113382978,2.1108790334578,.314193514488106,6.0685481480974,7.53323366045268,2.74657165283107,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,1.80585734803576,2.24171334431186,2.07415143290851,.625327304106952,.776254316565985,.173615533771693,3.86201434387853,4.79413786374787,4.43578923494485,3.27577043476888,4.06639998504474,1.43801908896753,4.43071582310304,5.50009932589306,4.76431769725595,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.72409573525399,3.3815748334408,3.12881140849955,1.66715454931021,2.06953367746009,.818062705137864,2.81340752869054,3.49244263704704,3.2313921487783,4.21756481949244,5.23550287189747,3.27222005765247,2.75487590413568,3.41978400515968,5.23550287189747,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,3.27976309580122,4.07135630399393,3.76703360991161,3.01515438576158,3.74288247400753,2.22245957594899,1.80736427869175,2.24358398248714,2.07588224620454,4.06994367638345,5.05225235844291,4.74332322095564,1.41847476579742,1.76083333137235,4.04940821803081,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,3.27976309580122,4.07135630399393,3.76703360991161,4.06994367638345,5.05225235844291,4.04940821803081,.991945941760336,1.23135886476496,1.13931817396847,3.01515438576158,3.74288247400753,4.74332322095564,.576746649127296,.715948389092241,2.22245957594899,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,2.72409573525399,3.3815748334408,3.12881140849955,4.21756481949244,5.23550287189747,5.23550287189747,.439603191878621,.545704422499163,.504914516765946,1.66715454931021,2.06953367746009,3.27222005765247,.170154848292167,.211222882205465,.818062705137864,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,.555278916264744,.689299272303129,.637776046087675,.540839285781897,1.80585734803576,2.24171334431186,2.07415143290851,3.27577043476888,4.06639998504474,4.76431769725595,.140734880491406,.174702204403594,.161643694787028,.625327304106952,.776254316565985,1.43801908896753,.0308216365902836,.0382606489337913,.173615533771693,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.844111317574503,1.04784334529504,.969519935107731,1.70046113382978,2.1108790334578,2.74657165283107,.0249321870529784,.0309497405586248,.0286363325160947,.123027723654541,.152721304374374,.314193514488106,.00229823078236534,.00285292446672455,.014376841680633,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.172756623071377,.21445261314721,.198422869712097,.368327957423731,.457226423641343,.629639404683915,.000932318622844532,.0011573400854766,.00107083209499401,.00486900211054309,.0060441689989464,.0131603422009587,1.66187980881431e-005,2.0629858427195e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,.172756623071377,.251259193339977,.272378587590661,.00486900211054309,.00708153192819011,.000110027806742895,5.33523806427656,7.7596307926582,8.41186046814944,.368327957423731,.535700361454612,.0131603422009587,7.19422103343416,10.463356720588,.629639404683915,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,.844111317574503,1.22768508073515,1.33087718642915,.123027723654541,.178932917532666,.014376841680633,4.7630793830775,6.92747671304058,7.50976033055426,1.70046113382978,2.47316997168416,.314193514488106,6.0685481480974,8.82616530052541,2.74657165283107,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,1.80585734803576,2.6264594230177,2.8472244079752,.625327304106952,.909483127972704,.173615533771693,3.86201434387853,5.6169574934443,6.08908644662382,3.27577043476888,4.76431769725595,1.43801908896753,4.43071582310304,6.44408337876027,4.76431769725595,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.72409573525399,3.9619557551666,4.29497483592082,1.66715454931021,2.42472849716033,.818062705137864,2.81340752869054,4.09185184033277,4.43578923494485,4.21756481949244,6.13407413888259,3.27222005765247,2.75487590413568,4.00672278096023,5.23550287189747,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,3.27976309580122,4.77012467103367,5.17107375557585,3.01515438576158,4.38527475783153,2.22245957594899,1.80736427869175,2.62865111950513,2.84960032604169,4.06994367638345,5.91937227298309,4.74332322095564,1.41847476579742,2.0630457981086,4.04940821803081,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,3.27976309580122,4.77012467103367,5.17107375557585,4.06994367638345,5.91937227298309,4.04940821803081,.991945941760336,1.44269743572906,1.5639622362693,3.01515438576158,4.38527475783153,4.74332322095564,.576746649127296,.838826872176601,2.22245957594899,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,2.72409573525399,3.9619557551666,4.29497483592082,4.21756481949244,6.13407413888259,5.23550287189747,.439603191878621,.639363871519146,.693105099882267,1.66715454931021,2.42472849716033,3.27222005765247,.170154848292167,.247475142499007,.818062705137864,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,.555278916264744,.807603957755601,.875486474339234,.869841424679861,1.80585734803576,2.6264594230177,2.8472244079752,3.27577043476888,4.76431769725595,4.76431769725595,.140734880491406,.204686407449048,.221891162762184,.625327304106952,.909483127972704,1.43801908896753,.0308216365902836,.0448273380617164,.173615533771693,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.844111317574503,1.22768508073515,1.33087718642915,1.70046113382978,2.47316997168416,2.74657165283107,.0249321870529784,.0362616558162599,.039309600832947,.123027723654541,.178932917532666,.314193514488106,.00229823078236534,.003342572933243,.014376841680633,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.172756623071377,.251259193339977,.272378587590661,.368327957423731,.535700361454612,.629639404683915,.000932318622844532,.00135597478636112,.00146995018268015,.00486900211054309,.00708153192819011,.0131603422009587,1.66187980881431e-005,2.4170568551556e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,.172756623071377,.279035159670341,.33592853869153,.00486900211054309,.00786437449721268,.000110027806742895,5.33523806427656,8.61743520079898,10.3744718692395,.368327957423731,.594920464031291,.0131603422009587,7.19422103343416,11.6200500941132,.629639404683915,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,.844111317574503,1.36340206292194,1.64139050859212,.123027723654541,.198713426363864,.014376841680633,4.7630793830775,7.69328893020963,9.26189843365359,1.70046113382978,2.74657165283107,.314193514488106,6.0685481480974,9.80187196802991,2.74657165283107,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,1.80585734803576,2.91680680307593,3.51152395332694,.625327304106952,1.01002381826467,.173615533771693,3.86201434387853,6.23789565750936,7.50976033055426,3.27577043476888,5.29099903451185,1.43801908896753,4.43071582310304,7.15645788167572,4.76431769725595,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.72409573525399,4.39993833480895,5.29705595843705,1.66715454931021,2.69277511548354,.818062705137864,2.81340752869054,4.5441940509206,5.47072443843503,4.21756481949244,6.81217803026546,3.27222005765247,2.75487590413568,4.44965422390436,5.23550287189747,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,3.27976309580122,5.29744795219613,6.37756170755723,3.01515438576158,4.87005401069696,2.22245957594899,1.80736427869175,2.919240785177,3.51445420820192,4.06994367638345,6.57374150328145,4.74332322095564,1.41847476579742,2.29110945565894,4.04940821803081,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,3.27976309580122,5.29744795219613,6.37756170755723,4.06994367638345,6.57374150328145,4.04940821803081,.991945941760336,1.6021834026584,1.92885774629333,3.01515438576158,4.87005401069696,4.74332322095564,.576746649127296,.931556720779814,2.22245957594899,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,2.72409573525399,4.39993833480895,5.29705595843705,4.21756481949244,6.81217803026546,5.23550287189747,.439603191878621,.710043670861402,.854816765967696,1.66715454931021,2.69277511548354,3.27222005765247,.170154848292167,.274832793115808,.818062705137864,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,1.80585734803576,2.91680680307593,3.51152395332694,3.27577043476888,5.29099903451185,4.76431769725595,.140734880491406,.227313888999126,.273661651286941,.625327304106952,1.01002381826467,1.43801908896753,.0308216365902836,.0497828687077028,.173615533771693,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.844111317574503,1.36340206292194,1.64139050859212,1.70046113382978,2.74657165283107,2.74657165283107,.0249321870529784,.0402702754319125,.0484811118273528,.123027723654541,.198713426363864,.314193514488106,.00229823078236534,.00371208455992771,.014376841680633,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.172756623071377,.279035159670341,.33592853869153,.368327957423731,.594920464031291,.629639404683915,.000932318622844532,.00150587381895025,.00181291129080671,.00486900211054309,.00786437449721268,.0131603422009587,1.66187980881431e-005,2.68425539597291e-005,.000110027806742895,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,.172756623071377,.295319361902928,.376281515036571,.00486900211054309,.00832333122831677,.000110027806742895,5.33523806427656,9.12033977275976,11.620691733626,.368327957423731,.629639404683915,.0131603422009587,7.19422103343416,12.2981841540248,.629639404683915,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,.844111317574503,1.44296879187171,1.83856039664088,.123027723654541,.210310135628253,.014376841680633,4.7630793830775,8.14226128523923,10.3744718692395,1.70046113382978,2.90685872308354,.314193514488106,6.0685481480974,10.3738990400802,2.74657165283107,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,1.80585734803576,3.08702861996377,3.93334117545275,.625327304106952,1.06896775972274,.173615533771693,3.86201434387853,6.60193277208899,8.41186046814944,3.27577043476888,5.59977624521248,1.43801908896753,4.43071582310304,7.57410133403447,4.76431769725595,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.72409573525399,4.65671416814689,5.93335787735631,1.66715454931021,2.84992267566634,.818062705137864,2.81340752869054,4.80938849808895,6.1278880563708,4.21756481949244,7.20972966791673,3.27222005765247,2.75487590413568,4.7093314249164,5.23550287189747,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,3.27976309580122,5.60660151503774,7.14365796638211,3.01515438576158,5.15426530682,2.22245957594899,1.80736427869175,3.08960464673337,3.93662342336207,4.06994367638345,6.95737823275544,4.74332322095564,1.41847476579742,2.42481622493138,4.04940821803081,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,3.27976309580122,5.60660151503774,7.14365796638211,4.06994367638345,6.95737823275544,4.04940821803081,.991945941760336,1.6956851626997,2.16055926028883,3.01515438576158,5.15426530682,4.74332322095564,.576746649127296,.985921403890809,2.22245957594899,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,2.72409573525399,4.65671416814689,5.93335787735631,4.21756481949244,7.20972966791673,5.23550287189747,.439603191878621,.751481082347236,.957500511953663,1.66715454931021,2.84992267566634,3.27222005765247,.170154848292167,.290871749599041,.818062705137864,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,1.80585734803576,3.08702861996377,3.93334117545275,3.27577043476888,5.59977624521248,4.76431769725595,.140734880491406,.240579691570783,.306534899222172,.625327304106952,1.06896775972274,1.43801908896753,.0308216365902836,.0526881452424316,.173615533771693,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.844111317574503,1.44296879187171,1.83856039664088,1.70046113382978,2.90685872308354,2.74657165283107,.0249321870529784,.0426204068987497,.0543048419765405,.123027723654541,.210310135628253,.314193514488106,.00229823078236534,.00392871796107236,.014376841680633,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.172756623071377,.295319361902928,.376281515036571,.368327957423731,.629639404683915,.629639404683915,.000932318622844532,.00159375505167201,.00203068488848474,.00486900211054309,.00832333122831677,.0131603422009587,1.66187980881431e-005,2.84090575416998e-005,.000110027806742895,2.74870870898337e-007}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,0.,0.,0.,0.,0.,0.,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,0.,0.,0.,1.66187980881431e-005,3.75545104141921e-007,0.,2.74870870898337e-007},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,0.,0.,0.,0.,0.,0.,.0249321870529784,.00056340781892775,9.4896263063424e-006,0.,0.,0.,.00229823078236534,5.1934521012163e-005,0.,.000196571591120254},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,0.,0.,0.,0.,0.,0.,.140734880491406,.00318027182678497,5.35661561215366e-005,0.,0.,0.,.0308216365902836,.000696495297778531,0.,.00626330408481417},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,0.,0.,0.,0.,0.,0.,.439603191878621,.00993398112262371,.000167320660844516,0.,0.,0.,.170154848292167,.00384509276081317,0.,.0611113213546646},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,0.,0.,0.,0.,0.,0.,.991945941760336,.0224156067156837,.000377551968601727,0.,0.,0.,.576746649127296,.013033095369547,0.,.31115444533374},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,0.,0.,0.,0.,0.,0.,1.80736427869175,.0408421116086564,.000687914444399609,0.,0.,0.,1.41847476579742,.0320541383810507,0.,1.03297875354965},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,0.,0.,0.,0.,0.,0.,2.81340752869054,.0635762837863472,.00107083209499401,0.,0.,0.,2.75487590413568,.0622536090052649,0.,2.50302559510192},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,0.,0.,0.,0.,0.,0.,3.86201434387853,.0872722907753235,.00146995018268015,0.,0.,0.,4.43071582310304,.100123584532725,0.,4.71658607960963},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,0.,0.,0.,0.,0.,0.,4.7630793830775,.107634206373357,.00181291129080671,0.,0.,0.,6.0685481480974,.13713467930593,0.,7.17423567654394},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,0.,0.,0.,0.,0.,0.,5.33523806427656,.120563624637786,.00203068488848474,0.,0.,0.,7.19422103343416,.16257219522684,0.,9.00134910987265},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,0.,0.,0.,0.,0.,0.,.000932318622844532,.00010894940458837,9.4896263063424e-006,0.,0.,0.,1.66187980881431e-005,1.9420486861126e-006,0.,2.74870870898337e-007},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,0.,0.,0.,0.,0.,0.,.0249321870529784,.00291353928576501,.000253772832951388,0.,0.,0.,.00229823078236534,.000268567922156809,0.,.000196571591120254},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,0.,0.,0.,0.,0.,0.,.140734880491406,.0164460743984419,.00143247318181469,0.,0.,0.,.0308216365902836,.00360177183250731,0.,.00626330408481417},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,0.,0.,0.,0.,0.,0.,.439603191878621,.0513713926084575,.00447451108643047,0.,0.,0.,.170154848292167,.0198840492440458,0.,.0611113213546646},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,0.,0.,0.,0.,0.,0.,.991945941760336,.115917366756987,.0100965443280308,0.,0.,0.,.576746649127296,.0673977784805412,0.,.31115444533374},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,0.,0.,0.,0.,0.,0.,1.80736427869175,.211205973165025,.0183962984155437,0.,0.,0.,1.41847476579742,.165760907653494,0.,1.03297875354965},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,0.,0.,0.,0.,0.,0.,2.81340752869054,.328770730954697,.0286363325160947,0.,0.,0.,2.75487590413568,.321930810017305,0.,2.50302559510192},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,0.,0.,0.,0.,0.,0.,3.86201434387853,.45130940535495,.039309600832947,0.,0.,0.,4.43071582310304,.517767036891475,0.,4.71658607960963},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,0.,0.,0.,0.,0.,0.,4.7630793830775,.556606561402958,.0484811118273528,0.,0.,0.,6.0685481480974,.709161751356258,0.,7.17423567654394},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,0.,0.,0.,0.,0.,0.,5.33523806427656,.623468196598569,.0543048419765405,0.,0.,0.,7.19422103343416,.840706255138427,0.,9.00134910987265},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,0.,0.,0.,0.,0.,0.,.000932318622844532,.000258848437177504,5.35661561215366e-005,0.,0.,0.,1.66187980881431e-005,4.61403409428573e-006,0.,2.74870870898337e-007},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,0.,0.,0.,0.,0.,0.,.0249321870529784,.00692215890141759,.00143247318181469,0.,0.,0.,.00229823078236534,.000638079548841519,0.,.000196571591120254},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,0.,0.,0.,0.,0.,0.,.140734880491406,.0390735559485202,.00808589080538567,0.,0.,0.,.0308216365902836,.00855730247849368,0.,.00626330408481417},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,0.,0.,0.,0.,0.,0.,.439603191878621,.122051191950714,.0252573022041014,0.,0.,0.,.170154848292167,.0472416998608471,0.,.0611113213546646},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,0.,0.,0.,0.,0.,0.,.991945941760336,.275403333686328,.0569920302764555,0.,0.,0.,.576746649127296,.160127627083755,0.,.31115444533374},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,0.,0.,0.,0.,0.,0.,1.80736427869175,.501795638836895,.103841706846431,0.,0.,0.,1.41847476579742,.393824565203833,0.,1.03297875354965},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,0.,0.,0.,0.,0.,0.,2.81340752869054,.781112941542527,.161643694787028,0.,0.,0.,2.75487590413568,.764862252961431,0.,2.50302559510192},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,0.,0.,0.,0.,0.,0.,3.86201434387853,1.07224756942001,.221891162762184,0.,0.,0.,4.43071582310304,1.23014153980692,0.,4.71658607960963},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,0.,0.,0.,0.,0.,0.,4.7630793830775,1.32241877857201,.273661651286941,0.,0.,0.,6.0685481480974,1.68486841886076,0.,7.17423567654394},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,0.,0.,0.,0.,0.,0.,5.33523806427656,1.48127260473935,.306534899222172,0.,0.,0.,7.19422103343416,1.99739962866362,0.,9.00134910987265},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,0.,0.,0.,0.,0.,0.,.000932318622844532,.00045748313806202,.000167320660844516,0.,0.,0.,1.66187980881431e-005,8.15474421864669e-006,0.,2.74870870898337e-007},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,0.,0.,0.,0.,0.,0.,.0249321870529784,.0122340741590526,.00447451108643047,0.,0.,0.,.00229823078236534,.00112772801535997,0.,.000196571591120254},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,0.,0.,0.,0.,0.,0.,.140734880491406,.0690577589939741,.0252573022041014,0.,0.,0.,.0308216365902836,.0151239916064188,0.,.00626330408481417},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,0.,0.,0.,0.,0.,0.,.439603191878621,.215710640970697,.0788943766349663,0.,0.,0.,.170154848292167,.083493960154389,0.,.0611113213546646},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,0.,0.,0.,0.,0.,0.,.991945941760336,.48674190465042,.178021811889788,0.,0.,0.,.576746649127296,.283006110168114,0.,.31115444533374},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,0.,0.,0.,0.,0.,0.,1.80736427869175,.886862775854878,.324362699711839,0.,0.,0.,1.41847476579742,.696037031940087,0.,1.03297875354965},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,0.,0.,0.,0.,0.,0.,2.81340752869054,1.38052214482826,.504914516765946,0.,0.,0.,2.75487590413568,1.35180102876198,0.,2.50302559510192},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,0.,0.,0.,0.,0.,0.,3.86201434387853,1.89506719911644,.693105099882267,0.,0.,0.,4.43071582310304,2.17412559267414,0.,4.71658607960963},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,0.,0.,0.,0.,0.,0.,4.7630793830775,2.33721439175002,.854816765967696,0.,0.,0.,6.0685481480974,2.9778000589335,0.,7.17423567654394},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,0.,0.,0.,0.,0.,0.,5.33523806427656,2.61796921368607,.957500511953663,0.,0.,0.,7.19422103343416,3.53016097005959,0.,9.00134910987265},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,0.,0.,0.,0.,0.,0.,.000932318622844532,.000687208741781061,.000377551968601727,0.,0.,0.,1.66187980881431e-005,1.22496569770466e-005,0.,2.74870870898337e-007},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,0.,0.,0.,0.,0.,0.,.0249321870529784,.018377426423439,.0100965443280308,0.,0.,0.,.00229823078236534,.0016940177375493,0.,.000196571591120254},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.409294114646486,.224865880330915,.113227330974451,0.,0.,0.,0.,0.,0.,.140734880491406,.103735179988284,.0569920302764555,0.,0.,0.,.0308216365902836,.0227185187340072,0.,.00626330408481417},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.723377579561695,.397423100941195,.200115539631623,0.,0.,0.,0.,0.,0.,.439603191878621,.324029949602564,.178021811889788,0.,0.,0.,.170154848292167,.125420533633389,0.,.0611113213546646},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,0.,0.,0.,0.,0.,0.,.991945941760336,.731159826532418,.401698661682273,0.,0.,0.,.576746649127296,.425117904255122,0.,.31115444533374},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,0.,0.,0.,0.,0.,0.,1.80736427869175,1.33220178323833,.731910663029089,0.,0.,0.,1.41847476579742,1.04555270600537,0.,1.03297875354965},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,0.,0.,0.,0.,0.,0.,2.81340752869054,2.07375268554642,1.13931817396847,0.,0.,0.,2.75487590413568,2.03060923305097,0.,2.50302559510192},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,0.,0.,0.,0.,0.,0.,3.86201434387853,2.8466770404089,1.5639622362693,0.,0.,0.,4.43071582310304,3.26586487830957,0.,4.71658607960963},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,0.,0.,0.,0.,0.,0.,4.7630793830775,3.51084887681561,1.92885774629333,0.,0.,0.,6.0685481480974,4.47310526119946,0.,7.17423567654394},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,0.,0.,0.,0.,0.,0.,5.33523806427656,3.93258500625844,2.16055926028883,0.,0.,0.,7.19422103343416,5.3028347422728,0.,9.00134910987265},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,0.,0.,0.,0.,0.,0.,.000932318622844532,.000927614481757561,.000687914444399609,0.,0.,0.,1.66187980881431e-005,1.65349456687952e-005,0.,2.74870870898337e-007},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.232537853099397,.172448954960703,.117210661865929,0.,0.,0.,0.,0.,0.,.0249321870529784,.0248063882942385,.0183962984155437,0.,0.,0.,.00229823078236534,.00228663474453522,0.,.000196571591120254},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.552477180456442,.409714423434642,.278476020673683,0.,0.,0.,0.,0.,0.,.140734880491406,.140024783409284,.103841706846431,0.,0.,0.,.0308216365902836,.0306661218062029,0.,.00626330408481417},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,.976436238050563,.724120424237357,.49217250527751,0.,0.,0.,0.,0.,0.,.439603191878621,.437385113867296,.324362699711839,0.,0.,0.,.170154848292167,.169296308726466,0.,.0611113213546646},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,0.,0.,0.,0.,0.,0.,.991945941760336,.986940942882965,.731910663029089,0.,0.,0.,.576746649127296,.573836595005234,0.,.31115444533374},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,0.,0.,0.,0.,0.,0.,1.80736427869175,1.79824497510369,1.33356983668368,0.,0.,0.,1.41847476579742,1.41131765730707,0.,1.03297875354965},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,0.,0.,0.,0.,0.,0.,2.81340752869054,2.79921209632888,2.07588224620454,0.,0.,0.,2.75487590413568,2.7409758008707,0.,2.50302559510192},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,0.,0.,0.,0.,0.,0.,3.86201434387853,3.84252802245541,2.84960032604169,0.,0.,0.,4.43071582310304,4.40836004025762,0.,4.71658607960963},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,0.,0.,0.,0.,0.,0.,4.7630793830775,4.73904661479697,3.51445420820192,0.,0.,0.,6.0685481480974,6.03792845818671,0.,7.17423567654394},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,0.,0.,0.,0.,0.,0.,5.33523806427656,5.30831839113911,3.93662342336207,0.,0.,0.,7.19422103343416,7.1579216069788,0.,9.00134910987265},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,0.,0.,0.,0.,0.,0.,.000932318622844532,.0011573400854766,.00107083209499401,0.,0.,0.,1.66187980881431e-005,2.0629858427195e-005,0.,2.74870870898337e-007},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.290126323030972,.268440177733536,.227639458852635,0.,0.,0.,0.,0.,0.,.0249321870529784,.0309497405586248,.0286363325160947,0.,0.,0.,.00229823078236534,.00285292446672455,0.,.000196571591120254},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.689299272303129,.637776046087675,.540839285781897,0.,0.,0.,0.,0.,0.,.140734880491406,.174702204403594,.161643694787028,0.,0.,0.,.0308216365902836,.0382606489337913,0.,.00626330408481417},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,0.,0.,0.,0.,0.,0.,.439603191878621,.545704422499163,.504914516765946,0.,0.,0.,.170154848292167,.211222882205465,0.,.0611113213546646},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,0.,0.,0.,0.,0.,0.,.991945941760336,1.23135886476496,1.13931817396847,0.,0.,0.,.576746649127296,.715948389092241,0.,.31115444533374},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,0.,0.,0.,0.,0.,0.,1.80736427869175,2.24358398248714,2.07588224620454,0.,0.,0.,1.41847476579742,1.76083333137235,0.,1.03297875354965},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,0.,0.,0.,0.,0.,0.,2.81340752869054,3.49244263704704,3.2313921487783,0.,0.,0.,2.75487590413568,3.41978400515968,0.,2.50302559510192},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,0.,0.,0.,0.,0.,0.,3.86201434387853,4.79413786374787,4.43578923494485,0.,0.,0.,4.43071582310304,5.50009932589306,0.,4.71658607960963},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,0.,0.,0.,0.,0.,0.,4.7630793830775,5.91268109986256,5.47072443843503,0.,0.,0.,6.0685481480974,7.53323366045268,0.,7.17423567654394},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,0.,0.,0.,0.,0.,0.,5.33523806427656,6.62293418371148,6.1278880563708,0.,0.,0.,7.19422103343416,8.93059537919201,0.,9.00134910987265},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,0.,0.,0.,0.,0.,0.,.000932318622844532,.00135597478636112,.00146995018268015,0.,0.,0.,1.66187980881431e-005,2.4170568551556e-005,0.,2.74870870898337e-007},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.339920809644861,.368492586412728,.366116582887396,0.,0.,0.,0.,0.,0.,.0249321870529784,.0362616558162599,.039309600832947,0.,0.,0.,.00229823078236534,.003342572933243,0.,.000196571591120254},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.807603957755601,.875486474339234,.869841424679861,0.,0.,0.,0.,0.,0.,.140734880491406,.204686407449048,.221891162762184,0.,0.,0.,.0308216365902836,.0448273380617164,0.,.00626330408481417},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,0.,0.,0.,0.,0.,0.,.439603191878621,.639363871519146,.693105099882267,0.,0.,0.,.170154848292167,.247475142499007,0.,.0611113213546646},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,0.,0.,0.,0.,0.,0.,.991945941760336,1.44269743572906,1.5639622362693,0.,0.,0.,.576746649127296,.838826872176601,0.,.31115444533374},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,0.,0.,0.,0.,0.,0.,1.80736427869175,2.62865111950513,2.84960032604169,0.,0.,0.,1.41847476579742,2.0630457981086,0.,1.03297875354965},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,0.,0.,0.,0.,0.,0.,2.81340752869054,4.09185184033277,4.43578923494485,0.,0.,0.,2.75487590413568,4.00672278096023,0.,2.50302559510192},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,0.,0.,0.,0.,0.,0.,3.86201434387853,5.6169574934443,6.08908644662382,0.,0.,0.,4.43071582310304,6.44408337876027,0.,4.71658607960963},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,0.,0.,0.,0.,0.,0.,4.7630793830775,6.92747671304058,7.50976033055426,0.,0.,0.,6.0685481480974,8.82616530052541,0.,7.17423567654394},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,0.,0.,0.,0.,0.,0.,5.33523806427656,7.7596307926582,8.41186046814944,0.,0.,0.,7.19422103343416,10.463356720588,0.,9.00134910987265},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,0.,0.,0.,0.,0.,0.,.000932318622844532,.00150587381895025,.00181291129080671,0.,0.,0.,1.66187980881431e-005,2.68425539597291e-005,0.,2.74870870898337e-007},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.377498057419076,.454467354307313,.50145308337853,0.,0.,0.,0.,0.,0.,.0249321870529784,.0402702754319125,.0484811118273528,0.,0.,0.,.00229823078236534,.00371208455992771,0.,.000196571591120254},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,0.,0.,0.,0.,0.,0.,.140734880491406,.227313888999126,.273661651286941,0.,0.,0.,.0308216365902836,.0497828687077028,0.,.00626330408481417},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,0.,0.,0.,0.,0.,0.,.439603191878621,.710043670861402,.854816765967696,0.,0.,0.,.170154848292167,.274832793115808,0.,.0611113213546646},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,0.,0.,0.,0.,0.,0.,.991945941760336,1.6021834026584,1.92885774629333,0.,0.,0.,.576746649127296,.931556720779814,0.,.31115444533374},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,0.,0.,0.,0.,0.,0.,1.80736427869175,2.919240785177,3.51445420820192,0.,0.,0.,1.41847476579742,2.29110945565894,0.,1.03297875354965},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,0.,0.,0.,0.,0.,0.,2.81340752869054,4.5441940509206,5.47072443843503,0.,0.,0.,2.75487590413568,4.44965422390436,0.,2.50302559510192},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,0.,0.,0.,0.,0.,0.,3.86201434387853,6.23789565750936,7.50976033055426,0.,0.,0.,4.43071582310304,7.15645788167572,0.,4.71658607960963},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,0.,0.,0.,0.,0.,0.,4.7630793830775,7.69328893020963,9.26189843365359,0.,0.,0.,6.0685481480974,9.80187196802991,0.,7.17423567654394},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,0.,0.,0.,0.,0.,0.,5.33523806427656,8.61743520079898,10.3744718692395,0.,0.,0.,7.19422103343416,11.6200500941132,0.,9.00134910987265},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,0.,0.,0.,0.,0.,0.,.000932318622844532,.00159375505167201,.00203068488848474,0.,0.,0.,1.66187980881431e-005,2.84090575416998e-005,0.,2.74870870898337e-007},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.233717104755142,.399528452143108,.509059650839749,.594469094037471,0.,0.,0.,0.,0.,0.,.0249321870529784,.0426204068987497,.0543048419765405,0.,0.,0.,.00229823078236534,.00392871796107236,0.,.000196571591120254},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,0.,0.,0.,0.,0.,0.,.140734880491406,.240579691570783,.306534899222172,0.,0.,0.,.0308216365902836,.0526881452424316,0.,.00626330408481417},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,0.,0.,0.,0.,0.,0.,.439603191878621,.751481082347236,.957500511953663,0.,0.,0.,.170154848292167,.290871749599041,0.,.0611113213546646},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,0.,0.,0.,0.,0.,0.,.991945941760336,1.6956851626997,2.16055926028883,0.,0.,0.,.576746649127296,.985921403890809,0.,.31115444533374},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,0.,0.,0.,0.,0.,0.,1.80736427869175,3.08960464673337,3.93662342336207,0.,0.,0.,1.41847476579742,2.42481622493138,0.,1.03297875354965},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,0.,0.,0.,0.,0.,0.,2.81340752869054,4.80938849808895,6.1278880563708,0.,0.,0.,2.75487590413568,4.7093314249164,0.,2.50302559510192},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,0.,0.,0.,0.,0.,0.,3.86201434387853,6.60193277208899,8.41186046814944,0.,0.,0.,4.43071582310304,7.57410133403447,0.,4.71658607960963},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,0.,0.,0.,0.,0.,0.,4.7630793830775,8.14226128523923,10.3744718692395,0.,0.,0.,6.0685481480974,10.3738990400802,0.,7.17423567654394},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,0.,0.,0.,0.,0.,0.,5.33523806427656,9.12033977275976,11.620691733626,0.,0.,0.,7.19422103343416,12.2981841540248,0.,9.00134910987265}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[100][35] = 
		{
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.41890639763464,.0772591105674499,.00130129555052609,2.00882372363126e-005,5.33523806427656,.120563624637786,.00203068488848474,7.19422103343416,.16257219522684,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,3.23038451124486,.0729989666585529,.00122954082448108,1.89805519306677e-005,4.7630793830775,.107634206373357,.00181291129080671,6.0685481480974,.13713467930593,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.90882269973526,.0657324385176018,.00110714877688771,1.70911729291799e-005,3.86201434387853,.0872722907753235,.00146995018268015,4.43071582310304,.100123584532725,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,2.48271365058504,.0561033927604955,.000944964222761956,1.45875127898467e-005,2.81340752869054,.0635762837863472,.00107083209499401,2.75487590413568,.0622536090052649,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.9899087271236,.0449671797023573,.000757394052773859,1.16919722098047e-005,1.80736427869175,.0408421116086564,.000687914444399609,1.41847476579742,.0320541383810507,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,1.4741928888764,.0333132347461305,.000561103588047708,8.66181551630733e-006,.991945941760336,.0224156067156837,.000377551968601727,.576746649127296,.013033095369547,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.981387965414959,.0221770216879923,.000373533418059611,5.76627493626533e-006,.439603191878621,.00993398112262371,.000167320660844516,.170154848292167,.00384509276081317,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.555278916264744,.012547975930886,.000211348863933852,3.26261479693214e-006,.140734880491406,.00318027182678497,5.35661561215366e-005,.0308216365902836,.000696495297778531,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.233717104755142,.0052814477899349,8.89568163404885e-005,1.3732357954443e-006,.0249321870529784,.00056340781892775,9.4896263063424e-006,.00229823078236534,5.1934521012163e-005,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319578453782266,.000538274407846207,8.30939904407154e-006,1.22925990446765e-007,.0451952183653594,.00102130388103784,1.72020902954776e-005,2.65550489799432e-007,.000932318622844532,2.10681718666077e-005,3.54856768499235e-007,1.66187980881431e-005,3.75545104141921e-007,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.41890639763464,.399528452143108,.0347994165105669,.00277802311184525,5.33523806427656,.623468196598569,.0543048419765405,7.19422103343416,.840706255138427,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,3.23038451124486,.377498057419076,.0328805422031642,.00262484016485325,4.7630793830775,.556606561402958,.0484811118273528,6.0685481480974,.709161751356258,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.90882269973526,.339920809644861,.029607517992745,.00236355598787826,3.86201434387853,.45130940535495,.039309600832947,4.43071582310304,.517767036891475,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,2.48271365058504,.290126323030972,.0252703573467095,.00201732223678035,2.81340752869054,.328770730954697,.0286363325160947,2.75487590413568,.321930810017305,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.9899087271236,.232537853099397,.020254331227405,.00161689493407496,1.80736427869175,.211205973165025,.0183962984155437,1.41847476579742,.165760907653494,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,1.4741928888764,.172272046833645,.0150051058409842,.00119785142975833,.991945941760336,.115917366756987,.0100965443280308,.576746649127296,.0673977784805412,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.981387965414959,.11468357690207,.00998907972167963,.000797424127052949,.439603191878621,.0513713926084575,.00447451108643047,.170154848292167,.0198840492440458,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.555278916264744,.0648890902881815,.00565191907564414,.000451190375955034,.140734880491406,.0164460743984419,.00143247318181469,.0308216365902836,.00360177183250731,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.233717104755142,.0273118425139666,.002378894865225,.000189906198980042,.0249321870529784,.00291353928576501,.000253772832951388,.00229823078236534,.000268567922156809,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.165262949625141,.0143946049058925,.00114911539118267,8.79094880195249e-005,.0451952183653594,.0052814477899349,.000460020557822313,3.67232519880407e-005,.000932318622844532,.00010894940458837,9.4896263063424e-006,1.66187980881431e-005,1.9420486861126e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.41890639763464,.949223319172042,.196432495608131,.0372561447917688,5.33523806427656,1.48127260473935,.306534899222172,7.19422103343416,1.99739962866362,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,3.23038451124486,.896882204814746,.185601013165114,.0352018040526908,4.7630793830775,1.32241877857201,.273661651286941,6.0685481480974,1.68486841886076,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.90882269973526,.807603957755601,.167125751844475,.0316977147282818,3.86201434387853,1.07224756942001,.221891162762184,4.43071582310304,1.23014153980692,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,2.48271365058504,.689299272303129,.142643752576027,.027054364315645,2.81340752869054,.781112941542527,.161643694787028,2.75487590413568,.764862252961431,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.9899087271236,.552477180456442,.11432975689879,.0216842226834324,1.80736427869175,.501795638836895,.103841706846431,1.41847476579742,.393824565203833,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,1.4741928888764,.409294114646486,.0846994197823303,.0160644186564959,.991945941760336,.275403333686328,.0569920302764555,.576746649127296,.160127627083755,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.981387965414959,.272472022799799,.0563854241050934,.0106942770242832,.439603191878621,.122051191950714,.0252573022041014,.170154848292167,.0472416998608471,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.555278916264744,.154167337347327,.0319034248366451,.00605092661164645,.140734880491406,.0390735559485202,.00808589080538567,.0308216365902836,.00855730247849368,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.233717104755142,.0648890902881815,.0134281635160067,.00254683728723745,.0249321870529784,.00692215890141759,.00143247318181469,.00229823078236534,.000638079548841519,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.392641487077741,.0812533211325227,.0154108182951418,.0028010347388895,.0451952183653594,.012547975930886,.0025966810729899,.000492496548159484,.000932318622844532,.000258848437177504,5.35661561215366e-005,1.66187980881431e-005,4.61403409428573e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.41890639763464,1.67763679592427,.613581734368196,.205677386612004,5.33523806427656,2.61796921368607,.957500511953663,7.19422103343416,3.53016097005959,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,3.23038451124486,1.58513024071019,.579748229567528,.194336131718731,4.7630793830775,2.33721439175002,.854816765967696,6.0685481480974,2.9778000589335,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.90882269973526,1.42734179481246,.522038415063935,.174991351448854,3.86201434387853,1.89506719911644,.693105099882267,4.43071582310304,2.17412559267414,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,2.48271365058504,1.21825264852133,.445565795167568,.149357132360091,2.81340752869054,1.38052214482826,.504914516765946,2.75487590413568,1.35180102876198,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.9899087271236,.976436238050563,.357123450021221,.119710567939023,1.80736427869175,.886862775854878,.324362699711839,1.41847476579742,.696037031940087,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,1.4741928888764,.723377579561695,.264569345968595,.0886857098386406,.991945941760336,.48674190465042,.178021811889788,.576746649127296,.283006110168114,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.981387965414959,.481561169090927,.176127000822248,.0590391454175724,.439603191878621,.215710640970697,.0788943766349663,.170154848292167,.083493960154389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.555278916264744,.272472022799799,.099654380925881,.0334049263288088,.140734880491406,.0690577589939741,.0252573022041014,.0308216365902836,.0151239916064188,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.233717104755142,.11468357690207,.0419445664222885,.0140601460589324,.0249321870529784,.0122340741590526,.00447451108643047,.00229823078236534,.00112772801535997,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.693946085208483,.25380502115662,.0850774241460835,.027329813743018,.0451952183653594,.0221770216879923,.00811106162162033,.00271889116565961,.000932318622844532,.00045748313806202,.000167320660844516,1.66187980881431e-005,8.15474421864669e-006,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.41890639763464,2.52006374830898,1.38452113767375,.697151710458745,5.33523806427656,3.93258500625844,2.16055926028883,7.19422103343416,5.3028347422728,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,3.23038451124486,2.38110493622146,1.30817727028933,.658710074371113,4.7630793830775,3.51084887681561,1.92885774629333,6.0685481480974,4.47310526119946,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.90882269973526,2.14408286840862,1.17795752358561,.59314016960062,3.86201434387853,2.8466770404089,1.5639622362693,4.43071582310304,3.26586487830957,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,2.48271365058504,1.82999940349341,1.00540030297533,.506251960943448,2.81340752869054,2.07375268554642,1.13931817396847,2.75487590413568,2.03060923305097,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.9899087271236,1.46675464678917,.805833905441279,.405763747650646,1.80736427869175,1.33220178323833,.731910663029089,1.41847476579742,1.04555270600537,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,1.4741928888764,1.08662233626593,.596989498475245,.300603752924425,.991945941760336,.731159826532418,.401698661682273,.576746649127296,.425117904255122,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.981387965414959,.723377579561695,.397423100941195,.200115539631623,.439603191878621,.324029949602564,.178021811889788,.170154848292167,.125420533633389,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.555278916264744,.409294114646486,.224865880330915,.113227330974451,.140734880491406,.103735179988284,.0569920302764555,.0308216365902836,.0227185187340072,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.233717104755142,.172272046833645,.0946461336271963,.047657426203958,.0249321870529784,.018377426423439,.0100965443280308,.00229823078236534,.0016940177375493,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.04241178833429,.572700256471795,.288373324563648,.139152498224195,.0451952183653594,.0333132347461305,.0183022662427706,.0092157901163265,.000932318622844532,.000687208741781061,.000377551968601727,1.66187980881431e-005,1.22496569770466e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.41890639763464,3.40165583724254,2.52265163047543,1.71460399590457,5.33523806427656,5.30831839113911,3.93662342336207,7.19422103343416,7.1579216069788,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,3.23038451124486,3.2140851638455,2.38355012000105,1.62005903265462,4.7630793830775,4.73904661479697,3.51445420820192,6.0685481480974,6.03792845818671,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.90882269973526,2.89414583648845,2.14628465152711,1.45879367384687,3.86201434387853,3.84252802245541,2.84960032604169,4.43071582310304,4.40836004025762,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,2.48271365058504,2.47018677889433,1.8318786507244,1.24509718924304,2.81340752869054,2.79921209632888,2.07588224620454,2.75487590413568,2.7409758008707,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.9899087271236,1.97986837015572,1.46826087384217,.99795228596256,1.80736427869175,1.79824497510369,1.33356983668368,1.41847476579742,1.41131765730707,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,1.4741928888764,1.46675464678917,1.08773820111958,.739317408557992,.991945941760336,.986940942882965,.731910663029089,.576746649127296,.573836595005234,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.981387965414959,.976436238050563,.724120424237357,.49217250527751,.439603191878621,.437385113867296,.324362699711839,.170154848292167,.169296308726466,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.555278916264744,.552477180456442,.409714423434642,.278476020673683,.140734880491406,.140024783409284,.103841706846431,.0308216365902836,.0306661218062029,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.233717104755142,.232537853099397,.172448954960703,.117210661865929,.0249321870529784,.0248063882942385,.0183962984155437,.00229823078236534,.00228663474453522,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.40707795466571,1.04348225278067,.709237382898708,.461962142352726,.0451952183653594,.0449671797023573,.0333474444863263,.0226656986159778,.000932318622844532,.000927614481757561,.000687914444399609,1.66187980881431e-005,1.65349456687952e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.41890639763464,4.24408278962725,3.92684926504155,3.33000018565495,5.33523806427656,6.62293418371148,6.1278880563708,7.19422103343416,8.93059537919201,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,3.23038451124486,4.01005985935677,3.71031890564764,3.14638067588647,4.7630793830775,5.91268109986256,5.47072443843503,6.0685481480974,7.53323366045268,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.90882269973526,3.61088691008461,3.3409830372935,2.83318084895721,3.86201434387853,4.79413786374787,4.43578923494485,4.43071582310304,5.50009932589306,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,2.48271365058504,3.08193353386641,2.85156747223424,2.41815246041719,2.81340752869054,3.49244263704704,3.2313921487783,2.75487590413568,3.41978400515968,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.9899087271236,2.47018677889433,2.28554710594336,1.93816257600455,1.80736427869175,2.24358398248714,2.07588224620454,1.41847476579742,1.76083333137235,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,1.4741928888764,1.82999940349341,1.69321197743782,1.43585755873455,.991945941760336,1.23135886476496,1.13931817396847,.576746649127296,.715948389092241,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.981387965414959,1.21825264852133,1.12719161114693,.955867674321917,.439603191878621,.545704422499163,.504914516765946,.170154848292167,.211222882205465,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.555278916264744,.689299272303129,.637776046087675,.540839285781897,.140734880491406,.174702204403594,.161643694787028,.0308216365902836,.0382606489337913,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.233717104755142,.290126323030972,.268440177733536,.227639458852635,.0249321870529784,.0309497405586248,.0286363325160947,.00229823078236534,.00285292446672455,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.75554365779152,1.62432159395843,1.37743795206784,1.11938707577824,.0451952183653594,.0561033927604955,.0519098183396273,.0440199490841533,.000932318622844532,.0011573400854766,.00107083209499401,1.66187980881431e-005,2.0629858427195e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.41890639763464,4.97249626637947,5.3904555359237,5.35569841507852,5.33523806427656,7.7596307926582,8.41186046814944,7.19422103343416,10.463356720588,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,3.23038451124486,4.69830789525221,5.09322047653873,5.06037989777605,4.7630793830775,6.92747671304058,7.50976033055426,6.0685481480974,8.82616530052541,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.90882269973526,4.23062474714147,4.58622658861223,4.55665505598359,3.86201434387853,5.6169574934443,6.08908644662382,4.43071582310304,6.44408337876027,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,2.48271365058504,3.61088691008461,3.91439717424508,3.88915753082078,2.81340752869054,4.09185184033277,4.43578923494485,2.75487590413568,4.00672278096023,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.9899087271236,2.89414583648845,3.13741099245285,3.11718127860418,1.80736427869175,2.62865111950513,2.84960032604169,1.41847476579742,2.0630457981086,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,1.4741928888764,2.14408286840862,2.32430207049861,2.30931520205927,.991945941760336,1.44269743572906,1.5639622362693,.576746649127296,.838826872176601,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.981387965414959,1.42734179481246,1.54731588870638,1.53733894984267,.439603191878621,.639363871519146,.693105099882267,.170154848292167,.247475142499007,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.555278916264744,.807603957755601,.875486474339234,.869841424679861,.140734880491406,.204686407449048,.221891162762184,.0308216365902836,.0448273380617164,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.233717104755142,.339920809644861,.368492586412728,.366116582887396,.0249321870529784,.0362616558162599,.039309600832947,.00229823078236534,.003342572933243,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.05684825592226,2.22973502095477,2.21535791155152,2.10932141870311,.0451952183653594,.0657324385176018,.07125752702776,.0707980655849325,.000932318622844532,.00135597478636112,.00146995018268015,1.66187980881431e-005,2.4170568551556e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.41890639763464,5.52219113340841,6.64812850041597,7.3354543591177,5.33523806427656,8.61743520079898,10.3744718692395,7.19422103343416,11.6200500941132,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,3.23038451124486,5.21769204264788,6.28154410760334,6.93097013741927,4.7630793830775,7.69328893020963,9.26189843365359,6.0685481480974,9.80187196802991,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.90882269973526,4.69830789525221,5.65626105065233,6.24104133632783,3.86201434387853,6.23789565750936,7.50976033055426,4.43071582310304,7.15645788167572,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,2.48271365058504,4.01005985935677,4.82768390215228,5.32680060595541,2.81340752869054,4.5441940509206,5.47072443843503,2.75487590413568,4.44965422390436,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.9899087271236,3.2140851638455,3.86941535783765,4.2694601574129,1.80736427869175,2.919240785177,3.51445420820192,1.41847476579742,2.29110945565894,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,1.4741928888764,2.38110493622146,2.86659610407301,3.1629630633849,.991945941760336,1.6021834026584,1.92885774629333,.576746649127296,.931556720779814,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.981387965414959,1.58513024071019,1.90832755975838,2.10562261484238,.439603191878621,.710043670861402,.854816765967696,.170154848292167,.274832793115808,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.555278916264744,.896882204814746,1.07975041125833,1.19138188446996,.140734880491406,.227313888999126,.273661651286941,.0308216365902836,.0497828687077028,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.233717104755142,.377498057419076,.454467354307313,.50145308337853,.0249321870529784,.0402702754319125,.0484811118273528,.00229823078236534,.00371208455992771,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.28422679337486,2.74996516387086,3.0342740740487,3.20841573119568,.0451952183653594,.0729989666585529,.0878829614946884,.0969688616801027,.000932318622844532,.00150587381895025,.00181291129080671,1.66187980881431e-005,2.68425539597291e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.41890639763464,5.84446047498406,7.4467262413552,8.69612941222294,5.33523806427656,9.12033977275976,11.620691733626,7.19422103343416,12.2981841540248,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,3.23038451124486,5.52219113340841,7.03610637781643,8.21661621987109,4.7630793830775,8.14226128523923,10.3744718692395,6.0685481480974,10.3738990400802,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.90882269973526,4.97249626637947,6.33571201146496,7.39871049163852,3.86201434387853,6.60193277208899,8.41186046814944,4.43071582310304,7.57410133403447,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,2.48271365058504,4.24408278962725,5.40760311670811,6.31488455311819,2.81340752869054,4.80938849808895,6.1278880563708,2.75487590413568,4.7093314249164,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.9899087271236,3.40165583724254,4.33422381683951,5.06141490786373,1.80736427869175,3.08960464673337,3.93662342336207,1.41847476579742,2.42481622493138,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,1.4741928888764,2.52006374830898,3.21094221181666,3.74967040604483,.991945941760336,1.6956851626997,2.16055926028883,.576746649127296,.985921403890809,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.981387965414959,1.67763679592427,2.13756291194806,2.49620076079037,.439603191878621,.751481082347236,.957500511953663,.170154848292167,.290871749599041,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.555278916264744,.949223319172042,1.20945401719121,1.41237482227004,.140734880491406,.240579691570783,.306534899222172,.0308216365902836,.0526881452424316,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.233717104755142,.399528452143108,.509059650839749,.594469094037471,.0249321870529784,.0426204068987497,.0543048419765405,.00229823078236534,.00392871796107236,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.41753189762177,3.08030113246585,3.59711051671708,4.02552569892883,.0451952183653594,.0772591105674499,.098439787300976,.114955901685624,.000932318622844532,.00159375505167201,.00203068488848474,1.66187980881431e-005,2.84090575416998e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{-5.02239597033184e-005,-.000112582670872584,-.000165039185570316,-.000202840374019418,-.000222620323957012,-.000222620323957012,-.000202840374019418,-.000165039185570316,-.000112582670872584,-5.02239597033184e-005,-.000582196454262715,-.00130505902323575,-.00191313526892416,-.00235132688129378,-.00258061618439108,-.00258061618439108,-.00235132688129378,-.00191313526892416,-.00130505902323575,-.000582196454262715,-.00202770964317194,-.00454533988819102,-.00666318526175589,-.00818934597856097,-.00898792887538633,-.00898792887538633,-.00818934597856097,-.0066631852617559,-.00454533988819102,-.00202770964317194,-.00440456112754404,-.0098733205961796,-.0144736732344206,-.0177887771450126,-.0195234471932234,-.0195234471932234,-.0177887771450126,-.0144736732344206,-.00987332059617959,-.00440456112754404,-.00726150487522216,-.016277482266154,-.0238617754892464,-.0293271652321865,-.0321869995374054,-.0321869995374054,-.0293271652321865,-.0238617754892464,-.016277482266154,-.00726150487522216,-.0098017919040896,-.0219718221824466,-.03220932325003,-.0395866663566687,-.0434469544404215,-.0434469544404215,-.0395866663566687,-.03220932325003,-.0219718221824466,-.0098017919040896,-.0111426514503529,-.0249775101026156,-.0366154745926699,-.0450020189787675,-.0493903843956318,-.0493903843956318,-.0450020189787675,-.0366154745926699,-.0249775101026156,-.0111426514503529,-.0106221354093669,-.0238107146832105,-.0349050251580164,-.0428998018485296,-.0470831698638901,-.0470831698638901,-.0428998018485296,-.0349050251580163,-.0238107146832105,-.0106221354093669,-.00804698659228353,-.0180382281362673,-.0264429193024774,-.0324995038175014,-.0356686882642096,-.0356686882642096,-.0324995038175014,-.0264429193024774,-.0180382281362673,-.00804698659228353,-.00379931823220995,-.00851660037567366,-.0124848058669686,-.0153443722038775,-.0168406764553548,-.0168406764553548,-.0153443722038775,-.0124848058669686,-.00851660037567366,-.00379931823220995},
			{-8.45935382972169e-007,-1.89625958134751e-006,-2.77979847617661e-006,-3.41649385058338e-006,-3.74965275769605e-006,-3.74965275769605e-006,-3.41649385058338e-006,-2.77979847617661e-006,-1.89625958134751e-006,-8.45935382972169e-007,-5.07100227590783e-005,-.000113672236039353,-.000166636420263325,-.00020480344528281,-.000224774823832688,-.000224774823832688,-.00020480344528281,-.000166636420263325,-.000113672236039353,-5.07100227590783e-005,-.000419614707621413,-.000940613699128858,-.00137888111587085,-.00169470517141149,-.00185996410278327,-.00185996410278327,-.00169470517141149,-.00137888111587085,-.000940613699128858,-.000419614707621413,-.00161093167609136,-.00361108506296509,-.00529362585909322,-.006506097600196,-.00714053877317793,-.00714053877317793,-.006506097600196,-.00529362585909322,-.00361108506296509,-.00161093167609136,-.00398946534464676,-.00894283657654873,-.0131096415979721,-.0161123226329801,-.017683513460247,-.017683513460247,-.0161123226329801,-.0131096415979721,-.00894283657654873,-.00398946534464676,-.00726896179664528,-.016294197796655,-.023886279416173,-.0293572816296078,-.0322200527309955,-.0322200527309955,-.0293572816296078,-.023886279416173,-.016294197796655,-.00726896179664528,-.0103097688766848,-.0231105098656633,-.0338785684935517,-.0416382417346981,-.045698588896762,-.045698588896762,-.0416382417346981,-.0338785684935517,-.0231105098656633,-.0103097688766848,-.0115149706602782,-.0258121056087878,-.0378389396388598,-.0465057110062776,-.0510407087352223,-.0510407087352223,-.0465057110062776,-.0378389396388598,-.0258121056087878,-.0115149706602782,-.00968771265055556,-.021716100667658,-.0318344514317794,-.0391259325039141,-.0429412921921929,-.0429412921921929,-.0391259325039141,-.0318344514317794,-.021716100667658,-.00968771265055556,-.00484090582829009,-.0108514364620189,-.0159075301916269,-.0195510500287827,-.0214575678641308,-.0214575678641308,-.0195510500287827,-.0159075301916269,-.0108514364620189,-.00484090582829009},
			{-1.30587941023263e-008,-2.92727599954231e-008,-4.2912043492978e-008,-5.27407774218851e-008,-5.7883786756776e-008,-5.7883786756776e-008,-5.27407774218851e-008,-4.2912043492978e-008,-2.92727599954231e-008,-1.30587941023263e-008,-4.04816026703356e-006,-9.07440786578996e-006,-1.33025169157498e-005,-1.63493748303836e-005,-1.79436817686314e-005,-1.79436817686314e-005,-1.63493748303836e-005,-1.33025169157498e-005,-9.07440786578996e-006,-4.04816026703356e-006,-7.95857439793226e-005,-.000178400422298479,-.000261523911036573,-.00032142431960244,-.000352767965960796,-.000352767965960796,-.00032142431960244,-.000261523911036573,-.000178400422298479,-7.95857439793226e-005,-.000539996871821585,-.00121046389913573,-.00177446470693278,-.00218089469839968,-.00239356433166203,-.00239356433166203,-.00218089469839968,-.00177446470693278,-.00121046389913573,-.000539996871821585,-.00200882638285304,-.0045030109301789,-.00660113364491101,-.00811308182877234,-.00890422783798524,-.00890422783798524,-.00811308182877234,-.00660113364491101,-.0045030109301789,-.00200882638285304,-.00494059139703164,-.0110748929087561,-.0162351034290897,-.0199536518580156,-.0218994293529137,-.0218994293529137,-.0199536518580156,-.0162351034290897,-.0110748929087561,-.00494059139703164,-.00874276804581894,-.0195979007476136,-.0287293022381269,-.0353095683983229,-.0387527758890071,-.0387527758890071,-.0353095683983229,-.0287293022381268,-.0195979007476136,-.00874276804581894,-.0114407232759916,-.0256456717219007,-.0375949579236433,-.0462058468122232,-.0507116033273556,-.0507116033273556,-.0462058468122232,-.0375949579236432,-.0256456717219007,-.0114407232759916,-.0106892900748741,-.0239612494409041,-.0351256997397901,-.0431710205566739,-.0473808364254027,-.0473808364254027,-.0431710205566739,-.0351256997397901,-.0239612494409041,-.0106892900748741,-.00565310744519306,-.0126720780015972,-.0185764773269063,-.0228313027353587,-.0250576939422306,-.0250576939422306,-.0228313027353587,-.0185764773269063,-.0126720780015972,-.00565310744519306},
			{-1.93186678290668e-010,-4.33049730595595e-010,-6.3482394133133e-010,-7.80226376246059e-010,-8.56310039258135e-010,-8.56310039258135e-010,-7.80226376246059e-010,-6.3482394133133e-010,-4.33049730595595e-010,-1.93186678290668e-010,-3.09691872038552e-007,-6.94209264016414e-007,-1.01766755630007e-006,-1.25075791566716e-006,-1.37272539416124e-006,-1.37272539416124e-006,-1.25075791566716e-006,-1.01766755630007e-006,-6.94209264016414e-007,-3.09691872038552e-007,-1.44653210137677e-005,-3.24256487218327e-005,-4.75339820521491e-005,-5.84213419665859e-005,-6.41182907371273e-005,-6.41182907371273e-005,-5.84213419665859e-005,-4.75339820521491e-005,-3.24256487218327e-005,-1.44653210137677e-005,-.000173465688278746,-.000388842906981744,-.000570019490439081,-.000700578873068661,-.000768895721248536,-.000768895721248536,-.000700578873068661,-.000570019490439081,-.000388842906981744,-.000173465688278746,-.000969344894307206,-.00217289592143802,-.00318533012606401,-.00391490997676346,-.00429667186083074,-.00429667186083074,-.00391490997676346,-.00318533012606401,-.00217289592143802,-.000969344894307206,-.00321805680650706,-.00721363732442475,-.010574743162472,-.0129968216385873,-.01426420483388,-.01426420483388,-.0129968216385873,-.010574743162472,-.00721363732442475,-.00321805680650706,-.00710488740889465,-.0159264065491336,-.023347120347727,-.0286946315642816,-.0314927844397286,-.0314927844397286,-.0286946315642816,-.023347120347727,-.0159264065491336,-.00710488740889465,-.0108931213951503,-.0244181603372121,-.0357955026643469,-.0439942376324665,-.0482843293960987,-.0482843293960987,-.0439942376324665,-.0357955026643469,-.0244181603372121,-.0108931213951503,-.0113027648830266,-.0253364224225416,-.0371416177061049,-.0456486718663717,-.050100095546689,-.050100095546689,-.0456486718663717,-.0371416177061049,-.0253364224225416,-.0113027648830266,-.00632639147606921,-.0141813201023726,-.0207889322741007,-.0255505065864879,-.0280420605663306,-.0280420605663306,-.0255505065864879,-.0207889322741007,-.0141813201023726,-.00632639147606921},
			{-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262448,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-.000233400654593986,-.00270558184440079,-.0094231669672911,-.0204688649887572,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813583,-.0373959351122988,-.017656181783494,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000314832681450395,-.00364954240783074,-.0127108509152548,-.0276103238093599,-.0455192912851962,-.0614432722233051,-.0698485514743261,-.0665856573917093,-.0504431427033939,-.0238163130465199,-.000286859607981103,-.00332527836563145,-.0115814841517041,-.0251571299005444,-.041474874823967,-.055984000459719,-.0636424655841521,-.0606694816070398,-.0459612390764785,-.0217002192803048,-.000233400654593986,-.00270558184440079,-.00942316696729111,-.0204688649887571,-.0337456465246067,-.0455508617823574,-.0517821007699867,-.0493631599813582,-.0373959351122988,-.017656181783494,-.000159215940061732,-.00184563217065341,-.00642808131650616,-.0139629838950143,-.0230198361857748,-.0310728489254499,-.0353235335470956,-.0336734356401934,-.0255099068755807,-.0120442917585213,-7.10274049799049e-005,-.000823350121715917,-.00286761447838841,-.00622899008387319,-.0102693186794246,-.0138618270485456,-.0157580888044127,-.0150219679596996,-.0113801575768671,-.00537304737262447},
			{-1.60504953819679e-006,-1.86057442617858e-005,-6.4801231236956e-005,-.000140760283448081,-.000232062049974506,-.000313244150044994,-.000356095132090808,-.000339460560939038,-.000257164353231333,-.000121418024586148,-1.86057442617858e-005,-.000215677903576663,-.000751176276838271,-.00163169408403043,-.0026900647312993,-.00363112814186739,-.00412785699311923,-.00393502894053338,-.00298105078728298,-.00140747849861324,-6.4801231236956e-005,-.000751176276838271,-.00261624296938716,-.00568296458123414,-.00936912301075426,-.0126467165764205,-.014376754391578,-.0137051609821032,-.0103825871557633,-.00490205274061561,-.000140760283448081,-.00163169408403043,-.00568296458123414,-.0123444522582422,-.0203514714995352,-.0274710121398013,-.0312289748912649,-.0297701495437364,-.0225529034413806,-.0106481670189776,-.000232062049974506,-.0026900647312993,-.00936912301075426,-.0203514714995352,-.033552107742962,-.0452896174679006,-.0514851189152542,-.0490800512895761,-.0371814611144508,-.0175549196574778,-.000313244150044994,-.00363112814186739,-.0126467165764205,-.0274710121398013,-.0452896174679006,-.0611332517796597,-.0694961210432554,-.0662496903395319,-.0501886249195276,-.0236961445777924,-.000356095132090808,-.00412785699311923,-.014376754391578,-.0312289748912649,-.0514851189152542,-.0694961210432554,-.0790030089920025,-.075312475042366,-.0570542977980853,-.0269377153005406,-.000339460560939038,-.00393502894053338,-.0137051609821032,-.0297701495437364,-.0490800512895761,-.0662496903395319,-.075312475042366,-.0717943401064784,-.0543890724391648,-.0256793511684561,-.000257164353231333,-.00298105078728298,-.0103825871557633,-.0225529034413806,-.0371814611144508,-.0501886249195276,-.0570542977980853,-.0543890724391648,-.0412034040065755,-.0194538467631362,-.000121418024586148,-.00140747849861324,-.00490205274061561,-.0106481670189776,-.0175549196574778,-.0236961445777924,-.0269377153005406,-.0256793511684561,-.0194538467631362,-.00918497301395746},
			{-2.70342721665942e-008,-3.13381451640588e-007,-1.09146420736547e-006,-2.3708625325406e-006,-3.90868223643537e-006,-5.27605373253693e-006,-5.99780411074261e-006,-5.71762364703134e-006,-4.33148694252304e-006,-2.04507577148045e-006,-1.62058306631703e-006,-1.878580901668e-005,-6.54283718476915e-005,-.000142122549078592,-.000234307926063146,-.00031627569935075,-.000359541389428153,-.000342745830361303,-.000259653167200857,-.000122593097538308,-1.34099819434037e-005,-.000155448594362042,-.000541405931790853,-.00117603401917912,-.00193884850644768,-.00261711448526345,-.00297512891523367,-.00283614921805116,-.00214857501357473,-.00101443194029589,-5.14819054148423e-005,-.000596778568799888,-.00207849712916292,-.00451488095923928,-.00744338179155761,-.0100472946912819,-.0114217383779821,-.0108881851148137,-.00824853725324792,-.00389447871149425,-.00012749471661469,-.00147791955052476,-.00514738917162678,-.0111810832137794,-.0184335028884922,-.0248820819487376,-.0282858858081064,-.0269645434541597,-.0204274669151108,-.00964465972428348,-.000232300357136792,-.00269282718941148,-.00937874426988752,-.0203723706574197,-.033586562784559,-.0453361258919233,-.0515379895702794,-.0491304521532631,-.0372196431803629,-.0175729470044231,-.00032947799961708,-.00381931102740118,-.0133021315121917,-.0288946948441916,-.0476367477719961,-.0643014769902532,-.0730977511924463,-.0696830745129148,-.0527896458390131,-.0249241951142798,-.000367993651863299,-.004265784708565,-.0148571375279271,-.0322724560897768,-.0532054364658069,-.0718182560454652,-.0816428060009037,-.0778289569952232,-.0589607032197774,-.0278378088689286,-.000309598422927644,-.00358886685037487,-.0124995263493677,-.0271512876888108,-.0447625091834733,-.0604217455831262,-.0686872826563242,-.065478635899881,-.049604498988284,-.0234203543456317,-.000154704919937309,-.00179334039722375,-.00624595630961966,-.0135673746279838,-.0223676216885663,-.0301924706996799,-.0343227218781454,-.0327193757277722,-.0247871419109467,-.0117030442522349},
			{-4.17330922690109e-010,-4.83770265983857e-009,-1.68490485682806e-008,-3.65992560176709e-008,-6.03387416602869e-008,-8.14470002666849e-008,-9.2588737297087e-008,-8.82635618042971e-008,-6.6865622688263e-008,-3.15700512824513e-008,-1.29370479862118e-007,-1.49966340979313e-006,-5.22312002294925e-006,-1.13455846575704e-005,-1.8704705399135e-005,-2.52481590386612e-005,-2.87020412883381e-005,-2.7361258713725e-005,-2.07280055781616e-005,-9.78655657087546e-006,-2.54338890004804e-006,-2.94829799993105e-005,-.000102685137321483,-.000223051148247843,-.000367729486213021,-.000496372027946554,-.000564274425659415,-.000537915000222613,-.000407507024506811,-.000192401074639172,-1.72571365319964e-005,-.000200044834358053,-.000696728461200002,-.00151342333799558,-.00249507967511344,-.00336793160368666,-.00382865585555338,-.00364980463714188,-.00276497407043368,-.00130545966206281,-6.41977629259773e-005,-.000744180868411951,-.00259187893037573,-.0056300413732684,-.00928187206303727,-.0125289426924045,-.0142428693476609,-.0135775302227536,-.0102858982161417,-.00485640185665492,-.00015789065592134,-.00183026946862863,-.00637457515237077,-.0138467585905541,-.0228282233120914,-.0308142042579516,-.0350295069642396,-.0333931441681891,-.025297567112414,-.0119440372937759,-.000279400029347063,-.00323880688356021,-.0112803159519127,-.0245029367570041,-.0403963504117509,-.054528176627927,-.0619874888523968,-.0590918151941233,-.0447660499753643,-.0211359206213315,-.000365620865418288,-.00423827935328899,-.0147613401837082,-.0320643665046187,-.0528623731064046,-.0713551790777345,-.0811163813127856,-.0773271236259448,-.0585805304731015,-.0276583134477209,-.000341606679368254,-.00395990675875596,-.0137918069785563,-.0299583607055857,-.0493903424240987,-.0666685303985192,-.0757886113280989,-.0722482342377723,-.0547329279680274,-.0258416997153403,-.000180661133614816,-.00209422498813209,-.00729389568128642,-.0158436931512049,-.0261204355502155,-.0352581287359617,-.0400813487105939,-.0382089949857042,-.0289459000949523,-.0136665675675336},
			{-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779044,-.0447078783667129,-.0376133897003122,-.0187952392881457,-4.81475219575139e-006,-.000288622746300124,-.00238829215039021,-.00916882894461163,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619087,-.0551388873133415,-.0275526505160409,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550778,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-6.49458908735487e-006,-.000389321415157618,-.00322155232641603,-.0123677759491016,-.0306287437708179,-.0558067683550777,-.0791522778068628,-.088405100787603,-.0743764998227976,-.0371655977491602,-5.91754093321632e-006,-.000354729972810523,-.00293531546086304,-.0112688916030352,-.0279073614294862,-.050848303356817,-.0721195502254918,-.0805502543085171,-.0677681029932591,-.0338634119926563,-4.81475219575139e-006,-.000288622746300124,-.00238829215039022,-.00916882894461162,-.0227065653177011,-.041372249554411,-.0586794019211047,-.0655389659619086,-.0551388873133415,-.0275526505160409,-3.28441793937709e-006,-.000196886088238747,-.00162919071725787,-.00625458279978264,-.0154894473150464,-.0282223784536199,-.0400285772779045,-.044707878366713,-.0376133897003122,-.0187952392881457,-1.46520306329223e-006,-8.7832335875546e-005,-.00072679399323529,-.00279021551063458,-.00690995667226622,-.0125902111506184,-.0178570435094928,-.0199445142321406,-.0167796105206252,-.00838469484962211},
			{-3.31100861878508e-008,-1.98480079913645e-006,-1.64238066106426e-005,-6.30521996263144e-005,-.000156148500303414,-.000284508671025808,-.000403526490267504,-.000450698337824376,-.000379179080671673,-.000189474057272324,-3.8381232568618e-007,-2.30078232481278e-005,-.000190384868710001,-.000730901491494302,-.00181007438983512,-.00329802628977656,-.00467768159306037,-.00522449794427726,-.00439544626910526,-.00219638445415982,-1.33676519027876e-006,-8.01330628639868e-005,-.000663084138302037,-.00254562869914522,-.00630423848900844,-.0114865689446415,-.0162917173481382,-.0181962029908938,-.0153087307912168,-.00764970295714498,-2.90370172750447e-006,-.000174063863093159,-.0014403416335721,-.00552957729977356,-.0136939743226831,-.0249509564807709,-.0353886293208596,-.0395255250831765,-.0332534003485589,-.0166165724938767,-4.7871385229751e-006,-.000286967430772327,-.00237459476467777,-.00911624367502254,-.0225763381244699,-.0411349705180729,-.0583428625235449,-.0651630854416681,-.0548226535530947,-.0273946299483073,-6.46181975011305e-006,-.000387357040725739,-.00320529754367204,-.0123053726445113,-.030474202256271,-.0555251876749137,-.0787529041667412,-.0879590407639987,-.0740012230234841,-.0369780736440852,-7.34577982424567e-006,-.000440346472755273,-.00364377388066238,-.0139887155008144,-.0346429935762571,-.0631208884080524,-.0895260958832347,-.0999916079356561,-.0841243971631835,-.0420365775924714,-7.00263023825107e-006,-.000419776197925987,-.0034735592093105,-.0133352488781344,-.033024686304908,-.0601722693038774,-.0853439881325168,-.0953206159256855,-.0801946235041213,-.0400728876177358,-5.30496641835636e-006,-.00031800888486993,-.00263145622867657,-.010102353697401,-.0250184353398111,-.0455845671001885,-.0646538480039295,-.0722118188820862,-.0607528557338689,-.0303579249318969,-2.50469606272866e-006,-.000150145267478098,-.0012424203162536,-.00476974282864777,-.0118122475336369,-.02152237671891,-.030525780139787,-.0340942136430102,-.028683958870995,-.014333243427598},
			{-5.57681903477072e-010,-3.34305226934559e-008,-2.76630501080463e-007,-1.06200480743303e-006,-2.63005032304177e-006,-4.79205449098542e-006,-6.79670297198979e-006,-7.59123082633588e-006,-6.3866131386048e-006,-3.19136145764326e-006,-3.34305226934559e-008,-2.00400952691774e-006,-1.65827547682883e-005,-6.36624132755433e-005,-.000157659691772648,-.000287262120951629,-.00040743178419295,-.000455060156746716,-.000382848742505824,-.000191307770554455,-2.76630501080463e-007,-1.65827547682883e-005,-.000137218786643261,-.000526792998299499,-.0013046005871693,-.00237703326355233,-.00337141179786177,-.00376552650213992,-.00316799232989985,-.00158303131884398,-1.06200480743303e-006,-6.36624132755433e-005,-.000526792998299499,-.00202239700441925,-.00500845745477193,-.00912560525126807,-.0129430974646002,-.0144561327553171,-.0121621551894091,-.00607737347965237,-2.63005032304177e-006,-.000157659691772648,-.0013046005871693,-.00500845745477193,-.0124034232751763,-.0225995220276466,-.0320535250214294,-.0358005503901207,-.030119524846696,-.0150505891984046,-4.79205449098542e-006,-.000287262120951629,-.00237703326355233,-.00912560525126807,-.0225995220276466,-.0411772124958644,-.0584027754849978,-.0652300021690509,-.0548789515711727,-.0274227618111799,-6.79670297198979e-006,-.00040743178419295,-.00337141179786177,-.0129430974646002,-.0320535250214294,-.0584027754849978,-.0828342662751547,-.0925175100657338,-.0778363296880565,-.0388944589534263,-7.59123082633588e-006,-.000455060156746716,-.00376552650213992,-.0144561327553171,-.0358005503901207,-.0652300021690509,-.0925175100657338,-.103332715477079,-.0869353196354016,-.0434411827319366,-6.3866131386048e-006,-.000382848742505824,-.00316799232989985,-.0121621551894091,-.030119524846696,-.0548789515711727,-.0778363296880565,-.0869353196354016,-.0731399515169607,-.0365477002003157,-3.19136145764326e-006,-.000191307770554455,-.00158303131884398,-.00607737347965237,-.0150505891984046,-.0274227618111799,-.0388944589534263,-.0434411827319366,-.0365477002003157,-.0182627191053361},
			{-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596958,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.15767573513552e-007,-3.58873635372629e-005,-.000705535931921592,-.00478712866332407,-.0178084556759705,-.0437988587058275,-.0775055517780142,-.101423206654711,-.0947616728508053,-.0501153878844612,-1.0548155484377e-007,-3.26987496607673e-005,-.00064284863920488,-.00436178939679936,-.0162261636575447,-.0399073037160311,-.0706191367966457,-.0924116936244465,-.0863420411133478,-.0456626054707174,-8.58240869859561e-008,-2.66050338314996e-005,-.000523047822073146,-.00354892941386556,-.013202267289822,-.0324702068581795,-.0574586044762537,-.0751899158472865,-.0702513994795803,-.0371529546538127,-5.85455199908462e-008,-1.81488157315799e-005,-.000356800844596959,-.00242092779827145,-.00900602186035781,-.0221497858175122,-.0391958014952271,-.0512913434438014,-.0479224988818083,-.0253441560031944,-2.61175882046526e-008,-8.09632053406712e-006,-.000159171487958645,-.00107999374364317,-.00401765276570609,-.00988118279406327,-.0174855360916379,-.0228814465519832,-.0213785801497482,-.0113062148903861},
			{-5.90195050823199e-010,-1.82957487178445e-007,-3.59689507657609e-006,-2.44052765295024e-005,-9.07893469973411e-005,-.000223291106959738,-.000395131310801382,-.000517065986523608,-.000483104798924764,-.000255493425333229,-6.8415447117756e-009,-2.12084433297021e-006,-4.16952301711737e-005,-.000282906117811307,-.00105243067689039,-.00258839190514417,-.00458036462030578,-.00599383214211218,-.00560015384355916,-.0029616813806621,-2.38281529966161e-008,-7.38660717382125e-006,-.000145218713843646,-.000985322839048874,-.00366547033510042,-.00901501063419505,-.0159527758058908,-.0208756874850881,-.0195045604772907,-.0103151261942612,-5.17591642291992e-008,-1.60450796946238e-005,-.0003154419589321,-.00214030381004998,-.00796208086622,-.0195822737922466,-.0346523854772107,-.0453458619764426,-.0423675200132299,-.0224063657308857,-8.53318667863116e-008,-2.64524480537304e-005,-.000520048026649204,-.00352857551549082,-.0131265493548066,-.0322839830104985,-.0571290666185282,-.0747586849808788,-.0698484921013427,-.0369398742075263,-1.15183452383392e-007,-3.57062889348363e-005,-.000701976053853711,-.00476297455073299,-.0177186006765078,-.0435778655721675,-.0771144869130937,-.10091146198997,-.0942835398662266,-.049862523838675,-1.30940247999038e-007,-4.05908160548146e-005,-.000798004545609922,-.00541453703618982,-.0201424589971083,-.0495392038284753,-.0876635474261488,-.114715886574844,-.107181282005842,-.0566835869406137,-1.24823526157934e-007,-3.86946631537408e-005,-.000760726688663539,-.00516160321748368,-.0192015273831553,-.0472250373695059,-.0835684464667091,-.109357066963176,-.102174432709151,-.0540356789095071,-9.45622704554012e-008,-2.9313826607454e-005,-.000576301960778015,-.00391026382973348,-.0145464567574012,-.0357761625028231,-.0633087550044376,-.0828453806800583,-.0774040490351568,-.0409356844864075,-4.46467946852167e-008,-1.38402810304592e-005,-.000272096209150128,-.00184619875908625,-.00686798936951685,-.0168914195291221,-.029890705593759,-.0391147619872944,-.0365456822095551,-.0193274452036918},
			{-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807,-9.68328634942698e-010,-1.55229910445862e-006,-7.2505954733558e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039878,-.0236458445483912,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.0641631467467009,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.9147674569896e-009,-3.06950729481149e-006,-.000143372856643853,-.00171930309977533,-.00960765035478182,-.0318957316434237,-.0704200067856531,-.107967042743436,-.112027219286111,-.0627039536356004,-1.74463921457135e-009,-2.79677972194095e-006,-.000130634091932628,-.00156654198328503,-.00875400483105963,-.029061776666107,-.064163146746701,-.0983741059332837,-.102073533343426,-.0571326695688283,-1.41950948611112e-009,-2.27557383366154e-006,-.00010628921507616,-.00127460232871745,-.00712261469039877,-.0236458445483913,-.0522057481598334,-.0800411772208763,-.0830511819588279,-.0464854657297896,-9.68328634942698e-010,-1.55229910445862e-006,-7.25059547335579e-005,-.000869479172304126,-.00485874298682691,-.0161301834173291,-.0356125276698691,-.0546006663821835,-.0566539628254899,-.0317103957495351,-4.31978544868352e-010,-6.92492077737805e-007,-3.23454410928852e-005,-.000387881070632081,-.00216752107662602,-.00719579377252414,-.0158870112137336,-.0243577599189904,-.0252737506041315,-.0141462413882807},
			{-.00537304737262447,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.017656181783494,-.0373959351122988,-.0493631599813582,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887572,-.00942316696729111,-.00270558184440079,-.000233400654593986,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0238163130465199,-.0504431427033939,-.0665856573917093,-.0698485514743261,-.0614432722233051,-.0455192912851962,-.0276103238093599,-.0127108509152548,-.00364954240783074,-.000314832681450395,-.0217002192803048,-.0459612390764785,-.0606694816070398,-.0636424655841521,-.055984000459719,-.041474874823967,-.0251571299005444,-.0115814841517041,-.00332527836563145,-.000286859607981103,-.017656181783494,-.0373959351122988,-.0493631599813583,-.0517821007699867,-.0455508617823574,-.0337456465246067,-.0204688649887571,-.0094231669672911,-.00270558184440079,-.000233400654593986,-.0120442917585213,-.0255099068755807,-.0336734356401934,-.0353235335470956,-.0310728489254499,-.0230198361857748,-.0139629838950143,-.00642808131650616,-.00184563217065341,-.000159215940061732,-.00537304737262448,-.0113801575768671,-.0150219679596996,-.0157580888044127,-.0138618270485456,-.0102693186794246,-.00622899008387319,-.00286761447838841,-.000823350121715917,-7.10274049799049e-005},
			{-.000121418024586148,-.000257164353231333,-.000339460560939038,-.000356095132090808,-.000313244150044994,-.000232062049974506,-.000140760283448081,-6.4801231236956e-005,-1.86057442617858e-005,-1.60504953819679e-006,-.00140747849861324,-.00298105078728298,-.00393502894053338,-.00412785699311923,-.00363112814186739,-.0026900647312993,-.00163169408403043,-.000751176276838271,-.000215677903576663,-1.86057442617858e-005,-.00490205274061561,-.0103825871557633,-.0137051609821032,-.014376754391578,-.0126467165764205,-.00936912301075426,-.00568296458123414,-.00261624296938716,-.000751176276838271,-6.4801231236956e-005,-.0106481670189775,-.0225529034413806,-.0297701495437364,-.031228974891265,-.0274710121398013,-.0203514714995352,-.0123444522582422,-.00568296458123414,-.00163169408403043,-.000140760283448081,-.0175549196574778,-.0371814611144508,-.0490800512895761,-.0514851189152542,-.0452896174679006,-.033552107742962,-.0203514714995352,-.00936912301075425,-.0026900647312993,-.000232062049974506,-.0236961445777924,-.0501886249195276,-.0662496903395319,-.0694961210432554,-.0611332517796597,-.0452896174679006,-.0274710121398013,-.0126467165764205,-.00363112814186739,-.000313244150044994,-.0269377153005406,-.0570542977980853,-.075312475042366,-.0790030089920025,-.0694961210432554,-.0514851189152542,-.031228974891265,-.014376754391578,-.00412785699311923,-.000356095132090808,-.0256793511684561,-.0543890724391648,-.0717943401064784,-.075312475042366,-.0662496903395319,-.0490800512895761,-.0297701495437364,-.0137051609821032,-.00393502894053338,-.000339460560939038,-.0194538467631362,-.0412034040065755,-.0543890724391648,-.0570542977980853,-.0501886249195276,-.0371814611144508,-.0225529034413806,-.0103825871557633,-.00298105078728298,-.000257164353231333,-.00918497301395746,-.0194538467631362,-.0256793511684561,-.0269377153005406,-.0236961445777924,-.0175549196574778,-.0106481670189776,-.00490205274061561,-.00140747849861324,-.000121418024586148},
			{-2.04507577148045e-006,-4.33148694252304e-006,-5.71762364703134e-006,-5.99780411074261e-006,-5.27605373253693e-006,-3.90868223643537e-006,-2.3708625325406e-006,-1.09146420736547e-006,-3.13381451640588e-007,-2.70342721665942e-008,-.000122593097538308,-.000259653167200857,-.000342745830361303,-.000359541389428153,-.00031627569935075,-.000234307926063146,-.000142122549078592,-6.54283718476915e-005,-1.878580901668e-005,-1.62058306631703e-006,-.00101443194029589,-.00214857501357473,-.00283614921805116,-.00297512891523367,-.00261711448526345,-.00193884850644768,-.00117603401917912,-.000541405931790853,-.000155448594362042,-1.34099819434037e-005,-.00389447871149424,-.00824853725324792,-.0108881851148137,-.0114217383779821,-.0100472946912819,-.00744338179155761,-.00451488095923928,-.00207849712916292,-.000596778568799888,-5.14819054148423e-005,-.00964465972428348,-.0204274669151108,-.0269645434541598,-.0282858858081064,-.0248820819487376,-.0184335028884922,-.0111810832137794,-.00514738917162678,-.00147791955052476,-.00012749471661469,-.0175729470044231,-.0372196431803629,-.0491304521532631,-.0515379895702794,-.0453361258919233,-.033586562784559,-.0203723706574197,-.00937874426988752,-.00269282718941148,-.000232300357136792,-.0249241951142798,-.0527896458390131,-.0696830745129148,-.0730977511924463,-.0643014769902532,-.0476367477719961,-.0288946948441916,-.0133021315121917,-.00381931102740118,-.000329477999617081,-.0278378088689286,-.0589607032197774,-.0778289569952232,-.0816428060009037,-.0718182560454652,-.0532054364658069,-.0322724560897768,-.0148571375279271,-.004265784708565,-.000367993651863299,-.0234203543456317,-.049604498988284,-.065478635899881,-.0686872826563242,-.0604217455831262,-.0447625091834733,-.0271512876888108,-.0124995263493677,-.00358886685037487,-.000309598422927644,-.0117030442522349,-.0247871419109467,-.0327193757277722,-.0343227218781454,-.0301924706996799,-.0223676216885663,-.0135673746279838,-.00624595630961966,-.00179334039722375,-.000154704919937309},
			{-3.15700512824513e-008,-6.6865622688263e-008,-8.82635618042971e-008,-9.2588737297087e-008,-8.14470002666849e-008,-6.03387416602869e-008,-3.65992560176709e-008,-1.68490485682806e-008,-4.83770265983857e-009,-4.17330922690109e-010,-9.78655657087546e-006,-2.07280055781616e-005,-2.7361258713725e-005,-2.87020412883381e-005,-2.52481590386612e-005,-1.8704705399135e-005,-1.13455846575704e-005,-5.22312002294925e-006,-1.49966340979313e-006,-1.29370479862118e-007,-.000192401074639172,-.000407507024506811,-.000537915000222613,-.000564274425659415,-.000496372027946554,-.000367729486213021,-.000223051148247843,-.000102685137321483,-2.94829799993105e-005,-2.54338890004804e-006,-.00130545966206281,-.00276497407043368,-.00364980463714188,-.00382865585555338,-.00336793160368666,-.00249507967511344,-.00151342333799558,-.000696728461200002,-.000200044834358053,-1.72571365319964e-005,-.00485640185665492,-.0102858982161417,-.0135775302227536,-.0142428693476609,-.0125289426924045,-.00928187206303727,-.0056300413732684,-.00259187893037573,-.000744180868411951,-6.41977629259773e-005,-.0119440372937759,-.025297567112414,-.0333931441681891,-.0350295069642396,-.0308142042579516,-.0228282233120914,-.0138467585905541,-.00637457515237077,-.00183026946862863,-.00015789065592134,-.0211359206213315,-.0447660499753643,-.0590918151941233,-.0619874888523968,-.054528176627927,-.0403963504117509,-.0245029367570041,-.0112803159519127,-.00323880688356021,-.000279400029347063,-.0276583134477209,-.0585805304731015,-.0773271236259449,-.0811163813127856,-.0713551790777345,-.0528623731064046,-.0320643665046187,-.0147613401837082,-.00423827935328899,-.000365620865418288,-.0258416997153403,-.0547329279680274,-.0722482342377723,-.0757886113280989,-.0666685303985192,-.0493903424240987,-.0299583607055857,-.0137918069785563,-.00395990675875596,-.000341606679368254,-.0136665675675336,-.0289459000949523,-.0382089949857042,-.0400813487105939,-.0352581287359617,-.0261204355502155,-.0158436931512049,-.00729389568128642,-.00209422498813209,-.000180661133614816},
			{-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.0012034333021735,-.0131809781472584,-.0413377545099444,-.0766395839246475,-.101270663310552,-.101270663310552,-.0766395839246475,-.0413377545099444,-.0131809781472584,-.0012034333021735,-.00109650752807023,-.0120098402958391,-.0376648701110469,-.0698301107088946,-.0922727038483408,-.0922727038483408,-.0698301107088946,-.0376648701110469,-.0120098402958391,-.00109650752807023,-.000892163161694358,-.00977169496377035,-.0306456715917109,-.0568166207314213,-.0750768281074222,-.0750768281074222,-.0568166207314213,-.0306456715917109,-.00977169496377035,-.000892163161694358,-.000608595535966736,-.00666583220325412,-.0209051659250277,-.0387578676530293,-.0512141998263455,-.0512141998263455,-.0387578676530293,-.0209051659250277,-.00666583220325412,-.000608595535966735,-.000271498956607684,-.00297367690223135,-.00932594868172963,-.0172901705751442,-.02284703221535,-.02284703221535,-.0172901705751442,-.00932594868172963,-.00297367690223135,-.000271498956607684},
			{-6.13522731444136e-006,-6.71979884668803e-005,-.000210744143550815,-.000390716517328129,-.000516288304945811,-.000516288304945811,-.000390716517328129,-.000210744143550815,-6.71979884668803e-005,-6.13522731444136e-006,-7.11195933109096e-005,-.000778959501602571,-.00244294742701878,-.0045291883068169,-.00598481725240988,-.00598481725240988,-.0045291883068169,-.00244294742701878,-.000778959501602571,-7.11195933109096e-005,-.000247699696758929,-.00271300809456183,-.00850844765415349,-.0157745357915389,-.0208442898723963,-.0208442898723963,-.0157745357915389,-.00850844765415349,-.00271300809456183,-.000247699696758929,-.000538049645975034,-.00589315636598079,-.0184819251214903,-.0342652151339464,-.0452776605429554,-.0452776605429554,-.0342652151339464,-.0184819251214903,-.00589315636598079,-.000538049645975034,-.000887046408080575,-.00971565212579472,-.0304699118679336,-.0564907648096207,-.0746462458462128,-.0746462458462128,-.0564907648096207,-.0304699118679336,-.00971565212579472,-.000887046408080575,-.00119736121515858,-.0131144717225921,-.0411291792262685,-.0762528884413686,-.100759688353677,-.100759688353677,-.0762528884413686,-.0411291792262685,-.0131144717225921,-.00119736121515858,-.00136115710384714,-.014908497220736,-.0467555435824129,-.0866840845325749,-.11454335070129,-.11454335070129,-.0866840845325749,-.0467555435824129,-.014908497220736,-.00136115710384714,-.00129757222819962,-.0142120640616299,-.0445714125837813,-.0826347380463795,-.109192590902875,-.109192590902875,-.0826347380463795,-.0445714125837813,-.0142120640616299,-.00129757222819962,-.000982998796422244,-.0107666005511246,-.0337658621019407,-.0626014076726695,-.082720778930882,-.082720778930882,-.0626014076726695,-.0337658621019407,-.0107666005511246,-.000982998796422244,-.000464114759811927,-.00508336149236723,-.0159422728047302,-.0295567373954438,-.039055932300929,-.039055932300929,-.0295567373954438,-.0159422728047302,-.00508336149236723,-.000464114759811927},
			{-1.03337249781483e-007,-1.13183342085305e-006,-3.54961912345676e-006,-6.58094122280528e-006,-8.69597991941751e-006,-8.69597991941751e-006,-6.58094122280528e-006,-3.54961912345676e-006,-1.13183342085305e-006,-1.03337249781483e-007,-6.19460350490148e-006,-6.78483246903409e-005,-.000212783706840731,-.000394497837425948,-.000521284897762289,-.000521284897762289,-.000394497837425948,-.000212783706840731,-6.78483246903409e-005,-6.19460350490148e-006,-5.12590331676487e-005,-.000561430529479367,-.00176074014710611,-.00326438612530762,-.0043135222202759,-.0043135222202759,-.00326438612530762,-.00176074014710611,-.000561430529479367,-5.12590331676487e-005,-.000196787192431025,-.00215537303015378,-.00675961072102864,-.0125322180484573,-.0165599285581662,-.0165599285581662,-.0125322180484573,-.00675961072102864,-.00215537303015378,-.000196787192431025,-.000487342632915848,-.00533777200871002,-.0167401467828093,-.0310359839203022,-.0410105916179671,-.0410105916179671,-.0310359839203022,-.0167401467828093,-.00533777200871002,-.000487342632915848,-.000887957326235526,-.00972562924066487,-.0305012017707475,-.0565487758254857,-.0747229009342903,-.0747229009342903,-.0565487758254857,-.0305012017707475,-.00972562924066487,-.000887957326235526,-.00125941435131387,-.0137941280277271,-.0432606952017087,-.080204687454718,-.105981550045219,-.105981550045219,-.080204687454718,-.0432606952017087,-.0137941280277271,-.00125941435131387,-.00140663864321037,-.0154066479494616,-.048317827678703,-.0895805360822827,-.118370688411972,-.118370688411972,-.0895805360822827,-.048317827678703,-.0154066479494616,-.00140663864321037,-.00118342559270231,-.0129618374762797,-.0406504926725559,-.0753654106685247,-.0995869854483788,-.0995869854483788,-.0753654106685247,-.0406504926725559,-.0129618374762797,-.00118342559270231,-.00059135237137032,-.00647697107125416,-.020312865791922,-.0376597519885992,-.0497631624376525,-.0497631624376525,-.0376597519885992,-.020312865791922,-.00647697107125416,-.00059135237137032},
			{-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474138,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512756,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.3917769310411e-005,-.00192110412837146,-.0143143236721634,-.0469036492894419,-.0931002257512757,-.125669410778967,-.118656774340934,-.0749854323824078,-.0265530630600676,-.00256579529169265,-3.09041550678753e-005,-.00175041286888237,-.013042487387887,-.0427362317853027,-.0848282145898218,-.114503607901172,-.108114048436568,-.068322931527997,-.0241938074059463,-.00233782283384287,-2.51448786159929e-005,-.00142420716629368,-.010611898675073,-.0347719378959308,-.0690196885948747,-.0931647966250701,-.0879659909371525,-.0555903183952796,-.0196850665920198,-.00190214782619252,-1.71527603191504e-005,-.000971533191363834,-.00723898322535009,-.0237199282474139,-.0470822784174149,-.0635530380200689,-.0600066352208138,-.0379213366690509,-.0134283101651034,-.0012975638599234,-7.65197286929417e-006,-.000433408121119428,-.00322936379982008,-.0105816349108646,-.0210037516044799,-.0283514789248146,-.0267694024835557,-.0169169879342966,-.00599046819007428,-.000578852806636916},
			{-1.72916292281226e-007,-9.79398733224426e-006,-7.29759010167018e-005,-.000239119650097383,-.000474634569865839,-.000640675640139909,-.000604924495042971,-.000382283481676095,-.000135370258905769,-1.308068949145e-005,-2.00444673906772e-006,-.000113531962266802,-.000845937099932474,-.00277187647596092,-.00550196689545652,-.00742671601811804,-.00701228852107834,-.00443143250491902,-.00156921288590388,-.000151631434204312,-6.9812104698908e-006,-.000395416105699515,-.00294628179627482,-.00965406199030398,-.019162588931357,-.0258662236377037,-.0244228300443723,-.0154340658681427,-.00546535121888794,-.000528111291459933,-1.51644829240855e-005,-.000858917061539686,-.00639987007723783,-.0209704117690521,-.041624694440043,-.0561862313643539,-.0530509129846532,-.0335256513631973,-.0118717557063213,-.00114715559656894,-2.50006671481639e-005,-.00141603902163566,-.0105510370774588,-.0345725131033195,-.0686238453397381,-.0926304758086336,-.0874614864268777,-.0552714955631914,-.0195721683597038,-.00189123858561784,-3.37466325590279e-005,-.00191141093432354,-.0142420987911849,-.0466669905017148,-.0926304758086336,-.12503532855168,-.118058075331442,-.0746070831112178,-.026419085942947,-.00255284922006524,-3.83630838022064e-005,-.00217288696066628,-.0161903807287985,-.0530509129846532,-.105302083100265,-.142139835110194,-.134208111862577,-.0848131373294056,-.0300331479365882,-.00290207232951189,-3.65709968299352e-005,-.00207138306607571,-.0154340658681427,-.0505727011047881,-.100383018401244,-.135499937544735,-.12793873555065,-.0808511899721722,-.0286301842585129,-.0027665053860117,-2.77050056147299e-005,-.00156921288590388,-.0116923496377051,-.0383122443879716,-.0760469314348448,-.102650374774569,-.0969222524410603,-.0612502492768556,-.0216893025727857,-.0020958150965657,-1.308068949145e-005,-.000740890898631601,-.00552044627469675,-.0180888096370989,-.0359049303476018,-.0484655263124446,-.0457610406806555,-.028918799122657,-.0102404250042771,-.000989521781403059},
			{-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.0551532224231419,-.0216893025727857,-.0022181246897901,-5.68214139596138e-007,-.000166430787646874,-.00294628179627482,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-7.66460494801417e-007,-.000224497447284009,-.00397422106567472,-.023015338460633,-.0686238453397381,-.12503532855168,-.147295396613918,-.10905966387144,-.0428883017939756,-.0043861069664707,-6.98360017958869e-007,-.000204550713807632,-.00362110912907025,-.0209704117690521,-.0625265753276467,-.113925864261887,-.134208111862577,-.0993696470156369,-.0390776503345112,-.0039963987193725,-5.68214139596138e-007,-.000166430787646874,-.00294628179627483,-.0170623806831829,-.0508741670314013,-.0926947208812389,-.109197183183113,-.0808511899721722,-.0317951670932725,-.0032516326843114,-3.87611373882146e-007,-.000113531962266802,-.00200982737900492,-.0116392260548274,-.0347041799983487,-.0632323724607349,-.0744896461530335,-.055153222423142,-.0216893025727857,-.0022181246897901,-1.72916292281226e-007,-5.06474455947119e-005,-.000896598814998893,-.00519234457509943,-.0154817906189702,-.028208427654087,-.0332303804636016,-.0246042592436186,-.00967575782281376,-.000989521781403059},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461163,-.00238829215039022,-.000288622746300124,-4.81475219575139e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0371655977491602,-.0743764998227976,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641603,-.000389321415157618,-6.49458908735487e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030352,-.00293531546086304,-.000354729972810523,-5.91754093321632e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461162,-.00238829215039021,-.000288622746300124,-4.81475219575139e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978264,-.00162919071725787,-.000196886088238747,-3.28441793937709e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.00072679399323529,-8.7832335875546e-005,-1.46520306329223e-006},
			{-.000189474057272324,-.000379179080671673,-.000450698337824376,-.000403526490267504,-.000284508671025808,-.000156148500303414,-6.30521996263144e-005,-1.64238066106426e-005,-1.98480079913645e-006,-3.31100861878508e-008,-.00219638445415982,-.00439544626910526,-.00522449794427726,-.00467768159306037,-.00329802628977656,-.00181007438983512,-.000730901491494302,-.000190384868710001,-2.30078232481278e-005,-3.8381232568618e-007,-.00764970295714498,-.0153087307912168,-.0181962029908938,-.0162917173481382,-.0114865689446415,-.00630423848900844,-.00254562869914522,-.000663084138302037,-8.01330628639868e-005,-1.33676519027876e-006,-.0166165724938767,-.0332534003485589,-.0395255250831765,-.0353886293208596,-.0249509564807709,-.0136939743226831,-.00552957729977355,-.0014403416335721,-.000174063863093159,-2.90370172750447e-006,-.0273946299483073,-.0548226535530947,-.0651630854416681,-.0583428625235449,-.0411349705180729,-.0225763381244699,-.00911624367502254,-.00237459476467777,-.000286967430772327,-4.7871385229751e-006,-.0369780736440852,-.0740012230234841,-.0879590407639987,-.0787529041667412,-.0555251876749137,-.030474202256271,-.0123053726445112,-.00320529754367204,-.000387357040725739,-6.46181975011305e-006,-.0420365775924714,-.0841243971631835,-.0999916079356561,-.0895260958832346,-.0631208884080524,-.0346429935762571,-.0139887155008144,-.00364377388066238,-.000440346472755273,-7.34577982424566e-006,-.0400728876177358,-.0801946235041213,-.0953206159256855,-.0853439881325168,-.0601722693038774,-.033024686304908,-.0133352488781344,-.0034735592093105,-.000419776197925987,-7.00263023825107e-006,-.0303579249318969,-.0607528557338689,-.0722118188820862,-.0646538480039295,-.0455845671001885,-.0250184353398111,-.010102353697401,-.00263145622867657,-.00031800888486993,-5.30496641835636e-006,-.014333243427598,-.028683958870995,-.0340942136430102,-.030525780139787,-.02152237671891,-.0118122475336369,-.00476974282864777,-.0012424203162536,-.000150145267478098,-2.50469606272866e-006},
			{-3.19136145764326e-006,-6.3866131386048e-006,-7.59123082633588e-006,-6.79670297198979e-006,-4.79205449098542e-006,-2.63005032304177e-006,-1.06200480743303e-006,-2.76630501080463e-007,-3.34305226934559e-008,-5.57681903477072e-010,-.000191307770554455,-.000382848742505824,-.000455060156746716,-.00040743178419295,-.000287262120951629,-.000157659691772648,-6.36624132755433e-005,-1.65827547682883e-005,-2.00400952691774e-006,-3.34305226934559e-008,-.00158303131884398,-.00316799232989985,-.00376552650213992,-.00337141179786177,-.00237703326355233,-.0013046005871693,-.000526792998299499,-.000137218786643261,-1.65827547682883e-005,-2.76630501080463e-007,-.00607737347965237,-.0121621551894091,-.0144561327553171,-.0129430974646002,-.00912560525126807,-.00500845745477193,-.00202239700441925,-.000526792998299499,-6.36624132755433e-005,-1.06200480743303e-006,-.0150505891984046,-.030119524846696,-.0358005503901207,-.0320535250214294,-.0225995220276466,-.0124034232751763,-.00500845745477193,-.0013046005871693,-.000157659691772648,-2.63005032304177e-006,-.0274227618111799,-.0548789515711727,-.0652300021690509,-.0584027754849978,-.0411772124958644,-.0225995220276466,-.00912560525126807,-.00237703326355233,-.000287262120951629,-4.79205449098542e-006,-.0388944589534263,-.0778363296880565,-.0925175100657338,-.0828342662751547,-.0584027754849978,-.0320535250214294,-.0129430974646002,-.00337141179786177,-.00040743178419295,-6.7967029719898e-006,-.0434411827319366,-.0869353196354016,-.103332715477079,-.0925175100657338,-.0652300021690509,-.0358005503901207,-.0144561327553171,-.00376552650213992,-.000455060156746716,-7.59123082633588e-006,-.0365477002003157,-.0731399515169607,-.0869353196354016,-.0778363296880565,-.0548789515711727,-.030119524846696,-.0121621551894091,-.00316799232989985,-.000382848742505824,-6.3866131386048e-006,-.0182627191053361,-.0365477002003157,-.0434411827319366,-.0388944589534263,-.0274227618111799,-.0150505891984046,-.00607737347965237,-.00158303131884398,-.000191307770554455,-3.19136145764326e-006},
			{-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474138,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512757,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00256579529169265,-.0265530630600677,-.0749854323824079,-.118656774340934,-.125669410778967,-.0931002257512756,-.0469036492894419,-.0143143236721634,-.00192110412837146,-3.3917769310411e-005,-.00233782283384287,-.0241938074059463,-.068322931527997,-.108114048436568,-.114503607901172,-.0848282145898218,-.0427362317853027,-.013042487387887,-.00175041286888237,-3.09041550678753e-005,-.00190214782619252,-.0196850665920198,-.0555903183952796,-.0879659909371525,-.0931647966250701,-.0690196885948747,-.0347719378959308,-.010611898675073,-.00142420716629368,-2.51448786159929e-005,-.0012975638599234,-.0134283101651034,-.0379213366690509,-.0600066352208138,-.0635530380200689,-.0470822784174149,-.0237199282474139,-.00723898322535009,-.000971533191363834,-1.71527603191504e-005,-.000578852806636916,-.00599046819007428,-.0169169879342966,-.0267694024835557,-.0283514789248146,-.0210037516044799,-.0105816349108646,-.00322936379982008,-.000433408121119428,-7.65197286929417e-006},
			{-1.308068949145e-005,-.000135370258905769,-.000382283481676095,-.000604924495042971,-.000640675640139909,-.000474634569865839,-.000239119650097383,-7.29759010167018e-005,-9.79398733224426e-006,-1.72916292281226e-007,-.000151631434204312,-.00156921288590388,-.00443143250491902,-.00701228852107834,-.00742671601811804,-.00550196689545652,-.00277187647596092,-.000845937099932474,-.000113531962266802,-2.00444673906772e-006,-.000528111291459933,-.00546535121888794,-.0154340658681427,-.0244228300443723,-.0258662236377037,-.019162588931357,-.00965406199030398,-.00294628179627483,-.000395416105699515,-6.9812104698908e-006,-.00114715559656894,-.0118717557063213,-.0335256513631973,-.0530509129846532,-.0561862313643539,-.041624694440043,-.0209704117690521,-.00639987007723783,-.000858917061539686,-1.51644829240855e-005,-.00189123858561784,-.0195721683597038,-.0552714955631914,-.0874614864268777,-.0926304758086336,-.0686238453397381,-.0345725131033195,-.0105510370774588,-.00141603902163566,-2.50006671481639e-005,-.00255284922006524,-.026419085942947,-.0746070831112178,-.118058075331442,-.12503532855168,-.0926304758086336,-.0466669905017148,-.0142420987911849,-.00191141093432354,-3.37466325590279e-005,-.00290207232951189,-.0300331479365882,-.0848131373294056,-.134208111862577,-.142139835110194,-.105302083100265,-.0530509129846532,-.0161903807287985,-.00217288696066627,-3.83630838022064e-005,-.0027665053860117,-.0286301842585129,-.0808511899721722,-.12793873555065,-.135499937544735,-.100383018401244,-.0505727011047881,-.0154340658681427,-.00207138306607571,-3.65709968299352e-005,-.0020958150965657,-.0216893025727857,-.0612502492768556,-.0969222524410603,-.102650374774569,-.0760469314348448,-.0383122443879716,-.0116923496377051,-.00156921288590388,-2.77050056147299e-005,-.000989521781403059,-.0102404250042771,-.028918799122657,-.0457610406806555,-.0484655263124446,-.0359049303476018,-.0180888096370989,-.00552044627469675,-.000740890898631601,-1.308068949145e-005},
			{-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-9.16757562810742e-005,-.00490619869945355,-.0329175945957259,-.0920604883605493,-.146461642152594,-.146461642152594,-.0920604883605493,-.0329175945957259,-.00490619869945355,-9.16757562810742e-005,-8.35303100904523e-005,-.00447027998846552,-.0299928464793087,-.0838808584928213,-.133448436987573,-.133448436987573,-.0838808584928213,-.0299928464793087,-.00447027998846552,-8.35303100904523e-005,-6.79636606588214e-005,-.00363720177570271,-.0244034008505957,-.0682488811091861,-.108579080779807,-.108579080779807,-.0682488811091861,-.0244034008505957,-.00363720177570271,-6.79636606588214e-005,-4.63619013436546e-005,-.00248114342661157,-.0166469559131689,-.04655646653118,-.0740680031402153,-.0740680031402153,-.04655646653118,-.0166469559131689,-.00248114342661157,-4.63619013436546e-005,-2.06823860795433e-005,-.00110685637949844,-.00742632979379346,-.0207691830444954,-.0330422824062289,-.0330422824062289,-.0207691830444954,-.00742632979379346,-.00110685637949844,-2.06823860795433e-005},
			{-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596958,-1.81488157315799e-005,-5.85455199908462e-008,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0501153878844612,-.0947616728508053,-.101423206654711,-.0775055517780142,-.0437988587058275,-.0178084556759705,-.00478712866332407,-.000705535931921591,-3.58873635372629e-005,-1.15767573513552e-007,-.0456626054707174,-.0863420411133478,-.0924116936244465,-.0706191367966457,-.0399073037160311,-.0162261636575447,-.00436178939679936,-.00064284863920488,-3.26987496607673e-005,-1.0548155484377e-007,-.0371529546538127,-.0702513994795803,-.0751899158472865,-.0574586044762537,-.0324702068581795,-.013202267289822,-.00354892941386556,-.000523047822073146,-2.66050338314996e-005,-8.5824086985956e-008,-.0253441560031944,-.0479224988818083,-.0512913434438014,-.0391958014952271,-.0221497858175122,-.00900602186035781,-.00242092779827145,-.000356800844596959,-1.81488157315799e-005,-5.85455199908462e-008,-.0113062148903861,-.0213785801497482,-.0228814465519832,-.0174855360916379,-.00988118279406327,-.00401765276570609,-.00107999374364317,-.000159171487958645,-8.09632053406712e-006,-2.61175882046526e-008},
			{-.000255493425333229,-.000483104798924764,-.000517065986523608,-.000395131310801382,-.000223291106959738,-9.07893469973411e-005,-2.44052765295024e-005,-3.59689507657609e-006,-1.82957487178445e-007,-5.90195050823199e-010,-.0029616813806621,-.00560015384355916,-.00599383214211218,-.00458036462030578,-.00258839190514417,-.00105243067689039,-.000282906117811307,-4.16952301711737e-005,-2.12084433297021e-006,-6.8415447117756e-009,-.0103151261942612,-.0195045604772907,-.0208756874850881,-.0159527758058908,-.00901501063419505,-.00366547033510042,-.000985322839048874,-.000145218713843646,-7.38660717382125e-006,-2.38281529966161e-008,-.0224063657308857,-.0423675200132299,-.0453458619764426,-.0346523854772107,-.0195822737922466,-.00796208086622,-.00214030381004998,-.0003154419589321,-1.60450796946238e-005,-5.17591642291991e-008,-.0369398742075263,-.0698484921013427,-.0747586849808788,-.0571290666185282,-.0322839830104985,-.0131265493548066,-.00352857551549082,-.000520048026649204,-2.64524480537304e-005,-8.53318667863116e-008,-.049862523838675,-.0942835398662266,-.10091146198997,-.0771144869130937,-.0435778655721675,-.0177186006765078,-.00476297455073299,-.000701976053853711,-3.57062889348363e-005,-1.15183452383392e-007,-.0566835869406137,-.107181282005842,-.114715886574844,-.0876635474261487,-.0495392038284753,-.0201424589971083,-.00541453703618982,-.000798004545609921,-4.05908160548146e-005,-1.30940247999038e-007,-.0540356789095071,-.102174432709151,-.109357066963176,-.0835684464667091,-.0472250373695059,-.0192015273831553,-.00516160321748368,-.000760726688663539,-3.86946631537408e-005,-1.24823526157934e-007,-.0409356844864075,-.0774040490351568,-.0828453806800583,-.0633087550044376,-.0357761625028231,-.0145464567574012,-.00391026382973348,-.000576301960778015,-2.9313826607454e-005,-9.45622704554012e-008,-.0193274452036918,-.0365456822095551,-.0391147619872944,-.029890705593759,-.0168914195291221,-.00686798936951685,-.00184619875908625,-.000272096209150128,-1.38402810304592e-005,-4.46467946852167e-008},
			{-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007,-.0022181246897901,-.0216893025727857,-.0551532224231419,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627483,-.000166430787646874,-5.68214139596138e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0043861069664707,-.0428883017939756,-.10905966387144,-.147295396613918,-.12503532855168,-.0686238453397381,-.023015338460633,-.00397422106567472,-.000224497447284009,-7.66460494801417e-007,-.0039963987193725,-.0390776503345112,-.0993696470156369,-.134208111862577,-.113925864261887,-.0625265753276467,-.0209704117690521,-.00362110912907025,-.000204550713807632,-6.98360017958869e-007,-.0032516326843114,-.0317951670932725,-.0808511899721722,-.109197183183113,-.0926947208812389,-.0508741670314013,-.0170623806831829,-.00294628179627482,-.000166430787646874,-5.68214139596138e-007,-.0022181246897901,-.0216893025727857,-.055153222423142,-.0744896461530335,-.0632323724607349,-.0347041799983487,-.0116392260548274,-.00200982737900492,-.000113531962266802,-3.87611373882146e-007,-.000989521781403059,-.00967575782281376,-.0246042592436186,-.0332303804636016,-.028208427654087,-.0154817906189702,-.00519234457509943,-.000896598814998893,-5.06474455947119e-005,-1.72916292281226e-007},
			{-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.2505954733558e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483912,-.00712261469039878,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.064163146746701,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0627039536356004,-.112027219286111,-.107967042743436,-.0704200067856531,-.0318957316434237,-.00960765035478182,-.00171930309977533,-.000143372856643853,-3.06950729481149e-006,-1.9147674569896e-009,-.0571326695688283,-.102073533343426,-.0983741059332837,-.0641631467467009,-.029061776666107,-.00875400483105963,-.00156654198328503,-.000130634091932628,-2.79677972194095e-006,-1.74463921457135e-009,-.0464854657297896,-.0830511819588279,-.0800411772208763,-.0522057481598334,-.0236458445483913,-.00712261469039877,-.00127460232871745,-.00010628921507616,-2.27557383366154e-006,-1.41950948611112e-009,-.0317103957495352,-.0566539628254899,-.0546006663821835,-.0356125276698691,-.0161301834173291,-.00485874298682691,-.000869479172304126,-7.25059547335579e-005,-1.55229910445862e-006,-9.68328634942698e-010,-.0141462413882807,-.0252737506041315,-.0243577599189904,-.0158870112137336,-.00719579377252414,-.00216752107662602,-.000387881070632081,-3.23454410928852e-005,-6.92492077737805e-007,-4.31978544868352e-010}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036796,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.00022262032405145,-.00258061618548581,-.0089879288791991,-.0195234472015054,-.0321869995510595,-.0434469544588521,-.0493903844165837,-.0470831698838632,-.0356686882793406,-.0168406764624987,-.000202840374105465,-.00235132688229124,-.00818934598203497,-.0177887771525588,-.0293271652446273,-.0395866663734618,-.0450020189978578,-.0428998018667281,-.032499503831288,-.0153443722103867,-.000165039185640327,-.00191313526973573,-.00666318526458248,-.0144736732405605,-.0238617754993688,-.0322093232636935,-.0366154746082026,-.0349050251728234,-.0264429193136947,-.0124848058722647,-.000112582670920343,-.00130505902378937,-.0045453398901192,-.00987332060036795,-.0162774822730591,-.0219718221917673,-.0249775101132113,-.0238107146933112,-.0180382281439193,-.00851660037928648,-5.02239597246239e-005,-.000582196454509688,-.00202770964403212,-.0044045611294125,-.00726150487830256,-.00980179190824761,-.0111426514550797,-.0106221354138729,-.00804698659569713,-.00379931823382166},
			{-1.1349414128987e-006,-1.31562479400023e-005,-4.58213900489777e-005,-9.95325509741368e-005,-.000164092649236315,-.000221497062722267,-.000251797282715359,-.000240034864648708,-.000181842658097301,-8.58555085657902e-005,-1.31562479400023e-005,-.000152507308211391,-.000531161839358935,-.00115378195194428,-.00190216301383433,-.00256759533314912,-.00291883567237296,-.00278248564875054,-.0021079212273054,-.000995237591006242,-4.58213900489777e-005,-.000531161839358935,-.00184996314537334,-.00401846279369377,-.00662497041642304,-.00894257905329049,-.0101659005244202,-.00969101227025514,-.00734159778603745,-.00346627473553786,-9.95325509741368e-005,-.00115378195194428,-.00401846279369377,-.00872884590413951,-.0143906635082426,-.0194249389752361,-.0220822199209436,-.0210506746248672,-.0159473109630529,-.00752939110831236,-.000164092649236315,-.00190216301383433,-.00662497041642304,-.0143906635082426,-.0237249229144092,-.0320245956373459,-.0364054767247764,-.0347048370969986,-.0262912632953882,-.0124132027362624,-.000221497062722267,-.00256759533314912,-.00894257905329049,-.0194249389752361,-.0320245956373459,-.0432277369007862,-.0491411784688113,-.0468456053029506,-.0354887170283886,-.0167557045233543,-.000251797282715359,-.00291883567237296,-.0101659005244202,-.0220822199209436,-.0364054767247764,-.0491411784688113,-.0558635634071245,-.053253961824449,-.0403434808795061,-.0190478411637101,-.000240034864648708,-.00278248564875054,-.00969101227025514,-.0210506746248672,-.0347048370969986,-.0468456053029506,-.053253961824449,-.0507662647534973,-.038458881954326,-.0181580433524764,-.000181842658097301,-.0021079212273054,-.00734159778603745,-.0159473109630529,-.0262912632953882,-.0354887170283886,-.0403434808795061,-.038458881954326,-.0291352063887049,-.0137559469700066,-8.58555085657902e-005,-.000995237591006242,-.00346627473553786,-.00752939110831236,-.0124132027362624,-.0167557045233543,-.0190478411637101,-.0181580433524764,-.0137559469700066,-.0064947567048982},
			{-1.91161171784846e-008,-2.21594149611604e-007,-7.71781742654136e-007,-1.67645297446285e-006,-2.763855715616e-006,-3.73073337316569e-006,-4.24108796005352e-006,-4.04297045415503e-006,-3.06282379048705e-006,-1.44608694643564e-006,-1.14592527597117e-006,-1.32835729492742e-005,-4.62648454277032e-005,-.000100495818239503,-.000165680723448716,-.000223640691794434,-.000254234154648952,-.00024235790093583,-.000183602515332728,-8.66864106188711e-005,-9.48228917027153e-006,-.000109918755228315,-.000382831805844932,-.000831581630087009,-.00137097292696425,-.00185057940015954,-.00210373383142091,-.00200546034506998,-.00151927196238748,-.000717311704224692,-3.64032044368432e-005,-.000421986172976531,-.00146971941509561,-.00319250294337036,-.0052632657411594,-.00710451021065933,-.00807638866214051,-.00769910953153036,-.00583259662818021,-.00275381230681073,-9.01523787074882e-005,-.00104504693649993,-.00363975378962373,-.00790621976356029,-.0130344548969134,-.0175942888806334,-.020001141672057,-.0190668115330659,-.0144444003819494,-.00681980429507682,-.000164261157846817,-.00190411636669871,-.00663177367400153,-.0144054414445077,-.0237492863079749,-.0320574820593633,-.0364428619234804,-.0347404758894987,-.0263182620931215,-.0124259499955376,-.00023297612784248,-.00270066072764841,-.00940602739898747,-.0204316346700341,-.0336842673921381,-.0454680104321141,-.051687915571302,-.0492733745350088,-.0373279165590501,-.0176240673855734,-.000260210806734789,-.00301636529530396,-.01050558269779,-.0228200725526466,-.0376219249308874,-.0507831758761377,-.057730181773567,-.0550333832785173,-.0416915130812311,-.0196843034297875,-.000218919144354553,-.00253771208734517,-.00883849984558959,-.01919885964777,-.0316518737949094,-.0427246260442284,-.0485692433603775,-.0463003874798657,-.0350756776212305,-.016560691379957,-.000109392897999453,-.00126808315618723,-.00441655806269213,-.00959358260487664,-.0158162969791731,-.0213493007781521,-.0242698293952193,-.023136092459402,-.0175271561360877,-.00827530195346473},
			{-2.95097525510867e-010,-3.42077235703851e-009,-1.19140765023158e-008,-2.58795821233052e-008,-4.26659334075081e-008,-5.75917262110692e-008,-6.54701240215422e-008,-6.24117630999618e-008,-4.72811352436055e-008,-2.23233973501177e-008,-9.14787436199951e-008,-1.06042216684182e-006,-3.69330358815301e-006,-8.02253985001058e-006,-1.32262240313143e-005,-1.78531444734238e-005,-2.02954080342344e-005,-1.93473315833786e-005,-1.46569133086574e-005,-6.92014051755746e-006,-1.79844753889302e-006,-2.08476150925998e-005,-7.26093569462482e-005,-.000157720979519105,-.000260024013412071,-.000350988027044924,-.000399002272939181,-.000380363344459719,-.000288150980485938,-.000136048104620829,-1.2202638268856e-005,-.000141453058953237,-.000492661419690163,-.00107015190538498,-.0017642877583389,-.0023814872761676,-.0027072685190056,-.00258080160960999,-.00195513191552442,-.000923099379853643,-4.53946735139408e-005,-.00052621533862221,-.00183273516816672,-.00398104043444918,-.00656327467961113,-.00885930034123408,-.010071229501942,-.00960076369480724,-.00727322838114722,-.00343399468591358,-.000111645553517426,-.00129419595300744,-.0045075053186138,-.00979113689941693,-.0161419915106792,-.0217889327934133,-.0247696019225699,-.023612518692696,-.0178880812574289,-.00844570976740211,-.00019756565546715,-.00229018231092328,-.00797638790562856,-.0173261927444337,-.0285645333188729,-.0385572434695171,-.0438317737278189,-.0417842232474103,-.031654377512867,-.0149453528019069,-.000258532993348771,-.00299691607206422,-.0104378437460552,-.0226729309958482,-.0373793425030134,-.0504557310119577,-.0573579433067165,-.0546785334999813,-.0414226903539633,-.0195573810002261,-.000241552399543637,-.00280007692272149,-.00975228024192592,-.0211837600137409,-.0349242460624195,-.0471417699489713,-.0535906410209485,-.0510872163717606,-.0387020245305973,-.0182728411109243,-.000127746712709587,-.00148084069082919,-.00515756309886556,-.0112031828692115,-.0184699371099763,-.0249312619277241,-.0283417934798407,-.027017839463842,-.0204678422500889,-.00966372260509664},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590033e-006,-.000275291812791722,-.00227798149655705,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293013e-006,-.000250831969330614,-.00207558146784555,-.00796830967106448,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.4045439282468e-006,-.000204087101167344,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779698e-006,-.00013921948815164,-.00115201180432314,-.00442265791238582,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214257e-006,-6.21068403214379e-005,-.000513920961277905,-.0019729803090621,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{-2.3412366475277e-008,-1.40346610474412e-006,-1.16133850303456e-005,-4.45846379362571e-005,-.000110413663465783,-.000201178010641796,-.000285336317731836,-.000318691851029199,-.000268120099297756,-.000133978390791535,-2.71396298267277e-007,-1.62689878433847e-005,-.000134622431735668,-.00051682540115133,-.00127991587584218,-.00233205675464775,-.00330762037555709,-.00369427792566825,-.0031080498640454,-.00155307834203885,-9.45235731149608e-007,-5.66626321633216e-005,-.000468871290814305,-.00180003131602355,-.00445776978697114,-.00812223079546563,-.0115199838170822,-.0128666585301024,-.0108249073566845,-.00540915683648703,-2.05322718260318e-006,-.000123081737985171,-.00101847533659284,-.00391000160679661,-.00968310210751827,-.0176429905292981,-.0250235397762786,-.0279487668236465,-.0235137048901805,-.0117496910935978,-3.38501811296807e-006,-.000202916616332326,-.0016790920611167,-.00644615772325802,-.0159638817863846,-.0290868166049117,-.0412546337351168,-.0460772596109979,-.0387654701002611,-.0193709286096546,-4.56919656531553e-006,-.000273902790309784,-.00226648762944902,-.00870121244425688,-.0215485150723445,-.0392622367419453,-.0556867125891282,-.062196434207296,-.0523267666298084,-.0261474466358464,-5.19425072819777e-006,-.000311371977038976,-.00257653722080653,-.00989151559332482,-.024496295684836,-.0446332082396283,-.0633045095088931,-.0707047440517022,-.0594849317130106,-.0297243490813528,-4.95160732891562e-006,-.00029682659621248,-.00245617727240442,-.00942944491302682,-.0233519796389196,-.0425482196753798,-.0603473127579275,-.0674018539257125,-.0567061621094263,-.0283358105837029,-3.75117772937632e-006,-.000224866239028424,-.00186072054418367,-.00714344280726179,-.0176907052881247,-.0322331565225002,-.0457171743654439,-.0510614668268089,-.0429587562772,-.0214662945877596,-1.77108757123392e-006,-.000106168736824841,-.000878523830938623,-.00337271749954253,-.00835252033429224,-.0152186185292077,-.0215849861435475,-.0241082496725556,-.0202826218343075,-.0101351336267259},
			{-3.94340655797694e-010,-2.36389493013898e-008,-1.95607303248633e-007,-7.50950801186721e-007,-1.85972641877534e-006,-3.38849422728519e-006,-4.80599476247265e-006,-5.36781079627058e-006,-4.51601746031396e-006,-2.25663332851228e-006,-2.36389493013898e-008,-1.41704872641982e-006,-1.17257783505037e-005,-4.50161241457132e-005,-.000111482237201632,-.000203124993756515,-.000288097777549774,-.000321776122768307,-.00027071494206603,-.000135275021888423,-1.95607303248633e-007,-1.17257783505037e-005,-9.70283345672377e-005,-.000372498901477441,-.000922491922170731,-.00168081634020727,-.00238394814506931,-.00266262932510325,-.00224010885981014,-.00111937218068057,-7.50950801186721e-007,-4.50161241457132e-005,-.000372498901477441,-.00143005063645349,-.00354151423048786,-.00645277735730557,-.0091521519891917,-.0102220295037144,-.00859994241054315,-.0042973520003992,-1.85972641877534e-006,-.000111482237201632,-.000922491922170731,-.00354151423048786,-.00877054471011803,-.0159802752815396,-.0226652649095649,-.0253148119577435,-.0212977202708341,-.0106423736858525,-3.38849422728519e-006,-.000203124993756515,-.00168081634020727,-.00645277735730557,-.0159802752815396,-.0291166861938667,-.0412969985964523,-.0461245768827176,-.0388052788106219,-.0193908208406645,-4.80599476247265e-006,-.000288097777549774,-.00238394814506931,-.0091521519891917,-.0226652649095649,-.0412969985964523,-.0585726714132265,-.0654197587632339,-.0550385965596167,-.0275025356838052,-5.36781079627058e-006,-.000321776122768307,-.00266262932510325,-.0102220295037144,-.0253148119577435,-.0461245768827176,-.0654197587632339,-.0730672638515391,-.0614725540550301,-.0307175549006202,-4.51601746031396e-006,-.00027071494206603,-.00224010885981014,-.00859994241054315,-.0212977202708341,-.0388052788106219,-.0550385965596167,-.0614725540550301,-.0517177557069423,-.025843126655234,-2.25663332851228e-006,-.000135275021888423,-.00111937218068057,-.0042973520003992,-.0106423736858525,-.0193908208406645,-.0275025356838052,-.0307175549006202,-.025843126655234,-.0129136925256951},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166929,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-8.1860036294541e-008,-2.53761981228001e-005,-.000498889241964144,-.00338501114114197,-.0125924797742609,-.0309704700072926,-.0548047012562974,-.0717170372141512,-.0670066214870632,-.0354369306242455,-7.45867227398079e-008,-2.31215076275497e-005,-.000454562632178236,-.00308425086139809,-.0114736303579195,-.0282187250839213,-.0499352705236225,-.0653449352400189,-.0610530427888428,-.0322883379835085,-6.06867939129239e-008,-1.8812599840914e-005,-.000369850661970349,-.00250947205515878,-.00933541273013334,-.0229599034620059,-.0406293688733935,-.0531672993864886,-.0496752409729615,-.026271106183758,-4.13979342045415e-008,-1.28331506776909e-005,-.000252296296814157,-.00171185446337238,-.00636821913065334,-.0156622637575245,-.0277156170386288,-.0362684567748484,-.0338863239396712,-.0179210245780364,-1.84679237326198e-008,-5.7249631538091e-006,-.000112551238536804,-.000763670899970602,-.00284090951583315,-.00698705136166928,-.0123641411463411,-.0161796260245334,-.0151169390000152,-.00799470122065349},
			{-4.17330922769922e-010,-1.2937047988686e-007,-2.54338890053445e-006,-1.72571365352968e-005,-6.4197762938255e-005,-.000157890655951536,-.000279400029400497,-.000365620865488212,-.000341606679433586,-.000180661133649367,-4.83770266076376e-009,-1.49966341007993e-006,-2.9482980004949e-005,-.000200044834396311,-.000744180868554274,-.00183026946897867,-.00323880688417962,-.00423827935409955,-.00395990675951328,-.0020942249885326,-1.68490485715029e-008,-5.22312002394816e-006,-.000102685137341121,-.00069672846133325,-.00259187893087141,-.00637457515358989,-.01128031595407,-.0147613401865313,-.0137918069811939,-.00729389568268136,-3.65992560246704e-008,-1.13455846597402e-005,-.000223051148290501,-.00151342333828502,-.00563004137434513,-.0138467585932022,-.0245029367616903,-.0320643665107509,-.0299583607113151,-.015843693154235,-6.03387416718265e-008,-1.87047054027123e-005,-.000367729486283348,-.00249507967559062,-.0092818720648124,-.0228282233164573,-.0403963504194766,-.0528623731165144,-.0493903424335445,-.026120435555211,-8.14470002822614e-008,-2.52481590434899e-005,-.000496372028041484,-.00336793160433076,-.0125289426948006,-.0308142042638448,-.0545281766383553,-.071355179091381,-.0666685304112694,-.0352581287427047,-9.25887373147944e-008,-2.87020412938273e-005,-.000564274425767331,-.0038286558562856,-.0142428693503848,-.0350295069709389,-.0619874888642517,-.0811163813282989,-.0757886113425933,-.0400813487182593,-8.82635618211773e-008,-2.73612587189578e-005,-.000537915000325487,-.0036498046378399,-.0135775302253503,-.0333931441745755,-.0590918152054245,-.0773271236407335,-.0722482342515896,-.0382089949930116,-6.68656227010509e-008,-2.07280055821258e-005,-.000407507024584746,-.00276497407096247,-.0102858982181088,-.0252975671172521,-.0447660499839257,-.0585805304843048,-.054732927978495,-.0289459001004882,-3.1570051288489e-008,-9.78655657274711e-006,-.000192401074675968,-.00130545966231248,-.00485640185758369,-.0119440372960602,-.0211359206253737,-.0276583134530104,-.0258416997202825,-.0136665675701473},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.05454958484093e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493428,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394505358986e-009,-2.17046942363539e-006,-.000101379919197701,-.00121573088108692,-.00679363471892785,-.0225536881419213,-.0497944643424747,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364621967283e-009,-1.97762190739121e-006,-9.23722522840775e-005,-.0011077124596865,-.0061900161802151,-.0205497793593546,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374478385261e-009,-1.60907368929725e-006,-7.5157824767176e-005,-.000901279950190045,-.0050364491486886,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711744365734e-010,-1.09764122348207e-006,-5.12694522820295e-005,-.000614814618998913,-.00343565011493427,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.05454958484094e-010,-4.89665844215546e-007,-2.28716807432831e-005,-.000274273335410201,-.00153266885205137,-.0050881945739141,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{-3.55137024936255e-005,-7.96079700390994e-005,-.000116700327309063,-.000143429804005386,-.000157416340741478,-.000157416340741478,-.000143429804005386,-.000116700327309063,-7.96079700390994e-005,-3.55137024936255e-005,-.000411675060900537,-.000922816085422149,-.00135279092234031,-.00166263918298768,-.0018247712041041,-.0018247712041041,-.00166263918298768,-.00135279092234031,-.000922816085422149,-.000411675060900537,-.0014338072393425,-.0032140406585855,-.00471158348413285,-.00579074207645094,-.00635542545828473,-.00635542545828473,-.00579074207645094,-.00471158348413286,-.0032140406585855,-.0014338072393425,-.00311449504225872,-.00698149194822924,-.0102344324954371,-.0125785649515731,-.0138051619061078,-.0138051619061078,-.0125785649515731,-.0102344324954371,-.00698149194822924,-.00311449504225872,-.00513465934024338,-.0115099180940778,-.0168728232640484,-.0207374374141283,-.022759645644952,-.022759645644952,-.0207374374141283,-.0168728232640484,-.0115099180940778,-.00513465934024338,-.00693091352498966,-.0155364244643318,-.0227754308935343,-.0279920002327546,-.03072163611483,-.03072163611483,-.0279920002327546,-.0227754308935343,-.0155364244643318,-.00693091352498966,-.00787904440302126,-.0176617667753745,-.0258910503876712,-.0318212327953672,-.0349242757407752,-.0349242757407752,-.0318212327953672,-.0258910503876712,-.0176617667753745,-.00787904440302126,-.00751098398062662,-.0168367178218381,-.0246815799932319,-.0303347408066573,-.033292828699298,-.033292828699298,-.0303347408066573,-.0246815799932318,-.0168367178218381,-.00751098398062663,-.00569007878902205,-.0127549534391095,-.0186979675580832,-.0229806195406161,-.0252215713543055,-.0252215713543055,-.0229806195406161,-.0186979675580832,-.0127549534391096,-.00569007878902205,-.00268652368659009,-.00602214587988349,-.00882809089266007,-.0108501096412746,-.0119081565244916,-.0119081565244916,-.0108501096412746,-.00882809089266007,-.00602214587988349,-.00268652368659009},
			{-5.98166645903067e-007,-1.34085800921453e-006,-1.96561435335508e-006,-2.41582597026698e-006,-2.6514048927612e-006,-2.6514048927612e-006,-2.41582597026698e-006,-1.96561435335508e-006,-1.34085800921453e-006,-5.98166645903067e-007,-3.58574009765282e-005,-8.03784089572694e-005,-.000117829742791934,-.000144817905008049,-.000158939802214037,-.000158939802214037,-.000144817905008049,-.000117829742791934,-8.03784089572694e-005,-3.58574009765282e-005,-.000296712405322989,-.000665114325306448,-.000975016187739576,-.00119833751913312,-.00131519323018857,-.00131519323018857,-.00119833751913312,-.000975016187739576,-.000665114325306448,-.000296712405322989,-.00113910071249293,-.00255342273613771,-.00374315874301679,-.00460050573337381,-.00504912338917163,-.00504912338917163,-.00460050573337381,-.00374315874301679,-.00255342273613771,-.00113910071249293,-.00282097799925267,-.00632354038798896,-.00926991647529692,-.0113931325974514,-.012504132286243,-.012504132286243,-.0113931325974514,-.00926991647529692,-.00632354038798896,-.00282097799925267,-.00513993217994983,-.0115217377590493,-.0168901501569485,-.0207587329229754,-.0227830177822856,-.0227830177822856,-.0207587329229754,-.0168901501569485,-.0115217377590493,-.00513993217994983,-.00729010748709311,-.0163415982470003,-.0239557655250033,-.0294426830950572,-.0323137821080817,-.0323137821080817,-.0294426830950572,-.0239557655250033,-.0163415982470003,-.00729010748709311,-.00814231384119493,-.0182519149174924,-.026756170818605,-.0328845036251162,-.0360912312727646,-.0360912312727646,-.0328845036251162,-.026756170818605,-.0182519149174924,-.00814231384119493,-.00685024731120176,-.0153556020470818,-.0225103564887036,-.0276662122010637,-.0303640789100232,-.0303640789100232,-.0276662122010637,-.0225103564887036,-.0153556020470818,-.00685024731120176,-.00342303733917246,-.00767312430993282,-.0112483224733966,-.0138246800583169,-.0151727917485003,-.0151727917485003,-.0138246800583169,-.0112483224733966,-.00767312430993282,-.00342303733917246},
			{-9.23396186630991e-009,-2.06989671022707e-008,-3.0343396956462e-008,-3.7293361369904e-008,-4.09300181472705e-008,-4.09300181472705e-008,-3.7293361369904e-008,-3.0343396956462e-008,-2.06989671022707e-008,-9.23396186630991e-009,-2.86248157690455e-006,-6.41657533884544e-006,-9.406299920457e-006,-1.15607538137749e-005,-1.26880990614001e-005,-1.26880990614001e-005,-1.15607538137749e-005,-9.406299920457e-006,-6.41657533884543e-006,-2.86248157690455e-006,-5.6275619268402e-005,-.000126148148407079,-.000184925330985175,-.000227281316089118,-.000249444620982072,-.000249444620982072,-.000227281316089118,-.000184925330985175,-.000126148148407079,-5.6275619268402e-005,-.000381835449985301,-.000855927231686191,-.00125473602757939,-.00154212543069905,-.00169250557057098,-.00169250557057098,-.00154212543069904,-.00125473602757939,-.000855927231686191,-.000381835449985301,-.00142045475791657,-.00318410956532667,-.00466770636506667,-.00573681517895975,-.00629623988713046,-.00629623988713046,-.00573681517895975,-.00466770636506667,-.00318410956532667,-.00142045475791657,-.00349352568083464,-.00783113187876223,-.0114799517310029,-.0141093625419607,-.0154852350036463,-.0154852350036463,-.0141093625419607,-.0114799517310029,-.00783113187876223,-.00349352568083464,-.00618207057317057,-.0138578085193144,-.0203146844366967,-.0249676352618112,-.0274023506281487,-.0274023506281487,-.0249676352618112,-.0203146844366967,-.0138578085193144,-.00618207057317057,-.00808981301226668,-.0181342283874242,-.0265836496932443,-.0326724676200095,-.0358585186070756,-.0358585186070756,-.0326724676200095,-.0265836496932443,-.0181342283874242,-.00808981301226668,-.0075584695000076,-.0169431619698356,-.0248376204864808,-.0305265213944214,-.0335033107435316,-.0335033107435316,-.0305265213944214,-.0248376204864808,-.0169431619698356,-.0075584695000076,-.00399735061032675,-.00896051228901818,-.013135553091879,-.0161441689917542,-.0177184653121228,-.0177184653121228,-.0161441689917542,-.013135553091879,-.00896051228901818,-.00399735061032675},
			{-1.36603610218197e-010,-3.06212401014368e-010,-4.48888313656529e-010,-5.51703361358645e-010,-6.05502635397803e-010,-6.05502635397803e-010,-5.51703361358645e-010,-4.48888313656529e-010,-3.06212401014368e-010,-1.36603610218197e-010,-2.18985222739045e-007,-4.90880078019026e-007,-7.19599629863477e-007,-8.84419403557672e-007,-9.7066343466231e-007,-9.7066343466231e-007,-8.84419403557672e-007,-7.19599629863477e-007,-4.90880078019026e-007,-2.18985222739045e-007,-1.02285265781769e-005,-2.29283960895319e-005,-3.36116010370069e-005,-4.13101270596927e-005,-4.53384781663522e-005,-4.53384781663522e-005,-4.13101270596927e-005,-3.36116010370069e-005,-2.29283960895319e-005,-1.02285265781769e-005,-.000122658764452733,-.000274953456270543,-.000403064646991639,-.000495384071772189,-.000543691378376726,-.000543691378376726,-.000495384071772189,-.000403064646991639,-.000274953456270543,-.000122658764452733,-.000685430347892353,-.00153646944045606,-.00225236853186345,-.00276825939157399,-.00303820580852531,-.00303820580852531,-.00276825939157399,-.00225236853186345,-.00153646944045606,-.000685430347892353,-.00227550978952434,-.00510081186777544,-.00747747259751734,-.00919014071209263,-.0100863159636095,-.0100863159636095,-.00919014071209263,-.00747747259751734,-.00510081186777544,-.00227550978952434,-.00502391406507093,-.0112616700678552,-.0165089071147008,-.0202901685574002,-.0222687614299034,-.0222687614299034,-.0202901685574002,-.0165089071147008,-.0112616700678552,-.00502391406507093,-.00770260000476695,-.0172662467539879,-.0253112426632633,-.0311086237548425,-.034142176732019,-.034142176732019,-.0311086237548425,-.0253112426632633,-.0172662467539879,-.00770260000476695,-.00799226169283676,-.0179155561012596,-.0262630897372964,-.0322784854203554,-.0354261172898117,-.0354261172898117,-.0322784854203554,-.0262630897372964,-.0179155561012596,-.00799226169283676,-.00447343431198914,-.0100277076079193,-.0146999949807664,-.0180669364652907,-.0198287311796651,-.0198287311796651,-.0180669364652907,-.0146999949807664,-.0100277076079193,-.00447343431198914},
			{-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036796,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0168406764624987,-.0356686882793406,-.0470831698838632,-.0493903844165837,-.0434469544588521,-.0321869995510595,-.0195234472015054,-.0089879288791991,-.00258061618548581,-.00022262032405145,-.0153443722103867,-.032499503831288,-.0428998018667281,-.0450020189978578,-.0395866663734618,-.0293271652446273,-.0177887771525588,-.00818934598203496,-.00235132688229124,-.000202840374105465,-.0124848058722647,-.0264429193136947,-.0349050251728234,-.0366154746082026,-.0322093232636935,-.0238617754993688,-.0144736732405605,-.00666318526458248,-.00191313526973573,-.000165039185640327,-.00851660037928648,-.0180382281439193,-.0238107146933112,-.0249775101132113,-.0219718221917673,-.0162774822730591,-.00987332060036795,-.0045453398901192,-.00130505902378937,-.000112582670920343,-.00379931823382166,-.00804698659569713,-.0106221354138729,-.0111426514550797,-.00980179190824761,-.00726150487830256,-.0044045611294125,-.00202770964403212,-.000582196454509688,-5.02239597246239e-005},
			{-8.58555085657902e-005,-.000181842658097301,-.000240034864648708,-.000251797282715359,-.000221497062722267,-.000164092649236315,-9.95325509741368e-005,-4.58213900489777e-005,-1.31562479400023e-005,-1.1349414128987e-006,-.000995237591006242,-.0021079212273054,-.00278248564875054,-.00291883567237296,-.00256759533314912,-.00190216301383433,-.00115378195194428,-.000531161839358935,-.000152507308211391,-1.31562479400023e-005,-.00346627473553786,-.00734159778603745,-.00969101227025514,-.0101659005244202,-.00894257905329049,-.00662497041642304,-.00401846279369377,-.00184996314537334,-.000531161839358935,-4.58213900489777e-005,-.00752939110831236,-.0159473109630529,-.0210506746248672,-.0220822199209436,-.0194249389752361,-.0143906635082426,-.00872884590413951,-.00401846279369377,-.00115378195194428,-9.95325509741368e-005,-.0124132027362624,-.0262912632953882,-.0347048370969986,-.0364054767247764,-.0320245956373459,-.0237249229144092,-.0143906635082426,-.00662497041642304,-.00190216301383433,-.000164092649236315,-.0167557045233543,-.0354887170283886,-.0468456053029506,-.0491411784688114,-.0432277369007862,-.0320245956373459,-.0194249389752361,-.00894257905329049,-.00256759533314912,-.000221497062722267,-.0190478411637101,-.0403434808795061,-.053253961824449,-.0558635634071245,-.0491411784688113,-.0364054767247764,-.0220822199209436,-.0101659005244202,-.00291883567237296,-.000251797282715359,-.0181580433524764,-.038458881954326,-.0507662647534973,-.053253961824449,-.0468456053029506,-.0347048370969986,-.0210506746248672,-.00969101227025514,-.00278248564875054,-.000240034864648708,-.0137559469700066,-.0291352063887049,-.038458881954326,-.0403434808795061,-.0354887170283886,-.0262912632953882,-.0159473109630529,-.00734159778603745,-.0021079212273054,-.000181842658097301,-.0064947567048982,-.0137559469700066,-.0181580433524764,-.0190478411637101,-.0167557045233543,-.0124132027362624,-.00752939110831236,-.00346627473553786,-.000995237591006242,-8.58555085657902e-005},
			{-1.44608694643564e-006,-3.06282379048705e-006,-4.04297045415503e-006,-4.24108796005352e-006,-3.73073337316569e-006,-2.763855715616e-006,-1.67645297446285e-006,-7.71781742654136e-007,-2.21594149611604e-007,-1.91161171784847e-008,-8.66864106188711e-005,-.000183602515332728,-.00024235790093583,-.000254234154648952,-.000223640691794434,-.000165680723448716,-.000100495818239503,-4.62648454277032e-005,-1.32835729492742e-005,-1.14592527597117e-006,-.000717311704224692,-.00151927196238748,-.00200546034506998,-.00210373383142091,-.00185057940015954,-.00137097292696425,-.000831581630087009,-.000382831805844932,-.000109918755228315,-9.48228917027153e-006,-.00275381230681073,-.00583259662818021,-.00769910953153036,-.00807638866214051,-.00710451021065933,-.0052632657411594,-.00319250294337036,-.00146971941509561,-.000421986172976531,-3.64032044368432e-005,-.00681980429507682,-.0144444003819494,-.0190668115330659,-.020001141672057,-.0175942888806334,-.0130344548969134,-.00790621976356029,-.00363975378962373,-.00104504693649993,-9.01523787074882e-005,-.0124259499955376,-.0263182620931215,-.0347404758894987,-.0364428619234804,-.0320574820593634,-.0237492863079749,-.0144054414445077,-.00663177367400152,-.00190411636669871,-.000164261157846817,-.0176240673855734,-.0373279165590501,-.0492733745350088,-.051687915571302,-.0454680104321141,-.0336842673921381,-.0204316346700341,-.00940602739898747,-.00270066072764841,-.00023297612784248,-.0196843034297875,-.0416915130812311,-.0550333832785173,-.057730181773567,-.0507831758761377,-.0376219249308874,-.0228200725526466,-.01050558269779,-.00301636529530396,-.000260210806734789,-.016560691379957,-.0350756776212305,-.0463003874798657,-.0485692433603775,-.0427246260442284,-.0316518737949094,-.01919885964777,-.00883849984558959,-.00253771208734517,-.000218919144354553,-.00827530195346473,-.0175271561360877,-.023136092459402,-.0242698293952193,-.0213493007781521,-.0158162969791731,-.00959358260487664,-.00441655806269213,-.00126808315618723,-.000109392897999453},
			{-2.23233973501177e-008,-4.72811352436055e-008,-6.24117630999618e-008,-6.54701240215422e-008,-5.75917262110692e-008,-4.26659334075081e-008,-2.58795821233052e-008,-1.19140765023158e-008,-3.42077235703851e-009,-2.95097525510867e-010,-6.92014051755746e-006,-1.46569133086574e-005,-1.93473315833786e-005,-2.02954080342344e-005,-1.78531444734238e-005,-1.32262240313143e-005,-8.02253985001058e-006,-3.69330358815301e-006,-1.06042216684182e-006,-9.14787436199951e-008,-.000136048104620829,-.000288150980485938,-.000380363344459719,-.000399002272939181,-.000350988027044924,-.000260024013412071,-.000157720979519105,-7.26093569462482e-005,-2.08476150925998e-005,-1.79844753889302e-006,-.000923099379853643,-.00195513191552442,-.00258080160960999,-.0027072685190056,-.0023814872761676,-.0017642877583389,-.00107015190538498,-.000492661419690163,-.000141453058953237,-1.2202638268856e-005,-.00343399468591358,-.00727322838114722,-.00960076369480724,-.010071229501942,-.00885930034123408,-.00656327467961113,-.00398104043444918,-.00183273516816672,-.000526215338622209,-4.53946735139408e-005,-.00844570976740211,-.0178880812574289,-.023612518692696,-.0247696019225699,-.0217889327934133,-.0161419915106792,-.00979113689941693,-.0045075053186138,-.00129419595300744,-.000111645553517426,-.0149453528019069,-.031654377512867,-.0417842232474103,-.0438317737278189,-.0385572434695171,-.0285645333188729,-.0173261927444337,-.00797638790562856,-.00229018231092328,-.00019756565546715,-.0195573810002261,-.0414226903539633,-.0546785334999813,-.0573579433067165,-.0504557310119577,-.0373793425030134,-.0226729309958482,-.0104378437460552,-.00299691607206422,-.000258532993348771,-.0182728411109243,-.0387020245305973,-.0510872163717606,-.0535906410209485,-.0471417699489713,-.0349242460624195,-.0211837600137409,-.00975228024192592,-.00280007692272149,-.000241552399543637,-.00966372260509664,-.0204678422500889,-.027017839463842,-.0283417934798407,-.0249312619277241,-.0184699371099763,-.0112031828692115,-.00515756309886556,-.00148084069082919,-.000127746712709587},
			{-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910325,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106448,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252843,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0262800462022158,-.052592127399495,-.062511846274886,-.0559691123983598,-.0394613443503929,-.0216577924252842,-.0087453382441128,-.00227798149655705,-.000275291812791722,-4.59236798590033e-006,-.0239450482604383,-.0479192851873238,-.0569576310628798,-.0509962230340228,-.0359551801249216,-.0197334845170196,-.00796830967106449,-.00207558146784555,-.000250831969330614,-4.18433332293013e-006,-.0194826660246954,-.0389890811366307,-.0463430472758461,-.0414926030253305,-.029254598220584,-.0160559663178376,-.0064833411239848,-.00168877757544105,-.000204087101167344,-3.4045439282468e-006,-.0132902411581778,-.0265966829275197,-.0316132439739062,-.0283044784419232,-.0199562351910326,-.0109526932361906,-.00442265791238582,-.00115201180432314,-.00013921948815164,-2.32243419779698e-006,-.00592887458791186,-.0118649763879334,-.0141029012647388,-.0126268365608368,-.00890262368352143,-.00488607722195371,-.0019729803090621,-.000513920961277905,-6.21068403214379e-005,-1.03605502214257e-006},
			{-.000133978390791535,-.000268120099297756,-.000318691851029199,-.000285336317731836,-.000201178010641796,-.000110413663465783,-4.45846379362571e-005,-1.16133850303456e-005,-1.40346610474412e-006,-2.3412366475277e-008,-.00155307834203885,-.0031080498640454,-.00369427792566825,-.00330762037555709,-.00233205675464775,-.00127991587584218,-.00051682540115133,-.000134622431735668,-1.62689878433847e-005,-2.71396298267277e-007,-.00540915683648703,-.0108249073566845,-.0128666585301024,-.0115199838170822,-.00812223079546563,-.00445776978697114,-.00180003131602355,-.000468871290814306,-5.66626321633216e-005,-9.45235731149608e-007,-.0117496910935979,-.0235137048901805,-.0279487668236465,-.0250235397762786,-.0176429905292981,-.00968310210751827,-.00391000160679661,-.00101847533659284,-.000123081737985171,-2.05322718260318e-006,-.0193709286096546,-.0387654701002611,-.0460772596109979,-.0412546337351168,-.0290868166049117,-.0159638817863846,-.00644615772325802,-.0016790920611167,-.000202916616332326,-3.38501811296807e-006,-.0261474466358464,-.0523267666298084,-.062196434207296,-.0556867125891282,-.0392622367419453,-.0215485150723445,-.00870121244425688,-.00226648762944902,-.000273902790309784,-4.56919656531553e-006,-.0297243490813528,-.0594849317130106,-.0707047440517022,-.0633045095088931,-.0446332082396283,-.024496295684836,-.00989151559332482,-.00257653722080653,-.000311371977038976,-5.19425072819777e-006,-.0283358105837029,-.0567061621094263,-.0674018539257125,-.0603473127579275,-.0425482196753798,-.0233519796389196,-.00942944491302682,-.00245617727240442,-.00029682659621248,-4.95160732891562e-006,-.0214662945877596,-.0429587562772,-.0510614668268089,-.0457171743654439,-.0322331565225002,-.0176907052881247,-.00714344280726179,-.00186072054418367,-.000224866239028424,-3.75117772937632e-006,-.0101351336267259,-.0202826218343075,-.0241082496725556,-.0215849861435475,-.0152186185292077,-.00835252033429224,-.00337271749954253,-.000878523830938623,-.000106168736824841,-1.77108757123392e-006},
			{-2.25663332851228e-006,-4.51601746031396e-006,-5.36781079627058e-006,-4.80599476247265e-006,-3.38849422728519e-006,-1.85972641877534e-006,-7.50950801186721e-007,-1.95607303248633e-007,-2.36389493013898e-008,-3.94340655797694e-010,-.000135275021888423,-.00027071494206603,-.000321776122768307,-.000288097777549774,-.000203124993756515,-.000111482237201632,-4.50161241457132e-005,-1.17257783505037e-005,-1.41704872641982e-006,-2.36389493013898e-008,-.00111937218068057,-.00224010885981014,-.00266262932510325,-.00238394814506931,-.00168081634020727,-.000922491922170731,-.000372498901477441,-9.70283345672377e-005,-1.17257783505037e-005,-1.95607303248633e-007,-.0042973520003992,-.00859994241054315,-.0102220295037144,-.0091521519891917,-.00645277735730557,-.00354151423048786,-.00143005063645349,-.000372498901477441,-4.50161241457133e-005,-7.50950801186721e-007,-.0106423736858525,-.0212977202708341,-.0253148119577435,-.0226652649095649,-.0159802752815396,-.00877054471011803,-.00354151423048786,-.000922491922170731,-.000111482237201632,-1.85972641877534e-006,-.0193908208406645,-.0388052788106218,-.0461245768827176,-.0412969985964523,-.0291166861938667,-.0159802752815396,-.00645277735730557,-.00168081634020727,-.000203124993756515,-3.38849422728519e-006,-.0275025356838052,-.0550385965596167,-.0654197587632339,-.0585726714132265,-.0412969985964523,-.0226652649095649,-.0091521519891917,-.00238394814506931,-.000288097777549774,-4.80599476247265e-006,-.0307175549006202,-.0614725540550301,-.0730672638515391,-.0654197587632339,-.0461245768827176,-.0253148119577435,-.0102220295037144,-.00266262932510325,-.000321776122768307,-5.36781079627058e-006,-.025843126655234,-.0517177557069423,-.0614725540550301,-.0550385965596167,-.0388052788106219,-.0212977202708341,-.00859994241054315,-.00224010885981014,-.00027071494206603,-4.51601746031396e-006,-.0129136925256951,-.025843126655234,-.0307175549006202,-.0275025356838052,-.0193908208406645,-.0106423736858525,-.0042973520003992,-.00111937218068057,-.000135275021888423,-2.25663332851228e-006},
			{-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166929,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0354369306242455,-.0670066214870632,-.0717170372141512,-.0548047012562974,-.0309704700072926,-.0125924797742609,-.00338501114114197,-.000498889241964144,-2.53761981228001e-005,-8.1860036294541e-008,-.0322883379835085,-.0610530427888428,-.0653449352400189,-.0499352705236225,-.0282187250839213,-.0114736303579195,-.00308425086139809,-.000454562632178237,-2.31215076275497e-005,-7.45867227398079e-008,-.026271106183758,-.0496752409729615,-.0531672993864886,-.0406293688733935,-.0229599034620059,-.00933541273013334,-.00250947205515878,-.000369850661970349,-1.8812599840914e-005,-6.06867939129239e-008,-.0179210245780364,-.0338863239396712,-.0362684567748484,-.0277156170386288,-.0156622637575245,-.00636821913065334,-.00171185446337238,-.000252296296814157,-1.28331506776909e-005,-4.13979342045415e-008,-.00799470122065349,-.0151169390000152,-.0161796260245334,-.0123641411463411,-.00698705136166928,-.00284090951583315,-.000763670899970602,-.000112551238536804,-5.7249631538091e-006,-1.84679237326198e-008},
			{-.000180661133649367,-.000341606679433586,-.000365620865488212,-.000279400029400497,-.000157890655951536,-6.4197762938255e-005,-1.72571365352968e-005,-2.54338890053445e-006,-1.2937047988686e-007,-4.17330922769922e-010,-.0020942249885326,-.00395990675951328,-.00423827935409955,-.00323880688417962,-.00183026946897867,-.000744180868554274,-.000200044834396311,-2.9482980004949e-005,-1.49966341007993e-006,-4.83770266076376e-009,-.00729389568268136,-.0137918069811939,-.0147613401865313,-.01128031595407,-.00637457515358989,-.00259187893087141,-.00069672846133325,-.000102685137341121,-5.22312002394816e-006,-1.68490485715029e-008,-.015843693154235,-.0299583607113152,-.0320643665107509,-.0245029367616903,-.0138467585932022,-.00563004137434513,-.00151342333828502,-.000223051148290501,-1.13455846597402e-005,-3.65992560246704e-008,-.026120435555211,-.0493903424335445,-.0528623731165144,-.0403963504194766,-.0228282233164573,-.0092818720648124,-.00249507967559062,-.000367729486283348,-1.87047054027123e-005,-6.03387416718265e-008,-.0352581287427047,-.0666685304112694,-.071355179091381,-.0545281766383553,-.0308142042638448,-.0125289426948006,-.00336793160433076,-.000496372028041484,-2.52481590434899e-005,-8.14470002822614e-008,-.0400813487182593,-.0757886113425933,-.0811163813282989,-.0619874888642517,-.0350295069709389,-.0142428693503848,-.0038286558562856,-.000564274425767331,-2.87020412938273e-005,-9.25887373147944e-008,-.0382089949930116,-.0722482342515896,-.0773271236407335,-.0590918152054245,-.0333931441745755,-.0135775302253503,-.0036498046378399,-.000537915000325487,-2.73612587189578e-005,-8.82635618211772e-008,-.0289459001004882,-.054732927978495,-.0585805304843048,-.0447660499839257,-.0252975671172521,-.0102858982181088,-.00276497407096247,-.000407507024584746,-2.07280055821258e-005,-6.68656227010509e-008,-.0136665675701473,-.0258416997202825,-.0276583134530104,-.0211359206253737,-.0119440372960602,-.00485640185758369,-.00130545966231248,-.000192401074675968,-9.78655657274711e-006,-3.1570051288489e-008},
			{-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484094e-010,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493428,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0443383908346372,-.0792152064555801,-.0763442280886821,-.0497944643424747,-.0225536881419213,-.00679363471892785,-.00121573088108692,-.000101379919197701,-2.17046942363539e-006,-1.35394505358986e-009,-.0403988980900667,-.0721768876258489,-.0695609974169401,-.0453701961788293,-.0205497793593546,-.0061900161802151,-.0011077124596865,-9.23722522840775e-005,-1.97762190739121e-006,-1.23364621967283e-009,-.0328701880528208,-.058726053964138,-.0565976592019673,-.0369150385504742,-.0167201370314615,-.0050364491486886,-.000901279950190045,-7.5157824767176e-005,-1.60907368929725e-006,-1.00374478385261e-009,-.0224226358745209,-.0400604013055631,-.0386085014663319,-.0251818598172014,-.0114057620791853,-.00343565011493427,-.000614814618998913,-5.12694522820295e-005,-1.09764122348207e-006,-6.84711744365734e-010,-.010002903216594,-.0178712404429137,-.0172235372177759,-.0112338133649814,-.0050881945739141,-.00153266885205137,-.000274273335410201,-2.28716807432831e-005,-4.89665844215546e-007,-3.05454958484093e-010},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.,3.354101966},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,2.645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370350,0.,-7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,-4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,7.245688375,2.049390154,0.,13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-3.354101966,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370350,0.,7.245688375,-5.477225576},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,4.183300132,0.,0.,-3.240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246899,0.,0.,.4082482906,0.,0.,0.,0.,-5.916079782,1.673320053,0.,0.,4.242640686},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.160246900,0.,0.,.4082482906,0.,0.,0.,0.,0.,1.673320053,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,-.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,-.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,7.245688375,2.049390154,0.,-13.74772709,5.196152424},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,2.645751311,0.,0.,.5000000000,0.,0.,-5.291502622,0.,0.,2.049390152,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,2.598076212,-1.963961012},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,3.354101966,0.,0.,0.,1.936491673,0.,-10.60660172,-.5000000000,0.,10.06230590,-2.535462764},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,-13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,-5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,.5000000000,0.,3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,-2.645751311,.5000000000},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370350,0.,-5.477225576,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,-7.245688375,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,-3.240370349,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-4.183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.581138830,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246899,0.,-5.916079782,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,1.673320053,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,-2.160246900,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,-.3872983346,0.,-2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,-.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390154,0.,5.196152424,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,13.74772709,.5000000000,0.,0.,2.645751311,0.,7.245688375,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,2.049390152,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,5.291502622,0.,0.,.5000000000,0.,0.,2.645751311,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,-1.963961012,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.598076212,0.,0.,0.,.5000000000,0.,2.738612788,.3872983346,0.,2.598076212,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3872983346,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,.5000000000,0.,0.,.3872983346,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,0.,0.,0.,.3585685828},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,-2.535462764,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,-10.06230590,0.,0.,0.,-1.936491673,0.,-10.60660172,-.5000000000,0.,-3.354101966,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-1.936491673,0.,0.,-.5000000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3779644731,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.645751311,0.,0.,0.,0.,0.,4.183300132,0.,0.,2.645751311,.5000000000}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,-1.811422094,.7905694150,0.,0.,1.045825033,0.,0.,.8100925875,0.,1.811422094,.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,-1.045825033,0.,0.,.7905694150,0.,0.,1.045825033,0.,0.,.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,.8385254914,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,-.8385254914,.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,.5303300858,0.,0.,0.,0.,0.,0.,0.,.5400617248,0.,1.403121520,-.2041241452,0.,0.,-.5400617248,0.,0.,-.4183300132,0.,-1.403121520,-.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5400617250,0.,0.,-.2041241452,0.,0.,-.5400617250,0.,0.,-.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.6495190530,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,.6495190530,.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[35][35] = 
		{
			{1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,.4192627457,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,2.218529919,.9682458365,0.,0.,2.561737692,0.,3.507803799,.6614378278,0.,2.218529919,.4192627457},
			{0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,.6614378278,0.,0.,1.224744872,0.,0.,0.,2.371708244,0.,0.,2.561737692,0.,0.,.9682458365,0.,0.,2.561737692,0.,0.,.6614378278,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,2.371708245,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925875,0.,-.6846531970,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,-3.622844188,-.7905694150,0.,0.,-3.137475099,0.,-5.728219620,-.8100925875,0.,-3.622844188,-.6846531970},
			{0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694150,0.,0.,-.8100925872,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,-3.137475099,0.,0.,-.7905694150,0.,0.,-3.137475099,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,3.436931772,.2500000000,0.,0.,1.984313483,0.,5.434266281,.5123475385,0.,3.436931772,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,1.984313483,0.,0.,.2500000000,0.,0.,1.984313483,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,-.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-1.677050983,0.,0.,0.,-.4841229182,0.,-2.651650429,-.1250000000,0.,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,-.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,.3307189139,.6250000000e-1},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925875,0.,.6846531970,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,1.811422094,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925875,0.,-1.811422094,-.6846531970},
			{0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694150,0.,0.,.8100925872,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,1.045825033,0.,0.,-.7905694150,0.,0.,-1.045825033,0.,0.,-.8100925872,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,.7905694148,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7905694148,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,-.5303300858,0.,0.,0.,0.,0.,0.,0.,-.5400617248,0.,-1.403121520,.2041241452,0.,0.,.5400617248,0.,0.,.4183300132,0.,1.403121520,.5303300858},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,-.4183300132,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5400617250,0.,0.,.2041241452,0.,0.,.5400617250,0.,0.,.4183300132,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2041241452,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.6495190530,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475385,0.,.6495190530,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,-1.811422094,.5123475385,0.,0.,.6495190530},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,.5123475380,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,-.6614378278,0.,0.,.2500000000,0.,0.,-.6614378278,0.,0.,.5123475380,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,-.2454951265,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,.6846531969,-.9682458365e-1,0.,0.,-.2454951265},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,.1250000000,0.,0.,-.9682458365e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,.4482107285e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,.3169328455,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,-.8385254914,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,.8385254914,-.3169328455},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4841229182,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.1250000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,.5229125166,0.,0.,-.3307189139,.6250000000e-1}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[35][35] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2672612419,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2357022604},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1290994449,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1091089451,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4303314828e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3333333333e-1,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8164965809e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7715167495e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[35][35] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,3.162277660,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,3.741657387,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.165151390,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,23.23790008,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,30.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.24744872,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,12.96148140,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,1.500000000,1.290994449,1.145643924,1.039230485,1.414213562,1.224744872,1.054092553,.9354143465,1.118033988,.9682458365,.8333333333,.9165151392,.7937253933,.7745966692,1.414213562,1.224744872,1.054092553,.9354143465,1.369306394,1.185854122,1.020620726,1.183215957,1.024695077,1.024695077,1.118033988,.9682458365,.8333333333,1.183215957,1.024695077,1.080123450,.9165151392,.7937253933,1.024695077,.7745966692},
			{2.236067977,2.581988897,2.500000000,2.366431913,2.236067977,1.825741858,2.108185107,2.041241452,1.932183566,1.443375673,1.666666667,1.613743061,1.183215956,1.366260101,1.,1.825741858,2.108185107,2.041241452,1.932183566,1.767766952,2.041241452,1.976423538,1.527525232,1.763834207,1.322875656,1.443375673,1.666666667,1.613743061,1.527525232,1.763834207,1.394433378,1.183215956,1.366260101,1.322875656,1.},
			{2.645751311,3.436931772,3.549647869,3.500000000,3.401680258,2.160246900,2.806243040,2.898275350,2.857738034,1.707825129,2.218529918,2.291287848,1.400000000,1.818653348,1.183215956,2.160246900,2.806243040,2.898275350,2.857738034,2.091650066,2.717133141,2.806243040,1.807392228,2.347871376,1.565247584,1.707825129,2.218529918,2.291287848,1.807392228,2.347871376,1.649915823,1.400000000,1.818653348,1.565247584,1.183215956},
			{3.,4.156921939,4.472135954,4.535573675,4.500000000,2.449489743,3.394112549,3.651483716,3.703280398,1.936491673,2.683281572,2.886751347,1.587450787,2.199636334,1.341640786,2.449489743,3.394112549,3.651483716,3.703280398,2.371708244,3.286335344,3.535533905,2.049390154,2.839718296,1.774823935,1.936491673,2.683281572,2.886751347,2.049390154,2.839718296,1.870828693,1.587450787,2.199636334,1.774823935,1.341640786},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,1.732050808,1.500000000,1.290994449,1.145643924,1.643167672,1.423024947,1.224744872,1.496662955,1.296148140,1.355261854,.8660254040,.7500000000,.6454972245,.5728219619,1.341640786,1.161895004,1.,1.449137675,1.254990040,1.434274331,.5477225575,.4743416490,.4082482906,.9660917834,.8366600265,1.133893419,.3741657387,.3240370349,.7171371655,.2710523708},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,2.236067977,2.581988897,2.500000000,2.366431913,2.121320343,2.449489743,2.371708244,1.932183566,2.231093404,1.749635531,1.118033988,1.290994449,1.250000000,1.183215956,1.732050808,2.,1.936491673,1.870828694,2.160246900,1.851640200,.7071067810,.8164965809,.7905694148,1.247219129,1.440164600,1.463850110,.4830458916,.5577733510,.9258201003,.3499271062},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,2.645751311,3.436931771,3.549647869,3.500000000,2.509980080,3.260559769,3.367491647,2.286190427,2.969848480,2.070196679,1.322875656,1.718465886,1.774823935,1.750000000,2.049390152,2.662235902,2.749545417,2.213594362,2.875543427,2.190890230,.8366600265,1.086853256,1.122497216,1.475729575,1.917028952,1.732050808,.5715476066,.7424621200,1.095445115,.4140393357},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,1.643167673,1.423024947,1.224744872,1.086853256,1.732050808,1.500000000,1.290994449,1.690308509,1.463850109,1.607142857,.5477225576,.4743416490,.4082482906,.3622844188,1.060660172,.9185586536,.7905694150,1.309307341,1.133893419,1.417366774,.2886751347,.2500000000,.2151657415,.6546536709,.5669467095,.8964214566,.1690308509,.1463850109,.4252100321,.1071428571},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,2.121320343,2.449489743,2.371708244,2.244994432,2.236067977,2.581988897,2.500000000,2.182178903,2.519763153,2.074812507,.7071067810,.8164965809,.7905694148,.7483314772,1.369306394,1.581138830,1.530931089,1.690308509,1.951800147,1.829812638,.3726779962,.4303314828,.4166666667,.8451542550,.9759000734,1.157275125,.2182178903,.2519763153,.5489437912,.1383208338},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,1.496662955,1.296148140,1.115546702,.9899494934,1.690308509,1.463850109,1.259881577,1.732050808,1.500000000,1.707825127,.3741657387,.3240370349,.2788866754,.2474873734,.8280786708,.7171371654,.6172133998,1.118033988,.9682458365,1.290994449,.1690308509,.1463850109,.1259881577,.4472135954,.3872983346,.6804138176,.8660254040e-1,.7500000000e-1,.2581988897,.4879500362e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.224744872,1.054092553,.9354143468,.8485281372,.8660254040,.7500000000,.6454972245,.5728219619,.5477225575,.4743416490,.4082482906,.3741657387,.3240370349,.2710523708,1.732050808,1.500000000,1.290994449,1.145643924,1.341640786,1.161895004,1.,.9660917834,.8366600265,.7171371655,1.643167672,1.423024947,1.224744872,1.449137675,1.254990040,1.133893419,1.496662955,1.296148140,1.434274331,1.355261854},
			{1.825741857,2.108185106,2.041241452,1.932183566,1.825741857,1.118033988,1.290994449,1.250000000,1.183215956,.7071067810,.8164965809,.7905694148,.4830458916,.5577733510,.3499271062,2.236067977,2.581988897,2.500000000,2.366431913,1.732050808,2.,1.936491673,1.247219129,1.440164600,.9258201003,2.121320343,2.449489743,2.371708244,1.870828694,2.160246900,1.463850110,1.932183566,2.231093404,1.851640200,1.749635531},
			{2.160246899,2.806243040,2.898275350,2.857738034,2.777460299,1.322875656,1.718465886,1.774823935,1.750000000,.8366600265,1.086853256,1.122497216,.5715476066,.7424621200,.4140393357,2.645751311,3.436931771,3.549647869,3.500000000,2.049390152,2.662235902,2.749545417,1.475729575,1.917028952,1.095445115,2.509980080,3.260559769,3.367491647,2.213594362,2.875543427,1.732050808,2.286190427,2.969848480,2.190890230,2.070196679},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.369306393,1.185854122,1.020620726,.9057110469,.8215838360,1.341640786,1.161895004,1.,.8874119673,1.060660172,.9185586536,.7905694148,.8280786714,.7171371654,.6561133241,1.341640786,1.161895004,1.,.8874119673,1.732050808,1.500000000,1.290994449,1.603567452,1.388730150,1.388730150,1.060660172,.9185586536,.7905694148,1.603567452,1.388730150,1.646831374,.8280786714,.7171371654,1.388730150,.6561133241},
			{1.767766952,2.041241452,1.976423538,1.870828694,1.767766952,1.732050808,2.,1.936491673,1.833030278,1.369306394,1.581138830,1.530931089,1.069044968,1.234426800,.8470386591,1.732050808,2.,1.936491673,1.833030278,2.236067977,2.581988897,2.500000000,2.070196679,2.390457218,1.792842914,1.369306394,1.581138830,1.530931089,2.070196679,2.390457218,2.126050161,1.069044968,1.234426800,1.792842914,.8470386591},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,1.449137675,1.254990040,1.080123450,.9585144758,1.309307342,1.133893419,.9759000734,1.118033988,.9682458365,.9449111824,.9660917834,.8366600265,.7200822999,.6390096507,1.603567452,1.388730150,1.195228610,1.732050808,1.500000000,1.666666667,.6546536711,.5669467095,.4879500367,1.299038106,1.125000000,1.581138830,.4472135954,.3872983346,1.,.3149703940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.118033988,.9682458365,.8333333333,.7395099728,.6708203931,.5477225576,.4743416490,.4082482906,.3622844188,.2886751347,.2500000000,.2151657415,.1690308509,.1463850109,.1071428571,1.643167673,1.423024947,1.224744872,1.086853256,1.060660172,.9185586536,.7905694150,.6546536709,.5669467095,.4252100321,1.732050808,1.500000000,1.290994449,1.309307341,1.133893419,.8964214566,1.690308509,1.463850109,1.417366774,1.607142857},
			{1.443375673,1.666666667,1.613743061,1.527525232,1.443375673,.7071067810,.8164965809,.7905694148,.7483314772,.3726779962,.4303314828,.4166666667,.2182178903,.2519763153,.1383208338,2.121320343,2.449489743,2.371708244,2.244994432,1.369306394,1.581138830,1.530931089,.8451542550,.9759000734,.5489437912,2.236067977,2.581988897,2.500000000,1.690308509,1.951800147,1.157275125,2.182178903,2.519763153,1.829812638,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.183215957,1.024695077,.8819171036,.7826237920,.7099295740,.9660917834,.8366600265,.7200822999,.6390096507,.6546536711,.5669467095,.4879500367,.4472135954,.3872983346,.3149703940,1.449137675,1.254990040,1.080123450,.9585144758,1.603567452,1.388730150,1.195228610,1.299038106,1.125000000,1.,1.309307342,1.133893419,.9759000734,1.732050808,1.500000000,1.581138830,1.118033988,.9682458365,1.666666667,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{.9165151390,.7937253933,.6831300507,.6062177828,.5499090834,.3741657387,.3240370349,.2788866754,.2474873734,.1690308509,.1463850109,.1259881577,.8660254040e-1,.7500000000e-1,.4879500362e-1,1.496662955,1.296148140,1.115546702,.9899494934,.8280786708,.7171371654,.6172133998,.4472135954,.3872983346,.2581988897,1.690308509,1.463850109,1.259881577,1.118033988,.9682458365,.6804138176,1.732050808,1.500000000,1.290994449,1.707825127},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,-.2397923765,-.4153325792,-.5361920542,-.6344309944,-.4964996319,-.8599625886,-1.110206928,-.5734033834,-.9931637928,-.5309563362,.2624521439e-1,.4545804478e-1,.5868608345e-1,.7424497583e-1,.1285960703,.1087015236,-.2482151352e-2,-.4299212254e-2,-.8901689590e-2,.2178212518e-3},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.851946997,3.207666290,4.141079375,4.899791194,1.565440216,2.711421991,3.500430739,1.143425313,1.980470737,.7749487345,1.851946997,3.207666290,4.141079375,4.899791194,3.834529752,6.641600355,8.574269190,4.428467199,7.670330186,4.100643259,1.565440216,2.711421991,3.500430739,4.428467199,7.670330186,6.483686290,1.143425313,1.980470737,4.100643259,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.2397923765,-.4153325792,-.5361920542,-.6344309944,.2624521439e-1,.4545804478e-1,.5868608345e-1,-.2482151352e-2,-.4299212254e-2,.2178212518e-3,1.851946997,3.207666290,4.141079375,4.899791194,-.4964996319,-.8599625886,-1.110206928,.7424497583e-1,.1285960703,-.8901689590e-2,1.565440216,2.711421991,3.500430739,-.5734033834,-.9931637928,.1087015236,1.143425313,1.980470737,-.5309563362,.7749487345},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,.6999612018,1.212368365,1.565160828,1.851923267,1.081581075,1.873353374,2.418488807,.9321846605,1.614591194,.6441723235,.2236285858,.3873360728,.5000487198,.4721128436,.8177234318,.5158410843,.6173679058e-1,.1069312580,.1652303366,.1581446975e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.382070207,2.393815818,3.090402932,3.656614062,.8718456225,1.510080914,1.949506078,.4752394227,.8231388257,.2403694019,1.382070207,2.393815818,3.090402932,3.656614062,2.135576909,3.698927711,4.775295141,1.840594370,3.188002963,1.271915321,.8718456225,1.510080914,1.949506078,1.840594370,3.188002963,2.011074701,.4752394227,.8231388257,1.271915321,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.6999612018,1.212368365,1.565160828,1.851923267,.2236285858,.3873360728,.5000487198,.6173679058e-1,.1069312580,.1581446975e-1,1.382070207,2.393815818,3.090402932,3.656614062,1.081581075,1.873353374,2.418488807,.4721128436,.8177234318,.1652303366,.8718456225,1.510080914,1.949506078,.9321846605,1.614591194,.5158410843,.4752394227,.8231388257,.6441723235,.2403694019},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,2.055529306,3.560281195,4.596303257,5.438419356,1.618556373,2.803421873,3.619202076,.7108692826,1.231261715,.2503276966,1.928531457,3.340314468,4.312327437,2.074742821,3.593559979,1.155190024,1.563483000,2.708031992,2.132347336,1.176124927},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7042861545,1.219859402,1.574831717,1.863366016,.2264006569,.3921374405,.5062472589,.6288826255e-1,.1089256659,.1620896801e-1,.7042861545,1.219859402,1.574831717,1.863366016,.5545660866,.9605366384,1.240047468,.2435651936,.4218672900,.8576979675e-1,.2264006569,.3921374405,.5062472589,.2435651936,.4218672900,.1356139561,.6288826255e-1,.1089256659,.8576979675e-1,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.055529306,3.560281195,4.596303257,5.438419356,1.928531457,3.340314468,4.312327437,1.563483000,2.708031992,1.176124927,.7042861545,1.219859402,1.574831717,1.863366016,1.618556373,2.803421873,3.619202076,2.074742821,3.593559979,2.132347336,.2264006569,.3921374405,.5062472589,.7108692826,1.231261715,1.155190024,.6288826255e-1,.1089256659,.2503276966,.1620896801e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,2.637214240,4.567789052,5.896990310,6.977413031,1.219036711,2.111433520,2.725848953,.3143011126,.5443854957,.6497290549e-1,3.174462530,5.498330390,7.098314011,2.004818052,3.472446724,.6552866779,3.301857911,5.718985660,2.643567355,3.186694352},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.4134436882,.7161054738,.9244881913,1.093869180,.7802110799e-1,.1351365231,.1744605012,.1272246435e-1,.2203595465e-1,.1924971158e-2,.4134436882,.7161054738,.9244881913,1.093869180,.1911119037,.3310155272,.4273392081,.4927389259e-1,.8534488540e-1,.1018598993e-1,.7802110799e-1,.1351365231,.1744605012,.4927389259e-1,.8534488540e-1,.1610546421e-1,.1272246435e-1,.2203595465e-1,.1018598993e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.637214240,4.567789052,5.896990310,6.977413031,3.174462530,5.498330390,7.098314011,3.301857911,5.718985660,3.186694352,.4134436882,.7161054738,.9244881913,1.093869180,1.219036711,2.111433520,2.725848953,2.004818052,3.472446724,2.643567355,.7802110799e-1,.1351365231,.1744605012,.3143011126,.5443854957,.6552866779,.1272246435e-1,.2203595465e-1,.6497290549e-1,.1924971158e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,3.239870679,5.611620623,7.244571074,8.571892094,.4061142418,.7034105006,.9080990514,.2839399656e-1,.4917984466e-1,.1591703120e-2,4.791094442,8.298418997,10.71321286,.8205190389,1.421180663,.7272675482e-1,6.122166972,10.60390425,1.329187787,7.258877438},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1121154690,.1941896887,.2506978100,.2966296492,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2536990019e-3,.4394195611e-3,.1040928537e-4,.1121154690,.1941896887,.2506978100,.2966296492,.1405355127e-1,.2434146484e-1,.3142469598e-1,.9825720096e-3,.1701864642e-2,.5508076085e-4,.5737338284e-2,.9937361407e-2,.1282907841e-1,.9825720096e-3,.1701864642e-2,.8709032970e-4,.2536990019e-3,.4394195611e-3,.5508076085e-4,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.239870679,5.611620623,7.244571074,8.571892094,4.791094442,8.298418997,10.71321286,6.122166972,10.60390425,7.258877438,.1121154690,.1941896887,.2506978100,.2966296492,.4061142418,.7034105006,.9080990514,.8205190389,1.421180663,1.329187787,.5737338284e-2,.9937361407e-2,.1282907841e-1,.2839399656e-1,.4917984466e-1,.7272675482e-1,.2536990019e-3,.4394195611e-3,.1591703120e-2,.1040928537e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,.1738221700,.3010688299,.3886781879,.4598902340,.2400807184,.4158320022,.5368368063,.1849554605,.3203522547,.1142442288,.1379080812e-1,.2388638034e-1,.3083718442e-1,.2602406990e-1,.4507501127e-1,.2541630033e-1,.9454479976e-3,.1637563968e-2,.2261780150e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,1.235370164,2.139723890,2.762371663,3.268482230,2.838211585,4.915926669,6.346434042,3.637057692,6.299568709,3.736911807,.6965841650,1.206519166,1.557609545,2.186525993,3.787174111,3.552117365,.3394012747,.5878602519,1.350584487,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,2.054909283,3.559207282,4.594916842,5.436778930,.3993492080,.6916931185,.8929719767,.4328832311e-1,.7497757498e-1,.3762235130e-2,1.927368199,3.338299645,4.309726311,.5117504685,.8863778119,.7032397725e-1,1.562068614,2.705582205,.5258003943,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.1738221700,.3010688299,.3886781879,.4598902340,.1379080812e-1,.2388638034e-1,.3083718442e-1,.9454479976e-3,.1637563968e-2,.6014221679e-4,1.235370164,2.139723890,2.762371663,3.268482230,.2400807184,.4158320022,.5368368063,.2602406990e-1,.4507501127e-1,.2261780150e-2,.6965841650,1.206519166,1.557609545,.1849554605,.3203522547,.2541630033e-1,.3394012747,.5878602519,.1142442288,.1534430404},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.235370164,2.139723890,2.762371663,3.268482230,.6965841650,1.206519166,1.557609545,.3394012747,.5878602519,.1534430404,2.054909283,3.559207282,4.594916842,5.436778930,2.838211585,4.915926669,6.346434042,2.186525993,3.787174111,1.350584487,1.927368199,3.338299645,4.309726311,3.637057692,6.299568709,3.552117365,1.562068614,2.705582205,3.736911807,1.174706518},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.054909283,3.559207282,4.594916842,5.436778930,1.927368199,3.338299645,4.309726311,1.562068614,2.705582205,1.174706518,.1738221700,.3010688299,.3886781879,.4598902340,.3993492080,.6916931185,.8929719767,.5117504685,.8863778119,.5258003945,.1379080812e-1,.2388638034e-1,.3083718442e-1,.4328832311e-1,.7497757498e-1,.7032397725e-1,.9454479976e-3,.1637563968e-2,.3762235130e-2,.6014221679e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,.7282240147e-1,.1261320992,.1628358399,.1926699641,.4836674308e-1,.8377365642e-1,.1081513254,.1791787294e-1,.3103466628e-1,.5322103273e-2,.2420523895e-2,.4192470366e-2,.5412455971e-2,.2196466622e-2,.3804391785e-2,.1031552789e-2,.6952117983e-4,.1204142157e-3,.7997598703e-4,.1852758219e-5},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.5940552538,1.028933882,1.328347929,1.571722466,1.857843328,3.217879037,4.154263974,3.240776599,5.613189724,4.532588975,.1610768260,.2789932465,.3601787325,.6882539233,1.192090763,1.522004526,.3774006705e-1,.6536771361e-1,.2044304294,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,2.797223960,4.844934019,6.254782921,7.400758958,.2277441565,.3944644502,.5092514154,.1034248755e-1,.1791371391e-1,.3765823900e-3,3.571361895,6.185780254,7.985807971,.3972713534,.6880941682,.2287137395e-1,3.940069325,6.824400256,.5556284741,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2420523895e-2,.4192470366e-2,.5412455971e-2,.6952117983e-4,.1204142157e-3,.1852758219e-5,.5940552538,1.028933882,1.328347929,1.571722466,.4836674308e-1,.8377365642e-1,.1081513254,.2196466622e-2,.3804391785e-2,.7997598703e-4,.1610768260,.2789932465,.3601787325,.1791787294e-1,.3103466628e-1,.1031552789e-2,.3774006705e-1,.6536771361e-1,.5322103273e-2,.8204764624e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.5940552538,1.028933882,1.328347929,1.571722466,.1610768260,.2789932465,.3601787325,.3774006705e-1,.6536771361e-1,.8204764624e-2,2.797223960,4.844934019,6.254782921,7.400758958,1.857843328,3.217879037,4.154263974,.6882539233,1.192090763,.2044304294,3.571361895,6.185780254,7.985807971,3.240776599,5.613189724,1.522004526,3.940069325,6.824400256,4.532588975,4.033366817},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.797223960,4.844934019,6.254782921,7.400758958,3.571361895,6.185780254,7.985807971,3.940069325,6.824400256,4.033366817,.7282240147e-1,.1261320992,.1628358399,.1926699641,.2277441565,.3944644502,.5092514154,.3972713534,.6880941682,.5556284741,.2420523895e-2,.4192470366e-2,.5412455971e-2,.1034248755e-1,.1791371391e-1,.2287137395e-1,.6952117983e-4,.1204142157e-3,.3765823900e-3,.1852758219e-5}
		};
		for (int j = 0; j < 27; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[27][35] = 
		{
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,-.2397923765,-0.,-0.,-0.,-.4964996319,-0.,-0.,-.5734033834,-0.,-.5309563362,.2624521439e-1,0.,0.,.7424497583e-1,0.,.1087015236,-.2482151352e-2,-0.,-.8901689590e-2,.2178212518e-3},
			{1.414213562,0.,0.,0.,0.,1.851946997,0.,0.,0.,1.565440216,0.,0.,1.143425313,0.,.7749487345,1.851946997,0.,0.,0.,3.834529752,0.,0.,4.428467199,0.,4.100643259,1.565440216,0.,0.,4.428467199,0.,6.483686290,1.143425313,0.,4.100643259,.7749487345},
			{1.414213562,0.,0.,0.,0.,-.2397923765,-0.,-0.,-0.,.2624521439e-1,0.,0.,-.2482151352e-2,-0.,.2178212518e-3,1.851946997,0.,0.,0.,-.4964996319,-0.,-0.,.7424497583e-1,0.,-.8901689590e-2,1.565440216,0.,0.,-.5734033834,-0.,.1087015236,1.143425313,0.,-.5309563362,.7749487345},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,.6999612018,0.,0.,0.,1.081581075,0.,0.,.9321846605,0.,.6441723235,.2236285858,0.,0.,.4721128436,0.,.5158410843,.6173679058e-1,0.,.1652303366,.1581446975e-1},
			{1.414213562,0.,0.,0.,0.,1.382070207,0.,0.,0.,.8718456225,0.,0.,.4752394227,0.,.2403694019,1.382070207,0.,0.,0.,2.135576909,0.,0.,1.840594370,0.,1.271915321,.8718456225,0.,0.,1.840594370,0.,2.011074701,.4752394227,0.,1.271915321,.2403694019},
			{1.414213562,0.,0.,0.,0.,.6999612018,0.,0.,0.,.2236285858,0.,0.,.6173679058e-1,0.,.1581446975e-1,1.382070207,0.,0.,0.,1.081581075,0.,0.,.4721128436,0.,.1652303366,.8718456225,0.,0.,.9321846605,0.,.5158410843,.4752394227,0.,.6441723235,.2403694019},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,2.055529306,0.,0.,0.,1.618556373,0.,0.,.7108692826,0.,.2503276966,1.928531457,0.,0.,2.074742821,0.,1.155190024,1.563483000,0.,2.132347336,1.176124927},
			{1.414213562,0.,0.,0.,0.,.7042861545,0.,0.,0.,.2264006569,0.,0.,.6288826255e-1,0.,.1620896801e-1,.7042861545,0.,0.,0.,.5545660866,0.,0.,.2435651936,0.,.8576979675e-1,.2264006569,0.,0.,.2435651936,0.,.1356139561,.6288826255e-1,0.,.8576979675e-1,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,2.055529306,0.,0.,0.,1.928531457,0.,0.,1.563483000,0.,1.176124927,.7042861545,0.,0.,0.,1.618556373,0.,0.,2.074742821,0.,2.132347336,.2264006569,0.,0.,.7108692826,0.,1.155190024,.6288826255e-1,0.,.2503276966,.1620896801e-1},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,2.637214240,0.,0.,0.,1.219036711,0.,0.,.3143011126,0.,.6497290549e-1,3.174462530,0.,0.,2.004818052,0.,.6552866779,3.301857911,0.,2.643567355,3.186694352},
			{1.414213562,0.,0.,0.,0.,.4134436882,0.,0.,0.,.7802110799e-1,0.,0.,.1272246435e-1,0.,.1924971158e-2,.4134436882,0.,0.,0.,.1911119037,0.,0.,.4927389259e-1,0.,.1018598993e-1,.7802110799e-1,0.,0.,.4927389259e-1,0.,.1610546421e-1,.1272246435e-1,0.,.1018598993e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,2.637214240,0.,0.,0.,3.174462530,0.,0.,3.301857911,0.,3.186694352,.4134436882,0.,0.,0.,1.219036711,0.,0.,2.004818052,0.,2.643567355,.7802110799e-1,0.,0.,.3143011126,0.,.6552866779,.1272246435e-1,0.,.6497290549e-1,.1924971158e-2},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,3.239870679,0.,0.,0.,.4061142418,0.,0.,.2839399656e-1,0.,.1591703120e-2,4.791094442,0.,0.,.8205190389,0.,.7272675482e-1,6.122166972,0.,1.329187787,7.258877438},
			{1.414213562,0.,0.,0.,0.,.1121154690,0.,0.,0.,.5737338284e-2,0.,0.,.2536990019e-3,0.,.1040928537e-4,.1121154690,0.,0.,0.,.1405355127e-1,0.,0.,.9825720096e-3,0.,.5508076085e-4,.5737338284e-2,0.,0.,.9825720096e-3,0.,.8709032970e-4,.2536990019e-3,0.,.5508076085e-4,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,3.239870679,0.,0.,0.,4.791094442,0.,0.,6.122166972,0.,7.258877438,.1121154690,0.,0.,0.,.4061142418,0.,0.,.8205190389,0.,1.329187787,.5737338284e-2,0.,0.,.2839399656e-1,0.,.7272675482e-1,.2536990019e-3,0.,.1591703120e-2,.1040928537e-4},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,.1738221700,0.,0.,0.,.2400807184,0.,0.,.1849554605,0.,.1142442288,.1379080812e-1,0.,0.,.2602406990e-1,0.,.2541630033e-1,.9454479976e-3,0.,.2261780150e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,1.235370164,0.,0.,0.,2.838211585,0.,0.,3.637057692,0.,3.736911807,.6965841650,0.,0.,2.186525993,0.,3.552117365,.3394012747,0.,1.350584487,.1534430404},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,2.054909283,0.,0.,0.,.3993492080,0.,0.,.4328832311e-1,0.,.3762235130e-2,1.927368199,0.,0.,.5117504685,0.,.7032397725e-1,1.562068614,0.,.5258003943,1.174706518},
			{1.414213562,0.,0.,0.,0.,.1738221700,0.,0.,0.,.1379080812e-1,0.,0.,.9454479976e-3,0.,.6014221679e-4,1.235370164,0.,0.,0.,.2400807184,0.,0.,.2602406990e-1,0.,.2261780150e-2,.6965841650,0.,0.,.1849554605,0.,.2541630033e-1,.3394012747,0.,.1142442288,.1534430404},
			{1.414213562,0.,0.,0.,0.,1.235370164,0.,0.,0.,.6965841650,0.,0.,.3394012747,0.,.1534430404,2.054909283,0.,0.,0.,2.838211585,0.,0.,2.186525993,0.,1.350584487,1.927368199,0.,0.,3.637057692,0.,3.552117365,1.562068614,0.,3.736911807,1.174706518},
			{1.414213562,0.,0.,0.,0.,2.054909283,0.,0.,0.,1.927368199,0.,0.,1.562068614,0.,1.174706518,.1738221700,0.,0.,0.,.3993492080,0.,0.,.5117504685,0.,.5258003945,.1379080812e-1,0.,0.,.4328832311e-1,0.,.7032397725e-1,.9454479976e-3,0.,.3762235130e-2,.6014221679e-4},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,.7282240147e-1,0.,0.,0.,.4836674308e-1,0.,0.,.1791787294e-1,0.,.5322103273e-2,.2420523895e-2,0.,0.,.2196466622e-2,0.,.1031552789e-2,.6952117983e-4,0.,.7997598703e-4,.1852758219e-5},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.5940552538,0.,0.,0.,1.857843328,0.,0.,3.240776599,0.,4.532588975,.1610768260,0.,0.,.6882539233,0.,1.522004526,.3774006705e-1,0.,.2044304294,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,2.797223960,0.,0.,0.,.2277441565,0.,0.,.1034248755e-1,0.,.3765823900e-3,3.571361895,0.,0.,.3972713534,0.,.2287137395e-1,3.940069325,0.,.5556284741,4.033366817},
			{1.414213562,0.,0.,0.,0.,.7282240147e-1,0.,0.,0.,.2420523895e-2,0.,0.,.6952117983e-4,0.,.1852758219e-5,.5940552538,0.,0.,0.,.4836674308e-1,0.,0.,.2196466622e-2,0.,.7997598703e-4,.1610768260,0.,0.,.1791787294e-1,0.,.1031552789e-2,.3774006705e-1,0.,.5322103273e-2,.8204764624e-2},
			{1.414213562,0.,0.,0.,0.,.5940552538,0.,0.,0.,.1610768260,0.,0.,.3774006705e-1,0.,.8204764624e-2,2.797223960,0.,0.,0.,1.857843328,0.,0.,.6882539233,0.,.2044304294,3.571361895,0.,0.,3.240776599,0.,1.522004526,3.940069325,0.,4.532588975,4.033366817},
			{1.414213562,0.,0.,0.,0.,2.797223960,0.,0.,0.,3.571361895,0.,0.,3.940069325,0.,4.033366817,.7282240147e-1,0.,0.,0.,.2277441565,0.,0.,.3972713534,0.,.5556284741,.2420523895e-2,0.,0.,.1034248755e-1,0.,.2287137395e-1,.6952117983e-4,0.,.3765823900e-3,.1852758219e-5}
		};
		for (int j = 0; j < 27; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-.2973526952e-2,-.2973526952e-2,.3850159296e-3,-.1846817770,-.1846817770,-.9353365550e-1,-.7236569171e-1,-.7236569171e-1,-.2112064808,-.2591194752e-1,-.2591194752e-1,-.1652833479,-.2652578911e-2,-.2652578911e-2,-.7665322818e-1,-.1119869688,-.1862786300,-.1575707306e-1,-.1575707306e-1,-.1119869688,-.1862786300,-.2130681262e-1,-.1003272446,-.2611900580e-2,-.2611900580e-2,-.2130681262e-1,-.1003272446},
			{-.3838806789e-2,-.3838806789e-2,.4970534278e-3,-.2384231489,-.2384231489,-.1207514300,-.9342370632e-1,-.9342370632e-1,-.2726663942,-.3345218040e-1,-.3345218040e-1,-.2133798847,-.3424464647e-2,-.3424464647e-2,-.9895889209e-1,-.1445745550,-.2404846772,-.2034229384e-1,-.2034229384e-1,-.1445745550,-.2404846772,-.2750697681e-1,-.1295219158,-.3371949150e-2,-.3371949150e-2,-.2750697681e-1,-.1295219158},
			{-.4542137448e-2,-.4542137448e-2,.5881215471e-3,-.2821060742,-.2821060742,-.1428750188,-.1105404201,-.1105404201,-.3226232284,-.3958115364e-1,-.3958115364e-1,-.2524744844,-.4051881214e-2,-.4051881214e-2,-.1170897402,-.1710629204,-.2845453076,-.2406932666e-1,-.2406932666e-1,-.1710629204,-.2845453076,-.3254669388e-1,-.1532523976,-.3989744040e-2,-.3989744040e-2,-.3254669388e-1,-.1532523976},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-.2513505346e-2,-.2513505346e-2,-.4213989523e-4,-.1165020402,-.1165020402,-.2988279787e-1,-.2326276051e-1,-.2326276051e-1,-.1981573996,-.4889852992e-2,-.4889852992e-2,-.1989545587,-.1357416837e-3,-.1357416837e-3,-.1133541712,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.5777297190e-2,-.1280930321,-.8681624939e-4,-.8681624939e-4,-.5777297190e-2,-.1280930321},
			{-.3244921449e-2,-.3244921449e-2,-.5440237079e-4,-.1504034872,-.1504034872,-.3857852615e-1,-.3003209470e-1,-.3003209470e-1,-.2558201030,-.6312773071e-2,-.6312773071e-2,-.2568492309,-.1752417601e-3,-.1752417601e-3,-.1463396058,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.7458458603e-2,-.1653673934,-.1120792961e-3,-.1120792961e-3,-.7458458603e-2,-.1653673934},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-.1835908907e-2,-.1835908907e-2,.3985396969e-5,-.6350477757e-1,-.6350477757e-1,-.8249696816e-2,-.6461794816e-2,-.6461794816e-2,-.1606485206,-.7973608935e-3,-.7973608935e-3,-.2069388682,-.6002353001e-5,-.6002353001e-5,-.1448464796,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.1353612365e-2,-.1413173578,-.2493496593e-5,-.2493496593e-5,-.1353612365e-2,-.1413173578},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{.3850159296e-3,-.2973526952e-2,-.2973526952e-2,-.9353365550e-1,-.1846817770,-.1846817770,-.2112064808,-.7236569171e-1,-.7236569171e-1,-.1652833479,-.2591194752e-1,-.2591194752e-1,-.7665322818e-1,-.2652578911e-2,-.2652578911e-2,-.1575707306e-1,-.1119869688,-.1862786300,-.1119869688,-.1862786300,-.1575707306e-1,-.2611900580e-2,-.2130681262e-1,-.1003272446,-.2130681262e-1,-.1003272446,-.2611900580e-2},
			{.4970534278e-3,-.3838806789e-2,-.3838806789e-2,-.1207514300,-.2384231489,-.2384231489,-.2726663942,-.9342370632e-1,-.9342370632e-1,-.2133798847,-.3345218040e-1,-.3345218040e-1,-.9895889209e-1,-.3424464647e-2,-.3424464647e-2,-.2034229384e-1,-.1445745550,-.2404846772,-.1445745550,-.2404846772,-.2034229384e-1,-.3371949150e-2,-.2750697681e-1,-.1295219158,-.2750697681e-1,-.1295219158,-.3371949150e-2},
			{.5881215471e-3,-.4542137448e-2,-.4542137448e-2,-.1428750188,-.2821060742,-.2821060742,-.3226232284,-.1105404201,-.1105404201,-.2524744844,-.3958115364e-1,-.3958115364e-1,-.1170897402,-.4051881214e-2,-.4051881214e-2,-.2406932666e-1,-.1710629204,-.2845453076,-.1710629204,-.2845453076,-.2406932666e-1,-.3989744040e-2,-.3254669388e-1,-.1532523976,-.3254669388e-1,-.1532523976,-.3989744040e-2},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{.7971907625e-3,-.6156805564e-2,.7971907625e-3,-.1445283416,-.2853705526,-.1445283416,-.1663073324,-.5698189327e-1,-.1663073324,-.7640125172e-1,-.1197764475e-1,-.7640125172e-1,-.9608398221e-2,-.3324978618e-3,-.9608398221e-2,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.2176344606e-1,-.2572854045,-.3620122019e-1,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2,-.1734756364e-2,-.6663474388e-1,-.8168435577e-2},
			{.1029168849e-2,-.7948401806e-2,.1029168849e-2,-.1865852866,-.3684117992,-.1865852866,-.2147018429,-.7356330789e-1,-.2147018429,-.9863359184e-1,-.1546307288e-1,-.9863359184e-1,-.1240438876e-1,-.4292528938e-3,-.1240438876e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2809648804e-1,-.3321540290,-.4673557432e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1,-.2239560836e-2,-.8602508443e-1,-.1054540498e-1},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{.9206691217e-3,-.7110444628e-2,-.1192093711e-3,-.1245649596,-.2459529459,-.6308698245e-1,-.7304211084e-1,-.2502642370e-1,-.2131806773,-.1969833902e-1,-.3088165462e-2,-.1256488891,-.6717834486e-3,-.2324701322e-4,-.1941294556e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-.4213989523e-4,-.2513505346e-2,-.2513505346e-2,-.2988279787e-1,-.1165020402,-.1165020402,-.1981573996,-.2326276051e-1,-.2326276051e-1,-.1989545587,-.4889852992e-2,-.4889852992e-2,-.1133541712,-.1357416837e-3,-.1357416837e-3,-.1250144162e-2,-.6314572866e-1,-.1747169623,-.6314572866e-1,-.1747169623,-.1250144162e-2,-.8681624939e-4,-.5777297190e-2,-.1280930321,-.5777297190e-2,-.1280930321,-.8681624939e-4},
			{-.5440237079e-4,-.3244921449e-2,-.3244921449e-2,-.3857852615e-1,-.1504034872,-.1504034872,-.2558201030,-.3003209470e-1,-.3003209470e-1,-.2568492309,-.6312773071e-2,-.6312773071e-2,-.1463396058,-.1752417601e-3,-.1752417601e-3,-.1613929173e-2,-.8152078517e-1,-.2255586285,-.8152078517e-1,-.2255586285,-.1613929173e-2,-.1120792961e-3,-.7458458603e-2,-.1653673934,-.7458458603e-2,-.1653673934,-.1120792961e-3},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-.1192093711e-3,-.7110444628e-2,.9206691217e-3,-.6308698245e-1,-.2459529459,-.1245649596,-.2131806773,-.2502642370e-1,-.7304211084e-1,-.1256488891,-.3088165462e-2,-.1969833902e-1,-.1941294556e-1,-.2324701322e-4,-.6717834486e-3,-.2359095912e-2,-.1982097554,-.4639045478e-1,-.1676631182e-1,-.3297012329,-.3924109737e-2,-.7878005023e-4,-.2468540979e-1,-.1424881984e-1,-.6426553063e-3,-.1162360223,-.3709510905e-3},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{.3985396969e-5,-.1835908907e-2,-.1835908907e-2,-.8249696816e-2,-.6350477757e-1,-.6350477757e-1,-.1606485206,-.6461794816e-2,-.6461794816e-2,-.2069388682,-.7973608935e-3,-.7973608935e-3,-.1448464796,-.6002353001e-5,-.6002353001e-5,-.8570536869e-4,-.3076690782e-1,-.1416023588,-.3076690782e-1,-.1416023588,-.8570536869e-4,-.2493496593e-5,-.1353612365e-2,-.1413173578,-.1353612365e-2,-.1413173578,-.2493496593e-5},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 27; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1716766587e-2,-.1716766587e-2,.2222890507e-3,-.1066260737,-.1066260737,-.5400168118e-1,-.4178035162e-1,-.4178035162e-1,-.1219401185,-.1496026988e-1,-.1496026988e-1,-.9542638548e-1,-.1531467148e-2,-.1531467148e-2,-.4425576195e-1,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1451172989e-2,-.1451172989e-2,-.2432947984e-4,-.6726248431e-1,-.6726248431e-1,-.1725284139e-1,-.1343076105e-1,-.1343076105e-1,-.1144062280,-.2823157942e-2,-.2823157942e-2,-.1148664680,-.7837049758e-4,-.7837049758e-4,-.6544506128e-1,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1059962502e-2,-.1059962502e-2,.2300970013e-5,-.3666450043e-1,-.3666450043e-1,-.4762964678e-2,-.3730718977e-2,-.3730718977e-2,-.9275046664e-1,-.4603565265e-3,-.4603565265e-3,-.1194762113,-.3465460121e-5,-.3465460121e-5,-.8362715398e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.7183823816e-3,-.7183823816e-3,-.2019216790e-6,-.1854438756e-1,-.1854438756e-1,-.1220078986e-2,-.9615642428e-3,-.9615642428e-3,-.6977123241e-1,-.6965419683e-4,-.6965419683e-4,-.1153090709,-.1421880381e-6,-.1421880381e-6,-.9915431315e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1},
			{.2222890507e-3,-.1716766587e-2,-.1716766587e-2,-.5400168118e-1,-.1066260737,-.1066260737,-.1219401185,-.4178035162e-1,-.4178035162e-1,-.9542638548e-1,-.1496026988e-1,-.1496026988e-1,-.4425576195e-1,-.1531467148e-2,-.1531467148e-2,-.9097350370e-2,-.6465570658e-1,-.1075480172,-.6465570658e-1,-.1075480172,-.9097350370e-2,-.1507981504e-2,-.1230149400e-1,-.5792396170e-1,-.1230149400e-1,-.5792396170e-1,-.1507981504e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4602583012e-3,-.3554633349e-2,.4602583012e-3,-.8344347689e-1,-.1647587653,-.8344347689e-1,-.9601758310e-1,-.3289851142e-1,-.9601758310e-1,-.4411028324e-1,-.6915296421e-2,-.4411028324e-1,-.5547411298e-2,-.1919677300e-3,-.5547411298e-2,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1256513144e-1,-.1485437975,-.2090078421e-1,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2,-.1001562054e-2,-.3847158731e-1,-.4716048478e-2},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.5315485655e-3,-.4105217121e-2,-.6882556248e-4,-.7191761300e-1,-.1420009996,-.3642328630e-1,-.4217088237e-1,-.1444901247e-1,-.1230799215,-.1137284134e-1,-.1782953162e-2,-.7254341999e-1,-.3878543549e-3,-.1342166934e-4,-.1120806935e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.4921998841e-3,-.3801322254e-2,.8251922589e-5,-.4969759516e-1,-.9812767548e-1,-.1274744362e-1,-.1485018428e-1,-.5088119713e-2,-.1264971928,-.2351014731e-2,-.3685753652e-3,-.9565627002e-1,-.2174223645e-4,-.7523883765e-6,-.1815634761e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1183750410e-3,-.7068569148e-1,-.2751887412e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1},
			{-.2432947984e-4,-.1451172989e-2,-.1451172989e-2,-.1725284139e-1,-.6726248431e-1,-.6726248431e-1,-.1144062280,-.1343076105e-1,-.1343076105e-1,-.1148664680,-.2823157942e-2,-.2823157942e-2,-.6544506128e-1,-.7837049758e-4,-.7837049758e-4,-.7217710685e-3,-.3645720343e-1,-.1008728852,-.3645720343e-1,-.1008728852,-.7217710685e-3,-.5012338496e-4,-.3335524088e-2,-.7395454658e-1,-.3335524088e-2,-.7395454658e-1,-.5012338496e-4},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.6882556248e-4,-.4105217121e-2,.5315485655e-3,-.3642328630e-1,-.1420009996,-.7191761300e-1,-.1230799215,-.1444901247e-1,-.4217088237e-1,-.7254341999e-1,-.1782953162e-2,-.1137284134e-1,-.1120806935e-1,-.1342166934e-4,-.3878543549e-3,-.1362024660e-2,-.1144364557,-.2678354156e-1,-.9680034646e-2,-.1903530956,-.2265585814e-2,-.4548368323e-4,-.1425212799e-1,-.8226559970e-2,-.3710372144e-3,-.6710889878e-1,-.2141687120e-3},
			{-0.,-0.,0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{-.1007670004e-3,-.6010418228e-2,-.1007670004e-3,-.3979689975e-1,-.1551534778,-.3979689975e-1,-.6852931167e-1,-.8045023649e-2,-.6852931167e-1,-.2371124734e-1,-.5827688220e-3,-.2371124734e-1,-.9934279074e-3,-.1189630476e-5,-.9934279074e-3,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.1330215756e-2,-.1859075642,-.3680553869e-2,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3,-.2136104405e-4,-.3151715169e-1,-.4736126273e-3},
			{.2300970013e-5,-.1059962502e-2,-.1059962502e-2,-.4762964678e-2,-.3666450043e-1,-.3666450043e-1,-.9275046664e-1,-.3730718977e-2,-.3730718977e-2,-.1194762113,-.4603565265e-3,-.4603565265e-3,-.8362715398e-1,-.3465460121e-5,-.3465460121e-5,-.4948201769e-4,-.1776328251e-1,-.8175415996e-1,-.1776328251e-1,-.8175415996e-1,-.4948201769e-4,-.1439620930e-5,-.7815084635e-3,-.8158961455e-1,-.7815084635e-3,-.8158961455e-1,-.1439620930e-5},
			{0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.,-0.},
			{.8251922589e-5,-.3801322254e-2,.4921998841e-3,-.1274744362e-1,-.9812767548e-1,-.4969759516e-1,-.1264971928,-.5088119713e-2,-.1485018428e-1,-.9565627002e-1,-.3685753652e-3,-.2351014731e-2,-.1815634761e-1,-.7523883765e-6,-.2174223645e-4,-.1183750410e-3,-.7068569148e-1,-.2751887410e-1,-.5979212992e-2,-.1955791717,-.1969045212e-3,-.1656115518e-5,-.4233275753e-2,-.1150576533e-1,-.1102083031e-3,-.9385930977e-1,-.7798139958e-5},
			{-.2019216790e-6,-.7183823816e-3,-.7183823816e-3,-.1220078986e-2,-.1854438756e-1,-.1854438756e-1,-.6977123241e-1,-.9615642428e-3,-.9615642428e-3,-.1153090709,-.6965419683e-4,-.6965419683e-4,-.9915431315e-1,-.1421880381e-6,-.1421880381e-6,-.3147669934e-5,-.8030765587e-2,-.6148074654e-1,-.8030765587e-2,-.6148074654e-1,-.3147669934e-5,-.3836628659e-7,-.1699014733e-3,-.8352158724e-1,-.1699014733e-3,-.8352158724e-1,-.3836628659e-7}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 27; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
